/***************************************************************************
 *   Copyright (C) 2012 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"

#define NUMBER_OF_POINTS        20

// Obtains area under the curve (AUC) for the ROC
scalar ClassifierAUROC(const t_Parameters *parm, t_Classifier classifier, scalar **test_data, int test_instances, int *test_labels, scalar **train_data, int train_instances, int *subset,
                             int *subspace, int *train_labels, int inputs, int minority, bool write_outputs, t_Learner *learner, bool fast, FILE *out_file, int classes, int **test_index,
                             int **train_index, scalar *sc_pos, scalar *sc_neg, t_Ensemble *ensemble, bool single)
{
    scalar auc = 0.0, **roc;
    bool free_sc = false;

    // Count positive and negative instances
    int n_positives = 0, n_negatives = 0;

    for (int i = 0; i < test_instances; i++)
        if (test_labels[i] == minority)
            n_positives++;
        else
            n_negatives++;

    // No instances? then return
    if (n_positives == 0 || n_negatives == 0)
        return 0.0;

    // Already available?
    if (!sc_pos) {
        MALLOC(sc_pos, scalar, n_positives, "ClassifierAUROC: sc_pos");
        MALLOC(sc_neg, scalar, n_negatives, "ClassifierAUROC: sc_neg");
        free_sc = true;

        GetPositiveAndNegativeInstancesOutputs(parm, classifier, train_data, train_instances, train_labels, test_data, test_instances, test_labels, sc_pos, sc_neg, minority, inputs, learner, classes,
                                               test_index, train_index, subset, subspace, ensemble, single);
    }

    // Sort ratios
    qsort(sc_pos, n_positives, sizeof(scalar), CompareDouble);
    qsort(sc_neg, n_negatives, sizeof(scalar), CompareDouble);

    // Obtain ROC curve points
    MALLOC(roc, scalar *, NUMBER_OF_POINTS+2, "ClassifierAUROC: roc");
    for (int i = 0; i < NUMBER_OF_POINTS+2; i++)
        MALLOC(roc[i], scalar, 2, "ClassifierAUROC: roc[i]");

    if (write_outputs) {
        fprintf(out_file, "Error\t1-Sp\tSn\tG\tF\tPrec\tFPrate\tFNrate\n");
        fprintf(out_file, "==============================================================\n");
        fprintf(out_file, "0.0000\t0.0000\t0.0000\t0.0000\t0.0000\t0.0000\t0.0000\t0.0000\n");
    }
    roc[0][0] = roc[0][1] = 0.0;
    roc[NUMBER_OF_POINTS+1][0] = roc[NUMBER_OF_POINTS+1][1] = 1.0;

    // Limits of the threshold depending on the classifier
    scalar max_value = 1.0, min_value = -1.0, delta, threshold;

    // If k-NN limits are variable
    if (classifier == kNN) {
        max_value = MAX(sc_neg[n_negatives-1], sc_pos[n_positives-1]);
        min_value = MIN(sc_neg[0], sc_pos[0]);
    }

    delta = (max_value-min_value)/NUMBER_OF_POINTS;
    threshold = max_value;

    for (int next = 1; next < NUMBER_OF_POINTS+1; next++, threshold -= delta) {
        scalar specificity, sensitivity, TN, TP, FN, FP, precision;

        // Obtain sensitivity
        // Positive sequences
        TP = 0.0;
        for (int i = n_positives - 1; i >= 0 && sc_pos[i] >= threshold; i--)
            TP++;
        FN = n_positives - TP;

        // Obtain specificity
        // Negative sequences
        TN = 0.0;
        for (int i = 0; i < n_negatives && sc_neg[i] < threshold; i++)
            TN++;
        FP = n_negatives - TN;

        sensitivity = TP/n_positives;
        specificity = TN/n_negatives;
        precision = (TP+FP > 0) ? TP/(TP+FP) : 0.0;

        roc[next][0] = 1.0 - specificity;
        roc[next][1] = sensitivity;

        if (write_outputs)
            fprintf(out_file, "%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\t%6.4lf\n", (FP+FN)/(n_positives+n_negatives), 1.0-specificity, sensitivity,
                    pow(sensitivity*specificity, 0.5), ((sensitivity+precision > 0.0) ? (2.0*sensitivity*precision)/(sensitivity+precision) : 0.0),
                    precision, FP/n_negatives, FN/n_positives);
    }

    if (write_outputs) {
        fprintf(out_file, "1.0000\t1.0000\t1.0000\t1.0000\t1.0000\t1.0000\t1.0000\t1.0000\n");
        fprintf(out_file, "==============================================================\n");
    }

    // If second point is 0,1, AUC is 1
    if (fabs(roc[1][0]-0.0) < ZERO && fabs(roc[1][1]-1.0) < ZERO)
        auc = 1.0;
    else
        // Integrate ROC curve
        auc = fast ? FastAUC(roc, NUMBER_OF_POINTS+2) : qtrap(roc, 0.0, 1.0, NUMBER_OF_POINTS+2);

    if (free_sc) {
        FREE(sc_neg);
        FREE(sc_pos);
    }
    for (int i = 0; i < NUMBER_OF_POINTS+2; i++)
        FREE(roc[i]);
    FREE(roc);

    return auc;
}


// Get positive and negative instances outputs for NN
void GetPositiveAndNegativeInstancesOutputs(const t_Parameters *parm, t_Classifier classifier, scalar **train_data, int train_instances, int *train_labels, scalar **test_data,
        int test_instances, int *test_labels, scalar *sc_pos, scalar *sc_neg, int minority, int inputs, t_Learner *learner, int classes, int **test_index, int **train_index, int *subset,
        int *subspace, t_Ensemble *ensemble, bool single)
{
    int n_positives = 0, n_negatives = 0;
    t_Learner llearner;
    t_Ensemble lensemble;

    // Need to train?
    if (single) {
        if (!learner)
            ClassifierTraining(parm, classifier, train_data, train_instances, train_labels, inputs, classes, train_index, subspace, subset, true, &llearner, NULL);
        else
            llearner = *learner;
    } else {
        if (!ensemble)
            ClassifierTraining(parm, classifier, train_data, train_instances, train_labels, inputs, classes, train_index, subspace, subset, true, NULL, &lensemble);
        else
            lensemble = *ensemble;
    }

    for (int i = 0; i < test_instances; i++) {
        scalar output;

        Classifier2ClassOutput(parm, classifier, test_data[i], inputs, classes, test_index ? test_index[i] : NULL, subspace, train_data, train_instances, train_labels, subset, learner,
                                     minority, &output, train_index, &lensemble);

        if (test_labels[i] == minority) {
            sc_pos[n_positives] = output;
            n_positives++;
        } else {
            sc_neg[n_negatives] = output;
            n_negatives++;
        }
    }
}

// Interpolate value
scalar f(scalar x, scalar **data, int n)
{
    scalar x1, x2, y1, y2, y;

    // If outside limits return 0
    if (x < data[0][0] && x > data[n-1][0])
        return 0.0;

    // Otherwise linear interpolation
    // Get points to interpolate
    int i = 0;
    while (x > data[i][0])
        i++;

    // Check especial cases
    if (i == 0)
        return data[0][1];

    x1 = data[i-1][0];
    y1 = data[i-1][1];

    // Check especial cases
    if (x == x1)
        y = y1;
    else {
        x2 = data[i][0];
        y2 = data[i][1];

        if (x==x2)
            y=y2;
        // If x1=x2, return y1
        else if (x1==x2)
            y = y1;
        else {
            // Equation is: y = y1+[(y2-y1)/(x2-x1)]·(x-x1) if x1 != x2
            y = y1+((y2-y1)/(x2-x1)) * (x-x1);
        }
    }

    return y;
}

scalar trapzd(scalar **data, scalar a, scalar b, int n, int nvalues)
{
    scalar  x, tnm, sum, del;
    static scalar s;
    int it, j;
    if (n == 1) {
        return (s = 0.5* (b-a) * (f(a, data, nvalues)+f(b, data, nvalues)));
    } else {
        for (it = 1,j = 1; j < n-1; j++)
            it <<= 1;
        tnm = it;
        del = (b - a)/tnm;   /*This is the spacing of points to be added. */
        x = a+0.5 * del;
        for (sum = 0.0, j = 1; j <= it; j++, x += del)
            sum += f(x, data, nvalues);
        s = 0.5 * (s+(b-a) * sum/tnm);     /*This replaces s by its refined value.*/
        return s;
    }
}

#define EPS 1.0e-5
#define JMAX 25
scalar qtrap(scalar **data, scalar a, scalar b, int n)
{
    int j;
    scalar s, olds;
    olds = -1.0e30;
    for (j = 1; j <= JMAX; j++) {
        s = trapzd(data, a, b, j, n);
        if (j > 5)
            if (fabs(s-olds) < EPS*fabs(olds) || (s == 0.0 && olds == 0.0))
                return s;
        olds = s;
    }

    return s;
}

// Fast integral just adding trapezoids
scalar FastAUC(scalar **curve, int n)
{
    scalar auc = 0.0;

    for (int i = 0; i < n-1; i++)
        auc += ((curve[i+1][0]-curve[i][0])*(curve[i][1]+curve[i+1][1]))/2.0;

    return auc;
}

#define N_POINTS 10
// Obtains area under the curve (AUC) for the PRC
scalar ClassifierAUPRC(const t_Parameters *parm, t_Classifier classifier, scalar **test_data, int test_instances, int *test_labels, scalar **train_data, int train_instances, int *train_labels,
                             int inputs, int minority, bool write_outputs, t_Learner *learner, FILE *out_file, int classes, int **test_index, int **train_index, int *subspace, int *subset,
                             scalar *sc_pos, scalar *sc_neg, t_Ensemble *ensemble, bool single)
{
    scalar **points;
    bool free_sc = false;

    // Count positive and negative instances
    int n_positives = 0, n_negatives = 0;

    for (int i = 0; i < test_instances; i++)
        if (test_labels[i] == minority)
            n_positives++;
        else
            n_negatives++;

    // Already available?
    if (!sc_pos) {
        MALLOC(sc_pos, scalar, n_positives, "ClassifierAUROC: sc_pos");
        MALLOC(sc_neg, scalar, n_negatives, "ClassifierAUROC: sc_neg");
        free_sc = true;

        GetPositiveAndNegativeInstancesOutputs(parm, classifier, train_data, train_instances, train_labels, test_data, test_instances, test_labels, sc_pos, sc_neg, minority, inputs, learner, classes,
                                               test_index, train_index, subset, subspace, ensemble, single);
    }


    // Sort ratios
    qsort(sc_pos, n_positives, sizeof(scalar), CompareDouble);
    qsort(sc_neg, n_negatives, sizeof(scalar), CompareDouble);

    // Get number of possible thresholds
    int n_thresholds = 1;
    scalar *thresholds;
    ALLOCA(thresholds, scalar, n_positives+1, "ObtainAUPRC: thresholds");
    thresholds[0] = sc_pos[n_positives-1]; // TP = 0
    for (int i = n_positives-2; i >= 0; i--) {
        if (sc_pos[i] < thresholds[n_thresholds-1]) {
            thresholds[n_thresholds] = sc_pos[i];
            n_thresholds++;
        }
    }
    thresholds[n_thresholds] = sc_pos[0] - 0.001;
    n_thresholds++;

    // Get TP and FP values for all thresholds
    int *TP, *FP;
    ALLOCA(TP, int, n_positives+1, "ObtainAUPRC: TP");
    ALLOCA(FP, int, n_positives+1, "ObtainAUPRC: FP");
    for (int i = 0; i < n_thresholds; i++) {
        // Obtain TP
        TP[i] = 0;
        for (int j = n_positives - 1; j >= 0 && sc_pos[j] > thresholds[i] && j >= 0; j--)
            TP[i]++;

        // Obtain FP
        FP[i] = 0;
        for (int j = n_negatives - 1; j >= 0 && sc_neg[j] > thresholds[i] && j >= 0; j--)
            FP[i]++;
    }

    // As many points as number of positives+1
    int n_points = 0;
    ALLOCA(points, scalar *, n_positives+1, "AURPCFromOutputs: points");
    for (int i = 0; i < n_positives+1; i++)
        ALLOCA(points[i], scalar, 2, "AURPCFromOutputs: points[i]");

    for (int i = 1; i < n_thresholds-1 && i < n_positives; i++) {
        if (n_points == 0) {
            // First point (0,1) if largest values is a positive instance and (0,0) otherwise
            if (sc_pos[n_positives-1] > sc_neg[n_negatives-1]) {
                points[n_points][0] = 0.0;
                points[n_points][1] = 1.0;
            } else
                points[n_points][0] = points[0][1] = 0.0;
        } else {
            // Get Recall/Precision for current point
            points[n_points][0] = TP[i]/((scalar) n_positives);
            points[n_points][1] = TP[i]/((scalar) TP[i]+FP[i]);
        }
        n_points++;

        // Follow J. Davis & M. Goadrich (2006). "The Relationship Between Precision-Recall and ROC Curves"
        // Proceedings of the Twenty-Third International Conference on Machine Learning (ICML'06), Pittsburgh, PA.
        // Interpolate?
        for (int j = 1; TP[i]+j < TP[i+1] && n_points < n_positives; j++) {
            scalar interpolated_value;
            points[n_points][0] = (TP[i]+j)/((scalar) n_positives);
            interpolated_value = TP[i]+j+FP[i]+j*((FP[i+1]-FP[i])/((scalar)(TP[i+1]-TP[i])));
            points[n_points][1] = (TP[i]+j)/interpolated_value;
            n_points++;
        }
    }
    points[n_points][0] = TP[n_thresholds-1]/((scalar) n_positives);
    points[n_points][1] = TP[n_thresholds-1]/((scalar) TP[n_thresholds-1]+FP[n_thresholds-1]);

    if (n_points <= 1)
        return 0.0;

    if (write_outputs) {
        fprintf(out_file, "Prec\tRecall\n");
        fprintf(out_file, "==============\n");
        for (int i = 0; i < n_points; i++)
            fprintf(out_file, "%6.4lf\t%6.4lf\n", points[i][0], points[i][1]);
        fprintf(out_file, "==============\n");
    }

    if (free_sc) {
        FREE(sc_neg);
        FREE(sc_pos);
    }

    // Return auPRC
    return qtrap(points, points[0][0], points[n_points-1][0], n_points);
}
