/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <alloca.h>
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"
#include "classifiers.h"
#include "data.h"

Tree TrainC45(const t_Parameters *parm, scalar **data, int *subspace, int instances, int *label, int inputs, bool releasedata, int *subset, int classes)
{
    int best;
    Tree tree;
    extern Tree *Raw, *Pruned;

    // Global variables
    MaxClass = classes - 1;
    MaxAtt = inputs - 1;

    // Create dummy names file
    FILE *fp = fopen("dummy.names", "w");
    for (int i = 0; i < classes - 1; i++)
        fprintf(fp, "o%d, ", i+1);
    fprintf(fp, "o%d\n\n", classes);
    for (int i = 0; i < inputs; i++)
        fprintf(fp, "x%d: continuous\n", i);
    fclose(fp);

    //  Initialize  values for C4.5
    MALLOC(FileName, char, 100, "main: FileName");

    sprintf(FileName, "dummy");
    GetNames();
    remove("dummy.names");

    // Only selected instances are counted
    MaxItem = -1;

    // Auxiliary memory
    MALLOC(Item, Description, instances+1, "main: Item");
    for (int i = 0; i < instances; i++)
        Item[i] = (Description) calloc(MaxAtt+2, sizeof(AttValue));

    // Create data for tree taking into account sampling
    for (int j = 0; j < instances; j++)
        if (!subset || subset[j]) {
            MaxItem++;  // Item added
            if (subspace)
                for (int i = 0; i < inputs; i++)
                    CVal(Item[MaxItem], i) = data[j][i] * subspace[i];
            else
                for (int i = 0; i < inputs; i++)
                    CVal(Item[MaxItem], i) = data[j][i];

            Class(Item[MaxItem]) = label[j];
        }

    if (MaxItem < 0)
        OnErrorExit(ERR_UNSPECTED_RESULT, "No items");

    // Create tree
    if (TRIALS == 1) {
        best = 0;
        OneTree(parm->prune_tree, parm->out_file, parm->c45_weights);
    } else
        best = BestTree(parm->out_file, parm->c45_weights);

    // Copy tree to classifier
    if (parm->prune_tree)
        tree = Pruned[best];
    else
        tree = Raw[best];

    if (parm->soften_thresholds)
        SoftenThresh(tree, parm->out_file, parm->c45_weights);

    if (releasedata)
        ReleaseTreeData();

    if (parm->prune_tree) {
        for (int t = 0; t < TRIALS; t++) {
            ReleaseTree(Raw[t]);
            if (t != best)
                ReleaseTree(Pruned[t]);
        }
        free(Raw);
        free(Pruned);
    } else {
        for (int t = 0; t < TRIALS; t++)
            if (t != best)
                ReleaseTree(Raw[t]);
        free(Raw);
    }

    for (int i = 0; i < instances; i++)
        FREE(Item[i]);
    FREE(Item);

    return tree;
}

scalar C45Margin(Tree tree, scalar *pattern, int label, int classes, int inputs, int *subspace)
{
    int max_index;
    scalar max_value, margin, output[classes];

    C45OutputValues(tree, pattern, subspace, inputs, classes, output);

    // Get classification.
    max_index = 0;
    max_value = output[0];
    for (int j = 1; j < classes; j++)
        if (output[j] > max_value) {
            max_value = output[j];
            max_index = j;
        }

    // Test if correctly Classified.
    if (label == max_index) {
        // Difference with second one
        margin = max_value;

        // Find second highest
        output[max_index] = -INFINITE;
        max_index = 0;
        max_value = output[0];
        for (int j = 1; j < classes; j++)
            if (output[j] > max_value) {
                max_value = output[j];
                max_index = j;
            }
        margin -= max_value;

    } else
        // Difference with correct one
        margin = output[label]-max_value;

    return margin;
}

// Evaluates a selection of features, for a specific instance and returns its predicted class
int C45OutputValues(Tree tree, scalar *instance, int *subspace, int inputs, int classes, scalar *output)
{
    extern float *ClassSum;

    static Description desc = NULL;

    if (!desc)
        desc = (Description) malloc((inputs + 2) * sizeof(AttValue));

    MaxAtt = inputs - 1;

    // Create description
    if (subspace)
        for (int i = 0; i < inputs; i++)
            CVal(desc, i) = instance[i] * subspace[i];
    else
        for (int i = 0; i < inputs; i++)
            CVal(desc, i) = instance[i];


    int category = Category(desc, tree);

    // Set output to output vector of tree
    for (int i = 0; i < classes; i++)
        output[i] = ClassSum[i];

    return category;
}

int C45Output(Tree tree, scalar *instance, int *subspace, int inputs)
{
    Description desc = NULL;

    desc = (Description) alloca((inputs+2)*sizeof(AttValue));

    MaxAtt = inputs - 1;

    // Create description
    if (!subspace)
        for (int i = 0; i < inputs; i++)
            CVal(desc, i) = instance[i];
    else
        // Create description
        for (int i = 0; i < inputs; i++)
            CVal(desc, i) = instance[i] * subspace[i];

    // Get output
    return Category(desc, tree);
}

#define C45_PARAM_SETS 16

void CrossValidatedParametersC45(t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, int inputs, int classes, int minority, int *subset)
{
    scalar best_so_far, accuracy = 0.0, **data_v, **data_t, * **ext_str_data = NULL;
    int *label_v = NULL, *label_t, **M, best_index = 0, instances_v, instances_t, majority, n_external_strata, *ext_str_instances = NULL, **ext_str_labels = NULL, **ext_str_list = NULL;
    t_PerformanceInfo perf;

    int c45_p[C45_PARAM_SETS][3] = {{1, false, false}, {1, true, false}, {1, false, true}, {1, true, true}, {3, false, false}, {3, true, false}, {3, false, true}, {3, true, true},
        {5, false, false}, {5, true, false}, {5, false, true}, {5, true, true}, {10, false, false}, {10, true, false}, {10, false, true}, {10, true, true}
    };

    // Sampling?
    int *s_labels = NULL, *n_labels = NULL, n_instances = 0, s_instances = 0;
    scalar **s_data = NULL, **n_data = NULL;
    if (parm->cv_sampling_size && parm->cv_sampling_size < instances)
        switch (parm->cv_sampling_method) {
            case sNone:
                break;

            case sKMeans:
                if (subset) {
                    // Only selected instances
                    MALLOC(n_data, scalar *, instances, "CrossValidatedParametersC45: n_data");
                    MALLOC(n_labels, int, instances, "CrossValidatedParametersC45: n_labels");
                    for (int i = 0; i < instances; i++)
                        if (subset[i]) {
                            n_data[n_instances] = data[i];
                            n_labels[n_instances] = labels[i];
                            n_instances++;
                        }
                } else {
                    n_data = data;
                    n_instances = instances;
                    n_labels = labels;
                }

                if (parm->stratified)
                    kMeansStratifiedSample((const t_Parameters *) &parm, &n_data, n_instances, &n_labels, &s_data, &s_labels, &s_instances, classes, inputs, parm->cv_sampling_size, NULL, false);
                else
                    kMeansSample((const t_Parameters *) &parm, &data, instances, &labels, &s_data, &s_labels, &s_instances, classes, inputs, parm->cv_sampling_size, NULL, false);

                // Overwrite
                data = s_data;
                instances = s_instances;
                labels = s_labels;

                if (subset) {
                    FREE(n_data);
                    FREE(n_labels);
                    subset = NULL;
                }
                break;

            case sStratified:
                n_external_strata = instances%parm->cv_sampling_size ? instances/parm->cv_sampling_size+1 : instances/parm->cv_sampling_size;

                // Memory
                MALLOC(ext_str_data, scalar **, n_external_strata, "CrossValidatedParametersC45: ext_str_data");
                MALLOC(ext_str_instances, int, n_external_strata, "CrossValidatedParametersC45: ext_str_list");
                MALLOC(ext_str_labels, int *, n_external_strata, "CrossValidatedParametersC45: ext_str_labels");
                MALLOC(ext_str_list, int *, n_external_strata, "CrossValidatedParametersC45: ext_str_list");
                for (int i = 0; i < n_external_strata; i++) {
                    MALLOC(ext_str_data[i], scalar *, parm->cv_sampling_size, "CrossValidatedParametersC45: ext_str_data[i]");
                    MALLOC(ext_str_labels[i], int, parm->cv_sampling_size, "CrossValidatedParametersC45: ext_str_labels[i]");
                    MALLOC(ext_str_list[i], int, parm->cv_sampling_size, "CrossValidatedParametersC45: ext_str_labels[i]");
                }

                if (subset) {
                    // Only selected instances
                    MALLOC(n_data, scalar *, instances, "CrossValidatedParametersC45: n_data");
                    MALLOC(n_labels, int, instances, "CrossValidatedParametersC45: n_labels");
                    for (int i = 0; i < instances; i++)
                        if (subset[i]) {
                            n_data[n_instances] = data[i];
                            n_labels[n_instances] = labels[i];
                            n_instances++;
                        }
                } else {
                    n_data = data;
                    n_instances = instances;
                    n_labels = labels;
                }

                // Get first stratum as training set
                DataPartition(parm, n_data, n_instances, n_labels, NULL, &n_external_strata, &ext_str_data, &ext_str_instances, &ext_str_labels, classes, &ext_str_list, NULL, NULL, inputs,
                              parm->cv_sampling_size, pRANDOM, parm->stratified, NULL);

                // Forget original data
                data = ext_str_data[0];
                instances = ext_str_instances[0];
                labels = ext_str_labels[0];

                if (subset) {
                    FREE(n_data);
                    FREE(n_labels);
                    subset = NULL;
                }
                break;
        }

    // Allocate memory
    MALLOC(label_v, int, instances, "SetCrossValidatedParametersSubspace: label_v");
    MALLOC(label_t, int, instances, "SetCrossValidatedParametersSubspace: label_t");
    MALLOC(data_v, scalar *, instances, "SetCrossValidatedParametersSubspace: data_v");
    MALLOC(data_t, scalar *, instances, "SetCrossValidatedParametersSubspace: data_t");

    // Size of validation set
    instances_v = instances/parm->cv_folds;
    instances_t = instances - instances_v;

    StratifyDataset(data, instances, labels, NULL, classes);

    AllocateIntegerMatrix(&M, classes, classes);

    // Do validation runs
    best_so_far = -1.0;
    for (int i = 0; i < C45_PARAM_SETS; i++) {
        scalar margin = 0.0, auroc = 0.0;
        TRIALS = (int) c45_p[i][0];
        parm->soften_thresholds = (bool) c45_p[i][1];
        parm->prune_tree = (bool) c45_p[i][2];

        ZeroIntegerMatrix(M, classes, classes);
        for (int fold = 0; fold < parm->cv_folds; fold ++) {
            // Set fold
            // Validation set
            for (int k = (fold * instances_v) % instances, c = 0; c < instances_v; k++, c++) {
                data_v[c] = data[k];
                label_v[c] = labels[k];
            }

            // Train set
            int next = 0;
            for (int k = 0, c = ((fold+1) * instances_v) % instances; k < instances_t; k++, c = (c+1) % instances)
                if (!subset || subset[c]) {
                    data_t[next] = data[c];
                    label_t[next] = labels[c];
                    next++;
                }

            // Train classifier
            Tree tree;
            if (parm->random_undersampling_for_training && !parm->weigthed_sampling_for_training && classes == 2) {
                scalar **u_data = NULL;
                int u_instances = 0, *u_labels = NULL;
                TemporalUndersampleDataForTraining(data_t, next, classes, label_t, parm->random_undersampling_ratio, &u_data, &u_labels, &u_instances, minority);
                tree = TrainC45(parm, u_data, subspace, u_instances, u_labels, inputs, false, NULL, classes);
                MaxItem = u_instances - 1;
                FREE(u_data);
                FREE(u_labels);
            } else {
                tree = TrainC45(parm, data_t, subspace, next, label_t, inputs, false, NULL, classes);
                MaxItem = next - 1;
            }

            // If margin as performance it must be obtained here
            if (parm->cv_performance_measure == pfMargin) {
                scalar minimum_margin;
                t_Learner learner;

                learner.tree = tree;
                margin += ClassifierDatasetMargin(parm, C45, data_t, instances_t, label_t, data_v, instances_v, label_v, inputs, classes, subspace, &minimum_margin, &learner, NULL, subset, NULL,
                                                  true);
            } else if (parm->cv_performance_measure == pfauROC) {
                t_Learner learner;

                learner.tree = tree;
                auroc += ClassifierAUROC(parm, C45, data_v, instances_v, label_v, data_t, instances_t, subset, subspace, label_t, inputs, minority, false, &learner, false, parm->out_file,
                                               classes, NULL, NULL, NULL, NULL, NULL, true);
            }

            // Better than current best?
            if (tree != NULL) {
                // Global variables
                MaxClass = classes - 1;
                MaxAtt = inputs - 1;

                // Auxiliary memory
                MALLOC(Item, Description, MaxItem+1, "main: Item");
                for (int i = 0; i <= MaxItem; i++)
                    Item[i] = (Description) calloc(MaxAtt+2, sizeof(AttValue));

                if (parm->soften_thresholds)
                    SoftenThresh(tree, parm->out_file, parm->c45_weights);

                if (parm->prune_tree)
                    Prune(tree, parm->out_file, parm->c45_weights);

                // Get M matrix and metric for performance
                for (int i = 0; i < instances_v; i++)
                    M[label_v[i]][C45Output(tree, data_v[i], subspace, inputs)]++;

                ReleaseTree(tree);
                ReleaseTreeData();

                for (int i = 0; i <= MaxItem; i++)
                    FREE(Item[i]);
                FREE(Item);
            }
        }

        PrintMulticlassMeasures(M, classes, instances, NULL, NULL, &perf);
        switch (parm->cv_performance_measure) {
            case pfAccuracy:
                accuracy = perf.accuracy;
                break;

            case pfGMean:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.gmean;
                break;

            case pfMargin:
                accuracy = margin/parm->cv_folds;
                break;

            case pfauROC:
                accuracy = auroc/parm->cv_folds;
                break;

            case pfauPRC:
            case pfPrecisionMacro:
            case pfF1Macro:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Metric for cross-validation of parameters");
                break;

            case pfCC:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.mcc;
                break;

            case pfRecallMacro:
                accuracy = perf.recall_macro;
                break;

            case pfRecallMacroGMean:
                accuracy = perf.recall_macro_g_mean;
                break;

            case pfKappa:
                accuracy = perf.kappa;
                break;

            case pfAverageAccuracy:
                accuracy = perf.average_accuracy;
                break;

            case pfF1:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.f1_score;
                break;
        }

        // Best so far
        if (accuracy > best_so_far) {
            best_so_far = accuracy;
            best_index = i;
        }

        SHOWPROGRESS(parm->show_progress > 1, "BEST Accuracy: %lf\tTRIALS = %d\tSoften thresholds: %d\tPrune tree: %d\n", accuracy, TRIALS, parm->soften_thresholds, parm->prune_tree);
    }

    TRIALS = (int) c45_p[best_index][0];
    parm->soften_thresholds = (bool) c45_p[best_index][1];
    parm->prune_tree = (bool) c45_p[best_index][2];
    SAVEPROGRESS(parm->show_progress > 0, "Best parameters (accuracy: %lf): TRIALS = %d\tSoften thresholds: %d\tPrune tree: %d\n", best_so_far, TRIALS, parm->soften_thresholds, parm->prune_tree);

    FreeIntegerMatrix(&M, classes);
    FREE(label_v);
    FREE(label_t);
    FREE(data_v);
    FREE(data_t)

    if (ext_str_data) {
        for (int i = 0; i < n_external_strata; i++) {
            FREE(ext_str_data[i]);
            FREE(ext_str_labels[i]);
            FREE(ext_str_list[i]);
        }
        FREE(ext_str_list);
        FREE(ext_str_data);
        FREE(ext_str_labels);
        FREE(ext_str_instances);
    } else if (s_data) {
        FREE(s_labels);
        for (int i = 0; i < s_instances; i++)
            FREE(s_data[i]);
        FREE(s_data);
    }
}
