/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Basic functions for dealing with classifiers
#include <stdbool.h>
#include "macros.h"
#include "classifiers.h"
#include "ensemble.h"

// Return the category of an unseen instance
int ClassifierOutput(const t_Parameters *parm, t_Classifier model, scalar *instance, int inputs, int classes, int *index, int *subspace, scalar **data, int instances, int *labels,
                           int *subset, t_Learner *learner, int label, cache_t **neighbors_cache, int inst_order, int **data_index, t_Ensemble *ensemble)
{
    int category = -1.0, *nearest;
    scalar output[classes];

    // Base classifier to use
    switch (model) {
        case SVM:
            category = SupportVectorMachineOutput(learner->svm, instance, inputs, subspace, index, output);
            break;

        case C45:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for C4.5");
            category = C45Output(learner->tree, instance, subspace, inputs);
            break;

        case kNN:
            ALLOCA(nearest, int, learner->knn->k, "ClassifierOutput: nearest");
            category = kNNCategory(instance, data, subspace, instances, learner->knn->k, inputs, labels, classes, nearest, subset, NULL, neighbors_cache, inst_order, data_index, index);
            break;

        case MLP:
            category = NetworkActualOutput(parm, learner->mlp, instance, output, classes, index);
            break;

        case ELM:
            category = ELMPredictClass(learner->elm, instance, inputs, classes, subspace, NULL, index);
            break;

        case NBC:
            category = NBCOutput(learner->nbc, instance, inputs, classes, NULL, subspace, index);
            break;

        case GNBC:
            category = GaussianNBCOutput(parm, learner->gnbc, instance, inputs, classes, NULL, subspace, index);
            break;

        case DT_CA:
        case DT_CART:
        case DT_C45:
        case DT_ID3:
            category = DecisionTreeClassify(learner->dt, instance, NULL, classes, index);
            break;

        case RForest:
        case ExtraTree:
            category = RForestClassify(learner->forest, instance, index, inputs, subspace, classes, NULL);
            break;

        case IBT:
            category = IBTClassify(learner->dt, instance, NULL, classes, parm->ibt_do_not_use_internal_nodes, parm->ibt_predict_with_probabilities, index);
            break;

        case OVO:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for OVO");
            category = OneVsOneOutput(parm, learner->ocoder, instance, subspace, inputs, classes, data, instances, labels, subset);
            break;

        case RF:
            category = RandomForestClassify(learner->rf, instance, index, inputs, subspace, parm->rf_ntree, classes, NULL);
            break;

        case LR:
            category = LROutput(learner->lr, instance, inputs, classes, NULL, subspace, index);

        case NN:
            ALLOCA(nearest, int, 1, "ClassifierOutput: nearest");
            category = kNNCategory(instance, data, subspace, instances, 1, inputs, labels, classes, nearest, subset, NULL, neighbors_cache, inst_order, data_index, index);
            break;

        case AdaBoostM1:
        case MadaBoost:
        case MultiBoost:
        case Bagging:
        case GentleBoost:
        case FloatBoost:
        case ARCING:
            category = AdaBoostOutput(parm, ensemble->boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, label, data_index, NULL);
            break;

        case ReweightBoost:
            category = ReweightBoostOutput(parm, ensemble->boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, label, data_index, NULL);
            break;

        case FeatureBoost:
            category = FeatureBoostOutput(parm, ensemble->featboost, instance, inputs, classes, index, data, instances, labels, subset, label, data_index, NULL);
            break;
    }

    return category;
}

void ClassifierTraining(const t_Parameters *parm, t_Classifier model, scalar **data, int instances, int *labels, int inputs, int classes, int **index, int *subspace, int *subset,
                              bool releasedata, t_Learner *learner, t_Ensemble *ensemble)
{
    scalar **t_data = NULL;
    int *t_labels = NULL, t_instances;

    // Sampling?
    if (parm->training_max_size && parm->training_max_size < instances) {
        t_instances = parm->training_max_size;
        StratifiedRandomSample(data, instances, labels, &t_data, &t_labels, t_instances, classes, NULL, NULL, NULL);

        data = t_data;
        labels = t_labels;
        instances = t_instances;
    }

    // Base classifier to use
    switch (model) {
        case SVM:
            learner->svm = TrainSupportVectorMachine(parm, data, subspace, instances, labels, inputs, subset, index);
            break;

        case C45:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for C4.5");
            learner->tree = TrainC45(parm, data, subspace, instances, labels, inputs, releasedata, subset, classes);
            break;

        case kNN:
            MALLOC(learner->knn, t_kNN, 1, "ClassifierTraining: learner->knn");
            // Cross-validate k?
            if (parm->k <= 0) {
                int minority = -1;
                if (classes == 2) {
                    int count[] = {0, 0};
                    for (int i = 0; i < instances; i++)
                        count[labels[i]]++;
                    minority = count[0] < count[1] ? 0 : 1;
                }
                CrossValidatedParameterskNN((t_Parameters *) parm, data, instances, labels, subspace, inputs, classes, minority, subset, index);
            }
            learner->knn->k = parm->k;
            learner->knn->localk = NULL;
            learner->knn->distance_to_furthest_neighbor = learner->knn->r = NULL;
            learner->knn->labels = NULL;
            break;

        case MLP:
            learner->mlp = TrainMultiLayerPerceptron(parm, data, instances, classes, inputs, labels, subspace, subset, index);
            break;

        case ELM:
            MALLOC(learner->elm, t_ELM, 1, "ClassifierTraining: learner->elm");
            CreateAndInitializeELM(parm, learner->elm, inputs, classes);
            TrainELM(learner->elm, data, labels, instances, inputs, classes, subset, subspace, index);
            break;

        case NBC:
            MALLOC(learner->nbc, t_NaiveBayes, 1, "ClassifierTraining: learner->nbc");
            TrainNBCClassifier(parm, learner->nbc, instances, labels, inputs, data, classes, subset, index);
            break;

        case LR:
            MALLOC(learner->lr, t_LR, 1, "ClassifierTraining: learner->lr");
            TrainLogisticRegressionClassifier(parm, learner->lr, instances, labels, inputs, data, classes, subset, subspace, index);
            break;

        case GNBC:
            MALLOC(learner->gnbc, t_GaussianNaiveBayes, 1, "ClassifierTraining: learner->gnbc");
            TrainGaussianNBCClassifier(learner->gnbc, instances, labels, inputs, data, classes, subset, index);
            break;

        case DT_CA:
            learner->dt = DecisionTreeTrain(parm, data, instances, labels, inputs, subset, subspace, classes, model == IBT, index, imAccuracy, false, false);
            break;

        case DT_CART:
            learner->dt = DecisionTreeTrain(parm, data, instances, labels, inputs, subset, subspace, classes, model == IBT, index, imGini, false, false);
            break;

        case DT_C45:
            learner->dt = DecisionTreeTrain(parm, data, instances, labels, inputs, subset, subspace, classes, model == IBT, index, imGainRatio, false, false);
            break;

        case DT_ID3:
            learner->dt = DecisionTreeTrain(parm, data, instances, labels, inputs, subset, subspace, classes, model == IBT, index, imInformationGain, false, false);
            break;

        case IBT:
            learner->dt = DecisionTreeTrain(parm, data, instances, labels, inputs, subset, subspace, classes, model == IBT, index, parm->ibt_split_measure, false, false);
            break;

        case OVO:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for OVO");
            learner->ocoder = CreateOneVsOneClassifier(parm, classes);
            OutputCodingTraining(parm, data, instances, labels, inputs, classes, index, subspace, subset, releasedata, learner);
            break;

        case RF:
            learner->rf = TrainRandomForest(parm, data, subspace, instances, labels, inputs, subset, index);
            break;

        case NN:
            MALLOC(learner->knn, t_kNN, 1, "ClassifierTraining: learner->knn");
            learner->knn->k = 1;
            learner->knn->localk = NULL;
            learner->knn->distance_to_furthest_neighbor = learner->knn->r = NULL;
            learner->knn->labels = NULL;
            break;

        case RForest:
        case ExtraTree:
            learner->forest = TrainRForest(parm, data, subspace, instances, labels, inputs, subset, index, classes, model == ExtraTree);
            break;

        case MultiBoost:
        case MadaBoost:
        case AdaBoostM1:
        case GentleBoost:
        case ReweightBoost:
        case ARCING:
            ensemble->boost = TrainAdaBoost(parm, data, instances, labels, inputs, classes, index, subset, subspace, model == MadaBoost, model == MultiBoost, model == GentleBoost,
                                            model == ReweightBoost, model == ARCING);
            break;

        case Bagging:
            ensemble->boost = TrainBagging(parm, data, instances, labels, inputs, classes, index, subset, subspace);
            break;

        case FloatBoost:
            ensemble->boost = TrainFloatBoost(parm, data, instances, labels, inputs, classes, index, subset, subspace);
            break;

        case FeatureBoost:
            ensemble->featboost = TrainFeatureBoostEnsemble(parm, inputs, data, instances, labels, classes, index, subset);
            break;
    }

    if (t_data) {
        FREE(t_data);
        FREE(t_labels);
    }
}

// Return the classifier margin
scalar ClassifierMargin(const t_Parameters *parm, t_Classifier model, scalar *instance, int inputs, int classes, int *index, int *subspace, int label, scalar **data, int instances,
                              int *labels, int *subset, t_Learner *learner, t_Ensemble *ensemble)
{
    scalar margin = -1.0;

    // Base classifier to use
    switch (model) {
        case SVM:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for SVMs");
            margin = SupportVectorMachineMargin(learner->svm, instance, label, subspace, classes, inputs);
            break;

        case C45:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for C4.5");
            margin = C45Margin(learner->tree, instance, label, classes, inputs, subspace);
            break;

        case kNN:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for kNN");
            margin = kNearestNeighbourMargin(instance, label, subspace, data, instances, inputs, learner->knn->k, labels, classes, subset);
            break;

        case NN:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for kNN");
            margin = kNearestNeighbourMargin(instance, label, subspace, data, instances, inputs, 1, labels, classes, subset);
            break;

        case MLP:
            margin = MLPMargin(parm, learner->mlp, instance, classes, label, index);
            break;

        case ELM:
            margin = ELMMargin(learner->elm, instance, inputs, classes, subspace, label, index);
            break;

        case LR:
            margin = LRMargin(learner->lr, instance, inputs, classes, subspace, label, index);
            break;

        case NBC:
            margin = NBCMargin(learner->nbc, instance, inputs, classes, subspace, label, index);
            break;

        case GNBC:
            margin = GaussianNBCMargin(parm, learner->gnbc, instance, inputs, classes, subspace, label, index);
            break;

        case IBT:
        case DT_C45:
        case DT_CA:
        case DT_ID3:
        case DT_CART:
            margin = DecisionTreeMargin(learner->dt, instance, label, classes, model == IBT, parm->ibt_do_not_use_internal_nodes, parm->ibt_predict_with_probabilities, index);
            break;

        case RForest:
        case ExtraTree:
            margin = RForestMargin(learner->forest, instance, label, classes, inputs, subspace, index);
            break;

        case OVO:
            // Not yet implemented
            margin = 0.0;
            break;

        case RF:
            margin = RandomForestMargin(learner->rf, instance, label, classes, inputs, subspace, index, parm->rf_ntree);
            break;

        case AdaBoostM1:
        case MadaBoost:
        case MultiBoost:
        case Bagging:
        case GentleBoost:
        case ReweightBoost:
        case FloatBoost:
        case ARCING:
            margin = AdaBoostMargin(parm, ensemble->boost, instance, label, classes, inputs, subspace, index, data, instances, labels, subset, NULL, model == ReweightBoost);
            break;

        case FeatureBoost:
            margin = FeatureBoostMargin(parm, ensemble->featboost, instance, label, classes, inputs, index, data, instances, labels, subset, NULL);
            break;
    }

    return margin;
}

void SetCrossValidatedParameters(t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, int inputs, int classes, t_Classifier model, int minority, int *subset, int **index)
{
    extern char *learner[];

    switch (model) {
        case SVM:
            CrossValidatedParametersSupportVectorMachine(parm, data, instances, labels, subspace, inputs, classes, minority, subset, index);
            break;

        case C45:
            CrossValidatedParametersC45(parm, data, instances, labels, subspace, inputs, classes, minority, subset);
            break;

        case kNN:
            CrossValidatedParameterskNN(parm, data, instances, labels, subspace, inputs, classes, minority, subset, index);
            break;

        case NN:
            // Nothing to do for 1-NN
            break;

        case MLP:
        case ELM:
        case NBC:
        case DT_C45:
        case DT_CA:
        case DT_CART:
        case DT_ID3:
        case IBT:
        case OVO:
        case RF:
        case GNBC:
        case LR:
        case RForest:
        case ExtraTree:
        case AdaBoostM1:
        case MultiBoost:
        case MadaBoost:
        case Bagging:
        case GentleBoost:
        case ReweightBoost:
        case FloatBoost:
        case FeatureBoost:
        case ARCING:
            if (parm->verbose > 0)
                printf("WARNING: No hyper-parameters CV available for %s\n", learner[model]);
            break;
    }
}

// Return output for 2-class problems
int Classifier2ClassOutput(const t_Parameters *parm, t_Classifier model, scalar *instance, int inputs, int classes, int *index, int *subspace, scalar **data, int instances, int *labels,
                                 int *subset, t_Learner *learner, int minority, scalar *output, int **data_index, t_Ensemble *ensemble)
{
    int category = -1.0, votes[classes], *nearest, ioutputs[2];
    scalar outputs[classes];
    extern float *ClassSum;

    if (classes != 2)
        OnErrorExit(ERR_PARAM, "Using 2-class classifier for M-class problem");

    // Check if null
    if (!learner->svm) { // Is an union, you can check any of them
        *output = 0.0;
        return 0;
    }

    // Base classifier to use
    switch (model) {
        case SVM:
            category = SupportVectorMachine2ClassOutput(learner->svm, instance, inputs, (double *) output, subspace, minority, index);
            break;

        case C45:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for C4.5");
            category = C45OutputValues(learner->tree, instance, subspace, inputs, classes, outputs);
            *output = ClassSum[minority] - ClassSum[(minority+1)%2];
            break;

        case kNN:
            ALLOCA(nearest, int, learner->knn->k, "Classifier2ClassOutput: nearest");
            category = kNNCategory(instance, data, subspace, instances, learner->knn->k, inputs, labels, classes, nearest, subset, votes, NULL, -1, data_index, index);
            *output = ((scalar) votes[minority] - votes[(minority+1)%2])/learner->knn->k;
            break;

        case MLP:
            category = NetworkActualOutput(parm, learner->mlp, instance, outputs, classes, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case ELM:
            category = ELMPredictClass(learner->elm, instance, inputs, classes, subspace, outputs, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case NBC:
            category = NBCOutput(learner->nbc, instance, inputs, classes, outputs, subspace, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case LR:
            category = LROutput(learner->lr, instance, inputs, classes, outputs, subspace, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case GNBC:
            category = GaussianNBCOutput(parm, learner->gnbc, instance, inputs, classes, outputs, subspace, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case DT_C45:
        case DT_CA:
        case DT_CART:
        case DT_ID3:
            category = DecisionTreeClassify(learner->dt, instance, outputs, classes, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case RForest:
        case ExtraTree:
            category = RForestClassify(learner->forest, instance, index, inputs, subspace, classes, outputs);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case IBT:
            category = IBTClassify(learner->dt, instance, outputs, classes, parm->ibt_do_not_use_internal_nodes, parm->ibt_predict_with_probabilities, index);
            *output = outputs[minority] - outputs[(minority+1)%2];
            break;

        case OVO:
            OnErrorExit(ERR_UNSPECTED_RESULT, "OVO should not be here");
            break;

        case RF:
            RandomForestClassify(learner->rf, instance, index, inputs, subspace, parm->rf_ntree, classes, ioutputs);
            *output = (scalar)(ioutputs[minority] - ioutputs[(minority+1)%2]);
            break;

        case NN:
            ALLOCA(nearest, int, learner->knn->k, "Classifier2ClassOutput: nearest");
            category = kNNCategory(instance, data, subspace, instances, 1, inputs, labels, classes, nearest, subset, votes, NULL, -1, data_index, index);
            *output = ((scalar) votes[minority] - votes[(minority+1)%2]);
            break;

        case AdaBoostM1:
        case MadaBoost:
        case MultiBoost:
        case Bagging:
        case GentleBoost:
        case FloatBoost:
        case ARCING:
            category = AdaBoostOutput(parm, ensemble->boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, -1, data_index, outputs);
            *output = ((scalar) outputs[minority] - outputs[(minority+1)%2]);
            break;

        case ReweightBoost:
            category = ReweightBoostOutput(parm, ensemble->boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, -1, data_index, outputs);
            *output = ((scalar) outputs[minority] - outputs[(minority+1)%2]);
            break;

        case FeatureBoost:
            category = FeatureBoostOutput(parm, ensemble->featboost, instance, inputs, classes, index, data, instances, labels, subset, -1, data_index, outputs);
            *output = ((scalar) outputs[minority] - outputs[(minority+1)%2]);
            break;
    }

    return category;
}

void ClassifierFree(const t_Parameters *parm, t_Learner *learner, t_Classifier classifier, int classes, int inputs, t_Ensemble *ensemble)
{
    switch (classifier) {
        case SVM:
            svm_free_and_destroy_model(&(learner->svm));
            break;

        case C45:
            ReleaseTree(learner->tree);
            break;

        case ELM:
            DestroyELM(learner->elm, classes);
            break;

        case MLP:
            FreeClassifier(parm, learner->mlp, classes);
            break;

        case RF:
            SHOWPROGRESS(parm->show_progress > 1, "WARNING: Random forest should implement a free functions%c", '\n');
            break;

        case LR:
            FreeLR(learner->lr);
            break;

        case kNN:
        case NN:
            FREE(learner->knn);
            break;

        case NBC:
            FreeNBC(learner->nbc, inputs, classes);
            break;

        case GNBC:
            GaussianFreeNBC(learner->gnbc, classes);
            break;

        case RForest:
        case ExtraTree:
            FreeRForest(learner->forest);
            break;

        case DT_C45:
        case DT_CA:
        case DT_CART:
        case DT_ID3:
        case IBT:
            FreeDecisionTree(learner->dt);
            FREE(learner->dt);
            break;

        case OVO:
            for (int b = 0; b < learner->ocoder->n_binary; b++)
                ClassifierFree(parm, learner->ocoder->binary[b], learner->ocoder->binary_learner, 2, inputs, NULL);
            FreeIntegerMatrix(&(learner->ocoder->M), classes);
            FREE(learner->ocoder->binary);
            break;

        case AdaBoostM1:
        case MadaBoost:
        case MultiBoost:
        case Bagging:
        case GentleBoost:
        case ReweightBoost:
        case FloatBoost:
        case ARCING:
            FreeAdaBoost(ensemble->boost, parm, classes, inputs);
            break;

        case FeatureBoost:
            FreeFeatureBoost(ensemble->featboost, parm, classes, inputs);
            break;
    }
}


// Return the category of an unseen instance together with probabilities

int ClassifierProbabilities(const t_Parameters *parm, t_Classifier model, scalar *instance, int inputs, int classes, int *index, int *subspace, scalar **data, int instances, int *labels,
                                  int *subset, t_Learner *learner, int label, cache_t **neighbors_cache, int inst_order, int **data_index, scalar *probabilities, t_Ensemble *ensemble)
{
    int category = -1.0, *nearest, votes[classes];

    // Base classifier to use
    switch (model) {
        case SVM:
            category = SupportVectorMachineOutput(learner->svm, instance, inputs, subspace, index, probabilities);
            break;

        case C45:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for C4.5");
            category = C45OutputValues(learner->tree, instance, subspace, inputs, classes, probabilities);
            break;

        case kNN:
            ALLOCA(nearest, int, learner->knn->k, "ClassifierOutput: nearest");
            category = kNNCategory(instance, data, subspace, instances, learner->knn->k, inputs, labels, classes, nearest, subset, votes, neighbors_cache, inst_order, data_index, index);
            for (int i = 0; i < classes; i++)
                probabilities[i] = votes[i];
            NormalizeVector(probabilities, classes);
            break;

        case MLP:
            category = NetworkActualOutput(parm, learner->mlp, instance, probabilities, classes, index);
            break;

        case ELM:
            category = ELMPredictClass(learner->elm, instance, inputs, classes, subspace, probabilities, index);
            break;

        case NBC:
            category = NBCOutput(learner->nbc, instance, inputs, classes, probabilities, subspace, index);
            break;

        case GNBC:
            category = GaussianNBCOutput(parm, learner->gnbc, instance, inputs, classes, probabilities, subspace, index);
            break;

        case LR:
            category = LROutput(learner->lr, instance, inputs, classes, probabilities, subspace, index);
            break;

        case DT_C45:
        case DT_CA:
        case DT_CART:
        case DT_ID3:
            category = DecisionTreeClassify(learner->dt, instance, probabilities, classes, index);
            break;

        case IBT:
            category = IBTClassify(learner->dt, instance, probabilities, classes, parm->ibt_do_not_use_internal_nodes, parm->ibt_predict_with_probabilities, index);
            break;

        case OVO:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Probabilities for OVO");
            break;

        case RF:
            category = RandomForestClassify(learner->rf, instance, index, inputs, subspace, parm->rf_ntree, classes, votes);
            if (probabilities) {
                for (int i = 0; i < classes; i++)
                    probabilities[i] = votes[i];
                NormalizeVector(probabilities, classes);
            }
            break;

        case NN:
            ALLOCA(nearest, int, 1, "ClassifierOutput: nearest");
            category = kNNCategory(instance, data, subspace, instances, 1, inputs, labels, classes, nearest, subset, votes, neighbors_cache, inst_order, data_index, index);
            for (int i = 0; i < classes; i++)
                probabilities[i] = votes[i];
            NormalizeVector(probabilities, classes);
            break;

        case RForest:
        case ExtraTree:
            category = RForestClassify(learner->forest, instance, index, inputs, subspace, classes, probabilities);
            break;

        case AdaBoostM1:
        case MadaBoost:
        case MultiBoost:
        case Bagging:
        case GentleBoost:
        case FloatBoost:
        case ARCING:
            category = AdaBoostOutput(parm, ensemble->boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, label, data_index, probabilities);
            break;

        case ReweightBoost:
            category = ReweightBoostOutput(parm, ensemble->boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, label, data_index, probabilities);
            break;

        case FeatureBoost:
            category = FeatureBoostOutput(parm, ensemble->featboost, instance, inputs, classes, index, data, instances, labels, subset, -1, data_index, probabilities);
            break;
    }

    return category;
}


void PrintClassifierPerformanceMetrics(const t_Parameters *parm, int instances, int inputs, scalar **data, int classes, int **index, int *subspace, int *subset, int *labels, t_Classifier classifier,
                                       t_Learner *model, t_Ensemble *ensemble, int minority)
{
    int **M;
    t_PerformanceInfo perf;
    extern char *learner[];

    // Create confusion matrix
    AllocateIntegerMatrix(&M, classes, classes);
    ZeroIntegerMatrix(M, classes, classes);
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < instances; i++) {
        int class = ClassifierOutput(parm, classifier, data[i], inputs, classes, index ? index[i] : NULL, subspace, data, instances, labels, subset, model, labels[i], NULL, i, index, ensemble);
#ifdef _OPENMP
        #pragma omp critical
#endif
        M[labels[i]][class]++;
    }

#ifdef _OPENMP
    #pragma omp critical
#endif
    PrintMulticlassMeasures(M, classes, instances, parm->out_file, learner[classifier], &perf);

    if (classes == 2) {
        int majority = (minority+1) % 2;
#ifdef _OPENMP
        #pragma omp critical
#endif
        PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], parm->out_file, learner[classifier], &perf);
    }

    FreeIntegerMatrix(&M, classes);
}
