/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Definitions and prototypes
#pragma once

#include <stdbool.h>
#include <stdio.h>
#include <time.h>
#include <gsl/gsl_matrix_double.h>
#include "typex.h"
#include "common.h"
#include "svm.h"
// #include "ensemble.h"
#include "data.h"
#include "model.h"

// Old definitions
// Information for scaling output values
typedef struct {
    scalar a, b;
} t_scale;

typedef struct {
    int nodes, outputs, inputs;
    scalar **prototypes, * *y, *nodes_output;
    scalar sigma2;
    scalar *mask; // Input mask, overlaps subspace
} t_GRNN;

// Network for projection of non-linear methods
typedef struct {
    int hidden1, hidden2, inputs, outputs;
    scalar **ih, * *ho, * *hh, A;
    scalar *mask;
    t_scale *scaling;
} t_ProjNet;


// Structure for classifiers
typedef struct {
    int nodes;                      // Number of hidden nodes
    scalar beta, w;
    scalar theta;                   // Possible threshold. 0 by default
    int inputs;                     // Actual number of inputs to the classifier
    scalar *subspace;               // Subspace mask

    scalar *sample;                 // Weights for patterns
    scalar **ih, **ho;

    scalar **nlp;                   // Non-linear/NDA projection
    scalar a, b;                    // Coefficients for NL projection
    t_Transfer h;
    t_GRNN *grnn;                   // Non-linear projection by means of a GRNN
    int n_grnns;
    t_ProjNet *projnet;             // Non-linear projection by means of a MLP
    int n_projnets;
    char *scca_file, *scca_data;

    Tree tree;                      // C4.5 classifier

    struct svm_model *smodel;       // SVM model
    struct svm_problem prob;        // SVM data

    // k-NN classifier
    int k, n;
    scalar **protos;
    scalar **desired;               // Desired only needed for IS without k-NN
    int *labels;
    bool weight_by_distance;
} t_ClassModel;

typedef t_ClassModel *t_OldEnsemble;
typedef t_ClassModel **t_DemocraticEnsembleMember;
typedef t_ClassModel ***t_DemocraticEnsemble;

///////////////////////////
// Single-label classifiers
///////////////////////////

// Extreme learning machine with a hidden layer
typedef struct {
    int n_nodes;
    scalar **ih_w;
    scalar **ho_w;
    t_Transfer h;
    scalar a, b;
} t_ELM;

// k-NN method
typedef struct {
    int k;
    int *localk;
    scalar *r, *distance_to_furthest_neighbor;
    int *labels;    // Only used for k-NN as base classifier for BR approach
} t_kNN;

// Naive Bayes classifier
typedef struct {
    scalar ** *freq, **limit;
    scalar *class_freq;
    int *n_intervals;
} t_NaiveBayes;

// Naive Bayes classifier
typedef struct {
    scalar **mean, **variance;
    scalar *class_freq;
} t_GaussianNaiveBayes;

// Local decision trees
struct dt_node {
    bool leaf;
    int depth, feat, class;
    scalar cut_point, score, *probabilities, alpha;
    struct dt_node *left, *right;
};

typedef struct dt_node *t_DTree;

typedef struct {
    int n_trees;
    t_DTree *trees;
} t_RForest;

// Output coding model
union learners; // Incomplete definition for cross-reference
typedef struct {
    int **M;
    int n_binary;
    t_Classifier binary_learner;
    union learners **binary;
} t_OutputCoder;

// Logistic regression
typedef struct {
    int n;
    scalar **w;
} t_LR;

// Classification model
typedef union learners {
    struct svm_model *svm;
    Tree tree;
    t_ClassModel *mlp;
    t_ELM *elm;
    t_kNN *knn;
    t_NaiveBayes *nbc;
    t_GaussianNaiveBayes *gnbc;
    t_DTree dt;
    t_OutputCoder *ocoder;
    struct Node **rf;
    t_LR *lr;
    t_RForest *forest;
} t_Learner;

///////////////////////////
// Ensembles of classifiers
///////////////////////////

typedef struct {
    int T, r;
    scalar *alpha;
    t_Learner *bcs;
    t_Classifier base;
} t_Boosting;

typedef struct {
    int T, **subspace;
    scalar *alpha;
    t_Learner *bcs;
    t_Classifier base;
} t_FeatureBoost;

typedef union {
    t_Boosting *boost;
    t_FeatureBoost *featboost;
} t_Ensemble;

//////////////////////////
// Multi-label classifiers
//////////////////////////

typedef struct {
    int k;
    scalar *prior, **post_0, **post_1;
} t_MLkNN;

typedef struct {
    int *ks;
    scalar *prior, **post_0, **post_1;
} t_MLksNN;

typedef struct {
    int k, n_clusters, *partition;
    scalar **centroids, **prior, ***post_0, ***post_1;
    int **centroids_index;
    bool clustered;
} t_LAMLkNN;

typedef struct {
    int **ks;
    t_LAMLkNN *lamlknn;
} t_LAMLksNN;

typedef struct {
    t_Classifier method;
    t_Learner *models;
    t_Ensemble *ensembles;
} t_BR;

typedef struct {
    t_Classifier method;
    t_Learner *models;
    t_Ensemble *ensembles;
    int *chain, n_diff, **diff;
} t_CC;

typedef struct {
    t_Classifier method;
    t_Learner *models;
    t_Ensemble *ensembles;
    int *k, m;
    int **h;
    scalar *threshold;
} t_RAkEL;

typedef struct {
    t_Classifier method;
    t_Learner model;
    t_Ensemble ensemble;
    int **decoded;
    int n_classes;
} t_LP;

typedef struct {
    t_Classifier method;
    bool *mask; // Some classifier may have no instances to train
    t_Learner **models;
    t_Ensemble **ensembles;
} t_CLR;

struct t_Leaf {
    int *labels, n, n_children, level;
    struct t_Leaf *children;
    t_Learner model;
    t_Ensemble ensemble;
};

typedef struct {
    t_Classifier method;
    int k;
    struct t_Leaf *hierarchy;
} t_HOMER;

typedef struct {
    int n_hidden;
    t_Transfer transfer_h, transfer_o;
    scalar **ih, **ho, a, b, a_output, b_output;
} t_MLMLP;

typedef struct {
    int M;
    scalar **w, **v, *w_t;
} t_BPMLL;

// ML decision trees
struct mldt_node {
    bool leaf;
    int depth, feat;
    scalar cut_point, score, *probabilities;
    struct mldt_node *left, *right;
};

typedef struct {
    scalar *w;
    struct mldt_node *nodes;
} t_MLDTree;

typedef struct {
    int size;
    t_MLDTree *trees;
    scalar threshold;
} t_MLRandomForest;

typedef struct {
    t_Classifier method;
    t_Learner *models;
    t_Ensemble *mensembles;
    t_Learner *predictors;
    t_Ensemble *pensembles;
} t_DBR;

typedef struct {
    int T;
    t_CC *ccs;
} t_ECC;

typedef struct {
    int k;
    t_LR *lrs;
} t_IBLR;

struct MLTreeNode {
    int n;
    t_Learner *models;
    t_Ensemble *ensembles;
    int *b;
    scalar *p;
    struct MLTreeNode *children;
};

typedef struct {
    t_Classifier method;    // Binary classifiers
    struct MLTreeNode *root;
} t_MLTree;

typedef struct {
    int size;
    t_MLTree *trees;
} t_MLForest;

typedef struct {
    scalar **probs;
    int k;
    scalar beta, t;
} t_MLRS;

typedef struct {
    scalar ***probs;
    int k, n_clusters, *partition;
    scalar beta, t;
} t_MLRSLC;

// Option ML decision trees
struct omldt_node {
    bool leaf;
    int depth, feat, n_options;
    scalar cut_point, score, *probabilities;
    struct omldt_node *left, *right;
    struct omldt_node *options;
};

typedef struct {
    struct omldt_node *root;
} t_OMLDTree;

typedef struct {
    t_MLkNN first_layer, second_layer;
    scalar **data;  // Data for second layer
} t_StackedMLkNN;

typedef struct {
    int k, **c;
    scalar *prior, *C, *Cprime;
} t_DMLkNN;

typedef struct {
    int k;
    scalar f, th, **u;
} t_MrKNN;

typedef struct {
    int k;
} t_DWkNN;

typedef struct {
    int k;
    scalar *prior, *ranges;
} t_FRONEC;

typedef struct {
    int k;
    scalar **fuzzy_labels;
} t_FVkNN;

typedef union {
    t_MLkNN mlknn;
    t_LAMLkNN lamlknn;
    t_MLksNN mlksnn;
    t_LAMLksNN lamlksnn;
    t_CC cc;
    t_BR br;
    t_RAkEL rakel;
    t_LP lp;
    t_CLR clr;
    t_HOMER homer;
    t_MLDTree mldt;
    t_MLMLP mlmlp;
    t_BPMLL bpmll;
    t_MLRandomForest mlrf;
    t_DBR dbr;
    t_ECC ecc;
    t_IBLR iblr;
    t_MLTree mltree;
    t_MLForest forest;
    t_MLRS mlrs;
    t_MLRSLC mlrslc;
    t_OMLDTree otree;
    t_StackedMLkNN smlknn;
    t_DMLkNN dmlknn;
    t_MrKNN mrknn;
    t_DWkNN dwknn;
    t_FRONEC fronec;
    t_FVkNN fvknn;
} t_MLLearner;

typedef struct {
    int T;
    scalar *alpha;
    t_MLLearner *bcs;
    t_MLClassifier base;
} t_MLBoosting;

typedef struct {
    int T;
    scalar *alpha;
    t_Learner **bcs;
    t_Classifier base;
} t_MLAdaBoost;

typedef struct {
    t_MLBoosting *boost;
    t_MLAdaBoost *adaboost;
    scalar th;
} t_MLEnsemble;

// Prototypes
/* moorepenrose.c */
gsl_matrix *moore_penrose_pinv(gsl_matrix *, const scalar);
void MoorePenroseGeneralizedInverse(scalar **, scalar **, int, int);
/* elm.c */
void CreateAndInitializeELM(const t_Parameters *, t_ELM *, int, int);
int ELMPredictClass(t_ELM *, scalar *, int, int, int *, scalar *, int *);
void TrainELM(t_ELM *, scalar **, int *, int, int, int, int *, int *, int **);
scalar ELMMargin(t_ELM *, scalar *, int, int, int *, int, int *);
void DestroyELM(t_ELM *, int);
/* svm.c */
struct svm_model *TrainSupportVectorMachine(const t_Parameters *, scalar **, int *, int, int *, int, int *, int **);
int SupportVectorMachineOutput(struct svm_model *, scalar *, int, int *, int *, scalar *);
scalar SupportVectorMachineMargin(struct svm_model *, scalar *, int, int *, int, int);
void CrossValidatedParametersSupportVectorMachine(t_Parameters *, scalar **, int, int *, int *, int, int, int, int *, int **);
int SupportVectorMachine2ClassOutput(struct svm_model *, scalar *, int, double *, int *, int, int *);
/* classifier.c */
void ClassifierTraining(const t_Parameters *, t_Classifier, scalar **, int, int *, int, int, int **, int *, int *, bool, t_Learner *, t_Ensemble *);
int ClassifierProbabilities(const t_Parameters *, t_Classifier, scalar *, int, int, int *, int *, scalar **, int, int *, int *, t_Learner *, int, cache_t **, int, int **, scalar *, t_Ensemble *);
int ClassifierOutput(const t_Parameters *, t_Classifier, scalar *, int, int, int *, int *, scalar **, int, int *, int *, t_Learner *, int, cache_t **, int, int **, t_Ensemble *);
scalar ClassifierMargin(const t_Parameters *, t_Classifier, scalar *, int, int, int *, int *, int, scalar **, int, int *, int *, t_Learner *, t_Ensemble *);
void SetCrossValidatedParameters(t_Parameters *, scalar **, int, int *, int *, int, int, t_Classifier, int, int *, int **);
int Classifier2ClassOutput(const t_Parameters *, t_Classifier, scalar *, int, int, int *, int *, scalar **, int, int *, int *, t_Learner *, int, scalar *, int **, t_Ensemble *);
void ClassifierFree(const t_Parameters *, t_Learner *, t_Classifier, int, int, t_Ensemble *);
void PrintClassifierPerformanceMetrics(const t_Parameters *, int, int, scalar **, int, int **, int *, int *, int *, t_Classifier, t_Learner *, t_Ensemble *, int);
/* c45.c */
Tree TrainC45(const t_Parameters *, scalar **, int *, int, int *, int, bool, int *, int);
int C45Output(Tree, scalar *, int *, int);
int C45OutputValues(Tree, scalar *, int *, int, int, scalar *);
scalar C45Margin(Tree, scalar *, int, int, int, int *);
void CrossValidatedParametersC45(t_Parameters *, scalar **, int, int *, int *, int, int, int, int *);
/* knn.c */
int kNNCategory(scalar *, scalar **, int *, int, int, int, int *, int, int *, int *, int *, cache_t **, int, int **, int *);
scalar kNearestNeighbourMargin(scalar *, int, int *, scalar **, int, int, int, int *, int, int *);
scalar NearestNeighbourMargin(scalar *, int, int *, scalar **, int, int, int *, int *);
void CrossValidatedParameterskNN(t_Parameters *, scalar **, int, int *, int *, int, int, int, int *, int **);
int ObtainkCrossValidation(t_Parameters *, int *, scalar **, int, int *, int, int, int **, int *subspace);
void kNNOfOneClass(scalar *, scalar **, int, int, int, int, int *, int *);
/* mlp.c */
scalar MLPMargin(const t_Parameters *, t_ClassModel *, scalar *, int, int, int *);
t_ClassModel * TrainMultiLayerPerceptron(const t_Parameters *, scalar **, int, int, int, int *, int *, int *, int **);
/* knn-cache.c */
void WeightVotes(scalar *, int, scalar *, int *, int *, int);
int kNearestNeighbourCached(int, int *, scalar **, int, int, int *, int, bool);
int kNearestNeighbourWithListCached(int, int *, int, int, int *, int *, scalar **, int, bool);
int kNearestNeighbourCache(int, int *, int, int *, int *, int, cache_t **, scalar **, int, int, bool, scalar *, scalar *, int **, int *);
/* evaluation.c */
void ClassifierEvaluation(const t_Parameters *, scalar **, int, int *, scalar **, int, int *, int *, int, int *, bool, int, t_Classifier, t_Learner *, int, int **, int **, t_PerformanceInfo *,
                              t_PerformanceInfo *, FILE *, char *, char *, cache_t **, int **, int **, t_Ensemble *);
scalar SingleClassifierClassifyDataset(const t_Parameters *, scalar **, int, int *, scalar **, int, int *, int *, int, int *, bool, int, t_Classifier, t_Learner *, int, bool *, int *, t_Ensemble *);
scalar SingleClassifierWeightedError(const t_Parameters *, t_Classifier, int *, scalar **, int, int, int *, int, bool *, scalar *, t_Learner *, scalar *, scalar **, int , int *);
void kFoldClassifierEvaluation(const t_Parameters *, scalar **, int, int *, int *, int, int *, int, t_Classifier, int, int **, t_PerformanceInfo *, FILE *, char *,  int);
/* auc.c */
scalar f(scalar, scalar **, int);
scalar trapzd(scalar **, scalar a, scalar b, int, int nvalues);
scalar qtrap(scalar **, scalar a, scalar b, int);
scalar FastAUC(scalar **curve, int);
scalar ObtainAUROC(const t_Parameters *, t_Classifier,  scalar **, int, int *, scalar **, int, int *, int *, int *, int, int, bool, t_Learner *, bool, FILE *, scalar *, scalar *, int, int **, int **);
void GetPositiveAndNegativeInstancesOutputs(const t_Parameters *, t_Classifier, scalar **, int, int *, scalar **, int, int *, scalar *, scalar *, int, int, t_Learner *, int, int **, int **, int *,
                                            int *, t_Ensemble *, bool);
scalar ClassifierAUPRC(const t_Parameters *, t_Classifier, scalar **, int, int *, scalar **, int, int *, int, int, bool, t_Learner *, FILE *, int, int **, int **, int *, int *, scalar *,
                             scalar *, t_Ensemble *, bool);
scalar ClassifierAUROC(const t_Parameters *, t_Classifier, scalar **, int, int *, scalar **, int, int *, int *, int *, int, int, bool, t_Learner *, bool, FILE *, int, int **, int **, scalar *,
                             scalar *, t_Ensemble *, bool);
/* margin.c */
scalar ClassifierDatasetMargin(const t_Parameters *, t_Classifier, scalar **, int, int *, scalar **, int, int *, int, int, int *, scalar *, t_Learner *, scalar *, int *, t_Ensemble *, bool);
/* test.c */
void TestAllClassifiers(const t_Parameters *, scalar **, int, int *, int, scalar **, int, int *, int, int *, int *, char *, int, bool, t_PerformanceInfo *, char *, cache_t **, int **,
                        int **, char *);
/* nbc. */
int NBCOutput(t_NaiveBayes *, scalar *, int, int, scalar *, int *, int *);
void TrainNBCClassifier(const t_Parameters *, t_NaiveBayes *, int, int *, int, scalar **, int, int *, int **);
scalar NBCMargin(t_NaiveBayes *, scalar *, int, int, int *, int, int *);
void FreeNBC(t_NaiveBayes *, int, int);
/* trees.c */
int DecisionTreeClassify(t_DTree, scalar *, scalar *, int, int *);
void DecisionTreeBuild(const t_Parameters *, t_DTree, scalar **, int *, int, int *, int, int *, scalar, int, int, scalar *, int, int **, t_DTSplitMeasure split_measure, bool, bool);
t_DTree DecisionTreeTrain(const t_Parameters *, scalar **, int, int *, int, int *, int *, int, bool, int **, t_DTSplitMeasure, bool, bool);
int GetMajorityClass(int *, int *, int, int, scalar *);
void PrintDTree(const t_Parameters *, t_DTree, FILE *, bool, t_DTSplitMeasure);
void PrintDTNode(t_DTree, FILE *, bool);
void FreeDecisionTree(t_DTree);
scalar DecisionTreeMargin(t_DTree, scalar *, int, int, bool, bool, bool, int *);
void CountAllNodes(t_DTree, int *);
int CountDTNodes(t_DTree);
/* split.c */
int CompareFeature(const void *, const void *);
int MajorityClass(scalar *, int);
void GetBestCupPoint(const t_Parameters *, scalar **, int *, int, int *, int, scalar *, scalar *, int, scalar *, int **, t_DTSplitMeasure, bool);
scalar InformationGain(int, int, scalar, scalar *, scalar *, int);
scalar GainRatio(int, int, scalar, scalar *, scalar *);
scalar DTClassEntropy(int, int, scalar *);
scalar Gini(int, int, scalar *);
scalar GiniSplit(int, int, scalar *, scalar *);
/* ibt.c */
scalar *AllocateCopyAndUpdateInstanceWeights(const t_Parameters *, scalar *, int, int *, int, int *, int);
int IBTClassify(t_DTree, scalar *, scalar *, int, bool, bool, int *);
void InternallyBoostedTreeClassify(t_DTree, scalar *, scalar *, int, bool, int *);
scalar GetAlpha(const t_Parameters *, scalar *, int, int *, int, int *);
/* ovo.c */
int CreateOVOMatrix(int ***, int);
int OneVsOneOutput(const t_Parameters *, t_OutputCoder *, scalar *, int *, int, int, scalar **, int, int *, int *);
void OutputCodingTraining(const t_Parameters *, scalar **, int, int *, int, int, int **, int *, int *, bool, t_Learner *);
t_OutputCoder *CreateOneVsOneClassifier(const t_Parameters *, int);
/* measures.c */
void PrintMulticlassMeasures(int **, int, int, FILE *, char *, t_PerformanceInfo *);
scalar ConfusionEntropy(int **, int);
scalar h(scalar, scalar);
/* rf.c */
int RandomForestClassify(struct Node **, scalar *, int *, int, int *, int, int, int *);
struct Node **TrainRandomForest(const t_Parameters *, scalar **, int *, int, int *, int, int *, int **);
int RFClassifyInstance(struct Node **, int, float *, int, int *);
scalar RandomForestMargin(struct Node **, scalar *, int, int, int, int *, int *, int);
/* gnbc.c */
void TrainGaussianNBCClassifier(t_GaussianNaiveBayes *, int, int *, int, scalar **, int, int *, int **);
int GaussianNBCOutput(const t_Parameters *, t_GaussianNaiveBayes *, scalar *, int, int, scalar *, int *, int *);
scalar GaussianNBCMargin(const t_Parameters *, t_GaussianNaiveBayes *, scalar *, int, int, int *, int, int *);
void GaussianFreeNBC(t_GaussianNaiveBayes *, int);
/* lr.h */
scalar hx(scalar *, scalar *, int, int *, int *);
void TrainLogisticRegressionClassifier(const t_Parameters *, t_LR *, int, int *, int, scalar **, int, int *, int *, int **);
void FreeLR(t_LR *);
int LROutput(t_LR *lr, scalar *, int, int, scalar *, int *, int *);
scalar LRMargin(t_LR *, scalar *, int, int, int *, int, int *);
void TrainLogisticRegressionClassifierFor2Classes(const t_Parameters *,scalar *, int, int *, int, scalar **, int *, int *, int **);
int LR2ClassOutput(scalar *, scalar *, int, scalar *, int *, int *);
/* rforest.c */
t_RForest *TrainRForest(const t_Parameters *, scalar **, int *, int, int *, int, int *, int **, int, bool);
int RForestClassify(t_RForest *, scalar *, int *, int, int *, int , scalar *);
scalar RForestMargin(t_RForest *, scalar *, int, int, int, int *, int *);
void FreeRForest(t_RForest *);
