/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "common.h"
#include "typex.h"
#include "macros.h"
#include "typex.h"
#include "classifiers.h"
#include "ensemble.h"

void CreateAndInitializeELM(const t_Parameters *parm, t_ELM *elm, int inputs, int classes)
{
    // Check normalized inputs
    if (!parm->typify)
        OnErrorExit(ERR_PARAM, "ELM must work with normalized inputs");

    elm->n_nodes = parm->elm_n_hidden;

    // Allocate memory
    MALLOC(elm->ih_w, scalar *, elm->n_nodes, "CreateAndInitializeELM: elm->ih_w");
    for (int i = 0; i < elm->n_nodes; i++)
        MALLOC(elm->ih_w[i], scalar, inputs+1, "CreateAndInitializeELM: elm->ih_w[i]");

    MALLOC(elm->ho_w, scalar *, classes, "CreateAndInitializeELM: elm->ho_w");
    for (int i = 0; i < classes; i++)
        MALLOC(elm->ho_w[i], scalar, elm->n_nodes, "CreateAndInitializeELM: elm->ho_w[i]");

    elm->h = parm->elm_tf;
    elm->a = parm->elm_a;
    elm->b = parm->elm_b;

    // Initialize weights
    for (int i = 0; i < elm->n_nodes; i++)
        for (int j = 0; j < inputs+1; j++)
            elm->ih_w[i][j] = RANDOM(-1.0, 1.0);

    for (int i = 0; i < classes; i++)
        for (int j = 0; j < elm->n_nodes; j++)
            elm->ho_w[i][j] = RANDOM(-1.0, 1.0);
}

int ELMPredictClass(t_ELM *elm, scalar *instance, int inputs, int classes, int *subspace, scalar *routputs, int *index)
{
    scalar h_outputs[elm->n_nodes], outputs[classes];

    // Hidden layer output
    for (int i = 0; i < elm->n_nodes; i++) {
        h_outputs[i] = elm->ih_w[i][inputs];    // Bias input
        if (index) {
            if (subspace)
                for (int j = 0; index[j] != -1; j++)
                    h_outputs[i] += elm->ih_w[i][index[j]-1] * instance[j] * subspace[index[j]-1];
            else
                for (int j = 0; index[j] != -1; j++)
                    h_outputs[i] += elm->ih_w[i][index[j]-1] * instance[j];
        } else {
            if (subspace)
                for (int j = 0; j < inputs; j++)
                    h_outputs[i] += elm->ih_w[i][j] * instance[j] * subspace[j];
            else
                for (int j = 0; j < inputs; j++)
                    h_outputs[i] += elm->ih_w[i][j] * instance[j];
        }

        h_outputs[i] = TransferFunction(h_outputs[i], elm->h, elm->a, elm->b);
    }

    // Output layer
    ZeroVector(outputs, NULL, classes);
    for (int i = 0; i < classes; i++)
        for (int j = 0; j < elm->n_nodes; j++)
            outputs[i] += elm->ho_w[i][j] * h_outputs[j];

    // Get maximum
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (outputs[i] > outputs[category])
            category = i;

    if (routputs)
        memcpy(routputs, outputs, classes*sizeof(scalar));

    return category;
}

void TrainELM(t_ELM *elm, scalar **data, int *labels, int instances, int inputs, int classes, int *subset, int *subspace, int **index)
{
    scalar **H, **H_pinv, **T;
    int c = 0;

    // Create H matrix
    MALLOC(H, scalar *, instances, "TrainELM: H");
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            MALLOC(H[c], scalar, elm->n_nodes, "TrainELM: H[c]");

            for (int j = 0; j < elm->n_nodes; j++) {
                H[c][j] = elm->ih_w[j][inputs];
                if (index) {
                    if (subspace)
                        for (int k = 0; index[c][k] != -1; k++)
                            H[c][j] += elm->ih_w[j][index[c][k]-1] * data[c][k] * subspace[index[c][k]-1];
                    else
                        for (int k = 0; index[c][k] != -1; k++)
                            H[c][j] += elm->ih_w[j][index[c][k]-1] * data[c][k];
                } else {
                    if (subspace)
                        for (int k = 0; k < inputs; k++)
                            H[c][j] += elm->ih_w[j][k] * data[c][k] * subspace[k];
                    else
                        for (int k = 0; k < inputs; k++)
                            H[c][j] += elm->ih_w[j][k] * data[c][k];
                }

                H[c][j] = TransferFunction(H[c][j], elm->h, elm->a, elm->b);
            }

            c++;
        }

    // Get pseudo-inverse
    AllocateScalarMatrix(&H_pinv, elm->n_nodes, c);
    MoorePenroseGeneralizedInverse(H, H_pinv, c, elm->n_nodes);

    // T matrix (desired outputs)
    AllocateScalarMatrix(&T, c, classes);
    c = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            for (int j = 0; j < classes; j++)
                T[c][j] = (j == labels[c]) ? 1.0 : -1.0;
            c++;
        }

    // Store output weights
    for (int i = 0; i < elm->n_nodes; i++)
        for (int j = 0; j < classes; j++) {
            elm->ho_w[j][i] = 0.0;
            for (int k = 0; k < c; k++)
                elm->ho_w[j][i] += H_pinv[i][k] * T[k][j];
        }

    // Free memory
    FreeScalarMatrix(&H, c);
    FreeScalarMatrix(&H_pinv, elm->n_nodes);
    FreeScalarMatrix(&T, c);
}

scalar ELMMargin(t_ELM *elm, scalar *instance, int inputs, int classes, int *subspace, int label, int *index)
{
    scalar h_outputs[elm->n_nodes], outputs[classes];

    // Hidden layer output
    for (int i = 0; i < elm->n_nodes; i++) {
        h_outputs[i] = elm->ih_w[i][inputs];    // Bias input
        if (index) {
            if (subspace)
                for (int j = 0; index[j] != -1; j++)
                    h_outputs[i] += elm->ih_w[i][index[j]-1] * instance[j] * subspace[index[j]-1];
            else
                for (int j = 0; index[j] != -1; j++)
                    h_outputs[i] += elm->ih_w[i][index[j]-1] * instance[j];
        } else {
            if (subspace)
                for (int j = 0; j < inputs; j++)
                    h_outputs[i] += elm->ih_w[i][j] * instance[j] * subspace[j];
            else
                for (int j = 0; j < inputs; j++)
                    h_outputs[i] += elm->ih_w[i][j] * instance[j];
        }

        h_outputs[i] = TransferFunction(h_outputs[i], elm->h, elm->a, elm->b);
    }

    // Output layer
    ZeroVector(outputs, NULL, classes);
    for (int i = 0; i < classes; i++)
        for (int j = 0; j < elm->n_nodes; j++)
            outputs[i] += elm->ho_w[i][j] * h_outputs[j];

    // Get maximum
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (outputs[i] > outputs[category])
            category = i;

    // Wrong classification? Return negative margin
    if (category != label)
        return outputs[label] - outputs[category];

    // Otherwise get second maximum
    int second = (category != 0) ? 0 : 1;
    for (int i = 0; i < classes; i++)
        if (i != category && outputs[i] > outputs[second])
            second = i;

    return outputs[category] - outputs[second];
}

void DestroyELM(t_ELM *elm, int classes)
{
    for (int i = 0; i < elm->n_nodes; i++)
        FREE(elm->ih_w[i]);
    FREE(elm->ih_w);

    for (int i = 0; i < classes; i++)
        FREE(elm->ho_w[i]);
    FREE(elm->ho_w);
}
