/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdbool.h>
#include "macros.h"
#include "classifiers.h"
#include "svm.h"
#include <omp.h>

void ClassifierEvaluation(const t_Parameters *parm, scalar **data, int instances, int *labels, scalar **train_data, int train_instances, int *train_labels, int *subspace, int inputs,
                                int *subset, bool cv, int classes, t_Classifier classifier, t_Learner *return_learner, int minority, int **M, int **M_training, t_PerformanceInfo *perf,
                                t_PerformanceInfo *perf_training, FILE *out_file, char *text, char *text_training, cache_t **neighbors_cache, int **data_index, int **train_index,
                                t_Ensemble *return_ensemble)
{
    int majority;
    t_Learner learner;
    t_Ensemble ensemble;

    // Check number of selected instances
    int selected = train_instances;
    if (subset) {
        selected = 0;
        for (int i = 0; i < train_instances; i++)
            if (subset[i])
                selected++;
    }

    // Cross-validate?
    if (cv && selected > 10) // With less instances do not CV
        SetCrossValidatedParameters((t_Parameters *) parm, train_data, train_instances, train_labels, subspace, inputs, classes, classifier, minority, subset, train_index);
    else if (classifier == SVM)
        printf("WARNING: SVM should be used with cross-validation of hyper-parameters\n");

    // Train
    ClassifierTraining(parm, classifier, train_data, train_instances, train_labels, inputs, classes, train_index, subspace, subset, true, &learner, &ensemble);

    // Test data
    // Create confusion matrix
    ZeroIntegerMatrix(M, classes, classes);
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < instances; i++) {
        int class = ClassifierOutput(parm, classifier, data[i], inputs, classes, data_index ? data_index[i] : NULL, subspace, train_data, train_instances, train_labels, subset, &learner,
                                               labels[i], neighbors_cache, i, train_index, &ensemble);
#ifdef _OPENMP
        #pragma omp critical
#endif
        M[labels[i]][class]++;
    }

    PrintMulticlassMeasures(M, classes, instances, out_file, text, perf);
    if (classes == 2) {
        majority = (minority+1) % 2;
        PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], out_file, text, perf);
    }

    // Training error
    if (M_training && perf_training) {
        ZeroIntegerMatrix(M_training, classes, classes);

        // Create confusion matrix
        for (int i = 0; i < train_instances; i++)
            M_training[train_labels[i]][ClassifierOutput(parm, classifier, train_data[i], inputs, classes, train_index ? train_index[i] : NULL, subspace, train_data, train_instances,
                                        train_labels, subset, &learner, train_labels[i], neighbors_cache, i, train_index, &ensemble)]++;

        PrintMulticlassMeasures(M_training, classes, train_instances, out_file, text_training, perf_training);
        if (classes == 2)
            PrintTwoClassMeasures(M_training[minority][minority], M_training[majority][majority], M_training[majority][minority], M_training[minority][majority], out_file, text_training,
                                  perf_training);
    }

    // Return models?
    if (return_learner) {
        *return_learner = learner;
        *return_ensemble = ensemble;
    } else
        ClassifierFree(parm, &learner, classifier, classes, inputs, NULL);
}

// Return error
scalar SingleClassifierClassifyDataset(const t_Parameters *parm, scalar **data, int instances, int *labels, scalar **train_data, int train_instances, int *train_labels, int *subspace, int inputs,
                                       int *subset, bool cv, int classes, t_Classifier classifier, t_Learner *return_learner, int minority, bool *ok, int *prediction, t_Ensemble *return_ensemble)
{
    t_Learner learner;
    t_Ensemble ensemble;
    scalar acc = 0.0;

    // Cross-validate?
    if (cv)
        SetCrossValidatedParameters((t_Parameters *) parm, train_data, train_instances, train_labels, subspace, inputs, classes, classifier, minority, subset, NULL);

    // Train
    ClassifierTraining(parm, classifier, train_data, train_instances, train_labels, inputs, classes, NULL, subspace, subset, true, &learner, &ensemble);

    // Test data
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < instances; i++) {
        int class = ClassifierOutput(parm, classifier, data[i], inputs, classes, NULL, subspace, train_data, train_instances, train_labels, subset, &learner, labels[i], NULL, -1, NULL, &ensemble);

        if (ok)
            ok[i] = (class == labels[i]);

        if (prediction)
            prediction[i] = class;
    }

    // Return models?
    if (return_learner) {
        *return_learner = learner;
        *return_ensemble = ensemble;
    } else
        ClassifierFree(parm, &learner, classifier, classes, inputs, NULL);

    return 1.0 - acc/instances;
}

// Get weighted error and vector of classified instances
scalar SingleClassifierWeightedError(const t_Parameters *parm, t_Classifier classifier, int *subspace, scalar **data, int inputs, int instances, int *labels, int classes, bool *ok, scalar *w,
                                     t_Learner *learner, scalar *h, scalar **t_data, int t_instances, int *t_labels)
{
    scalar error = 0.0, *local_w = NULL, lh;
    t_Learner llearner;
    t_Ensemble lensemble;

    // If no weights given get equal distribution
    if (!w) {
        MALLOC(local_w, scalar, instances, "GetWeightedErrorFS: local_w");
        for (int i = 0; i < instances; i++)
            local_w[i] = 1.0/instances;
    } else
        local_w = w;

    if (parm->weigthed_sampling_for_training)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted undersampling for boost FS method");

    // Train classifier?
    if (!learner)
        ClassifierTraining(parm, classifier, t_data ? t_data : data, t_instances ? t_instances : instances, t_labels ? t_labels : labels, inputs, classes, NULL, subspace, NULL, true,
                                 &llearner, &lensemble);
    else
        llearner = *learner;

    // Get margin
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic) reduction(+: error)
#endif
    for (int k = 0; k < instances; k++) {
        lh = ClassifierMargin(parm, classifier, data[k], inputs, classes, NULL, subspace, labels[k], data, instances, labels, NULL, &llearner, &lensemble);

        if (h)
            h[k] = lh;

        ok[k] = (lh > 0.0);
        if (!ok[k])
            error += local_w[k];
    }

    return error;
}

void kFoldClassifierEvaluation(const t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, int inputs, int *subset, int classes, t_Classifier classifier, int minority,
                                     int **M, t_PerformanceInfo *perf, FILE *out_file, char *text,  int folds)
{
    int majority, *labels_t, *labels_v, instances_v, instances_t;
    t_Learner learner;
    t_Ensemble lensemble;
    scalar **data_v, **data_t;

    if (folds == 1)
        OnErrorExit(ERR_PARAM, "CV with just one fold");

    // If first call allocate memory
    MALLOC(labels_v, int, instances, "kFoldClassifierEvaluation: labels_v");
    MALLOC(labels_t, int, instances, "kFoldClassifierEvaluation: labels_t");
    MALLOC(data_v, scalar *, instances, "kFoldClassifierEvaluation: data_v");
    MALLOC(data_t, scalar *, instances, "kFoldClassifierEvaluation: data_t");

    // Size of validation set
    instances_v = instances/parm->cv_folds;
    instances_t = instances - instances_v;

    ZeroIntegerMatrix(M, classes, classes);

    for (int fold = 0; fold < parm->cv_folds; fold ++) {
        // Validation set
        for (int k = (fold * instances_v) % instances, c = 0; c < instances_v; k++, c++) {
            data_v[c] = data[k];
            labels_v[c] = labels[k];
        }

        // Train set
        int next = 0;
        for (int k = 0, c = ((fold+1) * instances_v) % instances; k < instances_t; k++, c = (c+1) % instances)
            if (!subset || subset[c]) {
                data_t[next] = data[c];
                labels_t[next] = labels[c];
                next++;
            }

        // Train
        ClassifierTraining(parm, classifier, data_t, instances_t, labels_t, inputs, classes, NULL, subspace, subset, true, &learner, &lensemble);

        // Update confusion matrix
        for (int i = 0; i < instances_v; i++)
            M[labels_v[i]][ClassifierOutput(parm, classifier, data_v[i], inputs, classes, NULL, subspace, data_t, instances_t, labels_t, subset, &learner, labels_v[i], NULL, -1, NULL, &lensemble)]++;

        ClassifierFree(parm, &learner, classifier, classes, inputs, NULL);
    }

    PrintMulticlassMeasures(M, classes, instances, out_file, text, perf);
    if (classes == 2) {
        majority = (minority+1) % 2;
        PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], out_file, text, perf);
    }

    FREE(labels_v);
    FREE(labels_t);
    FREE(data_v);
    FREE(data_t)
}
