/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "typex.h"
#include "macros.h"
#include "classifiers.h"
#include "data.h"

#define PI      3.14159265358979323846
#define EPSILON 0.1

// Implements Gaussian Naive Bayes Classifier

void TrainGaussianNBCClassifier(t_GaussianNaiveBayes *cl, int instances, int *labels, int inputs, scalar **data, int classes, int *subset, int **index)
{
    int count = 0;

    // Memory for info
    // Last value is for general average
    MALLOC(cl->class_freq, scalar, classes, "TrainGaussianNBCClassifier: cl->class_freq");
    AllocateScalarMatrix(&(cl->mean), classes, inputs);
    AllocateScalarMatrix(&(cl->variance), classes, inputs);
    for (int i = 0; i < classes; i++) {
        cl->class_freq[i] = 0.0;
        for (int j = 0; j < inputs; j++)
            cl->mean[i][j] = cl->variance[i][j] = 0.0;
    }

    // Count class frequencies and other statistics
    // Compute first mean and then variance to avoid numerical problems
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            count++;

            cl->class_freq[labels[i]]++;
            if (index)
                for (int j = 0; index[i][j] != -1; j++)
                    cl->mean[labels[i]][index[i][j]-1] += data[i][j];
            else
                for (int j = 0; j < inputs; j++)
                    cl->mean[labels[i]][j] += data[i][j];
        }

    // Mean
    for (int i = 0; i < classes; i++)
        if (cl->class_freq[i] > 0.0)
            for (int j = 0; j < inputs; j++)
                cl->mean[i][j] /= cl->class_freq[i];

    // Variance
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            if (index)
                for (int j = 0; index[i][j] != -1; j++)
                    cl->variance[labels[i]][index[i][j]-1] += pow(data[i][j] - cl->mean[labels[i]][index[i][j]-1], 2.0);
            else
                for (int j = 0; j < inputs; j++)
                    cl->variance[labels[i]][j] += pow(data[i][j] - cl->mean[labels[i]][j], 2.0);
        }

    for (int i = 0; i < classes; i++) {
        if (cl->class_freq[i] > 1)
            for (int j = 0; j < inputs; j++)
                cl->variance[i][j] = cl->variance[i][j]/cl->class_freq[i] + EPSILON;
        else
            FillVector(cl->variance[i], NULL, inputs, EPSILON);

        cl->class_freq[i] /= count;
    }
}

int GaussianNBCOutput(const t_Parameters *parm, t_GaussianNaiveBayes *cl, scalar *instance, int inputs, int classes, scalar *probabilities, int *subspace, int *index)
{
    scalar probs[classes], Z = 0.0;

    // Obtain probability of each class
    for (int i = 0; i < classes; i++)
        probs[i] =  parm->gnbc_use_logs ? log(cl->class_freq[i]) : cl->class_freq[i];

    if (parm->gnbc_use_logs) {
        for (int i = 0; i < classes; i++)
            if (index) {
                for (int j = 0; index[j] != -1; j++)
                    if (!subspace || subspace[index[j]-1])
                        probs[i] += log(MAX(0.0000000001, (1.0/sqrt(2.0*PI*cl->variance[i][index[j]-1]))*exp(-(pow(instance[j] - cl->mean[i][index[j]-1], 2.0)/(2.0*cl->variance[i][index[j]-1])))));
            } else
                for (int j = 0; j < inputs; j++)
                    if (!subspace || subspace[j])
                        probs[i] += log(MAX(0.0000000001, (1.0/sqrt(2.0*PI*cl->variance[i][j]))*exp(-(pow(instance[j] - cl->mean[i][j], 2.0)/(2.0*cl->variance[i][j])))));

        scalar A = probs[0];
        for (int i = 1; i < classes; i++)
            if (probs[i] > A)
                A = probs[i];

        for (int i = 0; i < classes; i++)
            Z += exp(probs[i] - A);
        Z = A + log(Z);

        for (int i = 0; i < classes; i++)
            probs[i] = exp(probs[i] - Z);
    } else {
        for (int i = 0; i < classes; i++)
            if (index) {
                for (int j = 0; index[j] != -1; j++)
                    if (!subspace || subspace[index[j]-1])
                        probs[i] *= (1.0/sqrt(2.0*PI*cl->variance[i][index[j]-1]))*exp(-(pow(instance[j] - cl->mean[i][index[j]-1], 2.0)/(2.0*cl->variance[i][index[j]-1])));
            } else
                for (int j = 0; j < inputs; j++)
                    if (!subspace || subspace[j])
                        probs[i] *= (1.0/sqrt(2.0*PI*cl->variance[i][j]))*exp(-(pow(instance[j] - cl->mean[i][j], 2.0)/(2.0*cl->variance[i][j])));

        for (int i = 0; i < classes; i++)
            Z += probs[i];

        for (int i = 0; i < classes; i++)
            probs[i] = Z > ZERO ? probs[i]/Z : cl->class_freq[i];
    }

    int category = 0;
    for (int i = 1; i < classes; i++)
        if (probs[i] > probs[category] || (probs[i] == probs[category] && cl->class_freq[i] > cl->class_freq[category]))
            category = i;

    if (probabilities)
        for (int i = 0; i < classes; i++)
            probabilities[i] = probs[i];

    return category;
}

scalar GaussianNBCMargin(const t_Parameters *parm, t_GaussianNaiveBayes *cl, scalar *instance, int inputs, int classes, int *subspace, int label, int *index)
{
    scalar probabilities[classes];

    GaussianNBCOutput(parm, cl, instance, inputs, classes, probabilities, subspace, index);

    scalar wrong_probs = -1.0;
    for (int j = 0; j < classes; j++)
        if (wrong_probs < probabilities[j] && j != label)
            wrong_probs = probabilities[j];

    if (isnan(probabilities[label] - wrong_probs) || isinf(probabilities[label] - wrong_probs))
        OnErrorExit(ERR_NUMERICAL, "Margin infinite or NaN");

    return probabilities[label] - wrong_probs;
}

void GaussianFreeNBC(t_GaussianNaiveBayes *cl, int classes)
{
    FreeScalarMatrix(&(cl->mean), classes);
    FreeScalarMatrix(&(cl->variance), classes);
    FREE(cl->class_freq);
    FREE(cl);
}
