/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <stdbool.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"

// Specific functions for internally boosted trees (IBTs)

// Update weights following AdaBoost scheme or undo a previous update

scalar *AllocateCopyAndUpdateInstanceWeights(const t_Parameters *parm, scalar *w, int instances, int *labels, int majority, int *list, int train_instances)
{
    scalar e_i = 0.0, sum = 0.0, *new_w;
    static int t = 0;

    MALLOC(new_w, scalar, train_instances, "AllocateCopyAndUpdateInstanceWeights: train_instances");

    // Consider different ensemble methods
    switch (parm->ibt_ens_method) {
        case NoBoost:
            // All instances equally weighted
            for (int i = 0; i < instances; i++)
                new_w[list[i]] = 1.0/instances;
            break;

        case eMultiBoost:
            t++;
            if (!(t % parm->ibt_multiboost_I)) {    // Restart weights to Poisson distribution
                sum = 0.0;
                for (int k = 0; k < instances; k++)
                    sum += (new_w[list[k]] = Poisson());

                // Normalize w[] to sum 1
                for (int k = 0; k < instances; k++)
                    new_w[list[k]] /= sum;

                SHOWPROGRESS(parm->show_progress > 0, "eMultiBoost reinit weights at node %d\n", t);
                break;
            }

        case AdaBoost:
            // Copy current weights
            for (int i = 0; i < instances; i++)
                sum += (new_w[list[i]] = w[list[i]]);

            // Renormalize and get error
            for (int i = 0; i < instances; i++) {
                new_w[list[i]] /= sum;
                if (labels[list[i]] != majority)
                    e_i += new_w[list[i]];
            }

            // Update weights
            sum = 0.0;
            if (e_i >= 0.5 || e_i <= 0.0) {
                for (int k = 0; k < instances; k++)
                    sum += (new_w[list[k]] = Poisson());
                SHOWPROGRESS(parm->show_progress > 0, "AdaBoost reinit weights at node %d for %s\n", t, e_i >= 0.5 ? "e > 0.5" : "e = 0");
            } else {
                scalar sbeta = sqrt((1.0-e_i)/e_i);

                for (int k = 0; k < instances; k++)
                    sum += (new_w[list[k]] *= (labels[list[k]] == majority) ? 1.0/sbeta : sbeta);
            }

            // Normalize w[] to sum 1
            for (int k = 0; k < instances; k++)
                new_w[list[k]] /= sum;
            break;

        case RealAdaBoost:
        case DiscreteAdaBoost:
        case Wagging:
        case eBagging:
        case RandomSubsampling:
        case eGentleBoost:
        case eMadaBoost:
        case eReweightBoost:
        case eFloatBoost:
        case BalanceCascade:
        case EasyEnsemble:
        case AsymmetricAdaBoost:
        case AdaCost:
        case AdaC1:
        case AdaC2:
        case AdaC3:
        case CSB0:
        case CSB1:
        case CSB2:
        case Arcing:
        case BoostingProj:
        case RandomProj:
        case AdaBoostMH:
        case eFeatureBoost:
        case FreundNN:
        case LearningFromBites:
        case FilterBoost:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ensemble method for IBTs");
            break;

        case RotationForest:
            OnErrorExit(ERR_PARAM, "Ensemble method not suitable for IBTs");
            break;
    }

    return new_w;
}

void InternallyBoostedTreeClassify(t_DTree tree, scalar *instance, scalar *probabilities, int classes, bool predict_with_probabilities, int *index)
{
    if (predict_with_probabilities)
        for (int i = 0; i < classes; i++)
            probabilities[i] += tree->probabilities[i]*tree->alpha;
    else
        probabilities[tree->class] += tree->alpha;

    // Leaf node? Do nothing
    if (!tree->leaf) {
        if (index ? SparseFeature(instance, index, tree->feat) : instance[tree->feat] <= tree->cut_point)
            InternallyBoostedTreeClassify(tree->left, instance, probabilities, classes, predict_with_probabilities, index);
        else
            InternallyBoostedTreeClassify(tree->right, instance, probabilities, classes, predict_with_probabilities, index);
    }
}

int IBTClassify(t_DTree tree, scalar *instance, scalar *probabilities, int classes, bool do_not_use_internal_nodes, bool predict_with_probabilities, int *index)
{
    if (do_not_use_internal_nodes)
        return DecisionTreeClassify(tree, instance, probabilities, classes, NULL);

    scalar *votes = NULL;

    if (!probabilities) {
        MALLOC(votes, scalar, classes, "IBTClassify: votes");
        probabilities = votes;
    }

    ZeroVector(probabilities, NULL, classes);

    InternallyBoostedTreeClassify(tree, instance, probabilities, classes, predict_with_probabilities, index);

    int category = MaxVectorIndex(probabilities, NULL, classes);

#ifdef VERBOSE_MODE
    printf("%2d --> %2d  %c  ", label+1, category+1, label == category ? ' ' : '*');
    printvec(classes, probabilities);
#endif

    if (votes)
        FREE(votes);

    return category;
}


scalar GetAlpha(const t_Parameters *parm, scalar *w, int instances, int *labels, int majority, int *list)
{
    scalar e_i = 0.0, alpha = 0.0;

    if (parm->ibt_fixed_node_weight)
        return 1.0;

    // Consider different ensemble methods
    switch (parm->ibt_alpha_method) {
        case AdaBoost:
        case eMultiBoost:
            // Get weighted error
            for (int i = 0; i < instances; i++)
                if (labels[list[i]] != majority)
                    e_i += w[list[i]];

            alpha = AssignWithinBoundsScalar(0.0, 0.05*log((1.0 - e_i)/e_i), 10.0);
            break;

        case NoBoost:
            // Get accuracy
            for (int i = 0; i < instances; i++)
                if (labels[list[i]] == majority)
                    alpha += 1.0;

            alpha /= instances;
            break;

        case RealAdaBoost:
        case DiscreteAdaBoost:
        case Wagging:
        case eBagging:
        case RandomSubsampling:
        case eGentleBoost:
        case eMadaBoost:
        case eReweightBoost:
        case eFloatBoost:
        case BalanceCascade:
        case EasyEnsemble:
        case AsymmetricAdaBoost:
        case AdaCost:
        case AdaC1:
        case AdaC2:
        case AdaC3:
        case CSB0:
        case CSB1:
        case CSB2:
        case Arcing:
        case BoostingProj:
        case RandomProj:
        case AdaBoostMH:
        case eFeatureBoost:
        case FreundNN:
        case LearningFromBites:
        case FilterBoost:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ensemble method for IBTs");
            break;

        case RotationForest:
            OnErrorExit(ERR_PARAM, "Ensemble method not suitable for IBTs");
            break;
    }

    return alpha;
}
