/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdbool.h>
#include "macros.h"
#include "classifiers.h"
#include "maths.h"

// Functions that implement Nearest Neighbor classification with cache

// k-NN with a cached matrix of distances
int kNearestNeighbourCached(int instance, int *S, scalar **dcache, int k, int instances, int *label, int classes, bool knn_with_weighted_votes)
{
    int j, prediction, nearest[k];
    scalar value, dmin, dist[k], votes[classes];

    for (int i = 0; i < classes; i++)
        votes[i] = 0.0;

    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            value = dcache[instance][i];

            // Test if is one of the nearest
            if (value < dmin /* && value > threshold */) {
                int count = S[i];
                while (count--) {
                    // Look for its place
                    j = 0;
                    while (value > dist[j])
                        j++;

                    // Sets in j-th place
                    for (int l = k - 1; l > j; l--) {
                        nearest[l] = nearest[l - 1];
                        dist[l] = dist[l - 1];
                    }
                    nearest[j] = i;
                    dist[j] = value;
                    dmin = dist[k - 1];
                }
            }
        }

    // Vote neighbors
    if (!knn_with_weighted_votes) {
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                votes[label[nearest[i]]] += 1.0;
    } else
        WeightVotes(votes, k, dist, nearest, label, classes);

    // Get maximum votes
    prediction = 0;
    for (int i = 1; i < classes; i++)
        if (votes[prediction] < votes[i])
            prediction = i;

    return prediction;
}

int kNearestNeighbourWithListCached(int instance, int *list, int n_protos, int k, int *nearest, int *label, scalar **distances, int classes, bool knn_with_weighted_votes)
{
    int prediction;
    scalar value, dmin, dist[k], votes[classes];

    for (int i = 0; i < classes; i++)
        votes[i] = 0;

    // Select k first as k initial neighbors
    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    for (int i = 0; i < n_protos; i++) {
        value = distances[instance][list[i]];

        // Test if is one of the nearest
        if (value < dmin) {
            // Look for its place
            int j = 0;
            while (value > dist[j])
                j++;

            // Sets in j-th place
            for (int l = k - 1; l > j; l--) {
                nearest[l] = nearest[l - 1];
                dist[l] = dist[l - 1];
            }
            nearest[j] = list[i];
            dist[j] = value;
            dmin = dist[k - 1];
        }
    }

    // Vote neighbors
    if (!knn_with_weighted_votes) {
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                votes[label[nearest[i]]] += 1.0;
    } else
        WeightVotes(votes, k, dist, nearest, label, classes);

    // Get maximum votes
    prediction = 0;
    for (int i = 1; i < classes; i++)
        if (votes[prediction] < votes[i])
            prediction = i;

    return prediction;
}


// Classify a pattern using k-NN rule with cache of neighbors
int kNearestNeighbourCache(int instance, int *S, int n_protos, int *nearest, int *label, int k, cache_t **neighbors_cache, scalar **data, int inputs, int classes, bool knn_with_weighted_votes,
                           scalar *r, scalar *distance_to_furthest_neighbor, int **index, int *subspace)
{
    int c = 0;
    scalar distances[k], votes[classes];

    if (neighbors_cache == NULL)
        return kNNCategory(data[instance], data, subspace, n_protos, k, inputs, label, classes, nearest, S, NULL, neighbors_cache, instance, index, index ? index[instance] : NULL);

    for (int i = 0; i < classes; i++)
        votes[i] = 0;
    for (int i = 0; i < k; i++)
        nearest[i] = -1;

    // Find first k neighbors in the sample
    for (int i = 0; i < n_protos && c < k; i++)
        if (!S || (neighbors_cache[instance][i] != -1 && S[neighbors_cache[instance][i]])) {
            int count = S ? S[neighbors_cache[instance][i]] : 1;
            while (count-- && c < k) {
                nearest[c] = neighbors_cache[instance][i];
                // Vote neighbor's class
                votes[label[nearest[c]]] += 1.0;

                c++;
            }
        }

    if (knn_with_weighted_votes) {
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                distances[i] = EuclideanDistanceSparseDataSubspace(data[instance], data[nearest[i]], inputs, index ? index[instance] : NULL, index ? index[nearest[i]]: NULL, subspace)/r[nearest[i]];

        WeightVotes(votes, k, distances, nearest, label, classes);
    }

    // Get maximum votes
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (votes[category] < votes[i])
            category = i;

    return category;
}

// Weighting of votes using [Dudani, 76] method
void WeightVotes(scalar *votes, int k, scalar *distances, int *nearest, int *labels, int classes)
{
    scalar margin;

    margin = distances[k-1] - distances[0];

    for (int i = 0; i < classes; i++)
        votes[i] = 0.0;

    // Vote
    if (margin <= ZERO) {
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                votes[labels[nearest[i]]] += 1.0;
    } else {
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                votes[labels[nearest[i]]] += (distances[k-1]-distances[i])/margin;
    }
}
