/***************************************************************************
 *   Copyright (C) 20089 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <string.h>
#include "macros.h"
#include "classifiers.h"
#include "data.h"
#include "maths.h"

scalar kNearestNeighbourMargin(scalar *instance, int instance_class, int *subspace, scalar **protos, int n_protos, int dim, int k, int *label, int classes, int *subset)
{
    int nearest[k];
    scalar value, dmin, dist[k], votes[classes];

    // If 1-NN obtain
    if (k == 1)
        return NearestNeighbourMargin(instance, instance_class, subspace, protos, n_protos, dim, label, subset);

    ZeroVector(votes, NULL, classes);
    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    for (int i = 0; i < n_protos && dmin > 0.0; i++)
        if (!subset || subset[i])  {
            value = EuclideanDistanceSubspace(instance, protos[i], dim, subspace);

            // Test if is one of the nearest
            if (value < dmin) {
                // Look for its place
                int j = 0;
                while (j < k && value > dist[j])
                    j++;

                if (j == k)
                    break;

                // Sets in j-th place
                for (int l = k - 1; l > j; l--) {
                    nearest[l] = nearest[l - 1];
                    dist[l] = dist[l - 1];
                }
                nearest[j] = i;
                dist[j] = value;
                dmin = dist[k - 1];
            }
        }

    // Vote neighbors
    for (int i = 0; i < k; i++)
        if (nearest[i] != -1)
            votes[label[nearest[i]]]++;

    // Votes of the correct class
    int ok_votes = votes[instance_class];

    // Get maximum of votes of wrong class
    int wrong_votes = 0;
    for (int i = 0; i < classes; i++)
        if (wrong_votes < votes[i] && i != instance_class)
            wrong_votes = votes[i];

    scalar margin = ((scalar)(ok_votes-wrong_votes))/k;

    return margin;
}

scalar NearestNeighbourMargin(scalar *instance, int instance_class, int *subspace, scalar **protos, int n_protos, int inputs, int *label, int *subset)
{
    scalar ok_nearest = INFINITE, wrong_nearest = INFINITE, value, margin;
    bool at_least_one_ok, at_least_one_wrong;

    at_least_one_ok = at_least_one_wrong = false;

    // Iterate for every prototype
    for (int i = 0; i < n_protos; i++)
        if (!subset || subset[i]) {
            if (label[i] == instance_class) {
                if ((value = EuclideanDistanceSubspace(instance, protos[i], inputs, subspace)) < ok_nearest) {
                    ok_nearest = value;
                    at_least_one_ok = true;
                }
            } else {
                if ((value = EuclideanDistanceSubspace(instance, protos[i], inputs, subspace)) < wrong_nearest) {
                    wrong_nearest = value;
                    at_least_one_wrong = true;
                }
            }
        }

    if (at_least_one_ok && at_least_one_wrong)
        margin = wrong_nearest - ok_nearest;
    else if (at_least_one_ok)
        margin = ok_nearest > ZERO ? 1.0/ok_nearest : 1.0;
    else if (at_least_one_wrong)
        margin = wrong_nearest > ZERO ? -1.0/wrong_nearest : -1.0;
    else
        margin = 0.0;

    return margin;
}

void CrossValidatedParameterskNN(t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, int inputs, int classes, int minority, int *subset, int **index)
{
    scalar best_so_far, accuracy = 0.0, **data_v, **data_t, * **ext_str_data = NULL;
    int *label_v = NULL, *label_t, **M, instances_v, instances_t, best_k = -1, majority, n_external_strata, *ext_str_instances = NULL, **ext_str_labels = NULL, **ext_str_list = NULL,
         **index_t = NULL, **index_v = NULL;
    t_PerformanceInfo perf;

    // Sampling?
    int *s_labels = NULL, *n_labels = NULL, n_instances = 0, s_instances = 0;
    scalar **s_data = NULL, **n_data = NULL;
    if (parm->cv_sampling_size && parm->cv_sampling_size < instances) {
        switch (parm->cv_sampling_method) {
            case sNone:
                break;

            case sKMeans:
                if (subset) {
                    // Only selected instances
                    MALLOC(n_data, scalar *, instances, "CrossValidatedParameterskNN: n_data");
                    MALLOC(n_labels, int, instances, "CrossValidatedParameterskNN: n_labels");
                    for (int i = 0; i < instances; i++)
                        if (subset[i]) {
                            n_data[n_instances] = data[i];
                            n_labels[n_instances] = labels[i];
                            n_instances++;
                        }
                } else {
                    n_data = data;
                    n_instances = instances;
                    n_labels = labels;
                }

                if (parm->stratified)
                    kMeansStratifiedSample((const t_Parameters *) &parm, &n_data, n_instances, &n_labels, &s_data, &s_labels, &s_instances, classes, inputs, parm->cv_sampling_size, NULL, false);
                else
                    kMeansSample((const t_Parameters *) &parm, &data, instances, &labels, &s_data, &s_labels, &s_instances, classes, inputs, parm->cv_sampling_size, NULL, false);

                // Overwrite
                data = s_data;
                instances = s_instances;
                labels = s_labels;

                if (subset) {
                    FREE(n_data);
                    FREE(n_labels);
                    subset = NULL;
                }
                break;

            case sStratified:
                n_external_strata = instances%parm->cv_sampling_size ? instances/parm->cv_sampling_size+1 : instances/parm->cv_sampling_size;

                // Memory
                MALLOC(ext_str_data, scalar **, n_external_strata, "CrossValidatedParameterskNN: ext_str_data");
                MALLOC(ext_str_instances, int, n_external_strata, "CrossValidatedParameterskNN: ext_str_list");
                MALLOC(ext_str_labels, int *, n_external_strata, "CrossValidatedParameterskNN: ext_str_labels");
                MALLOC(ext_str_list, int *, n_external_strata, "CrossValidatedParameterskNN: ext_str_list");
                for (int i = 0; i < n_external_strata; i++) {
                    MALLOC(ext_str_data[i], scalar *, parm->cv_sampling_size, "CrossValidatedParameterskNN: ext_str_data[i]");
                    MALLOC(ext_str_labels[i], int, parm->cv_sampling_size, "CrossValidatedParameterskNN: ext_str_labels[i]");
                    MALLOC(ext_str_list[i], int, parm->cv_sampling_size, "CrossValidatedParameterskNN: ext_str_labels[i]");
                }

                if (subset) {
                    // Only selected instances
                    MALLOC(n_data, scalar *, instances, "CrossValidatedParameterskNN: n_data");
                    MALLOC(n_labels, int, instances, "CrossValidatedParameterskNN: n_labels");
                    for (int i = 0; i < instances; i++)
                        if (subset[i]) {
                            n_data[n_instances] = data[i];
                            n_labels[n_instances] = labels[i];
                            n_instances++;
                        }
                } else {
                    n_data = data;
                    n_instances = instances;
                    n_labels = labels;
                }

                // Get first stratum as training set
                DataPartition(parm, n_data, n_instances, n_labels, NULL, &n_external_strata, &ext_str_data, &ext_str_instances, &ext_str_labels, classes, &ext_str_list, NULL, NULL, inputs,
                              parm->cv_sampling_size, pRANDOM, parm->stratified, NULL);

                // Forget original data
                data = ext_str_data[0];
                instances = ext_str_instances[0];
                labels = ext_str_labels[0];

                if (subset) {
                    FREE(n_data);
                    FREE(n_labels);
                    subset = NULL;
                }
                break;
        }
    }

    // If first call allocate memory
    MALLOC(label_v, int, instances, "SetCrossValidatedParametersSubspace: label_v");
    MALLOC(label_t, int, instances, "SetCrossValidatedParametersSubspace: label_t");
    MALLOC(data_v, scalar *, instances, "SetCrossValidatedParametersSubspace: data_v");
    MALLOC(data_t, scalar *, instances, "SetCrossValidatedParametersSubspace: data_t");
    if (index) {
        MALLOC(index_v, int *, instances, "SetCrossValidatedParametersSubspace: index_v");
        MALLOC(index_t, int *, instances, "SetCrossValidatedParametersSubspace: index_t");
    }

    // Size of validation set
    instances_v = instances/parm->cv_folds;
    instances_t = instances - instances_v;

    StratifyDataset(data, instances, labels, index, classes);

    AllocateIntegerMatrix(&M, classes, classes);

    // Do validation runs
    best_so_far = -1.0;

    // K in [min-k, max-k]
    for (int i = parm->min_k; i <= parm->max_k; i++) {
        int c;
        scalar margin = 0.0, auroc = 0.0;

        parm->k = i;

        ZeroIntegerMatrix(M, classes, classes);

        // For each fold
        for (int fold = 0; fold < parm->cv_folds; fold ++) {
            int nearest[parm->k];

            // Set fold
            // Validation set
            for (int k = (fold * instances_v) % instances, c = 0; c < instances_v; k++, c++) {
                data_v[c] = data[k];
                label_v[c] = labels[k];
                if (index)
                    index_v[c] = index[k];
            }

            // Train set
            int next = 0;
            for (int k = 0, c = ((fold+1) * instances_v) % instances; k < instances_t; k++, c = (c+1) % instances)
                if (!subset || subset[c]) {
                    data_t[next] = data[c];
                    label_t[next] = labels[c];
                    if (index)
                        index_t[next] = index[c];
                    next++;
                }

            // Undersample?
            scalar **u_data = NULL;
            int u_instances = 0, *u_labels = NULL, **u_index = NULL;
            if (parm->random_undersampling_for_training && !parm->weigthed_sampling_for_training && classes == 2) {
                if (index)
                    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Undersampling for training with sparse data");
                TemporalUndersampleDataForTraining(data_t, next, classes, label_t, parm->random_undersampling_ratio, &u_data, &u_labels, &u_instances, minority);
            } else {
                u_data = data_t;
                u_labels = label_t;
                u_index = index_t;
                u_instances = next;
            }

            // If margin as performance it must be obtained here
            if (parm->cv_performance_measure == pfMargin) {
                scalar minimum_margin;
                t_Learner learner;

                MALLOC(learner.knn, t_kNN, 1, "CrossValidatedParameterskNN: learner.knn");
                learner.knn->k = i;
                learner.knn->localk = NULL;
                learner.knn->distance_to_furthest_neighbor = learner.knn->r = NULL;
                margin += ClassifierDatasetMargin(parm, kNN, data_t, next, label_t, data_v, instances_v, label_v, inputs, classes, subspace, &minimum_margin, &learner, NULL, subset, NULL, true);
                FREE(learner.knn);
            } else if (parm->cv_performance_measure == pfauROC) {
                t_Learner learner;

                MALLOC(learner.knn, t_kNN, 1, "CrossValidatedParameterskNN: learner.knn");
                learner.knn->k = i;
                learner.knn->localk = NULL;
                learner.knn->distance_to_furthest_neighbor = learner.knn->r = NULL;;
                auroc += ClassifierAUROC(parm, kNN, data_v, instances_v, label_v, data_t, next, subset, subspace, label_t, inputs, minority, false, &learner, false, parm->out_file, classes, NULL, NULL,
                                         NULL, NULL, NULL, true);
                FREE(learner.knn);
            }

            // Get M matrix and metric for performance
            for (int i = 0; i < instances_v; i++)
                M[label_v[i]][c = kNNCategory(data_v[i], u_data, subspace, u_instances, parm->k, inputs, u_labels, classes, nearest, NULL, NULL, NULL, -1, u_index, index_v ? index_v[i] : NULL)]++;

            if (parm->random_undersampling_for_training && !parm->weigthed_sampling_for_training && classes == 2) {
                FREE(u_data);
                FREE(u_labels);
            }
        }

        PrintMulticlassMeasures(M, classes, instances, NULL, NULL, &perf);

        switch (parm->cv_performance_measure) {
            case pfAccuracy:
                accuracy = perf.accuracy;
                break;

            case pfGMean:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.gmean;
                break;

            case pfMargin:
                accuracy = margin/parm->cv_folds;
                break;

            case pfauROC:
                accuracy = auroc/parm->cv_folds;
                break;

            case pfauPRC:
            case pfPrecisionMacro:
            case pfF1Macro:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Metric for cross-validation of parameters");
                break;

            case pfCC:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.mcc;
                break;

            case pfRecallMacro:
                accuracy = perf.recall_macro;
                break;

            case pfRecallMacroGMean:
                accuracy = perf.recall_macro_g_mean;
                break;

            case pfKappa:
                accuracy = perf.kappa;
                break;

            case pfAverageAccuracy:
                accuracy = perf.average_accuracy;
                break;

            case pfF1:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.f1_score;
                break;
        }

        // Better than current best?
        if (accuracy > best_so_far) {
            best_so_far = accuracy;
            best_k = i;
        }

        SHOWPROGRESS(parm->show_progress > 1, "Accuracy: %lf\tk = %d\n", accuracy, i);
    }

    parm->k = best_k;

    SAVEPROGRESS(parm->show_progress > 0, "Best k: %d (from cross-validation)\n", parm->k);

    FreeIntegerMatrix(&M, classes);
    FREE(label_v);
    FREE(label_t);
    FREE(data_v);
    FREE(data_t);
    if (index) {
        FREE(index_t);
        FREE(index_v);
    }

    if (ext_str_data) {
        for (int i = 0; i < n_external_strata; i++) {
            FREE(ext_str_data[i]);
            FREE(ext_str_labels[i]);
            FREE(ext_str_list[i]);
        }
        FREE(ext_str_list);
        FREE(ext_str_data);
        FREE(ext_str_labels);
        FREE(ext_str_instances);
    } else if (s_data) {
        FREE(s_labels);
        for (int i = 0; i < s_instances; i++)
            FREE(s_data[i]);
        FREE(s_data);
    }
}

int ObtainkCrossValidation(t_Parameters *parm, int *S, scalar **data, int instances, int *labels, int inputs, int classes, int **index, int *subspace)
{
    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for cross-validated k");

    int old_k = parm->k;

    CrossValidatedParameterskNN(parm, data, instances, labels, subspace, inputs, classes, -1, S, NULL);

    int new_k = parm->k;
    parm->k = old_k;

    return new_k;
}

// Classify a pattern using k nearest neighbors rule, return k instances from each category

void kNNOfOneClass(scalar *instance, scalar **data, int category, int n, int k, int inputs, int *labels, int *nearest)
{
    scalar value, dmin;
    scalar *dist;

    // Allocate memory
    MALLOC(dist, scalar, n, "kNNOfOneClass: dist");

    // Select k first as k initial neighbors
    dmin = INFINITE;
    for (int i = 0; i < k && i < n; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    // Test if sampling
    for (int i = 0; i < n && dmin > 0.0; i++)
        if (labels[i] == category) {
            value = EuclideanDistance(instance, data[i], inputs);

            // Test if is one of the nearest
            if (value < dmin) {
                // Look for its place
                int j = 0;
                while (value > dist[j])
                    j++;

                // Sets in j-th place
                for (int l = k - 1; l > j; l--) {
                    nearest[l] = nearest[l - 1];
                    dist[l] = dist[l - 1];
                }
                nearest[j] = i;
                dist[j] = value;
                dmin = dist[k - 1];
            }
        }

    FREE(dist);
}

// Classify a pattern using k nearest neighbors rule
int kNNCategory(scalar *instance, scalar **data, int *subspace, int instances, int k, int inputs, int *labels, int classes, int *nearest, int *subset, int *rvotes, cache_t **neighbors_cache,
                int inst_order, int **index, int *instance_index)
{
    int aclass, votes[classes];

    for (int i = 0; i < classes; i++)
        votes[i] = 0;

    for (int i = 0; i < k; i++)
        nearest[i] = -1;

    // Use cache?
    if (neighbors_cache) {
        int c = 0;
        for (int i = 0; i < instances && c < k; i++)
            if (!subset || subset[neighbors_cache[inst_order][i]]) {
                votes[labels[neighbors_cache[inst_order][i]]]++;
                nearest[c] = neighbors_cache[inst_order][i];
                c++;
            }
    } else {
        scalar dmin = INFINITE, dist[k], value;
        for (int i = 0; i < k; i++)
            dist[i] = INFINITE;

        // Get k nearest neighbors
        for (int i = 0; i < instances && dmin > 0.0; i++)
            if (!subset || subset[i]) {
                value = EuclideanDistanceSparseDataSubspace(instance, data[i], inputs, instance_index, index ? index[i] : NULL, subspace);

                // Test if is one of the nearest
                if (value < dmin) {
                    // Look for its place
                    int j = 0;
                    while (value > dist[j])
                        j++;

                    // Sets in j-th place
                    for (int l = k - 1; l > j; l--) {
                        nearest[l] = nearest[l - 1];
                        dist[l] = dist[l - 1];
                    }
                    nearest[j] = i;
                    dist[j] = value;
                    dmin = dist[k - 1];
                }
            }

        // Vote neighbors
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                votes[labels[nearest[i]]]++;
    }

    // Get maximum votes
    aclass = 0;
    for (int i = 1; i < classes; i++)
        if (votes[aclass] < votes[i])
            aclass = i;

    if (rvotes)
        memcpy(rvotes, votes, classes*sizeof(int));

    return aclass;
}
