/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "typex.h"
#include "macros.h"
#include "classifiers.h"
#include "data.h"
#include "ensemble.h"
#include "maths.h"

// Implements Logistic Regression Classifier
// If more than two classes uses one-vs-one method

void TrainLogisticRegressionClassifier(const t_Parameters *parm, t_LR *lr, int instances, int *labels, int inputs, scalar **data, int classes, int *subset, int *subspace, int **index)
{
    // Number of OVO classifier
    lr->n = (classes*(classes-1))/2;

    AllocateScalarMatrix(&lr->w, lr->n, inputs);

    // If only two classes train with original dataset
    if (classes == 2)
        TrainLogisticRegressionClassifierFor2Classes(parm, lr->w[0], instances, labels, inputs, data, subset, subspace, index);
    else {
        // One-vs-one method
        int *n_labels, **n_index = NULL, n_instances, *n_subset = NULL;
        scalar **n_data;
        MALLOC(n_labels, int, instances, "TrainLogisticRegressionClassifier: n_labels");
        if (subset)
            MALLOC(n_subset, int, instances, "TrainLogisticRegressionClassifier: n_subset");
        if (index)
            MALLOC(n_index, int *, instances, "TrainLogisticRegressionClassifier: n_index");
        MALLOC(n_data, scalar *, instances, "TrainLogisticRegressionClassifier: n_data");

        int n = 0;
        for (int i = 1; i < classes; i++)
            for (int j = 0; j < i; j++, n++) {
                n_instances = 0;

                // i: +1, j: 0
                for (int k = 0; k < instances; k++)
                    if (labels[k] == i || labels[k] == j) {
                        n_labels[n_instances] = (labels[k] == i);
                        n_data[n_instances] = data[k];
                        if (index)
                            n_index[n_instances] = index[k];
                        if (subset)
                            n_subset[n_instances] = subset[k];
                        n_instances++;
                    }

                TrainLogisticRegressionClassifierFor2Classes(parm, lr->w[n], n_instances, n_labels, inputs, n_data, n_subset, subspace, n_index);
            }

        FREE(n_labels);
        if (subset)
            FREE(n_subset);
        if (index)
            FREE(n_index);
        FREE(n_data);
    }

}

void TrainLogisticRegressionClassifierFor2Classes(const t_Parameters *parm, scalar *w, int instances, int *labels, int inputs, scalar **data, int *subset, int *subspace, int **index)
{
    scalar *delta;

    MALLOC(delta, scalar, inputs, "TrainLogisticRegressionClassifier: delta");

    // Train for number of epochs
    ZeroVector(w, NULL, inputs);
    for (int e = 0; e < parm->lr_epochs; e++) {
        ZeroVector(delta, NULL, inputs);

        int c = 0;

        for (int i = 0; i < instances; i++)
            if (!subset || subset[i]) {
                scalar error = hx(data[i], w, inputs, subspace, index ? index[i] : NULL) - labels[i];
                if (!index)
                    for (int j = 0; j < inputs; j++)
                        delta[j] += error*data[i][j];
                else {
                    int j = 0, x = 0;
                    for (; index[i][j] != -1; x++)
                        if (index[i][j]-1 == x){
                            delta[x] += error*data[i][j];
                            j++;
                        }
                }
                c++;
            }

        for (int j = 0; j < inputs; j++)
            w[j] -= parm->lr_alpha*(1.0/c)*delta[j];
    }

    FREE(delta);
}

scalar hx(scalar *x, scalar *theta, int n, int *subspace, int *index)
{
    scalar ret = 0.0;

    if (index) {
        if (subspace)
            for (int i = 0; index[i] != -1; i++)
                ret += theta[index[i]-1]*x[i]*subspace[index[i]-1];
        else
            for (int i = 0; index[i] != -1; i++)
                ret += theta[index[i]-1]*x[i];
    } else {
        if (subspace)
            for (int i = 0; i < n; i++)
                ret += theta[i]*x[i]*subspace[i];
        else
            for (int i = 0; i < n; i++)
                ret += theta[i]*x[i];
    }

    return 1.0/(1.0 + exp(-ret));
}

void FreeLR(t_LR *lr)
{
    FreeScalarMatrix(&lr->w, lr->n);
}

int LROutput(t_LR *lr, scalar *instance, int inputs, int classes, scalar *probabilities, int *subspace, int *index)
{
    int category;
    // Two classes?
    if (classes == 2)
        category = LR2ClassOutput(lr->w[0], instance, inputs, probabilities, subspace, index);
    else {
        scalar *votes, *probs;
        CALLOC(votes, scalar, classes, "LROutput: votes");
        MALLOC(probs, scalar, classes, "LROutput: votes");
        if (probabilities)
            ZeroVector(probabilities, NULL, classes);

        int n = 0;
        for (int i = 1; i < classes; i++)
            for (int j = 0; j < i; j++, n++) {
                // i: +1, j: 0
                category = LR2ClassOutput(lr->w[n], instance, inputs, probs, subspace, index);

                if (probabilities)  {
                    probabilities[j] += probs[0];
                    probabilities[i] += probs[1];
                }

                if (category == 1)
                    votes[i]++;
                else
                    votes[j]++;
           }

        FREE(probs);

        if (probabilities)
            NormalizeVector(probabilities, classes);

        category = IndexOfMinimumOfVector(votes, NULL, classes);

        FREE(votes);
    }

    return category;
}

int LR2ClassOutput(scalar *w, scalar *instance, int inputs, scalar *probabilities, int *subspace, int *index)
{
    // Two classes?
    scalar r = hx(instance, w, inputs, subspace, index);

    if (probabilities) {
        probabilities[0] = 1.0 - r;
        probabilities[1] = r;
    }

    if (r >= 0.5)
        return 1;
    else
        return 0;
}

scalar LRMargin(t_LR *cl, scalar *instance, int inputs, int classes, int *subspace, int label, int *index)
{
    scalar probabilities[classes];

    LROutput(cl, instance, inputs, classes, probabilities, subspace, index);

    scalar wrong_probs = -1.0;
    for (int j = 0; j < classes; j++)
        if (wrong_probs < probabilities[j] && j != label)
            wrong_probs = probabilities[j];

    return probabilities[label] - wrong_probs;
}
