/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "typex.h"
#include "macros.h"
#include "classifiers.h"
#include "data.h"

// Implements Multimodal Naive Bayes Classifier

void TrainNBCClassifier(const t_Parameters *parm, t_NaiveBayes *cl, int instances, int *labels, int inputs, scalar **data, int classes, int *subset, int **index)
{
    scalar *to_sort, max, min, interval;

    // For sorting values
    MALLOC(to_sort, scalar, instances, "TrainNBCClassifier");

    // Memory for info
    MALLOC(cl->freq, scalar **, classes, "TrainNBCClassifier: cl->freq");
    MALLOC(cl->limit, scalar *, inputs, "TrainNBCClassifier: cl->limit");
    MALLOC(cl->class_freq, scalar, classes, "TrainNBCClassifier: cl->class_freq");
    MALLOC(cl->n_intervals, int, inputs, "TrainNBCClassifier: cl->n_intervals");

    // Count class frequencies
    FillVector(cl->class_freq, NULL, classes, 1.0);
    int count = 0;
    for (int k = 0; k < instances; k++)
        if (!subset || subset[k]) {
            cl->class_freq[labels[k]]++;
            count++;
        }

    for (int i = 0; i < classes; i++)
        cl->class_freq[i] /= count+classes;

    if (parm->NBC_discretizer == PKID)
        for (int j = 0; j < inputs; j++)
            // PKID adjust n intervals = size = sqrt (n) and then does EFD
            cl->n_intervals[j] = MAX(2, (int) sqrt((scalar) instances));
    // WPKID
    else if (parm->NBC_discretizer == WPKID)
        for (int j = 0; j < inputs; j++)
            cl->n_intervals[j] = MAX(2, -15 + sqrt(225.0 + instances));
    else
        for (int j = 0; j < inputs; j++)
            cl->n_intervals[j] = parm->NBC_intervals;

    // More memory
    for (int j = 0; j < inputs; j++)
        MALLOC(cl->limit[j], scalar, cl->n_intervals[j], "TrainNBCClassifier: cl->limit[j]");
    for (int i = 0; i < classes; i++) {
        MALLOC(cl->freq[i], scalar *, inputs, "TrainNBCClassifier: cl->freq[i]");
        for (int j = 0; j < inputs; j++)
            MALLOC(cl->freq[i][j], scalar, cl->n_intervals[j], "TrainNBCClassifier: cl->freq[i][j]");
    }

    // Choose method for obtaining intervals
    if (parm->NBC_discretizer == FANDI)  {
        FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, cl->limit, cl->n_intervals, subset, index);

        // Reallocate memory for frequencies
        for (int i = 0; i < classes; i++)
            for (int j = 0; j < inputs; j++)
                REALLOC(cl->freq[i][j], scalar, cl->n_intervals[j], "TrainNBCClassifier: cl->freq[i][j]");
    } else {
        // Create frequencies and bins info for each class, negative and positive
        // Each variable in turn
        for (int j = 0; j < inputs; j++) {
            switch (parm->NBC_discretizer) {
                case FANDI:
                    OnErrorExit(1, "Something wrong happened in NBC");
                    break;

                case EFD:
                case PKID:
                case WPKID:
                    // Get equally distributed intervals
                    // Sort values
                    count = 0;
                    for (int i = 0; i < instances; i++)
                        if (!subset || subset[i]) {
                            to_sort[count] = index ? SparseFeature(data[i], index[i], j) : data[i][j];
                            count++;
                        }
                    // Sort values
                    qsort(to_sort, count, sizeof(scalar), CompareDouble);
                    interval = count/cl->n_intervals[j];

                    // Assign limits
                    for (int i = 0; i < cl->n_intervals[j]; i++)
                        cl->limit[j][i] = to_sort[(int)((i + 1) * interval)];
                    break;

                case EWD:
                    // Get equal size interval
                    // Get maximum and minimum for the variable
                    max = -INFINITE, min = INFINITE;
                    for (int i = 0; i < instances; i++)
                        if (!subset || subset[i]) {
                            if (max < (index ? SparseFeature(data[i], index[i], j) : data[i][j]))
                                max = index ? SparseFeature(data[i], index[i], j) : data[i][j];
                            if (min > (index ? SparseFeature(data[i], index[i], j) : data[i][j]))
                                min = index ? SparseFeature(data[i], index[i], j) : data[i][j];
                        }

                    interval = (max - min)/cl->n_intervals[j];

                    // Assign limits
                    cl->limit[j][0] = min + interval;
                    for (int i = 1; i < cl->n_intervals[j]; i++)
                        cl->limit[j][i] = cl->limit[j][i - 1] + interval;
            }
        }
    }

    // Assign frequencies
    for (int k = 0; k < classes; k++)
        for (int j = 0; j < inputs; j++) {
            // Select values for the class and sort them
            count = 0;
            for (int i = 0; i < instances; i++)
                if ((!subset || subset[i]) && labels[i] == k)
                    to_sort[count++] = index ? SparseFeature(data[i], index[i], j) : data[i][j];

            // Sort values
            qsort(to_sort, count, sizeof(scalar), CompareDouble);

            // Assign frequencies using pseudo-counts
            FillVector(cl->freq[k][j], NULL, cl->n_intervals[j], 1.0);

            int index = 0;
            for (int i = 0; i < count; i++) {
                while (index < cl->n_intervals[j] - 1 && (to_sort[i] > cl->limit[j][index]))
                    index++;
                cl->freq[k][j][index]++;
            }

            for (int i = 0; i < cl->n_intervals[j]; i++)
                cl->freq[k][j][i] /= count + cl->n_intervals[j];
        }

    FREE(to_sort);
}

int NBCOutput(t_NaiveBayes *cl, scalar *instance, int inputs, int classes, scalar *probabilities, int *subspace, int *index)
{
    scalar probs[classes];

    // Obtain probability of each class
    // Use logarithms to avoid numerical problems;
    for (int i = 0; i < classes; i++)
        probs[i] = log(cl->class_freq[i]);

    // Apply Bayes' theorem for both classes
    // Use logarithms to avoid numerical problems
    for (int i = 0; i < inputs; i++) {
        scalar x = index ? SparseFeature(instance, index, i) : instance[i];

        if (!subspace || subspace[i]) {
            int index = 0;
            while (index < cl->n_intervals[i] - 1 && (x > cl->limit[i][index]))
                index++;

            for (int j = 0; j < classes; j++)
                probs[j] += log(cl->freq[j][i][index]);
        }
    }

    for (int i = 0; i < classes; i++)
        probs[i] = exp(probs[i]);

    int category = 0;
    for (int i = 1; i < classes; i++)
        if (probs[i] > probs[category])
            category = i;

    if (probabilities)
        memcpy(probabilities, probs, sizeof(scalar)*classes);

    return category;
}

scalar NBCMargin(t_NaiveBayes *cl, scalar *instance, int inputs, int classes, int *subspace, int label, int *index)
{
    scalar probabilities[classes];

    NBCOutput(cl, instance, inputs, classes, probabilities, subspace, index);

    scalar wrong_probs = -1.0;
    for (int j = 0; j < classes; j++)
        if (wrong_probs < probabilities[j] && j != label)
            wrong_probs = probabilities[j];

    return probabilities[label] - wrong_probs;
}

void FreeNBC(t_NaiveBayes *cl, int inputs, int classes)
{
    for (int j = 0; j < inputs; j++)
        FREE(cl->limit[j]);
    for (int i = 0; i < classes; i++) {
        for (int j = 0; j < inputs; j++)
            FREE(cl->freq[i][j]);
        FREE(cl->freq[i]);
    }

    FREE(cl->freq);
    FREE(cl->limit);
    FREE(cl->class_freq);
    FREE(cl->n_intervals);
}
