/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"

// Functions that deal with one-vs-one (OVO) method

int CreateOVOMatrix(int ***M, int classes)
{
    // Allocate matrix
    AllocateIntegerMatrix(M, classes, (classes*(classes-1))/2);
    ZeroIntegerMatrix(*M, classes, (classes*(classes-1))/2);

    // Create One-vs-one matrix
    int c = 0;
    for (int i = 0; i < classes-1; i++)
        for (int j = i+1; j < classes; j++) {
            (*M)[i][c] = 1;
            (*M)[j][c] = -1;
            c++;
        }

    return c;
}

int OneVsOneOutput(const t_Parameters *parm, t_OutputCoder *coder, scalar *instance, int *subspace, int inputs, int classes, scalar **data, int instances, int *labels, int *subset)
{
    int votes[classes];
    scalar outputs[classes];

    ZeroVector(NULL, votes, classes);

    // Every classifier cast a vote
    for (int i = 0; i < coder->n_binary; i++) {
        Classifier2ClassOutput(parm, coder->binary_learner, instance, inputs, 2, NULL, subspace, data,  instances, labels, subset, coder->binary[i], 0, outputs, NULL, NULL);

        // If positive search +1 in the column, if negative search for -1
        for (int j = 0; j < classes; j++)
            if (outputs[0] >= ZERO && coder->M[j][i] == 1) {
                votes[j]++;
                break;
            } else if (outputs[0] < ZERO && coder->M[j][i] == -1)  {
                votes[j]++;
                break;
            }
    }

    return MaxVectorIndex(NULL, votes, classes);
}

void OutputCodingTraining(const t_Parameters *parm, scalar **data, int instances, int *labels, int inputs, int classes, int **data_index, int *subspace, int *subset, bool releasedata,
                          t_Learner *learner)
{
    int *local_labels, *local_subset;

    if (data_index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for output coding");

    MALLOC(local_subset, int, instances, "OutputCodingTraining: local_subset");
    MALLOC(local_labels, int, instances, "OutputCodingTraining: local_labels");

    // Create labels for classifiers and train them
    for (int b = 0; b < learner->ocoder->n_binary; b++) {
        // Create labels
        for (int i = 0; i < instances; i++)
            if (!subset || subset[i]) {
                // Check whether instance is positive, negative, or not
                if (learner->ocoder->M[labels[i]][b] == 1) {    // Positive (aka minority) is set arbitrarily to 1
                    local_subset[i] = 1;
                    local_labels[i] = 0;
                } else if (learner->ocoder->M[labels[i]][b] == -1) {
                    local_subset[i] = 1;
                    local_labels[i] = 1;
                } else {
                    local_subset[i] = 0;
                    local_labels[i] = -1;
                }
            } else {
                local_subset[i] = 0;
                local_labels[i] = -1;
            }

        MALLOC(learner->ocoder->binary[b], t_Learner, 1, "OutputCodingTraining: learner->ocoder->binary[b]");
        ClassifierTraining(parm, learner->ocoder->binary_learner, data, instances, local_labels, inputs, 2, data_index, subspace, local_subset, releasedata, learner->ocoder->binary[b], NULL);
    }

    FREE(local_labels);
    FREE(local_subset);
}

t_OutputCoder *CreateOneVsOneClassifier(const t_Parameters *parm, int classes)
{
    t_OutputCoder *coder;

    MALLOC(coder, t_OutputCoder, 1, "CreateOneVsOneClassifier");

    coder->n_binary = CreateOVOMatrix(&(coder->M), classes);
    MALLOC(coder->binary, union learners *, coder->n_binary, "CreateOneVsOneClassifier: coder->binary");

    if (parm->base_learner == kNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "k-NN for OVO method");
    coder->binary_learner = parm->base_learner;

    return coder;
}
