/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "eval.h"
#include "model.h"
#include "utils.h"

// Implement native decision trees
// Only binary trees are considered

int RandomForestClassify(struct Node **rf, scalar *instance, int *index, int inputs, int *subspace, int n_estimators, int classes, int *output)
{
    float *row;
    int category;

    CALLOC(row, float, inputs, "RandomForestClassify: data");

    // Create description
    if (index) {
        int c = 0;
        while (index[c] != -1) {
            if (!subspace || subspace[index[c]-1])
                row[index[c]-1] = instance[c];
            c++;
        }
    } else {
        if (!subspace)
            for (int i = 0; i < inputs; i++)
                row[i] = instance[i];
        else {
            for (int i = 0; i < inputs; i++)
                row[i] = instance[i] * subspace[i];
        }
    }

    // float *class = get_predictions(data, 1, rf, n_estimators);  // Internal only valid for binary problems
    category = RFClassifyInstance(rf, n_estimators, row, classes, output);
    FREE(row);

    // Return category
    return category;
}

int RFClassifyInstance(struct Node **trees, int n_estimators, float *row, int classes, int *outputs)
{
    int *votes;

    if (outputs)
        votes = outputs;
    else
        ALLOCA(votes, int, classes, "RFClassifyInstance");

    ZeroVector(NULL, votes, classes);


    for (int i = 0; i < n_estimators; i++)
        votes[(int) predict(trees[i], row)]++;

    return MaxVectorIndex(NULL, votes, classes);
}


void RandomForest2ClassOutput()
{
    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Random forest 2 class output");
}

void CrossValidatedParametersRandomForest()
{
    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Random forest parameters cross-validation");
}


struct Node **TrainRandomForest(const t_Parameters *parm, scalar **data, int *subspace, int instances, int *labels, int inputs, int *subset, int **index)
{
    struct Node **model;       // Random forest model
    float **ldata;

    // Count selected inputs
    int features = 0;
    for (int i = 0; i < inputs; i++)
        if (!subspace || subspace[i])
            features++;

    // Count selected instances
    int selected = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            selected++;

    // If no inputs or instances do nothing
    if (!features || !selected)
        return NULL;

    // Parameter values
    struct RF_params parameters;
    parameters.n_estimators = parm->rf_ntree;
    parameters.min_samples_leaf = parm->dt_min_samples_split;
    parameters.sampling_ratio = 1;
    parameters.max_depth = parm->dt_max_depth;
    parameters.max_features = parm->rf_percent > 0 ? parm->rf_percent*features : ceil(sqrt((scalar) features));

    // Create data structure for the classifier
    // Allocate memory
    MALLOC(ldata, float *, selected, "TrainRandomForest: data");

    // Create data for RF taking into account selected instances
    int next = 0;
    for (int j = 0; j < instances; j++)
        if (!subset || subset[j]) {
            // Sparse format
            if (index) {
                CALLOC(ldata[next], float, inputs+1, "TrainRandomForest: ldata[next]");
                int c = 0;
                while (index[j][c] != -1) {
                    if (!subspace || subspace[index[j][c]-1])
                        ldata[next][index[j][c]-1] = data[j][c];

                    c++;
                }
            } else {
                MALLOC(ldata[next], float, inputs+1, "TrainRandomForest: ldata[next]");
                if (subspace)
                    for (int i = 0; i < inputs; i++)
                        ldata[next][i] = data[j][i] * subspace[i];
                else
                    for (int i = 0; i < inputs; i++)
                        ldata[next][i] = data[j][i];

            }

            ldata[next][inputs] = (float) labels[j];

            next++;
        }

    // Create classifier
    model = fit_model(ldata, parameters, selected, inputs+1, parm->show_progress);

    // Free memory
    for (int i = 0; i < selected; i++)
        FREE(ldata[i]);
    FREE(ldata);

    return model;
}

scalar RandomForestMargin(struct Node **rf, scalar *instance, int label, int classes, int inputs, int *subspace, int *index, int n_estimators)
{
    int max_index;
    scalar max_value, margin;
    int output[classes];

    float *row;

    CALLOC(row, float, inputs, "RandomForestClassify: data");

    // Create description
    if (index) {
        int c = 0;
        while (index[c] != -1) {
            if (!subspace || subspace[index[c]-1])
                row[index[c]-1] = instance[c];
            c++;
        }
    } else {
        if (!subspace)
            for (int i = 0; i < inputs; i++)
                row[i] = instance[i];
        else {
            for (int i = 0; i < inputs; i++)
                row[i] = instance[i] * subspace[i];
        }
    }

    RFClassifyInstance(rf, n_estimators, row, classes, output);
    FREE(row);

    // Get classification.
    max_index = 0;
    max_value = output[0];
    for (int j = 1; j < classes; j++)
        if (output[j] > max_value) {
            max_value = output[j];
            max_index = j;
        }

    // Test if correctly Classified.
    if (label == max_index) {
        // Difference with second one
        margin = max_value;

        // Find second highest
        output[max_index] = -100000;
        max_index = 0;
        max_value = output[0];
        for (int j = 1; j < classes; j++)
            if (output[j] > max_value) {
                max_value = output[j];
                max_index = j;
            }
        margin -= max_value;

    } else
        // Difference with correct one
        margin = output[label]-max_value;

    return margin;
}
