/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "eval.h"
#include "utils.h"
#include "ensemble.h"

// Implement native random forest

// Implement extremely randomized trees
// P. Geurts, D. Ernst and L. Wehenkel (2006) "Extremely randomized trees," Machine Learning 63:3–42.

int RForestClassify(t_RForest *forest, scalar *instance, int *index, int inputs, int *subspace, int classes, scalar *outputs)
{
    scalar votes[classes];
    ZeroVector(votes, NULL, classes);

    for (int i = 0; i < forest->n_trees; i++)
        votes[DecisionTreeClassify(forest->trees[i], instance, NULL, classes, index)]++;

    int category = IndexOfMaximumOfScalarVector(votes, classes);

    if (outputs) {
        memcpy(outputs, votes, sizeof(scalar)*classes);
        NormalizeVector(outputs, classes);
    }

    // Return category
    return category;
}

t_RForest *TrainRForest(const t_Parameters *parm, scalar **data, int *subspace, int instances, int *labels, int inputs, int *subset, int **index, int classes, bool ert)
{
    t_RForest *forest;       // Random forest model

    // Count selected inputs
    int features = 0;
    for (int i = 0; i < inputs; i++)
        if (!subspace || subspace[i])
            features++;

    // Count selected instances
    int selected = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            selected++;

    // If no inputs or instances do nothing
    if (!features || !selected)
        return NULL;

    MALLOC(forest, t_RForest, 1, "TrainRForest");
    forest->n_trees = parm->rf_ntree;
    MALLOC(forest->trees, t_DTree, forest->n_trees, "TrainRForest: forest->trees");

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < forest->n_trees; i++) {
        // Create bagging sample
        scalar **bdata;
        int **bindex = NULL, *blabels;

        MALLOC(bdata, scalar *, instances, "TrainRForest: bdata");
        MALLOC(blabels, int, instances, "TrainRForest: blabels");
        if (index)
            MALLOC(bindex, int *, instances, "TrainRForest: bindex");

        for (int j = 0; j < instances; j++) {
            int next;
            do {
                next = RANDOM(0, instances);
            } while (subset && !subset[next]);

            bdata[j] = data[next];
            blabels[j] = labels[next];
            if (index)
                bindex[j] = index[next];
        }

        SHOWPROGRESS(parm->show_progress > 0, "RForest BEGIN  training %0*d/%d tree\r", WIDTH(forest->n_trees), i+1, forest->n_trees);
        forest->trees[i] = DecisionTreeTrain(parm, bdata, instances, blabels, features, NULL, subspace, classes, false, bindex, parm->rforest_split_measure, true, ert);
        SHOWPROGRESS(parm->show_progress > 0, "RForest FINISH training %0*d/%d tree\r", WIDTH(forest->n_trees), i+1, forest->n_trees);

        FREE(bdata);
        FREE(blabels);
        if (index)
            FREE(bindex);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    return forest;
}

scalar RForestMargin(t_RForest *forest, scalar *instance, int label, int classes, int inputs, int *subspace, int *index)
{
    int max_index;
    scalar max_value, margin, output[classes];

    RForestClassify(forest, instance, index, inputs, subspace, classes, output);

    // Get classification
    max_index = IndexOfMaximumOfScalarVector(output, classes);
    max_value = output[max_index];

    // Test if correctly Classified.
    if (label == max_index) {
        // Difference with second one
        margin = max_value;

        // Find second highest
        output[max_index] = -100000;
        max_index = 0;
        max_value = output[0];
        for (int j = 1; j < classes; j++)
            if (output[j] > max_value) {
                max_value = output[j];
                max_index = j;
            }
        margin -= max_value;

    } else
        // Difference with correct one
        margin = output[label]-max_value;

    return margin;
}

void FreeRForest(t_RForest *forest)
{
    for (int i = 0; i < forest->n_trees; i++)
        FreeDecisionTree(forest->trees[i]);
    FREE(forest->trees);
}
