/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"

// Implement native decision trees
// Split measures

typedef struct {
    int class;
    scalar value, w;
} t_FeatureInfo;


#define EPSILON 0.1

void GetBestCupPoint(const t_Parameters *parm, scalar **data, int *labels, int instances, int *i_list, int feature, scalar *cut_point, scalar *score, int classes, scalar *w, int **index,
                     t_DTSplitMeasure split_measure, bool ert)
{
    t_FeatureInfo *info;
    int n_cuts, next;
    scalar *cuts, pentropy = 0.0, new_score = 0.0, cleft[classes], cright[classes];

    MALLOC(info, t_FeatureInfo, instances, "GetBestCupPoint: info");

    // Sort values
    for (int i = 0; i < instances; i++) {
        info[i].class = labels[i_list[i]];
        info[i].value = index ? SparseFeature(data[i_list[i]], index[i_list[i]], feature) : data[i_list[i]][feature];
        info[i].w = w ? w[i_list[i]] : 0.0;
        if (w && (split_measure == imGainRatio || split_measure == imInformationGain))
            info[i].w *= instances;
    }

    qsort(info, instances, sizeof(t_FeatureInfo), CompareFeature);

    // If constant values return 0 score
    if (info[0].value == info[instances-1].value) {
        *score = *cut_point = 0.0;
        FREE(info);
        return;
    }

    // Get cut point scores
    // For extremely randomized trees only a random cut point is considered
    MALLOC(cuts, scalar, instances+1, "GetBestCupPoint: cuts");
    if (ert) {
        n_cuts = 1;
        cuts[0] = RANDOM(info[0].value, info[instances-1].value);
    } else {
        cuts[0] = info[0].value - EPSILON;
        n_cuts = 1;
        for (int i = 0; i < instances-1; i++)
            if (info[i].value != info[i+1].value) {
                cuts[n_cuts] = (info[i].value+info[i+1].value)/2.0;
                n_cuts++;
            }

        cuts[n_cuts] = info[instances-1].value + EPSILON;
        n_cuts++;
    }

    // First cut point, all in the right child
    ZeroVector(cleft, NULL, classes);
    ZeroVector(cright, NULL, classes);
    if (w)
        for (int i = 0; i < instances; i++)
            cright[labels[i_list[i]]] += w[i_list[i]];
    else
        for (int i = 0; i < instances; i++)
            cright[labels[i_list[i]]]++;

    // Maximum best
    *score = (split_measure == imGini) ? INFINITE : -INFINITE;

    // Obtain parent entropy if needed
    if (split_measure == imGainRatio || split_measure == imInformationGain)
        pentropy = DTClassEntropy(instances, classes, cright);

    // All possible thresholds
    next = 0;
    for (int i = 0; i < n_cuts; i++) {
        // IBT decision trees
        if (w)
            while (next < instances && info[next].value <= cuts[i]) {
                cleft[info[next].class] += info[next].w;
                cright[info[next].class] -= info[next].w;
                next++;
            }
        else // Standard decision trees
            while (next < instances && info[next].value <= cuts[i]) {
                cleft[info[next].class]++;
                cright[info[next].class]--;
                next++;
            }

        switch (split_measure) {
            case imAccuracy:
                if (w) {
                    // If w get weighted error
                    int left = MajorityClass(cleft, classes);
                    int right = MajorityClass(cright, classes);
                    new_score = 0.0;
                    for (int j = 0; j < instances; j++)
                        if ((info[j].value <= cuts[i] && info[j].class == left) || (info[j].value > cuts[i] && info[j].class == right))
                            new_score += info[j].w;
                } else
                    new_score = ((scalar)(cleft[MajorityClass(cleft, classes)]+cright[MajorityClass(cright, classes)]))/instances;

                if (new_score > *score) {
                    *score = new_score;
                    *cut_point = cuts[i];
                }
                break;

            case imGainRatio:
                new_score = GainRatio(instances, classes, pentropy, cleft, cright);
                if (new_score > *score) {
                    *score = new_score;
                    *cut_point = cuts[i];
                }
                break;

            case imInformationGain:
                new_score = InformationGain(instances, classes, pentropy, cleft, cright, SumVector(cleft, NULL, classes));
                if (new_score > *score) {
                    *score = new_score;
                    *cut_point = cuts[i];
                }
                break;

            case imGini:
                new_score = GiniSplit(instances, classes, cleft, cright);
                if (new_score < *score) {
                    *score = new_score;
                    *cut_point = cuts[i];
                }
                break;
        }
    }

    FREE(cuts);
    FREE(info);
}

int CompareFeature(const void *a, const void *b)
{
    if (((t_FeatureInfo *)a)->value < ((t_FeatureInfo *)b)->value)
        return -1;
    else if (((t_FeatureInfo *)a)->value > ((t_FeatureInfo *)b)->value)
        return +1;
    else
        return 0;
}

int MajorityClass(scalar *count, int classes)
{
    int majority = 0;

    for (int i = 1; i < classes; i++)
        if (count[majority] < count[i])
            majority = i;

    return majority;
}

scalar DTClassEntropy(int instances, int classes, scalar *count)
{
    scalar entropy = 0.0;

    // Use weights? count must include the weights
    for (int i = 0; i < classes; i++)
        if (count[i] > ZERO)
            entropy -= (count[i]/instances)*log2(count[i]/instances);

    return entropy;
}

scalar InformationGain(int instances, int classes, scalar pentropy, scalar *lcount, scalar *rcount, int linstances)
{
    scalar gain = pentropy;

    gain -= (((scalar)linstances)/instances)*DTClassEntropy(linstances, classes, lcount);
    gain -= (((scalar)instances-linstances)/instances)*DTClassEntropy(instances-linstances, classes, rcount);

    return gain;
}

scalar GainRatio(int instances, int classes, scalar pentropy, scalar *lcount, scalar *rcount)
{
    int linstances;
    scalar gain, split_info;

    linstances = SumVector(lcount, NULL, classes);

    if (!linstances || linstances == instances)
        return -INFINITE;

    gain = InformationGain(instances, classes, pentropy, lcount, rcount, linstances);
    split_info = -(((scalar)linstances)/instances)*log2((((scalar)linstances)/instances)) -(((scalar)instances-linstances)/instances) * log2((((scalar)instances-linstances)/instances));

    return gain/split_info;
}

scalar Gini(int instances, int classes, scalar *count)
{
    scalar gini = 1.0;

    for (int i = 0; i < classes; i++)
        gini -= pow(((scalar) count[i])/instances, 2.0);

    return gini;
}

scalar GiniSplit(int instances, int classes, scalar *lcount, scalar *rcount)
{
    int linstances;
    scalar gini_split;

    linstances = SumVector(lcount, NULL, classes);

    if (!linstances || linstances == instances)
        return INFINITE;

    gini_split = (((scalar)linstances)/instances) * Gini(linstances, classes, lcount) +
                 (((scalar)instances-linstances)/instances) * Gini(instances-linstances, classes, rcount);

    return gini_split;
}
