/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "macros.h"
#include "common.h"
#include "classifiers.h"
#include "data.h"
#include "svm.h"

struct svm_model *TrainSupportVectorMachine(const t_Parameters *parm, scalar **data, int *subspace, int instances, int *labels, int inputs, int *subset, int **index)
{
    char *tempnam(const char *, const char *);

    struct svm_model *smodel;       // SVM model
    struct svm_problem prob;        // SVM data
    struct svm_parameter param;     // SVM parameters
    struct svm_node *x = NULL;  // Used for sparse data

    int in;
    const char *error_msg;

    // Count selected inputs
    int features = 0;
    for (int i = 0; i < inputs; i++)
        if (!subspace || subspace[i])
            features++;

    // If no inputs do nothing
    if (!features)
        return NULL;

    // Parameter values
    param.svm_type = parm->svm_type;
    param.kernel_type = parm->kernel;
    param.degree = parm->svm_d;
    // default 1/k
    if (parm->gamma == 0.0)
        param.gamma = 1.0/inputs;
    else
        param.gamma = parm->gamma;  // 1/k
    param.coef0 = parm->svm_r;
    param.cache_size = parm->cache_size;
    param.C = parm->svm_c;
    param.eps = parm->tolerance;
    param.p = 0.1;
    param.shrinking = 1;
    param.nr_weight = 0;
    param.weight_label = NULL;
    param.weight = NULL;
    param.nu = 0.5;
    param.probability = 1;

    // Create problem structure for the classifier
    // Allocate memory
    MALLOC(prob.x, struct svm_node *, instances, "TrainSVMClassifierSubspace: prob.x");
    MALLOC(prob.y, double, instances, "TrainSVMClassifierSubspace: prob.y");
    if (!index) {
        for (int i = 0; i < instances; i++)
            MALLOC(prob.x[i], struct svm_node, inputs+1, "TrainSVMClassifierSubspace: prob.x[i]");
    } else
        // If sparse reserve memory for each instance separately
        MALLOC(x, struct svm_node, inputs+1, "TrainSVMClassifierSubspace: x");

    // Create data for SVM taking into account selected instances
    prob.l = 0;
    for (int j = 0; j < instances; j++)
        if (!subset || subset[j]) {
            // Set inputs
            in = 0;

            // Sparse format
            if (index) {
                int c = 0;
                while (index[j][c] != -1) {
                    if (!subspace || subspace[index[j][c]-1]) {
                        x[in].index = index[j][c];
                        x[in].value  = data[j][c];
                        in++;
                    }
                    c++;
                }

                MALLOC(prob.x[prob.l], struct svm_node, in+1, "TrainSVMClassifierSubspace: prob.x[prob.l]")
                if (in > 0) // Might a null instance
                    memcpy(prob.x[prob.l], x, sizeof(struct svm_node)*in);
            } else {
                for (int i = 0; i < inputs; i++)
                    if ((!subspace || subspace[i]) && data[j][i] * data[j][i] > ZERO2) {
                        prob.x[prob.l][in].index = i;
                        prob.x[prob.l][in].value = data[j][i];
                        in++;
                    }
            }

            prob.x[prob.l][in].index = -1;

            // Set class label
            prob.y[prob.l] = labels[j];

            prob.l++;
        }

    // Check parm
    error_msg = svm_check_parameter(&prob, &param);
    if (error_msg)
        OnErrorExit(1, (char *) error_msg);

    // Create classifier
    smodel = svm_train(&prob, &param);

    // Free memory
    char *file_template = tempnam(".", NULL);
    svm_save_model(file_template, smodel);
    svm_free_and_destroy_model(&smodel);
    svm_destroy_param(&param);

#ifdef _OPENMP
    #pragma omp critical
    {
#endif
        if ((smodel = svm_load_model(file_template)) == NULL) {
            remove(file_template);
            OnErrorExit(1, "Error loading SVM model");
        }
        remove(file_template);
#ifdef _OPENMP
    }
#endif

    // Free memory
    for (int i = 0; i < (index ? prob.l : instances); i++)
        FREE(prob.x[i]);
    FREE(prob.x);
    FREE(prob.y);
    if (x)
        FREE(x);

    return smodel;
}

int SupportVectorMachineOutput(struct svm_model *smodel, scalar *instance, int inputs, int *subspace, int *index, scalar *probabilities)
{
    scalar category;
    struct svm_node *x;
    int in;

    MALLOC(x, struct svm_node, inputs + 1, "SupportVectorMachineOutput: x");

    // Create description
    in = 0;
    if (index) {
        int c = 0;
        while (index[c] != -1) {
            if (!subspace || subspace[index[c]-1]) {
                x[in].index = index[c];
                x[in].value = instance[c];
                in++;
            }
            c++;
        }
    } else {
        for (int i = 0; i < inputs; i++)
            if ((!subspace || subspace[i]) && instance[i] * instance[i] > ZERO2) {
                x[in].index = i;
                x[in].value = instance[i];
                in++;
            }
    }

    x[in].index = -1;

    // Get output
    if (!probabilities)
        category = svm_predict(smodel, x);
    else {
        double probs[smodel->nr_class];
        category = svm_predict_probability(smodel, x, probs);
        for (int i = 0; i < smodel->nr_class; i++)
            probabilities[smodel->label[i]] = probs[i];
    }

    FREE(x);

    // Return category
    return (int) category;
}

int SupportVectorMachine2ClassOutput(struct svm_model *smodel, scalar *instance, int inputs, double *output, int *subspace, int minority, int *index)
{
    struct svm_node *x;
    int in, category;

    MALLOC(x, struct svm_node, inputs + 1, "SupportVectorMachineOutput: x");

    // String kernel?
    // Create description
    in = 0;
    if (index) {
        int c = 0;
        while (index[c] != -1) {
            if (!subspace || subspace[index[c]-1]) {
                x[in].index = index[c];
                x[in].value = instance[c];
                in++;
            }
            c++;
        }
    } else {
        if (subspace) {
            for (int i = 0; i < inputs; i++)
                if (subspace[i] && instance[i] * instance[i] > ZERO2) {
                    x[in].index = i;
                    x[in].value = instance[i];
                    in++;
                }
        } else {
            for (int i = 0; i < inputs; i++)
                if (instance[i] * instance[i] > ZERO2) {
                    x[in].index = i;
                    x[in].value = instance[i];
                    in++;
                }
        }
    }

    x[in].index = -1;

    // Get output
    category = svm_predict_values(smodel, x, output);

    FREE(x);

    // Change sign?
    if (minority != smodel->label[0])
        output[0] *= -1.0;

    return category;
}

scalar SupportVectorMachineMargin(struct svm_model *svm, scalar *instance, int label, int *subspace, int classes, int inputs)
{
    double dec_values[classes*(classes-1)/2];
    scalar margin;
    int max_index, in, category, index;
    struct svm_node svm_x[inputs+1];

    int nr_class = svm->nr_class;
    scalar output[nr_class];

    in = 0;
    for (int i = 0; i < inputs; i++)
        if (instance[i]*instance[i]*subspace[i] > ZERO2) {
            svm_x[in].index = i;
            svm_x[in].value = instance[i];
            in++;
        }
    svm_x[in].index = -1;

    svm_predict_values(svm, svm_x, dec_values);

    if (classes == 2) {
        if (svm->label[0] > 0.5)
            return dec_values[0] - 0.5;
        else
            return 0.5 - dec_values[0];
    }

    ZeroVector(output, NULL, nr_class);

    int pos = 0;
    for (int i = 0; i < nr_class; i++)
        for (int j = i+1; j < nr_class; j++) {
            // Voting
            if (dec_values[pos] > 0.0)
                output[i]++;
            else
                output[j]++;

            pos++;
        }

    // Get maximum
    max_index = 0;
    for (int i = 1; i < nr_class; i++)
        if (output[i] > output[max_index])
            max_index = i;

    category = svm->label[max_index];

    // Test if correctly classified
    if (label == category) {
        // Difference with second one
        margin = output[max_index];

        // Find second highest
        output[max_index] = -INFINITE;
        max_index = 0;
        for (int j = 1; j < nr_class; j++)
            if (output[j] > output[max_index])
                max_index = j;

        margin -= output[max_index];
    } else {
        // Get index of classes
        index = 0;
        for (int i = 0; i < nr_class && index != -1; i++)
            if (svm->label[i] == label)
                index = i;

        // Difference with correct one
        margin = output[index]-output[max_index];
    }

    return margin;
}

#define SVM_PARAM_SETS  21

void CrossValidatedParametersSupportVectorMachine(t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, int inputs, int classes, int minority, int *subset, int **index)
{
    t_PerformanceInfo perf;
    scalar best_so_far, accuracy = 0.0, **data_v, **data_t, * **ext_str_data = NULL, **s_data = NULL, **n_data = NULL;
    int *label_v = NULL, *label_t, **M, best_index = 0, instances_v, instances_t, majority, n_external_strata, *ext_str_instances = NULL, **ext_str_labels = NULL, **ext_str_list = NULL, *s_labels,
         s_instances, n_instances = 0, *n_labels, **index_t = NULL, **index_v = NULL, ***ext_str_index = NULL, **n_index = NULL;

    // Sets of parm
    scalar svm_p[SVM_PARAM_SETS][3] = {{2.0, 0.0001, 0.1}, {2.0, 0.0001, 1.0}, {2.0, 0.0001, 10.0}, {2.0, 0.001, 0.1}, {2.0, 0.001, 1.0}, {2.0, 0.001, 10.0}, {2.0, 0.01, 0.1}, {2.0, 0.01, 1.0},
        {2.0, 0.01, 10.0}, {2.0, 0.1, 0.1}, {2.0, 0.1, 1.0}, {2.0, .1, 10.0}, {2.0, 1.0, 0.1}, {2.0, 1.0, 1.0}, {2.0, 1.0, 10.0}, {2.0, 10.0, 0.1}, {2.0, 10.0, 1.0}, {2.0, 10.0, 10.0},
        {0.0, 0.0, 0.1}, {0.0, 0.0, 1.0}, {0.0, 0.0, 10.0}
    };

    // Count effective number of instances if a subset given
    int e_instances = instances;
    if (subset) {
        e_instances = 0;
        for (int i = 0; i < instances; i++)
            if (subset[i])
                e_instances++;
    }

    // Sampling?
    if (parm->cv_sampling_size && parm->cv_sampling_size < e_instances)
        switch (parm->cv_sampling_method) {
            case sNone:
                OnErrorExit(ERR_PARAM, "None sampling and sampling size makes no sense");
                break;

            case sKMeans:
                if (index)
                    OnErrorExit(ERR_PARAM, "K-Means sampling for sparse data is not adequate");
                if (subset) {
                    // Only selected instances
                    MALLOC(n_data, scalar *, instances, "CrossValidatedParametersSupportVectorMachine: n_data");
                    MALLOC(n_labels, int, instances, "CrossValidatedParametersSupportVectorMachine: n_labels");
                    for (int i = 0; i < instances; i++)
                        if (subset[i]) {
                            n_data[n_instances] = data[i];
                            n_labels[n_instances] = labels[i];
                            n_instances++;
                        }
                } else {
                    n_data = data;
                    n_instances = instances;
                    n_labels = labels;
                }

                if (parm->stratified)
                    kMeansStratifiedSample((const t_Parameters *) &parm, &n_data, n_instances, &n_labels, &s_data, &s_labels, &s_instances, classes, inputs, parm->cv_sampling_size, NULL, false);
                else
                    kMeansSample((const t_Parameters *) &parm, &data, instances, &labels, &s_data, &s_labels, &s_instances, classes, inputs, parm->cv_sampling_size, NULL, false);

                // Overwrite
                data = s_data;
                instances = s_instances;
                labels = s_labels;

                if (subset) {
                    FREE(n_data);
                    FREE(n_labels);
                    subset = NULL;
                }
                break;

            case sStratified:
                n_external_strata = instances%parm->cv_sampling_size ? instances/parm->cv_sampling_size+1 : instances/parm->cv_sampling_size;

                // Memory
                MALLOC(ext_str_data, scalar **, n_external_strata, "CrossValidatedParametersSupportVectorMachine: ext_str_data");
                MALLOC(ext_str_instances, int, n_external_strata, "CrossValidatedParametersSupportVectorMachine: ext_str_list");
                MALLOC(ext_str_labels, int *, n_external_strata, "CrossValidatedParametersSupportVectorMachine: ext_str_labels");
                MALLOC(ext_str_list, int *, n_external_strata, "CrossValidatedParametersSupportVectorMachine: ext_str_list");
                if (index)
                    MALLOC(ext_str_index, int **, n_external_strata, "CrossValidatedParametersSupportVectorMachine: ext_str_index");
                for (int i = 0; i < n_external_strata; i++) {
                    MALLOC(ext_str_data[i], scalar *, parm->cv_sampling_size, "CrossValidatedParametersSupportVectorMachine: ext_str_data[i]");
                    MALLOC(ext_str_labels[i], int, parm->cv_sampling_size, "CrossValidatedParametersSupportVectorMachine: ext_str_labels[i]");
                    MALLOC(ext_str_list[i], int, parm->cv_sampling_size, "CrossValidatedParametersSupportVectorMachine: ext_str_labels[i]");
                    if (index)
                        MALLOC(ext_str_index[i], int *, parm->cv_sampling_size, "CrossValidatedParametersSupportVectorMachine: ext_str_index[i]");
                }

                if (subset) {
                    // Only selected instances
                    MALLOC(n_data, scalar *, instances, "CrossValidatedParametersSupportVectorMachine: n_data");
                    MALLOC(n_labels, int, instances, "CrossValidatedParametersSupportVectorMachine: n_labels");
                    if (index)
                        MALLOC(n_index, int *, instances, "CrossValidatedParametersSupportVectorMachine: n_index");
                    for (int i = 0; i < instances; i++)
                        if (subset[i]) {
                            n_data[n_instances] = data[i];
                            n_labels[n_instances] = labels[i];
                            if (index)
                                n_index[n_instances] = index[i];
                            n_instances++;
                        }
                } else {
                    n_data = data;
                    n_instances = instances;
                    n_labels = labels;
                    n_index = index;
                }

                // Get first stratum as training set
                DataPartition(parm, n_data, n_instances, n_labels, NULL, &n_external_strata, &ext_str_data, &ext_str_instances, &ext_str_labels, classes, &ext_str_list, n_index, &ext_str_index,
                              inputs, parm->cv_sampling_size, pRANDOM, parm->stratified, NULL);

                // Forget original data
                data = ext_str_data[0];
                instances = ext_str_instances[0];
                labels = ext_str_labels[0];
                if (index)
                    index = ext_str_index[0];

                if (subset) {
                    FREE(n_data);
                    FREE(n_labels);
                    subset = NULL;
                    if (index)
                        FREE(n_index);
                }
                break;
        }

    // If first call allocate memory
    MALLOC(label_v, int, instances, "CrossValidatedParametersSupportVectorMachine: label_v");
    MALLOC(label_t, int, instances, "CrossValidatedParametersSupportVectorMachine: label_t");
    MALLOC(data_v, scalar *, instances, "CrossValidatedParametersSupportVectorMachine: data_v");
    MALLOC(data_t, scalar *, instances, "CrossValidatedParametersSupportVectorMachine: data_t");
    if (index) {
        MALLOC(index_v, int *, instances, "CrossValidatedParametersSupportVectorMachine: index_v");
        MALLOC(index_t, int *, instances, "CrossValidatedParametersSupportVectorMachine: index_t");
    }

    // Size of validation set
    instances_v = instances/parm->cv_folds;
    instances_t = instances - instances_v;

    AllocateIntegerMatrix(&M, classes, classes);

    // Do validation runs
    best_so_far = -1.0;
    for (int i = 0; i < SVM_PARAM_SETS; i++) {
        scalar margin = 0.0, auroc = 0.0, auprc = 0.0;
        struct svm_model *svm = NULL;

        parm->kernel = (int) svm_p[i][0];
        parm->gamma = svm_p[i][1];
        parm->svm_c = svm_p[i][2];
        if (parm->verbose > 2)
            fprintf(parm->out_file, "Kernel = %d\tC = %lf\tgamma = %lf\t", parm->kernel, parm->svm_c, parm->gamma);

        ZeroIntegerMatrix(M, classes, classes);

        // For each fold
        for (int fold = 0; fold < parm->cv_folds; fold ++) {
            // Set fold
            // Validation set
            for (int k = (fold * instances_v) % instances, c = 0; c < instances_v; k++, c++) {
                data_v[c] = data[k];
                label_v[c] = labels[k];
                if (index)
                    index_v[c] = index[k];
            }

            // Train set
            int next = 0;
            for (int k = 0, c = ((fold+1) * instances_v) % instances; k < instances_t; k++, c = (c+1) % instances)
                if (!subset || subset[c]) {
                    data_t[next] = data[c];
                    label_t[next] = labels[c];
                    if (index)
                        index_t[next] = index[c];
                    next++;
                }

            // None? Continue with next fold
            if (!next)
                continue;

            // Train and evaluate classifier
            // Learn model
            if (parm->random_undersampling_for_training && !parm->weigthed_sampling_for_training && classes == 2) {
                if (index)
                    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Temporal undersampling for training with sparse data");
                scalar **u_data = NULL;
                int u_instances = 0, *u_labels = NULL;
                TemporalUndersampleDataForTraining(data_t, next, classes, label_t, parm->random_undersampling_ratio, &u_data, &u_labels, &u_instances, minority);
                svm = TrainSupportVectorMachine(parm, u_data, subspace, u_instances, u_labels, inputs, NULL, NULL);
                FREE(u_data);
                FREE(u_labels);
            } else
                svm = TrainSupportVectorMachine(parm, data_t, subspace, next, label_t, inputs, NULL, index_t);

            // If margin as performance it must be obtained here
            if (parm->cv_performance_measure == pfMargin) {
                scalar minimum_margin;
                t_Learner learner;

                learner.svm = svm;
                margin += ClassifierDatasetMargin(parm, SVM, data_t, instances_t, label_t, data_v, instances_v, label_v, inputs, classes, subspace, &minimum_margin, &learner, NULL, subset, NULL, true);
            } else if (parm->cv_performance_measure == pfauROC) {
                t_Learner learner;

                learner.svm = svm;
                auroc += ClassifierAUROC(parm, SVM, data_v, instances_v, label_v, data_t, instances_t, subset, subspace, label_t, inputs, minority, false, &learner, false, parm->out_file,
                                               classes, NULL, NULL, NULL, NULL, NULL, true);
            } else if (parm->cv_performance_measure == pfauPRC) {
                t_Learner learner;

                learner.svm = svm;
                auprc += ClassifierAUPRC(parm, SVM, data_v, instances_v, label_v, data_t, instances_t, label_t, inputs, minority, false, &learner, parm->out_file, classes, NULL, NULL, subspace,
                                               subset, NULL, NULL, NULL, true);
            }

            // Get M matrix and metric for performance
            for (int i = 0; i < instances_v; i++)
                M[label_v[i]][SupportVectorMachineOutput(svm, data_v[i], inputs, subspace, index_v ? index_v[i] : NULL, NULL)]++;

            svm_free_and_destroy_model(&svm);
        }

        PrintMulticlassMeasures(M, classes, instances, NULL, NULL, &perf);
        switch (parm->cv_performance_measure) {
            case pfAccuracy:
                accuracy = perf.accuracy;
                break;

            case pfGMean:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.gmean;
                break;

            case pfMargin:
                accuracy = margin/parm->cv_folds;
                break;

            case pfauROC:
                accuracy = auroc/parm->cv_folds;
                break;

            case pfauPRC:
                accuracy = auprc/parm->cv_folds;
                break;

            case pfPrecisionMacro:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Metric for cross-validation of parameters");
                break;

            case pfCC:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.mcc;
                break;

            case pfRecallMacro:
                accuracy = perf.recall_macro;
                break;

            case pfRecallMacroGMean:
                accuracy = perf.recall_macro_g_mean;
                break;

            case pfKappa:
                accuracy = perf.kappa;
                break;

            case pfAverageAccuracy:
                accuracy = perf.average_accuracy;
                break;

            case pfF1:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "Two-class measure for multi-class problem");
                majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
                accuracy = perf.f1_score;
                break;

            case pfF1Macro:
                accuracy = perf.f1_score_macro;
                break;
        }

        // Better than current best?
        if (accuracy > best_so_far) {
            best_so_far = accuracy;
            best_index = i;
        }

        SHOWPROGRESS(parm->show_progress > 1, "Accuracy: %lf\tKernel: %s\tC: %4.1lf\tGamma: %7.4lf\n", accuracy, parm->kernel == 0 ? "Linear" : "RBF", parm->svm_c, parm->gamma);
    }

    parm->kernel = (int) svm_p[best_index][0];
    parm->gamma = svm_p[best_index][1];
    parm->svm_c = svm_p[best_index][2];
    if (parm->verbose > 2)
        fprintf(parm->out_file, "Best parameters (accuracy: %lf): Kernel = %d\tC = %lf\tGamma = %lf\n", best_so_far, parm->kernel, parm->svm_c, parm->gamma);
    SHOWPROGRESS(parm->show_progress > 1 && parm->verbose > 0, "Best parameters. Accuracy %lf\tKernel: %s\tC: %4.1lf\tGamma: %7.4lf\n", best_so_far, parm->kernel == 0 ? "Linear" : "RBF",
                 parm->svm_c, parm->gamma);
    SAVEPROGRESS(parm->show_progress > 0, "Best parameters. Accuracy %lf\tKernel: %s\tC: %4.1lf\tGamma: %7.4lf\n", best_so_far, parm->kernel == 0 ? "Linear" : "RBF", parm->svm_c, parm->gamma);

    FreeIntegerMatrix(&M, classes);
    FREE(label_v);
    FREE(label_t);
    FREE(data_v);
    FREE(data_t);
    if (index) {
        FREE(index_t);
        FREE(index_v);
    }

    if (ext_str_data) {
        for (int i = 0; i < n_external_strata; i++) {
            FREE(ext_str_data[i]);
            FREE(ext_str_labels[i]);
            FREE(ext_str_list[i]);
            if (index)
                FREE(ext_str_index[i]);
        }
        FREE(ext_str_list);
        FREE(ext_str_data);
        FREE(ext_str_labels);
        FREE(ext_str_instances);
        if (index)
            FREE(ext_str_index);
    } else if (s_data) {
        FREE(s_labels);
        for (int i = 0; i < s_instances; i++)
            FREE(s_data[i]);
        FREE(s_data);
    }
}

