/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdbool.h>
#include <time.h>
#include <string.h>
#include "typex.h"
#include "macros.h"
#include "classifiers.h"
#include "external.h"
#include "ensemble.h"

// Test of all internal and external classifiers

void TestAllClassifiers(const t_Parameters *parm, scalar **train_data, int train_instances, int *train_labels, int classes, scalar **test_data, int test_instances, int *test_labels, int features,
                        int *subspace, int *subset, char *out_file_name, int minority, bool training_error, t_PerformanceInfo *test_perf, char *file_for_testing_classification,
                        cache_t **neighbors_cache, int **train_index, int **test_index, char *prefix)
{
    int **M, **M_training = NULL;
    scalar average_margin, minimum_margin;
    t_PerformanceInfo perf, perf_training;
    extern char *learner[], *external[];
    t_Learner rlearner;
    t_Ensemble rensemble;
    char text[1024];

    AllocateIntegerMatrix(&M, classes, classes);
    if (training_error)
        AllocateIntegerMatrix(&M_training, classes, classes);

    char *learner_training[] = {"SVM Training", "k-NN Training", "C4.5 Training", "ELM Training", "MLP Training", "NBC Training", "DT training", "IBT training", "OVO training", "RF training"};
    for (t_Classifier evaluator = 0; evaluator < N_CLASSIFIERS; evaluator++) {
        // Text for output messages
        if (prefix)
            sprintf(text, "%s %s", prefix, learner[evaluator]);
        else
            strcpy(text, learner[evaluator]);

        time_t init_time = time(NULL);

        // Use this evaluator?
        if (!parm->evaluate_with[evaluator]) {
            SHOWPROGRESS(parm->show_progress > 1, "NOT evaluating %s\n", text);
            continue;
        }

        SHOWPROGRESS(parm->show_progress > 0, "Training and evaluating %s", text);
        ClassifierEvaluation(parm, test_data, test_instances, test_labels, train_data, train_instances, train_labels, subspace, features, subset, parm->cv_hyperparameters_for_testing, classes,
                                   evaluator, &rlearner, minority, M, M_training, &perf, &perf_training, parm->out_file, text, learner_training[evaluator], neighbors_cache, test_index,
                                   train_index, &rensemble);

        if (evaluator == C45) {
            if (parm->print_dt) {
                fprintf(parm->out_file, "─────────────────────── C4.5 DECISION TREE ─────────────────────\n");
                PrintTree(rlearner.tree, parm->out_file);
                fprintf(parm->out_file,
                        "────────────────────────────────────────────────────────────────\n");
            }
            FREE(FileName);
            if (Item) {
                for (int i = 0; i < train_instances; i++)
                    FREE(Item[i]);
                FREE(Item);
            }
        }

        // Print cross-validated k if apply
        if (evaluator == kNN && (parm->cv_hyperparameters_for_testing || parm->cv_k_for_testing))
            fprintf(parm->out_file, "@CV k: %d\n", parm->k);

        if (classes == 2 && parm->obtain_auc && !parm->minimum_info && !train_index) {
            // auROC
            perf.auroc = ClassifierAUROC(parm, evaluator, test_data, test_instances, test_labels, train_data, train_instances, subset, subspace, train_labels, features, minority, false,
                                               &rlearner, false, parm->out_file, classes, NULL, NULL, NULL, NULL, &rensemble, !IsEnsemble(evaluator));
            fprintf(parm->out_file, "@%s auROC: %lf\n", text, perf.auroc);

            // auPRC
            perf.auprc = ClassifierAUPRC(parm, evaluator, test_data, test_instances, test_labels, train_data, train_instances, train_labels, features, minority, false, &rlearner,
                                               parm->out_file, classes, NULL, NULL, subspace, subset, NULL, NULL, &rensemble, !IsEnsemble(evaluator));
            fprintf(parm->out_file, "@%s auPRC: %lf\n", text, perf.auprc);
        } else
            perf.auprc = perf.auroc = 0.0;

        if (evaluator == SVM && rlearner.svm != NULL)
            fprintf(parm->out_file, "@SVM Support vectors: %d\n", rlearner.svm->l);

        else if (evaluator == C45 && rlearner.tree != NULL)
            fprintf(parm->out_file, "@C4.5 Nodes: %d\n", TreeSize(rlearner.tree));
        else if ((evaluator == DT_C45 || evaluator == DT_CA || evaluator == DT_CART || evaluator == DT_ID3 || evaluator == IBT) && rlearner.dt != NULL)
            fprintf(parm->out_file, "@%s Nodes: %d\n", text, CountDTNodes(rlearner.dt));

        // Get margin
        if (parm->evaluate_margin && !parm->minimum_info && !train_index) {
            average_margin = ClassifierDatasetMargin(parm, evaluator, train_data, train_instances, train_labels, test_data, test_instances, test_labels, features, classes, subspace,
                             &minimum_margin, &rlearner, NULL, subset, &rensemble, !IsEnsemble(evaluator));

            fprintf(parm->out_file, "@%s Average margin: %lf\n", text, average_margin);
            fprintf(parm->out_file, "@%s Minimum margin: %lf\n", text, minimum_margin);
        }

        fprintf(parm->out_file, "%s evaluation time: %ld\n", text, time(NULL) - init_time);

        // Partial Ok
        SHOWPROGRESS(parm->show_progress > 0, " ---> @ok%c", '\n');
        fprintf(parm->out_file, "@ok\n");
        fflush(parm->out_file);

        // Return values of performance
        if (test_perf)
            test_perf[evaluator] = perf;

        if (file_for_testing_classification) {
            FILE *f;

            if ((f = fopen(file_for_testing_classification, "w")) == NULL)
                OnErrorExit(ERR_FILE, "Cannot open file for testing instances output");

            for (int i = 0; i < test_instances; i++)
                fprintf(f, "%d\n", ClassifierOutput(parm, evaluator, test_data[i], features, classes, test_index[i], subspace, train_data, train_instances, train_labels, subset, &rlearner,
                        test_labels[i], neighbors_cache, i, train_index, &rensemble));

            fclose(f);
        }

        ClassifierFree(parm, &rlearner, evaluator, classes, features, &rensemble);
    }

    scalar *sc_pos = NULL, *sc_neg = NULL;

    if (parm->obtain_auc && classes == 2) {
        MALLOC(sc_pos, scalar, test_instances, "TestAllClassifiers: sc_pos");
        MALLOC(sc_neg, scalar, test_instances, "TestAllClassifiers: sc_neg");
    }

    // External classifiers
    for (t_ExternalClassifier evaluator = 0; evaluator < N_EXTERNAL; evaluator++) {
        time_t init_time = time(NULL);

        // Text for output messages
        if (prefix)
            sprintf(text, "%s %s", prefix, external[evaluator]);
        else
            strcpy(text, external[evaluator]);

        if (!parm->evaluate_with_external[evaluator]) {
            SHOWPROGRESS(parm->show_progress > 1, "NOT evaluating %s\n", text);
            continue;
        }

        // Confusion matrix
        SHOWPROGRESS(parm->show_progress > 0, "Training and evaluating %s", text);
        GetExternalClassifierConfusionMatrix(parm, evaluator, test_data, test_instances, test_labels, subset, subspace, classes, features, train_data, train_instances, train_labels, test_index,
                                             train_index, M, out_file_name, &average_margin, &minimum_margin, NULL, minority, sc_pos, sc_neg);

        PrintMulticlassMeasures(M, classes, test_instances, parm->out_file, text, &perf);

        if (classes == 2) {
            int majority = (minority+1)%2;
            PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], parm->out_file, text, &perf);

            if (parm->obtain_auc && !parm->minimum_info) {
                // auROC
                perf.auroc = ClassifierAUROC(parm, (t_Classifier) -1, test_data, test_instances, test_labels, train_data, train_instances, subset, subspace, train_labels, features, minority,
                                                   false, &rlearner, false, parm->out_file, classes, test_index, train_index, sc_pos, sc_neg, NULL, true);
                fprintf(parm->out_file, "@%s auROC: %lf\n", text, perf.auroc);

                // auPRC
                perf.auprc = ClassifierAUPRC(parm, (t_Classifier) -1, test_data, test_instances, test_labels, train_data, train_instances, train_labels, features, minority, false, &rlearner,
                                                   parm->out_file, classes, test_index, train_index, subspace, subset, sc_pos, sc_neg, NULL, true);
                fprintf(parm->out_file, "@%s auPRC: %lf\n", text, perf.auprc);
            } else
                perf.auprc = perf.auroc = 0.0;
        }

        fprintf(parm->out_file, "@%s Average margin: %lf\n", text, average_margin);
        fprintf(parm->out_file, "@%s Minimum margin: %lf\n", text, minimum_margin);

        if (training_error) {
            // Text for output messages
            if (prefix)
                sprintf(text, "%s %s training", prefix, external[evaluator]);
            else
                sprintf(text, "%s training", external[evaluator]);

            // Confusion matrix
            GetExternalClassifierConfusionMatrix(parm, evaluator, train_data, train_instances, train_labels, subset, subspace, classes, features, train_data, train_instances, train_labels,
                                                 test_index, train_index, M, out_file_name, NULL, NULL, NULL, -1, NULL, NULL);

            PrintMulticlassMeasures(M, classes, train_instances, parm->out_file, text, &perf);

            if (classes == 2) {
                int majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], parm->out_file, text, &perf);
            }
        }

        fprintf(parm->out_file, "%s evaluation time: %ld\n", text, time(NULL) - init_time);

        // Partial Ok
        SHOWPROGRESS(parm->show_progress > 0, " ---> @ok%c", '\n');
        fprintf(parm->out_file, "@ok\n");
        fflush(parm->out_file);

        // Return values of performance
        if (test_perf)
            test_perf[evaluator+N_CLASSIFIERS] = perf;
    }

    if (parm->obtain_auc && classes == 2) {
        FREE(sc_pos);
        FREE(sc_neg);
    }

    FreeIntegerMatrix(&M, classes);
    if (M_training)
        FreeIntegerMatrix(&M_training, classes);
}
