/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"

// Implement native decision trees
// Only binary trees are considered

int DecisionTreeClassify(t_DTree tree, scalar *instance, scalar *probabilities, int classes, int *index)
{
    // Leaf node?
    if (tree->leaf) {
        if (probabilities)
            memcpy(probabilities, tree->probabilities, sizeof(scalar)*classes);
        return (tree->class);
    } else {
        if ((index ? SparseFeature(instance, index, tree->feat) : instance[tree->feat]) <= tree->cut_point)
            return DecisionTreeClassify(tree->left, instance, probabilities, classes, index);
        else
            return DecisionTreeClassify(tree->right, instance, probabilities, classes, index);
    }
}

t_DTree DecisionTreeTrain(const t_Parameters *parm, scalar **data, int instances, int *labels, int features, int *subset, int *subspace, int classes, bool ibt, int **index,
                          t_DTSplitMeasure split_measure, bool for_rforest, bool ert)
{
    t_DTree tree;
    int category;
    scalar parent_sm = -1.0, count[classes];

    int *f_list, fs = 0;
    MALLOC(f_list, int, features, "DecisionTreeTrain: f_list");
    for (int i = 0; i < features; i++)
        if (!subspace || subspace[i]) {
            f_list[fs] = i;
            fs++;
        }

    int *i_list, is = 0;
    MALLOC(i_list, int, instances, "DecisionTreeTrain: i_list");
    ZeroVector(count, NULL, classes);
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            i_list[is] = i;
            is++;
            count[labels[i]]++;
        }

    // Build root node with majority class and no children
    category = MaxVectorIndex(count, NULL, classes);

    MALLOC(tree, struct dt_node, 1, "DecisionTreeTrain: tree");
    MALLOC(tree->probabilities, scalar, classes, "DecisionTreeTrain: tree->probabilities");
    tree->class = category;
    tree->leaf = true;
    tree->depth = 0;
    tree->cut_point = 0.0;
    tree->feat = -1;
    tree->left = tree->right = NULL;
    scalar all = SumVector(count, NULL, classes);
    for (int i = 0; i < classes; i++)
        tree->probabilities[i] = count[i]/all;

    scalar *w = NULL;
    // Internally boosted trees use vector of weights
    if (ibt) {
        MALLOC(w, scalar, is, "DecisionTreeTrain: w");
        for (int i = 0; i < is; i++)
            w[i] = 1.0/is;
    }

    tree->alpha = w ? GetAlpha(parm, w, is, labels, category, i_list) : 1.0;

    // Split measure
    switch (split_measure) {
        case imAccuracy:
            tree->score = parent_sm = count[category]/is;
            break;

        case imGainRatio:
        case imInformationGain:
            tree->score = parent_sm = DTClassEntropy(is, classes, count);
            break;

        case imGini:
            tree->score = parent_sm = Gini(is, classes, count);
            break;
    }

    DecisionTreeBuild(parm, tree, data, labels, is, i_list, fs, f_list, parent_sm, classes, 0, w, is, index, split_measure, for_rforest, ert);

    FREE(i_list);
    FREE(f_list);

    if (parm->print_dt)
        PrintDTree(parm, tree, parm->out_file, ibt, split_measure);

    if (ibt)
        FREE(w);

    return tree;
}

typedef struct {
    int feature;
    scalar cut_point, score;
} t_BestSplitInfo;

void DecisionTreeBuild(const t_Parameters *parm, t_DTree tree, scalar **data, int *labels, int instances, int *i_list, int features, int *f_list, scalar parent_score, int classes, int depth,
                       scalar *w, int train_instances, int **index, t_DTSplitMeasure split_measure, bool for_rforest, bool ert)
{
    t_BestSplitInfo *cutpoint;
    bool improves;
    scalar *new_w = NULL;

    // Test if all instances from the same class
    bool one_class = true;
    for (int i = 1; i < instances && one_class; i++)
        if (labels[i_list[0]] != labels[i_list[i]])
            one_class = false;

    SHOWPROGRESS(parm->show_progress > 1, "Depth %d, memory: %.2lfMB\n", depth, ((scalar) MemoryUsage())/pow(2.0, 20));

    //  No minimum samples or maximum depth exceeded?
    if (one_class || instances < parm->dt_min_samples_split || (parm->dt_max_depth && depth >= parm->dt_max_depth)) {
        // Leaf node
        tree->cut_point = 0.0;
        tree->feat = -1;
        tree->leaf = true;
        tree->depth = depth;

        // Get majority class
        tree->class = GetMajorityClass(labels, i_list, instances, classes, tree->probabilities);
        tree->alpha = w ? GetAlpha(parm, w, instances, labels, tree->class, i_list) : 1.0;

        tree->left = tree->right = NULL;
        return;
    }

    // Evaluate every remaining feature
    int efeatures = features;
    if (!for_rforest) {
        MALLOC(cutpoint, t_BestSplitInfo, features, "DecisionTreeBuild: cutpoint");
        for (int i = 0; i < features; i++) {
            cutpoint[i].feature = f_list[i];
            GetBestCupPoint(parm, data, labels, instances, i_list, f_list[i], &(cutpoint[i].cut_point), &(cutpoint[i].score), classes, w, index, split_measure, ert);
        }
    } else {
        // For RForest only a random subset of features is used
        efeatures = parm->rf_percent > 0 ? parm->rf_percent*features : ceil(sqrt((scalar) features));

        RandomizeVector(f_list, features);
        MALLOC(cutpoint, t_BestSplitInfo, efeatures, "DecisionTreeBuild: cutpoint");
        for (int i = 0; i < efeatures; i++) {
            cutpoint[i].feature = f_list[i];
            GetBestCupPoint(parm, data, labels, instances, i_list, f_list[i], &(cutpoint[i].cut_point), &(cutpoint[i].score), classes, w, index, split_measure, ert);
        }
    }

    // Test if best improves
    int best_index = 0;
    // Gini searches minimum
    if (split_measure == imGini) {
        for (int i = 1; i < efeatures; i++)
            if (cutpoint[i].score < cutpoint[best_index].score)
                best_index = i;
    } else
        for (int i = 1; i < efeatures; i++)
            if (cutpoint[i].score > cutpoint[best_index].score)
                best_index = i;

    // If best improves create children and follow up, else return
    // Improve criterion depends on measure
    if (split_measure == imAccuracy)  // Classification accuracy
        improves = cutpoint[best_index].score > parent_score;
    else if (split_measure == imGini) // Gini: CART
        improves = cutpoint[best_index].score < parent_score;
    else    // Gain ratio: C4.5, Information gain: ID3
        improves = cutpoint[best_index].score > 0.0;

    // If IBT check whether to construct complete tree
    if ((w && parm->ibt_no_stop) || parm->dt_grow_full_tree)
        improves = true;

    if (improves) {
        tree->cut_point = cutpoint[best_index].cut_point;
        tree->feat = cutpoint[best_index].feature;
        tree->class = GetMajorityClass(labels, i_list, instances, classes, tree->probabilities);
        tree->alpha = w ? GetAlpha(parm, w, instances, labels, tree->class, i_list) : 1.0;
        tree->leaf = false;
        tree->depth = depth;
        tree->score = cutpoint[best_index].score;

        MALLOC(tree->left, struct dt_node, 1, "DecisionTreeBuild: tree->left");
        MALLOC(tree->left->probabilities, scalar, classes, "DecisionTreeBuild: tree->left->probabilities");
        tree->left->left = tree->left->right = NULL;

        int *new_i_list, is;
        MALLOC(new_i_list, int, instances, "DecisionTreeBuild: new_i_list");

        // Instances for left node
        is = 0;
        if (index) {
            for (int i = 0; i < instances; i++)
                if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->feat) <= tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
        } else {
            for (int i = 0; i < instances; i++)
                if (data[i_list[i]][tree->feat] <= tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
        }

        if (is) {
            // If internally boosted tree update weights
            if (w)
                new_w = AllocateCopyAndUpdateInstanceWeights(parm, w, is, labels, GetMajorityClass(labels, i_list, instances, classes, NULL), new_i_list, train_instances);

            DecisionTreeBuild(parm, tree->left, data, labels, is, new_i_list, features, f_list, tree->score, classes, depth+1, new_w, train_instances, index, split_measure, for_rforest, ert);

            if (w)
                FREE(new_w);
        } else {    // If no instances treat as leaf node
            // Leaf node
            tree->left->cut_point = 0.0;
            tree->left->feat = -1;
            tree->left->leaf = true;
            tree->left->depth = depth;

            // Get majority class
            tree->left->class = GetMajorityClass(labels, i_list, instances, classes, tree->left->probabilities);

            tree->left->left = tree->left->right = NULL;
        }

        // Instances for right node
        is = 0;
        if (index) {
            for (int i = 0; i < instances; i++) {
                if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->feat) > tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
            }
        } else {
            for (int i = 0; i < instances; i++)
                if (data[i_list[i]][tree->feat] > tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
        }
        MALLOC(tree->right, struct dt_node, 1, "DecisionTreeBuild: tree->right");
        MALLOC(tree->right->probabilities, scalar, classes, "DecisionTreeBuild: tree->right->probabilities");
        tree->right->left = tree->right->right = NULL;
        if (is) {
            // If internally boosted tree update weights
            if (w)
                new_w = AllocateCopyAndUpdateInstanceWeights(parm, w, is, labels, GetMajorityClass(labels, i_list, instances, classes, NULL), new_i_list, train_instances);

            DecisionTreeBuild(parm, tree->right, data, labels, is, new_i_list, features, f_list, tree->score, classes, depth+1, new_w, train_instances, index, split_measure, for_rforest, ert);

            if (w)
                FREE(new_w);
        } else {    // If no instances treat as leaf node
            // Leaf node
            tree->right->cut_point = 0.0;
            tree->right->feat = -1;
            tree->right->leaf = true;
            tree->right->depth = depth;

            // Get majority class
            tree->right->class = GetMajorityClass(labels, i_list, instances, classes, tree->right->probabilities);

            tree->right->left = tree->right->right = NULL;
        }

        FREE(new_i_list);
    } else {
        // Leaf node
        tree->cut_point = 0.0;
        tree->feat = -1;
        tree->leaf = true;
        tree->depth = depth;

        // Get majority class
        tree->class = GetMajorityClass(labels, i_list, instances, classes, tree->probabilities);
        tree->alpha = w ? GetAlpha(parm, w, instances, labels, tree->class, i_list) : 1.0;

        tree->left = tree->right = NULL;
    }

    FREE(cutpoint);
}

int GetMajorityClass(int *labels, int *list, int instances, int classes, scalar *freq)
{
    int count[classes], majority;

    ZeroVector(NULL, count, classes);
    for (int i = 0; i < instances; i++)
        count[labels[list[i]]]++;

    majority = MaxVectorIndex(NULL, count, classes);

    if (freq)
        for (int i = 0; i < classes; i++)
            freq[i] = ((scalar) count[i])/instances;

    return majority;
}

void PrintDTree(const t_Parameters *parm, t_DTree tree, FILE *f, bool ibt, t_DTSplitMeasure split_measure)
{
    char *method[] = {"Accuracy ", "CART ────", "C4.5 ────", "ID3 ─────"};

    fprintf(f, "──────────────────── DECISION TREE: %s───────────────────\n", method[split_measure]);
    PrintDTNode(tree, f, ibt);
    fprintf(f,
            "────────────────────────────────────────────────────────────────\n");
}

void PrintDTNode(t_DTree tree, FILE *f, bool ibt)
{
    if (tree) {
        for (int i = 0; i < tree->depth; i++)
            fprintf(f, "%2c   ", '|');
        fprintf(f, "%2d\t", tree->depth);
        if (tree->leaf) {
            fprintf(f, "Class: %2d", tree->class+1);
            if (ibt)
                fprintf(f, "\tAlpha: %6.4lf\n", tree->alpha);
            else
                fprintf(f, "\n");
        } else {
            fprintf(f, "===>\tFeat: %3d (<= %lf)", tree->feat+1, tree->cut_point);
            if (ibt)
                fprintf(f, "\tClass: %2d\tAlpha: %6.4lf\n", tree->class+1, tree->alpha);
            else
                fprintf(f, "\n");
        }
        PrintDTNode(tree->left, f, ibt);
        PrintDTNode(tree->right, f, ibt);
    }
}

void FreeDecisionTree(t_DTree tree)
{
    if (tree->left) {
        FreeDecisionTree(tree->left);
        FREE(tree->left);
    }
    if (tree->right) {
        FreeDecisionTree(tree->right);
        FREE(tree->right);
    }
    if (tree)
        FREE(tree->probabilities);
}

scalar DecisionTreeMargin(t_DTree tree, scalar *pattern, int label, int classes, bool ibt, bool do_not_use_internal_nodes, bool predict_with_probabilities, int *index)
{
    int max_index;
    scalar max_value, margin, output[classes];

    if (ibt)
        IBTClassify(tree, pattern, output, classes, do_not_use_internal_nodes, predict_with_probabilities, index);
    else
        DecisionTreeClassify(tree, pattern, output, classes, index);

    // Get classification.
    max_index = 0;
    max_value = output[0];
    for (int j = 1; j < classes; j++)
        if (output[j] > max_value) {
            max_value = output[j];
            max_index = j;
        }

    // Test if correctly Classified.
    if (label == max_index) {
        // Difference with second one
        margin = max_value;

        // Find second highest
        output[max_index] = -INFINITE;
        max_index = 0;
        max_value = output[0];
        for (int j = 1; j < classes; j++)
            if (output[j] > max_value) {
                max_value = output[j];
                max_index = j;
            }
        margin -= max_value;

    } else
        // Difference with correct one
        margin = output[label]-max_value;

#ifdef AUTOTEST
    if (isnan(margin))
        OnErrorExit(ERR_NUMERICAL, "Margin NaN");
#endif

    return margin;
}

int CountDTNodes(t_DTree tree)
{
    int nodes = 0;

    CountAllNodes(tree, &nodes);

    return nodes;
}

void CountAllNodes(t_DTree tree, int *nodes)
{
    if (tree) {
        (*nodes)++;
        CountAllNodes(tree->left, nodes);
        CountAllNodes(tree->right, nodes);
    }
}
