/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <assert.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include "svm.h"
#include "macros.h"

// TODO: Allow some mismatches in the comparison

// WD kernel: S. Sonnenburg, G. Schweikert, P. Philips, J. Behr and G. Rätsch, "Accurate splice site prediction using support vector machines", BMC Bioinformatics vol. 8 (suppl. 10), S7, 2007.

double WeightedDegreeKernel(const struct svm_node *x, const struct svm_node *y, int degree)
{
    double Kwd = 0.0, Kxx = 0.0, wd;
    int length = 0;
    bool equal = true;

    // Get length
    for (int i = 0; x[i].index != -1 && y[i].index != -1; i++) {
        length++;
        if (x[i].value != y[i].value)
            equal = false;
    }

    // If equal, WD = 1
    if (equal)
        wd = 1.0;
    else {
        // Compute k(x,y) and k(x,x) and k(y,y) for normalization
        for (int delta = 1; delta <= degree ; delta++) {
            double sum = 0.0, sum_xx = 0.0, w_delta;
            for (int l = 0; l < length - delta + 1; l++) {
                equal = true;
                for (int j = 0; j < delta; j++)
                    if (x[l+j].value != y[l+j].value) {
                        equal = false;
                        break;
                    }

                sum += (equal == true);
            }

            sum_xx += length - delta + 1;

            w_delta = degree - delta + 1;

            Kwd += w_delta * sum;
            Kxx += w_delta * sum_xx;
        }

        // Compute normalization of k
        // Being x and y of the same length k(x,x) = k(y,y)
        wd = Kwd/Kxx;
    }

    return wd;
}

// WDS kernel: S. Sonnenburg, G. Schweikert, P. Philips, J. Behr and G. Rätsch, "Accurate splice site prediction using support vector machines", BMC Bioinformatics vol. 8 (suppl. 10), S7, 2007.

double WeightedDegreeKernelWithShifts(const struct svm_node *x, const struct svm_node *y, int degree, double sigma)
{
    double WDSf(const struct svm_node *, const struct svm_node *, int, int, double);

    int length = 0;
    bool equal = true;

    // Get length
    for (int i = 0; x[i].index != -1 && y[i].index != -1; i++) {
        length++;
        if (x[i].value != y[i].value)
            equal = false;
    }

    // If equal, WDS = 1
    if (equal)
        return 1.0;

    double K_xy = WDSf(x, y, degree, length, sigma);
    double K_xx = WDSf(x, x, degree, length, sigma);
    double K_yy = WDSf(y, y, degree, length, sigma);

    return K_xy/sqrt(K_xx * K_yy);
}

double WDSf(const struct svm_node *x, const struct svm_node *y, int degree, int length, double sigma)
{
    int S(int, double);

    double sum = 0.0, Kwd = 0.0;
    bool equal_l, equal_r;

    // Compute k(x,y)
    for (int delta = 1; delta <= degree ; delta++) {
        double w_delta = degree - delta + 1;

        for (int l = 0; l < length - delta + 1; l++) {
            sum = 0.0;

            for (int s = 0; s <= S(l, sigma) && s+l < length; s++) {
                double delta_s;

                delta_s = 1.0 / (2.0 * (s+1.0));

                // Compare x+s with y
                equal_l = equal_r = true;
                for (int j = 0; j < delta && l+j+s < length && (equal_l || equal_r); j++) {
                    if (x[l+j+s].value != y[l+j].value)
                        equal_l = false;

                    if (x[l+j].value != y[l+j+s].value)
                        equal_r = false;
                }

                sum += delta_s * ((equal_l == true) + (equal_r == true));
            }
        }

        Kwd += w_delta * sum;
    }

    return Kwd;
}

int S(int l, double sigma)
{
    return (int)(sigma * fabs((scalar) l));
}

// A. Zien, G. Rätsch, S. Mika, B. Schölkopf, T. Lengauer, and K. R. Müller, “Engineering support vector machine kernels that recognize
// translation initiation sites.” Bioinformatics, vol. 16, no. 9, pp. 799–807, 2000.


double LocalityImprovedKernel(const struct svm_node *x, const struct svm_node *y, int degree, double gamma)
{
    double WinP(const struct svm_node *x, const struct svm_node *y, int p, int l, int d);

    int length = 0, l;
    double lik = 0.0, wp;

    l = (int) gamma;

    // Get length
    for (int i = 0; x[i].index != -1 ; i++)
        length++;

    for (int p = l+1; p < length - l; p++) {
        wp = (p <= length/2) ? p-l : length-p-l+1;
        lik += wp * WinP(x, y, p, l, degree);
    }

    return lik;
}

double WinP(const struct svm_node *x, const struct svm_node *y, int p, int l, int d)
{
    double winp = 0.0;

    for (int j = -l; j < l; j++)
        if (x[p+j].value == y[p+j].value)
            winp++;

    return pow((1./(2.*l+1))*winp, (double) d);
}

// Spectrum kernel implementation
// Degree is the k-mer
// Gamma is the number of different tokens
// Normalized version

// Spectrum kernel: C. Leslie, E. Eskin, and W. Noble, The Spectrum Kernel: A String Kernel for SVM Protein Classification. Pacific Symposium on Biocomputing, 2002.

double SpectrumKernel(const struct svm_node *x, const struct svm_node *y, int k, int n_letters, int hybrid, double gamma)
{
    void InsertOrdered(double *, int *, double, double *, double);
    double dot(const struct svm_node *, const struct svm_node *);

    bool update = false;

    if (x == y)
        return 1.0;

    // Bases for numbers first time of when k changes
    static double *powers = NULL;
    static int old_k = 0;
#ifdef _OPENMP
    #pragma omp critical
#endif
    if (old_k != k) {
        REALLOC(powers, double, k, "SpectrumKernel: powers");
        for (int i = old_k; i < k; i++)
            powers[i] = pow((double) n_letters, (double) i);
        old_k = k;
        update = true;
    }

    // If not cached calculate
    double *list_x = NULL;
    double *count_x = NULL;
    int n_k_mers_x = 0;
    if (x[0].spectrum->n_k_mers == -1 || update) {
        // All k-mers in x sequence
        // Count length
        int length_x = -1;
        while (x[++length_x].index != -1)
            ;

        REALLOC(x[0].spectrum->list, double, length_x-k+1, "SpectrumKernel: x[0].spectrum->list");
        REALLOC(x[0].spectrum->count, double, length_x-k+1, "SpectrumKernel: x[0].spectrum->count");
        list_x = x[0].spectrum->list;
        count_x = x[0].spectrum->count;
        for (int i = 0; i <= length_x-k; i++) {
            double u = 0;
            for (int j = 0; j < k; j++)
                u += x[i+j].value * powers[j];

            // Check if new
            bool new = true;
            for (int j = 0; j < n_k_mers_x && list_x[j] <= u; j++)
                if (list_x[j] == u) {
                    new = false;
                    count_x[j] += 1.0;
                    break;
                }

            if (new)
                InsertOrdered(list_x, &n_k_mers_x, u, count_x, 1.0);
        }
        x[0].spectrum->n_k_mers = n_k_mers_x;

        // k_xx for normalization
        x[0].spectrum->k_xx = 0.0;
        for (int i = 0; i < n_k_mers_x; i++)
            x[0].spectrum->k_xx += count_x[i] * count_x[i];
        x[0].spectrum->k_xx = sqrt(x[0].spectrum->k_xx);
    } else {
        list_x = x[0].spectrum->list;
        count_x = x[0].spectrum->count;
        n_k_mers_x = x[0].spectrum->n_k_mers;
    }

    // If not cached calculate
    double *list_y = NULL;
    double *count_y = NULL;
    int n_k_mers_y = 0;
    if (y[0].spectrum->n_k_mers == -1 || update) {
        // All k-mers in y sequence
        int length_y = -1;
        while (y[++length_y].index != -1)
            ;

        MALLOC(y[0].spectrum->list, double, length_y-k+1, "SpectrumKernel: y[0].spectrum->list");
        MALLOC(y[0].spectrum->count, double, length_y-k+1, "SpectrumKernel: y[0].spectrum->count");
        list_y = y[0].spectrum->list;
        count_y = y[0].spectrum->count;
        for (int i = 0; i <= length_y-k; i++) {
            double u = 0;
            for (int j = 0; j < k; j++)
                u += y[i+j].value * powers[j];

            // Check if new
            bool new = true;
            for (int j = 0; j < n_k_mers_y && list_y[j] <= u; j++)
                if (list_y[j] == u) {
                    new = false;
                    count_y[j] += 1.0;
                    break;
                }

            if (new)
                InsertOrdered(list_y, &n_k_mers_y, u, count_y, 1.0);
        }
        y[0].spectrum->n_k_mers = n_k_mers_y;

        // k_xx for normalization
        y[0].spectrum->k_xx = 0.0;
        for (int i = 0; i < n_k_mers_y; i++)
            y[0].spectrum->k_xx += count_y[i] * count_y[i];
        y[0].spectrum->k_xx = sqrt(y[0].spectrum->k_xx);
    } else {
        list_y = y[0].spectrum->list;
        count_y = y[0].spectrum->count;
        n_k_mers_y = y[0].spectrum->n_k_mers;
    }

    // k-mers in both sequences
    double k_xy = 0.0;
    int index_x = 0, index_y = 0;
    do {
        while (index_x < n_k_mers_x && list_x[index_x] < list_y[index_y])
            index_x++;
        if (index_x == n_k_mers_x)
            break;
        while (index_y < n_k_mers_y && list_x[index_x] > list_y[index_y])
            index_y++;
        if (index_y == n_k_mers_y)
            break;
        if (list_x[index_x] == list_y[index_y]) {
            k_xy += count_x[index_x]*count_y[index_y];
            index_x++;
            index_y++;
        }
    } while (index_x < n_k_mers_x && index_y < n_k_mers_y);

    // Normalized version
    k_xy /= x[0].spectrum->k_xx*y[0].spectrum->k_xx;

    if (hybrid == 0)
        return k_xy;

    // Linear or Gaussian kernels
    double *converted;
    int c, index;
    struct svm_node *x_freq, *y_freq;

    MALLOC(converted, double, n_letters, "SpectrumKernel: converted");

    // Convert each sequence to the count of tokens
    for (int j = 0; j < n_letters; j++)
        converted[j] = 0.0;
    index = 0;
    while (x[index].index != -1) {
        converted[(int) x[index].value]++;
        index++;
    }

    c = 0;
    for (int j = 0; j < n_letters; j++)
        if (converted[j] > 0.0)
            c++;

    MALLOC(x_freq, struct svm_node, c+1, "SpectrumKernel: x_freq");
    c = 0;
    for (int j = 0; j < n_letters; j++)
        if (converted[j] > 0.0) {
            x_freq[c].index = j;
            x_freq[c].value = converted[j]/n_k_mers_x;
            c++;
        }
    x_freq[c].index = -1;

    for (int j = 0; j < n_letters; j++)
        converted[j] = 0.0;
    index = 0;
    while (y[index].index != -1) {
        converted[(int) y[index].value]++;
        index++;
    }

    c = 0;
    for (int j = 0; j < n_letters; j++)
        if (converted[j] > 0.0)
            c++;

    MALLOC(y_freq, struct svm_node, c+1, "SpectrumKernel: y_freq");
    c = 0;
    for (int j = 0; j < n_letters; j++)
        if (converted[j] > 0.0) {
            y_freq[c].index = j;
            y_freq[c].value = converted[j]/n_k_mers_y;
            c++;
        }
    y_freq[c].index = -1;

    FREE(converted);

    if (hybrid == 1 || hybrid == 2) {
        // Linear
        double linear = dot(x_freq, y_freq)/sqrt(dot(x_freq, x_freq)*dot(y_freq, y_freq));

        FREE(x_freq);
        FREE(y_freq);

        return (hybrid == 1) ? k_xy * linear : k_xy + linear;
    } else {
        // RBF kernel
        double x_sq_x = dot(x_freq, x_freq);
        double x_sq_y = dot(y_freq, y_freq);

        double gaussian =  exp(-gamma*(x_sq_x+x_sq_y-2.0*dot(x_freq,y_freq)));

        FREE(x_freq);
        FREE(y_freq);

        // Hybrid kernel is the product of both kernels
        return (hybrid == 3) ? k_xy * gaussian : k_xy + gaussian;
    }

    return 0.0;
}

// Insert a new element in an ordered list keeping the order and do the same in the count list
// The element is not yet present in the list
void InsertOrdered(double *v, int *n, double value, double *count, double init_count)
{
    int index = 0;

    while (index < *n && v[index] < value)
        index++;

    // Shift right
    memmove(v+index+1, v+index, sizeof(double) * (*n-index));
    memmove(count+index+1, count+index, sizeof(double) * (*n-index));

    v[index] = value;
    count[index] = init_count;
    (*n)++;
}

// Hybrid kernels: Instance must be in two part: sequence + real inputs
// The sequence is [0, offset-1] and the real part is [offset, end]

// Hybrid kernels:
// 0    WD x LINEAR
// 1    WD + LINEAR (normalized)

double HybridKernel(struct svm_node *x, struct svm_node *y, int ktype, int offset, int degree, double gamma)
{
    double WeightedDegreeTimesLinearHybridKernel(struct svm_node *, struct svm_node *, int, int, double);
    double WeightedDegreePlusLinearHybridKernel(struct svm_node *, struct svm_node *, int, int, double);
    double WeightedDegreeTimesGaussianHybridKernel(struct svm_node *, struct svm_node *, int, int, double);

    // Function pointer for each type of hybrid kernel
    static double (*kernels[])(struct svm_node *, struct svm_node *, int, int, double) = {
        WeightedDegreeTimesLinearHybridKernel,
        WeightedDegreePlusLinearHybridKernel,
        WeightedDegreeTimesGaussianHybridKernel
    };

    return (*kernels[ktype])(x, y, offset, degree, gamma);
}

double WeightedDegreeTimesLinearHybridKernel(struct svm_node *x, struct svm_node *y, int offset, int degree, double dummy)
{
    double dot(const struct svm_node *, const struct svm_node *);

    double wd, linear;

    // Temporally cut sequence
    x[offset].index = y[offset].index = -1;

    // Construct the instance for WD kernel
    wd = WeightedDegreeKernel(x, y, degree);

    x[offset].index = y[offset].index = offset;

    // Construct the instance for the linear kernel
    linear = dot(x+offset, y+offset);

    // Hybrid kernel is the product of both kernels
    return wd * linear;
}

double dot(const struct svm_node *px, const struct svm_node *py)
{
    double sum = 0;
    while (px->index != -1 && py->index != -1) {
        if (px->index == py->index) {
            sum += px->value * py->value;
            ++px;
            ++py;
        } else {
            if (px->index > py->index)
                ++py;
            else
                ++px;
        }
    }
    return sum;
}

double WeightedDegreePlusLinearHybridKernel(struct svm_node *x, struct svm_node *y, int offset, int degree, double dummy)
{
    double dot(const struct svm_node *, const struct svm_node *);

    double wd, linear;

    // Temporally cut sequence
    x[offset].index = y[offset].index = -1;

    // Construct the instance for WD kernel
    wd = WeightedDegreeKernel(x, y, degree);

    x[offset].index = y[offset].index = offset;

    // Construct the instance for the linear kernel
    linear = dot(x+offset, y+offset);

    // Normalize linear in [0:1]. Assume inputs in [0:1]
    int size = 0;
    while (x[size+offset].index != -1)
        size++;

    // Hybrid kernel is the product of both kernels
    return wd + (linear/size);
}

double WeightedDegreeTimesGaussianHybridKernel(struct svm_node *x, struct svm_node *y, int offset, int degree, double gamma)
{
    double dot(const struct svm_node *, const struct svm_node *);

    double wd, gaussian, x_sq_x, x_sq_y;

    // Temporally cut sequence
    x[offset].index = y[offset].index = -1;

    // Construct the instance for WD kernel
    wd = WeightedDegreeKernel(x, y, degree);

    x[offset].index = y[offset].index = offset;

    // RBF kernel
    x_sq_x = dot(x+offset, x+offset);
    x_sq_y = dot(y+offset, y+offset);

    gaussian =  exp(-gamma*(x_sq_x+x_sq_y-2.0*dot(x+offset,y+offset)));

    // Hybrid kernel is the product of both kernels
    return wd * gaussian;
}

// Spectrum kernel implementation allowing for miss-matches
// Degree is the k-mer
// Gamma is the number of different tokens
// Offset is the allowed distance: offset = 0 is equivalent to the Spectrum kernel

// Normalized version

// C. Leslie, E. Eskin, J. Weston and W. Noble, Mismatch String Kernels for SVM Protein Classification. Neural Information Processing Systems 2002.

double MismatchKernel(const struct svm_node *x, const struct svm_node *y, int k, int n_letters, int m)
{
    void InsertOrdered(double *, int *, double, double *, double);
    bool InNeighborhood(double, double, int, int);

    bool update = false;

    if (x == y)
        return 1.0;

    // Bases for numbers first time of when k changes
    static double *powers = NULL;
    static int old_k = 0;
#ifdef _OPENMP
    #pragma omp critical
#endif
    if (old_k != k) {
        REALLOC(powers, double, k, "MismatchKernel: powers");
        for (int i = old_k; i < k; i++)
            powers[i] = pow((double) n_letters, (double) i);
        old_k = k;
        update = true;
    }

    static int old_m = 0;
    if (old_m != m) {
        old_m = m;
        update = true;
    }

    // If not cached calculate
    double *list_x = NULL;
    double *count_x = NULL;
    int n_k_mers_x = 0;
    if (x[0].spectrum->n_k_mers == -1 || update) {
        // All k-mers in x sequence
        // Count length
        int length_x = -1;
        while (x[++length_x].index != -1)
            ;

        REALLOC(x[0].spectrum->list, double, length_x-k+1, "MismatchKernel: x[0].spectrum->list");
        REALLOC(x[0].spectrum->count, double, length_x-k+1, "MismatchKernel: x[0].spectrum->count");
        list_x = x[0].spectrum->list;
        count_x = x[0].spectrum->count;
        for (int i = 0; i <= length_x-k; i++) {
            double u = 0;
            for (int j = 0; j < k; j++)
                u += x[i+j].value * powers[j];

            // Check if new
            bool new = true;
            for (int j = 0; j < n_k_mers_x && list_x[j] <= u; j++)
                if (list_x[j] == u) {
                    new = false;
                    count_x[j] += 1.0;
                    break;
                }

            if (new)
                InsertOrdered(list_x, &n_k_mers_x, u, count_x, 1.0);
        }
        x[0].spectrum->n_k_mers = n_k_mers_x;

        // Add mismatches within given neighborhood
        double *add_to;
        CALLOC(add_to, double, n_k_mers_x, "MismatchKernel: add_to");
        for (int i = 0; i < n_k_mers_x; i++)
            for (int j = i+1; j < n_k_mers_x; j++)
                if (InNeighborhood(list_x[i], list_x[j], n_letters, m)) {
                    add_to[i] += count_x[j];
                    add_to[j] += count_x[i];
                }
        for (int i = 0; i < n_k_mers_x; i++)
            count_x[i] += add_to[i];

        FREE(add_to);

        // k_xx for normalization
        x[0].spectrum->k_xx = 0.0;
        for (int i = 0; i < n_k_mers_x; i++)
            x[0].spectrum->k_xx += count_x[i] * count_x[i];
        x[0].spectrum->k_xx = sqrt(x[0].spectrum->k_xx);
    } else {
        list_x = x[0].spectrum->list;
        count_x = x[0].spectrum->count;
        n_k_mers_x = x[0].spectrum->n_k_mers;
    }

    // If not cached calculate
    double *list_y = NULL;
    double *count_y = NULL;
    int n_k_mers_y = 0;
    if (y[0].spectrum->n_k_mers == -1 || update) {
        // All k-mers in y sequence
        int length_y = -1;
        while (y[++length_y].index != -1)
            ;

        MALLOC(y[0].spectrum->list, double, length_y-k+1, "MismatchKernel: y[0].spectrum->list");
        MALLOC(y[0].spectrum->count, double, length_y-k+1, "MismatchKernel: y[0].spectrum->count");
        list_y = y[0].spectrum->list;
        count_y = y[0].spectrum->count;
        for (int i = 0; i <= length_y-k; i++) {
            double u = 0;
            for (int j = 0; j < k; j++)
                u += y[i+j].value * powers[j];

            // Check if new
            bool new = true;
            for (int j = 0; j < n_k_mers_y && list_y[j] <= u; j++)
                if (list_y[j] == u) {
                    new = false;
                    count_y[j] += 1.0;
                    break;
                }

            if (new)
                InsertOrdered(list_y, &n_k_mers_y, u, count_y, 1.0);
        }
        y[0].spectrum->n_k_mers = n_k_mers_y;

        // Add mismatches within given neighborhood
        double *add_to;
        CALLOC(add_to, double, n_k_mers_y, "MismatchKernel: add_to");
        for (int i = 0; i < n_k_mers_y; i++)
            for (int j = i+1; j < n_k_mers_y; j++)
                if (InNeighborhood(list_y[i], list_y[j], n_letters, m)) {
                    add_to[i] += count_y[j];
                    add_to[j] += count_y[i];
                }
        for (int i = 0; i < n_k_mers_y; i++)
            count_y[i] += add_to[i];

        FREE(add_to);

        // k_xx for normalization
        y[0].spectrum->k_xx = 0.0;
        for (int i = 0; i < n_k_mers_y; i++)
            y[0].spectrum->k_xx += count_y[i] * count_y[i];
        y[0].spectrum->k_xx = sqrt(y[0].spectrum->k_xx);
    } else {
        list_y = y[0].spectrum->list;
        count_y = y[0].spectrum->count;
        n_k_mers_y = y[0].spectrum->n_k_mers;
    }

    // k-mers in both sequences
    double k_xy = 0.0;
    int index_x = 0, index_y = 0;
    do {
        while (index_x < n_k_mers_x && list_x[index_x] < list_y[index_y])
            index_x++;
        if (index_x == n_k_mers_x)
            break;
        while (index_y < n_k_mers_y && list_x[index_x] > list_y[index_y])
            index_y++;
        if (index_y == n_k_mers_y)
            break;
        if (list_x[index_x] == list_y[index_y]) {
            k_xy += count_x[index_x]*count_y[index_y];
            index_x++;
            index_y++;
        }
    } while (index_x < n_k_mers_x && index_y < n_k_mers_y);

    // Normalized version
    k_xy /= x[0].spectrum->k_xx*y[0].spectrum->k_xx;

    return k_xy;
}

bool InNeighborhood(double u, double v, int base, int n_mismatches)
{
    int gap = 0;

    // Get digits
    while (u > 0.0 || v > 0.0) {
        if ((int) fmod(u, (double) base) != (int) fmod(v, (double) base))
            gap++;

        if (gap > n_mismatches)
            return false;

        u = floor(u/base);
        v = floor(v/base);
    }

    return true;
}

// Based on Smith - Waterman algorithm for genes or amino acids sequences

// #define STATIC_MEMORY
#define MEMORY_LIMIT    10       // Memory limit in GB
#define LENGTH_LIMIT    50000
#define ONEGB           1073741824.0

double SmithWatermanKernel(const struct svm_node *x, const struct svm_node *y, double mu, double delta, int use_minimum)
{
    double FindArrayMax(double *, int);
#ifdef STATIC_MEMORY
    static int pre_N_a = -1, pre_N_b = 0;
    static float **H = NULL;
#else
    float **H;
#endif

    if (x == y)
        return 1.0;

    // Count length
    int N_a = -1, N_b = -1;

    while (x[++N_a].index != -1);
    while (y[++N_b].index != -1);

#ifdef STATIC_MEMORY
    // Check memory limit is not reached, in such a case limit sequences
    double size = (MAX(N_a+1, pre_N_a+1)*sizeof(float *) + ((double) MAX(N_a+1, pre_N_a+1))*MAX(N_b+1, pre_N_b)*sizeof(float))/ONEGB;
    if (size > MEMORY_LIMIT) {
        N_a = MIN(N_a, LENGTH_LIMIT);
        N_b = MIN(N_b, LENGTH_LIMIT);
    }
    // Reallocate H if necessary
    if (N_a > pre_N_a) {
        REALLOC(H, float *, N_a+1, "SmithWatermanKernel: H");
        for (int i = pre_N_a+1; i < N_a+1; i++)
            MALLOC(H[i], float, N_b+1, "SmithWatermanKernel: H[i]");
    }
    if (N_b > pre_N_b)
        for (int i = 0; i < pre_N_a+1; i++)
            REALLOC(H[i], float, N_b+1, "SmithWatermanKernel: H[i]");

    pre_N_a = N_a;
    pre_N_b = N_b;
#else
    // Check memory limit is not reached, in such a case limit sequences
    double size = ((N_a+1)*sizeof(float *) + ((double)(N_a+1))*(N_b+1)*sizeof(float))/ONEGB;
    if (size > MEMORY_LIMIT) {
        N_a = MIN(N_a, LENGTH_LIMIT);
        N_b = MIN(N_b, LENGTH_LIMIT);
    }
    MALLOC(H, float *, N_a+1, "SmithWatermanKernel: H");
    for (int i = 0; i < N_a+1; i++)
        CALLOC(H[i], float, N_b+1, "SmithWatermanKernel: H[i]");
#endif

    // Initialize H
    for (int i=0; i <= N_a; i++)
        for (int j = 0; j <= N_b; j++)
            H[i][j]=0.;

    // Here comes the actual algorithm
    double temp[4];
    for (int i = 1; i <= N_a; i++) {
        for (int j = 1; j <= N_b; j++) {
            temp[0] = H[i-1][j-1] + (x[i-1].value == y[j-1].value ? 1.0 : -mu);
            temp[1] = H[i-1][j]-delta;
            temp[2] = H[i][j-1]-delta;
            temp[3] = 0.0;
            H[i][j] = FindArrayMax(temp, 4);
        }
    }

    // Search H for the maximal score
    double H_max = 0.;
    for (int i = 1; i <= N_a; i++)
        for (int j = 1; j <= N_b; j++)
            if (H[i][j] > H_max)
                H_max = H[i][j];

#ifndef STATIC_MEMORY
    for (int i = 0; i < N_a+1; i++)
        FREE(H[i]);
    FREE(H);
#endif

    if (use_minimum)
        return H_max/MIN(N_a, N_b);
    else
        return H_max/(sqrt(N_a)*sqrt(N_b));
}

double FindArrayMax(double *array, int length)
{

    double max = array[0];            // start with max = first element

    for (int i = 1; i < length; i++)
        if (array[i] > max)
            max = array[i];

    return max;                    // return highest value in array
}


// Wildcard kernel
// Spectrum kernel implementation allowing for wildcard
// C. Leslie and R. Kuang, "Fast String Kernels using Inexact Matching for Protein Sequences", Journal of Machine Learning Research 5 (2004) 1435–1455.

double WildcardKernel(const struct svm_node *x, const struct svm_node *y, int k, int n_letters, int m, double lambda)
{
    void InsertOrdered(double *, int *, double, double *, double);
    bool InNeighborhood(double, double, int, int);
    void AddWildCardPatterns(double **, double **, int *, int, int, int, double, double *);

    bool update = false;

    if (x == y)
        return 1.0;

    // Bases for numbers first time of when k changes
    static double *powers = NULL;
    static int old_k = 0;
#ifdef _OPENMP
    #pragma omp critical
#endif
    if (old_k != k) {
        REALLOC(powers, double, k, "MismatchKernel: powers");
        for (int i = old_k; i < k; i++)
            powers[i] = pow((double) n_letters+1, (double) i);  // The wildcard increments the base +1
        old_k = k;
        update = true;
    }

    static int old_m = 0;
    if (old_m != m) {
        old_m = m;
        update = true;
    }

    // If not cached calculate
    double *list_x = NULL;
    double *count_x = NULL;
    int n_k_mers_x = 0;
    if (x[0].spectrum->n_k_mers == -1 || update) {
        // All k-mers in x sequence
        // Count length
        int length_x = -1;
        while (x[++length_x].index != -1)
            ;

        REALLOC(x[0].spectrum->list, double, length_x-k+1, "MismatchKernel: x[0].spectrum->list");
        REALLOC(x[0].spectrum->count, double, length_x-k+1, "MismatchKernel: x[0].spectrum->count");
        list_x = x[0].spectrum->list;
        count_x = x[0].spectrum->count;
        for (int i = 0; i <= length_x-k; i++) {
            double u = 0;
            for (int j = 0; j < k; j++)
                u += x[i+j].value * powers[j];

            // Check if new
            bool new = true;
            for (int j = 0; j < n_k_mers_x && list_x[j] <= u; j++)
                if (list_x[j] == u) {
                    new = false;
                    count_x[j] += 1.0;
                    break;
                }

            if (new)
                InsertOrdered(list_x, &n_k_mers_x, u, count_x, 1.0);
        }
        x[0].spectrum->n_k_mers = n_k_mers_x;

        // Add wildcard patterns with at most m "*"s
        AddWildCardPatterns(&list_x, &count_x, &n_k_mers_x, k, n_letters, m, lambda, powers);
        x[0].spectrum->n_k_mers = n_k_mers_x;
        x[0].spectrum->list = list_x;
        x[0].spectrum->count = count_x;

        // k_xx for normalization
        x[0].spectrum->k_xx = 0.0;
        for (int i = 0; i < n_k_mers_x; i++)
            x[0].spectrum->k_xx += count_x[i] * count_x[i];
        x[0].spectrum->k_xx = sqrt(x[0].spectrum->k_xx);

    } else {
        list_x = x[0].spectrum->list;
        count_x = x[0].spectrum->count;
        n_k_mers_x = x[0].spectrum->n_k_mers;
    }

    // If not cached calculate
    double *list_y = NULL;
    double *count_y = NULL;
    int n_k_mers_y = 0;
    if (y[0].spectrum->n_k_mers == -1 || update) {
        // All k-mers in y sequence
        int length_y = -1;
        while (y[++length_y].index != -1)
            ;

        MALLOC(y[0].spectrum->list, double, length_y-k+1, "MismatchKernel: y[0].spectrum->list");
        MALLOC(y[0].spectrum->count, double, length_y-k+1, "MismatchKernel: y[0].spectrum->count");
        list_y = y[0].spectrum->list;
        count_y = y[0].spectrum->count;
        for (int i = 0; i <= length_y-k; i++) {
            double u = 0;
            for (int j = 0; j < k; j++)
                u += y[i+j].value * powers[j];

            // Check if new
            bool new = true;
            for (int j = 0; j < n_k_mers_y && list_y[j] <= u; j++)
                if (list_y[j] == u) {
                    new = false;
                    count_y[j] += 1.0;
                    break;
                }

            if (new)
                InsertOrdered(list_y, &n_k_mers_y, u, count_y, 1.0);
        }
        y[0].spectrum->n_k_mers = n_k_mers_y;

        // Add wildcard patterns with at most m "*"s
        AddWildCardPatterns(&list_y, &count_y, &n_k_mers_y, k, n_letters, m, lambda, powers);
        y[0].spectrum->n_k_mers = n_k_mers_y;
        y[0].spectrum->list = list_y;
        y[0].spectrum->count = count_y;

        // k_xx for normalization
        y[0].spectrum->k_xx = 0.0;
        for (int i = 0; i < n_k_mers_y; i++)
            y[0].spectrum->k_xx += count_y[i] * count_y[i];
        y[0].spectrum->k_xx = sqrt(y[0].spectrum->k_xx);
    } else {
        list_y = y[0].spectrum->list;
        count_y = y[0].spectrum->count;
        n_k_mers_y = y[0].spectrum->n_k_mers;
    }

    // k-mers in both sequences
    double k_xy = 0.0;
    int index_x = 0, index_y = 0;
    do {
        while (index_x < n_k_mers_x && list_x[index_x] < list_y[index_y])
            index_x++;
        if (index_x == n_k_mers_x)
            break;
        while (index_y < n_k_mers_y && list_x[index_x] > list_y[index_y])
            index_y++;
        if (index_y == n_k_mers_y)
            break;
        if (list_x[index_x] == list_y[index_y]) {
            k_xy += count_x[index_x]*count_y[index_y];
            index_x++;
            index_y++;
        }
    } while (index_x < n_k_mers_x && index_y < n_k_mers_y);

    // Normalized version
    k_xy /= x[0].spectrum->k_xx*y[0].spectrum->k_xx;

    return k_xy;
}

// Add new segments containing wildcards from existing ones, otherwise there will be no occurrences
void AddWildCardPatterns(double **list_x, double **count_x, int *n_k_mers, int k, int base, int m, double lambda, double *powers)
{
    void InsertOrdered(double *, int *, double, double *, double);
    bool MatchWithWildcard(int [], int [], double, int, int);

    int digits[k], added = 0;
    double *list = NULL, *count = NULL;

    // For every found k-mer
    for (int i1 = 0; i1 < *n_k_mers; i1++) {
        // Get digits
        double u = (*list_x)[i1];
        for (int i2 = 0; i2 < k; i2++) {
            digits[i2] = (int) fmod(u, (double) base);
            u = floor(u/base);
        }

        // Generate new k-mers with wildcard
        // Generate all possible binary numbers with k digits, count numbers of 1's, if less of equal than m construct the corresponding pattern substituting
        // the ones with a wildcard
        for (int i2 = 1; i2 < (int) pow(2.0, k); i2++) {
            int kept[k], mask[k], cc = 0;

            int number = i2;
            for (int i3 = 0; i3 < k; i3++) {
                mask[i3] = number % 2;
                number /= 2;
                if (mask[i3])
                    cc++;
            }

            // If more 1's than m next
            if (cc > m)
                continue;

            // Save pattern with no wild cards
            memcpy(kept, digits, k*sizeof(int));

            // Substitute 1's by wild cards
            for (int i3 = 0; i3 < k; i3++)
                if (mask[i3])
                    digits[i3] = base;

            // Convert to number for the list
            u = 0.0;
            for (int i3 = 0; i3 < k; i3++)
                u += digits[i3] * powers[i3];

            bool new = true;
            for (int i3 = 0; i3 < added && new; i3++)
                if (list[i3] == u)
                    new = false;

            // Add if new
            if (new) {
                added++;

                REALLOC(list, double, added, "AddWildCardPatterns: list");
                REALLOC(count, double, added, "AddWildCardPatterns: count");

                list[added-1] = u;
                count[added-1] = 0;

                // Search for matching sequences with the wildcard
                for (int i3 = 0; i3 < *n_k_mers; i3++)
                    if (MatchWithWildcard(digits, mask, (*list_x)[i3], k, base))
                        count[added-1] += (*count_x)[i3] * pow(lambda, cc);
            }

            // Restore
            memcpy(digits, kept, k*sizeof(int));
        }
    }

    // Add new k-mers
    REALLOC(*list_x, double, *n_k_mers+added, "AddWildCardPatterns: *list_x");
    REALLOC(*count_x, double, *n_k_mers+added, "AddWildCardPatterns: *count_x");
    for (int i = 0; i < added; i++)
        InsertOrdered(*list_x, n_k_mers, list[i], *count_x, count[i]);

    FREE(count);
    FREE(list);
}

bool MatchWithWildcard(int digits[], int mask[], double u, int k, int base)
{
    int u_digits[k];

    // Convert u to digits
    for (int i = 0; i < k; i++) {
        u_digits[i] = (int) fmod(u, (double) base);
        u = floor(u/base);
    }

    // Check matching
    for (int i = 0; i < k; i++)
        if (!mask[i] && digits[i] != u_digits[i])
            return false;

    return true;
}
