/***************************************************************************
*   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#include <stdio.h>
#include <math.h>
#include "common.h"
#include "macros.h"
#include "classifiers.h"

/* Accuracy measures for multiclass datasets from the confusion matrix

Confusion matrix must be:

                Predicted
           -------------------
           |
Actual     |
           |


Measures in: Marina Sokolova and Guy Lapalme, "Systematic analysis of performance measures for classification tasks", Information Processing and Management 45 (2009) 427–437
*/


void PrintMulticlassMeasures(int **M, int classes, int instances, FILE *out_file, char *text, t_PerformanceInfo *perf)
{
    scalar tp[classes], tn[classes], fp[classes], fn[classes], TP = 0.0, theta_1 = 0.0, theta_2 = 0.0;

    // Get TP, TN, FP and FN for every class
    for (int i = 0; i < classes; i++) {
        TP += (tp[i] = M[i][i]);

        fp[i] = fn[i] = -M[i][i];
        for (int j = 0; j < classes; j++) {
            fn[i] += M[i][j];
            fp[i] += M[j][i];
        }

        tn[i] = instances - tp[i] - fp[i] - fn[i];
    }

    // Accuracy
    perf->accuracy = TP/instances;

    // Average perf->accuracy
    perf->average_accuracy = 0.0;
    for (int i = 0; i < classes; i++)
        perf->average_accuracy += (tp[i]+tn[i])/instances;
    perf->average_accuracy /= classes;

    // Error error
    perf->error_rate = 0.0;
    for (int i = 0; i < classes; i++)
        perf->error_rate += (fp[i]+fn[i])/instances;
    perf->error_rate /= classes;

    // Micro-averaged precision
    perf->precision_micro = 0.0;
    for (int i = 0; i < classes; i++)
        perf->precision_micro += tp[i]+fp[i];
    perf->precision_micro = TP/perf->precision_micro;

    // Micro-averaged recall
    perf->recall_micro = 0.0;
    for (int i = 0; i < classes; i++)
        perf->recall_micro += tp[i]+fn[i];
    perf->recall_micro = TP/perf->recall_micro;

    // Micro-averaged F1-score
    perf->f1_score_micro = (2.0*perf->precision_micro*perf->recall_micro)/(perf->precision_micro+perf->recall_micro);

    // Macro-averaged precision
    perf->precision_macro = 0.0;
    for (int i = 0; i < classes; i++)
        perf->precision_macro += (tp[i]+fp[i] > 0) ? tp[i]/(tp[i]+fp[i]) : 0.0;
    perf->precision_macro /= classes;

    // Macro-averaged recall
    perf->recall_macro = 0.0;
    perf->recall_macro_g_mean = 1.0;
    for (int i = 0; i < classes; i++) {
        perf->recall_macro += (tp[i]+fn[i] > 0) ? tp[i]/(tp[i]+fn[i]) : 0.0;
        perf->recall_macro_g_mean *= (tp[i]+fn[i] > 0) ? tp[i]/(tp[i]+fn[i]) : 1.0;
    }
    perf->recall_macro /= classes;
    perf->recall_macro_g_mean = pow(perf->recall_macro_g_mean, 1.0/classes);

    // Macro-averaged F1-score
    perf->f1_score_macro = (2.0*perf->precision_macro*perf->recall_macro)/(perf->precision_macro+perf->recall_macro);

    // Kappa
    // Sum of the diagonal
    for (int j = 0; j < classes; j++)
        theta_1 += M[j][j];

    if (theta_1 == instances)
        perf->kappa = 1.0;
    else {
        theta_1 *= instances;

        for (int i = 0; i < classes; i++) {
            // Rows/columns total counts
            scalar row_i = 0, column_i = 0;
            for (int j = 0; j < classes; j++) {
                row_i += M[i][j];
                column_i += M[j][i];
            }

            theta_2 += row_i * column_i;
        }

        perf->kappa = (theta_1 - theta_2)/(((scalar) instances)*instances-theta_2);
    }

    // Multi-class Matthews Correlation coefficient
    // Jurman G, Riccadonna S, Furlanello C (2012) A Comparison of MCC and CEN Error Measures in Multi-Class Prediction. PLoS ONE 7(8): e41882. doi:10.1371/journal.pone.0041882
    scalar cov_xy = 0.0;
    for (int k = 0; k < classes; k++)
        for (int l = 0; l < classes; l++)
            for (int m = 0; m < classes; m++)
                cov_xy += M[k][k]*M[m][l] - M[l][k]*M[k][m];

    scalar cov_xx = 0.0, cov_yy = 0.0;
    for (int k = 0; k < classes; k++) {
        scalar sum_xx_l = 0.0, sum_yy_l = 0.0;
        for (int l = 0; l < classes; l++) {
            sum_xx_l += M[l][k];
            sum_yy_l += M[k][l];
        }
        scalar sum_xx_fg = 0.0, sum_yy_fg = 0.0;
        for (int f = 0; f < classes; f++)
            if (f != k)
                for (int g = 0; g < classes; g++) {
                    sum_xx_fg += M[g][f];
                    sum_yy_fg += M[f][g];
                }
        cov_xx += sum_xx_l*sum_xx_fg;
        cov_yy += sum_yy_l*sum_yy_fg;
    }
    cov_xx = sqrt(cov_xx);
    cov_yy = sqrt(cov_yy);

    perf->mcc = cov_xx*cov_yy > ZERO ? cov_xy/(cov_xx*cov_yy) : -1.0;

    // Confusion entropy (CEN)
    perf->cen = ConfusionEntropy(M, classes);

    if (text && out_file) {
        fprintf(out_file, "@%s Accuracy: %lf\n", text, perf->accuracy);
        fprintf(out_file, "@%s Average accuracy: %lf\n", text, perf->average_accuracy);
        fprintf(out_file, "@%s Error rate (-): %lf\n", text, perf->error_rate);
        fprintf(out_file, "@%s Precision_\u03BC: %lf\n", text, perf->precision_micro);
        fprintf(out_file, "@%s Recall_\u03BC: %lf\n", text, perf->recall_micro);
        fprintf(out_file, "@%s F1-score_\u03BC: %lf\n", text, perf->f1_score_micro);
        fprintf(out_file, "@%s Precision_M: %lf\n", text, perf->precision_macro);
        fprintf(out_file, "@%s Recall_M: %lf\n", text, perf->recall_macro);
        fprintf(out_file, "@%s Recall_M G-mean: %lf\n", text, perf->recall_macro_g_mean);
        fprintf(out_file, "@%s F1-score_M: %lf\n", text, perf->f1_score_macro);
        fprintf(out_file, "@%s Kappa: %lf\n", text, perf->kappa);
        fprintf(out_file, "@%s MCC: %lf\n", text, perf->mcc);
        fprintf(out_file, "@%s CEN (-): %lf\n", text, perf->cen);

        // Print confusion matrix
        fprintf(out_file, "%s contingency table\n", text);
        for (int i = 0; i < 8+(WIDTH(instances)+3)*classes; i++)
            fprintf(out_file, "-");
        fprintf(out_file, "\n");
        fprintf(out_file, "       | Predicted\n");
        fprintf(out_file, "Actual |");
        for (int i = 0; i < classes; i++)
            fprintf(out_file, "   %*d", WIDTH(instances), i+1);
        fprintf(out_file, "\n");
        for (int i = 0; i < 8+(WIDTH(instances)+3)*classes; i++)
            fprintf(out_file, "-");
        fprintf(out_file, "\n");
        for (int i = 0; i < classes; i++) {
            fprintf(out_file, "%3d    |", i+1);
            for (int j = 0; j < classes; j++)
                fprintf(out_file, "   %*d", WIDTH(instances), (int) M[i][j]);
            fprintf(out_file, "\n");
        }
        for (int i = 0; i < 8+(WIDTH(instances)+3)*classes; i++)
            fprintf(out_file, "-");
        fprintf(out_file, "\n");
        fflush(out_file);
    }
}

// Confusion entropy (CEN)
// Wei JM, Yuan XJ, Hu QH, Wang SQ (2010) A novel measure for evaluating classifiers. Expert Systems with Applications 37: 3799–3809.


scalar ConfusionEntropy(int **M, int classes)
{
    scalar cen = 0.0, *P, /***Pi,*/ **Pj;

    //     AllocateScalarMatrix(&Pi, classes, classes);
    AllocateScalarMatrix(&Pj, classes, classes);
    MALLOC(P, scalar, classes, "ConfusionEntropy: P");

    // P
    for (int j = 0; j < classes; j++) {
        scalar sum_k = 0.0, sum_kl = 0.0;
        for (int k = 0; k < classes; k++) {
            sum_k += M[j][k]+M[k][j];
            for (int l = 0; l < classes; l++)
                sum_kl += M[k][l];
        }
        P[j] = sum_k/(2.0*sum_kl);
    }

    // Pj, Pi
    for (int i = 0; i < classes; i++)
        for (int j = 0; j < classes; j++) {
            scalar sum_j = 0.0/*, sum_i = 0.0*/;
            for (int k = 0; k < classes; k++) {
                sum_j += M[j][k]+M[k][j];
                //                 sum_i += M[i][k]+M[k][i];
            }

            Pj[i][j] = M[i][j]/sum_j;
            //             Pi[i][j] = M[i][j]/sum_i;
        }

    for (int j = 0; j < classes; j++) {
        scalar cen_j = 0.0;
        for (int k = 0; k < classes; k++)
            if (k != j)
                cen_j += h(2.0*(classes-1), Pj[j][k])+h(2.0*(classes-1), Pj[k][j]);

        cen += P[j] * cen_j;
    }

    //     FreeScalarMatrix(&Pi, classes);
    FreeScalarMatrix(&Pj, classes);
    FREE(P);

    return cen;
}

scalar h(scalar b, scalar x)
{
    if (x > ZERO)
        return -x*(log(x)/log(b));
    else
        return 0.0;
}
