/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas   *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>
#include <stdbool.h>
#include <string.h>
#include <strings.h>
#include "typex.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

void OnErrorExit(int id, char *message)
{
    char *codes[] = {"No error", "Not yet implemented", "Unspecified error", "File error", "Memory error", "Read failed", "Invalid arguments", "PVM error", "Numerical error", "Unexpected result",
                     "Deprecated", "MPI error"
                    };

    printf("%s: %s\n", codes[id], message);

    fflush(NULL);

    exit(id);
}

// Select a member of the population following the standard Montecarlo method.
int RouletteSelection(scalar *cumulative, int no_indiv)
{
    int being;
    scalar uniform;

    // An individual is selected using a random value
    uniform = RANDOM(0, cumulative[no_indiv-1]);
    being = 0;
    while (uniform > cumulative[being])
        being++;

    return (MIN(being, no_indiv-1));
}

void RandomizeVector(int *v, int n)
{
    int l, r;

    for (int i = 0; i < n; i++) {
        // Exchange two random positions
        l = RANDOM(0, n);
        r = RANDOM(0, n);

        int aux = v[l];
        v[l] = v[r];
        v[r] = aux;
    }
}

void RandomizeScalarVector(scalar *v, int n)
{
    int l, r;

    for (int i = 0; i < n; i++) {
        // Exchange two random positions
        l = RANDOM(0, n);
        r = RANDOM(0, n);

        scalar aux = v[l];
        v[l] = v[r];
        v[r] = aux;
    }
}

void ZeroVector(scalar *v, int *w, int n)
{
    if (v)
        bzero((void *)v, sizeof(scalar)*n);
    else
        bzero((void *)w, sizeof(int)*n);
}

void ZeroMatrix(scalar **v, int **w, int f, int c)
{
    if (v)
        for (int i = 0; i < f; i++)
            bzero((void *)v[i], sizeof(scalar)*c);
    else
        for (int i = 0; i < f; i++)
            bzero((void *)w[i], sizeof(int)*c);
}

// Returns a vector with its elements randomized
void GetRandomisedVector(int *v, int n)
{
    int aux, r;

    // Order
    for (int i = 0; i < n; i++)
        v[i] = i;

    // Randomize
    for (int i = 0; i < n; i++) {
        r = RANDOM(0, n);

        aux = v[r];
        v[r] = v[i];
        v[i] = aux;
    }

}

// Generates a random population of pop_size individuals and n_feat features selected with a feat_sel_prob probability
void GenerateRandomPopulation(int **new_pop, int pop_size, int n_feat, scalar feat_sel_prob)
{
    int p1 = 0, null_ind;
    scalar local_pfeat = 0.0;

    for (int i = 0; i < pop_size; i++) {
        null_ind =  1;
        for (int j = 0; j < n_feat; j++) {
            //Get scalar random value between (0,1) in order to decide if we are going to select a feature
            local_pfeat = RANDOM(0, 1);

            // Select feature
            if (local_pfeat < feat_sel_prob) {
                new_pop[i][j] = 1;
                null_ind = 0;
            } else
                new_pop[i][j] = 0;
        }
        if (null_ind) {
            p1 = (int) RANDOM(0, n_feat);
            new_pop[i][p1] = 1;
        }
    }
}

// Auxiliary function for qsort()
// Ascending order
int CompareDouble(const void *x, const void *y)
{
    scalar diff;

    if ((diff = * (scalar *) x - * (scalar *) y) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

// Auxiliary function for qsort()
// Ascending order
int CompareInt(const void *x, const void *y)
{
    if (* (int *) x < * (int *) y)
        return -1;
    else {
        if (* (int *) x > * (int *) y)
            return +1;
        else
            return 0;
    }
}

char *ToLower(char *p)
{
    for (int i = 0; i < strlen(p); i++)
        p[i] = tolower(p[i]);

    return p;
}

scalar EuclideanDistanceSubspace(scalar *v, scalar *w, int n, int *subspace)
{
    scalar dist = 0.0;

    for (int i = 0; i < n; i++)
        if (!subspace || subspace[i])
            dist += (v[i] - w[i]) * (v[i] - w[i]);

    return sqrt(dist);
}

void AllocateScalarMatrix(scalar ***m, int r, int c)
{
    MALLOC(*m, scalar *, r, "AllocateScalarMatrix: *m");
    for (int i = 0; i < r; i++)
        MALLOC((*m)[i], scalar, c, "AllocateScalarMatrix: (*m)[i]");
}

void FreeScalarMatrix(scalar ***m, int r)
{
    for (int i = 0; i < r; i++)
        FREE((*m)[i]);
    FREE(*m);

    *m = NULL;
}

void AllocateIntegerMatrix(int ***m, int r, int c)
{
    MALLOC(*m, int *, r, "AllocateIntegerMatrix: *m");
    for (int i = 0; i < r; i++)
        MALLOC((*m)[i], int, c, "AllocateIntegerMatrix: (*m)[i]");
}

void FreeIntegerMatrix(int ***m, int r)
{
    for (int i = 0; i < r; i++)
        FREE((*m)[i]);
    FREE(*m);

    *m = NULL;
}

void AllocateCacheMatrix(cache_t ***m, int r, int c)
{
    MALLOC(*m, cache_t *, r, "AllocateCacheMatrix: *m");
    for (int i = 0; i < r; i++)
        MALLOC((*m)[i], cache_t, c, "AllocateCacheMatrix: (*m)[i]");
}

void FreeCacheMatrix(cache_t ***m, int r)
{
    if (*m != NULL) {
        for (int i = 0; i < r; i++)
            FREE((*m)[i]);
        FREE(*m);

        *m = NULL;
    }
}

// Check vector is sorted
bool CheckSorted(int *v, int n)
{
    for (int i = 0; i < n-1; i++)
        if (v[i] >= v[i+1])
            return false;

    return true;
}

// Auxiliary function for qsort()
int CompareFloat(const void *x, const void *y)
{
    scalar diff;

    if ((diff = * (float *) x - * (float *) y) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

void ZeroFloatVector(float *v, int n)
{
    bzero((void *)v, sizeof(float)*n);
}

void ZeroFloatMatrix(float **v, int f, int c)
{
    for (int i = 0; i < f; i++)
        bzero((void *)v[i], sizeof(float)*c);
}

void AllocateFloatMatrix(float ***m, int r, int c)
{
    MALLOC(*m, float *, r, "AllocateFloatMatrix: *m");
    for (int i = 0; i < r; i++)
        MALLOC((*m)[i], float, c, "AllocateFloatMatrix: (*m)[i]");
}

void FreeFloatMatrix(float ***m, int r)
{
    for (int i = 0; i < r; i++)
        FREE((*m)[i]);
    FREE(*m);

    *m = NULL;
}

void CopyScalarMatrix(scalar **orig, scalar **dest, int r, int c)
{
    for (int i = 0; i < r; i++)
        memcpy(dest[i], orig[i], sizeof(scalar)*c);
}

void CopyScalarMatrixExcludeColumn(scalar **orig, scalar **dest, int r, int c, int exclude)
{
    for (int i = 0; i < r; i++) {
        memcpy(dest[i], orig[i], sizeof(scalar)*exclude);
        memcpy(dest[i]+exclude, orig[i]+exclude+1, sizeof(scalar)*(c-exclude-1));
    }
}

void ZeroIntegerMatrix(int **v, int f, int c)
{
    for (int i = 0; i < f; i++)
        bzero((void *) v[i], sizeof(int)*c);
}

// Auxiliary function for qsort
int Compare(const void *element_1, const void *element_2)
{
    scalar diff;

    if ((diff = ((RANK *) element_2)->fitness - ((RANK *) element_1)->fitness) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

char *GetTime(void)
{
    time_t timer;
    static char buffer[20];
    struct tm *tm_info;

    time(&timer);
    tm_info = localtime(&timer);

    strftime(buffer, 10, "%H:%M:%S (%b %d)", tm_info);

    return buffer;
}

void Allocate3DScalarMatrix(scalar ****m, int r, int c, int d)
{
    MALLOC(*m, scalar **, r, "Allocate3DScalarMatrix: *m");
    for (int i = 0; i < r; i++) {
        MALLOC((*m)[i], scalar *, c, "Allocate3DScalarMatrix: (*m)[i]");
        for (int j = 0; j < c; j++)
            MALLOC((*m)[i][j], scalar, d, "Allocate3DScalarMatrix: (*m)[i][j]");
    }
}

void Zero3DScalarMatrix(scalar ***m, int r, int c, int d)
{
    for (int i = 0; i < r; i++)
        for (int j = 0; j < c; j++)
            bzero((void *)m[i][j], sizeof(scalar)*d);
}

void Free3DScalarMatrix(scalar ****m, int r, int c)
{
    for (int i = 0; i < r; i++) {
        for (int j = 0; j < c; j++)
            FREE((*m)[i][j]);
        FREE((*m)[i]);
    }
    FREE(*m);

    *m = NULL;
}

void Allocate3DIntegerMatrix(int ****m, int r, int c, int d)
{
    MALLOC(*m, int **, r, "Allocate3DIntegerMatrix: *m");
    for (int i = 0; i < r; i++) {
        MALLOC((*m)[i], int *, c, "Allocate3DIntegerMatrix: (*m)[i]");
        for (int j = 0; j < c; j++)
            MALLOC((*m)[i][j], int, d, "Allocate3DIntegerMatrix: (*m)[i][j]");
    }
}

void Free3DIntegerMatrix(int ****m, int r, int c)
{
    for (int i = 0; i < r; i++) {
        for (int j = 0; j < c; j++)
            FREE((*m)[i][j]);
        FREE((*m)[i]);
    }
    FREE(*m);

    *m = NULL;
}

void CopyScalarVector(scalar *orig, scalar *dest, int s)
{
    memcpy(dest, orig, sizeof(scalar)*s);
}

void CopyIntegerVector(int *orig, int *dest, int s)
{
    memcpy(dest, orig, sizeof(int)*s);
}

uint64_t get_posix_clock_time(void)
{
    struct timespec ts;

    if (clock_gettime(CLOCK_MONOTONIC, &ts) == 0)
        return (uint64_t)(ts.tv_sec * 1000000 + ts.tv_nsec / 1000);
    else
        return 0;
}

void BLXAlphaCrossover(int *son, int *daughter, int *father, int *mother, int size, int min_value, int max_value)
{
    // BLX-alpha crossover adapted to integers
    for (int j = 0; j < size; j++) {
        int min = MIN(father[j], mother[j]);
        int max = MAX(father[j], mother[j]);
        scalar I = max - min;
        son[j] = (int) AssignWithinBoundsScalar((scalar) min_value, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), (scalar) max_value);
        daughter[j] = (int) AssignWithinBoundsScalar((scalar) min_value, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), (scalar) max_value);
    }
}

scalar AssignWithinBoundsScalar(scalar lbound, scalar x, scalar rbound)
{
    if (x <= lbound)
        return lbound;
    else if (x >= rbound)
        return rbound;
    else
        return x;
}

int AssignWithinBoundsInteger(int lbound, int x, int rbound)
{
    if (x <= lbound)
        return lbound;
    else if (x >= rbound)
        return rbound;
    else
        return x;
}

void InitVectorRandomly(scalar *w, int *v, int n, scalar lbound, scalar ubound)
{
    if (w)
        for (int i = 0; i < n; i++)
            w[i] = RANDOM(lbound, ubound);
    else
        for (int i = 0; i < n; i++)
            v[i] = (int) RANDOM(lbound, ubound);
}

void FillVector(scalar *v, int *w, int n, scalar val)
{
    if (v)
        for (int i = 0; i < n; i++)
            v[i] = val;
    else
        for (int i = 0; i < n; i++)
            w[i] = (int) val;
}
