/***************************************************************************
*  Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
*  npedrajas@uco.es                                                      *
*                                                                        *
*  This program is FREE software; you can redistribute it and/or modify  *
*  it under the terms of the GNU General Public License as published by  *
*  the Free Software Foundation; either version 2 of the License, or     *
*  (at your option) any later version.                                   *
*                                                                        *
*  This program is distributed in the hope that it will be useful,       *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*  GNU General Public License for more details.                          *
*                                                                        *
*  You should have received a copy of the GNU General Public License     *
*  along with this program; if not, write to the                         *
*  Free Software Foundation, Inc.,                                       *
*  59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*
*Scale the input values.
 */
#include <math.h>
#include <stdio.h>
#include "typex.h"
#include "common.h"
#include "macros.h"

void TypifyInputData(scalar **data, scalar **test_data, int instances, int test_instances, int inputs)
{
    scalar *mean = NULL, *sigma = NULL;
    scalar sq_sum;     // Typify input data

    MALLOC(mean, scalar, inputs, "TypifyInputData: mean");
    MALLOC(sigma, scalar, inputs, "TypifyInputData: sigma");

    if (data != NULL)
        // Scale input
        for (int i = 0; i < inputs; i++) {
            // Get the mean[i] and variance.
            mean[i] = sigma[i] = sq_sum = 0.;

            for (int j = 0; j < instances; j++) {
                mean[i] += data[j][i];
                sq_sum += data[j][i]*data[j][i];
            }

            mean[i] /= instances;
            sigma[i] = sqrt(sq_sum/instances - mean[i]*mean[i]);
        }

    for (int i = 0; i < inputs; i++)
        // Typify: z = (x - mean[i])/std. dev.
        // If standard deviation is 0 do nothing.
        if (sigma[i] > ZERO) {
            if (data != NULL)
                for (int j = 0; j < instances; j++)
                    data[j][i] = (data[j][i] - mean[i])/sigma[i];

            if (test_data != NULL)
                for (int j = 0; j < test_instances; j++)
                    test_data[j][i] = (test_data[j][i] - mean[i])/sigma[i];
        }

    FREE(sigma);
    FREE(mean);
}

// Make the linear transformation x' = ax + b,where a=(MAX - MIN)/(max_val-min_val), b=(MIN*max_val-MAX*min_val)/(max_val-min_val)
void ScaleInputData(scalar **data, scalar **test_data, int training_instances, int test_instances, int inputs, scalar lbound, scalar ubound, char *filename)
{
    scalar max_val, min_val;
    scalar *a = NULL, *b = NULL;

    MALLOC(a, scalar, inputs, "TypifyInputData: a");
    MALLOC(b, scalar, inputs, "TypifyInputData: b");

    // Scale output
    if (data != NULL)
        for (int i = 0; i < inputs; i++) {
            // Get the max and min values of the column
            min_val = max_val = data[0][i];
            for (int j = 0; j < training_instances; j++)
                if (max_val < data[j][i])
                    max_val = data[j][i];
                else {
                    if (min_val > data[j][i])
                        min_val = data[j][i];
                }

            // Calculate a and b coefficients
            // If constant do nothing.
            if (fabs(max_val - min_val) > ZERO) {
                a[i] = (ubound - lbound)/(max_val - min_val);
                b[i] = (lbound*max_val - ubound*min_val)/(max_val - min_val);
            } else
                a[i] = b[i] = 0.0;
        }

    for (int i = 0; i < inputs; i++) {
        if (data != NULL)
            // Scale column
            for (int j = 0; j < training_instances; j++)
                data[j][i] = a[i]*data[j][i] + b[i];
        if (test_data != NULL)
            for (int j = 0; j < test_instances; j++)
                test_data[j][i] = a[i]*test_data[j][i] + b[i];
    }

    // If filename is not NULL, write scaling values
    if (filename) {
        FILE *f;

        if ((f = fopen(filename, "w"))== NULL)
            OnErrorExit(1, "Cannot open file for writing scaling values");

        // Write a and b
        for (int i = 0; i < inputs; i++)
            fprintf(f, "%lf %lf\n", a[i], b[i]);

        fclose(f);
    }

    FREE(a);
    FREE(b);
}

// Standardize sparse data.
// Mean = 0 to keep data sparse

void TypifySparseData(scalar **data, scalar **test_data, int instances, int test_instances, int inputs, int **train_index, int **test_index)
{
    scalar *sigma, *mean;
    int *count;

    MALLOC(sigma, scalar, inputs, "TypifySparseData: sigma");
    MALLOC(count, int, inputs, "TypifySparseData: count");
    MALLOC(mean, scalar, inputs, "TypifySparseData: mean");

    ZeroVector(NULL, count, inputs);
    ZeroVector(sigma, NULL, inputs);
    ZeroVector(mean, NULL, inputs);

    for (int i = 0; i < instances; i++) {
        int j = 0;
        while (train_index[i][j] != -1) {
            mean[train_index[i][j]-1] += data[i][j];
            sigma[train_index[i][j]-1] += data[i][j]*data[i][j];
            count[train_index[i][j]-1]++;
            j++;
        }
    }

    for (int i = 0; i < inputs; i++)
        if (count[i]) {
            mean[i] /= count[i];
            sigma[i] = sqrt(sigma[i]/count[i] - mean[i]*mean[i]);
        }

    for (int i = 0; i < instances; i++) {
        int j = 0;
        while (train_index[i][j] != -1) {
            if (sigma[train_index[i][j]-1] > ZERO)
                data[i][j] /= sigma[train_index[i][j]-1];
            j++;
        }
    }

    if (test_data != NULL)
        for (int i = 0; i < test_instances; i++) {
            int j = 0;
            while (test_index[i][j] != -1) {
                if (sigma[test_index[i][j]-1] > ZERO)
                    test_data[i][j] /= sigma[test_index[i][j]-1];
                j++;
            }
        }

    FREE(sigma);
    FREE(count);
    FREE(mean);
}
