/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdarg.h>
#include <string.h>
#include "maths.h"
#include "data.h"
#include "macros.h"
#include "common.h"

void gram_schmidt(int, scalar *, scalar *);
void getVec2D(int OP, ...);

void BidimensionalGrandTour(const t_Parameters *parm,  int *partition, int *count, scalar **data, int instances, int *subsets, int *label, int inputs, int classes)
{
    int *class_count = NULL, columns;
    RANK *to_sort;
    scalar *old[2], *base[2];

    // Number of subsets
    int rows = (int) sqrt(*subsets);
    if ((rows * rows) == *subsets)
        columns = rows;
    else
        columns = rows+1;

    *subsets = rows * columns;

    // Memory allocation
    MALLOC(class_count, int, classes, "BidimensionalGrandTour: class_count");
    MALLOC(to_sort, RANK, instances, "BidimensionalGrandTour: to_sort");
    MALLOC(old[0], scalar, inputs, "BidimensionalGrandTour: old[0]");
    MALLOC(old[1], scalar, inputs, "BidimensionalGrandTour: old[1]");
    MALLOC(base[0], scalar, inputs, "BidimensionalGrandTour: base[0]");
    MALLOC(base[1], scalar, inputs, "BidimensionalGrandTour: base[1]");

    if (parm->use_pcomponents)
        // First two principal components
        PrincipalComponentsAnalysis((scalar **) data, instances, inputs, 2, NULL, base);
    else {
        for (int i = 0; i < inputs; i++) {
            base[0][i] = RANDOM(-1.0, 1.0);
            base[1][i] = RANDOM(-1.0, 1.0);
        }
        normalize(inputs, base[0]);
        gram_schmidt(inputs, base[0], base[1]);
    }

    getVec2D(INIT, inputs, parm->angle_step);         // Init the vec generation dim: inputs, step: degrees
    getVec2D(SET, base[0], base[1]);
    getVec2D(NEXT, base[0], base[1]);       // Ignore first step

    // First vector
    scalar value;
    for (int j = 0; j < instances; j++) {
        // Projection
        value = 0.0;
        for (int i = 0; i < inputs; i++)
            value += data[j][i] * base[0][i];
        to_sort[j].fitness = value;
        to_sort[j].being = j;
    }
    qsort(to_sort, instances, sizeof(RANK), Compare);

    // Partition in rows number of subsets
    int size = instances/rows;
    for (int j = 0; j < rows * size; j++)
        partition[to_sort[j].being] = j/size;
    // Assign remaining randomly
    for (int j = size * rows; j < instances; j++) {
        int subset = RANDOM(0, rows);
        partition[to_sort[j].being] = subset;
    }

    // Second vector
    int next = rows;
    for (int i = 0; i < rows; i++) {
        int n = 0;
        for (int j = 0; j < instances; j++)
            if (partition[j] == i) {
                scalar value = 0.0;

                for (int k = 0; k < inputs; k++)
                    value += data[j][k] * base[1][k];

                to_sort[n].fitness = value;
                to_sort[n].being = j;
                n++;
            }
        qsort(to_sort, n, sizeof(RANK), Compare);

        // Partition into columns subsets
        int size = n/columns;
        for (int j = size; j < columns * size; j++)
            partition[to_sort[j].being] = next+j/size - 1;
        // Assign remaining randomly
        for (int j = size * columns; j < n; j++) {
            int subset = RANDOM(0, columns);
            if (subset != 0)
                partition[to_sort[j].being] = next+subset - 1;
        }

        next += columns - 1;
    }

    // Count
    for (int i = 0; i < *subsets; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    if (parm->verbose > 0) {
        fprintf(parm->out_file, "Partition info: ");
        for (int i = 0; i < *subsets; i++)
            fprintf(parm->out_file, "%d ", count[i]);
        fprintf(parm->out_file, "\n");

        // Class distribution
        fprintf(parm->out_file, "Class distribution info:\n");
        for (int i = 0; i < *subsets; i++) {
            for (int j = 0; j < classes; j++)
                class_count[j] = 0;

            for (int j = 0; j < instances; j++)
                if (partition[j] == i)
                    class_count[label[j]]++;
            fprintf(parm->out_file, "Subset %4d: ", i);
            for (int j = 0; j < classes; j++)
                fprintf(parm->out_file, "%3d ", class_count[j]);
            fprintf(parm->out_file, "\n");
        }
    }

    FREE(class_count);
    FREE(to_sort);
    FREE(old[0]);
    FREE(old[1]);
    FREE(base[0]);
    FREE(base[1]);
}

void gram_schmidt(int dim, scalar *vec1, scalar *vec2)
{
    int i;
    scalar *x1, *x2;
    scalar ip = 0.0;
    for (i = 0, x1 = vec1, x2 = vec2; i < dim; i++, x1++, x2++)
        ip += (*x1) * (*x2);

    for (i = 0, x1 = vec1, x2 = vec2; i < dim; i++, x1++, x2++)
        *x2 = *x2 - ip * (*x1);

    normalize(dim, vec2);
}


void getVec2D(int OP, ...)
{
    va_list ap;

    static int dim;
    static int p;
    static scalar *last;
    static scalar *next;
    static scalar *alpha;
    static scalar *ivec1;
    static scalar *ivec2;
    static scalar stepSize;
    static int initialized = 0;

    scalar *vec;


    va_start(ap, OP);

    switch (OP) {
        case INIT:                      // getVec2D(INIT, int dim, scalar step)
            if (initialized) {
                fprintf(stderr, "getVec2D error: already initialized\n");
                exit(-1);
            }
            dim = va_arg(ap, int);
            stepSize = va_arg(ap, double);

            p = dim * (dim - 1)/2;
            alpha = (scalar *) malloc(sizeof(scalar) * p);
            last = (scalar *) malloc(sizeof(scalar) * p);
            next = (scalar *) malloc(sizeof(scalar) * p);

            nextAlpha(dim, p, stepSize, last, next, NULL);          // INITIALIZATION

            // The unit vector that is going to be rotate if not initial direction is provided.
            //ivec=(scalar *)malloc(sizeof(scalar)*dim);for (i=1;i<dim;i++) ivec[i]=0; ivec[0]=1;
            ivec1 = (scalar *) calloc(dim, sizeof(scalar));
            ivec1[0] = 1;
            ivec2 = (scalar *) calloc(dim, sizeof(scalar));
            ivec2[1] = 1;

            initialized = 1;
            break;
        case SET:                       // getVec2D(SET, scalar * ivec1, scalar * ivec2)
            if (!initialized) {
                fprintf(stderr, "getVec2D error: no initialization\n");
                exit(-1);
            }
            vec = va_arg(ap, scalar *);
            memcpy(ivec1, vec, dim * sizeof(scalar));       // for (i=0;i<dim;i++) ivec[i]=vec[i];
            normalize(dim, ivec1);
            vec = va_arg(ap, scalar *);
            memcpy(ivec2, vec, dim * sizeof(scalar));       // for (i=0;i<dim;i++) ivec[i]=vec[i];
            normalize(dim, ivec2);
            break;
        case NEXT:                      // getVec2D(NEXT, scalar * nvec1, scalar * nvec2)
            if (!initialized) {
                fprintf(stderr, "getVec2D error: no initialization\n");
                exit(-1);
            }
            nextAlpha(dim, p, stepSize, last, next, alpha);
            vec = va_arg(ap, scalar *);
            memcpy(vec, ivec1, dim * sizeof(scalar));       // for (i=0;i<dim;i++) vec[i]=ivec[i];
            applyRotation(dim, alpha, vec);
            vec = va_arg(ap, scalar *);
            memcpy(vec, ivec2, dim * sizeof(scalar));       // for (i=0;i<dim;i++) vec[i]=ivec[i];
            applyRotation(dim, alpha, vec);
            break;
        case CLEAR:             // getVec2D(CLEAR)
            free(alpha);
            free(ivec1);
            free(ivec2);
            free(last);
            free(next);
            initialized = 0;
            break;
    }
    va_end(ap);
}
