/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include "fs.h"
#include "macros.h"
#include "svm.h"
#include "data.h"

// Implements the method WPKID for discretization
void WPKIDMethod(scalar **data, int n, int n_feat, scalar **limits, int *intervals, int **index)
{
    int interval = 0;
    scalar *to_sort = NULL;
    MALLOC(to_sort, scalar, n, "WPKIDMethod: to_sort");

    for (int j = 0; j < n_feat; j++)  {
        intervals[j] = MAX(2, -15 + sqrt(225.0 + n));

        // Get equally distributed intervals
        // Sort values
        if (index)
            for (int i = 0; i < n; i++)
                to_sort[i] = SparseFeature(data[i], index[i], j);
        else
            for (int i = 0; i < n; i++)
                to_sort[i] = data[i][j];

        // Sort values
        qsort(to_sort, n, sizeof(scalar), CompareDouble);
        interval = n/intervals[j];
        // Assign limits
        REALLOC(limits[j], scalar, intervals[j] + 1, "WPKIDMethod: limits[j]");
        for (int i = 0; i < intervals[j]; i++)
            limits[j][i] = to_sort[(int)((i + 1) * interval)];
    }
    FREE(to_sort);
}


// Implements the method for discretization described in
// U. M. Fayyad and K. B. Irani, "Multi-interval discretization of continuous-valued attributes for classification learning",
// Proceedings of the International Joint Conference on Uncertainty in AI (Q334 .I571 1993), pp. 1022-1027.

void FayyadAndIraniMDLMethod(scalar **data, int instances, int outputs, int *labels, int inputs, scalar **limits, int *intervals, int *subset, int **index)
{
    t_info *feature;

    MALLOC(feature, t_info, instances, "FayyadAndIraniMDLMethod: feature");
    // Repeat for each instance
    for (int i = 0; i < inputs; i++) {
        int count = 0;
        // Sort instances using feature i-th
        for (int j = 0; j < instances; j++)
            if (!subset || subset[j]) {
                scalar v = index ? SparseFeature(data[j], index[j], i) : data[j][i];

                feature[count].value = v;
                feature[count].label = labels[j];
                count++;
            }

        qsort(feature, count, sizeof(t_info), CompareInfo);

        // Recursively discretize feature
        intervals[i] = 1;
        DiscretizeFeature(feature, count, outputs, &intervals[i], &limits[i]);

        // Sort limits
        qsort(limits[i], intervals[i], sizeof(scalar), CompareDouble);
    }

    FREE(feature);
}

// Auxiliary function for qsort.
int CompareInfo(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_info *) y)->value - ((t_info *) x)->value) > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}

void DiscretizeFeature(t_info *feature, int instances, int outputs, int *intervals, scalar **limits)
{
    int k1, k2, best_k1 = 0, best_k2 = 0;
    scalar EntS1, EntS2, best_EntS1 = 0.0, best_EntS2 = 0.0;

    // Get next cut point
    scalar best_E = INFINITE, best_cut_point = 0.0;
    bool all_same_class = true;

    for (int i = 0; i < instances - 1; i++) {
        // Cut point can only be boundaries between classes
        if (feature[i].label != feature[i + 1].label) {
            // Cut point between values
            scalar cut_point = (feature[i].value + feature[i + 1].value) / 2.0;

            // Evaluate class information entropy of the partition induced by cut point
            scalar E = EvaluateCutPointCIF(cut_point, feature, instances, outputs, &k1, &k2, &EntS1, &EntS2);

            if (E <= best_E) {
                best_E = E;
                best_cut_point = cut_point;
                best_k1 = k1;
                best_k2 = k2;
                best_EntS1 = EntS1;
                best_EntS2 = EntS2;

                all_same_class = false;
            }
        }
    }

    // If all in the same class do nothing
    if (all_same_class && *intervals > 1)
        return;

    // Evaluate criterion, always accepting the first cut point
    if ((*intervals) == 1 || EvaluateMDLCriterion(feature, instances, outputs, best_E, best_EntS1, best_EntS2, best_k1, best_k2)) {
        // Accept cut point
        REALLOC(*limits, scalar, *intervals + 1, "DiscretizeFeature: *limits");
        (*limits) [*intervals - 1] = best_cut_point;
        (*intervals) ++;

        // Count instances in each subset
        int n = 1;
        while (n < instances && feature[n].value <= best_cut_point)
            n++;

        // Do a recursively call DiscretizeFeature
        DiscretizeFeature(feature, n, outputs, intervals, limits);
        DiscretizeFeature(feature + n, instances - n, outputs, intervals, limits);
    } else
        // Don't store cut point and return
        return;
}

scalar EvaluateCutPointCIF(scalar T, t_info *feature, int n, int outputs, int *k1, int *k2, scalar *EntS1, scalar *EntS2)
{
    scalar E, P[outputs];
    int j;

    // Count number of classes in each subset
    *k1 = *k2 = 0;

    // Class information entropy of the partition
    // S1: Values <= T
    for (int i = 0; i < outputs; i++)
        P[i] = 0.0;
    j = 0;
    while (j < n && feature[j].value <= T) {
        P[feature[j].label]++;
        j++;
    }
    scalar E1 = 0.0;
    for (int i = 0; i < outputs; i++)
        if (P[i] > 0.0) {
            E1 -= (P[i]/j) * log2(P[i]/j);
            (*k1) ++;
        }
    E1 *= j / ((scalar) n);

    // S2: Values > T
    for (int i = 0; i < outputs; i++)
        P[i] = 0.0;
    for (int i = j; i < n; i++)
        P[feature[i].label]++;

    scalar E2 = 0.0;
    for (int i = 0; i < outputs; i++)
        if (P[i] > 0.0) {
            E2 -= (P[i]/ (n-j)) * log2(P[i]/ (n-j));
            (*k2) ++;
        }
    E2 *= (n - j) / ((scalar) n);

    *EntS1 = E1;
    *EntS2 = E2;

    return E = E1 + E2;
}

// Criterion for accepting a cut point
// Gain(A,T;S) > log2(N-1)/N + Delta(A,T;S)/N

bool EvaluateMDLCriterion(t_info *feature, int n, int outputs, scalar E, scalar EntS1, scalar EntS2, int k1, int k2)
{
    scalar P[outputs];
    int k = 0;

    // Gain(A,T;S) = Ent(S) - E(A,T;S)
    // Get entropy of S
    for (int i = 0; i < outputs; i++)
        P[i] = 0.0;
    for (int i = 0; i < n; i++)
        P[feature[i].label]++;

    scalar EntS = 0.0;
    for (int i = 0; i < outputs; i++)
        if (P[i] > 0.0) {
            EntS -= (P[i]/n) * log2(P[i]/n);
            k++;
        }

    scalar gain = EntS - E;

    scalar Delta =
        log2(pow(3.0, k) - 2.0) - (k * EntS - k1 * EntS1 - k2 * EntS2);

    bool accept = (gain > (log2(n - 1.0) / n + Delta / n));

    return accept;
}
