/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Partition following a k-means algorithm
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "macros.h"
#include "common.h"
#include "maths.h"
#include "data.h"
#include "classifiers.h"

void DataDivisionkMeans(const t_Parameters *parm, int *partition, int **count, scalar **data, int instances, int *subsets, int inputs, int kmeans_max_iterations, int **index, int subset_size)
{
    int *cl_partition = NULL, *cl_count = NULL, cl_subsets, **cl_index = NULL, *list = NULL, previous_subsets;
    scalar **cl_data = NULL;

    previous_subsets = *subsets;

    // k-Means with all data
    kMeans(partition, *count, data, instances, *subsets, inputs, kmeans_max_iterations, index, NULL);
    RemoveEmptyClusters(subsets, *count, partition, instances);

    // Recursively divide large clusters
    for (int i = 0; i < *subsets; i++) {
        // If larger than size perform division
        if ((*count)[i] > subset_size) {
            cl_subsets = ceil(((scalar)(*count)[i])/subset_size);

            // Memory for temporal data
            REALLOC(cl_partition, int, (*count)[i], "DataDivisionkMeans: cl_partition");
            REALLOC(cl_count, int, cl_subsets, "DataDivisionkMeans: cl_count");
            REALLOC(list, int, (*count)[i], "DataDivisionkMeans: list");
            if (index)
                REALLOC(cl_index, int *, (*count)[i], "DataDivisionkMeans: cl_index");
            REALLOC(cl_data, scalar *, (*count)[i], "DataDivisionkMeans: cl_data");

            // Copy data
            int c = 0;
            for (int j = 0; j < instances; j++)
                if (partition[j] == i) {
                    list[c] = j;
                    cl_data[c] = data[j];
                    if (index)
                        cl_index[c] = index[j];
                    c++;
                }
            if (c != (*count)[i])
                OnErrorExit(ERR_UNSPECTED_RESULT, "Count and set size do not match");

            kMeans(cl_partition, cl_count, cl_data, (*count)[i], cl_subsets, inputs, kmeans_max_iterations, cl_index, NULL);

            // Remove i-th clusters as it has been divided into others
            (*count)[i] = (*count)[*subsets-1];
            for (int j = 0; j < instances; j++)
                if (partition[j] == *subsets-1)
                    partition[j] = i;
            (*subsets)--;

            // Add new clusters
            int added = 0;
            for (int j = 0; j < cl_subsets; j++)
                if (cl_count[j] > 0) {
                    // Add as last one
                    for (int k = 0; k < c; k++)
                        if (cl_partition[k] == j)
                            partition[list[k]] = *subsets;

                    // Add new one
                    (*count)[*subsets] = cl_count[j];
                    (*subsets)++;
                    if (*subsets > previous_subsets-1)
                        REALLOC((*count), int, *subsets+1, "DataDivisionkMeans: (*count)");
                    added++;
                }

            // If not divided into at least 2 clusters, perform random partition
            if (added == 1) {
                (*count)[*subsets-1] = (*count)[*subsets] = 0;
                for (int j = 0; j < instances; j++)
                    if (partition[j] == *subsets-1) {
                        if (RANDOM(0.0, 1.0) < 0.5)
                            (*count)[*subsets-1]++;
                        else {
                            (*count)[*subsets]++;
                            partition[j] = *subsets;
                        }
                    }
                (*subsets)++;
                if (*subsets > previous_subsets-1)
                    REALLOC((*count), int, *subsets+1, "DataDivisionkMeans: (*count)");
            }

            // Revisit
            i--;
        }
    }

    // Memory for temporal data
    if (cl_partition) {
        FREE(cl_partition);
        FREE(cl_count);
        FREE(list);
        if (index)
            FREE(cl_index);
        FREE(cl_data);
    }

    // Average subset size
    scalar average_size = 0.0;
    for (int i = 0; i < *subsets; i++)
        average_size += (*count)[i];
    fprintf(parm->out_file, "@kMeans average subset size: %.1lf\n", average_size/(*subsets));
}

void RemoveEmptyClusters(int *subsets, int *count, int *partition, int instances)
{
    // Remove empty initial clusters
    int i = 0;
    do {
        if (count[i] == 0) {
            count[i] = count[*subsets-1];
            for (int j = 0; j < instances; j++)
                if (partition[j] == *subsets-1)
                    partition[j] = i;
            (*subsets)--;
        } else
            i++;
    } while (i < *subsets);
}

void kMeans(int *partition, int *count, scalar **data, int instances, int n_clusters, int inputs, int kmeans_max_iterations, int **index, int *subset)
{
    scalar **centroids;
    scalar dist_to_nearest, dist;
    int *old_belong, **centroids_index = NULL;
    bool finished;

    // Allocate memory
    MALLOC(centroids, scalar *, n_clusters, "kMeans: centroids");
    if (index)
        MALLOC(centroids_index, int *, n_clusters, "kMeans: centroids_index");
    MALLOC(old_belong, int, instances, "kMeans: old_belong");
    for (int i = 0; i < n_clusters; i++) {
        MALLOC(centroids[i], scalar, inputs, "kMeans: centroids[i]");
        if (index)
            MALLOC(centroids_index[i], int, inputs+1, "kMeans: centroids_index[i]");
    }

    // Assign random instances as initial centers
    for (int i = 0; i < instances; i++)
        partition[i] = i;
    RandomizeVector(partition, instances);
    for (int i = 0; i < n_clusters; i++)
        if (!index)
            for (int j = 0; j < inputs; j++)
                centroids[i][j] = data[partition[i]][j];
        else {
            int j = 0;
            for (; index[partition[i]][j] != -1; j++) {
                centroids[i][j] = data[partition[i]][j];
                centroids_index[i][j] = index[partition[i]][j];
            }
            centroids_index[i][j] = -1;
        }

    for (int i = 0; i < instances; i++)
        partition[i] = -1;

    // Do while not converged
    int iteration = 0;
    do {
        // Assign instance to cluster
        for (int i = 0; i < instances; i++)
            if (!subset || subset[i]) {
                old_belong[i] = partition[i];
                partition[i] = 0;
                dist_to_nearest = EuclideanDistanceSparseData(data[i], centroids[0], inputs, index ? index[i] : NULL, centroids_index ? centroids_index[0] : NULL);
                for (int j = 1; j < n_clusters; j++)
                    if ((dist = EuclideanDistanceSparseData(data[i], centroids[j], inputs, index ? index[i] : NULL, centroids_index ? centroids_index[j] : NULL)) < dist_to_nearest) {
                        dist_to_nearest = dist;
                        partition[i] = j;
                    }
            }

        // Recalculate centroids
        for (int i = 0; i < n_clusters; i++) {
            count[i] = 0;
            for (int j = 0; j < inputs; j++)
                centroids[i][j] = 0.0;
        }

        for (int i = 0; i < instances; i++)
            if (!subset || subset[i]) {
                count[partition[i]]++;
                if (!index)
                    for (int j = 0; j < inputs; j++)
                        centroids[partition[i]][j] += data[i][j];
                else {
                    for (int j = 0; index[i][j] != -1; j++)
                        centroids[partition[i]][index[i][j]-1] += data[i][j];
                }
            }

        for (int i = 0; i < n_clusters; i++)
            if (count[i])
                for (int j = 0; j < inputs; j++)
                    centroids[i][j] /= count[i];

        // Convert to sparse if needed
        if (index) {
            for (int i = 0; i < n_clusters; i++) {
                int c = 0;
                for (int j = 0; j < inputs; j++)
                    if (fabs(centroids[i][j]) > ZERO) {
                        centroids[i][c] = centroids[i][j];
                        centroids_index[i][c] = j+1;
                        c++;
                    }
                centroids_index[i][c] = -1;
            }
        }

        // Check if something changed
        finished = true;
        for (int i = 0; i < instances && finished; i++)
            if ((!subset || subset[i]) && partition[i] != old_belong[i])
                    finished = false;

        iteration++;

    } while (!finished && iteration < kmeans_max_iterations);

    // Count
    for (int i = 0; i < n_clusters; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            count[partition[i]]++;

    FREE(old_belong);
    for (int i = 0; i < n_clusters; i++) {
        FREE(centroids[i]);
        if (index)
            FREE(centroids_index[i]);
    }
    FREE(centroids);
    if (index)
        FREE(centroids_index);
}

// Evaluate a partition using Dunn's index
// The larger the better
scalar DunnIndex(scalar **data, int instances, int v, int *cluster)
{
    scalar distance, dunn;

    // Obtain minimum delta and maximum diameter
    scalar max_diameter = 0.0;
    scalar min_delta = INFINITE;

    for (int i = 1; i < instances; i++)
        for (int j = 0; j < i; j++) {
            distance = EuclideanDistance(data[i], data[j], v);
            if (cluster[j] != cluster[i]) {
                if (distance < min_delta)
                    min_delta = distance;
            } else {
                if (distance > max_diameter)
                    max_diameter = distance;
            }
        }

    if (max_diameter > 0.0)
        dunn = min_delta/max_diameter;
    else
        dunn = 10.0;

    return dunn;
}

void kMeansStratifiedSample(const t_Parameters *parm, scalar ***data, int instances, int **labels, scalar ***sampled_data, int **sampled_labels, int *sampled_instances, int classes, int inputs, int k,
                            int **index, bool free_data)
{
    int *partition = NULL, *count = NULL, cinstances, ck;
    scalar **centroids, **cdata = NULL;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for k-Means sampling");

    // Count classes
    int ccount[classes];
    for (int i = 0; i < classes; i++)
        ccount[i] = 0;
    for (int i = 0; i < instances; i++)
        ccount[(*labels)[i]]++;

    AllocateScalarMatrix(&centroids, k, inputs);
    MALLOC(partition, int, instances, "kMeansStratifiedSample: partition");
    MALLOC(count, int, k, "kMeansStratifiedSample: count");

    *sampled_instances = 0;
    *sampled_data = NULL;
    *sampled_labels = NULL;
    // Partition per class
    for (int c = 0; c < classes; c++)
        if (ccount[c]) {
            // Create data for class
            cinstances = ccount[c];
            REALLOC(cdata, scalar *, cinstances, "kMeansStratifiedSample: cdata");

            int j = 0;
            for (int i = 0; i < instances; i++)
                if ((*labels)[i] == c)
                    cdata[j++] = (*data)[i];

            // Adjust k
            ck = MAX(1, (((scalar) cinstances)/instances)*k);

            kMeans(partition, count, cdata, cinstances, ck, inputs, parm->kmeans_max_iterations, index, NULL);

            // For every cluster select nearest instance
            REALLOC(*sampled_data, scalar *, *sampled_instances+ck, "kMeansStratifiedSample: *sampled_data");
            REALLOC(*sampled_labels, int, *sampled_instances+ck, "kMeansStratifiedSample: *sampled_labels");

            // Calculate centroids
            CalculateCentroids(cdata, cinstances, centroids, ck, inputs, count, partition, index);

            int added = 0;
            for (int i = 0; i < ck; i++) {
                int nearest;

                // Get nearest neighbor of non empty clusters and add to sample
                if (count[i]) {
                    kNNCategory(centroids[i], cdata, NULL, cinstances, 1, inputs, *labels, classes, &nearest, NULL, NULL, NULL, -1, NULL, NULL);

                    MALLOC((*sampled_data)[*sampled_instances], scalar, inputs, "kMeansStratifiedSample: (*sampled_data)[*sampled_instances]");
                    memcpy((*sampled_data)[*sampled_instances], cdata[nearest], inputs*sizeof(scalar));

                    (*sampled_labels)[*sampled_instances] = c;

                    (*sampled_instances)++;
                    added++;
                }
            }

            SHOWPROGRESS(parm->show_progress > 1, "Sampled using k-Means to %*d instances for class %*d\n", WIDTH(k), added, WIDTH(classes), c+1);
        }

    SHOWPROGRESS(parm->show_progress > 0, "Sampled using k-Means to %d instances\n", *sampled_instances);

    FreeScalarMatrix(&centroids, k);
    FREE(partition);
    FREE(count);
    FREE(cdata);

    // Free previous data?
    if (free_data) {
        FreeScalarMatrix(data, instances);
        FREE(*labels);
    }
}
