/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "data.h"
#include "macros.h"
#include "common.h"
#include "maths.h"

// KCD: Keep Class Distribution

// Functions that deal with partition which keep the class distribution
void RandomPartitionKCD(int *partition, int *count, int n, int n_subsets, int *label, int classes, int subset_max_size)
{
    int *list = NULL, **class_list, *class_count, *class_c;
    scalar size;

    MALLOC(list, int, n, "RandomPartition: list");
    MALLOC(class_list, int *, classes, "RandomPartition: class_list");
    MALLOC(class_count, int, classes, "RandomPartition: class_count");
    MALLOC(class_c, int, classes, "RandomPartition: class_c");

    // Number of instances per class
    for (int i = 0; i < classes; i++)
        class_count[i] = 0;
    for (int i = 0; i < n; i++)
        class_count[label[i]]++;
    for (int i = 0; i < classes; i++)
        MALLOC(class_list[i], int, class_count[i], "RandomPartition: class_list[i]");

    // List of instances in each class
    for (int i = 0; i < classes; i++)
        class_count[i] = 0;
    for (int i = 0; i < n; i++) {
        class_list[label[i]][class_count[label[i]]] = i;
        class_count[label[i]]++;
    }

    // Random partition
    for (int i = 0; i < classes; i++)
        RandomizeVector(class_list[i], class_count[i]);

    size = n/n_subsets;
    for (int i = 0; i < classes; i++)
        class_c[i] = 0;
    for (int i = 0; i < n_subsets; i++) {
        count[i] = 0;
        // For each class
        for (int k = 0; k < classes; k++) {
            for (int j = 0; j < (size * class_count[k])/n - 1 && class_c[k] < class_count[k]; j++) {
                partition[class_list[k][class_c[k]]] = i;
                class_c[k]++;
                count[i]++;
            }
        }
    }

    // Assign remaining randomly
    for (int i = 0; i < classes; i++)
        for (; class_c[i] < class_count[i];) {
            int subset = RANDOM(0, n_subsets);
            if (count[subset] < subset_max_size) {
                partition[class_list[i][class_c[i]]] = subset;
                count[subset]++;
                class_c[i]++;
            }
        }

    FREE(list);
    FREE(class_list);
    FREE(class_count);
    FREE(class_c);
}

// Perform a partition keeping all the subsets local
void PartitionateEqualKCD(int *partition, int *count, scalar **data, int instances, int subsets, int *label, int inputs, int classes)
{
    int *features = NULL, init;
    RANK *to_sort;

    // Memory allocation
    init = 0;
    MALLOC(features, int, inputs, "PartitionateEqual: features");
    MALLOC(to_sort, RANK, instances, "PartitionateEqual: to_sort");
    for (int i = 0; i < inputs; i++)
        features[i] = i;
    // Random order of variables
    RandomizeVector(features, inputs);

    // Initialize partition
    for (int i = 0; i < instances; i++)
        partition[i] = 0;

    // Divide by every variable, taken at random
    for (int i = 0; i < log2(subsets); i++) {
        // Pick random variable
        int variable = features[i];

        for (int k = 0; k < pow(2.0, i); k++)
            // For each class
            for (int c = 0; c < classes; c++) {
                int n = 0;

                for (int j = 0; j < instances; j++)
                    if (partition[j] == k && label[j] == c) {
                        to_sort[n].fitness = data[j][variable];
                        to_sort[n].being = j;
                        n++;
                    }
                qsort(to_sort, n, sizeof(RANK), Compare);

                // Divide using order
                int j = init;
                for (int l = n/2; l < n; l++, j = (j+1) % n)
                    partition[to_sort[j].being] += (int) pow(2.0, i);
            }
    }

    // Count
    for (int i = 0; i < subsets; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    FREE(features);
    FREE(to_sort);
}

void PartitionateOneInputKCD(const t_Parameters *parm, int *partition, int *count, scalar **data, int instances, int subsets, int *label, int inputs, int classes, t_Partition partition_method)
{
    int /**class_count = NULL,*/ init = -1, c, variable = -1;
    RANK *to_sort;
    scalar *coeffs = NULL, value;
    static bool initialized = false;

    // Memory allocation
    // MALLOC(class_count, int, classes, "PartitionateEqual: class_count");
    MALLOC(to_sort, RANK, instances, "PartitionateEqual: to_sort");

    if (partition_method == pGrandTour) {
        // Random projection
        MALLOC(coeffs, scalar, inputs, "PartitionateOneInputKCD: coeffs");
        if (parm->use_pcomponents)
            PrincipalComponentsAnalysis((scalar **) data, instances, inputs, 1, NULL, &coeffs);
        else {
            // or random coefficients for the linear projection
            for (int i = 0; i < inputs; i++)
                coeffs[i] = RANDOM(-1.0, 1.0);

            // Normalize vector
            normalize(inputs, coeffs);
        }

        if (!initialized) {
            getVec(INIT, inputs, parm->angle_step);           // Init the vec generation dim: inputs, step: degrees
            getVec(SET, coeffs);    // Set the initial direction of the vector
            initialized = true;
        }
        getVec(NEXT, coeffs);           // Ignore first one
    } else
        variable = RANDOM(0, inputs);

    init = 0;

    for (int i = 0; i < classes; i++) {
        c = 0;
        if (partition_method == pGrandTour) {
            for (int j = 0; j < instances; j++)
                if (label[j] == i) {
                    // Projection
                    value = 0.0;
                    for (int k = 0; k < inputs; k++)
                        value += data[j][k] * coeffs[k];
                    to_sort[c].fitness = value;
                    to_sort[c].being = j;
                    c++;
                }
        } else {
            for (int j = 0; j < instances; j++)
                if (label[j] == i) {
                    to_sort[c].fitness = data[j][variable];
                    to_sort[c].being = j;
                    c++;
                }
        }

        qsort(to_sort, c, sizeof(RANK), Compare);
        int size = c/subsets;
        int j = init;
        for (int l = 0; l < subsets * size; l++, j = (j+1) % c)
            partition[to_sort[j].being] = j/size;
        // Assign remaining randomly
        for (int l = size * subsets; l < c; l++, j = (j+1) % c) {
            int subset = RANDOM(0, subsets);
            partition[to_sort[j].being] = subset;
        }
    }

    // Count
    for (int i = 0; i < subsets; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    // FREE(class_count);
    FREE(to_sort);
    if (partition_method == pGrandTour)
        FREE(coeffs);
}
