/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "macros.h"
#include "typex.h"
#include "data.h"
#include "classifiers.h"

// Sampling using k-means for scaling up algorithms

void kMeansSample(const t_Parameters *parm, scalar ***data, int instances, int **labels, scalar ***sampled_data, int **sampled_labels, int *sampled_instances, int classes, int inputs, int k,
                  int **index, bool free_data)
{
    int *partition = NULL, *count = NULL;
    scalar **centroids;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse data for k-Means sampling");

    AllocateScalarMatrix(&centroids, k, inputs);
    MALLOC(partition, int, instances, "kMeansSample: partition");
    MALLOC(count, int, k, "kMeansSample: count");
    kMeans(partition, count, *data, instances, k, inputs, parm->kmeans_max_iterations, index, NULL);

    // For every cluster select nearest instance
    *sampled_instances = 0;
    MALLOC(*sampled_data, scalar *, k, "kMeansSample: *sampled_data");
    MALLOC(*sampled_labels, int, k, "kMeansSample: *sampled_labels");

    // Calculate centroids
    CalculateCentroids(*data, instances, centroids, k, inputs, count, partition, index);

    for (int i = 0; i < k; i++) {
        int nearest;

        // Get nearest neighbor of non empty clusters and add to sample
        if (count[i]) {
            kNNCategory(centroids[i], *data, NULL, instances, 1, inputs, *labels, classes, &nearest, NULL, NULL, NULL, -1, NULL, NULL);

            MALLOC((*sampled_data)[*sampled_instances], scalar, inputs, "kMeansSample: (*sampled_data)[*sampled_instances]");
            memcpy((*sampled_data)[*sampled_instances], (*data)[nearest], inputs*sizeof(scalar));

            (*sampled_labels)[*sampled_instances] = (*labels)[nearest];

            (*sampled_instances)++;
        }
    }

    // If any empty realloc
    if (*sampled_instances != k) {
        REALLOC(*sampled_data, scalar *, *sampled_instances, "kMeansSample: *sampled_data");
        REALLOC(*sampled_labels, int, *sampled_instances, "kMeansSample: *sampled_labels");
    }

    // Check all classes present
    int ccount[classes];
    ZeroVector(NULL, ccount, classes);
    for (int i = 0; i < *sampled_instances; i++)
        ccount[(*sampled_labels)[i]]++;

    // If not present add a new member randomly
    for (int i = 0; i < classes; i++)
        if (!ccount[i]) {
            for (int j = 0; j < instances; j++)
                if ((*labels)[j] == i) {
                    // Add
                    REALLOC(*sampled_data, scalar *, *sampled_instances+1, "kMeansSample: *sampled_data");
                    REALLOC(*sampled_labels, int, *sampled_instances+1, "kMeansSample: *sampled_labels");

                    MALLOC((*sampled_data)[*sampled_instances], scalar, inputs, "kMeansSample: (*sampled_data)[*sampled_instances]");
                    memcpy((*sampled_data)[*sampled_instances], (*data)[j], inputs*sizeof(scalar));

                    (*sampled_labels)[*sampled_instances] = (*labels)[j];

                    (*sampled_instances)++;

                    break;  // Next class
                }
        }

    SHOWPROGRESS(parm->show_progress > 0, "Sampled using k-Means to %d instances\n", *sampled_instances);

    FreeScalarMatrix(&centroids, k);
    FREE(partition);
    FREE(count);

    // Free previous data?
    if (free_data) {
        FreeScalarMatrix(data, instances);
        FREE(*labels);
    }
}

void CalculateCentroids(scalar **data, int instances, scalar **centroids, int k, int inputs, int *count, int *partition, int **index)
{
    ZeroVector(NULL, count, k);
    ZeroMatrix(centroids, NULL, k, inputs);

    for (int i = 0; i < instances; i++) {
        count[partition[i]]++;
        if (!index)
            for (int j = 0; j < inputs; j++)
                centroids[partition[i]][j] += data[i][j];
        else {
            for (int j = 0; index[i][j] != -1; j++)
                centroids[partition[i]][index[i][j]-1] += data[i][j];
        }
    }

    for (int i = 0; i < k; i++)
        if (count[i])
            for (int j = 0; j < inputs; j++)
                centroids[i][j] /= count[i];
}

void StratifiedSample(scalar **data, int instances, int *labels, scalar **sampled_data, int *sampled_labels, int sampled_instances, int classes)
{
    int *class_list, class_count[classes], class_sample_count[classes];

    MALLOC(class_list, int, instances, "DataPartition: class_list");

    // Adjust sampling of every class
    ZeroVector(NULL, class_count, classes);
    ZeroVector(NULL, class_sample_count, classes);
    for (int i = 0; i < instances; i++)
        class_count[labels[i]]++;

    scalar ratio = ((scalar) sampled_instances)/instances;
    int c = 0;
    for (int i = 0; i < classes; i++)
        c += (class_sample_count[i] = class_count[i]*ratio);

    // Randomly fill
    while (c < sampled_instances) {
        int next;
        do {
            next = RANDOM(0, classes);
        } while (class_sample_count[next] == class_count[next]);

        class_sample_count[next]++;
        c++;

    }

    int next = 0;
    for (int i = 0; i < classes; i++) {
        int count = 0;
        for (int j = 0; j < instances; j++)
            if (labels[j] == i) {
                class_list[count] = j;
                count++;
            }
        RandomizeVector(class_list, count);

        for (int j = 0; j < class_sample_count[i]; j++) {
            sampled_data[next] = data[class_list[j]];
            sampled_labels[next] = labels[class_list[j]];
            next++;
        }
    }

    FREE(class_list);
}
