/***************************************************************************
 *   Copyright (C) 2014 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "data.h"
#include "macros.h"
#include "common.h"
#include "imbalance.h"
#include "maths.h"

// Perform a stratified partition

void DataPartition(const t_Parameters *parm, scalar **data, int instances, int *labels, int *list, int *n_strata, scalar ****stratum_data, int **stratum_instances, int ***stratum_labels, int classes,
                   int ***stratum_list, int **index, int ****stratum_index, int inputs, int stratum_size, t_Partition partition_method, bool stratified_approach, int ***stratum_S)
{
    int *class_list, *partition, *count;

    // Only one stratum? Do nothing
    if (*n_strata == 1) {
        for (int j = 0; j < instances; j++) {
            (*stratum_data)[0][j] = data[j];
            (*stratum_labels)[0][j] = labels[j];
            (*stratum_list)[0][j] = j;

            // Index?
            if (index)
                (*stratum_index)[0][j] = index[j];
        }

        (*stratum_instances)[0] = instances;

        return;
    }

    switch (partition_method) {
        case pRANDOM:
            // Initialize instance count per stratum
            for (int i = 0; i < *n_strata; i++)
                (*stratum_instances)[i] = 0;

            if (stratified_approach) {
                // Assign instances to stratum in a round-robin fashion starting from latest stratum to get balanced strata
                MALLOC(class_list, int, instances, "DataPartition: class_list");

                int next = RANDOM(0, *n_strata);
                for (int i = 0; i < classes; i++) {
                    int count = 0;
                    for (int j = 0; j < instances; j++)
                        if (labels[j] == i) {
                            class_list[count] = j;
                            count++;
                        }
                    RandomizeVector(class_list, count);

                    for (int j = 0; j < count; j++) {
                        (*stratum_data)[next][(*stratum_instances)[next]] = data[class_list[j]];
                        (*stratum_labels)[next][(*stratum_instances)[next]] = labels[class_list[j]];
                        (*stratum_list)[next][(*stratum_instances)[next]] = list ? list[class_list[j]] : class_list[j];

                        // Index?
                        if (index)
                            (*stratum_index)[next][(*stratum_instances)[next]] = index[class_list[j]];

                        (*stratum_instances)[next]++;

                        next = (next+1) % *n_strata;
                    }
                }

                FREE(class_list);
            } else {
                int size, c;

                MALLOC(class_list, int, instances, "RandomPartition: list");

                // Random partition
                for (int i = 0; i < instances; i++)
                    class_list[i] = i;
                RandomizeVector(class_list, instances);

                size = instances/(*n_strata);
                if (size > stratum_size)
                    OnErrorExit(ERR_UNSPECTED_RESULT, "Stratum size exceeded");
                c = 0;
                for (int i = 0; i < *n_strata; i++) {
                    for (int j = 0; j < size; j++) {
                        (*stratum_data)[i][j] = data[class_list[c]];
                        (*stratum_labels)[i][j] = labels[class_list[c]];
                        (*stratum_list)[i][j] = list ? list[class_list[c]] : class_list[c];

                        // Index?
                        if (index)
                            (*stratum_index)[i][j] = index[class_list[c]];

                        c++;
                    }
                    (*stratum_instances)[i] = size;
                }

                // Assign remaining randomly
                for (; c < instances; c++) {
                    int subset;

                    do {
                        subset = RANDOM(0, *n_strata);
                    } while ((*stratum_instances)[subset] == stratum_size);

                    (*stratum_data)[subset][(*stratum_instances)[subset]] = data[class_list[c]];
                    (*stratum_labels)[subset][(*stratum_instances)[subset]] = labels[class_list[c]];
                    (*stratum_list)[subset][(*stratum_instances)[subset]] = list ? list[class_list[c]] : class_list[c];

                    // Index?
                    if (index)
                        (*stratum_index)[subset][(*stratum_instances)[subset]] = index[class_list[c]];

                    (*stratum_instances)[subset]++;
                }

                FREE(class_list);
            }
            break;

        case pkMeans:
            if (stratified_approach)
                SHOWPROGRESS(parm->show_progress > 0, "WARNING: k-Means partition has no stratified version%c", '\n');
            MALLOC(partition, int, instances, "DataPartition: partition");
            MALLOC(count, int, *n_strata, "DataPartition: count");

            int previous_n_strata = *n_strata;
            DataDivisionkMeans(parm, partition, &count, data, instances, n_strata, inputs, parm->kmeans_max_iterations, index, stratum_size);

            // Resize if different for previous
            if (previous_n_strata != *n_strata) {
                // Free?
                for (int i = *n_strata; i < previous_n_strata; i++) {
                    FREE((*stratum_labels)[i]);
                    FREE((*stratum_list)[i]);
                    FREE((*stratum_data)[i]);
                    if (index)
                        FREE((*stratum_index)[i]);
                    if (stratum_S)
                        FREE((*stratum_S)[i]);
                }

                REALLOC(*stratum_instances, int, *n_strata, "DataPartition: *stratum_instances");
                REALLOC(*stratum_labels, int *, *n_strata, "DataPartition: *stratum_labels");
                REALLOC(*stratum_list, int *, *n_strata, "DataPartition: *stratum_list");
                REALLOC(*stratum_data, scalar **, *n_strata, "DataPartition: *stratum_data");
                if (index)
                    REALLOC(*stratum_index, int **, *n_strata, "DataPartition: *stratum_index");
                if (stratum_S)
                    REALLOC(*stratum_S, int *, *n_strata, "DataPartition: *stratum_S");

                // New?
                for (int i = previous_n_strata; i < *n_strata; i++) {
                    MALLOC((*stratum_labels)[i], int, stratum_size, "DataPartition: (*stratum_labels)[i]");
                    MALLOC((*stratum_list)[i], int, stratum_size, "DataPartition: (*stratum_list)[i]");
                    MALLOC((*stratum_data)[i], scalar *, stratum_size, "DataPartition: (*stratum_data)[i]");
                    if (index) {
                        MALLOC((*stratum_index)[i], int *, stratum_size, "DataPartition: (*stratum_index)[i]");
                    } else
                        (*stratum_index)[i] = NULL;
                    if (stratum_S)
                        MALLOC((*stratum_S)[i], int, stratum_size, "DataPartition: (*stratum_S)[i]");
                }
            }

            // Initialize instance count per stratum
            for (int i = 0; i < *n_strata; i++)
                (*stratum_instances)[i] = 0;

            for (int i = 0; i < instances; i++) {
                int next = partition[i];

                // If stratum already full assign to another one randomly
                while ((*stratum_instances)[next] >= stratum_size)
                    next = RANDOM(0, *n_strata);

                (*stratum_data)[next][(*stratum_instances)[next]] = data[i];
                (*stratum_labels)[next][(*stratum_instances)[next]] = labels[i];
                (*stratum_list)[next][(*stratum_instances)[next]] = list ? list[i] : i;

                // Index?
                if (index)
                    (*stratum_index)[next][(*stratum_instances)[next]] = index[i];

                (*stratum_instances)[next]++;
            }

            FREE(partition);
            FREE(count);
            break;

        case peBagging:
        case pkClustering:
        case pDecoupled:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "eBagging/k Clustering/Decoupled partitions");
            break;

        case pBOARD:
        case pRPROTOS:
        case p2DGrandTour:
            if (stratified_approach)
                SHOWPROGRESS(parm->show_progress > 0, "Board/Random prototypes/2D Grand tour partitions have no stratified version%c", '\n');
        case pOneInput:
        case pGrandTour:
        case pEQUAL:
            MALLOC(partition, int, instances, "DataPartition: partition");
            MALLOC(count, int, *n_strata, "DataPartition: count");

            if (stratified_approach) {
                if (partition_method == pEQUAL)
                    PartitionateEqualKCD(partition, count, data, instances, *n_strata, labels, inputs, classes);
                else
                    PartitionateOneInputKCD(parm, partition, count, data, instances, *n_strata, labels, inputs, classes, partition_method);
            } else {
                if (partition_method == pEQUAL)
                    PartitionateEqual(partition, count, data, instances, *n_strata, labels, inputs, classes, parm->use_pcomponents);
                else if (partition_method == pBOARD)
                    PartitionateBoard(partition, count, data, instances, *n_strata, inputs);
                else if (partition_method == pRPROTOS)
                    PartitionateRandomPrototypes(partition, count, data, instances, *n_strata, inputs, index);
                else if (partition_method == p2DGrandTour)
                    BidimensionalGrandTour((const t_Parameters *) parm, partition, count, data, instances, n_strata, labels, inputs, classes);
                else
                    PartitionateOneInput(parm, partition, count, data, instances, *n_strata, labels, inputs, classes, partition_method);
            }

            // Initialize instance count per stratum
            for (int i = 0; i < *n_strata; i++)
                (*stratum_instances)[i] = 0;

            for (int i = 0; i < instances; i++) {
                int next = partition[i];

                // If stratum already full assign to another one randomly
                while ((*stratum_instances)[next] >= stratum_size)
                    next = RANDOM(0, *n_strata);

                (*stratum_data)[next][(*stratum_instances)[next]] = data[i];
                (*stratum_labels)[next][(*stratum_instances)[next]] = labels[i];
                (*stratum_list)[next][(*stratum_instances)[next]] = list ? list[i] : i;

                // Index?
                if (index)
                    (*stratum_index)[next][(*stratum_instances)[next]] = index[i];

                (*stratum_instances)[next]++;
            }

            FREE(partition);
            FREE(count);

            break;
    }

    // Check if any stratum has empty classes, in such a case remove the stratum
    MALLOC(count, int, classes, "DataPartition: count");
    int removed = 0, orig = *n_strata;
    for (int i = 0; i < *n_strata; i++) {
        for (int j = 0; j < classes; j++)
            count[j] = 0;

        for (int j = 0; j < (*stratum_instances)[i]; j++)
            count[(*stratum_labels)[i][j]]++;

        bool remove = false;
        for (int j = 0; j < classes && !remove; j++)
            if (!count[j])
                remove = true;

        if (remove) {
            (*n_strata)--;
            (*stratum_data)[i] = (*stratum_data)[*n_strata];
            (*stratum_labels)[i] = (*stratum_labels)[*n_strata];
            (*stratum_list)[i] = (*stratum_list)[*n_strata];
            (*stratum_instances)[i] = (*stratum_instances)[*n_strata];

            if (index)
                (*stratum_index)[i] = (*stratum_index)[*n_strata];

            if (stratum_S)
                (*stratum_S)[i] = (*stratum_S)[*n_strata];

            // Revisit
            i--;
            removed++;
        }
    }

    FREE(count);
    fprintf(parm->out_file, "Removed %d strata (of %d) with empty classes\n", removed, orig);
}

void StratifiedRandomSample(scalar **data, int instances, int *labels, scalar ***s_data, int **s_labels, int s_instances, int classes, int **list, int **index, int ***s_index)
{
    int *class_list, class_count[classes], class_sample_count[classes];

    // Assign instances to stratum to get balanced strata
    MALLOC(class_list, int, instances, "DataPartition: class_list");

    // Adjust sampling of every class
    ZeroVector(NULL, class_count, classes);
    ZeroVector(NULL, class_sample_count, classes);
    for (int i = 0; i < instances; i++)
        class_count[labels[i]]++;

    scalar ratio = ((scalar) s_instances)/instances;
    int c = 0;
    for (int i = 0; i < classes; i++)
        c += (class_sample_count[i] = class_count[i]*ratio);

    // Randomly fill
    while (c < s_instances) {
        int next;
        do {
            next = RANDOM(0, classes);
        } while (class_sample_count[next] == class_count[next]);

        class_sample_count[next]++;
        c++;
    }

    MALLOC(*s_data, scalar *, s_instances, "StratifiedRandomSample: *s_data");
    MALLOC(*s_labels, int, s_instances, "StratifiedRandomSample: *s_labels");
    if (list)
        MALLOC(*list, int, s_instances, "StratifiedRandomSample: *list");
    if (index)
        MALLOC(*s_index, int *, s_instances, "StratifiedRandomSample: *s_index");

    int next = 0;
    for (int i = 0; i < classes; i++) {
        int count = 0;
        for (int j = 0; j < instances; j++)
            if (labels[j] == i) {
                class_list[count] = j;
                count++;
            }
        RandomizeVector(class_list, count);

        for (int j = 0; j < class_sample_count[i]; j++) {
            (*s_data)[next] = data[class_list[j]];
            (*s_labels)[next] = labels[class_list[j]];
            if (index)
                (*s_index)[next] = index[class_list[j]];
            if (list)
                (*list)[next] = class_list[j]; // Index of original instance
            next++;
        }
    }

    FREE(class_list);
}

void StratifiedPartition(int instances, int *labels, int n_strata, int classes, int *partition)
{
    int *list;

    MALLOC(list, int, instances, "StratifiedPartition: list");

    // Distribute by class
    for (int i = 0; i < classes; i++) {
        // List of instances of the class
        int count = 0;
        for (int j = 0; j < instances; j++)
            if (labels[j] == i) {
                list[count] = j;
                count++;
            }

        RandomizeVector(list, count);

        int next = RANDOM(0, n_strata);
        for (int j = 0; j < count; j++) {
            partition[list[j]] = next;
            next = (next+1)%n_strata;
        }
    }

    FREE(list);
}

// Reorder the dataset with samples distributed by classes

void StratifyDataset(scalar **data, int instances, int *labels, int **index, int classes)
{
    int *count, **list, *order, *selected;

    // Info about classes
    AllocateIntegerMatrix(&list, classes, instances);
    CALLOC(count, int, classes, "StratifyDataset: count");
    CALLOC(selected, int, classes, "StratifyDataset: selected");
    MALLOC(order, int, instances, "StratifyDataset; order");
    for (int i = 0; i < instances; i++)
        list[labels[i]][count[labels[i]]++] = i;

    int minority = MinimumOfIntegerVector(count, classes);

    // Assign in a round robin fashion
    int next = 0;
    for (int i = 0; i < minority; i++)
        for (int j = 0; j < classes; j++)
            for (int k = 0; k < count[j]/minority; k++)
                order[next++] = list[j][selected[j]++];

    // Assign remaining per class
    for (int j = 0; j < classes; j++)
        for (int k = selected[j]; k < count[j]; k++)
            order[next++] = list[j][k];

    // Reorder dataset
    int *o_labels, **o_index = NULL;
    scalar **o_data;
    MALLOC(o_labels, int, instances, "StratifyDataset: o_labels");
    if (index)
        MALLOC(o_index, int *, instances, "StratifyDataset: o_index");
    MALLOC(o_data, scalar *, instances, "StratifyDataset: o_data");
    memcpy(o_labels, labels, sizeof(int)*instances);
    if (index)
        memcpy(o_index, index, sizeof(int *)*instances);
    memcpy(o_data, data, sizeof(scalar *)*instances);
    for (int i = 0; i < instances; i++) {
        data[i] = o_data[order[i]];
        if (index)
            index[i] = o_index[order[i]];
        labels[i] = o_labels[order[i]];
    }

    FreeIntegerMatrix(&list, classes);
    FREE(count);
    FREE(order);
    FREE(selected);
    FREE(o_data);
    if (index)
        FREE(o_index);
    FREE(o_labels);
}
