/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "data.h"
#include "macros.h"
#include "common.h"
#include "data.h"

// Stratified subsampling of data
// Overwrites old data

void StratifiedRandomSubsampling(scalar ***data, int *instances, int **label, scalar ratio, int inputs, int classes)
{
    int selected, *list, *S;

    MALLOC(list, int, *instances, "StratifiedRandomSubsampling: list");
    CALLOC(S, int, *instances, "StratifiedRandomSubsampling: S");

    // Select the same ratio for each class
    int count;
    for (int i = 0; i < classes; i++) {
        // List of instances of the class
        count = 0;
        for (int j = 0; j < *instances; j++)
            if ((*label)[j] == i) {
                list[count] = j;
                count++;
            }

        RandomizeVector(list, count);

        for (int j = 0; j < ratio * count; j++)
            S[list[j]] = 1;
    }

    FREE(list);

    // Copy new sample onto old data
    selected = 0;
    for (int i = 0; i < *instances; i++)
        if (S[i]) {
            for (int j = 0; j < inputs; j++)
                (*data)[selected][j] = (*data)[i][j];
            (*label)[selected] = (*label)[i];
            selected++;
        }

    FREE(S);

    for (int i = selected; i < *instances; i++)
        FREE(data[i]);

    *instances = selected;
    REALLOC(*data, scalar *, selected, "StratifiedRandomSubsampling: *data");
    REALLOC(*label, int, selected, "StratifiedRandomSubsampling: *label");
}

// Undersample randomly the majority class until both classes have the desired ratio
// Data points to original vectors that cannot be freed
// Labels are copied

void TemporalUndersampleDataForTraining(scalar **data, int instances, int classes, int *labels, int ratio, scalar ***new_data, int **new_labels, int *new_instances, int minority)
{
    int count[] = {0, 0};

    // Only valid for two class problems
    if (classes > 2)
        OnErrorExit(ERR_PARAM, "Undersampling not valid for more than two classes");

    // Count instances
    for (int i = 0; i < instances; i++)
        count[labels[i]]++;

    int majority = (minority+1) % 2;

    // Check ratio is Ok and there are instances from both classes
    // Otherwise just copy all data.
    if (ratio*count[minority] >= count[majority] || count[0] == 0 || count[1] == 0) {
        MALLOC(*new_data, scalar *, instances, "TemporalUndersampleDataForTraining: new_data");
        MALLOC(*new_labels, int, instances, "TemporalUndersampleDataForTraining: new_labels");
        for (int i = 0; i < instances; i++) {
            (*new_data)[i] = data[i];
            (*new_labels)[i] = labels[i];
        }
        *new_instances = instances;
        return;
    }

    // List of majority and minority instances
    int *list_m, *list_M, M = 0, m = 0;
    MALLOC(list_M, int, count[majority], "TemporalUndersampleDataForTraining: list_M");
    MALLOC(list_m, int, count[minority], "TemporalUndersampleDataForTraining: list_m");
    for (int i = 0; i < instances; i++)
        if (labels[i] == majority) {
            list_M[M] = i;
            M++;
        } else {
            list_m[m] = i;
            m++;
        }

    // Space for new data
    MALLOC(*new_data, scalar *, (ratio+1) * m, "TemporalUndersampleDataForTraining: new_data");
    MALLOC(*new_labels, int, (ratio+1) * m, "TemporalUndersampleDataForTraining: new_labels");

    // Sample majority class and copy complete minority class
    RandomizeVector(list_M, M);

    *new_instances = 0;
    for (int i = 0; i < m; i++) {
        // Minority class instances
        (*new_data)[*new_instances] = data[list_m[i]];
        (*new_labels)[*new_instances] = minority;
        (*new_instances)++;

        // Majority class instances
        for (int j = 0; j < ratio; j++) {
            (*new_data)[*new_instances] = data[list_M[ratio*i+j]];
            (*new_labels)[*new_instances] = majority;
            (*new_instances)++;
        }
    }

    FREE(list_M);
    FREE(list_m);
}
