/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "data.h"
#include "macros.h"
#include "common.h"
#include "maths.h"

// Perform a partition keeping all the subsets local
void PartitionateEqual(int *partition, int *count, scalar **data, int instances, int subsets, int *label, int inputs, int classes, bool use_pcomponents)
{
    int *class_count, init, *features;
    RANK *to_sort = NULL;
    scalar **pc;

    // Memory allocation
    init = 0;
    MALLOC(class_count, int, classes, "PartitionateEqual: class_count");
    MALLOC(to_sort, RANK, instances, "PartitionateEqual: to_sort");
    MALLOC(features, int, inputs, "PartitionateEqual: features");
    if (use_pcomponents) {
        for (int i = 0; i < log2(subsets); i++)
            features[i] = i;
        RandomizeVector(features, log2(subsets));
    } else {
        for (int i = 0; i < inputs; i++)
            features[i] = i;
        RandomizeVector(features, inputs);
    }

    // Initialize partition
    for (int i = 0; i < instances; i++)
        partition[i] = 0;

    /////////////////////////////////////////////////////////////////////////////
    // Use principal components to partition
    /////////////////////////////////////////////////////////////////////////////
    if (use_pcomponents)
        PrincipalComponentsAnalysis((scalar **) data, instances, inputs, log2(subsets), &pc, NULL);

    /////////////////////////////////////////////////////////////////////////////
    // END Use principal components to partition
    /////////////////////////////////////////////////////////////////////////////

    // Divide by every variable, taken at random
    for (int i = 0; i < log2(subsets); i++) {
        // Pick random variable
        int variable = features[i];

        for (int k = 0; k < pow(2.0, i); k++) {
            int n = 0;

            for (int j = 0; j < instances; j++)
                if (partition[j] == k) {
                    if (use_pcomponents)
                        to_sort[n].fitness = pc[j][variable];
                    else
                        to_sort[n].fitness = data[j][variable];

                    to_sort[n].being = j;
                    n++;
                }
            qsort(to_sort, n, sizeof(RANK), Compare);

            // Divide using order
            int j = init+n/2;
            for (int l = 0; l < n/2; l++, j = (j+1) % n)
                partition[to_sort[j].being] += (int) pow(2.0, i);
        }
    }

    // Count
    for (int i = 0; i < subsets; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    FREE(class_count);
    FREE(to_sort);
    FREE(features);
}

// Perform a partition taking +/- planes parallel to axis
// Threshold is dynamically set to balance subsets
void PartitionateBoard(int *partition, int *count, scalar **data, int instances, int subsets, int inputs)
{
    int *features = NULL;
    RANK *to_sort;

    // Memory allocation
    MALLOC(features, int, inputs, "PartitionateBoard: features");
    MALLOC(to_sort, RANK, instances, "PartitionateBoard: to_sort");
    for (int i = 0; i < inputs; i++)
        features[i] = i;

    // Random order of variables
    RandomizeVector(features, inputs);

    // Initialize partition
    for (int i = 0; i < instances; i++)
        partition[i] = 0;

    // Divide by every variable, taken at random
    for (int i = 0; i < log2(subsets); i++) {
        // Pick random variable
        int variable = features[i];

        // Divide all previous subsets in two
        for (int k = 0; k < pow(2.0, (scalar) i); k++) {
            // Order instances
            int n = 0;

            for (int j = 0; j < instances; j++)
                if (partition[j] == k) {
                    to_sort[n].fitness = data[j][variable];
                    to_sort[n].being = j;
                    n++;
                }

            qsort(to_sort, n, sizeof(RANK), Compare);

            // Divide using threshold
            for (int j = 0; j < n/2; j++)
                partition[to_sort[j].being] += (int) pow(2.0, i);
        }
    }

    // Count
    for (int i = 0; i < subsets; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    FREE(features);
    FREE(to_sort);
}

// Perform a partition using random prototypes as centers
void PartitionateRandomPrototypes(int *partition, int *count, scalar **data, int instances, int subsets, int inputs, int **index)
{
    scalar **prototypes;
    RANK *to_sort;
    int **protos_index = NULL;

    // Set approximate size of subsets
    int size = instances/subsets;
    if (instances % size)
        size++;

    for (int i = 0; i < subsets; i++)
        count[i] = 0;

    MALLOC(to_sort, RANK, instances, "PartitionateRandomPrototypes: to_sort");

    // All initially unassigned
    for (int i = 0; i < instances; i++)
        partition[i] = -1;

    MALLOC(prototypes, scalar *, subsets, "PartitionateRandomPrototypes: prototypes");
    for (int i = 0; i < subsets; i++)
        MALLOC(prototypes[i], scalar, inputs, "PartitionateRandomPrototypes: prototypes[i]");

    if (index) {
        MALLOC(protos_index, int *, subsets, "PartitionateRandomPrototypes: protos_index");
        for (int i = 0; i < subsets; i++)
            MALLOC(protos_index[i], int, inputs+1, "PartitionateRandomPrototypes: protos_index[i]");
    }

    // Create prototypes selecting random values
    for (int k = 0; k < inputs; k++) {
        // Get boundaries
        scalar min = data[0][k];
        scalar max = data[0][k];
        for (int i = 1; i < instances; i++)
            if (data[i][k] < min)
                min = data[i][k];
            else if (data[i][k] > max)
                max = data[i][k];

        for (int i = 0; i < subsets; i++)
            prototypes[i][k] = (scalar) RANDOM(min, max);
    }

    // Convert to sparse?
    if (index) {
        for (int i = 0; i < subsets; i++) {
            int c = 0;
            for (int j = 0; j < inputs; j++)
                if (fabs(prototypes[i][j]) > ZERO) {
                    prototypes[i][c] = prototypes[i][j];
                    protos_index[i][c] = j+1;
                    c++;
                }
            protos_index[i][c] = -1;
        }
    }

    // Assign each instance
    // Shrink or expand neighborhood to obtain equal sized subsets
    for (int j = 0; j < subsets; j++) {
        int c = 0;
        for (int i = 0; i < instances; i++)
            if (partition[i] == -1) {
                to_sort[c].fitness = EuclideanDistanceSparseData(data[i], prototypes[j], inputs, index ? index[i] : NULL, index ? protos_index[j] : NULL);
                to_sort[c].being = i;
                c++;
            }

        qsort(to_sort, c, sizeof(RANK), Compare);

        for (int i = 0; i < size && i < c; i++) {
            partition[to_sort[i].being] = j;
            count[j]++;
        }
    }

    FREE(to_sort);
    for (int i = 0; i < subsets; i++)
        FREE(prototypes[i]);
    FREE(prototypes);
    if (index) {
        for (int i = 0; i < subsets; i++)
            FREE(protos_index[i]);
        FREE(protos_index);
    }
}


void FusionPartitions(int *old_partition, int *new_partition, int *n_subsets, int n, int *count)
{
    int size, c, old_n_subsets;

    // Count subsets
    old_n_subsets = 0;
    for (int i = 0; i < n; i++)
        if (old_partition[i] > old_n_subsets)
            old_n_subsets = old_partition[i];
    old_n_subsets++;

    // Set approximate size of subsets
    size = n/(*n_subsets);

    int subset = 0, current = 0;
    while (subset < old_n_subsets) {
        c = 0;
        while (c < size && subset < old_n_subsets) {
            // Add subset to current one
            for (int i = 0; i < n; i++)
                if (old_partition[i] == subset) {
                    new_partition[i] = current;
                    c++;
                }
            subset++;
        }
        count[current] = c;
        current++;
    }

    *n_subsets = current;
}

void RandomPartition(int *partition, int *count, int n, int n_subsets)
{
    int *list;
    int size, c;

    MALLOC(list, int, n, "RandomPartition: list");

    // Random partition
    for (int i = 0; i < n; i++)
        list[i] = i;
    RandomizeVector(list, n);

    size = n/n_subsets;
    c = 0;
    for (int i = 0; i < n_subsets; i++) {
        for (int j = 0; j < size; j++) {
            partition[list[c]] = i;
            c++;
        }
        count[i] = size;
    }

    // Assign remaining Round Robin
    int subset = RANDOM(0, n_subsets);
    for (; c < n; c++) {
        partition[list[c]] = subset;
        count[subset]++;
        subset = (subset+1) % n_subsets;
    }

    FREE(list);
}

void PartitionateOneInput(const t_Parameters *parm, int *partition, int *count, scalar **data, int instances, int subsets, int *label, int inputs, int classes, t_Partition partition_method)
{
    int *class_count = NULL,  init = -1;
    RANK *to_sort;
    static scalar *coeffs = NULL, *old;

    // Memory allocation
    MALLOC(class_count, int, classes, "PartitionateOneInput: class_count");
    MALLOC(to_sort, RANK, instances, "PartitionateOneInput: to_sort");
    MALLOC(old, scalar, inputs, "PartitionateOneInput");

    if (partition_method == pGrandTour) {
        // Random projection
        if (coeffs == NULL) {
            MALLOC(coeffs, scalar, inputs, "PartitionateOneInput: coeffs");
            // Use first component
            if (parm->use_pcomponents)
                PrincipalComponentsAnalysis((scalar **) data, instances, inputs, 1, NULL, &coeffs);
            else {
                // or random coefficients for the linear projection
                for (int i = 0; i < inputs; i++)
                    coeffs[i] = RANDOM(-1.0, 1.0);

                // Normalize vector
                normalize(inputs, coeffs);
            }
            getVec(INIT, inputs, parm->angle_step);       // Init the vec generation dim: inputs, step: degrees
            getVec(SET, coeffs);        // Set the initial direction of the vector
            getVec(NEXT, coeffs);       // Ignore first one
        } else {
            if (parm->use_pcomponents)
                PrincipalComponentsAnalysis((scalar **) data, instances, inputs, 1, NULL, &coeffs);
            else {
                // or random coefficients for the linear projection
                for (int i = 0; i < inputs; i++)
                    coeffs[i] = RANDOM(-1.0, 1.0);

                // Normalize vector
                normalize(inputs, coeffs);
            }

            getVec(CLEAR);
            getVec(INIT, inputs, parm->angle_step);
            getVec(SET, coeffs);    // Set the initial direction of the vector
            getVec(NEXT, coeffs);      // Ignore first one
        }

        scalar value;
        for (int j = 0; j < instances; j++) {
            // Projection
            value = 0.0;
            for (int i = 0; i < inputs; i++)
                value += data[j][i] * coeffs[i];
            to_sort[j].fitness = value;
            to_sort[j].being = j;
        }
        qsort(to_sort, instances, sizeof(RANK), Compare);
    } else {
        int variable = RANDOM(0, inputs);
        for (int j = 0; j < instances; j++) {
            to_sort[j].fitness = data[j][variable];
            to_sort[j].being = j;
        }
        qsort(to_sort, instances, sizeof(RANK), Compare);
    }

    int size = instances/subsets;
    init = 0;
    int j = init;
    for (int l = 0; l < subsets * size; l++, j = (j+1) % instances)
        partition[to_sort[j].being] = l/size;
    // Assign remaining randomly
    for (int l = size * subsets; l < instances; l++, j = (j+1) % instances) {
        int subset = RANDOM(0, subsets);
        partition[to_sort[j].being] = subset;
    }

    //////////////////////////////////////////////////////////
    // Partition following Cesar's idea
    //////////////////////////////////////////////////////////
    /*    int *sizes = Partition(subsets, instances, 1.0);
        int c = 0;
        for (int i = 0; i < subsets; i++)
            for (int j = 0; j < sizes[i]; j++) {
                partition[to_sort[c].being] = i;
                c++;
            }
        FREE(sizes); */

    //////////////////////////////////////////////////////////
    // END
    //////////////////////////////////////////////////////////

    // Count
    for (int i = 0; i < subsets; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[partition[i]]++;

    FREE(class_count);
    FREE(to_sort);
    FREE(old);
}

// Mix subsets of a previous partition pairwise

void MixSubsets(int *new_partition, int *old, int *count, int n_subsets, int n)
{
    int *list;

    ALLOCA(list, int, n_subsets, "MixSubsets: list");

    // Random order of subsets
    for (int i = 0; i < n_subsets; i++)
        list[i] = i;
    RandomizeVector(list, n_subsets);

    // For every pair of subsets
    for (int i = 0; i < n_subsets; i += 2)
        for (int j = 0; j < n; j++)
            if (old[j] == list[i] || old[j] == list[i+1])
                new_partition[j] = RANDOM(0.0, 1.0) < 0.5 ? list[i] : list[i+1];

    // Count
    for (int i = 0; i < n_subsets; i++)
        count[i] = 0;
    for (int i = 0; i < n; i++)
        count[new_partition[i]]++;
}
