/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ensemble.h"

t_Boosting *TrainAdaBoost(const t_Parameters *parm, scalar **data, int instances, int *labels, int inputs, int classes, int **index, int *subset, int *subspace, bool madaboost, bool multiboost,
                          bool gentleboost, bool reweightboost, bool arcing)
{
    t_Boosting *boost;
    scalar **l_data;
    int *l_labels,  **l_index = NULL, *l_subset = NULL, *I = NULL, k = 1;

    // Initialize?
    if (multiboost) {
        int n = sqrt((scalar) parm->ensemble_size);
        MALLOC(I, int, parm->ensemble_size+1, "UpdateMultiBoostWeightsIS");

        I[0] = 0;
        for (int i = 1; i < n; i++)
            I[i] = ceil(i * parm->ensemble_size/n);
        for (int i = n; i < parm->ensemble_size+1; i++)
            I[i] = parm->ensemble_size;
    }

    MALLOC(l_data, scalar *, instances, "TrainAdaBoost: l_data");
    MALLOC(l_labels, int, instances, "TrainAdaBoost: l_labels");
    if (subset)
        MALLOC(l_subset, int, instances, "TrainAdaBoost: l_subset");
    if (index)
        MALLOC(l_index, int *, instances, "TrainAdaBoost: l_index");

    // Initialize ensemble
    MALLOC(boost, t_Boosting, 1, "TrainAdaBoost: boost");
    MALLOC(boost->alpha, scalar, parm->ensemble_size, "TrainAdaBoost: boost->alpha");
    MALLOC(boost->bcs, t_Learner, parm->ensemble_size, "TrainAdaBoost: boost->bcs");

    boost->T = parm->ensemble_size;
    boost->r = reweightboost ? parm->reweightboost_r : 0;
    boost->base = parm->base_learner;
    FillVector(boost->alpha, NULL, boost->T, 1.0);

    // First round with original data
    for (int i = 0; i < instances; i++) {
        l_data[i] = data[i];
        l_labels[i] = labels[i];
        if (subset)
            l_subset[i] = subset[i];
        if (index)
            l_index[i] = index[i];
    }

    // Iterative training of base classifiers, Webb (2000) model
    scalar *w, *cdf;
    MALLOC(w, scalar, instances, "TrainAdaBoost: w");
    MALLOC(cdf, scalar, instances, "TrainAdaBoost: w");
    FillVector(w, NULL, instances, 1.0/instances);

    for (int t = 0; t < boost->T; t++) {
        ClassifierTraining(parm, parm->base_learner, l_data, instances, l_labels, inputs, classes, l_index, subspace, l_subset, true, &(boost->bcs[t]), NULL);

        // Update weights
        if (gentleboost)
            UpdateGentleBoostWeights(parm, boost->base, inputs, classes, index, subspace, data, instances, labels, subset, w, &(boost->bcs[t]));
        else if (reweightboost)
            GenerateReweightBoostSample(parm, boost, w, w, &(boost->alpha[t]), instances, classes, data, labels, inputs, index, subspace, subset, t);
        else if (arcing)
            GenerateArcingSample(parm, boost, w, instances, classes, data, labels, inputs, boost->base, index, t, subspace, subset);
        else
            GenerateAdaSample(parm, &(boost->bcs[t]), w, w, &(boost->alpha[t]), instances, classes, data, labels, inputs, boost->base, index, subspace, subset, madaboost);

        // CDF function
        cdf[0] = w[0];
        for (int i = 1; i < instances; i++)
            cdf[i] = cdf[i-1]+w[i];

        // With MultiBoost reinitialize at intervals
        if (multiboost && I[k] == t) {
            for (int i = 0; i < instances; i++)
                w[i] = Poisson();
            NormalizeVector(w, instances);
            k++;
        }

        // Resample for next round
        for (int i = 0; i < instances; i++) {
            int next = RouletteSelection(cdf, instances);
            l_data[i] = data[next];
            l_labels[i] = labels[next];
            if (subset)
                l_subset[i] = subset[next];
            if (index)
                l_index[i] = index[next];
        }
    }

    FREE(w);
    FREE(cdf);
    FREE(l_data);
    FREE(l_labels);
    if (subset)
        FREE(l_subset);
    if (index)
        FREE(l_index);
    if (multiboost)
        FREE(I);

    return boost;
}

void GenerateAdaSample(const t_Parameters *parm, t_Learner *learner, scalar *w_t, scalar *w_t_1, scalar *alpha, int instances, int classes, scalar **data, int *labels, int inputs, t_Classifier method,
                       int **index, int *subspace, int *subset, bool madaboost)
{
    bool *ok;

    MALLOC(ok, bool, instances, "GenerateAdaSample: ok");

    // Classify each pattern
    scalar e_i = 0.0;

    for (int j = 0; j < instances; j++)
        if (w_t[j] > 0.0) {
            ok[j] = (ClassifierOutput(parm, method, data[j], inputs, classes, index ? index[j] : NULL, subspace, data, instances, labels, subset, learner, labels[j], NULL, -1, index, NULL)
                == labels[j]);

            if (!ok[j])
                e_i += w_t[j];
        }

    e_i /= instances;

    // Update weights
    if (e_i > 0.5) {
        *alpha = MIN_ALPHA;
        // New bootstrap sample
        for (int k = 0; k < instances; k++)
            w_t_1[k] = Poisson();
    } else if (e_i == 0.0) {
        // New bootstrap sample
        *alpha = MAX_ALPHA;
        // New bootstrap sample
        for (int k = 0; k < instances; k++)
            w_t_1[k] = Poisson();
    } else {
        *alpha = (1.0-e_i)/e_i;
        *alpha = AssignWithinBoundsScalar(MIN_ALPHA, *alpha, MAX_ALPHA);

        // Get new weights
        if (madaboost) {
            // MadaBoost variant
            for (int i = 0; i < instances; i++)
                if (ok[i]) {
                    w_t_1[i] = w_t[i]*sqrt(1.0/(*alpha));
                    if (w_t_1[i] < 1.0e-10)
                        w_t_1[i] = 1.0e-10;
                }
        } else {
            for (int i = 0; i < instances; i++) {
                if (ok[i])
                    w_t_1[i] = w_t[i]*sqrt(1.0/(*alpha));
                else
                    w_t_1[i] = w_t[i]*sqrt(*alpha);
                if (w_t_1[i] < 1.0e-10)
                    w_t_1[i] = 1.0e-10;
            }
        }
    }

    // Normalize to sum N
    NormalizeVector(w_t_1, instances);

    FREE(ok);
}

void FreeAdaBoost(t_Boosting *boost, const t_Parameters *parm, int classes, int inputs)
{
    FREE(boost->alpha);
    for (int t = 0; t < boost->T; t++)
        ClassifierFree(parm, &(boost->bcs[t]), boost->base, classes, inputs, NULL);

    FREE(boost->bcs);
}

int AdaBoostOutput(const t_Parameters *parm, t_Boosting *boost, scalar *instance, int inputs, int classes, int *index, int *subspace, scalar **data, int instances, int *labels, int *subset, int label,
                   int **data_index, scalar *probs)
{
    scalar votes[classes];
    ZeroVector(votes, NULL, classes);

    for (int t = 0; t < boost->T; t++) {
        int category = ClassifierOutput(parm, boost->base, instance, inputs, classes, index, subspace, data, instances, labels, subset, &(boost->bcs[t]), label, NULL, -1, data_index, NULL);

        votes[category] += boost->alpha[t];
    }

    if (probs) {
        memcpy(probs, votes, sizeof(scalar)*classes);
        NormalizeVector(probs, classes);
    }

    return IndexOfMaximumOfScalarVector(votes, classes);
}

scalar AdaBoostMargin(const t_Parameters *parm, t_Boosting *boost, scalar *instance, int label, int classes, int inputs, int *subspace, int *index, scalar **data, int instances, int *labels,
                      int *subset, int **data_index, bool reweightboost)
{
    int max_index;
    scalar max_value, margin, outputs[classes];

    if (reweightboost)
        ReweightBoostOutput(parm, boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, label, data_index, outputs);
    else
        AdaBoostOutput(parm, boost, instance, inputs, classes, index, subspace, data, instances, labels, subset, label, data_index, outputs);

    // Get classification
    max_index = IndexOfMaximumOfScalarVector(outputs, classes);
    max_value = outputs[max_index];

    // Test if correctly Classified.
    if (label == max_index) {
        // Difference with second one
        margin = max_value;

        // Find second highest
        outputs[max_index] = -100000;
        max_index = 0;
        max_value = outputs[0];
        for (int j = 1; j < classes; j++)
            if (outputs[j] > max_value) {
                max_value = outputs[j];
                max_index = j;
            }
        margin -= max_value;

    } else
        // Difference with correct one
        margin = outputs[label]-max_value;

    return margin;
}

// ARCING Method: L. Breiman (1998) "Arcing classifiers," The Annals of Statistics 26(3):801-824.

void GenerateArcingSample(const t_Parameters *parm, t_Boosting *boost, scalar *w_t, int instances, int classes, scalar **data, int *labels, int inputs, t_Classifier method, int **index, int T,
                          int *subspace, int *subset)
{
    scalar *misses;

    MALLOC(misses, scalar, instances, "GenerateArcingSample: misses");
    ZeroVector(misses, NULL, instances);

    // Classify each pattern
    for (int j = 0; j < instances; j++)
        for (int t = 0; t <= T; t++)
            misses[j] += (ClassifierOutput(parm, parm->base_learner, data[j], inputs, classes, index ? index[j] : NULL, subspace, data, instances, labels, subset, &(boost->bcs[t]), labels[j], NULL,
                                           -1, index, NULL) != labels[j]);

    for (int j = 0; j < instances; j++)
        w_t[j] = 1.0 + pow(misses[j], 4.0);

    // Normalize to sum N
    NormalizeVector(w_t, instances);

    FREE(misses);
}
