/***************************************************************************
 *   Copyright (C) 2007 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <time.h>
#include "ensemble.h"
#include "macros.h"
#include "maths.h"

// Summary of changes.

void BP(const t_Parameters *parm, t_ClassModel *classifier, int index, int n_epochs, int epoch_size, scalar **data, int patterns, scalar **desired, int inputs, int classes, int **d_index)
{
    int pattern;
    scalar sum, change, weight, *w, **training_data = NULL, **mom_hid, **mom_out, **change_ih, **change_ho, *error, *output, *delta, *cum_probs;

    MALLOC(mom_hid, scalar *, classifier->nodes, "BP: mom_hid");
    for (int i = 0; i < classifier->nodes; i++)
        MALLOC(mom_hid[i], scalar, inputs, "BP: mom_hid[i]");
    MALLOC(mom_out, scalar *, classes, "BP: mom_out");
    for (int i = 0; i < classes; i++)
        MALLOC(mom_out[i], scalar, classifier->nodes, "BP: mom_out[i]");
    MALLOC(change_ih, scalar *, classifier->nodes, "BP: change_ih");
    for (int i = 0; i < classifier->nodes; i++)
        MALLOC(change_ih[i], scalar, inputs, "BP: change_ih[i]");
    MALLOC(change_ho, scalar *, classes, "BP: change_ho");
    for (int i = 0; i < classes; i++)
        MALLOC(change_ho[i], scalar, classifier->nodes, "BP: change_ho[i]");

    // Momentum term
    for (int i = 0; i < classifier->nodes; i++)
        for (int j = 0; j < classifier->inputs; j++)
            mom_hid[i][j] = 0.0;

    for (int i = 0; i < classes; i++)
        for (int j = 0; j < classifier->nodes; j++)
            mom_out[i][j] = 0.0;

    // Cumulative changes
    for (int i = 0; i < classifier->nodes; i++)
        for (int j = 0; j < classifier->inputs; j++)
            change_ih[i][j] = 0.0;

    for (int i = 0; i < classes; i++)
        for (int j = 0; j < classifier->nodes; j++)
            change_ho[i][j] = 0.0;

    MALLOC(error, scalar, classes+classifier->nodes, "BP: error");
    MALLOC(delta, scalar, classes+classifier->nodes, "BP: delta");
    MALLOC(output, scalar, classes+classifier->nodes, "BP: output");
    MALLOC(cum_probs, scalar, patterns, "BP: cum_probs");

    if (!index) {
        // Data for training the network
        MALLOC(training_data, scalar *, patterns, "BP: training_data");
        //     MALLOC(w, scalar, patterns, "BP: w");
        for (int i = 0; i < patterns; i++)
            MALLOC(training_data[i], scalar, inputs, "BP: training_data[i]");

    // If new data, recreate training data
#ifdef NON_STANDARD
        if (parm->projection && (index || parm->projection_algorithm == PCA)) {
            for (int k = 0; k < patterns; k++)
                // Projection
                ProjectInstance(parm, training_data[k], data[k], classifier, ensemble, inputs, index);
        } else {
#endif
            for (int j = 0; j < patterns; j++)
                for (int i = 0; i < classifier->inputs; i++)
                    training_data[j][i] = data[j][i]*classifier->subspace[i];
#ifdef NON_STANDARD
        }
#endif
    }

    // Select sampling weights
    scalar *equal = NULL;
    if (parm->projection) {
        MALLOC(equal, scalar, (unsigned) patterns, "BP: equal");
        for (int i = 0; i < patterns; i++)
            equal[i] = 1.0;
        w = equal;
        ((t_Parameters *)parm)->ensemble_technique = Reweighting;
    } else
        w = classifier->sample;

    // Cumulative probabilities for sampling.
    if (parm->ensemble_technique == Resampling) {
        cum_probs[0] = w[0];
        for (int i = 1; i < patterns; i++)
            cum_probs[i] = cum_probs[i-1]+w[i];
    }
    // Iterate
    weight = 1.0;
    for (int iter = 0; iter < n_epochs; iter++) {
        for (int k = 0; k < epoch_size; k++) {
            // Choose random pattern
            if (parm->ensemble_technique == Resampling)
                pattern = RouletteSelection(cum_probs, patterns);
            else {
                pattern = RANDOM(0, patterns);
                weight = w[pattern];
            }

            for (int i = 0; i < classifier->nodes-1; i++) {
                sum = 0.0;
                if (index)
                    for (int j = 0; d_index[pattern][j] != -1; j++)
                        sum += classifier->ih[i][d_index[pattern][j]-1]*training_data[pattern][j];
                else
                    for (int j = 0; j < classifier->inputs; j++)
                        sum += classifier->ih[i][j]*training_data[pattern][j];

                output[i] = TransferFunction(sum, parm->transfer_h, parm->a, parm->b);
            }

            // Bias input to output nodes
            output[classifier->nodes-1] = 0.0;

            for (int i = 0; i < classes; i++) {
                sum = 0.0;
                for (int j = 0; j < classifier->nodes; j++)
                    sum += output[j]*classifier->ho[i][j];
                output[classifier->nodes+i] = TransferFunction(sum, parm->transfer_o, parm->a_output, parm->b_output);
            }

            // Obtain error for output nodes.
            for (int i = 0; i < classes; i++)
                error[i] = weight* (desired[pattern][i]-output[classifier->nodes+i]);

            // Compute delta's for output and hidden nodes.
            if (parm->transfer_o == Log)
                for (int i = classifier->nodes+classes-1; i >= classifier->nodes; i--)
                    delta[i] = error[i-classifier->nodes]*(parm->flat_spot+parm->b_output*output[i]* (1.0-output[i]/parm->a_output));
            else
                for (int i = classifier->nodes+classes-1; i >= classifier->nodes; i--)
                    delta[i] = error[i-classifier->nodes]*(parm->flat_spot+ (parm->b_output/parm->a_output) * (parm->a_output-output[i]) * (parm->a_output+output[i]));

            if (parm->transfer_h == Log) {
                for (int i = classifier->nodes-2; i >= 0; i--) {
                    delta[i] = 0.0;
                    for (int j = 0; j < classes; j++)
                        delta[i] += delta[classifier->nodes+j]*classifier->ho[j][i];

                    delta[i] *= parm->flat_spot+parm->b*output[i]* (1.0-output[i]/parm->a);
                }
            } else {
                for (int i = classifier->nodes-2; i >= 0; i--) {
                    delta[i] = 0.0;
                    for (int j = 0; j < classes; j++)
                        delta[i] += delta[classifier->nodes+j]*classifier->ho[j][i];
                    delta[i] *= parm->flat_spot+ (parm->b/parm->a) * (parm->a-output[i]) * (parm->a+output[i]);
                }
            }

            // Obtain weight update
            // Input
            for (int i = 0; i < classifier->nodes-1; i++)
                if (index)
                    for (int j = 0; d_index[pattern][j] != -1; j++)
                        change_ih[i][j] += parm->eta*delta[i]*training_data[pattern][j]-parm->lambda*PENALTY(classifier->ih[i][d_index[pattern][j]-1]);
                else
                    for (int j = 0; j < classifier->inputs; j++)
                        change_ih[i][j] += parm->eta*delta[i]*training_data[pattern][j]-parm->lambda*PENALTY(classifier->ih[i][j]);

            // Output
            for (int i = 0; i < classes; i++)
                for (int j = 0; j < classifier->nodes; j++)
                    change_ho[i][j] += parm->eta*delta[i+classifier->nodes]*output[j]-parm->lambda*PENALTY(classifier->ho[i][j]);
        }

        // Perform weight update
        // Input -> Hidden
        for (int i = 0; i < classifier->nodes-1; i++)
            for (int j = 0; j < classifier->inputs; j++) {
                change = change_ih[i][j]+parm->alpha*mom_hid[i][j];
                mom_hid[i][j] = change;
                change_ih[i][j] = 0.0;
                classifier->ih[i][j] += change;
            }

        // Hidden -> Output
        for (int i = 0; i < classes; i++)
            for (int j = 0; j < classifier->nodes; j++) {
                change = change_ho[i][j]+parm->alpha*mom_out[i][j];
                mom_out[i][j] = change;
                change_ho[i][j] = 0.0;
                classifier->ho[i][j] += change;
            }
    }

    for (int i = 0; i < classifier->nodes; i++)
        FREE(mom_hid[i]);
    for (int i = 0; i < classes; i++)
        FREE(mom_out[i]);
    for (int i = 0; i < classifier->nodes; i++)
        FREE(change_ih[i]);
    for (int i = 0; i < classes; i++)
        FREE(change_ho[i]);
    if (!index) {
        for (int i = 0; i < patterns; i++)
            FREE(training_data[i]);
        FREE(training_data);
    }
    FREE(mom_hid);
    FREE(mom_out);
    FREE(change_ih);
    FREE(change_ho);
    if (equal != NULL)
        FREE(equal);
    FREE(error);
    FREE(delta);
    FREE(output);
    FREE(cum_probs);
}
