/***************************************************************************
 *   Copyright (C) 2007 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Basic functions for dealing with classifiers
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <features.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include "macros.h"
#include "ensemble.h"


// Allocate memory for a classifier
void AllocateClassifier(const t_Parameters *parm, t_ClassModel *classifier, int inputs, int nodes, bool alloc_net, int patterns, int classes, int dimension)
{
    if (alloc_net) {
        // Allocate memory for the new network
        MALLOC(classifier->ih, scalar *, nodes, "AllocateClassifier: classifier->ih");
        MALLOC(classifier->ho, scalar *, classes, "AllocateClassifier: classifier->ho");

        for (int i = 0; i < nodes; i++)
            MALLOC(classifier->ih[i], scalar, inputs, "AllocateClassifier: classifier->ih[i]");
        for (int i = 0; i < classes; i++)
            MALLOC(classifier->ho[i], scalar, nodes, "AllocateClassifier: classifier->ho[i]");

        classifier->nodes = nodes;
    } else
        classifier->ih = classifier->ho = NULL;

    MALLOC(classifier->subspace, scalar, inputs, "AllocateClassifier: classifier->subspace");
    for (int i = 0; i < inputs; i++)
        classifier->subspace[i] = 1.0;
    MALLOC(classifier->sample, scalar, patterns, "AllocateClassifier: classifier->sample");

#ifdef NON_STANDARD
    if ((parm->projection && dimension != -1 && (parm->projection_algorithm == BPStd || parm->projection_algorithm == EDBD || parm->projection_algorithm == RProp || parm->projection_algorithm == PCA ||
            parm->projection_algorithm == TryAll || parm->projection_algorithm == RCGA)) || parm->ensemble_method == RandomProj) {
        MALLOC(classifier->nlp, scalar *, dimension+1, "AllocateClassifier: classifier->nlp");
        for (int i = 0; i < dimension; i++)
            MALLOC(classifier->nlp[i], scalar, inputs+1, "AllocateClassifier: classifier->nlp[i]");
    }
    classifier->h = parm->projection_transfer;
#else
    classifier->nlp = NULL;
    classifier->grnn = NULL;
    classifier->projnet = NULL;
    classifier->scca_file = classifier->scca_data = NULL;
#endif

    if (dimension > 0)
        classifier->inputs = dimension;
    else
        classifier->inputs = inputs;

    classifier->theta = 0.0;
}

void FreeClassifier(const t_Parameters *parm, t_ClassModel *classifier, int classes)
{
    if (classifier->ih != NULL) {
        for (int i = 0; i < classifier->nodes; i++)
            FREE(classifier->ih[i]);
        for (int i = 0; i < classes; i++)
            FREE(classifier->ho[i]);
        FREE(classifier->ih);
        FREE(classifier->ho);
    }

    FREE(classifier->sample);
    FREE(classifier->subspace);
#ifdef NON_STANDARD
    if ((parm->projection && (parm->projection_algorithm == BPStd || parm->projection_algorithm == EDBD || parm->projection_algorithm == RProp ||
                              parm->projection_algorithm == TryAll || parm->projection_algorithm == RCGA)) || parm->ensemble_method == RandomProj) {
        for (int i = 0; i < classifier->inputs; i++)
            FREE(classifier->nlp[i]);
        FREE(classifier->nlp);
    }
#endif
}


// Generate a new classifier
void GenerateNewClassifier(const t_Parameters *parm, t_ClassModel *classifier, bool init_net, int patterns, int inputs, int classes)
{
    scalar init_weight;

    // Generate new structure
    classifier->beta = 1.0;
    classifier->w = 1.0;
    //     classifier->inputs = inputs;

#ifdef NON_STANDARD
    // NLP
    classifier->a = parm->projection_a;
    classifier->b = parm->projection_b;
#endif

    // Each connection is generated
    if (init_net) {
        if (parm->init_weight == -1.0) {
            for (int i = 0; i < classifier->nodes; i++)
                for (int j = 0; j < classifier->inputs; j++)
                    classifier->ih[i][j] = Exponential();
        } else {
            if (parm->init_weight <= NEARZERO)
                init_weight = sqrt(3.0) /classifier->inputs;
            else
                init_weight = parm->init_weight;
            for (int i = 0; i < classifier->nodes; i++)
                for (int j = 0; j < classifier->inputs; j++)
                    classifier->ih[i][j] = RANDOM(-init_weight, init_weight);
        }
        if (parm->init_weight == -1.0) {
            for (int i = 0; i < classes; i++)
                for (int j = 0; j < classifier->nodes; j++)
                    classifier->ho[i][j] = Exponential();
        } else {
            if (parm->init_weight <= NEARZERO)
                init_weight = sqrt(3.0) /classifier->inputs;
            else
                init_weight = parm->init_weight;
            for (int i = 0; i < classes; i++)
                for (int j = 0; j < classifier->nodes; j++)
                    classifier->ho[i][j] = RANDOM(-init_weight, init_weight);
        }
    }
    // Equal sample
    for (int i = 0; i < patterns; i++)
        classifier->sample[i] = 1.0;

    // Subspace ...
    if (parm->subspace_method == RSM)
        // random
        for (int i = 0; i < inputs; i++)
            classifier->subspace[i] = RANDOM(0.0, 1.0) < parm->subspace_prob ? 1.0 : 0.0;
    else
        // complete
        for (int i = 0; i < inputs; i++)
            classifier->subspace[i] = 1.0;
}

int NetworkActualOutput(const t_Parameters *parm, t_ClassModel *classifier, scalar *input, scalar *output, int classes, int *index)
{
    int max;
    scalar x[classifier->nodes];

#ifdef NON_STANDARD
    scalar nlp[classifier->inputs];
    if (parm->projection && (index || parm->projection_algorithm == PCA)) {
        // Projection
        ProjectInstance(parm, nlp, input, classifier, ensemble, inputs, index);

        for (int i = 0; i < classifier->nodes; i++) {
            x[i] = 0.0;
            for (int j = 0; j < classifier->inputs; j++)
                x[i] += classifier->ih[i][j]*nlp[j];

            x[i] = TransferFunction(x[i], parm->transfer_h, parm->a, parm->b);
        }
    } else {
#endif
        // Weighted output
        for (int i = 0; i < classifier->nodes-1; i++) {
            x[i] = 0.0;
            if (index)
                for (int j = 0; index[j] != -1; j++)
                    x[i] += classifier->ih[i][index[j]-1]*input[j]*classifier->subspace[j];
            else
                for (int j = 0; j < classifier->inputs; j++)
                    x[i] += classifier->ih[i][j]*input[j]*classifier->subspace[j];

            x[i] = TransferFunction(x[i], parm->transfer_h, parm->a, parm->b);
        }
#ifdef NON_STANDARD
    }
#endif

    // Bias input to output nodes
    x[classifier->nodes-1] = 0.0;

    // Transfer to output vector.
    for (int i = 0; i < classes; i++) {
        output[i] = 0.0;
        for (int j = 0; j < classifier->nodes; j++)
            output[i] += classifier->ho[i][j]*x[j];

        output[i] = TransferFunction(output[i], parm->transfer_o, parm->a_output, parm->b_output);
    }

    // Get max output
    max = 0;
    for (int i = 1; i < classes; i++)
        if (output[i] > output[max])
            max = i;

    return max;
}

// Reallocate memory for a neural network classifier
// changing the number of nodes
void ReallocateNetwork(t_ClassModel *classifier, int nodes, int classes)
{
    // Reallocate memory for the new network
    REALLOC(classifier->ih, scalar *, nodes, "ReallocateNetwork: classifier->ih");

    // Allocate new nodes
    for (int i = classifier->nodes; i < nodes; i++)
        MALLOC(classifier->ih[i], scalar, classifier->inputs, "ReallocateNetwork: classifier->ih[i]");
    for (int i = 0; i < classes; i++)
        REALLOC(classifier->ho[i], scalar, nodes, "ReallocateNetwork: classifier->ho[i]");

    classifier->nodes = nodes;
}
