/***************************************************************************
 *   Copyright (C) 2007 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "typex.h"
#include "classifiers.h"
#include "ensemble.h"
#include "macros.h"
#include "maths.h"

// Implements FeatureBoost algorithm
// J. O'Sullivan, J. Langford, R. Caruana and A. Blum, "FeatureBoost: A meta learning algorithm that improves model robustness"
// Proceedings of the Seventeenth International Conference on Machine Learning, 2000, pp. 703-710.

#define ITERATIONS 500
#define MAX_ATTEMPTS 100

t_FeatureBoost *TrainFeatureBoostEnsemble(const t_Parameters *parm, int inputs, scalar **data, int instances, int *labels, int classes, int **index, int *subset)
{
    t_FeatureBoost *boost;
    scalar w[inputs], p[inputs], I[inputs], efm, *equal;

    // Initialize weight for each input
    for (int i = 0; i < inputs; i++)
        w[i] = 1.0/inputs;

    MALLOC(equal, scalar, instances, "TrainFeatureBoostEnsemble: equal");
    FillVector(equal, NULL, instances, 1.0);

    MALLOC(boost, t_FeatureBoost, 1, "TrainFeatureBoostEnsemble: boost");
    boost->T = parm->ensemble_size;
    boost->base = parm->base_learner;
    MALLOC(boost->alpha, scalar, boost->T, "TrainFeatureBoostEnsemble: boost->alpha");
    MALLOC(boost->bcs, t_Learner, boost->T, "TrainFeatureBoostEnsemble: boost->bcs");
    AllocateIntegerMatrix(&(boost->subspace), boost->T, inputs);

    // First member
    FillVector(NULL, boost->subspace[0], inputs, 1.0);
    ClassifierTraining(parm, boost->base, data, instances, labels, inputs, classes, index, NULL, subset, true, &(boost->bcs[0]), NULL);

    // Evaluate error
    scalar e = 0.0;
    for (int i = 0; i < instances; i++)
        e += (ClassifierOutput(parm, boost->base, data[i], inputs, classes, index ? index[i] : NULL, NULL, data, instances, labels, subset, &(boost->bcs[0]), labels[i], NULL, -1, index, NULL)
                != labels[i]);
    e /= instances;

    // Weight member
    boost->alpha[0] = AssignWithinBoundsScalar(MIN_ALPHA, log((1.0-e)/e), MAX_ALPHA);

    // Construct boost
    for (int t = 1; t < boost->T; t++) {
        SHOWPROGRESS(parm->show_progress > 1, "FeatureBoost adding classifier %d\n", t+1);
        // Obtain probabilities
        memcpy(p, w, sizeof(scalar)*inputs);
        NormalizeVector(p, inputs);

        // Create subspace
        efm = Deemphasize(parm, p, t, boost->T, ITERATIONS, e, &(boost->bcs[t-1]), boost->subspace[t-1], inputs, data, classes, instances, index, labels, subset, boost->base);

        ClassifierTraining(parm, boost->base, data, instances, labels, inputs, classes, index, NULL, subset, true, &(boost->bcs[t]), NULL);

        // Relevance of each input
        Importance(parm, &(boost->bcs[t]), I, inputs, data, classes, instances, ITERATIONS, index, labels, subset, boost->base);

        // Evaluate error
        e = 0.0;
        for (int i = 0; i < instances; i++)
            e += (ClassifierOutput(parm, boost->base, data[i], inputs, classes, index ? index[i] : NULL, NULL, data, instances, labels, subset, &(boost->bcs[t]), labels[i], NULL, -1, index,
                                   NULL) != labels[i]);
        e /= instances;

        // Weight boost
        boost->alpha[t] = AssignWithinBoundsScalar(MIN_ALPHA, log((1.0-e)/e), MAX_ALPHA);

        // Update feature weights
        for (int i = 0; i < inputs; i++)
            w[i] += I[i]*(efm-e);
    }

    FREE(equal);

    return boost;
}

scalar Deemphasize(const t_Parameters *parm, scalar *p, int t, int T, int S, scalar e, t_Learner *h, int *subspace, int inputs, scalar **data, int classes, int instances, int **index, int *labels,
                   int *subset, t_Classifier base)
{
    scalar efm, em, cdf[inputs];
    int K;

    efm = Marginalize(parm, p, inputs, S, h, inputs, data, classes, instances, index, labels, subset, base);

    em = e;
    K = 0.15*inputs;

    while (em < ((T-t) *e + t*efm) /T && K < 0.9*inputs) {
        K++;
        em = Marginalize(parm, p, K, S, h, inputs, data, classes, instances, index, labels, subset, base);
    }

    // Select K inputs using p
    // Cumulative probabilities
    // If K = inputs select all
    if (K == inputs) {
        for (int i = 0; i < inputs; i++)
            subspace[i] = 1.0;
    } else {
        cdf[0] = p[0];
        subspace[0] = 0.0;
        for (int i = 1; i < inputs; i++) {
            subspace[i] = 0.0;
            cdf[i] = cdf[i-1] + p[i];
        }

        int count = 0, attempts = 0;
        do {
            attempts++;
            int selected = RouletteSelection(cdf, inputs);
            if (subspace[selected] <= NEARZERO) {
                subspace[selected] = 1.0;
                count++;
            }

        } while (count < K && attempts < MAX_ATTEMPTS);
    }

    return efm;
}

void Importance(const t_Parameters *parm, t_Learner *h, scalar *cn, int inputs, scalar **data, int classes, int instances, int T, int **index, int *labels, int *subset, t_Classifier base)
{
    // Initialize importance to 0
    ZeroVector(cn, NULL, inputs);

    // Main loop
    for (int n = 0; n < inputs; n++) {
        for (int t = 0; t < T; t++) {
            int x1 = RANDOM(0, instances);
            int x2 = RANDOM(0, instances);

            int v = ClassifierOutput(parm, base, data[x1], inputs, classes, index ? index[x1] : NULL, NULL, data, instances, labels, subset, h, labels[x1], NULL, -1, index, NULL);

            // Change n input
            scalar aux = data[x1][n];
            data[x1][n] = data[x2][n];

            if (ClassifierOutput(parm, base, data[x1], inputs, classes, index ? index[x1] : NULL, NULL, data, instances, labels, subset, h, labels[x1], NULL, -1, index, NULL) != v)
                cn[n]++;

            data[x1][n] = aux;
        }
    }

    // Update importance
    for (int i = 0; i < inputs; i++)
        cn[i] /= T;
}

scalar Marginalize(const t_Parameters *parm, scalar *p, int K, int S, t_Learner *h, int inputs, scalar **data, int classes, int instances, int **index, int *labels, int *subset, t_Classifier base)
{
    scalar em = 0.0, *aux, *cdf;
    int *subspace;

    MALLOC(aux, scalar, inputs, "Marginalize: aux");
    MALLOC(cdf, scalar, inputs, "Marginalize: cdf");
    MALLOC(subspace, int, inputs, "Marginalize: subspace");
    ZeroVector(NULL, subspace, inputs);

    // Cumulative probabilities
    cdf[0] = p[0];
    for (int i = 1; i < inputs; i++)
        cdf[i] = cdf[i-1] + p[i];

    for (int t = 0; t < S; t++) {
        int x1 = RANDOM(0, instances);

        // Save instance
        CopyScalarVector(data[x1], aux, inputs);

        int v = ClassifierOutput(parm, base, data[x1], inputs, classes, index ? index[x1] : NULL, NULL, data, instances, labels, subset, h, labels[x1], NULL, -1, index, NULL);

        // Get K random inputs without replacement
        if (K == inputs) {
            for (int i = 0; i < inputs; i++)
                data[x1][i] = data[(int) RANDOM(0, instances) ][i];
        } else {
            int count = 0, attempts = 0;
            do {
                attempts++;
                int selected = RouletteSelection(cdf, inputs);
                if (!subspace[selected]) {
                    subspace[selected] = 1;
                    count++;

                    data[x1][selected] = data[(int) RANDOM(0, instances)][selected];
                }
            } while (count < K && attempts < MAX_ATTEMPTS);
        }

        // Classify
        if (ClassifierOutput(parm, base, data[x1], inputs, classes, index ? index[x1] : NULL, NULL, data, instances, labels, subset, h, labels[x1], NULL, -1, index, NULL) != v)
            em++;

        // Restore instance
        CopyScalarVector(aux, data[x1], inputs);
    }

    em /= S;

    FREE(aux);
    FREE(cdf);
    FREE(subspace);

    return em;
}

int FeatureBoostOutput(const t_Parameters *parm, t_FeatureBoost *boost, scalar *instance, int inputs, int classes, int *index, scalar **data, int instances, int *labels, int *subset, int label,
                       int **data_index, scalar *probs)
{
    scalar votes[classes];
    ZeroVector(votes, NULL, classes);

    for (int t = 0; t < boost->T; t++) {
        int category = ClassifierOutput(parm, boost->base, instance, inputs, classes, index, boost->subspace[t], data, instances, labels, subset, &(boost->bcs[t]), label, NULL, -1, data_index,
                                        NULL);

        votes[category] += boost->alpha[t];
    }

    if (probs) {
        memcpy(probs, votes, sizeof(scalar)*classes);
        NormalizeVector(probs, classes);
    }

    return IndexOfMaximumOfScalarVector(votes, classes);
}

void FreeFeatureBoost(t_FeatureBoost *boost, const t_Parameters *parm, int classes, int inputs)
{
    FREE(boost->alpha);
    for (int t = 0; t < boost->T; t++) {
        ClassifierFree(parm, &(boost->bcs[t]), boost->base, classes, inputs, NULL);
        FREE(boost->subspace[t]);
    }

    FREE(boost->bcs);
    FREE(boost->subspace);
}


scalar FeatureBoostMargin(const t_Parameters *parm, t_FeatureBoost *boost, scalar *instance, int label, int classes, int inputs, int *index, scalar **data, int instances, int *labels, int *subset,
                          int **data_index)
{
    int max_index;
    scalar max_value, margin, outputs[classes];

    FeatureBoostOutput(parm, boost, instance, inputs, classes, index, data, instances, labels, subset, label, data_index, outputs);

    // Get classification
    max_index = IndexOfMaximumOfScalarVector(outputs, classes);
    max_value = outputs[max_index];

    // Test if correctly Classified.
    if (label == max_index) {
        // Difference with second one
        margin = max_value;

        // Find second highest
        outputs[max_index] = -100000;
        max_index = 0;
        max_value = outputs[0];
        for (int j = 1; j < classes; j++)
            if (outputs[j] > max_value) {
                max_value = outputs[j];
                max_index = j;
            }
        margin -= max_value;

    } else
        // Difference with correct one
        margin = outputs[label]-max_value;

    return margin;
}
