/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ensemble.h"

// FloatBoost method: S. Z. Li and Zh. Zhang, "FloatBoost learning and statistical face recognition", IEEE TPAMI vol. 26, no. 9, pp. 1-12, 2004

#define MAX_ATTEMPS 100   // Maximum number of attempts to construct the ensemble

t_Boosting *TrainFloatBoost(const t_Parameters *parm, scalar **data, int instances, int *labels, int inputs, int classes, int **index, int *subset, int *subspace)
{
    t_Boosting *boost;
    t_Ensemble ensemble;
    scalar **l_data, e_min[parm->ensemble_size+1];
    int *l_labels, **l_index = NULL, *l_subset = NULL;

    MALLOC(l_data, scalar *, instances, "TrainAdaBoost: l_data");
    MALLOC(l_labels, int, instances, "TrainAdaBoost: l_labels");
    if (subset)
        MALLOC(l_subset, int, instances, "TrainAdaBoost: l_subset");
    if (index)
        MALLOC(l_index, int *, instances, "TrainAdaBoost: l_index");

    // Initialize ensemble
    MALLOC(boost, t_Boosting, 1, "TrainFloatBoost: boost");
    MALLOC(boost->alpha, scalar, parm->ensemble_size, "TrainAdaBoost: boost->alpha");
    MALLOC(boost->bcs, t_Learner, parm->ensemble_size, "TrainAdaBoost: boost->bcs");

    boost->r = 0;
    FillVector(boost->alpha, NULL, parm->ensemble_size, 1.0);

    ensemble.boost = boost;

    // First round with original data
    for (int i = 0; i < instances; i++) {
        l_data[i] = data[i];
        l_labels[i] = labels[i];
        if (subset)
            l_subset[i] = subset[i];
        if (index)
            l_index[i] = index[i];
    }

    // Iterative training of base classifiers, Webb (2000) model
    scalar *w, *cdf;
    MALLOC(w, scalar, instances, "TrainAdaBoost: w");
    MALLOC(cdf, scalar, instances, "TrainAdaBoost: w");
    FillVector(w, NULL, instances, 1.0/instances);
    FillVector(e_min, NULL, parm->ensemble_size+1, 1.0);

    boost->T = 0;
    boost->base = parm->base_learner;
    for (int attempts = 0; boost->T < parm->ensemble_size && attempts < MAX_ATTEMPS; attempts++) {
        ClassifierTraining(parm, boost->base, l_data, instances, l_labels, inputs, classes, l_index, subspace, l_subset, true, &(boost->bcs[boost->T]), NULL);

        // Forward inclusion
        boost->T++;
        SHOWPROGRESS(parm->show_progress > 1, "FloatBoost forward inclusion: %d member\n", boost->T+1);

        scalar e_i = 0.0;
        for (int i = 0; i < instances; i++)
            e_i += (ClassifierOutput(parm, FloatBoost, data[i], inputs, classes, index ? index[i] : NULL, subspace, data, instances, labels, subset, NULL, labels[i], NULL, -1, index, &ensemble)
                != labels[i]);

        e_i /= instances;
        if (e_i < e_min[boost->T])
            e_min[boost->T] = e_i;

        // Conditional exclusion
        bool removed = true;
        while (removed && boost->T > 1) {
            int h_prime = -1;
            removed = false;

            // Get the member that less contribute to the classification accuracy
            scalar best_error = 1.0;
            for (int i = 0; i < boost->T; i++) {
                boost->alpha[i] = 0.0; // Temporarily remove classifier

                // Evaluate ensemble
                scalar e_i = 0.0;
                for (int j = 0; j < instances; j++)
                    e_i += (ClassifierOutput(parm, FloatBoost, data[j], inputs, classes, index ? index[j] : NULL, subspace, data, instances, labels, subset, NULL, labels[j], NULL, -1, index,
                        &ensemble) != labels[j]);

                if (e_i < best_error) {
                    best_error = e_i;
                    h_prime = i;
                }

                boost->alpha[i] = 1.0;
            }

            if (best_error < e_min[boost->T-1]) {
                e_min[boost->T-1] = best_error;

                // Remove h'
                removed = true;
                ClassifierFree(parm, &(boost->bcs[h_prime]), boost->base, classes, inputs, NULL);
                boost->T--;
                boost->bcs[h_prime] = boost->bcs[boost->T];

                SHOWPROGRESS(parm->show_progress > 1, "FloatBoost conditional exclusion: %d member\n", boost->T+1);
            }
        }

        // Update weights
        UpdateFloatBoostWeights(parm, inputs, classes, index, subspace, data, instances, labels, subset, w, &ensemble);

        // CDF function
        cdf[0] = w[0];
        for (int i = 1; i < instances; i++)
            cdf[i] = cdf[i-1]+w[i];

        // Resample for next round
        for (int i = 0; i < instances; i++) {
            int next = RouletteSelection(cdf, instances);
            l_data[i] = data[next];
            l_labels[i] = labels[next];
            if (subset)
                l_subset[i] = subset[next];
            if (index)
                l_index[i] = index[next];
        }
    }

    SAVEPROGRESS(true, "@FloatBoost final ensemble size: %d\n", boost->T);
    SHOWPROGRESS(parm->show_progress > 1, "FloatBoost final ensemble size: %d\n", boost->T+1);

    FREE(w);
    FREE(cdf);
    FREE(l_data);
    FREE(l_labels);
    if (subset)
        FREE(l_subset);
    if (index)
        FREE(l_index);

    return boost;
}

void UpdateFloatBoostWeights(const t_Parameters *parm, int inputs, int classes, int **index, int *subspace, scalar **data, int instances, int *labels, int *subset, scalar *w, t_Ensemble *ensemble)
{
    // WARNING: The implementation of FloatBoost is "ad hoc""
    for (int j = 0; j < instances; j++) {
        scalar margin = ClassifierMargin(parm, FloatBoost, data[j], inputs, classes, index ? index[j] : NULL, subspace, labels[j], data, instances, labels, subset, NULL, ensemble);

        // Update weights
        w[j] *= exp(-margin);
    }

    // Normalize
    NormalizeVector(w, instances);
}
