/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ensemble.h"

// Juan J. Rodrı́guez, Jesús Maudes (2008) "Boosting recombined weak classifiers", Pattern Recognition Letters 29 1049–1059

void GenerateReweightBoostSample(const t_Parameters *parm, t_Boosting *boost, scalar *w_t, scalar *w_t_1, scalar *alpha, int instances, int classes, scalar **data, int *labels, int inputs,
                                 int **index, int *subspace, int *subset, int round)
{
    bool *ok;
    scalar votes[classes];

    MALLOC(ok, bool, instances, "GenerateAdaSample: ok");

    // Classify each pattern
    scalar e_i = 0.0;

    for (int j = 0; j < instances; j++)  {
        // Out of each group of r classifiers
        ZeroVector(votes, NULL, classes);
        for (int t = MAX(0, round-boost->r); t <= round; t++)
            votes[ClassifierOutput(parm, parm->base_learner, data[j], inputs, classes, index ? index[j] : NULL, subspace, data, instances, labels, subset, &(boost->bcs[t]), labels[j], NULL, -1,
                                                          index, NULL)]++;

        ok[j] = (labels[j] == IndexOfMaximumOfScalarVector(votes, classes));
        if (!ok[j])
            e_i += w_t[j];
    }

    e_i /= instances;

    // Update weights
    if (e_i > 0.5) {
        *alpha = MIN_ALPHA;
        // New bootstrap sample
        for (int k = 0; k < instances; k++)
            w_t_1[k] = Poisson();
    } else if (e_i == 0.0) {
        // New bootstrap sample
        *alpha = MAX_ALPHA;
        // New bootstrap sample
        for (int k = 0; k < instances; k++)
            w_t_1[k] = Poisson();
    } else {
        *alpha = (1.0-e_i)/e_i;
        *alpha = AssignWithinBoundsScalar(MIN_ALPHA, *alpha, MAX_ALPHA);

        for (int i = 0; i < instances; i++) {
            if (ok[i])
                w_t_1[i] = w_t[i]*sqrt(1.0/(*alpha));
            else
                w_t_1[i] = w_t[i]*sqrt(*alpha);
            if (w_t_1[i] < 1.0e-10)
                w_t_1[i] = 1.0e-10;
        }
    }

    // Normalize to sum N
    NormalizeVector(w_t_1, instances);

    FREE(ok);
}


int ReweightBoostOutput(const t_Parameters *parm, t_Boosting *boost, scalar *instance, int inputs, int classes, int *index, int *subspace, scalar **data, int instances, int *labels, int *subset,
                        int label, int **data_index, scalar *probs)
{
    scalar votes[classes], lvotes[classes];
    int categories[boost->T];

    ZeroVector(votes, NULL, classes);

    // First obtain category of every classifier
    for (int t = 0; t < boost->T; t++)
        categories[t] = ClassifierOutput(parm, parm->base_learner, instance, inputs, classes, index, subspace, data, instances, labels, subset, &(boost->bcs[t]), label, NULL, -1, data_index, NULL);

    // Out of each group of r classifiers
    for (int t = 0; t < boost->T; t++) {
        ZeroVector(lvotes, NULL, classes);
        for (int i = MAX(0, t-boost->r); i <= t; i++)
            lvotes[categories[i]]++;

        votes[IndexOfMaximumOfScalarVector(lvotes, classes)] += boost->alpha[t];
    }

    if (probs) {
        memcpy(probs, votes, sizeof(scalar)*classes);
        NormalizeVector(probs, classes);
    }

    return IndexOfMaximumOfScalarVector(votes, classes);
}
