#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Boosting classifiers
"""

import sys
import os
import numpy as np
from sklearn.ensemble import AdaBoostClassifier
from sklearn.model_selection import GridSearchCV
from sklearn import svm
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis, LinearDiscriminantAnalysis
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn.linear_model import RidgeClassifier

try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
parser.add_argument("metric", help="Metric for CV of hyper-parameters", type=int)
parser.add_argument("estimators", help="Number of estimators", type=int)
parser.add_argument("base", help="Base learners (as in typex.h)", type=int)

# Optional
parser.add_argument("-cv", help="Validate parameters", action='store_true')
parser.add_argument("-r", help="Use Real AdaBoost (default = Discrete)", action='store_true')

args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

classifiers = [
    GaussianNB(),
    DecisionTreeClassifier(),
    None,
    None,
    None,
    None,
    svm.SVC(),
    LinearDiscriminantAnalysis(),
    QuadraticDiscriminantAnalysis(),
    GaussianProcessClassifier(kernel=1.0*RBF()),
    RidgeClassifier(),
    None,
    None,
    None
    ]

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

X = dtrain.loc[:, dtrain.columns != c-1]
y = dtrain[c-1]
X_test = dtest.loc[:, dtrain.columns != c-1]

if args.cv:
    from sklearn.metrics import make_scorer, recall_score, cohen_kappa_score
    from metrics import gmean

    gmean_scorer = make_scorer(gmean, greater_is_better=True)
    kappa_scorer = make_scorer(cohen_kappa_score, greater_is_better=True)

    metrics = ["accuracy", gmean_scorer, kappa_scorer, "roc_auc", None, "recall_macro", None, None, "balanced_accuracy", "precision_macro", "f1_macro", None, "f1"]
    names = ["Accuracy", "G-mean", "Kappa", "auROC", None, "Recall macro", None, None, "Balanced accuracy", "Precision macro", "F1 macro", None, "F1"]

    parameters = {'n_estimators': [10, 30, 50, 100, 250, 500]
                 }

    # Fit model with random search
    model = AdaBoostClassifier(n_estimators=args.estimators, base_estimator=classifiers[args.base], algorithm='SAMME.R' if args.r else 'SAMME')
    boost = GridSearchCV(model, parameters, scoring=metrics[args.metric], n_jobs=-1, cv=3, verbose=3).fit(X, y)

    print(boost.best_params_)
    # Save parameters
    f = open(args.trainfile + '.boost.parameters', 'w')
    f.write(str(boost.best_params_))
    f.close()
else:
    boost = AdaBoostClassifier(n_estimators=args.estimators, base_estimator=classifiers[args.base], algorithm='SAMME.R' if args.r else 'SAMME').fit(X, y)

# Save predictions
preds = boost.predict(X_test)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

# Save probabilities for margin
probs = boost.predict_proba(X_test)
np.savetxt(args.testfile+".probs", probs)

sys.exit(1)
