#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Linear and quadratic discriminant analysis
"""

import sys
import os
import numpy as np
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis, LinearDiscriminantAnalysis
from sklearn.model_selection import GridSearchCV

try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
parser.add_argument("metric", help="Metric for CV of hyper-parameters", type=int)

# Optional
parser.add_argument("-q", help="Quadratic discriminant analysis", action='store_true')
parser.add_argument("-cv", help="Validate parameters", action='store_true')

args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

X = dtrain.loc[:, dtrain.columns != c-1]
y = dtrain[c-1]

if args.cv:
    from metrics import gmean
    from sklearn.metrics import make_scorer, recall_score, cohen_kappa_score

    gmean_scorer = make_scorer(gmean, greater_is_better=True)
    kappa_scorer = make_scorer(cohen_kappa_score, greater_is_better=True)

    metrics = ["accuracy", gmean_scorer, kappa_scorer, "roc_auc", None, "recall_macro", None, None, "balanced_accuracy", "precision_macro", "f1_macro", None, "f1"]
    names = ["Accuracy", "G-mean", "Kappa", "auROC", None, "Recall macro", None, None, "Balanced accuracy", "Precision macro", "F1 macro", None, "F1"]

    if args.q:
        parameters = {'reg_param': [0.00001, 0.0001, 0.001,0.01, 0.1],
                      'store_covariance': [True, False],
                      'tol': [0.0001, 0.001,0.01, 0.1],
                     }
    else:
        parameters = [{'solver': ['lsqr', 'eigen'], 'shrinkage': [0.0, 0.1, 0.25, 0.5, 0.75, 1.0]},
                      {'solver': ['svd']}
                     ]

    # Fit model with random search
    model = QuadraticDiscriminantAnalysis() if args.q else LinearDiscriminantAnalysis()

    da = GridSearchCV(model, parameters, scoring=metrics[args.metric], n_jobs=-1, cv=3, verbose=3)
    da.fit(X, y)

    print(da.best_params_)
    # Save parameters
    f = open(args.trainfile + '.da.parameters', 'w')
    f.write(names[args.metric] + ", ")
    f.write(str(da.best_params_))
    f.close()
else:
    if args.q:
        da = QuadraticDiscriminantAnalysis().fit(X, y)
    else:
        da = LinearDiscriminantAnalysis().fit(X, y)

# Get probabilities and save them for margin
preds = da.predict(dtest.drop(c-1, axis=1))
probs = da.predict_proba(dtest.drop(c-1, axis=1))

np.savetxt(args.testfile+".probs", probs)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

sys.exit(1)
