#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Naive Bayes
"""

import sys
import os
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.model_selection import GridSearchCV
from sklearn.naive_bayes import GaussianNB, MultinomialNB, ComplementNB, BernoulliNB
try:
	import argparse
	import pandas as pd
except ImportError:
    print("Cannot import argparse or pandas")
    sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
parser.add_argument("metric", help="Metric for CV of hyper-parameters", type=int)

# Optional
parser.add_argument("-cv", help="Validate parameters", action='store_true')
args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
# dtrain = genfromtxt(args.trainfile, delimiter=' ')
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")
# dtest = genfromtxt(args.testfile, delimiter=' ')

# Scale
X_train = dtrain.loc[:, dtrain.columns != c-1]
X_test = dtest.drop(c-1, axis=1)
y_train = dtrain[c-1]

if args.cv:
    from metrics import gmean
    from sklearn.metrics import make_scorer, recall_score, cohen_kappa_score

    gmean_scorer = make_scorer(gmean, greater_is_better=True)
    kappa_scorer = make_scorer(cohen_kappa_score, greater_is_better=True)

    metrics = ["accuracy", gmean_scorer, kappa_scorer, "roc_auc", None, "recall_macro", None, None, "balanced_accuracy", "precision_macro", "f1_macro", None, "f1"]
    names = ["Accuracy", "G-mean", "Kappa", "auROC", None, "Recall macro", None, None, "Balanced accuracy", "Precision macro", "F1 macro", None, "F1"]

    # Some models do not accept negative values
    min = np.min(X_train, axis=0)
    max = np.max(X_train, axis=0)
    X_train = (X_train - min)#/(max - min)
    X_test = (X_test - min)#/(max - min)

    parameters = [{'classifier': [GaussianNB()]}, {'classifier': [MultinomialNB()]}, {'classifier': [ComplementNB()]}, {'classifier': [BernoulliNB()]}]
    pipeline = Pipeline([('classifier', GaussianNB)])

    nb = GridSearchCV(pipeline, parameters, cv=3, n_jobs=-1, scoring=metrics[args.metric], verbose=3).fit(X_train, y_train)
    print(nb.best_params_)
    # Save parameters
    f = open(args.trainfile + '.nbc.parameters', 'w')
    f.write(names[args.metric] + ", ")
    f.write(str(nb.best_params_))
    f.close()
else:
    nb = GaussianNB() # Other models are MultinomialNB(), ComplementNB(), BernoulliNB()

    # Train the model on training data
    nb.fit(X_train, y_train)

# Save predictions
preds = nb.predict(X_test)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

# Save probabilities for margin
probs = nb.predict_proba(dtest.drop(c-1, axis=1))
np.savetxt(args.testfile+".probs", probs)

sys.exit(1)
