/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <math.h>
#include "typex.h"
#include "common.h"
#include "macros.h"
#include "external.h"
#include "data.h"

// Stub for evaluating Extreme Gradient Boosting using xgboost library

// Cite as: T. Chen and C. Guestrin, "XGBoost: A Scalable Tree Boosting System", Proceedings of the 22nd ACM SIGKDD International Conference on Knowledge Discovery and Data Mining,
// KDD'16, 2016, an Francisco, California, USA, 785-794, ACM, New York, NY, USA (Bibtex: chen06)

#define MAX_COMMAND_LENGTH  10000

void GetConfusionMatrixWithPython(const t_Parameters *parm, scalar **test_data, int test_instances, int *test_labels, int *subset, int *subspace, int classes, int inputs, scalar **train_data,
                                  int train_instances, int *train_labels, int **test_index, int **train_index, int **M, char *out_file_name, scalar *average_margin, scalar *minimum_margin,
                                  scalar *margin, t_ExternalClassifier classifier, int minority, scalar *sc_pos, scalar *sc_neg)
{
    char f_train_name[MAX_FILENAME_LENGTH], f_test_name[MAX_FILENAME_LENGTH], f_name[MAX_FILENAME_LENGTH+12], *kernels[] = {"linear", "polynomial", "rbf", "sigmoid"};
    FILE *f;
    bool quadratic = false;

    // Check classifier
    char command[MAX_COMMAND_LENGTH];
    switch (classifier) {
        case XGBoost:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "XGBoost evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.xgb.training", out_file_name);
            SaveLIBsvmFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.xgb.test", out_file_name);
            SaveLIBsvmFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            // Call runxgboost to train XGBoost and test
            sprintf(command, "runxgboost.py %s %s %d", f_train_name, f_test_name, classes);
            break;

        case ERT:
        case RandomForest:
            sprintf(f_train_name, "%s.RandomForest.training", out_file_name);
            sprintf(f_test_name, "%s.RandomForest.test", out_file_name);

            // Create training and testing datasets
            if (train_index || test_index) {
                SaveLibSVMData(train_data, train_index, train_labels, train_instances, f_train_name);
                SaveLibSVMData(test_data, test_index, test_labels, test_instances, f_test_name);;
            } else {
                SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);
                SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);
            }

            sprintf(command, "randomforest.py %s %s %s %s %s %d %d %d %d", classifier == ERT ? "-x" : "", parm->cv_external_hyperparameters_for_testing ? "-cv" : "", train_index ? "-s" : "",
                    f_train_name, f_test_name, parm->rf_ntree, inputs, parm->dt_max_depth, parm->cv_performance_measure);
            break;

        case NaiveBayes:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Naïve Bayes evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.nb.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.nb.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "naivebayes.py %s %s %s %d", parm->cv_external_hyperparameters_for_testing ? "-cv" : "", f_train_name, f_test_name, parm->cv_performance_measure);
            break;

        case DeepLearning:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "DeepLearning evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.dl.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.dl.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "deeplearning.py %s %s -l %s", f_train_name, f_test_name, "10,5,10");
            break;

        case DecisionTree:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "DecisionTree evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.dt.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.dt.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "decisiontree.py %s %s %s", parm->print_dt ? "-p" : " ", f_train_name, f_test_name);
            break;

        case PLSDA:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "PLSDA evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.pls.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.pls.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "plsda.py %s %s", f_train_name, f_test_name);
            break;

        case GPML:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "GPML evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.gpml.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.gpml.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "gpml.py %s %s", f_train_name, f_test_name);
            break;

        case Ridge:
        case KRR:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ridge Regression evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.ridge.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.ridge.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "ridge.py %s %s %s %s %d", classifier == KRR ? "-k" : "", parm->cv_external_hyperparameters_for_testing ? "-cv" : "", f_train_name, f_test_name,
                    parm->cv_performance_measure);
            break;

        case QDA:
            quadratic = true;
        case LDA:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "LDA/QDA evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.da.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.da.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "da.py %s %s %s %s %d", parm->cv_external_hyperparameters_for_testing ? "-cv" : "", quadratic ? "-q" : "", f_train_name, f_test_name,
                    parm->cv_performance_measure);
            break;

        case SVC:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "SVC evaluation for sparse data");

            if (parm->kernel > 3)
                OnErrorExit(ERR_PARAM, "Kernel not supported in Python SVC");

            // Create training dataset
            sprintf(f_train_name, "%s.svc.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.svc.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "svc.py %s %s %s %s %lf -coef0 %lf -d %d -gamma %lf %d", parm->cv_external_hyperparameters_for_testing ? "-cv" : "", f_train_name, f_test_name,
                    kernels[parm->kernel], parm->svm_c, parm->svm_r, parm->svm_d, parm->gamma, parm->cv_performance_measure);
            break;

        case AdaBoostR:
        case AdaBoostD:
            if (train_index || test_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "AdaBoost evaluation for sparse data");

            // Create training dataset
            sprintf(f_train_name, "%s.boosting.training", out_file_name);
            SaveCSVFormat(train_data, train_instances, inputs, train_labels, subspace, subset, f_train_name);

            // Create testing dataset
            sprintf(f_test_name, "%s.boosting.test", out_file_name);
            SaveCSVFormat(test_data, test_instances, inputs, test_labels, subspace, NULL, f_test_name);

            sprintf(command, "boosting.py %s %s %s %s %d %d %d", classifier == AdaBoostR ? "-r" : "", parm->cv_external_hyperparameters_for_testing ? "-cv" : "", f_train_name, f_test_name,
                    parm->cv_performance_measure, parm->ensemble_size, (int) parm->external_base);
            break;
    }

    int status = system(command);
    if (!WEXITSTATUS(status)) {
        fprintf(parm->out_file, "@ok\n");

        // Remove files
        if (!parm->keep_files) {
            remove(f_train_name);
            remove(f_test_name);
            remove(f_name); // Prediction
            sprintf(f_name, "%s.preds", f_test_name);
            remove(f_name); // Probabilities
            sprintf(f_name, "%s.probs", f_test_name);
        }

        printf("Failed command: %s\n", command);
        OnErrorExit(ERR_UNSPECIFIED, "Python learning method failed");
    }

    // Build M
    ZeroIntegerMatrix(M, classes, classes);
    // Read predictions file
    sprintf(f_name, "%s.preds", f_test_name);
    if ((f = fopen(f_name, "r")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open predictions file");

    // Read prediction for every instance
    for (int i = 0; i < test_instances; i++) {
        int pred;

        fscanf(f, "%d", &pred);
        M[test_labels[i]][pred]++;
    }

    fclose(f);

    // If decision tree program return number of nodes and write file
    if (classifier == DecisionTree) {
        char line[MAX_LENGTH];

        if (parm->print_dt) {
            // Read tree file
            sprintf(f_name, "%s.tree", f_test_name);
            if ((f = fopen(f_name, "r")) == NULL)
                OnErrorExit(ERR_FILE, "Cannot open tree file");

            // Read tree and write to output
            fprintf(parm->out_file, "───────────────────── CART DECISION TREE ───────────────────────\n");
            fgets(line, MAX_LENGTH, f);
            do {
                fprintf(parm->out_file, "%s", line);
                fgets(line, MAX_LENGTH, f);
            } while (!feof(f));
            fprintf(parm->out_file,
                    "────────────────────────────────────────────────────────────────\n");

            fclose(f);
            remove(f_name);
        }

        // Number of nodes
        char f_nodes_name[FILENAME_LENGTH];
        FILE *f;
        scalar average;

        sprintf(f_nodes_name, "%s.nodes", f_test_name);
        if ((f = fopen(f_nodes_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open number of nodes info file");
        fscanf(f, "Number of nodes: %lf", &average);
        fclose(f);

        fprintf(parm->out_file, "@Decision Tree Nodes: %lf\n", average);
        remove(f_nodes_name);
    } else if (classifier == RandomForest) {
        char f_nodes_name[FILENAME_LENGTH];
        FILE *f;
        scalar average;

        sprintf(f_nodes_name, "%s.nodes", f_test_name);
        if ((f = fopen(f_nodes_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open number of nodes info file");
        fscanf(f, "Average number of nodes: %lf", &average);
        fclose(f);

        fprintf(parm->out_file, "@Random Forest Average Nodes: %lf\n", average);
        remove(f_nodes_name);
    }

    // Obtain margin?
    if ((average_margin && minimum_margin) || (sc_pos && sc_neg)) {
        // Margin is given by probabilities
        scalar **probs;
        AllocateScalarMatrix(&probs, test_instances, classes);

        // Read predictions file
        sprintf(f_name, "%s.probs", f_test_name);
        if ((f = fopen(f_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open probabilities file");

        // Read prediction for every tree
        for (int i = 0; i < test_instances; i++)
            for (int j = 0; j < classes; j++)
                fscanf(f, "%lf", &probs[i][j]);

        fclose(f);

        if (average_margin && minimum_margin) {
            // Margin is (probs[ok] - probs[best or second best])
            *average_margin = 0.0;
            *minimum_margin = INFINITE;

            for (int i = 0; i < test_instances; i++) {
                // Votes of the correct class
                scalar ok_probs = probs[i][test_labels[i]];

                // Get maximum of probabilities of wrong class
                scalar wrong_probs = 0;
                for (int j = 0; j < classes; j++)
                    if (wrong_probs < probs[i][j] && j != test_labels[i])
                        wrong_probs = probs[i][j];

                scalar m = ok_probs-wrong_probs;

                *average_margin += m;
                if (m < *minimum_margin)
                    *minimum_margin = m;

                if (margin)
                    margin[i] = m;
            }

            *average_margin /= test_instances;
        }

        if (sc_pos && sc_neg) {
            // Values for AUC
            int positives = 0, negatives = 0, majority;
            majority = (minority+1)%2;
            for (int i = 0; i < test_instances; i++)
                if (test_labels[i] == minority) {
                    sc_pos[positives] = probs[i][minority] - probs[i][majority];
                    positives++;
                } else {
                    sc_neg[negatives] = probs[i][minority] - probs[i][majority];
                    negatives++;
                }
        }

        FreeScalarMatrix(&probs, test_instances);
    }

    // Save parameters if apply
    char *postfix[N_EXTERNAL] = {"nbc", "-", "rf", "-", "-", "-", "svc", "da", "da", "", "ridge", "rf", "ridge", "boost", "boost"};
    extern char *external[N_EXTERNAL];
    sprintf(f_name, "%s.%s.parameters", f_train_name, postfix[classifier]);
    if (access(f_name, F_OK) == 0) {
        char line[1000];
        FILE *fi;
        if ((fi = fopen(f_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open parameters file");

        fgets(line, 1000, fi);
        fclose(fi);
        SAVEPROGRESS(true, "External %s parameters: %s\n", external[classifier], line);
        remove(f_name);
    }

    // Remove files
    if (!parm->keep_files) {
        remove(f_train_name);
        remove(f_test_name);
        // Probabilities
        sprintf(f_name, "%s.probs", f_test_name);
        remove(f_name);
        // Prediction
        sprintf(f_name, "%s.preds", f_test_name);
        remove(f_name);
    }
}

void SaveCSVFormat(scalar **data, int instances, int inputs, int *labels, int *subspace, int *subset, char *f_name)
{
    FILE *f;

    if ((f = fopen(f_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot create training file for XGBoost");

    // CSV format
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            for (int j = 0; j < inputs; j++)
                fprintf(f, "%lf ", !subspace ? data[i][j] : data[i][j]*subspace[j]);

            fprintf(f, "%d\n", labels[i]);
        }

    fclose(f);
}

void SaveLIBsvmFormat(scalar **data, int instances, int inputs, int *labels, int *subspace, int *subset, char *f_name)
{
    FILE *f;

    if ((f = fopen(f_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot create training file in LIBsvm format");

    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            fprintf(f, "%d ", labels[i]);

            if (!subspace) {
                for (int j = 0; j < inputs; j++)
                    if (fabs(data[i][j]) > ZERO)
                        fprintf(f, "%d:%lf ", j+1, data[i][j]);
            } else {
                for (int j = 0; j < inputs; j++)
                    if (fabs(data[i][j]*subspace[j]) > ZERO)
                        fprintf(f, "%d:%lf ", j+1, data[i][j]);
            }

            fprintf(f, "\n");
        }

    fclose(f);
}
