#!	/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Random Forest and Extremely Randomized Trees
"""

try:
    import sys
    import os
    import numpy as np
    from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier
    from sklearn.model_selection import RandomizedSearchCV, GridSearchCV
    from sklearn.metrics import cohen_kappa_score, make_scorer
    from sklearn.datasets import load_svmlight_file
    import argparse
except ImportError:
    print("Modules not all present")
    sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset (libsvm format)", type=str)
parser.add_argument("testfile", help="Test dataset (libsvm format)", type=str)
parser.add_argument("ntrees", help="Number of trees", type=int)
parser.add_argument("nfeatures", help="Number of features", type=int)
parser.add_argument("maxdepth", help="Maximum depth", type=int)
parser.add_argument("metric", help="Metric for CV of hyper-parameters", type=int)

# Optional
parser.add_argument("-cv", help="Validate parameters", action='store_true')
parser.add_argument("-s", help="Sparse format", action='store_true')
parser.add_argument("-x", help="Use Extremely randomized trees", action='store_true')
args = parser.parse_args()

if args.s:
    # Sparse data
    X, y = load_svmlight_file(args.trainfile, n_features=args.nfeatures)
    X_test, y_test = load_svmlight_file(args.testfile, n_features=args.nfeatures)
else:
    import pandas as pd # Not imported before because it is not available in Skynet

    # Dense data
    # Read train file
    if not os.path.isfile(args.trainfile):
        print("Train dataset file '%s' does not exist" % args.trainfile)
        sys.exit()

    dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
    r, c = dtrain.shape

    # Read test file
    if not os.path.isfile(args.testfile):
        print("Test dataset file '%s' does not exist" % args.testfile)
        sys.exit()

    dtest = pd.read_csv(args.testfile, header=None, sep=" ")

    X = dtrain.loc[:, dtrain.columns != c-1]
    y = dtrain[c-1]
    X_test = dtest.drop(c-1, axis=1)

# Validate?
if args.cv:
    from sklearn.metrics import make_scorer, recall_score
    from metrics import gmean

    gmean_scorer = make_scorer(gmean, greater_is_better=True)
    kappa_scorer = make_scorer(cohen_kappa_score, greater_is_better=True)

    metrics = ["accuracy", gmean_scorer, kappa_scorer, "roc_auc", None, "recall_macro", None, None, "balanced_accuracy", "precision_macro", "f1_macro", None, "f1"]
    names = ["Accuracy", "G-mean", "Kappa", "auROC", None, "Recall macro", None, None, "Balanced accuracy", "Precision macro", "F1 macro", None, "F1"]

    # Create the grid
    parameters = {'n_estimators': [50, 100, 500], 'max_features': ['log2', 'sqrt'], 'max_depth': [10, 25, None]}

    # Fit model with random search
    rforest = RandomForestClassifier(class_weight='balanced', min_samples_split=10, min_samples_leaf=4, bootstrap=True) if not args.x else ExtraTreesClassifier(class_weight='balanced',
              min_samples_split=10, min_samples_leaf=4, bootstrap=True)
    rf = GridSearchCV(rforest, parameters, scoring=metrics[args.metric], n_jobs=-1, cv=3, verbose=3)
    rf.fit(X, y)

    print(rf.best_params_)

    # Save parameters
    f = open(args.trainfile + '.rf.parameters', 'w')
    f.write(names[args.metric] + ", ")
    f.write(str(rf.best_params_))
    f.close()

    # Trained forest
    rf = rf.best_estimator_

else:
    # Create random forest
    # Instantiate model with 1000 decision trees
    max_depth = args.maxdepth if args.maxdepth > 0 else None
    rf = RandomForestClassifier(class_weight='balanced', n_estimators=args.ntrees, random_state=None, max_depth=max_depth) if not args.x else ExtraTreesClassifier(class_weight='balanced',
         min_samples_split=10, min_samples_leaf=4, bootstrap=True)

    # Train the model on training data
    rf.fit(X, y)

# Save predictions
preds = rf.predict(X_test)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

# Save probabilities for margin
probs = rf.predict_proba(X_test)
np.savetxt(args.testfile+".probs", probs)

# Save average nodes per tree
nodes = 0
for estimator in rf.estimators_:
    nodes += estimator.tree_.node_count

f = open(args.testfile+".nodes", "w")
f.write("Average number of nodes: %g\n"  % (nodes/args.ntrees))
f.close()

sys.exit(1)
