/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "typex.h"
#include "common.h"
#include "macros.h"
#include "external.h"

// Stub for evaluating Random Forest using ranger

// Cite as: Wright, M. N. & Ziegler, A. (2017). ranger: A fast implementation of random forests for high dimensional data in C++ and R. J Stat Softw 77:1-17.

#define MAX_COMMAND_LENGTH  10000

// Train: ranger --file dataset.train --depvarname Class --write --outprefix model
// Test: ranger --file iris.test --predict model.forest --depvarname Class

void GetRandomForestConfusionMatrix(const t_Parameters *parm, scalar **test_data, int test_instances, int *test_labels, int *subset, int *subspace, int classes, int inputs, scalar **train_data,
                                    int train_instances, int *train_labels, int **test_index, int **train_index, int **M, char *out_file_name, scalar *average_margin, scalar *minimum_margin,
                                    scalar *margin)
{
    char f_name[MAX_FILENAME_LENGTH], line[MAX_LENGTH];
    FILE *f;
    int prediction;

    if (train_index || test_index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Random forest evaluation for sparse data");

    // Create training dataset
    sprintf(f_name, "%s.RandomForest.training", out_file_name);
    if ((f = fopen(f_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot create training file for rotation forest");

    // Header line
    for (int i = 0; i < inputs; i++)
        if (!subspace || subspace[i])
            fprintf(f, "x%d ", i+1);
    fprintf(f, "Class\n");

    // Data
    for (int i = 0; i < train_instances; i++)
        if (!subset || subset[i]) {
            for (int j = 0; j < inputs; j++)
                if (!subspace || subspace[j])
                    fprintf(f, "%lf ", train_data[i][j]);
            fprintf(f, "%d\n", train_labels[i]+1);
        }

    fclose(f);

    // Count selected if apply
    int selected = 0;
    if (!subspace)
        selected = inputs;
    else
        for (int i = 0; i < inputs; i++)
            if (subspace[i])
                selected++;

    // Call ranger to train rotation forest
    char command[MAX_COMMAND_LENGTH];
    sprintf(command, "ranger --file %s --depvarname Class --write --outprefix %s --ntree %d --mtry %d --maxdepth %d", f_name, out_file_name, parm->rf_ntree, MAX(1, (int)(selected*parm->rf_percent)),
            parm->dt_max_depth);    // Maximum depth of 0 means unlimited

    int status = WEXITSTATUS(system(command));
    if (status)
        OnErrorExit(ERR_UNSPECIFIED, "Random forest training failed");

    // Create testing dataset
    sprintf(f_name, "%s.RandomForest.test", out_file_name);
    if ((f = fopen(f_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot create training file for rotation forest");

    // Header line
    for (int i = 0; i < inputs; i++)
        if (!subspace || subspace[i])
            fprintf(f, "x%d ", i+1);
    fprintf(f, "Class\n");

    // Data
    for (int i = 0; i < test_instances; i++) {
        for (int j = 0; j < inputs; j++)
            if (!subspace || subspace[j])
                fprintf(f, "%lf ", test_data[i][j]);
        fprintf(f, "%d\n", test_labels[i]+1);
    }

    fclose(f);

    // Call ranger to test file
    sprintf(command, "ranger --file %s --depvarname Class --predict %s.forest --outprefix %s", f_name, out_file_name, out_file_name);
    status = WEXITSTATUS(system(command));
    if (status)
        OnErrorExit(ERR_UNSPECIFIED, "Random forest testing failed");

    // Read result file
    sprintf(f_name, "%s.prediction", out_file_name);
    if ((f = fopen(f_name, "r")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open predictions file");

    // Ignore first line
    fgets(line, MAX_LENGTH, f);

    // Read prediction
    ZeroIntegerMatrix(M, classes, classes);

    for (int i = 0; i < test_instances; i++) {
        fscanf(f, "%d", &prediction);
        M[test_labels[i]][prediction-1]++;
    }

    fclose(f);

    // Obtain margin?
    if (average_margin && minimum_margin) {
        // Margin is given by number of votes
        int **votes;
        AllocateIntegerMatrix(&votes, test_instances, classes);
        ZeroIntegerMatrix(votes, test_instances, classes);

        // Generate output of every tree
        sprintf(f_name, "%s.RandomForest.test", out_file_name);
        sprintf(command, "ranger --file %s --depvarname Class --predict %s.forest --outprefix %s --predall", f_name, out_file_name, out_file_name);

        status = WEXITSTATUS(system(command));
        if (status)
            OnErrorExit(ERR_UNSPECIFIED, "Random forest testing failed");

        // Read result file
        sprintf(f_name, "%s.prediction", out_file_name);
        if ((f = fopen(f_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open prediction file");

        // Ignore first line
        fgets(line, MAX_LENGTH, f);

        // Read prediction for every tree
        char tag[MAX_LENGTH];
        for (int i = 0; i < parm->rf_ntree; i++) {
            // Read "Tree n:"
            fgets(line, MAX_LENGTH, f);
            sprintf(tag, "Tree %d:\n", i);
            if (strcmp(line, tag))
                OnErrorExit(ERR_FILE, "Error in tag of prediction file");

            for (int i = 0; i < test_instances; i++) {
                fgets(line, MAX_LENGTH, f);
                sscanf(line, "%d", &prediction);
                votes[i][prediction-1]++;
            }

            // Read empty line
            fgets(line, MAX_LENGTH, f);
        }

        fclose(f);

        // Margin is (votes[ok] - votes[best or second best])
        *average_margin = 0.0;
        *minimum_margin = INFINITE;

        for (int i = 0; i < test_instances; i++) {
            // Votes of the correct class
            int ok_votes = votes[i][test_labels[i]];

            // Get maximum of votes of wrong class
            int wrong_votes = 0;
            for (int j = 0; j < classes; j++)
                if (wrong_votes < votes[i][j] && j != test_labels[i])
                    wrong_votes = votes[i][j];

            scalar m = ((scalar)(ok_votes-wrong_votes))/parm->rf_ntree;

            *average_margin += m;
            if (m < *minimum_margin)
                *minimum_margin = m;

            if (margin)
                margin[i] = m;
        }

        *average_margin /= test_instances;

        FreeIntegerMatrix(&votes, test_instances);
    }

    // Remove files
    if (!parm->keep_files) {
        sprintf(f_name, "%s.RandomForest.training", out_file_name);
        remove(f_name); // Training data
        sprintf(f_name, "%s.RandomForest.test", out_file_name);
        remove(f_name); // Testing data
        sprintf(f_name, "%s.prediction", out_file_name);
        remove(f_name); // Prediction
        sprintf(f_name, "%s.confusion", out_file_name);
        remove(f_name); // Confusion
        sprintf(f_name, "%s.forest", out_file_name);
        remove(f_name); // Model
        sprintf(f_name, "%s.log", out_file_name);
        remove(f_name); // Log
    }
}
