#!	/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Ridge regression and Kernel ridge regression
"""

import sys
import os
import numpy as np
from sklearn.linear_model import RidgeClassifier
from sklearn.kernel_ridge import KernelRidge
from sklearn.model_selection import GridSearchCV

try:
	import argparse
	import pandas as pd
except ImportError:
	sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset", type=str)
parser.add_argument("testfile", help="Test dataset", type=str)
parser.add_argument("metric", help="Metric for CV of hyper-parameters", type=int)

# Optional
parser.add_argument("-cv", help="Validate parameters", action='store_true')
parser.add_argument("-k", help="Kernel ridge regression", action='store_true')

args = parser.parse_args()

# Read train file
if not os.path.isfile(args.trainfile):
    print("Train dataset file '%s' does not exist" % args.trainfile)
    sys.exit()

dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
r, c = dtrain.shape

# Read test file
if not os.path.isfile(args.testfile):
    print("Test dataset file '%s' does not exist" % args.testfile)
    sys.exit()

dtest = pd.read_csv(args.testfile, header=None, sep=" ")

X = dtrain.loc[:, dtrain.columns != c-1]
y = dtrain[c-1]

if args.cv:
    if args.k:  # Parameters for Kernel Ridge Regression
        parameters = [
                       {'alpha': [0.1, 0.5, 1, 1e3, 1e6], "kernel": ["linear"]},
                       {'alpha': [0.1, 0.5, 1, 1e3, 1e6], "kernel": ["rbf"], "gamma": [0.0001, 0.001, 0.01, 0.1, 1.0, 10.0]},
                       {'alpha': [0.1, 0.5, 1, 1e3, 1e6], "kernel": ["poly"], "degree": [2, 3, 4, 5]}
                     ]
    else:   # Parameters for Linear Ridge Regression
        parameters = {'solver': ['svd', 'cholesky', 'lsqr', 'sparse_cg', 'sag', 'saga'], # 'lbfgs' not usable for all cases
                      'alpha': [0.1, 0.5, 1, 1e3, 1e6]
                     }

    # Fit model with random search
    model = RidgeClassifier() if not args.k else KernelRidge()
    ridge = GridSearchCV(model, parameters, n_jobs=-1, cv=3, verbose=3)

    # If kernel, decode y
    if args.k:
        new_y = np.empty((y.shape[0], max(y)+1))
        new_y.fill(-1)
        for i in range(0, new_y.shape[0]):
            new_y[i][y[i]] = 1
        y = new_y

    ridge.fit(X, y)

    print(ridge.best_params_)
    # Save parameters
    f = open(args.trainfile + '.ridge.parameters', 'w')
    f.write(str(ridge.best_params_))
    f.close()
else:
    ridge = RidgeClassifier().fit(X, y) if not args.k else KernelRidge().fit(X, y)

# Get probabilities and save them for margin
# Different depending on Linear or Kernel regression
preds = ridge.predict(dtest.drop(c-1, axis=1))

if args.k:
    probs = preds
    preds = probs.argmax(axis=1)
else:
    probs = ridge.decision_function(dtest.drop(c-1, axis=1))

    # Shape depends on number of classes
    if len(probs.shape) == 1:
        probs = probs[:, np.newaxis]
        probs = np.insert(probs, ridge.classes_[0], -probs[0], axis=1)

        probs = np.column_stack((probs, -probs)) if ridge.classes_[0] == 1 else np.column_stack((-probs, probs))

np.savetxt(args.testfile+".probs", probs)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

sys.exit(1)
