#!	/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Support vector machines
"""

try:
    import sys
    import os
    import numpy as np
    from sklearn import svm
    from sklearn.model_selection import GridSearchCV
    from sklearn.metrics import cohen_kappa_score, make_scorer
    from sklearn.datasets import load_svmlight_file
    import argparse
except ImportError:
    print("Modules not all present")
    sys.exit(0)

parser = argparse.ArgumentParser()
parser.add_argument("trainfile", help="Train dataset (libsvm format)", type=str)
parser.add_argument("testfile", help="Test dataset (libsvm format)", type=str)
parser.add_argument("kernel", help="Kernel: linear, polynomial, rbf, sigmoid", choices=['linear', 'polynomial', 'rbf', 'sigmoid'])
parser.add_argument("C", help="Number of features", type=float)
parser.add_argument("metric", help="Metric for CV of hyper-parameters", type=int)

# Optional
parser.add_argument("-cv", help="Validate parameters", action='store_true')
parser.add_argument("-s", help="Sparse format", action='store_true')
parser.add_argument("-coef0", help="r for sigmoid kernel", type=float)
parser.add_argument("-d", help="Degree for polynomial kernel", type=float)
parser.add_argument("-gamma", help="Gamma for RBF kernel", type=float)
args = parser.parse_args()

if args.s:
    # Sparse data
    X, y = load_svmlight_file(args.trainfile, n_features=args.nfeatures)
    X_test, y_test = load_svmlight_file(args.testfile, n_features=args.nfeatures)
else:
    import pandas as pd # Not imported before because it is not available in Skynet

    # Dense data
    # Read train file
    if not os.path.isfile(args.trainfile):
        print("Train dataset file '%s' does not exist" % args.trainfile)
        sys.exit()

    dtrain = pd.read_csv(args.trainfile, header=None, sep=" ")
    r, c = dtrain.shape

    # Read test file
    if not os.path.isfile(args.testfile):
        print("Test dataset file '%s' does not exist" % args.testfile)
        sys.exit()

    dtest = pd.read_csv(args.testfile, header=None, sep=" ")

    X = dtrain.loc[:, dtrain.columns != c-1]
    y = dtrain[c-1]
    X_test = dtest.drop(c-1, axis=1)

# Validate?
if args.cv:
    from sklearn.metrics import make_scorer, recall_score, cohen_kappa_score
    from metrics import gmean

    score = make_scorer(gmean, greater_is_better=True)
    kappa_scorer = make_scorer(cohen_kappa_score, greater_is_better=True)

    metrics = ["accuracy", score, kappa_scorer, "roc_auc", None, "recall_macro", None, None, "balanced_accuracy", "precision_macro", "f1_macro", None, "f1"]
    names = ["Accuracy", "G-mean", "Kappa", "auROC", None, "Recall macro", None, None, "Balanced accuracy", "Precision macro", "F1 macro", None, "F1"]

    parameters = [
        {"kernel": ["rbf"], "gamma": [0.0001, 0.001, 0.01, 0.1, 1.0, 10.0], "C": [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]},
        {"kernel": ["linear"], "C": [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]},
    ]

    # Fit model with grid search
    classifier = svm.SVC(kernel=args.kernel, coef0=args.coef0 if args.coef0 else 0.0, degree=args.d if args.d else 0.0, probability=True)
    svm = GridSearchCV(classifier, parameters, scoring=metrics[args.metric], n_jobs=-1, cv=3, verbose=3)
    svm.fit(X, y)
    print(svm.best_params_)

    # Save parameters
    f = open(args.trainfile + '.svc.parameters', 'w')
    f.write(names[args.metric] + ", ")
    f.write(str(svm.best_params_))
    f.close()
else:
    # Create SVM using SVC
    # Instantiate model with 1000 decision trees
    svm = svm.SVC(C=args.C, kernel=args.kernel, coef0=args.coef0 if args.coef0 else 0.0, degree=args.d if args.d else 0.0, gamma=args.gamma if args.gamma and args.gamma > 0.0 else 'scale',
                  probability=True)

    # Train the model on training data
    svm.fit(X, y)

# Save predictions
preds = svm.predict(X_test)
np.savetxt(args.testfile+".preds", preds, fmt='%d')

# Save probabilities for margin
probs = svm.predict_proba(X_test)
np.savetxt(args.testfile+".probs", probs)

sys.exit(1)
