/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "typex.h"
#include "fs.h"
#include "macros.h"
#include "svm.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "is.h"
#include "data.h"
#include "classifiers.h"
#include "ensemble.h"

#define MAXIMUM_ALPHA   25.0   // Maximum value of classifier weight
#define MAXIMUM_ALPHA_ADACOST   25.0
#define MINIMUM_ALPHA   0.001   // Minimum value of classifier weight

scalar UpdateAdaBoostWeightsFS(const t_Parameters *parm, int *subspace, scalar **data, int instances, int *labels, scalar *w, scalar *alpha, int inputs, int classes, bool madaboost, int minority,
                               int t)
{
    scalar e_i, beta, sum, *h;
    bool *ok;

    MALLOC(ok, bool, instances, "UpdateAdaBoostWeightsFS: ok");
    MALLOC(h, scalar, instances, "UpdateAdaBoostWeightsFS: h");

    // Test and update weights
    e_i = GetWeightedErrorFS(parm, parm->boostfs_classifier, subspace, data, inputs, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, h, NULL, 0, NULL, NULL,
                             !IsEnsemble(parm->boostfs_classifier));

    // If above 50% random re-sampling
    sum = 0.0;
    if ((e_i >= 0.5 || fabs(e_i - 0.5) <= ZERO) && parm->ensemble_method != AdaCost && parm->ensemble_method != AdaC1 && parm->ensemble_method != AdaC2 && parm->ensemble_method != AdaC3) {
        // Weight classifier
        *alpha = MINIMUM_ALPHA; // Almost remove classifier assigning a small weight

        for (int k = 0; k < instances; k++)
            sum += (w[k] = Poisson());
    } else if (fabs(e_i) <= ZERO && parm->ensemble_method != AdaCost && parm->ensemble_method != AdaC1 && parm->ensemble_method != AdaC2 && parm->ensemble_method != AdaC3) {
        // If perfect, also random resampling
        // Weight classifier
        *alpha = MAXIMUM_ALPHA;

        for (int k = 0; k < instances; k++)
            sum += (w[k] = Poisson());
    } else {
        switch (parm->ensemble_method) {
            // Different variants of boosting
            case AsymmetricAdaBoost:
                // Asymmetric AdaBoost?
                // Weight classifier
                *alpha = MIN(MAXIMUM_ALPHA, log(1.0/(e_i/(1.0 - e_i))));

                for (int k = 0; k < instances; k++) {
                    scalar C = (labels[k] == minority) ? pow(sqrt(parm->asymmetric_adaboost_k), 1.0/t) : pow(sqrt(parm->asymmetric_adaboost_k), -1.0/t);
                    if (ok[k])
                        sum += (w[k] *= C*exp(-(*alpha)));
                    else
                        // For eMadaBoost do not increment weights
                        sum += (w[k] *= C*exp(*alpha));
                }
                break;

            case AdaCost:
                // Weight classifier
                if (fabs(e_i + 1.0) < ZERO)
                    *alpha = -MAXIMUM_ALPHA_ADACOST;
                else if (fabs(e_i - 1.0) < ZERO)
                    *alpha = MAXIMUM_ALPHA_ADACOST;
                else
                    *alpha = fabs(0.5*log((1.0+e_i)/(1.0-e_i)));

                for (int k = 0; k < instances; k++) {
                    beta = (ok[k] ? -1.0 : 1.0)*0.5*(labels[k] == minority ? parm->C_plus : parm->C_minus)+0.5;
                    sum += (w[k] *= exp(-(*alpha)*h[k]*(labels[k] == minority ? 1.0 : -1.0)*beta));
                }
                break;

            case AdaC1:
            case AdaC2:
            case AdaC3:
                if (fabs(e_i + 1.0) < ZERO)
                    *alpha = MAXIMUM_ALPHA;
                else if (e_i < 1.0)
                    *alpha = MINIMUM_ALPHA;
                else
                    *alpha = 0.5*log(e_i);

                if (parm->ensemble_method == AdaC1)
                    for (int k = 0; k < instances; k++)
                        sum += (w[k] *=  exp(-1.0 * (*alpha) * (labels[k] == minority ? parm->C_plus : -parm->C_minus) * h[k]));
                else if (parm->ensemble_method == AdaC2)
                    for (int k = 0; k < instances; k++)
                        sum += (w[k] *= (labels[k] == minority ? parm->C_plus : parm->C_minus) * exp(-(*alpha) * h[k] * (labels[k] == minority ? 1.0 : -1.0)));
                else // AdaC3
                    for (int k = 0; k < instances; k++)
                        sum += (w[k] *= (labels[k] == minority ? parm->C_plus : parm->C_minus) * exp(-1.0 * (*alpha) * (labels[k] == minority ? parm->C_plus : -parm->C_minus) * h[k]));
                break;

            case DiscreteAdaBoost:
            case AdaBoost:
            case eMadaBoost:
            case BalanceCascade:
            case EasyEnsemble:
                // Weight classifier
                *alpha = MIN(MAXIMUM_ALPHA, 0.5*log((1.0-e_i)/e_i));

                for (int k = 0; k < instances; k++)
                    if (ok[k])
                        sum += (w[k] *= sqrt(e_i/(1.0-e_i)));
                    else
                        // For eMadaBoost do not increment weights
                        sum += (w[k] /= (madaboost ? 1.0 : sqrt((1.0-e_i)/e_i)));
                break;

            case RealAdaBoost:
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "RealAdaBoost only implemented for two-class problems");
                *alpha = MIN(MAXIMUM_ALPHA, log(1.0/(e_i/(1.0 - e_i))));

                for (int k = 0; k < instances; k++)
                    sum += (w[k] *= exp(-1.0*h[k]*(labels[k] == minority ? 1.0 : -1.0)));
                break;

            case CSB0:
                // Weight classifier
                *alpha = MIN(MAXIMUM_ALPHA, log(1.0/(e_i/(1.0 - e_i))));

                for (int k = 0; k < instances; k++)
                    sum += (w[k] *= ok[k] ? 1.0 : (labels[k] == minority ? parm->C_plus : parm->C_minus));
                break;

            case CSB1:
                // Weight classifier
                *alpha = MIN(MAXIMUM_ALPHA, log(1.0/(e_i/(1.0 - e_i))));

                for (int k = 0; k < instances; k++)
                    sum += (w[k] *= (ok[k] ? 1.0 : (labels[k] == minority ? parm->C_plus : parm->C_minus)) * exp(-h[k] * (labels[k] == minority ? 1.0 : -1.0)));
                break;

            case CSB2:
                // Weight classifier
                *alpha = MIN(MAXIMUM_ALPHA, log(1.0/(e_i/(1.0 - e_i))));

                for (int k = 0; k < instances; k++)
                    sum += (w[k] *= (ok[k] ? 1.0 : (labels[k] == minority ? parm->C_plus : parm->C_minus)) * exp(- (*alpha) * h[k] * (labels[k] == minority ? 1.0 : -1.0)));
                break;

            default:
                OnErrorExit(ERR_PARAM, "No approppriate boosting method");
        }
    }

    // Normalize w[] to sum 1
    for (int k = 0; k < instances; k++)
        w[k] *= 1.0/sum;

    FREE(ok);
    FREE(h);

    return e_i;
}

#define MAXIMUM_WEIGHT  1.0

void UpdateGentleBoostWeightsFS(const t_Parameters *parm, int *subspace, scalar **data, int instances, int *labels, scalar *w, int inputs, int classes, bool cv, int minority)
{
    scalar *h, sum = 0.0;
    bool *ok;

    if (parm->weigthed_sampling_for_training)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weigthed sampling");

    MALLOC(h, scalar, instances, "UpdateGentleBoostWeightsFS: h");
    MALLOC(ok, bool, instances, "UpdateGentleBoostWeightsFS: ok");
    GetWeightedErrorFS(parm, parm->boostfs_classifier, subspace, data, inputs, instances, labels, classes, NULL, w, parm->cv_hyperparameters, NULL, minority, h, NULL, 0, NULL, NULL,
                       !IsEnsemble(parm->boostfs_classifier));

    for (int j = 0; j < instances; j++) {
        // Update weights
        w[j] = MIN(MAXIMUM_WEIGHT, w[j]*exp(-h[j]));

        sum += w[j];
    }

    // Normalize to sum 1
    for (int i = 0; i < instances; i++)
        w[i] *= 1.0/sum;
}

scalar UpdateeReweightBoostWeightsFS(const t_Parameters *parm, int **subspaces, scalar **data, int instances, int *labels, scalar *w, scalar *alpha, int inputs, int classes, int size, int iteration,
                                    t_Learner *learners, bool cv, int minority)
{
    //     scalar e_i, sum;
    //     bool *ok;

    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Needs updating");

    //     MALLOC(ok, bool, instances, "UpdateeReweightBoostWeightsFS: ok");
    //
    //     // Train new classifier if not k-NN
    //     if (parm->boostfs_classifier == C45) {
    //         if (parm->weigthed_sampling_for_training) {
    //             scalar *cumulative_probabilities, **s_data;
    //             int *s_labels;
    //
    //             MALLOC(cumulative_probabilities, scalar, instances, "UpdateeReweightBoostWeightsFS: cumulative_probabilities");
    //             MALLOC(s_data, scalar *, instances, "UpdateeReweightBoostWeightsFS: s_data");
    //             MALLOC(s_labels, int, instances, "UpdateeReweightBoostWeightsFS: s_labels");
    //
    //             cumulative_probabilities[0] = w[0];
    //             for (int k = 1; k < instances; k++)
    //                 cumulative_probabilities[k] = cumulative_probabilities[k-1] + w[k];
    //             for (int k = 0; k < instances; k++) {
    //                 int instance = RouletteSelection(cumulative_probabilities, instances);
    //
    //                 s_labels[k] = labels[instance];
    //                 s_data[k] = data[instance];
    //             }
    //
    //             if (cv)
    //                 SetCrossValidatedParametersSubspace((t_Parameters *) parm, s_data, instances, s_labels, subspaces[iteration-1], inputs, classes, C45, minority);
    //
    //             trees[iteration-1]  = TrainC45ClassifierSubspace(parm, s_data, subspaces[iteration], instances, s_labels, inputs, true);
    //
    //             FREE(s_data);
    //             FREE(cumulative_probabilities);
    //             FREE(s_labels);
    //         } else if (parm->random_undersampling_for_training) {
    //             scalar **u_data = NULL;
    //             int u_instances = 0, *u_labels = NULL;
    //             TemporalUndersampleDataForTraining(data, instances, classes, labels, parm->random_undersampling_ratio, &u_data, &u_labels, &u_instances, minority);
    //
    //             if (cv)
    //                 SetCrossValidatedParametersSubspace((t_Parameters *) parm, u_data, u_instances, u_labels, subspaces[iteration-1], inputs, classes, C45, minority);
    //
    //             trees[iteration-1] = TrainC45ClassifierSubspace(parm, u_data, subspaces[iteration-1], u_instances, u_labels, inputs, true);
    //             FREE(u_data);
    //             FREE(u_labels);
    //         } else {
    //             if (cv)
    //                 SetCrossValidatedParametersSubspace((t_Parameters *) parm, data, instances, labels, subspaces[iteration-1], inputs, classes, C45, minority);
    //
    //             trees[iteration-1] = TrainC45ClassifierSubspace(parm, data, subspaces[iteration-1], instances, labels, inputs, true);
    //         }
    //     } else if (parm->boostfs_classifier == SVM) {
    //         if (parm->weigthed_sampling_for_training)
    //             OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted undersampling for SVM method");
    //         else if (parm->random_undersampling_for_training) {
    //             scalar **u_data = NULL;
    //             int u_instances = 0, *u_labels = NULL;
    //             TemporalUndersampleDataForTraining(data, instances, classes, labels, parm->random_undersampling_ratio, &u_data, &u_labels, &u_instances, minority);
    //
    //             if (cv)
    //                 SetCrossValidatedParametersSubspace((t_Parameters *) parm, u_data, u_instances, u_labels, subspaces[iteration-1], inputs, classes, SVM, minority);
    //
    //             svms[iteration-1] = TrainSVMClassifierSubspace(parm, u_data, subspaces[iteration-1], u_instances, u_labels, inputs);
    //             FREE(u_data);
    //             FREE(u_labels);
    //         } else {
    //             if (cv)
    //                 SetCrossValidatedParametersSubspace((t_Parameters *) parm, data, instances, labels, subspaces[iteration-1], inputs, classes, SVM, minority);
    //             svms[iteration-1] = TrainSVMClassifierSubspace(parm, data, subspaces[iteration-1], instances, labels, inputs);
    //         }
    //     }
    //
    //     // Test and update weights
    //     // Get weighted error
    //     scalar *h;
    //     MALLOC(h, scalar, instances, "UpdateeReweightBoostWeightsFS: h");
    //     // MUST MODIFY
    //     e_i = GetWeightedErrorFS(parm, parm->boostfs_classifier, subspaces[iteration-1], data, inputs, instances, labels, classes, ok, w, cv, parm->boostfs_classifier == C45 ? trees[iteration-1] : NULL,
    //                              parm->boostfs_classifier == SVM ? svms[iteration-1] : NULL, minority, h, NULL, -1, NULL);
    //     GetEnsembleAccuracy(parm, parm->boostfs_classifier, data, labels, subspaces, instances, parm->k, classes, inputs, svms, trees, iteration, alpha, ok, iteration-size, parm->use_g_measure,
    //                         parm->sampling_size, minority);
    //     FREE(h);
    //     sum = 0.0;
    //
    //     // Weight classifier
    //     alpha[iteration] = MIN(MAXIMUM_ALPHA, 0.5*log((1.0-e_i) /e_i));
    //
    //     for (int k = 0; k < instances; k++)
    //         if (ok[k])      // Ok? If yes do not modify weight
    //             sum += w[k];
    //         else
    //             sum += (w[k] *= exp(-(*alpha)));
    //
    //     // Normalize w[] to sum 1
    //     for (int k = 0; k < instances; k++)
    //         w[k] *= 1.0/sum;
    //
    //     FREE(ok);
    //
    //     return e_i;
    return 0.0;
}
