/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include <omp.h>
#include "macros.h"
#include "common.h"
#include "maths.h"
#include "typex.h"

// Deal with cache function for Margin Based Feature Selection
// IMPORTANT: Cache stores squared distances

// Update cache after adding/removing feature
#ifdef _OPENMP

void UpdateCache(scalar **dist_cache, int added, int removed, scalar **data, int instances, int **index)
{
    // Added
    if (added != -1) {
        if (index)
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    dist_cache[j][i] += SQR(SparseFeature(data[i], index[i], added) - SparseFeature(data[j], index[j], added));
        else
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    dist_cache[j][i] += SQR(data[i][added] - data[j][added]);

    } else if (removed != -1) {
        if (index)
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    dist_cache[j][i] -= SQR(SparseFeature(data[i], index[i], removed) - SparseFeature(data[j], index[j], removed));
        else
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    dist_cache[j][i] -= SQR(data[i][removed] - data[j][removed]);
    } else
        OnErrorExit(ERR_PARAM, "Cannot update cache with no modifications");
}

#else

void UpdateCache(scalar **dist_cache, int added, int removed, scalar **data, int instances, int **index)
{
    // Added
    if (added != -1) {
        if (index)
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    dist_cache[j][i] = (dist_cache[i][j] += SQR(SparseFeature(data[i], index[i], added) - SparseFeature(data[j], index[j], added)));
        else
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    dist_cache[j][i] = (dist_cache[i][j] += SQR(data[i][added] - data[j][added]));

    } else if (removed != -1) {
        if (index)
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    dist_cache[j][i] = (dist_cache[i][j] -= SQR(SparseFeature(data[i], index[i], removed) - SparseFeature(data[j], index[j], removed)));
        else
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    dist_cache[j][i] = (dist_cache[i][j] -= SQR(data[i][removed] - data[j][removed]));
    } else
        OnErrorExit(ERR_PARAM, "Cannot update cache with no modifications");
}

#endif

void ReserveAndInitializeCache(scalar ***dist_cache, int instances)
{
    MALLOC(*dist_cache, scalar *, instances, "ReserveAndInitializeCache: *dist_cache");
    for (int i = 0; i < instances; i++)
        CALLOC((*dist_cache)[i], scalar, instances, "ReserveAndInitializeCache: (*dist_cache)[i]");
}

void InitializeCacheToActualSqDistance(scalar **dist_cache, int inputs, scalar **data, int instances, int **index)
{
    if (index) {
        for (int i = 0; i < instances; i++) {
            dist_cache[i][i] = 0.0;
            for (int j = 0; j < i; j++)
                dist_cache[i][j] = dist_cache[j][i] = SQR(EuclideanDistanceSparseData(data[i], data[j], inputs, index[i], index[j]));
        }
    } else {
        for (int i = 0; i < instances; i++) {
            dist_cache[i][i] = 0.0;
            for (int j = 0; j < i; j++)
                dist_cache[i][j] = dist_cache[j][i] = SQR(EuclideanDistance(data[i], data[j], inputs));
        }
    }
}
