/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "macros.h"
#include "common.h"
#include "svm.h"
#include "fs.h"
#include "is.h"
#include "data.h"
#include "classifiers.h"

// Get weighted error and vector of classified instances
scalar GetWeightedErrorFS(const t_Parameters *parm, t_Classifier classifier, int *subspace, scalar **data, int inputs, int instances, int *labels, int classes, bool *ok, scalar *w, bool cv,
                          t_Learner *learner, int minority, scalar *h, scalar **t_data, int t_instances, int *t_labels, t_Ensemble *ensemble, bool single)
{
    scalar error = 0.0, *local_w = NULL;
    t_Ensemble lensemble;
    t_Learner llearner;

    // If no weights given get equal distribution
    if (!w) {
        MALLOC(local_w, scalar, instances, "GetWeightedErrorFS: local_w");
        for (int i = 0; i < instances; i++)
            local_w[i] = 1.0/instances;
    } else
        local_w = w;

    if (parm->weigthed_sampling_for_training)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted undersampling for boost FS method");

    // Train classifier?
    if (single) {
        if (!learner)
            ClassifierTraining(parm, classifier, t_data ? t_data : data, t_instances ? t_instances : instances, t_labels ? t_labels : labels, inputs, classes, NULL, subspace, NULL, true, &llearner,
                               NULL);
        else
            llearner = *learner;
    } else {
        if (!ensemble)
            ClassifierTraining(parm, classifier, t_data ? t_data : data, t_instances ? t_instances : instances, t_labels ? t_labels : labels, inputs, classes, NULL, subspace, NULL, true, NULL,
                               &lensemble);
        else
            lensemble = *ensemble;
    }
    // Get margin
    for (int k = 0; k < instances; k++) {
        h[k] = ClassifierMargin(parm, classifier, data[k], inputs, classes, NULL, subspace, labels[k], data, instances, labels, NULL, &llearner, &lensemble);

        ok[k] = (h[k] > 0.0);
        if (!ok[k])
            error += local_w[k];
    }

    // For AdaCost error function is different
    if (parm->ensemble_method == AdaCost) {
        if (!h)
            OnErrorExit(ERR_PARAM, "h cannot be null");
        for (int i = 0; i < instances; i++)
            error += local_w[i] * (labels[i] == minority ? 1.0 : -1.0) * h[i] * ((ok[i] ? -1.0 : 1.0)*0.5*(labels[i] == minority ? parm->C_plus : parm->C_minus)+0.5);

    } else if (parm->ensemble_method == AdaC1 || parm->ensemble_method == AdaC2) {
        scalar err = 0.0, acc = 0.0;

        for (int i = 0; i < instances; i++)
            if (!ok[i])
                err += local_w[i] * (labels[i] == minority ? parm->C_plus : parm->C_minus);
            else
                acc += local_w[i] * (labels[i] == minority ? parm->C_plus : parm->C_minus);

        if (parm->ensemble_method == AdaC1)
            error = fabs(1.0-acc+err) > ZERO ? (1.0+acc-err)/(1.0-acc+err) : -1.0;
        else if (parm->ensemble_method == AdaC2)
            error = fabs(err) > ZERO ? acc/err : -1.0;    // -1 means all instances Ok
    } else if (parm->ensemble_method == AdaC3) {
        scalar err = 0.0, acc = 0.0, cost = 0.0;

        for (int i = 0; i < instances; i++) {
            cost += local_w[i] * (labels[i] == minority ? parm->C_plus : parm->C_minus);
            if (!ok[i])
                err += local_w[i] * pow(labels[i] == minority ? parm->C_plus : parm->C_minus, 2.0);
            else
                acc += local_w[i] * pow(labels[i] == minority ? parm->C_plus : parm->C_minus, 2.0);
        }

        error = fabs(cost-acc+err) > ZERO ? (cost+acc-err)/(cost-acc+err) : -1.0;    // -1 means all instances Ok
    } else
        for (int i = 0; i < instances; i++)
            if (!ok[i])
                error += local_w[i];

    if (!learner)
        ClassifierFree(parm, &llearner, classifier, classes, inputs, NULL);

    return error;
}
