/****************************************************************************
* The algorithm implemented here can be found in the book titled            *
*    FEATURE SELECTION FOR KNOWLEDGE DISCOVERY by Huan Liu &                *
* Hiroshi Motoda published by Kluwer in 1998, ISBN 0-7923-8198-X            *
*                                                                           *
* The program developed by:                                                 *
*                                                                           *
*                       Md. Farhad Hussain                                  *
*                           DISCS, NUS                                      *
*                    Kent Ridge, S'pore 119260                              *
*                  email: farhad@iscs.nus.edu.sg                            *
*                          liuh@iscs.nus.edu.sg                             *
*****************************************************************************/

#include <time.h>
#include "filter.h"
#include "fs.h"

/*------------------------------Global definitions---------------------------*/
long Delta;                       /* Contains the lowest inconsistancy count */
ATTRIBLIST *ParentLevel, *ChildLevel, *PrunedList;
int numCalculated = 0;

/*---------------------------------------------------------------------------*/
void InitABB(int COL, scalar **DataSet, int ROW)
{

    ParentLevel = (ATTRIBLIST *)malloc(sizeof(ATTRIBLIST));
    SUCCESS(ParentLevel);
    ParentLevel->mask = (char *)malloc(COL - 1);
    SUCCESS(ParentLevel->mask);
    IncludeAllFeatures(ParentLevel->mask, COL);
    ParentLevel->Next = NULL;
    LoadInstances(ParentLevel->mask, ROW, DataSet, COL);
    Delta  = InconsistancyCount();
    FreeHashTable();
    ChildLevel = NULL;
    PrunedList = NULL;
}
/*---------------------------------------------------------------------------*/
char *GetChildSubSet(ATTRIBLIST *Parent, long SubSetNo, int COL)
{
    char *mask;
    long FeatureNo = 0, i;

    mask = (char *)malloc(COL - 1);
    SUCCESS(mask);
    for (i = 0; i < COL - 1; i++) {
        mask[i] = Parent->mask[i];

        if (mask[i]) {
            FeatureNo ++;
            if (FeatureNo == SubSetNo)
                mask[i] = 0;
        }
    }
    return mask;
}
/*---------------------------------------------------------------------------*/
long HammingDistanceChar(char *SubSet1, char *SubSet2, int COL)
{
    long i, diff = 0;

    for (i = 0; i < COL - 1; i++)
        if (SubSet1[i] != SubSet2[i])
            diff++;
    return diff;
}
/*---------------------------------------------------------------------------*/
BOOL Legitimated(char *SubSet, int COL)
{
    ATTRIBLIST *PrunedAttrib;

    PrunedAttrib = PrunedList;
    while (PrunedAttrib != NULL) {
        if (HammingDistanceChar(PrunedAttrib->mask, SubSet, COL) == 1)
            return FALSE;
        PrunedAttrib = PrunedAttrib->Next;
    }
    return TRUE;
}
/*---------------------------------------------------------------------------*/
void InsertInPrunedList(char *SubSet, int COL)
{
    ATTRIBLIST *NewSubSet;
    long i;

    NewSubSet = PrunedList;                     /* Check already exists or not */
    while (NewSubSet != NULL) {
        if (HammingDistanceChar(NewSubSet->mask, SubSet, COL) == 0)
            return ;
        NewSubSet = NewSubSet->Next;
    }

    NewSubSet = (ATTRIBLIST *)malloc(sizeof(ATTRIBLIST));
    SUCCESS(NewSubSet);
    NewSubSet->mask = (char *)malloc(COL - 1);
    SUCCESS(NewSubSet->mask);
    for (i = 0; i < COL - 1; i++)
        NewSubSet->mask[i] = SubSet[i];
    NewSubSet->Next = PrunedList;
    PrunedList = NewSubSet;
}
/*---------------------------------------------------------------------------*/
char InsertInChildLevel(char *SubSet, int COL, scalar **DataSet, int ROW)
{
    ATTRIBLIST *NewSubSet;
    long i;

    NewSubSet = ChildLevel;                     /* Check already exists or not */
    while (NewSubSet != NULL) {
        if (HammingDistanceChar(NewSubSet->mask, SubSet, COL) == 0)
            return 1;
        NewSubSet = NewSubSet->Next;
    }

    if (InconCal(SubSet, ROW, DataSet, COL) <= Delta) {
    } else
        return 0;

    NewSubSet = (ATTRIBLIST *)malloc(sizeof(ATTRIBLIST));
    SUCCESS(NewSubSet);
    NewSubSet->mask = (char *)malloc(COL - 1);
    SUCCESS(NewSubSet->mask);
    for (i = 0; i < COL - 1; i++)
        NewSubSet->mask[i] = SubSet[i];
    NewSubSet->Next = ChildLevel;
    ChildLevel = NewSubSet;
    return 1;
}
/*---------------------------------------------------------------------------*/
long NumOfNewSubSet(ATTRIBLIST  *Attrib, int COL)
{

    return NumOfFeatures(Attrib->mask, COL);
}
/*---------------------------------------------------------------------------*/
long InconCal(char *mask, int ROW, scalar **DataSet, int COL)
{
    long InconCnt;

    /****added by DASH ****************/
    numCalculated++;
    /**********************************/
    LoadInstances(mask, ROW, DataSet, COL);                /* Load according to the mask settings */
    InconCnt = InconsistancyCount();                /* Calculate Inconsistancy */
    FreeHashTable();                                      /* Unload hash table */
    return InconCnt;
}
/*---------------------------------------------------------------------------*/
void PrintSubSet(ATTRIBLIST  *Attrib, int COL)
{

    DisplayFeatures(Attrib->mask, COL);
}
/*---------------------------------------------------------------------------*/
void GenerateChildSubSets(int COL, scalar **DataSet, int ROW)       /* Generates valid children of all parents */
{
    ATTRIBLIST *ParentPTR;
    char *SubSet;
    long NumOfSubSet, SubSetNo;

    ParentPTR = ParentLevel;                        /* Start with first parent */
    while (ParentPTR != NULL) {
        NumOfSubSet = NumOfNewSubSet(ParentPTR, COL);

        if (NumOfSubSet == 0)
            return;

        for (SubSetNo = 1; SubSetNo <= NumOfSubSet; SubSetNo++) {
            SubSet = GetChildSubSet(ParentPTR, SubSetNo, COL);

            if (Legitimated(SubSet, COL)) {          /* child of pruned set ? */
                if (! InsertInChildLevel(SubSet, COL, DataSet, ROW))
                    InsertInPrunedList(SubSet, COL);
            } else {
            }
            free(SubSet);
        }
        ParentPTR = ParentPTR->Next;
    }
}
/*---------------------------------------------------------------------------*/
void DeleteOldPrunedSet(ATTRIBLIST *PrunedSet)
{
    ATTRIBLIST *LastAttrib;

    /* PrunedSet points to the beginning of the old pruned list */

    if (PrunedSet == NULL)
        return;

    if (PrunedSet == PrunedList) {
        PrunedSet = PrunedList->Next;
        while (PrunedSet != NULL) {
            free(PrunedList->mask);
            free(PrunedList);
            PrunedList = PrunedSet;
            PrunedSet = PrunedList->Next;
        }
        free(PrunedList->mask);
        free(PrunedList);
        PrunedList = NULL;
        return;
    }

    LastAttrib = PrunedList;
    while (LastAttrib->Next != PrunedSet)
        LastAttrib = LastAttrib->Next;
    LastAttrib->Next = NULL;

    LastAttrib = PrunedSet->Next;
    while (LastAttrib != NULL) {
        free(PrunedSet->mask);
        free(PrunedSet);
        PrunedSet = LastAttrib;
        LastAttrib = PrunedSet->Next;
    }
    free(PrunedSet->mask);
    free(PrunedSet);
}
/*---------------------------------------------------------------------------*/
void SetNewParentLevel()
{
    ATTRIBLIST *AttribPTR;

    AttribPTR = ParentLevel->Next;
    while (AttribPTR != NULL) {
        free(ParentLevel->mask);
        free(ParentLevel);
        ParentLevel = AttribPTR;
        AttribPTR = ParentLevel->Next;
    }
    free(ParentLevel->mask);
    free(ParentLevel);
    ParentLevel = ChildLevel;
    ChildLevel = NULL;
}
/*---------------------------------------------------------------------------*/
void PrintParentList(int COL)
{
    ATTRIBLIST *AttribPTR;
    printf("\n----Parent List---\n");
    AttribPTR = ParentLevel;
    while (AttribPTR != NULL) {
        PrintSubSet(AttribPTR, COL);
        AttribPTR = AttribPTR->Next;
    }
}
/*---------------------------------------------------------------------------*/
void PrintChildList(int COL)
{
    ATTRIBLIST *AttribPTR;
    printf("\n----Child List---\n");
    AttribPTR = ChildLevel;
    while (AttribPTR != NULL) {
        PrintSubSet(AttribPTR, COL);
        AttribPTR = AttribPTR->Next;
    }
}
/*---------------------------------------------------------------------------*/
void PrintPrunedList(int COL)
{
    ATTRIBLIST *AttribPTR;
    printf("\n----Pruned List---\n");
    AttribPTR = PrunedList;
    while (AttribPTR != NULL) {
        PrintSubSet(AttribPTR, COL);
        AttribPTR = AttribPTR->Next;
    }
}
/*---------------------------------------------------------------------------*/
void RunABB(scalar **DataSet, int COL, int ROW, time_t time_limit)
{
    ATTRIBLIST *OldPrunedSetPTR = NULL, *NewPrunedSetPTR = NULL;

    time_t init_time = time(NULL);

    InitABB(COL, DataSet, ROW);
    OldPrunedSetPTR = PrunedList;
    FOREVER {
        NewPrunedSetPTR = PrunedList;
        DeleteOldPrunedSet(OldPrunedSetPTR);
        if (PrunedList == NULL)
            NewPrunedSetPTR = NULL;
        GenerateChildSubSets(COL, DataSet, ROW);
        if (ChildLevel == NULL || (time_limit && time(NULL) - init_time > time_limit))
        {
            /* No child generated from the parent level */
            //             printf("\nBest Selected Features\n");
            //             PrintSubSet(ParentLevel, COL);
            //             printf("\n");
            //             printf("NUMBER OF SUBSETS EVALUATED = %d\n", numCalculated);
            //             exit(0);
            return;
        }
        OldPrunedSetPTR =  NewPrunedSetPTR;
        SetNewParentLevel();
    }
}
