/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <setjmp.h>
#include "fs.h"
#include "common.h"
#include "macros.h"
#include "filter.h"

/* Selection of features using adaptive branch and bound algorithm
 *
 * H. Liu, H. Motoda and M. Dash, "A monotonic measure for optimal feature selection", Proc. of the Tenth European Conference on Machine Learning, 101-106, 1998.
 *
 */

void AutomaticBranchAndBound(scalar **data, int n, int *labels, int *subspace, int n_features, time_t time_limit)
{
    extern ATTRIBLIST *ParentLevel;
    int selected_features, *list;

    // Construct new data with labels in the last column
    scalar **dataset;

    // Count features, not all have to be selected
    MALLOC(list, int, n_features, "AutomaticBranchAndBound: list");
    selected_features = 0;
    for (int i = 0; i < n_features; i++)
        if (subspace[i]) {
            list[selected_features] = i;
            selected_features++;
        }

    MALLOC(dataset, scalar *, n, "AutomaticBranchAndBound: dataset");
    for (int i = 0; i < n; i++) {
        MALLOC(dataset[i], scalar, selected_features+1, "AutomaticBranchAndBound: dataset[i]");
        for (int j = 0; j < selected_features; j++)
            dataset[i][j] = data[i][list[j]];
        dataset[i][selected_features] = (scalar) labels[i];
    }

    // Call function
    CreateHashTableBB();
    RunABB(dataset, selected_features+1, n, time_limit);

    // Free dataset
    for (int i = 0; i < n; i++)
        FREE(dataset[i]);
    FREE(dataset);

    // Set features
    for (int i = 0; i < selected_features; i++)
        subspace[list[i]] = ParentLevel->mask[i] ? 1 : 0;

    FREE(list);
}

TREENODE *ResultNode;
void BranchAndBound(scalar **data, int n, int *labels, int *subspace, int n_features, time_t time_limit)
{
    extern long BestIncon;
    jmp_buf env_buffer;
    TREENODE *RootNode;
    time_t init_time;

    BestIncon = 0;

    // Construct new data with labels in the last column
    scalar **dataset;

    MALLOC(dataset, scalar *, n, "BranchAndBound: dataset");
    for (int i = 0; i < n; i++) {
        MALLOC(dataset[i], scalar, n_features+1, "BranchAndBound: dataset[i]");
        for (int j = 0; j < n_features; j++)
            dataset[i][j] = data[i][j];
        dataset[i][n_features] = (scalar) labels[i];
    }

    // Call function
    CreateHashTableBB();
    RootNode = InitRoot(n_features+1);

    /* Save calling environment for longjmp */
    int val = setjmp(env_buffer);

    // First call to setjmp always returns 0
    if (val == 0) {
        init_time = time(NULL);
        B_and_B(RootNode, n, dataset, n_features+1, env_buffer, time_limit, init_time);
    }

    // Free dataset
    for (int i = 0; i < n; i++)
        FREE(dataset[i]);
    FREE(dataset);

    // Set features
    for (int i = 0; i < n_features; i++)
        subspace[i] = ResultNode->mask[i] ? 1 : 0;

    FreeNode(ResultNode);
}
