/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "filter.h"
#include "data.h"

/*
 *  M. A. Hall, "Correlation-based feature selection for discrete and numeric class machine learning".
 */

void CorrelationBasedFilterSequentialForwardSelection(const t_Parameters *parm, scalar **data, int inputs, int instances, int classes, int *labels, int *S)
{
    int size = 0, *list;
    static int *intervals = NULL;
    bool improve = false;
    scalar overall_best_value = 0.0;
    static scalar **limits = NULL;
    static t_EntropyInfo *entropy = NULL;

    time_t init_time = time(NULL);

    // If discretizer is NONE use continuous attributes
    if (parm->discretizer != dNONE && limits == NULL) {
        //Memory allocation for auxiliary data structure
        MALLOC(limits, scalar *, inputs, "CorrelationBasedFilterSequentialForwardSelection: limits");
        //Set limits to NULL
        for (int k = 0; k < inputs; k++)
            limits[k] = NULL;
        MALLOC(intervals, int, inputs, "CorrelationBasedFilterSequentialForwardSelection: intervals");

        // Discretization of dataset
        switch (parm->discretizer) {
            case dFAYYAD:
                FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
                break;

            case dWPKID:
                WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
                break;

            case dNONE:
                //Do nothing
                break;
        }

        if (parm->verbose > 2)
            fprintf(parm->out_file, "Discretization time: %ld\n", time(NULL) - init_time);

        // Calculate symmetric uncertainty for every variable
        MALLOC(entropy, t_EntropyInfo, inputs, "CorrelationBasedFilterSequentialForwardSelection: entropy");
        for (int i = 0; i < inputs; i++) {
            scalar Hi = Entropy(data, i, instances, limits, intervals[i]);
            scalar Hc = ClassEntropy(instances, labels, classes);
            scalar Hic = ConditionalClassEntropy(data, i, instances, limits, intervals[i], classes, labels);

            entropy[i].index = i;
            entropy[i].Hi = Hi;
            entropy[i].SUic = (Hi-Hic)/(Hi+Hc);
        }
    }

    // Start with empty subset
    for (int i = 0; i < inputs; i++)
        S[i] = 0;

    init_time = time(NULL);

    // Step 1: Remove correlated features
    MALLOC(list, int, inputs, "CorrelationBasedFilterSequentialForwardSelection: list");
    do {
        int best_index = -1;
        scalar best_value = 0.0;

        // Add new unselected feature
        size++;
        for (int i = 0; i < inputs; i++)
            if (!S[i]) {
                list[size-1] = i;

                scalar value = MeritS(data, list, size, labels, instances, entropy, intervals, limits, parm->discretizer);

                if (value > best_value) {
                    best_value = value;
                    best_index = i;
                }
            }

        // Check whether improvement is made
        if (best_value > overall_best_value) {
            list[size-1] = best_index;
            S[best_index] = 1;
            overall_best_value = best_value;
            improve = true;
        } else
            improve = false;
    } while (improve && size < inputs && (!parm->time_limit || time(NULL) - init_time < parm->time_limit));

    for (int i = 0; i < size; i++)
        S[list[i]] = 1;

    // Test unselected instances
    // Recover if its correlation with the class is higher than the highest correlation between it and any one of the already selected features
    scalar r_cf = -1.0, r_ff = -1;
    for (int i = 0; i < inputs && (!parm->time_limit || time(NULL) - init_time < parm->time_limit); i++)
        if (!S[i]) {
            bool higher = true;
            if (parm->discretizer == dNONE)
                r_cf = FeatureClassCorrelation(data, i, labels, instances);
            else
                r_cf = entropy[i].SUic;
            for (int j = 0; j < inputs && higher; j++)
                if (S[j] && i != j) {
                    if (parm->discretizer == dNONE)
                        r_ff = FeatureFeatureCorrelation(data, i, j, instances);
                    else {
                        scalar Hij = ConditionalEntropy(data, i, j, instances, limits, intervals[i], intervals[j]);
                        r_ff = (entropy[i].Hi-Hij)/(entropy[i].Hi+entropy[j].Hi);
                    }

                    if (r_cf <= r_ff)
                        higher = false;
                }

            if (higher) {
                S[i] = 1;
                fprintf(parm->out_file, "Recovered feature %d\n", i+1);
            }
        }

    FREE(list);
}

scalar MeritS(scalar **data, int *s, int k, int *labels, int instances, t_EntropyInfo *entropy, int *intervals, scalar **limits, t_Discretizer discretizer)
{
    scalar r_cf = 0, r_ff = 0;

    // Average feature class correlation
    for (int i = 0; i < k; i++)
        if (discretizer == dNONE)
            r_cf += FeatureClassCorrelation(data, s[i], labels, instances);
        else
            r_cf += entropy[s[i]].SUic;
    r_cf /= k;

    if (k > 1) {
        for (int i = 1; i < k; i++)
            for (int j = 0; j < i; j++) {
                if (discretizer == dNONE)
                    r_ff += FeatureFeatureCorrelation(data, s[i], s[j], instances);
                else {
                    scalar Hij = ConditionalEntropy(data, s[i], s[j], instances, limits, intervals[s[i]], intervals[s[j]]);
                    r_ff += (entropy[s[i]].Hi-Hij)/(entropy[s[i]].Hi+entropy[s[j]].Hi);
                }
            }

        r_ff /= (k*(k-1))/2;
    } else
        r_ff = 1.0;

    return (k*r_cf)/sqrt(k+k*(k-1)*r_ff);
}


scalar FeatureClassCorrelation(scalar **data, int feature, int *labels, int n)
{
    scalar s_xy, s_x, s_y, s_x2, s_y2, r;

    s_xy = s_x = s_y = s_x2 = s_y2 = 0.0;
    for (int j = 0; j < n; j++) {
        s_xy += data[j][feature] * labels[j];
        s_x += data[j][feature];
        s_y += labels[j];
        s_x2 += data[j][feature] * data[j][feature];
        s_y2 += labels[j] * labels[j];
    }

    r = (n*s_xy - s_x*s_y)/(sqrt(n*s_x2 - s_x*s_x)*sqrt(n*s_y2 - s_y*s_y));

    return r;
}

scalar FeatureFeatureCorrelation(scalar **data, int x, int y, int n)
{
    scalar s_xy, s_x, s_y, s_x2, s_y2, r;

    s_xy = s_x = s_y = s_x2 = s_y2 = 0.0;
    for (int j = 0; j < n; j++) {
        s_xy += data[j][x] * data[j][y];
        s_x += data[j][x];
        s_y += data[j][y];
        s_x2 += data[j][x] * data[j][x];
        s_y2 += data[j][y] * data[j][y];
    }

    r = (n*s_xy - s_x*s_y)/(sqrt(n*s_x2 - s_x*s_x)*sqrt(n*s_y2 - s_y*s_y));

    return r;
}
