/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "fs.h"
#include "macros.h"
#include "filter.h"
#include "data.h"

/*
  Q. Song, J. Ni and G. Wang, "A fast clustering-based feature subset selection algorithm for high-dimensional data", IEEE Transactions on Knowledge and Data Engineering 25(1): 1-13, 2013.
 */

void FastclusteringBasedFeatureSubsetSelection(const t_Parameters *parm, scalar **data, int inputs, int instances, int classes, int *labels, int *S, scalar theta)
{
    int CompareDoubleDescending(const void *, const void *);

    scalar **limits = NULL;
    int *intervals = NULL;
    int n_relevant;

    // Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "FastCorrelationBasedFilter: limits");
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "FastCorrelationBasedFilter: intervals");

    // Discretization of dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
            break;

        case dNONE:
            //Do nothing
            break;
    }

    // Part 1: Irrelevant feature removal
    scalar *SUic, *Hi, Hc;
    MALLOC(SUic, scalar, inputs, "FastclusteringBasedFeatureSubsetSelection: SUic");
    MALLOC(Hi, scalar, inputs, "FastclusteringBasedFeatureSubsetSelection: Hi");
    Hc = ClassEntropy(instances, labels, classes);
    if (parm->verbose > 1)
        fprintf(parm->out_file, "Class entropy: %lf\n", Hc);
    for (int i = 0; i < inputs; i++) {
        Hi[i] = Entropy(data, i, instances, limits, intervals[i]);
        scalar Hic = ConditionalClassEntropy(data, i, instances, limits, intervals[i], classes, labels);

        SUic[i] = (2.0*(Hi[i]-Hic))/(Hi[i]+Hc);
    }

    // If theta = -1 adjust to the SU of sqrt(N)/log(N)-th feature
    if (theta < 0) {
        scalar *to_sort;
        MALLOC(to_sort, scalar, inputs, "FastclusteringBasedFeatureSubsetSelection: to_sort");

        for (int i = 0; i < inputs; i++)
            to_sort[i] = SUic[i];
        qsort(to_sort, inputs, sizeof(scalar), CompareDoubleDescending);

        int index = floor(sqrt(inputs)*log10(inputs));
        theta = to_sort[index];

        FREE(to_sort);
    }

    int *list;
    MALLOC(list, int, inputs, "FastclusteringBasedFeatureSubsetSelection: list");
    n_relevant = 0;
    for (int i = 0; i < inputs; i++)
        if (SUic[i] > theta) {
            list[n_relevant] = i;
            n_relevant++;
        }

    if (parm->verbose > 2) {
        fprintf(parm->out_file, "Relevant features (%d):", n_relevant);
        for (int i = 0; i < n_relevant; i++)
            fprintf(parm->out_file, " %d", list[i]+1);
        fprintf(parm->out_file, "\n");
        fflush(parm->out_file);
    }

    // If no relevant end
    if (n_relevant == 0) {
        // Start with no inputs selected
        for (int i = 0; i < inputs; i++)
            S[i] = 0;

        FREE(list);

        return;
    }

    // Part 2: Minimum spanning tree construction
    // Initialize G = NULL
    scalar **G;
    MALLOC(G, scalar *, n_relevant, "FastclusteringBasedFeatureSubsetSelection: G");
    G[0] = NULL;
    for (int i = 1; i < n_relevant; i++) {
        MALLOC(G[i], scalar, i, "FastclusteringBasedFeatureSubsetSelection: G[i]");
        for (int j = 0; j < i; j++) {
            scalar Hij = ConditionalEntropy(data, list[i], list[j], instances, limits, intervals[list[i]], intervals[list[j]]);

            G[i][j] = (2.0*(Hi[list[i]]-Hij))/(Hi[list[i]]+Hi[list[j]]);
        }
    }

    // Generate minimum spanning tree with Prim algorithm
    int n_edges;    // Number of vertices
    t_Edge *E;      // List of edges
    Prim(G, &E, n_relevant, &n_edges);

    if (parm->verbose > 2) {
        fprintf(parm->out_file, "Minimum spanning tree generated with %d edges\n", n_edges);
        fflush(parm->out_file);
    }

    // Part 3: Tree partition and representative feature selection
    // Remove edges
    int n_trees = 1;
    for (int i = 0; i < n_edges; i++) {
        // printf("%lf %lf %lf\n", E[i].value, SUic[list[E[i].origin]], SUic[list[E[i].destination]]);
        if (E[i].value < SUic[list[E[i].origin]] && E[i].value < SUic[list[E[i].destination]]) {
            E[i] = E[n_edges-1];
            n_edges--;
            i--;
            n_trees++;
        }
    }

    // Free G
    for (int i = 1; i < n_relevant; i++)
        FREE(G[i]);
    FREE(G);
    FREE(Hi);

    if (parm->verbose > 2) {
        fprintf(parm->out_file, "Edges removed to obtain %d trees (some of them are isolated nodes)\n", n_trees);
        fflush(parm->out_file);
    }

    // Set inputs corresponding to isolated selected vertices
    for (int i = 0; i < inputs; i++)
        S[i] = 0;
    for (int i = 0; i < n_relevant; i++)
        S[list[i]] = 1;
    for (int i = 0; i < n_edges; i++)
        S[list[E[i].origin]] = S[list[E[i].destination]] = 0;

    if (parm->verbose > 2) {
        fprintf(parm->out_file, "Set inputs for isolated vertices\n");
        fflush(parm->out_file);
    }

    // Now we have a forest, as the MST is broken and must get the trees
    int actual_trees = 0;
    int **trees;
    MALLOC(trees, int *, n_trees+1, "FastclusteringBasedFeatureSubsetSelection: trees");
    for (int i = 0; i < n_trees+1; i++) {
        MALLOC(trees[i], int, n_relevant, "FastclusteringBasedFeatureSubsetSelection: trees[i]");
        for (int j = 0; j < n_relevant; j++)
            trees[i][j] = 0;
    }

    for (int i = 0; i < n_edges; i++) {
        int found = -1;

        if (parm->verbose > 2) {
            fprintf(parm->out_file, "Actual trees: %d/%d  %d %d\n", actual_trees, n_trees, E[i].origin, E[i].destination);
            fflush(parm->out_file);
        }

        // Search in current trees
        for (int j = 0; j < actual_trees; j++)
            if (trees[j][E[i].origin] || trees[j][E[i].destination]) {
                found = j;
                break;
            }

        // Start new tree or add to existing
        if (found == -1) {
            if (actual_trees == n_trees) {
                REALLOC(trees, int *, actual_trees+1, "FastclusteringBasedFeatureSubsetSelection: trees");
                MALLOC(trees[actual_trees], int, n_relevant, "FastclusteringBasedFeatureSubsetSelection: trees[actual_trees]");
            }
            trees[actual_trees][E[i].origin] = trees[actual_trees][E[i].destination] = 1;
            actual_trees++;
        } else
            trees[found][E[i].origin] = trees[found][E[i].destination] = 1;
    }

    if (parm->verbose > 1)
        fprintf(parm->out_file, "Removed after MST: %d\n", n_edges-actual_trees);

    // Select an input from each tree
    for (int i = 0; i < actual_trees; i++) {
        scalar best = 0.0;
        int best_index = -1;
        for (int j = 0; j < n_relevant; j++)
            if (trees[i][j] && best < SUic[list[j]]) {
                best = SUic[list[j]];
                best_index = list[j];
            }

        // Set the input
        S[best_index] = 1;
    }

    FREE(list);
    for (int i = 0; i < n_trees; i++)
        FREE(trees[i]);
    FREE(trees);
    FREE(SUic);
    for (int k = 0; k < inputs; k++)
        FREE(limits[k]);
    FREE(limits);
    FREE(intervals);
}

// Prim's algorithm
void Prim(scalar **G, t_Edge **E, int n_vertices, int *n_edges)
{
    int n_v = 0, *V;

    // Initialize to 0 edges
    *E = NULL;
    *n_edges = 0;

    // Arbitrarily select one vertex
    MALLOC(V, int, n_vertices, "Prim: V");
    for (int i = 0; i < n_vertices; i++)
        V[i] = 0;
    V[(int) RANDOM(0, n_vertices)] = n_v = 1;

    // Repeat until all vertices are included
    while (n_v < n_vertices) {
        // Search for lowest edge from u -> v where u is in V and v is not or viceversa
        scalar min_value = INFINITE;
        int u, v;

        u = v = -1;

        for (int i = 1; i < n_vertices; i++)
            for (int j = 0; j < i; j++)
                if (((V[i] && !V[j]) || (!V[i] && V[j])) && G[i][j] < min_value) {
                    min_value = G[i][j];
                    u = i;
                    v = j;
                }

        // Add vertex and edge
        (*n_edges)++;
        n_v++;
        V[v] = V[u] = 1;
        REALLOC(*E, t_Edge, *n_edges, "Prim: *E");
        (*E)[*n_edges-1].value = min_value;
        (*E)[*n_edges-1].origin = u;
        (*E)[*n_edges-1].destination = v;
    }

    FREE(V);
}

int CompareDoubleDescending(const void *x, const void *y)
{
    scalar diff;

    if ((diff = * (scalar *) x - * (scalar *) y) > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}
