/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <time.h>
#include "macros.h"
#include "filter.h"
#include "fs.h"
#include "data.h"

/*
  L. Yu and H. Liu, "Efficient Feature Selection via Analysis of Relevance and Redundancy", Journal of Machine Learning Research 5 (2004) 1205–1224.
 */

void FastCorrelationBasedFilter(const t_Parameters *parm, scalar **data, int inputs, int instances, int classes, int *labels, int *S, scalar delta)
{
    scalar **limits = NULL;
    int *intervals = NULL;
    int n_relevant = 0;
    time_t init_time;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "FastCorrelationBasedFilter: limits");
    //Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "FastCorrelationBasedFilter: intervals");

    init_time = time(NULL);

    // Discretization of dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
            break;

        case dNONE:
            //Do nothing
            break;
    }

    if (parm->verbose > 2)
        fprintf(parm->out_file, "Discretization time: %ld\n", time(NULL) - init_time);

    // Calculate symmetric uncertainty for every variable
    t_EntropyInfo *entropy;
    MALLOC(entropy, t_EntropyInfo, inputs, "FastCorrelationBasedFilter: entropy");
    for (int i = 0; i < inputs; i++) {
        scalar Hi = Entropy(data, i, instances, limits, intervals[i]);
        scalar Hc = ClassEntropy(instances, labels, classes);
        scalar Hic = ConditionalClassEntropy(data, i, instances, limits, intervals[i], classes, labels);

        entropy[i].index = i;
        entropy[i].Hi = Hi;
        entropy[i].SUic = (Hi-Hic)/(Hi+Hc);
    }

    // Sort features in descending SU value
    qsort(entropy, inputs, sizeof(t_EntropyInfo), CompareDescending);

    // If delta = -1 adjust delta the SU of N/log(N)-th feature
    if (delta < 0)
        delta = floor(entropy[MIN(inputs-1, (int)(inputs/log10(inputs)))].SUic);

    // Check if above threshold
    while (n_relevant < inputs && entropy[n_relevant].SUic > delta)
        n_relevant++;

    init_time = time(NULL);

    // None selected for start
    for (int i = 0; i < inputs; i++)
        S[i] = 0;
    if (n_relevant != 0) {
        // Get first variable
        int Fj = 0;
        do {
            int Fi = Fj + 1;

            // Check if remaining
            while (Fi < n_relevant) {
                scalar Hij = ConditionalEntropy(data, entropy[Fi].index, entropy[Fj].index, instances, limits, intervals[entropy[Fi].index], intervals[entropy[Fj].index]);

                scalar SUij = (entropy[Fi].Hi-Hij)/(entropy[Fi].Hi+entropy[Fj].Hi);
                if (SUij >= entropy[Fi].SUic) {
                    // Remove from relevant list
                    for (int i = Fi; i < n_relevant-1; i++)
                        entropy[i] = entropy[i+1];
                    n_relevant--;

                    // One step back
                    Fi--;
                }

                // Next element
                Fi++;
            }

            // Next element
            Fj++;
        } while (Fj < n_relevant-1 && (!parm->time_limit || time(NULL) - init_time < parm->time_limit));

        // Set selected features
        for (int i = 0; i < n_relevant; i++)
            S[entropy[i].index] = 1;
    }

    FREE(entropy);
    for (int k = 0; k < inputs; k++)
        FREE(limits[k]);
    FREE(limits);
    FREE(intervals);
}

scalar ClassEntropy(int instances, int *labels, int classes)
{
    scalar H = 0.0, P[classes];

    for (int i = 0; i < classes; i++)
        P[i] = 0.0;

    for (int i = 0; i < instances; i++)
        P[labels[i]] += 1.0;

    for (int i = 0; i < classes; i++)
        H -= P[i] ? (P[i]/instances) * log2(P[i]/instances) : 0.0;

    return H;
}

scalar Entropy(scalar **data, int variable, int instances, scalar **limits, int n_intervals)
{
    scalar H = 0.0, *P;

    MALLOC(P, scalar, n_intervals+1, "Entropy: P");
    for (int i = 0; i <= n_intervals; i++)
        P[i] = 0.0;

    // Count the frequency of the intervals
    for (int i = 0; i < instances; i++) {
        int j = 0;
        while (j < n_intervals && data[i][variable] > limits[variable][j])
            j++;

        P[j] += 1.0;
    }

    for (int i = 0; i <= n_intervals; i++)
        if (P[i] > 0.0)
            H -= (P[i]/instances) * log2(P[i]/instances);

    FREE(P);

    return H;
}

scalar ConditionalClassEntropy(scalar **data, int x, int instances, scalar **limits, int n_intervals, int classes, int *labels)
{
    scalar *Px, Pc[classes], H = 0.0;

    MALLOC(Px, scalar, n_intervals+1, "ConditionalClassEntropy: Px");

    // Frequency of every class
    for (int i = 0; i < classes; i++)
        Pc[i] = 0.0;
    for (int i = 0; i < instances; i++)
        Pc[labels[i]] += 1.0;
    for (int i = 0; i < classes; i++)
        Pc[i] /= instances;

    // Conditional entropy
    // H(X|Y) = - sum_j P(y_j) sum_i P(x_i|y_j) log_2 P(X_i| y_j)
    for (int i = 0; i < classes; i++) {
        scalar h = 0.0;
        int n_instances = 0;

        for (int j = 0; j < n_intervals+1; j++)
            Px[j] = 0.0;

        // Count the frequency of the intervals in class i-th
        for (int j = 0; j < instances; j++)
            if (labels[j] == i) {
                int k = 0;
                while (k < n_intervals && data[j][x] > limits[x][k])
                    k++;

                Px[k] += 1.0;

                n_instances++;
            }

        for (int j = 0; j < n_intervals+1; j++)
            if (Px[j] > 0.0)
                h += (Px[j]/n_instances) * log2(Px[j]/n_instances);

        H -= Pc[i] * h;
    }

    FREE(Px);

    return H;
}

// Auxiliary function for qsort()
int CompareDescending(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_EntropyInfo *) x)->SUic - ((t_EntropyInfo *) y)->SUic) > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}

scalar ConditionalEntropy(scalar **data, int x, int y, int instances, scalar **limits, int n_intervals_x, int n_intervals_y)
{
    scalar *Px, *Py, H = 0.0;

    MALLOC(Px, scalar, n_intervals_x+1, "ConditionalEntropy: Px");
    MALLOC(Py, scalar, n_intervals_y+1, "ConditionalEntropy: Py");

    // Count the frequency of y, Py
    for (int i = 0; i < n_intervals_y+1; i++)
        Py[i] = 0.0;
    for (int i = 0; i < instances; i++) {
        int j = 0;
        while (j < n_intervals_y && data[i][y] > limits[y][j])
            j++;

        Py[j] += 1.0;
    }
    for (int i = 0; i < n_intervals_y+1; i++)
        Py[i] /= instances;

    // Conditional entropy
    // H(X|Y) = - sum_j P(y_j) sum_i P(x_i|y_j) log_2 P(X_i| y_j)
    for (int i = 0; i < n_intervals_y+1; i++) {
        scalar h = 0.0;
        int n_instances = 0;

        for (int j = 0; j < n_intervals_x+1; j++)
            Px[j] = 0.0;

        // Count the frequency of the intervals x_j in y_i
        for (int j = 0; j < instances; j++)
            if ((i == 0 && data[j][y] <= limits[y][0]) || (i == n_intervals_y && data[j][y] > limits[y][n_intervals_y-1]) || (data[j][y] > limits[y][i-1] && data[j][y] <= limits[y][i])) {
                int k = 0;
                while (k < n_intervals_x && data[j][x] > limits[x][k])
                    k++;

                Px[k] += 1.0;

                n_instances++;
            }

        for (int j = 0; j < n_intervals_x+1; j++)
            if (Px[j] > 0.0)
                h += (Px[j]/n_instances) * log2(Px[j]/n_instances);

        H -= Py[i] * h;
    }

    FREE(Px);
    FREE(Py);

    return H;
}
