/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "maths.h"

/*
 * P. Pudil, J. Novovicova and J. Kittle, "Floating search methods for feature selection", Pattern Recognition Letters 15, 1119-1125, 1994.
 */

void SequentialForwardFloatingSelection(const t_Parameters *parm, scalar **data, int inputs, int instances, int classes, int *labels, int *S)
{
    scalar **V, **inverseV, **centroids, J, J_best;
    int count[classes], k, attempts;

    // Initially unselect all features
    for (int i = 0; i < inputs; i++)
        S[i] = 0;

    // Covariance matrix
    MALLOC(V, scalar *, inputs, "SequentialForwardFloatingSelection: V");
    MALLOC(inverseV, scalar *, inputs, "SequentialForwardFloatingSelection: inverseV");
    for (int i = 0; i < inputs; i++) {
        MALLOC(V[i], scalar, inputs, "SequentialForwardFloatingSelection: V[i]");
        MALLOC(inverseV[i], scalar, inputs, "SequentialForwardFloatingSelection: inverseV[i]");
    }

    CovarianceMatrix(data, instances, inputs, V);
    // Tikhonov correction to avoid near singular matrix
    for (int i = 0; i < inputs; i++)
        V[i][i] += 0.0001;
    InvertMatrix(V, inverseV, inputs);

    for (int i = 0; i < inputs; i++)
        FREE(V[i]);
    FREE(V);

    // Centroids of classes
    MALLOC(centroids, scalar *, classes, "SequentialForwardFloatingSelection: centroids");
    for (int i = 0; i < classes; i++) {
        count[i] = 0;
        MALLOC(centroids[i], scalar, inputs, "SequentialForwardFloatingSelection: centroids[i]");
        for (int j = 0; j < inputs; j++)
            centroids[i][j] = 0.0;
    }

    for (int i = 0; i < instances; i++) {
        count[labels[i]]++;
        for (int j = 0; j < inputs; j++)
            centroids[labels[i]][j] += data[i][j];
    }

    for (int i = 0; i < classes; i++)
        if (count[i])
            for (int j = 0; j < inputs; j++)
                centroids[i][j] /= count[i];

    k = 0;
    J = 0.0;
    attempts = 0;
    do {
        J_best = J;

        if (parm->verbose > 2) {
            fprintf(parm->out_file, "Best J so far: %lf\n", J_best);
            fflush(parm->out_file);
        }

        // STEP 1: Inclusion
        // Add best input not yet selected
        scalar best = 0.0;
        int index_best = -1;
        for (int i = 0; i < inputs; i++)
            if (!S[i]) {
                // Evaluate S using Mahalanobis Distance
                S[i] = 1;

                scalar evaluation = EvaluateUsingMahalanobisDistance(inputs, classes, S, inverseV, centroids);
                if (evaluation > best) {
                    best = evaluation;
                    index_best = i;
                }

                S[i] = 0;
            }

        S[index_best] = 1;
        k++;
        J = best;

        if (parm->verbose > 2) {
            fprintf(parm->out_file, "Added feature %d\n", index_best);
            fflush(parm->out_file);
        }

        // STEP 2: Conditional exclusion
        // Remove worst input provided is not the last one added
        scalar worst = 0;
        int index_worst = -1;
        if (k > 1) {
            for (int i = 0; i < inputs; i++)
                if (S[i]) {
                    // Evaluate S using Mahalanobis Distance
                    S[i] = 0;

                    scalar evaluation = EvaluateUsingMahalanobisDistance(inputs, classes, S, inverseV, centroids);
                    if (evaluation > worst) {
                        worst = evaluation;
                        index_worst = i;
                    }

                    S[i] = 1;
                }

            // If the worst feature is not the lass added remove it
            if (index_worst != index_best) {
                S[index_worst] = 0;
                k--;
                J = worst;

                if (parm->verbose > 2) {
                    fprintf(parm->out_file, "Removed feature %d\n", index_worst);
                    fflush(parm->out_file);
                }
            } else {
                // Number of attempts that allow worst J
                if (J_best > J)
                    attempts++;
                continue;
            }
        }

        // STEP 3: Continuation of conditional exclusion
        while (k > 2) {
            // Find lest significant
            worst = 0.0;
            index_worst = -1;
            for (int i = 0; i < inputs; i++)
                if (S[i]) {
                    // Evaluate S using Mahalanobis Distance
                    S[i] = 0;

                    scalar evaluation = EvaluateUsingMahalanobisDistance(inputs, classes, S, inverseV, centroids);
                    if (evaluation > worst) {
                        worst = evaluation;
                        index_worst = i;
                    }

                    S[i] = 1;
                }

            // Better than before
            if (worst > J) {
                S[index_worst] = 0;
                J = worst;
                k--;

                if (parm->verbose > 2) {
                    fprintf(parm->out_file, "Removed feature %d\n", index_worst);
                    fflush(parm->out_file);
                }
            } else
                break;
        }

        // Number of attempts that allow worst J
        if (J_best > J)
            attempts++;
    } while ((J_best < J || attempts < parm->floating_attempts) && k < inputs);

    for (int i = 0; i < inputs; i++)
        FREE(inverseV[i]);
    for (int i = 0; i < classes; i++)
        FREE(centroids[i]);
    FREE(inverseV);
    FREE(centroids);
}

void SequentialBackwardFloatingSelection(const t_Parameters *parm, scalar **data, int inputs, int instances, int classes, int *labels, int *S)
{
    scalar **V, **inverseV, **centroids, J, J_best;
    int count[classes], k, attempts;

    // Initially select all features
    for (int i = 0; i < inputs; i++)
        S[i] = 1;

    // Covariance matrix
    MALLOC(V, scalar *, inputs, "SequentialBackwardFloatingSelection: V");
    MALLOC(inverseV, scalar *, inputs, "SequentialBackwardFloatingSelection: inverseV");
    for (int i = 0; i < inputs; i++) {
        MALLOC(V[i], scalar, inputs, "SequentialBackwardFloatingSelection: V[i]");
        MALLOC(inverseV[i], scalar, inputs, "SequentialBackwardFloatingSelection: inverseV[i]");
    }

    CovarianceMatrix(data, instances, inputs, V);
    // Tikhonov correction to avoid near singular matrix
    for (int i = 0; i < inputs; i++)
        V[i][i] += 0.0001;
    InvertMatrix(V, inverseV, inputs);

    for (int i = 0; i < inputs; i++)
        FREE(V[i]);
    FREE(V);

    // Centroids of classes
    MALLOC(centroids, scalar *, classes, "SequentialBackwardFloatingSelection: centroids");
    for (int i = 0; i < classes; i++) {
        count[i] = 0;
        MALLOC(centroids[i], scalar, inputs, "SequentialBackwardFloatingSelection: centroids[i]");
        for (int j = 0; j < inputs; j++)
            centroids[i][j] = 0.0;
    }

    for (int i = 0; i < instances; i++) {
        count[labels[i]]++;
        for (int j = 0; j < inputs; j++)
            centroids[labels[i]][j] += data[i][j];
    }

    for (int i = 0; i < classes; i++)
        if (count[i])
            for (int j = 0; j < inputs; j++)
                centroids[i][j] /= count[i];

    k = inputs;
    J = 0.0;
    attempts = 0;
    do {
        J_best = J;

        if (parm->verbose > 2) {
            fprintf(parm->out_file, "Best J so far: %lf\n", J_best);
            fflush(parm->out_file);
        }

        // STEP 1: Exclusion
        // Remove worst input selected
        scalar worst = 0.0;
        int index_worst = -1;
        for (int i = 0; i < inputs; i++)
            if (S[i]) {
                // Evaluate S using Mahalanobis Distance
                S[i] = 0;

                scalar evaluation = EvaluateUsingMahalanobisDistance(inputs, classes, S, inverseV, centroids);
                if (evaluation > worst) {
                    worst = evaluation;
                    index_worst = i;
                }

                S[i] = 1;
            }

        S[index_worst] = 0;
        k--;
        J = worst;

        if (parm->verbose > 2) {
            fprintf(parm->out_file, "Removed feature %d\n", index_worst);
            fflush(parm->out_file);
        }

        // STEP 2: Conditional inclusion
        // Add best input provided is not the last one removed
        scalar best = 0;
        int index_best = -1;
        for (int i = 0; i < inputs; i++)
            if (!S[i]) {
                // Evaluate S using Mahalanobis Distance
                S[i] = 1;

                scalar evaluation = EvaluateUsingMahalanobisDistance(inputs, classes, S, inverseV, centroids);
                if (evaluation > best) {
                    best = evaluation;
                    index_best = i;
                }

                S[i] = 0;
            }

        // If the worst feature is not the lass added remove it
        if (index_worst != index_best) {
            S[index_best] = 1;
            k++;
            J = best;

            if (parm->verbose > 2) {
                fprintf(parm->out_file, "Added feature %d\n", index_best);
                fflush(parm->out_file);
            }
        } else {
            // Number of attempts that allow worst J
            if (J_best > J)
                attempts++;
            continue;
        }

        // STEP 3: Continuation of conditional exclusion
        while (k < inputs) {
            // Find most significant
            best = 0.0;
            index_best = -1;
            for (int i = 0; i < inputs; i++)
                if (!S[i]) {
                    // Evaluate S using Mahalanobis Distance
                    S[i] = 1;

                    scalar evaluation = EvaluateUsingMahalanobisDistance(inputs, classes, S, inverseV, centroids);
                    if (evaluation > best) {
                        best = evaluation;
                        index_best = i;
                    }

                    S[i] = 0;
                }

            // Better than before
            if (best > J) {
                S[index_best] = 1;
                J = best;
                k++;

                if (parm->verbose > 2) {
                    fprintf(parm->out_file, "Added feature %d\n", index_best);
                    fflush(parm->out_file);
                }
            } else
                break;
        }

        // Number of attempts that allow worst J
        if (J_best > J)
            attempts++;
    } while ((J_best < J || attempts < parm->floating_attempts) && k < inputs);

    for (int i = 0; i < inputs; i++)
        FREE(inverseV[i]);
    for (int i = 0; i < classes; i++)
        FREE(centroids[i]);
    FREE(inverseV);
    FREE(centroids);
}

scalar EvaluateUsingMahalanobisDistance(int inputs, int classes, int *S, scalar **inverseV, scalar **centroids)
{
    scalar average_distance = 0.0;

    // Average Mahalanobis distance among all centroids
    for (int i = 1; i < classes; i++)
        for (int j = 0; j < i; j++)
            average_distance += MaskedMahalanobisDistance(centroids[i], centroids[j], inverseV, S, inputs);

    return average_distance/(classes*(classes-1));
}

void CovarianceMatrix(scalar **data, int instances, int features, scalar **V)
{
    scalar s_xy, *s_x;

    MALLOC(s_x, scalar, features, "CovarianceMatrix: s_x");

    for (int i = 0; i < features; i++) {
        s_x[i] = 0;
        for (int j = 0; j < instances; j++)
            s_x[i] += data[j][i];
        s_x[i] /= instances;
    }

    for (int i = 0; i < features; i++)
        for (int j = 0; j <= i; j++) {
            s_xy = 0.0;
            for (int k = 0; k < instances; k++)
                s_xy += data[k][i] * data[k][j];

            V[i][j] = V[j][i] = s_xy/instances - s_x[i]*s_x[j];
        }

    FREE(s_x);
}

scalar MaskedMahalanobisDistance(scalar *x, scalar *y, scalar **inverseV, int *S, int n)
{
    scalar x_y[n], product[n];

    for (int i = 0; i < n; i++)
        x_y[i] = x[i] - y[i];

    MaskedMatrixVectorProduct(inverseV, x_y, product, n, S);

    return MaskedDotProduct(x_y, product, n, S);
}
