/***************************************************************************
 *   Copyright (C) 2009 by Aida de Haro-García                             *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"

void FocusAlg(const t_Parameters *parm, int *S, int n_feat, scalar **data, int n, int *labels, int outputs, time_t time_limit)
{
    int i, j, size = 0, n_sel = 0, cont = 0, num = 0, *sel_feat = NULL, *intervals, *comb_ones = NULL, **comb = NULL;
    scalar best_inconsistency = 0.0, cur_incons = 0.0, **limits;

    // Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, n_feat, "FocusAlg: limits");

    // Set limits to NULL
    for (int k = 0; k < n_feat; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, n_feat, "FocusAlg: intervals");
    MALLOC(sel_feat, int, n_feat, "FocusAlg");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, n, outputs, labels, n_feat, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, n, n_feat, limits, intervals, NULL);
            break;

        case dNONE:
            //Do nothing
            break;
    }

    best_inconsistency = 1.0;

    // Number of features allows to use a cache (there is enough memory)
    if (n_feat <= MAX_FEAT_MEM)  {
        //Memory allocation
        MALLOC(comb_ones, int, pow(2, n_feat), "FocusAlg");
        MALLOC(comb, int *, pow(2, n_feat), "FocusAlg");
        for (i = 1; i < pow(2, n_feat); i++)
            MALLOC(comb[i], int, n_feat, "FocusAlg");
        for (i = 1; i < pow(2, n_feat); i++)  {
            n_sel = 0;
            cont = n_feat - 1;
            num = i;

            // Binarize all numbers from 1 to pow(2, n_feat) in matrix 'comb'
            while (num)  {
                if (num == 1)  {
                    comb[i][cont] = 1;
                    n_sel = n_sel + 1;
                    break;
                } else  {
                    comb[i][cont] = num%2;
                    n_sel = n_sel + comb[i][cont];
                    num /= 2;
                    cont--;
                }
            }

            // Count the number of ones in the rows of matrix 'comb' and store that information in vector 'comb_ones'
            comb_ones[i] = n_sel;
        }

        // In each iteration of 'size' only get the inconsistency of the rows of 'comb'
        // that have the same number of ones (elements of 'comb_ones') than 'size'
        for (size = 1; size <= n_feat; size++) {
            for (i = 1; i < pow(2, n_feat) - 1; i++)  {
                if (comb_ones[i] == size)  {
                    cur_incons = InconsistCount(n_feat, data, n, comb[i], labels, outputs, limits, intervals);

                    // If inconsistency = 0 return
                    if (cur_incons <= ZERO) {
                        for (j = 0; j < n_feat; j++)
                            S[j] = comb[i][j];

                        goto end;
                    } else {
                        // Store new best S: less or equal inconsistency value than all features selected
                        if (cur_incons <= best_inconsistency)  {
                            for (j = 0; j < n_feat; j++)
                                S[j] = comb[i][j];

                            best_inconsistency = cur_incons;
                        }
                    }
                }
            }
        }
    } else {
        // Get time for testing time limit
        time_t init_time = time(NULL);

        // Number of features doesn't allow the use of a cache (examine exhaustively number of ones)
        for (size = 1; size <= n_feat; size++) {
            for (i = 1; i < pow(2, n_feat); i++)  {
                n_sel = 0;
                cont = n_feat - 1;
                num = i;

                // Reset sel_feat vector
                for (j = 0; j < n_feat; j++)
                    sel_feat[j] = 0;

                // Binarize all numbers from 1 to pow(2, n_feat)
                while (num)  {
                    if (num == 1)  {
                        sel_feat[cont] = 1;
                        n_sel = n_sel + 1;
                        break;
                    } else  {
                        sel_feat[cont] = num%2;
                        n_sel = n_sel + sel_feat[cont];
                        num /= 2;
                        cont--;
                    }
                }

                // In each iteration of 'size' only get the inconsistency of the binary combinations that have the same number of ones ('n_sel') than 'size'
                if (n_sel == size)  {
                    cur_incons = InconsistCount(n_feat, data, n, sel_feat, labels, outputs, limits, intervals);

                    // If inconsistency = 0 return
                    if (cur_incons <= ZERO)  {
                        for (j = 0; j < n_feat; j++)
                            S[j] = sel_feat[j];

                        goto end;
                    } else {
                        // Store new best S: less or equal inconsistency value than all features selected
                        if (cur_incons <= best_inconsistency)  {
                            for (j = 0; j < n_feat; j++)
                                S[j] = sel_feat[j];

                            best_inconsistency = cur_incons;
                        }
                    }
                }
            }

            // Check time
            if (time_limit && (time(NULL) - init_time > time_limit))
                // If here return best current
                goto end;
        }
    }

end:

    if (comb_ones) {
        for (int i = 1; i < pow(2, n_feat); i++)
            FREE(comb[i]);
        FREE(comb);
        FREE(comb_ones);
    }
    for (int k = 0; k < n_feat; k++)
        FREE(limits[k]);
    FREE(limits);
    FREE(intervals);
    FREE(sel_feat);
    return;
}

// Variant of FOCUS using sequential forward selection

// Almuallim H. and Dietterich T.G., Learning boolean concepts in the presence of many irrelevant features, Artificial Intelligence, 69(1-2), pp 279-305, 1994.


void FocusSF(const t_Parameters *parm, int *S, int n_feat, scalar **data, int n, int *labels, int outputs)
{
    int *sel_feat = NULL;
    static int *intervals = NULL;
    scalar best_inconsistency = 0.0;
    static scalar **limits = NULL;

    time_t init_time = time(NULL);

    MALLOC(sel_feat, int, n_feat, "FocusAlg");

    if (limits == NULL) {
        // Memory allocation for auxiliary data structure
        MALLOC(limits, scalar *, n_feat, "FocusAlg: limits");

        // Set limits to NULL
        for (int k = 0; k < n_feat; k++)
            limits[k] = NULL;
        MALLOC(intervals, int, n_feat, "FocusAlg: intervals");

        // Discretize dataset
        switch (parm->discretizer) {
            case dFAYYAD:
                FayyadAndIraniMDLMethod(data, n, outputs, labels, n_feat, limits, intervals, NULL, NULL);
                break;

            case dWPKID:
                WPKIDMethod(data, n, n_feat, limits, intervals, NULL);
                break;

            case dNONE:
                //Do nothing
                break;
        }

        if (parm->verbose > 2)
            fprintf(parm->out_file, "Discretization time: %ld\n", time(NULL) - init_time);
    }

    // Unselect all features
    for (int i = 0; i < n_feat; i++)
        sel_feat[i] = 0;

    best_inconsistency = 1.1;

    init_time = time(NULL);

    bool improved;
    do {
        scalar next_best_inconsistency = 1.1;
        int next_best = -1;

        improved = false;

        // Add unselected one at a time and get best one
        for (int i = 0; i < n_feat; i++)
            if (!sel_feat[i]) {
                sel_feat[i] = 1;

                // Get inconsistency
                scalar inconsistency = InconsistCount(n_feat, data, n, sel_feat, labels, outputs, limits, intervals);

                // If inconsistency = 0 return, we have an optimal subset of features
                if (inconsistency <= ZERO)  {
                    for (int j = 0; j < n_feat; j++)
                        S[j] = sel_feat[j];

                    goto end;
                } else if (inconsistency < next_best_inconsistency) {
                    next_best_inconsistency = inconsistency;
                    next_best = i;
                }

                sel_feat[i] = 0;

                // If time limit exceeded finish
                if (parm->time_limit && (time(NULL) - init_time > parm->time_limit))
                    break;
            }

        // Test if improve over current best
        if (next_best_inconsistency < best_inconsistency) {
            best_inconsistency = next_best_inconsistency;
            sel_feat[next_best] = 1;
            improved = true;
        }

        // If time limit exceeded finish
        if (parm->time_limit && (time(NULL) - init_time > parm->time_limit))
            improved = false;
    } while (improved);

    for (int j = 0; j < n_feat; j++)
        S[j] = sel_feat[j];

end:
    FREE(sel_feat);
}
