
/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "filter.h"
#include "fs.h"

/*
 * R. Gilad-Bachrach, A. Navot and N. Tishby, "Margin Based Feature Selection - Theory and algorithms", Proc. of the 21st Int. Con. on Mach. Learn., Canada, 2004
 */

void GreedyFeatureFlip(const t_Parameters *parm, scalar **data, int inputs, int instances, int *labels, int *S, int max_iterations)
{
    int *list;
    scalar *w;

    for (int i = 0; i < inputs; i++)
        S[i] = 0;

    MALLOC(w, scalar, inputs, "GreedyFeatureFlip: w");
    MALLOC(list, int, inputs, "GreedyFeatureFlip: list");
    for (int i = 0; i < inputs; i++) {
        list[i] = i;
        w[i] = 1.0;
    }

    for (int i = 0; i < max_iterations; i++) {
        SHOWPROGRESS(parm->show_progress > 0, "G-flip iteration %*d/%d\r", WIDTH(max_iterations), i+1, max_iterations);

        // Random permutation
        RandomizeVector(list, inputs);

        int change = false;
        for (int j = 0; j < inputs; j++) {
            int backup = S[list[j]];

            // Error with list[j]
            S[list[j]] = 1;
            scalar e_1 = MarginBasedSubsetEvaluation(S, data, instances, inputs, labels, w);

            // Error without list[j]
            S[list[j]] = 0;
            scalar e_2 = MarginBasedSubsetEvaluation(S, data, instances, inputs, labels, w);

            if (e_1 > e_2) {
                // Add
                S[list[j]] = 1;
                if (!backup)
                    change = true;
            } else if (e_2 > e_1) {
                // Remove
                S[list[j]] = 0;
                if (backup)
                    change = true;
            } else
                S[list[j]] = backup;
        }

        if (!change)
            break;
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FREE(w);
    FREE(list);
}

scalar MarginBasedSubsetEvaluation(int *s, scalar **data, int instances, int inputs, int *labels, scalar *w)
{
    scalar evaluation = 0.0, *instance;
    int label;

    for (int i = 0; i < instances; i++) {
        instance = data[i];
        label = labels[i];

        // Remove instance temporarily
        scalar *sw_d = data[i];
        data[i] = data[instances-1];
        data[instances-1] = sw_d;

        int sw_l = labels[i];
        labels[i] = labels[instances-1];
        labels[instances-1] = sw_l;

        evaluation += InstanceMargin(instance, label, data, inputs, s, instances-1, labels, w);

        sw_d = data[i];
        data[i] = data[instances-1];
        data[instances-1] = sw_d;

        sw_l = labels[i];
        labels[i] = labels[instances-1];
        labels[instances-1] = sw_l;
    }

    return evaluation;
}

scalar InstanceMargin(scalar *instance, int instance_label, scalar **data, int inputs, int *subspace, int instances, int *labels, scalar *w)
{
    scalar dist_miss = 10.E20, dist_hit = 10.E20, distance;
    int miss = -1, hit = -1;

    // Nearest "friend" and nearest enemy
    for (int i = 0; i < instances; i++) {
        distance = EuclideanDistanceSubspace(instance, data[i], inputs, subspace);

        // Hit?
        if (distance < dist_hit && labels[i] == instance_label) {
            dist_hit = distance;
            hit = i;
        } else if (distance < dist_miss && labels[i] != instance_label) {
            dist_miss = distance;
            miss = i;
        }
    }

    // Missing hit or miss?
    if (miss == -1 || hit == -1)
        return 0.0;

    dist_miss = dist_hit = 0.0;
    for (int i = 0; i < inputs; i++) {
        dist_miss += subspace[i]*SQR(w[i])*SQR(instance[i]-data[miss][i]);
        dist_hit += subspace[i]*SQR(w[i])*SQR(instance[i]-data[hit][i]);
    }

    scalar margin = 0.5*(sqrt(dist_miss) - sqrt(dist_hit));

    return margin;
}
