/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "ranking.h"
#include "filter.h"
#include "fs.h"
#include "maths.h"

/*
 * M. Yang, F. Wang and P. Yang (2008) "A novel feature selection algorithm based on hypothesis-margin", Journal of Computers 3(2):27-35.
 *
 * It can be both a filter subset selection with parameter sigma or a ranking algorithm
 */

void ImprovedIterativeSearchMarginBasedAlgorithm(const t_Parameters *parm, scalar **data, int inputs, int instances, int *labels, t_RankInfo *ranking, int *subspace, int classes)
{
    scalar *w, *delta;
    int T, hit, miss;

    MALLOC(w, scalar, inputs, "ImprovedIterativeSearchMarginBasedAlgorithm: w");
    MALLOC(delta, scalar, inputs, "ImprovedIterativeSearchMarginBasedAlgorithm: delta");
    for (int i = 0; i < inputs; i++)
        w[i] = 1.0;

    // Get centermiss() and centerhit() for every class
    scalar **centermiss, **centerhit;
    AllocateScalarMatrix(&centermiss, classes, inputs);
    AllocateScalarMatrix(&centerhit, classes, inputs);

    CenterMissHit(data, instances, inputs, classes, labels, centerhit, centermiss);

    // If no iterations use number of instances
    T = parm->isimba_T ? parm->isimba_T : instances;

    for (int i = 0; i < T; i++) {
        SHOWPROGRESS(parm->show_progress > 0, "iSimba iteration %*d/%d\r", WIDTH(T), i+1, T);

        int r = RANDOM(0, instances);

        NearMissHitW(r, data, inputs, instances, labels, &hit, &miss, w);

        // Not found?
        if (miss == -1 || hit == -1)
            continue;

        // Miss and hit
        scalar dist_miss = 0.0, dist_hit = 0.0;
        for (int j = 0; j < inputs; j++) {
            dist_miss += SQR(w[j])*SQR(data[r][j]-data[miss][j]);
            dist_hit += SQR(w[j])*SQR(data[r][j]-data[hit][j]);
        }
        dist_miss = sqrt(dist_miss);
        dist_hit = sqrt(dist_hit);

        if (dist_hit <= ZERO || dist_miss <= ZERO)
            continue;

        // Center miss and hit
        scalar distcenter_miss = 0.0, distcenter_hit = 0.0;
        for (int j = 0; j < inputs; j++) {
            distcenter_miss += SQR(w[j])*SQR(data[r][j]-centermiss[labels[r]][j]);
            distcenter_hit += SQR(w[j])*SQR(data[r][j]-centerhit[labels[r]][j]);
        }
        distcenter_miss = sqrt(distcenter_miss);
        distcenter_hit = sqrt(distcenter_hit);

        if (distcenter_hit <= ZERO || distcenter_miss <= ZERO)
            continue;

        for (int j = 0; j < inputs; j++) {
            delta[j] = (0.5*(SQR(data[r][j]-data[miss][j])/dist_miss - SQR(data[r][j]-data[hit][j])/dist_hit)) +
                       (parm->isimba_lambda*(SQR(data[r][j]-centermiss[labels[r]][j])/distcenter_miss - SQR(data[r][j]-centerhit[labels[r]][j])/distcenter_hit)*w[j]);

            w[j] += delta[j];
        }
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // Normalize
    scalar max = MaximumOfVector(w, inputs);
    if (fabs(max) > ZERO) {
        for (int i = 0; i < inputs; i++)
            w[i] = SQR(w[i]/max);
    }

    // Ranking or subset or both
    if (ranking)
        for (int i = 0; i < inputs; i++) {
            ranking[i].feature = i;
            ranking[i].score = w[i];
        }

    if (subspace)
        for (int i = 0; i < inputs; i++)
            subspace[i] = (w[i] >= parm->isimba_delta);

    FREE(w);
    FREE(delta);
    FreeScalarMatrix(&centerhit, classes);
    FreeScalarMatrix(&centermiss, classes);
}

void CenterMissHit(scalar **data, int instances, int inputs, int classes, int *labels, scalar **centerhit, scalar **centermiss)
{
    int count[classes];

    // Center of every class
    ZeroMatrix(centerhit, NULL, classes, inputs);
    ZeroVector(NULL, count, classes);

    for (int i = 0; i < instances; i++) {
        count[labels[i]]++;
        for (int j = 0; j < inputs; j++)
            centerhit[labels[i]][j] += data[i][j];
    }

    for (int i = 0; i < classes; i++)
        if (count[i])
            for (int j = 0; j < inputs; j++)
                centerhit[i][j] /= count[i];

    // Center of every enemy of class
    ZeroMatrix(centermiss, NULL, classes, inputs);
    for (int i = 0; i < classes; i++)
        for (int j = 0; j < classes; j++)
            if (i != j)
                for (int k = 0; k < inputs; k++)
                    centermiss[i][k] += centerhit[j][k]*count[j];

    for (int i = 0; i < classes; i++)
        for (int j = 0; j < inputs; j++)
            centermiss[i][j] /= (instances - count[i]);
}
