/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "filter.h"

/*
 * Biesiada J., Duch W. (2005) Feature Selection for High-Dimensional Data: A Kolmogorov-Smirnov Correlation-Based Filter. In: Kurzyński M., Puchała E., Woźniak M., żołnierek A. (eds)
 * Computer Recognition Systems. Advances in Soft Computing, vol 30. Springer, Berlin, Heidelberg
 */

void  KolmogorovSmirnovCorrelationBasedFilter(const t_Parameters *parm, scalar **data, int inputs, int instances, int classes, int *labels, int *S)
{
    scalar *limits = NULL;
    t_RankInfo *ranking;

    // Discretization of dataset into equal bins
    if (!parm->typify)
        OnErrorExit(ERR_PARAM, "KSCBF needs typified features");

    // Memory allocation for auxiliary data structure
    MALLOC(limits, scalar, parm->kscbf_k+1, "KolmogorovSmirnovCorrelationBasedFilter: limits");

    scalar max = data[0][0], min = data[0][0];
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < inputs; j++)
            if (data[i][j] > max)
                max = data[i][j];
            else if (data[i][j] < min)
                min = data[i][j];

    limits[0] = min - 0.1;
    limits[parm->kscbf_k] = max + 0.1;
    scalar increment = (max-min)/parm->kscbf_k;
    for (int i = 1; i < parm->kscbf_k; i++)
        limits[i] = min + increment*i;

    // Obtain entropy (Hc) for class
    scalar *p;
    MALLOC(p, scalar, classes, "KolmogorovSmirnovCorrelationBasedFilter: p");
    ZeroVector(p, NULL, classes);
    for (int i = 0; i < instances; i++)
        p[labels[i]]++;

    scalar Hc = 0.0;
    for (int i = 0; i < classes; i++) {
        p[i] /= instances;
        if (p[i] > ZERO)
            Hc -= p[i]*log2(p[i]);
    }
    FREE(p);

    // Step 1: Relevance analysis
    // Obtain MI(X, C) for every feature
    scalar **px, **pxy;
    MALLOC(ranking, t_RankInfo, inputs, "KolmogorovSmirnovCorrelationBasedFilter: ranking")
    AllocateScalarMatrix(&px, inputs, parm->kscbf_k);

    for (int i = 0; i < inputs; i++) {
        AllocateScalarMatrix(&pxy, parm->kscbf_k, classes);
        ZeroVector(px[i], NULL, parm->kscbf_k);
        ZeroMatrix(pxy, NULL, parm->kscbf_k, classes);

        // Frequencies of x and joint probabilities x, y
        for (int j = 0; j < instances; j++) {
            int place = 0;
            while (data[j][i] > limits[place+1] || data[j][i] < limits[place])
                place++;

            px[i][place]++;
            pxy[place][labels[j]]++;
        }

        for (int j = 0; j < parm->kscbf_k; j++) {
            px[i][j] /= instances;
            for (int k = 0; k < classes; k++)
                pxy[j][k] /= instances;
        }

        // X entropy
        scalar Hx = 0.0;
        for (int j = 0; j < parm->kscbf_k; j++)
            if (px[i][j] > ZERO)
                Hx -= px[i][j]*log2(px[i][j]);

        // Shannon entropy per class
        scalar Hxc = 0.0;
        for (int j = 0; j < classes; j++)
            for (int k = 0; k < parm->kscbf_k; k++)
                if (pxy[k][j] > ZERO)
                    Hxc -= pxy[k][j]*log(pxy[k][j]);

        FreeScalarMatrix(&pxy, parm->kscbf_k);

        scalar MIxc = Hx + Hc - Hxc;
        scalar SUxc = 2.0*(MIxc/(Hx + Hc));

        ranking[i].feature = i;
        ranking[i].score = SUxc;
    }

    // Sort features from highest to lowest score
    qsort(ranking, inputs, sizeof(t_RankInfo), CompareScore);

    for (int i = 0; i < inputs; i++)
        S[i] = 1;

    // Step 2: Redundancy analysis
    for (int i = 0; i < inputs; i++) {
        int x = ranking[i].feature;
        if (S[x]) {
            // Apply K-S test and remove features in the next places
            for (int j = i+1; j < inputs; j++)
                if (S[ranking[j].feature]) {
                    scalar maxabd = fabs(px[x][0] - px[ranking[j].feature][0]);
                    for (int k = 1; k < parm->kscbf_k; k++)
                        if (fabs(px[x][k] - px[ranking[j].feature][k]) > maxabd)
                            maxabd = fabs(px[x][k] - px[ranking[j].feature][k]);

                    scalar lambda = sqrt(instances/2.0)*maxabd;
                    scalar c_alpha = sqrt(-0.5*log(parm->kscbf_alpha));

                    if (lambda <= c_alpha)
                        // Remove because both distributions are equal
                        S[ranking[j].feature] = 0;
                }
        }
    }

    FreeScalarMatrix(&px, inputs);
    FREE(ranking);
}

// Auxiliary function for qsort: from highest to lowest score
int CompareScore(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_RankInfo *) y)->score - ((t_RankInfo *) x)->score) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}
