/***************************************************************************
 *   Copyright (C) 2009 by Aida de Haro-García                             *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"

/*
 * Huan Liu and Rudy Setiono, "Scalable Feature Selection For Large Sized Databases", In Proceedings of the Fourth World Congress on Expert Systems, 1998, 521 - 528.
 */

void LvfAlg(const t_Parameters *parm, int *S, int n_feat, scalar **data, int n, int *labels, int outputs, int max_tries)
{
    int i, j, n_sel, max_sel = 0, *sel_feat = NULL;
    static int *intervals = NULL;
    scalar max_incons = 0.0, cur_incons = 0.0;
    static scalar **limits;

    if (limits == NULL) {
        //Memory allocation for auxiliary data structure
        MALLOC(limits, scalar *, n_feat, "LvfAlg: limits");
        //Set limits to NULL
        for (int k = 0; k < n_feat; k++)
            limits[k] = NULL;
        MALLOC(intervals, int, n_feat, "LvfAlg: intervals");

        // Discretize dataset
        switch (parm->discretizer) {
            case dFAYYAD:
                FayyadAndIraniMDLMethod(data, n, outputs, labels, n_feat, limits, intervals, NULL, NULL);
                break;

            case dWPKID:
                WPKIDMethod(data, n, n_feat, limits, intervals, NULL);
                break;

            case dNONE:
                //Do nothing
                break;
        }
    }

    time_t init_time = time(NULL);

    if (pow(2, n_feat) > max_tries)  {
        //Memory allocation
        MALLOC(sel_feat, int, n_feat, "LvfAlg");

        //Initially all features are selected
        for (j = 0; j < n_feat; j++)
            sel_feat[j] = S[j] = 1;

        //Set initial values
        max_sel = n_feat;
        max_incons = InconsistCount(n_feat, data, n, sel_feat, labels, outputs, limits, intervals);

        for (i = 0; i < max_tries; i++) {
            // Randomly select features and count number of selected features
            n_sel = 0;

            // Must be selected at least one feature
            while (!n_sel)  {
                for (j = 0; j < n_feat; j++)  {
                    sel_feat[j] = (int) RANDOM(0, 2);
                    n_sel += sel_feat[j];
                }
            }

            if (n_sel <= max_sel)  {
                cur_incons = InconsistCount(n_feat, data, n, sel_feat, labels, outputs, limits, intervals);

                if (cur_incons <= max_incons)  {
                    // Store new best: less inconsistency value and less number
                    if (n_sel < max_sel)  {
                        max_incons = cur_incons;
                        max_sel = n_sel;
                        for (j = 0; j < n_feat; j++)
                            S[j] = sel_feat[j];
                    }
                }
            }

            // Exit if time limit exceeded
            if (parm->time_limit && (time(NULL) - init_time > parm->time_limit))
                break;
        }
        FREE(sel_feat);
    } else {
        // If max_tries is more than all possible feature combination: run exhaustive search
        int *pop = NULL, num, cont, j, top = pow(2, n_feat);
        MALLOC(pop, int, n_feat, "LvfAlg: pop");

        // Initially all features are selected
        for (j = 0; j < n_feat; j++)
            S[j] = 1;

        // Set initial values
        max_sel = n_feat;
        max_incons = InconsistCount(n_feat, data, n, S, labels, outputs, limits, intervals);

        //Get all possible feature combinations
        for (i = 1; i < top; i++)  {
            n_sel = 0;
            num = i;

            for (cont = 0; cont < n_feat; cont++) {
                pop[cont] = num % 2;
                if (pop[cont])
                    n_sel++;
                num /= 2;
            }

            if (n_sel <= max_sel)  {
                cur_incons = InconsistCount(n_feat, data, n, pop, labels, outputs, limits, intervals);
                if (cur_incons <= max_incons)  {
                    // Store new best: less inconsistency value and less number
                    if (n_sel < max_sel)  {
                        max_incons = cur_incons;
                        max_sel = n_sel;
                        for (j = 0; j < n_feat; j++)
                            S[j] = pop[j];
                    }
                }
            }

            // Exit if time limit exceeded
            if (parm->time_limit && (time(NULL) - init_time > parm->time_limit))
                break;
        }
        FREE(pop);
    }
}
