/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "wrapper.h"
#include "filter.h"
#include "ranking.h"
#include "fs.h"
#include "is.h"
#include "classifiers.h"
#include "data.h"
#include "ensemble.h"
#include "maths.h"

#define PRECISION   0.000001
#define MAX_ATTEMPTS(x)    ((x)*10)
#define FOLDS   10

static inline void PRINTSEPARATOR(int width, bool show)
{
    if (!show)
        return;

    for (int i = 0; i < width; i++)
        printf("─");
    printf("\n");
}

int MarginBasedFeatureSelection(const t_Parameters *parm, int *selection, scalar **data, int instances, int *labels, int features, int classes, int minority, bool as_a_mutation, int **index)
{
    scalar *w, best_performance, best_accuracy = 0.0, best_reduction = 0.0, accuracy, averaged_performance = 0.0, **t_data, **sq_dists = NULL, weighted_sampling_ratio = 0.0, r = 0.0, best_r = 0.0;
    int second_chance = 0, selected = -1, *best_S = NULL, *S, *added_list = NULL, added = 0, *reweight_s = NULL, t_instances, *t_labels, *list, iteration_lenght, attempts = 0,
        *prediction, losing_streak = 0, best_second_chance= 0, best_losing_streak = 0, best_selected = 0, *forbidden, *unforbidden;
    bool *ok, filter;
    t_WeightInfo *w_info = NULL;
    time_t beggining = 0;

    time_t init_time = time(NULL);

    MALLOC(S, int, features, "MarginBasedFeatureSelection: S");
    MALLOC(best_S, int, features, "MarginBasedFeatureSelection: best_S");
    MALLOC(prediction, int, instances, "MarginBasedFeatureSelection: prediction");
    MALLOC(forbidden, int, features, "MarginBasedFeatureSelection: forbidden");
    MALLOC(unforbidden, int, features, "MarginBasedFeatureSelection: unforbidden");

    // A forbidden feature is no longer considered for addition
    for (int i = 0; i < features; i++)
        forbidden[i] = unforbidden[i] = parm->mabuse_feature_life;

    // Set all predictions to wrong in case needed
    for (int i = 0; i < instances; i++)
        prediction[i] = -1;

    // Initialize selection and weights
    MALLOC(w, scalar, instances, "MarginBasedFeatureSelection: w");
    MALLOC(ok, bool, instances, "MarginBasedFeatureSelection: ok");
    if (parm->mabuse_ensemble_method == eReweightBoost)
        MALLOC(reweight_s, int, instances, "MarginBasedFeatureSelection: reweight_s");

    // List of features in case of limit
    MALLOC(list, int, features, "MarginBasedFeatureSelection: list");
    for (int i = 0; i < features; i++)
        list[i] = i;

    for (int i = 0; i < instances; i++)
        w[i] = 1.0/instances;

    // Initially use all instances or sample
    bool fs_use_cache = parm->fs_use_cache;
    int max_instances = (parm->mabuse_max_instances < instances) ? parm->mabuse_max_instances : 0;
    if (max_instances) {
        t_instances = MIN(max_instances, instances);
        MALLOC(t_data, scalar *, t_instances, "MarginBasedFeatureSelection: t_data");
        MALLOC(t_labels, int, t_instances, "MarginBasedFeatureSelection: t_lables");
        fs_use_cache = false;
    } else {
        t_data = data;
        t_instances = instances;
        t_labels = labels;
    }

    // If 1-NN the consider it as a filter and speed up as much as possible
    filter = (parm->k == 1 && parm->mabuse_classifier == kNN) ? true : false;
    SHOWPROGRESS(parm->show_progress > 0, "Margin-Based Feature Selection version: %s\n", filter ? "FILTER" : "WRAPPER");

    // Iterate over all features or subset
    iteration_lenght = parm->mabuse_feat_per_iteration ? MIN(parm->mabuse_feat_per_iteration, features) : features;

    // Use cache?
    if (filter && fs_use_cache && !max_instances) {
        scalar size = (instances*sizeof(scalar *) + instances*instances*sizeof(scalar))/pow(2.0, 30.0);
        if (size > MAXIMUM_CACHE_SIZE) {
            SHOWPROGRESS(true, "Cache size of %lf.1GB exceeds limits\n", size);
        } else {
            ReserveAndInitializeCache(&sq_dists, instances);

            // If starting with all features selected the cache must be initialize to actual distances
            if (parm->search_heuristic == SBS || parm->search_heuristic == SFBS || (parm->search_heuristic == LRS && parm->plus_l < parm->minus_r))
                InitializeCacheToActualSqDistance(sq_dists, features, data, instances, index);
        }
    }

    // Initial reduction
    scalar reduction = 1.0 - ((scalar) selected/features);

    // Set l and r appropriately
    int plus_l = -1, minus_r = -1;
    switch (parm->search_heuristic) {
        case SFS:
            plus_l = 1;
            minus_r = 0;
            break;

        case SBS:
            plus_l = 0;
            minus_r = 1;
            break;

        case LRS:
            plus_l = parm->plus_l;
            minus_r = parm->minus_r;
            break;

        case SFFS:
        case SFBS:
            plus_l = minus_r = 1;
            break;
    }

    best_performance = -1.0;
    accuracy = 0.0;
    int iterations = 0;
    scalar previous_performance = 0.0, performance = 0.0;
    int last_added = -1, last_removed = -1;

    // Structure to order instances
    weighted_sampling_ratio = parm->mabuse_filter_weighted_sampling_ratio;
    MALLOC(w_info, t_WeightInfo, instances, "MarginBasedFeatureSelection: w_info");
    for (int i = 0; i < instances; i++) {
        w_info[i].instance = i;
        w_info[i].w = 1.0;
    }

    scalar *py = NULL, **limits = NULL, *ranks = NULL;
    int *intervals = NULL;
    t_RankInfo *rank = NULL;
    if ((parm->mabuse_sort_by_weighted_mi_limit > features && iteration_lenght < features) || parm->mabuse_use_rankings_for_perf) {
        // Obtain p(y) for class
        MALLOC(py, scalar, classes, "MarginBasedFeatureSelection: py");
        ZeroVector(py, NULL, classes);
        for (int i = 0; i < instances; i++)
            py[labels[i]]++;

        for (int i = 0; i < classes; i++)
            py[i] /= instances;

        // Memory allocation for auxiliary data structure
        MALLOC(limits, scalar *, features, "MarginBasedFeatureSelection: limits");
        // Set limits to NULL
        for (int k = 0; k < features; k++)
            limits[k] = NULL;
        MALLOC(intervals, int, features, "MarginBasedFeatureSelection: intervals");

        // Discretize dataset
        beggining = time(NULL);
        switch (parm->discretizer) {
            case dFAYYAD:
                FayyadAndIraniMDLMethod(data, instances, classes, labels, features, limits, intervals, NULL, index);
                break;

            case dWPKID:
                WPKIDMethod(data, instances, features, limits, intervals, NULL);
                break;

            case dNONE:
                OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
                break;
        }
        MALLOC(rank, t_RankInfo, features, "MarginBasedFeatureSelection: rank");
        SHOWTIME(parm->time_all, NULL, "TIME -> Discretization: %ld s\n", time(NULL) - beggining);

        if (parm->mabuse_use_rankings_for_perf) {
            t_RankInfo *ranking;
            MALLOC(ranking, t_RankInfo, features, "MarginBasedFeatureSelection: ranking");
            time_t init_time = time(NULL);

            MutualInformationFeatureRanking(parm, features, data, labels, classes, instances, ranking, limits, intervals, index);
            SHOWTIME(parm->time_all, NULL, "TIME -> MI algorithm: %ld s\n", time(NULL) - init_time);

            fprintf(parm->out_file, "@Feature ranking time: %ld\n", time(NULL) - init_time);

            // Sort ranking list
            beggining = time(NULL);
            qsort(ranking, features, sizeof(t_RankInfo), CompareRankingScore);

            MALLOC(ranks, scalar, features, "MarginBasedFeatureSelection: ranks");
            for (int i = 0; i < features; i++)
                ranks[ranking[i].feature] = 1.0 - ((scalar) i)/features;
        }
    }

    // Initialization depends on type of heuristic search or if the FS is used as a mutation operator
    if (as_a_mutation)
        for (int i = 0; i < features; i++)
            S[i] = selection[i];
    else {
        if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
            // Start with no features
            selected = 0;
            for (int i = 0; i < features; i++)
                S[i] = 0;

            // Use ranking to perform an initial selection?
            if (parm->mabuse_ranking_initial_selection_percentage > 0.0) {
                t_RankInfo *ranking;
                MALLOC(ranking, t_RankInfo, features, "MarginBasedFeatureSelection: ranking");
                time_t init_time = time(NULL);

                FeatureRankingAlgorithm(parm, parm->fralgorithm, features, data, labels, classes, instances, minority, ranking, limits, intervals, index);
                SHOWTIME(parm->time_all, NULL, "TIME -> Feature ranking algorithm: %ld s\n", time(NULL) - init_time);

                fprintf(parm->out_file, "@Feature ranking time: %ld\n", time(NULL) - init_time);

                // Sort ranking list
                beggining = time(NULL);
                qsort(ranking, features, sizeof(t_RankInfo), CompareRankingScore);

                for (int i = 0; i < features*parm->mabuse_ranking_initial_selection_percentage; i++) {
                    S[ranking[i].feature] = 1;
                    selected++;
                    SHOWPROGRESS(parm->show_progress > 0, "Initial selection based on ranking added feature: %*d\r", WIDTH(features), ranking[i].feature+1);

                    if (sq_dists)
                        UpdateCache(sq_dists, ranking[i].feature, -1, data, instances, index);
                }
                SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
                SHOWTIME(parm->time_all, NULL, "TIME -> Initial selection cache updating: %ld s\n", time(NULL) - beggining);
            }
        } else {
            // Start with all features
            selected = features;
            for (int i = 0; i < features; i++)
                S[i] = 1;

            SHOWPROGRESS(parm->show_progress > 0, "Added ALL features%c", '\n');
        }
    }

    do {
        // Sample for training
        if (max_instances)
            StratifiedSample(data, instances, labels, t_data, t_labels, t_instances, classes);

        scalar best_e_i, e_i = 0.0, previous = 0.0;

        previous_performance = parm->use_averaged_accuracy_for_stop ? averaged_performance : performance;

        for (int j = 0; j < plus_l && selected < features; j++) {
            int best_feature = -1;
            best_e_i = 1.1;

            // If only subset considered randomize
            if (iteration_lenght < features) {
                if (parm->mabuse_sort_by_weighted_mi_limit > features) {
                    beggining = time(NULL);
                    for (int i = 0; i < features; i++)
                        rank[rank[i].feature = i].score = WeightedMutualInformation(i, py, data, instances, classes, intervals, limits, labels, w);
                    qsort(rank, features, sizeof(t_RankInfo), CompareRankingScore);
                    for (int i = 0; i < features; i++)
                        list[i] = rank[i].feature;
                    SHOWTIME(parm->time_all, NULL, "TIME -> Weighted mutual information calculation: %ld s\n", time(NULL) - beggining);
                } else
                    RandomizeVector(list, features);
            }

            // If sampling of most difficult instances order instances
            if (parm->mabuse_filter_weighted_sampling_ratio < 1.0) {
                for (int i = 0; i < instances; i++)
                    w_info[w_info[i].instance = i].w = w[i];
                qsort(w_info, instances, sizeof(t_WeightInfo), CompareWeight);

                // If negative adjust to error
                if (parm->mabuse_filter_weighted_sampling_ratio <= 0.0) {
                    // Get error from last prediction
                    weighted_sampling_ratio = 0.0;
                    for (int i = 0; i < instances; i++)
                        if (prediction[i] != labels[i])
                            weighted_sampling_ratio++;
                    weighted_sampling_ratio /= instances;

                    // Minimum limited by parameter absolute value
                    weighted_sampling_ratio = MAX(weighted_sampling_ratio, fabs(parm->mabuse_filter_weighted_sampling_ratio));
                }
                SHOWPROGRESS(parm->show_progress > 1, "Weighted sampling: %6.2lf%%\n", weighted_sampling_ratio*100);
            }

            bool checked_at_least_one = false;
            beggining = time(NULL);
            for (int i = 0; i < iteration_lenght || (!checked_at_least_one && i < features); i++) {
                int next = list[i];

                if (!S[next] && forbidden[next] > 0) {
                    // Add feature and check
                    S[next] = 1;
                    checked_at_least_one = true;

                    SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "Evaluating for addition: %*d\r", WIDTH(features), next+1);

                    // Evaluate error
                    if (filter) {
                        if (weighted_sampling_ratio < 1.0)
                            NearestNeighborWithSemiCacheWihtWeights(data, S, instances, features, labels, sq_dists, prediction, next, -1, w_info, instances*weighted_sampling_ratio,
                                                                    parm->mabuse_folds, parm->mabuse_roulette_selection);
                        else
                            NearestNeighborWithSemiCache(data, S, instances, features, labels, sq_dists, prediction, next, -1, parm->mabuse_folds);
                        e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                    } else
                        e_i = GetWeightedErrorFS(parm, parm->mabuse_classifier, S, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL, t_data,
                                                 t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));
                    if (e_i < best_e_i) {
                        best_e_i = e_i;
                        best_feature = next;
                    }

                    S[next] = 0;
                }
            }
            SHOWTIME(parm->time_all, NULL, "TIME -> Selection of next best iteration: %ld s\n", time(NULL) - beggining);
            SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "%c", '\n');

            // Add best
            if (best_feature != -1) {
                S[best_feature] = 1;
                unforbidden[best_feature]--;
                if (best_feature == last_removed)
                    second_chance++;
                // Need to resize added list?
                if (!(added % features))
                    REALLOC(added_list, int, added+features, "MarginBasedFeatureSelection: added_list");
                added_list[added] = last_added = best_feature;
                added++;
                last_removed = -1;
                selected++;
                SHOWPROGRESS(parm->show_progress > 0, "%3d: Added %*d\n", iterations+1, WIDTH(features), best_feature+1);

                // Recalculate boosting weights
                if (filter) {
                    if (sq_dists)
                        UpdateCache(sq_dists, best_feature, -1, data, instances, index);
                    NearestNeighborWithSemiCache(data, S, instances, features, labels, sq_dists, prediction, -1, -1, parm->mabuse_folds);
                    e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                } else
                    e_i = GetWeightedErrorFS(parm, parm->mabuse_classifier, S, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL, t_data,
                                             t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));
                if (parm->mabuse_ensemble_method == eReweightBoost) {
                    // Evaluate elements added
                    for (int i = 0; i < features; i++)
                        reweight_s[i] = 0;
                    for (int i = 0; i < parm->reweightboost_r && i < added; i++)
                        reweight_s[added_list[i]] = S[added_list[i]];
                    if (filter) {
                        NearestNeighborWithSemiCache(data, S, instances, features, labels, NULL, prediction, -1, -1, parm->mabuse_folds);
                        e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                    } else
                        GetWeightedErrorFS(parm, parm->mabuse_classifier, reweight_s, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL, t_data,
                                           t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));
                }
                UpdateWeights(parm, instances, w, e_i, ok, parm->mabuse_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations), parm->mabuse_ensemble_method, data,
                              labels, NULL, features, parm->k, classes, NULL, index, false, true, NULL, parm->mabuse_classifier, S);

                // If sampling of most difficult instances order instances
                if (parm->mabuse_filter_weighted_sampling_ratio < 1.0) {
                    for (int i = 0; i < instances; i++)
                        w_info[w_info[i].instance = i].w = w[i];
                    qsort(w_info, instances, sizeof(t_WeightInfo), CompareWeight);

                    // If negative adjust to error
                    if (parm->mabuse_filter_weighted_sampling_ratio < 0.0) {
                        // Get error from last prediction
                        weighted_sampling_ratio = 0.0;
                        for (int i = 0; i < instances; i++)
                            if (prediction[i] != labels[i])
                                weighted_sampling_ratio++;
                        weighted_sampling_ratio /= instances;

                        // Minimum limited by parameters absolute value
                        weighted_sampling_ratio = MAX(weighted_sampling_ratio, fabs(parm->mabuse_filter_weighted_sampling_ratio));
                    }
                    SHOWPROGRESS(parm->show_progress > 1, "Weighted sampling: %6.2lf%%\n", weighted_sampling_ratio*100);
                }

                // Conditional exclusion for eFloatBoost
                if (parm->mabuse_ensemble_method == eFloatBoost) {
                    beggining = time(NULL);
                    int excluded = ConditionalExclusionMarginBasedFS(parm, S, selected, data, instances, features, labels, classes, minority, t_data, t_instances, t_labels, filter, prediction,
                                   sq_dists, forbidden, last_added, w_info, instances, index);
                    selected -= excluded;
                    SHOWTIME(parm->time_all, NULL, "TIME -> Conditional exclusion: %ld s\n", time(NULL) - beggining);
                }

                // For SFBS keep addition in the performance does not worsen, otherwise undo
                if (parm->search_heuristic == SFBS) {
                    accuracy = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, S, labels, instances, classes, data, features, parm->mabuse_condit_excl_with_margin ? pfMargin :
                               parm->mabuse_performance_measure, minority, t_data, t_instances, t_labels, filter, prediction, w_info, instances);
                    if (previous < accuracy) {
                        j--;
                        previous = accuracy;
                    } else {
                        S[best_feature] = 0;
                        selected--;
                        forbidden[best_feature]--;
                        SHOWPROGRESS(parm->show_progress > 0, "Undo added feature: %d\n", best_feature+1);
                        if (filter) {
                            if (sq_dists)
                                UpdateCache(sq_dists, -1, best_feature, data, instances, index);
                            NearestNeighborWithSemiCache(data, S, instances, features, labels, sq_dists, prediction, -1, -1, parm->mabuse_folds);
                            e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                        } else
                            e_i = GetWeightedErrorFS(parm, parm->mabuse_classifier, S, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL,
                                                     t_data, t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));
                        UpdateWeights(parm, instances, w, e_i, ok, parm->mabuse_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations), parm->mabuse_ensemble_method,
                                      data, labels, NULL, features, parm->k, classes, NULL, index, false, true, NULL, parm->mabuse_classifier, S);
                    }
                }
            } else if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l > minus_r))
                second_chance = 0;
        }

        beggining = time(NULL);
        if (parm->search_heuristic == SFFS)
            previous = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, S, labels, instances, classes, data, features, parm->mabuse_condit_excl_with_margin ? pfMargin :
                       parm->mabuse_performance_measure, minority, t_data, t_instances, t_labels, filter, prediction, w_info, instances);
        SHOWTIME(parm->time_all, NULL, "TIME -> SFFS backwards step: %ld s\n", time(NULL) - beggining);

        for (int j = 0; j < minus_r && selected > 1; j++) {
            int worst_feature = -1;
            best_e_i = 1.1;

            // If only subset considered randomize
            if (iteration_lenght < features)
                RandomizeVector(list, features);

            // If sampling of most difficult instances order instances
            if (parm->mabuse_filter_weighted_sampling_ratio < 1.0) {
                for (int i = 0; i < instances; i++)
                    w_info[w_info[i].instance = i].w = w[i];
                qsort(w_info, instances, sizeof(t_WeightInfo), CompareWeight);

                // If negative adjust to error
                if (parm->mabuse_filter_weighted_sampling_ratio < 0.0) {
                    // Get error from last prediction
                    weighted_sampling_ratio = 0.0;
                    for (int i = 0; i < instances; i++)
                        if (prediction[i] != labels[i])
                            weighted_sampling_ratio++;
                    weighted_sampling_ratio /= instances;

                    // Minimum limited by parameters absolute value
                    weighted_sampling_ratio = MAX(weighted_sampling_ratio, fabs(parm->mabuse_filter_weighted_sampling_ratio));
                }
                SHOWPROGRESS(parm->show_progress > 1, "Weighted sampling: %6.2lf%%\n", weighted_sampling_ratio*100);
            }

            for (int i = 0; i < iteration_lenght; i++) {
                int next = list[i];

                if (S[next] && unforbidden[next] > 0) {
                    // Remove instance and check
                    S[next] = 0;

                    SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "Evaluating for deletion: %*d\r", WIDTH(features), next+1);

                    if (filter) {
                        if (weighted_sampling_ratio < 1.0)
                            NearestNeighborWithSemiCacheWihtWeights(data, S, instances, features, labels, sq_dists, prediction, -1, next, w_info, instances*weighted_sampling_ratio,
                                                                    parm->mabuse_folds, parm->mabuse_roulette_selection);
                        else
                            NearestNeighborWithSemiCache(data, S, instances, features, labels, sq_dists, prediction, -1, next, parm->mabuse_folds);
                        e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                    } else
                        e_i = GetWeightedErrorFS(parm, parm->mabuse_classifier, S, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL, t_data,
                                                 t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));

                    if (e_i < best_e_i) {
                        best_e_i = e_i;
                        worst_feature = next;
                    }

                    S[next] = 1;
                }
            }
            SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "%c", '\n');

            // Remove worst
            if (worst_feature != -1) {
                S[worst_feature] = 0;
                // Forbid for the feature
                forbidden[worst_feature]--;
                if (worst_feature == last_added)
                    second_chance++;
                last_removed = worst_feature;
                last_added = -1;
                selected--;

                SHOWPROGRESS(parm->show_progress > 0, "%3d: Deleted %*d\n", iterations+1, WIDTH(features), worst_feature+1);

                // Recalculate Boost weights
                if (filter)  {
                    if (sq_dists)
                        UpdateCache(sq_dists, -1, worst_feature, data, instances, index);
                    NearestNeighborWithSemiCache(data, S, instances, features, labels, sq_dists, prediction, -1, -1, parm->mabuse_folds);
                    e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                } else
                    e_i = GetWeightedErrorFS(parm, parm->mabuse_classifier, S, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL, t_data,
                                             t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));
                UpdateWeights(parm, instances, w, e_i, ok, parm->mabuse_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations), parm->mabuse_ensemble_method, data,
                              labels, NULL, features, parm->k, classes, NULL, index, false, true, NULL, parm->mabuse_classifier, S);

                // If sampling of most difficult instances order instances
                if (parm->mabuse_filter_weighted_sampling_ratio < 1.0) {
                    for (int i = 0; i < instances; i++)
                        w_info[w_info[i].instance = i].w = w[i];
                    qsort(w_info, instances, sizeof(t_WeightInfo), CompareWeight);

                    // If negative adjust to error
                    if (parm->mabuse_filter_weighted_sampling_ratio < 0.0) {
                        // Get error from last prediction
                        weighted_sampling_ratio = 0.0;
                        for (int i = 0; i < instances; i++)
                            if (prediction[i] != labels[i])
                                weighted_sampling_ratio++;
                        weighted_sampling_ratio /= instances;

                        // Minimum limited by parameters absolute value
                        weighted_sampling_ratio = MAX(weighted_sampling_ratio, fabs(parm->mabuse_filter_weighted_sampling_ratio));
                    }
                    SHOWPROGRESS(parm->show_progress > 1, "Weighted sampling: %6.2lf%%\n", weighted_sampling_ratio*100);
                }

                // For SFFS keep deletion if the performance does not worsen, otherwise undo
                if (parm->search_heuristic == SFFS) {
                    accuracy = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, S, labels, instances, classes, data, features, parm->mabuse_condit_excl_with_margin ? pfMargin :
                               parm->mabuse_performance_measure, minority, t_data, t_instances, t_labels, filter, prediction, w_info, instances);
                    if (previous <= accuracy) {
                        j--;
                        previous = accuracy;
                    } else {
                        S[worst_feature] = 1;
                        selected++;
                        unforbidden[worst_feature]--;
                        SHOWPROGRESS(parm->show_progress > 0, "Undo removed instance: %d\n", worst_feature+1);
                        if (filter) {
                            if (sq_dists)
                                UpdateCache(sq_dists, worst_feature, -1, data, instances, index);
                            NearestNeighborWithSemiCache(data, S, instances, features, labels, sq_dists, prediction, -1, -1, parm->mabuse_folds);
                            e_i = GetFilterWeightedError(instances, labels, ok, w, prediction);
                        } else
                            e_i = GetWeightedErrorFS(parm, parm->mabuse_classifier, S, data, features, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, NULL,
                                                     t_data, t_instances, t_labels, NULL, !IsEnsemble(parm->mabuse_classifier));
                        UpdateWeights(parm, instances, w, e_i, ok, parm->mabuse_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations), parm->mabuse_ensemble_method,
                                      data, labels, NULL, features, parm->k, classes, NULL, index, false, true, NULL, parm->mabuse_classifier, S);
                    }
                }
            } else if (parm->search_heuristic == SBS || parm->search_heuristic == SFBS || (parm->search_heuristic == LRS && parm->plus_l < minus_r))
                second_chance = 0;
        }

        if (parm->search_heuristic == SFBS)
            previous = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, S, labels, instances, classes, data, features, parm->mabuse_condit_excl_with_margin ? pfMargin :
                       parm->mabuse_performance_measure, minority, t_data, t_instances, t_labels, filter, prediction, w_info, instances);

        // Get unweighted accuracy measure
        beggining = time(NULL);
        iterations++;
        accuracy = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, S, labels, instances, classes, data, features, parm->mabuse_performance_measure, minority, t_data, t_instances,
                   t_labels, filter, prediction, w_info, instances);
        reduction = 1.0 - ((scalar) selected/features);
        SHOWTIME(parm->time_all, NULL, "TIME -> Performance calculation: %ld s\n", time(NULL) - beggining);

        // Obtain performance as a combination of reduction and accuracy
        // If both ZERO use geometric mean
        if (parm->mabuse_use_rankings_for_perf) {
            // Average ranking
            r = 0.0;
            for (int i = 0; i < features; i++)
                if (S[i])
                    r += ranks[i];
            r /= selected;

            // Maximum average ranking
            scalar max_r = 0.0;
            for (int i = 0; i < selected; i++)
                max_r += 1.0 - ((scalar) i)/features;
            max_r /= selected;

            r = 1.0 - (max_r-r);

            SHOWPROGRESS(parm->show_progress > 0, "Avarage ranking of selected features: %lf\n", r);
            if (parm->ok_w <= ZERO && parm->reduction_w <= ZERO && parm->rank_w <= 0.0)
                performance = pow(accuracy*reduction*r, 1.0/3);
            else
                performance = accuracy*parm->ok_w+reduction*parm->reduction_w+parm->rank_w*r;
        } else {
            if (parm->ok_w <= ZERO && parm->reduction_w <= ZERO)
                performance = sqrt(accuracy*reduction);
            else
                performance = accuracy*parm->ok_w+reduction*parm->reduction_w;
        }
        averaged_performance = (averaged_performance*(iterations-1)+performance)/iterations;

        bool best_so_far = false;
        if (parm->use_averaged_accuracy_for_stop) {
            if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
                if (previous_performance > averaged_performance || fabs(previous_performance - averaged_performance) <= PRECISION) {
                    second_chance++;
                    losing_streak++;
                } else
                    losing_streak = 0;
            } else {
                if (previous_performance > averaged_performance) {
                    second_chance++;
                    losing_streak++;
                } else
                    losing_streak = 0;
            }
            best_so_far = performance > best_performance;
        } else {
            if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
                if (previous_performance > performance || fabs(previous_performance - performance) <= PRECISION) {
                    second_chance++;
                    losing_streak++;
                } else
                    losing_streak = 0;
            } else {
                if (previous_performance > performance) {
                    second_chance++;
                    losing_streak++;
                } else
                    losing_streak = 0;
            }
            best_so_far = performance >= best_performance;
        }

        SHOWPROGRESS(parm->show_progress > 0, "ACC,RED,PERF,RANK: [%6.4lf|%6.4lf|%6.4lf|%6.4lf] MEAN-PERF: %6.4lf CHANCE,STREAK,SELECTED: [%*d|%d|%*d]\n", accuracy, reduction, performance, r,
                     averaged_performance, WIDTH(parm->second_chances), second_chance, losing_streak, WIDTH(features), selected);

        if (best_so_far) {
            best_performance = performance;
            best_accuracy = accuracy;
            best_r = r;
            best_reduction = reduction;
            best_second_chance = second_chance;
            best_losing_streak = losing_streak;
            best_selected = selected;
            // Best selection so far
            for (int i = 0; i < features; i++)
                best_S[i] = S[i];
        }

        attempts++; // Avoid infinite recursion

        // Limited time execution
        if (parm->time_limit && time(NULL)-init_time > parm->time_limit) {
            fprintf(parm->out_file, "Time limit %ld exceeded, exiting after %ld second(s)\n", parm->time_limit, time(NULL)-init_time);
            break;
        }
    } while (!MarginBasedFeatureSelectionStop(parm, selected, second_chance, losing_streak, features, attempts));

    PRINTSEPARATOR(83+WIDTH(parm->second_chances)+WIDTH(losing_streak)+WIDTH(features), parm->show_progress);
    SHOWPROGRESS(parm->show_progress > 0, "ACC,RED,PERF,RANK: [%6.4lf|%6.4lf|%6.4lf|%6.4lf] MEAN-PERF: ------ CHANCE,STREAK,SELECTED: [%*d|%d|%*d]\n", best_accuracy, best_reduction, best_performance,
                 best_r, WIDTH(parm->second_chances), best_second_chance, best_losing_streak, WIDTH(features), best_selected);
    PRINTSEPARATOR(83+WIDTH(parm->second_chances)+WIDTH(losing_streak)+WIDTH(features), parm->show_progress);

    // Best selection
    selected = 0;
    for (int i = 0; i < features; i++)
        if ((selection[i] = best_S[i]))
            selected++;

    // Free memory
    FREE(best_S);
    FREE(forbidden);
    FREE(unforbidden);
    FREE(w);
    FREE(ok);
    FREE(added_list);
    if (parm->mabuse_ensemble_method == eReweightBoost)
        FREE(reweight_s);
    FREE(w_info);
    FREE(S);
    if (max_instances) {
        FREE(t_data);
        FREE(t_labels);
    }
    FREE(list);
    FREE(prediction);
    if (py) {
        FREE(py);
        FREE(rank);
        FREE(intervals);
        for (int i = 0; i < features; i++)
            FREE(limits[i]);
        FREE(limits);
    }

    if (ranks)
        FREE(ranks);

    return selected;
}

scalar SubspaceUnweightedPerformance(const t_Parameters *parm, t_Classifier classifier, int *subspace, int *labels, int instances, int classes, scalar **data, int features,
                                     t_PerformanceMeasure measure, int minority, scalar **t_data, int t_instances, int *t_labels, bool filter, int *prediction, t_WeightInfo *w_info, int top)
{
    int **M, majority;
    scalar performance = 0.0, theta_1 = 0.0, theta_2 = 0.0, sp, sn, minimum_margin;
    t_PerformanceInfo perf;

    AllocateIntegerMatrix(&M, classes, classes);

    if (filter && measure != pfMargin) {
        ZeroIntegerMatrix(M, classes, classes);

        for (int i = instances-top; i < instances; i++)
            M[labels[w_info[i].instance]][prediction[w_info[i].instance]]++;
    } else {
        if (top != instances)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted sampling without cache");

        ClassifierEvaluation(parm, data, instances, labels, t_data, t_instances, t_labels, subspace, features, NULL, parm->cv_hyperparameters, classes, classifier, NULL, minority, M,
                                   NULL, &perf, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    }

    switch (measure) {
        case pfAccuracy:
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            performance = theta_1/top;
            break;

        case pfKappa:
            // Obtain kappa
            // Sum of the diagonal
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            theta_1 *= top;

            for (int i = 0; i < classes; i++) {
                // Rows/columns total counts
                scalar row_i = 0, column_i = 0;

                for (int j = 0; j < classes; j++) {
                    row_i += M[i][j];
                    column_i += M[j][i];
                }

                theta_2 += row_i * column_i;
            }

            performance = (theta_1 - theta_2)/(((scalar) top)*top-theta_2);
            break;

        case pfGMean:
            if (classes != 2)
                OnErrorExit(ERR_PARAM, "G-mean measure for N-class problem");
            // Obtain G-mean
            majority = (minority+1) % 2;

            sp = (M[majority][majority]+M[majority][minority]) > ZERO ? ((scalar) M[majority][majority])/(M[majority][majority]+M[majority][minority]) : 0.0;
            sn = (M[minority][minority]+M[minority][majority]) > ZERO ? ((scalar) M[minority][minority])/(M[minority][minority]+M[minority][majority]) : 0.0;

            performance = sqrt(sn*sp);
            break;

        case pfMargin:
            if (top != instances)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted sampling for margin as performance measure");
            performance = ClassifierDatasetMargin(parm, classifier, t_data, t_instances, t_labels, data, instances, labels, features, classes, subspace, &minimum_margin, NULL, NULL, NULL, NULL,
                                                        true);
            break;

        case pfAverageAccuracy:
        case pfauPRC:
        case pfauROC:
        case pfRecallMacro:
        case pfRecallMacroGMean:
        case pfCC:
        case pfF1Macro:
        case pfPrecisionMacro:
        case pfF1:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Performance metric for Margin-Based FS");
    }

    FreeIntegerMatrix(&M, classes);

    return performance;
}

// Returns removed features from the current selection
int ConditionalExclusionMarginBasedFS(const t_Parameters *parm, int *subspace, int selected, scalar **data, int instances, int features, int *labels, int classes, int minority, scalar **t_data,
                                      int t_instances, int *t_labels, bool filter, int *prediction, scalar **sq_dists, int *forbidden, int last_added, t_WeightInfo *w_info, int top, int **index)
{
    int h_prime = -1, M, *new_prediction;
    scalar *h, best_perf, perf;
    bool removed, improves, *ok;

    M = selected;

    // Cannot remove if only one
    if (M == 1)
        return 0;

    // Current error is best
    MALLOC(ok, bool, instances, "ConditionalExclusionMarginBasedFS: ok");
    MALLOC(h, scalar, instances, "ConditionalExclusionMarginBasedFS: h");
    MALLOC(new_prediction, int, instances, "ConditionalExclusionMarginBasedFS: new_prediction");
    if (parm->mabuse_condit_excl_with_margin)
        best_perf = NearestNeighborMarginWithSemiCache(data, instances, labels, sq_dists, -1, -1, w_info, top);
    else
        best_perf = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, subspace, labels, instances, classes, data, features, parm->mabuse_performance_measure, minority, t_data,
                    t_instances, t_labels, filter, prediction, w_info, top);

    do {
        removed = improves = false;

        // Get the feature that less contribute to the classification accuracy
        for (int i = 0; i < features; i++)
            if (subspace[i] && i != last_added) {   // Just added cannot be removed to avoid infinite loops
                subspace[i] = 0;

                // Update prediction
                if (parm->mabuse_condit_excl_with_margin)
                    perf = NearestNeighborMarginWithSemiCache(data, instances, labels, sq_dists, -1, i, w_info, top);
                else {
                    NearestNeighborWithSemiCache(data, subspace, instances, features, labels, sq_dists, new_prediction, -1, i, parm->mabuse_folds);
                    perf = SubspaceUnweightedPerformance(parm, parm->mabuse_classifier, subspace, labels, instances, classes, data, features, parm->mabuse_performance_measure, minority,
                                                         t_data, t_instances, t_labels, filter, new_prediction, w_info, top);
                }
                subspace[i] = 1;

                if (perf >= best_perf) {
                    best_perf = perf;
                    h_prime = i;
                    improves = true;
                }
            }

        // Remove if reduction improves
        if (improves) {
            // Remove h prime from selected features
            subspace[h_prime] = 0;
            M--;

            removed = true;
            // Forbid feature
            forbidden[h_prime]--;

            SHOWPROGRESS(parm->show_progress > 0, "eFloatBoost Conditional Exclusion removed: %*d\n", WIDTH(features), h_prime+1);
            if (sq_dists)
                UpdateCache(sq_dists, -1, h_prime, data, instances, index);
        }
    } while (removed && M > 1);

    SHOWPROGRESS(parm->show_progress > 0, "eFloatBoost Conditional Exclusion removed %*d feature(s)\n", WIDTH(features), selected-M);

    FREE(ok);
    FREE(h);
    FREE(new_prediction);

    // If features removed update prediction?
    if (selected-M)
        NearestNeighborWithSemiCache(data, subspace, instances, features, labels, sq_dists, prediction, -1, -1, parm->mabuse_folds);

    return selected - M;
}

// Get weighted error and vector of classified instances for 1-NN rule
scalar GetFilterWeightedError(int instances, int *labels, bool *ok, scalar *w, int *prediction)
{
    scalar error = 0.0;

    // Test weighted error
    for (int k = 0; k < instances; k++) {
        if (prediction[k] == labels[k])
            // Ok
            ok[k] = true;
        else {
            // Misclassified
            ok[k] = false;
            error += w[k];
        }
    }

    return error;
}

// Classify a pattern using nearest neighbors rule with a cache of distance to all but one feature added or remove
void NearestNeighborWithSemiCache(scalar **data, int *S, int instances, int features, int *labels, scalar **sq_dists, int *prediction, int add, int remove, int folds)
{
    scalar value, dmin;
    int nearest = -1;
    int *fold;

    // Create random folds
    MALLOC(fold, int, instances, "NearestNeighborWithSemiCache: fold");
    // If folds = 1 just initialize fold[i] = i
    if (folds == 1)
        for (int i = 0; i < instances; i++)
            fold[i] = i;
    else
        for (int i = 0; i < instances; i++)
            fold[i] = RANDOM(0, folds);

    // Use cache?
    if (sq_dists) {
        // Add?
        if (add != -1) {
            // Get nearest neighbor
            for (int j = 0; j < instances; j++) {
                dmin = INFINITE;

                for (int i = 0; i < instances; i++)
                    // Test if nearest
                    if (fold[i] != fold[j] && (value = sq_dists[j][i] + SQR(data[j][add] - data[i][add])) < dmin) {
                        nearest = i;
                        dmin = value;
                    }

                prediction[j] = labels[nearest];
            }
        } else if (remove != -1) {
            // Get nearest neighbor
            for (int j = 0; j < instances; j++) {
                dmin = INFINITE;

                for (int i = 0; i < instances; i++)
                    // Test if nearest
                    if (fold[i] != fold[j] && (value = sq_dists[j][i] - SQR(data[j][remove] - data[i][remove])) < dmin) {
                        nearest = i;
                        dmin = value;
                    }

                prediction[j] = labels[nearest];
            }
        } else {    // Use current cache
            // Get nearest neighbor
            for (int j = 0; j < instances; j++) {
                dmin = INFINITE;

                for (int i = 0; i < instances; i++)
                    // Test if nearest
                    if (fold[i] != fold[j] && sq_dists[j][i] < dmin) {
                        nearest = i;
                        dmin = sq_dists[j][i];
                    }

                prediction[j] = labels[nearest];
            }
        }
    } else {
        for (int j = 0; j < instances; j++) {
            dmin = INFINITE;

            for (int i = 0; i < instances; i++)
                // Test if nearest
                if (fold[i] != fold[j] && (value = EuclideanDistanceSubspace(data[j], data[i], features, S)) < dmin) {
                    nearest = i;
                    dmin = value;
                }

            prediction[j] = labels[nearest];
        }
    }

    FREE(fold);
}

// Hierarchical stop criterion
bool MarginBasedFeatureSelectionStop(const t_Parameters *parm, int selected, int second_chance, int losing_streak, int features, int attempts)
{
    // Maximum number of attempts or all selected, finish right away
    if (attempts >= MAX_ATTEMPTS(features))
        return true;

    // Finish if no more features to add/delete
    if ((parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) && selected >= features)
        return true;

    if ((parm->search_heuristic == SBS || parm->search_heuristic == SFBS || (parm->search_heuristic == LRS && parm->plus_l < parm->minus_r)) && selected <= 1)
        return true;

    // If a minimum must be selected?
    if (selected < features*parm->mabuse_min_selected_percent)
        return false;

    // Losing streak and no second chance
    if (parm->mabuse_max_losing_streak && !parm->second_chances)
        return losing_streak > parm->mabuse_max_losing_streak;
    else if (!parm->mabuse_max_losing_streak && parm->second_chances)    // Second chance and no losing streak
        return second_chance > parm->second_chances;
    else if (parm->mabuse_max_losing_streak && parm->second_chances)    // Both
        return losing_streak > parm->mabuse_max_losing_streak || second_chance > parm->second_chances;
    else    // Don't allow steps back
        return losing_streak || second_chance;
}

// Auxiliary function for qsort: from best to worst weight
int CompareWeight(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_WeightInfo *) y)->w - ((t_WeightInfo *) x)->w) > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}

/*
 * Classify a pattern using nearest neighbors rule with a cache of distance to all but one feature added or remove
 * Only the n instances with largest weights are considered.
 * Weight info must be sorted from smallest to largest
 */

void NearestNeighborWithSemiCacheWihtWeights(scalar **data, int *S, int instances, int features, int *labels, scalar **sq_dists, int *prediction, int add, int remove, t_WeightInfo *w_info, int top,
        int folds, bool roulette_selection)
{
    scalar value, dmin, *cum_probs = NULL;
    int nearest = -1, *fold;

    // Create random folds
    // If folds = 1 just initialize fold[i] = i
    MALLOC(fold, int, instances, "NearestNeighborWithSemiCacheWihtWeights: fold");
    if (folds == 1)
        for (int i = 0; i < instances; i++)
            fold[i] = i;
    else
        for (int i = 0;  i < instances; i++)
            fold[i] = RANDOM(0, FOLDS);

    if (roulette_selection) {
        MALLOC(cum_probs, scalar, instances, "NearestNeighborWithSemiCacheWihtWeights: cum_probs");
        cum_probs[0] = w_info[0].w;
        for (int i = 1; i < instances; i++)
            cum_probs[i] = cum_probs[i-1] + w_info[i].w;
    }

    // Use cache?
    if (sq_dists) {
        // Add?
        if (add != -1) {
            // Get nearest neighbor
            for (int k = instances-top; k < instances; k++) {
                dmin = INFINITE;

                // Next worse
                int j =  !roulette_selection ? w_info[k].instance : RouletteSelection(cum_probs, instances);

                for (int i = 0; i < instances; i++)
                    // Test if nearest
                    if (fold[i] != fold[j] && (value = sq_dists[j][i] + SQR(data[j][add] - data[i][add])) < dmin) {
                        nearest = i;
                        dmin = value;
                    }

                prediction[j] = labels[nearest];
            }
        } else if (remove != -1) {
            // Get nearest neighbor
            for (int k = instances-top; k < instances; k++) {
                dmin = INFINITE;

                // Next worse
                int j =  !roulette_selection ? w_info[k].instance : RouletteSelection(cum_probs, instances);

                for (int i = 0; i < instances; i++)
                    // Test if nearest
                    if (fold[i] != fold[j] && (value = sq_dists[j][i] - SQR(data[j][remove] - data[i][remove])) < dmin) {
                        nearest = i;
                        dmin = value;
                    }

                prediction[j] = labels[nearest];
            }
        } else {    // Use current cache
            // Get nearest neighbor
            for (int k = instances-top; k < instances; k++) {
                dmin = INFINITE;

                // Next worse
                int j =  !roulette_selection ? w_info[k].instance : RouletteSelection(cum_probs, instances);

                for (int i = 0; i < instances; i++)
                    // Test if nearest
                    if (fold[i] != fold[j] && sq_dists[j][i] < dmin) {
                        nearest = i;
                        dmin = sq_dists[j][i];
                    }

                prediction[j] = labels[nearest];
            }
        }
    } else {
        for (int k = instances-top; k < instances; k++) {
            dmin = INFINITE;

            // Next worse
            int j =  !roulette_selection ? w_info[k].instance : RouletteSelection(cum_probs, instances);

            for (int i = 0; i < instances; i++)
                // Test if nearest
                if (fold[i] != fold[j] && (value = EuclideanDistanceSubspace(data[j], data[i], features, S)) < dmin) {
                    nearest = i;
                    dmin = value;
                }

            prediction[j] = labels[nearest];
        }
    }

    // The remaining are assumed correctly classified
    for (int k = 0; k < instances-top; k++)
        prediction[w_info[k].instance] = labels[w_info[k].instance];

    FREE(fold);
    if (roulette_selection)
        FREE(cum_probs);
}
