/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "filter.h"

/*
 * P. Mitra, C. A. Murthy and S. K. Pal, "Unsupervised feature selection using feature similarity", IEEE TPAMI 24(3):301-312, 2002.
 *
 */

void MaximalInformationCompressionIndexBasedMethod(int *micibm_k, scalar **data, int instances, int n_features, int *S)
{
    scalar **lambda2, **r_ik, epsilon;
    int **neighbors, selected, k;
    bool *visited;

    if (*micibm_k < 0)
        *micibm_k = n_features/2;
    else if (*micibm_k >= n_features)
        *micibm_k = n_features-1;

    MALLOC(lambda2, scalar *, n_features, "MaximalInformationCompressionIndexBasedMethod: lambda2");
    MALLOC(neighbors, int *, n_features, "MaximalInformationCompressionIndex: neighbors");
    MALLOC(r_ik, scalar *, n_features, "MaximalInformationCompressionIndexBasedMethod: r_ik");
    MALLOC(visited, bool, n_features, "MaximalInformationCompressionIndex: visited");
    for (int i = 0; i < n_features; i++) {
        MALLOC(lambda2[i], scalar, n_features, "MaximalInformationCompressionIndexBasedMethod: lambda2[i]");
        MALLOC(neighbors[i], int, *micibm_k, "MaximalInformationCompressionIndexBasedMethod: neighbors[i]");
        MALLOC(r_ik[i], scalar, *micibm_k, "MaximalInformationCompressionIndexBasedMethod: r_ik[i]")
    }

    // Get maximal information compression index for all pairs of features
    MaximalInformationCompressionIndex(lambda2, data, instances, n_features);

    // Start with all selected
    for (int i = 0; i < n_features; i++) {
        S[i] = 1;
        visited[i] = false;
    }
    selected = n_features;

    // Obtain r_i^k for every feature
    k = *micibm_k;
    epsilon = INFINITE;
    for (int i = 0; i < n_features; i++)
        GetKNeighbors(i, lambda2, n_features, k, neighbors[i], r_ik[i]);

    do {
        // Get minimum r_i^k of selected and unvisited instances
        scalar minimum = INFINITE;
        int minimum_index = -1;
        for (int i = 0; i < n_features; i++)
            if (!visited[i] && S[i] && minimum > r_ik[i][k-1]) {
                minimum = r_ik[i][k-1];
                minimum_index = i;
            }

        if (minimum_index == -1)
            break;

        // Remove neighbors of minimum r_i^k
        for (int i = 0; i < k; i++)
            if (S[neighbors[minimum_index][i]]) {
                S[neighbors[minimum_index][i]] = 0;
                selected--;
            }
        visited[minimum_index] = true;

        if (k > selected-1)
            k = selected-1;
        if (k == 1)
            break;

        epsilon = minimum;

        do {
            // Get new minimum
            minimum = INFINITE;
            for (int i = 0; i < n_features; i++)
                if (!visited[i] && S[i] && minimum > r_ik[i][k-1])
                    minimum = r_ik[i][k-1];
            if (minimum > epsilon)
                k--;
        } while (minimum > epsilon && k > 1);

        if (k == 1)
            break;
    } while (true);

    for (int i = 0; i < n_features; i++) {
        FREE(lambda2[i]);
        FREE(r_ik[i]);
        FREE(neighbors[i]);
    }
    FREE(lambda2);
    FREE(neighbors);
    FREE(r_ik);
    FREE(visited);
}


void MaximalInformationCompressionIndex(scalar **lambda2, scalar **data, int instances, int n_features)
{
    // Get covariance matrix
    CovarianceMatrix(data, instances, n_features, lambda2);

    // Get lambda 2 from covariance matrix
    for (int i = 1; i < n_features; i++)
        for (int j = 0; j < i; j++) {
            // Linear correlation coefficient
            scalar rho = lambda2[i][j]/sqrt(lambda2[i][i]*lambda2[j][j]);

            lambda2[i][j] = lambda2[j][i] = lambda2[i][i] + lambda2[j][j] - sqrt(pow(lambda2[i][i]+lambda2[j][j], 2.0)-4.0*lambda2[i][i]*lambda2[j][j]*(1.0-rho*rho));
        }

    for (int i = 0; i < n_features; i++)
        lambda2[i][i] = 0.0;
}

// Gte k nearest neighbors and return distance to last one
void GetKNeighbors(int feature, scalar **lambda2, int n_features, int k, int *nearest, scalar *r)
{
    scalar dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        r[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    for (int i = 0; i < n_features; i++) {
        // Test if is one of the nearest
        if (lambda2[feature][i] < dmin && feature != i) {
            // Look for its place
            int j = 0;
            while (lambda2[feature][i] > r[j])
                j++;

            // Sets in j-th place
            for (int l = k - 1; l > j; l--) {
                nearest[l] = nearest[l - 1];
                r[l] = r[l - 1];
            }
            nearest[j] = i;
            r[j] = lambda2[feature][i];
            dmin = r[k - 1];
        }
    }
}
