/***************************************************************************
 *   Copyright (C) 2009 by Aida de Haro-García                             *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "classifiers.h"

// Kononenko, Igor et al. Overcoming the myopia of inductive learning algorithms with RELIEFF (1997), Applied Intelligence, 7(1), p39-55

void ReliefAlg(const t_Parameters *parm, int *S, int n_feat, scalar **data, int n, int *labels, int classes, int m, int k_n, scalar threshold, scalar relief_alpha, t_RankInfo *ranking)
{
    int i, j, k, l, sel = 0, near = 0, sel_class = 0, *list = NULL, *nearest = NULL;
    scalar local_prob = 0.0, max_val = 0.0, min_val = 0.0, *weight, *hit, *miss, prob[classes], *max = NULL, *min = NULL;

    // Memory allocation
    MALLOC(list, int, n, "relief: ReliefAlg");
    MALLOC(weight, scalar, n_feat, "relief: ReliefAlg");
    MALLOC(hit, scalar, n_feat, "relief: ReliefAlg");
    MALLOC(miss, scalar, n_feat, "relief: ReliefAlg");
    MALLOC(max, scalar, n_feat, "relief: ReliefAlg");
    MALLOC(min, scalar, n_feat, "relief: ReliefAlg");

    // Set m to default value: number of instances
    if (!m)
        m = n;
    // If m is not zero, set iterations to a m% of n
    else  {
        m = n * m/100.0;
    }

    // Set threshold
    // To default value if relief_alpha == 0:  to default value: 0.05
    // Else: Calculate value if relief_alpha != 0
    if (relief_alpha > ZERO)
        threshold = 1/sqrt(relief_alpha * m);

    // Get probability of occurrence of each class
    for (i = 0; i < classes; i++)
        prob[i] = 0.0;
    for (i = 0; i < n; i++)
        prob[labels[i]]++;
    for (i = 0; i < classes; i++)
        prob[i] /= n;

    // Get max and min value of each feature
    for (i = 0; i < n_feat; i++)  {
        max_val = -INFINITE;
        min_val = INFINITE;
        for (j = 0; j < n; j++)  {
            if (data[j][i] > max_val)  {
                max_val = data[j][i];
            }

            if (data[j][i] < min_val)  {
                min_val = data[j][i];
            }
        }

        max[i] = max_val;
        min[i] = min_val;
    }

    // Randomize instances
    for (i = 0; i < n; i++)
        list[i] = i;
    RandomizeVector(list, n);

    for (int i = 0; i < n_feat; i++)
        miss[i] = hit[i] = 0.0;

    // Cannot be more neighbors than instances
    if (k_n > n)
        k_n = n;

    MALLOC(nearest, int, n, "kNNOfOneClass: nearest");
    for (i = 0; i < m; i++)  {
        //Get hits values
        sel = list[i];
        sel_class = labels[sel];
        kNNOfOneClass(data[sel], data, sel_class, n, k_n, n_feat, labels, nearest);
        for (l = 0; l < n_feat; l++) {
            scalar c = 0.0;
            for (k = 0; k < k_n; k++)
                if (nearest[k] != -1)
                    c++;

            for (k = 0; k < k_n; k++)
                if ((near = nearest[k]) != -1)  {
                    if (max[l] != min[l] && m*c > ZERO)
                        hit[l] -= (fabs(data[sel][l] - data[near][l])/(max[l] - min[l]))/(m*c);
                }
        }

        // Get misses values
        for (j = 0; j < sel_class; j++)  {
            kNNOfOneClass(data[sel], data, j, n, k_n, n_feat, labels, nearest);
            local_prob = prob[j]/(1.0 - prob[sel_class]);
            SHOWPROGRESS(parm->show_progress > 1, "Local probability of class %d: %lf\n", j+1, local_prob);
            for (l = 0; l < n_feat; l++) {
                scalar c = 0.0;
                for (k = 0; k < k_n; k++)
                    if (nearest[k] != -1)  {
                        c++;
                    }
                for (k = 0; k < k_n; k++)
                    if ((near = nearest[k]) != -1)  {
                        if (max[l] != min[l] && m*c > ZERO)
                            miss[l] += local_prob * ((fabs(data[sel][l] - data[near][l])/(max[l] - min[l]))/(m*c));
                    }
            }
        }

        for (j = sel_class + 1; j < classes; j++)  {
            kNNOfOneClass(data[sel], data, j, n, k_n, n_feat, labels, nearest);
            local_prob = prob[j]/(1.0 - prob[sel_class]);
            SHOWPROGRESS(parm->show_progress > 1, "Local probability of class %d: %lf\n", j+1, local_prob);
            for (l = 0; l < n_feat; l++) {
                scalar c = 0.0;
                for (k = 0; k < k_n; k++)
                    if (nearest[k] != -1)  {
                        c++;
                    }
                for (k = 0; k < k_n; k++)
                    if ((near = nearest[k]) != -1)  {
                        if (max[l] != min[l] && m*c > ZERO)
                            miss[l] += local_prob * ((fabs(data[sel][l] - data[near][l])/(max[l] - min[l]))/(m*c));
                    }
            }
        }

    }

    int selected = 0;
    for (i = 0; i < n_feat; i++)  {
        weight[i] = miss[i] + hit[i];
        SHOWPROGRESS(false, "Weight (threshold), miss, hit of feature %d: %lf(%lf), %lf, %lf\n", i+1, weight[i], threshold, miss[i], hit[i]);
        if (weight[i] >= threshold) {
            S[i] = 1;
            selected++;
        } else
            S[i] = 0;

        // Return score
        if (ranking) {
            ranking[i].score = weight[i];
            ranking[i].feature = i;

            SHOWPROGRESS(parm->show_progress > 0, "ReliefF Feature %3d score: %lf\n", i+1, ranking[i].score);
        }
    }

    // If none selected, set the maximum value
    if (!selected) {
        int maximum = 0;
        for (int i = 1; i < n_feat; i++)
            if (weight[i] > weight[maximum])
                maximum = i;

        S[maximum] = 1;
    }

    // Free memory
    FREE(list);
    FREE(weight);
    FREE(hit);
    FREE(miss);
    FREE(max);
    FREE(min);
    FREE(nearest);
}
