/***************************************************************************
 *   Copyright (C) 2015 by Nicolás García-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "fs.h"
#include "macros.h"
#include "filter.h"
#include "data.h"

/*
 * M. Dash, Feature selection via set cover, in: Proceedings of IEEE Knowledge and Data Engineering Exchange Workshop, Newport, CA, IEEE Computer Society, 1997,

 * pp. 165–171.
 */

void SetCover(const t_Parameters *parm, int *S, int n_features, scalar **data, int instances, int *labels, int classes)
{
    static int *intervals = NULL;
    static scalar **limits = NULL;

    if (limits == NULL) {
        // Memory allocation for auxiliary data structure
        MALLOC(limits, scalar *, n_features, "FocusAlg: limits");

        // Set limits to NULL
        for (int k = 0; k < n_features; k++)
            limits[k] = NULL;
        MALLOC(intervals, int, n_features, "FocusAlg: intervals");

        // Discretize dataset
        switch (parm->discretizer) {
            case dFAYYAD:
                FayyadAndIraniMDLMethod(data, instances, classes, labels, n_features, limits, intervals, NULL, NULL);
                break;

            case dWPKID:
                WPKIDMethod(data, instances, n_features, limits, intervals, NULL);
                break;

            case dNONE:
                //Do nothing
                break;
        }
    }

    time_t init_time = time(NULL);

    // Inconsistency with all features
    for (int i = 0; i < n_features; i++)
        S[i] = 1;
    scalar delta = InconsistCount(n_features, data, instances, S, labels, classes, limits, intervals);
    if (parm->verbose > 1)
        fprintf(parm->out_file, "Inconsistency with all features: %lf\n", delta);

    // Start with all features unselected
    for (int i = 0; i < n_features; i++)
        S[i] = 0;

    scalar lowest;
    do {
        lowest = 1.0;
        int index = -1;

        for (int i = 0; i < n_features; i++)
            // If not selected add and test
            if (!S[i]) {
                S[i] = 1;

                scalar inconsistency = InconsistCount(n_features, data, instances, S, labels, classes, limits, intervals);

                // If better than all return
                if (inconsistency < delta) {
                    fprintf(parm->out_file, "Best inconsistency: %lf\n", lowest);

                    return;
                }

                if (inconsistency < lowest) {
                    lowest = inconsistency;
                    index = i;
                }

                S[i] = 0;
            }

        // Add feature corresponding to lowest inconsistency
        S[index] = 1;

        if (parm->verbose > 2)
            fprintf(parm->out_file, "Best inconsistency so far: %7.5lf. Added instance: %d\n", lowest, index+1);
    } while (lowest > delta && (!parm->time_limit || time(NULL) - init_time < parm->time_limit));

    if (parm->verbose > 1)
        fprintf(parm->out_file, "Best inconsistency: %lf\n", lowest);
}
