/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "filter.h"

// D. C. Whitley, M. G. Ford, and D. J. Livingstone, "Unsupervised Forward Selection: A Method for Eliminating Redundant Variables", J. Chem. Inf. Comput. Sci. 2000, 40, 1160-1168

void UnsupervisedForwardSelection(const t_Parameters *parm, scalar **data, int inputs, int instances, int *S)
{
    scalar **local_data, *mean;

    // Initialize all features to selected
    for (int i = 0; i < inputs; i++)
        S[i] = 1;

    // Copy data to local
    AllocateScalarMatrix(&local_data, instances, inputs);
    MALLOC(mean, scalar, inputs, "UnsupervisedForwardSelection: mean");
    ZeroVector(mean, NULL, inputs);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < inputs; j++)
            mean[j] += (local_data[i][j] = data[i][j]);

    for (int i = 0; i < inputs; i++)
        mean[i] /= instances;

    // Mean-center columns
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < inputs; j++)
            local_data[i][j] -= mean[j];

    // Reject columns with length < epsilon
    scalar *length;
    MALLOC(length, scalar, inputs, "UnsupervisedForwardSelection: length")
    for (int i = 0; i < inputs; i++) {
        length[i] = ColumnLength(local_data, i, instances);

        if (length[i] < parm->ufs_epsilon) {
            S[i] = 0;
            SHOWPROGRESS(parm->show_progress > 0, "UFS: feature %d REMOVED for small length\n", i+1);
        }
    }

    // Normalize to unit length
    for (int i = 0; i < inputs; i++)
        if (S[i])
            for (int j = 0; j < instances; j++)
                local_data[j][i] /= length[i];

    // Obtain correlation matrix
    scalar **r, smallest = 2.0;
    int alpha = -1, beta = -1;
    AllocateScalarMatrix(&r, inputs, inputs);
    for (int i = 0; i < inputs; i++) {
        r[i][i] = 1.0;
        for (int j = 0; j < i; j++) {
            scalar product = 0.0;
            for (int k = 0; k < instances; k++)
                product += local_data[k][i]*local_data[k][j];

            r[i][j] = r[j][i] = product*product;    // Use r2

            if (product < smallest) {
                smallest = product;
                alpha = i;
                beta = j;
            }
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "UFS: features alpha, beta: %d, %d\n", alpha+1, beta+1);

    // Select columns alpha, beta with the smallest correlation
    // Done in previous step

    // Remove columns whose r2ij with alpha or beta excess R2max
    for (int i = 0; i < inputs; i++)
        if (S[i] && i != alpha && i != beta && (r[i][alpha] > parm->ufs_R2max || r[i][beta] > parm->ufs_R2max)) {
            S[i] = 0;
            SHOWPROGRESS(parm->show_progress > 0, "UFS: feature %d REMOVED for r2ij with alpha, beta exceeding R2max\n", i+1);
        }

    // Choose orthonormal bases {c1, c2}
    scalar **c;
    MALLOC(c, scalar *, inputs, "UnsupervisedForwardSelection: c");
    MALLOC(c[0], scalar, instances, "UnsupervisedForwardSelection: c1");
    MALLOC(c[1], scalar, instances, "UnsupervisedForwardSelection: c2");

    for (int i = 0; i < instances; i++)
        c[0][i] = local_data[i][alpha];

    scalar dot = 0.0;
    for (int i = 0; i < instances; i++)
        dot += local_data[i][alpha] * local_data[i][beta];

    scalar module = 0.0;
    for (int i = 0; i < instances; i++) {
        c[1][i] = local_data[i][beta] - dot*c[0][i];
        module += c[1][i]*c[1][i];
    }

    module = sqrt(module);
    for (int i = 0; i < instances; i++)
        c[1][i] /= module;

    // Start with alpha, beta selected
    // Remaining not previously selectionables possible selected
    int *selected, *selectionables;
    MALLOC(selected, int, inputs, "UnsupervisedForwardSelection: selected");
    MALLOC(selectionables, int, inputs, "UnsupervisedForwardSelection: selectionables");

    int l = 2, selectionable = 0;
    selected[0] = alpha;
    selected[1] = beta;
    for (int i = 0; i < inputs; i++)
        if (S[i] && i != alpha && i != beta) {
            selectionables[selectionable] = i;
            selectionable++;
        }

    // Repeat until no more selectionables
    while (selectionable) {
        // Obtain R2 with the selected columns
        scalar smallest_v = 2.0;
        int smallest_f = -1;

        for (int j = 0; j < selectionable; j++) {
            scalar R2 = LengthOrtogonalProjection(local_data, selectionables[j], c, l, instances);

            // Remove?
            if (R2 > parm->ufs_R2max) {
                SHOWPROGRESS(parm->show_progress > 0, "UFS: feature %d REMOVED for r2ij with current basis exceeding R2max\n", selectionables[j]+1);
                selectionables[j] = selectionables[selectionable-1];
                selectionable--;
                j--;
            } else {
                // Smallest?
                if (R2 < smallest_v) {
                    smallest_v = R2;
                    smallest_f = selectionables[j];
                }
            }
        }

        // Add new feature to basis?
        if (smallest_f != -1) {
            SHOWPROGRESS(parm->show_progress > 0, "UFS: feature %d ADDED for r2ij with current basis exceeding R2max\n", smallest_f+1);

            scalar dot[l];
            for (int i = 0; i < l; i++)
                dot[i] = ColumnVectorDotProduct(local_data, smallest_f, c[i], instances);

            // New member of the basis
            MALLOC(c[l], scalar, instances, "UnsupervisedForwardSelection: c[l]");

            scalar module = 0.0;
            for (int i = 0; i < instances; i++) {
                c[l][i] = local_data[i][smallest_f];
                for (int k = 0; k < l; k++)
                    c[l][i] -= dot[k]*c[k][i];
                module += c[l][i]*c[l][i];
            }

            module = sqrt(module);
            for (int i = 0; i < instances; i++)
                c[l][i] /= module;

            selected[l] = smallest_f;
            l++;

            // Do not choose again
            selectionables[smallest_f] = selectionables[selectionable-1];
            selectionable--;
        }
    }

    // Set final selection
    ZeroVector(NULL, S, inputs);
    for (int i = 0; i < l; i++)
        S[selected[i]] = 1;

    // Free memory
    FREE(mean);
    FREE(length);
    FreeScalarMatrix(&local_data, instances);
    FreeScalarMatrix(&r, inputs);
    for (int i = 0; i < l; i++)
        FREE(c[i]);
    FREE(c);
    FREE(selected);
    FREE(selectionables);
}

scalar ColumnVectorDotProduct(scalar **m, int column, scalar *v, int n)
{
    scalar dot = 0.0;
    for (int i = 0; i < n; i++)
        dot += m[i][column] * v[i];

    return dot;
}

scalar ColumnLength(scalar **m, int column, int n)
{
    scalar length = 0.0;
    for (int j = 0; j < n; j++)
        length += m[j][column]*m[j][column];

    return sqrt(length);
}

scalar LengthOrtogonalProjection(scalar **X, int j, scalar **c, int l, int n)
{
    scalar dot[l];
    for (int i = 0; i < l; i++)
        dot[i] = ColumnVectorDotProduct(X, j, c[i], n);

    scalar length = 0.0;
    for (int i = 0; i < n; i++) {
        scalar element = 0.0;
        for (int k = 0; k < l; k++)
            element += dot[k]*c[k][i];

        length += element*element;
    }

    return sqrt(length);
}
