/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "svm.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "is.h"

// eFloatBoost method: subspace. Z. Li and Zh. Zhang, "eFloatBoost learning and statistical face recognition", IEEE TPAMI vol. 26, no. 9, pp. 1-12, 2004
scalar UpdateeFloatBoostWeightsFS(const t_Parameters *parm, int **subspaces, scalar **data, int instances, int *labels, scalar *w, scalar *alpha, int inputs, int classes, int size, bool cv,
                                 t_Learner **learners, int minority)
{
    scalar e_i, sum, **u_data = NULL;
    int u_instances = 0, *u_labels = NULL;
    bool ok;

    if (parm->weigthed_sampling_for_training)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted sampling for training");

    if (parm->random_undersampling_for_training && parm->boostfs_classifier == kNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Random under-sampling for training");

    // Train new classifier
    ClassifierTraining(parm, parm->boostfs_classifier, data, instances, labels, inputs, classes, NULL, subspaces[size-1], NULL, true, learners[size-1], NULL);

    // Update weights using w(i) = exp (-margin(x_i))
    e_i = sum = 0.0;
    for (int k = 0; k < instances; k++) {
        // Get margin by last "r" classifiers
        scalar margin = GetEnsembleInstanceMargin(parm, parm->boostfs_classifier, data[k], labels[k], u_data, u_labels, subspaces, u_instances, parm->k, classes, inputs, learners, size, alpha, &ok, 0);

        sum += (w[k] = exp(-margin));
        if (!ok)
            e_i ++;
    }

    e_i /= instances;

    // Weight classifier
    // alpha[size-1] = 1.0;  Not needed, set in main function

    // Normalize w[] to sum 1
    for (int k = 0; k < instances; k++)
        w[k] *= 1.0/sum;

    if (parm->random_undersampling_for_training && parm->boostfs_classifier == kNN) {
        FREE(u_data);
        FREE(u_labels);
    }

    return e_i;
}

#define MAX_ATTEMPS 30   // Maximum number of attempts to construct the ensemble

// Returns removed members from the ensemble
int ConditionalExclusionFS(t_Parameters *parm, int **subspaces, int size, scalar **data, int instances, int inputs, int k, int *labels, int classes, scalar *w, scalar *alpha, scalar *epsilon,
                           t_Learner **learners, int minority)
{
    int h_prime = -1, M;
    scalar best_error;
    bool removed;
    static int attempts = 0;

    M = size;

    // Cannot remove if only one
    attempts++;
    if (M == 1 || attempts > MAX_ATTEMPS) {
        if (attempts > MAX_ATTEMPS) {
            fprintf(parm->out_file, "Maximum attempts reached. Final ensemble size: %d\n", size);
            parm->t = size;
        }
        return 0;
    }

    do {
        removed = false;

        // Get the member that less contribute to the classification accuracy
        best_error = INFINITE;
        for (int i = 0; i < M; i++) {
            scalar backup = alpha[i];
            alpha[i] = 0.;
            scalar error = 1.0 - GetEnsembleAccuracy(parm, parm->boostfs_classifier, data, labels, subspaces, instances, k, classes, inputs, learners, M, alpha, NULL, 0, parm->use_g_measure,
                           parm->sampling_size, minority);
            alpha[i] = backup;

            if (best_error > error) {
                best_error = error;
                h_prime = i;
            }
        }

        // Remove?
        if (best_error < epsilon[M-1]) {
            // Remove h prime from ensemble
            int *backup = subspaces[h_prime];
            subspaces[h_prime] = subspaces[M-1];
            subspaces[M-1] = backup;
            alpha[h_prime] = alpha[M-1];
            epsilon[M-1] = best_error;
            M--;

            removed = true;
        } else {// Update weight, if needed, and return
            if (size != M)
                UpdateeFloatBoostWeightsFS(parm, subspaces, data, instances, labels, w, alpha, inputs, classes, size, parm->cv_hyperparameters, learners, minority);
        }
    } while (removed && M > 1);

    return size - M;
}
