/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program minor FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program minor distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <search.h>
#include <string.h>
#include <unistd.h>
#include "typex.h"
#include "macros.h"
#include "common.h"
#include "isfs.h"
#include "ensemble.h"
#include "classifiers.h"
#include "maths.h"

// Get best subspace according to give measure

void GetBestSubspace(const t_Parameters *parm, scalar **t_data, int t_instances, int *t_labels, int inputs, int classes, scalar *votes, int *subspace, scalar max_votes, int **index, int minority)
{
    int instances, *labels, **s;
    scalar **data, gap, *coefficient, *thresholds = NULL;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Getting bets subspace for sparse format");

    switch (parm->feature_subset_combination) {
        case Union:
            // Select all features with at least one vote
            for (int i = 0; i < inputs; i++)
                subspace[i] = votes[i] > 0.0;
            break;

        case Intersection:
            // Features selected all the time
            for (int i = 0; i < inputs; i++)
                subspace[i] = fabs(votes[i]-max_votes) <= ZERO;
            break;

        case Majority:
            // At least half the times
            for (int i = 0; i < inputs; i++)
                subspace[i] = votes[i] >= max_votes/2;
            break;

        case Thresholding:
            // Estimate best threshold using classification performance and reduction

            // Subsample for evaluating performance?
            if (parm->subsampling_for_evaluation) {
                instances = t_instances;
                AllocateScalarMatrix(&data, t_instances, inputs);
                MALLOC(labels, int, t_instances, "GetBestThreshold: labels");

                CopyScalarMatrix(t_data, data, instances, inputs);
                CopyIntegerVector(t_labels, labels, instances);

                StratifiedRandomSubsampling(&data, &instances, &labels, parm->random_undersampling_ratio, inputs, classes);
            } else {
                data = t_data;
                instances = t_instances;
                labels = t_labels;
            }

            // Possible thresholds
            int *tosort;
            int CompareInt(const void *, const void *);

            if (parm->n_thresholds == 0)
                ((t_Parameters *) parm)->n_thresholds = inputs;
            else
                // Cannot be more thresholds than inputs
                ((t_Parameters *) parm)->n_thresholds = MIN(parm->n_thresholds, inputs);
            MALLOC(thresholds, scalar, parm->n_thresholds, "main: thresholds");
            MALLOC(tosort, int, inputs, "main: tosort");

            // Sort total votes and get thresholds
            for (int i = 0; i < inputs; i++)
                tosort[i] = votes[i];
            qsort(tosort, inputs, sizeof(int), CompareInt);

            // Ignore repeated values
            int n_thresholds = inputs;
            CompactVector(NULL, tosort, &n_thresholds);
            n_thresholds--; // One less than different values

            // If more than allowed get a few
            if (n_thresholds > parm->n_thresholds) {
                gap = (tosort[n_thresholds] - tosort[0])/n_thresholds;
                thresholds[0] = tosort[0]+gap;
                for (int j = 0; j < n_thresholds; j++)
                    thresholds[j] = thresholds[j-1]+gap;
            } else
                for (int i = 0; i < n_thresholds; i++)
                    thresholds[i] = (tosort[i]+tosort[i+1])/2.0;

            // Training error for every number of votes
            MALLOC(coefficient, scalar, n_thresholds, "main: coef")

            // Store each subspace to avoid repeated evaluation of the same subspaces
            AllocateIntegerMatrix(&s, n_thresholds, inputs);

            for (int j = 0; j < n_thresholds; j++) {
                // Count votes
                int selected = 0;
                for (int i = 0; i < inputs; i++)
                    if (votes[i] >= thresholds[j]) {
                        selected++;
                        s[j][i] = 1;
                    } else
                        s[j][i] = 0;

                // At least some instances
                if (selected > 0) {
                    // Check is previously evaluated
                    int same = GetSameSubspace(s[j], s, j, inputs);
                    if (same == -1)
                        coefficient[j] = EvaluateSubspaceWithCriterion(parm, data, labels, instances, inputs, minority, classes, parm->feature_subset_evaluation, NULL, s[j]);
                    else
                        coefficient[j] = coefficient[same];
                } else
                    coefficient[j] = 0.0;
            }

            if (parm->subsampling_for_evaluation) {
                FreeScalarMatrix(&data, t_instances);
                FREE(labels);
            }

            // Get best coefficient
            int best = 0;
            for (int i = 1; i < n_thresholds; i++)
                if (coefficient[i] > coefficient[best])
                    best = i;

            // Count votes
            int selected = 0;
            for (int i = 0; i < inputs; i++)
                if (votes[i] >= thresholds[best]) {
                    subspace[i] = 1;
                    selected++;
                } else
                    subspace[i] = 0;


            fprintf(parm->out_file, "Selected inputs: %d\nVotes:\t ", selected);
            for (int i = 0; i < inputs; i++)
                fprintf(parm->out_file, "%d ", (int) votes[i]);
            fprintf(parm->out_file, "\n");

            FREE(coefficient);
            FREE(thresholds);
            FREE(tosort);
            FreeIntegerMatrix(&s, n_thresholds);
            break;
    }
}


#define SIGMOID(x, y)        (tanh(10.*exp(-.025*(y))*(x))/tanh(10.*exp(-.025*(y))))

// Evaluates a threshold for votes of simultaneous features selection using the same method that for democratic selection

scalar EvaluateSubspaceWithCriterion(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, int minority, int classes, t_FeatureSubsetEvaluation evaluation_method, int *S,
                                     int *subspace)
{
    int **M;
    scalar ok = 0.0, margin = 0.0, voor, minimum_margin, r = 0.0, n2, mfdr, mfe, cur_mfe, coefficient = 0.0;
    t_Learner learner;
    t_Ensemble ensemble;
    t_PerformanceInfo perf;

    // Reduction
    for (int i = 0; i < inputs; i++)
        if (subspace[i])
            r++;
    r = 1.0 - r/inputs;

    AllocateIntegerMatrix(&M, classes, classes);

    // Performance
    switch (evaluation_method) {
        case ClassModel:
            // Evaluate for every subset
            ZeroMatrix(NULL, M, classes, classes);

            // Validate?
            if (parm->cv_hyperparameters)
                SetCrossValidatedParameters((t_Parameters *) parm, data, instances, labels, S, inputs, classes, parm->votes_evaluator_classifier, minority, NULL, NULL);

            // Train classifier
            ClassifierTraining(parm, parm->votes_evaluator_classifier, data, instances, labels, inputs, classes, NULL, subspace, S, true, &learner, &ensemble);

            if (parm->performance_measure != pfMargin)
                for (int i = 0; i < instances; i++)
                    M[labels[i]][ClassifierOutput(parm, parm->votes_evaluator_classifier, data[i], inputs, classes, NULL, subspace, data, instances, labels, S, &learner, labels[i], NULL, -1,
                                                        NULL, &ensemble)]++;
            else
                margin = ClassifierDatasetMargin(parm, parm->votes_evaluator_classifier, data, instances, labels, data, instances, labels, inputs, classes, subspace, &minimum_margin,
                                                       &learner, NULL, S, &ensemble, !IsEnsemble(parm->votes_evaluator_classifier));

            ClassifierFree(parm, &learner, parm->votes_evaluator_classifier, classes, inputs, NULL);

            PrintMulticlassMeasures(M, classes, instances, NULL, NULL, &perf);
            if (classes == 2) {
                int majority = (minority+1)%2;
                PrintTwoClassMeasures(M[minority][minority], M[majority][majority], M[majority][minority], M[minority][majority], NULL, NULL, &perf);
            }

            switch (parm->performance_measure) {
                case pfGMean:
                    ok = perf.gmean;
                    break;

                case pfAccuracy:
                    ok = perf.accuracy;
                    break;

                case pfauROC:
                case pfauPRC:
                    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation metric for thresholds");
                    break;

                case pfCC:
                    ok = perf.cc;
                    break;

                case pfF1Macro:
                    ok = perf.f1_score_macro;
                    break;

                case pfPrecisionMacro:
                    ok = perf.precision_macro;
                    break;

                case pfF1:
                    ok = perf.f1_score;
                    break;

                case pfKappa:
                    ok = perf.kappa;
                    break;

                case pfRecallMacro:
                    ok = perf.recall_macro;
                    break;

                case pfRecallMacroGMean:
                    ok = perf.recall_macro_g_mean;
                    break;

                case pfMargin:
                    ok = margin;
                    break;

                case pfAverageAccuracy:
                    ok = perf.average_accuracy;
                    break;
            }

            coefficient = parm->alpha_ok * ok + parm->alpha_red * r;
            break;

        case Mfdr:
            ok = MaximumFishersDiscriminantRatio(data, labels, instances, inputs, classes, S, NULL);
            coefficient = parm->alpha_ok * ok + parm->alpha_red * r;
            break;

        case Voor:
            voor = 1.0;

            for (int i = 1; i < classes; i++)
                for (int j = 0; j < i; j++)
                    // Sum instead of product following authors' recommendation
                    voor *= fabs(VolumeOfOverlapRegion(data, instances, labels, i, j, inputs, S, NULL));

            // Must be minimized
            ok = voor;
            coefficient = parm->alpha_ok * (-ok) + parm->alpha_red * r;
            break;

        case Loor:
            voor = 0.0;

            for (int i = 1; i < classes; i++)
                for (int j = 0; j < i; j++)
                    // Sum instead of product following authors' recommendation
                    voor += fabs(VolumeOfOverlapRegion(data, instances, labels, i, j, inputs, S, NULL));

            // Must be minimized
            ok = voor;
            coefficient = parm->alpha_ok * (-ok) + parm->alpha_red * r;
            break;

        case Riic:
            n2 = RIICComplexityMeasure(data, instances, labels, inputs, NULL, S);

            // Must be minimized
            ok = n2;
            coefficient = parm->alpha_ok * (-ok) + parm->alpha_red * r;
            break;

        case Mfe:
            ok = -1.0;

            for (int i = 1; i < classes; i++)
                for (int j = 0; j < i; j++)
                    if ((mfe = fabs(MaximumFeatureEfficiency(data, instances, labels, i, j, S, inputs, NULL))) > ok)
                        ok = mfe;
            coefficient = parm->alpha_ok * ok + parm->alpha_red * r;
            break;

        case Cf:
            mfdr = MaximumFishersDiscriminantRatio(data, labels, instances, inputs, classes, S, NULL);

            voor = 1.0;
            mfe = -1.0;

            for (int i = 1; i < classes; i++)
                for (int j = 0; j < i; j++) {
                    // Sum instead of product following authors' recommendation
                    voor *= fabs(VolumeOfOverlapRegion(data, instances, labels, i, j, inputs, S, NULL));
                    if ((cur_mfe = MaximumFeatureEfficiency(data, instances, labels, i, j, S, inputs, NULL)) > mfe)
                        mfe = cur_mfe;
                }

            ok = (mfe + (1.0-voor) + mfdr)/3.0;
            coefficient = parm->alpha_ok * ok + parm->alpha_red * r;
            break;
    }

    FreeIntegerMatrix(&M, classes);

    return coefficient;
}
