/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "common.h"
#include "macros.h"
#include "filter.h"
#include "genAlg.h"
#include "wrapper.h"
#include "ranking.h"
#include "fs.h"
#include "isfs.h"
#include "maths.h"

void FeatureSelection(const t_Parameters *parm, int *subspace, int inputs, scalar **train_data, int *train_labels, int classes, int instances, int minority, int **train_index)
{
    scalar **data;
    int *labels, **index = NULL;

    // Run FS with not partitions
    if (!parm->feature_partition && !parm->instance_partition) {
        // Perform selection
        FeatureSelectionAlgorithm(parm, subspace, inputs, train_data, train_labels, classes, instances, minority, train_index);
        return;
    } else {
        scalar *votes, max_votes = 0;
        int *S_local, *partition;

        MALLOC(votes, scalar, inputs, "FeatureSelection: votes");
        MALLOC(S_local, int, inputs, "FeatureSelection: S_local");

        ZeroVector(votes, NULL, inputs);

        if (parm->instance_partition) {    // Partition by rows
            if (parm->subset_size == 0)
                OnErrorExit(ERR_PARAM, "Subset size for partition by instances");

            MALLOC(partition, int, instances, "FeatureSelection: partition");

            int n_subsets = instances/parm->subset_size;

            // Add remainder
            if ((instances % parm->subset_size) || n_subsets <= 1)
                n_subsets = MAX(n_subsets + 1, 2);

            // Democracy: n_electors > 1
            // Standard partitioning: n_electors = 1

            for (int iteration = 0; iteration < parm->n_electors; iteration++) {
                int count[n_subsets];

                // Partition method?
                SHOWPROGRESS(parm->verbose >= 0, "PARTITIONING for training (%2d of %2d)\n", iteration + 1, parm->n_electors);
                // Random partition or stratified partition
                if (parm->keep_class_distribution)
                    RandomPartitionKCDFS(partition, count, instances, n_subsets, train_labels, classes);
                else
                    RandomPartition(partition, count, instances, n_subsets);

                // Local data
                int subset_size = MaximumOfIntegerVector(count, n_subsets);
                MALLOC(data, scalar *, subset_size, "FeatureSelection: data");
                MALLOC(labels, int, subset_size, "FeatureSelection: labels");
                if (train_index)
                    MALLOC(index, int *, subset_size, "FeatureSelection: index");

                // For every subset
                for (int subset = 0; subset < n_subsets; subset++) {
                    // Select instances in the subset
                    int cursor = 0;
                    for (int i = 0; i < instances; i++)
                        if (partition[i] == subset) {
                            data[cursor] = train_data[i];
                            labels[cursor] = train_labels[i];
                            if (train_index)
                                index[cursor] = train_index[i];
                            cursor++;
                        }
                    // Empty subset? skip
                    if (cursor == 0)
                        continue;

                    SHOWPROGRESS(parm->show_progress > 0, "BEGGINING subset %2d/%d ITERATION %2d/%d\n", subset+1, n_subsets, iteration+1, parm->n_electors);

                    FeatureSelectionAlgorithm(parm, S_local, inputs, data, labels, classes, cursor, minority, index);

                    // Store votes
                    for (int i = 0; i < inputs; i++)
                        if (!S_local[i])
                            votes[i]++;

                    max_votes++;

                    SHOWPROGRESS(parm->verbose >= 0, "Feature selection %d/%d ", subset + 1, n_subsets);
                }

                SHOWPROGRESS(parm->show_progress > 0, "\nFINISHED train iteration  %d/%d\n", iteration + 1, parm->n_electors);

                FREE(data);
                FREE(labels);
                if (train_index)
                    FREE(index);
            }

            FREE(partition);

        } else {    // Partition by columns
            if (parm->subspace_size == 0)
                OnErrorExit(ERR_PARAM, "Subset size for partition by instances");

            if (train_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for subspace partition");

            AllocateScalarMatrix(&data, instances, parm->subspace_size);

            int n_subspaces = inputs/parm->subspace_size;
            if (inputs % parm->subspace_size)
                n_subspaces++;

            // If subspaces 2 or less, make them equal size and at least two
            if (n_subspaces <= 2) {
                n_subspaces = 2;
                ((t_Parameters *)parm)->subspace_size = (inputs % 2) ? inputs/2 + 1 : inputs/2;

                fprintf(parm->out_file, "Subspace size modified to: %d\n", parm->subspace_size);
            }

            for (int iteration = 0; iteration < parm->n_electors; iteration++) {
                SHOWPROGRESS(parm->show_progress > 0, "PARTITIONING for training (%2d of %2d)\n", iteration + 1, parm->n_electors);

                int list[inputs];
                for (int i = 0; i < inputs; i++)
                    list[i] = i;
                RandomizeVector(list, inputs);

                // For every subset
                int cursor = 0;

                max_votes++;

                for (int subspaces = 0; subspaces < n_subspaces; subspaces++) {
                    SHOWPROGRESS(parm->show_progress > 0, "BEGGINING subset %2d/%d ITERATION %2d/%d\n", subspaces+1, n_subspaces, iteration+1, parm->n_electors);

                    // Select features in the subsets
                    int size;
                    for (size = 0; size < parm->subspace_size && cursor < inputs; size++, cursor++)
                        for (int j = 0; j < instances; j++)
                            data[j][size] = train_data[j][list[cursor]];

                    FeatureSelectionAlgorithm(parm, S_local, size, data, train_labels, classes, instances, minority, train_index);

                    // Store votes
                    for (int i = 0; i < size; i++)
                        if (!S_local[i])
                            votes[list[cursor-size+i]]++;

                    SHOWPROGRESS(parm->verbose >= 0, "Feature selection %d/%d ", subspaces + 1, n_subspaces);
                    SHOWPROGRESS(parm->show_progress > 0, "FINISHED subset %2d/%d ITERATION %2d/%d\n", subspaces+1, n_subspaces, iteration+1, parm->n_electors);
                }

                SHOWPROGRESS(parm->show_progress > 0, "\nFINISHED train iteration  %d/%d\n", iteration + 1, parm->n_electors);
            }

            FreeScalarMatrix(&data, instances);
        }

        // Get best subspace
        GetBestSubspace(parm, train_data, instances, train_labels, inputs, classes, votes, subspace, max_votes, train_index, minority);

        FREE(votes);
        FREE(S_local);
    }
}

void FeatureSelectionAlgorithm(const t_Parameters *parm, int *subspace, int inputs, scalar **original_data, int *labels, int classes, int instances, int minority, int **index)
{
    int *s, generations;
    scalar evaluations;
    bool fs, is;
    t_RankInfo *ranking;
    extern char *fsalgorithm[];

    if (!parm->feature_selection) {
        for (int i = 0; i < inputs; i++)
            subspace[i] = 1;
        return;
    }

    SHOWPROGRESS(parm->show_progress >= 0, "Feature selection algorithm: %s", fsalgorithm[parm->fsalgorithm]);

    switch (parm->fsalgorithm) {
        case fsLVF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "LVF for sparse format");
            LvfAlg(parm, subspace, inputs, original_data, instances, labels, classes, parm->lvf_iter);
            break;

        case fsFOCUS:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FOCUS for sparse format");
            FocusAlg(parm, subspace, inputs, original_data, instances, labels, classes, parm->time_limit);
            break;

        case fsGA:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FSGA for sparse format");
            GeneticAlgorithm(parm, subspace, parm->individuals, parm->elite, inputs, parm->generations, parm->init_setprob, parm->mutation, parm->mutation, parm->ga_alpha, original_data, instances, labels,
                             classes, minority);
            break;

        case fsRELIEF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ReliefF for sparse format");
            ReliefAlg(parm, subspace, inputs, original_data, instances, labels, classes, parm->m, parm->relief_k, parm->relief_threshold, parm->relief_alpha, NULL);
            break;

        case fsRANDOM:
            RandomFeatureSelection(subspace, inputs, parm->selection_probability);
            break;

        case fsABB:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ABB for sparse format");
            for (int i = 0; i < inputs; i++)
                subspace[i] = 1;
            AutomaticBranchAndBound(original_data, instances, labels, subspace, inputs, parm->time_limit);
            break;

        case fsBB:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "BB for sparse format");
            BranchAndBound(original_data, instances, labels, subspace, inputs, parm->time_limit);
            break;

        case fsFCBF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FCBF for sparse format");
            FastCorrelationBasedFilter(parm, original_data, inputs, instances, classes, labels, subspace, parm->fcbf_delta);
            break;

        case fsFOCUSSF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FOCUS-SF for sparse format");
            FocusSF(parm, subspace, inputs, original_data, instances, labels, classes);
            break;

        case fsFAST:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FAST for sparse format");
            FastclusteringBasedFeatureSubsetSelection(parm, original_data, inputs, instances, classes, labels, subspace, parm->fast_theta);
            break;

        case fsCFSSF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CSF-FS for sparse format");
            CorrelationBasedFilterSequentialForwardSelection(parm, original_data, inputs, instances, classes, labels, subspace);
            break;

        case fsSetCover:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "SetCover for sparse format");
            SetCover(parm, subspace, inputs, original_data, instances, labels, classes);
            break;

        case fsQBB:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "QBB for sparse format");
            QuickBranchAndBound(parm, subspace, inputs, original_data, instances, labels, classes);
            break;

        case fsSFFS:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "SFFS for sparse format");
            SequentialForwardFloatingSelection(parm, original_data, inputs, instances, classes, labels, subspace);
            break;

        case fsSBFS:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "SBFS for sparse format");
            SequentialBackwardFloatingSelection(parm, original_data, inputs, instances, classes, labels, subspace);
            break;

        case fsMICIBM:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "MICIBM for sparse format");
            // Check k value
            MaximalInformationCompressionIndexBasedMethod(&(((t_Parameters *)parm)->micibm_k), original_data, instances, inputs, subspace);
            break;

        case fsRFE:
            OnErrorExit(ERR_PARAM, "RFE is a instance weighting method");
            break;

        case fsCLASSIFIER:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Classifier feature selection");
            break;

        case fsNONE:
            for (int i = 0; i < inputs; i++)
                subspace[i] = 1;
            break;

        case fsCHC:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CHCFS for sparse format");
            fs = parm->feature_selection;
            is = parm->instance_selection;
            ((t_Parameters *) parm)->feature_selection = true;
            ((t_Parameters *) parm)->instance_selection = false;
            MALLOC(s, int, instances+inputs, "FeatureSelectionAlgorithm: s")

            CHCISplusFS(parm, s, original_data, instances, inputs, labels, NULL, true, &generations, &evaluations, classes, NULL, 0, NULL, NULL, NULL, NULL);

            for (int i = 0; i < inputs; i++)
                subspace[i] = s[instances+i];
            FREE(s);

            ((t_Parameters *) parm)->instance_selection = is;
            ((t_Parameters *) parm)->feature_selection = fs;
            break;

        case fsRanking:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Feature Ranking for sparse format");
            // Ranking algorithm + relevant threshold method
            MALLOC(ranking, t_RankInfo, inputs, "FeatureSelectionAlgorithm: ranking");
            FeatureRankingAlgorithm(parm, parm->fralgorithm, inputs, original_data, labels, classes, instances, minority, ranking, NULL, NULL, index);

            RankingToFeatureSubsetSelection(parm, original_data, labels, instances, inputs, classes, ranking, subspace);

            FREE(ranking);
            break;

        case fsUFS:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "UFS for sparse format");
            UnsupervisedForwardSelection(parm, original_data, inputs, instances, subspace);
            break;

        case fsKSCBF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "KSCBF for sparse format");
            KolmogorovSmirnovCorrelationBasedFilter(parm, original_data, inputs, instances, classes, labels, subspace);
            break;

        case fsMABUSE:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "MABUSE for sparse format");
            MarginBasedFeatureSelection(parm, subspace, original_data, instances, labels, inputs, classes, minority, false, index);
            break;

        case fsGFlip:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "G-flip for sparse format");
            GreedyFeatureFlip(parm, original_data, inputs, instances, labels, subspace, parm->gflip_max_iterations);
            break;

        case fsBMIFS:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "BMIFS for sparse format");
            BoostingMutualInformationFeatureSelection(parm, inputs, original_data, labels, classes, instances, subspace);
            break;

        case fswMI:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "wMI for sparse format");
            WeightedMutualInformationFS(parm, inputs, original_data, labels, classes, instances, subspace);
            break;

        case fsiSimba:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "wMI for sparse format");
            ImprovedIterativeSearchMarginBasedAlgorithm(parm, original_data, inputs, instances, labels, NULL, subspace, classes);
            break;

        case fsmRMR:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "mRMR for sparse format");
            mRMR(parm, original_data, instances, labels, classes, subspace, inputs, minority);
            break;
    }

    if (parm->verbose > 1) {
        fprintf(parm->out_file, "Selected features (%s): ", fsalgorithm[(int) parm->fsalgorithm]);
        if (parm->duplicate_features) {
            // Print info about duplicated features
            for (int i = 0; i < inputs/2; i++)
                if (subspace[i])
                    fprintf(parm->out_file, "%d ", i+1);

            for (int i = inputs/2; i < inputs; i++)
                if (subspace[i])
                    // Duplicated?
                    fprintf(parm->out_file, "%s%d ", subspace[i-inputs/2] ? "»" : "", subspace[i-inputs/2] ? i-inputs/2+1 : i+1);
        } else
            for (int i = 0; i < inputs; i++)
                if (subspace[i])
                    fprintf(parm->out_file, "%d ", i+1);

        fprintf(parm->out_file, "\n");
    }

    SHOWPROGRESS(parm->show_progress >= 0, " ---> @ok%c", '\n');
}
