/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "macros.h"
#include "fs.h"
#include "is.h"
#include "isfs.h"
#include "filter.h"
#include "classifiers.h"

// This mutation is based on feature selection algorithms.
// Now implements: Focus

// FOCUS mutation is based on focus algorithm, but instead of using inconsistency uses LOO error
int FeatureSelectionMutation(const t_Parameters *parm, int *individual, int instances, int *label, scalar *w, scalar **data, int inputs, t_FSAlgorithm method, scalar *evaluations, int classes,
                             bool ls_only_use_selected)
{
    scalar **l_data;
    int selected_inputs = 0, selected_instances = 0, *S, *l_label, *l_individual, l_instances;

    for (int i = 0; i < inputs; i++)
        if (individual[instances+i])
            selected_inputs++;

    // If no inputs, nothing to do
    if (selected_inputs == 0)
        return 0;

    // Create local data using selected features
    MALLOC(l_data, scalar *, instances, "FeatureSelectionMutation: l_data");
    ALLOCA(S, int, inputs, "FeatureSelectionMutation: S");

    // Modified 4/January/2011
    // Both possibilities are used
    if (ls_only_use_selected) {
        ALLOCA(l_label, int, instances, "FeatureSelectionMutation: l_label");
        ALLOCA(l_individual, int, instances, "FeatureSelectionMutation: l_individual");

        // Only selected individuals are used
        for (int i = 0; i < instances; i++)
            if (individual[i]) {
                MALLOC(l_data[selected_instances], scalar, selected_inputs, "FeatureSelectionMutation: l_data[selected_instances]");
                for (int j = 0, k = 0; j < inputs; j++)
                    if (individual[instances+j]) {
                        l_data[selected_instances][k] = data[i][j];
                        k++;
                    }
                l_label[selected_instances] = label[i];
                l_individual[selected_instances] = 1;

                selected_instances++;
            }

        l_instances = selected_instances;
    } else {
        l_label = label;
        l_individual = individual;

        // All individuals are used
        for (int i = 0; i < instances; i++) {
            if (individual[i])
                selected_instances++;

            MALLOC(l_data[i], scalar, selected_inputs, "FeatureSelectionMutation: l_data[i]");
            for (int j = 0, k = 0; j < inputs; j++)
                if (individual[instances+j]) {
                    l_data[i][k] = data[i][j];
                    k++;
                }
        }
        l_instances = instances;
    }

    // If no instances, nothing to do
    if (selected_instances == 0) {
        for (int i = 0; i < l_instances; i++)
            FREE(l_data[i]);
        FREE(l_data);
        return 0;
    }

    int selected_by = 0;
    FeatureSelectionAlgorithm(parm, S, selected_inputs, l_data, l_label, classes, l_instances, -1, NULL);

    // Restore individual
    for (int i = 0, j = 0; i < inputs; i++)
        if (individual[instances+i]) {
            individual[instances+i] = S[j];
            j++;
        }

    for (int i = 0; i < l_instances; i++)
        FREE(l_data[i]);
    FREE(l_data);

    return selected_inputs - selected_by;
}

int FocusMutation(const t_Parameters *parm, int *S, int n_feat, scalar **data, int n, int *labels, int outputs, int *S_instances, int inputs, scalar *evaluations, int classes,
                  int k, int max_features_for_focus)
{
    int size = 0, n_sel = 0, cont = 0, num = 0, *sel_feat = NULL, backup = 0;
    scalar Ok_with_all = 0.0, Ok = 0.0, fitness = 0.0, reduction, individual_evaluations = 0.0;

    // Nothing to do if just one feature
    if (n_feat == 1)
        return 1;

    // Memory allocation for auxiliary data structure
    ALLOCA(sel_feat, int, n_feat, "FocusMutation: sel_feat");

    // Get inconsistency value when all features are selected
    for (int j = 0; j < n_feat; j++)  {
        sel_feat[j] = 1;
        S[j] = 1;
    }

    int nearest[k];
    for (int i = 0; i < n; i++) {
        backup = S_instances[i];
        S_instances[i] = 0;

        if (kNNCategory(data[i], data, S, n, k, n_feat, labels, classes, nearest, S_instances, NULL, NULL, -1, NULL, NULL) == labels[i])
            Ok_with_all++;

        S_instances[i] = backup;
    }

    individual_evaluations += n * n * n_feat;

    Ok_with_all /= n;
    scalar fitness_with_all = Fitness(parm, Ok_with_all, n_feat, inputs, n, n, &reduction);

    // Due to the task heavy weight, a maximum number of features is set
    if (n_feat <= max_features_for_focus) {
        // Number of features doesn't allow the use of a cache (examine exhaustively number of ones)
        for (size = 1; size <= n_feat; size++) {
            for (int i = 1; i < pow(2, n_feat); i++)  {
                n_sel = 0;
                cont = n_feat - 1;
                num = i;

                for (int j = 0; j < n_feat; j++)
                    sel_feat[j] = 0;

                // Binarize all numbers from 1 to pow(2, n_feat)
                while (num)  {
                    if (num == 1)  {
                        sel_feat[cont] = 1;
                        n_sel++;
                        break;
                    } else  {
                        sel_feat[cont] = num%2;
                        n_sel = n_sel+sel_feat[cont];
                        num /= 2;
                        cont--;
                    }
                }

                // In each iteration of 'size' only get the inconsistency of the binary combinations
                // that have the same number of ones ('n_sel') than 'size'
                if (n_sel == size)  {
                    Ok = 0.0;
                    for (int j = 0; j < n; j++) {
                        backup = S_instances[j];
                        S_instances[j] = 0;

                        if (kNNCategory(data[j], data, sel_feat, n, k, n_feat, labels, classes, nearest, sel_feat, NULL, NULL, -1, NULL, NULL) == labels[j])
                            Ok++;

                        S_instances[j] = backup;
                    }

                    individual_evaluations += n * n * n_sel;

                    Ok /= n;

                    // If as good as with all return
                    fitness = Fitness(parm, Ok, n_sel, inputs, n, n, &reduction);
                    if (fitness >= fitness_with_all)  {
                        for (int j = 0; j < n_feat; j++)
                            S[j] = sel_feat[j];

                        *evaluations += individual_evaluations/(((scalar) n) *n*inputs);

                        return size;
                    }
                }
            }
        }

        // If not returned, best is with all features
        for (int j = 0; j < n_feat; j++)
            S[j] = 1;
    }

    *evaluations += individual_evaluations/(((scalar) n) *n*inputs);

    // Nothing done
    return n_feat;
}
