/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "classifiers.h"
#include "ensemble.h"

// eMultiBoost algorithm [Webb, 2000]

#define MAXIMUM_BETA    (10.0E+10)
#define MINIMUM_WEIGHT  (10.0E-8)

scalar UpdateMultiBoostWeightsFS(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, scalar *w, scalar *alpha, int inputs, int classes, bool initialize, int T, int t,
                                 int minority)
{
    scalar e_i, betta, sum, *h;
    bool *ok;
    static int k, *I;

    // Initialize?
    if (initialize) {
        int n = sqrt((scalar) T);
        k = 1;
        MALLOC(I, int, T+1, "UpdateMultiBoostWeightsFS");

        I[0] = 0;
        for (int i = 1; i < n; i++)
            I[i] = ceil(i * T/n);
        for (int i = n; i < T+1; i++)
            I[i] = T;

        return 0.0;
    }

    MALLOC(ok, bool, instances, "UpdateMultiBoostWeightsFS: ok");
    MALLOC(h, scalar, instances, "UpdateMultiBoostWeightsFS: h");

    // Test previous error for alpha
    e_i = GetWeightedErrorFS(parm, parm->boostfs_classifier, S, data, inputs, instances, labels, classes, ok, w, parm->cv_hyperparameters, NULL, minority, h, NULL, 0, NULL, NULL,
                             !IsEnsemble(parm->boostfs_classifier));

    // Adjust alpha
    if (e_i >= 0.5)
        *alpha = 0.00001;           // Remove classifier
    else if (e_i <= ZERO)
        // Weight classifier
        *alpha = log(1.0/MAXIMUM_BETA);
    else {
        betta = MIN(MAXIMUM_BETA, e_i/(1.0-e_i));

        // Weight classifier
        *alpha = log(1.0/betta);
    }

    sum = 0.0;

    // Initialize weights?
    if (e_i >= 0.5 || e_i <= 0.0 || I[k] == t) {
        if (classes == 2 && parm->boost_ci_initialization) {
            int ccount[] = {0, 0};
            // Count classes
            for (int i = 0; i < instances; i++)
                ccount[labels[i]]++;

            for (int i = 0; i < instances; i++)
                w[i] = 0.5/ccount[labels[i]];
        } else {
            for (int k = 0; k < instances; k++)
                sum += (w[k] = Poisson());
            k++;
        }

        if (parm->verbose > 0)
            printf("Reinitialize weights. Error: %lf\n", e_i);
    } else {
        for (int k = 0; k < instances; k++)
            if (ok[k]) {
                w[k] /= (2.0*(1.0-e_i));

                if (w[k] < MINIMUM_WEIGHT)
                    w[k] = MINIMUM_WEIGHT;

                sum += w[k];
            } else
                sum += (w[k] /= (2.0*e_i));
    }

    // Normalize w[] to sum 1
    for (int k = 0; k < instances; k++)
        w[k] *= 1.0/sum;

    FREE(h);
    FREE(ok);

    return e_i;
}
