/***************************************************************************
 *   Copyright (C) 2018 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"
#include "ranking.h"

void ChiSquareFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking, int **index)
{
    scalar **limits, *fy, *fx = NULL;
    int *intervals;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "ChiSquareFeatureRanking: limits");
    //Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "ChiSquareFeatureRanking: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, index);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, index);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Obtain p(y) for class
    MALLOC(fy, scalar, classes, "ChiSquareFeatureRanking: fy");
    ZeroVector(fy, NULL, classes);
    for (int i = 0; i < instances; i++)
        fy[labels[i]]++;

    // Obtain chi square statistics for every feature
    scalar **fo = NULL;
    for (int i = 0; i < inputs; i++) {
        REALLOC(fx, scalar, intervals[i]+1, "ChiSquareFeatureRanking: fx");
        AllocateScalarMatrix(&fo, intervals[i]+1, classes);
        ZeroVector(fx, NULL, intervals[i]+1);
        ZeroMatrix(fo, NULL, intervals[i]+1, classes);

        for (int j = 0; j < instances; j++) {
            int place;
            scalar v = index ? SparseFeature(data[j], index[j], i) : data[j][i];
            if (v < limits[i][0])
                place = 0;
            else if (v > limits[i][intervals[i]-1])
                place = intervals[i];
            else {
                place = 1;
                while (v > limits[i][place])
                    place++;
            }

            fx[place]++;
            fo[place][labels[j]]++;
        }

        // Mutual information
        scalar chi_sq = 0.0;

        for (int j = 0; j < intervals[i]+1; j++)
            for (int k = 0; k < classes; k++) {
                scalar fe = fx[j]*fy[k]/instances;
                if (fe > ZERO)
                    chi_sq += pow(fo[j][k]-fe, 2.0)/fe;
            }

        ranking[i].feature = i;
        ranking[i].score = chi_sq;

        SHOWPROGRESS(parm->show_progress > 0, "CHI2 feature %*d score: %lf\n", WIDTH(inputs), i+1, ranking[i].score);

        FreeScalarMatrix(&fo, intervals[i]+1);
    }

    FREE(fy);
    FREE(fx);
    FREE(intervals);
    for (int i = 0; i < inputs; i++)
        FREE(limits[i]);
    FREE(limits);
}
