/***************************************************************************
 *   Copyright (C) 2018 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"

// Jeffrey W. Godden and Jürgen Bajoratha, "An Information-Theoretic Approach to Descriptor Selection for Database Profiling and QSAR Modeling"", QSAR Comb. Sci. 22 (2003)


void DifferentialShannonEntropyFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking)
{
    scalar **limits,*px = NULL, **pxy, class_count[classes];
    int *intervals;

    // Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "DifferentialShannonEntropyFeatureRanking: limits");
    // Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "DifferentialShannonEntropyFeatureRanking: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Class counts
    for (int i = 0; i < classes; i++)
        class_count[i] = 0.0;
    for (int i = 0; i < instances; i++)
        class_count[labels[i]]++;

    // Obtain Shannon's entropy for every feature and for every class
    for (int i = 0; i < inputs; i++) {
        REALLOC(px, scalar, intervals[i]+1, "DifferentialShannonEntropyFeatureRanking: px");
        AllocateScalarMatrix(&pxy, intervals[i]+1, classes);
        ZeroVector(px, NULL, intervals[i]+1);
        ZeroMatrix(pxy, NULL, intervals[i]+1, classes);

        // Frequencies of x_i and Joint probabilities x, y
        for (int j = 0; j < instances; j++) {
            int place;
            if (data[j][i] < limits[i][0])
                place = 0;
            else if (data[j][i] > limits[i][intervals[i]-1])
                place = intervals[i];
            else {
                place = 1;
                while (data[j][i] > limits[i][place])
                    place++;
            }

            px[place]++;
            pxy[place][labels[j]]++;
        }

        for (int j = 0; j < intervals[i]+1; j++) {
            px[j] /= instances;
            for (int k = 0; k < classes; k++)
                pxy[j][k] /= class_count[k];
        }

        // Shannon entropy
        scalar se = 0.0;
        for (int j = 0; j < intervals[i]+1; j++)
            if (px[j] > ZERO)
                se -= px[j]*log2(px[j]);

        // Shannon entropy per class
        scalar se_average = 0.0;
        for (int k = 0; k < classes; k++) {
            scalar sec = 0.0;
            for (int j = 0; j < intervals[i]+1; j++)
                if (pxy[j][k] > ZERO)
                    sec -= pxy[j][k]*log2(pxy[j][k]);

            se_average += sec/classes;
        }

        ranking[i].feature = i;
        ranking[i].score = (se - se_average)/log2(intervals[i]+1);

        FreeScalarMatrix(&pxy, intervals[i]+1);

        SHOWPROGRESS(parm->show_progress > 0, "Differential Shannon entropy feature %*d score: %lf\n", WIDTH(inputs), i+1, ranking[i].score);
    }

    FREE(px);
    FREE(intervals);
    for (int i = 0; i < inputs; i++)
        FREE(limits[i]);
    FREE(limits);
}
