/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <typex.h>
#include "common.h"
#include "macros.h"
#include "ranking.h"
#include "fs.h"

// Weston, J., Perez-Cruz, F., Bousquet, O., Chapelle, O., Elisseeff,
// A., & Scholkopf, B. (2003) Bioinformatics 19, 764–771. http://dx.doi.org/10.1093/bioinformatics/btg054


void FisherScore(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, int minority, t_RankInfo *ranking, int **index)
{
    scalar std_dev_pos, std_dev_neg;

    if (classes != 2)
        OnErrorExit(ERR_PARAM, "Fisher score method only available for two-class problems");

    if (minority == -1)
        OnErrorExit(ERR_PARAM, "Minority class not specified");

    if (index) {
        // Sparse format
        scalar *mean_pos, *mean_neg, *sq_pos, *sq_neg;
        int n_pos = 0, n_neg = 0;

        CALLOC(mean_neg, scalar, inputs, "UnbalancedCorrelationScore: mean_neg");
        CALLOC(mean_pos, scalar, inputs, "UnbalancedCorrelationScore: mean_pos");
        CALLOC(sq_neg, scalar, inputs, "UnbalancedCorrelationScore: sq_neg");
        CALLOC(sq_pos, scalar, inputs, "UnbalancedCorrelationScore: sq_pos");

        for (int i = 0; i < instances; i++)
            if (labels[i] == minority) {
                int j = 0;
                n_pos++;
                while (index[i][j] != -1) {
                    mean_pos[index[i][j]-1] += data[i][j];
                    sq_pos[index[i][j]-1] += data[i][j]*data[i][j];
                    j++;
                }
            } else {
                int j = 0;
                n_neg++;
                while (index[i][j] != -1) {
                    mean_neg[index[i][j]-1] += data[i][j];
                    sq_neg[index[i][j]-1] += data[i][j]*data[i][j];
                    j++;
                }
            }

        for (int i = 0; i < inputs; i++) {
            std_dev_pos = sqrt(sq_pos[i]/n_pos - (mean_pos[i]/n_pos)*(mean_pos[i]/n_pos));
            std_dev_neg = sqrt(sq_neg[i]/n_neg - (mean_neg[i]/n_neg)*(mean_neg[i]/n_neg));

            ranking[i].feature = i;
            ranking[i].score = pow(std_dev_pos, 2.0)+pow(std_dev_neg, 2.0) > ZERO ? pow(mean_pos[i]/n_pos-mean_neg[i]/n_neg, 2.0)/(pow(std_dev_pos, 2.0)+pow(std_dev_neg, 2.0)) : 0.0;

            SHOWPROGRESS(parm->show_progress > 0, "Fisher Feature %*d score: %lf\r", WIDTH(inputs), i+1, ranking[i].score);
        }

        FREE(mean_neg);
        FREE(mean_pos);
        FREE(sq_neg);
        FREE(sq_pos);
    } else {
        for (int i = 0; i < inputs; i++) {
            scalar mean_pos = 0.0, mean_neg = 0.0, sq_pos = 0.0, sq_neg = 0.0;
            int n_pos = 0, n_neg = 0;

            for (int j = 0; j < instances; j++)
                if (labels[j] == minority) {
                    mean_pos += data[j][i];
                    sq_pos += data[j][i]*data[j][i];
                    n_pos++;
                } else {
                    mean_neg += data[j][i];
                    sq_neg += data[j][i]*data[j][i];
                    n_neg++;
                }

            mean_pos /= n_pos;
            sq_pos /= n_pos;
            mean_neg /= n_neg;
            sq_neg /= n_neg;

            std_dev_pos = sqrt(sq_pos - mean_pos*mean_pos);
            std_dev_neg = sqrt(sq_neg - mean_neg*mean_neg);

            ranking[i].feature = i;
            ranking[i].score = pow(std_dev_pos, 2.0)+pow(std_dev_neg, 2.0) > ZERO ? pow(mean_pos-mean_neg, 2.0)/(pow(std_dev_pos, 2.0)+pow(std_dev_neg, 2.0)) : 0.0;

            SHOWPROGRESS(parm->show_progress > 0, "Fisher Feature %*d score: %lf\r", WIDTH(inputs), i+1, ranking[i].score);
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}
