/***************************************************************************
 *   Copyright (C) 2018 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"
#include "ranking.h"

// Alhaj TA, Siraj MM, Zainal A, Elshoush HT, Elhaj F (2016) Feature Selection Using Information Gain for Improved Structural-Based Alert Correlation. PLoS ONE 11(11): e0166017.

void InformationGainFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking, int **index)
{
    scalar **limits, *p;
    int *intervals;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "InformationGainFeatureRanking: limits");
    //Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "InformationGainFeatureRanking: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, index);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, index);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Obtain entropy (Hc) for class
    MALLOC(p, scalar, classes, "InformationGainFeatureRanking: p");
    ZeroVector(p, NULL, classes);
    for (int i = 0; i < instances; i++)
        p[labels[i]]++;

    scalar Hc = 0.0;
    for (int i = 0; i < classes; i++) {
        p[i] /= instances;
        if (p[i] > ZERO)
            Hc += p[i]*log2(p[i]);
    }

    // Obtain information gain for every feature
    scalar *P[classes];
    for (int i = 0; i < classes; i++)
        P[i] = NULL;
    for (int i = 0; i < inputs; i++) {
        REALLOC(p, scalar, intervals[i]+1, "InformationGainFeatureRanking: p");
        ZeroVector(p, NULL, intervals[i]+1);
        for (int j = 0; j < instances; j++) {
            scalar v = index ? SparseFeature(data[j], index[j], i) : data[j][i];

            if (v < limits[i][0])
                p[0]++;
            else if (v > limits[i][intervals[i]-1])
                p[intervals[i]]++;
            else {
                int inter = 0;
                while (v > limits[i][inter])
                    inter++;

                p[inter]++;
            }
        }

        // Entropy (He) for feature
        scalar He = 0.0;
        for (int j = 0; j < intervals[i]+1; j++) {
            p[j] /= instances;
            if (p[j] > ZERO)
                He += p[j]*log2(p[j]);
        }

        // Conditional entropy class/feature
        scalar M = 0.0;
        for (int j = 0; j < classes; j++) {
            REALLOC(P[j], scalar, intervals[i]+1, "InformationGainFeatureRanking: P[j]");
            ZeroVector(P[j], NULL, intervals[i]+1);
        }

        for (int k = 0; k < instances; k++) {
            scalar v = index ? SparseFeature(data[k], index[k], i) : data[k][i];
            if (v < limits[i][0])
                P[labels[k]][0]++;
            else if (v > limits[i][intervals[i]-1])
                P[labels[k]][intervals[i]]++;
            else {
                int inter = 0;
                while (v > limits[i][inter])
                    inter++;

                P[labels[k]][inter]++;
            }
        }

        // Get conditional probabilities
        for (int j = 0; j < classes; j++)
            for (int k = 0; k < intervals[i]+1; k++)
                P[j][k] = (P[j][k]/instances)/(p[k] > ZERO ? p[k] : 1.0);

        for (int j = 0; j < classes; j++)
            for (int k = 0; k < intervals[i]+1; k++)
                if (P[j][k] > ZERO)
                    M += P[j][k]*log2(P[j][k]);

        ranking[i].feature = i;

        scalar Hce = -1.0*He*(-1.0)*M;
        ranking[i].score = Hc - Hce;

        // Gain ration? Then divide by H(e)
        if (parm->fralgorithm == frGR)
            ranking[i].score /= He > ZERO ? -He : 1.0;

        SHOWPROGRESS(parm->show_progress > 0, "%s feature %*d score: %lf\n", parm->fralgorithm == frGR ? "Gain Ratio" : "IG", WIDTH(inputs), i+1, ranking[i].score);
    }

    FREE(p);
    for (int i = 0; i < classes; i++)
        FREE(P[i]);
    FREE(intervals);
    for (int i = 0; i < inputs; i++)
        FREE(limits[i]);
    FREE(limits);
}
