/***************************************************************************
 *   Copyright (C) 2018 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"

// Jorge R. Vergara and Pablo A. Estévez, "A review of feature selection methods based on mutual information" Neural Comput & Applic (2014) 24:175–186

void MutualInformationFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking, scalar **lims, int *ints, int **index)
{
    scalar **limits, *py, *px = NULL, v;
    int *intervals, place;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "MutualInformationFeatureRanking: limits");
    //Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "MutualInformationFeatureRanking: intervals");

    // Discretize dataset or already done?
    if (lims == NULL) {
        switch (parm->discretizer) {
            case dFAYYAD:
                FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, index);
                break;

            case dWPKID:
                WPKIDMethod(data, instances, inputs, limits, intervals, index);
                break;

            case dNONE:
                OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
                break;
        }
    } else {
        limits = lims;
        intervals = ints;
    }

    // Obtain p(y) for class
    MALLOC(py, scalar, classes, "MutualInformationFeatureRanking: py");
    ZeroVector(py, NULL, classes);
    for (int i = 0; i < instances; i++)
        py[labels[i]]++;

    for (int i = 0; i < classes; i++)
        py[i] /= instances;

    // Obtain mutual information for every feature
    scalar **pxy = NULL;
    for (int i = 0; i < inputs; i++) {
        REALLOC(px, scalar, intervals[i]+1, "MutualInformationFeatureRanking: px");
        AllocateScalarMatrix(&pxy, intervals[i]+1, classes);
        ZeroVector(px, NULL, intervals[i]+1);
        ZeroMatrix(pxy, NULL, intervals[i]+1, classes);

        // Frequencies of x_i and Joint probabilities x, y
        for (int j = 0; j < instances; j++) {
            v = index ? SparseFeature(data[j], index[j], i) : data[j][i];

            if (v < limits[i][0])
                place = 0;
            else if (v > limits[i][intervals[i]-1])
                place = intervals[i];
            else {
                place = 1;
                while (v > limits[i][place])
                    place++;
            }

            px[place]++;
            pxy[place][labels[j]]++;
        }

        for (int j = 0; j < intervals[i]+1; j++) {
            px[j] /= instances;
            for (int k = 0; k < classes; k++)
                pxy[j][k] /= instances;
        }

        // Mutual information
        scalar mi = 0.0;

        for (int j = 0; j < intervals[i]+1; j++)
            for (int k = 0; k < classes; k++)
                if (px[j]*py[k] > ZERO && pxy[j][k] > ZERO)
                    mi += pxy[j][k]*log(pxy[j][k]/(px[j]*py[k]));

        ranking[i].feature = i;
        ranking[i].score = mi;

        SHOWPROGRESS(parm->show_progress > 0, "MI feature %*d score: %lf\r", WIDTH(inputs), i+1, ranking[i].score);

        FreeScalarMatrix(&pxy, intervals[i]+1);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FREE(py);
    FREE(px);

    if (!lims) {
        FREE(intervals);
        for (int i = 0; i < inputs; i++)
            FREE(limits[i]);
        FREE(limits);
    }
}
