/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <typex.h>
#include "common.h"
#include "macros.h"
#include "ranking.h"
#include "filter.h"
#include "fs.h"

void FeatureRankingAlgorithm(const t_Parameters *parm, t_FRAlgorithm fralgoritm, int inputs, scalar **data, int *labels, int classes, int instances, int minority, t_RankInfo *ranking, scalar **limits,
                             int *intervals, int **index)
{
    t_Feat *feat_info;
    int *mask, *subspace;

    switch (fralgoritm) {
        case frCUB:
            UnbalancedCorrelationScore(parm, inputs, data, labels, classes, instances, minority, ranking, index);
            break;

        case frFisher:
            FisherScore(parm, inputs, data, labels, classes, instances, minority, ranking, index);
            break;

        case frIG:
        case frGR:
            InformationGainFeatureRanking(parm, inputs, data, labels, classes, instances, ranking, index);
            break;

        case frRFE:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            MALLOC(feat_info, t_Feat, inputs, "FeatureRankingAlgorithm: feat_info");
            MALLOC(mask, int, inputs, "FeatureRankingAlgorithm: mask");
            for (int i = 0; i < inputs; i++)
                mask[i] = 1;
            RFE(parm, feat_info, inputs, data, instances, labels, mask);
            for (int i = 0; i < inputs; i++) {
                ranking[i].feature = feat_info[i].index;
                ranking[i].score = feat_info[i].weight;
                SHOWPROGRESS(parm->show_progress > 0, "SVM-RFE Feature %3d score: %lf\n", ranking[i].feature+1, ranking[i].score);
            }
            FREE(mask);
            FREE(feat_info);
            break;

        case frMI:
            MutualInformationFeatureRanking(parm, inputs, data, labels, classes, instances, ranking, limits, intervals, index);
            break;

        case frCHISQ:
            ChiSquareFeatureRanking(parm, inputs, data, labels, classes, instances, ranking, index);
            break;

        case frSE:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            ShannonEntropyFeatureRanking(parm, inputs, data, labels, classes, instances, ranking);
            break;

        case frDSE:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            DifferentialShannonEntropyFeatureRanking(parm, inputs, data, labels, classes, instances, ranking);
            break;

        case frReliefF:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            MALLOC(subspace, int, inputs, "FeatureRankingAlgorithm: subspace");
            ReliefAlg(parm, subspace, inputs, data, instances, labels, classes, parm->m, parm->relief_k, parm->relief_threshold, parm->relief_alpha, ranking);
            FREE(subspace);
            break;

        case frSVDEi:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            SingularValueDecompositionEntropy(parm, inputs, data, instances, ranking);
            break;

        case frEDSE:
            EuclideanDistanceBasedEntropy(parm, inputs, data, instances, ranking, index);
            break;

        case frFScore:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            FScore(parm, inputs, data, labels, classes, instances, ranking);
            break;

        case frtTest:
            tTestBasedFeatureRanking(parm, inputs, data, labels, classes, instances, ranking, index);
            break;

        case frWilcoxon:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            WilcoxonTestBasedFeatureRanking(parm, inputs, data, labels, classes, instances, ranking);
            break;

        case frSOMIFS:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            SecondOrderMutualInformationFeatureRanking(parm, inputs, data, labels, classes, instances, ranking);
            break;

        case frSimba:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            IterativeSearchMarginBasedAlgorithm(parm, data, inputs, instances, labels, ranking);
            break;

        case friSimba:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Ranking method for sparse format");
            ImprovedIterativeSearchMarginBasedAlgorithm(parm, data, inputs, instances, labels, ranking, NULL, classes);
            break;
    }
}
