/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "ranking.h"
#include "filter.h"
#include "fs.h"
#include "maths.h"

/*
 * R. Gilad-Bachrach, A. Navot and N. Tishby, "Margin Based Feature Selection - Theory and algorithms", Proc. of the 21st Int. Con. on Mach. Learn., Canada, 2004
 */

void IterativeSearchMarginBasedAlgorithm(const t_Parameters *parm, scalar **data, int inputs, int instances, int *labels, t_RankInfo *ranking)
{
    scalar *w, *delta;
    int T, hit, miss;

    MALLOC(w, scalar, inputs, "IterativeSearchMarginBasedAlgorithm: w");
    MALLOC(delta, scalar, inputs, "IterativeSearchMarginBasedAlgorithm: delta");
    for (int i = 0; i < inputs; i++)
        w[i] = 1.0;

    // If no iterations use number of instances
    T = parm->simba_T ? parm->simba_T : instances;

    for (int i = 0; i < T; i++) {
        SHOWPROGRESS(parm->show_progress > 0, "Simba iteration %*d/%d\r", WIDTH(T), i+1, T);

        int r = RANDOM(0, instances);

        NearMissHitW(r, data, inputs, instances, labels, &hit, &miss, w);

        // Not found?
        if (miss == -1 || hit == -1)
            continue;

        scalar dist_miss = 0.0, dist_hit = 0.0;
        for (int j = 0; j < inputs; j++) {
            dist_miss += SQR(w[j])*SQR(data[r][j]-data[miss][j]);
            dist_hit += SQR(w[j])*SQR(data[r][j]-data[hit][j]);
        }
        dist_miss = sqrt(dist_miss);
        dist_hit = sqrt(dist_hit);

        if (dist_hit <= ZERO || dist_miss <= ZERO)
            continue;

        for (int j = 0; j < inputs; j++) {
            delta[j] = 0.5*(SQR(data[r][j]-data[miss][j])/dist_miss - SQR(data[r][j]-data[hit][j])/dist_hit)*w[j];

            w[j] += delta[j];
        }
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // Normalize
    scalar max = MaximumOfVector(w, inputs);
    if (fabs(max) > ZERO) {
        for (int i = 0; i < inputs; i++)
            w[i] = SQR(w[i]/max);
    }

    // Scores
    for (int i = 0; i < inputs; i++) {
        ranking[i].feature = i;
        ranking[i].score = w[i];
    }

    FREE(w);
    FREE(delta);
}

void NearMissHitW(int x, scalar **data, int inputs, int instances, int *labels, int *hit, int *miss, scalar *w)
{
    scalar dist_miss = 10.E20, dist_hit = 10.E20, distance;
    *miss = *hit = -1;

    // Nearest "friend" and nearest enemy
    for (int i = 0; i < instances; i++)
        if (i != x) {
            distance = WeightedEuclideanDistance(data[x], data[i], inputs, w);

            // Hit?
            if (distance < dist_hit && labels[i] == labels[x]) {
                dist_hit = distance;
                *hit = i;
            } else if (distance < dist_miss && labels[i] != labels[x]) {
                dist_miss = distance;
                *miss = i;
            }
        }
}

// Get euclidean distance between vectors x and y and a weight vector w
scalar WeightedEuclideanDistance(scalar *x, scalar *y, int n, scalar *w)
{
    scalar dist = 0.0;

    for (int i = 0; i < n; i++)
        dist += SQR(w[i])*SQR(x[i] - y[i]);

    return sqrt(dist);
}
