/***************************************************************************
 *   Copyright (C) 2018 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "fs.h"
#include "filter.h"
#include "data.h"

// Jin-Jie Huanga, Yun-Ze Caib and Xiao-Ming Xu, "A parameterless feature ranking algorithm based on MI", Neurocomputing 71 (2008) 1656–1668.


void SecondOrderMutualInformationFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking)
{
    scalar **limits, *py, *px = NULL;
    int *intervals;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "SecondOrderMutualInformationFeatureRanking: limits");
    //Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "SecondOrderMutualInformationFeatureRanking: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Obtain p(y) for class
    MALLOC(py, scalar, classes, "SecondOrderMutualInformationFeatureRanking: py");
    ZeroVector(py, NULL, classes);
    for (int i = 0; i < instances; i++)
        py[labels[i]]++;

    for (int i = 0; i < classes; i++)
        py[i] /= instances;

    // Compute entropy for every feature and obtain mutual information for every feature and class
    scalar *H, *mi;
    MALLOC(H, scalar, inputs, "SecondOrderMutualInformationFeatureRanking: H");
    MALLOC(mi, scalar, inputs, "SecondOrderMutualInformationFeatureRanking: mi");

    ZeroVector(H, NULL, inputs);
    ZeroVector(mi, NULL, inputs);

    scalar **pxy = NULL;
    for (int i = 0; i < inputs; i++) {
        REALLOC(px, scalar, intervals[i]+1, "SecondOrderMutualInformationFeatureRanking: px");
        AllocateScalarMatrix(&pxy, intervals[i]+1, classes);
        ZeroVector(px, NULL, intervals[i]+1);
        ZeroMatrix(pxy, NULL, intervals[i]+1, classes);

        // Frequencies of x_i and joint probabilities x, y
        for (int j = 0; j < instances; j++) {
            int place;
            if (data[j][i] < limits[i][0])
                place = 0;
            else if (data[j][i] > limits[i][intervals[i]-1])
                place = intervals[i];
            else {
                place = 1;
                while (data[j][i] > limits[i][place])
                    place++;
            }

            px[place]++;
            pxy[place][labels[j]]++;
        }

        for (int j = 0; j < intervals[i]+1; j++) {
            px[j] /= instances;
            for (int k = 0; k < classes; k++)
                pxy[j][k] /= instances;
        }

        for (int j = 0; j < intervals[i]+1; j++)
            if (px[j] > ZERO)
                H[i] -= px[j]*log(px[j]);

        // Mutual information
        for (int j = 0; j < intervals[i]+1; j++)
            for (int k = 0; k < classes; k++)
                if (px[j]*py[k] > ZERO && pxy[j][k] > ZERO)
                    mi[i] += pxy[j][k]*log(pxy[j][k]/(px[j]*py[k]));

        FreeScalarMatrix(&pxy, intervals[i]+1);
    }

    // Compute mutual information of all pairs of features
    scalar **I;
    AllocateScalarMatrix(&I, inputs, inputs);
    for (int i = 0; i < inputs; i++) {
        I[i][i] = 0.0;
        for (int k = 0; k < i; k++) {
            REALLOC(px, scalar, intervals[i]+1, "SecondOrderMutualInformationFeatureRanking: px");
            REALLOC(py, scalar, intervals[k]+1, "SecondOrderMutualInformationFeatureRanking: py");

            AllocateScalarMatrix(&pxy, intervals[i]+1, intervals[k]+1);
            ZeroVector(px, NULL, intervals[i]+1);
            ZeroVector(py, NULL, intervals[k]+1);
            ZeroMatrix(pxy, NULL, intervals[i]+1, intervals[k]+1);

            // Frequencies of x, y and joint probabilities x, y
            for (int j = 0; j < instances; j++) {
                int placex, placey;
                if (data[j][i] < limits[i][0])
                    placex = 0;
                else if (data[j][i] > limits[i][intervals[i]-1])
                    placex = intervals[i];
                else {
                    placex = 1;
                    while (data[j][i] > limits[i][placex])
                        placex++;
                }

                if (data[j][k] < limits[k][0])
                    placey = 0;
                else if (data[j][k] > limits[k][intervals[k]-1])
                    placey = intervals[k];
                else {
                    placey = 1;
                    while (data[j][k] > limits[k][placey])
                        placey++;
                }

                px[placex]++;
                py[placey]++;
                pxy[placex][placey]++;
            }

            for (int j = 0; j < intervals[i]+1; j++)
                px[j] /= instances;

            for (int j = 0; j < intervals[k]+1; j++)
                py[j] /= instances;

            for (int j = 0; j < intervals[i]+1; j++)
                for (int l = 0; l < intervals[k]+1; l++)
                    pxy[j][l] /= instances;

            // Mutual information
            for (int j = 0; j < intervals[i]+1; j++)
                for (int l = 0; l < intervals[k]+1; l++)
                    if (px[j]*py[l] > ZERO && pxy[j][l] > ZERO)
                        I[i][k] += pxy[j][l]*log(pxy[j][l]/(px[j]*py[l]));

            I[k][i] = I[i][k];

            FreeScalarMatrix(&pxy, intervals[i]+1);
        }
    }

    // Get feature with largest MI
    int largest = 0;
    for (int i = 1; i < inputs; i++)
        if (mi[i] > mi[largest])
            largest = i;

    ranking[largest].feature = largest;
    ranking[largest].score = inputs;

    SHOWPROGRESS(parm->show_progress > 1, "Second order MI feature %*d score: %lf\r", WIDTH(inputs), largest+1, ranking[largest].score);

    // Repeat until all features are ordered
    int remain = inputs-1, *list;
    MALLOC(list, int, inputs, "SecondOrderMutualInformationFeatureRanking: list");

    for (int i = 0; i < inputs; i++)
        list[i] = i;
    list[largest] = inputs-1;

    while (remain > 1) {
        scalar I_max = -1.0;
        int next = -1;

        // Compute I and test if maximum
        for (int i = 0; i < remain; i++) {
            scalar IC;

            scalar sum = 0.0;
            for (int k = 0; k < remain; k++) {
                scalar phi_sum = 0.0;
                for (int j = 0; j < remain; j++)
                    if (j != k && H[list[j]] > ZERO && H[list[k]] > ZERO)
                        phi_sum += (I[list[i]][list[j]]/H[list[j]])*(I[list[j]][list[k]]/H[list[k]]);

                phi_sum = H[list[k]] > ZERO ? I[list[i]][list[k]]/H[list[k]] - 0.5*phi_sum : 0.0;

                sum += phi_sum*mi[list[k]];
            }

            IC = mi[list[i]] - sum;

            if (IC > I_max) {
                I_max = IC;
                next = i;
            }
        }

        ranking[list[next]].feature = list[next];
        ranking[list[next]].score = remain;

        SHOWPROGRESS(parm->show_progress > 1, "Second order MI feature %*d score: %lf\r", WIDTH(inputs), list[next]+1, ranking[list[next]].score);

        // Remove selected feature
        list[next] = list[remain-1];
        remain--;
    }

    // Last one
    ranking[list[0]].feature = list[0];
    ranking[list[0]].score = 1;
    SHOWPROGRESS(parm->show_progress > 1, "Second order MI feature %*d score: %lf\n", WIDTH(inputs), list[0]+1, ranking[list[0]].score);

    FREE(py);
    FREE(px);
    FREE(mi);
    FREE(H);
    FREE(list);
    FreeScalarMatrix(&I, inputs);
    FREE(intervals);
    for (int i = 0; i < inputs; i++)
        FREE(limits[i]);
    FREE(limits);
}
