/***************************************************************************
 *   Copyright (C) 2019 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "fs.h"
#include "filter.h"

// Singular Value Decomposition Entropy (SVDEi)
// Ricardo W. Pino Urias et al., "IMMAN: free software for information theory-based chemometric analysis", Mol Divers (2015) 19:305–319

void SingularValueDecompositionEntropy(const t_Parameters *parm, int inputs, scalar **data, int instances, t_RankInfo *ranking)
{
    if (instances < inputs)
        OnErrorExit(ERR_NUMERICAL, "SVDEi method only available for instances >= features");

    // Perform singular value decomposition of data matrix
    scalar **v, *w, **a;
    AllocateScalarMatrix(&v, inputs, inputs);
    AllocateScalarMatrix(&a, instances, inputs);
    MALLOC(w, scalar, inputs, "SingularValueDecompositionEntropy: w");

    CopyScalarMatrix(data, a, instances, inputs);
    if (!SingularValueDecomposition(a, instances, inputs, w, v))
        OnErrorExit(ERR_NUMERICAL, "Singular value decomposition failed");

    // Dataset entropy
    scalar ST = 0.0;
    for (int i = 0; i < inputs; i++)
        ST += w[i]*w[i];

    scalar EA = 0.0;
    for (int i = 0; i < inputs; i++)
        EA += ((w[i]*w[i])/ST)*log10((w[i]*w[i])/ST);
    EA *= -1.0/log10((scalar) inputs);

    // Relative contribution of every feature
    for (int i = 0; i < inputs; i++) {
        // Copy excluding feature
        CopyScalarMatrixExcludeColumn(data, a, instances, inputs, i);

        // Dataset entropy without i-th feature
        if (!SingularValueDecomposition(a, instances, inputs-1, w, v))
            OnErrorExit(ERR_NUMERICAL, "Singular value decomposition failed");

        // Dataset entropy
        ST = 0.0;
        for (int j = 0; j < inputs-1; j++)
            ST += w[j]*w[j];

        scalar EAi = 0.0;
        for (int j = 0; j < inputs-1; j++)
            EAi += ((w[j]*w[j])/ST)*log10((w[j]*w[j])/ST);
        EAi *= -1.0/log10((scalar) inputs-1);

        SHOWPROGRESS(parm->show_progress > 0, "SVDEi feature %*d score: %lf\n", WIDTH(inputs), i+1, EAi-EA);

        // Store score
        ranking[i].feature = i;
        ranking[i].score = EAi - EA;
    }

    FREE(w);
    FreeScalarMatrix(&v, inputs);
    FreeScalarMatrix(&a, instances);
}

