/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "fs.h"
#include <typex.h>
#include "common.h"
#include "macros.h"
#include "ranking.h"

// t-test Based Feature Ranking

// Ligang Zhou, Kin Keung Lai, Jerome Yen, Empirical models based on features ranking techniques for corporate financial distress prediction
// Computers and Mathematics with Applications 64 (2012) 2484–2496

// Multi-class version:
// Nina Zhou and Lipo Wang
// A Modified T-test Feature Selection Method and Its Application on the HapMap Genotype Data
// Geno. Prot. Bioinfo.Vol. 5 No. 3–4, 2007

void tTestBasedFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking, int **index)
{
    scalar v;
    if (classes == 2) {
        scalar var_pos, var_neg;

        for (int i = 0; i < inputs; i++) {
            // Mean and variance within class
            scalar mean_pos = 0.0, mean_neg = 0.0, sq_pos = 0.0, sq_neg = 0.0;
            int pos = 0, neg = 0;
            for (int j = 0; j < instances; j++) {
                v = index ? SparseFeature(data[j], index[j], i) : data[j][i];

                if (labels[j] == 1) {
                    mean_pos += v;
                    sq_pos += v*v;
                    pos++;
                } else {
                    mean_neg += v;
                    sq_neg += v*v;
                    neg++;
                }
            }

            if (pos == 0 || neg == 0)
                OnErrorExit(ERR_NUMERICAL, "Only one class");

            mean_pos /= pos;
            mean_neg /= neg;

            var_pos = sq_pos/pos - mean_pos*mean_pos;
            var_neg = sq_neg/neg - mean_neg*mean_neg;

            // Obtain z for i-th feature
            scalar z = fabs(mean_pos - mean_neg)/(var_neg > ZERO && var_pos > ZERO ? sqrt(var_pos/pos+var_neg/neg) : 1.0);

            ranking[i].feature = i;
            ranking[i].score = z;

            SHOWPROGRESS(parm->show_progress > 0, "t-Test Feature %3d score: %lf\n", i+1, ranking[i].score);
        }
    } else {
        scalar mean[classes], z, mean_all = 0.0;
        int count[classes];

        for (int i = 0; i < inputs; i++) {
            ZeroVector(mean, NULL, classes);
            ZeroVector(NULL, count, classes);

            // Z score for every class
            for (int j = 0; j < instances; j++) {
                v = index ? SparseFeature(data[j], index[j], i) : data[j][i];
                mean_all += v;
                mean[labels[j]] += v;
                count[labels[j]]++;
            }

            for (int j = 0; j < classes; j++) {
                if (!count[j])
                    OnErrorExit(ERR_NUMERICAL, "Class without instances");

                mean[j] /= count[j];
            }
            mean_all /= instances;

            scalar Si2 = 0.0;
            for (int j = 0; j < instances; j++)
                Si2 = pow((index ? SparseFeature(data[j], index[j], i) : data[j][i]) - mean[labels[j]], 2.0);
            Si2 /= 1.0/(instances - classes);

            z = 0.0;
            if (Si2 > ZERO) {
                // Obtain z as the maximum for every class
                scalar Mc = sqrt(1.0/count[0]+1.0/instances);
                z = fabs(mean[0] - mean_all)/(Mc*sqrt(Si2));

                for (int j = 1; j < classes; j++) {
                    Mc = sqrt(1.0/count[j]+1.0/instances);
                    if (fabs(mean[j] - mean_all)/(Mc*sqrt(Si2)) > z)
                        z = fabs(mean[j] - mean_all)/(Mc*sqrt(Si2));
                }
            }

            ranking[i].feature = i;
            ranking[i].score = z;

            SHOWPROGRESS(parm->show_progress > 1 || (parm->show_progress > 0 && !(i+1 % 1000)), "t-Test feature %*d score: %lf\r", WIDTH(inputs), i+1, ranking[i].score);
        }

        SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    }
}
