/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "fs.h"
#include <typex.h>
#include "common.h"
#include "macros.h"

// Wilcoxon test Based Feature Ranking

// L. Chen, S. Li, and Z. Luo, "Gene Selection Using Wilcoxon Rank Sum Test and Support Vector Machine"
// In Y. Wang, Y. Cheung, and H. Liu (Eds.): CIS 2006, LNAI 4456, pp. 57–66, 2007. Springer-Verlag Berlin Heidelberg 2007

void WilcoxonTestBasedFeatureRanking(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, t_RankInfo *ranking)
{
    if (classes != 2)
        OnErrorExit(ERR_PARAM, "Wilcoxon test ranking only available for 2-class problems");

    // List of instances of both classes
    int *list_0, *list_1, n_0 = 0, n_1 = 0;
    MALLOC(list_0, int, instances, "WilcoxonTestBasedFeatureRanking: list_0");
    MALLOC(list_1, int, instances, "WilcoxonTestBasedFeatureRanking: list_1");

    for (int i = 0; i < instances; i++)
        if (labels[i] == 0) {
            list_0[n_0] = i;
            n_0++;
        } else {
            list_1[n_1] = i;
            n_1++;
        }

    for (int i = 0; i < inputs; i++) {
        scalar sg = 0.0;

        for (int j = 0; j < n_0; j++)
            for (int k = 0; k < n_1; k++)
                if (data[list_1[k]][i] <= data[list_0[j]][i])
                    sg++;

        // Score
        scalar qg = MAX(sg, n_0*n_1-sg);

        ranking[i].feature = i;
        ranking[i].score = qg;

        SHOWPROGRESS(parm->show_progress > 0, "Wilcoxon test feature %*d score: %lf\n", WIDTH(inputs), i+1, ranking[i].score);
    }

    FREE(list_0);
    FREE(list_1);
}
