/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"
#include "typex.h"
#include "data.h"

// Implements redundancy measures

// Wr: I. Jo, S. Lee and S. Oh (2019) "Improved measures of redundancy and relevance for mRMR feature selection," Computers 8(2)):42.

void RedundancyMeasureOfSeletedFeatures(const t_Parameters *parm, scalar **data, int instances, int *labels, int *subspace, int classes, int features, scalar *Wr, scalar *MI, scalar *ave_corr)
{
    scalar *sigma, *mean, maxC, meanC;

    MALLOC(sigma, scalar, features, "RedundancyMeasureOfSeletedFeatures: sigma");
    MALLOC(mean, scalar, features, "RedundancyMeasureOfSeletedFeatures: mean");
    ZeroVector(sigma, NULL, features);
    ZeroVector(mean, NULL, features);

    // Correlation matrix
    for (int i = 0; i < features; i++)
        if (!subspace || subspace[i])
            for (int j = 0; j < instances; j++) {
                mean[i] += data[j][i];
                sigma[i] += data[j][i]*data[j][i];
            }

    for (int i = 0; i < features; i++)
        if (!subspace || subspace[i]) {
            mean[i] /= instances;
            sigma[i] = sqrt(sigma[i]/instances-mean[i]*mean[i]);
        }

    // Obtain max correlation and mean in absolute value for all pairs
    maxC = meanC = 0.0;
    int p = 0;
    for (int i = 1; i < features; i++)
        if ((!subspace || subspace[i]) && sigma[i] > ZERO)
            for (int j = 0; j < i; j++)
                if ((!subspace || subspace[j]) && sigma[j] > ZERO) {
                    scalar corr = 0.0;
                    for (int k = 0; k < instances; k++)
                        corr += (data[k][i]-mean[i])*(data[k][j]-mean[j]);

                    corr = fabs((corr/instances)/(sigma[i]*sigma[j]));

                    meanC += corr;
                    if (corr > maxC)
                        maxC = corr;

                    p++;    // Another couple
                }

    // If maxC >= 0.5 return maxC, otherwise return meanC
    if (p) {
        meanC /= p;
        *Wr = (maxC >= 0.5) ? maxC : meanC;
        *ave_corr = meanC;

        // Average mutual information
        *MI = AverageMutualInformationAllPairs(parm, features, data, labels, classes, instances, subspace);
    } else
        *Wr = *MI = *ave_corr = 0.0;

    FREE(sigma);
    FREE(mean);
}


scalar AverageMutualInformationAllPairs(const t_Parameters *parm, int inputs, scalar **data, int *labels, int classes, int instances, int *subspace)
{
    scalar **limits, *py = NULL, *px = NULL, **pxy = NULL, average_mi = 0.0;
    int *intervals;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, inputs, "AverageMutualInformationAllPairs: limits");
    //Set limits to NULL
    for (int k = 0; k < inputs; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, inputs, "AverageMutualInformationAllPairs: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, inputs, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, inputs, limits, intervals, NULL);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Compute mutual information of all pairs of features
    scalar I;
    int p = 0;
    for (int i = 0; i < inputs; i++)
        if (!subspace || subspace[i])
            for (int k = 0; k < i; k++)
                if (!subspace || subspace[k]) {
                    I = 0.0;
                    p++;

                    REALLOC(px, scalar, intervals[i]+1, "AverageMutualInformationAllPairs: px");
                    REALLOC(py, scalar, intervals[k]+1, "AverageMutualInformationAllPairs: py");

                    AllocateScalarMatrix(&pxy, intervals[i]+1, intervals[k]+1);
                    ZeroVector(px, NULL, intervals[i]+1);
                    ZeroVector(py, NULL, intervals[k]+1);
                    ZeroMatrix(pxy, NULL, intervals[i]+1, intervals[k]+1);

                    // Frequencies of x, y and joint probabilities x, y
                    for (int j = 0; j < instances; j++) {
                        int placex, placey;
                        if (data[j][i] < limits[i][0])
                            placex = 0;
                        else if (data[j][i] > limits[i][intervals[i]-1])
                            placex = intervals[i];
                        else {
                            placex = 1;
                            while (data[j][i] > limits[i][placex])
                                placex++;
                        }

                        if (data[j][k] < limits[k][0])
                            placey = 0;
                        else if (data[j][k] > limits[k][intervals[k]-1])
                            placey = intervals[k];
                        else {
                            placey = 1;
                            while (data[j][k] > limits[k][placey])
                                placey++;
                        }

                        px[placex]++;
                        py[placey]++;
                        pxy[placex][placey]++;
                    }

                    for (int j = 0; j < intervals[i]+1; j++)
                        px[j] /= instances;

                    for (int j = 0; j < intervals[k]+1; j++)
                        py[j] /= instances;

                    for (int j = 0; j < intervals[i]+1; j++)
                        for (int l = 0; l < intervals[k]+1; l++)
                            pxy[j][l] /= instances;

                    // Mutual information
                    for (int j = 0; j < intervals[i]+1; j++)
                        for (int l = 0; l < intervals[k]+1; l++)
                            if (px[j]*py[l] > ZERO && pxy[j][l] > ZERO)
                                I += pxy[j][l]*log(pxy[j][l]/(px[j]*py[l]));

                    FreeScalarMatrix(&pxy, intervals[i]+1);

                    average_mi += I;
                }

    FREE(py);
    FREE(px);
    FREE(intervals);
    for (int i = 0; i < inputs; i++)
        FREE(limits[i]);
    FREE(limits);

    // Return average MI
    return average_mi/p;
}
