/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "fs.h"
#include "macros.h"

scalar EvaluateThresholdFS(const t_Parameters *parm, scalar threshold, scalar *votes, scalar **data, int instances, int *labels, scalar *reduction, scalar *accuracy, int inputs, int classes,
                           bool reset, int *evaluations, int minority, scalar *auroc, scalar *g_mean, scalar *kappa, scalar *auprc)
{
    int *subspace, **M;
    t_PerformanceInfo perf;

    // Previous values
    static int sel;
    static scalar red, acc, gme, kap, aur, aup;

    MALLOC(subspace, int, inputs, "EvaluateThresholdFS: subspace");

    // Forget previous iteration
    if (reset) {
        sel = -1;
        red = acc = gme = kap = aur = aup = -1.0;
    }

    // Count selected
    int selected = 0;
    for (int i = 0; i < inputs; i++)
        if (votes[i] >= threshold) {
            subspace[i] = 1;
            selected++;
        } else
            subspace[i] = 0;

    // If same as previous assign values and return
    if (sel == selected) {
        *accuracy = parm->boostfs_w_acc > ZERO ? acc/parm->boostfs_w_acc : 0.0;
        *reduction = parm->boostfs_w_red > ZERO ? red/parm->boostfs_w_red : 0.0;
        *auroc = parm->boostfs_w_auroc > ZERO ? aur/parm->boostfs_w_auroc : 0.0;
        *auprc = parm->boostfs_w_auprc > ZERO ? aup/parm->boostfs_w_auprc : 0.0;
        *g_mean = parm->boostfs_w_gme > ZERO ? gme/parm->boostfs_w_gme : 0.0;
        *kappa = parm->boostfs_w_kap > ZERO ? kap/parm->boostfs_w_kap : 0.0;

        FREE(subspace);

        return red + acc + gme + aur + kap;
    }

    (*evaluations)++;
    sel = selected;

    gme = kap = acc = 0.0;

    // Performance
    if (parm->boostfs_w_gme > ZERO || parm->boostfs_w_acc > ZERO || parm->boostfs_w_kap > ZERO) {
        AllocateIntegerMatrix(&M, classes, classes);

        ClassifierEvaluation(parm, data, instances, labels, data, instances, labels, subspace, inputs, NULL, parm->cv_hyperparameters, classes, parm->boostfs_classifier, NULL, minority, M, NULL,
                                   &perf, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

        gme = (parm->boostfs_w_gme < ZERO || classes != 2) ? 0.0 : parm->boostfs_w_gme * perf.gmean;
        acc = (parm->boostfs_w_acc <= ZERO) ? 0.0 : parm->boostfs_w_acc * perf.accuracy;
        kap = (parm->boostfs_w_kap <= ZERO) ? 0.0 : parm->boostfs_w_kap * perf.kappa;

        FreeIntegerMatrix(&M, classes);
    }

    // AUC
    aur = (parm->boostfs_w_auroc < ZERO || classes != 2) ? 0.0 : parm->boostfs_w_auroc *
          ClassifierAUROC(parm, parm->boostfs_classifier, data, instances, labels, data, instances, NULL,  subspace, labels, inputs, minority, false, NULL, true, NULL, classes, NULL, NULL, NULL,
                                NULL, NULL, true);

    aup = (parm->boostfs_w_auprc < ZERO || classes != 2) ? 0.0 : parm->boostfs_w_auprc *
          ClassifierAUPRC(parm, parm->boostfs_classifier, data, instances, labels, data, instances, labels, inputs, minority, false, NULL, NULL, classes, NULL, NULL, subspace, NULL, NULL, NULL,
                                NULL, true);

    red = (1.0 - ((scalar) selected)/inputs) * parm->boostfs_w_red;

    // Store for next threshold
    *accuracy = parm->boostfs_w_acc > ZERO ? acc/parm->boostfs_w_acc : 0.0;
    *reduction = parm->boostfs_w_red > ZERO ? red/parm->boostfs_w_red : 0.0;
    *auroc = parm->boostfs_w_auroc > ZERO ? aur/parm->boostfs_w_auroc : 0.0;
    *auprc = parm->boostfs_w_auprc > ZERO ? aup/parm->boostfs_w_auprc : 0.0;
    *g_mean = parm->boostfs_w_gme > ZERO ? gme/parm->boostfs_w_gme : 0.0;
    *kappa = parm->boostfs_w_kap > ZERO ? kap/parm->boostfs_w_kap : 0.0;

    FREE(subspace);

    return red + acc + gme + aur + aup + kap;
}
