/***************************************************************************
 *   Copyright (C) 2019 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "wrapper.h"
#include "filter.h"
#include "fs.h"
#include "data.h"
#include "ensemble.h"

// H. Liu, L. Liu and H. Zhang, "Boosting feature selection using information metric for classification", Neurocomputing 73 (2009) 295–303

int BoostingMutualInformationFeatureSelection(const t_Parameters *parm, int features, scalar **data, int *labels, int classes, int instances, int *subspace)
{
    scalar **limits, *py;
    int *intervals, *S;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, features, "BoostingMutualInformationFeatureSelection: limits");
    //Set limits to NULL
    for (int k = 0; k < features; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, features, "BoostingMutualInformationFeatureSelection: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, features, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, features, limits, intervals, NULL);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Obtain p(y) for class
    MALLOC(py, scalar, classes, "BoostingMutualInformationFeatureSelection: py");
    ZeroVector(py, NULL, classes);
    for (int i = 0; i < instances; i++)
        py[labels[i]]++;

    for (int i = 0; i < classes; i++)
        py[i] /= instances;

    // All instances equally weighted
    scalar *w;
    MALLOC(w, scalar, instances, "BoostingMutualInformationFeatureSelection: w");
    for (int i = 0; i < instances; i++)
        w[i] = 1.0/instances;

    // Inconsistency with all features
    for (int i = 0; i < features; i++)
        subspace[i] = 1;
    scalar full_inconsistency = InconsistCount(features, data, instances, subspace, labels, classes, limits, intervals);
    SHOWPROGRESS(parm->show_progress > 1, "BMIFS inconsistency with ALL features: %lf\n", full_inconsistency);

    // All features unselected
    int selected = 0;
    for (int i = 0; i < features; i++)
        subspace[i] = 0;

    // Summary of correctly classified instances for every feature
    bool **ok;
    MALLOC(ok, bool *, features, "BoostingMutualInformationFeatureSelection: ok");
    for (int i = 0; i < features; i++)
        ok[i] = NULL;

    // For correctness evaluation
    MALLOC(S, int, features, "BoostingMutualInformationFeatureSelection: S");
    ZeroVector(NULL, S, features);

    // Obtain mutual information for every feature
    scalar inconsistency;
    do {
        // Get maximum MI of unseleted features
        int best_feature = -1;
        scalar best_mi = -1.0;

        for (int i = 0; i < features; i++)
            if (!subspace[i]) {
                scalar mi = WeightedMutualInformation(i, py, data, instances, classes, intervals, limits, labels, w);

                if (mi > best_mi) {
                    best_mi = mi;
                    best_feature = i;
                }
            }

        // Add best to selection
        subspace[best_feature] = 1;
        SHOWPROGRESS(parm->show_progress > 1, "BMIFS Selected feature: %*d\n", WIDTH(features), best_feature+1);

        // Obtain error per instance
        MALLOC(ok[best_feature], bool, instances, "BoostingMutualInformationFeatureSelection: ok[best_feature]");
        S[best_feature] = 1;
        selected++;

        // Get error
        scalar e = SingleClassifierClassifyDataset(parm, data, instances, labels, data, instances, labels, S, features, NULL, parm->cv_hyperparameters, classes, parm->wrapper_fs_classifier,
                   NULL, 0, ok[best_feature], NULL, NULL);
        S[best_feature] = 0;

        // Update weights: Strategy 2 in paper as in the reported experiments
        scalar g_e;

        if (e > 0.5)
            e = 1.0 - e;
        g_e = e > ZERO ? MIN((1.0-e)/e, 100.0) : 100.0;

        for (int i = 0; i < instances; i++) {
            bool correct = true;
            for (int j = 0; j < features && correct; j++)
                if (subspace[j] && !ok[j][i])
                    correct = false;

            w[i] += correct ? -g_e : g_e;
        }

        // Get current inconsistency value
        inconsistency = InconsistCount(features, data, instances, subspace, labels, classes, limits, intervals);

        SHOWPROGRESS(parm->show_progress > 1, "BMIFS inconsistency with %*d features: %lf\n", WIDTH(features), selected, inconsistency);
    } while (inconsistency > full_inconsistency);

    FREE(py);
    FREE(w);
    for (int i = 0; i < features; i++)
        if (ok[i])
            FREE(ok[i]);
    FREE(ok);
    FREE(S);

    return selected;
}

scalar WeightedMutualInformation(int feature, scalar *py, scalar **data, int instances, int classes, int *intervals, scalar **limits, int *labels, scalar *w)
{
    scalar *px, **pxy;

    MALLOC(px, scalar, intervals[feature]+1, "WeightedMutualInformation: px");
    AllocateScalarMatrix(&pxy, intervals[feature]+1, classes);
    ZeroVector(px, NULL, intervals[feature]+1);
    ZeroMatrix(pxy, NULL, intervals[feature]+1, classes);

    // Frequencies of x and joint probabilities x, y
    scalar sum = 0.0;
    for (int j = 0; j < instances; j++) {
        int place;
        if (data[j][feature] < limits[feature][0])
            place = 0;
        else if (data[j][feature] > limits[feature][intervals[feature]-1])
            place = intervals[feature];
        else {
            place = 1;
            while (data[j][feature] > limits[feature][place])
                place++;
        }

        px[place] += w[j];
        pxy[place][labels[j]] += w[j];

        sum += w[j];
    }

    for (int j = 0; j < intervals[feature]+1; j++) {
        px[j] /= sum;
        for (int k = 0; k < classes; k++)
            pxy[j][k] /= sum;
    }

    // Mutual information
    scalar mi = 0.0;

    for (int j = 0; j < intervals[feature]+1; j++)
        for (int k = 0; k < classes; k++)
            if (px[j]*py[k] > ZERO && pxy[j][k] > ZERO)
                mi += pxy[j][k]*log(pxy[j][k]/(px[j]*py[k]));

    FreeScalarMatrix(&pxy, intervals[feature]+1);
    FREE(px);

    return mi;
}
