/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include "macros.h"
#include "typex.h"
#include "ranking.h"
#include "filter.h"
#include "wrapper.h"

// H. Peng, F. Long and C. Ding (2005) "Feature selection based on mutual information: Criteria of max-dependency, max-relevance and min-redundancy", TPAMI 27(8):1226-1238.

void mRMR(const t_Parameters *parm, scalar **data, int instances, int *labels, int classes, int *subspace, int features, int minority)
{
    t_RankInfo *ranking;
    int *S, **M, *candidates, n_candidates;
    t_PerformanceInfo perf;
    bool end;

    MALLOC(ranking, t_RankInfo, features, "mRMR: ranking");
    MALLOC(S, int, features, "mRMR: S");
    AllocateIntegerMatrix(&M, classes, classes);

    // Order features
    SecondOrderMutualInformationFeatureRanking(parm, features, data, labels, classes, instances, ranking);

    // Select n best
    qsort(ranking, features, sizeof(t_RankInfo), CompareRankingScore);
    ZeroVector(NULL, subspace, features);

    // Get the size of the smallest sequential error
    int omega = 1;
    ZeroVector(NULL, S, features);
    S[ranking[0].feature] = 1;
    kFoldClassifierEvaluation(parm, data, instances, labels, S, features, NULL, classes, parm->mrmr_evaluator, minority, M, &perf, NULL, NULL, parm->cv_folds);

    scalar e_best = perf.accuracy;

    int n = MIN(parm->mrmr_n, features);
    for (int i = 1; i < n; i++) {
        S[ranking[i].feature] = 1;
        kFoldClassifierEvaluation(parm, data, instances, labels, S, features, NULL, classes, parm->mrmr_evaluator, minority, M, &perf, NULL, NULL, parm->cv_folds);
        if (e_best < perf.accuracy) {
            e_best = perf.accuracy;
            omega = i+1;
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "mRMR omega: %d\n", omega);

    // Two options: Forward or Backward selection
    MALLOC(candidates, int, omega, "mRMR: candidates");

    switch (parm->mrmr_search) {
        case SFS:
            // Select initially best feature
            ZeroVector(NULL, S, features);
            S[ranking[0].feature] = 1;

            // Search until error is worse
            kFoldClassifierEvaluation(parm, data, instances, labels, S, features, NULL, classes, parm->mrmr_evaluator, minority, M, &perf, NULL, NULL, parm->cv_folds);
            e_best = perf.accuracy;

            // Store best so far
            memcpy(subspace, S, sizeof(int)*features);

            n_candidates = omega - 1;
            for (int i = 1; i < omega; i++)
                candidates[i-1] = ranking[i].feature;

            end = false;
            do {
                int next = -1;
                scalar best = -1.0;
                // Test candidate
                for (int i = 0; i < n_candidates; i++) {
                    S[candidates[i]] = 1;

                    kFoldClassifierEvaluation(parm, data, instances, labels, S, features, NULL, classes, parm->mrmr_evaluator, minority, M, &perf, NULL, NULL, parm->cv_folds);

                    if (perf.accuracy > best) {
                        next = i;
                        best = perf.accuracy;
                    }

                    S[candidates[i]] = 0;
                }

                // Improves?
                if (best > e_best) {
                    // Store new best
                    e_best = best;
                    S[candidates[next]] = 1;
                    // Store best so far
                    memcpy(subspace, S, sizeof(int)*features);

                    SHOWPROGRESS(parm->show_progress > 1, "Added feature %*d (Remain: %*d)\r", WIDTH(features), candidates[next], WIDTH(features), n_candidates-1);
                } else if (fabs(best - e_best) < ZERO)
                    // Continue but do not consider best
                    S[candidates[next]] = 1;
                else    // End
                    end = true;

                // Remove best from list
                n_candidates--;
                candidates[next] = candidates[n_candidates];
            } while (!end);

            SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
            break;

        case SBS:
            ZeroVector(NULL, S, features);
            for (int i = 0; i < omega; i++)
                S[ranking[i].feature] = 1;

            // Search until error is worse
            kFoldClassifierEvaluation(parm, data, instances, labels, S, features, NULL, classes, parm->mrmr_evaluator, minority, M, &perf, NULL, NULL, parm->cv_folds);

            e_best = perf.accuracy;
            // Store best so far
            memcpy(subspace, S, sizeof(int)*features);

            n_candidates = omega;
            for (int i = 0; i < omega; i++)
                candidates[i-1] = ranking[i].feature;

            end = false;
            do {
                int next = -1;
                scalar best = -1.0;
                // Test candidate
                for (int i = 0; i < n_candidates; i++) {
                    S[candidates[i]] = 0;

                    kFoldClassifierEvaluation(parm, data, instances, labels, S, features, NULL, classes, parm->mrmr_evaluator, minority, M, &perf, NULL, NULL, parm->cv_folds);

                    if (perf.accuracy > best) {
                        next = i;
                        best = perf.accuracy;
                    }

                    S[candidates[i]] = 1;
                }

                // Improves?
                if (best >= e_best) {
                    // Store new best
                    e_best = best;
                    S[candidates[next]] = 1;
                    // Store best so far
                    memcpy(subspace, S, sizeof(int)*features);
                } else // End
                    end = true;

                // Remove best from list
                n_candidates--;
                candidates[next] = candidates[n_candidates];
            } while (!end);

            break;

        default:
            OnErrorExit(ERR_PARAM, "Search heuristic not suitable for mRMR");
    }

    FREE(ranking);
    FREE(S);
    FreeIntegerMatrix(&M, classes);
}
