/***************************************************************************
 *   Copyright (C) 2019 by Nicolás García-Pedrajas                         *
 *   adeharo@uco.es                                                        *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include "macros.h"
#include "wrapper.h"
#include "filter.h"
#include "fs.h"
#include "svm.h"
#include "classifiers.h"
#include "data.h"
#include "ensemble.h"

// E. Schaffernicht and H. M. Gross (2011) "Weighted Mutual Information for Feature Selection." 181-188
// In: Honkela T., Duch W., Girolami M., Kaski S. (eds) Artificial Neural Networks and Machine Learning – ICANN 2011. ICANN 2011.
// Lecture Notes in Computer Science, vol 6792. Springer, Berlin, Heidelberg

int WeightedMutualInformationFS(const t_Parameters *parm, int features, scalar **data, int *labels, int classes, int instances, int *subspace)
{
    scalar **limits, *py, min_margin;
    int *intervals, **M;
    bool *ok;
    t_PerformanceInfo perf;
    t_Learner learner;
    t_Ensemble ensemble;

    //Memory allocation for auxiliary data structure
    MALLOC(limits, scalar *, features, "BoostingMutualInformationFeatureSelection: limits");
    //Set limits to NULL
    for (int k = 0; k < features; k++)
        limits[k] = NULL;
    MALLOC(intervals, int, features, "BoostingMutualInformationFeatureSelection: intervals");

    // Discretize dataset
    switch (parm->discretizer) {
        case dFAYYAD:
            FayyadAndIraniMDLMethod(data, instances, classes, labels, features, limits, intervals, NULL, NULL);
            break;

        case dWPKID:
            WPKIDMethod(data, instances, features, limits, intervals, NULL);
            break;

        case dNONE:
            OnErrorExit(ERR_PARAM, "Information gain needs discretization method");
            break;
    }

    // Obtain p(y) for class
    MALLOC(py, scalar, classes, "BoostingMutualInformationFeatureSelection: py");
    ZeroVector(py, NULL, classes);
    for (int i = 0; i < instances; i++)
        py[labels[i]]++;

    for (int i = 0; i < classes; i++)
        py[i] /= instances;

    // All instances equally weighted
    scalar *w, *margin;
    MALLOC(margin, scalar, instances, "BoostingMutualInformationFeatureSelection: score");
    MALLOC(ok, bool, instances, "BoostingMutualInformationFeatureSelection: ok");
    MALLOC(w, scalar, instances, "BoostingMutualInformationFeatureSelection: w");
    for (int i = 0; i < instances; i++)
        w[i] = 1.0;

    // All features unselected
    int selected = 0;
    for (int i = 0; i < features; i++)
        subspace[i] = 0;

    AllocateIntegerMatrix(&M, classes, classes);

    // Obtain weighted mutual information for every feature and select highest
    scalar previous_acc = 0.0;
    bool finish = false;
    do {
        // Get maximum weighted MI of unseleted features
        int best_feature = -1;
        scalar best_mi = -1.0;

        for (int i = 0; i < features; i++)
            if (!subspace[i]) {
                scalar mi = WeightedMutualInformation(i, py, data, instances, classes, intervals, limits, labels, w);

                if (mi > best_mi) {
                    best_mi = mi;
                    best_feature = i;
                }
            }

        // Add best to selection
        subspace[best_feature] = 1;
        selected++;

        // Get accuracy
        ClassifierEvaluation(parm, data, instances, labels, data, instances, labels, subspace, features, NULL, parm->cv_hyperparameters, classes, parm->wrapper_fs_classifier, &learner, 0,
                                   M, NULL, &perf, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

        scalar acc = perf.accuracy;

        // Obtain margin per instance and update weights accordingly
        ClassifierDatasetMargin(parm, parm->wrapper_fs_classifier, data, instances, labels, data, instances, labels, features, classes, subspace, &min_margin, &learner, margin, NULL, &ensemble,
                                      !IsEnsemble(parm->wrapper_fs_classifier));

        ClassifierFree(parm, &learner, parm->wrapper_fs_classifier, classes, features, &ensemble);

        scalar max_margin = VectorMaxValue(margin, instances);
        for (int i = 0; i < instances; i++)
            w[i] = fabs(margin[i] - max_margin);

        // Check stop criterion
        if (previous_acc*(parm->wmi_epsilon+1.0) >= acc || selected == features)
            finish = true;
        else
            previous_acc = acc;

        SHOWPROGRESS(parm->show_progress > 0, "wMI Selected feature: %*d with current accuracy: %lf\n", WIDTH(features), best_feature+1, acc);
    } while (!finish);

    FREE(py);
    FREE(w);
    FREE(margin);
    FREE(ok);
    FREE(intervals);
    for (int i = 0; i < features; i++)
        FREE(limits[i]);
    FREE(limits);

    FreeIntegerMatrix(&M, classes);

    return selected;
}

// Return the maximum of the vector
scalar VectorMaxValue(scalar *v, int n)
{
    scalar max = v[0];

    for (int i = 1; i < n; i++)
        if (v[i] > max)
            max = v[i];

    return max;
}
