#!      /usr/bin/python3

import networkx as nx
import argparse
import re
from sklearn.cluster import SpectralClustering
from sklearn import metrics
import numpy as np
import scipy
from scipy.sparse import csgraph
# from scipy.sparse.linalg import eigsh
from numpy import linalg as LA
import community

import matplotlib
matplotlib.use('PDF')
import matplotlib.pyplot as plt

def eigenDecomposition(A, plot = True, topK = 5):
    """
    :param A: Affinity matrix
    :param plot: plots the sorted eigen values for visual inspection
    :return A tuple containing:
    - the optimal number of clusters by eigengap heuristic
    - all eigen values
    - all eigen vectors
    
    This method performs the eigen decomposition on a given affinity matrix,
    following the steps recommended in the paper:

    2. Find the eigenvalues and their associated eigen vectors
    3. Identify the maximum gap which corresponds to the number of clusters
    by eigengap heuristic
    
    References:
    https://papers.nips.cc/paper/2619-self-tuning-spectral-clustering.pdf
    http://www.kyb.mpg.de/fileadmin/user_upload/files/publications/attachments/Luxburg07_tutorial_4488%5b0%5d.pdf
    """
    L = csgraph.laplacian(A, normed=True)
    n_components = A.shape[0]
    
    # LM parameter : Eigenvalues with largest magnitude (eigs, eigsh), that is, largest eigenvalues in 
    # the euclidean norm of complex numbers.
    # eigenvalues, eigenvectors = eigsh(L, k=n_components, which="LM", sigma=1.0, maxiter=5000)
    eigenvalues, eigenvectors = LA.eig(L)

    eigenvalues = np.sort(eigenvalues)[::-1]
    np.set_printoptions(suppress=True)
    
    if plot:
        plt.title('Largest eigen values of input matrix')
        plt.scatter(np.arange(len(eigenvalues)), np.sort(eigenvalues)[::-1])
        plt.grid()
        plt.savefig("eigenvalues.png")
        
    # Identify the optimal number of clusters as the index corresponding
    # to the larger gap between eigen values
    index_largest_gap = np.argsort(np.diff(eigenvalues))[::][:topK]
    nb_clusters = index_largest_gap + 1
        
    return nb_clusters, eigenvalues, eigenvectors

# Positional arguments
parser = argparse.ArgumentParser()
parser.add_argument("infile", help="Graph description file", type=str)
parser.add_argument("outfile", help="Output file)", type=str)
parser.add_argument("-r", help="Precision of labels", type=int, default=0)

args = parser.parse_args()

# Read file
# Format
# Node, Node, Weight

gfile = open(args.infile, 'r')
nodes = gfile.readlines()
gfile.close()

# Create graph
g = nx.Graph()
for node in nodes:
    # Parse node
    parsed = node.strip().replace(' ','').split(',')
    g.add_edge(parsed[0], parsed[1], weight='%.*lf' % (args.r, float(parsed[2].replace("!",""))), label=parsed[2]) #  label='%.*lf' % (args.r, float(parsed[2])))

outf = open(args.outfile, "w")

n_connected = nx.number_connected_components(g)

for G, j in zip(nx.connected_component_subgraphs(g), range(0, n_connected)):
    adj_mat = nx.to_numpy_matrix(G)
    nodes = [int(i) for i in G.nodes]
    if G.number_of_nodes() > 1:
        G.add_edges_from(nx.non_edges(G), weight=0) # Complete graph with 0 edges 
    
        print("Clustering connected subgraph %d/%d" % (j+1, n_connected))
        # Optimal number
        k, _,  _ = eigenDecomposition(adj_mat, True, 1)
        print('Optimal number of clusters: %d' % k[0])

        # Clustering
        sc = SpectralClustering(k[0], affinity='precomputed', n_init=100)
        sc.fit(adj_mat)

        # Print clustering-results
        for i in range(0, k[0]):
            outf.write("Cluster: ")
            for j in range(0, len(sc.labels_)):
                if sc.labels_[j] == i:
                    outf.write("%d " % nodes[j])
            outf.write("\n")

outf.close()
