/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <omp.h>
#include "typex.h"
#include "common.h"
#include "macros.h"
#include "imbalance.h"
#include "data.h"
#include "classifiers.h"

// Stub for class-imbalanced methods

void ClassImbalanceMethodFor2ClassNoStratified(const t_Parameters *parm, scalar ***data, int *instances, int features, int classes, int **labels)
{
    extern char *class_imbalance_method[];
    char command[1000];
    int evaluations, generations_performed;

    if (classes != 2)
        OnErrorExit(ERR_PARAM, "Class-imbalance method not suitable for more than 2 classes");

    int previous_instances = *instances;

    time_t init_time = time(NULL);

    switch (parm->imbalance_method) {
        case iNone:
            // Do nothing
            break;

        case iOversampling:
            OversampleData(data, instances, features, classes, labels, parm->oversampling_ratio);
            break;

        case iUndersampling:
            UndersampleData(data, instances, features, classes, labels, parm->random_undersampling_ratio);
            break;

        case iSMOTE:
            SMOTEOverSampling(data, instances, features, classes, labels, parm->smote_N, parm->smote_k);
            fprintf(parm->out_file, "@SMOTE added instances: %d\n", *instances-previous_instances);
            fprintf(parm->out_file, "@SMOTE oversampling: %lf\n", ((scalar) *instances-previous_instances)/previous_instances);
            break;

        case iEvoSMOTE:
            EvolutionarySMOTE(parm, data, instances, NULL, features, classes, labels, NULL, &evaluations, &generations_performed);
            SAVEPROGRESS(true, "@Actual generations performed: %d\n", generations_performed+1);
            break;

        case iROSE:
            RunExternalCIMethod("rose.py", data, labels, instances, features, classes);
            break;

        case iADASYN:
            RunExternalCIMethod("adasyn.py", data, labels, instances, features, classes);
            break;

        case iBorderlineSMOTE:
            RunExternalCIMethod("border.py", data, labels, instances, features, classes);
            break;

        case iKMeansSMOTE:
            RunExternalCIMethod("kmeans.py", data, labels, instances, features, classes);
            break;

        case iSVMSMOTE:
            RunExternalCIMethod("svm.py", data, labels, instances, features, classes);
            break;

        case iSMOTETomek:
            RunExternalCIMethod("smotetomek.py", data, labels, instances, features, classes);
            break;

        case iSMOTEENN:
            RunExternalCIMethod("smoteenn.py", data, labels, instances, features, classes);
            break;

        case iIHT:
            RunExternalCIMethod("iht.py", data, labels, instances, features, classes);
            break;

        case iNCR:
            RunExternalCIMethod("ncr.py", data, labels, instances, features, classes);
            break;

        case iNearMiss:
            RunExternalCIMethod("nearmiss.py", data, labels, instances, features, classes);
            break;

        case iSMOTEVariant:
            sprintf(command, "variants.py %d", (int) parm->smote_variant);
            RunExternalCIMethod(command, data, labels, instances, features, classes);
            break;
    }

    fprintf(parm->out_file, "CI elapsed time: %ld\n", time(NULL) - init_time);
    fprintf(parm->out_file, "CI %s method: %d -> %d instances\n", class_imbalance_method[(int) parm->imbalance_method], previous_instances, *instances);
    fprintf(parm->out_file, "@CI method reduction: %lf\n", 1.0 - ((scalar) *instances)/previous_instances);
}


// Save file using ARFF format
void SaveDatasetAsARFFForSampling(char *out_file_name, int inputs, scalar **data, int instances, int **index, int *labels, int classes)
{
    FILE *f;

    if ((f = fopen(out_file_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open dataset file");

    // Relation name
    fprintf(f, "@relation TRAINING\n\n");

    // Attributes
    for (int i = 0; i < inputs; i++)
        fprintf(f, "@attribute INPUT%d numeric\n", i+1);

    // Labels
    fprintf(f, "@attribute LABELS {0");
    for (int i = 1; i < classes-1; i++)
        fprintf(f, ", %d", i);
    fprintf(f, ", %d}\n", classes-1);

    // Data
    fprintf(f, "\n@data\n");

    // Sparse or dense format?
    if (index) {
        // Variable index begins with 0
        for (int i = 0; i < instances; i++) {

            fprintf(f,"{");
            // Attributes
            int j = 0;
            while (index[i][j] != -1) {
                fprintf(f, "%d %g%s", index[i][j]-1, data[i][j], index[i][j+1] == -1 ? " " : ", ");
                j++;
            }

            // Label
            fprintf(f, ", %d %d}\n", classes+inputs, labels[i]);
        }
    } else {
        for (int i = 0; i < instances; i++) {
            // Number of repetitions or presence

            //  Attributes
            for (int j = 0; j < inputs; j++)
                fprintf(f, "%g, ", data[i][j]);

            fprintf(f, "%d\n", labels[i]);
        }
    }

    fclose(f);
}

#define MAX_COMMAND_LENGTH  10000
void RunExternalCIMethod(char *program, scalar ***data, int **labels, int *instances, int features, int classes)
{
    char *tempnam(const char *dir, const char *pfx);

    char command[MAX_COMMAND_LENGTH], *file_template;
    int old_instances = *instances;

    file_template = tempnam(".", NULL);
    SaveDatasetAsARFFForSampling(file_template, features, *data, *instances, NULL, *labels, classes);

    sprintf(command, "%s %s", program, file_template);
    int status = system(command);
    if (!WEXITSTATUS(status)) {
        unlink(file_template);
        printf("Failed command: %s\n", command);
        OnErrorExit(ERR_UNSPECIFIED, "Python class-imbalance method failed");
    }

    // Free previous data
    for (int i = 0; i < old_instances; i++)
        FREE((*data)[i]);
    FREE(*labels);

    // Read new data
    ReadDataAndLabels(data, labels, instances, &features, &classes, file_template, false);
    unlink(file_template);
}

// Wrapper for CI methods considering stratification

void ClassImbalanceMethodFor2Class(const t_Parameters *parm, scalar ***data, int *instances, int features, int classes, int **labels)
{
    // If k-NN and k = 0 get best CV k for the evaluation
    if (parm->imbalance_method == iEvoSMOTE && parm->evosmote_wrapper_eval_interval && parm->chc_wrapper == kNN && parm->k == 0) {
        int count[] = {0, 0};
        for (int i = 0; i < *instances; i++)
            count[(*labels)[i]]++;
        int minority = (count[0] > count[1]) ? 1 : 0;

        CrossValidatedParameterskNN((t_Parameters *) parm, *data, *instances, *labels, NULL, features, classes, minority, NULL, NULL);
        fprintf(parm->out_file, "&EvoSMOTE k: %d\n", parm->k);
    }

    if (parm->external_stratified_approach && parm->external_subset_size < *instances) {
        // Stratified partition for evaluating the performance
        int sets = 0, *set_instances = NULL, **set_labels = NULL, **set_list = NULL;
        scalar ***set_data = NULL;

        sets = (int) ceil(((scalar) *instances)/parm->external_subset_size);
        MALLOC(set_data, scalar **, sets, "ClassImbalanceMethodFor2Class: set_data");
        MALLOC(set_labels, int *, sets, "ClassImbalanceMethodFor2Class: set_labels");
        MALLOC(set_list, int *, sets, "ClassImbalanceMethodFor2Class: set_list");
        MALLOC(set_instances, int, sets, "ClassImbalanceMethodFor2Class: set_instances");
        for (int i = 0; i < sets; i++) {
            MALLOC(set_data[i], scalar *, parm->external_subset_size, "ClassImbalanceMethodFor2Class: set_data[i]");
            MALLOC(set_labels[i], int, parm->external_subset_size, "ClassImbalanceMethodFor2Class: set_labels[i]");
            MALLOC(set_list[i], int, parm->external_subset_size, "ClassImbalanceMethodFor2Class: set_labels[i]");
        }
        DataPartition(parm, *data, *instances, *labels, NULL, &sets, &set_data, &set_instances, &set_labels, classes, &set_list, NULL, NULL, features, parm->external_subset_size, pRANDOM, true, NULL);

        // Repeat for every stratum
        scalar **new_data = NULL;
        int new_instances = 0, *new_labels = NULL;

#ifdef _OPENMP
        // Some methods cannot be run concurrently
        int old = omp_get_max_threads();
        if (parm->imbalance_method == iEvoSMOTE && parm->chc_wrapper == C45) {
            printf("WARNING: Number of OPENMP threads set to 1 (<-- %d)\n", old);
            omp_set_num_threads(1);
        }
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < sets; i++) {
            ClassImbalanceMethodFor2ClassNoStratified(parm, &set_data[i], &set_instances[i], features, classes, &set_labels[i]);

#ifdef _OPENMP
            #pragma omp critical
#endif
            {
                // Add to new data
                REALLOC(new_data, scalar *, new_instances+set_instances[i], "ClassImbalanceMethodFor2ClassNoStratified: new_data");
                REALLOC(new_labels, int, new_instances+set_instances[i], "ClassImbalanceMethodFor2Class: new_labels");

                for (int j = new_instances; j < new_instances+set_instances[i]; j++) {
                    MALLOC(new_data[j], scalar, features, "ClassImbalanceMethodFor2Class: new_data[j]");
                    memcpy(new_data[j], set_data[i][j-new_instances], features*sizeof(scalar));
                    new_labels[j] = set_labels[i][j-new_instances];
                }

                new_instances += set_instances[i];
            }

            // Free data
            for (int j = 0; j < set_instances[i]; j++)
                FREE(set_data[i][j]);
            FREE(set_data[i]);
            FREE(set_labels[i]);
            FREE(set_list[i]);
        }

        // Free previous data
        FREE(*data);        // data[i] already freed
        FREE(*labels);
        FREE(set_data);
        FREE(set_labels);
        FREE(set_list);
        FREE(set_instances);

        // New data
        *data = new_data;
        *instances = new_instances;
        *labels = new_labels;

#ifdef _OPENMP
        if (parm->imbalance_method == iEvoSMOTE && parm->chc_wrapper == C45) {
            printf("WARNING: Number of OPENMP threads restored to %d\n", old);
            omp_set_num_threads(old);
        }
#endif
    } else
        ClassImbalanceMethodFor2ClassNoStratified(parm, data, instances, features, classes, labels);
}
