/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "imbalance.h"
#include "typex.h"
#include "maths.h"
#include "is.h"
#include "ml.h"

// Evolutionary operators for Evo SMOTE


void NonUniformMutationReal(scalar *individual, int n, int generation, int max_generations, double bit_mutation_rate, scalar min_value, scalar max_value)
{
    scalar delta;

    for (int i = 0; i < n; i++)
        if (RANDOM(0.0, 1.0) < bit_mutation_rate) {
            if (RANDOM(0.0, 1.0) < 0.5)
                individual[i] += (delta = DeltaNUM(generation, (double) max_value - individual[i], max_generations));
            else
                individual[i] -= (delta = DeltaNUM(generation, (double) individual[i] - min_value, max_generations));

            if (individual[i] < min_value)
                individual[i] = min_value;
            else if (individual[i] > max_value)
                individual[i] = max_value;
        }
}


void EvoSMOTECrossover(int *father, int *mother, int *son, int *daughter, int n, scalar *fatherw, scalar *motherw, scalar *sonw, scalar *daughterw, int inputs)
{
    for (int i = 0; i < n; i++)
        if (i%2) {
            son[i] = father[i];
            daughter[i] = mother[i];
            for (int j = 0; j < inputs; j++) {
                sonw[i*inputs+j] = fatherw[i*inputs+j];
                daughterw[i*inputs+j] = motherw[i+inputs+j];
            }
        } else {
            son[i] = mother[i];
            daughter[i] = father[i];
            for (int j = 0; j < inputs; j++) {
                sonw[i*inputs+j] = motherw[i*inputs+j];
                daughterw[i*inputs+j] = fatherw[i+inputs+j];
            }
        }
}


// Random undersampling of the majority class
// Keeps already undersampled instances removed
void RandomUndersamplingMutation(int *s, int instances, scalar ratio, int *labels, int majority, int *pairs, int N, int plength, int max_N)
{
    // Check current balance
    scalar Mc = 0.0, mc = 0.0;
    for (int i = 0; i < instances; i++)
        if (s[i]) {
            if (labels[i] == majority)
                Mc++;
            else
                mc++;
        }

    // Equivalent to obtain "added" without actually adding instances
    int mcount = plength/max_N;
    for (int i = 0; i < mcount; i++)
        for (int n = 0; n < N; n++)
            if (pairs[i*max_N+n] >= 0 && s[instances+(i*max_N+n)])
                mc++;

    // Remove probability to desired ratio
    scalar prob_remove = 1.0 - (mc/Mc)*ratio;

    // If ratio below desired do nothing
    if (prob_remove <= 0.0)
        return;

    /* // If ratio below desired add instances
    if (prob_remove < 0.0) {
        scalar prob_add = (ratio*mc-Mc)/(instances-Mc-mc); // unselected = instances - Mc - mc, needed = ratio*mc - Mc

        for (int i = 0; i < instances; i++)
            if (labels[i] == majority && !s[i] && RANDOM(0.0, 1.0) < prob_add)
                s[i] = 1;
    } else */

    for (int i = 0; i < instances; i++)
        if (labels[i] == majority && s[i] && RANDOM(0.0, 1.0) < prob_remove)
            s[i] = 0;
}

void BLXAlphaCrossoverReal(scalar *son, scalar *daughter, scalar *father, scalar *mother, int size, scalar min_value, scalar max_value)
{
    // BLX-alpha crossover adapted to integers
    for (int j = 0; j < size; j++) {
        scalar min = MIN(father[j], mother[j]);
        scalar max = MAX(father[j], mother[j]);
        scalar I = max - min;
        son[j] = AssignWithinBoundsScalar(min_value, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), max_value);
        daughter[j] = AssignWithinBoundsScalar(min_value, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), max_value);
    }
}

// Following [Dongxia and Yujian, 2022]
// Remove minority class samples wiht all k neighbors from majority class

int NoisyMinorityClassInstancesMutation(scalar **data, int instances, int *labels, int minority, int k, int *s, int *subspace, int inputs, int **index, int min_count)
{
    int neighbors[k], *to_remove, removed = 0;

    MALLOC(to_remove, int, min_count, "NoisyMinorityClassInstancesMutation: to_remove");

    for (int i = 0; i < instances; i++)
        if (s[i] && labels[i] == minority) {
            // Get neighbors
            GetkNeighborsForSMOTE(data[i], data, subspace, instances, k, inputs, labels, neighbors, index, index ? index[i] : NULL, false, minority, NULL, s);

            bool all_maj = true;
            for (int j = 0; j < k && neighbors[j] != -1 && all_maj; j++)
                if (labels[neighbors[j]] == minority)
                    all_maj =false;

            if (all_maj) {
                to_remove[removed] = i;
                removed++;
            }
        }

    for (int i = 0; i < removed; i++)
        s[to_remove[i]] = 0;

    FREE(to_remove);

    return removed;
}


// When distance is above a given threshold returns infinite

scalar EuclideanDistanceWithMinimum(scalar *v, scalar *u, int n, scalar minimum_distance)
{
    scalar dist = 0.0, m = minimum_distance*minimum_distance;

    for (int i = 0; i < n && dist < m; i++)
        dist += (v[i] - u[i]) * (v[i] - u[i]);

    return dist <= minimum_distance ? sqrt(dist) : INFINITE;
}
