/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "imbalance.h"
#include "typex.h"
#include "maths.h"
#include "is.h"
#include "ml.h"
#include "ensemble.h"

#undef WITH_CACHE

// Evolutionary oversampling minority class following SMOTE method [Chawla, 2002]

void EvolutionarySMOTE(const t_Parameters *parm, scalar ***data, int *instances, int ***index, int inputs, int classes, int **labels, int *subspace, int *evaluations, int *generations_performed)
{
    scalar *fitness, *reduction, *performance, *balance, delta_rm = 0.0, delta_um = 0.0, delta_om = 0.0, delta_cm = 0.0, delta_im = 0.0, delta_nm = 0.0, p_fitness;
    int c_rm = 0, c_um = 0, c_om = 0, c_cm = 0, c_im = 0, c_nm = 0, reinit_attempts = 0;
    bool *elite;

    // Check parameters
    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evo SMOTE for sparse data");

    // WARNING
    if (parm->auroc_w)
        printf("WARNING: With auROC in fitness no other metric is considered");

    // Re-scale data
    int count[] = {0, 0};

    // Count instances
    for (int i = 0; i < *instances; i++)
        count[(*labels)[i]]++;

    int minority = (count[0] > count[1]) ? 1 : 0;
    int majority = (minority+1) % 2;

    fprintf(parm->out_file, "Instances: %d, %d\n", count[minority], count[majority]);

    // If minority less than threshold use also majority class instances
    if (count[minority] <= parm->evosmote_use_majority_threshold)
        ((t_Parameters *) parm)->evosmote_only_minority = false;

    // If minimum N is 0 the adjust to balanced dataset, incompatible with undersampling
    int min_N = parm->evosmote_min_N, max_N = parm->evosmote_max_N;
    if (parm->evosmote_min_N == 0) {
        if (parm->smote_undersampling_ratio > 0.0)
            OnErrorExit(ERR_PARAM, "EvoSMOTE, balanced dataset and undersampling incompatible");
        min_N = (int) floor(((scalar) count[majority])/count[minority]);
        max_N = (int) ceil(((scalar) count[majority])/count[minority]);
    } else if (parm->evosmote_max_N <= 0)   // If negative absolute value is minimum
        // Adjust maximum N to balanced classes
        max_N = MAX((int) ceil(((scalar) count[majority])/count[minority]), abs(parm->evosmote_max_N));

    fprintf(parm->out_file, "N: [%d, %d]\n", min_N, max_N);

    int plength = count[minority] * max_N;
    int wlength = plength * inputs;

    REALLOC(*labels, int, *instances+plength, "EvolutionarySMOTE: *labels");
    REALLOC(*data, scalar *, *instances+plength, "EvolutionarySMOTE: *data");
    for (int i = *instances; i < *instances+plength; i++)
        MALLOC((*data)[i], scalar, inputs, "EvolutionarySMOTE: (*data)[i]");

    // List of minority instances
    int *minority_list, *majority_list, M = 0, m = 0;
    MALLOC(minority_list, int, count[minority], "OversampleData: minority_list");
    MALLOC(majority_list, int, count[majority], "OversampleData: minority_list");
    for (int i = 0; i < *instances; i++)
        if ((*labels)[i] == minority) {
            minority_list[m] = i;
            m++;
        } else {
            majority_list[M] = i;
            M++;
        }

    // Neighbors to consider
    int n_neighbors = 0, **neighbors;
    MALLOC(neighbors, int *, count[minority], "EvolutionarySMOTE: neighbors");

    // Use only minority class for SMOTE?
    if (parm->evosmote_only_minority)
        n_neighbors = (parm->smote_k == 0) ? count[minority]-1 : MIN(count[minority]-1, parm->smote_k);
    else
        n_neighbors = (parm->smote_k == 0) ? *instances-1 : MIN(*instances-1, parm->smote_k);

    // Cache of neighbors?
    cache_t **neighbors_cache = NULL;

#ifdef WITH_CACHE
    scalar **distance_cache = NULL;
    int neighbors_cache_size = 0;
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(*instances) < MAXIMUM_CACHE_SIZE) {
        SHOWPROGRESS(parm->show_progress > 0, "Cache size (maximum size): %.1lfGb (%.1lfGb)\n", NEIGHBORS_CACHE_SIZE(*instances), MAXIMUM_CACHE_SIZE);
        ConstructAllNeighborsCache(*data, *instances, inputs, &distance_cache, &neighbors_cache, NULL, index ? *index : NULL, &neighbors_cache_size, subspace, parm->LOO_for_knn_performance);

        // Resize cache for oversampled instances
        REALLOC(neighbors_cache, cache_t *, *instances+plength, "EvolutionarySMOTE: neighbors_cache");
        for (int i = 0; i < *instances; i++)
            REALLOC(neighbors_cache[i], cache_t, *instances+plength, "EvolutionarySMOTE: neighbors_cache[i]");
        for (int i = *instances; i < *instances+plength; i++)
            MALLOC(neighbors_cache[i], cache_t, *instances+plength, "EvolutionarySMOTE: neighbors_cache[i]");
    }
#endif

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < count[minority]; i++) {
        MALLOC(neighbors[i], int, n_neighbors, "EvolutionarySMOTE: neighbors[i]");
        GetkNeighborsForSMOTE((*data)[minority_list[i]], *data, subspace, *instances, n_neighbors, inputs, *labels, neighbors[i], index ? *index : NULL, index ? (*index)[minority_list[i]] :
                              NULL, parm->evosmote_only_minority, minority, neighbors_cache ? neighbors_cache[minority_list[i]] : NULL, NULL);
    }

    // Create initial population
    // Population is composed of four parts:
    // pair: Neighbors of instance to mate
    // weight: Weight of the combination of each mating
    // subset: Possible undersampling if SMOTE ratio is > 0
    // N: number of copies of each instance
    int **pair, **subset, *N;
    scalar **weight;
    MALLOC(pair, int *, 2*parm->population_size, "EvolutionarySMOTE: pair");
    MALLOC(subset, int *, 2*parm->population_size, "EvolutionarySMOTE: pair");
    MALLOC(N, int, 2*parm->population_size, "EvolutionarySMOTE: N");
    MALLOC(weight, scalar *, 2*parm->population_size, "EvolutionarySMOTE: weight");
    MALLOC(fitness, scalar, 2*parm->population_size, "EvolutionarySMOTE: fitness");
    MALLOC(balance, scalar, 2*parm->population_size, "EvolutionarySMOTE: balance");
    MALLOC(performance, scalar, 2*parm->population_size, "EvolutionarySMOTE: performance");
    MALLOC(reduction, scalar, 2*parm->population_size, "EvolutionarySMOTE: reduction");
    MALLOC(elite, bool, 2*parm->population_size, "EvolutionarySMOTE: elite");

    // Initial population
    for (int i = 0; i < 2*parm->population_size; i++) {
        MALLOC(pair[i], int, plength, "EvolutionarySMOTE: pair[i]");
        MALLOC(subset[i], int, *instances+plength, "EvolutionarySMOTE: subset[i]");
        MALLOC(weight[i], scalar, wlength, "EvolutionarySMOTE: weight[i]");
        elite[i] = false;
    }

    InitializePopulation(parm, pair, subset, weight, N, count[minority], count[majority], *instances, n_neighbors, inputs, subspace, neighbors, *data, plength, wlength, minority_list,
                         majority_list, parm->population_size, min_N, max_N);

    // Needed for evaluation
    int *orig_nns;
    scalar *orig_dist;
    MALLOC(orig_nns, int, *instances+plength, "EvolutionarySMOTE: orig_nns");
    MALLOC(orig_dist, scalar, *instances+plength, "EvolutionarySMOTE: orig_dist");
    orig_dist[0] = -1.0;    // Force first calculation

    // Evaluate initial population
    time_t init_time  = time(NULL);
    for (int k = 0; k < parm->population_size; k++) {
        SHOWPROGRESS(parm->show_progress > 0, "Evaluating initial population: %*d/%d\r", WIDTH(parm->population_size), k+1, parm->population_size);
        fitness[k] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[k], weight[k], inputs, neighbors, classes, neighbors_cache, subspace, subset[k],
                                             index ? *index : NULL, &reduction[k], orig_nns, orig_dist, N[k], 0, &performance[k], max_N, &balance[k]);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    *evaluations = parm->population_size;

    // Main loop of CHC algorithm
    SHOWPROGRESS(parm->show_progress > 1, "Generations: %d/%0*d/%0*d", parm->generations, WIDTH(parm->generations), 0, WIDTH(parm->generations), parm->generations);
    int *list;
    MALLOC(list, int, parm->population_size, "EvolutionarySMOTE: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;

    // Just in case for mutation
    int *s;
    MALLOC(s, int, *instances+plength, "EvolutionarySMOTE: s");

    scalar best_fitness = 0.0;
    int best_fitness_gs = 0;

    for (*generations_performed = 0; *generations_performed < parm->generations; (*generations_performed)++) {
        // CHC next generation
        int actual_size = parm->population_size, order, father, mother;;

        // Check evolutionary scheme
        switch (parm->evolutionary_scheme) {
            case schemeCHCGA:
                RandomizeVector(list, parm->population_size);
                for (int i = 0; i < parm->population_size; i += 2) {
                    // Random selection without replacement
                    father = list[i];
                    mother = list[i+1];

                    // Two crossovers are used, one more conservative
                    if (RANDOM(0.0, 1.0) < 0.5) {
                        HuxCrossover(pair[mother], pair[father], pair[actual_size], pair[actual_size+1], plength);
                        BLXAlphaCrossoverReal(weight[actual_size], weight[actual_size+1], weight[father], weight[mother], wlength, 0.0, 1.0);
                    } else
                        EvoSMOTECrossover(pair[mother], pair[father], pair[actual_size], pair[actual_size+1], plength, weight[mother], weight[father], weight[actual_size], weight[actual_size+1],
                                          inputs);

                    if (parm->biased_hux_ratio > 0.0)
                        BiasedHuxCrossover(subset[mother], subset[father], subset[actual_size], subset[actual_size+1], *instances+plength, parm->biased_hux_ratio);
                    else
                        HuxCrossover(subset[mother], subset[father], subset[actual_size], subset[actual_size+1], *instances+plength);

                    if (RANDOM(0.0, 1.0) < 0.5)
                        BLXAlphaCrossover(&N[actual_size], &N[actual_size+1], &N[father], &N[mother], 1, min_N, max_N);
                    else
                        HuxCrossover(&N[mother], &N[father], &N[actual_size], &N[actual_size+1], 1);

                    // Two new individuals in the population.
                    actual_size += 2;
                }

                break;

            case schemeSSGA:
            case schemeSSMA:
            case schemeGGA:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for Evo SMOTE");
        }

        // Evaluate new individuals
        for (int k = parm->population_size; k < actual_size; k++) {
            SHOWPROGRESS(parm->show_progress > 0, "Evaluating new population: %*d/%d\r", WIDTH(actual_size), k+1, actual_size);
            fitness[k] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[k], weight[k], inputs, neighbors, classes, neighbors_cache, subspace,
                                                 subset[k], index ? *index : NULL, &reduction[k], orig_nns, orig_dist, N[k], *generations_performed, &performance[k], max_N, &balance[k]);
        }
        SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
        *evaluations += actual_size - parm->population_size;

        // Elitism?
        if (parm->elite > 0.0) {
            for (int i = 0; i < actual_size; i++)
                elite[i] = false;
            for (int i = 0; i < parm->elite * parm->population_size; i++)
                elite[OrderedSelection(fitness, i, actual_size)] = true;
        }

        // Random mutations
        if (parm->mutation > ZERO)
            for (int i = 0; i < actual_size; i++) {
                if (elite[i])
                    continue;

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    NonUniformMutation(pair[i], plength, *generations_performed, parm->generations, parm->bit_mutation, parm->symmetric_selection ? -n_neighbors : 0, n_neighbors);
                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation && parm->smote_undersampling_ratio > 0) {
                    NonUniformMutation(subset[i], *instances+plength, *generations_performed, parm->generations, parm->bit_mutation, 0, 1);
                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    NonUniformMutationReal(weight[i], wlength, *generations_performed, parm->generations, parm->bit_mutation, 0.0, 1.0);
                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    NonUniformMutation(&N[i], 1, *generations_performed, parm->generations, 1.0, min_N, max_N);
                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                }
            }

        // Sampling mutations
        if (parm->evosmote_undersamplig_mutation_rate > 0.0 || parm->evosmote_oss_mutation_rate > 0.0 || parm->evosmote_cnntl_mutation_rate > 0.0 || parm->is_mutation > 0.0 ||
            parm->evosmote_noise_mutation_rate > 0.0)
            for (int i = 0; i < actual_size; i++) {
                if (elite[i])
                    continue;

                if (parm->smote_undersampling_ratio > 0.0 && RANDOM(0.0, 1.0) < parm->evosmote_undersamplig_mutation_rate) {
                    RandomUndersamplingMutation(subset[i], *instances, parm->smote_undersampling_ratio, *labels, majority, pair[i], N[i], plength, max_N);
                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_um += fitness[i] - p_fitness;
                    c_um++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }


                if (RANDOM(0.0, 1.0) < parm->evosmote_oss_mutation_rate) {
                    int added = AddInstances(*data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], N[i], inputs, neighbors, max_N, subset[i], parm->evosmote_one_w);

                    // Current selection
                    memcpy(s, subset[i], sizeof(int)*(*instances));
                    for (int j = 0; j < added; j++)
                        s[*instances+j] = 1;    // If added subset[j] must be 1

                    OneSidedSelection(s, *data, *instances+added, *labels, inputs, minority, parm->k, classes, index ? *index : NULL, neighbors_cache, true);

                    // Copy selection performed by the algorithm
                    memcpy(subset[i], s, sizeof(int)*(*instances));
                    int next = 0;
                    for (int j = 0; j < count[minority]; j++)
                        for (int n = 0; n < N[i]; n++)
                            // If negative do not add instance
                            if (pair[i][j*max_N+n] >= 0 && subset[i][*instances+(j*max_N+n)]) {
                                subset[i][*instances+(j*max_N+n)] = s[*instances+next];
                                next++;
                            }
                    if (next != added)
                        OnErrorExit(ERR_UNSPECTED_RESULT, "Added and OSS mutation do not match");

                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_om += fitness[i] - p_fitness;
                    c_om++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                // CNNTL mutation
                if (RANDOM(0.0, 1.0) < parm->evosmote_cnntl_mutation_rate) {
                    int added = AddInstances(*data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], N[i], inputs, neighbors, max_N, subset[i], parm->evosmote_one_w);

                    // Current selection
                    memcpy(s, subset[i], sizeof(int)*(*instances));
                    for (int j = 0; j < added; j++)
                        s[*instances+j] = 1;    // If added subset[j] must be 1

                    CNNAndTomekLinks(s, *data, *instances+added, *labels, inputs, minority, parm->k, classes, index ? *index : NULL);

                    // Copy selection performed by the algorithm
                    memcpy(subset[i], s, sizeof(int)*(*instances));
                    int next = 0;
                    for (int j = 0; j < count[minority]; j++)
                        for (int n = 0; n < N[i]; n++)
                            // If negative do not add instance
                            if (pair[i][j*max_N+n] >= 0 && subset[i][*instances+(j*max_N+n)]) {
                                subset[i][*instances+(j*max_N+n)] = s[*instances+next];
                                next++;
                            }
                    if (next != added)
                        OnErrorExit(ERR_UNSPECTED_RESULT, "Added and OSS mutation do not match");

                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N, &balance[i]);
                    delta_cm += fitness[i] - p_fitness;
                    c_cm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                }

                // IS mutation
                if (RANDOM(0.0, 1.0) < parm->is_mutation) {
                    int added = AddInstances(*data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], N[i], inputs, neighbors, max_N, subset[i], parm->evosmote_one_w);

                    // Current selection
                    memcpy(s, subset[i], sizeof(int)*(*instances));
                    for (int j = 0; j < added; j++)
                        s[*instances+j] = 1;    // If added subset[j] must be 1

                    int generations;
                    scalar evaluations;
                    *instances += added;
                    InstanceSelectionAlgorithm(parm, parm->isalgorithm, &s, data, instances, inputs, labels, classes, index ? *index : NULL, NULL, NULL, minority, &generations, parm->randomsel_ratio,
                                               NULL, NULL, &evaluations, NULL, true, NULL, false, NULL);
                    *instances -= added;

                    // Copy selection performed by the algorithm
                    memcpy(subset[i], s, sizeof(int)*(*instances));
                    int next = 0;
                    for (int j = 0; j < count[minority]; j++)
                        for (int n = 0; n < N[i]; n++)
                            // If negative do not add instance
                            if (pair[i][j*max_N+n] >= 0 && subset[i][*instances+(j*max_N+n)]) {
                                subset[i][*instances+(j*max_N+n)] = s[*instances+next];
                                next++;
                            }
                    if (next != added)
                        OnErrorExit(ERR_UNSPECTED_RESULT, "Added and OSS mutation do not match");

                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N,
                                                         &balance[i]);
                    delta_im += fitness[i] - p_fitness;
                    c_im++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                // IS mutation
                if (RANDOM(0.0, 1.0) < parm->evosmote_noise_mutation_rate) {
                    int added = AddInstances(*data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], N[i], inputs, neighbors, max_N, subset[i], parm->evosmote_one_w);

                    // Current selection
                    memcpy(s, subset[i], sizeof(int)*(*instances));
                    for (int j = 0; j < added; j++)
                        s[*instances+j] = 1;    // If added subset[j] must be 1

                    int removed = NoisyMinorityClassInstancesMutation(*data, *instances+added, *labels, minority, parm->evosmote_noise_mutation_k, s, subspace, inputs, index ? *index : NULL,
                                  count[minority]);
                    SHOWPROGRESS(parm->show_progress > 1, "Noise mutation removed %d minority class instances\n", removed);

                    // Copy selection performed by the algorithm
                    memcpy(subset[i], s, sizeof(int)*(*instances));
                    int next = 0;
                    for (int j = 0; j < count[minority]; j++)
                        for (int n = 0; n < N[i]; n++)
                            // If negative do not add instance
                            if (pair[i][j*max_N+n] >= 0 && subset[i][*instances+(j*max_N+n)]) {
                                subset[i][*instances+(j*max_N+n)] = s[*instances+next];
                                next++;
                            }
                    if (next != added)
                        OnErrorExit(ERR_UNSPECTED_RESULT, "Added and OSS mutation do not match");

                    p_fitness = fitness[i];
                    fitness[i] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[i], weight[i], inputs, neighbors, classes, neighbors_cache, subspace,
                                                         subset[i], index ? *index : NULL, &reduction[i], orig_nns, orig_dist, N[i], *generations_performed, &performance[i], max_N,
                                                         &balance[i]);
                    delta_nm += fitness[i] - p_fitness;
                    c_nm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                }
            }

        // Obtain new population selecting best N individuals
        switch (parm->evolutionary_scheme) {
            case schemeCHCGA:
                for (int i = 0; i < parm->population_size; i++) {
                    order = OrderedSelection(fitness+i, 0, actual_size - i)+i;
                    int *aux_pt = pair[order];
                    pair[order] = pair[i];
                    pair[i] = aux_pt;
                    aux_pt = subset[order];
                    subset[order] = subset[i];
                    subset[i] = aux_pt;
                    scalar *aux_sc = weight[order];
                    weight[order] = weight[i];
                    weight[i] = aux_sc;
                    int aux_i = N[order];
                    N[order] = N[i];
                    N[i] = aux_i;
                    scalar aux_f = fitness[order];
                    fitness[order] = fitness[i];
                    fitness[i] = aux_f;
                    aux_f = reduction[order];
                    reduction[order] = reduction[i];
                    reduction[i] = aux_f;
                    aux_f = performance[order];
                    performance[order] = performance[i];
                    performance[i] = aux_f;
                    aux_f = balance[order];
                    balance[order] = balance[i];
                    balance[i] = aux_f;
                }
                break;

            case schemeSSGA:
            case schemeSSMA:
            case schemeGGA:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for Evo SMOTE");
        }
        // Finished?
        if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed)) {
            fprintf(parm->out_file, "Reached limit at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
            fprintf(parm->out_file, "@Actual number of evaluations: %d\n", *evaluations);
            break;
        }

        int best = OrderedSelection(fitness, 0, parm->population_size);
        SHOWPROGRESS(parm->show_progress > 0, "%*d generation FITNESS(MEAN), PERF, RED, BAL, N, G's: %6.4lf(%6.4lf), %6.4lf, %6.4lf, %6.4lf, %2d %3d (t = %4lds)\n", WIDTH(parm->generations),
                     *generations_performed+1, fitness[best], AverageScalarVector(fitness, parm->population_size), performance[best], reduction[best], balance[best], N[best], best_fitness_gs+1,
                     time(NULL) - init_time);

        // Show mutation usefulness
        SHOWPROGRESS(parm->show_progress > 0, "MUTATIONS: Random: %7.4lf Undersampling: %7.4lf OSS: %7.4lf CNNTL: %7.4lf IS: %7.4lf Noise: %7.4lf\n", c_rm ? delta_rm/c_rm: 0.0,
                     c_um ? delta_um/c_um: 0.0, c_om ? delta_om/c_om: 0.0, c_cm ? delta_cm/c_cm: 0.0, c_im ? delta_im/c_im : 0.0, c_nm ? delta_nm/c_nm : 0.0)

        // Number of generations best is current
        if (fitness[best] == best_fitness) {
            best_fitness_gs++;

            // Reinitialize worst individuals?
            if (parm->best_individual_fitness_improvement_generations && best_fitness_gs >= parm->best_individual_fitness_improvement_generations) {
                int shift = parm->population_size * (1.0-parm->reinitalize_percent_addition);
                InitializePopulation(parm, pair+shift, subset+shift, weight+shift, N+shift, count[minority], count[majority], *instances, n_neighbors, inputs, subspace, neighbors, *data, plength,
                                     wlength, minority_list, majority_list, parm->population_size-shift, min_N, max_N);
                SHOWPROGRESS(parm->show_progress > 0, "%d worst individuals substituted\n", parm->population_size-shift);
                for (int k = shift; k < parm->population_size; k++) {
                    SHOWPROGRESS(parm->show_progress > 0, "Evaluating NEW population: %*d/%d\r", WIDTH(parm->population_size), k+1, parm->population_size);
                    fitness[k] = EvoSMOTEEvaluateFitness(parm, *data, *instances, plength, *labels, minority_list, minority, pair[k], weight[k], inputs, neighbors, classes, neighbors_cache, subspace, subset[k],
                                                         index ? *index : NULL, &reduction[k], orig_nns, orig_dist, N[k], 0, &performance[k], max_N, &balance[k]);
                }
                SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
                reinit_attempts++;

                if (parm->chc_max_reinit_attempts && (reinit_attempts > parm->chc_max_reinit_attempts)) {
                    fprintf(parm->out_file, "Reached limit of REINITILIZATIONS at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
                    fprintf(parm->out_file, "@Actual number of evaluations: %d\n", *evaluations);
                    break;
                }
            }
        } else {
            best_fitness = fitness[best];
            best_fitness_gs = 0;
        }
    }

    // Create new dataset with best individual
    int original = *instances;
    int best = OrderedSelection(fitness, 0, parm->population_size);
    int added = AddInstances(*data, *instances, plength, *labels, minority_list, minority, pair[best], weight[best], N[best], inputs, neighbors, max_N, subset[best], parm->evosmote_one_w);

    // Mutation benefits
    fprintf(parm->out_file, "Random mutation average fitness increment: %7.4lf\n", c_rm ? delta_rm/c_rm: 0.0);
    fprintf(parm->out_file, "Undersampling mutation average fitness increment: %7.4lf\n", c_um ? delta_um/c_um: 0.0);
    fprintf(parm->out_file, "OSS mutation average fitness increment: %7.4lf\n", c_om ? delta_om/c_om: 0.0);
    fprintf(parm->out_file, "CNNTL mutation average fitness increment: %7.4lf\n", c_cm ? delta_cm/c_cm: 0.0);
    fprintf(parm->out_file, "Instance selection mutation average fitness increment: %7.4lf\n", c_im ? delta_im/c_im: 0.0);
    fprintf(parm->out_file, "Noise mutation average fitness increment: %7.4lf\n", c_nm ? delta_nm/c_nm: 0.0);

    fprintf(parm->out_file, "@EvoSMOTE added instances: %d\n", added);
    fprintf(parm->out_file, "@EvoSMOTE oversampling: %lf\n", ((scalar) added)/(*instances));
    fprintf(parm->out_file, "@EvoSMOTE N: %d\n", N[best]);

    // Remove undersampled instances
    // Added instances must be kept
    for (int i = *instances; i < *instances+added; i++)
        subset[best][i] = 1;
    *instances += added;
    int removed = RemoveUnselectedInstances(data, instances, labels, index, subset[best]);
    fprintf(parm->out_file, "@Number of undersampled instances: %d\n", removed);
    fprintf(parm->out_file, "@EvoSMOTE undersampling: %lf\n", ((scalar) removed)/(*instances-added));

    fprintf(parm->out_file, "@EvoSMOTE reduction: %lf\n", 1.0 - ((scalar)(*instances))/original);

    // If  mutation rate is negative it is applied once after evolution
    if (parm->evosmote_noise_mutation_rate < 0.0) {
        NoisyMinorityClassInstancesMutation(*data, *instances, *labels, minority, parm->evosmote_noise_mutation_k, subset[best], subspace, inputs, index ? *index : NULL, count[minority]);
        removed = RemoveUnselectedInstances(data, instances, labels, index, subset[best]);
        fprintf(parm->out_file, "@Number of removed noisy instances: %d (%d -> %d)\n", removed, *instances+removed, *instances);
    }

    if (parm->evosmote_oss_mutation_rate < 0.0) {
        OneSidedSelection(subset[best], *data, *instances, *labels, inputs, minority, parm->k, classes, index ? *index : NULL, neighbors_cache, true);
        removed = RemoveUnselectedInstances(data, instances, labels, index, subset[best]);
        fprintf(parm->out_file, "@Number of removed instances by OSS: %d (%d -> %d)\n", removed, *instances+removed, *instances);
    }

    if (parm->evosmote_cnntl_mutation_rate < 0.0) {
        CNNAndTomekLinks(subset[best], *data, *instances, *labels, inputs, minority, parm->k, classes, index ? *index : NULL);
        removed = RemoveUnselectedInstances(data, instances, labels, index, subset[best]);
        fprintf(parm->out_file, "@Number of removed instances by CNNTL: %d (%d -> %d)\n", removed, *instances+removed, *instances);
    }

    if (parm->is_mutation < 0.0) {
        int generations;
        scalar evaluations;
        InstanceSelectionAlgorithm(parm, parm->isalgorithm, &subset[best], data, instances, inputs, labels, classes, index ? *index : NULL, NULL, NULL, minority, &generations, parm->randomsel_ratio,
                                   NULL, NULL, &evaluations, NULL, true, NULL, false, NULL);
        removed = RemoveUnselectedInstances(data, instances, labels, index, subset[best]);
        fprintf(parm->out_file, "@Number of removed instances by IS algorithm: %d (%d -> %d)\n", removed, *instances+removed, *instances);
    }

    if (parm->evosmote_undersamplig_mutation_rate < 0.0) {
        removed = *instances;
        UndersampleData(data, instances, inputs, classes, labels, parm->smote_undersampling_ratio);
        removed -= *instances;
        fprintf(parm->out_file, "@Number of removed instances by undersampling: %d (%d -> %d)\n", removed, *instances+removed, *instances);
    }

    FREE(list);
    FREE(s);
    for (int i = 0; i < 2*parm->population_size; i++) {
        FREE(pair[i]);
        FREE(weight[i]);
        FREE(subset[i]);
    }
    FREE(pair);
    FREE(weight);
    FREE(subset);
    FREE(fitness);
    FREE(balance);
    FREE(performance);
    FREE(reduction);
    FREE(minority_list);
    FREE(majority_list);
    FREE(elite);
    for (int i = 0; i < count[minority]; i++)
        FREE(neighbors[i]);
    FREE(neighbors);
    FREE(orig_dist);
    FREE(orig_nns);

#ifdef WITH_CACHE
    if (neighbors_cache) {
        for (int i = 0; i < *instances - plength; i++) {
            FREE(neighbors_cache[i]);
            FREE(distance_cache[i]);
        }
        for (int i = *instances-plength; i < *instances; i++)
            FREE(neighbors_cache[i]);
        FREE(distance_cache);
        FREE(neighbors_cache);
    }
#endif
}

// Add instances using an individual

#define WEIGHT(i, n, l) ((i)*max_N*inputs)+((n)*inputs+l)
int AddInstances(scalar **data, int instances, int plength, int *labels, int *list, int minority, int *pairs, scalar *weights, int N, int inputs, int **neighbors, int max_N, int *subset, bool one_w)
{
    int added = 0, f = one_w ? 0 : 1;

    // Add an instance SMOTE way
    int mcount = plength/max_N;
    for (int i = 0; i < mcount; i++)
        for (int n = 0; n < N; n++)
            // If negative do not add instance
            if (pairs[i*max_N+n] >= 0 && subset[instances+(i*max_N+n)]) {
                for (int l = 0; l < inputs; l++)
                    data[instances+added][l] = data[list[i]][l] + weights[WEIGHT(i, n, l*f)] * (data[neighbors[i][pairs[i*max_N+n]]][l] - data[list[i]][l]);
                // Labels
                labels[instances+added] = minority;

                // Instance actually added
                added++;
            }

    return added;
}

// Get k neighbors of a pattern
void GetkNeighborsForSMOTE(scalar *instance, scalar **data, int *subspace, int instances, int k, int inputs, int *labels, int *nearest, int **index, int *instance_index, bool same_class,
                           int instance_label, cache_t *nn_cache, int *s)
{
    for (int i = 0; i < k; i++)
        nearest[i] = -1;

    // Cache?
    if (nn_cache) {
        int j = 0;
        for (int i = 0; i < instances && j < k; i++)
            if ((!same_class || labels[nn_cache[i]] == instance_label) && (instance != data[nn_cache[i]])) {
                nearest[j] = nn_cache[i];
                j++;
            }

        return;
    }

    scalar dmin = INFINITE, dist[k], value;
    for (int i = 0; i < k; i++)
        dist[i] = INFINITE;

    // Get k nearest neighbors
    for (int i = 0; i < instances; i++)
        if ((!s || s[i]) && (!same_class || labels[i] == instance_label)) {
            value = EuclideanDistanceSparseDataSubspace(instance, data[i], inputs, instance_index, index ? index[i] : NULL, subspace);

            // Test if is one of the nearest
            if (value <= dmin && instance != data[i]) {
                // Look for its place
                int j = 0;
                while (value > dist[j])
                    j++;

                // Sets in j-th place
                for (int l = k -
                             1; l > j; l--) {
                    nearest[l] = nearest[l - 1];
                    dist[l] = dist[l - 1];
                }
                nearest[j] = i;
                dist[j] = value;
                dmin = dist[k - 1];
            }
        }
}

// Update cache after adding new oversampled instances

#ifdef WITH_CACHE
void UpdateCacheWithOversampledData(scalar **data, int instances, int added, int inputs, scalar **distance_cache, cache_t **neighbors_cache, int **index, int *subspace, bool loo)
{
    RANK *rank;
    MALLOC(rank, RANK, instances+added, "main: r");

    // Update neighbors of original instances
    for (int i = 0; i < instances; i++) {
        // Old instances distance already obtained
        for (int j = 0; j < instances; j++) {
            rank[j].fitness = -distance_cache[i][j];
            rank[j].being = j;
        }

        for (int j = instances; j < instances+added; j++) {
            rank[j].fitness = -EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace);
            rank[j].being = j;
        }

        rank[i].fitness = loo ? -INFINITE : 0.0; // Consider itself as neighbor?

        // Sort
        qsort(rank, instances+added, sizeof(RANK), Compare);

        // Store neighbors in ascending order of distance
        for (int j = 0; j < instances+added; j++)
            neighbors_cache[i][j] = rank[j].being;
    }

    // New instances
    for (int i = instances; i < instances+added; i++) {
        for (int j = 0; j < instances+added; j++) {
            rank[j].fitness = -EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace);
            rank[j].being = j;
        }

        rank[i].fitness = loo ? -INFINITE : 0.0; // Consider itself as neighbor?

        // Sort
        qsort(rank, instances+added, sizeof(RANK), Compare);

        // Store neighbors in ascending order of distance
        for (int j = 0; j < instances+added; j++)
            neighbors_cache[i][j] = rank[j].being;
    }

    FREE(rank);
}
#endif

void ListOfNearestNeighborsAndMinimumDistance(int *nns, scalar *dist, scalar **data, int instances, int added, int inputs, bool first_time, int *orig_nns, scalar *orig_dist,
        int **index, int *subspace, bool eval_new_instances, int *subset, int *labels, int minority)
{
    if (subspace || index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace of sparse format for Evo SMOTE");

    // First item? Obtain info for original instances only
    if (first_time) {
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int j = 0; j < instances; j++) {
            orig_dist[j] = INFINITE;

            for (int i = 0; i < instances; i++) {
                scalar value = EuclideanDistanceSparseDataSubspace(data[j], data[i], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace);

                // Test if nearest
                if (value < orig_dist[j] && i != j) {
                    orig_nns[j] = i;
                    orig_dist[j] = value;
                }
            }
        }
    } else {
        // Update nearest neighbor using new added instances
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int j = 0; j < instances; j++) {
            dist[j] = subset[orig_nns[j]] ? orig_dist[j] : INFINITE;
            nns[j] = subset[orig_nns[j]] ? orig_nns[j] : -1;

            // If current nearest is minority that cannot change
            if (nns[j] != -1 && labels[nns[j]] == minority)
                continue;

            for (int i = subset[orig_nns[j]] ? instances : 0; i < instances+added; i++)
                if (subset[i]) {
                    scalar value = EuclideanDistanceWithMinimum(data[j], data[i], inputs, dist[j]);

                    // Test if nearest
                    if (value < dist[j] && i != j) {
                        nns[j] = i;
                        dist[j] = value;
                    }

                    // If current nearest is minority and testing new instances that cannot change
                    if (labels[nns[j]] == minority && i >= instances)
                        break;
                }
        }

        // Nearest neighbor of new added instances, only if needed
        if (eval_new_instances)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int j = instances; j < instances+added; j++) {
                dist[j] = INFINITE;

                for (int i = 0; i < instances+added; i++)
                    if (subset[i]) {
                        scalar value = EuclideanDistanceWithMinimum(data[j], data[i], inputs, dist[j]);

                        // Test if nearest
                        if (value < dist[j] && i != j) {
                            nns[j] = i;
                            dist[j] = value;
                        }

                        // If current nearest is minority and testing new instances that cannot change
                        if (labels[nns[j]] == minority && i >= instances)
                            break;
                    }
            }
    }
}

scalar EvoSMOTEEvaluateFitness(const t_Parameters *parm, scalar **data, int instances, int plength, int *labels, int *minority_list, int minority, int *pair, scalar *weight, int inputs,
                               int **neighbors, int classes, cache_t **neighbors_cache, int *subspace, int *subset, int **index, scalar *reduction, int *orig_nns, scalar *orig_dist, int N,
                               int generation, scalar *performance, int max_N, scalar *balance)
{
    scalar fitness;
    int added, *nns, **M;
    scalar *dist;
    t_Learner learner;
    t_Ensemble ensemble;
    t_PerformanceInfo perf;

    added = AddInstances(data, instances, plength, labels, minority_list, minority, pair, weight, N, inputs, neighbors, max_N, subset, parm->evosmote_one_w);

    // Get subset
    int majority = (minority+1) % 2;

    // If auROC in fitness no other metric is considered
    int test_instances = parm->evosmote_eval_new_instances ? instances+added : instances;

    if (parm->auroc_w > 0.0) {
        ClassifierTraining(parm, parm->chc_wrapper, data, instances+added, labels, inputs, classes, NULL, subspace, subset, true, &learner, &ensemble);
        fitness = ClassifierAUROC(parm, parm->chc_wrapper, data, test_instances, labels, data, instances+added, subset, subspace, labels, inputs, minority, false, &learner, false, parm->out_file,
                                        classes, NULL,  NULL, NULL, NULL, NULL, !IsEnsemble(parm->chc_wrapper));
    } else {
        // If wrapper is NN the process can be faster
        AllocateIntegerMatrix(&M, classes, classes);
        if (parm->evosmote_wrapper_eval_interval && ((generation+1) % parm->evosmote_wrapper_eval_interval) == 0) {
#ifdef WITH_CACHE
            if (neighbors_cache)
                UpdateCacheWithOversampledData(*data, *instances, added, inputs, distance_cache, neighbors_cache, index ? *index : NULL, subspace, parm->LOO_for_knn_performance);
#endif

            ClassifierEvaluation(parm, data, test_instances, labels, data, instances+added, labels, subspace, inputs, subset, false, classes, parm->chc_wrapper, &learner, minority, M, NULL,
                                       &perf, NULL, NULL, NULL, NULL, neighbors_cache, index, index, &ensemble);
        } else {
            // First time?
            if (orig_dist[0] < 0.0)
                ListOfNearestNeighborsAndMinimumDistance(NULL, NULL, data, instances, added, inputs, true, orig_nns, orig_dist, index, subspace, parm->evosmote_eval_new_instances, NULL, labels,
                        minority);

            MALLOC(dist, scalar, instances+added, "EvoSMOTEEvaluateFitness: dist");
            MALLOC(nns, int, instances+added, "EvoSMOTEEvaluateFitness: nns");
            ListOfNearestNeighborsAndMinimumDistance(nns, dist, data, instances, added, inputs, false, orig_nns, orig_dist, index, subspace, parm->evosmote_eval_new_instances, subset, labels,
                    minority);

            for (int i = 0; i < classes; i++)
                for (int j = 0; j <= i; j++)
                    M[i][j] = M[j][i] = 0;

            for (int i = 0; i < test_instances; i++)
                M[labels[i]][labels[nns[i]]]++;

            FREE(dist);
            FREE(nns);
        }

        PrintTwoClassMeasures((scalar) M[minority][minority], (scalar) M[majority][majority], (scalar) M[majority][minority], (scalar) M[minority][majority], NULL, NULL, &perf);
        PrintMulticlassMeasures(M, classes, test_instances, NULL, NULL, &perf);

        FreeIntegerMatrix(&M, classes);

        if (!parm->combine_metrics_gmean) {
            fitness = parm->accuracy_w * perf.accuracy;
            fitness += parm->gmean_w * perf.gmean;
            fitness += parm->kappa_w * perf.kappa;
            fitness += parm->recall_macro_w * perf.recall_macro;
            fitness += parm->recall_macro_g_mean_w * perf.recall_macro_g_mean;
            fitness += parm->average_accuracy_w * perf.average_accuracy;
            fitness += parm->f1_w * perf.f1_score;
            fitness += parm->cc_w * perf.cc;
            fitness += parm->recall_w * perf.recall;
            fitness += parm->specificity_w * perf.specificity;
            fitness += parm->precision_w * perf.precision;
            fitness += parm->npv_w * perf.npv;

            *performance = fitness;
        } else {
            fitness = parm->accuracy_w > 0.0 ? pow(perf.accuracy, parm->accuracy_w) : 1.0;
            fitness *= parm->gmean_w > 0.0 ? pow(perf.gmean, parm->gmean_w) : 1.0;
            fitness *= parm->kappa_w > 0.0 ? pow(perf.kappa, parm->kappa_w) : 1.0;
            fitness *= parm->recall_macro_w > 0.0 ? pow(perf.recall_macro, parm->recall_macro_w) : 1.0;
            fitness *= parm->recall_macro_g_mean_w > 0.0 ? pow(perf.recall_macro_g_mean, parm->recall_macro_g_mean_w) : 1.0;
            fitness *= parm->average_accuracy_w > 0.0 ? pow(perf.average_accuracy, parm->average_accuracy_w) : 1.0;
            fitness *= parm->f1_w > 0.0 ? pow(perf.f1_score, parm->f1_w) : 1.0;
            fitness *= parm->cc_w > 0.0 ? pow(perf.cc, parm->cc_w) : 1.0;
            fitness *= parm->recall_w > 0.0 ? pow(perf.recall, parm->recall_w) : 1.0;
            fitness *= parm->specificity_w > 0.0 ? pow(perf.specificity, parm->specificity_w) : 1.0;
            fitness *= parm->precision_w > 0.0 ? pow(perf.precision, parm->precision_w) : 1.0;
            fitness *= parm->npv_w > 0.0 ? pow(perf.npv, parm->npv_w) : 1.0;

            *performance = fitness;

            scalar root = parm->accuracy_w + parm->gmean_w + parm->kappa_w + parm->recall_macro_w + parm->recall_macro_g_mean_w + parm->average_accuracy_w + parm->f1_w + parm->cc_w + parm->recall_w +
                          parm->specificity_w + parm->precision_w + parm->npv_w;
            fitness = pow(fitness, 1.0/root);
        }

        // Consider reduction?
        int selected = 0;
        for (int i = 0; i < instances+added; i++)
            if (subset[i])
                selected++;

        // Reduction term
        *reduction = 1.0 - ((scalar) selected)/(instances+added);
        scalar added_f = 1.0 - ((scalar) added)/(((scalar) plength/max_N)*N);

        fitness += parm->reduction_w*(*reduction) + parm->added_w * added_f;

        // Measure the "balancedness" of the dataset
        scalar mc = added, MC = 0.0;
        for (int i = 0; i < instances; i++)
            if (subset[i]) {
                if (labels[i] == minority)
                    mc++;
                else
                    MC++;
            }

        *balance = MIN(MC, mc) > 0.0 ? MIN(MC, mc)/MAX(MC, mc) : 0.0;
        fitness += parm->balance_w*(*balance);
    }

    if (isnan(fitness) || isinf(fitness))
        OnErrorExit(ERR_UNSPECTED_RESULT, "Fitness NaN of infinite");

    return fitness;
}


// Remove unselected instances if undersampling was used
int RemoveUnselectedInstances(scalar ***data, int *instances, int **labels, int ***index, int *subset)
{
    int removed = 0;
    // Instances must be removed from last to first
    for (int i = 0; i < *instances; i++)
        // Instance must be removed
        if (!subset[i]) {
            (*instances)--;
            (*data)[i] = (*data)[*instances];
            (*labels)[i] = (*labels)[*instances];
            if (index)
                (*index)[i] = (*index)[*instances];
            subset[i] = subset[*instances];
            i--;
            removed++;
        }

    REALLOC(*data, scalar *, *instances, "RemoveUnselectedInstances: *data");
    REALLOC(*labels, int, *instances, "RemoveUnselectedInstances: *labels");
    if (index)
        REALLOC(*index, int *, *instances, "RemoveUnselectedInstances: *index");

    return removed;
}

// Initialize population
// Initialization using distance to favor nearer neighbors
void InitializePopulation(const t_Parameters *parm, int **pair, int **subset, scalar **weight, int *N, int min_count, int maj_count, int instances, int n_neighbors, int inputs, int *subspace,
                          int **neighbors, scalar **data, int plength, int wlength, int *minority_list, int *majority_list, int size, int min_N, int max_N)
{
    scalar *roulette = NULL;
    if (parm->evosmote_roulette_initialization)
        MALLOC(roulette, scalar, n_neighbors, "EvolutionarySMOTE: roulette");

    // Initial population
    for (int i = 0; i < size; i++) {
        N[i] = RANDOM(min_N, max_N+1);

        // Roulette initialization
        if (parm->evosmote_roulette_initialization) {
            // Prepare roulette for each minority instance
            for (int k = 0; k < min_count; k++) {
                scalar farthest = EuclideanDistanceSparseDataSubspace(data[minority_list[k]], data[neighbors[k][n_neighbors-1]], inputs, NULL, NULL, subspace);
                roulette[0] = farthest - EuclideanDistanceSparseDataSubspace(data[minority_list[k]], data[neighbors[k][0]], inputs, NULL, NULL, subspace);
                for (int j = 1; j < n_neighbors; j++)
                    roulette[j] = roulette[j-1] + (farthest - EuclideanDistanceSparseDataSubspace(data[minority_list[k]], data[neighbors[k][j]], inputs, NULL, NULL, subspace));

                for (int j = 0; j < max_N; j++)
                    pair[i][k*max_N+j] = !parm->symmetric_selection ? RouletteSelection(roulette, n_neighbors) :
                                         (RANDOM(0.0, 1.0) < 0.5 ? RouletteSelection(roulette, n_neighbors) : -RouletteSelection(roulette, n_neighbors));
            }
        } else
            for (int j = 0; j < plength; j++)
                pair[i][j] = RANDOM(parm->symmetric_selection ? -n_neighbors : 0, n_neighbors);

        // Select all minority samples
        for (int j = 0; j < min_count; j++)
            subset[i][minority_list[j]] = 1;
        for (int j = instances; j < instances+plength; j++)
            subset[i][j] = 1;

        // Undersample majority class if apply
        if (parm->smote_undersampling_ratio <= 0.0)
            for (int j = 0; j < maj_count; j++)
                subset[i][majority_list[j]] = 1;
        else {
            scalar m_count = min_count * N[i] * (parm->symmetric_selection ? 0.5 : 1.0);  // Approximate number of instances of minority class for the individual
            scalar ratio = (parm->smote_undersampling_ratio*m_count)/maj_count;

            for (int j = 0; j < maj_count; j++)
                subset[i][majority_list[j]] = RANDOM(0.0, 1.0) < ratio ? 1 : 0;
        }

        for (int j = 0; j < wlength; j++)
            weight[i][j] = RANDOM(0.0, 1.0);
    }

    if (roulette)
        FREE(roulette);
}
