/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "imbalance.h"
#include "typex.h"

// Oversampling methods for class imbalance problems

// or: oversampling ratio. If or = 0, then oversample until both classes have the same number of instances

// Oversample with replacement of minority class until both classes have the same number of instances
void OversampleData(scalar  ***data, int *instances, int inputs, int classes, int **labels, scalar oversampling_ratio)
{
    int count[] = {0, 0};

    // Only valid for two class problems
    if (classes > 2)
        OnErrorExit(5, "Oversampling not valid for more than two classes");

    // Count instances
    for (int i = 0; i < *instances; i++)
        count[(*labels)[i]]++;

    int minority = (count[0] > count[1]) ? 1 : 0;

    // List of minority instances
    int *list, cursor = 0;
    MALLOC(list, int, count[minority], "OversampleData: list");
    for (int i = 0; i < *instances; i++)
        if ((*labels)[i] == minority) {
            list[cursor] = i;
            cursor++;
        }

    // Reallocate memory
    int majority = (minority+1)%2;
    int new_instances = *instances + (count[majority]*oversampling_ratio - count[minority]);

    REALLOC(*data, scalar *, new_instances, "OversampleData: *data");
    REALLOC(*labels, int, new_instances, "OversampleData: *labels");

    // Sample minority class
    for (; *instances < new_instances; (*instances)++) {
        int new = RANDOM(0, count[minority]);

        MALLOC((*data)[*instances], scalar, inputs, "OversampleData: (*data)[*instances]");

        for (int i = 0; i < inputs; i++)
            (*data)[*instances][i] = (*data)[list[new]][i];
        (*labels)[*instances] = minority;
    }
}

// Oversampling minority class following SMOTE method [Chawla, 2002]
// Create artificial instances from adjacent instances
// Each instance will be oversampled N times
void SMOTEOverSampling(scalar  ***data, int *instances, int inputs, int classes, int **labels, int N, int k)
{
    int count[2] = {0, 0};

    // Only valid for two class problems
    if (classes > 2)
        OnErrorExit(5, "SMOTE not valid for more than two classes");

    // Count instances
    for (int i = 0; i < *instances; i++)
        count[(*labels)[i]]++;

    int minority = (count[0] > count[1]) ? 1 : 0;

    // List of minority instances
    int *list, cursor = 0;
    MALLOC(list, int, count[minority], "SMOTEOverSampling: list");
    for (int i = 0; i < *instances; i++)
        if ((*labels)[i] == minority) {
            list[cursor] = i;
            cursor++;
        }

    // Check k below minority instances
    if (k >= count[minority])
        OnErrorExit(1, "Cannot apply SMOTE with k greater than minority samples");

    // Reallocate memory
    REALLOC(*data, scalar *, *instances + N * count[minority], "SMOTEOverSampling: *data");
    REALLOC(*labels, int, *instances + N * count[minority], "SMOTEOverSampling: *labels");

    // Oversample every minority class instance
    int old_instances = *instances;
    for (int j = 0; j < count[minority]; j++) {
        int nn[k];

        GetkNearestNeighbours((*data)[list[j]], list, *data, count[minority], inputs, k, nn);

        // Get k nearest neighbor for each pattern
        for (int i = 0; i < N; i++) {
            // Choose a random neighbor
            int r = nn[(int)RANDOM(0, k)];

            // Create artificial instance
            MALLOC((*data) [*instances], scalar, inputs + classes, "SMOTEOverSampling: (*data)[*instances]");
            for (int l = 0; l < inputs; l++)
                (*data) [*instances][l] = (*data) [list[j]][l] + RANDOM(0.0, 1.0) * ((*data) [r][l] - (*data) [list[j]][l]);
            // Labels
            (*labels)[*instances] = (*labels)[list[j]];

            (*instances) ++;
        }
    }

    if (*instances != old_instances + N * count[minority])
        OnErrorExit(1, "SMOTE failed");
}

void GetkNearestNeighbours(scalar *instance, int *list, scalar **protos, int n_protos, int dim, int k, int *nearest)
{
    scalar value, dmin, dist[k];

    // Select k first as k initial neighbors
    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    if (list == NULL) {
        for (int i = 0; i < n_protos; i++) {
            value = EuclideanDistance(instance, protos[i], dim);

            // Test if is one of the nearest
            if (value < dmin && value > 0.0) {
                // Look for its place
                int j = 0;
                while (value > dist[j])
                    j++;

                // Sets in j-th place
                for (int l = k - 1; l > j; l--) {
                    nearest[l] = nearest[l - 1];
                    dist[l] = dist[l - 1];
                }
                nearest[j] = i;
                dist[j] = value;
                dmin = dist[k - 1];
            }
        }
    } else {
        for (int i = 0; i < n_protos; i++) {
            value = EuclideanDistance(instance, protos[list[i]], dim);

            // Test if is one of the nearest
            if (value < dmin && value > 0.0) {
                // Look for its place
                int j = 0;
                while (value > dist[j])
                    j++;

                // Sets in j-th place
                for (int l = k - 1; l > j; l--) {
                    nearest[l] = nearest[l - 1];
                    dist[l] = dist[l - 1];
                }
                nearest[j] = list[i];
                dist[j] = value;
                dmin = dist[k - 1];
            }
        }
    }
}
