/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "imbalance.h"

// Undersampling methods for class imbalance problems

// Undersample randomly the majority class until both classes have the same number of instances
void UndersampleData(scalar  ***data, int *instances, int inputs, int classes, int **labels, int ratio)
{
    void RandomizeVector(int *, int);

    int count[2] = { 0, 0 }, *new_labels;
    scalar **new_data;

    // Only valid for two class problems
    if (classes > 2)
        OnErrorExit(5, "Undersampling not valid for more than two classes");

    // Count instances
    for (int i = 0; i < *instances; i++)
        count[(*labels)[i]]++;

    int majority = (count[0] < count[1]) ? 1 : 0;
    int minority = (majority+1) % 2;

    // Check ratio is Ok. Otherwise do nothing
    if (ratio*count[minority] > count[majority])
        return;

    // List of majority and minority instances
    int *list_m, *list_M, cursor_M = 0, cursor_m = 0;
    MALLOC(list_M, int, count[majority], "UndersampleData: list_M");
    MALLOC(list_m, int, count[minority], "UndersampleData: list_m");
    for (int i = 0; i < *instances; i++)
        if ((*labels)[i] == majority) {
            list_M[cursor_M] = i;
            cursor_M++;
        } else {
            list_m[cursor_m] = i;
            cursor_m++;
        }

    // Space for new data
    MALLOC(new_data, scalar *, (ratio+1) * cursor_m, "UndersampleData: new_data");
    MALLOC(new_labels, int, (ratio+1) * cursor_m, "UndersampleData: new_labels");

    // Sample majority class and copy complete minority class
    RandomizeVector(list_M, cursor_M);

    int old_instances = *instances;

    *instances = 0;
    int M = 0;
    for (int i = 0; i < cursor_m; (*instances) += ratio+1, i++, M += ratio) {
        MALLOC(new_data[*instances], scalar, inputs, "UndersampleData: new_data[*instances]");
        for (int j = 0; j < ratio; j++)
            MALLOC(new_data[*instances+1+j], scalar, inputs, "UndersampleData: new_data[*instances+1]");

        // Minority and majority class instances
        for (int j = 0; j < inputs; j++) {
            new_data[*instances][j] = (*data) [list_m[i]][j];
            for (int k = 0; k < ratio; k++)
                new_data[*instances+1+k][j] = (*data) [list_M[M+k]][j];
        }

        new_labels[*instances] = (*labels)[list_m[i]];
        for (int j = 0; j < ratio; j++)
            new_labels[*instances+1+j] = (*labels)[list_M[M+j]];
    }

    // Free data
    for (int i = 0; i < old_instances; i++)
        FREE((*data)[i]);
    FREE(*data);
    FREE(*labels);

    *data = new_data;
    *labels = new_labels;
}

// Undersample randomly all the majority classes until all classes have the same number of instances of the minority class
void MulticlassUndersampleData(scalar  ***data, int *instances, int inputs, int classes, int **labels, int ratio)
{
    void RandomizeVector(int *, int);

    int count[classes], *new_labels, new_instances = 0;
    scalar **new_data;

    // Count instances
    for (int i = 0; i < classes; i++)
        count[i] = 0;
    for (int i = 0; i < *instances; i++)
        count[(*labels)[i]]++;

    int minority = 0;
    for (int i = 1; i < classes; i++)
        if (count[i] < count[minority])
            minority = i;

    // List of instances to keep
    int *list, *class_list, class_count;
    MALLOC(list, int, *instances, "MulticlassUndersampleData: list");
    MALLOC(class_list, int, *instances, "MulticlassUndersampleData: class_list");
    for (int i = 0; i < classes; i++) {
        class_count = 0;
        for (int j = 0; j < *instances; j++)
            if ((*labels)[j] == i) {
                class_list[class_count] = j;
                class_count++;
            }

        // Randomize list
        RandomizeVector(class_list, class_count);

        // Copy enough instances to complete ratio
        for (int j = 0; j < count[minority]*ratio; j++, new_instances++)
            list[new_instances] = class_list[j];
    }
    FREE(class_list);

    // Space for new data
    MALLOC(new_data, scalar *, new_instances, "MulticlassUndersampleData: new_data");
    MALLOC(new_labels, int, new_instances, "MulticlassUndersampleData: new_labels");

    // Randomize instance order
    RandomizeVector(list, new_instances);

    for (int i = 0; i < new_instances; i++) {
        MALLOC(new_data[i], scalar, inputs, "MulticlassUndersampleData: new_data[i]");
        for (int j = 0; j < inputs; j++)
            new_data[i][j] = (*data)[list[i]][j];

        new_labels[i] = (*labels)[list[i]];
    }

    // Reallocate memory and free data
    for (int i = 0; i < *instances; i++)
        FREE((*data)[i]);
    FREE(*data);
    FREE(*labels);
    FREE(list);

    *data = new_data;
    *labels = new_labels;
    *instances = new_instances;
}

// Undersample randomly the majority class until both classes have the same number of instances
void ClusterBasedUndersampling(scalar ***data, int *instances, int inputs, int classes, int **labels)
{
    int count[2] = {0, 0}, *new_labels, *cluster_labels;
    scalar **new_data, **cluster_data;

    // Only valid for two class problems
    if (classes > 2)
        OnErrorExit(5, "Cluster Based Undersampling not valid for more than two classes");

    // Count instances
    for (int i = 0; i < *instances; i++)
        count[(*labels)[i]]++;

    int majority = (count[0] < count[1]) ? 1 : 0;
    int minority = (majority+1) % 2;

    // Check ratio is Ok, otherwise do nothing
    if (count[minority] > count[majority])
        return;

    // List of majority and minority instances
    int *list_m, *list_M, cursor_M = 0, cursor_m = 0;
    MALLOC(list_M, int, count[majority], "ClusterBasedUndersampling: list_M");
    MALLOC(list_m, int, count[minority], "ClusterBasedUndersampling: list_m");
    for (int i = 0; i < *instances; i++)
        if ((*labels)[i] == majority) {
            list_M[cursor_M] = i;
            cursor_M++;
        } else {
            list_m[cursor_m] = i;
            cursor_m++;
        }

    MALLOC(cluster_data, scalar *, cursor_M, "ClusterBasedUndersampling: cluster_data");
    MALLOC(cluster_labels, int, cursor_M, "ClusterBasedUndersampling: cluster_labels");
    for (int i = 0; i < cursor_M; i++) {
        cluster_data[i] = (*data)[list_M[i]];
        cluster_labels[i] = (*labels)[list_M[i]];
    }

    // Clustering
    int *partition, *cluster_count, n_clusters;
    n_clusters = cursor_m;
    MALLOC(partition, int, cursor_M, "ClusterBasedUndersampling: partition");
    MALLOC(cluster_count, int, n_clusters, "ClusterBasedUndersampling: cluster_count");
    OnErrorExit(1, "This function must be update");

    // Recalculate centroids
    scalar **centroids;
    MALLOC(centroids, scalar *, n_clusters, "ClusterBasedUndersampling: centroids");
    for (int i = 0; i < n_clusters; i++) {
        MALLOC(centroids[i], scalar, inputs, "ClusterBasedUndersampling: centroids[i]");
        for (int j = 0; j < inputs; j++)
            centroids[i][j] = 0.0;
    }

    for (int i = 0; i < cursor_M; i++)
        for (int j = 0; j < inputs; j++)
            centroids[partition[i]][j] += cluster_data[i][j];

    for (int i = 0; i < n_clusters; i++)
        for (int j = 0; j < inputs; j++)
            centroids[i][j] /= cluster_count[i];

    // Space for new data
    MALLOC(new_data, scalar *, cursor_m + n_clusters, "ClusterBasedUndersampling: new_data");
    MALLOC(new_labels, int, cursor_m + n_clusters, "ClusterBasedUndersampling: new_labels");

    // Copy minority instances
    for (int i = 0; i < cursor_m; i++) {
        MALLOC(new_data[i], scalar, inputs, "ClusterBasedUndersampling: new_data[i]");

        // Minority class instances
        for (int j = 0; j < inputs; j++)
            new_data[i][j] = (*data)[list_m[i]][j];
        new_labels[i] = (*labels)[list_m[i]];
    }

    // Copy instance closer to centroid for each cluster for majority class
    for (int i = 0; i < n_clusters; i++) {
        scalar closer_distance = 0.0, distance;
        int closer_instance = -1;

        for (int j = 0; j < cursor_M; j++)
            if (partition[j] == i) {
                if ((distance = EuclideanDistance(centroids[i], cluster_data[j], inputs)) < closer_distance || closer_instance == -1) {
                    closer_distance = distance;
                    closer_instance = j;
                }
            }

        // Copy to new data
        MALLOC(new_data[i+cursor_m], scalar, inputs, "ClusterBasedUndersampling: new_data[i+cursor_m]");
        for (int j = 0; j < inputs; j++)
            new_data[i+cursor_m][j] = cluster_data[closer_instance][j];
        new_labels[i+cursor_m] = cluster_labels[closer_instance];
    }

    // Free data
    for (int i = 0; i < *instances; i++)
        FREE((*data)[i]);
    FREE(*data);
    FREE(*labels);
    FREE(cluster_data);
    FREE(cluster_labels);
    FREE(partition);
    FREE(cluster_count);
    for (int i = 0; i < n_clusters; i++)
        FREE(centroids[i]);
    FREE(centroids);

    *data = new_data;
    *labels = new_labels;
    *instances = cursor_m + n_clusters;
}
