/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "macros.h"
#include "isfs.h"
#include "classifiers.h"

#define MAXIMUM_ALPHA   10.   // Maximum value of classifier weight
#define MINIMUM_ALPHA   0.1   // Minimum value of classifier weight

scalar UpdateAdaBoostWeightsISFS(const t_Parameters *parm, int *S, scalar **data, int n, int *labels, scalar *w, scalar *alpha, int inputs, int classes, bool madaboost, t_Classifier boostis_classifier,
                             int **index, int *subspace)
{
    scalar e_i, beta, sum;
    bool *ok;

    ALLOCA(ok, bool, n, "UpdateAdaBoostWeights: ok");

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    // Normalize w[] to sum 1
    sum = 0.0;
    for (int k = 0; k < n; k++)
        sum += w[k];
    for (int k = 0; k < n; k++)
        w[k] *= 1.0/sum;

    // Test and update weights
    e_i = GetWeightedError(parm, boostis_classifier, S, data, inputs, n, labels, classes, ok, w, parm->cv_hyperparameters, index, subspace);

    // If above 50% random re-sampling
    sum = 0.0;
    if (e_i >= 0.5 || fabs(e_i - 0.5) <= ZERO) {
        // Weight classifier
        *alpha = MINIMUM_ALPHA; // Almost remove classifier assigning a small weight

        for (int k = 0; k < n; k++)
            sum += (w[k] = Poisson());
    } else if (fabs(e_i) <= ZERO) {         // If perfect, also random resampling
        // Weight classifier
        *alpha = MAXIMUM_ALPHA;

        for (int k = 0; k < n; k++)
            sum += (w[k] = Poisson());
    } else {
        // beta = e_i/(1.0-e_i);
        beta = (1.0-e_i)/e_i;

        // Weight classifier
        // *alpha = MIN(MAXIMUM_ALPHA, log(1.0/beta));
        *alpha = MIN(MAXIMUM_ALPHA, 0.5*log(beta));

        for (int k = 0; k < n; k++)
            if (ok[k])
                // sum += (w[k] /= 2.0 * (1.0-e_i));
                sum += (w[k] *= sqrt(1.0/beta));
            else
                // For eMadaBoost do not increment weights
                sum += (w[k] *= (madaboost ? 1.0 : sqrt(beta)));
    }

    // Normalize w[] to sum 1
    for (int k = 0; k < n; k++)
        w[k] *= 1.0/sum;

    return e_i;
}

#define MAXIMUM_WEIGHT  1.0

void UpdateGentleBoostWeightsISFS(const t_Parameters *parm, int *S, scalar **data, int n, int *labels, scalar *w, int inputs, int classes, int **index, scalar **dist_cache, int *subspace)
{
    // WARNING: The implementation of eGentleBoost is "ad hoc""
    if (parm->boostisfs_classifier != kNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "eGentleBoost for C4.5 or SVM");

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    scalar margin, sum = 0.0;

    // Normalize w[] to sum 1
    for (int k = 0; k < n; k++)
        sum += w[k];
    for (int k = 0; k < n; k++)
        w[k] *= 1.0/sum;

    sum = 0.0;
    for (int j = 0; j < n; j++) {
        margin = kNearestNeighbourMarginSubspaceBoosting(j, labels[j], S, data, n, inputs, parm->k, labels, classes, NULL, dist_cache, index, subspace);

        // Update weights
        w[j] = MIN(MAXIMUM_WEIGHT, w[j]*exp(-margin));

        sum += w[j];
    }

    // Normalize to sum 1
    for (int i = 0; i < n; i++)
        w[i] *= 1.0/sum;
}

scalar kNearestNeighbourMarginSubspaceBoosting(int instance, int instance_class, int *S, scalar **protos, int n_protos, int dim, int k, int *label, int classes, scalar *r, scalar **dist_cache,
        int **index, int *subspace)
{
    int nearest[k];
    scalar value, dmin, dist[k], votes[classes];

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    // If 1-NN obtain
    if (k == 1)
        return NearestNeighbourMarginSubspaceBoosting(instance, instance_class, S, protos, n_protos, dim, label, r, dist_cache, index, subspace);

    ZeroVector(votes, NULL, classes);

    dmin = INFINITE;
    for (int i = 0; i < k; i++) {
        dist[i] = INFINITE;
        nearest[i] = -1;
    }

    // Get k nearest neighbors
    for (int i = 0; i < n_protos && dmin > 0.0; i++)
        if (S[i]) {
            value = dist_cache ? dist_cache[instance][i]/(r ? r[i] : 1.0) : EuclideanDistanceSubspace(protos[instance], protos[i], dim, subspace)/(r ? r[i] : 1.0);

            // Test if is one of the nearest
            if (value < dmin) {
                int count = S[i];
                while (count--) {
                    // Look for its place
                    int j = 0;
                    while (j < k && value > dist[j])
                        j++;

                    if (j == k)
                        break;

                    // Sets in j-th place
                    for (int l = k - 1; l > j; l--) {
                        nearest[l] = nearest[l - 1];
                        dist[l] = dist[l - 1];
                    }
                    nearest[j] = i;
                    dist[j] = value;
                    dmin = dist[k - 1];
                }
            }
        }

    // Vote neighbors
    for (int i = 0; i < k; i++)
        if (nearest[i] != -1)
            votes[label[nearest[i]]]++;

    // Votes of the correct class
    int ok_votes = votes[instance_class];

    // Get maximum of votes of wrong class
    int wrong_votes = 0;
    for (int i = 0; i < classes; i++)
        if (wrong_votes < votes[i] && i != instance_class)
            wrong_votes = votes[i];

    scalar margin = ((scalar)(ok_votes-wrong_votes)) /k;

    return margin;
}

scalar NearestNeighbourMarginSubspaceBoosting(int instance, int instance_class, int *S, scalar **protos, int n_protos, int n, int *label, scalar *r, scalar **dist_cache, int  **index, int *subspace)
{
    scalar ok_nearest = INFINITE, wrong_nearest = INFINITE, value, margin;
    bool at_least_one_ok, at_least_one_wrong;
    static scalar max_distance = -1.0;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    // First time?
    for (int i = 1; i < n_protos; i++)
        for (int j = 0; j < i; j++)
            if ((value = dist_cache ? dist_cache[instance][i]/(r ? r[i]: 1.0) : EuclideanDistanceSubspace(protos[instance], protos[i], n, subspace)/(r ? r[i]: 1.0)) > max_distance)
                max_distance = value;

    at_least_one_ok = at_least_one_wrong = false;

    // Iterate for every prototype
    for (int i = 0; i < n_protos; i++)
        if (S[i]) {
            if (label[i] == instance_class) {
                if ((value = dist_cache ? dist_cache[instance][i]/(r ? r[i]: 1.0) : EuclideanDistanceSubspace(protos[instance], protos[i], n, subspace)/(r ? r[i]: 1.0)) < ok_nearest) {
                    ok_nearest = value;
                    at_least_one_ok = true;
                }
            } else {
                if ((value = dist_cache ? dist_cache[instance][i]/(r ? r[i]: 1.0) : EuclideanDistanceSubspace(protos[instance], protos[i], n, subspace)/(r ? r[i] : 1.0)) < wrong_nearest) {
                    wrong_nearest = value;
                    at_least_one_wrong = true;
                }
            }
        }

    if (at_least_one_ok && at_least_one_wrong)
        margin = (wrong_nearest - ok_nearest)/max_distance;
    else if (at_least_one_ok)
        margin = 1.0/ok_nearest;
    else if (at_least_one_wrong)
        margin = -1.0/wrong_nearest;
    else
        margin = 0.0;

    return margin;
}

scalar UpdateeReweightBoostWeights(const t_Parameters *parm, int *S, int **S_all, scalar **data, int n, int *labels, scalar *w, scalar *alpha, int inputs, int classes, int r, int size,
                                  scalar *unweighted_accuracy, int **index, int *subspace)
{
    scalar e_i, sum;
    int nearest[parm->k];

    // Normalize w[] to sum 1
    sum = 0.0;
    for (int i = 0; i < n; i++)
        sum += w[i];
    for (int i = 0; i < n; i++)
        w[i] *= 1.0/sum;

    // Test and update weights
    e_i = sum = *unweighted_accuracy = 0.0;
    for (int k = 0; k < n; k++) {
        // Get error
        sum += w[k];
        if (kNNCategory(data[k], data, subspace, n, parm->k, inputs, labels, classes, nearest, S, NULL, NULL, -1, index, index[k]) != labels[k])
            e_i += w[k];
        else
            *unweighted_accuracy += 1.0;
    }
    e_i /= sum;
    *unweighted_accuracy /= n;

    // Weight classifier
    if (e_i >= 0.5) {
        alpha[size-1] = MINIMUM_ALPHA;
        for (int k = 0; k < n; k++)
            sum += (w[k] = Poisson());
    } else if (e_i <= ZERO) {
        alpha[size-1] = MAXIMUM_ALPHA;
        for (int k = 0; k < n; k++)
            sum += (w[k] = Poisson());
    } else {
        alpha[size-1] = MIN(MAXIMUM_ALPHA, 0.5*log((1.0-e_i)/e_i));

        scalar factor = exp(-alpha[size-1]);
        for (int k = 0; k < n; k++)
            // Get classification by last "r" classifiers
            if (kNearestNeighbourEnsembleSubspace(k, S_all, r, data, n, inputs, parm->k, nearest, labels, classes, size, alpha, parm->knn_with_weighted_votes) != labels[k])
                w[k] *= factor;
    }

    // Normalize w[] to sum 1
    for (int k = 0; k < n; k++)
        w[k] *= 1.0/sum;

    return e_i;
}
