/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "macros.h"
#include "isfs.h"
#include "fs.h"
#include "common.h"
#include "kd.h"
#include "svm.h"
#include "classifiers.h"
#include "maths.h"

scalar CHCISplusFS(const t_Parameters *parm, int *S, scalar **data, int instances, int inputs, int *label, scalar *w, bool free_cache, int *generations_performed, scalar *evaluations, int classes,
                   scalar **val_data, int val_instances, int *val_labels, scalar *r, int **index, scalar *hit_ratio)
{
    int min_dist, actual_size, order, father, mother, d, best, *aux_pt, mutations = 0, *aux = NULL, minority = -1, *list, **population, new = 0, used = 0;
    cache_t ***neighbors_cache = NULL;
    double *s_index = NULL;
    scalar aux_f, old_fitness = 0.0, new_fitness = 0.0, average_improvement = 0.0, *fitness, *ok, *reduction, val_accuracy = -1.0;
    bool *cache_done = NULL, first_time = false;
    t_SubspaceCache cache;

    // Test consistent parameters
    if (parm->multiselection &&(parm->k == 1 || parm->chc_wrapper != kNN))
        OnErrorExit(ERR_PARAM, "Multi-instance not valid for k=1 or wrapper not k-NN");

    if (parm->multiselect_only_minority_class)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Multi select only minority class for CHC IS+FS");

    if (index && val_data)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Validation data for sparse format");

    // Allocate memory for auxiliary variables
    MALLOC(list, int, 2 * parm->individuals, "main: list");
    // Allocate memory for individuals
    MALLOC(fitness, scalar, 2 * parm->individuals, "CHCISplusFS: fitness");
    MALLOC(ok, scalar, 2 * parm->individuals, "CHCISplusFS: ok");
    MALLOC(reduction, scalar, 2 * parm->individuals, "CHCISplusFS: ok");
    MALLOC(population, int *, 2 * parm->individuals, "CHCISplusFS: population");
    for (int i = 0; i < 2 * parm->individuals; i++)
        MALLOC(population[i], int, instances+inputs, "main: population[i]");

    *evaluations = 0.0;

    // Confusion matrix
    int **M;
    AllocateIntegerMatrix(&M, classes, classes);

    bool isfs_use_cache = parm->isfs_use_cache || parm->isfs_stepwise_cache;
    if (parm->chc_wrapper != kNN)
        isfs_use_cache = false;

    // Check if cache above limits
    if (isfs_use_cache) {
        scalar cache_size;

        if (parm->feature_selection) {
            cache_size = sizeof(cache_t **) * pow(2.0, inputs);
            cache_size += pow(2.0, (scalar) inputs) * sizeof(cache_t *) * instances;
            cache_size += pow(2.0, (scalar) inputs) * sizeof(cache_t) * instances * instances;
            cache_size += sizeof(bool) * pow(2.0, (scalar) inputs);
        } else
            cache_size = sizeof(cache_t *) * instances + sizeof(cache_t) * instances * instances;

        cache_size /= pow(2.0, 30.0); // GB

        if (cache_size > parm->isfs_cache_size && !parm->isfs_stepwise_cache) {
            isfs_use_cache = false;
            SHOWPROGRESS(parm->show_progress > 0, "\nIS+FS cache disabled: %.2lfGB > %.2lfGB\n", cache_size, parm->isfs_cache_size);
        } else
            SHOWPROGRESS(parm->show_progress > 0, "\nIS+FS maximum cache size: %.2lfG\n", cache_size);
    }

    isfs_use_cache &= !parm->isfs_stepwise_cache;  // If stepwise cache disable normal cache
    ((t_Parameters *) parm)->isfs_use_cache &= !parm->isfs_stepwise_cache;

    if (classes == 2) {
        int ccount[] = {0, 0};

        for (int i = 0; i < instances; i++)
            ccount[label[i]]++;

        minority = (ccount[0] < ccount[1]) ? 0 : 1;
    }

    if (parm->performance_measure == pfGMean && classes != 2)
        OnErrorExit(ERR_PARAM, "Cannot use G measure with more than 2 classes");

    // Time
    time_t init_time = time(NULL);

    // If initial percentage 0 adjust for uniform distribution in (0, k/2+1)
    scalar init_percent = parm->init_percent;
    if (init_percent < ZERO)
        init_percent = 1.0/(parm->k/2+2.0);

    // Create nearest neighbors cache
    if (isfs_use_cache) {
        // Allocate new cache
        AllocateCache(&neighbors_cache, instances, parm->feature_selection ? inputs : 0);

        // Allocate space for storing the value of the subspace in decimal
        MALLOC(s_index, double, 2 * parm->individuals, "CHCISplusFS: s_index");

        if (parm->feature_selection) {
            // Allocate space for storing the values already got
            MALLOC(cache_done, bool, (int) pow(2.0, (scalar) inputs), "CHCISplusFS: cache_done");
            for (int i = 0; i < (int) pow(2.0, (scalar) inputs); i++)
                cache_done[i] = false;
        } else {
            MALLOC(cache_done, bool, 1, "CHCISplusFS: cache_done");
            cache_done[0] = false;
        }
    } else if (parm->isfs_stepwise_cache) {
        // Allocate space for storing the value of the subspace in decimal
        MALLOC(s_index, double, 2 * parm->individuals, "CHCISplusFS: s_index");

        AllocateSubspaceCache(&cache);
    }

    // Create individuals
    for (int i = 0; i < parm->individuals; i++) {
        if (parm->instance_selection) {
            for (int k = 0; k < instances; k++)
                if (parm->multiselection)
                    population[i][k] = RANDOM(0.0, 1.0) < init_percent ? ((parm->multiselect_only_minority_class && label [k] != minority) ? 1 : RANDOM(1, (parm->k/2)+2)) : 0;
                else
                    population[i][k] = RANDOM(0.0, 1.0) < init_percent ? 1 : 0;
        } else
            for (int k = 0; k < instances; k++)
                population[i][k] = 1;

        if (parm->feature_selection)
            for (int k = instances; k < inputs+instances; k++)
                population[i][k] = RANDOM(0.0, 1.0) < parm->init_fs_percent ? 1 : 0;
        else
            for (int k = instances; k < inputs+instances; k++)
                population[i][k] = 1;

        if (isfs_use_cache)
            s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used, index);
        else if (parm->isfs_stepwise_cache)
            s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index, parm->isfs_cache_size);
    }

    // Temporal individual for mutations
    if (parm->only_accept_if_better)
        MALLOC(aux, int, instances+inputs, "CHCISplusFS: aux");

    for (int i = 0; i < 2 * parm->individuals; i++)
        list[i] = i;

    // Evaluate initial individuals
    if (parm->verbose > 3) {
        fprintf(parm->out_file, "Indiv\tOk\tReduc\tFitness\nInitial Population\n");
        fprintf(parm->out_file, "-------------------------------------------------------\n");
    }

    for (int k = 0; k < parm->individuals; k++) {
        fitness[k] = EvaluateISFSIndividual(parm, population[k], &ok[k], &reduction[k], data, instances, inputs, label, w, evaluations, minority,
                                            ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[k] : -1, cache_done), instances, classes, r, index);

        if (parm->verbose > 3)
            fprintf(parm->out_file, "%d\t%6.4lf\t%6.4lf\t%6.4lf\n", k, ok[k], reduction[k], fitness[k]);

    }

    if (parm->verbose > 3)
        fprintf(parm->out_file, "-------------------------------------------------------\n");

    // Main loop of CHC algorithm
    if (parm->instance_selection)
        min_dist = (instances+inputs)/4;
    else
        min_dist = inputs/4;

    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
    for (*generations_performed = 0; *generations_performed < parm->generations; (*generations_performed)++) {
        SHOWPROGRESS(parm->show_progress > 1, "CHC IS+FS Generation %*d/%d\r", WIDTH(parm->generations),  *generations_performed+1, parm->generations);

        // CHC next generation
        actual_size = parm->individuals;

        // Perform HUX crossover
        for (int i = 0; i < parm->individuals; i += 2) {
            // Random selection without replacement
            order = RANDOM(i, parm->individuals);
            father = list[order];
            list[order] = list[i];
            list[i] = father;
            order = RANDOM(i+1, parm->individuals);
            mother = list[order];
            list[order] = list[i+1];
            list[i+1] = mother;

            if (parm->instance_selection) {
                // If no minimum distance skip
                if ((d = HammingDistance(population[mother], population[father], instances+inputs)) < min_dist)
                    continue;

                HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], instances+inputs);
            } else {
                // If no minimum distance skip
                if ((d = HammingDistance(population[mother]+instances, population[father]+instances, inputs)) < min_dist)
                    continue;

                HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], instances+inputs);
            }

            // Two new individuals in the population.
            if (isfs_use_cache) {
                s_index[actual_size] = (double)
                                       ConvertToDecimalAndUpdateCache(population[actual_size]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                               index);
                s_index[actual_size+1] = (double)
                                         ConvertToDecimalAndUpdateCache(population[actual_size+1]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new,
                                                 &used, index);
            } else if (parm->isfs_stepwise_cache) {
                s_index[actual_size] = ConvertToDecimalAndUpdateSubspaceCache(population[actual_size]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                       parm->isfs_cache_size);
                s_index[actual_size+1] = ConvertToDecimalAndUpdateSubspaceCache(population[actual_size+1]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                         parm->isfs_cache_size);
            }

            actual_size += 2;
        }

        // Parents has their fitness calculated, offspring do not
        for (int i = parm->individuals; i < actual_size; i++)
            fitness[i] = -1.0;

        // If no crossover decrement threshold and retry.
        if (actual_size == parm->individuals)
            min_dist--;

        // Introduce random mutation if required
        // First thing to do, to introduce diversity
        if (parm->gga_mutation > ZERO && parm->bit_mutation > ZERO)
            for (int i = 1; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->gga_mutation) {
                    if (parm->instance_selection && parm->feature_selection) {
                        if (RandomMutation(parm, population[i], instances+inputs, parm->bit_mutation, label, minority, false, 0, 0, *generations_performed, parm->generations))
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } else if (parm->feature_selection) {
                        if (RandomMutation(parm, population[i]+instances, inputs, parm->bit_mutation, label, minority, false, 0, 0, *generations_performed, parm->generations))
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } else if (parm->instance_selection) {
                        if (RandomMutation(parm, population[i], instances, parm->bit_mutation, label, minority, false, 0, 0, *generations_performed, parm->generations))
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } else
                        OnErrorExit(ERR_PARAM, "Neither instance nor feature selection");

                    if (fitness[i] < -0.0) {
                        if (isfs_use_cache)
                            s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                         index);
                        else if (parm->isfs_stepwise_cache)
                            s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                         parm->isfs_cache_size);
                    }
                }

        // MODIFIED 27/January/2011
        // FS mutation before IS mutation

        // Introduce RFE mutation if required
        if (parm->rfe_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RFE mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rfe_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (RNNMutationISFS(parm, aux, instances, label, w, data, inputs, true, false, true, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority, neighbors_cache, cache_done,
                                           classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after RFE mutation: %6.4lf => %6.4lf\n", old_fitness,
                                new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                              ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes, r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce RNN mutation of features if required
        if (parm->rnn_fs_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RNN IS+FS mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rnn_fs_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (RNNMutationISFS(parm, aux, instances, label, w, data, inputs, true, false, false, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority, neighbors_cache, cache_done,
                                           classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after RNN FS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                           ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes,
                                                           r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce FOCUS mutation of features if required
        if (parm->focus_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FOCUS mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->focus_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (FeatureSelectionMutation(parm, aux, instances, label, w, data, inputs, fsFOCUS, evaluations, classes, parm->ls_only_use_selected)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index, parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after FOCUS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                           ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done),
                                                           instances, classes, r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce add Feature mutation if required
        if (parm->addF_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ADD feature mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->addF_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (AddIorFMutation(parm, aux, instances, label, w, data, inputs, true, false, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used, index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index, parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after Add FS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes, r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce RNN mutation if required
        if (parm->instance_selection && parm->rnn_is_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RNN IS mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rnn_is_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];
                    if (RNNMutationISFS(parm, aux, instances, label, w, data, inputs, false, true, false, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used, index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index, parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after RNN IS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                           ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes, r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce ICF mutation if required
        if (parm->instance_selection && parm->icf_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->icf_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, index);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (InstanceSelectionMutation(parm, aux, instances, label, w, data, inputs, ICF, NULL, classes, index, aux+instances)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after ICF mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                           ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes,
                                                           r, index));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }

        // Introduce Drop3 mutation if required
        if (parm->instance_selection && parm->drop3_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->drop3_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, index);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (InstanceSelectionMutation(parm, aux, instances, label, w, data, inputs, DROP3, NULL, classes, index, aux+instances)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after Drop3 mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                           ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes,
                                                           r, index));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }

        // Introduce MSS mutation if required
        if (parm->instance_selection && parm->mss_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->mss_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, index);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (InstanceSelectionMutation(parm, aux, instances, label, w, data, inputs, MSS, NULL, classes, index, aux+instances)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after MSS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes, r, index));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }

        // Introduce IB3 mutation if required
        if (parm->instance_selection && parm->ib3_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ib3_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, index);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (InstanceSelectionMutation(parm, aux, instances, label, w, data, inputs, IB3, evaluations, classes, index, aux+instances)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged


                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after IB3 mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes, r, index));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }

        // Introduce RNN standard mutation if required
        if (parm->instance_selection && parm->rnnstd_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rnnstd_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, index);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (InstanceSelectionMutation(parm, aux, instances, label, w, data, inputs, RNN, NULL, classes, index, aux+instances)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index, parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after RNN standard mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes, r,
                                                           index));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }

        // Introduce add Instance mutation if required
        if (parm->instance_selection && parm->addI_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ADD instance mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->addI_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (AddIorFMutation(parm, aux, instances, label, w, data, inputs, false, true, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index, parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after Add IS standard mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes,
                                                           r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce RNN mutation of instances and features if required
        if (parm->instance_selection && parm->rnn_isfs_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RNN FS mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rnn_isfs_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (RNNMutationISFS(parm, aux, instances, label, w, data, inputs, true, true, false, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after RNN IS+FS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes,
                                                           r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce IMOEA mutation of instances and features if required
        if (parm->instance_selection && parm->imoea_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "IMOEA mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->imoea_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (InstanceAndFeatureSelectionMutation(parm, aux, instances, label, w, data, inputs, IMOEA, classes)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after IMOEA mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes,
                                                           r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        // Introduce add I or F mutation if required
        if (parm->instance_selection && parm->addiorf_mutation > ZERO) {
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ADD instance or feature mutation for sparse data");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->addiorf_mutation * (i < parm->individuals ? parm->ls_factor_old_individuals : 1)) {
                    // End?
                    if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
                        break;

                    // If apply show fitness before mutation
                    if (parm->verbose >= 3)
                        old_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                             ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances,
                                                             classes, r, NULL);

                    if (parm->only_accept_if_better)
                        CopyIndividual(aux, population[i], instances+inputs);
                    else
                        aux = population[i];

                    if (AddIorFMutation(parm, aux, instances, label, w, data, inputs, true, true, evaluations, minority, classes, instances, r)) {
                        if (parm->only_accept_if_better)
                            AcceptIfBetter(parm, population[i], aux, &(fitness[i]), &(ok[i]), &(reduction[i]), data, instances, inputs, label, w, evaluations, minority,
                                           neighbors_cache, cache_done, classes, r, &new, &used, index, &cache);
                        else
                            // Fitness must be recalculated
                            fitness[i] = -1.0;
                    } // else: do nothing, individual unchanged

                    if (isfs_use_cache)
                        s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                     index);
                    else if (parm->isfs_stepwise_cache)
                        s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                     parm->isfs_cache_size);

                    if (parm->verbose >= 3) {
                        fprintf(parm->out_file, "Fitness before => after Add IS+FS mutation: %6.4lf => %6.4lf\n",
                                old_fitness, new_fitness = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w,
                                                           evaluations, minority, ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1,
                                                                   cache_done), instances, classes, r, NULL));
                        average_improvement += new_fitness - old_fitness;
                        mutations++;
                    }
                }
        }

        for (int k = 0; k < actual_size; k++)
            if (fitness[k] < 0.0)
                fitness[k] = EvaluateISFSIndividual(parm, population[k], &ok[k], &reduction[k], data, instances, inputs, label, w, evaluations, minority,
                                                    ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[k] : -1, cache_done), instances, classes, r,
                                                    index);

        // Obtain new population selecting best N individuals
        int new_accepted = 0;

        for (int i = 0; i < parm->individuals; i++) {
            order = OrderedSelection(fitness+i, 0, actual_size - i)+i;

            // Check if accepted is new individual or old
            if (order >= parm->individuals)
                new_accepted ++;

            aux_pt = population[order];
            population[order] = population[i];
            population[i] = aux_pt;

            aux_f = fitness[order];
            fitness[order] = fitness[i];
            fitness[i] = aux_f;
            aux_f = ok[order];

            ok[order] = ok[i];
            ok[i] = aux_f;

            aux_f = reduction[order];
            reduction[order] = reduction[i];
            reduction[i] = aux_f;

            if (isfs_use_cache) {
                int aux_i = s_index[order];
                s_index[order] = s_index[i];
                s_index[i] = aux_i;
            }
        }

        if (parm->verbose > 2)
            fprintf(parm->out_file, "Percentage of new individuals accepted: %lf\n", ((scalar) new_accepted)/parm->individuals);

        // Print individuals
        if (parm->verbose > 3) {
            fprintf(parm->out_file, "Indiv\tOk\tReduc\tFitness\n");
            fprintf(parm->out_file, "-------------------------------------------------------\n");

            for (int i = 0; i < parm->individuals; i++)
                fprintf(parm->out_file, "%02d\t%6.4lf\t%6.4lf\t%6.4lf\n", i, ok[i], reduction[i], fitness[i]);

            fprintf(parm->out_file, "-------------------------------------------------------\n");
        }

        // Print best individual so far
        if (parm->verbose > 1) {
            best = 0;
            fprintf(parm->out_file, "Best[%3d]\tAccuracy: %6.4lf\tReduction: %6.4lf\tFitness: %6.4lf\t", *generations_performed, ok[best], reduction[best], fitness[best]);
        }

        if (parm->verbose >= 3 && mutations > 0)
            fprintf(parm->out_file, "Mutation average improvement: %lf\n", average_improvement/mutations);

        // End?
        if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed))
            break;

        // Early stopping? Check when there are validation data
        if (val_data != NULL) {
            scalar accuracy = 0.0;
            int *i_list = NULL;
            t_PerformanceInfo perf;

            // Best always first individual
            best = 0;

            // List of selected instances
            MALLOC(i_list, int, instances, "CHCISplusFS: i_list");

            int selected = 0;
            for (int i = 0; i < instances; i++)
                if (population[best][i]) {
                    i_list[selected] = i;
                    selected++;
                }

            // Get accuracy on validation set
            ClassifierEvaluation(parm, val_data, val_instances, val_labels, data, instances, label, population[best]+instances, inputs, population[best], false, classes, parm->chc_wrapper,
                                       NULL, minority, M, NULL, &perf, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
            accuracy = perf.accuracy;

            if (accuracy < val_accuracy && !first_time) {
                // First time, give a second chance
                first_time = true;
                val_accuracy = accuracy;
            } else if (accuracy < val_accuracy && first_time) {
                FREE(i_list);
                break;
            } else
                val_accuracy = accuracy;

            FREE(i_list);
        }

        // If minimum distance 0 do as requested
        if (min_dist == 0) {
            if (parm->reinit_distance0) {
                if (parm->verbose > 2)
                    fprintf(parm->out_file, "Staled population reinitialized \n");
                // Reinitialize populations excepting best one
                best = OrderedSelection(fitness, 0, parm->individuals);

                for (int i = 0; i < parm->individuals; i++)
                    // Reseed changing 35%
                    if (i != best) {
                        if (parm->instance_selection)
                            for (int k = 0; k < instances+inputs; k++)
                                population[i][k] = RANDOM(0.0, 1.0) < 0.35 ? (population[best][k]+1)%2 : population[best][k];
                        else
                            for (int k = 0; k < instances; k++)
                                population[i][k] = 1;

                        if (parm->feature_selection)
                            for (int k = instances; k < instances+inputs; k++)
                                population[i][k] = RANDOM(0.0, 1.0) < 0.35 ? (population[best][k]+1)%2 : population[best][k];
                        else
                            for (int k = instances; k < instances+inputs; k++)
                                population[i][k] = 1;

                        if (isfs_use_cache)
                            s_index[i] = (double) ConvertToDecimalAndUpdateCache(population[i]+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, &new, &used,
                                         index);
                        else if (parm->isfs_stepwise_cache)
                            s_index[i] = ConvertToDecimalAndUpdateSubspaceCache(population[i]+instances, &cache, data, instances, inputs, parm->feature_selection, &new, &used, index,
                                         parm->isfs_cache_size);

                        fitness[i] = EvaluateISFSIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, inputs, label, w, evaluations, minority,
                                                            ReturnCache(isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, &cache, s_index ? s_index[i] : -1, cache_done), instances, classes,
                                                            r, index);
                    }
            } else
                break;
        }
    }

    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

    if (parm->verbose > 1)
        fprintf(parm->out_file, "Evaluations of the fitness function: %d\n", (int) *evaluations);

    if (parm->verbose >= 3 && mutations > 0)
        fprintf(parm->out_file, "FINAL mutation average improvement: %lf\n", average_improvement/mutations);

    best = OrderedSelection(fitness, 0, parm->individuals);

    // Copy best to S
    for (int i = 0; i < instances+inputs; i++)
        S[i] = population[best][i];

    // Obtain reduction
    // Instances
    scalar sel_instances = 0.0, sel_features = 0.0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            sel_instances++;

    // Features
    for (int i = instances; i < instances+inputs; i++)
        if (S[i])
            sel_features++;

    scalar reduction_ratio = 1.0 - (sel_instances/instances) * (sel_features/inputs);

    FREE(list);
    FREE(fitness);
    for (int i = 0; i < 2 * parm->individuals; i++)
        FREE(population[i]);
    FREE(population);
    FREE(ok);
    FREE(reduction);

    // Free cache
    if (free_cache && (isfs_use_cache || parm->isfs_stepwise_cache)) {
        if (isfs_use_cache) {
            FreeCache(&neighbors_cache, instances, parm->feature_selection ? inputs : 0);
            FREE(cache_done);
        } else if (parm->isfs_stepwise_cache)
            FreeSubspaceCache(&cache, instances);

        FREE(s_index);
    }

    if (parm->only_accept_if_better)
        FREE(aux);

    FreeIntegerMatrix(&M, classes);

    // If cache print usefulness
    if (hit_ratio)
        *hit_ratio = ((scalar) used)/(used+new);

    return reduction_ratio;
}

// Copy an individual
void CopyIndividual(int *target, int *source, int size)
{
    memcpy(target, source, size * sizeof(int));
}

// Accept an individual if fitness is better or equal than parent after mutation
// On return fitness is assigned so it does not need to be recalculated
bool AcceptIfBetter(const t_Parameters *parm, int *old_indiv, int *new_indiv, scalar *fitness, scalar *ok, scalar *reduction, scalar **data, int instances, int inputs, int *label, scalar *w,
                    scalar *evaluations, int minority, cache_t ***neighbors_cache, bool *cache_done, int classes, scalar *r, int *new, int *used, int **index, t_SubspaceCache *cache)
{
    scalar new_indiv_fitness, new_indiv_ok, new_indiv_reduction;
    double s_index = 0.0;

    if (*fitness < 0.0) {
        if (parm->isfs_use_cache)
            s_index = (double) ConvertToDecimalAndUpdateCache(old_indiv+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, new, used, index);
        else if (parm->isfs_stepwise_cache)
            s_index = ConvertToDecimalAndUpdateSubspaceCache(old_indiv+instances, cache, data, instances, inputs, parm->feature_selection, new, used, index, parm->isfs_cache_size);

        *fitness = EvaluateISFSIndividual(parm, old_indiv, ok, reduction, data, instances, inputs, label, w, evaluations, minority,
                                          ReturnCache(parm->isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, cache, s_index, cache_done), instances, classes, r, NULL);
    }

    if (parm->isfs_use_cache)
        s_index = (double) ConvertToDecimalAndUpdateCache(new_indiv+instances, cache_done, neighbors_cache, data, instances, inputs, parm->feature_selection, new, used, index);
    else if (parm->isfs_stepwise_cache)
        s_index = ConvertToDecimalAndUpdateSubspaceCache(new_indiv+instances, cache, data, instances, inputs, parm->feature_selection, new, used, index, parm->isfs_cache_size);

    new_indiv_fitness = EvaluateISFSIndividual(parm, new_indiv, &new_indiv_ok, &new_indiv_reduction, data, instances, inputs, label, w, evaluations, minority,
                        ReturnCache(parm->isfs_use_cache, parm->isfs_stepwise_cache, neighbors_cache, cache, s_index, cache_done), instances, classes, r, NULL);

    // Accept?
    if (new_indiv_fitness >= *fitness) {
        // Yes
        CopyIndividual(old_indiv, new_indiv, instances+inputs);
        *fitness = new_indiv_fitness;
        *ok = new_indiv_ok;
        *reduction = new_indiv_reduction;
        return true;
    } else
        // No
        return false;
}

// Convert binary subspace to decimal and obtain cache of the subspace
int ConvertToDecimalAndUpdateCache(int *subspace, bool *cache_done, cache_t ***neighbors_cache, scalar **data, int instances, int inputs, bool feature_selection, int *new, int *used,
                                   int **index)
{
    int decimal = 0, power = 1;

    // No feature selection
    if (!feature_selection) {
        // Is the cache calculated
        if (!cache_done[decimal]) {
            ObtainCache(neighbors_cache[decimal], data, instances, inputs, subspace, index);
            cache_done[decimal] = true;
        }

        return 0;
    }

    for (int i = 0; i < inputs; i++) {
        if (subspace[i])
            decimal += power;

        power *= 2;
    }

    // Is the cache calculated
    if (!cache_done[decimal]) {
        ObtainCache(neighbors_cache[decimal], data, instances, inputs, subspace, index);
        cache_done[decimal] = true;
        (*new)++;
    } else
        (*used)++;

    return decimal;
}

void FreeCache(cache_t ** **cache, int instances, int inputs)
{
    for (int j = 0; j < (int) pow(2.0, (scalar) inputs); j++) {
        for (int i = 0; i < instances; i++)
            FREE((*cache) [j][i]);
        FREE((*cache) [j]);
    }
    FREE(*cache);
}

void AllocateCache(cache_t ** **cache, int instances, int inputs)
{
    MALLOC(*cache, cache_t **, pow(2.0, inputs), "AllocateCache: *cache");
    for (int j = 0; j < (int) pow(2.0, (scalar) inputs); j++) {
        MALLOC((*cache) [j], cache_t *, instances, "AllocateCache: (*cache)[j]");

        for (int i = 0; i < instances; i++)
            MALLOC((*cache) [j][i], cache_t, instances, "AllocateCache: (*cache)[j][i]");
    }
}

void ObtainCache(cache_t **neighbors_cache, scalar **data, int instances, int inputs, int *subspace, int **index)
{
    RANK *rank;

    MALLOC(rank, RANK, instances, "ObtainCache: rank");

    for (int i = 0; i < instances; i++) {
        for (int k = 0; k < instances; k++) {
            rank[k].fitness = -EuclideanDistanceSparseDataSubspace(data[i], data[k], inputs, index ? index[i] : NULL, index ? index[k] : NULL, subspace);
            rank[k].being = k;
        }

        // Sort
        qsort(rank, instances, sizeof(RANK), Compare);

        // Store neighbors in ascending order of distance
        for (int k = 0; k < instances; k++)
            neighbors_cache[i][k] = rank[k].being;
    }

    FREE(rank);
}
