/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "isfs.h"
#include "macros.h"

// eFloatBoost method: S. Z. Li and Zh. Zhang, "eFloatBoost learning and statistical face recognition", IEEE TPAMI vol. 26, no. 9, pp. 1-12, 2004

scalar UpdateeFloatBoostWeights(int **S_all, scalar **data, int instances, int *labels, scalar *w, scalar *alpha, int inputs, int classes, int size, t_Classifier boostisfs_classifier, int k,
                               bool knn_with_weighted_votes, int **index)
{
    scalar e_i, sum;
    int nearest[k];
    bool ok;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    // Normalize w[] to sum 1
    sum = 0.0;
    for (int i = 0; i < instances; i++)
        sum += w[i];
    for (int i = 0; i < instances; i++)
        w[i] *= 1.0/sum;

    if (boostisfs_classifier != kNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "eReweightBoost for C4.5 or SVM");

    // Update weights using w(i) = exp (-margin(x_i))
    e_i = sum = 0.0;
    for (int i = 0; i < instances; i++) {
        // Get margin by last "r" classifiers
        scalar margin = kNearestNeighbourEnsembleMarginSubspace(i, S_all, size, data, instances, inputs, k, nearest, labels, classes, size, alpha, &ok, knn_with_weighted_votes);

        sum += (w[i] = exp(-margin));
        if (!ok)
            e_i ++;
    }

    e_i /= instances;

    // Weight classifier
    // alpha[size-1] = 1.0;  Not needed, set in main function

    // Normalize w[] to sum 1
    for (int i = 0; i < instances; i++)
        w[i] *= 1.0/sum;

    return e_i;
}

#define MAX_ATTEMPS 30   // Maximum number of attempts to construct the ensemble

// Returns removed members from the ensemble
int ConditionalExclusion(const t_Parameters *parm, int **S, int size, scalar **data, int instances, int inputs, int k, int *label, int classes, scalar *w, scalar *alpha, scalar *epsilon,
                         FILE *out_file, int *t, int **index)
{
    int h_prime = -1, M;
    scalar best_error;
    bool removed;
    static int attempts = 0;

    M = size;

    // Cannot remove if only one
    attempts++;
    if (M == 1 || attempts > MAX_ATTEMPS) {
        if (attempts > MAX_ATTEMPS) {
            fprintf(out_file, "Maximum attempts reached. Final ensemble size: %d\n", size);
            *t = size;
        }
        return 0;
    }

    do {
        removed = false;

        // Get the member that less contribute to the classification accuracy
        best_error = INFINITE;
        for (int i = 0; i < M; i++) {
            scalar backup = alpha[i];
            alpha[i] = 0.;

            scalar error = (1. - kNearestNeighbourEnsembleAccuracySubspace(S, M, data, instances, inputs, k, label, classes, M, alpha, parm->knn_with_weighted_votes));

            alpha[i] = backup;

            if (best_error > error) {
                best_error = error;
                h_prime = i;
            }
        }

        // Remove?
        if (best_error < epsilon[M-1]) {
            // Remove h prime from ensemble
            int *backup = S[h_prime];
            S[h_prime] = S[M-1];
            S[M-1] = backup;
            // alpha[h_prime] = alpha[M-1]; All alphas = 1
            epsilon[M-1] = best_error;
            M--;

            removed = true;
        } else {// Update weight, if needed, and return
            if (size != M)
                UpdateeFloatBoostWeights(S, data, instances, label, w, alpha, inputs, classes, size, parm->boostisfs_classifier, parm->k, parm->knn_with_weighted_votes, index);
        }
    } while (removed && M > 1);

    return size - M;
}
