/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "macros.h"
#include "isfs.h"

#define GBYTE   (pow(2.0, 30.0))

// To test if it is connected a DFS must visit all the nodes

bool Connected(t_Graph *graph, int nodes)
{
    bool *visited;

    MALLOC(visited, bool, nodes, "Connected: visited");
    for (int i = 0; i < nodes; i++)
        visited[i] = false;

    // Run DFS first on an arbitrary node
    DFS(0, visited, nodes, graph);

    for (int i = 1; i < nodes; i++)
        if (!visited[i]) {
            FREE(visited);
            return false; // DFS did not reach this node
        }

    FREE(visited);
    return true;
}

// Deep First Search

void DFS(int v, bool *visited, int nodes, t_Graph *graph)
{
    if (visited[v])
        // Don't visit the same node twice
        return;

    visited[v] = true;

    // DFS to all the neighbors
    for (int i = 0; i < nodes; i++)
        if (i != v && !visited[i] && GetValueFromEdge(graph, MAX(i, v), MIN(i, v)) > ZERO)
            DFS(i, visited, nodes, graph);
}

void PrintGraph(t_Graph *graph, int nodes, FILE *out_file, char *message)
{
    // Graph matrix
    fprintf(out_file, "%s\n", message);
    fprintf(out_file, "        ");
    for (int i = 0; i < nodes; i++)
        fprintf(out_file, "%4d ", i+1);
    fprintf(out_file, "\n        ");
    for (int i = 0; i < nodes; i++)
        fprintf(out_file, "-----");
    fprintf(out_file, "\n");
    for (int i = 0; i < nodes; i++) {
        fprintf(out_file, "%5d | ", i+1);
        for (int j = 0; j < i; j++)
            fprintf(out_file, "%4.1lf ", GetValueFromEdge(graph, i, j));
        fprintf(out_file, "%4.1lf\n", GetValueFromNode(graph, i));
    }
    fprintf(out_file, "        ");
    for (int i = 0; i < nodes; i++)
        fprintf(out_file, "-----");
    fprintf(out_file, "\n");
}


bool TestIfGraphOfSelectedFeaturesIsConnected(const t_Parameters *parm, t_Graph *graph, int nodes, int *subspace, scalar major_threshold, scalar minor_threshold, t_ThresholdEvaluation
        graph_selection, FILE *out_file)
{
    t_Graph local;

    int selected = SelectWithGraph(parm, graph, nodes, subspace, major_threshold, minor_threshold, graph_selection);

    // Construct a graph only with selected features and arcs above threshold
    local.use_dense_matrix = graph->use_dense_matrix;
    local.use_edge_list = graph->use_edge_list;
    AllocateAndInitializeGraph(&local, selected, graph->max_size);

    int next = 0;
    for (int i = 0; i < nodes; i++)
        if (subspace[i]) {
            int column = 0;
            for (int j = 0; j < i; j++)
                if (subspace[j]) {
                    AssignValueToEdge(&local, next, column, GetValueFromEdge(graph, i, j) >= minor_threshold ? GetValueFromEdge(graph, i, j) : 0.0);
                    column++;
                }

            AssignValueToNode(&local, next, GetValueFromNode(graph, i));
            next++;
        }

    if (!selected)
        return true;

    bool connected = Connected(&local, selected);
    if (parm->verbose >= 0)
        PrintGraph(&local, selected, out_file, "SELECTED Graph");

    FreeGraph(&local);

    return connected;
}

// Method to print get components in an undirected graph

int ConnectedComponents(t_Graph *graph, int nodes, int *size, int **ccomponentes, scalar minor_threshold, scalar major_threshold)
{
    bool *visited;
    int cc;

    MALLOC(visited, bool, nodes, "ConnectedComponents: visited");

    // Mark all the vertices as not visited, excepting those below threshold
    for (int v = 0; v < nodes; v++)
        visited[v] = GetValueFromNode(graph, v) >= major_threshold ? false : true;

    size[0] = cc = 0;

    for (int v = 0; v < nodes; v++) {
        if (!visited[v]) {
            // Get all reachable vertices from v
            DFSWithThreshold(v, visited, nodes, graph, &size[cc], ccomponentes[cc], minor_threshold);

            cc++;
            if (cc == nodes)
                break;
            size[cc] = 0;
        }
    }

    FREE(visited);

    return cc;
}

void DFSWithThreshold(int v, bool *visited, int nodes, t_Graph *graph, int *size, int *list, scalar minor_threshold)
{
    if (visited[v])
        // Don't visit the same node twice
        return;

    visited[v] = true;
    list[*size] = v;
    (*size)++;

    // DFS to all the neighbors
    for (int i = 0; i < nodes; i++)
        if (i != v && !visited[i] && GetValueFromEdge(graph, MAX(i, v), MIN(i, v)) >= minor_threshold)
            DFSWithThreshold(i, visited, nodes, graph, size, list, minor_threshold);
}

void AllocateAndInitializeGraph(t_Graph *graph, int nodes, scalar max_size)
{
    graph->nodes = nodes;
    graph->edges = 0;
    MALLOC(graph->node_list, float, nodes, "AllocateAndInitializeGraph: grap->node_list");
    ZeroFloatVector(graph->node_list, nodes);

    graph->size = ((scalar) sizeof(float)*nodes + sizeof(float *))/GBYTE;
    graph->max_size = max_size;

    if (graph->use_dense_matrix) {
        AllocateFloatMatrix(&(graph->adjacency_matrix), nodes, nodes);
        ZeroFloatMatrix(graph->adjacency_matrix, nodes, nodes);
        graph->edge_list = NULL;

        graph->size += ((scalar) sizeof(float *)*nodes+sizeof(float)*nodes*nodes)/GBYTE;
    } else if (graph->use_edge_list) {
        graph->adjacency_matrix = NULL;
        graph->edge_list = NULL;
    } else {
        // Use sparse matrix
        MALLOC(graph->adjacency_matrix, float *, nodes, "AllocateAndInitializeGraph: graph->adjacency_matrix");
        for (int i = 0; i < nodes; i++)
            graph->adjacency_matrix[i] = NULL;
        graph->edge_list = NULL;

        graph->size += ((scalar) sizeof(float *)*nodes)/GBYTE;
    }
}

void FreeGraph(t_Graph *graph)
{
    FREE(graph->node_list);

    if (graph->use_dense_matrix)
        FreeFloatMatrix(&(graph->adjacency_matrix), graph->nodes);
    else if (graph->use_edge_list) {
        FREE(graph->edge_list);
    } else {
        for (int i = 0; i < graph->nodes; i++)
            if (graph->adjacency_matrix[i])
                FREE(graph->adjacency_matrix[i]);
        FREE(graph->adjacency_matrix);
    }

    graph->nodes = graph->edges = 0;
}

void InitializeGraph(t_Graph *graph, int nodes)
{
    graph->nodes = nodes;
    graph->edges = 0;
    ZeroFloatVector(graph->node_list, nodes);

    if (graph->use_dense_matrix)
        ZeroFloatMatrix(graph->adjacency_matrix, nodes, nodes);
    else if (graph->use_edge_list) {
        FREE(graph->edge_list);
        graph->edge_list = NULL;
        graph->size = ((scalar) sizeof(float)*nodes)/GBYTE;
    } else {
        // Free if already reserved
        for (int i = 0; i < nodes; i++)
            if (graph->adjacency_matrix[i])
                FREE(graph->adjacency_matrix[i]);
        graph->size = ((scalar) sizeof(float)*nodes + sizeof(float *)*nodes)/GBYTE;
    }
}

float GetValueFromEdge(t_Graph *graph, int i, int j)
{
    if (i < j)
        return GetValueFromEdge(graph, j, i);
    else if (i == j)
        OnErrorExit(ERR_UNSPECTED_RESULT, "Self loop not allowed");

    if (graph->use_dense_matrix)
        return graph->adjacency_matrix[i][j];
    else if (graph->use_edge_list) {
        // Edges are ordered by first index, i, and second index, j
        int edge = 0;
        while (edge < graph->edges && (i < graph->edge_list[edge].origin || (i == graph->edge_list[edge].origin && j < graph->edge_list[edge].destination)))
            edge++;

        // Origin or destination not found?
        if (edge == graph->edges || graph->edge_list[edge].origin != i || graph->edge_list[edge].destination != j)
            return 0.0;
        else
            return graph->edge_list[edge].value;
    } else {
        if (graph->adjacency_matrix[i])
            return graph->adjacency_matrix[i][j];
        else
            return 0.0;
    }
}

float GetValueFromNode(t_Graph *graph, int i)
{
    return graph->node_list[i];
}

void AssignValueToEdge(t_Graph *graph, int i, int j, scalar value)
{
    if (i < j) {
        AssignValueToEdge(graph, j, i, value);
        return;
    } else if (i == j)
        OnErrorExit(ERR_UNSPECTED_RESULT, "Self loop not allowed");

    if (graph->use_dense_matrix) {
        if (graph->adjacency_matrix[i][j] <= ZERO)
            graph->edges++;
        graph->adjacency_matrix[i][j] = value;
    } else if (graph->use_edge_list) {
        // Edges are ordered by first index, i, and second index, i
        int edge = 0;
        while (edge < graph->edges && (i < graph->edge_list[edge].origin || (i == graph->edge_list[edge].origin && j < graph->edge_list[edge].destination)))
            edge++;

        // Origin or destination not found? Create new edge
        if (edge == graph->edges || graph->edge_list[edge].origin != i || graph->edge_list[edge].destination != j) {
            graph->edges++;
            graph->size += ((scalar) sizeof(t_Edge))/GBYTE;
            REALLOC(graph->edge_list, t_Edge, graph->edges, "AssignValueToEdge: graph->edge_list");
            // Move previous edges
            memmove(graph->edge_list+edge+1, graph->edge_list+edge, sizeof(t_Edge)*(graph->edges-edge-1));

            // Add new edge
            graph->edge_list[edge].origin = i;
            graph->edge_list[edge].destination = j;
            graph->edge_list[edge].value = value;
        } else
            // Assign new value
            graph->edge_list[edge].value = value;
    } else {
        // Ir reserved add value...
        if (graph->adjacency_matrix[i]) {
            if (graph->adjacency_matrix[i][j] <= ZERO)
                graph->edges++;
            graph->adjacency_matrix[i][j] = value;
            // Otherwise, reserve and initialize
        } else {
            graph->edges++;
            MALLOC(graph->adjacency_matrix[i], float, MIN(graph->nodes, i+1), "AssignValueToEdge: graph->adjacency_matrix[i]");
            graph->size += ((scalar) sizeof(float)*MIN(graph->nodes, i+1))/GBYTE;
            if (graph->size > graph->max_size)
                OnErrorExit(ERR_MEMORY, "Adjacency matrix exceeded maximum size");
            for (int k = 0; k <= i; k++)
                graph->adjacency_matrix[i][k] = 0.0;
            graph->adjacency_matrix[i][j] = value;
        }
    }
}

void AddValueToEdge(t_Graph *graph, int i, int j, scalar value)
{
    if (i < j) {
        AddValueToEdge(graph, j, i, value);
        return;
    } else if (i == j)
        OnErrorExit(ERR_UNSPECTED_RESULT, "Self loop not allowed");

    if (graph->use_dense_matrix) {
        if (graph->adjacency_matrix[i][j] <= ZERO)
            graph->edges++;
        graph->adjacency_matrix[i][j] += value;
    } else if (graph->use_edge_list) {
        // Edges are ordered by first index, i, and second index, i
        int edge = 0;
        while (edge < graph->edges && (i < graph->edge_list[edge].origin || (i == graph->edge_list[edge].origin && j < graph->edge_list[edge].destination)))
            edge++;

        // Origin or destination not found? Create new edge
        if (edge == graph->edges || graph->edge_list[edge].origin != i || graph->edge_list[edge].destination != j) {
            graph->edges++;
            REALLOC(graph->edge_list, t_Edge, graph->edges, "AssignValueToEdge: graph->edge_list");
            // Move previous edges
            memmove(graph->edge_list+edge+1, graph->edge_list+edge, sizeof(t_Edge)*(graph->edges-edge-1));

            // Add new edge with value
            graph->edge_list[edge].origin = i;
            graph->edge_list[edge].destination = j;
            graph->edge_list[edge].value = value;
        } else
            // Add value
            graph->edge_list[edge].value += value;
    } else {
        // If reserved add value...
        if (graph->adjacency_matrix[i]) {
            if (graph->adjacency_matrix[i][j] <= ZERO)
                graph->edges++;
            graph->adjacency_matrix[i][j] += value;
            // Otherwise, reserve and initialize
        } else {
            graph->edges++;
            MALLOC(graph->adjacency_matrix[i], float, MIN(graph->nodes, i+1), "AddValueToEdge: graph->adjacency_matrix[i]");
            graph->size += ((scalar) sizeof(float)*MIN(graph->nodes, i+1))/GBYTE;
            if (graph->size > graph->max_size)
                OnErrorExit(ERR_MEMORY, "Adjacency matrix exceeded maximum size");
            for (int k = 0; k <= i; k++)
                graph->adjacency_matrix[i][k] = 0.0;
            graph->adjacency_matrix[i][j] = value;
        }
    }
}

void AddValueToNode(t_Graph *graph, int i, scalar value)
{
    graph->node_list[i] += value;
}

void AssignValueToNode(t_Graph *graph, int i, scalar value)
{
    graph->node_list[i] = value;
}

scalar GraphMemorySizeInGB(t_Graph *graph)
{
    scalar size;

    if (!graph->use_edge_list) {
        size = graph->nodes * sizeof(float *);

        for (int i = 0; i < graph->nodes; i++)
            if (graph->adjacency_matrix[i])
                size += sizeof(float)*(i+1);
    } else
        size = graph->edges * sizeof(t_Edge);

    size += graph->nodes*sizeof(float)+sizeof(float *);

    return size/pow(2.0, 30.0);
}

void AddGraphToGraph(const t_Parameters *parm, t_Graph *addee, t_Graph *added, scalar factor)
{
    if (added->use_dense_matrix != addee->use_dense_matrix || added->use_edge_list != addee->use_edge_list)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Adding graphs with different representations (AddGraphToGraph)");

    if (added->nodes != addee->nodes)
        OnErrorExit(ERR_UNSPECTED_RESULT, "Incompatible number of nodes");

    // Nodes
    for (int i = 0; i < added->nodes; i++)
        addee->node_list[i] += added->node_list[i]*factor;

    // Add all non zero values of graph "added" to graph "addee"
    if (added->use_dense_matrix) {
        for (int i = 0; i < added->nodes; i++)
            for (int j = 0; j < i; j++)
                if (added->adjacency_matrix[i][j] >= ZERO) {
                    if (addee->adjacency_matrix[i][j] <= ZERO)
                        addee->edges++;
                    addee->adjacency_matrix[i][j] += added->adjacency_matrix[i][j]*factor;
                }
    } else if (added->use_edge_list) {
        // Edges are ordered by first index, i, and second index, j
        int addee_edge = 0;
        for (int i = 0; i < added->edges; i++) {
            while (addee_edge < addee->edges && addee->edge_list[addee_edge].origin < added->edge_list[i].origin)
                addee_edge++;
            if (addee_edge < addee->edges && addee->edge_list[addee_edge].origin == added->edge_list[i].origin)
                while (addee_edge < addee->edges && addee->edge_list[addee_edge].origin == added->edge_list[i].origin && addee->edge_list[addee_edge].destination < added->edge_list[i].destination)
                    addee_edge++;

            // New or existing?
            if (addee_edge < addee->edges && addee->edge_list[addee_edge].origin == added->edge_list[i].origin && addee->edge_list[addee_edge].destination == added->edge_list[i].destination) {
                if (addee->edge_list[addee_edge].value <= ZERO)
                    addee->edges++;
                addee->edge_list[addee_edge].value += added->edge_list[i].value*factor;
            } else
                AddValueToEdge(addee, added->edge_list[i].origin, added->edge_list[i].destination, added->edge_list[i].value*factor);
            SHOWPROGRESS(parm->show_progress > 1, "Copied edge: %*d/%d\r", WIDTH(added->edges), i+1, added->edges);
        }
    } else {
        for (int i = 0; i < added->nodes; i++)
            if (added->adjacency_matrix[i]) {
                if (addee->adjacency_matrix[i]) {
                    for (int j = 0; j < i; j++)
                        if (added->adjacency_matrix[i][j] >= ZERO) {
                            if (addee->adjacency_matrix[i][j] <= ZERO)
                                addee->edges++;
                            addee->adjacency_matrix[i][j] += added->adjacency_matrix[i][j]*factor;
                        }
                } else {
                    // New row
                    MALLOC(addee->adjacency_matrix[i], float, i+1, "AddGraphToGraph: addee->adjacency_matrix[i]");
                    for (int j = 0; j < i; j++)
                        if (added->adjacency_matrix[i][j] >= 0) {
                            addee->adjacency_matrix[i][j] = added->adjacency_matrix[i][j]*factor;
                            addee->edges++;
                        } else
                            addee->adjacency_matrix[i][j] = 0.0;
                }
            }
    }
    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
}

int GetListOfNodeValues(t_Graph *graph, scalar *values)
{
    int n = 0;

    for (int i = 0; i < graph->nodes; i++)
        if (graph->node_list[i] > ZERO) {
            values[n] = graph->node_list[i];
            n++;
        }

    return n;
}

int GetListOfEdgeValues(t_Graph *graph, float **values, scalar threshold)
{
    int n = 0;

    // Suppose all edges
    MALLOC(*values, float, graph->edges, "GetListOfNodeEdges: *values");

    if (graph->use_edge_list) {
        // Edges are ordered by first index, i, and second index, j
        for (int edge = 0; edge < graph->edges; edge++) {
            if (graph->edge_list[edge].origin != graph->edge_list[edge].destination && GetValueFromNode(graph, graph->edge_list[edge].origin) > threshold &&
                GetValueFromNode(graph, graph->edge_list[edge].destination) > threshold) {
                (*values)[n] = graph->edge_list[edge].value;
                n++;
            }
        }
    } else {
        for (int i = 1; i < graph->nodes; i++)
            if (graph->adjacency_matrix[i] && GetValueFromNode(graph, i) > threshold) {
                for (int j = 0; j < i; j++)
                    if (graph->adjacency_matrix[i][j] > ZERO && GetValueFromNode(graph, j) > threshold) {
                        (*values)[n] = graph->adjacency_matrix[i][j];
                        n++;
                    }
            }
    }

    // Reassign memory and add 0
    REALLOC(*values, float, n+1, "GetListOfNodeEdges: *values");
    (*values)[n] = 0.0;

    return n+1;
}

void SortGraphByValue(t_Graph *graph, t_Graph *sorted)
{
    sorted->use_edge_list = true;
    AllocateAndInitializeGraph(sorted, graph->nodes, graph->max_size);
    sorted->edges = graph->edges;

    // Copy all nodes as edge list
    MALLOC(sorted->edge_list, t_Edge, sorted->edges, "SortGraphByValue: sorted->edge_list");

    if (graph->use_edge_list)
        memcpy(sorted->edge_list, graph->edge_list, sizeof(t_Edge)*graph->edges);
    else {
        int n = 0;
        for (int i = 0; i < graph->nodes; i++)
            if (graph->adjacency_matrix[i]) {
                for (int j = 0; j < i; j++)
                    if (graph->adjacency_matrix[i][j] > ZERO) {
                        sorted->edge_list[n].value = graph->adjacency_matrix[i][j];
                        sorted->edge_list[n].origin = i;
                        sorted->edge_list[n].destination = j;
                        n++;
                    }
            }

        if (n != graph->edges)
            OnErrorExit(ERR_UNSPECTED_RESULT, "Edges count does not match");
    }

    // Sort by value
    qsort(sorted->edge_list, sorted->edges, sizeof(t_Edge), CompareEdgesHighestToLowest);

    // Copy nodes
    memcpy(sorted->node_list, graph->node_list, graph->nodes*sizeof(float));
}

// Auxiliary function for sorting, highest first

int CompareEdgesHighestToLowest(const void *edge_1, const void *edge_2)
{
    scalar diff;

    if ((diff = ((t_Edge *) edge_2)->value - ((t_Edge *) edge_1)->value) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}

// Save graph in text format as a list of edges
void SaveGraph(t_Graph *graph, char *filename, int *selection, bool save_null_nodes, bool save_null_edges, scalar major, scalar minor)
{
    scalar value;
    FILE *file;

    if ((file = fopen(filename, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open file for graph");

    // Save nodes as self loops
    for (int i = 0; i < graph->nodes; i++)
        if ((value = GetValueFromNode(graph, i)) > ZERO && value >= major)
            fprintf(file, "%d, %d, %c%g\n", i+1, i+1, !selection || selection[i] ? ' ' : '!',  value);
        else if (save_null_nodes)
            fprintf(file, "%d, %d, !%g\n", i+1, i+1, 0.0);

    // Graph matrix
    for (int i = 0; i < graph->nodes; i++)
        for (int j = 0; j < i; j++)
            if ((!selection || (selection[i] && selection[j])) && (value = GetValueFromEdge(graph, i, j)) > ZERO && value >= minor)
                fprintf(file, "%d, %d, %g\n", i+1, j+1, value);
            else if (save_null_edges)
                fprintf(file, "%d, %d, %g\n", i+1, j+1, 0.0);

    fclose(file);
}

void AddGraphToGraphWithConversion(const t_Parameters *parm, t_Graph *addee, t_Graph *added, int *list, scalar factor)
{
    if (added->use_dense_matrix != addee->use_dense_matrix || added->use_edge_list != addee->use_edge_list)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Adding graphs with different representations (AddGraphToGraphWithConversion)");

    // Nodes
    for (int i = 0; i < added->nodes; i++)
        addee->node_list[list[i]] += added->node_list[i]*factor;

    // Add all non zero values of graph "added" to graph "addee"
    if (added->use_dense_matrix) {
        for (int i = 0; i < added->nodes; i++)
            for (int j = 0; j < i; j++)
                if (added->adjacency_matrix[i][j] >= ZERO) {
                    if (addee->adjacency_matrix[list[i]][j] <= ZERO)
                        addee->edges++;
                    addee->adjacency_matrix[list[i]][j] += added->adjacency_matrix[i][j]*factor;
                }
    } else if (added->use_edge_list) {
        // Edges are ordered by first index, i, and second index, j
        int addee_edge = 0;
        for (int i = 0; i < added->edges; i++) {
            while (addee_edge < addee->edges && addee->edge_list[addee_edge].origin < list[added->edge_list[i].origin])
                addee_edge++;
            if (addee_edge < addee->edges && addee->edge_list[addee_edge].origin == list[added->edge_list[i].origin])
                while (addee_edge < addee->edges && addee->edge_list[addee_edge].origin == list[added->edge_list[i].origin] &&
                       addee->edge_list[addee_edge].destination < list[added->edge_list[i].destination])
                    addee_edge++;

            // New or existing?
            if (addee_edge < addee->edges && addee->edge_list[addee_edge].origin == list[added->edge_list[i].origin] &&
                addee->edge_list[addee_edge].destination == list[added->edge_list[i].destination]) {
                if (addee->edge_list[addee_edge].value <= ZERO)
                    addee->edges++;
                addee->edge_list[addee_edge].value += added->edge_list[i].value*factor;
            } else
                AddValueToEdge(addee, list[added->edge_list[i].origin], list[added->edge_list[i].destination], added->edge_list[i].value*factor);
            SHOWPROGRESS(parm->show_progress > 1, "Copied edge: %*d/%d\r", WIDTH(added->edges), i+1, added->edges);
        }
    } else {
        for (int i = 0; i < addee->nodes; i++)
            if (added->adjacency_matrix[i]) {
                if (addee->adjacency_matrix[list[i]]) {
                    for (int j = 0; j < i; j++)
                        if (added->adjacency_matrix[i][j] >= ZERO) {
                            if (addee->adjacency_matrix[list[i]][list[j]] <= ZERO)
                                addee->edges++;
                            addee->adjacency_matrix[list[i]][list[j]] += added->adjacency_matrix[i][j]*factor;
                        }
                } else {
                    // New row
                    MALLOC(addee->adjacency_matrix[list[i]], float, list[i]+1, "AddGraphToGraph: addee->adjacency_matrix[list[i]]");
                    for (int j = 0; j < i; j++)
                        if (added->adjacency_matrix[i][j] >= ZERO) {
                            addee->adjacency_matrix[list[i]][list[j]] = added->adjacency_matrix[i][j]*factor;
                            addee->edges++;
                        } else
                            addee->adjacency_matrix[list[i]][list[j]] = 0.0;
                }
            }
    }

    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
}
