/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "isfs.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

/*      MODIFICATIONS
        ---------------------
        2/1/2010        Added RNN mutation at Prnn ratio

*/

// Parameters of the algorithm
#define N_Objectives    3
#define Npop            30
#define P1              0.5             // Initial setting probability
// #define Nps          100
#define NE              30              // Maximum size for non-dominated population
#define Ps              0.4             // Selection proportion
#define Pm              0.05            // Mutation probability at bit level
#define gamma           7
#define Pc              0.6             // Probability of crossover
#define MaxEvaluations  1000000

#define Prnn            0.0     // RNN mutation ratio

// Maximum size of sets
#define max_E                   (Npop*(2+gamma))
#define max_Eprime              (Npop*(1+gamma))

int Neval;

// Chromosome [Instances+Inputs]

void IntelligentMOEA(const t_Parameters *parm, int *subspace, int *S, scalar **data, int instances, int *labels, int inputs, time_t time_limit, int classes)
{
    int **population, size, **E, **Eprime, **Ebar, n_E, n_Eprime, n_Ebar, **OA, omega, *parents[2], **by_products, max_Ebar = 100;
    scalar *fitness, **objectives, **E_objs, **Eprime_objs, **Ebar_objs, **bp_objs;
    time_t init_time;

    extern int Neval;

    size = instances+inputs;
    Neval = 0;

    // Allocate memory
    MALLOC(population, int *, Npop, "IntelligentMOEA: population");
    MALLOC(fitness, scalar, Npop, "IntelligentMOEA: fitness");
    MALLOC(objectives, scalar *, Npop, "IntelligentMOEA: objectives");
    MALLOC(Eprime, int *, max_Eprime, "IntelligentMOEA: Eprime");
    MALLOC(Eprime_objs, scalar *, max_Eprime, "IntelligentMOEA: Eprime_objs");
    MALLOC(E, int *, max_E, "IntelligentMOEA: E");
    MALLOC(E_objs, scalar *, max_E, "IntelligentMOEA: E_objs");
    MALLOC(Ebar, int *, max_Ebar, "IntelligentMOEA: Ebar");
    MALLOC(Ebar_objs, scalar *, max_Ebar, "IntelligentMOEA: Ebar_objs");
    for (int i = 0; i < max_Ebar; i++) {
        MALLOC(Ebar[i], int, size, "IntelligentMOEA: Ebar[i]");
        MALLOC(Ebar_objs[i], scalar, N_Objectives, "IntelligentMOEA: Ebar_objs[i]");
    }
    for (int i = 0; i < max_E; i++) {
        MALLOC(E[i], int, size, "IntelligentMOEA: E[i]");
        MALLOC(E_objs[i], scalar, N_Objectives, "IntelligentMOEA: E[i]");
    }
    for (int i = 0; i < max_Eprime; i++) {
        MALLOC(Eprime[i], int, size, "IntelligentMOEA: Eprime[i]");
        MALLOC(Eprime_objs[i], scalar, N_Objectives, "IntelligentMOEA: Eprime_objs[i]");
    }
    for (int i = 0; i < Npop; i++) {
        MALLOC(population[i], int, size, "IntelligentMOEA: population[i]");
        MALLOC(objectives[i], scalar, N_Objectives, "IntelligentMOEA: objectives[i]");
    }

    // Orthogonal array
    omega = (int) pow(2.0, floor(log2((scalar) gamma+1.0)));
    MALLOC(OA, int *, omega, "IntelligentMOEA: OA");
    for (int i = 0; i < omega; i++)
        MALLOC(OA[i], int, omega - 1, "IntelligentMOEA: OA[i]");
    OrthogonalArray(omega, OA);

    MALLOC(by_products, int *, omega, "IntelligentMOEA: by_products");
    MALLOC(bp_objs, scalar *, omega, "IntelligentMOEA: bp_objs");
    for (int i = 0; i < omega; i++) {
        MALLOC(by_products[i], int, size, "IntelligentMOEA: by_products");
        MALLOC(bp_objs[i], scalar, N_Objectives, "IntelligentMOEA: bp_objs[i]");
    }

    // Random initial population
    for (int i = 0; i < Npop; i++) {
        if (parm->multiselection)
            // Allow multiple instance selection for instances
            for (int j = 0; j < size - inputs; j++)
                population[i][j] = (RANDOM(0.0, 1.0) < P1) ? RANDOM(1, parm->k/2+2) : 0;
        else
            for (int j = 0; j < size - inputs; j++)
                population[i][j] = (RANDOM(0.0, 1.0) < P1) ? 1 : 0;

        for (int j = size - inputs; j < size; j++)
            population[i][j] = (RANDOM(0.0, 1.0) < P1) ? 1 : 0;
    }

    // Initialize E populations
    n_E = n_Ebar = n_Eprime = 0;

    init_time = time(NULL);

    // Evaluate objectives
    // Objectives are training error, size of subspace, size of selected training set
    for (int i = 0; i < Npop; i++)
        EvaluateObjectives(parm, population[i], objectives[i], data, instances, labels, inputs, classes, parm->k);

    while (Neval < MaxEvaluations) {
        // Evaluate fitness
        GPSIFF(fitness, objectives, Npop);

        // Best so far
        int best = OrderedSelection(fitness, 0, Npop);
        if (parm->verbose > 0)
            fprintf(parm->out_file, "Evaluations %4d. Fitness = %6.4lf Error = %6.4lf Reduction = %6.4lf Subspace = %6.4lf\n", Neval, fitness[best], objectives[best][0],
                    objectives[best][2], objectives[best][1]);

        // Update elite sets
        // Add non-dominated ones of population and E' to E
        for (int i = 0; i < Npop; i++)
            if (!TestIfDominated(objectives[i], objectives, Npop, N_Objectives)) {
                // Add to E
                for (int j = 0; j < size; j++)
                    E[n_E][j] = population[i][j];
                for (int j = 0; j < N_Objectives; j++)
                    E_objs[n_E][j] = objectives[i][j];
                n_E++;
            }

        for (int i = 0; i < n_Eprime; i++)
            if (!TestIfDominated(Eprime_objs[i], Eprime_objs, n_Eprime, N_Objectives)) {
                // Add to E
                for (int j = 0; j < size; j++)
                    E[n_E][j] = Eprime[i][j];
                for (int j = 0; j < N_Objectives; j++)
                    E_objs[n_E][j] = Eprime_objs[i][j];
                n_E++;
            }
        // Empty Eprime
        n_Eprime = 0;

        // Remove dominated ones in E
        for (int i = 0; i < n_E; i++)
            if (TestIfDominated(E_objs[i], E_objs, n_E, N_Objectives)) {
                // Remove, exchanging with last one
                int *auxi = E[i];
                E[i] = E[n_E - 1];
                E[n_E - 1] = auxi;
                scalar *auxd = E_objs[i];
                E_objs[i] = E_objs[n_E - 1];
                E_objs[n_E - 1] = auxd;
                n_E--;
            }
        // Add E to bar E
        for (int i = 0; i < n_E; i++) {
            // If max_Ebar reached reallocate
            if (n_Ebar >= max_Ebar) {
                max_Ebar += 100;
                REALLOC(Ebar, int *, max_Ebar, "IntelligentMOEA: Ebar");
                REALLOC(Ebar_objs, scalar *, max_Ebar, "IntelligentMOEA: Ebar_objs");
                for (int i = max_Ebar - 100; i < max_Ebar; i++) {
                    MALLOC(Ebar[i], int, size, "IntelligentMOEA: Ebar[i]");
                    MALLOC(Ebar_objs[i], scalar, N_Objectives, "IntelligentMOEA: Ebar_objs[i]");
                }
            }
            // Add to bar E
            for (int j = 0; j < size; j++)
                Ebar[n_Ebar][j] = E[i][j];
            for (int j = 0; j < N_Objectives; j++)
                Ebar_objs[n_Ebar][j] = E_objs[i][j];
            n_Ebar++;
        }

        // Remove dominated in bar E
        for (int i = 0; i < n_Ebar; i++)
            if (TestIfDominated(Ebar_objs[i], Ebar_objs, n_Ebar, N_Objectives)) {
                // Remove, exchanging with last one
                int *auxi = Ebar[i];
                Ebar[i] = Ebar[n_Ebar - 1];
                Ebar[n_Ebar - 1] = auxi;
                scalar *auxd = Ebar_objs[i];
                Ebar_objs[i] = Ebar_objs[n_Ebar - 1];
                Ebar_objs[n_Ebar - 1] = auxd;
                n_Ebar--;
            }
        // If n_E > NE, randomly remove individuals
        for (; n_E > NE; n_E--) {
            // Random selection of an individual
            int individual = RANDOM(0, n_E);

            // Remove, exchanging with last one
            int *auxi = E[individual];
            E[individual] = E[n_E - 1];
            E[n_E - 1] = auxi;
            scalar *auxd = E_objs[individual];
            E_objs[individual] = E_objs[n_E - 1];
            E_objs[n_E - 1] = auxd;
        }

        // Selection using binary tournament from old population to create new population
        for (int i = 0; i < Npop * (1.0 - Ps); i++) {
            int selected = BinaryTournament(fitness+i, Npop - i);
            selected += i;

            // Exchange selected with i-th
            int *auxi = population[selected];
            population[selected] = population[i];
            population[i] = auxi;

            scalar *auxd = objectives[selected];
            objectives[selected] = objectives[i];
            objectives[i] = auxd;

            fitness[selected] = fitness[i];
        }

        // Select randomly from E to complete new population
        for (int i = Npop * (1.0 - Ps); i < Npop; i++) {
            int selected = RANDOM(0, n_E);

            // Copy to population
            for (int j = 0; j < size; j++)
                population[i][j] = E[selected][j];
            for (int j = 0; j < N_Objectives; j++)
                objectives[i][j] = E_objs[selected][j];
        }

        // Crossover
        // Add offspring to E'
        for (int i = 0; i < Npop * Pc; i += 2) {
            // Random parents
            parents[0] = population[(int) RANDOM(0, Npop) ];
            parents[1] = population[(int) RANDOM(0, Npop) ];
            IntelligentCrossover(parm, OA, parents, Eprime[n_Eprime], Eprime[n_Eprime+1], size, omega, by_products, bp_objs, data, instances, labels, inputs, classes);

            // Obtain objectives
            EvaluateObjectives(parm, Eprime[n_Eprime], Eprime_objs[n_Eprime], data, instances, labels, inputs, classes, parm->k);
            EvaluateObjectives(parm, Eprime[n_Eprime+1], Eprime_objs[n_Eprime+1], data, instances, labels, inputs, classes, parm->k);

            n_Eprime += 2;

            // Add by-products of crossover if non-dominated
            for (int j = 0; j < omega; j++)
                if (TestIfDominated(bp_objs[j], objectives, Npop, N_Objectives)) {
                    for (int k = 0; k < size; k++)
                        Eprime[n_Eprime][k] = by_products[j][k];
                    for (int k = 0; k < N_Objectives; k++)
                        Eprime_objs[n_Eprime][k] = bp_objs[j][k];

                    n_Eprime++;
                }
        }

        // Mutation
        for (int i = 0; i < Npop; i++) {
            int removed = 0;

            bool modified = RandomMutation(parm, population[i], size, Pm, NULL, -1, false, 0, 0, 0, 0);
            if (RANDOM(0.0, 1.0) < Prnn)
                removed = RNNMutationIMOEA(parm, population[i], data, instances, labels, inputs, classes, parm->k);

            if (modified || removed)
                EvaluateObjectives(parm, population[i], objectives[i], data, instances, labels, inputs, classes, parm->k);
        }

        // Check whether exceeded time limit
        if (time_limit && (time(NULL) - init_time > time_limit))
            break;
    }

    fprintf(parm->out_file, "IMOEA number of evaluations: %d\n", Neval);

    // Final result is barE
    // Select best individual and copy to S
    REALLOC(fitness, scalar, n_Ebar, "IntelligentMOEA: fitness");
    GPSIFF(fitness, Ebar_objs, n_Ebar);

    int best = OrderedSelection(fitness, 0, n_Ebar);

    if (parm->verbose > 0)
        fprintf(parm->out_file, "FINAL BEST after %4d evaluations and %d seconds. Fitness = %3d Error = %6.4lf Reduction = %6.4lf Subspace = %6.4lf\n", Neval,
                (int)(time(NULL) - init_time), (int) fitness[best], Ebar_objs[best][0], Ebar_objs[best][2], Ebar_objs[best][1]);

    for (int i = 0; i < instances; i++)
        S[i] = Ebar[best][i];

    for (int i = 0; i < inputs; i++)
        subspace[i] = Ebar[best][instances+i];

    for (int i = 0; i < max_Ebar; i++) {
        FREE(Ebar[i]);
        FREE(Ebar_objs[i]);
    }
    for (int i = 0; i < max_E; i++) {
        FREE(E[i]);
        FREE(E_objs[i]);
    }
    for (int i = 0; i < max_Eprime; i++) {
        FREE(Eprime[i]);
        FREE(Eprime_objs[i]);
    }
    for (int i = 0; i < Npop; i++) {
        FREE(population[i]);
        FREE(objectives[i]);
    }
    FREE(population);
    FREE(fitness);
    FREE(objectives);
    FREE(Eprime);
    FREE(Eprime_objs);
    FREE(E);
    FREE(E_objs);
    FREE(Ebar);
    FREE(Ebar_objs);

    for (int i = 0; i < omega; i++)
        FREE(OA[i]);
    for (int i = 0; i < omega; i++) {
        FREE(by_products[i]);
        FREE(bp_objs[i]);
    }
    FREE(OA);
    FREE(by_products);
    FREE(bp_objs);
}


// Fitness function
// GPSIFF (X) = Individuals dominated by X - Individuals that dominate X+Total individuals
void GPSIFF(scalar *fitness, scalar **objectives, int pop_size)
{
    static int *dominates = NULL, *dominatedby = NULL, size = 0;

    // Allocate, or reallocate, if necessary
    if (size < pop_size) {
        REALLOC(dominates, int, pop_size, "GPSIFF: dominates");
        REALLOC(dominatedby, int, pop_size, "GPSIFF: dominatedby");
        size = pop_size;
    }
    // Obtain individuals domination
    for (int i = 0; i < pop_size; i++)
        dominates[i] = dominatedby[i] = 0;
    for (int i = 1; i < pop_size; i++)
        for (int j = 0; j < i; j++)
            if (DominatedBy(objectives[i], objectives[j], N_Objectives)) {
                dominatedby[i]++;
                dominates[j]++;
            } else if (DominatedBy(objectives[j], objectives[i], N_Objectives)) {
                dominatedby[j]++;
                dominates[i]++;
            }
    // Assign fitness
    for (int i = 0; i < pop_size; i++)
        fitness[i] = dominates[i] - dominatedby[i]+pop_size;
}


void IntelligentCrossover(const t_Parameters *parm, int **OA, int *parents[2], int *son, int *daughter, int size, int experiments, int **by_products, scalar **bp_objs,
                          scalar **data, int instances, int *labels, int inputs, int classes)
{
    static int *segments = NULL;
    static scalar *Y, **S;

    if (segments == NULL) {
        MALLOC(segments, int, gamma+1, "IntelligentCrossover: segments");
        segments++;
        MALLOC(Y, scalar, experiments, "IntelligentCrossover: Y");
        MALLOC(S, scalar *, gamma, "IntelligentCrossover: S");
        for (int i = 0; i < gamma; i++)
            MALLOC(S[i], scalar, 2, "IntelligentCrossover: S[i]");
    }

    // Randomly select gamma segments
    segments[0] = RANDOM(0, size);
    for (int i = 1; i < gamma - 1; i++)
        segments[i] = RANDOM(segments[i - 1], size);
    // Fixed limits
    segments[-1] = 0;
    segments[gamma - 1] = size;

    // Evaluate all experiments
    for (int i = 0; i < experiments; i++) {
        // Construct individual
        int k = 0;
        for (int j = 0; j < gamma; j++)
            for (; k < segments[j]; k++)
                by_products[i][k] = parents[OA[i][j]][k];

        // Evaluate individual
        EvaluateObjectives(parm, by_products[i], bp_objs[i], data, instances, labels, inputs, classes, parm->k);
    }

    // Obtain fitness values
    GPSIFF(Y, bp_objs, experiments);
    // Compute main effect of each level Sjk
    for (int i = 0; i < gamma; i++) {
        S[i][0] = S[i][1] = 0.0;
        for (int k = 0; k < experiments; k++)
            S[i][OA[k][i]] += Y[k];
    }

    // First child selecting best level
    // Randomly select gamma segments
    int k = 0;
    for (int j = 0; j < gamma; j++)
        for (; k < segments[j]; k++)
            if (S[j][0] > S[j][1])
                son[k] = daughter[k] = parents[0][k];
            else
                son[k] = daughter[k] = parents[1][k];
    // Select factor with lowest MED = |Sj1 - Sj2|
    int lowest = 0;
    for (int i = 1; i < gamma; i++)
        if (fabs(S[i][0] - S[i][1]) < fabs(S[lowest][0] - S[lowest][1]))
            lowest = i;
    // Change value of lowest factor in son
    for (int k = segments[lowest - 1]; k < segments[lowest]; k++)
        if (S[lowest][0] <= S[lowest][1])
            son[k] = parents[0][k];
        else
            son[k] = parents[1][k];
}

// All objectives are better if smaller
void EvaluateObjectives(const t_Parameters *parm, int *individual, scalar *objectives, scalar **data, int instances, int *labels, int inputs, int classes, int k)
{
    extern int Neval;
    int nearest[k];

    // First: training error and
    // Third: Instances selected
    objectives[0] = objectives[2] = 0.0;
    for (int i = 0; i < instances; i++) {
        if (individual[i])
            objectives[2]++;
        int backup = individual[i];
        individual[i] = 0;

        if (kNNCategory(data[i], data, individual+instances, instances, k, inputs, labels, classes, nearest, NULL, individual, NULL, -1, NULL, NULL) != labels[i])
            objectives[0]++;

        individual[i] = backup;
    }

    // At least one instance
    objectives[2] = (objectives[2] ? objectives[2]/instances : 1.0);
    objectives[0] /= instances;

    // Second: Subspace size
    objectives[1] = 0.0;
    for (int i = 0; i < inputs; i++)
        if (individual[instances+i])
            objectives[1]++;

    // At least one input
    objectives[1] = (objectives[1] ? objectives[1]/inputs : 1.0);

    // Add an evaluation
    Neval++;
}

//  Test if ind1 is dominated by ind2
bool DominatedBy(scalar *ind1, scalar *ind2, int no_objs)
{
    int i;
    bool one_better = false;
    for (i = 0; i < no_objs; i++)
        if (ind1[i] < ind2[i])
            return false;
        else {
            if (ind1[i] > ind2[i])
                one_better = true;
        }

    if (one_better)
        return true;
    else
        return false;
}

bool TestIfDominated(scalar *objs, scalar **all_objs, int n, int n_objs)
{
    for (int i = 0; i < n; i++)
        if (DominatedBy(objs, all_objs[i], n_objs))
            return true;
    return false;
}

void OrthogonalArray(int omega, int **A)
{
    int M, Q, J/*, N*/;
    M = omega;
    Q = 2;
    // N = omega - 1;
    J = log2((scalar) M);
    // Correct to indexes in 1,N
    for (int i = 0; i < M; i++)
        A[i]--;
    A--;
    if (M != pow((scalar) Q, (scalar) J))
        OnErrorExit(ERR_PARAM, "Invalid arguments to Orthogonal Array");
    // Basic columns
    for (int k = 1; k <= J; k++) {
        int j = (pow((scalar) Q, (scalar) k - 1.0) - 1.0)/(Q - 1.0)+1;
        for (int i = 1; i <= pow((scalar) Q, (scalar) J); i++)
            A[i][j] = ((int) floor(((scalar) i - 1.0)/pow((scalar) Q, (scalar) J - k))) % Q;
    }

    // Nonbasic columns
    for (int k = 2; k <= J; k++) {
        int j = (pow((scalar) Q, (scalar) k - 1.0) - 1.0)/(Q - 1.0)+1;
        for (int s = 1; s <= j - 1; s++)
            for (int t = 1; t <= Q - 1; t++)
                for (int i = 1; i <= M; i++)
                    A[i][j+(s - 1) * (Q - 1)+t] = (A[i][s] * t+A[i][j]) % Q;
    }

    //     // Increment all values in 1
    //     for (int i = 1; i <= M; i++)
    //      for (int j = 1; j <= N; j++)
    //          A[i][j]++;

    // Restore to indexes in 0,N-1
    A++;
    for (int i = 0; i < M; i++)
        A[i]++;
}


// This mutation is based on reduced nearest neighbor rule, but not identical
int RNNMutationIMOEA(const t_Parameters *parm, int *individual, scalar **data, int instances, int *labels, int inputs, int classes, int k)
{
    scalar all_bad, bad;
    int nearest[k], removed = 0, n_bad, n_ok;
    static int *list_bad = NULL, *list_ok;

    if (list_bad == NULL) {
        MALLOC(list_bad, int, instances, "RNNMutationIMOEA: list_bad");
        MALLOC(list_ok, int, instances, "RNNMutationIMOEA: list_ok");
    }

    // Store bad and good instances to start classifying bad ones
    n_bad = n_ok = 0;

    // Test ok when all present
    all_bad = 0.0;
    for (int i = 0; i < instances; i++)
        if (kNNCategory(data[i], data, individual+instances, instances, k, inputs, labels, classes, nearest, NULL, individual, NULL, -1, NULL, NULL) != labels[i]) {
            all_bad++;
            list_bad[n_bad] = i;
            n_bad++;
        } else {
            list_ok[n_ok] = i;
            n_ok++;
        }

    // Test if removing each instance affects the classification of any other instance
    // Remove each instance and test
    for (int i = 0; i < instances+inputs; i++) {
        if (individual[i]) {
            int backup = individual[i];
            individual[i] = 0;

            bad = 0.0;
            for (int j = 0; j < n_bad && all_bad >= bad; j++)
                if (kNNCategory(data[list_bad[j]], data, individual+instances, instances, k, inputs, labels, classes, nearest, NULL, individual, NULL, -1, NULL, NULL) != labels[list_bad[j]])
                    bad++;
            for (int j = 0; j < n_ok && all_bad >= bad; j++)
                if (kNNCategory(data[list_ok[j]], data, individual+instances, instances, k, inputs, labels, classes, nearest, NULL, individual, NULL, -1, NULL, NULL) != labels[list_ok[j]])
                    bad++;

            // If worse undo and leave for next instance
            if (all_bad < bad)
                individual[i] = backup;
            else
                // Removed, so count one less
                removed++;
        }
    }

    return removed;
}
