/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas and                     *
 *   Juan Antonio Romero del Castillo                                      *
 *   npedrajas@uco.es, aromero@uco.es                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements Democratic/Oligarchic Instance Selection wrapper

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <stdbool.h>
#include "macros.h"
#include "isfs.h"
#include "common.h"
#include "imbalance.h"

void InstanceAndFeatureSelectionAlgorithm(const t_Parameters *parm, int **S, scalar ***data, int *instances, int inputs, int **labels, int *generations, scalar *evaluations, int classes,
        int minority, int *fs, int **index, scalar *hit_ratio)
{
    int *s, l_instances, *l_labels, next, l_inputs, **l_index = NULL;
    scalar **l_data;

    switch (parm->isfsalgorithm) {
        case CHCGA:
            MALLOC(s, int, *instances+inputs, "InstanceAndFeatureSelectionAlgorithm: s");
            CHCISplusFS(parm, s, *data, *instances, inputs, *labels, NULL, true, generations, evaluations, classes, NULL, 0, NULL, NULL, index, hit_ratio);
            for (int i = 0; i < *instances; i++)
                (*S)[i] = s[i];
            for (int i = 0; i < inputs; i++)
                fs[i] = s[*instances+i];
            FREE(s);
            break;

        case IMOEA:
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "IMOEA for sparse format");
            IntelligentMOEA(parm, fs, *S, *data, *instances, *labels, inputs, parm->time_limit, classes);
            break;

        case SIFS2:
            ISFSWithFloatingSearch(parm, *S, fs, *data, *instances, *labels, inputs, classes, minority, index);
            break;

        case RMHC_PF1:
            RMHC_PF1Algorithm(parm, *S, *data, *instances, inputs, *labels, NULL, classes, NULL, index);
            break;

        case ISplusFS:
            l_instances = InstanceSelectionAlgorithm(parm, parm->isalgorithm, S, data, instances, inputs, labels, classes, index, NULL, NULL, -1, generations, parm->ratio, NULL, NULL, evaluations,
                          NULL, parm->cnn_ci, NULL, false, NULL);

            // Feature selection only with selected instances
            MALLOC(l_labels, int, l_instances, "InstanceAndFeatureSelectionAlgorithm: l_labels");
            MALLOC(l_data, scalar *, l_instances, "InstanceAndFeatureSelectionAlgorithm: l_data");
            if (index)
                MALLOC(l_index, int *, l_instances, "InstanceAndFeatureSelectionAlgorithm: l_index");
            next = 0;
            for (int i = 0; i < *instances; i++)
                if ((*S)[i]) {
                    l_labels[next] = (*labels)[i];
                    l_data[next] = (*data)[i];
                    if (index)
                        l_index[next] = index[i];
                    next++;
                }

            FeatureSelectionAlgorithm(parm, fs, inputs, l_data, l_labels, classes, l_instances, minority, index);

            FREE(l_labels);
            FREE(l_data);
            if (index)
                FREE(l_index);
            break;

        case FSplusIS:
            if (parm->isalgorithm == SMOTE)
                OnErrorExit(ERR_PARAM, "SMOTE not suitable for FS + IS");
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FSplusIS with sparse format");

            FeatureSelectionAlgorithm(parm, fs, inputs, *data, *labels, classes, *instances, minority, index);

            // Instance selection only with feature instances
            l_inputs = 0;
            for (int i = 0; i < inputs; i++)
                if (fs[i])
                    l_inputs++;

            // No inputs selected? Select one randomly
            if (!l_inputs) {
                fs[(int) RANDOM(0, inputs)] = 1;
                l_inputs = 1;
            }

            MALLOC(l_labels, int, *instances, "InstanceAndFeatureSelectionAlgorithm: l_labels");
            MALLOC(l_data, scalar *, *instances, "InstanceAndFeatureSelectionAlgorithm: l_data");

            for (int i = 0; i < *instances; i++) {
                MALLOC(l_data[i], scalar, l_inputs, "InstanceAndFeatureSelectionAlgorithm: l_data[i]");
                l_labels[i] = (*labels)[i];
                next = 0;
                for (int j = 0; j < inputs; j++)
                    if (fs[j]) {
                        l_data[i][next] = (*data)[i][j];
                        next++;
                    }
            }

            InstanceSelectionAlgorithm(parm, parm->isalgorithm, S, &l_data, instances, l_inputs, &l_labels, classes, index, NULL, NULL, -1, generations, parm->ratio, NULL, NULL, evaluations, NULL,
                                       parm->cnn_ci, NULL, false, fs);

            if (parm->show_progress > 1) {
                int count = 0;
                for (int i = 0; i < inputs; i++)
                    if (fs[i])
                        count++;
                SHOWPROGRESS(parm->show_progress > 1, "Selected features: %*d/%d\n", WIDTH(inputs), count, inputs);
            }

            FREE(l_labels);
            for (int i = 0; i < *instances; i++)
                FREE(l_data[i]);
            FREE(l_data);
            break;

        case NoISFS:
            for (int i = 0; i < *instances; i++)
                (*S)[i] = 1;
            for (int i = 0; i < inputs; i++)
                fs[i] = 1;
            break;
    }
}
