/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that deal with the individuals
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include "macros.h"
#include "isfs.h"
#include "common.h"
#include "classifiers.h"

scalar EvaluateISFSIndividual(const t_Parameters *parm, int *individual, scalar *ok, scalar *reduction, scalar **data, int instances, int inputs, int *label, scalar *w, scalar *evaluations,
                              int minority, cache_t **neighbors_cache, int max_size, int classes, scalar *r, int **index)
{
    int sel_instances, sel_features, **M;
    t_PerformanceInfo perf;

    if (w != NULL)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "IS+FS individual fitness evaluation with weights");

    // Evaluate reduction
    // Count the prototypes on the individual
    sel_instances = sel_features = 0;

    for (int i = 0; i < instances; i++)
        if (individual[i])
            sel_instances++;

    for (int i = instances; i < instances+inputs; i++)
        if (individual[i])
            sel_features++;

    // If reduction is 100% return 0 fitness
    if (sel_instances == 0 || sel_features == 0) {
        *ok = *reduction = 0.0;
        return 0.0;
    }

    (*evaluations)++;

    AllocateIntegerMatrix(&M, classes, classes);
    ClassifierEvaluation(parm, data, instances, label, data, instances, label, individual+instances, inputs, individual, false, classes, parm->chc_wrapper, NULL, minority, M, NULL, &perf, NULL,
                               NULL, NULL, NULL, neighbors_cache, index, index, NULL);
    FreeIntegerMatrix(&M, classes);

    // Evaluate performance using a wrapper approach
    switch (parm->performance_measure) {
        case pfGMean:
            *ok = perf.gmean;
            break;

        case pfAccuracy:
            *ok = perf.accuracy;
            break;

        case pfauROC:
        case pfauPRC:
        case pfMargin:
        case pfAverageAccuracy:
        case pfCC:
        case pfF1Macro:
        case pfPrecisionMacro:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "auROC/auPRC/Margin/Average accuracy/CC evaluation for IS+FS individual fitness");
            break;

        case pfRecallMacro:
            *ok = perf.recall_macro;
            break;

        case pfRecallMacroGMean:
            *ok = perf.recall_macro_g_mean;
            break;

        case pfKappa:
            *ok = perf.kappa;
            break;

        case pfF1:
            *ok = perf.f1_score;
            break;
    }

    return Fitness(parm, *ok, sel_features, inputs, sel_instances, instances, reduction);
}

// This mutation is based on reduced nearest neighbor rule, but not identical
int RNNMutationISFS(const t_Parameters *parm, int *individual, int instances, int *label, scalar *w, scalar **data, int inputs, bool mutate_features, bool mutate_instances, bool rfe,
                    scalar *evaluations, int minority, int classes, int max_size, scalar *r)
{
    scalar accuracy = 0.0, Sn, Sp, **l_data, reduction, current = 0.0, TP, TN, FP, FN, individual_evaluations = 0.0;
    int nearest[parm->k], *l_label, l_instances, *l_individual, l_inputs, backup = 0;
    bool *Ok, *Ok_so_far;

    // SVM
    if (parm->chc_wrapper == SVM)
        return RNNMutationISFSForLearners(parm, individual, instances, label, w, data, inputs, mutate_features, mutate_instances, rfe, evaluations, minority, classes, max_size, false, r);

    // Get reduction
    // The reduction improvement in the fitness of removing:
    // an instance is: selected features/(instances*features)
    // a feature is: selected instances/(instances*features)
    scalar sel_i = 0.0, sel_f = 0.0;

    for (int i = 0; i < instances; i++)
        if (individual[i])
            sel_i++;

    for (int i = instances; i < instances+inputs; i++)
        if (individual[i])
            sel_f++;

    // Nothing to do?
    if (sel_f == 0 || sel_i == 0)
        return 0;

    if (parm->ls_only_use_selected) {
        ALLOCA(l_label, int, sel_i, "RNNMutationISFS: l_label");
        ALLOCA(l_individual, int, sel_i+sel_f, "RNNMutationISFS: l_individual");
        ALLOCA(l_data, scalar *, sel_i, "RNNMutationISFS: l_data");

        // Only selected individuals are used
        l_instances = 0;
        l_inputs = sel_f;

        for (int i = 0; i < instances; i++)
            if (individual[i]) {
                MALLOC(l_data[l_instances], scalar, l_inputs, "RNNMutationISFS: l_data[l_instances]");

                for (int j = 0, k = 0; j < inputs; j++)
                    if (individual[instances+j]) {
                        l_data[l_instances][k] = data[i][j];
                        k++;
                    }

                l_label[l_instances] = label[i];
                l_individual[l_instances] = individual[i];
                l_instances++;
            }

        // Inputs
        for (int i = l_instances; i < l_instances+l_inputs; i++)
            l_individual[i] = 1;
    } else {
        l_instances = instances;
        l_inputs = inputs;
        l_label = label;
        l_data = data;
        l_individual = individual;
    }

    MALLOC(Ok_so_far, bool, l_instances, "RNNMutationISFS: Ok_so_far");
    MALLOC(Ok, bool, l_instances, "RNNMutationISFS: Ok");

    int *n_associates = NULL, **associates = NULL;
    if (mutate_instances) {
        MALLOC(n_associates, int, l_instances, "main: n_associates");
        MALLOC(associates, int *, l_instances, "main: associates");
        for (int i = 0; i < l_instances; i++)
            MALLOC(associates[i], int, l_instances, "main: associates[i]");

        for (int i = 0; i < l_instances; i++)
            n_associates[i] = 0;
    }

    for (int i = 0; i < l_instances; i++) {
        // MODIFIED 13/December/2010
        // Temporarily remove instance for evaluation: LOO
        backup = l_individual[i];
        l_individual[i] = 0;

        if (kNNCategory(l_data[i], l_data, l_individual+l_instances, l_instances, parm->k, inputs, l_label, classes, nearest, l_individual, NULL, NULL, -1, NULL, NULL) == l_label[i]) {
            accuracy++;
            Ok_so_far[i] = true;
        } else
            Ok_so_far[i] = false;

        if (mutate_instances) {
            // Add neighbors to associates list
            for (int j = 0; j < parm->k && nearest[j] != -1; j++) {
                bool found = false;
                for (int l = 0; l < n_associates[nearest[j]] && !found; l++)
                    if (associates[nearest[j]][l] == i)
                        found = true;

                if (!found) {
                    associates[nearest[j]][n_associates[nearest[j]]] = i;
                    n_associates[nearest[j]]++;
                }
            }
        }

        l_individual[i] = backup;
    }

    accuracy /= l_instances;

    if (mutate_instances) {
        // Instances with no associates can be removed
        for (int i = 0; i < l_instances; i++)
            if (n_associates[i] == 0) {
                l_individual[i] = 0;
                sel_i--;
            }
    }

    // Obtain G
    if (parm->performance_measure == pfGMean) {
        TP = TN = FP = FN = 0.0;
        // Store G value with all selected instances
        for (int i = 0; i < l_instances; i++) {
            if (Ok_so_far[i]) {
                if (l_label[i] != minority)
                    TN++;
                else
                    TP++;

            } else {
                if (l_label[i] != minority)
                    FP++;
                else
                    FN++;
            }
        }
        Sp = TN/(TN+FP);
        Sn = TP/(TP+FN);
        accuracy = pow(Sn * Sp, 0.5);
    }

    individual_evaluations += l_instances;

    scalar best_so_far;
    if (parm->isfs_use_fitness)
        best_so_far = Fitness(parm, accuracy, sel_f, inputs, sel_i, instances, &reduction);
    else
        best_so_far = accuracy;

    // Test if removing each instance affects the classification of any other instance
    // Remove each instance and test
    int removed_instances = 0, removed_features = 0;

    // Mutate features, instances or both?
    int begin = mutate_instances ? 0 : l_instances;

    int end = mutate_features ? l_instances+l_inputs : l_instances;

    // MODIFIED 24/December/2010
    // Add the option of randomizing the order of checking selected instances
    int *order;

    ALLOCA(order, int, l_instances+l_inputs, "RNNMutationISFS: order");

    if (mutate_instances && mutate_features && parm->isfsmutation_random_order) {
        for (int i = begin; i < end; i++)
            order[i] = i;

        RandomizeVector(order, l_instances+l_inputs);
    } else if (mutate_instances && mutate_features && parm->isfs_inputs_first) {
        for (int i = 0; i < l_inputs; i++)
            order[i] = i+l_instances;
        for (int i = 0; i < l_instances; i++)
            order[i+l_inputs] = i;
        RandomizeVector(order, l_inputs);
        RandomizeVector(order+l_inputs, l_instances);
    } else {
        for (int i = begin; i < end; i++)
            order[i] = i;

        if (mutate_instances)
            RandomizeVector(order, l_instances);

        if (mutate_features && !rfe)
            RandomizeVector(order+l_instances, l_inputs);

        // If RFE mutation order features accordingly
        else if (mutate_features && rfe) {
            // Order instances using RFE
            t_Feat *f_info;

            ALLOCA(f_info, t_Feat, l_inputs, "RNNMutationISFS: f_info");
            RFE(parm, f_info, l_inputs, l_data, l_instances, l_label, l_individual+l_instances);

            //Sort features from best to worst ranking
            qsort(f_info, l_inputs, sizeof(t_Feat), CompareRanking);

            // Order of features

            for (int i = l_inputs - 1, j = 0 ; i >= 0; i--, j++)
                order[j] = f_info[i].index;
        }
    }

    // MODIFIED 26/December/2010
    // Use as criterion for acceptance fitness instead of error
    for (int i = begin; i < end; i++) {
        // Do not do for last instance/feature
        if (l_individual[order[i]] && (order[i] >= l_instances || sel_i > 1) && (order[i] < l_instances || sel_f > 1)) {
            // Backup to undo if necessary
            int yes = l_individual[order[i]];
            l_individual[order[i]] = 0;

            // Get accuracy
            accuracy = 0.0;

            // For instances reuse
            if (order[i] < l_instances) {
                // Recalculate for associates of order[i] only
                for (int j = 0; j < l_instances; j++)
                    accuracy += (Ok[j] = Ok_so_far[j]);
                for (int j = 0; j < n_associates[order[i]]; j++) {
                    backup = l_individual[associates[order[i]][j]];
                    l_individual[associates[order[i]][j]] = 0;

                    if (kNNCategory(l_data[associates[order[i]][j]], l_data, l_individual+l_instances, l_instances, parm->k, inputs, l_label, classes, nearest, l_individual, NULL, NULL, -1, NULL, NULL)
                        == l_label[associates[order[i]][j]]) {
                        Ok[associates[order[i]][j]] = true;
                        if (!Ok_so_far[associates[order[i]][j]])
                            accuracy++;
                    } else {
                        Ok[associates[order[i]][j]] = false;
                        if (Ok_so_far[associates[order[i]][j]])
                            accuracy--;
                    }

                    l_individual[associates[order[i]][j]] = backup;

                    individual_evaluations++;
                }
            } else {
                for (int j = 0; j < l_instances; j++) {
                    backup = l_individual[j];
                    l_individual[j] = 0;

                    if ((Ok[j] = (kNNCategory(l_data[j], l_data, l_individual+l_instances, l_instances, parm->k, inputs, l_label, classes, nearest, l_individual, NULL, NULL, -1, NULL, NULL) ==
                                  l_label[j])))
                        accuracy++;

                    l_individual[j] = backup;

                    individual_evaluations++;
                }
            }

            accuracy /= l_instances;

            if (parm->performance_measure == pfGMean) {
                TP = TN = FP = FN = 0.0;
                // Store G value with all selected instances
                for (int j = 0; j < l_instances; j++) {
                    if (Ok[j]) {
                        if (l_label[j] != minority)
                            TN++;
                        else
                            TP++;

                    } else {
                        if (l_label[j] != minority)
                            FP++;
                        else
                            FN++;
                    }
                }
                Sp = TN/(TN+FP);
                Sn = TP/(TP+FN);
                accuracy = pow(Sn * Sp, 0.5);
            }

            // Get fitness
            if (parm->isfs_use_fitness) {
                if (order[i] < l_instances)
                    current = Fitness(parm, accuracy, sel_f, inputs, sel_i-1, instances, &reduction);
                else
                    current = Fitness(parm, accuracy, sel_f-1, inputs, sel_i, instances, &reduction);
            } else
                current = accuracy;


            // Fitness improvement is: accuracy improvement+reduction improvement
            // If negative improvement, or last instance/feature, undo
            if (current < best_so_far)
                // Undo
                l_individual[order[i]] = yes;
            else {
                if (order[i] < l_instances) {
                    removed_instances++;
                    sel_i--;
                } else {
                    removed_features++;
                    sel_f--;
                }

                // Update best fitness so far and Ok list
                best_so_far = current;

                if (order[i] < l_instances) {
                    bool *tempi = Ok;
                    Ok = Ok_so_far;
                    Ok_so_far = tempi;

                    // Update associates
                    for (int j = 0; j < n_associates[order[i]]; j++) {
                        kNNCategory(l_data[associates[order[i]][j]], l_data, l_individual+l_instances, l_instances, parm->k, inputs, l_label, classes, nearest, l_individual, NULL, NULL, -1, NULL, NULL);

                        // Add neighbors to associates list
                        for (int k = 0; k < parm->k && nearest[k] != -1; k++) {
                            // Test if previously in the list
                            bool found = false;
                            for (int l = 0; l < n_associates[nearest[k]] && !found; l++)
                                if (associates[nearest[k]][l] == order[i])
                                    found = true;

                            if (!found) {
                                associates[nearest[k]][n_associates[nearest[k]]] = order[i];
                                n_associates[nearest[k]]++;
                            }
                        }
                    }
                }
            }
        }
    }

    if (parm->ls_only_use_selected) {
        for (int i = 0; i < l_instances; i++)
            FREE(l_data[i]);

        // Transfer selection
        // Instances?
        if (mutate_instances) {
            for (int i = 0, j = 0; i < instances; i++)
                if (individual[i]) {
                    individual[i] = l_individual[j];
                    j++;
                }
        }

        // Features?
        if (mutate_features) {
            for (int i = instances, j = l_instances; i < instances+inputs; i++)
                if (individual[i]) {
                    individual[i] = l_individual[j];
                    j++;
                }
        }
    }

    if (parm->verbose >= 3) {
        char *method;
        if (mutate_features && mutate_instances)
            method = "RNN IS+FS";
        else if (mutate_features)
            method = "RNN FS";
        else
            method = "RNN IS";

        fprintf(parm->out_file, "%s removed %3d/%3d instances and %d/%d features\n", method, removed_instances, l_instances, removed_features, l_inputs);
    }

    // Fractional evaluations of the fitness functions
    *evaluations += (individual_evaluations*sel_i*l_inputs)/(((scalar) instances) *instances*inputs);

    FREE(Ok_so_far);
    FREE(Ok);
    if (mutate_instances) {
        FREE(n_associates);
        for (int i = 0; i < l_instances; i++)
            FREE(associates[i]);
        FREE(associates);
    }

    return removed_instances+removed_features;
}

// Evaluate error for 2-class problems with class imbalance measures
scalar EvaluateISFS2ClassError(const t_Parameters *parm, int *individual, int instances, int inputs, int *label, scalar *w, scalar *Sn, scalar *Sp, scalar **data, int minority,
                               cache_t **neighbors_cache, int classes, scalar *ratio, int **index)
{
    int nearest[parm->k], backup = 0;
    scalar ok, TP, TN, FP, FN;
    bool r;

    ok = TP = TN = FP = FN = 0.0;

    // Check if classifier available
    switch (parm->chc_wrapper) {
        case kNN:
            for (int i = 0; i < instances; i++) {
                // Evaluate using leave-one out
                backup = individual[i];
                individual[i] = 0;

                if (parm->isfs_use_cache)
                    r = (kNearestNeighbourCache(i, individual, instances, nearest, label, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, ratio, NULL, index, NULL)
                         == label[i]);
                else
                    r = (kNNCategory(data[i], data, NULL, instances, parm->k, inputs, label, classes, nearest, individual, NULL, NULL, -1, NULL, NULL) == label[i]);

                if (r) {
                    ok += (w == NULL) ? 1.0 : w[i];

                    if (label[i] != minority)
                        TN++;
                    else
                        TP++;

                } else {
                    if (label[i] != minority)
                        FP++;
                    else
                        FN++;
                }

                individual[i] = backup;
            }

            break;

        default:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Classifier for 2 class IS+FS evaluation");
            break;

    }

    *Sp = TN/(TN+FP);
    *Sn = TP/(TP+FN);

    if (w == NULL)
        ok /= instances;

    return ok;
}

// This mutation is based on instance selection algorithms.
// Now implements: Drop3, ICF, IB3, RNN and MSS
int InstanceSelectionMutation(const t_Parameters *parm, int *individual, int instances, int *label, scalar *w, scalar **data, int inputs, t_ISAlgorithm method, scalar *evaluations, int classes,
                              int **index, int *subspace)
{
    scalar **l_data;
    int selected_inputs = 0, selected_instances = 0, *l_label, *l_individual, l_instances, **l_index = NULL;
    char *name[] = { "DROP3", "ICF", NULL, "MSS", "RNN", "IB3" };

    // Count features
    if (parm->feature_selection) {
        for (int i = 0; i < inputs; i++)
            if (individual[instances+i])
                selected_inputs++;
    } else
        selected_inputs = inputs;

    // If no inputs, nothing to do
    if (selected_inputs == 0)
        return 0;

    // Create local data using selected features
    MALLOC(l_data, scalar *, instances, "InstanceSelectionMutation: l_data");
    if (index)
        MALLOC(l_index, int *, instances, "InstanceSelectionMutation: l_index");

    // Modified 24/December/2010
    // Modified 3/January/2011
    // Both possibilities are used
    if (parm->ls_only_use_selected) {
        MALLOC(l_label, int, instances, "InstanceSelectionMutation: l_label");
        MALLOC(l_individual, int, instances, "InstanceSelectionMutation: l_individual");

        // Only selected individuals are used
        for (int i = 0; i < instances; i++)
            if (individual[i]) {
                if (parm->feature_selection) {
                    MALLOC(l_data[selected_instances], scalar, selected_inputs, "InstanceSelectionMutation: l_data[selected_instances]");
                    if (index)
                        MALLOC(l_index[selected_instances], int, selected_inputs, "InstanceSelectionMutation: l_index[selected_instances]");

                    for (int j = 0, k = 0; j < inputs; j++)
                        if (individual[instances+j]) {
                            l_data[selected_instances][k] = data[i][j];
                            if (index)
                                l_index[selected_instances][k] = index[i][j];
                            k++;
                        }
                } else {
                    l_data[selected_instances] = data[i];
                    if (index)
                        l_index[selected_instances] = index[i];
                }

                l_label[selected_instances] = label[i];
                l_individual[selected_instances] = individual[i];

                selected_instances++;
            }

        l_instances = selected_instances;
    } else {
        l_label = label;
        l_individual = individual;

        // All individuals are used
        for (int i = 0; i < instances; i++) {
            if (individual[i])
                selected_instances++;

            if (parm->feature_selection) {
                MALLOC(l_data[i], scalar, selected_inputs, "InstanceSelectionMutation: l_data[i]");
                if (index)
                    MALLOC(l_index[i], int, selected_inputs, "InstanceSelectionMutation: l_index[i]");

                for (int j = 0, k = 0; j < inputs; j++)
                    if (individual[instances+j]) {
                        l_data[i][k] = data[i][j];
                        if (index)
                            l_index[i][k] = index[i][j];
                        k++;
                    }
            } else {
                l_data[i] = data[i];
                if (index)
                    l_index[i] = index[i];
            }
        }

        l_instances = instances;
    }

    // If no instances, nothing to do
    if (selected_instances == 0) {
        FREE(l_data);
        return 0;
    }

    int selected_by = 0;

    InstanceSelectionAlgorithm(parm, parm->is_threshold_step, &l_individual, &l_data, &l_instances, inputs, &l_label, classes, l_index, NULL, NULL, -1, NULL, 0.0, NULL, NULL, NULL, NULL, false, NULL,
                               false, subspace);

    if (parm->ls_only_use_selected)
        // Transfer selection
        for (int i = 0, j = 0; i < instances; i++)
            if (individual[i]) {
                individual[i] = l_individual[j];
                j++;
            }

    if (parm->feature_selection)
        for (int i = 0; i < l_instances; i++) {
            FREE(l_data[i]);
            if (index)
                FREE(l_index[i]);
        }

    FREE(l_data);
    FREE(l_label);
    FREE(l_individual);
    if (index)
        FREE(l_index);

    if (parm->verbose >= 3)
        fprintf(parm->out_file, "Instance selection method (%s) removed %3d instances\n", name[method], selected_instances - selected_by);

    return selected_instances - selected_by;
}

// This mutation adds instances and/or features no selected if they improve the error
int AddIorFMutation(const t_Parameters *parm, int *individual, int instances, int *label, scalar *w, scalar **data, int inputs, bool mutate_features, bool mutate_instances, scalar *evaluations,
                    int minority, int classes, int max_size, scalar *r)
{
    scalar accuracy = 0.0, Sn, Sp, fitness = 0.0, red, TP, TN, FP, FN, individual_evaluations = 0.0;
    int nearest[parm->k], backup = 0;

    // SVM or C4.5
    if (parm->chc_wrapper == SVM || parm->chc_wrapper == C45)
        return RNNMutationISFSForLearners(parm, individual, instances, label, w, data, inputs, mutate_features, mutate_instances, false, evaluations, minority, classes, max_size, true, r);

    // Get reduction
    // The reduction improvement in the fitness of removing:
    // an instance is: selected features/(instances*features)
    // a feature is: selected instances/(instances*features)
    scalar sel_i = 0.0, sel_f = 0.0;

    for (int i = 0; i < instances; i++)
        if (individual[i])
            sel_i++;

    if (mutate_features) {
        for (int i = instances; i < instances+inputs; i++)
            if (individual[i])
                sel_f++;
    } else
        sel_f = inputs;

    // Nothing to do?
    if (sel_f == 0 || sel_i == 0)
        return 0;

    // If adding instances store furthest nearest neighbor distance
    scalar *furthest = NULL;
    if (mutate_instances)
        MALLOC(furthest, scalar, instances, "AddIorFMutation: furthest");

    bool *Ok, *Ok_so_far;
    MALLOC(Ok_so_far, bool, instances, "AddIorFMutation: Ok_so_far");
    MALLOC(Ok, bool, instances, "AddIorFMutation: Ok");

    for (int i = 0; i < instances; i++) {
        // MODIFIED 13/December/2010
        // Temporarily remove instance for evaluation: LOO
        backup = individual[i];
        individual[i] = 0;

        if (kNNCategory(data[i], data, NULL, instances, parm->k, inputs, label, classes, nearest, individual, NULL, NULL, -1, NULL, individual+instances) == label[i]) {
            accuracy++;
            Ok_so_far[i] = true;
        } else
            Ok_so_far[i] = false;

        // Store furthest
        if (mutate_instances) {
            if (nearest[parm->k-1] != -1)
                furthest[i] = EuclideanDistanceSubspace(data[i], data[nearest[parm->k-1]], inputs, individual+instances);
            else
                furthest[i] = INFINITE;
        }

        individual[i] = backup;
    }

    accuracy /= instances;

    // Obtain G
    if (parm->performance_measure == pfGMean) {
        TP = TN = FP = FN = 0.0;
        // Store G value with all selected instances
        for (int i = 0; i < instances; i++) {
            if (Ok_so_far[i]) {
                if (label[i] != minority)
                    TN++;
                else
                    TP++;

            } else {
                if (label[i] != minority)
                    FP++;
                else
                    FN++;
            }
        }
        Sp = TN/(TN+FP);
        Sn = TP/(TP+FN);
        accuracy = pow(Sn * Sp, 0.5);
    }

    scalar best_so_far = Fitness(parm, accuracy, sel_f, inputs, sel_i, instances, &red);

    individual_evaluations += instances;

    // Instances/features are added in random order
    int *order;
    ALLOCA(order, int, instances+inputs, "AddIorFMutation: order");

    for (int i = 0; i < instances+inputs; i++)
        order[i] = i;

    if (mutate_instances && mutate_features && parm->isfsmutation_random_order)
        RandomizeVector(order, instances+inputs);
    else {
        RandomizeVector(order, instances);
        RandomizeVector(order+instances, inputs);
    }

    // Test if added each instance/feature affects the classification of any other instance
    // Add each unselected instance/feature and test
    int added_instances = 0, added_features = 0;

    // Mutate features, instances or both?
    int end = mutate_features ? instances+inputs : instances;

    int begin = mutate_instances ? 0 : instances;

    for (int i = begin; i < end; i++) {
        if (!individual[order[i]]) {
            if (order[i] >= instances)
                individual[order[i]] = 1;
            else
                individual[order[i]] = (!parm->multiselection || parm->k == 1) ? 1 : RANDOM(1, (parm->k/2)+2);

            // Get accuracy
            scalar accuracy;

            accuracy = 0.0;

            if (order[i] < instances) {
                // Recalculate for affected instances
                for (int j = 0; j < instances; j++) {
                    if (order[i] != j && EuclideanDistanceSubspace(data[order[i]], data[j], inputs, individual+instances) < furthest[j]) {
                        backup = individual[j];
                        individual[j] = 0;

                        if ((Ok[j] = (kNNCategory(data[j], data, individual+instances, instances, parm->k, inputs, label, classes, nearest, individual, NULL, NULL, -1, NULL, individual+instances)
                                      == label[j])))
                            accuracy++;

                        // Recalculate furthest
                        if (nearest[parm->k-1] != -1)
                            furthest[j] = EuclideanDistanceSubspace(data[j], data[nearest[parm->k-1]], inputs, individual+instances);
                        else
                            furthest[j] = INFINITE;
                        individual[j] = backup;

                        individual_evaluations++;
                    } else
                        accuracy += (Ok[j] = Ok_so_far[j]);
                }
            } else {
                for (int j = 0; j < instances; j++) {
                    backup = individual[j];
                    individual[j] = 0;

                    if ((Ok[j] = (kNNCategory(data[j], data, individual+instances, instances, parm->k, inputs, label, classes, nearest, individual, NULL, NULL, -1, NULL, individual+instances)
                                  == label[j])))
                        accuracy++;

                    individual[j] = backup;

                    individual_evaluations ++;
                }
            }

            accuracy /= instances;

            if (parm->performance_measure == pfGMean) {
                TP = TN = FP = FN = 0.0;
                // Store G value with all selected instances
                for (int j = 0; j < instances; j++) {
                    if (Ok[j]) {
                        if (label[j] != minority)
                            TN++;
                        else
                            TP++;

                    } else {
                        if (label[j] != minority)
                            FP++;
                        else
                            FN++;
                    }
                }
                Sp = TN/(TN+FP);
                Sn = TP/(TP+FN);
                accuracy = pow(Sn * Sp, 0.5);
            }

            // Get fitness
            if (order[i] < instances)
                fitness = Fitness(parm, accuracy, sel_f, inputs, sel_i+1, instances, &red);
            else
                fitness = Fitness(parm, accuracy, sel_f+1, inputs, sel_i, instances, &red);

            // Fitness improvement is: accuracy improvement+reduction improvement
            // If negative improvement undo
            if (fitness <= best_so_far)
                individual[order[i]] = 0;
            else {
                if (order[i] < instances) {
                    added_instances++;
                    sel_i++;
                } else {
                    added_features++;
                    sel_f++;
                }

                // Update
                best_so_far = fitness;
            }
        }
    }

    if (parm->verbose >= 3) {
        char *method;
        if (mutate_features && mutate_instances)
            method = "ADD IS+FS";
        else if (mutate_features)
            method = "ADD FS";
        else
            method = "ADD IS";
        fprintf(parm->out_file, "%s mutation added %d/%d instances and %d/%d features\n", method, added_instances, instances, added_features, inputs);
    }

    // Fractional evaluations of the fitness functions
    *evaluations += (individual_evaluations*instances*inputs)/(((scalar) instances) *instances*inputs);

    if (mutate_instances)
        FREE(furthest);

    return added_instances+added_features;
}

// #define SIGMOID(x, y)        (tanh(10.*exp(-.01*(y))*(x))/tanh(10.*exp(-.01*(y))))
#define SIGMOID(x, y)        (tanh(10.*exp(-.025*(y))*(x))/tanh(10.*exp(-.025*(y))))

scalar Fitness(const t_Parameters *parm, scalar accuracy, int selected_inputs, int inputs, int selected_instances, int instances, scalar *reduction)
{
    scalar fitness, ired, fred;

    *reduction = 1.0 - (((scalar) selected_instances)/instances) * (((scalar) selected_inputs)/inputs);
    if (parm->exp_reduction)
        *reduction = SIGMOID(*reduction, inputs*instances);

    ired = 1.0 - (((scalar) selected_instances)/instances);
    if (parm->exp_redinstance)
        ired = SIGMOID(ired, instances);

    fred = 1.0 - (((scalar) selected_inputs)/inputs);
    if (parm->exp_redfeature)
        fred =  SIGMOID(fred, inputs);

    fitness = parm->ok_w*accuracy + parm->ired_w*ired + parm->fred_w*fred + parm->reduction_w*(*reduction);

    return fitness;
}
