/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that deal with the individuals
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include "isfs.h"
#include "macros.h"
#include "common.h"

// IS+FS Mutation for a support vector machines as wrapper

int RNNMutationISFSForLearners(const t_Parameters *parm, int *individual, int instances, int *label, scalar *w, scalar **data, int inputs, bool mutate_features, bool mutate_instances, bool rfe,
                               scalar *evaluations, int minority, int classes, int max_size, bool add_instances, scalar *r)
{
    scalar accuracy, reduction, current_fitness, best_fitness;
    int *order, changed_instances = 0, changed_features = 0;

    // Best fitness so far
    best_fitness = EvaluateISFSIndividual(parm, individual, &accuracy, &reduction, data, instances, inputs, label, w, evaluations, minority, NULL, max_size, classes, r, NULL);

    // Test if removing each instance negatively affects the fitness
    // Remove each instance and test
    // Mutate features, instances or both?
    int begin = mutate_instances ? 0 : instances;
    int end = mutate_features ? instances+inputs : instances;

    // Randomize the order of checking selected instances?
    MALLOC(order, int, instances+inputs, "RNNMutationISFSForSVM: order");

    if (mutate_instances && mutate_features && parm->isfsmutation_random_order) {
        for (int i = begin; i < end; i++)
            order[i] = i;

        RandomizeVector(order, instances+inputs);
    } else if (mutate_instances && mutate_features && parm->isfs_inputs_first) {
        for (int i = 0; i < inputs; i++)
            order[i] = i+instances;
        for (int i = 0; i < instances; i++)
            order[i+inputs] = i;
        RandomizeVector(order, inputs);
        RandomizeVector(order+inputs, instances);
    } else {
        for (int i = begin; i < end; i++)
            order[i] = i;

        if (mutate_instances)
            RandomizeVector(order, instances);

        if (mutate_features && !rfe)
            RandomizeVector(order+instances, inputs);

        // If RFE mutation order features accordingly
        else if (mutate_features && rfe) {
            // Order instances using RFE
            t_Feat *f_info;

            ALLOCA(f_info, t_Feat, inputs, "RNNMutationISFS: f_info");
            RFE(parm, f_info, inputs, data, instances, label, individual+instances);

            //Sort features from best to worst ranking
            qsort(f_info, inputs, sizeof(t_Feat), CompareRanking);

            // Order of features

            for (int i = inputs - 1, j = 0 ; i >= 0; i--, j++)
                order[j] = f_info[i].index;
        }
    }

    // Use as criterion for acceptance fitness
    for (int i = begin; i < end; i++) {
        // Do not do for last instance/feature
        if ((individual[order[i]] && !add_instances) || (!individual[order[i]] && add_instances)) {
            // Backup to undo if necessary
            int backup = individual[order[i]];
            individual[order[i]] = add_instances ? 1 : 0;

            // Get fitness
            current_fitness = EvaluateISFSIndividual(parm, individual, &accuracy, &reduction, data, instances, inputs, label, w, evaluations, minority, NULL, max_size, classes, r, NULL);

            // Fitness improvement?
            // If negative improvement
            if (current_fitness < best_fitness)
                // Undo
                individual[order[i]] = backup;
            else {
                if (order[i] < instances)
                    changed_instances++;
                else
                    changed_features++;

                // Update best fitness so far
                best_fitness = current_fitness;
            }

            // Evaluations of the fitness functions
            (*evaluations)++;
        }
    }

    if (parm->verbose >= 3) {
        char *method;
        if (mutate_features && mutate_instances)
            method = "RNN IS+FS";
        else if (mutate_features)
            method = "RNN FS";
        else
            method = "RNN IS";

        fprintf(parm->out_file, "%s %s %3d/%3d instances and %d/%d features\n", method, add_instances ? "added" : "removed", changed_instances, instances, changed_features, inputs);
    }

    FREE(order);

    return changed_instances+changed_features;
}
