/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <search.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "isfs.h"

// Evaluate selection threshold using MFDR
// B. Seijo-Pardo, V. Bolón-Canedo and A. Alonso-Betanzos, "On developing an automatic threshold applied to feature selection", Information Fusion (45):227-245, 2019.

void MaximumFishersDiscriminantRatioEvaluation(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, scalar *votes_fs, scalar *coefficient, scalar *r, scalar *complexity,
        int classes, scalar *threshold_fs, int n_thresholds_fs, int *subset)
{
    int *S;

    MALLOC(S, int, inputs, "MaximumFishersDiscriminantRatioEvaluation: S");

    // Get coefficient for every threshold
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    for (int fs = 0; fs < n_thresholds_fs; fs++) {
        int selected = 0;

        scalar th_fs = threshold_fs[fs];

        SHOWPROGRESS(parm->show_progress > 0, "Evaluating thresholds FS: %*d/%d\r", WIDTH(n_thresholds_fs), fs+1, n_thresholds_fs);

        // Select features
        for (int i = 0; i < inputs; i++)
            if (votes_fs[i] >= th_fs) {
                S[i] = 1;
                selected++;
            } else
                S[i] = 0;

        r[fs] = 1.0 - ((scalar) selected)/inputs;

        complexity[fs] = MaximumFishersDiscriminantRatio(data, labels, instances, inputs, classes, S, subset);
        coefficient[fs] = complexity[fs] * parm->alpha_ok + parm->alpha_red * r[fs];
    }

    FREE(S);

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}

scalar FishersDiscriminantRatio(scalar **data, int *labels, int instances, int feature, int classes, int *subset)
{
    scalar f1 = 0.0, sum = 0.0, p[classes], sigma[classes], mu[classes];

    ZeroVector(p, NULL, classes);
    ZeroVector(sigma, NULL, classes);
    ZeroVector(mu, NULL, classes);

    // Get mean, variance and class probabilities
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            p[labels[i]]++;
            mu[labels[i]] += data[i][feature];
            sigma[labels[i]] += SQR(data[i][feature]);
        }

    for (int i = 0; i < classes; i++)
        if (p[i] > 1.0)  {
            sigma[i] = (sigma[i] - SQR(mu[i])/p[i])/(p[i]-1.0);
            mu[i] /= p[i];
            p[i] /= instances;
        } else if (p[i] == 1)
            p[i] = 1.0/instances;

    for (int i = 0; i < classes; i++)
        if (p[i] > 0) {
            sum += p[i]*sigma[i];

            for (int j = 0; j < classes; j++)
                if (i != j)
                    f1 += p[i]*p[j]*SQR(mu[i]-mu[j]);
        }

    f1 /= sum;

    return f1;
}

// scalar AverageFishersDiscriminantRatio(scalar **data, int *labels, int instances, int features, int classes, int *subpace)
// {
//     scalar f1 = 0.0;
//     int count = 0;
//
//     for (int i = 0; i < features; i++)
//         if (subpace[i]) {
//             f1 += FishersDiscriminantRatio(data, labels, instances, i, classes);
//             count++;
//         }
//
//     if (count > 1)
//         f1 /= count;
//
//     return f1;
// }

scalar MaximumFishersDiscriminantRatio(scalar **data, int *labels, int instances, int features, int classes, int *subpace, int *subset)
{
    scalar f1 = -1.0;

    for (int i = 0; i < features; i++)
        if (subpace[i]) {
            scalar value = FishersDiscriminantRatio(data, labels, instances, i, classes, subset);
            if (value > f1)
                f1 = value;
        }

    return f1;
}
