/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <search.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "isfs.h"

// Evaluate selection threshold using Maximum (individual) feature efficiency

// B. Seijo-Pardo, V. Bolón-Canedo and A. Alonso-Betanzos, "On developing an automatic threshold applied to feature selection", Information Fusion (45):227-245, 2019.

void MulticlassMaximumFeatureEfficiency(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, scalar *votes_fs, scalar *coefficient, scalar *r, scalar *complexity,
                                        int classes, scalar *threshold_fs, int n_thresholds_fs)
{
    int *S;

    MALLOC(S, int, inputs, "MulticlassMaximumFeatureEfficiency: S");

    // Get coefficient for every threshold
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    for (int fs = 0; fs < n_thresholds_fs; fs++) {
        int selected = 0;

        scalar th_fs = threshold_fs[fs];

        SHOWPROGRESS(parm->show_progress > 0, "Evaluating thresholds FS: %*d/%d\r", WIDTH(n_thresholds_fs), fs+1, n_thresholds_fs);

        // Select features
        for (int i = 0; i < inputs; i++)
            if (votes_fs[i] >= th_fs) {
                S[i] = 1;
                selected++;
            } else
                S[i] = 0;

        r[fs] = 1.0 - ((scalar) selected)/inputs;

        complexity[fs] = -1.0;

        for (int i = 1; i < classes; i++)
            for (int j = 0; j < i; j++) {
                scalar value = fabs(MaximumFeatureEfficiency(data, instances, labels, i, j, S, inputs, NULL));
                if (value > complexity[fs])
                    complexity[fs] = value;
            }

        coefficient[fs] = complexity[fs] * parm->alpha_ok + parm->alpha_red * r[fs];
    }

    FREE(S);

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}

scalar MaximumFeatureEfficiency(scalar **data, int instances, int *labels, int c1, int c2, int *subspace, int features, int *subset)
{
    scalar f3 = 0.0, minmax, maxmin;
    int subspace_size = 0;

    for (int i = 0; i < features; i++)
        if (subspace[i]) {
            subspace_size++;

            minmax = MIN(MaxValueOfFeatureInClass(data, instances, labels, i, c1, subset), MaxValueOfFeatureInClass(data, instances, labels, i, c2, subset));
            maxmin = MAX(MinValueOfFeatureInClass(data, instances, labels, i, c1, subset), MinValueOfFeatureInClass(data, instances, labels, i, c2, subset));

            // Count instances of both classes in the overlapping region
            for (int j = 0; j < instances; j++)
                if ((labels[j] == c1 || labels[j] == c2) && (data[j][i] >= minmax || data[j][i] <= maxmin))
                    f3++;
        }

    // Count size
    int data_size = 0;
    for (int j = 0; j < instances; j++)
        if (labels[j] == c1 || labels[j] == c2)
            data_size++;

    return f3/(subspace_size*data_size);
}

void ComplexityFusion(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, scalar *votes_fs, scalar *coefficient, scalar *r, scalar *complexity, int classes,
                      scalar *threshold_fs, int n_thresholds_fs, int *subset)
{
    int *S;

    MALLOC(S, int, inputs, "ComplexityFusion: S");

    // Get coefficient for every threshold
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    for (int fs = 0; fs < n_thresholds_fs; fs++) {
        int selected = 0;

        scalar th_fs = threshold_fs[fs];

        SHOWPROGRESS(parm->show_progress > 0, "Evaluating thresholds FS: %*d/%d\r", WIDTH(n_thresholds_fs), fs+1, n_thresholds_fs);

        // Select features
        for (int i = 0; i < inputs; i++)
            if (votes_fs[i] >= th_fs) {
                S[i] = 1;
                selected++;
            } else
                S[i] = 0;

        r[fs] = 1.0 - ((scalar) selected)/inputs;

        scalar mfe = -1.0, cur_mfe, voor = 1.0;
        for (int i = 1; i < classes; i++)
            for (int j = 0; j < i; j++) {
                if ((cur_mfe = MaximumFeatureEfficiency(data, instances, labels, i, j, S, inputs, subset)) > mfe)
                    mfe = cur_mfe;

                // Sum instead of product following authors' recommendation
                voor *= fabs(VolumeOfOverlapRegion(data, instances, labels, i, j, inputs, S, subset));
            }

        scalar mfdr = MaximumFishersDiscriminantRatio(data, labels, instances, inputs, classes, S, subset);

        complexity[fs] = (mfe + (1.0-voor) + mfdr)/3.0;

        coefficient[fs] = complexity[fs] * parm->alpha_ok + parm->alpha_red * r[fs];
    }

    FREE(S);

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}
