/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "isfs.h"
#include "macros.h"

// eMultiBoost algorithm [Webb, 2000]

#define MAXIMUM_BETA    (10.0E+10)
#define MINIMUM_WEIGHT  (10.0E-8)

scalar UpdateMultiBoostWeights(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, scalar *w, scalar *alpha, int inputs, int classes, bool initialize, int T, int t,
                               t_Classifier boostis_classifier, int **index, bool advance, int *subspace)
{
    scalar e_i, betta, sum;
    bool *ok;
    static int k, *I;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for IS/FS boosting");

    // Initialize?
    if (initialize) {
        int n = sqrt((scalar) T);
        k = 1;
        MALLOC(I, int, T+1, "UpdateMultiBoostWeights");

        I[0] = 0;
        for (int i = 1; i < n; i++)
            I[i] = ceil(i * T/n);
        for (int i = n; i < T+1; i++)
            I[i] = T;

        return 0.0;
    }

    MALLOC(ok, bool, instances, "UpdateMultiBoostWeights: ok");

    // Normalize w[] to sum 1
    sum = 0.0;
    for (int i = 0; i < instances; i++)
        sum += w[i];
    for (int i = 0; i < instances; i++)
        w[i] *= 1.0/sum;

    // Test previous error for alpha
    e_i = GetWeightedError(parm, boostis_classifier, S, data, inputs, instances, labels, classes, ok, w, true, index, subspace);

    // Adjust alpha
    if (e_i >= 0.5)
        *alpha = 0.00001;           // Remove classifier
    else if (e_i <= ZERO)
        // Weight classifier
        *alpha = log(1.0/MAXIMUM_BETA);
    else {
        betta = MIN(MAXIMUM_BETA, e_i/(1.0-e_i));

        // Weight classifier
        *alpha = log(1.0/betta);
    }

    sum = 0.0;

    // Initialize weights?
    if (e_i >= 0.5 || e_i <= 0.0 || I[k] == t) {
        for (int i = 0; i < instances; i++)
            sum += (w[i] = Poisson());
        if (advance)
            k++;
    } else {
        for (int i = 0; i < instances; i++)
            if (ok[i]) {
                w[i] /= (2.0*(1.0-e_i));

                if (w[i] < MINIMUM_WEIGHT)
                    w[i] = MINIMUM_WEIGHT;

                sum += w[i];
            } else
                sum += (w[i] /= (2.0*e_i));
    }

    // Normalize w[] to sum 1
    for (int i = 0; i < instances; i++)
        w[i] *= 1.0/sum;

    FREE(ok);

    return e_i;
}
