/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "common.h"
#include "macros.h"
#include "fs.h"
#include "isfs.h"


// RIIC measure from L. Morán-Fernández et al. (2017) "Centralized vs. distributed feature selection methods based on data complexity measures", Knowledge-Based Systems 117:27-45.

scalar RIICComplexityMeasure(scalar **data, int instances, int *labels, int features, int *subset, int *subspace)
{
    scalar intra_dist = 0.0, inter_dist = 0.0;

    // Nearest friend and enemy for every instance
    if (subset) {
        scalar enemy_d, friend_d;

        for (int i = 0; i < instances; i++)
            if (subset[i]) {
                enemy_d = friend_d = INFINITE;

                for (int j = 0; j < instances; j++)
                    if (subset[j] && i != j) {
                        scalar dist = EuclideanDistanceSubspace(data[i], data[j], features, subspace);

                        if (labels[i] == labels[j] && dist < friend_d)
                            friend_d = dist;
                        else if (labels[i] != labels[j] && dist < enemy_d)
                            enemy_d = dist;
                    }

                intra_dist += friend_d;
                inter_dist += enemy_d;
            }
    } else {
        // Faster implementation
        scalar *enemy_d, *friend_d;

        MALLOC(enemy_d, scalar, instances, "RIICComplexityMeasure: enemy_d");
        MALLOC(friend_d, scalar, instances, "RIICComplexityMeasure: friend_d");
        for (int i = 0; i < instances; i++)
            enemy_d[i] = friend_d[i] = INFINITE;

        for (int i = 0; i < instances; i++)
            for (int j = 0; j < i; j++) {
                scalar dist = EuclideanDistanceSubspace(data[i], data[j], features, subspace);

                if (labels[i] == labels[j]) {
                    if (dist < friend_d[i])
                        friend_d[i] = dist;
                    if (dist < friend_d[j])
                        friend_d[j] = dist;
                } else {
                    if (dist < enemy_d[i])
                        enemy_d[i] = dist;
                    if (dist < enemy_d[j])
                        enemy_d[j] = dist;
                }
            }

        for (int j = 0; j < instances; j++) {
            intra_dist += friend_d[j];
            inter_dist += enemy_d[j];
        }

        FREE(friend_d);
        FREE(enemy_d);
    }

    return intra_dist/inter_dist;
}

void RIICThresholdEvaluation(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, scalar *votes_fs, scalar *coefficient, scalar *r, scalar *complexity, int classes,
                             scalar *threshold_fs, int n_thresholds_fs)
{
    int *S;

    MALLOC(S, int, inputs, "RIICThresholdEvaluation: S");

    // Get coefficient for every threshold
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    for (int fs = 0; fs < n_thresholds_fs; fs++) {
        int selected = 0;

        scalar th_fs = threshold_fs[fs];

        SHOWPROGRESS(parm->show_progress > 0, "RIIC Evaluating thresholds FS: %*d/%d\r", WIDTH(n_thresholds_fs), fs+1, n_thresholds_fs);

        // Select features
        for (int i = 0; i < inputs; i++)
            if (votes_fs[i] >= th_fs) {
                S[i] = 1;
                selected++;
            } else
                S[i] = 0;

        r[fs] = ((scalar) selected)/inputs;
        complexity[fs] = RIICComplexityMeasure(data, instances, labels, inputs, NULL, S);

        coefficient[fs] = complexity[fs] * parm->alpha_ok + parm->alpha_red * r[fs];
    }

    FREE(S);

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}
