/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <pvm3.h>
#include <string.h>
#include <unistd.h>
#include "isfs.h"
#include "macros.h"
#include "common.h"
#include "pvm-local.h"

time_t ParallelInstanceFeatureSelectionForAllStrata(const t_Parameters *parm, int n_slaves, bool spawn, bool kill, char *parameters, scalar ***is_data,  int *is_instances, int **is_labels,
        int **is_isel, int **is_fsel, int ***is_index, int n_strata, int minority, char *out_file, int inputs, int classes, scalar ratio, scalar *average_is, scalar *average_fs, int thread)
{
    int *tids = NULL/*, guardian = 0*/;
    int *fs_selection, *is_selection, selection_times = 0, *running_stratum, *runnning_subs, restarted = 0;
    time_t init_time, sending_time = 0;

    SHOWPROGRESS(parm->show_progress > 0, "Internal strata: %d/%0*d/%0*d", n_strata, WIDTH(n_strata), 0, WIDTH(n_strata), n_strata);

    // Partition into disjoint subspaces
    int ***subspaces, **subspace_size, n_subspaces;
    scalar **data;

    // Create subspaces
    int max_subspace_size = inputs;
    if (parm->internal_subspace_size) {
        n_subspaces = (int) ceil(((scalar) inputs)/parm->internal_subspace_size);
        max_subspace_size = parm->internal_subspace_size;
    } else
        n_subspaces = 1;

    MALLOC(subspaces, int **, n_strata, "ParallelInstanceFeatureSelectionForAllStrata subspaces");
    MALLOC(subspace_size, int *, n_strata, "ParallelInstanceFeatureSelectionForAllStrata subspace_size");
    for (int i = 0; i < n_strata; i++) {
        MALLOC(subspaces[i], int *, n_subspaces, "ParallelInstanceFeatureSelectionForAllStrata subspaces[i]");
        MALLOC(subspace_size[i], int, n_subspaces, "ParallelInstanceFeatureSelectionForAllStrata subspace_size[i]");
        for (int j = 0; j < n_subspaces; j++)
            MALLOC(subspaces[i][j], int, max_subspace_size, "ParallelInstanceFeatureSelectionForAllStrata subspaces[i][j]");

        FeaturePartition(inputs, n_subspaces, subspace_size[i], subspaces[i]);
    }

    MALLOC(fs_selection, int, max_subspace_size, "ParallelInstanceFeatureSelectionForAllStrata fs_selection");
    MALLOC(is_selection, int, parm->internal_subset_size, "ParallelInstanceFeatureSelectionForAllStrata is_selection");
    MALLOC(data, scalar *, parm->internal_subset_size, "ParallelInstanceFeatureSelectionForAllStrata data");
    MALLOC(running_stratum, int, n_slaves, "ParallelInstanceFeatureSelectionForAllStrata: running_stratum");
    MALLOC(runnning_subs, int, n_slaves, "ParallelInstanceFeatureSelectionForAllStrata: running_subs");
    for (int i = 0; i < parm->internal_subset_size; i++)
        MALLOC(data[i], scalar, max_subspace_size, "ParallelInstanceFeatureSelectionForAllStrata data[i]");

    //  Too many slaves?
    if (n_slaves > n_strata*n_subspaces) {
        n_slaves = n_strata * n_subspaces;
        SHOWPROGRESS(parm->show_progress > 2, "Number of slaves adjusted to %d\n", n_slaves);
    }

    if (spawn) {
        // PVM Spawning processes if not already running
        char *args[5];
        MALLOC(args[0], char, strlen(parameters)+1, "ParallelInstanceFeatureSelectionForAllStrata args[0]");
        strcpy(args[0], parameters);       // Parameters file name
        MALLOC(args[1], char, 5, "ParallelInstanceFeatureSelectionForAllStrata args[1]");
        strcpy(args[1], "none");
        args[2] = getcwd(NULL, 0);      // Current directory for slaves
        MALLOC(args[3], char, 5, "ParallelInstanceFeatureSelectionForAllStrata: args[3]");
        sprintf(args[3], "%d", thread);
        args[4] = NULL;

        // Start slaves
        MALLOC(tids, int, n_slaves, "ParallelInstanceFeatureSelectionForAllStrata tids");
        PVMSpawn(n_slaves, "wrapper-is+fs", tids, args, NULL /*&guardian*/, out_file, parm->show_progress);

        SHOWPROGRESS(parm->show_progress > 2, "Started %d slaves for performing instance selection\n", n_slaves);
    } else
        OnErrorExit(ERR_PVM, "Slaves must be spawned every time");

    // Instance selection made by slaves
    // Send strata to first slaves
    int stratum = 0, subs = 0;
    for (int p = 0; p < n_slaves && stratum < n_strata; p++) {
        init_time = time(NULL);

        // Get data ready
        if ((parm->internal_subset_size && is_instances[stratum] > parm->internal_subset_size) || (parm->internal_subspace_size && subspace_size[stratum][subs] > parm->internal_subspace_size))
            OnErrorExit(ERR_UNSPECTED_RESULT, "Subset/subspace too big");
        for (int i = 0; i < is_instances[stratum]; i++)
            for (int j = 0; j < subspace_size[stratum][subs]; j++)
                data[i][j] = is_data[stratum][i][subspaces[stratum][subs][j]];

        SendDataToSlaveSubspace(tids[p], stratum, is_instances[stratum], subspace_size[stratum][subs], classes, minority, ratio, is_labels[stratum], is_index ? is_index[stratum] : NULL, data, subs,
                                thread);
        running_stratum[p] = stratum;
        runnning_subs[p] = subs;

        sending_time += time(NULL) - init_time;

        SHOWPROGRESS(parm->show_progress > 2, "\nSENT data to slave %d\n", tids[p]);

        subs = (subs+1) % n_subspaces;
        if (!subs)
            stratum++;
    }

    // Receive results and send more strata if needed
    int received = 0;

    // Start with empty selection
    for (int i = 0; i < n_strata; i++)
        ZeroVector(NULL, is_isel[i], is_instances[i]);

    while (received < n_strata*n_subspaces) {
        int stratum_received, bytes, mstag, tid, subspace_received;
        time_t tid_time;

        int bufid = ReceiveSelection(&received, &stratum_received, &subspace_received, is_selection, fs_selection, is_instances, subspace_size, &tid_time, thread);

        // If no message returned check dead slaves and continue
        if (bufid == 0) {
            restarted += RestartDeadSlaves(tids, n_slaves, parameters, parm->show_progress, out_file, running_stratum, runnning_subs, is_instances, is_data, subspace_size, is_index, subspaces,
                                           is_labels, classes, minority, ratio, &sending_time, thread);
            continue;
        }

        // Obtain average instance and feature selection
        selection_times++;
        AverageFeatureAndInstanceSelection(average_is, average_fs, is_selection, fs_selection, is_instances[stratum_received], subspace_size[stratum_received][subspace_received], selection_times);

        // Update selections
        int i_selected = 0;
        for (int i = 0; i < is_instances[stratum_received]; i++) {
            is_isel[stratum_received][i] += is_selection[i];
            i_selected += (is_selection[i] != 0);
        }

        int f_selected = 0;
        for (int i = 0; i < subspace_size[stratum_received][subspace_received]; i++)
            f_selected += (is_fsel[stratum_received][subspaces[stratum_received][subspace_received][i]] = fs_selection[i]);

        SHOWPROGRESS(parm->show_progress > 2, "\rReceived subsets: %*d/%d (%*d/%*d) IS: %*d/%*d FS: %*d/%*d", WIDTH(n_strata*n_subspaces), received, n_strata*n_subspaces,
                     WIDTH(n_strata), stratum_received+1,  WIDTH(n_subspaces), subspace_received+1,
                     WIDTH(parm->internal_subset_size), i_selected, WIDTH(parm->internal_subset_size), is_instances[stratum_received],
                     WIDTH(max_subspace_size), f_selected, WIDTH(max_subspace_size), subspace_size[stratum_received][subspace_received]);

        // Get tid of sending slave
        pvm_bufinfo(bufid, &bytes, &mstag, &tid);

        // Send more?
        if (stratum < n_strata) {
            init_time = time(NULL);

            // Get data ready
            for (int i = 0; i < is_instances[stratum]; i++)
                for (int j = 0; j < subspace_size[stratum][subs]; j++)
                    data[i][j] = is_data[stratum][i][subspaces[stratum][subs][j]];

            SendDataToSlaveSubspace(tid, stratum, is_instances[stratum], subspace_size[stratum][subs], classes, minority, ratio, is_labels[stratum], is_index ? is_index[stratum] : NULL, data, subs,
                                    thread);

            for (int p = 0; p < n_slaves; p++)
                if (tids[p] == tid) {
                    running_stratum[p] = stratum;
                    runnning_subs[p] = subs;
                    break;
                }

            sending_time += time(NULL) - init_time;

            SHOWPROGRESS(parm->show_progress > 2, "\nSENT data to slave %d\n", tid);

            subs = (subs+1) % n_subspaces;
            if (!subs)
                stratum++;

            sending_time += time(NULL) - init_time;
        }
    }

    SHOWPROGRESS(parm->show_progress > 2, "%c", '\n');

    if (kill) {
        // Kill spawned processes and guardian
        // pvm_kill(guardian);
        for (int i = 0; i < n_slaves; i++)
            pvm_kill(tids[i]);
    } else
        OnErrorExit(ERR_PVM, "Slaves must be killed every time");

    // Free memory
    FREE(fs_selection);
    FREE(is_selection);
    for (int i = 0; i < n_strata; i++)
        FREE(subspace_size[i]);
    FREE(subspace_size);
    for (int i = 0; i < n_strata; i++) {
        for (int j = 0; j < n_subspaces; j++)
            FREE(subspaces[i][j]);
        FREE(subspaces[i]);
    }
    for (int i = 0; i < parm->internal_subset_size; i++)
        FREE(data[i]);
    FREE(subspaces);
    FREE(data);
    FREE(running_stratum);
    FREE(runnning_subs);

    if (parm->verbose > 1)
        fprintf(parm->out_file, "Restarted slaves: %d\n", restarted);

    return sending_time;
}

void SendDataToSlaveSubspace(int tid, int stratum, int instances, int inputs, int classes, int minority, scalar ratio, int *labels, int **index, scalar **data, int subspace, int thread)
{
    int use_index = 0;

    // Send info to slave
    if (pvm_initsend(PvmDataDefault) < 0)
        pvm_perror("Sent by master: INITSEND");
    if (pvm_pkint(&stratum, 1, 1))
        pvm_perror("Sent by master: stratum");
    if (pvm_pkint(&subspace, 1, 1))
        pvm_perror("Sent by master: subspace");
    if (pvm_pkint(&instances, 1, 1))
        pvm_perror("Sent by master: instances");
    if (pvm_pkint(&inputs, 1, 1))
        pvm_perror("Sent by master: inputs");
    if (pvm_pkint(&classes, 1, 1))
        pvm_perror("Sent by master: classes");
    if (pvm_pkint(&minority, 1, 1))
        pvm_perror("Sent by master: minority");
    if (pvm_pkscalar(&ratio, 1, 1))
        pvm_perror("Sent by master: ratio");

    // Labels
    if (pvm_pkint(labels, instances, 1))
        pvm_perror("Sent by master: labels");

    // Index?
    use_index = (index != NULL);
    if (pvm_pkint(&use_index, 1, 1))
        pvm_perror("Sent by master: use_index");

    if (index) {
        for (int i = 0; i < instances; i++) {
            int count = 0;

            // Count size and send data and index
            for (; index[i][count] != -1; count++);
            if (pvm_pkint(&count, 1, 1))
                pvm_perror("Sent by master: count");

            if (pvm_pkint(index[i], count+1, 1))
                pvm_perror("Sent by master: index[i]");
            if (pvm_pkscalar(data[i], count, 1))
                pvm_perror("Sent by master: data[i]");
        }
    } else
        // Dense data
        for (int i = 0; i < instances; i++)
            if (pvm_pkscalar(data[i], inputs, 1))
                pvm_perror("Sent by master: data[i]");

    if (pvm_send(tid, SENT_BY_MASTER+thread))
        pvm_perror("Sent by master: SEND");
}

int ReceiveSelection(int *received, int *stratum_received, int *subspace_received, int *is_selection, int *fs_selection, int *is_instances, int **subspace_size, time_t *tid_time, int thread)
{
    int inputs, instances;
    struct timeval tmout;

    // Wait a limited time to recover from dead slaves
    tmout.tv_sec = 60;
    tmout.tv_usec = 0;
    int bufid = pvm_trecv(-1, SENT_BY_SLAVE+thread, &tmout);

    if (bufid < 0)
        pvm_perror("Received by master: RECEIVE");
    else if (bufid == 0)
        return bufid;

    (*received)++;

    if (pvm_upkint(stratum_received, 1, 1))
        pvm_perror("Received from slave: stratum_received");
    if (pvm_upkint(subspace_received, 1, 1))
        pvm_perror("Received from slave: subspace_received");
    if (pvm_upkint(&instances, 1, 1))
        pvm_perror("Received from slave: stratum_received");
    if (instances != is_instances[*stratum_received])
        OnErrorExit(ERR_UNSPECTED_RESULT, "Instances count does not match");
    if (pvm_upkint(&inputs, 1, 1))
        pvm_perror("Received from slave: subspace_received");
    if (inputs != subspace_size[*stratum_received][*subspace_received])
        OnErrorExit(ERR_UNSPECTED_RESULT, "Features count does not match");
    if (pvm_upkint(is_selection, is_instances[*stratum_received], 1))
        pvm_perror("Received from slave: is_instances[stratum_received]");
    if (pvm_upkint(fs_selection, subspace_size[*stratum_received][*subspace_received], 1))
        pvm_perror("Received from slave: subspace_size[stratum_received][subspace_received]");
    if (pvm_upkint((int *) tid_time, 1, 1))
        pvm_perror("Received from slave: tid_time");

    return bufid;
}

int RestartDeadSlaves(int *tids, int ntids, char *parameters, int show_progress, char *out_file, int *stratum, int *subs, int *is_instances, scalar ***is_data, int **subspace_size, int ***is_index,
                      int ***subspaces, int **is_labels, int classes, int minority, scalar ratio, time_t *sending_time, int thread)
{
    scalar **data = NULL;
    int restarted = 0;

    for (int p = 0; p < ntids; p++)
        if (pvm_pstat(tids[p]) == PvmNoTask) {
            // Dead slave start a new one
            char *args[5];
            MALLOC(args[0], char, strlen(parameters)+1, "ParallelInstanceFeatureSelectionForAllStrata args[0]");
            strcpy(args[0], parameters);       // Parameters file name
            MALLOC(args[1], char, 5, "ParallelInstanceFeatureSelectionForAllStrata args[1]");
            strcpy(args[1], "none");
            args[2] = getcwd(NULL, 0);      // Current directory for slaves
            MALLOC(args[3], char, 5, "ParallelInstanceFeatureSelectionForAllStrata: args[3]");
            sprintf(args[3], "%d", thread);
            args[4] = NULL;

            // Start slave
            PVMSpawn(1, "wrapper-is+fs", &tids[p], args, NULL, out_file, show_progress);
            restarted++;

            SHOWPROGRESS(show_progress > 0, "\nRestarted dead slave %d for performing instance selection\n", tids[p]);

            // Send the data of the dead slave
            // Get data ready
            MALLOC(data, scalar *, is_instances[stratum[p]], "RestartDeadSlave: data");
            for (int i = 0; i < is_instances[stratum[p]]; i++)
                MALLOC(data[i], scalar, subspace_size[stratum[p]][subs[p]], "RestartDeadSlave: data[i]");

            for (int i = 0; i < is_instances[stratum[p]]; i++)
                for (int j = 0; j < subspace_size[stratum[p]][subs[p]]; j++)
                    data[i][j] = is_data[stratum[p]][i][subspaces[stratum[p]][subs[p]][j]];

            time_t init_time = time(NULL);
            SendDataToSlaveSubspace(tids[p], stratum[p], is_instances[stratum[p]], subspace_size[stratum[p]][subs[p]], classes, minority, ratio, is_labels[stratum[p]], is_index ? is_index[stratum[p]]
                                    : NULL,  data, subs[p], thread);
            *sending_time += time(NULL) - init_time;

            for (int i = 0; i < is_instances[stratum[p]]; i++)
                FREE(data[i]);
            FREE(data);

            SHOWPROGRESS(show_progress > 0, "\nSENT data to slave %d\n", tids[p]);
        }

    return restarted;
}
