/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "macros.h"
#include "isfs.h"
#include "fs.h"
#include "common.h"
#include "kd.h"
#include "svm.h"
#include "classifiers.h"
#include "maths.h"

// Skalak, D.B. Prototype and Feature Selection by Sampling and Random Mutation Hill Climbing
// Algorithms . in Eleventh International Conference on Machine Learning. 1994.

scalar RMHC_PF1Algorithm(const t_Parameters *parm, int *S, scalar **data, int instances, int features, int *label, scalar *w, int classes, scalar *r, int **index)
{
    int  minority = -1;
    scalar fitness, ok, reduction, evaluations = 0;

    if (classes == 2) {
        int ccount[] = {0, 0};

        for (int i = 0; i < instances; i++)
            ccount[label[i]]++;

        minority = (ccount[0] < ccount[1]) ? 0 : 1;
    }

    if (parm->performance_measure == pfGMean && classes != 2)
        OnErrorExit(ERR_PARAM, "Cannot use G measure with more than 2 classes");

    // Time
    time_t init_time = time(NULL);


    // Create initial individual
    for (int k = 0; k < instances+features; k++)
        S[k] = RANDOM(0.0, 1.0) < parm->init_percent ? 1 : 0;

    // Evaluate initial individual
    fitness = EvaluateISFSIndividual(parm, S, &ok, &reduction, data, instances, features, label, w, &evaluations, minority, NULL, instances, classes, r, index);
    SHOWPROGRESS(parm->show_progress > 1, "RMHC-PF1 step %*d. Fitness: %6.4lf, accuracy: %6.4lf, reduction: %6.4lf\n", WIDTH(parm->rmhc_fp1_iterations), 0, fitness, ok, reduction);

    for (int iteration = 0; iteration < parm->rmhc_fp1_iterations; iteration++) {
        // Random mutation of 1 bit
        int next = RANDOM(0, instances+features);
        S[next] = !S[next];

        // Check mutated individual
        scalar new_ok, new_reduction;
        scalar new_fitness = EvaluateISFSIndividual(parm, S, &new_ok, &new_reduction, data, instances, features, label, w, &evaluations, minority, NULL, instances, classes, r, index);

        // Keep?
        if (new_fitness > fitness || (new_fitness == fitness && new_reduction > reduction)) {
            SHOWPROGRESS(parm->show_progress > 1, "RMHC-PF1 step %*d. Fitness: %6.4lf, accuracy: %6.4lf, reduction: %6.4lf\n", WIDTH(parm->rmhc_fp1_iterations), iteration+1, new_fitness, new_ok,
                         new_reduction);
            fitness = new_fitness;
        } else {    // Undo
            SHOWPROGRESS(parm->show_progress > 1, "RMHC-PF1 step %*d. Fitness: %6.4lf, accuracy: %6.4lf, reduction: %6.4lf\n", WIDTH(parm->rmhc_fp1_iterations), iteration+1, fitness, ok, reduction);
            S[next] = !S[next];
        }

        // If time limit end
        if (parm->time_limit && (time(NULL) - init_time > parm->time_limit)) {
            fprintf(parm->out_file, "Reached time limit at iteration: %d\n", iteration+1);
            break;
        }
    }

    // Obtain reduction
    // Instances
    scalar sel_instances = 0.0, sel_features = 0.0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            sel_instances++;

    // Features
    for (int i = instances; i < instances+features; i++)
        if (S[i])
            sel_features++;

    scalar reduction_ratio = 1.0 - (sel_instances/instances) * (sel_features/features);

    return reduction_ratio;
}
