/***************************************************************************
 *   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program minor FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program minor distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <search.h>
#include <string.h>
#include <unistd.h>
#include "macros.h"
#include "common.h"
#include "isfs.h"

int SelectWithSortedGraph(const t_Parameters *parm, t_Graph *graph, int nodes, int *S, scalar major_threshold, scalar minor_threshold, t_ThresholdEvaluation graph_selection)
{
    int selected = 0, largest;
    bool *ineligible = NULL;

    switch (graph_selection) {
        case GraphMajorMinorThreshold:
            MALLOC(ineligible, bool, nodes, "SelectWithSortedGraph: ineligible");
            for (int i = 0; i < nodes; i++)
                ineligible[i] = (graph->node_list[i] < major_threshold);

        case GraphMinorThreshold:
            if (!ineligible) {
                MALLOC(ineligible, bool, nodes, "SelectWithSortedGraph: ineligible");
                for (int i = 0; i < nodes; i++)
                    ineligible[i] = false;
            }

            if (minor_threshold <= ZERO) {
                selected = 0;
                for (int i = 0; i < nodes; i++)
                    selected += (S[i] = !ineligible[i]);
                SHOWPROGRESS(parm->show_progress > 1, "Selected before/after: %d/%d\n", selected, selected);
            } else {
                bool finished = false;
                // Check is previously assigned for major threshold
                do {
                    selected = 0;
                    ZeroVector(NULL, S, nodes);

                    for (int i = 0; graph->edge_list[i].value >= minor_threshold && i < graph->edges; i++)
                        if (graph->edge_list[i].origin != graph->edge_list[i].destination && !ineligible[graph->edge_list[i].origin] && !ineligible[graph->edge_list[i].destination]) {
                            S[graph->edge_list[i].origin]++;
                            S[graph->edge_list[i].destination]++;
                        }

                    // Count selected and get selected with minimum frequency
                    int freq_ind = -1, freq_min = nodes;
                    for (int i = 0; i < nodes; i++)
                        if (S[i]) {
                            selected++;
                            if (S[i] < freq_min) {
                                freq_min = S[i];
                                freq_ind = i;
                            }
                        }

                    // Remove least frequently selected if not all the time-1
                    if (freq_min < selected-1)
                        ineligible[freq_ind] = true; // Don't choose again
                    else
                        finished = true;
                } while (!finished);

                selected = 0;
                for (int i = 0; i < nodes; i++)
                    selected += (S[i] = (S[i] > 0));
            }

            FREE(ineligible);
            break;

        case GraphMinorConnectedSubgraph:
        case GraphMajorMinorConnectedSubgraph:
        case GraphMinorThresholdChain:
        case GraphMajorMinorThresholdChain:
        case GraphMinorThresholdIFEdges:
        case GraphMajorMinorThresholdIFEdges:
            OnErrorExit(ERR_PARAM, "Sorted list is not suited for connected subgraphs, chains or only IF edges");
            break;

        default:
            OnErrorExit(ERR_UNSPECIFIED, "Bad selection method");
            break;
    }

    // If none, select highest node
    if (!selected) {
        largest = 0;
        for (int i = 1; i < nodes; i++)
            if (GetValueFromNode(graph, i) > GetValueFromNode(graph, largest))
                largest = i;

        S[largest] = 1;
        selected = 1;
    }

    return selected;
}

void SelectWithThreshold(int instances, int features, int *isel, int *fsel, int *selected_i, int *selected_f, scalar fthreshold, scalar ithreshold, scalar *votes_f, scalar *votes_i)
{
    *selected_f = *selected_i = 0;

    for (int i = 0; i < features; i++)
        if (votes_f[i] >= fthreshold) {
            fsel[i] = 1;
            (*selected_f)++;
        } else
            fsel[i] = 0;

    for (int i = 0; i < instances; i++)
        if (votes_i[i] >= ithreshold) {
            isel[i] = 1;
            (*selected_i)++;
        } else
            isel[i] = 0;
}
