/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include <omp.h>
#include <unistd.h>
#include "macros.h"
#include "common.h"
#include "wrapper.h"
#include "filter.h"
#include "ranking.h"
#include "fs.h"
#include "is.h"
#include "classifiers.h"
#include "data.h"
#include "isfs.h"
#include "maths.h"

// Simultaneous instance and feature selection using floating search: SIFS²

// Instance + Feature selection using floating search

// IS must be kept to next step

// Hardcoded maximum size of cache depending on the machine to avoid shutdown (in GB)
#define CAIRO_MC    32.0
#define POTEMKIN_MC 192.0

#define FR_INSTANCES_THRESHOLD  100000

int ISFSWithFloatingSearch(const t_Parameters *parm, int *subset, int *subspace, scalar **data, int instances, int *labels, int features, int classes, int minority, int **index)
{
    extern char *isalgorithm[];

    scalar best_performance, averaged_performance = 0.0, **sq_dists = NULL, reduction, **sq_copy, red, accuracy, evaluations = 0.0;
    int selected_f = -1, *S, *best_S, *F, attempts = 0, losing_streak = 0, *cur_subset, generations;
    cache_t **neighbors_cache = NULL;
    time_t beggining = 0;
    bool use_cache = true;

    // For random selection of instance selection algorithm
    t_ISAlgorithm isalgs[] = {DROP3, ICF, CHC, MSS, GCNN, BoostBased, RMHC, HMNEI /* PSA, CNN, RNN, DROP2, RNGE, PSRCG, IB3, */}, isalg = DROP3;
    int n_isalgs = 8;

    time_t init_time = time(NULL);

    MALLOC(S, int, instances+features, "ISFSWithFloatingSearch: S");
    MALLOC(best_S, int, instances, "ISFSWithFloatingSearch: S");
    F = S+instances;
    MALLOC(cur_subset, int, instances, "ISFSWithFloatingSearch: cur_subset");

    isalg = parm->isalgorithm;
    SHOWPROGRESS(parm->show_progress > 0, "Instance selection algorithm: %s\n", isalgorithm[isalg]);

    // Check IS algorithm
    switch (isalg) {
        case NoIS:
        case RANDOMSELECTION:
        case SMOTE:
        case Undersampling:
            OnErrorExit(ERR_PARAM, "Instance selection algorithm not appropriate for SIFS²");
            break;

        case BoostBased:
        case CHC:
        case DROP2:
        case DROP3:
        case ICF:
        case MSS:
        case RNN:
        case IB3:
        case CNN:
        case GCNN:
        case RMHC:
        case HMNEI:
        case CFilter:
        case PSRCG:
        case RC45:
        case PSA:
        case RNGE:
        case MCS:
            // Ok, do nothing
            break;

        case CNNTL:
            if (classes != 2)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm only valid for 2-class problems");
        case ENTROPY:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Instance selection algorithm with subspace");
            break;

        case OSS:
            if (classes != 2)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm only valid for 2-class problems");
            break;

        case SimulatedAnnealing:
        case kNNCHC:
        case LVQPRU:
            OnErrorExit(ERR_PARAM, "Unsuitable instance selection method");
            break;
    }

    // Iterate over all features or subset: Use all possible features of only a subset?
    int iteration_lenght = parm->mabuse_feat_per_iteration ? MIN(parm->mabuse_feat_per_iteration, features) : features, *list;
    MALLOC(list, int, features, "ISFSWithFloatingSearch: list");
    for (int i = 0; i < features; i++)
        list[i] = i;

    // Roulette selection of features
    // Features are selected using MI score as probability
    bool roulette_selection = parm->mabuse_sort_by_weighted_mi_limit > features && iteration_lenght < features;
    scalar *cum_probabilities = NULL;

    if (roulette_selection) {
        t_RankInfo *ranking;
        MALLOC(ranking, t_RankInfo, features, "ISFSWithFloatingSearch: ranking");
        time_t i_time = time(NULL);

        t_FRAlgorithm fralgorithm = parm->fralgorithm;

        // If too many instances select and scalable FR algorithm
        if (instances > FR_INSTANCES_THRESHOLD) {
            fralgorithm = (classes == 2) ? frFisher : frtTest;

            SHOWPROGRESS(parm->show_progress > 0, "Ranking algorithm set to %s due to large dataset\n", classes == 2 ? "Fisher" : "t-Test");
        }

        // Check ranking algorithm
        if (fralgorithm != frMI && fralgorithm != frFisher && fralgorithm != frCHISQ && fralgorithm != frtTest && fralgorithm != frEDSE)
            OnErrorExit(ERR_PARAM, "Feature ranking method not appropriate for SIFS2");

        FeatureRankingAlgorithm(parm, fralgorithm, features, data, labels, classes, instances, minority, ranking, NULL, NULL, index);
        fprintf(parm->out_file, "@Feature ranking time: %ld\n", time(NULL) - i_time);

        // The minimum non zero value is added to all features to avoid removing any
        scalar min_non_zero = 1.0E10;
        for (int i = 0; i < features; i++)
            if (ranking[i].score > ZERO && ranking[i].score < min_non_zero)
                min_non_zero = ranking[i].score;

        MALLOC(cum_probabilities, scalar, features, "ISFSWithFloatingSearch: cum_probabilities");

        cum_probabilities[0] = ranking[0].score+min_non_zero;
        for (int i = 1; i < features; i++)
            cum_probabilities[i] = cum_probabilities[i-1]+ranking[i].score+min_non_zero;

        FREE(ranking);
    }

    // Some machines have a hardcoded maximum size
    scalar max_cache_size = parm->isfs_cache_size;
    char machine[1000];
    gethostname(machine, 1000);
    if (!strcmp(machine, "cairo"))
        max_cache_size = MIN(max_cache_size, CAIRO_MC);
    else if (!strcmp(machine, "potemkin"))
        max_cache_size = MIN(max_cache_size, POTEMKIN_MC);

    // Use cache?
    scalar size = instances;
    size = (size*sizeof(scalar *) + size*size*sizeof(scalar))/pow(2.0, 30.0) + // Distances cache
           (size*sizeof(cache_t *) + size*size*sizeof(cache_t))/pow(2.0, 30.0);  // Neighbors cache

#ifdef _OPENMP
    // Set maximum number of threads avoiding exceeding maximum cache size if cache is used
    if (parm->isfs_use_cache && parm->isfs_cache_size) {
        int max_threads = omp_get_max_threads();

        if (size*max_threads > max_cache_size) {
            max_threads = max_cache_size/size;
            if (!max_threads) {
                SHOWPROGRESS(true, "Cache size of %.1lfGB exceeds limits\n", size);
                use_cache = false;
            } else {
                omp_set_num_threads(max_threads);
                SHOWPROGRESS(parm->show_progress > 0, "Set maximum number of threads to: %d\n", max_threads);
            }
        }
    }

    // Allow nested loops
    if (parm->omp_allow_nested_loops)
        omp_set_nested(1);
#endif

    if (size > max_cache_size || !use_cache || (parm->sifs2_stratum_size && parm->sifs2_stratum_size < instances)) {    // With stratification cache is useless
        SHOWPROGRESS(true, "Cache size of %.1lfGB exceeds limits\n", size);
        use_cache = false;
        sq_dists = sq_copy = NULL;
        neighbors_cache = NULL;
    } else {
        ReserveAndInitializeCache(&sq_dists, instances);
        ReserveAndInitializeCache(&sq_copy, instances);

        MALLOC(neighbors_cache, cache_t *, instances, "ISFSWithFloatingSearch: neighbors_cache");
        for (int i = 0; i < instances; i++)
            MALLOC(neighbors_cache[i], cache_t, instances, "ISFSWithFloatingSearch: neighbors_cache[i]");

        // If starting with all features selected the cache must be initialized to actual distances
        if (parm->search_heuristic == LRS && parm->plus_l < parm->minus_r)
            InitializeCacheToActualSqDistance(sq_dists, features, data, instances, index);
    }

    // Set l and r appropriately
    int plus_l = -1, minus_r = -1;
    switch (parm->search_heuristic) {
        case SFS:
            plus_l = 1;
            minus_r = 0;
            break;

        case SBS:
            OnErrorExit(ERR_PARAM, "SFBS not suitable for SIFS²");
            plus_l = 0;
            minus_r = 1;
            break;

        case LRS:
            plus_l = parm->plus_l;
            minus_r = parm->minus_r;
            break;

        case SFBS:
            OnErrorExit(ERR_PARAM, "SFBS not suitable for SIFS²");
        case SFFS:
            plus_l = minus_r = 1;
            break;
    }

    best_performance = -1.0;
    int iterations = 0;
    scalar previous_performance = -1.0, performance = -1.0;

    // Initialization depends on type of heuristic search
    if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
        // Start with no features
        selected_f = 0;
        reduction = 1.0;
        for (int i = 0; i < instances; i++)
            S[i] = 1;
        for (int i = 0; i < features; i++)
            F[i] = 0;

        SHOWPROGRESS(parm->show_progress > 0, "Begin with NO features%c", '\n');
    } else {
        // Start with all features
        selected_f = features;
        reduction = 0.0;
        for (int i = 0; i < instances; i++)
            S[i] = 1;
        for (int i = 0; i < features; i++)
            F[i] = 1;

        SHOWPROGRESS(parm->show_progress > 0, "Begin with ALL features%c", '\n');
    }

    bool sifs2_when_stratified_evaluate_with_training_set = parm->sifs2_when_stratified_evaluate_with_training_set;
    int *group;

    MALLOC(group, int, parm->sifs2_add_group_size, "ISFSWithFloatingSearch: group");

#ifdef _OPENMP
    // Reserve private copies
    scalar ***pc_sq_copy = NULL;
    cache_t ***pc_neighbors_cache = NULL;
    int **pc_F, **pc_S, **pc_t_S, **pc_group;

    int max_threads = omp_get_max_threads();

    if (use_cache) {
        MALLOC(pc_sq_copy, scalar **, max_threads, "ISFSWithFloatingSearch: pc_sq_copy");
        MALLOC(pc_neighbors_cache, cache_t **, max_threads, "ISFSWithFloatingSearch: neighbors_cache");
    }
    MALLOC(pc_F, int *, max_threads, "ISFSWithFloatingSearch: pc_F");
    MALLOC(pc_S, int *, max_threads, "ISFSWithFloatingSearch: pc_S");
    MALLOC(pc_t_S, int *, max_threads, "ISFSWithFloatingSearch: pc_t_S");
    MALLOC(pc_group, int *, max_threads, "ISFSWithFloatingSearch: pc_group");

    for (int i = 0; i < max_threads; i++) {
        if (use_cache) {
            ReserveAndInitializeCache(&pc_sq_copy[i], instances);
            MALLOC(pc_neighbors_cache[i], cache_t *, instances, "ISFSWithFloatingSearch: pc_neighbors_cache[i]");
            for (int j = 0; j < instances; j++)
                MALLOC(pc_neighbors_cache[i][j], cache_t, instances, "ISFSWithFloatingSearch: pc_neighbors_cache[i][j]");
        }
        MALLOC(pc_F[i], int, features, "ISFSWithFloatingSearch: pc_F[i]");
        MALLOC(pc_S[i], int, instances, "ISFSWithFloatingSearch: pc_S[i]");
        MALLOC(pc_t_S[i], int, instances, "ISFSWithFloatingSearch: pc_t_S[i]");
        MALLOC(pc_group[i], int, parm->sifs2_add_group_size, "ISFSWithFloatingSearch: pc_group[i]");
    }
#endif

    SHOWPROGRESS(parm->show_progress > 1, "Starting SIFS² algorithm iteration%c", '\n');

    // Stratified sampling?
    // Only consider and stratified sample
    int *t_labels = NULL, *t_list = NULL, **t_index = NULL, *t_S = NULL, t_instances = 0, stratum_size = 0;
    scalar **t_data = NULL;
    bool random_subsample = false;
    if (parm->sifs2_stratum_size && parm->sifs2_stratum_size < instances) {
        stratum_size = parm->sifs2_stratum_size;
        t_S = S;
        t_instances = instances;
    } else if (parm->sifs2_sampling_size && parm->sifs2_sampling_size < instances) {
        t_instances = parm->sifs2_sampling_size;
        random_subsample = true;
        MALLOC(t_S, int, t_instances, "ISFSWithFloatingSearch: t_S");
    } else {
        t_data = data;
        t_index = index;
        t_instances = instances;
        t_labels = labels;
        t_S = S;

        MALLOC(t_list, int, instances, "ISFSWithFloatingSearch: t_list");
        for (int i = 0; i < instances; i++)
            t_list[i] = i;

        // If not stratified True has no meaning
        sifs2_when_stratified_evaluate_with_training_set = false;
    }

    // Stratified partition for evaluating the performance
    int sets = 0, *set_instances = NULL, **set_labels = NULL, **set_list = NULL, * **set_index = NULL, *set_subset = NULL;
    scalar ***set_data = NULL;

    if (parm->sifs2_sampling_size_for_performance && parm->sifs2_sampling_size_for_performance < instances) {
        sets = (int) ceil(((scalar) instances)/parm->sifs2_sampling_size_for_performance);
        MALLOC(set_subset, int, parm->sifs2_sampling_size_for_performance, "ISFSWithFloatingSearch: set_subset");
        MALLOC(set_data, scalar **, sets, "ISFSWithFloatingSearch: set_data");
        MALLOC(set_index, int **, sets, "ISFSWithFloatingSearch: set_data");
        MALLOC(set_labels, int *, sets, "ISFSWithFloatingSearch: set_labels");
        MALLOC(set_list, int *, sets, "ISFSWithFloatingSearch: set_list");
        MALLOC(set_instances, int, sets, "ISFSWithFloatingSearch: set_instances");
        for (int i = 0; i < sets; i++) {
            MALLOC(set_data[i], scalar *, parm->sifs2_sampling_size_for_performance, "ISFSWithFloatingSearch: set_data[i]");
            if (index) {
                MALLOC(set_index[i], int *, parm->sifs2_sampling_size_for_performance, "ISFSWithFloatingSearch: set_index[i]");
            } else
                set_index[i] = NULL;
            MALLOC(set_labels[i], int, parm->sifs2_sampling_size_for_performance, "ISFSWithFloatingSearch: set_labels[i]");
            MALLOC(set_list[i], int, parm->sifs2_sampling_size_for_performance, "ISFSWithFloatingSearch: set_labels[i]");
        }
        DataPartition((const t_Parameters *) &parm, data, instances, labels, NULL, &sets, &set_data, &set_instances, &set_labels, classes, &set_list, index, &set_index, features,
                      parm->sifs2_sampling_size_for_performance, pRANDOM, true, NULL);
    }

    if (parm->sifs2_stratified_incremental_selection)
        ZeroVector(NULL, cur_subset, instances);

    // Group of features added every step
    do {
        scalar previous = 0.0;

        if (random_subsample)
            StratifiedRandomSample(data, instances, labels, &t_data, &t_labels, t_instances, classes, &t_list, index, &t_index);

        previous_performance = parm->use_averaged_accuracy_for_stop ? averaged_performance : performance;

        for (int j = 0; j < plus_l && selected_f < features; j++) {
            int best_group[parm->sifs2_add_group_size], best_count = -1;
            scalar best_fitness = 0.0;

            RandomizeVector(list, features);

            beggining = time(NULL);

#ifdef _OPENMP
            // Copy F to all private copies
            for (int i = 0; i < max_threads; i++)
                memcpy(pc_F[i], F, sizeof(int)*features);

            #pragma omp parallel for schedule(dynamic) private(accuracy, red, sq_copy, neighbors_cache, S, F, t_S, group)
#endif
            for (int i = 0; i < iteration_lenght; i += parm->sifs2_add_group_size) {
                scalar fitness;

#ifdef _OPENMP
                // Use reserved private copies
                int thread_id = omp_get_thread_num();

                if (use_cache) {
                    sq_copy = pc_sq_copy[thread_id];
                    neighbors_cache = pc_neighbors_cache[thread_id];
                } else {
                    sq_copy = NULL;
                    neighbors_cache = NULL;
                }
                S = pc_S[thread_id];
                F = pc_F[thread_id];
                t_S = pc_t_S[thread_id];
                group = pc_group[thread_id];
#endif

                // Select group
                for (int j = 0; j < parm->sifs2_add_group_size && i+j < features; j++)
                    group[j] = roulette_selection ? RouletteSelection(cum_probabilities, features) : list[i+j];

                // Check if all already selected
                int count = parm->sifs2_add_group_size;
                for (int j = 0; j < count; j++)
                    if (F[group[j]]) {
                        count--;
                        group[j] = group[count];
                        j--;
                    }

                if (!count)
                    continue;

                // Stratified approach
                if (stratum_size) {
                    fitness = StratifiedEvaluationOfNextFeature(parm, data, instances, features, labels, index, F, t_S, classes, stratum_size, minority, &accuracy, &red, group, count, -1,
                              isalg);
                    SHOWPROGRESS(parm->show_progress > 1, "Evaluation of features%c", ' ');
                    for (int j = 0; j < count; j++)
                        SHOWPROGRESS(parm->show_progress > 1, "%d ", group[j]);
                    SHOWPROGRESS(parm->show_progress > 1, "for addition (P, A, R): (%6.4lf, %6.4lf, %6.4lf)\n", fitness, accuracy, red);
                } else {
                    // Add feature and check
                    SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "Evaluating features for addition:%c", ' ');
                    for (int j = 0; j < count; j++) {
                        F[group[j]] = 1;

                        SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "%d ", group[j]+1);
                    }
                    SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "%c", '\n');

                    // Stratified random sub-sampling?
                    if (random_subsample) {
                        if (use_cache) {
                            // Copy cache of sampled instances
                            CopyDistanceCacheWithList(sq_dists, sq_copy, t_list, t_instances);

                            // Update NN cache with next feature
                            for (int j = 0; j < count; j++)
                                UpdateCache(sq_copy, group[j], -1, t_data, t_instances, t_index);

                            // Use distance cache to construct neighbors cache
                            ConstructNeighborsCacheFromDistances(t_instances, neighbors_cache, sq_copy);
                        }
                    } else {
                        if (use_cache) {
                            // Update NN cache with next feature
                            CopyAndUpdateCache(sq_dists, sq_copy, group[0], -1, data, instances, index);
                            for (int j = 1; j < count; j++)
                                UpdateCache(sq_copy, group[j], -1, data, instances, index);

                            // Use distance cache to construct neighbors cache
                            ConstructNeighborsCacheFromDistances(instances, neighbors_cache, sq_copy);
                        }
                    }

                    // Apply IS algorithm
                    if (parm->sifs2_random_is_alg)
                        isalg = isalgs[(int) RANDOM(0, n_isalgs)];
                    InstanceSelectionAlgorithm(parm, isalg, &t_S, &t_data, &t_instances, features, &t_labels, classes, t_index, neighbors_cache, sq_copy, minority, &generations, 0.0, NULL,
                                               NULL, &evaluations, NULL, false, NULL, false, F);

                    // Evaluate fitness
                    if (random_subsample && sifs2_when_stratified_evaluate_with_training_set) {
                        ZeroVector(NULL, S, instances);
                        for (int ii = 0; ii < t_instances; ii++)
                            S[t_list[ii]] = t_S[ii];

                        // Update cache for evaluation
                        if (use_cache) {
                            CopyAndUpdateCache(sq_dists, sq_copy, group[0], -1, data, instances, index);
                            for (int j = 1; j < count; j++)
                                UpdateCache(sq_copy, group[j], -1, data, instances, index);
                            ConstructNeighborsCacheFromDistancesWithSubset(instances, neighbors_cache, sq_copy, S);
                        }

                        // Evaluate fitness
                        fitness = SIFS2Performance(parm, F, S, labels, instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, data, index);
                    } else
                        fitness = SIFS2Performance(parm, F, t_S, t_labels, t_instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, t_data, t_index);

                    for (int j = 0; j < count; j++)
                        F[group[j]] = 0;

                    SHOWPROGRESS(parm->show_progress > 1, "Evaluation of features%c", ' ');
                    for (int j = 0; j < count; j++)
                        SHOWPROGRESS(parm->show_progress > 1, "%d ", group[j]);
                    SHOWPROGRESS(parm->show_progress > 1, "for addition (P, A, R): (%6.4lf, %6.4lf, %6.4lf)\n", fitness, accuracy, red);
                }

#ifdef _OPENMP
                #pragma omp critical
#endif
                if (fitness > best_fitness) {
                    best_fitness = fitness;
                    best_count = count;
                    for (int j = 0; j < count; j++)
                        best_group[j] = group[j];

                    memcpy(best_S, t_S, sizeof(int)*t_instances);
                }
            }

            SHOWTIME(parm->time_all, parm->out_file, "TIME -> Selection of next best feature: %ld s\n", time(NULL) - beggining);
            SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "%c", '\n');

            // Add best
            if (best_count != -1) {
                for (int j = 0; j < best_count; j++) {
                    F[best_group[j]] = 1;
                    SHOWPROGRESS(parm->show_progress > 0, "%3d: Added %*d\t", iterations+1, WIDTH(features), best_group[j]+1);
                    SAVEPROGRESS(parm->show_progress > 0, "%3d: Added %*d\t", iterations+1, WIDTH(features), best_group[j]+1);

                    // If roulette selection remove from possible next features
                    if (roulette_selection)
                        cum_probabilities[best_group[j]] = best_group[j] ? cum_probabilities[best_group[j]-1] : -1.0;
                }
                selected_f += best_count;

                if (stratum_size)
                    memcpy(cur_subset, best_S, sizeof(int)*instances);
                else {
                    // Keep best S if needed
                    if (!parm->sifs2_stratified_incremental_selection)
                        ZeroVector(NULL, cur_subset, instances);

                    if (parm->sifs2_is_additive)
                        // Add newly selected, but do not remove previous ones
                        for (int j = 0; j < t_instances; j++)
                            cur_subset[t_list[j]] |= best_S[j];
                    else
                        for (int j = 0; j < t_instances; j++)
                            cur_subset[t_list[j]] = best_S[j];
                }

                if (parm->show_progress > 1) {
                    int count = 0;
                    for (int i = 0; i < instances; i++)
                        if (cur_subset[i])
                            count++;

                    SHOWPROGRESS(true, "\nCurrent selection stratified voting count after %d attempts: %d\n", attempts+1, count);
                }

                // Recalculate cache
                time_t cache_time = time(NULL);
                if (use_cache)
                    for (int j = 0; j < best_count; j++)
                        UpdateCache(sq_dists, best_group[j], -1, data, instances, index);
                SHOWTIME(parm->time_all, parm->out_file, "TIME -> Cache updating: %ld s\n", time(NULL) - cache_time);
            }
        }

        if (parm->search_heuristic == SFFS) {
            beggining = time(NULL);
            if (random_subsample) {
                for (int i = 0; i < t_instances; i++)
                    t_S[i] = cur_subset[t_list[i]];
                if (use_cache) {
                    CopyDistanceCacheWithList(sq_dists, sq_copy, t_list, t_instances);
                    ConstructNeighborsCacheFromDistancesWithSubset(t_instances, neighbors_cache, sq_copy, t_S);
                }

                previous = SIFS2Performance(parm, F, t_S, t_labels, t_instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, t_data, t_index);
            } else {
                if (use_cache)
                    ConstructNeighborsCacheFromDistancesWithSubset(instances, neighbors_cache, sq_dists, cur_subset);
                previous = SIFS2Performance(parm, F, cur_subset, labels, instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, data, index);
            }

            SHOWTIME(parm->time_all, NULL, "TIME -> Performance calculation for SFFS backwards step: %ld s\n", time(NULL) - beggining);
        }

        for (int j = 0; j < minus_r && selected_f > 1; j++) {
            int worst_feature = -1;
            scalar worst_fitness = 1.1, fitness;

            RandomizeVector(list, features);

            beggining = time(NULL);

#ifdef _OPENMP
            // Copy F to all private copies
            for (int i = 0; i < max_threads; i++)
                memcpy(pc_F[i], F, sizeof(int)*features);

            #pragma omp parallel for schedule(dynamic) private(accuracy, red, sq_copy, neighbors_cache, S, F, t_S, isalg)
#endif
            for (int i = 0; i < iteration_lenght; i++) {
                int next = list[i];
                isalg = parm->isalgorithm;

#ifdef _OPENMP
                // Use reserved private copies
                int thread_id = omp_get_thread_num();

                sq_copy = use_cache ? pc_sq_copy[thread_id] : NULL;
                neighbors_cache = use_cache ? pc_neighbors_cache[thread_id] : NULL;
                S = pc_S[thread_id];
                F = pc_F[thread_id];
                t_S = pc_t_S[thread_id];
#endif

                if (!F[next])
                    continue;
                SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "Evaluating for deletion: %*d\r", WIDTH(features), next+1);

                // Stratified approach
                if (stratum_size)
                    fitness = StratifiedEvaluationOfNextFeature(parm, data, instances, features, labels, index, F, t_S, classes, stratum_size, minority, &accuracy, &red, NULL, 0, next, isalg);
                else {
                    // Remove instance and check
                    F[next] = 0;

                    // Stratified sampling?
                    if (random_subsample) {
                        if (use_cache) {
                            // Copy cache of sampled instances
                            CopyDistanceCacheWithList(sq_dists, sq_copy, t_list, t_instances);

                            // Update NN cache with next feature
                            UpdateCache(sq_copy, -1, next, t_data, t_instances, t_index);

                            // Use distance cache to construct neighbors cache
                            ConstructNeighborsCacheFromDistances(t_instances, neighbors_cache, sq_copy);
                        }
                    } else {
                        if (use_cache) {
                            // Update NN cache with next feature
                            CopyAndUpdateCache(sq_dists, sq_copy, -1, next, data, instances, index);

                            // Use distance cache to construct neighbors cache
                            ConstructNeighborsCacheFromDistances(instances, neighbors_cache, sq_copy);
                        }
                    }

                    // Apply IS algorithm
                    if (parm->sifs2_random_is_alg)
                        isalg = isalgs[(int) RANDOM(0, n_isalgs)];
                    InstanceSelectionAlgorithm(parm, isalg, &t_S, &t_data, &t_instances, features, &t_labels, classes, t_index, neighbors_cache, sq_copy, minority, &generations, 0.0, NULL,
                                               NULL, &evaluations, NULL, false, NULL, false, F);

                    // Evaluate fitness
                    if (random_subsample && sifs2_when_stratified_evaluate_with_training_set) {
                        ZeroVector(NULL, S, instances);
                        for (int ii = 0; ii < t_instances; ii++)
                            S[t_list[ii]] = t_S[ii];

                        // Update cache for evaluation
                        if (use_cache) {
                            GetSqCacheWithSubsetSubspace(sq_copy, features, data, instances, index, S, F);
                            ConstructNeighborsCacheFromDistancesWithSubset(instances, neighbors_cache, sq_copy, S);
                        }

                        // Evaluate fitness
                        fitness = SIFS2Performance(parm, F, S, labels, instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, data, index);
                    } else
                        fitness = SIFS2Performance(parm, F, t_S, t_labels, t_instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, t_data, t_index);

                    F[next] = 1;
                }
#ifdef _OPENMP
                #pragma omp critical
#endif
                if (fitness < worst_fitness) {
                    worst_fitness = fitness;
                    worst_feature = next;

                    memcpy(best_S, t_S, sizeof(int)*t_instances);
                }
            }

            SHOWTIME(parm->time_all, parm->out_file, "TIME -> Selection of next worst feature: %ld s\n", time(NULL) - beggining);
            SHOWPROGRESS(parm->show_progress > 1 && parm->verbose >= 0, "%c", '\n');

            // Remove worst
            if (worst_feature != -1) {
                F[worst_feature] = 0;
                // Forbid for the feature
                selected_f--;

                SHOWPROGRESS(parm->show_progress > 0, "%3d: Deleted %*d\n", iterations+1, WIDTH(features), worst_feature+1);
                SAVEPROGRESS(parm->show_progress > 0, "%3d: Deleted %*d\t", iterations+1, WIDTH(features), worst_feature+1);

                if (stratum_size)
                    memcpy(cur_subset, best_S, sizeof(int)*instances);
                else {
                    // Keep best S if needed
                    if (!parm->sifs2_stratified_incremental_selection)
                        ZeroVector(NULL, cur_subset, instances);

                    if (parm->sifs2_is_additive)
                        // Add newly selected, but do not remove previous ones
                        for (int j = 0; j < t_instances; j++)
                            cur_subset[t_list[j]] |= best_S[j];
                    else
                        for (int j = 0; j < t_instances; j++)
                            cur_subset[t_list[j]] = best_S[j];
                }

                if (parm->show_progress > 1) {
                    int count = 0;
                    for (int i = 0; i < instances; i++)
                        if (cur_subset[i])
                            count++;

                    SHOWPROGRESS(true, "\nCurrent selection stratified voting count after %d attempts: %d\n", attempts+1, count);
                }

                // Recalculate cache
                time_t cache_time = time(NULL);
                if (use_cache)
                    UpdateCache(sq_dists, -1, worst_feature, data, instances, index);
                SHOWTIME(parm->time_all, parm->out_file, "TIME -> Cache updating: %ld s\n", time(NULL) - cache_time);

                // For SFFS keep deletion if the performance does not worsen, otherwise undo
                if (parm->search_heuristic == SFFS) {
                    scalar fitness;
                    time_t cache_time = time(NULL);
                    if (random_subsample) {
                        for (int i = 0; i < t_instances; i++)
                            t_S[i] = cur_subset[t_list[i]];
                        if (use_cache) {
                            CopyDistanceCacheWithList(sq_dists, sq_copy, t_list, t_instances);
                            ConstructNeighborsCacheFromDistancesWithSubset(t_instances, neighbors_cache, sq_copy, t_S);
                        }

                        fitness = SIFS2Performance(parm, F, t_S, t_labels, t_instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, t_data, t_index);
                    } else {
                        if (use_cache)
                            ConstructNeighborsCacheFromDistancesWithSubset(instances, neighbors_cache, sq_dists, cur_subset);
                        fitness = SIFS2Performance(parm, F, cur_subset, labels, instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &red, data, index);
                    }
                    SHOWTIME(parm->time_all, parm->out_file, "TIME -> SFFS backward step: %ld s\n", time(NULL) - cache_time);

                    if (previous <= fitness) {
                        j--;
                        previous = fitness;

                        SHOWPROGRESS(parm->show_progress > 1, "Do not undo removed instance: %d\n", worst_feature+1);
                    } else {
                        F[worst_feature] = 1;
                        selected_f++;
                        SHOWPROGRESS(parm->show_progress > 0, "Undo removed instance: %d\n", worst_feature+1);

                        // Recalculate cache
                        cache_time = time(NULL);
                        if (use_cache)
                            UpdateCache(sq_dists, worst_feature, -1, data, instances, index);
                        SHOWTIME(parm->time_all, parm->out_file, "TIME -> Cache updating for SFFS backward step: %ld s\n", time(NULL) - cache_time);

                        // Apply IS algorithm again with removed instance
                        cache_time = time(NULL);
                        if (!random_subsample) {
                            if (use_cache)
                                ConstructNeighborsCacheFromDistances(instances, neighbors_cache, sq_dists);
                            if (parm->sifs2_random_is_alg)
                                isalg = isalgs[(int) RANDOM(0, n_isalgs)];
                            InstanceSelectionAlgorithm(parm, isalg, &cur_subset, &data, &instances, features, &labels, classes, index, neighbors_cache, sq_copy, minority, &generations, 0.0,
                                                       NULL, NULL, &evaluations, NULL, false, NULL, false, F);
                        } else {
                            if (use_cache) {
                                // Copy cache of sampled instances
                                CopyDistanceCacheWithList(sq_dists, sq_copy, t_list, t_instances);
                                // Use distance cache to construct neighbors cache
                                ConstructNeighborsCacheFromDistances(t_instances, neighbors_cache, sq_copy);
                            }

                            if (parm->sifs2_random_is_alg)
                                isalg = isalgs[(int) RANDOM(0, n_isalgs)];
                            InstanceSelectionAlgorithm(parm, isalg, &t_S, &t_data, &t_instances, features, &t_labels, classes, t_index, neighbors_cache, sq_copy, minority, &generations, 0.0,
                                                       NULL, NULL, &evaluations, NULL, false, NULL, false, F);

                            if (parm->sifs2_is_additive)
                                // Add newly selected, but do not remove previous ones
                                for (int j = 0; j < t_instances; j++)
                                    cur_subset[t_list[j]] |= t_S[j];
                            else
                                for (int j = 0; j < t_instances; j++)
                                    cur_subset[t_list[j]] = t_S[j];
                        }
                        SHOWTIME(parm->time_all, parm->out_file, "TIME -> Instance selection for SFFS backward step: %ld s\n", time(NULL) - cache_time);
                    }
                }
            }
        }

        // Obtain performance as a combination of reduction and fitness
        beggining = time(NULL);
        iterations++;
        if (parm->sifs2_sampling_size_for_performance && parm->sifs2_sampling_size_for_performance < instances) {
            scalar a, r;

            performance = 0.0;
            reduction = 0.0;
            accuracy = 0.0;

            // Evaluate all strata
            for (int i = 0; i < sets; i++) {
                for (int j = 0; j < set_instances[i]; j++)
                    set_subset[j] = cur_subset[set_list[i][j]];

                performance += SIFS2Performance(parm, F, set_subset, set_labels[i], set_instances[i], classes, features, parm->performance_measure, minority, NULL, &a, &r, set_data[i], set_index[i]);

                accuracy += a;
                reduction += r;
            }

            performance /= sets;
            reduction /= sets;
            accuracy /= sets;
        } else {
            if (use_cache)
                ConstructNeighborsCacheFromDistancesWithSubset(instances, neighbors_cache, sq_dists, cur_subset);
            performance = SIFS2Performance(parm, F, cur_subset, labels, instances, classes, features, parm->performance_measure, minority, neighbors_cache, &accuracy, &reduction, data, index);
        }

        if (parm->sifs2_convergence_with_only_acc)
            performance = accuracy;
        SHOWTIME(parm->time_all, parm->out_file, "TIME -> Performance calculation: %ld s\n", time(NULL) - beggining);

        averaged_performance = (averaged_performance*(iterations-1)+performance)/iterations;

        bool best_so_far = false;
        if (parm->use_averaged_accuracy_for_stop) {
            if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
                if (previous_performance > averaged_performance || fabs(previous_performance - averaged_performance) <= parm->precision_for_zero)
                    losing_streak++;
                else
                    losing_streak = 0;
            } else {
                if (previous_performance > averaged_performance)
                    losing_streak++;
                else
                    losing_streak = 0;
            }
        } else {
            if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
                if (previous_performance > performance || fabs(previous_performance - performance) <= parm->precision_for_zero)
                    losing_streak++;
                else
                    losing_streak = 0;
            } else {
                if (previous_performance > performance)
                    losing_streak++;
                else
                    losing_streak = 0;
            }
        }

        if (parm->sifs2_best_with_only_acc)
            best_so_far = accuracy >= best_performance;
        else
            best_so_far = performance >= best_performance;

        SHOWPROGRESS(parm->show_progress > 0, "Acc,Red,Perf: [%6.4lf|%6.4lf|%6.4lf] Average: %7.5lf Streak,#Sel: [%*d|%*d]\n", accuracy, reduction, performance, averaged_performance,
                     WIDTH(parm->mabuse_max_losing_streak), losing_streak, WIDTH(features), selected_f);
        SAVEPROGRESS(parm->show_progress > 0, "Acc,Red,Perf: [%6.4lf|%6.4lf|%6.4lf] Average: %7.5lf Streak,#Sel: [%*d|%*d]\n", accuracy, reduction, performance, averaged_performance,
                     WIDTH(parm->mabuse_max_losing_streak), losing_streak, WIDTH(features), selected_f);

        if (random_subsample) {
            FREE(t_labels);
            FREE(t_data);
            FREE(t_list);
            if (index)
                FREE(t_index);
        }

        if (best_so_far) {
            best_performance = parm->sifs2_best_with_only_acc ? accuracy : performance;

            // Best selection so far
            for (int i = 0; i < instances; i++)
                subset[i] = cur_subset[i];

            for (int i = 0; i < features; i++)
                subspace[i] = F[i];

            if (parm->show_progress > 0) {
                int n_f = 0, n_i = 0;

                for (int i = 0; i < instances; i++)
                    if (subset[i])
                        n_i++;

                for (int i = 0; i < features; i++)
                    if (subspace[i])
                        n_f++;

                SHOWPROGRESS(true, "Current best Features: %d Instances: %d\n", n_f, n_i);
                SAVEPROGRESS(true, "Current best Features: %d Instances: %d\n", n_f, n_i);
            }
        }

        attempts++; // Avoid infinite recursion

        // Limited time execution
        if (parm->time_limit && time(NULL)-init_time > parm->time_limit) {
            fprintf(parm->out_file, "Time limit %ld exceeded, exiting after %ld second(s)\n", parm->time_limit, time(NULL)-init_time);
            break;
        } else
            SHOWPROGRESS(parm->show_progress > 0, "Spent time so far: %ld/%ld (s)\n", time(NULL) - init_time, parm->time_limit);

    } while (!MarginBasedFeatureSelectionStop(parm, selected_f, -1, losing_streak, features, attempts));

    if (roulette_selection)
        FREE(cum_probabilities);

    if (parm->sifs2_sampling_size_for_performance && parm->sifs2_sampling_size_for_performance < instances) {
        FREE(set_subset);
        FREE(set_instances);
        for (int i = 0; i < sets; i++) {
            FREE(set_data[i]);
            if (index)
                FREE(set_index[i]);
            FREE(set_labels[i]);
            FREE(set_list[i]);
        }
        FREE(set_data);
        FREE(set_index);
        FREE(set_labels);
        FREE(set_list);
    }

    // Free memory
    if (!random_subsample && !stratum_size)
        FREE(t_list);
    FREE(cur_subset);
    FREE(list);
    FREE(S);
    FREE(group);

    if (use_cache) {
        FreeScalarMatrix(&sq_dists, instances);
        FreeScalarMatrix(&sq_copy, instances);

        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }

#ifdef _OPENMP
    for (int i = 0; i < max_threads; i++) {
        if (use_cache) {
            for (int j = 0; j < instances; j++) {
                FREE(pc_neighbors_cache[i][j]);
                FREE(pc_sq_copy[i][j])
            }
            FREE(pc_neighbors_cache[i]);
            FREE(pc_sq_copy[i])
        }
        FREE(pc_F[i]);
        FREE(pc_S[i]);
        FREE(pc_t_S[i]);
        FREE(pc_group[i]);
    }
    if (use_cache) {
        FREE(pc_sq_copy);
        FREE(pc_neighbors_cache);
    }
    FREE(pc_F);
    FREE(pc_group);
    FREE(pc_S);
    FREE(pc_t_S);
#endif

    return selected_f;
}

// Create k neighbors cache
void ConstructNeighborsCacheFromDistances(int instances, cache_t **neighbors_cache, scalar **sq_distance)
{
    RANK *rank = NULL;

#ifdef _OPENMP
    RANK **cp_rank = NULL;
    int n_threads = omp_get_max_threads();
    MALLOC(cp_rank, RANK *, n_threads, "ConstructNeighborsCacheFromDistances: cp_rank");
    for (int i = 0; i < n_threads; i++)
        MALLOC(cp_rank[i], RANK, instances, "ConstructNeighborsCacheFromDistances: cp_rank[i]");
#else
    MALLOC(rank, RANK, instances, "ConstructNeighborsCacheFromDistances: rank");
#endif

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic) private(rank)
#endif
    for (int i = 0; i < instances; i++) {
#ifdef _OPENMP
        rank = cp_rank[omp_get_thread_num()];
#endif

        for (int j = 0; j < instances; j++) {
            rank[j].fitness = sq_distance[i][j];
            rank[j].being = j;
        }

        // Sort
        qsort(rank, instances, sizeof(RANK), CompareAscending);

        // Store neighbors in ascending order of distance
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int j = 0; j < instances; j++)
            neighbors_cache[i][j] = rank[j].being;
    }

#ifdef _OPENMP
    for (int i = 0; i < n_threads; i++)
        FREE(cp_rank[i]);
    FREE(cp_rank);
#else
    FREE(rank);
#endif
}

#define SIGMOID(red, M)        (tanh(10.*exp(-.025*(M))*(red))/tanh(10.*exp(-.025*(M))))

scalar SIFS2Performance(const t_Parameters *parm, int *subspace, int *S, int *labels, int instances, int classes, int features, t_PerformanceMeasure measure, int minority, cache_t **nns,
                        scalar *performance, scalar *reduction, scalar **data, int **index)
{
    int **M, majority;
    scalar theta_1 = 0.0, theta_2 = 0.0, sp, sn, tp, fp, tn, fn, evaluations = 0.0;

    *performance = 0.0;

    if (parm->sifs2_evaluator != kNN)
        return EvaluateISFSIndividual(parm, S, performance, reduction, data, instances, features, labels, NULL, &evaluations, minority, nns, 0, classes, NULL, index);

    AllocateIntegerMatrix(&M, classes, classes);
    ZeroIntegerMatrix(M, classes, classes);

    // Cache of neighbors?
    if (!nns) {
#ifdef _OPENMP
        #pragma omp parallel for
#endif
        for (int i = 0; i < instances; i++) {
            int nearest[parm->k],
                category = kNNCategory(data[i], data, subspace, instances, parm->k, features, labels, classes, nearest, S, NULL, NULL, i, index, index ? index[i] : NULL);
#ifdef _OPENMP
            #pragma omp critical
#endif
            M[labels[i]][category]++;
        }
    } else {
#ifdef _OPENMP
        #pragma omp parallel for
#endif
        for (int i = 0; i < instances; i++) {
            int j = 0;
            while ((j < instances && !S[nns[i][j]]) || i == j)
                j++;

            if (j < instances)
#ifdef _OPENMP
                #pragma omp critical
#endif
                M[labels[i]][labels[nns[i][j]]]++;
        }
    }

    switch (measure) {
        case pfAccuracy:
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            *performance = theta_1/instances;
            break;

        case pfKappa:
            // Obtain kappa
            // Sum of the diagonal
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            theta_1 *= instances;

            for (int i = 0; i < classes; i++) {
                // Rows/columns total counts
                scalar row_i = 0, column_i = 0;

                for (int j = 0; j < classes; j++) {
                    row_i += M[i][j];
                    column_i += M[j][i];
                }

                theta_2 += row_i * column_i;
            }

            *performance = (theta_1 - theta_2)/(((scalar) instances)*instances-theta_2);
            break;

        case pfGMean:
            if (classes != 2)
                OnErrorExit(ERR_PARAM, "G-mean measure for N-class problem");
            // Obtain G-mean
            majority = (minority+1) % 2;

            sp = (M[majority][majority]+M[majority][minority]) > ZERO ? ((scalar) M[majority][majority])/(M[majority][majority]+M[majority][minority]) : 0.0;
            sn = (M[minority][minority]+M[minority][majority]) > ZERO ? ((scalar) M[minority][minority])/(M[minority][minority]+M[minority][majority]) : 0.0;

            *performance = sqrt(sn*sp);
            break;

        case pfAverageAccuracy:
            *performance = 0.0;
            for (int i = 0; i < classes; i++) {
                tp = M[i][i];

                fp = fn = -M[i][i];
                for (int j = 0; j < classes; j++) {
                    fn += M[i][j];
                    fp += M[j][i];
                }

                tn = instances - tp - fp - fn;
                *performance += (tp+tn)/instances;
            }
            *performance /= classes;
            break;

        case pfRecallMacro:
            // Macro-averaged recall
            *performance = 0.0;
            for (int i = 0; i < classes; i++) {
                tp = M[i][i];
                fn = -M[i][i];
                for (int j = 0; j < classes; j++)
                    fn += M[i][j];

                *performance += (tp+fn > 0) ? tp/(tp+fn) : 0.0;
            }

            *performance /= classes;
            break;

        case pfRecallMacroGMean:
            // Macro-averaged recall
            *performance = 1.0;
            for (int i = 0; i < classes; i++) {
                tp = M[i][i];
                fn = -M[i][i];
                for (int j = 0; j < classes; j++)
                    fn += M[i][j];

                *performance *= (tp+fn > 0) ? tp/(tp+fn) : 1.0;
            }

            *performance = pow(*performance, 1.0/classes);;
            break;

        case pfMargin:
        case pfauPRC:
        case pfauROC:
        case pfCC:
        case pfF1Macro:
        case pfPrecisionMacro:
        case pfF1:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Performance METRIC for SIFS²");
            break;
    }

    FreeIntegerMatrix(&M, classes);

    // Reduction
    scalar s_i = 0.0, s_f = 0.0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            s_i++;
    for (int i = 0; i < features; i++)
        if (subspace[i])
            s_f++;

    // Consider reduction of features using and smoothing function?
    if (parm->exp_redfeature)
        *reduction = ((1.0-s_i/instances)+SIGMOID(s_f/features, features))/2.0;
    else
        *reduction = 1.0 - (s_i/instances)*(s_f/features);

    if (parm->ok_w <= ZERO && parm->reduction_w <= ZERO)
        return sqrt((*performance > ZERO ? *performance : 0.0) * *reduction);
    else
        return *performance*parm->ok_w + *reduction*parm->reduction_w;
}

// Auxiliary function for qsort
int CompareAscending(const void *element_1, const void *element_2)
{
    scalar diff;

    if ((diff = ((RANK *) element_2)->fitness - ((RANK *) element_1)->fitness) > 0)
        return -1;
    else {
        if (diff < 0)
            return +1;
        else
            return 0;
    }
}
#ifdef _OPENMP

void CopyDistanceCacheWithList(scalar **origin, scalar **destination, int *list, int instances)
{
#if __GNUC_PREREQ(4,4)
    #pragma omp parallel for schedule(dynamic) collapse(2)
#else
    // collapse not available for Skynet's gcc version
    #pragma omp parallel for schedule(dynamic)
#endif
    // Copy cache using only sampled instances
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < instances; j++)
            destination[i][j] = origin[list[i]][list[j]];
}

void GetSqCacheWithSubsetSubspace(scalar **dist_cache, int inputs, scalar **data, int instances, int **index, int *selected, int *subspace)
{
    if (index) {
#if __GNUC_PREREQ(4,4)
        #pragma omp parallel for schedule(dynamic) collapse(2)
#else
        // collapse not available for Skynet's gcc version
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < instances; j++)
                if (selected[j])
                    dist_cache[i][j] = SQR(EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index[i], index[j], subspace));
    } else {
#if __GNUC_PREREQ(4,4)
        #pragma omp parallel for schedule(dynamic) collapse(2)
#else
        // collapse not available for Skynet's gcc version
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < instances; j++)
                if (selected[j])
                    dist_cache[i][j] = SQR(EuclideanDistanceSubspace(data[i], data[j], inputs, subspace));
    }
}

#else

void CopyDistanceCacheWithList(scalar **origin, scalar **destination, int *list, int instances)
{
    // Copy cache using only sampled instances
    for (int i = 0; i < instances; i++) {
        destination[i][i] = 0.0;
        for (int j = 0; j < i; j++)
            destination[i][j] = destination[j][i] = origin[list[i]][list[j]];
    }
}

void GetSqCacheWithSubsetSubspace(scalar **dist_cache, int inputs, scalar **data, int instances, int **index, int *selected, int *subspace)
{
    if (index) {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < instances; j++)
                if (selected[j])
                    dist_cache[i][j] = SQR(EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index[i], index[j], subspace));
    } else {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < instances; j++)
                if (selected[j])
                    dist_cache[i][j] = SQR(EuclideanDistanceSubspace(data[i], data[j], inputs, subspace));
    }
}

#endif

// Create k neighbors cache
void ConstructNeighborsCacheFromDistancesWithSubset(int instances, cache_t **neighbors_cache, scalar **sq_distance, int *selected)
{
    RANK *rank = NULL;
    int *list;

    // List of selected instances
    MALLOC(list, int, instances, "ConstructNeighborsCacheFromDistancesWithSubset: list");
    int c = 0;
    for (int j = 0; j < instances; j++)
        if (selected[j]) {
            list[c] = j;
            c++;
        }

    MALLOC(rank, RANK, c, "ConstructNeighborsCacheFromDistancesWithSubset: rank");

    for (int i = 0; i < instances; i++) {
        for (int j = 0; j < c; j++) {
            rank[j].fitness = sq_distance[i][list[j]];
            rank[j].being = list[j];
        }

        // Sort
        qsort(rank, c, sizeof(RANK), CompareAscending);

        // Store neighbors in ascending order of distance
        for (int j = 0; j < c; j++)
            neighbors_cache[i][j] = rank[j].being;
    }

    FREE(rank);
    FREE(list);
}

#ifdef _OPENMP

// Update cache after adding/removing feature
void CopyAndUpdateCache(scalar **origin, scalar **destination, int added, int removed, scalar **data, int instances, int **index)
{
    // Added
    if (added != -1) {
        if (index) {
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    destination[i][j] = (origin[i][j] + SQR(SparseFeature(data[i], index[i], added) - SparseFeature(data[j], index[j], added)));
        } else {
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    destination[i][j] = (origin[i][j] + SQR(data[i][added] - data[j][added]));
        }
    } else if (removed != -1) {
        if (index) {
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    destination[i][j] = (origin[i][j] - SQR(SparseFeature(data[i], index[i], removed) - SparseFeature(data[j], index[j], removed)));
        } else {
#if __GNUC_PREREQ(4,4)
            #pragma omp parallel for schedule(dynamic) collapse(2)
#else
            // collapse not available for Skynet's gcc version
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < instances; j++)
                    destination[i][j] = (origin[i][j] - SQR(data[i][removed] - data[j][removed]));
        }
    } else
        OnErrorExit(ERR_PARAM, "Cannot update cache with no modifications");
}

#else

// Update cache after adding/removing feature
void CopyAndUpdateCache(scalar **origin, scalar **destination, int added, int removed, scalar **data, int instances, int **index)
{
    // Added
    if (added != -1) {
        if (index) {
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    destination[i][j] = destination[j][i] = (origin[i][j] + SQR(SparseFeature(data[i], index[i], added) - SparseFeature(data[j], index[j], added)));
        } else {
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    destination[i][j] = destination[j][i] = (origin[i][j] + SQR(data[i][added] - data[j][added]));
        }
    } else if (removed != -1) {
        if (index) {
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    destination[i][j] = destination[j][i] = (origin[i][j] - SQR(SparseFeature(data[i], index[i], removed) - SparseFeature(data[j], index[j], removed)));
        } else {
            for (int i = 1; i < instances; i++)
                for (int j = 0; j < i; j++)
                    destination[i][j] = destination[j][i] = (origin[i][j] - SQR(data[i][removed] - data[j][removed]));
        }
    } else
        OnErrorExit(ERR_PARAM, "Cannot update cache with no modifications");
}

#endif
