
/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "isfs.h"

// Stratified evaluation of next feature for addition or deletion

scalar StratifiedEvaluationOfNextFeature(const t_Parameters *parm, scalar **data, int instances, int features, int *labels, int **index, int *F, int *S, int classes, int size, int minority,
        scalar *accuracy, scalar *reduction, int *add, int count, int delete, t_ISAlgorithm isalgorithm)
{
    // Stratified partition for evaluating the performance
    int sets = 0, *set_instances, **set_labels, **set_list, * **set_index, *set_subset, *set_S, generations = 0;
    scalar ***set_data, evaluations = 0.0, performance = 0.0;

    sets = (int) ceil(((scalar) instances)/size);
    MALLOC(set_subset, int, size, "StratifiedEvaluationOfNextFeature: set_subset");
    MALLOC(set_data, scalar **, sets, "StratifiedEvaluationOfNextFeature: set_data");
    MALLOC(set_index, int **, sets, "StratifiedEvaluationOfNextFeature: set_data");
    MALLOC(set_labels, int *, sets, "StratifiedEvaluationOfNextFeature: set_labels");
    MALLOC(set_list, int *, sets, "StratifiedEvaluationOfNextFeature: set_list");
    MALLOC(set_instances, int, sets, "StratifiedEvaluationOfNextFeature: set_instances");
    MALLOC(set_S, int, size, "StratifiedEvaluationOfNextFeature: set_S");
    for (int i = 0; i < sets; i++) {
        MALLOC(set_data[i], scalar *, size, "StratifiedEvaluationOfNextFeature: set_data[i]");
        if (index) {
            MALLOC(set_index[i], int *, size, "StratifiedEvaluationOfNextFeature: set_index[i]");
        } else
            set_index[i] = NULL;
        MALLOC(set_labels[i], int, size, "StratifiedEvaluationOfNextFeature: set_labels[i]");
        MALLOC(set_list[i], int, size, "StratifiedEvaluationOfNextFeature: set_labels[i]");
    }

    DataPartition((const t_Parameters *) &parm, data, instances, labels, NULL, &sets, &set_data, &set_instances, &set_labels, classes, &set_list, index, &set_index, features,
                  size, pRANDOM, true, NULL);

    // Add feature
    if (count)
        for (int i = 0; i < count; i++)
            F[add[i]] = 1;
    else
        F[delete] = 0;

    *accuracy = *reduction = 0.0;

    // Apply IS algorithm to each stratum
    ZeroVector(NULL, S, instances);
    for (int set = 0; set < sets; set++) {
        cache_t **cache = NULL;
        int cache_size = 0;
        scalar **dist_cache = NULL;

        ConstructAllNeighborsCache(set_data[set], set_instances[set], features, &dist_cache, &cache, NULL, set_index[set], &cache_size, F, false);

        InstanceSelectionAlgorithm(parm, isalgorithm, &set_S, &set_data[set], &set_instances[set], features, &set_labels[set], classes, set_index[set], cache, dist_cache, minority, &generations,
                                   0.0, NULL, NULL, &evaluations, NULL, false, NULL, false, F);

        for (int i = 0; i < set_instances[set]; i++) {
            FREE(cache[i]);
            FREE(dist_cache[i]);
        }
        FREE(cache);
        FREE(dist_cache);

        for (int i = 0; i < set_instances[set]; i++)
            S[set_list[set][i]] = set_S[i];

        // Evaluate with the same stratum
        scalar a, r;
        performance += SIFS2Performance(parm, F, set_S, set_labels[set], set_instances[set], classes, features, parm->performance_measure, minority, NULL, &a, &r, set_data[set], set_index[set]);

        *accuracy += a;
        *reduction += r;
    }

    // Restore subspace
    // Add feature
    if (count)
        for (int i = 0; i < count; i++)
            F[add[i]] = 0;
    else
        F[delete] = 1;

    performance /= sets;
    *reduction /= sets;
    *accuracy /= sets;

    // Free memory
    FREE(set_subset);
    FREE(set_instances);
    for (int i = 0; i < sets; i++) {
        FREE(set_data[i]);
        if (index)
            FREE(set_index[i]);
        FREE(set_labels[i]);
        FREE(set_list[i]);
    }
    FREE(set_S);
    FREE(set_data);
    FREE(set_index);
    FREE(set_labels);
    FREE(set_list);

    return performance;
}
