/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "macros.h"
#include "isfs.h"
#include "common.h"

// Functions that deal with cache using subspaces

void AllocateSubspaceCache(t_SubspaceCache *cache)
{
    cache->dec_value = NULL;
    cache->nns = NULL;
    cache->size = 0;
    cache->mem_size = 0.0;
}

// Convert binary subspace to decimal and obtain cache of the subspace
double ConvertToDecimalAndUpdateSubspaceCache(int *subspace, t_SubspaceCache *cache, scalar **data, int instances, int inputs, bool feature_selection, int *new, int *used, int **index,
        scalar max_size)
{
    double decimal = 0.0, power = 1.0;

    // No feature selection
    if (!feature_selection) {
        // Is the cache calculated
        if (!cache->size) {
            AddNewSubspace(cache, instances, max_size, -1.0);
            ObtainCache(cache->nns[0], data, instances, inputs, subspace, index);
        }

        return 0.0;
    }

    for (int i = 0; i < inputs; i++) {
        if (subspace[i])
            decimal += power;

        power *= 2;
    }
    decimal = floor(decimal);

    // Is the cache calculated?
    bool found = false;
    for (int i = 0; i < cache->size; i++)
        if (cache->dec_value[i] == decimal) {
            found = true;
            break;
        }

    if (!found) {
        AddNewSubspace(cache, instances, max_size, decimal);
        ObtainCache(cache->nns[cache->size-1], data, instances, inputs, subspace, index);

        (*new)++;
    } else
        (*used)++;

    return decimal;
}

// Return the appropriate cache

cache_t **ReturnCache(bool use_cache, bool use_stepwise_cache, cache_t ***nn_cache, t_SubspaceCache *cache, double decimal, bool *done)
{
    // No cache
    if (!use_cache && !use_stepwise_cache)
        return NULL;
    else if (use_cache) {
        if (!done[(int) decimal])
            OnErrorExit(ERR_UNSPECTED_RESULT, "Cache for this subspace should be present");

        return nn_cache[(int) decimal];
    } else {    // Step wise cache
        for (int i = 0; i < cache->size; i++)
            if (cache->dec_value[i] == decimal)
                return cache->nns[i];

        OnErrorExit(ERR_UNSPECTED_RESULT, "Cache for this subspace should be present");
    }

    return NULL;
}

void AddNewSubspace(t_SubspaceCache *cache, int instances, scalar max_size, scalar value)
{
    REALLOC(cache->dec_value, double, cache->size+1, "AddNewSubspace: cache->dec_value");
    REALLOC(cache->nns, cache_t **, cache->size+1, "AddNewSubspace: cache->nns");
    MALLOC(cache->nns[cache->size], cache_t *, instances, "AddNewSubspace: (*cache)[j]");
    for (int i = 0; i < instances; i++)
        MALLOC(cache->nns[cache->size][i], cache_t, instances, "AddNewSubspace: (*cache)[j][i]");
    cache->dec_value[cache->size] = value;
    cache->size++;

    cache->mem_size += (sizeof(cache_t *) * instances + sizeof(cache_t) * instances * instances)/pow(2.0, 30);   // In GB

    if (cache->mem_size > max_size)
        OnErrorExit(ERR_MEMORY, "Maximum IS+FS cache size exceeded");
}

void FreeSubspaceCache(t_SubspaceCache *cache, int instances)
{
    FREE(cache->dec_value);
    for (int i = 0; i < cache->size; i++) {
        for (int j = 0; j < instances; j++)
            FREE(cache->nns[i][j]);
        FREE(cache->nns[i])
    }
    FREE(cache->nns);
}
