/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <search.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "isfs.h"

// Evaluate selection threshold using VOOR
// B. Seijo-Pardo, V. Bolón-Canedo and A. Alonso-Betanzos, "On developing an automatic threshold applied to feature selection", Information Fusion (45):227-245, 2019.

void MulticlassVolumeOfOverlapRegion(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, scalar *votes_fs, scalar *coefficient, scalar *r, scalar *complexity, int classes,
                                     scalar *threshold_fs, int n_thresholds_fs)
{
    int *S;

    MALLOC(S, int, inputs, "MulticlassVolumeOfOverlapRegion: S");

    // Get coefficient for every threshold
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    for (int fs = 0; fs < n_thresholds_fs; fs++) {
        int selected = 0;

        scalar th_fs = threshold_fs[fs];

        SHOWPROGRESS(parm->show_progress > 0, "Evaluating thresholds FS: %*d/%d\r", WIDTH(n_thresholds_fs), fs+1, n_thresholds_fs);

        // Select features
        for (int i = 0; i < inputs; i++)
            if (votes_fs[i] >= th_fs) {
                S[i] = 1;
                selected++;
            } else
                S[i] = 0;

        r[fs] = ((scalar) selected)/inputs;

        complexity[fs] = 1.0;

        for (int i = 1; i < classes; i++)
            for (int j = 0; j < i; j++)
                complexity[fs] *= fabs(VolumeOfOverlapRegion(data, instances, labels, i, j, inputs, S, NULL));

        coefficient[fs] = complexity[fs] * parm->alpha_ok + parm->alpha_red * r[fs];
    }

    FREE(S);

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}

void LOORThresholdEvaluation(const t_Parameters *parm, scalar **data, int *labels, int instances, int inputs, scalar *votes_fs, scalar *coefficient, scalar *r, scalar *complexity, int classes,
                             scalar *threshold_fs, int n_thresholds_fs)
{
    int *S;

    MALLOC(S, int, inputs, "LOORThresholdEvaluation: S");

    // Get coefficient for every threshold
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    for (int fs = 0; fs < n_thresholds_fs; fs++) {
        int selected = 0;

        scalar th_fs = threshold_fs[fs];

        SHOWPROGRESS(parm->show_progress > 0, "Evaluating thresholds FS: %*d/%d\r", WIDTH(n_thresholds_fs), fs+1, n_thresholds_fs);

        // Select features
        for (int i = 0; i < inputs; i++)
            if (votes_fs[i] >= th_fs) {
                S[i] = 1;
                selected++;
            } else
                S[i] = 0;

        r[fs] = ((scalar) selected)/inputs;

        complexity[fs] = 0.0;

        for (int i = 1; i < classes; i++)
            for (int j = 0; j < i; j++)
                // Sum instead of product following authors' recommendation
                complexity[fs] += fabs(VolumeOfOverlapRegion(data, instances, labels, i, j, inputs, S, NULL));

        coefficient[fs] = complexity[fs] * parm->alpha_ok + parm->alpha_red * r[fs];
    }

    FREE(S);

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}

#define VERYSMALL   -10.0E20
#define VERYLARGE   +10.0E20

scalar MaxValueOfFeatureInClass(scalar **data, int instances, int *labels, int feature, int class, int *subset)
{
    scalar max = VERYSMALL;

    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            if (labels[i] == class && data[i][feature] > max)
                max = data[i][feature];

    return max;
}

scalar MinValueOfFeatureInClass(scalar **data, int instances, int *labels, int feature, int class, int *subset)
{
    scalar min = VERYLARGE;

    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            if (labels[i] == class && data[i][feature] < min)
                min = data[i][feature];

    return min;
}

// Volume of overlap region for a pair of classes

scalar VolumeOfOverlapRegion(scalar **data, int instances, int *labels, int c1, int c2, int features, int *subspace, int *subset)
{
    scalar f2 = 1.0, minmax, maxmin, maxmax, minmin;

    for (int i = 0; i < features; i++)
        if (subspace[i]) {
            minmax = MIN(MaxValueOfFeatureInClass(data, instances, labels, i, c1, subset), MaxValueOfFeatureInClass(data, instances, labels, i, c2, subset));
            maxmin = MAX(MinValueOfFeatureInClass(data, instances, labels, i, c1, subset), MinValueOfFeatureInClass(data, instances, labels, i, c2, subset));
            maxmax = MAX(MaxValueOfFeatureInClass(data, instances, labels, i, c1, subset), MaxValueOfFeatureInClass(data, instances, labels, i, c2, subset));
            minmin = MIN(MinValueOfFeatureInClass(data, instances, labels, i, c1, subset), MinValueOfFeatureInClass(data, instances, labels, i, c2, subset));

            // Length instead of volume following authors' recommendation
            // With volume too small values are obtained
            if (fabs(maxmax - minmin) > ZERO && fabs(minmax - VERYSMALL) > ZERO && fabs(maxmin - VERYLARGE) > ZERO)
                f2 *= (minmax - maxmin)/(maxmax-minmin);
        }

    return f2;
}
