/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <pvm3.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include "pvm-local.h"
#include "macros.h"
#include "is.h"

// Get best round of electors
int GetBestRound(const t_Parameters *parm, byte **S, int n_subsets, int processors, FILE *binary, char *parameters_file, scalar *best_coef, int inputs, int classes, int instances,
                 scalar **data, int *train_labels)
{
    int count[n_subsets], tids[processors], *partition;

    // Evaluation of rounds time in seconds
    time_t begin_evaluation = time(NULL);

    if (!parm->democratic_evaluation_of_threshold)
        OnErrorExit(1, "Best round only implemented with democratic evaluation");

    // Values per round
    scalar r[parm->n_electors], ok[parm->n_electors], coef[parm->n_electors];
    // Values per evaluation
    scalar round_r[parm->n_electors], round_ok[parm->n_electors], round_coef[parm->n_electors];

    // Initialize values
    for (int i = 0; i < parm->n_electors; i++)
        r[i] = ok[i] = coef[i] = 0.0;

    MALLOC(partition, int, instances, "GetBestRound: partition");

    if (parm->external_partitioning_method == pDecoupled)
        fprintf(parm->out_file, "For best round evaluation random partition is used\n");

    // Perform the stated rounds
    for (int round = 0; round < parm->demo_eval_rounds; round++) {
        switch (parm->external_partitioning_method) {
            case pRANDOM:
            case pDecoupled:
            case peBagging:
                // Random partition
                if (parm->external_stratified_approach)
                    RandomPartitionKCD(partition, count, instances, n_subsets, train_labels, classes, instances);
                else
                    RandomPartition(partition, count, instances, n_subsets);
                break;

            case pEQUAL:
                if (parm->external_stratified_approach)
                    PartitionateEqualKCD(partition, count, data, instances, n_subsets, train_labels, inputs, classes);
                else
                    PartitionateEqual(partition, count, data, instances, n_subsets, train_labels, inputs, classes, parm->use_pcomponents);
                break;

            case pBOARD:
                PartitionateBoard(partition, count, data, instances, n_subsets, inputs);
                break;

            case pRPROTOS:
                PartitionateRandomPrototypes(partition, count, data, instances, n_subsets, inputs, NULL);
                break;

            case p2DGrandTour:
                BidimensionalGrandTour(parm, partition, count, data, instances, &n_subsets, train_labels, inputs, classes);
                break;

            case pOneInput:
            case pGrandTour:
                if (parm->external_stratified_approach)
                    PartitionateOneInputKCD(parm, partition, count, data, instances, n_subsets, train_labels, inputs, classes, parm->external_partitioning_method);
                else
                    PartitionateOneInput(parm, partition, count, data, instances, n_subsets, train_labels, inputs, classes, parm->external_partitioning_method);
                break;

            case pkMeans:
                kMeans(partition, count, data, instances, n_subsets, inputs, parm->kmeans_max_iterations, NULL, NULL);
                break;

            case pkClustering:
                OnErrorExit(1, "kClustering not yet allowed");
                break;
        }
        // Get maximum subset size
        int max_size = count[0];
        for (int i = 1; i < n_subsets; i++)
            if (count[i] > max_size)
                max_size = count[i];

        ParallelDemocraticRoundEvaluation(parm, data, train_labels, instances, partition, n_subsets, S, round_ok, round_r, round_coef, processors, tids, parameters_file, binary, max_size,
                                          inputs, classes);
        if (parm->verbose > 0)
            printf("Finished evaluation of round %d/%d\n", round+1, parm->demo_eval_rounds);

        for (int i = 0; i < parm->n_electors; i++) {
            r[i] += round_r[i]/parm->demo_eval_rounds;
            ok[i] += round_ok[i]/parm->demo_eval_rounds;
            coef[i] += round_coef[i]/parm->demo_eval_rounds;
        }
    }

    FREE(partition);

    // Kill spawned processes
    for (int i = 0; i < processors; i++)
        pvm_kill(tids[i]);

    // Get best round
    int best_round = 0;
    for (int i = 1; i < parm->n_electors; i++)
        if (coef[i] > coef[best_round])
            best_round = i;

    fprintf(parm->out_file, "Evaluation of rounds in %ld seconds\n", time(NULL) - begin_evaluation);

    if (parm->verbose > 0) {
        fprintf(parm->out_file, "Training info\nRound\tT acc\tReduct\tCoefficient\n");
        for (int i = 0; i < parm->n_electors; i++)
            fprintf(parm->out_file, "%3d:\t%6.4lf\t%6.4lf\t%6.4lf %s\n", i, ok[i], r[i], coef[i], i == best_round ? "(*)" : "");
    }

    *best_coef = coef[best_round];

    return best_round;
}

// Evaluates a round for votes using the same method that for democratic selection
// using a parallel implementation
void ParallelDemocraticRoundEvaluation(const t_Parameters *parm, scalar **data, int *labels, int n, int *partition, int n_subsets, byte **S, scalar *ok, scalar *r, scalar *coef,
                                       int processors, int *tids, char *parameters_file, FILE *binary, int max_instances, int inputs, int classes)
{
    int *local_labels;
    scalar **local_data, *r_slave, *ok_slave;
    static bool first_time = true;
    byte **local_S;

    // If first time run processes
    if (first_time) {
        // Parameters for spawning process
        char *args[3];
        MALLOC(args[0], char, strlen(parameters_file)+1, "main");
        strcpy(args[0], parameters_file);       // Parameters file name
        args[1] = getcwd(NULL, 0);      // Current directory for slaves
        args[2] = NULL;

        // Avoid running processes in master node
        int numt = pvm_spawn("eval-round-wrapper", args, PvmTaskHost | PvmHostCompl, "skynet.uco.es", processors, tids);
        // int numt = pvm_spawn("/home/frodo/scripts/eval-round-wrapper", args, PvmTaskDefault+PvmTaskDebug, "skynet.uco.es", processors, tids);
        if (numt != processors)
            OnErrorExit(1, "ParallelDemocraticRoundEvaluation: Not enough processes spawned");

        first_time = false;
    }

    // Initialize values
    for (int i = 0; i < parm->n_electors; i++)
        r[i] = coef[i] = ok[i] = 0.0;

    // Obtain maximum size for subset
    MALLOC(local_data, scalar *, max_instances, "ParallelDemocraticRoundEvaluation: local_data");
    MALLOC(local_labels, int, max_instances, "ParallelDemocraticRoundEvaluation: local_labels");
    MALLOC(local_S, byte *, parm->n_electors, "ParallelDemocraticRoundEvaluation: local_S");
    for (int i = 0; i < parm->n_electors; i++)
        MALLOC(local_S[i], byte, max_instances, "ParallelDemocraticRoundEvaluation: local_S[i]");

    ALLOCA(r_slave, scalar, parm->n_electors,  "ParallelDemocraticRoundEvaluation: r_slave");
    ALLOCA(ok_slave, scalar, parm->n_electors,  "ParallelDemocraticRoundEvaluation: ok_slave");

    // If huge dataset allocate memory
    if (parm->huge_dataset)
        for (int i = 0; i < max_instances; i++)
            MALLOC(local_data[i], scalar, inputs, "ParallelDemocraticRoundEvaluation: local_data[i]");

    // Evaluate for every subset
    for (int subset = 0; subset < n_subsets; subset += processors) {
        for (int p = 0; p < processors && subset+p < n_subsets; p++) {
            // Get subset
            int cursor = 0;
            if (parm->huge_dataset) {
                for (int j = 0; j < n; j++)
                    if (partition[j] == subset+p) {
                        // Get instance from file
                        GetInstanceFromBinaryFile(local_data[cursor], &local_labels[cursor], j, inputs, binary);
                        for (int i = 0; i < parm->n_electors; i++)
                            local_S[i][cursor] = S[i][j];
                        cursor++;
                    }
            } else {
                for (int j = 0; j < n; j++)
                    if (partition[j] == subset+p) {
                        local_data[cursor] = data[j];
                        local_labels[cursor] = labels[j];
                        for (int i = 0; i < parm->n_electors; i++)
                            local_S[i][cursor] = S[i][j];
                        cursor++;
                    }
            }

            // Send data
            pvm_initsend(PvmDataDefault);     // PvmDataDefault 0 XDR
            pvm_pkint(&cursor, 1, 1);     // Number of instances
            pvm_pkint(&inputs, 1, 1);     // Number of instances
            pvm_pkint(&((t_Parameters *)parm)->k, 1, 1);
            pvm_pkint(&classes, 1, 1);
            pvm_pkint(&((t_Parameters *)parm)->n_electors, 1, 1);
            for (int i = 0; i < parm->n_electors; i++)
                pvm_pkbyte((char *) local_S[i], cursor, 1);      // S for each round
            pvm_pkint(local_labels, cursor, 1);    // Labels
            for (int i = 0; i < cursor; i++)
                pvm_pkscalar(local_data[i], inputs, 1);
            pvm_send(tids[p], SENT_BY_MASTER);
        }

        // Check no problem with processes
        for (int p = 0; p < processors; p++)
            if (pvm_pstat(tids[p]) == PvmNoTask) {
                printf("Died after %d/%d subsets processed\n", subset, n_subsets);
                OnErrorExit(ERR_PVM, "PVM Task died unexpectedly");
            }

        // Get values
        for (int p = 0; p < processors && subset+p < n_subsets; p++) {
            // Receive data
            pvm_recv(-1, SENT_BY_SLAVE);    // Receive from any slave

            pvm_upkscalar(r_slave, parm->n_electors, 1);
            pvm_upkscalar(ok_slave, parm->n_electors, 1);

            // Get coefficient for every threshold
            for (int th = 0; th < parm->n_electors; th++) {
                // Evaluate
                r[th] += r_slave[th];
                ok[th] += ok_slave[th];

                coef[th] += parm->alpha_acc * ok_slave[th]+(1.0 - parm->alpha_acc) * r_slave[th];
            }
            if (parm->verbose > 0) {
                printf("Evaluation of rounds %d/%d\n", subset+p+1, n_subsets);
                fflush(stdout);
            }
        }
    }

    // Average
    for (int th = 0; th < parm->n_electors; th++) {
        r[th] /= n_subsets;
        coef[th] /= n_subsets;
        ok[th] /= n_subsets;
    }

    // If huge dataset free memory
    if (parm->huge_dataset)
        for (int i = 0; i < max_instances; i++)
            FREE(local_data[i]);

    FREE(local_data);
    FREE(local_labels);
    FREE(local_S);
}
