/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdbool.h>
#include <time.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "is.h"
#include "fs.h"
#include "classifiers.h"
#include "maths.h"
#include "ensemble.h"

#define PRECISION   0.000001

int BoostBasedInstanceSelection(const t_Parameters *parm, int *selection, scalar **data, int instances, int *label, bool use_cache, int inputs, int classes, bool knn_with_weighted_votes, scalar *r,
                                scalar *distance_to_furthest_neighbor, int minority, bool as_a_mutation, int **index, scalar **distance_cache, cache_t **nn_cache, int *subspace)
{
    cache_t **neighbors_cache = NULL;
    scalar *w, best_performance, accuracy, **dist_cache = NULL, *previous_dist_to_nn, averaged_performance = 0.0, best_acc[classes];
    int chance = parm->second_chances+1, selected, *best_S = NULL, *previous_prediction, boost_based_filter_noise_k, *S, best_instance[classes], *added_list, added = 0, best_selection[classes],
        *reweight_s = NULL;
    bool *ok, *marked;

    if (knn_with_weighted_votes)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Weighted votes with Boosted IS");

    MALLOC(S, int, instances, "BoostBasedInstanceSelection: S");

    // Create nearest neighbors cache
    if (use_cache) {
        RANK *rank;

        if (distance_cache)
            dist_cache = distance_cache;
        else {
            MALLOC(dist_cache, scalar *, instances, "BoostBasedInstanceSelection: dist_cache");
            for (int i = 0; i < instances; i++)
                MALLOC(dist_cache[i], scalar, instances, "BoostBasedInstanceSelection: dist_cache[i]");

            for (int i = 0; i < instances; i++) {
                dist_cache[i][i] = 0.0;
                for (int j = 0; j < i; j++)
                    dist_cache[i][j] = dist_cache[j][i] = EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace)/(r ? r[j] : 1.0);
            }
        }

        if (nn_cache)
            neighbors_cache = nn_cache;
        else {
            MALLOC(rank, RANK, instances, "BoostBasedInstanceSelection: rank");
            MALLOC(neighbors_cache, cache_t *, instances, "BoostBasedInstanceSelection: neighbors_cache");

            for (int i = 0; i < instances; i++) {
                MALLOC(neighbors_cache[i], cache_t, instances, "BoostBasedInstanceSelection: neighbors_cache[i]");

                for (int j = 0; j < instances; j++) {
                    rank[j].fitness = -dist_cache[i][j];
                    rank[j].being = j;
                }

                // Sort
                qsort(rank, instances, sizeof(RANK), Compare);

                // Store neighbors in ascending order of distance
                for (int j = 0; j < instances; j++)
                    neighbors_cache[i][j] = rank[j].being;

            }
            FREE(rank);
        }
    } else
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Boost based instance selection without cache");

    // Initialize selection and weights
    MALLOC(w, scalar, instances, "BoostBasedInstanceSelection: w");
    MALLOC(ok, bool, instances, "BoostBasedInstanceSelection: ok");
    MALLOC(best_S, int, instances, "BoostBasedInstanceSelection: best_S");
    MALLOC(previous_prediction, int, instances, "BoostBasedInstanceSelection: previous_prediction");
    MALLOC(marked, bool, instances, "BoostBasedInstanceSelection: marked");
    MALLOC(previous_dist_to_nn, scalar, instances, "BoostBasedInstanceSelection: previous_dist_to_nn");
    MALLOC(added_list, int, instances, "BoostBasedInstanceSelection: added_list");
    if (parm->boostedis_ensemble_method == eReweightBoost)
        MALLOC(reweight_s, int, instances, "BoostBasedInstanceSelection: reweight_s");
    for (int i = 0; i < instances; i++) {
        w[i] = 1.0/instances;
        S[i] = 0;
        // Initially none marked
        marked[i] = false;
        // Initial previous prediction is unknown and distance to far far away
        previous_prediction[i] = -1;
        previous_dist_to_nn[i] = INFINITE;
    }

    // Filter noise marking all noise instances to be avoided
    if (parm->filter_noise) {
        time_t init_time = time(NULL);
        for (int i = 0; i < instances; i++)
            S[i] = selection[i];

        // Get k?
        if (parm->boost_based_filter_noise_k <= 0) {
            boost_based_filter_noise_k = ObtainkCrossValidation((t_Parameters *) parm, S, data, instances, label, inputs, classes, index, subspace);
            fprintf(parm->out_file, "@Wilson's editing k: %d\n", boost_based_filter_noise_k);
        } else
            boost_based_filter_noise_k = parm->boost_based_filter_noise_k;

        int c = 0, nearest[boost_based_filter_noise_k];

        for (int i = 0; i < instances; i++)
            if (S[i]) {
                S[i] = 0;

                if (kNearestNeighbourCache(i, S, instances, nearest, label, boost_based_filter_noise_k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, NULL, subspace)
                    != label[i]) {
                    marked[i] = true;
                    c++;
                }

                S[i] = 1;
            } else
                marked[i] = true;

        if (parm->verbose > 0) {
            fprintf(parm->out_file, "@Wilson's editing reduction: %6.4lf\n", ((scalar) c)/instances);
            fprintf(parm->out_file, "@Wilson's editing time: %ld\n", time(NULL) - init_time);
            fflush(parm->out_file);
        }
    }

    // CV k?
    int k;
    if (parm->k <= 0) {
        for (int i = 0; i < instances; i++)
            S[i] = marked[i] ? 0 : 1;
        k = ObtainkCrossValidation((t_Parameters *) parm, S, data, instances, label, inputs, classes, index, subspace);
        fprintf(parm->out_file, "@CV k: %d\n", k);
    } else
        k = parm->k;

    // Multi-selection of instances?
    int max_sel = parm->multiselection ? k/2+1 : 1;

    // Initialization depends on type of heuristic search or if the IS is used as a mutation operator
    if (as_a_mutation)
        for (int i = 0; i < instances; i++)
            S[i] = selection[i];
    else {
        if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
            for (int i = 0; i < instances; i++)
                S[i] = 0;

            if (parm->bbis_select_best_instance_from_every_class) {
                // Select best instance from every class
                for (int i = 0; i < classes; i++) {
                    best_acc[i] = -1.0;
                    best_instance[i] = -1;
                }

                for (int l = 0; l < classes; l++) {
                    // Set unmarked instances of the remaining classes
                    for (int i = 0; i < instances; i++)
                        S[i] = (marked[i] || label[i] == l) ? 0 : selection[i];

                    for (int i = 0; i < instances; i++)
                        if (!marked[i] && label[i] == l) {
                            scalar acc;

                            for (int j = 1; j <= max_sel; j++) {
                                // Add instance and check
                                S[i] = j;

                                if ((acc = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok,
                                                                false, previous_prediction, false, dist_cache, previous_dist_to_nn, i, -1, subspace)) > best_acc[l]) {
                                    best_acc[l] = acc;
                                    best_instance[l] = i;
                                    best_selection[l] = j;
                                }
                            }

                            S[i] = 0;
                        }
                }

                for (int i = 0; i < classes; i++)
                    if (best_instance[i] != -1) {
                        S[best_instance[i]] = best_selection[i];
                        added_list[added] = best_instance[i];
                        added++;
                    }
            }
        } else {
            // Start with all non-marked instances
            for (int i = 0; i < instances; i++)
                S[i] = marked[i] ? 0 : 1;
        }
    }

    // Set l and r appropriately
    int plus_l = -1, minus_r = -1;
    switch (parm->search_heuristic) {
        case SFS:
            plus_l = 1;
            minus_r = 0;
            break;

        case SBS:
            plus_l = 0;
            minus_r = 1;
            break;

        case LRS:
            plus_l = parm->plus_l;
            minus_r = parm->minus_r;
            break;

        case SFFS:
        case SFBS:
            plus_l = minus_r = 1;
            break;
    }

    // Iterate over all features or subset
    int iteration_lenght = parm->bbis_instances_per_iteration ? MIN(parm->bbis_instances_per_iteration, instances) : instances;

    // Count selected so far
    selected = 0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            selected++;

    best_performance = accuracy = -1.0;
    int iterations = 0;
    scalar previous_performance = -1.0, performance = -1.0, reduction;
    int last_added = -1, last_removed = -1;
    bool all_marked;
    do {
        scalar best_acc, acc, accuracy_w = 0.0, previous = 0.0;
        int best_instance = -1, best_n_copies = -1;

        previous_performance = parm->use_averaged_accuracy_for_stop ? averaged_performance : performance;

        for (int j = 0; j < plus_l && selected < instances; j++) {
            best_instance = -1;
            best_acc = -1.0;
            for (int i = 0; i < iteration_lenght; i++) {
                int next = (iteration_lenght == instances) ? i : RANDOM(0, instances);
                if (!S[next] && !marked[next]) {
                    for (int j = 1; j <= max_sel; j++) {
                        // Add instance and check
                        S[next] = j;

                        if ((acc = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                                        previous_prediction, false, dist_cache, previous_dist_to_nn, next, -1, subspace)) > best_acc) {
                            best_acc = acc;
                            best_instance = next;
                            best_n_copies = j;
                        }
                    }

                    S[next] = 0;
                }
            }

            // Add best
            if (best_instance != -1) {
                S[best_instance] = best_n_copies;
                if (best_instance == last_removed)
                    chance--;
                added_list[added] = last_added = best_instance;
                added++;
                last_removed = -1;
                selected++;
                SHOWPROGRESS(parm->show_progress > 1, "Added instance: %d\n", best_instance);
                // Recalculate Boost weights
                accuracy_w = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                                  previous_prediction, true, dist_cache, previous_dist_to_nn, best_instance, -1, subspace);
                if (parm->boostedis_ensemble_method == eReweightBoost) {
                    // Evaluate with last 'k' elements added
                    for (int i = 0; i < instances; i++)
                        reweight_s[i] = 0;
                    for (int i = 0; i < parm->reweightboost_r && i < added; i++)
                        reweight_s[added_list[i]] = S[added_list[i]];
                    EvaluatekWithWeights(k, reweight_s, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                         previous_prediction, false, dist_cache, previous_dist_to_nn, -1, -1, subspace);
                }
                UpdateWeights(parm, instances, w, 1.0 - accuracy_w, ok, parm->boostedis_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations),
                              parm->boostedis_ensemble_method, data, label, S, inputs, k, classes, dist_cache, index, true, false, NULL, kNN, subspace);

                // For SFFS repeat deletion in the performance does not worsen, otherwise undo
                if (parm->search_heuristic == SFBS) {
                    accuracy = PerformanceWithCache(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, previous_prediction,
                                                    parm->evaluate_with_loo, parm->performance_measure, minority, subspace);
                    if (previous < accuracy) {
                        j--;
                        previous = accuracy;
                    } else {
                        S[best_instance] = 0;
                        selected--;
                        SHOWPROGRESS(parm->show_progress > 1, "Undo added instance: %d\n", best_instance);
                        accuracy_w = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                                          previous_prediction, true, dist_cache, previous_dist_to_nn, -1, best_instance, subspace);
                        UpdateWeights(parm, instances, w, 1.0 - accuracy_w, ok, parm->boostedis_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations),
                                      parm->boostedis_ensemble_method, data, label, S, inputs, k, classes, dist_cache, index, true, false, NULL, kNN, subspace);
                    }
                }

            } else if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l > minus_r))
                chance = 0;
        }

        if (parm->search_heuristic == SFFS)
            previous = PerformanceWithCache(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, previous_prediction,
                                            parm->evaluate_with_loo, parm->performance_measure, minority, subspace);

        for (int j = 0; j < minus_r && selected > 1; j++) {
            best_instance = -1;
            best_acc = -1.0;
            for (int i = 0; i < instances; i++)
                if (S[i]) {
                    int backup = S[i];
                    // Remove instance and check
                    S[i] = 0;

                    if ((acc = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                                    previous_prediction, false, dist_cache, previous_dist_to_nn, -1, i, subspace)) > best_acc) {
                        best_acc = acc;
                        best_instance = i;
                    }

                    S[i] = backup;
                }

            // Add best
            if (best_instance != -1) {
                int backup = S[best_instance];
                S[best_instance] = 0;
                if (best_instance == last_added)
                    chance--;
                last_removed = best_instance;
                last_added = -1;
                selected--;
                SHOWPROGRESS(parm->show_progress > 1, "Removed instance: %d\n", best_instance);
                // Recalculate Boost weights
                accuracy_w = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                                  previous_prediction, true, dist_cache, previous_dist_to_nn, -1, best_instance, subspace);
                UpdateWeights(parm, instances, w, 1.0 - accuracy_w, ok, parm->boostedis_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations),
                              parm->boostedis_ensemble_method, data, label, S, inputs, k, classes, dist_cache, index, true, false, NULL, kNN, subspace);

                // For SFFS repeat deletion in the performance does not worsen, otherwise undo
                if (parm->search_heuristic == SFFS) {
                    accuracy = PerformanceWithCache(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, previous_prediction,
                                                    parm->evaluate_with_loo, parm->performance_measure, minority, subspace);
                    if (previous <= accuracy) {
                        j--;
                        previous = accuracy;
                    } else {
                        S[best_instance] = backup;
                        selected++;
                        SHOWPROGRESS(parm->show_progress > 1, "Undo removed instance: %d\n", best_instance);
                        accuracy_w = EvaluatekWithWeights(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, w, ok, false,
                                                          previous_prediction, true, dist_cache, previous_dist_to_nn, best_instance, -1, subspace);
                        UpdateWeights(parm, instances, w, 1.0 - accuracy_w, ok, parm->boostedis_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations),
                                      parm->boostedis_ensemble_method, data, label, S, inputs, k, classes, dist_cache, index, true, false, NULL, kNN, subspace);
                    }
                }
                UpdateWeights(parm, instances, w, 1.0 - accuracy_w, ok, parm->boostedis_ensemble_method == eMultiBoost && !(iterations % parm->multiboost_reinit_iterations),
                              parm->boostedis_ensemble_method, data, label, S, inputs, k, classes, dist_cache, index, true, false, NULL, kNN, subspace);
            } else if (parm->search_heuristic == SBS || parm->search_heuristic == SFBS || (parm->search_heuristic == LRS && parm->plus_l < minus_r))
                chance = 0;
        }

        if (parm->search_heuristic == SFBS)
            previous = PerformanceWithCache(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, previous_prediction,
                                            parm->evaluate_with_loo, parm->performance_measure, minority, subspace);

        // Get unweighted Kappa measure
        iterations++;
        accuracy = PerformanceWithCache(k, S, neighbors_cache, label, instances, classes, data, inputs, knn_with_weighted_votes, r, distance_to_furthest_neighbor, previous_prediction,
                                        parm->evaluate_with_loo, parm->performance_measure, minority, subspace);
        reduction = 1.0 - ((scalar) selected/instances);

        performance = accuracy*parm->ok_w+reduction*parm->reduction_w;
        averaged_performance = (averaged_performance*(iterations-1)+performance)/iterations;

        SHOWPROGRESS(parm->show_progress > 1, "Accuracy: %8.6lf Reduction: %8.6lf Performance: %8.6lf Averaged: %8.6lf (%3d, %5d, %5d)\n", accuracy, reduction, performance, averaged_performance,
                     chance, iterations, selected);

        bool best_so_far = false;
        if (parm->use_averaged_accuracy_for_stop) {
            if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
                if ((previous_performance > averaged_performance || fabs(previous_performance - averaged_performance) <= PRECISION))
                    chance--;
            } else {
                if ((previous_performance > averaged_performance || fabs(previous_performance - averaged_performance) < PRECISION))
                    chance--;
            }
            best_so_far = performance > best_performance;
        } else {
            if (parm->search_heuristic == SFS || parm->search_heuristic == SFFS || (parm->search_heuristic == LRS && parm->plus_l >= parm->minus_r)) {
                if ((previous_performance > performance || fabs(previous_performance - performance) <= PRECISION))
                    chance--;
            } else {
                if ((previous_performance > performance || fabs(previous_performance - performance) < PRECISION))
                    chance--;
            }
            best_so_far = performance >= best_performance;
        }

        // If none selected and all marked return
        all_marked = false;
        if (!selected) {
            all_marked = true;
            for (int i = 0; i < instances && all_marked; i++)
                if (!marked[i])
                    all_marked = false;
        }

        if (best_so_far) {
            best_performance = performance;
            // Best selection so far
            for (int i = 0; i < instances; i++)
                best_S[i] = S[i];
        }
    } while (!all_marked && ((chance > 0 && selected > 1 && selected < instances) || selected < k));

    // Best selection
    selected = 0;
    for (int i = 0; i < instances; i++)
        if ((selection[i] = best_S[i]))
            selected++;
    FREE(best_S);

    FREE(w);
    FREE(ok);
    FREE(marked);
    FREE(previous_prediction);
    FREE(previous_dist_to_nn);
    FREE(added_list);
    if (use_cache && (!distance_cache || !nn_cache)) {
        for (int i = 0; i < instances; i++) {
            if (!distance_cache)
                FREE(dist_cache[i]);
            if (!nn_cache)
                FREE(neighbors_cache[i]);
        }
        if (!distance_cache)
            FREE(dist_cache);
        if (!nn_cache)
            FREE(neighbors_cache);
    }
    if (parm->boostedis_ensemble_method == eReweightBoost)
        FREE(reweight_s);
    FREE(S);

    return selected;
}

scalar EvaluatekWithWeights(int k, int *S, cache_t  **neighbors_cache, int *label, int instances, int classes, scalar **data, int inputs, bool knn_with_weighted_votes, scalar *r,
                            scalar *distance_to_furthest_neighbor, scalar *w, bool *ok, bool loo, int *previous_prediction, bool update, scalar **dist_cache, scalar *previous_dist_to_nn,
                            int last_added, int last_removed, int *subspace)
{
    int nearest[k], prediction;
    scalar accuracy = 0.0;

    // Evaluation depends of whether instance is added or removed
    if (last_added != -1) {
        for (int i = 0; i < instances; i++) {
            if (previous_prediction[i] == label[last_added])
                prediction = previous_prediction[i];
            else if (dist_cache[i][last_added] >= previous_dist_to_nn[i])
                prediction = previous_prediction[i];
            else {
                int backup = S[i];
                if (loo)
                    S[i] = 0;
                prediction = kNearestNeighbourCache(i, S, instances, nearest, label, k, neighbors_cache, data, inputs, classes, knn_with_weighted_votes, r, distance_to_furthest_neighbor, NULL,
                                                    subspace);

                S[i] = backup;

                if (update) {
                    previous_prediction[i] = prediction;
                    previous_dist_to_nn[i] = dist_cache[i][nearest[0]];
                }
            }

            if (prediction == label[i]) {
                accuracy += w[i];
                ok[i] = true;
            } else
                ok[i] = false;
        }
    } else if (last_removed != -1) {
        for (int i = 0; i < instances; i++) {
            if (previous_prediction[i] != -1 && previous_prediction[i] != label[last_removed])
                prediction = previous_prediction[i];
            else if (previous_prediction[i] != -1 && dist_cache[i][last_removed] > previous_dist_to_nn[i])
                prediction = previous_prediction[i];
            else {
                int backup = S[i];
                if (loo)
                    S[i] = 0;
                prediction = kNearestNeighbourCache(i, S, instances, nearest, label, k, neighbors_cache, data, inputs, classes, knn_with_weighted_votes, r, distance_to_furthest_neighbor, NULL,
                                                    subspace);

                S[i] = backup;

                if (update) {
                    previous_prediction[i] = prediction;
                    previous_dist_to_nn[i] = dist_cache[i][nearest[0]];
                }
            }

            if (prediction == label[i]) {
                accuracy += w[i];
                ok[i] = true;
            } else
                ok[i] = false;
        }
    } else {
        // Recalculate always
        for (int i = 0; i < instances; i++) {
            int backup = S[i];
            if (loo)
                S[i] = 0;

            if (label[i] == kNearestNeighbourCache(i, S, instances, nearest, label, k, neighbors_cache, data, inputs, classes, knn_with_weighted_votes, r, distance_to_furthest_neighbor, NULL,
                                                   subspace)) {
                accuracy += w[i];
                ok[i] = true;
            } else
                ok[i] = false;

            S[i] = backup;
        }
    }

    return accuracy;
}

void UpdateWeights(const t_Parameters *parm, int instances, scalar *w, scalar e_i, bool *ok, bool reinit_w, t_EnsembleMethod e_method, scalar **data, int *labels, int *S, int inputs, int k,
                   int classes, scalar **dist_cache, int **index, bool is, bool fs, scalar *h, t_Classifier classifier, int *subspace)
{
    scalar sum = 0.0;

    switch (e_method) {
        case eFloatBoost:
            if (is)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "eFloatBoost for Boosted-Based IS");
        // Else, same weights that eGentleBoost
        case eGentleBoost:
            if (is)
                for (int j = 0; j < instances; j++)
                    sum += (w[j] *= exp(-kNearestNeighbourMargin(data[j], labels[j], subspace, data, instances, inputs, k, labels, classes, S)));
            else if (fs) {
                scalar *margin, min_margin;

                // Margin already obtained?
                if (h)
                    margin = h;
                else {
                    MALLOC(margin, scalar, instances, "UpdateWeights: margin");
                    ClassifierDatasetMargin(parm, classifier, data, instances, labels, data, instances, labels, inputs, classes, subspace, &min_margin, NULL, margin, S, NULL, !IsEnsemble(classifier));
                }

                for (int j = 0; j < instances; j++)
                    sum += (w[j] *= exp(-margin[j]));

                if (!h)
                    FREE(margin);
            }
            break;

        case AdaBoost:
        case eMadaBoost:
        case eMultiBoost:
            // If above 50% random re-sampling
            if (e_i >= 0.5 || e_i <= 0.0 || reinit_w) {
                for (int k = 0; k < instances; k++)
                    sum += (w[k] = Poisson());
                SHOWPROGRESS(parm->show_progress > 1 && parm->verbose > 0, "Reinitializing weights with error %lf\n", e_i);
            } else {
                for (int k = 0; k < instances; k++)
                    if (ok[k])
                        sum += (w[k] *= sqrt(e_i/(1.0-e_i)));
                    else if (e_method != eMadaBoost)
                        sum += (w[k] *= sqrt((1.0-e_i)/e_i));
            }
            break;

        case Wagging:
            for (int j = 0; j < instances; j++)
                sum += (w[j] = Poisson());
            break;

        case eReweightBoost:
            if (e_i >= 0.5 || e_i <= ZERO)
                for (int k = 0; k < instances; k++)
                    sum += (w[k] = Poisson());
            else {
                scalar alpha = 0.5*log((1.0-e_i)/e_i);
                scalar factor = exp(-alpha);
                for (int i = 0; i < instances; i++) {
                    if (!ok[i])
                        w[i] *= factor;
                    sum += w[i];
                }
            }
            break;

        case eBagging:
        case RandomSubsampling:
            OnErrorExit(ERR_PARAM, "Ensemble method not valid");
            break;

        case NoBoost:
            break;

        default:
            OnErrorExit(ERR_PARAM, "Boosting method not valid");
    }

    // Normalize w[] to sum 1
    for (int k = 0; k < instances; k++)
        w[k] *= 1.0/sum;
}

// Calculates kappa as a measure of classifier performance
scalar PerformanceWithCache(int k, int *S, cache_t  **neighbors_cache, int *label, int instances, int classes, scalar **data, int inputs, bool knn_with_weighted_votes, scalar *r,
                            scalar *distance_to_furthest_neighbor, int *prediction, bool loo, t_PerformanceMeasure measure, int minority, int *subspace)
{
    scalar M[classes][classes];
    int nearest[k], predicted, majority;        // Confusion matrix
    scalar performance = 0.0, theta_1 = 0.0, theta_2 = 0.0, sp, sn;

    for (int i = 0; i < classes; i++)
        for (int j = 0; j <= i; j++)
            M[i][j] = M[j][i] = 0;

    // Create confusion matrix
    if (prediction)
        for (int i = 0; i < instances; i++)
            M[label[i]][prediction[i]]++;
    else
        for (int i = 0; i < instances; i++) {
            int backup;
            if (loo) {
                backup = S[i];
                S[i] = 0;
            }
            predicted = kNearestNeighbourCache(i, S, instances, nearest, label, k, neighbors_cache, data, inputs, classes, knn_with_weighted_votes, r, distance_to_furthest_neighbor, NULL, subspace);
            M[label[i]][predicted]++;
            if (loo)
                S[i] = backup;
        }

    switch (measure) {
        case pfAccuracy:
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            performance = theta_1/instances;
            break;

        case pfKappa:
            // Obtain kappa
            // Sum of the diagonal
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            theta_1 *= instances;

            for (int i = 0; i < classes; i++) {
                // Rows/columns total counts
                scalar row_i = 0, column_i = 0;

                for (int j = 0; j < classes; j++) {
                    row_i += M[i][j];
                    column_i += M[j][i];
                }

                theta_2 += row_i * column_i;
            }

            performance = (((scalar) instances)*instances-theta_2) > ZERO ? (theta_1 - theta_2)/(((scalar) instances)*instances-theta_2) : 0.0;
            break;

        case pfGMean:
            // Obtain G-mean
            majority = (minority+1) % 2;

            sp = (M[majority][majority]+M[majority][minority]) > ZERO ? ((scalar) M[majority][majority])/(M[majority][majority]+M[majority][minority]) : 0.0;
            sn = (M[minority][minority]+M[minority][majority]) > ZERO ? ((scalar) M[minority][minority])/(M[minority][minority]+M[minority][majority]) : 0.0;

            performance = sqrt(sn*sp);
            break;

        case pfAverageAccuracy:
        case pfMargin:
        case pfauPRC:
        case pfauROC:
        case pfRecallMacro:
            // Macro-averaged recall
            performance = 0.0;
            for (int i = 0; i < classes; i++) {
                scalar tp = M[i][i];
                scalar fn = -M[i][i];
                for (int j = 0; j < classes; j++)
                    fn += M[i][j];

                performance += (tp+fn > 0) ? tp/(tp+fn) : 0.0;
            }

            performance /= classes;
            break;

        case pfRecallMacroGMean:
        case pfCC:
        case pfF1Macro:
        case pfPrecisionMacro:
        case pfF1:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Performance metric with cache");
    }

    return performance;
}
