/***************************************************************************
 *   Copyright (C) 2014 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"

// Create neighbors and distances cache
// If neighbors cache is NULL only distance cache is constructed

void ConstructAllNeighborsCache(scalar **data, int instances, int inputs, scalar ***distance_cache, cache_t ***neighbors_cache, scalar *r, int **index, int *cache_size, int *subspace, bool loo)
{
    int size = cache_size ? *cache_size : 0;

    if (!neighbors_cache && !distance_cache)    // Nothing to do
        return;

    // Cache of distances?
    if (distance_cache != NULL) {
        // Need to resize?
        if (instances != size) {
            // Free
            for (int i = instances; i < size; i++)
                FREE((*distance_cache)[i]);

            // Resize
            REALLOC(*distance_cache, scalar *, instances, "main: distance_cache");

            // Resize
            for (int i = 0; i < instances && i < size; i++)
                REALLOC((*distance_cache)[i], scalar, instances, "main: distance_cache[i]");

            // Allocate new
            for (int i = size; i < instances; i++)
                MALLOC((*distance_cache)[i], scalar, instances, "main: distance_cache[i]");
        }
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < instances; j++) {
                scalar value = EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace);
                (*distance_cache)[i][j] = r ? value/r[j] : value;
            }
            if (loo) // Consider itself as neighbor?
                (*distance_cache)[i][i] = INFINITE;
        }
    }

    // Need to resize?
    if (instances != size) {
        // Free
        for (int i = instances; i < size; i++)
            FREE((*neighbors_cache)[i]);

        // Resize
        REALLOC(*neighbors_cache, cache_t *, instances, "main: neighbors_cache");

        // Resize
        for (int i = 0; i < instances && i < size; i++)
            REALLOC((*neighbors_cache)[i], cache_t, instances, "main: distance_cache[i]");

        // Allocate new
        for (int i = size; i < instances; i++)
            MALLOC((*neighbors_cache)[i], cache_t, instances, "main: distance_cache[i]");

        if (cache_size)
            *cache_size = instances;
    }

    SHOWPROGRESS(false, "Cache successfully reserved%c", '\n');

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < instances; i++) {
        RANK *rank;

        MALLOC(rank, RANK, instances, "main: r");

        for (int j = 0; j < instances; j++) {
            rank[j].fitness = (distance_cache && *distance_cache) ? -(*distance_cache)[i][j] :
                              -EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace)/(r ? r[j] : 1.0);
            rank[j].being = j;
        }
        rank[i].fitness = loo ? -INFINITE : 0.0; // Consider itself as neighbor?

        // Sort
        qsort(rank, instances, sizeof(RANK), Compare);

        // Store neighbors in ascending order of distance
        if (neighbors_cache)
            for (int j = 0; j < instances; j++)
                (*neighbors_cache)[i][j] = rank[j].being;

        FREE(rank);
    }
}

// Create k neighbors cache
void ConstructkNeighborsCache(scalar **data, int instances, int inputs, cache_t ***neighbors_cache, scalar *r, int **index, int k, int *cache_size)
{
    RANK *rank;

    MALLOC(rank, RANK, instances, "main: r");

    // Need to resize?
    if (instances != *cache_size) {
        // Free
        if (*neighbors_cache)
            FREE((*neighbors_cache)[0]);

        REALLOC((*neighbors_cache), cache_t *, instances, "ConstructkNeighborsCache: (*neighbors_cache)");
        MALLOC((*neighbors_cache)[0], cache_t, (k+1) * instances, "ConstructkNeighborsCache: (*neighbors_cache)[0]");
        for (int i = 1; i < instances; i++)
            (*neighbors_cache)[i] = (*neighbors_cache)[i-1] + k+1;

        *cache_size = instances;
    }

    for (int i = 0; i < instances; i++) {
        for (int j = 0; j < instances; j++) {
            rank[j].fitness = -EuclideanDistanceSparseData(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL)/(r ? r[j] : 1.0);
            rank[j].being = j;
        }

        // Sort
        qsort(rank, instances, sizeof(RANK), Compare);

        // Store neighbors in ascending order of distance
        for (int j = 0; j <= k; j++)
            (*neighbors_cache)[i][j] = rank[j].being;
    }

    FREE(rank);
}
