/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "defns.i"
#include "types.i"
#include "extern.i"
#include "classifiers.h"

// Consensus filter
// C. E. Brodley and M. A. Friedl, "Identifying mislabeled training data", Journal of Artificial Intelligence Research, vol. 11, pp. 131-167, 1999.

#define THRESHOLD 3     // Majority voting
// Consensus filter threshold = 5

int ConsensusFilter(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, int inputs, int classes, int **index, int *subspace)
{
    int Np, *s, *nearest, votes;
    struct svm_model *svm, *svm_rbf;
    t_Parameters local_parameters;

    MALLOC(s, int, instances, "ConsensusFilter: s");
    for (int i = 0; i < instances; i++)
        s[i] = 1;

    Np = instances;

    // Local parm to avoid modifying global ones
    local_parameters = *parm;

    // At least k = 3 to avoid two very similar classifiers
    local_parameters.k = MAX(parm->k, 3);

    if (parm->verbose > 0)
        fprintf(parm->out_file, "K for filter: %d\n",local_parameters.k);
    ALLOCA(nearest, int, local_parameters.k, "ConsensusFilter: nearest");

    // Train tree with all instances
    // Always prune data
    local_parameters.prune_tree = false;
    local_parameters.soften_thresholds = false;
    TRIALS = 1;

    t_DTree tree = DecisionTreeTrain(parm, data, instances, labels, inputs, NULL, subspace, classes, false, index, imInformationGain, false, false);

    // Train SVM with all instances
    // SVM's parm not subject to cross-validation
    local_parameters.svm_type = 0;
    local_parameters.tolerance = 0.1;
    local_parameters.cache_size = 10000;
    // Linear kernel
    local_parameters.kernel = 0;
    local_parameters.svm_c = 1.0;

    svm = TrainSupportVectorMachine(&local_parameters, data, subspace, instances, labels, inputs, NULL, index);

    local_parameters.kernel = 2;
    local_parameters.gamma = 1.0/inputs;
    svm_rbf =  TrainSupportVectorMachine(&local_parameters, data, subspace, instances, labels, inputs, NULL, index);

    // Remove misclassified instances by all three classifiers: C4.5, SVM and 1-NN
    for (int i = 0; i < instances; i++) {
        votes = 0;
        votes += (DecisionTreeClassify(tree, data[i], NULL, classes, index ? index[i] : NULL) != labels[i]);
        votes += (SupportVectorMachineOutput(svm, data[i], inputs, subspace, index ? index[i] : NULL, NULL) != labels[i]);
        votes += (kNNCategory(data[i], data, subspace, instances, 1, inputs, labels, classes, nearest, NULL, NULL, NULL, -1, index, index ? index[i] : NULL) != labels[i]);
        votes += (SupportVectorMachineOutput(svm_rbf, data[i], inputs, subspace, index ? index[i] : NULL, NULL) != labels[i]);
        votes += (kNNCategory(data[i], data, subspace, instances, local_parameters.k, inputs, labels, classes, nearest, NULL, NULL, NULL, -1, index, index ? index[i] : NULL) != labels[i]);
        if (votes >= THRESHOLD) {
            S[i] = 0;
            Np--;
        } else
            S[i] = 1;
    }

    if (parm->verbose > 0)
        fprintf(parm->out_file, "CF removed %d instances\n", instances - Np);

    FreeDecisionTree(tree);
    svm_free_and_destroy_model(&svm);
    svm_free_and_destroy_model(&svm_rbf);
    FREE(s);

    return Np;
}
