/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that deal with the individuals
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

scalar EvaluateCHCIndividual(const t_Parameters *parm, int *individual, scalar *ok, scalar *reduction, scalar **data, int instances, int *label, scalar *w, int minority, cache_t **neighbors_cache,
                             int inputs, int classes, bool evolve_local_k, scalar *r, int **index, scalar ok_w, scalar reduction_w, int *subspace, int test_instances)
{
    int selected = 0;
    scalar n_minus = 0.0, n_plus = 0.0;

    // Evaluate reduction
    // Count the prototypes on the individual
    int count[classes];
    for (int i = 0; i < classes; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        if (individual[i]) {
            selected++;
            count[label[i]]++;
        }

    // If reduction is 100% or only one class selected return 0 fitness
    int selected_clases = 0;
    for (int i = 0; i < classes; i++)
        if (count[i])
            selected_clases++;

    if (selected == 0 || selected_clases == 1) {
        *ok = *reduction = 0.0;
        return 0.0;
    }

    *reduction = 1.0 - ((scalar) selected)/instances;

    t_Learner learner;
    t_Ensemble ensemble;
    int **M;
    t_PerformanceInfo perf;

    // Test instances could be less than total instances
    int t_instances = test_instances ? test_instances : instances;

    AllocateIntegerMatrix(&M, classes, classes);

    // Accuracy or Kappa measure
    // Evaluate performance using a wrapper approach
    switch (parm->chc_wrapper) {
        case kNN:
            // Evaluate error
            *ok = 0.0;
            for (int i = 0; i < classes; i++)
                for (int j = 0; j <= i; j++)
                    M[i][j] = M[j][i] = 0.0;

            for (int i = 0; i < t_instances; i++) {
                // Evaluate using leave-one out
                int backup = -1, k = parm->k, category;

                if (parm->LOO_for_knn_performance) {
                    backup = individual[i];
                    individual[i] = 0;
                }

                // If evolving local k values, get k value of the nearest neighbor
                if (evolve_local_k) {
                    int l = 0;
                    // Get nearest neighbor k value
                    while (l < instances && !individual[neighbors_cache[i][l]])
                        l++;

                    k = (l == instances) ? 1 : individual[instances+neighbors_cache[i][l]];
                }

                int nearest[k];

                category = kNearestNeighbourCache(i, individual, instances, nearest, label, k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace);
                M[label[i]][category]++;

                if (parm->LOO_for_knn_performance)
                    individual[i] = backup;
            }
            break;


        default:
            ClassifierEvaluation(parm, data, t_instances, label, data, instances, label, subspace, inputs, individual, false, classes, parm->chc_wrapper, &learner, minority, M, NULL, &perf, NULL,
                                       NULL, NULL, NULL, NULL, index, index, &ensemble);
            break;
    }

    PrintMulticlassMeasures(M, classes, t_instances, NULL, NULL, &perf);
    if (classes == 2) {
        int majority = (minority+1) % 2;
        PrintTwoClassMeasures((scalar) M[minority][minority], (scalar) M[majority][majority], (scalar) M[majority][minority], (scalar) M[minority][majority], NULL, NULL, &perf);
    }

    switch (parm->performance_measure) {
        case pfAccuracy:
            *ok = perf.accuracy;
            break;

        case pfGMean:
            *ok = perf.gmean;
            break;

        case pfKappa:
            *ok = perf.kappa;
            break;

        case pfRecallMacro:
            *ok = perf.recall_macro;
            break;

        case pfRecallMacroGMean:
            *ok = perf.recall_macro_g_mean;
            break;

        case pfAverageAccuracy:
            *ok = perf.average_accuracy;
            break;

        case pfF1:
            *ok = perf.f1_score;
            break;

        case pfauROC:
        case pfauPRC:
        case pfMargin:
        case pfCC:
        case pfF1Macro:
        case pfPrecisionMacro:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Performance measure");
    }

    FreeIntegerMatrix(&M, classes);


    scalar fitness = 0.0, P = 0.2;
    if (parm->performance_measure == pfGMean && parm->ebus_gs_gm)
        // Fitness following S. García and F. Herrera, Evol. Comp., 2009.
        fitness = (n_minus > 0) ? (*ok - (fabs(1.0 - n_plus/n_minus) * P)) : *ok - P;
    else
        fitness = ok_w **ok+reduction_w **reduction;

    return fitness;
}

// This mutation is based on reduced nearest neighbor rule, but not identical
int RNNMutationCHC(const t_Parameters *parm, int *individual, int n, int instances, int *label, scalar *w, int minority, cache_t **neighbors_cache, scalar **data, int inputs, int classes,
                   scalar *evaluations, bool evolve_local_k, scalar *r, int *subspace)
{
    scalar all_bad, bad, individual_evaluations;
    int removed = 0, n_bad, n_ok, *list_bad = NULL, *list_ok;

    if (parm->performance_measure == pfGMean)
        return RNNMutationCHCUsingGMeasure(parm, individual, n, instances, label, w, minority, neighbors_cache, inputs, data, evolve_local_k, r);

    MALLOC(list_bad, int, instances, "RNNMutationCHC: list_bad");
    MALLOC(list_ok, int, instances, "RNNMutationCHC: list_ok");

    // Store bad and good instances to start classifying bad ones
    n_bad = n_ok = 0;

    // Test Ok when all present
    all_bad = 0.0;

    for (int i = 0; i < instances; i++) {
        int k = parm->k;
        // If evolving local k values, get k value of the nearest neighbor
        if (evolve_local_k) {
            int l = 0;
            // Get nearest neighbor k value
            while (l < instances && !individual[neighbors_cache[i][l]])
                l++;

            k = (l == instances) ? 1 : individual[instances+neighbors_cache[i][l]];
        }

        int nearest[k];

        if (kNearestNeighbourCache(i, individual, instances, nearest, label, k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, NULL, subspace) != label[i]) {
            all_bad += ((w == NULL) ? 1.0 : w[i]);
            list_bad[n_bad] = i;
            n_bad++;

        } else {
            list_ok[n_ok] = i;
            n_ok++;
        }
    }

    individual_evaluations = instances;

    // Test if removing each instance affects the classification of any other instance
    // Remove each instance and test

    // Modified 2/December/2010
    // Add the option of randomizing the order of checking selected instances
    int *order;
    MALLOC(order, int, instances, "RNNMutationCHC: order");
    for (int i = 0; i < instances; i++)
        order[i] = i;
    RandomizeVector(order, instances);

    for (int i = 0; i < instances; i++) {
        // Modified 9/September/2010
        // Consider decrement until reaching 0
        // Modified 13/September/2010
        // Undo previous change

        if (individual[order[i]]) {
            // Backup to undo if necessary
            int best = individual[order[i]];

            individual[order[i]] = 0;

            bad = 0.0;

            if (parm->k == 1 && !evolve_local_k) {
                int nearest[parm->k];

                for (int j = 0; j < n_bad && all_bad >= bad; j++)
                    if (kNearestNeighbourCache(list_bad[j], individual, instances, nearest, label, 1, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, NULL, subspace)
                        != label[list_bad[j]])
                        bad += ((w == NULL) ? 1.0 : w[list_bad[j]]);

                for (int j = 0; j < n_ok && all_bad >= bad; j++)
                    if (kNearestNeighbourCache(list_ok[j], individual, instances, nearest, label, 1, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, NULL, subspace)
                        != label[list_ok[j]])
                        bad += ((w == NULL) ? 1.0 : w[list_ok[j]]);

            } else {
                for (int j = 0; j < n_bad && all_bad >= bad; j++) {
                    // If evolving local k values, get k value of the nearest neighbor
                    int l, k = parm->k;
                    if (evolve_local_k) {
                        l = 0;
                        // Get nearest neighbor k value
                        while (l < instances && !individual[neighbors_cache[list_bad[j]][l]])
                            l++;

                        k = (l == instances) ? 1 : individual[instances+neighbors_cache[list_bad[j]][l]];
                    }

                    int nearest[k];

                    if (kNearestNeighbourCache(list_bad[j], individual, instances, nearest, label, k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, NULL, subspace)
                        != label[list_bad[j]])
                        bad += ((w == NULL) ? 1.0 : w[list_bad[j]]);
                }

                for (int j = 0; j < n_ok && all_bad >= bad; j++) {
                    // If evolving local k values, get k value of the nearest neighbor
                    int l, k = parm->k;
                    if (evolve_local_k) {
                        l = 0;
                        // Get nearest neighbor k value
                        while (l < instances && !individual[neighbors_cache[list_ok[j]][l]])
                            l++;

                        k = (l == instances) ? 1 : individual[instances+neighbors_cache[list_ok[j]][l]];
                    }

                    int nearest[k];

                    if (kNearestNeighbourCache(list_ok[j], individual, instances, nearest, label, k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, NULL, subspace)
                        != label[list_ok[j]])
                        bad += ((w == NULL) ? 1.0 : w[list_ok[j]]);
                }
            }

            // If worse undo
            if (all_bad < bad)
                individual[order[i]] = best;
            else
                removed ++;

            individual_evaluations++;

        }
    }

    if (parm->verbose > 3)
        fprintf(parm->out_file, "Individual %d: RNN removed %3d instances\n", n, removed);

    *evaluations += individual_evaluations/instances;

    FREE(list_bad);
    FREE(list_ok);
    FREE(order);

    return removed;
}

// This mutation is based on reduced nearest neighbor rule, but not identical
int RNNMutationCHCUsingGMeasure(const t_Parameters *parm, int *individual, int n, int instances, int *label, scalar *w, int minority, cache_t **neighbors_cache, int inputs, scalar **data,
                                bool evolve_local_k, scalar *r)
{
    int removed = 0;

    OnErrorExit(ERR_UNSPECIFIED, "Need updating");

    // Store G value with all selected instances
    int *S;
    MALLOC(S, int, (instances * (evolve_local_k ? 2 : 1)), "RNNMutationCHCUsingGMeasure: S");
    for (int i = 0; i < instances; i++)
        S[i] = 1;

    // Copy k values?
    if (evolve_local_k)
        for (int i = instances; i < 2*instances; i++)
            S[i] = individual[i];

    scalar G_with_all = 0.0;

    FREE(S);

    // Test if removing each instance affects the classification of any other instance
    // Remove each instance and test
    // Modified 2/December/2010
    // Add the option of randomizing the order of checking selected instances
    int *order, selected = 0;
    ALLOCA(order, int, instances, "RNNMutationCHCUsingGMeasure: order");
    for (int i = 0; i < instances; i++)
        if (individual[i]) {
            order[selected] = i;
            selected++;
        }
    RandomizeVector(order, selected);

    for (int i = 0; i < selected; i++) {
        // Modified 25/October/2010
        // Added a different version for G measure
        // Backup to undo if necessary
        int backup = individual[order[i]];

        individual[order[i]] = 0;

        // Evaluate G without the instance
        scalar G = 0.0;

        // If worse undo
        if (G < G_with_all)
            individual[order[i]] = backup;
        else
            removed++;
    }

    if (parm->verbose > 3)
        fprintf(parm->out_file, "Individual %d: RNN removed %3d instances\n", n, removed);

    return removed;
}
