/***************************************************************************
 *   Copyright (C) 2012 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "is.h"
#include "macros.h"
#include "maths.h"

// Randomly subsample data taking into account minority class
int ClassImbalanceRandomSubsampling(int instances, int *labels, scalar ratio, int minority, int count[2], int *list)
{
    int n, *M_list, *m_list, M_count, m_count;

    n = instances*ratio;

    // List of minority and majority samples
    MALLOC(M_list, int, count[(minority+1)%2], "ClassImbalanceRandomSubsampling: m_list");
    MALLOC(m_list, int, count[minority], "ClassImbalanceRandomSubsampling: m_list");

    M_count = m_count = 0;
    for (int i = 0; i < instances; i++)
        if (labels[i] == minority) {
            m_list[m_count] = i;
            m_count++;
        } else {
            M_list[M_count] = i;
            M_count++;
        }

    RandomizeVector(M_list, M_count);
    RandomizeVector(m_list, m_count);

    // Test whether the sampling can be balanced
    if (2*count[minority] >= n) {
        // Balanced sampling
        if (n % 2)
            n++;

        for (int i = 0; i < n; i += 2) {
            list[i] = M_list[i/2];
            list[i+1] = m_list[i/2];
        }
    } else {
        // Imbalanced sampling
        // Select all minority class samples
        for (int i = 0; i < m_count; i++)
            list[i] = m_list[i];

        for (int i = m_count; i < n; i++)
            list[i] = M_list[i - m_count];

        // Avoid having all the minority class instances at the beginning
        RandomizeVector(list, n);
    }

    FREE(M_list);
    FREE(m_list);

    return n;
}


// Subsample data taking into account minority class
// Roulette selection is applied to both classes
int ClassImbalanceSubsampling(int instances, int *labels, scalar ratio, int minority, int count[2], int *list, scalar *w)
{
    int n, *M_list, *m_list, M_count, m_count;
    scalar *M_probs, *m_probs;

    n = instances*ratio;

    // List of minority and majority samples and probabilities for roulette sampling
    MALLOC(M_list, int, count[(minority+1)%2], "ClassImbalanceSubsampling: m_list");
    MALLOC(m_list, int, count[minority], "ClassImbalanceSubsampling: m_list");
    MALLOC(M_probs, scalar, count[(minority+1)%2], "ClassImbalanceSubsampling: M_probs");
    MALLOC(m_probs, scalar, count[minority], "ClassImbalanceSubsampling: m_probs");

    M_count = m_count = 0;
    for (int i = 0; i < instances; i++) {
        if (labels[i] == minority) {
            m_list[m_count] = i;
            if (m_count == 0)
                m_probs[0] = w[i];
            else
                m_probs[m_count] = m_probs[m_count-1]+w[i];
            m_count++;
        } else {
            M_list[M_count] = i;
            if (M_count == 0)
                M_probs[0] = w[i];
            else
                M_probs[M_count] = M_probs[M_count-1]+w[i];
            M_count++;
        }

        list[i] = 0;
    }

    // Balanced roulette sampling
    // Minority class
    for (int i = 0; i < n/2; i++)
        list[m_list[(int) RouletteSelection(m_probs, m_count)]] = 1;

    // Majority class
    for (int i = n/2; i < n; i++)
        list[M_list[(int) RouletteSelection(M_probs, M_count)]] = 1;

    // Avoid having all the minority class instances at the beginning
    RandomizeVector(list, n);

    FREE(M_list);
    FREE(m_list);
    FREE(M_probs);
    FREE(m_probs);

    return n;
}
