/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <string.h>
#include <pvm3.h>
#include <unistd.h>
#include "pvm-local.h"
#include "is.h"
#include "common.h"
#include "macros.h"

// Threshold for minority class
// #define MINORITY_THRESHOLD(x, M_ratio, max_votes)    (M_ratio*(x)*n_subsets*minority_sample_probability)
// #define MINORITY_THRESHOLD(x, M_ratio, max_votes)    (max_votes - 1)
// #define MINORITY_THRESHOLD(x, M_ratio, max_votes)    (max_votes*(((scalar) x)/parm->n_electors))

// Functions that deal with thresholds for class-imbalance datasets
int SelectInstancesCIUsingThreshold(const t_Parameters *parm, int *votes, int instances, int *S, int *list, int minority_threshold, int majority_threshold, int *labels, int minority,
                                    int *sel_minority,
                                    int *sel_majority)
{
    int sel_instances = 0;

    *sel_minority = *sel_majority = 0;

    for (int i = 0; i < instances; i++)
        if ((labels[i] == minority && votes[i] >= minority_threshold) ||
            (labels[i] != minority && votes[i] >= majority_threshold)) {
            list[sel_instances] = i;
            sel_instances++;

            if (labels[i] == minority)
                (*sel_minority) ++;
            else
                (*sel_majority) ++;
        }

    // Force balanced classes
    if (*sel_majority > *sel_minority && (parm->force_balanced || parm->force_undersampled)) {
        // Undersample majority class
        // Sort instances using votes
        t_VoteRank *M_list;

        MALLOC(M_list, t_VoteRank, *sel_majority, "SelectInstancesCIUsingThreshold: M_list");
        int c = 0;
        for (int i = 0; i < sel_instances; i++)
            if (labels[list[i]] != minority) {
                M_list[c].instance = list[i];
                M_list[c].votes = votes[list[i]];
                c++;
            }

        if (c != *sel_majority)
            OnErrorExit(1, "Something wrong in selected instances");

        qsort(M_list, c, sizeof(t_VoteRank), CompareVotes);

        // Select minority and the most voted among the majority
        c = 0;
        for (int i = 0; i < sel_instances; i++)
            if (labels[list[i]] == minority) {
                list[c] = list[i];
                c++;
            }

        if (c != *sel_minority)
            OnErrorExit(1, "Something wrong in selected instances");

        *sel_majority = c;
        for (int i = 0; i < c; i++)
            list[i+c] = M_list[i].instance;
        RandomizeVector(list, 2*c);

        FREE(M_list);
    }

    if (*sel_minority > *sel_majority && parm->force_balanced) {
        // Undersample minority class
        // Sort instances using votes
        t_VoteRank *m_list;

        MALLOC(m_list, t_VoteRank, *sel_minority, "SelectInstancesCIUsingThreshold: m_list");
        int c = 0;
        for (int i = 0; i < sel_instances; i++)
            if (labels[list[i]] == minority) {
                m_list[c].instance = list[i];
                m_list[c].votes = votes[list[i]];
                c++;
            }

        if (c != *sel_minority)
            OnErrorExit(1, "Something wrong in selected instances");

        qsort(m_list, c, sizeof(t_VoteRank), CompareVotes);

        // Select majority and the most voted among the minority
        c = 0;
        for (int i = 0; i < sel_instances; i++)
            if (labels[list[i]] != minority) {
                list[c] = list[i];
                c++;
            }

        if (c != *sel_majority)
            OnErrorExit(1, "Something wrong in selected instances");

        *sel_minority = c;
        for (int i = 0; i < c; i++)
            list[i+c] = m_list[i].instance;
        RandomizeVector(list, 2*c);

        FREE(m_list);
    }

    // If S non NULL store selection
    if (S != NULL) {
        for (int i = 0; i < instances; i++)
            S[i] = 0;

        for (int i = 0; i < sel_instances; i++)
            S[list[i]]++;
    }

    return sel_instances;
}

int GetThresholdForImbalancedClass(const t_Parameters *parm, scalar **data, int instances, int *labels, int minority, int classtoget, int *votes, int other_threshold, scalar *best_coefficient,
                                   int inputs,
                                   int classes)
{
    int best_threshold = 0, *list, s_min, s_maj, *local_votes, *local_labels, *partition;
    scalar **local_data, *G_mean, *r, *coef, *round_g_mean, *round_r, *round_coef;

    if (parm->votes_evaluator != kNN)
        OnErrorExit(1, "Not yet implemented");

    // Get maximum and minimum number of votes
    int min_threshold = 1000000, max_threshold = -1000000;

    for (int i = 0; i < instances; i++)
        if (labels[i] == classtoget) {
            if (min_threshold > votes[i])
                min_threshold = votes[i];
            else if (max_threshold < votes[i])
                max_threshold = votes[i];
        }
    int n_thresholds = max_threshold - min_threshold+1;

    CALLOC(r, scalar, n_thresholds, "GetThresholdForImbalancedClass: r");
    CALLOC(coef, scalar, n_thresholds, "GetThresholdForImbalancedClass: coef");
    CALLOC(G_mean, scalar, n_thresholds, "GetThresholdForImbalancedClass: G_mean");

    r -= min_threshold;
    coef -= min_threshold;
    G_mean -= min_threshold;

    MALLOC(round_r, scalar, n_thresholds, "GetThresholdForImbalancedClass: round_r");
    MALLOC(round_coef, scalar, n_thresholds, "GetThresholdForImbalancedClass: round_coef");
    MALLOC(round_g_mean, scalar, n_thresholds, "GetThresholdForImbalancedClass: round_g_mean");

    round_r -= min_threshold;
    round_coef -= min_threshold;
    round_g_mean -= min_threshold;

    // If no selection of minority return
    if (classtoget == minority && !parm->selection_minority)
        return min_threshold;

    MALLOC(list, int, 2*parm->subset_size, "GetThresholdForImbalancedClass: list");
    MALLOC(partition, int, instances, "GetThresholdForImbalancedClass: partition");
    MALLOC(local_data, scalar *, 2*parm->subset_size, "GetThresholdForImbalancedClass: local_data");
    MALLOC(local_labels, int, 2*parm->subset_size, "GetThresholdForImbalancedClass: local_labels");
    MALLOC(local_votes, int, 2*parm->subset_size, "GetThresholdForImbalancedClass: local_votes");

    int n_subsets = instances/parm->subset_size;
    // Add remainder
    if (instances % parm->subset_size)
        n_subsets++;

    if (n_subsets == 1)
        n_subsets++;

    for (int round = 0; round < parm->demo_eval_rounds; round++) {
        // New random partition
        int count[n_subsets];
        RandomPartition(partition, count, instances, n_subsets, labels, classes);

        // Evaluate for every subset
        for (int i = min_threshold; i <= max_threshold; i++)
            round_r[i] = round_g_mean[i] = round_coef[i] = 0.0;
        for (int subset = 0; subset < n_subsets; subset++) {
            int nearest[parm->k];

            // Get subset
            int cursor = 0;
            for (int j = 0; j < instances; j++)
                if (partition[j] == subset) {
                    local_data[cursor] = data[j];
                    local_labels[cursor] = labels[j];
                    local_votes[cursor] = votes[j];
                    cursor++;
                }

            // Get coefficient for every threshold
            for (int th = min_threshold; th <= max_threshold; th++) {
                scalar TN = 0.0, TP = 0.0, FN = 0.0, FP = 0.0;
                int selected;

                if (classtoget == minority)
                    selected = SelectInstancesCIUsingThreshold(parm, local_votes, cursor, NULL, list, th, other_threshold, local_labels, minority, &s_min, &s_maj);
                else
                    selected = SelectInstancesCIUsingThreshold(parm, local_votes, cursor, NULL, list, other_threshold, th, local_labels, minority, &s_min, &s_maj);

                // Evaluate
                scalar RED = 1.0 - ((scalar) selected)/cursor;

                // Leave-one-out error
                int next = 0;
                for (int i = 0; i < cursor; i++) {
                    // If selected exchange with last
                    bool restore = false;
                    if (list[next] == i) {
                        int swap = list[next];
                        list[next] = list[selected-1];
                        list[selected-1] = swap;
                        selected--;
                        restore = true;
                    }

                    bool ok = (kNearestNeighbourWithList(local_data[i], list, local_data, selected, inputs, parm->k, nearest, local_labels, classes, parm->knn_with_weighted_votes, NULL, NULL) ==
                               local_labels[i]);

                    if (ok) {
                        if (local_labels[i] == minority)
                            TP++;
                        else
                            TN++;
                    } else {
                        if (local_labels[i] == minority)
                            FN++;
                        else
                            FP++;
                    }

                    if (restore) {
                        selected++;
                        int swap = list[next];
                        list[next] = list[selected-1];
                        list[selected-1] = swap;
                        next++;
                    }
                }

                round_r[th] += RED;

                // Evaluate using k-fold cross-validation and 2-class imbalanced measures
                scalar specificity = (TN+FP) > ZERO ? TN/(TN+FP) : 0.0;
                scalar sensitivity = (TP+FN) > ZERO ? TP/(TP+FN) : 0.0;

                scalar G = pow(specificity*sensitivity, 0.5);

                round_g_mean[th] += G;

                round_coef[th] += parm->alpha_acc * G+(1.0 - parm->alpha_acc) * RED;
            }
        }

        // Average
        for (int th = min_threshold; th <= max_threshold; th++) {
            r[th] += round_r[th]/n_subsets;
            coef[th] += round_coef[th]/n_subsets;
            G_mean[th] += round_g_mean[th]/n_subsets;
        }
    }

    for (int th = min_threshold; th <= max_threshold; th++) {
        r[th] /= parm->demo_eval_rounds;
        coef[th] /= parm->demo_eval_rounds;
        G_mean[th] /= parm->demo_eval_rounds;
    }

    // Get best error
    best_threshold = min_threshold;
    for (int i = min_threshold+1; i <= max_threshold; i++)
        if (coef[i] > coef[best_threshold])
            best_threshold = i;

    *best_coefficient = coef[best_threshold];

    if (parm->verbose >= 1) {
        fprintf(parm->out_file, "Training info (%s threshold)\nVotes\tG\tReduct\tCoefficient\n", classtoget == minority ? "minority" : "majority");
        for (int i = min_threshold; i <= max_threshold; i++)
            fprintf(parm->out_file, "%3d:\t%6.4lf\t%6.4lf\t%6.4lf %s\n", i, G_mean[i], r[i], coef[i], i == best_threshold ? "(*)" : "");
    }

    r += min_threshold;
    coef += min_threshold;
    G_mean += min_threshold;
    round_r += min_threshold;
    round_coef += min_threshold;
    round_g_mean += min_threshold;

    FREE(r);
    FREE(coef);
    FREE(G_mean);
    FREE(round_r);
    FREE(round_coef);
    FREE(round_g_mean);
    FREE(local_data);
    FREE(local_labels);
    FREE(list);
    FREE(partition);
    FREE(local_votes);

    return best_threshold;
}

#define USE_CACHE

void GetBothThresholdsForImbalancedClass(const t_Parameters *parm, scalar **data, int instances, int *labels, int minority, int *votes, int *minority_th, int *majority_th, scalar *best_coefficient,
        int inputs, int classes, scalar *rr)
{
    int *list, s_min, s_maj, *local_votes, *local_labels, *partition, size;
    scalar **local_data, **G_mean, **r, **coef, **round_g_mean, **round_r, **round_coef;
#ifdef USE_CACHE
    scalar **distances = NULL;
    int max_cursor = 0;
#endif

    // Get maximum and minimum number of votes
    int min_min_threshold = 1000000, min_max_threshold = -1000000, maj_min_threshold = 1000000, maj_max_threshold = -1000000;

    for (int i = 0; i < instances; i++)
        if (labels[i] == minority) {
            if (min_min_threshold > votes[i])
                min_min_threshold = votes[i];
            else if (min_max_threshold < votes[i])
                min_max_threshold = votes[i];
        } else {
            if (maj_min_threshold > votes[i])
                maj_min_threshold = votes[i];
            else if (maj_max_threshold < votes[i])
                maj_max_threshold = votes[i];
        }

    int min_n_thresholds = min_max_threshold - min_min_threshold+1;
    int maj_n_thresholds = maj_max_threshold - maj_min_threshold+1;

    ALLOCA(r, scalar *, min_n_thresholds, "GetBothThresholdsForImbalancedClass: r");
    ALLOCA(coef, scalar *, min_n_thresholds, "GetBothThresholdsForImbalancedClass: coef");
    ALLOCA(G_mean, scalar *, min_n_thresholds, "GetBothThresholdsForImbalancedClass: G_mean");
    ALLOCA(round_r, scalar *, min_n_thresholds, "GetBothThresholdsForImbalancedClass: round_r");
    ALLOCA(round_coef, scalar *, min_n_thresholds, "GetBothThresholdsForImbalancedClass: round_coef");
    ALLOCA(round_g_mean, scalar *, min_n_thresholds, "GetBothThresholdsForImbalancedClass: round_g_mean");

    r -= min_min_threshold;
    coef -= min_min_threshold;
    G_mean -= min_min_threshold;
    round_r -= min_min_threshold;
    round_coef -= min_min_threshold;
    round_g_mean -= min_min_threshold;

    for (int i = min_min_threshold; i <= min_max_threshold; i++) {
        ALLOCA(r[i], scalar, maj_n_thresholds, "GetBothThresholdsForImbalancedClass: r[i]");
        ALLOCA(coef[i], scalar, maj_n_thresholds, "GetBothThresholdsForImbalancedClass: coef[i]");
        ALLOCA(G_mean[i], scalar, maj_n_thresholds, "GetBothThresholdsForImbalancedClass: G_mean[i]");
        ALLOCA(round_r[i], scalar, maj_n_thresholds, "GetBothThresholdsForImbalancedClass: round_r[i]");
        ALLOCA(round_coef[i], scalar, maj_n_thresholds, "GetBothThresholdsForImbalancedClass: round_coef[i]");
        ALLOCA(round_g_mean[i], scalar, maj_n_thresholds, "GetBothThresholdsForImbalancedClass: round_g_mean[i]");

        r[i] -= maj_min_threshold;
        coef[i] -= maj_min_threshold;
        G_mean[i] -= maj_min_threshold;
        round_r[i] -= maj_min_threshold;
        round_coef[i] -= maj_min_threshold;
        round_g_mean[i] -= maj_min_threshold;

        for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
            r[i][j] = G_mean[i][j] = coef[i][j] = 0.0;
    }


    ALLOCA(list, int, 2*parm->subset_size, "GetBothThresholdsForImbalancedClass: list");
    MALLOC(partition, int, instances, "GetBothThresholdsForImbalancedClass: partition");
    ALLOCA(local_data, scalar *, 2*parm->subset_size, "GetBothThresholdsForImbalancedClass: local_data");
    ALLOCA(local_labels, int, 2*parm->subset_size, "GetBothThresholdsForImbalancedClass: local_labels");
    ALLOCA(local_votes, int, 2*parm->subset_size, "GetBothThresholdsForImbalancedClass: local_votes");

    int n_subsets = instances/parm->subset_size;
    // Add remainder
    if (instances % parm->subset_size)
        n_subsets++;

    if (n_subsets == 1)
        n_subsets++;

    for (int round = 0; round < parm->demo_eval_rounds; round++) {
        // New random partition
        int count[n_subsets];
        RandomPartition(partition, count, instances, n_subsets, labels, classes);

        // Evaluate for every subset
        for (int i = min_min_threshold; i <= min_max_threshold; i++)
            for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
                round_r[i][j] = round_g_mean[i][j] = round_coef[i][j] = 0.0;

        for (int subset = 0; subset < n_subsets; subset++) {
            int nearest[parm->k];

            // Get subset
            int cursor = 0;
            for (int j = 0; j < instances; j++)
                if (partition[j] == subset) {
                    local_data[cursor] = data[j];
                    local_labels[cursor] = labels[j];
                    local_votes[cursor] = votes[j];
                    cursor++;
                }
#ifdef USE_CACHE
            // Create cache
            if (cursor > max_cursor) {
                REALLOC(distances, scalar *, cursor, "GetBothThresholdsForImbalancedClass: distances");
                for (int i = 0; i < max_cursor; i++) {
                    FREE(distances[i]);
                    MALLOC(distances[i], scalar, cursor, "GetBothThresholdsForImbalancedClass: distances[i]");
                }
                for (int i = max_cursor; i < cursor; i++)
                    MALLOC(distances[i], scalar, cursor, "GetBothThresholdsForImbalancedClass: distances[i]");

                max_cursor = cursor;
            }

            for (int i = 0; i < cursor; i++) {
                distances[i][i] = 0.0;
                for (int j = 0; j < i; j++) {
                    scalar value = EuclideanDistance(local_data[i], local_data[j], inputs);
                    if (rr == NULL)
                        distances[i][j] = distances[j][i] = value;
                    else {
                        distances[i][j] = value/rr[j];
                        distances[j][i] = value/rr[i];
                    }
                }
            }
#endif

            // Get coefficient for every threshold
            for (int min_th = min_min_threshold; min_th <= min_max_threshold; min_th++)
                for (int maj_th = maj_min_threshold; maj_th <= maj_max_threshold; maj_th++) {
                    scalar TN = 0.0, TP = 0.0, FN = 0.0, FP = 0.0, G = 0.0;
                    int selected;

                    selected = SelectInstancesCIUsingThreshold(parm, local_votes, cursor, NULL, list, min_th, maj_th, local_labels, minority, &s_min, &s_maj);

                    // Evaluate
                    scalar RED = 1.0 - ((scalar) selected)/cursor;

                    if (!selected)
                        G = 0.0;
                    else if (parm->evaluate_oligois_votes_with_auc)
                        // Obtain AUC
                        G = ObtainAUROCForClassifier(parm, parm->votes_evaluator, local_data, selected, local_labels, list, local_data, cursor, local_labels, inputs, minority, false, NULL, NULL,
                                                     NULL, parm->k, false, parm->out_file, NULL, NULL, classes);
                    else {
                        int next;
                        switch (parm->votes_evaluator) {
                            case kNN:
                                // Leave-one-out error
                                next = 0;
                                for (int i = 0; i < cursor; i++) {
                                    // If selected exchange with last
                                    bool restore = false;
                                    if (list[next] == i) {
                                        int swap = list[next];
                                        list[next] = list[selected-1];
                                        list[selected-1] = swap;
                                        selected--;
                                        restore = true;
                                    }
#ifdef USE_CACHE
                                    bool ok = (kNearestNeighbourWithListCached(i, list, selected, inputs, parm->k, nearest, local_labels, distances, classes, parm->knn_with_weighted_votes) == local_labels[i]);
#else
                                    bool ok = (kNearestNeighbourWithList(local_data[i], list, local_data, selected, inputs, parm->k, nearest, local_labels, parm->knn_with_weighted_votes, NULL, NULL) ==
                                               local_labels[i]);
#endif

                                    if (ok) {
                                        if (local_labels[i] == minority)
                                            TP++;
                                        else
                                            TN++;
                                    } else {
                                        if (local_labels[i] == minority)
                                            FN++;
                                        else
                                            FP++;
                                    }

                                    if (restore) {
                                        selected++;
                                        int swap = list[next];
                                        list[next] = list[selected-1];
                                        list[selected-1] = swap;
                                        next++;
                                    }
                                }

                                // Evaluate using k-fold cross-validation and 2-class imbalanced measures
                                scalar specificity = (TN+FP) > ZERO ? TN/(TN+FP) : 0.0;
                                scalar sensitivity = (TP+FN) > ZERO ? TP/(TP+FN) : 0.0;

                                G = pow(specificity*sensitivity, 0.5);
                                break;

                            case SVM:
                            case C45:
                                G = EvaluateWithListG(parm, local_data, selected, local_labels, list, local_data, selected, local_labels, NULL, &size, minority, &TN, &TP, &FN, &FP, inputs,
                                                      classes, parm->votes_evaluator, parm->k, NULL, NULL);
                                break;
                        }
                    }
                    round_r[min_th][maj_th] += RED;
                    round_g_mean[min_th][maj_th] += G;
                    round_coef[min_th][maj_th] += parm->alpha_acc * G+(1.0 - parm->alpha_acc) * RED;

                }
        }
        // Average
        for (int min_th = min_min_threshold; min_th <= min_max_threshold; min_th++)
            for (int maj_th = maj_min_threshold; maj_th <= maj_max_threshold; maj_th++) {
                r[min_th][maj_th] += round_r[min_th][maj_th]/n_subsets;
                coef[min_th][maj_th] += round_coef[min_th][maj_th]/n_subsets;
                G_mean[min_th][maj_th] += round_g_mean[min_th][maj_th]/n_subsets;
            }

    }

    // Get best error
    *minority_th = min_min_threshold;
    *majority_th = maj_min_threshold;
    for (int i = min_min_threshold; i <= min_max_threshold; i++)
        for (int j = maj_min_threshold; j <= maj_max_threshold; j++) {
            // First average
            r[i][j] /= parm->demo_eval_rounds;
            coef[i][j] /= parm->demo_eval_rounds;
            G_mean[i][j] /= parm->demo_eval_rounds;

            if (coef[i][j] > coef[*minority_th][*majority_th]) {
                *minority_th = i;
                *majority_th = j;
            }
        }

    *best_coefficient = coef[*minority_th][*majority_th];

    if (parm->verbose >= 1) {
        fprintf(parm->out_file, "THRESHOLDS TRAINING INFO\nMinority\tMajority\tG\tReduct\tCoefficient\n");
        for (int i = min_min_threshold; i <= min_max_threshold; i++)
            for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
                fprintf(parm->out_file, "%3d\t\t%3d\t\t%6.4lf\t%6.4lf\t%6.4lf %s\n", i, j, G_mean[i][j], r[i][j], coef[i][j], (i == *minority_th && j == *majority_th) ? "(*)" : "");
    }

#ifdef USE_CACHE
    for (int i = 0; i < max_cursor; i++)
        FREE(distances[i]);
    FREE(distances);
#endif

    FREE(partition);
}

// Auxiliary function for qsort: from worst to best weight
int CompareVotes(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_VoteRank *) y)->votes - ((t_VoteRank *) x)->votes) > 0)
        return +1;
    else {
        if (diff > 0)
            return -1;
        else
            return 0;
    }
}
