/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

// FUNCTIONS RELATED TO CONDENSED NEAREST NEIGHBOR RULE

// Condensed nearest neighbor rule [Hart, 68]
// For class-imbalanced datasets all minority class samples are kept
int CondensedNearestNeighborRule(int *S, scalar **data, int instances, int *label, int inputs, bool ci, int minority, cache_t **neighbors_cache, bool *consider_list, int k, int classes, int **index,
                                 int *subspace)
{
    int nearest, selected = 0, count[] = {0, 0};

    if (k != 1)
        OnErrorExit(ERR_PARAM, "CNN is for k = 1");

    // Remove all instances
    for (int i = 0; i < instances; i++)
        S[i] = 0;

    // One random instance...
    if (!ci) {
        S[(int) RANDOM(0, instances)] = 1;
        selected = 1;
        // ... or, for class-imbalanced datasets, one random majority class instance and all minority class instances
    } else {
        for (int i = 0; i < instances; i++) {
            count[label[i]]++;

            if (label[i] == minority) {
                S[i] = 1;
                selected++;
            }
        }

        // If no majority class instances do nothing
        if (!count[(minority+1) %2])
            return selected;

        int r;
        do {
            r = RANDOM(0, instances);
        }  while (label[r] == minority);
        S[r] = 1;
        selected++;
    }

    // Boosting
    if (consider_list != NULL) {
        int *list, n = 0;
        MALLOC(list, int, instances, "CondensedNearestNeighborRule: list");

        for (int i = 0; i < instances; i++)
            if (consider_list[i]) {
                list[n] = i;
                n++;
            }
        RandomizeVector(list, n);

        for (int i = 0; i < n; i++)
            if (!S[list[i]]) {
                if (kNearestNeighbourCache(list[i], S, instances, &nearest, label, 1, neighbors_cache, data, inputs, classes, false, NULL, NULL, index, subspace) != label[list[i]]) {
                    S[list[i]] = 1;
                    selected++;
                }
            }

        FREE(list);
    } else {
        // Check all instances and add instances miss-classified
        for (int i = 0; i < instances; i++)
            if (!S[i] && (consider_list == NULL || consider_list[i])) {
                // Check classification
                if (kNearestNeighbourCache(i, S, instances, &nearest, label, 1, neighbors_cache, data, inputs, classes, false, NULL, NULL, index, subspace) != label[i]) {
                    S[i] = 1;
                    selected++;
                }
            }
    }

    return selected;
}

// Following: G. E. A. P. A. Batista and R. C. Prati and M. C. Monard, "A study of the behavior of several methods for balancing machine learning training
// data", SIGKDD Explorations, 6(1), 20-29, 2004.

int CNNAndTomekLinks(int *S, scalar **data, int instances, int *label, int inputs, int minority, int k, int classes, int **index)
{
    int *origin, *dest, n_links;

    // First: CNN rule
    int selected = CondensedNearestNeighborRule(S, data, instances, label, inputs, true, minority, NULL, NULL, k, classes, index, NULL);

    // Second: Obtain Tomek's links
    MALLOC(origin, int, 2*selected, "CNNAndTomekLinks: origin");
    MALLOC(dest, int, 2*selected, "CNNAndTomekLinks: dest");
    GetTomekLinks(S, data, instances, label, inputs, origin, dest, &n_links, classes, index, NULL);

    // From every link remove majority class sample
    for (int i = 0; i < n_links; i++) {
        if (label[origin[i]] == minority) {
            // Maybe already removed
            if (S[dest[i]]) {
                S[dest[i]] = 0;
                selected--;
            }
        } else {
            // Maybe already removed
            if (S[origin[i]]) {
                S[origin[i]] = 0;
                selected--;
            }
        }
    }

    FREE(origin);
    FREE(dest);

    return selected;
}

// Following: M. Kubat and S. Matwin (1997). "Addressing the course of imbalanced training sets: One-sided selection", in Proceedings of the International
// Conference on Machine Learning, ICML, pp. 179–186.

int OneSidedSelection(int *S, scalar **data, int instances, int *label, int inputs, int minority, int k, int classes, int **index, cache_t **neighbors_cache, bool as_mutation)
{
    int *origin, *dest, n_links, selected;

    if (classes != 2)
        OnErrorExit(ERR_PARAM, "OSS only available for 2-class problems");

    if (!as_mutation)
        for (int i = 0; i < instances; i++)
            S[i] = 1;

    // First: Obtain Tomek's links
    MALLOC(origin, int, 2*instances, "OneSidedSelection: origin");
    MALLOC(dest, int, 2*instances, "OneSidedSelection: dest");
    GetTomekLinks(S, data, instances, label, inputs, origin, dest, &n_links, classes, index, neighbors_cache);

    selected = instances;

    // From every link remove majority class sample
    for (int i = 0; i < n_links; i++) {
        if (label[origin[i]] == minority) {
            // Maybe already removed
            if (S[dest[i]]) {
                S[dest[i]] = 0;
                selected--;
            }
        } else {
            // Maybe already removed
            if (S[origin[i]]) {
                S[origin[i]] = 0;
                selected--;
            }
        }
    }

    FREE(origin);
    FREE(dest);

    // Second: CNN rule
    bool *consider_list;
    MALLOC(consider_list, bool, instances, "OneSidedSelection: consider_list");
    for (int i = 0; i < instances; i++)
        consider_list[i] = S[i];

    selected = CondensedNearestNeighborRule(S, data, instances, label, inputs, true, minority, neighbors_cache, consider_list, k, classes, index, NULL);

    FREE(consider_list);

    return selected;
}

// Obtain all Tomek's links
// Each link is stored in pairs i,i+1 in the vector "links"
// Links may be duplicated

void GetTomekLinks(int *S, scalar **data, int instances, int *label, int inputs, int *origin, int *dest, int *n_links, int classes, int **index, cache_t **neighbors_cache)
{
    int nearest;

    *n_links = 0;

    // Get nearest neighbor of each instance
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            S[i] = 0;
            kNNCategory(data[i], data, NULL, instances, 1, inputs, label, classes, &nearest, S, NULL, neighbors_cache, i, index, index ? index[i] : NULL);
            S[i] = 1;

            // If different classes, they form a Tomek's link
            if (label[i] != label[nearest]) {
                origin[*n_links] = i;
                dest[*n_links] = nearest;
                (*n_links) ++;
            }
        }
}
