/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es   *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <pvm3.h>
#include <unistd.h>
#include <string.h>
#include "macros.h"
#include "is.h"

// Data partitioning strategy following: Dara et al., "Filter-based data partitioning for training multiple classifier systems", IEEE TKDE, vol. 22, no. 4, pp. 508-522, 2010.

void DecoupledDataPartitioning(int instances, int *labels, int size, int *partition, scalar r, int n_classes, int *n_subsets, int **count, bool *first_time)
{
    static int *decoupled_partition;
    int list[instances];

    // If first time obtain a decoupled partition, where all subsets contain instances from just one class
    if (*first_time) {
        int ccount[n_classes];

        *first_time = false;
        MALLOC(decoupled_partition, int, instances, "DecoupledDataPartitioning: decoupled_partition");

        for (int i = 0; i < instances; i++)
            ccount[decoupled_partition[i] = labels[i]]++;
        *n_subsets = n_classes;

        // Divide classes with more than "size" instances
        for (int i = 0; i < n_classes; i++)
            if (ccount[i] > size) {
                int l, n;

                n = (ccount[i] % size == 0) ? ccount[i]/size : ccount[i]/size+1;
                l = 0;
                for (int j = 0; j < instances; j++)
                    if (labels[j] == i) {
                        list[l] = j;
                        l++;
                    }
                RandomizeVector(list, ccount[i]);

                // Reassign new subsets
                l = 0;
                for (int j = 0; j < ccount[i]; j++)
                    if (l == 0)     // One subset left alone
                        l++;
                    else {
                        decoupled_partition[list[j]] = *n_subsets+l - 1;
                        l = (l+1) % n;
                    }

                // Add new subsets
                *n_subsets += n - 1;
            }

        REALLOC(*count, int, *n_subsets,    "DecoupledDataPartitioning: count");

    }

    // Get r% random instances and redistribute
    int n = r * instances;

    GetRandomisedVector(list, instances);

    // Assign to n first to random groups
    for (int i = 0; i < n; i++)
        partition[list[i]] = RANDOM(0, *n_subsets);

    // Remaining to decoupled partition
    for (int i = n; i < instances; i++)
        partition[list[i]] = decoupled_partition[list[i]];

    // Count
    for (int i = 0; i < *n_subsets; i++)
        (*count) [i] = 0;
    for (int i = 0; i < instances; i++)
        (*count) [partition[i]]++;
}
