/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"
#include "maths.h"

//  D. R. Wilson, T. R. Martinez, Reduction techniques for instance-based learning algorithms, Machine Learning 38 (2000) 257–286.


int Drop3(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, bool filter_noise, scalar *w, int inputs, int classes, cache_t **neighbors_cache, scalar **distances_cache,
          bool *consider_list, int **index, int *subspace, bool drop2)
{
    bool *remove;
    cache_t **associates;
    int *nearest, **knn, *n_associates, *list, selected;
    scalar value, with, without, *enemy;

    MALLOC(remove, bool, instances, "main: remove");
    MALLOC(enemy, scalar, instances, "main: enemy");
    MALLOC(list, int, instances, "main: list");
    MALLOC(knn, int *, instances, "main: knn");
    MALLOC(associates, cache_t *, instances, "main: associates");
    MALLOC(n_associates, int, instances, "main: n_associates");
    MALLOC(nearest, int, parm->k+1, "main: nearest");
    for (int i = 0; i < instances; i++) {
        MALLOC(knn[i], int, parm->k+1, "main: knn[i]");
        MALLOC(associates[i], cache_t, instances, "main: associates[i]");
    }

    for (int i = 0; i < instances; i++)
        n_associates[i] = 0;

    // Initially S = T (however, some methods can pass S without all instances selected
    selected = 0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            selected++;

    // Noise filtering
    // Wilson editing (only k != 1)
    // Mark for removal instances incorrectly classified by their neighbors
    int count[classes];
    // Do not let empty classes
    for (int i = 0; i < classes; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[labels[i]]++;

    scalar *r;
    MALLOC(r, scalar, instances, "Drop3: r");
    for (int i = 0; i < instances; i++)
        r[i] = 1.0;

    if (filter_noise) {
        for (int i = 0; i < instances; i++)
            remove[i] = false;

        for (int i = 0; i < instances; i++)
            if (S[i]) {
                int backup = S[i];
                S[i] = 0;
                if (kNearestNeighbourCache(i, S, instances, nearest, labels, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace) != labels[i])
                    remove[i] = true;
                S[i] = backup;
            }

        // Remove all marked
        for (int i = 0; i < instances; i++)
            if (remove[i] && (count[labels[i]] > 1)) {
                S[i] = 0;
                selected--;
                count[labels[i]]--;
            }

        if (parm->verbose > 0)
            fprintf(parm->out_file, "Wilson's editing reduction: %d (%6.4lf%%) \n", instances - selected, 1.0 - ((scalar) selected)/instances);
    }

    for (int i = 0; i < instances; i++)
        if (S[i] && (consider_list == NULL || consider_list[i])) {
            kNearestNeighbourCache(i, S, instances, knn[i], labels, parm->k+1, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace);
            // Add neighbors to associates list
            // Consider repeated entries
            for (int j = 1; j < parm->k+1; j++)
                if (knn[i][j] == knn[i][j-1])
                    knn[i][j-1] = -1;
            for (int j = 0; j < parm->k+1; j++)
                if (knn[i][j] != -1) {
                    associates[knn[i][j]][n_associates[knn[i][j]]] = i;
                    n_associates[knn[i][j]]++;
                }
        }

    // Sort instances in order of distance from its nearest enemy
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            // Compute nearest enemy
            enemy[i] = INFINITE;
            if (distances_cache == NULL) {
                for (int j = 0; j < instances; j++)
                    if (S[j] && labels[j] != labels[i] && (value = EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace))
                        < enemy[i])
                        enemy[i] = value;
            } else {
                for (int j = 0; j < instances; j++)
                    if (S[j] && labels[j] != labels[i] && distances_cache[i][j] < enemy[i])
                        enemy[i] = distances_cache[i][j];
            }
        } else
            enemy[i] = 0.0;

    // Sort instances
    OrderedList(enemy, list, instances, true);

    // Filter instances
    for (int i = 0; i < instances; i++)
        if (S[list[i]]) {
            // Associates of P classified correctly with P
            with = 0.0;
            for (int j = 0; j < n_associates[list[i]]; j++)
                if (kNearestNeighbourCache(associates[list[i]][j], S, instances, nearest, labels, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index,
                                           subspace) == labels[associates[list[i]][j]])
                    with += ((w == NULL) ? 1.0 : w[associates[list[i]][j]]);

            // Associates of P classified correctly without P
            S[list[i]]--;
            without = 0.0;
            for (int j = 0; j < n_associates[list[i]]; j++)
                if (kNearestNeighbourCache(associates[list[i]][j], S, instances, nearest, labels, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index,
                                           subspace) == labels[associates[list[i]][j]])
                    without += ((w == NULL) ? 1.0 : w[associates[list[i]][j]]);

            if (without >= with && (count[labels[list[i]]] > 1 || S[list[i]] > 1)) {
                // Remove P
                selected--;
                if (S[list[i]] == 0)
                    count[labels[list[i]]]--;

                for (int j = 0; j < n_associates[list[i]]; j++) {
                    // Remake k-nearest neighbors of P associates
                    kNearestNeighbourCache(associates[list[i]][j], S, instances, knn[associates[list[i]][j]], labels, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r,
                                           NULL, index, subspace);

                    // Add new k+1 neighbor to associates
                    if (knn[associates[list[i]][j]][parm->k] != -1 && (consider_list == NULL || consider_list[associates[list[i]][j]])) {
                        // Test if previously in the list
                        bool found = false;
                        for (int k = 0; k < n_associates[knn[associates[list[i]][j]][parm->k]] && !found; k++)
                            if (associates[knn[associates[list[i]][j]][parm->k]][k] == associates[list[i]][j])
                                found = true;
                        if (!found) {
                            associates[knn[associates[list[i]][j]][parm->k]][n_associates[knn[associates[list[i]][j]][parm->k]]] = associates[list[i]][j];
                            n_associates[knn[associates[list[i]][j]][parm->k]]++;
                        }
                    }
                }

                // DROP2 version
                if (drop2)
                    // Remove P from the list of associates of its neighbors
                    for (int j = 0; j < parm->k+1; j++) {
                        // Look for P in its neighbors' associates
                        for (int k = 0; k < n_associates[knn[list[i]][j]]; k++)
                            if (knn[list[i]][j] != -1)
                                if (associates[knn[list[i]][j]][k] == list[i]) {
                                    associates[knn[list[i]][j]][k] = associates[knn[list[i]][j]][n_associates[knn[list[i]][j]] - 1];
                                    n_associates[knn[list[i]][j]]--;
                                    k--;
                                }
                    }
            } else
                S[list[i]]++;
        }

    FREE(remove);
    FREE(enemy);
    FREE(list);
    FREE(n_associates);
    FREE(nearest);
    for (int i = 0; i < instances; i++) {
        FREE(knn[i]);
        FREE(associates[i]);
    }
    FREE(knn);
    FREE(associates);
    FREE(r);

    return selected;
}
