/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"
#include "isfs.h"
#include "classifiers.h"

scalar EvolutionaryInstanceSelection(const t_Parameters *parm, int *S, scalar **data, int instances, int *label, scalar *w, bool redo_cache, bool free_cache, bool filter_noise, bool use_cache,
                                     int inputs, int *generations_performed, scalar *evaluations, int classes, int minority, t_EvolutionaryScheme evolutionary_scheme, bool evolve_local_k, int min_k,
                                     int max_k, scalar *r, int **index, cache_t **nncache, int *subspace)
{
    int min_dist, actual_size, father, mother, d, best, *aux_pt, *list, **population, localk_reps = 0, *individual_backup = NULL;
    scalar aux_f, *fitness, *ok, *reduction, localk_average_improvement = 0.0;
    cache_t **neighbors_cache = NULL;
    bool redo_localk_cache = true;

    // Test parameters
    if (parm->multiselection && (parm->k == 1 || parm->chc_wrapper != kNN) && !parm->evolutionary_local_k_and_is)
        OnErrorExit(1, "Multi-selection of instances not valid for k=1 or wrapper not k-NN");

    if (evolve_local_k && parm->chc_wrapper != kNN)
        OnErrorExit(1, "Local k not valid for other method than k-NN");

    // Allocate memory for auxiliary variables
    MALLOC(list, int, 2 * parm->individuals, "EvolutionaryInstanceSelection: list");

    // Allocate memory for individuals
    MALLOC(fitness, scalar, 2 * parm->individuals, "EvolutionaryInstanceSelection: fitness");
    MALLOC(ok, scalar, 2 * parm->individuals, "EvolutionaryInstanceSelection: ok");
    MALLOC(reduction, scalar, 2 * parm->individuals, "EvolutionaryInstanceSelection: reduction");
    MALLOC(population, int *, 2 * parm->individuals, "EvolutionaryInstanceSelection: population");
    if (parm->localk_mutation > 0.0 || parm->boost_based_is_mutation)
        MALLOC(individual_backup, int, 2*instances, "EvolutionaryInstanceSelection: individual_backup");
    for (int i = 0; i < 2 * parm->individuals; i++)
        MALLOC(population[i], int, instances * (evolve_local_k ? 2 : 1), "EvolutionaryInstanceSelection: population[i]");

    // If more than two classes ignore
    if (parm->performance_measure == pfGMean && classes != 2)
        OnErrorExit(1, "Cannot use G measure with more than 2 classes");

    // Evaluate size with all instances
    int max_size = 0;
    if (parm->chc_wrapper == SVM || parm->chc_wrapper == C45) {
        int S[instances];
        for (int i = 0; i < instances; i++)
            S[i] = 1;

        if (parm->chc_wrapper == SVM) {
            // Learn model
            struct svm_model *svm;
            svm = TrainSupportVectorMachine(parm, data, subspace, instances, label, inputs, S, index);
            max_size = svm->l;

            svm_free_and_destroy_model(&svm);
        } else {
            Tree tree;

            tree = TrainC45(parm, data, subspace, instances, label, inputs, true, S, classes);
            max_size = TreeSize(tree);
            ReleaseTree(tree);
        }
    }

    if (parm->evolutionary_scheme == schemeSSMA && !use_cache)
        OnErrorExit(ERR_PARAM, "Steady state memetic algorithm needs a neighbors cache");

    // Create nearest neighbors cache
    if (nncache)
        neighbors_cache = nncache;
    else if (use_cache && parm->chc_wrapper == kNN) {
        RANK *rank;
        MALLOC(rank, RANK, instances, "EvolutionaryInstanceSelection: rank");

        MALLOC(neighbors_cache, cache_t *, instances, "EvolutionaryInstanceSelection: neighbors_cache");

        for (int i = 0; i < instances; i++) {
            MALLOC(neighbors_cache[i], cache_t, instances, "EvolutionaryInstanceSelection: neighbors_cache[i]");

            for (int j = 0; j < instances; j++) {
                rank[j].fitness = -EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace)/(r ? r[j] : 1.0);
                rank[j].being = j;
            }

            // Sort
            qsort(rank, instances, sizeof(RANK), Compare);

            // Store neighbors in ascending order of distance
            for (int j = 0; j < instances; j++)
                neighbors_cache[i][j] = rank[j].being;
        }

        FREE(rank);
    }

    // Noise removal
    if (filter_noise) {
        int *count, nearest[parm->k], selection;
        bool *remove;
        selection = instances;
        MALLOC(count, int, classes, "EvolutionaryInstanceSelection: count");
        MALLOC(remove, bool, instances, "EvolutionaryInstanceSelection: remove");

        // Do not let empty classes
        for (int i = 0; i < classes; i++)
            count[i] = 0;

        for (int i = 0; i < instances; i++) {
            S[i] = 1;
            remove[i] = false;
            count[label[i]]++;
        }

        for (int i = 0; i < instances; i++) {
            S[i] = 0;
            if (kNearestNeighbourCache(i, S, instances, nearest, label, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace) != label[i])
                remove[i] = true;
            S[i] = 1;
        }

        // Remove all marked
        for (int i = 0; i < instances; i++)
            if (remove[i] && count[label[i]] > 1) {
                S[i] = 0;
                selection--;
                count[label[i]]--;
            }
        if (parm->verbose > 0)
            fprintf(parm->out_file, "Wilson's editing reduction: %d (%6.4lf%%) \n", instances - selection, 1.0 - ((scalar) selection)/instances);
    } else
        for (int i = 0; i < instances; i++)
            S[i] = 1;

    // Time
    time_t init_time = time(NULL);

    // If initial percentage 0 adjust for uniform distribution in (0, k/2+1)
    double min_init_percent, max_init_percent, init_percent;
    min_init_percent = max_init_percent = (parm->init_percent <= ZERO ? 1.0/(parm->k/2+2.0) : parm->init_percent);

    // Adjust for multi-selection of instances
    if (parm->multiselection && parm->multiis_adjust_initial_percentage) {
        scalar sum = 0.0;
        int n = evolve_local_k ? parm->localk_max_k/2+1 : parm->k/2+1;

        for (int i = 1; i <= n; i++)
            sum += i;

        fprintf(parm->out_file, "Initial percentage adjusted: %lf", parm->init_percent);

        // With this adjustment, multi-selection selects the same number of instances, although some of them repeated
        min_init_percent /= sum/n;

        fprintf(parm->out_file, " => %lf\n", min_init_percent);
    }

    // Create individuals
    for (int i = 0; i < parm->individuals; i++)
        for (int k = 0; k < instances; k++) {
            if (S[k]) {
                init_percent = RANDOM(min_init_percent, max_init_percent);
                if (parm->multiselection)
                    population[i][k] = (RANDOM(0.0, 1.0) < init_percent) ? RANDOM(1, evolve_local_k ? (parm->localk_max_k/2)+2 : (parm->k/2)+2) : 0;
                else
                    population[i][k] = (RANDOM(0.0, 1.0) < init_percent) ? 1 : 0;
            } else
                population[i][k] = 0;

            // Initialize local k values if apply
            if (evolve_local_k)
                population[i][instances+k] = S[k] ? RANDOM(min_k, max_k+1) : 0;
        }

    if (parm->multiselection && parm->multiselect_only_minority_class)
        for (int i = 0; i < parm->individuals; i++)
            for (int k = 0; k < instances; k++)
                if (population[i][k] > 1 && label[k] != minority)
                    population[i][k] = 1;

    for (int i = 0; i < 2 * parm->individuals; i++)
        list[i] = i;

    // Evaluate initial individuals
    if (parm->verbose > 3) {
        fprintf(parm->out_file, "Indiv\tOk\tReduc\tFitness\nInitial Population\n");
        fprintf(parm->out_file, "-------------------------------------------------------\n");
    }

    for (int k = 0; k < parm->individuals; k++) {
        fitness[k] = EvaluateCHCIndividual(parm, population[k], &ok[k], &reduction[k], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k, r, NULL, parm->ok_w,
                                           parm->reduction_w, subspace, 0);

        if (parm->verbose > 3)
            fprintf(parm->out_file, "%d\t%6.4lf\t%6.4lf\t%6.4lf\n", k, ok[k], reduction[k], fitness[k]);

    }
    *evaluations = parm->individuals;

    // For SSMA, SSGA sort by fitness
    if (parm->evolutionary_scheme == schemeSSMA || parm->evolutionary_scheme == schemeSSGA) {
        for (int i = 0; i < parm->individuals; i++) {
            int order = OrderedSelection(fitness+i, 0, parm-> individuals - i)+i;
            aux_pt = population[order];
            population[order] = population[i];
            population[i] = aux_pt;
            aux_f = fitness[order];
            fitness[order] = fitness[i];
            fitness[i] = aux_f;
            aux_f = ok[order];
            ok[order] = ok[i];
            ok[i] = aux_f;
            aux_f = reduction[order];
            reduction[order] = reduction[i];
            reduction[i] = aux_f;
        }
    }

    if (parm->verbose > 3)
        fprintf(parm->out_file, "-------------------------------------------------------\n");

    // For minimum distance take into account initial percentage of 1's
    init_percent = (max_init_percent+min_init_percent)/2.0;
    min_dist = instances * (init_percent/2);

    // Main loop of CHC algorithm
    SHOWPROGRESS(parm->show_progress > 1, "Generations: %d/%0*d/%0*d", parm->generations, WIDTH(parm->generations), 0, WIDTH(parm->generations), parm->generations);
    for (*generations_performed = 0; *generations_performed < parm->generations; (*generations_performed)++) {

        // CHC next generation
        actual_size = parm->individuals;

        // Check evolutionary scheme
        switch (evolutionary_scheme) {
            case schemeCHCGA:
                // Perform HUX crossover
                RandomizeVector(list, parm->population_size);
                for (int i = 0; i < parm->individuals; i += 2) {
                    // Random selection without replacement
                    father = list[i];
                    mother = list[i+1];

                    // If no minimum distance skip
                    if ((d = HammingDistance(population[mother], population[father], instances * (evolve_local_k ? 2 : 1))) < min_dist)
                        continue;

                    HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], instances);
                    if (evolve_local_k)
                        UniformCrossover(population[mother]+instances, population[father]+instances, population[actual_size]+instances, population[actual_size+1]+instances, instances);

                    // Check no multi-selection of majority class if apply
                    if (parm->multiselection && parm->multiselect_only_minority_class) {
                        CheckMultiselectionOfMajorityClass(population[actual_size], instances, (minority+1)%2, label);
                        CheckMultiselectionOfMajorityClass(population[actual_size+1], instances, (minority+1)%2, label);
                    }

                    // Two new individuals in the population.
                    actual_size += 2;
                }

                // If no crossover decrease threshold and retry
                if (actual_size == parm->individuals) {
                    min_dist--;
                    continue;
                }
                break;

            case schemeSSGA:
            case schemeSSMA:
                // Create just one individual
                father = BinaryTournament(fitness, parm->individuals);
                mother = BinaryTournament(fitness, parm->individuals);

                HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], instances);
                if (evolve_local_k)
                    UniformCrossover(population[mother]+instances, population[father]+instances, population[actual_size]+instances, population[actual_size+1]+instances, instances);

                // Check no multi-selection of majority class if apply
                if (parm->multiselection && parm->multiselect_only_minority_class) {
                    CheckMultiselectionOfMajorityClass(population[actual_size], instances, (minority+1)%2, label);
                    CheckMultiselectionOfMajorityClass(population[actual_size+1], instances, (minority+1)%2, label);
                }

                // Mutate the offspring
                RandomMutation(parm, population[actual_size], instances, parm->bit_mutation, label, minority, evolve_local_k, min_k, max_k, *generations_performed, parm->generations);
                RandomMutation(parm, population[actual_size+1], instances, parm->bit_mutation, label, minority, evolve_local_k, min_k, max_k, *generations_performed, parm->generations);

                // Evaluate two new individuals
                scalar fitness_1 = EvaluateCHCIndividual(parm, population[actual_size], &ok[actual_size], &reduction[actual_size], data, instances, label, w, minority, neighbors_cache,
                                   inputs, classes, evolve_local_k, r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                scalar fitness_2 = EvaluateCHCIndividual(parm, population[actual_size+1], &ok[actual_size+1], &reduction[actual_size+1], data, instances, label, w, minority, neighbors_cache,
                                   inputs, classes, evolve_local_k, r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                *evaluations += 2;

                // If memetic version perform local search
                if (parm->evolutionary_scheme == schemeSSMA) {
                    // Adaptive Probability of Local Search: Salvador García, José Ramón Cano and Francisco Herrera, "A memetic algorithm for evolutionary prototype selection: A scaling up approach",
                    // Pattern Recognition 41 (2008), 2693 – 2709
                    scalar pls;

                    // Local search?
                    pls = (fitness_1 > fitness[actual_size-1]) ? 1.0 : 0.0625;
                    if (RANDOM(0.0, 1.0) < pls) {
                        MemeOptimization(population[actual_size], neighbors_cache, instances, label);

                        // Evaluate again
                        fitness_1 = EvaluateCHCIndividual(parm, population[actual_size], &ok[actual_size], &reduction[actual_size], data, instances, label, w, minority, neighbors_cache,
                                                          inputs, classes, evolve_local_k, r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }

                    pls = (fitness_2 > fitness[actual_size-1]) ? 1.0 : 0.0625;
                    if (RANDOM(0.0, 1.0) < pls) {
                        MemeOptimization(population[actual_size+1], neighbors_cache, instances, label);

                        // Evaluate again
                        fitness_2 = EvaluateCHCIndividual(parm, population[actual_size+1], &ok[actual_size+1], &reduction[actual_size+1], data, instances, label, w, minority, neighbors_cache,
                                                          inputs, classes, evolve_local_k, r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }

                    /*
                    RNNMutationCHC(parm, population[actual_size], actual_size, instances, label, w, minority, neighbors_cache, data, inputs, classes, evaluations, evolve_local_k, r);
                    OptimalLocalK(parm, population[actual_size]+instances, data, instances, label, w, use_cache, inputs, classes, minority, population[actual_size], redo_localk_cache, r, index);
                    redo_localk_cache = false;
                    RNNMutationCHC(parm, population[actual_size+1], actual_size+1, instances, label, w, minority, neighbors_cache, data, inputs, classes, evaluations, evolve_local_k, r);
                    OptimalLocalK(parm, population[actual_size+1]+instances, data, instances, label, w, use_cache, inputs, classes, minority, population[actual_size+1], false, r, index);
                    */
                }

                // Sort by fitness
                if (fitness_1 < fitness_2) {
                    int *swi = population[actual_size];
                    population[actual_size] = population[actual_size+1];
                    population[actual_size+1] = swi;

                    scalar sws = ok[actual_size];
                    ok[actual_size] = ok[actual_size+1];
                    ok[actual_size+1] = sws;

                    sws = reduction[actual_size];
                    reduction[actual_size] = reduction[actual_size+1];
                    reduction[actual_size+1] = sws;

                    sws = fitness[actual_size];
                    fitness[actual_size] = fitness[actual_size+1];
                    fitness[actual_size+1] = sws;
                }

                // Replace two worst individuals if worse than these two
                if (fitness_1 > fitness[actual_size-2]) {
                    // Replace second worst
                    int *swi = population[actual_size-2];
                    population[actual_size-2] = population[actual_size];
                    population[actual_size] = swi;

                    ok[actual_size-2] = ok[actual_size];
                    reduction[actual_size-2] = reduction[actual_size];
                    fitness[actual_size-2] = fitness[actual_size];

                    if (fitness_2 > fitness[actual_size-1]) {
                        // Replace worst
                        int *swi = population[actual_size-1];
                        population[actual_size-1] = population[actual_size+1];
                        population[actual_size+1] = swi;

                        ok[actual_size-1] = ok[actual_size+1];
                        reduction[actual_size-1] = reduction[actual_size+1];
                        fitness[actual_size-1] = fitness[actual_size+1];
                    }
                } else if (fitness_1 > fitness[actual_size-1]) {
                    // Replace worst
                    int *swi = population[actual_size-1];
                    population[actual_size-1] = population[actual_size];
                    population[actual_size] = swi;

                    ok[actual_size-1] = ok[actual_size];
                    reduction[actual_size-1] = reduction[actual_size];
                    fitness[actual_size-1] = fitness[actual_size];
                }

                // NO new individuals in the population
                actual_size += 0;
                break;

            case schemeGGA:
                // First step: Elitism selection
                for (int i = 0; i < parm->gga_elitism*parm->individuals; i++) {
                    int next = OrderedSelection(fitness, i, parm->individuals);
                    // Copy best i-th individual
                    for (int j = 0; j < instances * (evolve_local_k ? 2 : 1); j++)
                        population[i+parm->individuals][j] = population[next][j];
                }

                // Second step: Binary tournament selection
                for (int i = parm->gga_elitism*parm->individuals; i < parm->individuals; i++) {
                    list[i] = i+parm->individuals;
                    int next = TournamentSelection(fitness, 2, parm->individuals);
                    // Copy best i-th individual
                    for (int j = 0; j < instances * (evolve_local_k ? 2 : 1); j++)
                        population[i+parm->individuals][j] = population[next][j];
                }

                // Third step: Crossover
                RandomizeVector(list, parm->individuals-parm->gga_elitism*parm->individuals);
                for (int i = parm->gga_elitism*parm->individuals; i < parm->individuals; i += 2) {
                    // Crossover?
                    if (RANDOM(0.0, 1.0) < parm->gga_crossover_prob) {
                        HuxCrossover(population[list[i]], population[list[i+1]], population[list[i]], population[list[i+1]], instances);
                        if (evolve_local_k)
                            UniformCrossover(population[list[i]]+instances, population[list[i+1]]+instances, population[list[i]]+instances, population[list[i+1]]+instances, instances);
                    }
                }

                // Fourth step: Mutation
                // Keep best
                for (int i = 1; i < parm->individuals; i++)
                    if (RANDOM(0.0, 1.0) < parm->gga_mutation)
                        if (RandomMutation(parm, population[i+parm->individuals], instances, parm->bit_mutation, label, minority, evolve_local_k, min_k, max_k, *generations_performed,
                                           parm->generations))
                            if (parm->multiselection && parm->multiselect_only_minority_class)
                                CheckMultiselectionOfMajorityClass(population[i+parm->individuals], instances, (minority+1)%2, label);

                actual_size = 2*parm->individuals;
                break;

            default:
                OnErrorExit(1, "Bad evolutionary scheme");
        }

        for (int k = parm->individuals; k < actual_size; k++)
            fitness[k] = EvaluateCHCIndividual(parm, population[k], &ok[k], &reduction[k], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k, r, NULL,
                                               parm->ok_w, parm->reduction_w, subspace, 0);
        *evaluations += actual_size-parm->individuals;

        // For generational genetic algorithm copy new population into old one
        if (parm->evolutionary_scheme == schemeGGA) {
            for (int i = 0; i < parm->individuals; i++) {
                fitness[i] = fitness[i+parm->individuals];
                reduction[i] = reduction[i+parm->individuals];
                ok[i] = ok[i+parm->individuals];
                for (int j = 0; j < instances * (evolve_local_k ? 2 : 1); j++)
                    population[i][j] = population[i+parm->individuals][j];
            }

            actual_size = parm->individuals;
        }

        // Introduce Reduced Nearest Neighbor mutation if required
        if (parm->rnn_is_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rnn_is_mutation) {
                    // If any removed recalculate fitness
                    if (RNNMutationCHC(parm, population[i], i, instances, label, w, minority, neighbors_cache, data, inputs, classes, evaluations, evolve_local_k, r, subspace)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);

                        (*evaluations)++;
                    }
                }

        // Introduce Drop3 mutation if required
        if (parm->drop3_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->drop3_mutation) {
                    // If any removed recalculate fitness
                    if (InstanceSelectionMutation(parm, population[i], instances, label, w, data, inputs, DROP3, NULL, classes, index, subspace)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
                }

        // Introduce ICF mutation if required
        if (parm->icf_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->icf_mutation) {
                    // If any removed recalculate fitness
                    if (InstanceSelectionMutation(parm, population[i], instances, label, w, data, inputs, ICF, NULL, classes, index, subspace)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
                }

        // Introduce MSS mutation if required
        if (parm->mss_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->mss_mutation) {
                    // If any removed recalculate fitness
                    if (InstanceSelectionMutation(parm, population[i], instances, label, w, data, inputs, MSS, NULL, classes, index, subspace)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
                }

        // Introduce IB3 mutation if required
        if (parm->ib3_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ib3_mutation) {
                    // If any removed recalculate fitness
                    if (InstanceSelectionMutation(parm, population[i], instances, label, w, data, inputs, IB3, evaluations, classes, index, subspace)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
                }

        // Introduce RNN standard mutation if required
        if (parm->rnnstd_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->rnnstd_mutation) {
                    // If any removed recalculate fitness
                    if (InstanceSelectionMutation(parm, population[i], instances, label, w, data, inputs, RNN, NULL, classes, index, subspace)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
                }

        // Introduce RNN standard mutation if required
        if (parm->boost_based_is_mutation > ZERO)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->boost_based_is_mutation) {
                    scalar fit, oki, red;

                    for (int k = 0; k < 2*instances; k++)
                        individual_backup[k] = population[i][k];

                    ((t_Parameters *) parm)->search_heuristic = (RANDOM(0.0, 1.0) < 0.5) ? SFS : SBS;

                    BoostBasedInstanceSelection(parm, population[i], data, instances, label, parm->is_use_cache, inputs, classes, parm->knn_with_weighted_votes, NULL, NULL, minority, true, index,
                                                NULL, NULL, subspace);
                    if (parm->multiselection && parm->multiselect_only_minority_class)
                        CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);

                    fit = EvaluateCHCIndividual(parm, individual_backup, &oki, &red, data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k, r, NULL, parm->ok_w,
                                                parm->reduction_w, subspace, 0);
                    (*evaluations)++;

                    if (fit >= fitness[i]) {
                        fitness[i] = fit;
                        ok[i] = oki;
                        reduction[i] = red;

                        for (int k = 0; k < 2*instances; k++)
                            population[i][k] = individual_backup[k];
                    }
                }

        // Introduce add I or F mutation if required
        if (parm->addI_mutation > ZERO) {
            if (parm->evolutionary_local_k_and_is)
                OnErrorExit(1, "Add instance mutation for evolutionary local k");
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->addI_mutation) {
                    // If any removed recalculate fitness
                    if (AddIorFMutation(parm, population[i], instances, label, w, data, inputs, false, true, evaluations, minority, classes, max_size, r)) {
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
                }
        }

        // Introduce evolutionary local k mutation if required and if k is part of the evolution
        if (parm->localk_mutation > ZERO && evolve_local_k) {
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Local k with subspaces");

            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->localk_mutation) {
                    // Mutation with no selection?
                    if (RANDOM(0.0, 1.0) < parm->localk_mutation_with_no_selection) {
                        OptimalLocalK(parm, population[i]+instances, data, instances, label, w, use_cache, inputs, classes, minority, NULL, redo_localk_cache, r, index);
                        EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k, r, NULL, parm->ok_w,
                                              parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    } else {
                        scalar fit, oki, red, previous_fitness = fitness[i];

                        for (int k = 0; k < 2*instances; k++)
                            individual_backup[k] = population[i][k];

                        OptimalLocalK(parm, individual_backup+instances, data, instances, label, w, use_cache, inputs, classes, minority, individual_backup, redo_localk_cache, r, index);

                        fit = EvaluateCHCIndividual(parm, individual_backup, &oki, &red, data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k, r, NULL, parm->ok_w,
                                                    parm->reduction_w, subspace, 0);
                        (*evaluations)++;

                        if (fit >= fitness[i]) {
                            fitness[i] = fit;
                            ok[i] = oki;
                            reduction[i] = red;

                            for (int k = 0; k < 2*instances; k++)
                                population[i][k] = individual_backup[k];

                            localk_reps++;
                            localk_average_improvement += fitness[i] - previous_fitness;

                            if (parm->verbose > 1)
                                fprintf(parm->out_file, "@Local k mutation fitness increment: %lf\n", fitness[i] - previous_fitness);
                        }
                    }

                    redo_localk_cache = false;
                }
        }

        // Obtain new population selecting best N individuals
        if (parm->evolutionary_scheme != schemeGGA) {
            for (int i = 0; i < parm->individuals; i++) {
                int order = OrderedSelection(fitness+i, 0, actual_size - i)+i;
                aux_pt = population[order];
                population[order] = population[i];
                population[i] = aux_pt;
                aux_f = fitness[order];
                fitness[order] = fitness[i];
                fitness[i] = aux_f;
                aux_f = ok[order];
                ok[order] = ok[i];
                ok[i] = aux_f;
                aux_f = reduction[order];
                reduction[order] = reduction[i];
                reduction[i] = aux_f;
            }
        }

        // Print individuals
        if (parm->verbose > 3) {
            fprintf(parm->out_file, "Indiv\tOk\tReduc\tFitness\n");
            fprintf(parm->out_file, "-------------------------------------------------------\n");
            for (int i = 0; i < parm->individuals; i++)
                fprintf(parm->out_file, "%02d\t%6.4lf\t%6.4lf\t%6.4lf\n", i, ok[i], reduction[i], fitness[i]);

            fprintf(parm->out_file, "-------------------------------------------------------\n");
        }

        // Print best individual so far
        if (parm->verbose > 1) {
            best = OrderedSelection(fitness, 0, parm->individuals);
            fprintf(parm->out_file, "Best[%3d]\tAccuracy: %6.4lf\tReduction: %6.4lf\tFitness: %6.4lf\n", *generations_performed, ok[best], reduction[best],
                    fitness[best]);
        }

        // Finished?
        if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed)) {
            fprintf(parm->out_file, "Reached limit at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
            break;
        }

        // If minimum distance 0 do as requested
        if (evolutionary_scheme == schemeCHCGA && min_dist == 0) {
            if (parm->reinit_distance0) {
                if (parm->verbose > 2)
                    fprintf(parm->out_file, "Staled population reinitialized\n");
                // Reinitialize populations excepting best one
                best = OrderedSelection(fitness, 0, parm->individuals);
                for (int i = 0; i < parm->individuals; i++)
                    // Reseed changing 35%
                    if (i != best) {
                        for (int k = 0; k < instances; k++)
                            population[i][k] = RANDOM(0.0, 1.0) < 0.35 ? (population[best][k]+1) %2 : population[best][k];
                        if (evolve_local_k)
                            for (int k = 0; k < instances; k++)
                                population[i][instances+k] = RANDOM(0.0, 1.0) < 0.35 ? RANDOM(min_k, max_k+1) : population[best][instances+k];
                        if (parm->multiselection && parm->multiselect_only_minority_class)
                            CheckMultiselectionOfMajorityClass(population[i], instances, (minority+1)%2, label);
                        fitness[i] = EvaluateCHCIndividual(parm, population[i], &ok[i], &reduction[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, evolve_local_k,
                                                           r, NULL, parm->ok_w, parm->reduction_w, subspace, 0);
                        (*evaluations)++;
                    }
            } else {
                // Finish
                fprintf(parm->out_file, "Staled population finishing\n");
                break;
            }
        }

        SHOWPROGRESS(parm->show_progress > 1, "\rGenerations: %d/%0*d/%0*d", parm->generations, WIDTH(parm->generations), *generations_performed+1, WIDTH(parm->generations),
                     parm->generations-*generations_performed-1);
    }
    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

    if (localk_reps)
        fprintf(parm->out_file, "Local k mutation average fitness increment: %lf\n", localk_average_improvement/localk_reps);

    if (individual_backup)
        FREE(individual_backup);

    // Free distance cache
    if (use_cache && free_cache && parm->chc_wrapper == kNN && !nncache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }

    best = OrderedSelection(fitness, 0, parm->individuals);

    // Copy best to S
    scalar reduction_ratio = 0.0;
    for (int i = 0; i < instances; i++) {
        S[i] = population[best][i];
        if (S[i])
            reduction_ratio++;
    }
    reduction_ratio = 1.0 - reduction_ratio/instances;

    // Local k values?
    if (evolve_local_k)
        for (int i = instances; i < 2*instances; i++)
            S[i] = population[best][i];

    // Free memory
    FREE(list);
    FREE(ok);
    FREE(reduction);
    for (int i = 0; i < 2 * parm->individuals; i++)
        FREE(population[i]);
    FREE(population);

    return reduction_ratio;
}

// Check majority class instances are not selected more than once
void CheckMultiselectionOfMajorityClass(int *individual, int instances, int majority, int *labels)
{
    for (int i = 0; i < instances; i++)
        if (individual[i] > 1 && labels[i] == majority)
            individual[i] = 1;
}
