/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas and Juan Antonio Romero *
 *   del Castillo                                                          *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <pvm3.h>
#include "pvm-local.h"
#include "macros.h"
#include "is.h"

#define kFOLDS  10
// Evaluates a threshold for votes using the same method that for democratic selection

void DemocraticThresholdEvaluation(const t_Parameters *parm, scalar **data, int *labels, int n, int *partition, int n_subsets, scalar *votes, scalar *ok, scalar *r, scalar *coef, int inputs,
                                   int classes)
{
    int *list, *local_labels, selected, nearest[parm->k];
    scalar **local_data, *local_votes, RED, ACC;

    MALLOC(list, int, n, "DemocraticThresholdEvalution: list");
    MALLOC(local_data, scalar *, n, "DemocraticThresholdEvalution: local_data");
    MALLOC(local_labels, int, n, "DemocraticThresholdEvalution: local_labels");
    MALLOC(local_votes, scalar, n, "DemocraticThresholdEvalution: local_votes");

    if (parm->performance_measure == pfGMean)
        OnErrorExit(1, "Not yet implemented");

    // Initialize values
    // Get maximum and minimum number of votes
    int min_threshold, max_threshold;
    min_threshold = max_threshold = votes[0];
    for (int i = 1; i < n; i++)
        if (min_threshold > votes[i])
            min_threshold = votes[i];
        else if (max_threshold < votes[i])
            max_threshold = votes[i];

    for (int i = min_threshold; i <= max_threshold; i++)
        r[i] = coef[i] = ok[i] = 0.0;

    // Evaluate for every subset
    for (int subset = 0; subset < n_subsets; subset++) {
        // Get subset
        int cursor = 0;
        for (int j = 0; j < n; j++)
            if (partition[j] == subset) {
                local_data[cursor] = data[j];
                local_labels[cursor] = labels[j];
                local_votes[cursor] = votes[j];
                cursor++;
            }

        // Get size with all instances
        int max_size;
        for (int i = 0; i < cursor; i++)
            list[i] = i;

        if (parm->votes_evaluator != kNN)
            EvaluateWithList((t_Parameters *) parm, local_data, cursor, local_labels, list, local_data, cursor, local_labels, NULL, &max_size, inputs, classes, parm->votes_evaluator, NULL,
                             NULL, NULL, NULL, NULL);

        // Get coefficient for every threshold
        for (int th = min_threshold; th <= max_threshold; th++) {
            // Select instances
            SelectInstancesUsingThreshold(local_votes, cursor, NULL, list, &selected, th, parm->multiselection, parm->k);

            // Evaluate
            if (!parm->democratic_evaluation_of_threshold_with_cv) {
                int size;

                ACC = EvaluateWithList((t_Parameters *) parm,local_data, cursor, local_labels, list, local_data, selected, local_labels, NULL, &size, inputs, classes, parm->votes_evaluator,
                                       NULL, NULL, NULL, NULL, NULL);

                if (parm->votes_evaluator == kNN)
                    r[th] += (RED = 1.0 - ((scalar) selected)/cursor);
                else
                    r[th] += (RED = MAX(0.0, 1.0 - ((scalar) size)/max_size));
            } else {
                ACC = 0.0;

                for (int fold = 0; fold < kFOLDS; fold ++) {
                    int first = (fold * cursor)/kFOLDS;
                    // Account for remaining instances
                    int last = ((fold == kFOLDS - 1) ? cursor : ((fold+1) * cursor)/kFOLDS);

                    // Remove selected from fold
                    int fold_selected = selected;
                    for (int i = 0; i < fold_selected; i++)
                        if (list[i] >= first && list[i] < last) {
                            // Exchange with last
                            fold_selected--;
                            int sw = list[i];
                            list[i] = list[fold_selected];
                            list[fold_selected] = sw;
                        }

                    // Evaluate fold
                    for (int i = first; i < last; i++)
                        if (kNearestNeighbourWithList(local_data[i], list, local_data, fold_selected, inputs, parm->k, nearest, local_labels, classes, parm->knn_with_weighted_votes, NULL,
                                                      NULL, NULL, NULL) == local_labels[i])
                            ACC++;

                }

                r[th] += (RED = 1.0 - ((scalar) selected)/cursor);

                // Evaluate using k-fold cross-validation
                ACC /=  cursor;
            }

            ok[th] += ACC;

            coef[th] += parm->alpha_acc * ACC+(1.0 - parm->alpha_acc) * RED;

            SHOWPROGRESS(parm->show_progress > 0, "Democratic evaluation of thresholds (%2d of %2d)\n", th+1, max_threshold-min_threshold+1);
        }

        SHOWPROGRESS(parm->show_progress > 0, "Democratic evaluation of thresholds subset done (%2d of %2d)\n", subset+1, n_subsets);
    }

    // Average
    for (int th = min_threshold; th <= max_threshold; th++) {
        r[th] /= n_subsets;
        coef[th] /= n_subsets;
        ok[th] /= n_subsets;
    }

    FREE(local_data);
    FREE(local_labels);
    FREE(list);
    FREE(local_votes);
}

// Evaluates a threshold for votes using the same method that for democratic selection using a parallel implementation

int ParallelDemocraticThresholdEvaluation(const t_Parameters *parm, scalar **data, int *labels, int n, int *partition, int n_subsets, scalar *votes, scalar *ok, scalar *r, scalar *coef,
        int processors, int *tids, char *parameters_file, FILE *binary, int max_instances, char *out, int inputs, int classes)
{
    int *local_labels;
    scalar **local_data, *local_votes, *r_slave, *ok_slave;
    static bool first_time = true;
    static int guardian;

    // If first time run processes
    if (first_time) {
        // Parameters for spawning process
        char *args[3];
        MALLOC(args[0], char, strlen(parameters_file)+1, "main");
        strcpy(args[0], parameters_file);       // Parameters file name
        args[1] = getcwd(NULL, 0);      // Current directory for slaves
        args[2] = NULL;

        // Avoid running processes in master node
        int numt = pvm_spawn("eval-wrapper", args, PvmTaskHost | PvmHostCompl, "skynet.uco.es", processors, tids);
        // PvmTaskDefault+PvmTaskDebug, "skynet.uco.es", processors, tids);
        if (numt != processors)
            OnErrorExit(1, "ParallelDemocraticThresholdEvaluation: Not enough processes spawned");

        // Start guardian and send info
        guardian = InitGuardian(numt, tids, "eval-wrapper", out);

        first_time = false;
    }

    // Initialize values
    // Get maximum and minimum number of votes
    int min_threshold, max_threshold;
    min_threshold = max_threshold = votes[0];
    for (int i = 1; i < n; i++)
        if (min_threshold > votes[i])
            min_threshold = votes[i];
        else if (max_threshold < votes[i])
            max_threshold = votes[i];
    for (int i = min_threshold; i <= max_threshold; i++)
        r[i] = coef[i] = ok[i] = 0.0;

    int n_thresholds = max_threshold - min_threshold+1;

    MALLOC(local_data, scalar *, n, "ParallelDemocraticThresholdEvaluation: local_data");
    MALLOC(local_labels, int, n, "ParallelDemocraticThresholdEvaluation: local_labels");
    MALLOC(local_votes, scalar, n, "ParallelDemocraticThresholdEvaluation: local_votes");

    // If huge dataset allocate memory
    if (parm->huge_dataset)
        for (int i = 0; i < max_instances; i++)
            MALLOC(local_data[i], scalar, inputs, "ParallelDemocraticThresholdEvaluation: local_data[i]");

    ALLOCA(r_slave, scalar, n_thresholds,  "ParallelDemocraticThresholdEvaluation: r_slave");
    ALLOCA(ok_slave, scalar, n_thresholds,  "ParallelDemocraticThresholdEvaluation: ok_slave");

    r_slave -= min_threshold;
    ok_slave -= min_threshold;

    // Evaluate for every subset
    for (int subset = 0; subset < n_subsets; subset += processors) {
        for (int p = 0; p < processors && subset+p < n_subsets; p++) {
            int cursor = 0;
            if (parm->huge_dataset) {
                for (int j = 0; j < n; j++)
                    if (partition[j] == subset+p) {
                        // Get instance from file
                        GetInstanceFromBinaryFile(local_data[cursor], &local_labels[cursor], j, inputs, binary);
                        local_votes[cursor] = votes[j];
                        cursor++;
                    }
            } else {
                // Get subset
                for (int j = 0; j < n; j++)
                    if (partition[j] == subset+p) {
                        local_data[cursor] = data[j];
                        local_labels[cursor] = labels[j];
                        local_votes[cursor] = votes[j];
                        cursor++;
                    }
            }

            // Send data
            pvm_initsend(PvmDataDefault);  // PvmDataDefault 0 XDR
            pvm_pkint(&cursor, 1, 1);     // Number of instances
            pvm_pkint(&inputs, 1, 1);     // Number of instances
            int k = parm->k;
            pvm_pkint(&k, 1, 1);
            pvm_pkint(&classes, 1, 1);
            pvm_pkint(&min_threshold, 1, 1);
            pvm_pkint(&max_threshold, 1, 1);
            pvm_pkscalar(local_votes, cursor, 1);    // Votes
            pvm_pkint(local_labels, cursor, 1);    // Labels
            for (int i = 0; i < cursor; i++)
                pvm_pkscalar(local_data[i], inputs, 1);
            pvm_send(tids[p], SENT_BY_MASTER);
        }

        // Check no problem with processes
        for (int p = 0; p < processors; p++)
            if (pvm_pstat(tids[p]) == PvmNoTask) {
                printf("Died after %d/%d subsets processed\n", subset, n_subsets);
                OnErrorExit(ERR_PVM, "PVM Task died unexpectedly");
            }

        // Get values
        for (int p = 0; p < processors && subset+p < n_subsets; p++) {
            // Receive data
            pvm_recv(-1, SENT_BY_SLAVE);    // Receive from any slave

            pvm_upkscalar(r_slave+min_threshold, n_thresholds, 1);
            pvm_upkscalar(ok_slave+min_threshold, n_thresholds, 1);

            // Get coefficient for every threshold
            for (int th = min_threshold; th <= max_threshold; th++) {
                // Evaluate
                r[th] += r_slave[th];
                ok[th] += ok_slave[th];

                coef[th] += parm->alpha_acc * ok_slave[th]+(1.0 - parm->alpha_acc) * r_slave[th];
            }
            if (parm->verbose >= 0) {
                printf("Evaluation %d/%d\n", subset+p+1, n_subsets);
                fflush(stdout);
            }
        }
    }

    // Average
    for (int th = min_threshold; th <= max_threshold; th++) {
        r[th] /= n_subsets;
        coef[th] /= n_subsets;
        ok[th] /= n_subsets;
    }

    // If huge dataset free memory
    if (parm->huge_dataset)
        for (int i = 0; i < max_instances; i++)
            FREE(local_data[i]);

    FREE(local_data);
    FREE(local_labels);
    FREE(local_votes);

    return guardian;
}
