/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "classifiers.h"

// Local k for k-NN
// Each instance has an associated number which is the number of neighbors when the instances is the NN

void EvolutionaryLocalK(t_Parameters *parm, int *localk, scalar **data, int instances, int *label, scalar *w, bool redo_cache, bool free_cache, bool use_cache, int inputs, int *generations_performed,
                        scalar *evaluations, int classes, int minority, t_EvolutionaryScheme evolutionary_scheme, scalar *r, int **index)
{
    int actual_size, order, father, mother, best, *aux_pt, *list, **population, minimum_distance;
    scalar aux_f, *fitness, *accuracy, *regularization, *margin, **distance_cache = NULL;
    int **neighbors_cache = NULL;

    // Allocate memory for auxiliary variables
    MALLOC(list, int, 2 * parm->individuals, "EvolutionaryLocalK: list");
    // Allocate memory for individuals
    MALLOC(fitness, scalar, 2 * parm->individuals, "EvolutionaryLocalK: fitness");
    MALLOC(accuracy, scalar, 2 * parm->individuals, "EvolutionaryLocalK: accuracy");
    MALLOC(regularization, scalar, 2 * parm->individuals, "EvolutionaryLocalK: regularization");
    MALLOC(margin, scalar, 2 * parm->individuals, "EvolutionaryLocalK: margin");
    MALLOC(population, int *, 2 * parm->individuals, "EvolutionaryLocalK: population");
    for (int i = 0; i < 2 * parm->individuals; i++)
        MALLOC(population[i], int, instances, "EvolutionaryLocalK: population[i]");

    // If more than two classes ignore
    if (parm->performance_measure == pfGMean && classes != 2)
        OnErrorExit(1, "Cannot use G measure with more than 2 classes");

    // If lambda = 0 obtain k by cross-validation and adjust the parameter
    if (parm->localk_lambda <= 0) {
        for (int i = 0; i < instances; i++)
            population[0][i] = 1;

        int k = ObtainkCrossValidation(parm, NULL, data, instances, label, inputs, classes, index, NULL);

        parm->localk_lambda = k;

        fprintf(parm->out_file, "Minimum k: %d Lambda: %d Maximum k: %d\n", parm->localk_min_k, parm->localk_lambda, parm->localk_max_k);
    }

    // Create nearest neighbors cache
    if (use_cache && redo_cache && parm->chc_wrapper == kNN) {
        int size;
        RANK *rank;
        MALLOC(rank, RANK, instances, "EvolutionaryLocalK: rank");
        MALLOC(neighbors_cache, int *, instances, "EvolutionaryLocalK: neighbors_cache");
        MALLOC(distance_cache, scalar *, instances, "EvolutionaryLocalK: distance_cache");

        if (parm->localk_min_k == 0)
            size = instances;
        else if (parm->localk_error_estimation_method == HOLDOUT)
            size = MIN(parm->localk_max_k+instances*parm->localk_holdout_percent+1, instances);
        else if (parm->localk_error_estimation_method != kFOLDCV && parm->localk_error_estimation_method != BS && parm->localk_error_estimation_method != E632)
            size = MIN(parm->localk_max_k+1, instances);
        else
            size = MIN(parm->localk_max_k+instances/parm->localk_nfolds+2, instances);

        for (int i = 0; i < instances; i++) {
            MALLOC(neighbors_cache[i], int, size+1, "EvolutionaryLocalK: neighbors_cache[i]");
            MALLOC(distance_cache[i], scalar, size+1, "EvolutionaryLocalK: distance_cache[i]");

            for (int j = 0; j < instances; j++) {
                rank[j].fitness = -EuclideanDistanceSparseData(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL)/r[j];
                rank[j].being = j;
            }

            // Sort
            qsort(rank, instances, sizeof(RANK), Compare);

            // Store neighbors in ascending order of distance
            for (int j = 0; j < size; j++) {
                neighbors_cache[i][j] = rank[j].being;
                distance_cache[i][j] = -rank[j].fitness;
            }

            // For error checking
            neighbors_cache[i][size] = -1;
            distance_cache[i][size] = -1.0;

        }
        FREE(rank);
    }

    // Time
    time_t init_time = time(NULL);

    // Create individuals using a Poisson distribution around the optimal value of k
    for (int i = 0; i < parm->individuals; i++)
        for (int k = 0; k < instances; k++) {
            population[i][k] = PoissonRandomNumber(parm->localk_lambda);
            population[i][k] = MAX(parm->localk_min_k, population[i][k]);
            population[i][k] = MIN(population[i][k], parm->localk_max_k);

            // population[i][k] = RANDOM(parm->localk_min_k, parm->localk_max_k);
        }

    for (int i = 0; i < 2 * parm->individuals; i++)
        list[i] = i;

    // Evaluate initial individuals
    if (parm->verbose > 3) {
        fprintf(parm->out_file, "Individual\tFitness\nInitial Population\n");
        fprintf(parm->out_file, "-------------------------------------------------------\n");
    }

    for (int k = 0; k < parm->individuals; k++) {
        fitness[k] = EvaluateLocalKIndividual(parm, population[k], data, instances, label, w, minority, neighbors_cache, inputs, classes, &accuracy[k], &regularization[k], distance_cache, &margin[k],
                                              r, index);
        if (parm->verbose > 3)
            fprintf(parm->out_file, "%d\t%6.4lf\n", k, fitness[k]);

    }
    *evaluations = parm->individuals;

    if (parm->verbose > 3)
        fprintf(parm->out_file, "-------------------------------------------------------\n");

    // For CHC like algorithm
    minimum_distance = instances/4;

    // Main loop of genetic algorithm
    for (*generations_performed = 0; *generations_performed < parm->generations; (*generations_performed)++) {
        if (parm->verbose > 1)
            printf("Generation: %d\n", *generations_performed);

        // CHC next generation
        actual_size = parm->individuals;

        // Check evolutionary scheme
        switch (evolutionary_scheme) {
            case schemeCHCGA:
                // Perform crossover
                for (int i = 0; i < parm->individuals; i += 2) {
                    // Random selection without replacement
                    order = RANDOM(i, parm->individuals);
                    father = list[order];
                    list[order] = list[i];
                    list[i] = father;
                    order = RANDOM(i+1, parm->individuals);
                    mother = list[order];
                    list[order] = list[i+1];
                    list[i+1] = mother;

                    // Check whether to do crossover
                    if (PseudoHammingDistance(population[father], population[mother], instances) < minimum_distance)
                        continue;

                    Crossover(parm, population[actual_size], population[actual_size+1], population[father], population[mother], instances);

                    // Two new individuals in the population.
                    actual_size += 2;
                }

                // If no crossover decrement threshold and retry
                if (actual_size == parm->individuals)
                    minimum_distance--;
                break;

            case schemeSSGA:
            case schemeSSMA:
                // Create just one individual
                father = BinaryTournament(fitness, parm->individuals);
                mother = BinaryTournament(fitness, parm->individuals);

                Crossover(parm, population[actual_size], population[actual_size+1], population[father], population[mother], instances);

                // Mutate one of the offspring
                RandomIntegerMutation(population[actual_size], instances, parm->bit_mutation, parm->localk_min_k, parm->localk_max_k);

                // For memetic algorithm apply local search
                if (parm->evolutionary_scheme == schemeSSMA)
                    MemeticIntegerMutation(parm, population[actual_size+1], instances, parm->localk_memetic_bit_mutation_ratio, parm->localk_min_k, parm->localk_max_k, data, instances, w, label,
                                           neighbors_cache, inputs, classes, minority, evaluations, distance_cache, parm->localk_lambda, r, index);

                // Two new individuals in the population
                actual_size += 2;
                break;

            case schemeGGA:
                // First step: Elitism selection
                for (int i = 0; i < parm->gga_elitism*parm->individuals; i++) {
                    list[i] = i+parm->individuals;
                    int next = OrderedSelection(fitness, i, parm->individuals);
                    // Copy best i-th individual
                    for (int j = 0; j < instances; j++)
                        population[i+parm->individuals][j] = population[next][j];
                }

                // Second step: Binary tournament selection
                for (int i = parm->gga_elitism*parm->individuals; i < parm->individuals; i++) {
                    list[i] = i+parm->individuals;
                    int next = TournamentSelection(fitness, 2, parm->individuals);
                    // Copy best i-th individual
                    for (int j = 0; j < instances; j++)
                        population[i+parm->individuals][j] = population[next][j];
                }

                // Third step: Crossover
                RandomizeVector(list, parm->individuals-parm->gga_elitism*parm->individuals);
                for (int i = parm->gga_elitism*parm->individuals; i < parm->individuals; i += 2)
                    // Crossover?
                    if (RANDOM(0.0, 1.0) < parm->gga_crossover_prob)
                        Crossover(parm, population[list[i]], population[list[i+1]], population[list[i]], population[list[i+1]], instances);

                // Fourth step: Mutation
                // Keep best from random mutation
                for (int i = 1; i < parm->individuals; i++)
                    if (RANDOM(0.0, 1.0) < parm->gga_mutation)
                        RandomIntegerMutation(population[i+parm->individuals], instances, parm->bit_mutation, parm->localk_min_k, parm->localk_max_k);

                // Fifth step: Memetic mutation
                for (int i = 0; i < parm->individuals; i++)
                    if (RANDOM(0.0, 1.0) < parm->localk_memetic_mutation_ratio)
                        MemeticIntegerMutation(parm, population[i+parm->individuals], instances, parm->localk_memetic_bit_mutation_ratio, parm->localk_min_k, parm->localk_max_k, data, instances, w, label,
                                               neighbors_cache, inputs, classes, minority, evaluations, distance_cache, parm->localk_lambda, r, index);

                actual_size = 2*parm->individuals;
                break;

            default:
                OnErrorExit(1, "Bad evolutionary scheme");
        }

        // Reevaluate old individuals?
        for (int k = parm->localk_reevaluate ? 0 : parm->individuals; k < actual_size; k++)
            fitness[k] = EvaluateLocalKIndividual(parm, population[k], data, instances, label, w, minority, neighbors_cache, inputs, classes, &accuracy[k], &regularization[k], distance_cache, &margin[k],
                                                  r, index);
        *evaluations += actual_size - (parm->localk_reevaluate ? 0 : parm->individuals);

        // For generational genetic algorithm copy new population into old one
        if (parm->evolutionary_scheme == schemeGGA) {
            for (int i = 0; i < parm->individuals; i++) {
                fitness[i] = fitness[i+parm->individuals];
                accuracy[i] = accuracy[i+parm->individuals];
                regularization[i] = regularization[i+parm->individuals];
                margin[i] = margin[i+parm->individuals];

                // Exchange with all population
                int *aux_pt = population[i];
                population[i] = population[i+parm->individuals];
                population[i+parm->individuals] = aux_pt;
            }

            actual_size = parm->individuals;
        } else {
            for (int i = 0; i < parm->individuals; i++) {
                order = OrderedSelection(fitness+i, 0, actual_size - i)+i;
                aux_pt = population[order];
                population[order] = population[i];
                population[i] = aux_pt;
                aux_f = fitness[order];
                fitness[order] = fitness[i];
                fitness[i] = aux_f;
                aux_f = accuracy[order];
                accuracy[order] = accuracy[i];
                accuracy[i] = aux_f;
                aux_f = regularization[order];
                regularization[order] = regularization[i];
                regularization[i] = aux_f;
                aux_f = margin[order];
                margin[order] = margin[i];
                margin[i] = aux_f;
            }
        }

        // Print best individual so far
        if (parm->verbose > 1) {
            best = OrderedSelection(fitness, 0, parm->individuals);
            scalar average_k = 0.0;
            for (int i = 0; i < instances; i++)
                average_k += population[best][i];
            fprintf(parm->out_file, "Best[%3d]\tFitness: %6.4lf\tAccuracy: %6.4lf\tRegularization: %6.4lf\tMargin: %6.4lf\tAverage k: %4.2lf\n", *generations_performed, fitness[best],
                    accuracy[best], regularization[best], margin[best], average_k/instances);
        }

        // If minimum distance 0 reinitialize
        if (parm->evolutionary_scheme == schemeCHCGA && minimum_distance == 0) {
            if (parm->verbose > 2)
                fprintf(parm->out_file, "Staled population reinitialized\n");

            // Reinitialize populations excepting best one
            best = OrderedSelection(fitness, 0, parm->individuals);

            for (int i = 0; i < parm->individuals; i++)
                // Reseed changing 35%
                if (i != best) {
                    for (int k = 0; k < instances; k++)
                        if (RANDOM(0.0, 1.0) < 0.35) {
                            population[i][k] = PoissonRandomNumber(parm->localk_lambda);
                            population[i][k] = MAX(parm->localk_min_k, population[i][k]);
                            population[i][k] = MIN(population[i][k], parm->localk_max_k);
                            // population[i][k] = RANDOM(parm->localk_min_k, parm->localk_max_k);
                        } else
                            population[i][k] = population[best][k];

                    fitness[i] = EvaluateLocalKIndividual(parm, population[i], data, instances, label, w, minority, neighbors_cache, inputs, classes, &accuracy[i], &regularization[i],
                                                          distance_cache, &margin[i], r, index);

                    (*evaluations)++;
                }
        }

        // Finished?
        if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed)) {
            fprintf(parm->out_file, "Reached limit at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
            break;
        }
    }

    best = OrderedSelection(fitness, 0, parm->individuals);
    // Copy best to local k
    for (int i = 0; i < instances; i++)
        localk[i] = population[best][i];

    // Training accuracy for k=value assigned
    if (parm->verbose >= 1) {
        char *estimation[] = {"Leave-One-Out", "k-FOLDCV", "Resubstitution", "Bootstrap", "E632", "HoldOut"};

        fprintf(parm->out_file, "@kNN with local k best (%s) training accuracy: %lf\n", estimation[parm->localk_error_estimation_method], accuracy[best]);

        // Leave one out training error for k=value assigned
        for (int i = 0; i < instances; i++)
            population[0][i] = parm->k;

        fitness[0] = EvaluateLocalKIndividual(parm, population[0], data, instances, label, w, minority, neighbors_cache, inputs, classes, &accuracy[0], &regularization[0], distance_cache, &margin[0],
                                              r, index);
        fprintf(parm->out_file, "@kNN with k = %d (%s) training accuracy: %lf\n", parm->k, estimation[parm->localk_error_estimation_method], accuracy[0]);

        // Leave one out training error for k=CV assigned
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "k validation for repeated instances");
        int k = ObtainkCrossValidation(parm, population[0], data, instances, label, inputs, classes, index, NULL);
        for (int i = 0; i < instances; i++)
            population[0][i] = k;

        fitness[0] = EvaluateLocalKIndividual(parm, population[0], data, instances, label, w, minority, neighbors_cache, inputs, classes, &accuracy[0], &regularization[0], distance_cache, &margin[0],
                                              r, index);
        fprintf(parm->out_file, "@kNN with k = %d CV (%s) training accuracy: %lf\n", k, estimation[parm->localk_error_estimation_method], accuracy[0]);
    }

    // Free distance cache
    if (use_cache && free_cache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }

    // Free memory
    FREE(fitness);
    FREE(accuracy);
    FREE(regularization);
    FREE(list);
    for (int i = 0; i < 2 * parm->individuals; i++)
        FREE(population[i]);
    FREE(population);
}

// #define VARIANCE 2500     // For regularization term
// #define SMOOTHER(x)        exp(-pow((x) - ((scalar) best_cv_k), 2.0)/(2.0*VARIANCE));

scalar EvaluateLocalKIndividual(const t_Parameters *parm, int *individual, scalar **data, int instances, int *label, scalar *w, int minority, int **neighbors_cache, int inputs, int classes,
                                scalar *accuracy, scalar *regularization, scalar **distance_cache, scalar *margin, scalar *r, int **index)
{
    int nearest[parm->localk_max_k+1];
    scalar fitness, ind_margin, rs_margin, bs_margin;
    int /* best_cv_k = -1, */ *S = NULL;  // For regularization term
    scalar *acc_by_k = NULL;

    if (neighbors_cache == NULL)
        OnErrorExit(1, "Algorithm with no neighbor cache not yet implemented");

    // Get cross-validation value of k for regularization term
    MALLOC(S, int, instances, "EvaluateLocalKIndividual: S");
    MALLOC(acc_by_k, scalar, parm->localk_max_k+1, "EvaluateLocalKIndividual: acc_by_k");
    for (int i = 0; i < instances; i++)
        S[i] = 1;

    ((t_Parameters *) parm)->min_k = parm->localk_min_k;
    ((t_Parameters *) parm)->max_k = parm->localk_max_k;
    ((t_Parameters *) parm)->k = ObtainkCrossValidation((t_Parameters *) parm, NULL, data, instances, label, inputs, classes, index, NULL);

    *regularization = 0.0;
    for (int i = 0; i < instances; i++)
        // Sum all k values
        // *regularization += SMOOTHER(individual[i]);
        *regularization += acc_by_k[individual[i]];
    *regularization /= instances;

    if (parm->performance_measure == pfGMean) {
        // Set S[i]=0 of k values 0, rest S[i]=1
        for (int i = 0; i < instances; i++)
            S[i] = individual[i] ? 1 : 0;
        *accuracy = *margin = 0.0;

        switch (parm->localk_error_estimation_method) {
                scalar FP, TP, FN, TN, sp, sn;

            case LOO:
                TP = TN = FP = FN = 0.0;

                // Leave one out
                for (int i = 0; i < instances; i++) {
                    int l = 1, k;
                    // Evaluate using leave-one out
                    S[i] = 0;

                    // Get nearest neighbor k value
                    while (!S[neighbors_cache[i][l]])
                        l++;

                    k = individual[neighbors_cache[i][l]];

                    int category = kNearestNeighbourLargeCache(parm, i, S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[i], 0.0, NULL, NULL, index);
                    if (category == label[i]) {
                        if (label[i] == minority)
                            TP += 1.0;
                        else
                            TN += 1.0;
                    } else {
                        if (label[i] == minority)
                            FN += 1.0;
                        else
                            FP += 1.0;

                    }

                    *margin += ind_margin;

                    if (individual[i])
                        S[i] = 1;
                }

                sp = TN+TP > 0.0 ? TN/(TN+FP) : 1.0;
                sn = TP+FN > 0.0 ? TP/(TP+FN) : 1.0;

                *accuracy = sqrt(sp*sn);

                *margin /= instances;
                break;

            default:
                OnErrorExit(1, "Evaluate with G measure not yet implemented for this evaluation method");
        }
    } else if (parm->performance_measure == pfauROC)
        OnErrorExit(1, "Evaluate with AUC not yet implemented");
    else {
        int fold_size, n;
        scalar resubstitution, bootstrap;
        int *list = NULL;

        // Set S[i]=0 of k values 0, rest S[i]=1
        for (int i = 0; i < instances; i++)
            S[i] = individual[i] ? 1 : 0;
        *accuracy = *margin = 0.0;

        switch (parm->localk_error_estimation_method) {


            case LOO:
                // Leave one out
                for (int i = 0; i < instances; i++) {
                    int l = 1, k;
                    // Evaluate using leave-one out
                    S[i] = 0;

                    // Get nearest neighbor k value
                    while (!S[neighbors_cache[i][l]])
                        l++;

                    k = individual[neighbors_cache[i][l]];

                    if (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[i], 0.0, NULL, NULL, index)
                        == label[i])
                        *accuracy += (w == NULL) ? 1.0 : w[i];

                    *margin += ind_margin;

                    if (individual[i])
                        S[i] = 1;
                }

                if (w == NULL)
                    *accuracy /= instances;
                *margin /= instances;
                break;

            case kFOLDCV:
                // k-fold cross-validation
                fold_size = instances%parm->localk_nfolds ? instances/parm->localk_nfolds+1 : instances/parm->localk_nfolds;

                // Random vector to avoid order effects
                MALLOC(list, int, instances, "EvaluateLocalKIndividual: list");
                for (int i = 0; i < instances; i++)
                    list[i] = i;

                RandomizeVector(list, instances);

                // Test accuracy
                for (int fold = 0; fold < parm->localk_nfolds; fold++) {
                    // i-th fold
                    for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++)
                        S[list[j]] = 0;

                    // Test fold
                    for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++) {
                        int l = 1, k;

                        // Get nearest neighbor k value
                        while (!S[neighbors_cache[list[j]][l]])
                            l++;

                        k = individual[neighbors_cache[list[j]][l]];

                        if (kNearestNeighbourLargeCache(parm, list[j], S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[list[j]], 0.0, NULL, NULL,
                                                        index) == label[list[j]])
                            *accuracy += (w == NULL) ? 1.0 : w[list[j]];

                        *margin += ind_margin;
                    }

                    for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++)
                        if (individual[list[j]])
                            S[list[j]] = 1;
                }

                if (w == NULL)
                    *accuracy /= instances;
                *margin /= instances;

                FREE(list);
                break;

            case RESUBS:
                // Training accuracy
                for (int i = 0; i < instances; i++) {
                    int l = 0, k;

                    // Get nearest neighbor k value
                    while (!S[neighbors_cache[i][l]])
                        l++;

                    k = individual[neighbors_cache[i][l]];

                    if (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[i], 0.0, NULL, NULL, index)
                        == label[i])
                        *accuracy += (w == NULL) ? 1.0 : w[i];

                    *margin += ind_margin;
                }

                if (w == NULL)
                    *accuracy /= instances;
                *margin /= instances;
                break;

            case ArtificialData:
                EvaluateWithArtificialPopulation(parm, individual, data, instances, label, inputs, accuracy, margin, classes, w, -1, -1, NULL, parm->localk_lambda, r);
                break;

            case  BS:
                n = 0;  // Number of tested instances
                // Bootstrap samples
                for (int sample = 0; sample < parm->localk_bs_n; sample++) {
                    // Generate sample with replacement
                    for (int i = 0; i < instances; i++)
                        S[i] = 0;
                    for (int i = 0; i < instances; i++)
                        S[(int) RANDOM(0, instances)]++;

                    // Remove individuals with k=0
                    for (int i = 0; i < instances; i++)
                        if (!individual[i])
                            S[i] = 0;

                    // Accuracy of unselected instances
                    for (int i = 0; i < instances; i++)
                        if (!S[i]) {
                            int l = 1, k;

                            // Get nearest neighbor k value
                            while (!S[neighbors_cache[i][l]])
                                l++;

                            // Get nearest neighbor k value
                            k = individual[neighbors_cache[i][l]];

                            if (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[i], 0.0, NULL, NULL, index)
                                == label[i])
                                *accuracy += (w == NULL) ? 1.0 : w[i];

                            *margin += ind_margin;

                            // Instance tested
                            n++;
                        }
                }
                if (w == NULL)
                    *accuracy /= n;
                *margin /= n;
                break;

            case E632:
                // Combination of resubstitution and bootstrap method
                // Resubstitution
                resubstitution = rs_margin = 0.0;
                for (int i = 0; i < instances; i++) {
                    int l = 0, k;

                    // Get nearest neighbor k value
                    while (!S[neighbors_cache[i][l]])
                        l++;

                    k = individual[neighbors_cache[i][l]];

                    if (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[i], 0.0, NULL, NULL, index)
                        == label[i])
                        resubstitution += (w == NULL) ? 1.0 : w[i];

                    rs_margin += ind_margin;
                }
                if (w == NULL)
                    resubstitution /= instances;
                rs_margin /= instances;

                // Bootstrap method
                n = 0;
                bootstrap = bs_margin = 0.0;
                // Bootstrap samples
                for (int sample = 0; sample < parm->localk_bs_n; sample++) {
                    // Generate sample with replacement
                    for (int i = 0; i < instances; i++)
                        S[i] = 0;
                    for (int i = 0; i < instances; i++)
                        S[(int) RANDOM(0, instances)]++;

                    // Remove individuals with k=0
                    for (int i = 0; i < instances; i++)
                        if (!individual[i])
                            S[i] = 0;

                    // Accuracy of unselected instances
                    for (int i = 0; i < instances; i++)
                        if (!S[i]) {
                            int l = 1, k;

                            // Get nearest neighbor k value
                            while (!S[neighbors_cache[i][l]])
                                l++;

                            // Get nearest neighbor k value
                            k = individual[neighbors_cache[i][l]];

                            if (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[i], 0.0, NULL, NULL, index)
                                == label[i])
                                bootstrap += (w == NULL) ? 1.0 : w[i];

                            bs_margin += ind_margin;

                            // Instance tested
                            n++;
                        }
                }
                if (w == NULL)
                    bootstrap /= n;
                bs_margin /= n;

                *accuracy = 0.368 * resubstitution+0.632 * bootstrap;
                *margin = 0.368 * rs_margin+0.632 * bs_margin;
                break;

            case HOLDOUT:
                n = instances * parm->localk_holdout_percent;

                // Random vector to choose a new evaluation set each iteration
                MALLOC(list, int, instances, "EvaluateLocalKIndividual: list");
                for (int i = 0; i < instances; i++)
                    list[i] = i;

                RandomizeVector(list, instances);

                // Unselect samples for hold out evaluation
                for (int i = 0; i < n; i++)
                    S[list[i]] = 0;
                for (int i = n; i < instances; i++)
                    S[list[i]] = individual[list[i]] ? 1 : 0;

                // Accuracy of unselected instances
                for (int i = 0; i < n; i++) {
                    int l = 1, k;

                    // Get nearest neighbor k value
                    while (!S[neighbors_cache[list[i]][l]])
                        l++;

                    // Get nearest neighbor k value
                    k = individual[neighbors_cache[list[i]][l]];

                    if (kNearestNeighbourLargeCache(parm, list[i], S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs, label[list[i]], 0.0, NULL, NULL, index)
                        == label[list[i]])
                        *accuracy += (w == NULL) ? 1.0 : w[list[i]];

                    *margin += ind_margin;
                }

                if (w == NULL)
                    *accuracy /= n;
                *margin /= n;

                FREE(list);
                break;

            default:
                OnErrorExit(1, "Bad error estimation method");

        }
    }

    FREE(S);
    FREE(acc_by_k);

    // Consider accuracy and sum
    fitness = parm->localk_alpha_acc * (*accuracy)+parm->localk_alpha_reg * (*regularization)+parm->localk_alpha_mar * (*margin);

    return fitness;
}

#define SIGMOID(x) (2.0*(1.0/(1.0+exp(-0.057*(x)))-0.5))

// Classify a pattern using k-NN rule with cache of neighbors
int kNearestNeighbourLargeCache(const t_Parameters *parm, int instance, int *S, int n_protos, int *nearest, int *label, int k, int **neighbors_cache, scalar **distance_cache, int classes,
                                scalar *margin, int inputs, int instance_label, int max_distance_value, scalar **protos, scalar *r, int **index)
{
    int c = 0;
    scalar votes[classes], distances[k], wvotes[classes];

    for (int i = 0; i < classes; i++)
        votes[i] = wvotes[i] = 0;
    for (int i = 0; i < k; i++)
        nearest[i] = -1;

    // Find first k neighbors in the sample
    scalar homogeneous_neighborhood = 0.0, heterogeneous_neighborhood = 0.0;
    int n_homo = 0, n_hetero = 0;
    for (int i = 0; i < n_protos && c < k; i++)
        if (S[neighbors_cache[instance][i]]) {
            int count = S[neighbors_cache[instance][i]];
            while (count-- && c < k) {
                nearest[c] = neighbors_cache[instance][i];

                // Vote neighbor's class
                votes[label[nearest[c]]] += 1.0;
                wvotes[label[nearest[c]]] += k-c;
                distances[c] = distance_cache[instance][i];

                // Homogeneous neighborhood
                if (label[nearest[c]] == instance_label) {
                    homogeneous_neighborhood += distance_cache[instance][i];
                    n_homo++;
                } else { // Heterogeneous neighborhood
                    heterogeneous_neighborhood += distance_cache[instance][i];
                    n_hetero++;
                }
                c++;
            }
        }

    if (parm->knn_with_weighted_votes)
        WeightVotes(votes, k, distances, nearest, label, classes);

    // Symmetrical nearest neighbor
    if (parm->symmetric_knn)
        for (int i = 0; i < n_protos; i++)
            if (S[i]) {
                if (EuclideanDistanceSparseData(protos[instance], protos[i], inputs, index ? index[instance] : NULL, index ? index[i] : NULL)/r[i] < distance_cache[i][neighbors_cache[i][k-1]])
                    votes[label[i]]++;
            }

    // Get maximum votes
    int category = 0;
    for (int i = 1; i < classes; i++)
        if (votes[category] < votes[i])
            category = i;

    // Obtain margin of classification
    switch (parm->margin_method) {
            scalar second, first;

        case AverageNeighborhoodMargin:
        case NormalizedAverageNeighborhoodMargin:
            // Total average neighborhood margin [Wang et al., PR 42, 2863-2875, 2009]
            if (n_hetero && n_homo)
                *margin = heterogeneous_neighborhood/n_hetero - homogeneous_neighborhood/n_homo;
            else {
                // At least distance to k+1 neighbor
                c = 0;
                for (int i = 0; i < n_protos && c < k; i++)
                    if (S[neighbors_cache[instance][i]])
                        c++;

                if (n_hetero == 0)
                    *margin = distance_cache[instance][c] - homogeneous_neighborhood/n_homo;
                else
                    *margin = heterogeneous_neighborhood/n_hetero - distance_cache[instance][c];
            }

            // Normalize?
            if (parm->margin_method == NormalizedAverageNeighborhoodMargin)
                (*margin) /= max_distance_value;
            break;

        case NormalizedVotesDifference:
        case VotesDifference:
        case SmoothedVotesDifference:
            // Obtain maximum value excepting the category
            second = -1.0;
            for (int i = 0; i < classes; i++)
                if (i != instance_label && votes[i] > second)
                    second = votes[i];

            *margin = votes[instance_label] - second;

            if (parm->margin_method == NormalizedVotesDifference)
                *margin /= k;
            else if (parm->margin_method == SmoothedVotesDifference)
                // Adjust margin for compensating for k value
                *margin = ((*margin)/k) * SIGMOID((scalar) k);
            break;

        case NormalizedOrderedVotesDifference:
        case OrderedVotesDifference:
            // Obtain maximum value excepting the category
            second = -1.0;
            for (int i = 0; i < classes; i++)
                if (i != instance_label && wvotes[i] > second)
                    second = wvotes[i];

            *margin = wvotes[instance_label] - second;

            if (parm->margin_method == NormalizedOrderedVotesDifference) {
                int sum = 0;
                for (int i = 1; i <= k; i++)
                    sum += i;
                *margin /= sum;
            }
            break;

        case NearestNeighbourEnemyDistance:
            first = second = -1.0;
            for (int i = 0, c = 0; i < n_protos && c < k && (first < 0.0 || second < 0.0); i++)
                if (S[neighbors_cache[instance][i]]) {
                    if (label[neighbors_cache[instance][i]] == instance_label && first < 0.0)
                        first = distance_cache[instance][i];
                    else if (label[neighbors_cache[instance][i]] != instance_label && second < 0.0)
                        second = distance_cache[instance][i];

                    c++;
                }

            if (first > 0.0 && second > 0.0)
                *margin = second - first;
            else {
                // Get k+1 instance
                int index = -1;
                for (int i = 0, c = 0; c <= k; c++, i++)
                    if (S[neighbors_cache[instance][i]])
                        index = i;

                if (second < 0.0)       // No enemies among the k neighbors
                    *margin = distance_cache[instance][index] - first;
                else     // No friends among the k neighbors
                    *margin = second - distance_cache[instance][index];
            }
            break;

        default:
            OnErrorExit(1, "Bad margin method");
    }

    return category;
}

bool RandomIntegerMutation(int *individual, int size, scalar bit_mutation, int min, int max)
{
    bool modified = false;

    for (int j = 0; j < size; j++)
        if (RANDOM(0.0, 1.0) < bit_mutation) {
            modified = true;

            individual[j] = RANDOM(min, max);
        }

    return modified;
}

scalar EvaluateLocalK(const t_Parameters *parm, int *localk, scalar **prototypes, int n_prototypes, int *prototype_labels, int inputs, int classes, scalar **data, int instances, int *labels,
                      int minority, scalar *r, int **data_index, int **protos_index)
{
    int nearest[parm->localk_max_k+1], *S;
    scalar accuracy = 0.0, **distance_to_furthest_neighbor = NULL;

    MALLOC(S, int, n_prototypes, "EvaluateLocalK: S");
    for (int i = 0; i < n_prototypes; i++)
        S[i] = localk[i] ? 1 : 0;

    // Vector of distances to furthest neighbor if needed
    if (parm->symmetric_knn) {
        MALLOC(distance_to_furthest_neighbor, scalar *, parm->max_k, "EvaluateLocalK: distance_to_furthest_neighbor");

        for (int i = 0; i < parm->max_k; i++)
            distance_to_furthest_neighbor[i] = NULL;
    }

    if (parm->performance_measure == pfGMean) {
        // Evaluate G-mean
        scalar FP, TP, FN, TN, sp, sn;

        TP = TN = FP = FN = 0.0;

        for (int i = 0; i < instances; i++) {
            // Get nearest neighbor
            int nn_index = -1;
            scalar nn_value = INFINITE, value;

            // Iterate for every prototype
            for (int j = 0; j < n_prototypes; j++)
                if (S[j] && (value = EuclideanDistanceSparseData(data[i], prototypes[j], inputs, data_index ? data_index[i] : NULL, protos_index ? protos_index[j] : NULL)/r[j]) < nn_value) {
                    nn_value = value;
                    nn_index = j;
                }

            int k = localk[nn_index];

            if (parm->symmetric_knn && distance_to_furthest_neighbor[k] == NULL) {
                int neighbors[k];

                MALLOC(distance_to_furthest_neighbor[k], scalar, n_prototypes, "EvaluateLocalK: distance_to_furthest_neighbor[k]");

                for (int j = 0; j < n_prototypes; j++) {
                    int backup = S[j];
                    S[j] = 0;
                    kNNCategory(prototypes[j], prototypes, NULL, n_prototypes, k, inputs, prototype_labels, classes, neighbors, S, NULL, NULL, -1, protos_index, protos_index[j]);
                    S[j] = backup;

                    // Distance to furthest neighbor
                    distance_to_furthest_neighbor[k][j] = EuclideanDistanceSparseData(prototypes[j], prototypes[neighbors[k-1]], inputs, protos_index ? protos_index[j] : NULL,
                                                          protos_index ? protos_index[neighbors[k-1]] : NULL)/r[neighbors[k-1]];
                }
            }

            int category = kNNCategory(data[i], prototypes, NULL, n_prototypes, k, inputs, prototype_labels, classes, nearest, S, NULL, NULL, -1, protos_index, data_index[i]);

            if (category == labels[i]) {
                if (labels[i] == minority)
                    TP += 1.0;
                else
                    TN += 1.0;
            } else {
                if (labels[i] == minority)
                    FN += 1.0;
                else
                    FP += 1.0;
            }
        }

        sp = TN+FP > 0.0 ? TN/(TN+FP) : 1.0;
        sn = TP+FN > 0.0 ? TP/(TP+FN) : 1.0;

        accuracy = sqrt(sp*sn);
    } else if (parm->performance_measure == pfauROC)
        OnErrorExit(1, "Evaluate with AUC not yet implemented");
    else {
        int *list;
        MALLOC(list, int, n_prototypes, "EvaluateLocalK: list");
        for (int i = 0; i < n_prototypes; i++)
            list[i] = i;
        RandomizeVector(list, instances);
        // Evaluate error
        for (int i = 0; i < instances; i++) {
            // Get nearest neighbor
            int nn_index = -1;
            scalar nn_value = INFINITE, value;

            // Iterate for every prototype
            for (int j = 0; j < n_prototypes; j++)
                if (S[list[j]] && (value = EuclideanDistanceSparseData(data[i], prototypes[list[j]], inputs, data_index ? data_index[i] : NULL, protos_index ? protos_index[list[j]] : NULL)/r[list[j]])
                    < nn_value) {
                    nn_value = value;
                    nn_index = list[j];
                }

            int k = localk[nn_index];

            if (parm->symmetric_knn && distance_to_furthest_neighbor[k] == NULL) {
                int neighbors[k];

                MALLOC(distance_to_furthest_neighbor[k], scalar, n_prototypes, "EvaluateLocalK: distance_to_furthest_neighbor[k]");

                for (int j = 0; j < n_prototypes; j++) {
                    int backup = S[j];
                    S[j] = 0;
                    kNNCategory(prototypes[j], prototypes, NULL, n_prototypes, k, inputs, prototype_labels, classes, neighbors, S, NULL, NULL, -1, protos_index, protos_index[j]);
                    S[j] = backup;

                    // Distance to furthest neighbor
                    distance_to_furthest_neighbor[k][j] = EuclideanDistanceSparseData(prototypes[j], prototypes[neighbors[k-1]], inputs, protos_index ? protos_index[j] : NULL,
                                                          protos_index ? protos_index[neighbors[k-1]] : NULL)/r[neighbors[k-1]];
                }
            }

            if (kNNCategory(data[i], prototypes, NULL, n_prototypes, k, inputs, prototype_labels, classes, nearest, S, NULL, NULL, -1, protos_index, data_index[i]) == labels[i])
                accuracy += 1.0;
        }

        FREE(list);

        accuracy /= instances;
    }

    FREE(S);
    if (parm->symmetric_knn) {
        for (int i = 1; i < parm->max_k; i++)
            if (distance_to_furthest_neighbor[i])
                FREE(distance_to_furthest_neighbor[i]);
        FREE(distance_to_furthest_neighbor);
    }

    return accuracy;
}

#define ALPHA   0.5
void Crossover(const t_Parameters *parm, int *son, int *daughter, int *father, int *mother, int s)
{
    // BLX-alpha crossover adapted to integers
    scalar I;

    for (int i = 0; i < s; i++) {
        I = fabs((scalar)(father[i] - mother[i]));
        son[i] = (int) RANDOM(MIN(father[i], mother[i]) - I * ALPHA, MAX(father[i], mother[i])+I * ALPHA);
        daughter[i] = (int) RANDOM(MIN(father[i], mother[i]) - I * ALPHA, MAX(father[i], mother[i])+I * ALPHA);

        // Check limits
        daughter[i] = MAX(parm->localk_min_k, daughter[i]);
        daughter[i] = MIN(daughter[i], parm->localk_max_k);
        son[i] = MAX(parm->localk_min_k, son[i]);
        son[i] = MIN(son[i], parm->localk_max_k);
    }
}

void MemeticIntegerMutation(const t_Parameters *parm, int *individual, int size, scalar bit_mutation, int min, int max, scalar **data, int instances, scalar *w, int *labels,
                            int **neighbors_cache, int inputs, int classes, int minority, scalar *evaluations, scalar **distance_cache, int localk_lambda, scalar *r, int **index)
{
    scalar accuracy, regularization, margin;
    int evals = 0, list[max+1];

    // Order the k randomly
    for (int i = min; i <= max; i++)
        list[i] = i;

    list[min-1] = localk_lambda;

    // Select random points
    for (int j = 0; j < size*bit_mutation; j++) {
        int position = RANDOM(0, size);

        // Get optimum value for this position
        int best_value;
        scalar best_fitness;

        RandomizeVector(list+min, max-min+1);

        best_value = individual[position] = list[min-1];
        best_fitness = PartialEvaluationOfLocalKIndividual(parm, individual, data, instances, labels, w, minority, neighbors_cache, inputs, classes, &accuracy, &regularization,
                       distance_cache, &margin, -1, -1, &evals, r, index);

        for (int i = min; i <= max; i++) {
            individual[position] = list[i];

            scalar fitness = PartialEvaluationOfLocalKIndividual(parm, individual, data, instances, labels, w, minority, neighbors_cache, inputs, classes, &accuracy, &regularization,
                             distance_cache, &margin, list[i-1], position, &evals, r, index);

            if (fitness >= best_fitness) {
                best_fitness = fitness;
                best_value = list[i];
            }
        }

        individual[position] = best_value;
    }

    *evaluations += evals/((scalar) instances);
}

// Pseudo Hamming distance between to integer vectors. It counts the number of different values
int PseudoHammingDistance(int *x, int *y, int length)
{
    int distance = 0;

    for (int i = 0; i < length; i++)
        if (x[i] != y[i])
            distance++;

    return distance;
}
