/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdlib.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include <unistd.h>

// FUNCTIONS RELATED TO CONDENSED NEAREST NEIGHBOR RULE

// Generalized Condensed nearest neighbor rule [Chou, 2006]
int GeneralizedCondensedNearestNeighborRule(int *S, scalar **data, int instances, int *label, int inputs, cache_t **neighbors_cache, scalar rho, int classes, scalar **distance_cache, int k,
        int **index, int *subspace)
{
    int selected = 0, cache_size = 0;
    cache_t **ncache = NULL;
    scalar **dcache = NULL;

    if (k != 1)
        OnErrorExit(1, "GCNN is for k=1");

    // Cache already done?
    if (neighbors_cache == NULL)
        ConstructAllNeighborsCache(data, instances, inputs, &dcache, &ncache, NULL, index, &cache_size, subspace, false);
    else {
        ncache = neighbors_cache;
        dcache = distance_cache;
    }

    // Get minimum distance between instances of different classes
    scalar delta = INFINITE;
    for (int i = 1; i < instances; i++)
        for (int j = 0; j < i; j++)
            if (label[i] != label[j] && dcache[i][j] < delta)
                delta = dcache[i][j];

    // Remove all instances
    for (int i = 0; i < instances; i++)
        S[i] = 0;

    // Add a random prototype for each class
    int count[classes];
    for (int i = 0; i < classes; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[label[i]]++;

    // List of instances of each class
    int *list[classes];
    for (int i = 0; i < classes; i++)
        if (count[i]) {
            MALLOC(list[i], int, count[i], "GeneralizedCondensedNearestNeighborRule: list[i]");
            count[i] = 0;
        } else
            list[i] = NULL;

    for (int i = 0; i < instances; i++) {
        list[label[i]][count[label[i]]] = i;
        count[label[i]]++;
    }

    // One random instance per class
    for (int i = 0; i < classes; i++)
        if (count[i]) {
            S[list[i][(int) RANDOM(0, count[i])]] = 1;
            selected++;
        }

    // Repeat until no new instances are added
    bool finish;
    do {
        finish = true;

        // Check all instances and add instances unabsorbed
        // Unabsorbed instance per class
        for (int i = 0; i < classes; i++)
            count[i] = 0;

        for (int i = 0; i < instances; i++) {
            if (!S[i]) {
                // Check if absorbed
                // Find homogeneous nearest neighbor
                int nearest_homogeneous = -1;
                for (int j = 0; j < instances; j++)
                    if (S[ncache[i][j]] && label[i] == label[ncache[i][j]]) {
                        nearest_homogeneous = ncache[i][j];
                        break;
                    }

                // Find heterogeneous nearest neighbor
                int nearest_heterogeneous = -1;
                for (int j = 0; j < instances; j++)
                    if (S[ncache[i][j]] && label[i] != label[ncache[i][j]]) {
                        nearest_heterogeneous = ncache[i][j];
                        break;
                    }

                // Check absorption
                if (dcache[i][nearest_heterogeneous] - dcache[i][nearest_homogeneous] <= rho * delta) {
                    list[label[i]][count[label[i]]] = i;
                    count[label[i]]++;
                }
            }
        }

        // Randomly add one unabsorbed prototype per class
        for (int i = 0; i < classes; i++)
            if (count[i]) {
                S[list[i][(int) RANDOM(0, count[i])]] = 1;
                selected++;
                finish = false;
            }
    } while (!finish);

    for (int i = 0; i < classes; i++)
        if (list[i] != NULL)
            FREE(list[i]);

    // Free ncache?
    if (neighbors_cache == NULL) {
        for (int i = 0; i < instances; i++) {
            FREE(ncache[i]);
            FREE(dcache[i]);
        }
        FREE(dcache);
        FREE(ncache);
    }

    return selected;
}
