/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that deal with the individuals
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"


// Implementation of Elena Marchiori, "Hit Miss Networks with Applications to Instance Selection", Journal of Machine Learning Research 9 (2008) 997-1017.

int HitMissNetworkForEditingIterate(const t_Parameters *parm, int *S, int instances, cache_t **neighbors_cache, int classes, int *labels, scalar **data, int minority, int inputs, int **index,
                                    int *subspace)
{
    scalar accuracy = 0.0, ok, reduction;
    bool finished = false;
    cache_t **cache = NULL;
    int selected = instances, cache_size = 0;

    // Cache already done?
    if (neighbors_cache == NULL)
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &cache, NULL, index, &cache_size, subspace, false);
    else
        cache = neighbors_cache;

    for (int i = 0; i < instances; i++)
        S[i] = 1;

    // Apply HMNE iteratively until accuracy decreases
    int iteration = 1;
    do {
        scalar acc;
        int sel;

        sel = HitMissNetworkForEditing(S, instances, cache, classes, labels, parm->hmnei_epsilon);
        acc = EvaluateCHCIndividual(parm, S, &ok, &reduction, data, instances, labels, NULL, minority, cache, inputs, classes, false, NULL, index, 1.0, 0.0, subspace, 0);

        SAVEPROGRESS(parm->show_progress > 1, "HMNEI iteration %2d: %*d instances selected, %6.4lf accuracy\n", iteration, WIDTH(instances), sel, acc);
        iteration++;

        if (acc < accuracy || (acc == accuracy && sel >= selected))
            finished = true;
        else {
            accuracy = acc;
            selected = sel;
        }
    } while (!finished);

    // Free cache?
    if (neighbors_cache == NULL) {
        for (int i = 0; i < instances; i++)
            FREE(cache[i]);
        FREE(cache);
    }

    return selected;
}

int HitMissNetworkForEditing(int *S, int instances, cache_t **neighbors_cache, int classes, int *labels, scalar epsilon)
{
    bool ***hmn, *flag;
    int *miss, *hit, selected = 0;
    scalar w[classes], left[classes];

    hmn = HitMissNetwork(instances, neighbors_cache, classes, labels, S);

    // Relative ratio of classes
    for (int c = 0; c < classes; c++)
        w[c] = left[c] = 0.0;

    // Compute misses and hits
    MALLOC(miss, int, instances, "HitMissNetworkForEditingIterated: miss");
    MALLOC(hit, int, instances, "HitMissNetworkForEditingIterated: hit");
    for (int i = 0; i < instances; i++) {
        miss[i] = hit[i] = 0;

        w[labels[i]]++;

        for (int j = 0; j < instances; j++)
            for (int c = 0; c < classes; c++)
                if (hmn[j][i][c]) {
                    if (labels[i] == labels[j])
                        hit[i]++;
                    else
                        miss[i]++;
                }
    }

    // Relative ratio of classes
    for (int c = 0; c < classes; c++)
        w[c] /= instances;

    // R1
    // Flag for removal
    CALLOC(flag, bool, instances, "HitMissNetworkForEditingIterated: flag");
    for (int i = 0; i < instances; i++)
        if (w[labels[i]]*miss[i] + epsilon > (1.0 - w[labels[i]]) * hit[i])
            flag[i] = true;

    // R2
    // Count remaining on every class for possibly recovering some
    for (int i = 0; i < instances; i++)
        if (!flag[i])
            left[labels[i]]++;

    for (int i = 0; i < instances; i++)
        if (flag[i] && left[labels[i]] < 4 && hit[i]+miss[i] > 0)
            flag[i] = false;

    // R3 and R4
    // Convert w to count
    for (int c = 0; c < classes; c++)
        w[c] *= instances;
    for (int i = 0; i < instances; i++)
        if (flag[i]) {
            // R3
            if (classes > 3 && miss[i] < classes/2 && miss[i]+hit[i] > 0)
                flag[i] = false;

            // R4
            if (hit[i] >= w[labels[i]]/4)
                flag[i] = false;
        }

    // Perform Instance selection
    for (int i = 0; i < instances; i++)
        if (S[i])
            selected += (S[i] = flag[i] ? 0 : 1);

    // Free memory
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < instances; j++)
            FREE(hmn[i][j]);
    for (int i = 0; i < instances; i++)
        FREE(hmn[i]);
    FREE(hmn);
    FREE(miss);
    FREE(hit);
    FREE(flag);

    return selected;
}


bool ***HitMissNetwork(int instances, cache_t **neighbors_cache, int classes, int *labels, int *S)
{
    bool ***hmn;

    // Allocate memory
    MALLOC(hmn, bool **, instances, "HitMissNetwork: hmn");
    for (int i = 0; i < instances; i++) {
        MALLOC(hmn[i], bool *, instances, "HitMissNetwork: hmn[i]");
        for (int j = 0; j < instances; j++)
            CALLOC(hmn[i][j], bool, classes, "HitMissNetwork: hmn[i][j]");
    }

    // hmn[i][j][k] points to nearest neighbor of x_i in class k
    for (int i = 0; i < instances; i++)
        // Nearest neighbor for every class
        for (int c = 0; c < classes; c++)
            for (int j = 0; j < instances; j++)
                if (S[neighbors_cache[i][j]] && neighbors_cache[i][j] != i && labels[neighbors_cache[i][j]] == c) {
                    hmn[i][neighbors_cache[i][j]][c] = true;
                    break;
                }

    return hmn;
}
