/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"
#include "classifiers.h"

int IB3Algorithm(int *S, scalar **data, int instances, int *label, int inputs, scalar *evaluations, int classes, cache_t **neighbors_cache, scalar **distances_cache, int **d_index, int *subspace)
{
    int nearest, r, index, selected, k, individual_evaluations = 0, *acceptable;
    scalar *attempts, *accuracy, *freq, nearest_dist, distance;
    bool no_acceptable;

    // Memory allocation
    MALLOC(acceptable, int, instances, "main: acceptable");
    MALLOC(attempts, scalar, instances, "main: attempts");
    MALLOC(accuracy, scalar, instances, "main: accuracy");
    MALLOC(freq, scalar, classes, "main: freq");

    for (int i = 0; i < instances; i++) {
        S[i] = acceptable[i] = 0;
        attempts[i] = accuracy[i] = 0.0;
    }

    for (int i = 0; i < classes; i++)
        freq[i] = 0.0;

    scalar *ratio;
    MALLOC(ratio, scalar, instances, "IB3Algorithm: ratio");
    for (int i = 0; i < instances; i++)
        ratio[i] = 1.0;

    // IB3 algorithm
    // For each instance test if correctly classified
    S[0] = 1;
    attempts[0] = 1.0;
    accuracy[0] = 1.0;
    selected = 1;
    freq[label[0]]++;
    for (int t = 1; t < instances; t++) {
        // Update frequency of the classified
        freq[label[t]]++;

        // Search for acceptable prototypes
        no_acceptable = true;
        for (int j = 0; j < t; j++)
            if (S[j]) {
                if ((acceptable[j] = IsAcceptable(attempts[j], accuracy[j]/attempts[j], freq[label[j]]/(t+1), (scalar) t)))
                    no_acceptable = false;
            } else
                acceptable[j] = 0;

        // Test if there are acceptable instances
        if (no_acceptable) {
            // Select random sample
            r = RANDOM(0, selected);
            index = 0;
            for (k = 0; k < t && index < r; k++)
                if (S[k])
                    index++;

            nearest = k;
        } else {
            kNearestNeighbourCache(t, acceptable, instances, &nearest, label, 1, neighbors_cache, data, inputs, classes, false, NULL, NULL, d_index, subspace);

            individual_evaluations += selected;
        }

        // Add sample
        if (label[t] != label[nearest]) {
            S[t] = 1;
            selected++;
            attempts[t]++;
            accuracy[t]++;
        }

        // Update classification record and remove poor instances
        if (nearest != -1) {
            nearest_dist = (distances_cache == NULL) ? EuclideanDistanceSparseDataSubspace(data[nearest], data[t], inputs, d_index ? d_index[nearest] : NULL, d_index ? d_index[t]: NULL, subspace)
                           : distances_cache[nearest][t];
            for (int j = 0; j < t; j++)
                if (S[j]) {
                    distance = (distances_cache == NULL) ? EuclideanDistanceSparseDataSubspace(data[j], data[t], inputs, d_index ? d_index[j] : NULL, d_index ? d_index[t]: NULL, subspace)
                               : distances_cache[j][t];
                    // Test if as near as "nearest"
                    if (distance <= nearest_dist) {
                        // Update record
                        attempts[j]++;
                        if (label[j] == label[t])
                            accuracy[j]++;

                        // Test if "droppable"
                        if (IsDroppable(attempts[j], accuracy[j], freq[label[j]]/(t+1), (scalar) t)) {
                            selected--;
                            S[j] = 0;
                        }
                    }
                }
            individual_evaluations += selected;
        }
    }

    // Remove all non-acceptable instances
    for (int i = 0; i < instances; i++)
        if (S[i] && !IsAcceptable(attempts[i], accuracy[i]/attempts[i], freq[label[i]]/instances, (scalar) instances)) {
            selected--;
            S[i] = 0;
        }

    // Fractional evaluations of the fitness functions
    if (evaluations != NULL)
        *evaluations += (individual_evaluations*inputs)/(((scalar) instances)*instances*inputs);

    // Memory allocation
    FREE(acceptable);
    FREE(attempts);
    FREE(accuracy);
    FREE(freq);
    FREE(ratio);

    return selected;
}
