/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

// H. Brighton, C. Mellish, Advances in instance selection for instance-based learning algorithms, Data Mining and Knowledge Discovery 6 (2002) 153–172.

int ICFAlgorithm(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, bool filter_noise, int inputs, int classes, cache_t **neighbors_cache, scalar **distances_cache,
                 bool *consider_list, int **index, int *subspace)
{
    scalar *reachable = NULL, *coverage, *r;
    bool *remove, progress;
    int selected, iter, nearest[parm->k];

    MALLOC(reachable, scalar, instances, "ICFAlgorithm: reachable");
    MALLOC(coverage, scalar, instances, "ICFAlgorithm: coverage");
    MALLOC(remove, bool, instances, "ICFAlgorithm: remove");
    MALLOC(r, scalar, instances, "ICFAlgorithm: r");

    // Initially S = T (however, some methods can pass S without all instances selected
    selected = 0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            selected++;

    // Noise filtering
    // Wilson editing
    // Mark for removal instances incorrectly classified by their neighbors
    int count[classes];
    // Do not let empty classes
    for (int i = 0; i < classes; i++)
        count[i] = 0;
    for (int i = 0; i < instances; i++)
        count[labels[i]]++;

    for (int i = 0; i < instances; i++)
        r[i] = 1.0;

    if (filter_noise) {
        for (int i = 0; i < instances; i++)
            remove[i] = false;

        for (int i = 0; i < instances; i++)
            if (S[i]) {
                int backup = S[i];
                S[i] = 0;

                if (kNearestNeighbourCache(i, S, instances, nearest, labels, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace) != labels[i])
                    remove[i] = true;
                S[i] = backup;
            }

        // Remove all marked
        for (int i = 0; i < instances; i++)
            if (remove[i] && count[labels[i]] > 1) {
                S[i] = 0;
                selected--;
                count[labels[i]]--;
            }

        if (parm->verbose > 1) {
            fprintf(parm->out_file, "Wilson's editing reduction: %d (%6.4lf%%) \n", instances - selected, 1.0 - ((scalar) selected)/instances);
            fflush(parm->out_file);
        }
    }

    // Iterative case filtering algorithm
    iter = 1;
    do {
        progress = false;

        // Compute reachable (i) and coverage (i)
        ComputeReachableAndCoverage(parm, reachable, coverage, S, data, labels, instances, inputs, classes, neighbors_cache, distances_cache, consider_list, parm->k, r, index, subspace);

        for (int i = 0; i < instances; i++)
            if (S[i]) {
                // Remove if |reachable(i)| > |coverage(i)|
                if (reachable[i] > coverage[i])
                    remove[i] = true;
                else
                    remove[i] = false;
            }

        // Delete marked for removal
        for (int i = 0; i < instances; i++)
            if (S[i] && remove[i] && (count[labels[i]] > 1 || S[i] > 1)) {
                S[i]--;
                selected--;
                progress = true;
                if (S[i] == 0)
                    count[labels[i]]--;
            }

        // Report last selected
        if (parm->verbose > 2) {
            fprintf(parm->out_file, "Iteration: %3d\tReduction:  %6.4lf\n", iter, 1.0 - ((scalar) selected)/instances);
            fflush(parm->out_file);
        }

        iter++;
    } while (progress);

    if (selected == instances) {
        // No instances removed
        // Check if reachable = coverage
        bool equal = true;
        for (int i = 0; i < instances && equal; i++)
            if (reachable[i] != coverage[i])
                equal = false;

        if (equal) {
            // If equal remove all but one of each class
            int *list;
            MALLOC(list, int, instances, "ICFAlgorithm: list");
            for (int i = 0; i < instances; i++)
                list[i] = i;
            RandomizeVector(list, instances);
            for (int i = 0; i < instances; i++) {
                if (S[list[i]] && (count[labels[list[i]]] > 1 || S[list[i]] > 1)) {
                    S[list[i]]--;
                    selected--;
                    if (S[list[i]] == 0)
                        count[labels[list[i]]]--;
                }

            }
            FREE(list);
        }
    }

    FREE(reachable);
    FREE(coverage);
    FREE(remove);
    FREE(r);

    return selected;
}
