/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas and                     *
 *   Juan Antonio Romero del Castillo                                      *
 *   npedrajas@uco.es, aromero@uco.es                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements Democratic/Oligarchic Instance Selection wrapper

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "imbalance.h"

int InstanceSelectionAlgorithm(const t_Parameters *parm, t_ISAlgorithm isalgorithm, int **S, scalar ***data, int *instances, int inputs, int **labels, int classes, int **index,
                               cache_t **neighbors_cache, scalar **distance_cache, int minority, int *generations, scalar ratio, scalar *w, bool *considerlist, scalar *evaluations, scalar *r,
                               bool cnn_ci, scalar *distance_to_furthest_neighbor, bool keep_previous_selection, int *subspace)
{
    int selected = 0;
    bool drop2 = false;

    switch (isalgorithm) {
        case DROP2:
            drop2 = true;
        case DROP3:
            if (!parm->resample)
                OnErrorExit(1, "Re-weighting not possible");
            if (!keep_previous_selection)
                for (int i = 0; i < *instances; i++)
                    (*S)[i] = 1;

            // Filter noise only for first round
            selected = Drop3(parm, *S, *data, *instances, *labels, parm->filter_noise, w, inputs, classes, neighbors_cache, distance_cache, considerlist, index, subspace, drop2);
            break;

        case IB3:
            if (!parm->resample)
                OnErrorExit(1, "Re-weighting not possible");
            selected = IB3Algorithm(*S, *data, *instances, *labels, inputs, evaluations, classes, neighbors_cache, distance_cache, index, subspace);
            break;

        case ICF:
            if (!keep_previous_selection)
                for (int i = 0; i < *instances; i++)
                    (*S)[i] = 1;
            selected = ICFAlgorithm(parm, *S, *data, *instances, *labels, parm->filter_noise, inputs, classes, neighbors_cache, distance_cache, considerlist, index, subspace);
            break;

        case MSS:
            if (!keep_previous_selection)
                for (int i = 0; i < *instances; i++)
                    (*S)[i] = 1;
            selected = MSSAlgorithm(*S, *data, *instances, *labels, inputs, neighbors_cache, distance_cache, index, subspace);
            break;

        case RNN:
            selected = ReducedNearestNeighbor(parm, *S, *data, *instances, *labels, 1, inputs, classes, neighbors_cache, index, subspace);
            break;

        case CHC:
            selected = *instances * EvolutionaryInstanceSelection(parm, *S, *data, *instances, *labels, w, true, parm->resample, parm->filter_noise, parm->is_use_cache, inputs, generations,
                       evaluations, classes, minority,  parm->evolutionary_scheme, parm->evolutionary_local_k, parm->localk_min_k, parm->localk_max_k, r, index, neighbors_cache, subspace);
            break;

        case RANDOMSELECTION:
            selected = RandomSelection(*S, *data, *instances, *labels, ratio, parm->randomsel_withreplacement, inputs, classes);
            break;

        case CNN:
            selected = CondensedNearestNeighborRule(*S, *data, *instances, *labels, inputs, cnn_ci, minority, neighbors_cache, considerlist,  parm->k, classes, index, subspace);
            break;

        case MCS:
            if (parm->k != 1)
                OnErrorExit(ERR_PARAM, "MCS only valid for k = 1");
            selected = MinimalConsistentSet(parm, *S, *data, *instances, *labels, inputs, neighbors_cache, classes, index, subspace, distance_cache);
            break;

        case GCNN:
            selected = GeneralizedCondensedNearestNeighborRule(*S, *data, *instances, *labels, inputs, neighbors_cache, parm->rho, classes, distance_cache, parm->k, index, subspace);
            break;

        case ENTROPY:
            if (subspace)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm does not accept subspaces");
            selected = EntropyInstanceSelection(*S, *data, *instances, *labels, inputs, classes, parm->k, parm->knn_with_weighted_votes, index);
            break;

        case CNNTL:
            if (subspace)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm does not accept subspaces");
            selected = CNNAndTomekLinks(*S, *data, *instances, *labels, inputs, minority, parm->k, classes, index);
            break;

        case OSS:
            if (subspace)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm does not accept subspaces");
            selected = OneSidedSelection(*S, *data, *instances, *labels, inputs, minority, parm->k, classes, index, neighbors_cache, false);
            break;

        case Undersampling:
            selected = RandomUndersampling(*S, *instances, *labels, minority, parm->random_undersampling_ratio);
            break;

        case RC45:
            // selected = RobustC45(parm, *S, *data, *instances, *labels, inputs, classes, subspace, index);
            selected = RobustDT(parm, *S, *data, *instances, *labels, inputs, classes, subspace, index);
            break;

        case PSRCG:
            selected = PSRCGAlgorithm(parm, *S, *data, *instances, *labels, parm->k, inputs, classes, index, subspace);
            break;

        case CFilter:
            selected = ConsensusFilter(parm, *S, *data, *instances, *labels, inputs, classes, index, subspace);
            break;

        case RMHC:
            selected = RandomMutationHillClimbingInstanceSelection(parm, *S, *data, *labels, *instances, classes, neighbors_cache, inputs, minority, index, subspace);
            break;

        case HMNEI:
            selected = HitMissNetworkForEditingIterate(parm, *S, *instances, neighbors_cache, classes, *labels, *data, minority, inputs, index, subspace);
            break;

        case RNGE:
            selected = RelativeNeighborhoodGraphEditing(parm->rnge_pg, *S, *data, *instances, *labels, classes, inputs, index, distance_cache, subspace);
            break;

        case BoostBased:
            if (!keep_previous_selection)
                for (int i = 0; i < *instances; i++)
                    (*S)[i] = 1;
            selected = BoostBasedInstanceSelection(parm, *S, *data, *instances, *labels, parm->is_use_cache, inputs, classes, parm->knn_with_weighted_votes, distance_to_furthest_neighbor, r,
                                                   minority, false, index, distance_cache, neighbors_cache, subspace);
            break;

        case PSA:
            selected = PatternSelectionAlgorithm(*S, *data, *instances, *labels, parm->k, inputs, classes, parm->knn_with_weighted_votes, subspace, index);
            break;

        case SMOTE:
            if (subspace)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm does not accept subspaces");
            if (index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "SMOTE for sparse data");
            selected = *instances;
            // First under-sample
            if (parm->smote_undersampling_ratio > 0)
                UndersampleData(data, instances, inputs, classes, &*labels, parm->smote_undersampling_ratio);
            // The over-sample for 1:1 with SMOTE
            SMOTEOverSampling(data, instances, inputs, classes, labels, parm->smote_N, parm->smote_k);
            REALLOC(*S, int, *instances+inputs, "main: *S");
            for (int i = 0; i < *instances+inputs; i++)
                (*S)[i] = 1;
            fprintf(parm->out_file, "@SMOTE added instances: %d\n", *instances-selected);
            fprintf(parm->out_file, "@SMOTE oversampling: %lf\n", ((scalar) *instances-selected)/selected);
            selected = *instances;
            break;

        case LVQPRU:
            if (subspace)
                OnErrorExit(ERR_PARAM, "Instance selection algorithm does not accept subspaces");
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "LVQPRU convert results to instance selection");
            LVQPruning(parm, *data, *instances, *labels, inputs, classes);
            break;

        case NoIS:
            if (!keep_previous_selection)
                for (int i = 0; i < *instances; i++)
                    (*S)[i] = 1;
            selected = *instances;
            break;

        case SimulatedAnnealing:
        case kNNCHC:
            OnErrorExit(ERR_PARAM, "Unsuitable Instance Selection method");
            break;
    }

    // If less selected than class return one random sample from every class
    if (selected < classes)
        selected = SelectOneRandomSampleFromEveryClass(*S, *instances, *labels, classes);

    return selected;
}
