/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"
#include "classifiers.h"
#include "ensemble.h"

// Local k for k-NN
// Each instance has an associated number which is the number of neighbors when the instances is the NN

void OptimalLocalK(const t_Parameters *parm, int *localk, scalar **data, int instances, int *labels, scalar *w, bool use_cache, int inputs, int classes, int minority, int *S, bool redo_cache,
                   scalar *r, int **data_index)
{
    static scalar **distance_cache = NULL;
    scalar max_distance = 0.0, **test_data = NULL;
    static int **neighbors_cache = NULL, instances_allocated = -1;
    int *list, no_neighbors_count = 0, *no_neighbors_list, *train_list, test_instances, *test_labels = NULL, localk_lambda, fold_size, *instance_list, *s_backup = NULL, **test_index = NULL;

    // If more than two classes ignore
    if (parm->performance_measure == pfGMean && classes != 2)
        OnErrorExit(ERR_PARAM, "Cannot use G measure with more than 2 classes");

    if (parm->performance_measure == pfauROC && parm->localk_error_estimation_method != RESUBS)
        OnErrorExit(ERR_PARAM, "auROC only compatible with resubstitution error estimation");

    if (parm->localk_error_estimation_method != LOO && parm->localk_error_estimation_method != kFOLDCV && parm->symmetric_knn)
        OnErrorExit(ERR_PARAM, "Symmetric k-NN only compatible with LOO and k-FOLDS error estimations");

    MALLOC(list, int, instances, "OptimalLocalK: list");
    for (int i = 0; i < instances; i++)
        list[i] = i;

    // If no S given select all instances
    bool local_s = false;
    if (S == NULL) {
        MALLOC(S, int, instances, "OptimalLocalK: S");
        for (int i = 0; i < instances; i++)
            S[i] = 1;
        local_s = true;
    } else if (parm->localk_error_estimation_method != LOO && parm->localk_error_estimation_method != kFOLDCV)
        OnErrorExit(ERR_PARAM, "S only taken into account for LOO or kFOLDCV");

    MALLOC(no_neighbors_list, int, instances, "OptimalLocalK: no_neighbors_list");

    // Get cross-validation value of k for regularization term and local k lambda for regularization in case SMOOTHER is used
    scalar acc_by_k[parm->localk_max_k+1];
    if (parm->localk_lambda) {
        localk_lambda = parm->localk_lambda;
        for (int i = parm->localk_min_k; i <= parm->localk_max_k; i++)
            acc_by_k[i] = parm->localk_lambda;
    } else {
        if (parm->performance_measure == pfauROC)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "auROC for performance evaluation");

        ((t_Parameters *) parm)->cv_performance_measure = parm->performance_measure;
        localk_lambda = ObtainkCrossValidation((t_Parameters *) parm, S, data, instances, labels, inputs, classes, data_index, NULL);

        if (parm->verbose >= 0)
            fprintf(parm->out_file, "Minimum k: %d Lambda: %d Maximum k: %d\n", parm->localk_min_k, localk_lambda, parm->localk_max_k);
    }

    // Add artificial data to test?
    int ad_instances = 0, *ad_labels = NULL, **ad_neighbors_cache = NULL, *ad_list = NULL, **ad_index = NULL;
    scalar **ad_data = NULL, **ad_distance_cache  = NULL;

    // Populate using CV k value
    int populate_number_of_neighbors = parm->populate_number_of_neighbors;
    if (parm->populate_number_of_neighbors == 0)
        populate_number_of_neighbors = localk_lambda;

    // Create nearest neighbors cache
    if (use_cache && (neighbors_cache == NULL || redo_cache)) {
        int size = -1;
        RANK *rank;

        // If redo cache free previous cache
        if (neighbors_cache != NULL) {
            for (int i = 0; i < instances_allocated; i++) {
                FREE(neighbors_cache[i]);
                FREE(distance_cache[i]);
            }
            FREE(neighbors_cache);
            FREE(distance_cache);
        }

        MALLOC(neighbors_cache, int *, instances, "OptimalLocalK: neighbors_cache");
        MALLOC(distance_cache, scalar *, instances, "OptimalLocalK: distance_cache");
        MALLOC(rank, RANK, instances, "OptimalLocalK: rank");

        // Remember for freeing
        instances_allocated = instances;

        size = instances;
        //         if (local_s == false)
        //             size = instances;
        //         else {
        //             if (parm->localk_error_estimation_method == kFOLDCV)
        //                 size = MIN(parm->localk_max_k+instances/parm->localk_nfolds+2, instances);
        //             else if (parm->localk_error_estimation_method == LOO || parm->localk_error_estimation_method == RESUBS || parm->performance_measure == pfauROC)
        //                 size = MIN(parm->localk_max_k+1, instances);
        //             else
        //                 OnErrorExit(ERR_PARAM, "Bad error estimation method");
        //         }

        for (int i = 0; i < instances; i++) {
            MALLOC(neighbors_cache[i], int, size+1, "OptimalLocalK: neighbors_cache[i]");
            MALLOC(distance_cache[i], scalar, size+1, "OptimalLocalK: distance_cache[i]");

            // Use random order for repeated entries
            RandomizeVector(list, instances);

            for (int j = 0; j < instances; j++) {
                rank[j].fitness = -EuclideanDistanceSparseData(data[i], data[list[j]], inputs, data_index ? data_index[i] : NULL, data_index ? data_index[list[j]] : NULL)/r[list[j]];
                rank[j].being = list[j];
            }

            // Sort
            qsort(rank, instances, sizeof(RANK), Compare);

            // Store neighbors in ascending order of distance
            for (int j = 0; j < size; j++) {
                neighbors_cache[i][j] = rank[j].being;
                distance_cache[i][j] = -rank[j].fitness;
            }

            // Maximum distance
            for (int j = 0; j < parm->localk_max_k+1 && j < size; j++)
                if (-rank[j].fitness > max_distance)
                    max_distance = -rank[j].fitness;

            // For error checking
            neighbors_cache[i][size] = -1;
            distance_cache[i][size] = -1.0;

        }
        FREE(rank);
    }

    if (parm->localk_populate) {
        // Construct no neighbors vector
        bool *no_neighbors = NULL;

        if (data_index)
            OnErrorExit(ERR_PARAM, "Adding artificial instance incompatible with sparse format");
        SMOTEPopulate(&ad_data, &ad_labels, &ad_instances, data, instances, labels, inputs, parm->populate_copies_per_instance, classes, populate_number_of_neighbors, localk_lambda,
                      no_neighbors, parm->knn_with_weighted_votes, r);
        if (parm->verbose > 1)
            fprintf(parm->out_file, "Added %d artificial instances\n", ad_instances);

        // Create nearest neighbors cache for artificial data
        if (use_cache) {
            int size = -1;
            RANK *rank;
            MALLOC(ad_neighbors_cache, int *, ad_instances, "OptimalLocalK: ad_neighbors_cache");
            MALLOC(ad_distance_cache, scalar *, ad_instances, "OptimalLocalK: ad_distance_cache");
            MALLOC(rank, RANK, instances, "OptimalLocalK: rank");

            if (parm->localk_error_estimation_method == kFOLDCV)
                size = MIN(parm->localk_max_k+instances/parm->localk_nfolds+2, instances);
            else if (parm->localk_error_estimation_method == LOO || parm->localk_error_estimation_method == RESUBS || parm->performance_measure == pfauROC)
                size = MIN(parm->localk_max_k+1, instances);
            else
                OnErrorExit(ERR_PARAM, "Bad error estimation method");

            if (parm->evolutionary_local_k_and_is)
                size = instances;

            for (int i = 0; i < ad_instances; i++) {
                MALLOC(ad_neighbors_cache[i], int, size+1, "OptimalLocalK: ad_neighbors_cache[i]");
                MALLOC(ad_distance_cache[i], scalar, size+1, "OptimalLocalK: ad_distance_cache[i]");

                // Use random order for repeated entries
                RandomizeVector(list, instances);

                for (int j = 0; j < instances; j++) {
                    rank[j].fitness = -EuclideanDistanceSparseData(ad_data[i], data[list[j]], inputs, ad_index[i], data_index[list[j]])/r[list[j]];
                    rank[j].being = list[j];
                }

                // Sort
                qsort(rank, instances, sizeof(RANK), Compare);

                // Store neighbors in ascending order of distance
                for (int j = 0; j < size; j++) {
                    ad_neighbors_cache[i][j] = rank[j].being;
                    ad_distance_cache[i][j] = -rank[j].fitness;
                }

                // Maximum distance
                for (int j = 0; j < parm->localk_max_k+1 && j < size; j++)
                    if (-rank[j].fitness > max_distance)
                        max_distance = -rank[j].fitness;

                // For error checking
                ad_neighbors_cache[i][size] = -1;
                ad_distance_cache[i][size] = -1.0;

            }
            FREE(rank);
        }

        MALLOC(ad_list, int, ad_instances, "OptimalLocalK: ad_list");
    }

    switch (parm->localk_error_estimation_method) {
        case RESUBS:
            if (parm->performance_measure == pfauROC) {
                if (classes != 2)
                    OnErrorExit(ERR_PARAM, "auROC only valid for 2 classes");

                MALLOC(train_list, int, instances, "OptimalLocalK: train_list");
                MALLOC(test_labels, int, instances, "OptimalLocalK: test_labels");
                MALLOC(test_data, scalar *, instances, "OptimalLocalK: test_data");
                if (data_index)
                    MALLOC(test_index, int *, instances, "OptimalLocalK: test_index");

                for (int j = 0; j < instances; j++)
                    train_list[j] = j;

                for (int i = 0; i < instances; i++) {
                    int count[2] = {0, 0};
                    test_instances = 0;
                    for (int neighbor = 0; neighbor <= parm->k; neighbor++) {
                        // Get list of instances which nearest neighbor is i including itself
                        for (int j = 0; j < instances && test_instances < instances; j++)
                            if (neighbors_cache[j][neighbor] == i) {
                                test_data[test_instances] = data[j];
                                count[test_labels[test_instances] = labels[j]]++;
                                if (data_index)
                                    test_index[test_instances] = data_index[j];
                                test_instances++;
                            }

                        // Get list of instances which nearest neighbor is i for artificial data
                        for (int j = 0; j < ad_instances && test_instances < instances; j++)
                            if (ad_neighbors_cache[j][neighbor] == i) {
                                test_data[test_instances] = ad_data[j];
                                count[test_labels[test_instances] = ad_labels[j]]++;
                                test_instances++;
                            }
                    }

                    // If no neighbors of all of them from the same class assign cross-validated value
                    if (!parm->leave_no_neighbors_unchanged && (test_instances == 0 || count[0] == 0 || count[1] == 0)) {
                        localk[i] = localk_lambda;
                        no_neighbors_list[no_neighbors_count] = i;
                        no_neighbors_count++;
                        continue;
                    }

                    int best_k = -1;
                    scalar best_fitness = 0.0;

                    for (int k = parm->localk_min_k; k <= parm->localk_max_k; k++) {
                        scalar auroc = 0.0, regularization, fitness;

                        // Test auROC
                        auroc = ClassifierAUROC(parm, kNN, test_data, test_instances, test_labels, data, instances, NULL, NULL, labels, inputs, minority, false, NULL, true, NULL, classes,
                                                      test_index, data_index, NULL, NULL, NULL, !IsEnsemble(kNN));
                        regularization = acc_by_k[k];
                        fitness = parm->localk_alpha_acc * auroc+parm->localk_alpha_reg * regularization;

                        if (fitness > best_fitness) {
                            best_fitness = fitness;
                            best_k = k;
                        }
                    }

                    // If all fitness 0 set to best cross-validated value
                    if (best_k == -1)
                        localk[i] = localk_lambda;
                    else
                        localk[i] = best_k;
                }
                FREE(test_data);
                FREE(test_labels);
                FREE(train_list);
                if (data_index)
                    FREE(test_index);
            }

            OnErrorExit(ERR_PARAM, "Resubstitution not available for other measure than auROC");
            break;

        case LOO:
            // Optimal value of k for each instance
            // It is a local value as it only affects nearest neighbor
            for (int i = 0; i < instances; i++)
                if (S[i]) {
                    int n = 0, ad_n = 0, nearest[parm->localk_max_k+1];

                    list[n] = i;
                    n++;

                    // Get list of instances which their nearest neighbor is i including itself
                    for (int neighbor = 1; neighbor <= parm->k; neighbor++)
                        for (int j = 0; j < instances; j++)
                            if (neighbors_cache[j][neighbor] == i) {
                                list[n] = j;
                                n++;
                            }

                    // Get list of instances which their nearest neighbor is i for artificial data
                    if (ad_instances)
                        for (int neighbor = 0; neighbor <= parm->k; neighbor++)
                            for (int j = 0; j < ad_instances; j++)
                                if (ad_neighbors_cache[j][neighbor] == i) {
                                    ad_list[ad_n] = j;
                                    ad_n++;
                                }

                    // If no neighbors assign cross-validated value
                    if (!parm->leave_no_neighbors_unchanged && n+ad_n == 0) {
                        localk[i] = localk_lambda;
                        no_neighbors_list[no_neighbors_count] = i;
                        no_neighbors_count++;
                        continue;
                    }

                    int best_k = -1;
                    scalar best_accuracy = -1.0, best_margin = 0.0, best_regularization = 0.0;

                    for (int k = parm->localk_min_k; k <= parm->localk_max_k; k++) {
                        scalar accuracy = 0.0, regularization, margin = 0.0, ind_margin, TP, TN, FP, FN, count = 0.0;

                        int C[classes][classes], label;

                        for (int i = 0; i < classes; i++)
                            for (int j = 0; j <= i; j++)
                                C[i][j] = C[j][i] = 0;

                        TP = TN = FP = FN = 0.0;

                        for (int j = 0; j < n; j++) {
                            int backup = S[list[j]];
                            S[list[j]] = 0;

                            if ((label = kNearestNeighbourLargeCache(parm, list[j], S, instances, nearest, labels, k, neighbors_cache, distance_cache, classes,  &ind_margin, inputs, labels[list[j]],
                                         max_distance, data, r, data_index)) == labels[list[j]]) {
                                accuracy += 1.0;
                                if (labels[list[j]] == minority)
                                    TP += 1.0;
                                else
                                    TN += 1.0;
                            } else {
                                if (labels[list[j]] == minority)
                                    FN += 1.0;
                                else
                                    FP += 1.0;
                            }
                            margin += ind_margin;

                            // Leave one out error
                            S[list[j]] = backup;

                            C[labels[list[j]]][label]++;
                            count++;
                        }

                        // Artificial data
                        for (int j = 0; j < ad_n; j++) {
                            if ((label = kNearestNeighbourLargeCache(parm, ad_list[j], S, instances, nearest, labels, k, ad_neighbors_cache, ad_distance_cache, classes, &ind_margin, inputs,
                                         ad_labels[ad_list[j]], max_distance, data, r, data_index)) == ad_labels[ad_list[j]])
                                accuracy += 1.0;
                            margin += ind_margin;

                            C[ad_labels[ad_list[j]]][label]++;
                            count++;
                        }

                        if (parm->performance_measure == pfGMean) {
                            scalar sn = TP+FN > 0.0 ? TP/(TP+FN) : 1.0;
                            scalar sp = TN+FP > 0.0 ? TN/(TN+FP) : 1.0;

                            accuracy = pow(sn*sp, 0.5);
                        } else if (parm->performance_measure == pfAccuracy)
                            accuracy /= count;
                        else if (parm->performance_measure == pfKappa) {
                            scalar theta_1 = 0.0, theta_2 = 0.0;

                            for (int i = 0; i < classes; i++)
                                theta_1 += C[i][i];
                            theta_1 *= count;

                            for (int i = 0; i < classes; i++) {
                                scalar row_i = 0.0, column_i = 0.0;
                                for (int j = 0; j < classes; j++) {
                                    row_i += C[i][j];
                                    column_i += C[j][i];
                                }

                                theta_2 += row_i * column_i;
                            }

                            // Kappa measure
                            accuracy = (theta_1 - theta_2)/(((scalar) count)*count-theta_2);
                        }

                        margin = (margin/(n+ad_n)) * parm->localk_alpha_mar;
                        regularization = acc_by_k[k] * parm->localk_alpha_reg;
                        accuracy *= parm->localk_alpha_acc;

                        if ((accuracy > best_accuracy) || (fabs(accuracy-best_accuracy) < ZERO && margin > best_margin) ||
                            (fabs(accuracy-best_accuracy) < ZERO && fabs(margin-best_margin) < ZERO && regularization > best_regularization)) {
                            best_accuracy = accuracy;
                            best_margin = margin;
                            best_regularization = regularization;

                            best_k = k;
                        }
                    }

                    // If all fitness 0 set to best cross-validated value
                    if (best_k == -1)
                        localk[i] = localk_lambda;
                    else
                        localk[i] = best_k;
                }
            break;

        case kFOLDCV:
            // k-fold cross-validation
            fold_size = instances%parm->localk_nfolds ? instances/parm->localk_nfolds+1 : instances/parm->localk_nfolds;
            MALLOC(instance_list, int, instances, "OptimalLocalK: instance_list");
            MALLOC(s_backup, int, instances, "OptimalLocalK: s_backup");
            for (int i = 0; i < instances; i++)
                instance_list[i] = i;
            RandomizeVector(instance_list, instances);

            for (int i = 0; i < instances; i++)
                if (S[i]) {
                    int n = 0, ad_n = 0, nearest[parm->localk_max_k+1];

                    // Get list of instances which their nearest neighbor is i including itself
                    for (int neighbor = 0; neighbor <= parm->k; neighbor++)
                        for (int j = 0; j < instances; j++)
                            if (neighbors_cache[j][neighbor] == i) {
                                list[n] = j;
                                n++;
                            }

                    // Get list of instances which their nearest neighbor is i for artificial data
                    for (int neighbor = 0; neighbor <= parm->k; neighbor++)
                        for (int j = 0; j < ad_instances; j++)
                            if (ad_neighbors_cache[j][neighbor] == i) {
                                ad_list[ad_n] = j;
                                ad_n++;
                            }

                    // If no neighbors assign cross-validated value
                    if (!parm->leave_no_neighbors_unchanged && n+ad_n == 0) {
                        localk[i] = localk_lambda;
                        no_neighbors_list[no_neighbors_count] = i;
                        no_neighbors_count++;
                        continue;
                    }

                    int best_k = -1;
                    scalar best_fitness = 0.0;
                    for (int k = parm->localk_min_k; k <= parm->localk_max_k; k++) {
                        scalar accuracy = 0.0, regularization, margin = 0.0, ind_margin, fitness, TP, TN, FP, FN, count = 0.0;

                        int C[classes][classes];

                        for (int i = 0; i < classes; i++)
                            for (int j = 0; j <= i; j++)
                                C[i][j] = C[j][i] = 0;

                        TP = TN = FP = FN = 0.0;

                        // Test accuracy
                        for (int fold = 0; fold < parm->localk_nfolds; fold++) {
                            int label;

                            // i-th fold
                            for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++) {
                                s_backup[instance_list[j]] = S[instance_list[j]];
                                S[instance_list[j]] = 0;
                            }

                            for (int j = 0; j < n; j++)
                                if (!S[list[j]]) {
                                    // k-Fold cross-validation error
                                    if ((label = kNearestNeighbourLargeCache(parm, list[j], S, instances, nearest, labels, k, neighbors_cache, distance_cache, classes, &ind_margin, inputs,
                                                 labels[list[j]], max_distance, data, r, data_index)) == labels[list[j]]) {
                                        accuracy += 1.0;
                                        if (labels[list[j]] == minority)
                                            TP += 1.0;
                                        else
                                            TN += 1.0;
                                    } else {
                                        if (labels[list[j]] == minority)
                                            FN += 1.0;
                                        else
                                            FP += 1.0;
                                    }
                                    margin += ind_margin;

                                    C[labels[list[j]]][label]++;
                                    count++;
                                }

                            // Artificial data
                            for (int j = 0; j < ad_n; j++) {
                                if ((label = kNearestNeighbourLargeCache(parm, ad_list[j], S, instances, nearest, labels, k, ad_neighbors_cache, ad_distance_cache, classes, &ind_margin, inputs,
                                             ad_labels[ad_list[j]], max_distance, data, r, data_index)) == ad_labels[ad_list[j]])
                                    accuracy += 1.0;
                                margin += ind_margin;

                                C[ad_labels[ad_list[j]]][label]++;
                                count++;
                            }

                            for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++)
                                S[instance_list[j]] = s_backup[instance_list[j]];
                        }

                        if (parm->performance_measure == pfGMean) {
                            scalar sn = TP+FN > 0.0 ? TP/(TP+FN) : 1.0;
                            scalar sp = TN+FP > 0.0 ? TN/(TN+FP) : 1.0;

                            accuracy = pow(sn*sp, 0.5);
                        } else if (parm->performance_measure == pfAccuracy)
                            accuracy /= count;
                        else if (parm->performance_measure == pfKappa) {
                            scalar theta_1 = 0.0, theta_2 = 0.0;

                            for (int i = 0; i < classes; i++)
                                theta_1 += C[i][i];
                            theta_1 *= count;

                            for (int i = 0; i < classes; i++) {
                                scalar row_i = 0.0, column_i = 0.0;
                                for (int j = 0; j < classes; j++) {
                                    row_i += C[i][j];
                                    column_i += C[j][i];
                                }

                                theta_2 += row_i * column_i;
                            }

                            // Kappa measure
                            accuracy = (theta_1 - theta_2)/(((scalar) count)*count-theta_2);
                        }

                        margin /= n+ad_n*parm->localk_nfolds;
                        regularization = acc_by_k[k];

                        fitness = parm->localk_alpha_acc * accuracy+parm->localk_alpha_reg * regularization+parm->localk_alpha_mar * margin;

                        if (fitness > best_fitness) {
                            best_fitness = fitness;
                            best_k = k;
                        }
                    }

                    // If all fitness 0 set to best cross-validated value
                    if (best_k == -1)
                        localk[i] = localk_lambda;
                    else
                        localk[i] = best_k;
                }
            FREE(instance_list);
            FREE(s_backup);
            break;

        case BS:
        case E632:
        case HOLDOUT:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "BS, E632 or HOLDOUT error estimation");
            break;

        case ArtificialData:
            OnErrorExit(ERR_PARAM, "Not applicable to optimal k");
            break;

        default:
            OnErrorExit(ERR_PARAM, "Bad error estimation method");
    }

    // Assign nearest neighbor k value to no neighbors instances?
    if (parm->localk_no_neigbors_assign_nearest_value && !parm->leave_no_neighbors_unchanged)
        for (int i = 0; i < no_neighbors_count; i++)
            localk[no_neighbors_list[i]] = localk[neighbors_cache[no_neighbors_list[i]][1]];

    // Free distance cache
    if (use_cache) {
        if (parm->localk_populate) {
            for (int i = 0; i < ad_instances; i++) {
                FREE(ad_neighbors_cache[i]);
                FREE(ad_distance_cache[i]);
            }
            FREE(ad_neighbors_cache);
            FREE(ad_distance_cache);
        }
    }

    if (parm->verbose >= 0 && !parm->leave_no_neighbors_unchanged)
        fprintf(parm->out_file, "@Instances with no neighbors: %lf\n", ((scalar)no_neighbors_count)/instances);

    // Free memory
    if (local_s)
        FREE(S);
    FREE(list);
    FREE(no_neighbors_list);
    if (ad_list)
        FREE(ad_list);
    if (parm->localk_populate) {
        FREE(ad_labels);
        for (int i = 0; i < ad_instances; i++)
            FREE(ad_data[i]);
        FREE(ad_data);
    }
}

// Calculates kappa as a measure of classifier performance
scalar KappaForLocalK(const t_Parameters *parm, scalar **data, int instances, int *label, int *localk, int classes, int inputs, scalar **protos, int n_protos, int *p_labels, FILE *out_file, scalar *r,
                      scalar *accuracy, scalar *gmean, int *selection, int **M, int **data_index, int **protos_index)
{
    int predicted, *S;        // Confusion matrix
    scalar kappa, theta_1 = 0.0, theta_2 = 0.0, **distance_to_furthest_neighbor = NULL;

    // Also selection?
    if (!selection) {
        MALLOC(S, int, n_protos, "KappaForLocalK: S");
        for (int i = 0; i < n_protos; i++)
            S[i] = localk[i] ? 1 : 0;
    } else
        S = selection;

    for (int i = 0; i < classes; i++)
        for (int j = 0; j <= i; j++)
            M[i][j] = M[j][i] = 0;

    // Vector of distances to furthest neighbor if needed
    if (parm->symmetric_knn) {
        MALLOC(distance_to_furthest_neighbor, scalar *, parm->max_k, "KappaForLocalK: distance_to_furthest_neighbor");

        for (int i = 0; i < parm->max_k; i++)
            distance_to_furthest_neighbor[i] = NULL;
    }

    // Create confusion matrix
    for (int i = 0; i < instances; i++) {
        // Get nearest neighbor
        int nn_index = -1;
        scalar nn_value = INFINITE, value;

        // Iterate for every prototype
        for (int j = 0; j < n_protos; j++)
            if (S[j] && (value = EuclideanDistanceSparseData(data[i], protos[j], inputs, data_index ? data_index[i] : NULL, protos_index ? protos_index[j] : NULL)/r[j]) < nn_value) {
                nn_value = value;
                nn_index = j;
            }

        int k = localk[nn_index];
        int nearest[k];

        if (parm->symmetric_knn && distance_to_furthest_neighbor[k] == NULL) {
            int neighbors[k];

            MALLOC(distance_to_furthest_neighbor[k], scalar, n_protos, "KappaForLocalK: distance_to_furthest_neighbor[k]");

            for (int j = 0; j < n_protos; j++) {
                kNNCategory(protos[j], protos, NULL, n_protos, k, inputs, p_labels, classes, neighbors, S, NULL, NULL, -1, protos_index, protos_index[j]);

                // Distance to furthest neighbor
                distance_to_furthest_neighbor[k][j] = EuclideanDistanceSparseData(protos[j], protos[neighbors[k-1]], inputs, protos_index ? protos_index[j] : NULL,
                                                      protos_index ? protos_index[neighbors[k-1]] : NULL)/r[neighbors[k-1]];
            }
        }

        predicted = kNNCategory(data[i], protos, NULL, n_protos, k, inputs, p_labels, classes, nearest, S, NULL, NULL, -1, protos_index, data_index[i]);

        M[label[i]][predicted]++;
    }

    // Print confusion matrix
    fprintf(out_file, "Test contingency table\n");
    for (int i = 0; i <= 8*(classes+1); i++)
        fprintf(out_file, "-");
    fprintf(out_file, "\n");
    fprintf(out_file, "        | Predicted\n");
    fprintf(out_file, "Actual  |");
    for (int i = 0; i < classes; i++)
        fprintf(out_file, "   %5d", i+1);
    fprintf(out_file, "\n");
    for (int i = 0; i <= 8*(classes+1); i++)
        fprintf(out_file, "-");
    fprintf(out_file, "\n");
    for (int i = 0; i < classes; i++) {
        fprintf(out_file, "%3d     |", i+1);
        for (int j = 0; j < classes; j++)
            fprintf(out_file, "   %5d", (int) M[i][j]);
        fprintf(out_file, "\n");
    }
    for (int i = 0; i <= 8*(classes+1); i++)
        fprintf(out_file, "-");
    fprintf(out_file, "\n");

    // Obtain kappa
    // Sum of the diagonal
    for (int j = 0; j < classes; j++)
        theta_1 += M[j][j];
    theta_1 *= instances;

    for (int i = 0; i < classes; i++) {
        // Rows/columns total counts
        scalar row_i = 0.0, column_i = 0.0;
        for (int j = 0; j < classes; j++) {
            row_i += M[i][j];
            column_i += M[j][i];
        }

        theta_2 += row_i * column_i;
    }

    // Accuracy?
    if (accuracy) {
        *accuracy = 0.0;
        for (int i = 0; i < classes; i++)
            *accuracy += M[i][i];

        *accuracy /= instances;
    }

    // G-Mean
    if (gmean && classes == 2) {
        // Obtain minority prediction
        int count[2] = { 0, 0 }, minority, majority;

        for (int i = 0; i < instances; i++)
            count[label[i]]++;

        minority = (count[0] > count[1]) ? 1 : 0;
        majority = (minority+1) % 2;

        scalar TP = M[minority][minority];
        scalar TN = M[majority][majority];
        scalar FN = M[minority][majority];
        scalar FP = M[majority][minority];

        scalar specificity = (TN+FP) > ZERO ? TN/(TN+FP) : 0.0;
        scalar sensitivity = (TP+FN) > ZERO ? TP/(TP+FN) : 0.0;

        *gmean = sqrt(specificity*sensitivity);
    }

    if (!selection)
        FREE(S);
    if (parm->symmetric_knn) {
        for (int i = 1; i < parm->max_k; i++)
            if (distance_to_furthest_neighbor[i])
                FREE(distance_to_furthest_neighbor[i]);
        FREE(distance_to_furthest_neighbor);
    }

    kappa = (theta_1-theta_2)/((((scalar) instances)*instances)-theta_2);

    return kappa;
}
