/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

// Learning vector quantization algorithms

#define ALPHA           0.03
#define W               0.3     // also 0.2

t_LVQ_net *LVQCreate(int mi, int inputs)
{
    t_LVQ_net *lvq;

    // Allocate LVQ network
    MALLOC(lvq, t_LVQ_net, 1, "LVQLearning: lvq");
    MALLOC(lvq->labels, int, mi, "LVQLearning: lvq->labels");
    MALLOC(lvq->vectors, scalar *, mi, "LVQLearning: lvq->vectors");
    for (int i = 0; i < mi; i++)
        MALLOC(lvq->vectors[i], scalar, inputs, "LVQLearning: lvq->vectors[i]");

    return lvq;
}

// LVQ2.1 method
// Train a LVQ network
void LVQLearning(t_LVQ_net *lvq, scalar **data, int instances, int *labels, int mi, int iterations, int *label, int inputs, int classes)
{
    int nearest[2], *sample;
    scalar alpha = ALPHA, s, di, dj;

    ALLOCA(sample, int, mi, "LVQLearning: sample");
    for (int i = 0; i < mi; i++)
        sample[i] = 1;

    s = (1.0 - W)/(1.0+W);

    // Training algorithm
    for (int iter = 0; iter < iterations; iter++) {

        // Linear adaptation of alpha
        alpha *= ((scalar)(iterations - iter))/((scalar) iterations);

        int c = RANDOM(0, instances);

        // Find two nearest neighbors
        kNNCategory(data[c], lvq->vectors, NULL, mi, 2, inputs, label, classes, nearest, sample, NULL, NULL, -1, NULL, NULL);

        // Corrections are made only if the two nearest codebook vectors
        // belong to different classes, one of them correct, and if the
        // input entry is located inside a window between the nearest codebook vectors

        // Compute window
        di = MAX(ZERO, EuclideanDistance(data[c], lvq->vectors[nearest[0]], inputs));
        dj = MAX(ZERO, EuclideanDistance(data[c], lvq->vectors[nearest[1]], inputs));

        if ((lvq->labels[nearest[0]] != lvq->labels[nearest[1]])
            && (MIN(di/dj, dj/di) > s)) {
            // Test whether any of the two neighbors is correct
            if (lvq->labels[nearest[0]] == labels[c]) {
                // Update nearest neighbor ...
                for (int i = 0; i < inputs; i++)
                    lvq->vectors[nearest[0]][i] += alpha * (data[c][i] - lvq->vectors[nearest[0]][i]);

                // and nearest enemy
                for (int i = 0; i < inputs; i++)
                    lvq->vectors[nearest[1]][i] -= alpha * (data[c][i] - lvq->vectors[nearest[1]][i]);

            } else if (lvq->labels[nearest[1]] == labels[c]) {
                for (int i = 0; i < inputs; i++)
                    lvq->vectors[nearest[1]][i] += alpha * (data[c][i] - lvq->vectors[nearest[1]][i]);

                // and nearest enemy
                for (int i = 0; i < inputs; i++)
                    lvq->vectors[nearest[0]][i] -= alpha * (data[c][i] - lvq->vectors[nearest[0]][i]);
            }

        }
    }
}

void NearestNeighbourAndEnemy(scalar *instance, scalar **data, int instances, int label, int *labels, int *nearest, int *enemy, int inputs)
{
    scalar d_nearest = INFINITE, d_enemy = INFINITE, value;

    // Get nearest instance form other class
    for (int i = 0; i < instances; i++) {
        if (labels[i] != label) {
            // Compute nearest enemy
            if ((value = EuclideanDistance(data[i], instance, inputs)) < d_enemy) {
                *enemy = i;
                d_enemy = value;
            }
        } else {
            // Compute nearest neighbor
            if ((value = EuclideanDistance(data[i], instance, inputs)) < d_nearest) {
                *nearest = i;
                d_nearest = value;
            }
        }
    }
}

// Test accuracy
scalar TestLVQNetwork(t_LVQ_net *lvq, scalar **data, int instances, int *labels, int inputs)
{
    scalar ok = 0.0;

    for (int i = 0; i < instances; i++)
        if (OutputLVQNetwork(lvq, data[i], inputs) == labels[i])
            ok++;

    return ok/instances;
}


// Classifies instance
int OutputLVQNetwork(t_LVQ_net *lvq, scalar *x, int inputs)
{
    int nearest = -1;
    scalar distance = INFINITE, d;

    // Check nearest prototype
    for (int i = 0; i < lvq->mi; i++)
        if ((d = EuclideanDistance(lvq->vectors[i], x, inputs)) < distance) {
            distance = d;
            nearest = i;
        }

    return lvq->labels[nearest];
}
