/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"
#include "maths.h"

// Minimal Consistent Set
// Belur V. Dasarathy, "Minimal Consistent Set (MCS) Identification for Optimal Nearest Neighbor Decision Systems Design",
// IEEE TRANSACTIONS ON SYSTEMS, MAN, AND CYBERNETICS, VOL. 24, NO. 3, MARCH 1994, 511-517.

int MinimalConsistentSet(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, int inputs, cache_t **neighbors_cache, int classes, int **index, int *subspace,
                         scalar **dist_cache)
{
    int selected = 0, *neighbors, *votes, **voters_list, *candidate, *new_candidate, new_selected;
    scalar *nun_distance;
    // char running[] = {'|', '/', '-', '\\'}; int next = 0;

    // Select all instances
    for (int i = 0; i < instances; i++)
        S[i] = 1;
    selected = instances;

    // Get all instances Nearest Unlike Neighbor (NUN) distance
    MALLOC(neighbors, int, instances, "MinimalConsistentSet: neighbors");
    MALLOC(nun_distance, scalar, instances, "MinimalConsistentSet: nun_distance");
    CALLOC(votes, int, instances, "MinimalConsistentSet: votes");
    CALLOC(voters_list, int *, instances, "MinimalConsistentSet: voters_list");
    for (int i = 0; i < instances; i++)
        MALLOC(voters_list[i], int, instances, "MinimalConsistentSet: voters_list[i]");
    MALLOC(candidate, int, instances, "MinimalConsistentSet: candidate");
    MALLOC(new_candidate, int, instances, "MinimalConsistentSet: new_candidate");

    bool finish;
    int iter = 0;
    do {
        iter++;
        SHOWPROGRESS(parm->show_progress > 1, "MCS iteration %3d with %*d selected instances\n", iter, WIDTH(instances), selected);

        for (int i = 0; i < instances; i++)
            votes[i] = 0;

        for (int i = 0; i < instances; i++) {
            // Obtain all neighbors sorted
            kNearestNeighbourCache(i, S, instances, neighbors, labels, instances, neighbors_cache, data, inputs, classes, false, NULL, NULL, index, subspace);

            int nun = -1;
            for (int j = 0; j < instances && nun == -1; j++)
                if (neighbors[j] != -1 && S[neighbors[j]]) {
                    if (labels[i] != labels[neighbors[j]])
                        nun = neighbors[j];
                    else {
                        // Update list of voters
                        votes[neighbors[j]]++;
                        voters_list[neighbors[j]][votes[neighbors[j]]-1] = i;
                    }
                }

            if (nun != -1)
                nun_distance[i] = dist_cache ? dist_cache[i][nun] : EuclideanDistanceSparseDataSubspace(data[i], data[nun], inputs, index ? index[i] : NULL, index ? index[nun] : NULL, subspace);
            else
                nun_distance[i] = INFINITE;
        }

        // Candidate consistent set instances: either a) or b)
        // a) Already in the consistent set
        // b) Whose inclusion will not create an inconsistency
        for (int i = 0; i < instances; i++) {
            if (S[i])
                candidate[i] = 1;
            else {
                // Check no instance in the selected set is nearer than NUN distance
                bool b = true;
                for (int j = 0; j < instances && b; j++)
                    if (candidate[j] && labels[i] != labels[j] &&
                        (dist_cache ? dist_cache[i][j] : EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace) < nun_distance[j]))
                        b = false;

                candidate[i] = b;
            }

            new_candidate[i] = 0;
        }

        int total_votes = 0;
        new_selected = 0;
        do {
            // SHOWPROGRESS(parm->show_progress > 1, "%c\r", running[next = (next+1)%4]);

            // Get candidate most voted
            int most_voted = -1, most_votes = -1;
            for (int i = 0; i < instances; i++)
                if (candidate[i] && votes[i] > most_votes) {
                    most_voted = i;
                    most_votes = votes[i];
                }

            new_candidate[most_voted] = 1;
            new_selected++;

            // Delete voters of most voted from every votes
            for (int i = 0; i < instances; i++) {
                if (i == most_voted)
                    continue;

                for (int k = 0; k < votes[most_voted]; k++)
                    for (int j = 0; j < votes[i]; j++)
                        if (voters_list[i][j] == voters_list[most_voted][k]) {
                            voters_list[i][j] = voters_list[i][votes[i]-1];
                            votes[i]--;

                            j--;    // Revisit
                        }
            }
            // Most voted
            votes[most_voted] = 0;

            total_votes = 0;
            for (int i = 0; i < instances; i++)
                if (candidate[i])
                    total_votes += votes[i];

            SHOWPROGRESS(parm->show_progress > 1, "MCS inner loop %*d votes remaining\r", WIDTH(instances), total_votes);
        } while (total_votes);

        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        // Next S is the new candidate list
        for (int i = 0; i < instances; i++)
            S[i] = new_candidate[i];

        finish = (selected <= new_selected);
        selected = new_selected;

    } while (!finish); // No change


    FREE(neighbors);
    FREE(nun_distance);
    FREE(votes);
    for (int i = 0; i < instances; i++)
        FREE(voters_list[i]);
    FREE(voters_list);
    FREE(candidate);
    FREE(new_candidate);

    return selected;
}
