/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*
 * Miscellaneous functions for multi-label methods
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "macros.h"
#include "is.h"

// Return the index of the maximum of the vector
int VectorMaxIndex(int *v, int n)
{
    int m = 0;

    for (int i = 1; i < n; i++)
        if (v[i] > v[m])
            m = i;

    return m;
}

void CopyInstance(scalar *dest, scalar *orig, int n)
{
    for (int i = 0; i < n; i++)
        dest[i] = orig[i];
}

// Angle between two vectors
// Use dot product
scalar AngleBetweenVectors(scalar *u, scalar *v, int n)
{
    scalar mod_u = 0.0, mod_v = 0.0, dot = 0.0;

    // Get modules and dot product
    for (int i = 0; i < n; i++) {
        dot += u[i] * v[i];
        mod_u += u[i] * u[i];
        mod_v += v[i] * v[i];
    }

    mod_u = sqrt(mod_u);
    mod_v = sqrt(mod_v);

    scalar cangle = dot/(mod_u * mod_v);

    return acos(cangle);
}

scalar VectorModule(scalar *u, int n)
{
    scalar mod_u = 0.0;

    // Get modules and dot product
    for (int i = 0; i < n; i++)
        mod_u += u[i] * u[i];

    mod_u = sqrt(mod_u);

    return mod_u;
}

void RandomizeForValidation(int *v, int v_ins, int t_ins, int *labels, int classes)
{
    int c, rest, n, m, *list = NULL, *remaining;
    scalar v_percent, t_percent;

    v_percent = ((scalar) v_ins)/(v_ins+t_ins);
    t_percent = 1.0 - v_percent;

    MALLOC(list, int, t_ins+v_ins, "RandomizeForValidation");
    MALLOC(remaining, int, t_ins+v_ins, "RandomizeForValidation");

    // Partition class-wise
    rest = n = m = 0;
    for (int i = 0; i < classes; i++) {
        c = 0;
        for (int j = 0; j < t_ins+v_ins; j++)
            if (labels[j] == i) {
                list[c] = j;
                c++;
            }

        RandomizeVector(list, c);

        // Assign to training set
        int j;
        for (j = 0; j < ((int)(c * t_percent)); j++) {
            v[n] = list[j];
            n++;
        }

        // Assign to validation set
        for (; j < ((int)(c * v_percent))+((int)(c * t_percent)); j++) {
            v[m+t_ins] = list[j];
            m++;
        }

        for (; j < c; j++) {
            remaining[rest] = list[j];
            rest++;
        }
    }

#ifdef AUTOTEST
    if (m+n+rest != t_ins+v_ins)
        OnErrorExit(1, "Validation partition not valid");
#endif

    // Complete with the remaining
    rest = 0;
    for (int i = n; i < t_ins; i++) {
        v[i] = remaining[rest];
        rest++;
    }

    for (int i = m; i < v_ins; i++) {
        v[i+t_ins] = remaining[rest];
        rest++;
    }

#ifdef AUTOTEST
    if (m > v_ins || n > t_ins)
        OnErrorExit(1, "Validation partition not valid");
#endif

    FREE(list)
    FREE(remaining);
}
