/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements Modified Selective Subset Algorithm from:
// R. Barandela, F. J. Ferri and J. S. Sánchez, "Decision boundary preserving prototype selection for nearest neighbor classification", International Journal
// of Pattern Recognition and Artificial Intelligence, vol. 19, no. 6, pp. 787-806, 2005.

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"

int MSSAlgorithm(int *S, scalar **data, int instances, int *label, int inputs, cache_t **neighbors_cache, scalar **distances_cache, int **index, int *subspace)
{
    int *MSS = NULL, *nearest, *list, in_S;
    scalar selected, value, *enemy;

    // Memory
    MALLOC(MSS, int, instances, "MSSAlgorithm: MSS");
    MALLOC(nearest, int, instances, "MSSAlgorithm: nearest");
    MALLOC(list, int, instances, "MSSAlgorithm: list");
    MALLOC(enemy, scalar, instances, "MSSAlgorithm: enemy");

    // Initially S = T (or a previously given subset)
    selected = 0;
    in_S = instances;
    for (int i = 0; i < instances; i++)
        MSS[i] = 0;

    // Modified selective subset
    // Order instances in order of distance from its nearest enemy
    // Compute nearest enemy
    if (neighbors_cache == NULL || distances_cache == NULL) {
        for (int i = 0; i < instances; i++) {
            enemy[i] = INFINITE;
            for (int j = 0; j < instances; j++)
                if (label[j] != label[i] && (value = EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace)) < enemy[i])
                    enemy[i] = value;
        }
    } else {
        for (int i = 0; i < instances; i++) {
            enemy[i] = INFINITE;
            int j = 0;
            while (label[i] == label[neighbors_cache[i][j]])
                j++;
            enemy[i] = distances_cache[i][neighbors_cache[i][j]];
        }
    }

    // Sort instances
    OrderedList(enemy, list, instances, false);

    for (int i = 0; i < instances; i++) {
        bool add = false;

        if (distances_cache == NULL) {
            for (int j = i; j < instances; j++)
                if (S[list[j]] && EuclideanDistanceSparseDataSubspace(data[list[i]], data[list[j]], inputs, index ? index[list[i]] : NULL, index ? index[list[j]] : NULL, subspace) < enemy[list[j]]) {
                    S[list[j]] = 0;
                    in_S--;
                    add = true;
                }
        } else {
            for (int j = i; j < instances; j++)
                if (S[list[j]] && distances_cache[list[i]][list[j]] < enemy[list[j]]) {
                    S[list[j]] = 0;
                    in_S--;
                    add = true;
                }
        }

        if (add) {
            MSS[list[i]] = 1;
            selected++;
        }

        if (in_S == 0)
            break;
    }

    // Copy MSS to S and count selected
    selected = 0;
    for (int i = 0; i < instances; i++)
        if ((S[i] = MSS[i]))
            selected++;

    FREE(MSS);
    FREE(nearest);
    FREE(list);
    FREE(enemy);

    return selected;
}
