/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <string.h>
#include <pvm3.h>
#include <stdio.h>
#include <unistd.h>
#include "pvm-local.h"
#include "is.h"
#include "common.h"
#include "macros.h"

// Evaluates a threshold for votes using the same method that for democratic selection
// using a parallel implementation

void ParallelGetBothThresholdsForImbalancedClass(scalar **data, int instances, int *labels, int minority, int *votes, int *minority_th, int *majority_th,
        char *parameters_file, int processors, scalar *best_coefficient, char *out, FILE *binary, int inputs, int classes)
{
    int *local_labels = NULL, tids[processors], *partition, *local_votes = NULL, previous_largest = 0, largest = 0, **sel_slave, **TP_slave,
         **TN_slave, **FP_slave, **FN_slave, **round_sel, **round_TP, **round_TN, **round_FP, **round_FN;
    scalar **local_data = NULL, **G_mean, **r, **coef;

    // Number of subsets
    int n_subsets = instances/gl_eval_subset_size;
    // Add remainder
    if (instances % gl_eval_subset_size)
        n_subsets++;

    if (n_subsets == 1)
        n_subsets++;

    processors = MIN(n_subsets, processors);

    // Parameters for spawning process
    char *args[3];
    MALLOC(args[0], char, strlen(parameters_file)+1, "main");
    strcpy(args[0], parameters_file);       // Parameters file name
    args[1] = getcwd(NULL, 0);      // Current directory for slaves
    args[2] = NULL;

    // Avoid running processes in master node
    int numt = pvm_spawn("eval-ci-wrapper", args, PvmTaskDefault | PvmTaskHost | PvmHostCompl, "skynet.uco.es", processors, tids);
    //                    PvmTaskDebug | PvmTaskDefault/*| PvmTaskHost*/, NULL /*"skynet.uco.es"*/, processors, tids);

    if (numt != processors)
        OnErrorExit(1, "ParallelGetBothThresholdsForImbalancedClass: Not enough processes spawned");

    // Start guardian and send info
    int guardian = InitGuardian(numt, tids, "eval-ci-wrapper", out);

    // Initialize values
    // Get maximum and minimum number of votes
    int min_min_threshold = 1000000, min_max_threshold = -1000000, maj_min_threshold = 1000000, maj_max_threshold = -1000000;

    for (int i = 0; i < instances; i++) {
        if (labels[i] == minority) {
            if (min_min_threshold > votes[i])
                min_min_threshold = votes[i];
            else if (min_max_threshold < votes[i])
                min_max_threshold = votes[i];
        } else {
            if (maj_min_threshold > votes[i])
                maj_min_threshold = votes[i];
            else if (maj_max_threshold < votes[i])
                maj_max_threshold = votes[i];
        }
    }

    int min_n_thresholds = min_max_threshold - min_min_threshold+1;
    int maj_n_thresholds = maj_max_threshold - maj_min_threshold+1;

    // Thresholds to evaluate
    if (gl_verbose >= 0)
        printf("%d thresholds to evaluate: %d x %d\n", (min_max_threshold-min_min_threshold+1) * (maj_max_threshold-maj_min_threshold+1),
               (min_max_threshold-min_min_threshold+1), (maj_max_threshold-maj_min_threshold+1));
    fprintf(gl_out_file, "%d thresholds to evaluate: %d x %d\n", (min_max_threshold-min_min_threshold+1) * (maj_max_threshold-maj_min_threshold+1),
            (min_max_threshold-min_min_threshold+1), (maj_max_threshold-maj_min_threshold+1));
    fflush(gl_out_file);

    MALLOC(partition, int, instances, "ParallelGetBothThresholdsForImbalancedClass: partition");

    MALLOC(r, scalar *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: r");
    MALLOC(coef, scalar *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: coef");
    MALLOC(G_mean, scalar *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: G_mean");
    MALLOC(sel_slave, int *, min_n_thresholds,  "ParallelGetBothThresholdsForImbalancedClass: sel_slave");
    MALLOC(TP_slave, int *, min_n_thresholds,  "ParallelGetBothThresholdsForImbalancedClass: TP_slave");
    MALLOC(TN_slave, int *, min_n_thresholds,  "ParallelGetBothThresholdsForImbalancedClass: TN_slave");
    MALLOC(FP_slave, int *, min_n_thresholds,  "ParallelGetBothThresholdsForImbalancedClass: FP_slave");
    MALLOC(FN_slave, int *, min_n_thresholds,  "ParallelGetBothThresholdsForImbalancedClass: FN_slave");
    MALLOC(round_sel, int *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_sel");
    MALLOC(round_TP, int *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_TP");
    MALLOC(round_TN, int *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_TN");
    MALLOC(round_FP, int *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_FP");
    MALLOC(round_FN, int *, min_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_FN");

    r -= min_min_threshold;
    coef -= min_min_threshold;
    G_mean -= min_min_threshold;
    sel_slave -= min_min_threshold;
    TP_slave -= min_min_threshold;
    TN_slave -= min_min_threshold;
    FP_slave -= min_min_threshold;
    FN_slave -= min_min_threshold;
    round_sel -= min_min_threshold;
    round_TP -= min_min_threshold;
    round_TN -= min_min_threshold;
    round_FP -= min_min_threshold;
    round_FN -= min_min_threshold;

    for (int i = min_min_threshold; i <= min_max_threshold; i++) {
        MALLOC(r[i], scalar, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: r[i]");
        MALLOC(coef[i], scalar, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: coef[i]");
        MALLOC(G_mean[i], scalar, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: G_mean[i]");
        MALLOC(sel_slave[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: sel_slave[i]");
        MALLOC(TP_slave[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: TP_slave[i]");
        MALLOC(TN_slave[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: TN_slave[i]");
        MALLOC(FP_slave[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: FP_slave[i]");
        MALLOC(FN_slave[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: FN_slave[i]");
        MALLOC(round_sel[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_sel[i]");
        MALLOC(round_TP[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_TP[i]");
        MALLOC(round_TN[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_TN[i]");
        MALLOC(round_FP[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_FP[i]");
        MALLOC(round_FN[i], int, maj_n_thresholds, "ParallelGetBothThresholdsForImbalancedClass: round_FN[i]");

        r[i] -= maj_min_threshold;
        coef[i] -= maj_min_threshold;
        G_mean[i] -= maj_min_threshold;
        sel_slave[i] -= maj_min_threshold;
        TP_slave[i] -= maj_min_threshold;
        TN_slave[i] -= maj_min_threshold;
        FP_slave[i] -= maj_min_threshold;
        FN_slave[i] -= maj_min_threshold;
        round_sel[i] -= maj_min_threshold;
        round_TP[i] -= maj_min_threshold;
        round_TN[i] -= maj_min_threshold;
        round_FP[i] -= maj_min_threshold;
        round_FN[i] -= maj_min_threshold;

        for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
            r[i][j] = G_mean[i][j] = 0.0;
    }

    // Perform rounds
    for (int round = 0; round < gl_demo_eval_rounds; round++) {
        int count[n_subsets];
        if (gl_huge_dataset)
            RandomHugePartition(partition, count, instances, n_subsets, gl_subset_size);
        else
            // Perform partition
            switch (gl_partition) {
                case pRANDOM:
                case pDecoupled:
                case peBagging:
                case pkMeans:
                case pRPROTOS:
                case pkClustering:
                case pBOARD:
                    // Random partition[0]
                    if (gl_external_stratified_approach)
                        RandomPartitionKCD(partition, count, instances, n_subsets, labels, classes);
                    else
                        RandomPartition(partition, count, instances, n_subsets, labels, classes);
                    break;

                case pEQUAL:
                    if (gl_external_stratified_approach)
                        PartitionateEqualKCD(partition, count, data, instances, n_subsets, labels, inputs, classes);
                    else
                        PartitionateEqual(partition, count, data, instances, n_subsets, labels, inputs, classes);
                    break;

                case p2DGrandTour:
                case pOneInput:
                case pGrandTour:
                    if (gl_external_stratified_approach)
                        PartitionateOneInputKCD(partition, count, data, instances, n_subsets, labels, inputs, classes);
                    else
                        PartitionateOneInput(partition, count, data, instances, n_subsets, labels, inputs, classes);
                    break;
            }

        // Allocate memory for largest subset
        largest = count[0];
        for (int i = 1; i < n_subsets; i++)
            if (count[i] > largest)
                largest = count[i];

        // Only reallocate if necessary
        if (largest != previous_largest) {
            // Free memory no longer needed
            if (gl_huge_dataset)
                for (int i = largest; i < previous_largest; i++)
                    FREE(local_data[i]);

            REALLOC(local_data, scalar *, largest, "ParallelGetBothThresholdsForImbalancedClass: local_data");
            REALLOC(local_labels, int, largest, "ParallelGetBothThresholdsForImbalancedClass: local_labels");
            REALLOC(local_votes, int, largest, "ParallelGetBothThresholdsForImbalancedClass: local_votes");

            // If huge dataset need memory for reading from file
            if (gl_huge_dataset)
                for (int i = previous_largest; i < largest; i++)
                    MALLOC(local_data[i], scalar, inputs, "ParallelGetBothThresholdsForImbalancedClass: local_data[i]");
        }

        for (int i = min_min_threshold; i <= min_max_threshold; i++)
            for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
                round_sel[i][j] = round_TP[i][j] = round_TN[i][j] = round_FP[i][j] = round_FN[i][j] = 0;

        // Evaluate for every subset
        for (int subset = 0; subset < n_subsets; subset += processors) {
            for (int p = 0; p < processors && subset+p < n_subsets; p++) {
                int cursor = 0;
                // Get subset
                if (gl_huge_dataset) {
                    cursor = ReadBlockFromBinaryFile(local_data, local_labels, inputs, partition[subset+p], partition[subset+p+1], instances, binary,
                                                     NULL);
                    int next = 0;
                    for (int i = partition[subset+p]; i < partition[subset+p+1]; i = (i+1) % instances) {
                        local_votes[next] = votes[i];
                        next++;
                    }
                } else {
                    for (int j = 0; j < instances; j++)
                        if (partition[j] == subset+p) {
                            local_data[cursor] = data[j];
                            local_labels[cursor] = labels[j];
                            local_votes[cursor] = votes[j];
                            cursor++;
                        }
                }

                // Send data
                pvm_initsend(PvmDataDefault);           // PvmDataDefault 0 XDR
                pvm_pkint(&cursor, 1, 1);     // Number of instances
                pvm_pkint(&inputs, 1, 1);     // Number of instances
                pvm_pkint(&gl_k, 1, 1);
                pvm_pkint(&classes, 1, 1);
                pvm_pkint(&minority, 1, 1);
                pvm_pkint(&min_min_threshold, 1, 1);
                pvm_pkint(&min_max_threshold, 1, 1);
                pvm_pkint(&maj_min_threshold, 1, 1);
                pvm_pkint(&maj_max_threshold, 1, 1);
                pvm_pkint(local_votes, cursor, 1);    // Votes
                pvm_pkint(local_labels, cursor, 1);    // Labels
                for (int i = 0; i < cursor; i++)
                    pvm_pkscalar(local_data[i], inputs, 1);
                pvm_send(tids[p], SENT_BY_MASTER);
            }

            // Get values
            for (int p = 0; p < processors && subset+p < n_subsets; p++) {
                // Receive data
                pvm_recv(-1, SENT_BY_SLAVE);    // Receive from any slave

                for (int i = min_min_threshold; i <= min_max_threshold; i++) {
                    pvm_upkint(sel_slave[i]+maj_min_threshold, maj_n_thresholds, 1);
                    pvm_upkint(TP_slave[i]+maj_min_threshold, maj_n_thresholds, 1);
                    pvm_upkint(TN_slave[i]+maj_min_threshold, maj_n_thresholds, 1);
                    pvm_upkint(FP_slave[i]+maj_min_threshold, maj_n_thresholds, 1);
                    pvm_upkint(FN_slave[i]+maj_min_threshold, maj_n_thresholds, 1);
                }

                // Get coefficient for every threshold
                for (int min_th = min_min_threshold; min_th <= min_max_threshold; min_th++)
                    for (int max_th = maj_min_threshold; max_th <= maj_max_threshold; max_th++) {
                        // Evaluate
                        round_sel[min_th][max_th] += sel_slave[min_th][max_th];
                        round_TP[min_th][max_th] += TP_slave[min_th][max_th];
                        round_TN[min_th][max_th] += TN_slave[min_th][max_th];
                        round_FP[min_th][max_th] += FP_slave[min_th][max_th];
                        round_FN[min_th][max_th] += FN_slave[min_th][max_th];
                    }
                if (gl_verbose >= 0) {
                    printf("Evaluation %d/%d\n", subset+p+1, n_subsets);
                    fflush(stdout);
                }
            }
        }

        // Average
        for (int min_th = min_min_threshold; min_th <= min_max_threshold; min_th++)
            for (int maj_th = maj_min_threshold; maj_th <= maj_max_threshold; maj_th++) {
                r[min_th][maj_th] += 1.0 - ((scalar) round_sel[min_th][maj_th]) /instances;

                scalar specificity = (round_TN[min_th][maj_th]+round_FP[min_th][maj_th]) > ZERO
                                     ? ((scalar) round_TN[min_th][maj_th])/(round_TN[min_th][maj_th]+round_FP[min_th][maj_th]) : 0.0;
                scalar sensitiviy = (round_TP[min_th][maj_th]+round_FN[min_th][maj_th]) > ZERO
                                    ? ((scalar) round_TP[min_th][maj_th])/(round_TP[min_th][maj_th]+round_FN[min_th][maj_th]) : 0.0;
                G_mean[min_th][maj_th] += pow(specificity*sensitiviy, 0.5);
            }

        for (int i = min_min_threshold; i <= min_max_threshold; i++)
            for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
                if (round_TP[i][j]+round_TN[i][j]+round_FP[i][j]+round_FN[i][j] != instances)
                    OnErrorExit(1, "Bad count of instances");
    }

    // Get best one
    *minority_th = min_min_threshold;
    *majority_th = maj_min_threshold;
    for (int i = min_min_threshold; i <= min_max_threshold; i++)
        for (int j = maj_min_threshold; j <= maj_max_threshold; j++) {
            // First average
            r[i][j] /= gl_demo_eval_rounds;
            G_mean[i][j] /= gl_demo_eval_rounds;

            coef[i][j] = gl_alpha_acc * G_mean[i][j]+(1.0 - gl_alpha_acc) * r[i][j];

            if (coef[i][j] > coef[*minority_th][*majority_th]) {
                *minority_th = i;
                *majority_th = j;
            }
        }

    *best_coefficient = coef[*minority_th][*majority_th];

    if (gl_verbose > 0) {
        fprintf(gl_out_file, "THRESHOLDS TRAINING INFO\nMinority\tMajority\tG\tReduct\tCoefficient\n");
        for (int i = min_min_threshold; i <= min_max_threshold; i++)
            for (int j = maj_min_threshold; j <= maj_max_threshold; j++)
                fprintf(gl_out_file, "%3d\t\t%3d\t\t%6.4lf\t%6.4lf\t%6.4lf %s\n", i, j, G_mean[i][j], r[i][j], coef[i][j],
                        (i == *minority_th && j == *majority_th) ? "(*)" : "");
    }

    // Free memory
    if (gl_huge_dataset)
        for (int i = 0; i < largest; i++)
            FREE(local_data[i]);
    FREE(local_data);
    FREE(local_labels);
    FREE(local_votes);
    FREE(partition);
    for (int i = min_min_threshold; i <= min_max_threshold; i++) {
        free(r[i]+maj_min_threshold);
        free(coef[i]+maj_min_threshold);
        free(G_mean[i]+maj_min_threshold);
        free(sel_slave[i]+maj_min_threshold);
        free(TP_slave[i]+maj_min_threshold);
        free(TN_slave[i]+maj_min_threshold);
        free(FP_slave[i]+maj_min_threshold);
        free(FN_slave[i]+maj_min_threshold);
        free(round_sel[i]+maj_min_threshold);
        free(round_TP[i]+maj_min_threshold);
        free(round_TN[i]+maj_min_threshold);
        free(round_FP[i]+maj_min_threshold);
        free(round_FN[i]+maj_min_threshold);
    }
    free(r+min_min_threshold);
    free(coef+min_min_threshold);
    free(G_mean+min_min_threshold);
    free(sel_slave+min_min_threshold);
    free(TP_slave+min_min_threshold);
    free(TN_slave+min_min_threshold);
    free(FP_slave+min_min_threshold);
    free(FN_slave+min_min_threshold);
    free(round_sel+min_min_threshold);
    free(round_TP+min_min_threshold);
    free(round_TN+min_min_threshold);
    free(round_FP+min_min_threshold);
    free(round_FN+min_min_threshold);

    // Kill spawned processes and guardian
    pvm_kill(guardian);
    for (int i = 0; i < processors; i++)
        if (pvm_kill(tids[i]))
            printf("Process %d not killed\n", tids[i]);
}
