
/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas and                     *
 *   Juan Antonio Romero del Castillo                                      *
 *   npedrajas@uco.es, aromero@uco.es                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements Democratic/Oligarchic Instance Selection wrapper

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <stdbool.h>
#include <string.h>
#include <pvm3.h>
#include "macros.h"
#include "common.h"
#include "pvm-local.h"
#include "is.h"
#include "imbalance.h"

#define BLOCK   100

// Obtain the confusion matrix for NN in a parallel way

void GetNNConfusionMatrix(const t_Parameters *parm, int **M, scalar **test_data, int test_instances, int *test_labels, scalar **train_data, int train_instances, int *train_labels, int classes,
                          int inputs, int slaves, char *out_file, int **train_index, int **test_index, int *prediction)
{
    int n_slaves, guardian;

    // Adjust number of slaves to a minimum of BLOCK
    if (train_instances/slaves < BLOCK) {
        n_slaves = MAX(1, (int) ceil(((scalar) train_instances)/BLOCK));
        SHOWPROGRESS(true, "Number of slaves adapted to: %d\n", n_slaves);
    } else
        n_slaves = slaves;

    // Parallel construction of the M matrix
    int *tids, *partition, category[BLOCK][n_slaves], count;
    scalar distance[BLOCK][n_slaves];

    // Start slaves
    MALLOC(tids, int, n_slaves, "main: tids");
    PVMSpawn(n_slaves, "wrapper-knn", tids, NULL, &guardian, out_file, parm->show_progress);
    SHOWPROGRESS(parm->show_progress > 1, "Started %d slaves for performing instance selection\n", n_slaves);

    // Sent a portion of training data to every slave
    MALLOC(partition, int, train_instances, "GetNNConfusionMatrix: partition");
    int pcount[n_slaves];
    RandomPartition(partition, pcount, train_instances, n_slaves);

    // Broadcast whether to use sparse format
    pvm_initsend(PvmDataDefault);
    int sparse = (int)(train_index != NULL);
    pvm_pkint(&sparse, 1, 1);
    pvm_mcast(tids, n_slaves, SENT_BY_MASTER);

    for (int i = 0; i < n_slaves; i++) {
        // Send size
        pvm_initsend(PvmDataDefault);
        pvm_pkint(&pcount[i], 1, 1);
        pvm_pkint(&inputs, 1, 1);
        pvm_pkint(&classes, 1, 1);
        for (int j = 0; j < train_instances; j++)
            if (partition[j] == i) {
                if (train_index == NULL) {
                    pvm_pkscalar(train_data[j], inputs, 1);

                } else {
                    int c = 0;
                    while (train_index[j][c] != -1)
                        c++;
                    pvm_pkint(&c, 1, 1);
                    pvm_pkscalar(train_data[j], c, 1);
                    pvm_pkint(train_index[j], c+1, 1);
                }
                pvm_pkint(&train_labels[j], 1, 1);
            }
        pvm_send(tids[i], SENT_BY_MASTER);
    }

    FREE(partition);

    SHOWPROGRESS(true, "Test instances: %d/%0*d/%0*d", test_instances, WIDTH(test_instances), 0, WIDTH(test_instances), test_instances);
    fflush(stdout);

    for (int i = 0; i < classes; i++) {
        M[i][i] = 0;
        for (int j = 0; j < i; j++)
            M[i][j] = M[j][i] = 0;
    }

    // Test instances
    for (int i = 0; i < test_instances; i += BLOCK) {
        pvm_initsend(PvmDataDefault);
        // Send a block of at most BLOCK size
        count = MIN(BLOCK, test_instances-i);
        pvm_pkint(&count, 1, 1);
        if (train_index == NULL) {
            for (int c = 0; c < count; c++)
                pvm_pkscalar(test_data[i+c], inputs, 1);
        } else {
            for (int c = 0; c < count; c++) {
                int k = 0;
                while (test_index[i+c][k] != -1)
                    k++;
                pvm_pkint(&k, 1, 1);
                pvm_pkscalar(test_data[i+c], k, 1);
                pvm_pkint(test_index[i+c], k+1, 1);
            }
        }
        pvm_mcast(tids, n_slaves, SENT_BY_MASTER);

        // Receive nearest distance and
        for (int j = 0; j < n_slaves; j++) {
            pvm_recv(tids[j], SENT_BY_SLAVE);
            for (int c = 0; c < count; c++) {
                pvm_upkint(&category[c][j], 1, 1);
                pvm_upkscalar(&distance[c][j], 1, 1);
            }
        }

        // Assign to nearest neighbor
        for (int c = 0; c < count; c++) {
            int pred = category[c][IndexOfMinimumValue(distance[c], n_slaves)];
            M[test_labels[i+c]][pred]++;
            if (prediction)
                prediction[i+c] = pred;
            SHOWPROGRESS(true, "\rTest instances: %d/%0*d/%0*d", test_instances, WIDTH(test_instances), i+c+1, WIDTH(test_instances), test_instances-i-c-1);
        }

        fflush(stdout);
    }
    SHOWPROGRESS(true, "%c", '\n');

    // Kill spawned processes and guardian
    pvm_kill(guardian);
    for (int i = 0; i < n_slaves; i++)
        pvm_kill(tids[i]);
}

int IndexOfMinimumValue(scalar *v, int n)
{
    int index = 0;

    for (int i = 1; i < n; i++)
        if (v[index] > v[i])
            index = i;

    return index;
}
