/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdbool.h>
#include <pvm3.h>
#include <string.h>
#include <unistd.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "pvm-local.h"

time_t ParallelInstanceSelectionForAllStrata(const t_Parameters *parm, int n_slaves, bool spawn, bool kill, char *parameters, scalar ***is_data,  int *is_instances, int **is_labels, int **is_S,
        int ***is_index, int n_strata, int minority, char *out_file, int inputs, int classes, scalar ratio)
{
    int *tids = NULL, numt, guardian = 0;
    time_t init_time, sending_time = 0;

    SHOWPROGRESS(parm->show_progress > 0, "Internal strata: %d/%0*d/%0*d", n_strata, WIDTH(n_strata), 0, WIDTH(n_strata), n_strata);

    if (spawn) {
        // PVM Spawning processes if not already running
        char *args[4];
        MALLOC(args[0], char, strlen(parameters)+1, "main: args[0]");
        strcpy(args[0], parameters);       // Parameters file_name
        MALLOC(args[1], char, 5, "main: args[1]");
        strcpy(args[1], "none");
        args[2] = getcwd(NULL, 0);      // Current directory for slaves
        args[3] = NULL;

        // Start slaves
        MALLOC(tids, int, n_slaves, "main: tids");
        // Avoid running processes in Skynet master node
        numt = pvm_spawn("wrapper-is-strata", args, PvmTaskHost | PvmHostCompl, "skynet.uco.es", n_slaves, tids);
        // numt = pvm_spawn("wrapper-is-strata", args, PvmTaskDefault+PvmTaskDebug, NULL, n_slaves, tids);
        if (numt != n_slaves)
            OnErrorExit(ERR_PVM, "Processes not spawned correctly");

        // Start guardian and send info
        guardian = InitGuardian(numt, tids, "wrapper-is-strata", out_file);

        SHOWPROGRESS(parm->show_progress > 1, "Started %d slaves for performing instance selection\n", n_slaves);
    }

    // Instance selection made by slaves
    // Send strata to first slaves
    for (int p = 0; p < n_slaves && p < n_strata; p++) {
        init_time = time(NULL);

        SendDataToSlave(tids[p], p, is_instances[p], inputs, classes, minority, ratio, is_labels[p], is_index ? is_index[p] : NULL, is_data[p]);

        sending_time += time(NULL) - init_time;

        SHOWPROGRESS(parm->show_progress > 1, "SENT data to slave %d\n", tids[p]);
    }

    // Receive results and send more strata if needed
    int sent = MIN(n_slaves, n_strata);
    int received = 0;

    while (received < n_strata) {
        int slave_sent, selected, bytes, mstag, tid, instances;
        // Wait for selection of any slave
        int bufid = pvm_recv(-1, SENT_BY_SLAVE);

        received++;

        pvm_upkint(&slave_sent, 1, 1);
        pvm_upkint(&instances, 1, 1);
        // Check ok
        if (instances != is_instances[slave_sent])
            OnErrorExit(ERR_UNSPECTED_RESULT, "Instances sent by slave do not match with stratum");
        pvm_upkint(is_S[slave_sent], is_instances[slave_sent], 1);
        pvm_upkint(&selected, 1, 1);

        // Send time expended in seconds
        time_t tid_time;
        pvm_upkint((int *) &tid_time, 1, 1);

        SHOWPROGRESS(parm->show_progress > 0, "\rInternal strata: %*d/%d, %d/%0*d/%0*d", WIDTH(is_instances[slave_sent]), selected, is_instances[slave_sent],  n_strata, WIDTH(n_strata), received,
                     WIDTH(n_strata), n_strata-received);

        // Get tid of sending slave
        pvm_bufinfo(bufid, &bytes, &mstag, &tid);
        SHOWPROGRESS(parm->show_progress > 1, "RECEIVED selection from slave %d\n", tid);

        // Send more?
        if (sent < n_strata) {
            init_time = time(NULL);

            // Send info to slave
            SendDataToSlave(tid, sent, is_instances[sent], inputs, classes, minority, ratio, is_labels[sent], is_index ? is_index[sent] : NULL, is_data[sent]);

            SHOWPROGRESS(parm->show_progress > 1, "SENT data to slave %d\n", tid);

            sending_time += time(NULL) - init_time;

            sent++;
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    if (kill) {
        // Kill spawned processes and guardian
        pvm_kill(guardian);
        for (int i = 0; i < n_slaves; i++)
            pvm_kill(tids[i]);
    }

    return sending_time;
}

void SendDataToSlave(int tid, int stratum, int instances, int inputs, int classes, int minority, scalar ratio, int *labels, int **index, scalar **data)
{
    int use_index = 0;

    // Send info to slave
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&stratum, 1, 1);
    pvm_pkint(&instances, 1, 1);
    pvm_pkint(&inputs, 1, 1);
    pvm_pkint(&classes, 1, 1);
    pvm_pkint(&minority, 1, 1);
    pvm_pkscalar(&ratio, 1, 1);

    // Labels
    pvm_pkint(labels, instances, 1);

    // Index?
    use_index = (index != NULL);
    pvm_pkint(&use_index, 1, 1);

    if (index) {
        for (int i = 0; i < instances; i++) {
            int count = 0;

            // Count size and send data and index
            for (; index[i][count] != -1; count++);
            pvm_pkint(&count, 1, 1);

            pvm_pkint(index[i], count+1, 1);
            pvm_pkscalar(data[i], count, 1);
        }
    } else
        // Dense data
        for (int i = 0; i < instances; i++)
            pvm_pkscalar(data[i], inputs, 1);

    pvm_send(tid, SENT_BY_MASTER);
}
