/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements parallel k-NN using knn-wrapper

#include <stdio.h>
#include <pvm3.h>
#include <unistd.h>
#include <string.h>
#include "pvm-local.h"
#include "is.h"
#include "macros.h"
#include "common.h"

scalar ParallelkNN(const t_Parameters *parm, int *list, int p_instances, int instances, int n_subsets, int processors, char *testing_file, char *training_file, scalar *sensitivity,
                   scalar *specificity, scalar *precision, scalar *false_negatives, char *out, int inputs, int classes)
{
    int tids[processors], *labels, ok_slave, minority, TP_subset, TN_subset, FN_subset, FP_subset;
    scalar *instance, ok = 0.0, TP, TN, FP, FN;
    bool first_time = true;

    // Parameters for spawning process
    char *args[2];
    MALLOC(args[0], char, 1000, "main");
    strcpy(args[0], training_file);
    args[1] = NULL;

    // Avoid running processes in master node
    int numt = pvm_spawn("knn-wrapper", args, PvmTaskHost | PvmHostCompl, "skynet.uco.es", processors, tids);
    //      PvmTaskDefault+PvmTaskDebug, "skynet.uco.es", processors, tids);
    if (numt != processors)
        OnErrorExit(1, "ParallelkNN: Not enough processes spawned");

    // Start guardian and send info
    int guardian = InitGuardian(numt, tids, "wrapper", out);

    // Send training info to slaves
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&p_instances, 1, 1);
    pvm_pkint(&inputs, 1, 1);
    pvm_pkint(&((t_Parameters *) parm)->k, 1, 1);
    pvm_pkint(&classes, 1, 1);
    pvm_pkint(list, p_instances, 1);
    pvm_mcast(tids, processors, SENT_BY_MASTER);

    // Confirm training data read ok
    for (int p = 0; p < processors; p++) {
        int bufid, bytes, type, source, done;

        // Receive data
        bufid = pvm_recv(-1, SENT_BY_SLAVE);    // Receive from any slave

        pvm_upkint(&done, 1, 1);

        if (parm->verbose >= 0) {
            // Received from?
            if (pvm_bufinfo(bufid, &bytes, &type, &source))
                OnErrorExit(1, "ParallelkNN: Error reading PVM buffer");
            printf("Task %d finished reading training data\n", source);
            fflush(stdout);
        }
    }

    // Open the test file
    FILE *test;
    if ((test = fopen(testing_file, "rt")) == NULL) {
        printf("Testing file: %s\n", testing_file);
        OnErrorExit(1, "Cannot open testing file");
    }

    n_subsets = instances/parm->subset_size;
    // Add remainder
    if (instances % parm->subset_size)
        n_subsets++;
    int size =  instances/n_subsets;

    MALLOC(instance, scalar, inputs, "ParallelkNN: instance");
    MALLOC(labels, int, size, "ParallelkNN: labels");

    // If two classes evaluate also 2-class measures
    if (classes == 2) {
        // Obtain minority prediction
        int count[2] = { 0, 0 }, label;

        for (int i = 0; i < instances; i++) {
            ReadNextInstanceAndLabel(instance, &label, &inputs, &classes, test, &first_time);
            count[label]++;
        }

        minority = (count[0] > count[1]) ? 1 : 0;
        TP = TN = FP = FN = 0.0;

        first_time = true;

        // Send to slaves
        pvm_initsend(PvmDataDefault);
        pvm_pkint(&minority, 1, 1);
        pvm_mcast(tids, processors, SENT_BY_MASTER);
    }

    // Evaluate for every subset
    int cursor = 0;
    for (int subset = 0; subset < n_subsets; subset += processors) {
        for (int p = 0; p < processors && subset+p < n_subsets; p++) {
            // Send subset to slave
            pvm_initsend(PvmDataDefault);
            pvm_pkint(&size, 1, 1);
            for (int i = 0; i < size && cursor < instances; i++) {
                ReadNextInstanceAndLabel(instance, &labels[i], &inputs, &classes, test, &first_time);
                pvm_pkscalar(instance, inputs, 1);
                cursor++;
            }
            pvm_pkint(labels, size, 1);
            pvm_send(tids[p], SENT_BY_MASTER);
        }

        // Check no problem with processes
        for (int p = 0; p < processors; p++)
            if (pvm_pstat(tids[p]) == PvmNoTask) {
                printf("Died after %d/%d subsets processed\n", subset, n_subsets);
                OnErrorExit(ERR_PVM, "PVM Task died unexpectedly");
            }

        if (parm->verbose >= 0)
            printf("Sent subsets [%d, %d], Waiting...\n", subset+1, subset+processors > n_subsets ? n_subsets : subset+processors);

        // Get k-NN error
        for (int p = 0; p < processors && subset+p < n_subsets; p++) {
            // Receive data
            pvm_recv(-1, SENT_BY_SLAVE);    // Receive from any slave

            // Get error for every subset
            pvm_upkint(&ok_slave, 1, 1);
            ok += ok_slave;

            if (classes == 2) {
                pvm_upkint(&TP_subset, 1, 1);
                pvm_upkint(&TN_subset, 1, 1);
                pvm_upkint(&FP_subset, 1, 1);
                pvm_upkint(&FN_subset, 1, 1);

                TP += TP_subset;
                TN += TN_subset;
                FP += FP_subset;
                FN += FN_subset;
            }

            if (parm->verbose >= 0) {
                printf("Evaluation %d/%d: %d/%d\n", subset+p+1, n_subsets, ok_slave, size);
                fflush(stdout);
            }
        }
    }

    // Evaluate remaining
    if (cursor < instances) {
        size = instances - cursor;
        // Send subset to slave
        pvm_initsend(PvmDataDefault);    // PvmDataDefault 0 XDR
        pvm_pkint(&size, 1, 1);
        for (int i = 0; i < size; i++) {
            ReadNextInstanceAndLabel(instance, &labels[i], &inputs, &classes, test, &first_time);
            pvm_pkscalar(instance, inputs, 1);
            cursor++;
        }
        pvm_pkint(labels, size, 1);
        pvm_send(tids[0], SENT_BY_MASTER);

        // Receive data
        pvm_recv(-1, SENT_BY_SLAVE);    // Receive from any slave

        // Get error for every subset
        pvm_upkint(&ok_slave, 1, 1);
        ok += ok_slave;

        if (classes == 2) {
            pvm_upkint(&TP_subset, 1, 1);
            pvm_upkint(&TN_subset, 1, 1);
            pvm_upkint(&FP_subset, 1, 1);
            pvm_upkint(&FN_subset, 1, 1);

            TP += TP_subset;
            TN += TN_subset;
            FP += FP_subset;
            FN += FN_subset;
        }

        if (parm->verbose >= 0) {
            printf("Evaluation LAST: %d/%d\n", ok_slave, size);
            fflush(stdout);
        }
        n_subsets++;
    }

    fclose(test);

    ok /= instances;

    // Kill processes and guardian
    pvm_kill(guardian);
    for (int i = 0; i < processors; i++)
        pvm_kill(tids[i]);

    if (classes == 2) {
        *specificity = (TN+FP > ZERO) ? TN/(TN+FP) : 0.0;
        *sensitivity = (TP+FN > ZERO) ? TP/(TP+FN) : 0.0;
        *precision = (TP+FP > ZERO) ? TP/(TP+FP) : 0.0;
        *false_negatives = (TP+FN > ZERO) ?  FN/(TP+FN) : 0.0;

        if (parm->verbose >= 0) {
            fprintf(parm->out_file, "Minority prediction: %d\n", minority);
            fprintf(parm->out_file, "Positive instances: %d\nNegative instances: %d\n", (int)(TP+FN), (int)(TN+FP));
            fprintf(parm->out_file, "TP: %d\tTN: %d\tFP: %d\tFN: %d\n", (int) TP, (int) TN, (int) FP, (int) FN);
        }
    }

    return ok;
}
