/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pvm3.h>
#include <string.h>
#include <unistd.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "pvm-local.h"

// Local k for k-NN
// Each instance has an associated number which is the number of neighbors when the instances is the NN

void ParallelOptimalLocalK(const t_Parameters *parm, int *localk, scalar **data, int instances, int *labels, scalar *w, bool use_cache, int inputs, int classes, int minority, int *S, bool redo_cache,
                           scalar *r, int processors, char *parm_file_name, char *out_file_name)
{
    static scalar **distance_cache = NULL;
    scalar max_distance = 0.0;
    static int **neighbors_cache = NULL, instances_allocated = -1;
    int *list, no_neighbors_count = 0, localk_lambda;

    // If more than two classes ignore
    if (parm->performance_measure == pfGMean && classes != 2)
        OnErrorExit(1, "Cannot use G measure with more than 2 classes");

    if (parm->performance_measure == pfauROC && parm->localk_error_estimation_method != RESUBS)
        OnErrorExit(1, "auROC only compatible with resubstitution error estimation");

    if (parm->localk_error_estimation_method != LOO && parm->localk_error_estimation_method != kFOLDCV && parm->symmetric_knn)
        OnErrorExit(1, "Symmetric k-NN only compatible with LOO and k-FOLDS error estimations");

    MALLOC(list, int, instances, "OptimalLocalK: list");
    for (int i = 0; i < instances; i++)
        list[i] = i;

    // If no S given select all instances
    bool local_s = false;
    if (S == NULL) {
        MALLOC(S, int, instances, "OptimalLocalK: S");
        for (int i = 0; i < instances; i++)
            S[i] = 1;
        local_s = true;
    } else if (parm->localk_error_estimation_method != LOO)
        fprintf(parm->out_file, "WARNING: S only taken into account for LOO\n");

    // Get cross-validation value of k for regularization term and local k lambda for regularization in case SMOOTHER is used
    scalar acc_by_k[parm->localk_max_k+1];
    if (parm->performance_measure == pfGMean || parm->performance_measure == pfauROC)
        localk_lambda = ObtainkCrossValidationUsingGMean(S, data, instances, labels, inputs, classes, MAX(1, parm->localk_min_k), parm->localk_max_k, acc_by_k, redo_cache, parm->knn_with_weighted_votes,
                        r, NULL, NULL);
    else
        localk_lambda = ObtainkCrossValidation(S, data, instances, labels, inputs, classes, MAX(1, parm->localk_min_k), parm->localk_max_k, acc_by_k, redo_cache, parm->performance_measure,
                                               parm->knn_with_weighted_votes, r, NULL, NULL, NULL);

    if (parm->verbose >= 0)
        fprintf(parm->out_file, "Minimum k: %d Lambda: %d Maximum k: %d\n", parm->localk_min_k, localk_lambda, parm->localk_max_k);

    // Create nearest neighbors cache
    int size = -1;
    if (use_cache && (neighbors_cache == NULL || redo_cache)) {
        RANK *rank;

        // If redo cache free previous cache
        if (neighbors_cache != NULL) {
            for (int i = 0; i < instances_allocated; i++) {
                FREE(neighbors_cache[i]);
                FREE(distance_cache[i]);
            }
            FREE(neighbors_cache);
            FREE(distance_cache);
        }

        MALLOC(neighbors_cache, int *, instances, "OptimalLocalK: neighbors_cache");
        MALLOC(distance_cache, scalar *, instances, "OptimalLocalK: distance_cache");
        MALLOC(rank, RANK, instances, "OptimalLocalK: rank");

        // Remember for freeing
        instances_allocated = instances;

        if (local_s == false)
            size = instances;
        else {
            if (parm->localk_error_estimation_method == kFOLDCV)
                size = MIN(parm->localk_max_k+instances/parm->localk_nfolds+2, instances);
            else if (parm->localk_error_estimation_method == LOO || parm->localk_error_estimation_method == RESUBS || parm->performance_measure == pfauROC)
                size = MIN(parm->localk_max_k+1, instances);
            else
                OnErrorExit(1, "Bad error estimation method");
        }

        for (int i = 0; i < instances; i++) {
            MALLOC(neighbors_cache[i], int, size+1, "OptimalLocalK: neighbors_cache[i]");
            MALLOC(distance_cache[i], scalar, size+1, "OptimalLocalK: distance_cache[i]");

            // Use random order for repeated entries
            RandomizeVector(list, instances);

            for (int j = 0; j < instances; j++) {
                rank[j].fitness = -EuclideanDistance(data[i], data[list[j]], inputs)/r[list[j]];
                rank[j].being = list[j];
            }

            // Sort
            qsort(rank, instances, sizeof(RANK), Compare);

            // Store neighbors in ascending order of distance
            for (int j = 0; j < size; j++) {
                neighbors_cache[i][j] = rank[j].being;
                distance_cache[i][j] = -rank[j].fitness;
            }

            // Maximum distance
            for (int j = 0; j < parm->localk_max_k+1 && j < size; j++)
                if (-rank[j].fitness > max_distance)
                    max_distance = -rank[j].fitness;

            // For error checking
            neighbors_cache[i][size] = -1;
            distance_cache[i][size] = -1.0;

        }
        FREE(rank);
    }

    if (parm->localk_populate)
        OnErrorExit(1, "Not valid for parallel implementation");

    // Run parallel processes
    // Not more processors that subsets
    if (processors > instances)
        processors = instances;

    int tids[processors];

    // PVM specific
    // PVM Spawning process
    // Copy parm file
    char *args[4];
    MALLOC(args[0], char, strlen(parm_file_name)+1, "main");
    MALLOC(args[1], char, strlen(out_file_name)+1, "main");
    strcpy(args[0], parm_file_name);       // Parameters file_name
    strcpy(args[1], out_file_name);       // Output file (slave appends pid)
    args[2] = getcwd(NULL, 0);      // Current directory for slaves
    args[3] = NULL;

    // Avoid running processes in master node
    int numt = pvm_spawn("localk-wrapper", args, PvmTaskDefault | PvmTaskHost | PvmHostCompl, "skynet.uco.es", processors, tids);
    // Debug version
    // int numt = pvm_spawn("localk-wrapper", args, PvmTaskDefault | PvmTaskDebug, ".", processors, tids);

    if (numt != processors)
        OnErrorExit(1, "Processes not spawned correctly");

    // Start guardian and send info
    int guardian = InitGuardian(numt, tids, "localk-wrapper", out_file_name);


    // Send data to all slaves
    pvm_initsend(PvmDataDefault);

    // Pack data
    pvm_pkint(&instances, 1, 1);
    pvm_pkint(&classes, 1, 1);
    pvm_pkint(S, instances, 1);
    pvm_pkint(&size, 1, 1);
    int send_use_cache = use_cache ? 1 : 0;
    pvm_pkint(&send_use_cache, 1, 1);
    if (use_cache) {
        for (int i = 0; i < instances; i++)
            pvm_pkint(neighbors_cache[i], size+1, 1);
        for (int i = 0; i < instances; i++)
            pvm_pkscalar(distance_cache[i], size+1, 1);
    }
    pvm_pkint(&localk_lambda, 1, 1);
    pvm_pkint(&inputs, 1, 1);
    pvm_pkint(labels, instances, 1);
    pvm_pkint(&minority, 1, 1);
    pvm_pkscalar(&max_distance, 1, 1);
    pvm_pkscalar(acc_by_k, parm->localk_max_k+1, 1);
    pvm_pkscalar(r, instances, 1);
    if (parm->symmetric_knn)
        for (int i = 0; i < instances; i++)
            pvm_pkscalar(data[i], inputs, 1);
    pvm_mcast(tids, processors, SENT_BY_MASTER);

    // Send interval for each slave
    int first = 0;
    int step = instances/processors;
    if (instances % processors)
        step++;

    for (int p = 0; p < processors; p++, first += step) {
        int last;

        pvm_initsend(PvmDataDefault);
        pvm_pkint(&first, 1, 1);
        last = first+step;
        if (last > instances)
            last = instances;
        pvm_pkint(&last, 1, 1);

        pvm_send(tids[p], SENT_BY_MASTER);
    }

    // Receive results from each slave
    first = no_neighbors_count = 0;
    for (int p = 0; p < processors; p++, first += step) {
        int count = 0;

        pvm_recv(tids[p], SENT_BY_SLAVE);
        pvm_upkint(&count, 1, 1);
        no_neighbors_count += count;
        if (first+step > instances)
            step = instances - first;
        pvm_upkint(localk+first, step, 1);
    }

    if (parm->verbose >= 0)
        fprintf(parm->out_file, "@Instances with no neighbors: %lf\n", ((scalar)no_neighbors_count)/instances);

    // Free memory
    if (local_s)
        FREE(S);
    FREE(list);

    // Kill spawned processes and guardian
    pvm_kill(guardian);
    for (int i = 0; i < processors; i++)
        if (pvm_kill(tids[i]))
            printf("Process %d not killed\n", tids[i]);
}
