/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

// Partial evaluation of a local k individual when only a value is modified.
// Previous is the previous value and modified individual is the individual modified
// If previous = -1 then a whole evaluation is made


scalar PartialEvaluationOfLocalKIndividual(const t_Parameters *parm, int *individual, scalar **data, int instances, int *label, scalar *w, int minority, int **neighbors_cache, int inputs,
        int classes, scalar *accuracy, scalar *regularization, scalar **distance_cache, scalar *margin, int previous, int modified, int *evaluations, scalar *r, int **index)
{
    int nearest[parm->localk_max_k+1];
    scalar fitness;
    static int *S = NULL;  // For regularization term
    static scalar *acc_by_k = NULL;

    if (parm->n_electors || parm->external_partitioning_approach)
        OnErrorExit(1, "This functions is not adapted to stratified/democratic approaches");

    if (w != NULL)
        OnErrorExit(1, "Weighted error not yet implemented");

    if (neighbors_cache == NULL)
        OnErrorExit(1, "Algorithm with no neighbor cache not yet implemented");

    // Get cross-validation value of k for regularization term
    if (acc_by_k == NULL) {
        MALLOC(S, int, instances, "PartialEvaluationOfLocalKIndividual: S");
        MALLOC(acc_by_k, scalar, parm->localk_max_k+1, "PartialEvaluationOfLocalKIndividual: acc_by_k");
        for (int i = 0; i < instances; i++)
            S[i] = 1;

        ObtainkCrossValidation((t_Parameters *) parm, S, data, instances, label, inputs, classes, index, NULL);
    }

    if (previous == -1) {
        *regularization = 0.0;
        for (int i = 0; i < instances; i++)
            // Sum all k values
            *regularization += acc_by_k[individual[i]];
        *regularization /= instances;
    } else
        *regularization += (acc_by_k[individual[modified]] - acc_by_k[previous])/((scalar)instances);

    if (parm->performance_measure == pfGMean || parm->performance_measure == pfauROC)
        OnErrorExit(1, "Evaluate with AUC or G-mean not yet implemented");
    else {
        // Set S[i]=0 of k values 0, rest S[i]=1
        for (int i = 0; i < instances; i++)
            S[i] = individual[i] ? 1 : 0;

        if (previous == -1)
            *accuracy = *margin = 0.0;

        int fold_size;
        static int *list = NULL;
        switch (parm->localk_error_estimation_method) {
            case LOO:
            case RESUBS:
            case BS:
            case E632:
            case HOLDOUT:
                OnErrorExit(1, "Not yet implemented with partial evaluation");
                break;

            case ArtificialData:
                EvaluateWithArtificialPopulation(parm, individual, data, instances, label, inputs, accuracy, margin, classes, w, previous, modified, evaluations, parm->localk_lambda, r);
                break;

            case kFOLDCV:
                // Ten-fold cross-validation
                fold_size = instances%parm->localk_nfolds ? instances/parm->localk_nfolds+1 : instances/parm->localk_nfolds;

                // Random vector to avoid order effects
                if (list == NULL) {
                    MALLOC(list, int, instances, "PartialEvaluationOfLocalKIndividual: list");
                    for (int i = 0; i < instances; i++)
                        list[i] = i;
                }

                if (previous == -1)
                    RandomizeVector(list, instances);

                // Test accuracy
                for (int fold = 0; fold < parm->localk_nfolds; fold++) {
                    // i-th fold
                    for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++)
                        S[list[j]] = 0;

                    // Test fold
                    for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++) {
                        int l = 1, k;

                        // Get nearest neighbor k value
                        while (!S[neighbors_cache[list[j]][l]])
                            l++;

                        // Is the modified instance? Then evaluate with current value and previous one
                        if (previous == -1 || neighbors_cache[list[j]][l] == modified) {
                            int now, before;
                            scalar margin_now, margin_before;

                            k = individual[neighbors_cache[list[j]][l]];

                            now = (kNearestNeighbourLargeCache(parm, list[j], S, instances, nearest, label, k, neighbors_cache, distance_cache, classes, &margin_now, inputs, label[list[j]],
                                                               0.0, NULL, NULL, index) == label[list[j]]);

                            if (previous == -1) {
                                before = 0;
                                margin_before = 0.0;
                            } else
                                before = (kNearestNeighbourLargeCache(parm, list[j], S, instances, nearest, label, previous, neighbors_cache, distance_cache, classes, &margin_before, inputs,
                                                                      label[list[j]], 0.0, NULL, NULL, index) == label[list[j]]);

                            *accuracy += (now-before)/((scalar) instances);
                            *margin += (margin_now-margin_before)/instances;

                            (*evaluations)++;
                        }
                    }

                    for (int j = fold*fold_size; j < (fold+1)*fold_size && j < instances; j++)
                        if (individual[list[j]])
                            S[list[j]] = 1;
                }
                break;

            default:
                OnErrorExit(1, "Bad error estimation method");
        }
    }

    // Consider accuracy and sum
    fitness = parm->localk_alpha_acc * (*accuracy)+parm->localk_alpha_reg * (*regularization)+parm->localk_alpha_mar * (*margin);

    return fitness;
}
