/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

#define THRESHOLD       0.1     // Threshold to consider values equal

// Get input with lowest entropy considering only indicated  subset, and the subspace given
int LowestEntropy(scalar **data, int instances, int *label, int *part_info, int subset, int *subspace, int inputs, int classes)
{
    scalar *entropy = NULL, *values, *cardinal, I;
    int diff_values;
    static int *partition, **class_count;
    bool newval, null_space;

    // Memory allocation
    MALLOC(entropy, scalar, inputs, "EntropyPartitioning: entropy");
    MALLOC(values, scalar, instances, "EntropyPartitioning: values");
    MALLOC(partition, int, instances, "EntropyPartitioning: partition");
    MALLOC(cardinal, scalar, instances, "EntropyPartitioning: cardinal");
    MALLOC(class_count, int *, instances, "EntropyPartitioning: class_count");
    for (int i = 0; i < instances; i++)
        MALLOC(class_count[i], int, classes, "EntropyPartitioning: class_count[i]");

    // Obtain entropy of each attribute
    null_space = true;
    for (int i = 0; i < inputs; i++)
        if (subspace[i]) {
            null_space = false;

            // Count different values of the attribute and partition of set using that attribute
            diff_values = 0;
            for (int j = 0; j < instances; j++)
                if (part_info[j] == subset) {
                    newval = true;
                    for (int k = 0; k < diff_values && newval; k++)
                        if (fabs(data[j][i] - values[k]) < THRESHOLD) {
                            newval = false;
                            partition[j] = k;
                            cardinal[k]++;
                            class_count[k][label[j]]++;
                        }
                    // Add value
                    if (newval) {
                        values[diff_values] = data[j][i];
                        cardinal[diff_values] = 1.0;
                        partition[j] = diff_values;
                        for (int k = 0; k < classes; k++)
                            class_count[diff_values][k] = 0;
                        class_count[diff_values][label[j]] = 1;
                        diff_values++;
                    }
                }
            // Obtain entropy
            entropy[i] = 0.0;
            for (int j = 0; j < diff_values; j++) {
                I = 0.0;
                for (int k = 0; k < classes; k++)
                    if (class_count[j][k])
                        I -= (class_count[j][k]/cardinal[j]) * log2(class_count[j][k]/cardinal[j]);

                entropy[i] += (cardinal[j]/instances) * I;
            }
        } else
            entropy[i] = INFINITE;

    // No inputs left
    if (null_space)
        return -1;

    // Get lowest entropy
    int minimum = 0;
    for (int i = 1; i < inputs; i++)
        if (entropy[minimum] > entropy[i])
            minimum = i;

    FREE(entropy);
    FREE(values);
    FREE(partition);
    FREE(cardinal);
    for (int i = 0; i < instances; i++)
        FREE(class_count[i]);
    FREE(class_count);

    return minimum;
}

// Add a new partition using the designed attribute
// Return the number of added subsets
void PerformPartition(scalar **data, int instances, int *part_info, int subset, int *n_subsets, int *subspace, int *label, int *current_sp, int inputs, int classes)
{
    scalar *values = NULL;
    int cardinality, lowest, *local_sp;
    bool newval;

    MALLOC(local_sp, int, inputs, "PerformPartition: subsp");
    if (current_sp == NULL) {
        for (int i = 0; i < inputs; i++)
            local_sp[i] = true;
    } else
        for (int i = 0; i < inputs; i++)
            local_sp[i] = current_sp[i];

    // Memory allocation
    MALLOC(values, scalar, instances, "AddPartition: values");

    // Get lowest entropy
    lowest = LowestEntropy(data, instances, label, part_info, subset, local_sp, inputs, classes);

    // If no inputs left return
    if (lowest == -1) {
        FREE(local_sp);
        return;
    }

    // Count different values
    cardinality = 0;
    for (int j = 0; j < instances; j++)
        if (part_info[j] == subset) {
            newval = true;
            for (int k = 0; k < cardinality && newval; k++)
                if (fabs(data[j][lowest] - values[k]) < THRESHOLD) {
                    newval = false;
                    part_info[j] = k+*n_subsets;
                }
            // Add value
            if (newval) {
                values[cardinality] = data[j][lowest];
                part_info[j] = cardinality+*n_subsets;
                cardinality++;
            }
        }
    // Added subsets
    *n_subsets += cardinality;

    // Perform partition of each new subset
    subspace[lowest] = false;
    local_sp[lowest] = false;

    // Subsets to repartionate
    int first = *n_subsets - cardinality;

    for (int j = 0; j < cardinality; j++) {
        // Check is the partition is pure, that is, all the members are of the same class
        bool pure = true;
        int i = 0;
        while (part_info[i] != (first+j))
            i++;
        int category = label[i];
        for (i++; i < instances; i++)
            if (part_info[i] == (first+j) && label[i] != category) {
                pure = false;
                break;
            }
        if (!pure)
            PerformPartition(data, instances, part_info, first+j, n_subsets, subspace, label, local_sp, inputs, classes);
    }

    FREE(values);
    FREE(local_sp);
}

// Return the ordered list of indexes
int ReorderSubsetsLables(int *partition, int n)
{
    RANK ranking[n];
    int n_subsets = 1, old_label, new_label, *list = NULL;

    MALLOC(list, int, n, "ReorderSubsetsLables: list");

    // Ordered by partition label
    for (int i = 0; i < n; i++) {
        ranking[i].being = i;
        ranking[i].fitness = (scalar) -partition[i];
    }

    qsort(ranking, n, sizeof(RANK), Compare);

    // Ordered list
    for (int i = 0; i < n; i++)
        list[i] = ranking[i].being;

    // Reorder
    old_label = new_label = 0;
    for (int i = 0; i < n - 1; i++) {
        if (partition[list[i]] != partition[list[i+1]]) {
            new_label = old_label+1;
            n_subsets++;
        }
        partition[list[i]] = old_label;
        old_label = new_label;
    }

    partition[list[n - 1]] = old_label;

    FREE(list);

    return n_subsets;
}

// Obtain the center of a subset as the instance whose sum of distances to all remaining members
// is minimum and select the K nearest instances to the center

#define K       3               // No info given in the paper to set k

void RepresentativesOfSubset(scalar **data, int instances, int *part_info, int subset, int *subspace, int *S, int *label, int inputs, int classes, bool knn_with_weighted_votes, int **index)
{
    int *list = NULL, *nearest, *sample;
    scalar *distances;

    MALLOC(list, int, instances, "RepresentativesOfSubset: list");
    MALLOC(distances, scalar, instances, "RepresentativesOfSubset: list");
    MALLOC(nearest, int, K, "RepresentativesOfSubset: nearest");
    MALLOC(sample, int, instances, "RepresentativesOfSubset: sample");
    for (int i = 0; i < instances; i++)
        sample[i] = 0;

    // Get list of members
    int n = 0;
    for (int i = 0; i < instances; i++)
        if (part_info[i] == subset) {
            list[n] = i;
            n++;
        }
    // Obtain distances to all remaining members
    int minimum = 0;
    for (int i = 0; i < n; i++) {
        distances[i] = 0.0;
        for (int j = 0; j < n; j++)
            distances[i] += EuclideanDistanceSubspace(data[list[i]], data[list[j]], inputs, subspace);

        if (distances[i] < distances[minimum])
            minimum = i;
    }

    // Set minimum
    S[list[minimum]] = 1;

    // Select k nearest instances
    // Set sample
    for (int i = 0; i < n; i++)
        sample[list[i]] = 1;
    sample[list[minimum]] = 0;

    kNNCategory(data[list[minimum]], data, NULL, instances, K, inputs, label, classes, nearest, sample, NULL, NULL, -1, index, index[list[minimum]]);

    // Select K nearest instances
    for (int i = 0; i < K; i++)
        if (nearest[i] != -1)
            S[nearest[i]] = 1;

    // Unset sample
    for (int i = 0; i < n; i++)
        sample[list[i]] = 0;

    FREE(list);
    FREE(distances);
    FREE(nearest);
    FREE(sample);
}
