/***************************************************************************
 *   Copyright (C) 2013 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "is.h"
#include "classifiers.h"

// Populate data creating artificial instances from adjacent instances
void SMOTEPopulate(scalar ***data, int **labels, int *instances, scalar **orig_data, int orig_instances, int *orig_labels, int inputs, int reps_per_instance, int classes, int number_of_neighbors,
                   int best_k, bool *no_neighbors, bool knn_with_weighted_votes, scalar *r)
{
    int n = 0, *S;

    // Allocate memory for new data
    *instances = reps_per_instance*number_of_neighbors*orig_instances;
    MALLOC(*data, scalar *, *instances, "SMOTEPopulate: *data");
    MALLOC(*labels, int, *instances, "SMOTEPopulate: *labels");
    MALLOC(S, int, orig_instances, "SMOTEPopulate: S");
    for (int i = 0; i < orig_instances; i++)
        S[i] = 1;

    // Get nearest neighbor for each pattern
    for (int i = 0; i < orig_instances; i++)
        // Generate neighbors only for instance that are nearest to nobody?
        if (!no_neighbors || no_neighbors[i]) {
            int nearest[number_of_neighbors];

            // Get nearest neighbors
            S[i] = 0;
            kNNCategory(orig_data[i], orig_data, NULL, orig_instances, number_of_neighbors, inputs, orig_labels, classes, nearest, S, NULL, NULL, -1, NULL, NULL);
            S[i] = 1;

            // Add instances to new data
            bool enemy_reached = false;
            for (int k = 0; k < number_of_neighbors && !enemy_reached; k++) {
                // If nearest enemy do not get a new artificial instances
                if (orig_labels[nearest[k]] != orig_labels[i])
                    enemy_reached = true;

                for (int j = 0; j < reps_per_instance; j++) {
                    MALLOC((*data)[n], scalar, inputs, "Populate: (*data)[n]");

                    // Create artificial instance
                    for (int j = 0; j < inputs; j++) {
                        scalar limit = enemy_reached ? 0.5 : 1.0;
                        (*data)[n][j] = orig_data[i][j]+RANDOM(0.0, limit) * (orig_data[nearest[k]][j] - orig_data[i][j]);
                    }

                    (*labels)[n] = orig_labels[i];

                    n++;
                }
            }
        }

    // Adjust number of instances?
    if (*instances != n) {
        *instances = n;
        REALLOC(*data, scalar *, *instances, "SMOTEPopulate");
    }

    FREE(S);
}

void EvaluateWithArtificialPopulation(const t_Parameters *parm,int *individual, scalar **data, int instances, int *labels, int inputs, scalar *accuracy, scalar *margin, int classes, scalar *w,
                                      int previous, int modified, int *evaluations, int best_k, scalar *r)
{
    static scalar **artificial_data = NULL, **distance_cache;
    static int *artificial_labels, artificial_instances, **neighbors_cache, *S;

    if (parm->n_electors || parm->external_partitioning_approach)
        OnErrorExit(1, "This functions is not adapted to stratified/democratic approaches");

    // First time? Create artificial data
    if (artificial_data == NULL) {
        SMOTEPopulate(&artificial_data, &artificial_labels, &artificial_instances, data, instances, labels, inputs, parm->populate_copies_per_instance, classes, parm->populate_number_of_neighbors,
                      best_k, NULL, parm->knn_with_weighted_votes, r);

        // And create neighbors cache
        int size = -1;
        RANK *rank;
        MALLOC(rank, RANK, instances, "EvaluateWithArtificialPopulation: rank");
        MALLOC(neighbors_cache, int *, artificial_instances, "EvaluateWithArtificialPopulation: neighbors_cache");
        MALLOC(distance_cache, scalar *, artificial_instances, "EvaluateWithArtificialPopulation: distance_cache");

        if (parm->localk_min_k == 0)
            OnErrorExit(1, "Instance selection not implemented for artificial data");
        else
            size = MIN(parm->localk_max_k, instances);

        for (int i = 0; i < artificial_instances; i++) {
            MALLOC(neighbors_cache[i], int, size+1, "EvaluateWithArtificialPopulation: neighbors_cache[i]");
            MALLOC(distance_cache[i], scalar, size+1, "EvaluateWithArtificialPopulation: distance_cache[i]");

            for (int j = 0; j < instances; j++) {
                rank[j].fitness = -EuclideanDistance(artificial_data[i], data[j], inputs)/r[j];
                rank[j].being = j;
            }

            // Sort
            qsort(rank, instances, sizeof(RANK), Compare);

            // Store neighbors in ascending order of distance
            for (int j = 0; j < size; j++) {
                neighbors_cache[i][j] = rank[j].being;
                distance_cache[i][j] = -rank[j].fitness;
            }

            // For error checking
            neighbors_cache[i][size] = -1;
            distance_cache[i][size] = -1.0;
        }
        FREE(rank);

        MALLOC(S, int, instances, "EvaluateWithArtificialPopulation: S");
        for (int i = 0; i < instances; i++)
            S[i] = 1;
    }

    if (previous == -1)
        *accuracy = *margin = 0.0;

    // Evaluate accuracy on artificial data
    for (int i = 0; i < artificial_instances; i++) {
        int k, nearest[parm->localk_max_k+1];

        // Modified
        // Is the instance modified? Then evaluate with current value and previous one
        if (previous == -1 || neighbors_cache[i][0] == modified) {
            int now, before;
            scalar margin_now, margin_before;

            k = individual[neighbors_cache[i][0]];

            now = (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, labels, k, neighbors_cache, distance_cache, classes, &margin_now, inputs, artificial_labels[i], 0.0, NULL, NULL, NULL) ==
                   artificial_labels[i]);

            if (previous == -1) {
                before = 0;
                margin_before = 0.0;
            } else
                before = (kNearestNeighbourLargeCache(parm, i, S, instances, nearest, labels, previous, neighbors_cache, distance_cache, classes, &margin_before, inputs, artificial_labels[i], 0.0, NULL,
                                                      NULL, NULL) == artificial_labels[i]);

            *accuracy += (now-before)/((scalar) artificial_instances);
            *margin += (margin_now-margin_before)/artificial_instances;

            if (evaluations != NULL)
                (*evaluations)++;
        }
    }
}
