/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

// Implements LVQPRU algorithm functions

t_LVQ_net *LVQPruning(const t_Parameters *parm, scalar **data, int n, int *labels, int inputs, int classes)
{
    scalar **protos, **v_data;
    scalar bsf_accuracy = 0.0;
    int v_instances, *v_label, *list, *count, nearest, *sample, most;

    // Allocate memory
    // At most lvq->mi = Npc * Nc
    int Ntc = parm->Npc * classes;
    MALLOC(protos, scalar *, Ntc, "LVQPruning: protos");
    MALLOC(list, int, n, "LVQPruning: list");
    MALLOC(sample, int, Ntc, "LVQPruning: sample");
    MALLOC(count, int, Ntc, "LVQPruning: count");
    for (int i = 0; i < Ntc; i++)
        MALLOC(protos[i], scalar, inputs, "LVQPruning: protos[i]");

    // Create random validation set
    v_instances = n;
    MALLOC(v_data, scalar *, v_instances, "LVQPruning: 	v_data");
    MALLOC(v_label, int, v_instances, "LVQPruning: v_label");

    v_label = labels;
    v_data = data;

    // Randomly initialize prototypes. Npc per class at most
    t_LVQ_net *lvq = LVQCreate(Ntc, inputs);
    t_LVQ_net *best_so_far = LVQCreate(Ntc, inputs);
    lvq->mi = 0;

    for (int i = 0; i < classes; i++) {
        // Get instances of class i
        int instances = 0;
        for (int j = 0; j < n; j++)
            if (labels[j] == i) {
                list[instances] = j;
                instances++;
            }
        // If instances less than Npc select all, otherwise random selection
        if (instances <= parm->Npc) {
            for (int j = 0; j < instances; j++) {
                lvq->labels[lvq->mi] = labels[list[j]];
                for (int k = 0; k < inputs; k++)
                    lvq->vectors[lvq->mi][k] = data[list[j]][k];
                lvq->mi++;
            }
        } else {
            // Randomize list
            RandomizeVector(list, instances);

            for (int j = 0; j < parm->Npc; j++) {
                lvq->labels[lvq->mi] = labels[list[j]];
                for (int k = 0; k < inputs; k++)
                    lvq->vectors[lvq->mi][k] = data[list[j]][k];
                lvq->mi++;
            }
        }
    }

    fprintf(parm->out_file, "Initialization: %d prototypes\n", lvq->mi);

    // Train SOM network
    // MODIFIED: Not enough info, just remove prototypes without effect
    // Delete k prototype if it has no members
    for (int i = 0; i < lvq->mi; i++) {
        count[i] = 0;
        sample[i] = 1;
    }

    // Classify all prototypes and
    // construct Tjk, distribution of classes of the neighbors of each instance
    int **T;
    MALLOC(T, int *, lvq->mi, "LVQPruning: T");
    for (int i = 0; i < lvq->mi; i++) {
        MALLOC(T[i], int, classes, "LVQPruning: T[i]");
        for (int j = 0; j < classes; j++)
            T[i][j] = 0;
    }
    for (int i = 0; i < n; i++) {
        kNNCategory(data[i], lvq->vectors, NULL, lvq->mi, 1, inputs, labels, classes, &nearest, sample, NULL, NULL, -1, NULL, NULL);
        count[nearest]++;
        T[nearest][labels[i]]++;
    }

    // Change class label of prototype if it disagrees with neighbors
    for (int i = 0; i < lvq->mi; i++)
        if ((most = VectorMaxIndex(T[i], classes)) != lvq->labels[i])
            lvq->labels[i] = most;

    // Remove
    for (int i = 0; i < lvq->mi; i++)
        if (count[i] == 0) {
            // Remove
            lvq->mi--;
            lvq->labels[i] = lvq->labels[lvq->mi];
            for (int j = 0; j < inputs; j++)
                lvq->vectors[i][j] = lvq->vectors[lvq->mi][j];
            count[i] = count[lvq->mi];

            i--;
        }
    fprintf(parm->out_file, "After removing empty prototypes: %d remaining\n", lvq->mi);

    // Use LVQ2.1 to refine prototypes
    LVQLearning(lvq, data, n, labels, lvq->mi, 10 * n, labels, inputs, classes);

    // Apply basic condensing algorithm
    int *S;
    ALLOCA(S, int, lvq->mi, "LVQPruning: S");
    CondensedNN(S, lvq->vectors, lvq->mi, lvq->labels, inputs, classes);

    // Remove unselected instances
    for (int i = lvq->mi - 1; i >= 0; i--)
        if (S[i] == 0) {
            lvq->mi--;
            lvq->labels[i] = lvq->labels[lvq->mi];
            for (int j = 0; j < inputs; j++)
                lvq->vectors[i][j] = lvq->vectors[lvq->mi][j];
        }
    fprintf(parm->out_file, "After Condensed Nearest Neighbour: %d remaining\n", lvq->mi);

    // Remove instances stepwise
    while (lvq->mi > classes) {
        // Prune the prototype with least testing error increment
        // Test each one
        scalar worst_accuracy = 0.0;
        int worst = 0;

        lvq->mi--;

        for (int i = 0; i < lvq->mi; i++) {
            // Temporally
            int auxi = lvq->labels[i];
            lvq->labels[i] = lvq->labels[lvq->mi];
            lvq->labels[lvq->mi] = auxi;

            scalar *auxp = lvq->vectors[i];
            lvq->vectors[i] = lvq->vectors[lvq->mi];
            lvq->vectors[lvq->mi] = auxp;

            // Test accuracy with prototype removed
            scalar accuracy = TestLVQNetwork(lvq, v_data, v_instances, v_label, inputs);
            if (accuracy > worst_accuracy) {
                worst_accuracy = accuracy;
                worst = i;
            }
            // Restore prototype
            auxi = lvq->labels[i];
            lvq->labels[i] = lvq->labels[lvq->mi];
            lvq->labels[lvq->mi] = auxi;

            auxp = lvq->vectors[i];
            lvq->vectors[i] = lvq->vectors[lvq->mi];
            lvq->vectors[lvq->mi] = auxp;
        }

        // Remove worst
        lvq->labels[worst] = lvq->labels[lvq->mi];
        lvq->vectors[worst] = lvq->vectors[lvq->mi];

        // Use LVQ2.1 to refine prototypes
        LVQLearning(lvq, data, n, labels, lvq->mi, n, labels, inputs, classes);

        // Testing accuracy and test if best so far
        scalar accuracy = TestLVQNetwork(lvq, v_data, v_instances, v_label, inputs);
        if (accuracy >= bsf_accuracy) {
            bsf_accuracy = accuracy;

            // Copy LVQ network to best_so_far
            best_so_far->mi = lvq->mi;
            for (int i = 0; i < lvq->mi; i++) {
                best_so_far->labels[i] = lvq->labels[i];
                for (int j = 0; j < inputs; j++)
                    best_so_far->vectors[i][j] = lvq->vectors[i][j];
            }

        }

        fprintf(parm->out_file, "%d prototypes with training error: %6.4lf (best %6.4lf)\n", lvq->mi, accuracy, bsf_accuracy);
    }

    return best_so_far;
}

// Condensed nearest neighbor rule
// Start with empty set, add misclassified individual
void CondensedNN(int *S, scalar **instances, int n, int *label, int inputs, int classes)
{
    int nearest;
    bool added;

    for (int i = 0; i < n; i++)
        S[i] = 0;

    do {
        added = false;

        for (int i = 0; i < n; i++)
            if (S[i] == 0 && kNNCategory(instances[i], instances, NULL, n, 1, inputs, label, classes, &nearest, S, NULL, NULL, -1, NULL, NULL) != label[i]) {
                S[i] = 1;
                added = true;
            }
    } while (added);

}
