/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <unistd.h>
#include "is.h"
#include "macros.h"
#include "classifiers.h"

// Implementation of the pattern selection algorithm in
// H. Shin and S. Cho, "Pattern selection for support vector classifiers", IDEAL 2002, LNCS 2412, pp. 469-474, 2002.
int PatternSelectionAlgorithm(int *S, scalar **data, int instances, int *labels, int k, int inputs, int classes, bool knn_with_weighted_votes, int *subspace, int **index)
{
    scalar *proximity, *correctness, P[classes];
    int nearest[k];

    MALLOC(proximity, scalar, instances, "PatternSelectionAlgorithm: proximity");
    MALLOC(correctness, scalar, instances, "PatternSelectionAlgorithm: correctness");

    for (int i = 0; i < instances; i++)
        S[i] = 1;

    // Find proximity and correctness of each individual
    for (int i = 0; i < instances; i++) {
        // Get k neighbors
        kNNCategory(data[i], data, subspace, instances, k, inputs, labels, classes, nearest, S, NULL, NULL, -1, index, index ? index[i] : NULL);

        // Probability for each class
        for (int j = 0; j < classes; j++)
            P[j] = 0.0;
        for (int j = 0; j < k; j++)
            P[labels[nearest[j]]] += 1.0/k;

        // Calculate proximity
        proximity[i] = 0.0;
        for (int j = 0; j < classes; j++)
            if (P[j] > ZERO)
                proximity[i] += P[j] * Logarithm((scalar) classes, 1.0/P[j]);

        // Calculate correctness
        correctness[i] = P[labels[i]];
    }

    // Select the instances fulfilling:
    // proximity > 0 and correctness >= 1/#classes
    int Np = instances;
    for (int i = 0; i < instances; i++)
        if (!(proximity[i] > ZERO && correctness[i] >= 1.0/classes)) {
            // Unselect
            S[i] = 0;
            Np--;
        }

    FREE(proximity);
    FREE(correctness);

    return Np;
}

// Logarithm in any base
scalar Logarithm(scalar base, scalar x)
{
    return log(x)/log(base);
}
