/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "classifiers.h"

// Implementation of PSRCG instance selection algorithm
// M. Sebban, R. Nock, J. H. Chauchat and R. Rakotomalala, "Impact of learning set quality and size on decision tree performances",
// International Journal of Computers, Systems and Signals, vol. 1, no, 1, pp. 85-105, 2000.

int PSRCGAlgorithm(const t_Parameters *parm, int *S, scalar **data, int instances, int *labels, int k, int inputs, int classes, int **index, int *subspace)
{
    int Np, nearest[k];
    scalar RCG_t, RCG_t_plus_1, *Uloc;

    // All selected initially
    Np = instances;
    for (int i = 0; i < instances; i++)
        S[i] = 1;

    // Build k-NN graph on the learning set
    unsigned char **knn_graph;
    MALLOC(Uloc, scalar, instances, "PSRCG: Uloc");
    MALLOC(knn_graph, unsigned char *, instances, "PSRCG: knn_graph");
    for (int i = 0; i < instances; i++)
        CALLOC(knn_graph[i], unsigned char, instances, "PSRCG: knn_graph[i]");

    for (int i = 0; i < instances; i++) {
        // First add himself
        knn_graph[i][i] = 1;

        // Get k neighbors
        kNNCategory(data[i], data, subspace, instances, k, inputs, labels, classes, nearest, S, NULL, NULL, -1, index, index ? index[i] : NULL);

        // Set edges
        for (int j = 0; j < k; j++)
            if (nearest[j] >= 0)
                knn_graph[i][nearest[j]] = knn_graph[nearest[j]][i] = 1;
    }

    // Compute RCG(1)
    RCG_t_plus_1 = ComputeRCG(knn_graph, instances, labels, Uloc, S, classes);

    int omega, count, *list;

    MALLOC(list, int, instances, "PSRCGAlgorithm: list");
    do {
        RCG_t = RCG_t_plus_1;

        // Select the instance w = max(Uloc(wj))
        // If two instances have the same Uloc select the instance having the smallest number of neighbors
        omega = 0;
        while (!S[omega])
            omega++;
        list[0] = omega;
        count = 1;
        for (int i = omega+1; i < instances; i++)
            if (S[i]) {
                if (Uloc[i] > Uloc[omega]) {
                    omega = i;
                    list[0] = omega;
                    count = 1;
                } else if (Uloc[i] == Uloc[omega]) {
                    list[count] = i;
                    count++;
                }
            }

        // Check whether there is more than a maximum
        if (count > 1) {
            int least_neighbors = instances+1;
            for (int i = 0; i < count; i++) {
                // Count neighbors of list[i]
                int c = 0;
                for (int j = 0; j < instances; j++)
                    if (knn_graph[list[i]][j])
                        c++;

                if (c < least_neighbors) {
                    least_neighbors = c;
                    omega = list[i];
                }
            }
        }

        // Remove omega
        S[omega] = 0;
        Np--;
        knn_graph[omega][omega] = 0;

        // Local modifications of the k-NN graph
        // Remove edges of omega
        for (int i = 0; i < instances; i++)
            // Recompute neighbors of omega's friends
            if (knn_graph[omega][i]) {
                knn_graph[omega][i] = knn_graph[i][omega] = 0;

                // Get k neighbors
                kNNCategory(data[i], data, subspace, instances, k, inputs, labels, classes, nearest, S, NULL, NULL, -1, index, index ? index[i] : NULL);
                if (nearest[k-1] >= 0)
                    knn_graph[i][nearest[k-1]] = knn_graph[nearest[k-1]][i] = 1;
            }

        // Compute RCG(t+1) after removing omega
        RCG_t_plus_1 = ComputeRCG(knn_graph, instances, labels, Uloc, S, classes);
    } while (Np > 1 && RCG_t_plus_1 >= RCG_t && RCG_t_plus_1 > 0.0);

    FREE(list);

    // Remove instances having a null uncertainty with the (k+1)-nn graph
    // Add k+1 neighbor
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            kNNCategory(data[i], data, subspace, instances, k+1, inputs, labels, classes, nearest, S, NULL, NULL, -1, index, index ? index[i] : NULL);
            if (nearest[k] >= 0)
                knn_graph[i][nearest[k]] = knn_graph[nearest[k]][i] = 1;
        }

    VERBOSE(parm->verbose > 0, parm->out_file, "PSRCG first step removed %d instances\n", instances - Np);

    // Recompute RCG
    ComputeRCG(knn_graph, instances, labels, Uloc, S, classes);

    // Remove instances with null uncertainty
    for (int i = 0; i < instances; i++)
        if (S[i] && Uloc[i] <= ZERO) {
            S[i] = 0;
            Np--;
        }

    VERBOSE(parm->verbose > 0, parm->out_file, "PSRCG second step removed %d instances\n", instances - Np);

    // Free memory
    FREE(Uloc);
    for (int i = 0; i < instances; i++)
        FREE(knn_graph[i]);
    FREE(knn_graph);

    // Return |S|
    return Np;
}

scalar ComputeRCG(unsigned char **graph, int instances, int *labels, scalar *Uloc, int *S, int classes)
{
    scalar U0 = 0.0, n_ij[classes], Utot, n_i, count[classes];
    int cardE = 0, Np = 0;

    Utot = 0.0;

    for (int i = 0; i < classes; i++)
        count[i] = 0.0;

    // Get n_i and n_ij
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            // Count instance
            count[labels[i]]++;
            Np++;

            n_i = 0.0;
            for (int j = 0; j < classes; j++)
                n_ij[j] = 0.0;

            for (int j = 0; j < instances; j++)
                if (graph[i][j]) {
                    n_i++;
                    n_ij[labels[j]]++;
                }

            cardE += n_i - 1; // Not consider autolink

            // Compute Uloc(i)
            Uloc[i] = 0.0;
            for (int j = 0; j < classes; j++)
                Uloc[i] += (n_ij[j]/n_i) * (1.0 - (n_ij[j]/n_i));

            Utot += n_i * Uloc[i];
        }
    // Each edge twice counted
    cardE /= 2;

    Utot /= Np+2.0 * cardE;

    for (int i = 0; i < classes; i++)
        U0 += (count[i]/Np) * (1.0 - (count[i]/Np));

    scalar RCG = (U0 - Utot)/U0;

    return RCG;
}
