/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "is.h"
#include <unistd.h>
#include "macros.h"
#include "common.h"

// Random Mutation Hill Climbing Instance Selection
// Skalak, David B. "Prototype and feature selection by sampling and random mutation hill climbing algorithms." Proceedings of the eleventh
// international conference on machine learning. 1994.

int RandomMutationHillClimbingInstanceSelection(const t_Parameters *parm, int *S, scalar **data, int *labels, int instances, int classes, cache_t **neighbors_cache, int inputs, int minority,
        int **index, int *subspace)
{
    int n_selected, n_unselected, cache_size = 0, *s_list, *u_list;
    cache_t **cache = NULL;

    // S = empty set
    for (int i = 0; i < instances; i++)
        S[i] = 0;

    // List of selected and unselected instances
    MALLOC(s_list, int, instances, "RandomMutationHillClimbingInstanceSelection: s_list");
    MALLOC(u_list, int, instances, "RandomMutationHillClimbingInstanceSelection: u_list");

    // Randomly select m instances
    // If m = 0, use number of classes
    for (int i = 0; (i < parm->rmhc_m || i < classes) && i < instances; i++) {
        int s;
        do {
            s = RANDOM(0, instances);
        } while (S[s]);

        S[s] = 1;
    }

    n_unselected = n_selected = 0;
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            s_list[n_selected] = i;
            n_selected++;
        } else {
            u_list[n_unselected] = i;
            n_unselected++;
        }

    // Cache already done?
    if (neighbors_cache == NULL)
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &cache, NULL, index, &cache_size, subspace, false);
    else
        cache = neighbors_cache;

    scalar ok, reduction, acc_s, acc_new;
    acc_s = EvaluateCHCIndividual(parm, S, &ok, &reduction, data, instances, labels, NULL, minority, cache, inputs, classes, false, NULL, index, 1.0, 0.0, subspace, 0);

    for (int i = 0; i < parm->rmhc_iterations && n_selected && n_unselected; i++) {
        int selected, unselected;

        // Random selected instance
        selected = RANDOM(0, n_selected);

        // Random unselected instance
        unselected = RANDOM(0, n_unselected);

        // Exchange
        S[s_list[selected]] = 0;
        S[u_list[unselected]] = 1;

        // Evaluate
        acc_new = EvaluateCHCIndividual(parm, S, &ok, &reduction, data, instances, labels, NULL, minority, cache, inputs, classes, false, NULL, index, 1.0, 0.0, subspace, 0);

        // If better keep
        if (acc_new > acc_s) {
            acc_s = acc_new;

            // Remove from list and update number
            int sw = s_list[selected];
            s_list[selected] = u_list[unselected];
            u_list[unselected] = sw;
        } else { // Or undo
            S[s_list[selected]] = 1;
            S[u_list[unselected]] = 0;
        }
    }

    // Free cache?
    if (neighbors_cache == NULL) {
        for (int i = 0; i < instances; i++)
            FREE(cache[i]);
        FREE(cache);
    }

    FREE(s_list);
    FREE(u_list);

    return n_selected;
}
