/***************************************************************************
 *   Copyright (C) 2017 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
// Functions that deal with the individuals
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include "is.h"
#include "macros.h"
#include "common.h"
#include "maths.h"

// Relative Neighborhood Graph Editing:  J. S. Sanchez, F. Pla, and F. J. Ferri, “Prototype selection for the nearest neighbour rule through proximity graphs,”
// Pattern Recognition Letters, vol. 18, pp. 507–513, 1997.

int RelativeNeighborhoodGraphEditing(t_ProximityGraph pg, int *S, scalar **data, int instances, int *labels, int classes, int inputs, int **index, scalar **dist_cache, int *subspace)
{
    int votes[classes], selected = 0;
    bool **gg = NULL;

    if (pg == Gabriel)
        // Construct proximity graph
        gg = GabrielGraph(data, inputs, instances, index, dist_cache, subspace);
    else if (pg == RelativeNeighborhood)
        // Construct proximity graph
        gg = RelativeNeighborhoodGraph(data, inputs, instances, index, dist_cache, subspace);
    else
        OnErrorExit(ERR_UNSPECIFIED, "Bad PG option");

    // Apply editing
    // Test whether instance is misclassified by its graph neighbors
    for (int i = 0; i < instances; i++) {
        for (int j = 0; j < classes; j++)
            votes[j] = 0;

        for (int j = 0; j < instances; j++)
            if (gg[i][j])
                votes[labels[j]]++;

        // Classify
        selected += (S[i] = VectorMaxIndex(votes, classes) == labels[i] ? true : false);
    }

    for (int i = 0; i < instances; i++)
        FREE(gg[i]);
    FREE(gg);

    return selected;
}


bool **GabrielGraph(scalar **data, int inputs, int instances, int **index, scalar **distance_cache, int *subspace)
{
    bool **gg;
    scalar **distances;

    // Gabriel Graph and cache of distances
    MALLOC(gg, bool *, instances, "GabrielGraph: gg");
    for (int i = 0; i < instances; i++)
        MALLOC(gg[i], bool, instances, "GabrielGraph: gg[i]");

    if (distance_cache)
        distances = distance_cache;
    else {
        MALLOC(distances, scalar *, instances, "GabrielGraph: distances");
        for (int i = 0; i < instances; i++)
            MALLOC(distances[i], scalar, instances, "GabrielGraph: distances");

        for (int i = 0; i < instances; i++) {
            distances[i][i] = 0.0;
            for (int j = 0; j < i; j++)
                distances[i][j] = distances[j][i] = subspace ? EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace) :
                                                    EuclideanDistanceSparseData(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL);
        }
    }

    for (int i = 0; i < instances; i++) {
        // Initially all other points can be candidates
        for (int j = 0; j < instances; j++)
            gg[i][j] = (i != j);

        // Every point in Ni
        for (int r = 0; r < instances; r++)
            if (gg[i][r]) {
                for (int k = 0; k < instances; k++) {
                    if (!GabrielLink(i, r, k, distances)) {
                        gg[i][r] = false;
                        break;
                    }

                    if (!GabrielLink(i, k, r, distances))
                        gg[i][k] = false;
                }
            }
    }

    if (!distance_cache) {
        for (int i = 0; i < instances; i++)
            FREE(distances[i]);
        FREE(distances);
    }

    return gg;
}

bool **RelativeNeighborhoodGraph(scalar **data, int inputs, int instances, int **index, scalar **distance_cache, int *subspace)
{
    bool **rng;
    scalar **distances;

    // Gabriel Graph and cache of distances
    MALLOC(rng, bool *, instances, "RelativeNeighborhoodGraph: rng");
    for (int i = 0; i < instances; i++)
        MALLOC(rng[i], bool, instances, "RelativeNeighborhoodGraph: rng[i]");

    if (distance_cache)
        distances = distance_cache;
    else {
        MALLOC(distances, scalar *, instances, "GabrielGraph: distances");
        for (int i = 0; i < instances; i++)
            MALLOC(distances[i], scalar, instances, "GabrielGraph: distances");

        for (int i = 0; i < instances; i++) {
            distances[i][i] = 0.0;
            for (int j = 0; j < i; j++)
                distances[i][j] = distances[j][i] = subspace ? EuclideanDistanceSparseDataSubspace(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL, subspace) :
                                                    EuclideanDistanceSparseData(data[i], data[j], inputs, index ? index[i] : NULL, index ? index[j] : NULL);
        }
    }

    for (int i = 0; i < instances; i++) {
        // Initially all other points can be candidates
        for (int j = 0; j < instances; j++)
            rng[i][j] = (i != j);

        // Every point in Ni
        for (int r = 0; r < instances; r++)
            if (rng[i][r]) {
                for (int k = 0; k < instances; k++) {
                    if (!RelativeNeighborhoodLink(i, r, k, distances)) {
                        rng[i][r] = false;
                        break;
                    }

                    if (!RelativeNeighborhoodLink(i, k, r, distances))
                        rng[i][k] = false;
                }
            }
    }

    if (!distance_cache) {
        for (int i = 0; i < instances; i++)
            FREE(distances[i]);
        FREE(distances);
    }

    return rng;
}

// Test whether i and j can form a Gabriel link considering k
bool GabrielLink(int i, int j, int k, scalar **distances)
{
    if (distances[i][j]*distances[i][j] <= distances[i][k]*distances[i][k] + distances[j][k]*distances[j][k])
        return true;
    else
        return false;
}

// Test whether i and j can form a Relative Neighborhood link considering k
bool RelativeNeighborhoodLink(int i, int j, int k, scalar **distances)
{
    if (distances[i][j] <= MAX(distances[i][k], distances[j][k]))
        return true;
    else
        return false;
}
