/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "is.h"
#include "macros.h"
#include "classifiers.h"

// MODIFICATIONS:       11/14/2009      Considered k != 1

// G. W. Gates, "The Reduced Nearest Neighbor Rule", IEEE Transactions on Information Theory, Vol. 13, 1972.

int ReducedNearestNeighbor(const t_Parameters *parm, int *S, scalar **data, int instances, int *label, int value, int inputs, int classes, cache_t **neighbors_cache, int **index, int *subspace)
{
    int nearest[parm->k], selected;

    // First create Condensed Nearest Neighbor Rule subset
    // S starts with first instance
    S[0] = 1;
    selected = 1;
    for (int i = 1; i < instances; i++)
        S[i] = 0;

    scalar *r;
    MALLOC(r, scalar, instances, "ReducedNearestNeighbor: r");
    for (int i = 0; i < instances; i++)
        r[i] = 1.0;

    // Test whether all instances are correctly classified
    for (int j = 1; j <= value; j++)
        for (int i = 0; i < instances; i++)
            if (S[i] < j && kNearestNeighbourCache(i, S, instances, nearest, label, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace)
                != label[i]) {
                // Add to S
                S[i]++;
                if (S[i] == 1)
                    selected++;

                // Restart
                i = 0;
            }

    if (parm->verbose > 1)
        fprintf(parm->out_file, "CNN removed %4d instances out of %4d\n", instances - selected, instances);

    if (parm->verbose > 1) {
        // Count multiple selected instances
        int c = 0;
        for (int i = 0; i < instances; i++)
            if (S[i] > 1)
                c++;

        fprintf(parm->out_file, "Multiple selected instances by CNN: %d of %d selected\n", c, selected);
    }

    // Obtain accuracy using whole CNN set and list of correctly classified
    scalar all_ok = 0.0;
    for (int i = 0; i < instances; i++)
        if (kNearestNeighbourCache(i, S, instances, nearest, label, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace) == label[i])
            all_ok ++;

    if (parm->verbose > 1)
        fprintf(parm->out_file, "Accuracy after CNN: %lf\n", all_ok/instances);

    int cnn_left = selected;

    // MODIFIED TO ACCOUNT FOR k != 1
    // Apply Reduced Nearest Neighbor rule
    // Remove each instance and test
    for (int i = 0; i < instances; i++) {
        bool noworse = true;
        // If selected, test whether removing it miss-classifies any instance
        while (S[i] && noworse) {
            // Temporally remove instance
            S[i]--;
            noworse = true;
            for (int j = 0; j < instances; j++)
                // If any miss-classified restore i-th instance
                if (kNearestNeighbourCache(j, S, instances, nearest, label, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace) != label[j]) {
                    noworse = false;    // Worse that using the instance
                    S[i]++;
                    break;
                }
        }
    }

    if (parm->verbose > 1) {
        scalar all_ok = 0.0;
        for (int i = 0; i < instances; i++)
            if (kNearestNeighbourCache(i, S, instances, nearest, label, parm->k, neighbors_cache, data, inputs, classes, parm->knn_with_weighted_votes, r, NULL, index, subspace) == label[i])
                all_ok ++;

        fprintf(parm->out_file, "Accuracy after RNN: %lf\n", all_ok/instances);
    }

    // Count selected
    selected = 0;
    for (int i = 0; i < instances; i++)
        if (S[i])
            selected++;

    if (parm->verbose > 1)
        fprintf(parm->out_file, "RNN removed %4d instances out of %4d\n", cnn_left - selected, cnn_left);

    FREE(r);

    return selected;
}
