/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "is.h"
#include "macros.h"
#include "svm.h"
#include "classifiers.h"

scalar EvaluateThresholdIS(const t_Parameters *parm, scalar threshold, scalar *votes, scalar **data, int instances, int *labels, scalar *reduction, scalar *accuracy, int inputs, int classes,
                           bool reset, int *evaluations, int minority, cache_t **neighbors_cache, int **index)
{
    int *S = NULL, **M, majority;
    scalar evaluation, Sp, Sn, theta_1, theta_2, average_margin, minimum_margin;
    t_Learner learner;

    if (index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for boosting instance selection");

    // Previous values
    static int sel;
    static scalar red, acc, eval;

    // Forget previous iteration
    if (reset) {
        sel = -1;
        red = acc = -1.0;
    }

    // Count selected
    int selected = 0;
    MALLOC(S, int, instances, "EvaluateThresholdIS: S");
    for (int i = 0; i < instances; i++)
        if (votes[i] >= threshold) {
            S[i] = 1;
            selected++;
        } else
            S[i] = 0;

    // At least as many as number of classes
    if (selected < classes) {
        *accuracy = 0.0;
        *reduction = 1.0;

        return 0.0;
    }

    // If same as previous assign values and return
    if (sel == selected) {
        *accuracy = acc;
        *reduction = red;

        return eval;
    } else
        (*evaluations)++;

    // Obtain error. i-th instance is not used classify i-th instance
    *accuracy = 0.0;

    // Evaluate fold
    AllocateIntegerMatrix(&M, classes, classes);
    ZeroIntegerMatrix(M, classes, classes);

    // Validate?
    if (parm->cv_hyperparameters)
        SetCrossValidatedParameters((t_Parameters *) parm, data, instances, labels, NULL, inputs, classes, parm->boostis_classifier, minority, S, index);

    // Train classifier
    ClassifierTraining(parm, parm->boostis_classifier, data, instances, labels, inputs, classes, NULL, NULL, S, true, &learner);

    if (parm->votes_threshold_measure != pfMargin)
        for (int i = 0; i < instances; i++)
            M[labels[i]][ClassifierOutput(parm, parm->boostis_classifier, data[i], inputs, classes, index ? index[i] : NULL, NULL, data, instances, labels, S, &learner, labels[i], NULL, -1, index)]++;
    else
        average_margin = ClassifierDatasetMargin(parm, parm->boostis_classifier, data, instances, labels, data, instances, labels, inputs, classes, NULL, &minimum_margin, &learner, NULL, S);

    ClassifierFree(parm, &learner, parm->boostis_classifier, classes, inputs, NULL);

    // Accuracy...
    switch (parm->votes_threshold_measure) {
        case pfAccuracy:
            *accuracy = 0.0;
            for (int i = 0; i < classes; i++)
                *accuracy += M[i][i];
            *accuracy /= instances;
            break;
        case pfGMean:
            majority = (minority+1)%2;
            Sp = M[majority][majority]+M[majority][minority] > ZERO ? M[majority][majority]/(M[majority][majority]+M[majority][minority]) : 0.0;
            Sn = M[minority][minority]+M[minority][majority] > ZERO ? M[minority][minority]/(M[minority][minority]+M[minority][majority]) : 0.0;

            *accuracy= pow(Sp*Sn, 0.5);
            break;

        case pfauROC:
            if (classes != 2)
                OnErrorExit(1, "AUC with more than two classes");

            *accuracy = ClassifierAUROC(parm, parm->boostis_classifier, data, instances, labels, data, instances, S, NULL, labels, inputs, minority, false, &learner, true, NULL, classes,
                                              NULL, NULL, NULL, NULL);
            ClassifierFree(parm, &learner, parm->boostis_classifier, classes, inputs, NULL);
            break;

        case pfMargin:
            *accuracy = average_margin;

        case pfKappa:
            // Obtain kappa
            theta_1 = theta_2 = 0.0;

            // Sum of the diagonal
            for (int j = 0; j < classes; j++)
                theta_1 += M[j][j];
            theta_1 *= instances;

            for (int i = 0; i < classes; i++) {
                // Rows/columns total counts
                scalar row_i = 0, column_i = 0;
                for (int j = 0; j < classes; j++) {
                    row_i += M[i][j];
                    column_i += M[j][i];
                }

                theta_2 += row_i*column_i;
            }

            *accuracy = ((scalar) instances)*instances-theta_2 > ZERO ? (theta_1-theta_2)/(((scalar) instances)*instances-theta_2) : 0.0;
            break;

        default:
            OnErrorExit(1, "I should not be here");
    }

    if (selected == instances) {
        *reduction = 0.0;
        *accuracy = 0.0;
    }

    *reduction = 1. - ((scalar) selected)/instances;
    evaluation = parm->alpha_acc **accuracy+(1.-parm->alpha_acc) **reduction;

    FREE(S);
    FreeIntegerMatrix(&M, classes);

    // Store for next threshold
    red = *reduction;
    acc = *accuracy;
    sel = selected;
    eval = evaluation;

    return evaluation;
}
