/***************************************************************************
 *   Copyright (C) 2015 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "typex.h"
#include "common.h"
#include "macros.h"

#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>

void InvertMatrix(scalar **m, scalar **inverse, int f)
{
    gsl_matrix *im, *lu;
    gsl_permutation *p;

    int i, j, signum;

    lu = gsl_matrix_alloc(f, f);
    im = gsl_matrix_alloc(f, f);
    p = gsl_permutation_alloc(f);

    for (i = 0; i < f; i++)
        for (j = 0; j < f; j++)
            gsl_matrix_set(lu, i, j, m[i][j]);

    /* LU decomposition. */
    gsl_linalg_LU_decomp(lu, p, &signum);
    /* Invert matrix. */
    gsl_linalg_LU_invert(lu, p, im);

    for (i = 0; i < f; i++)
        for (j = 0; j < f; j++)
            inverse[i][j] = gsl_matrix_get(im, i, j);

    // Free memory
    gsl_matrix_free(lu);
    gsl_matrix_free(im);
    gsl_permutation_free(p);
}


/* Determinant of an square matrix. */
scalar MatrixDeterminant(scalar **m, int f)
{
    gsl_matrix *lu;
    gsl_permutation *p;

    int i, j, signum;

    lu = gsl_matrix_alloc(f, f);
    p = gsl_permutation_alloc(f);

    for (i = 0; i < f; i++)
        for (j = 0; j < f; j++)
            gsl_matrix_set(lu, i, j, m[i][j]);

    /* LU decomposition. */
    gsl_linalg_LU_decomp(lu, p, &signum);

    /* Determinant. */
    scalar determinant = gsl_linalg_LU_det(lu, signum);

    // Free memory
    gsl_matrix_free(lu);
    gsl_permutation_free(p);

    return determinant;
}

void MaskedMatrixVectorProduct(scalar **m, scalar *v, scalar *res, int n, int *S)
{
    for (int i = 0; i < n; i++) {
        res[i] = 0.0;
        if (S[i])
            for (int j = 0; j < n; j++)
                res[i] += m[i][j] * v[j] * S[j];
    }
}

scalar MaskedDotProduct(scalar *x, scalar *y, int n, int *S)
{
    scalar dot = 0.0;

    for (int i = 0; i < n; i++)
        dot += x[i] * y[i] * S[i];

    return dot;
}

void PrintMatrix(scalar **m, int f, int c, FILE *file, int format_i, int format_d)
{
    for (int i = 0; i < f; i++) {
        for (int j = 0; j < c; j++)
            fprintf(file, "%*.*lf ", format_i+format_d+1, format_d, m[i][j]);
        fprintf(file, "\n");
    }
}

scalar MaximumOfVector(scalar *v, int n)
{
    scalar maximum = v[0];

    for (int i = 1; i < n; i++)
        if (v[i] > maximum)
            maximum = v[i];

    return maximum;
}

int MinimumOfIntegerVector(int *v, int n)
{
    int minimum = v[0];

    for (int i = 1; i < n; i++)
        if (v[i] < minimum)
            minimum = v[i];

    return minimum;
}

void PrintVector(scalar *v, int *w, int c, FILE *file)
{
    if (v)
        for (int j = 0; j < c; j++)
            fprintf(file, "%6.4lf ", v[j]);
    else
        for (int j = 0; j < c; j++)
            fprintf(file, "%d ", w[j]);
    fprintf(file, "\n");
}


int IndexOfMinimumOfVector(scalar *w, int *v, int n)
{
    int minimum = 0;

    if (w) {
        for (int i = 1; i < n; i++)
            if (w[i] < w[minimum])
                minimum = i;
    } else {
        for (int i = 1; i < n; i++)
            if (v[i] < v[minimum])
                minimum = i;
    }

    return minimum;
}

bool CheckMatrix(scalar **m, int r, int c)
{
    for (int i = 0; i < r; i++)
        for (int j = 0; j < c; j++)
            if (isnan(m[i][j]) || isinf(m[i][j]))
                return false;

    return true;
}

bool CheckVector(scalar *v, int c)
{
    for (int j = 0; j < c; j++)
        if (isnan(v[j]) || isinf(v[j]))
            return false;

    return true;
}

int MaximumOfIntegerVector(int *v, int n)
{
    int maximum = v[0];

    for (int i = 1; i < n; i++)
        if (v[i] > maximum)
            maximum = v[i];

    return maximum;
}

int IndexOfMaximumOfIntegerVector(int *v, int n)
{
    int maximum = 0;

    for (int i = 1; i < n; i++)
        if (v[i] > v[maximum])
            maximum = i;

    return maximum;
}

// IF more than one choose randomly
int IndexOfMaximumOfScalarVector(scalar *v, int n)
{
    int maximum = 0;

    for (int i = 1; i < n; i++)
        if (v[i] > v[maximum])
            maximum = i;
        else if (fabs(v[i]-v[maximum]) < ZERO && RANDOM(0.0, 1.0) < 0.5)
            maximum = i;

    return maximum;
}

// Compact vector removing duplicated values
// The vector must be sorted

void CompactVector(scalar *v, int *w, int *n)
{
    if (v) {
        for (int i = 0; i < *n-1; i++)
            if (fabs(v[i]-v[i+1]) <= ZERO) {
                for (int j = i+1; j < *n-1; j++)
                    v[j] = v[j+1];
                (*n)--;
            }
    } else {
        for (int i = 0; i < *n-1; i++)
            if (w[i] == w[i+1]) {
                for (int j = i+1; j < *n-1; j++)
                    w[j] = w[j+1];
                (*n)--;
            }
    }
}

void ZeroScalarMatrix(scalar **m, int r, int c)
{
    for (int i = 0; i < r; i++)
        for (int j = 0; j < c; j++)
            m[i][j] = 0.0;
}

void FillMatrix(scalar **M, int **m, int r, int c, scalar value)
{
    if (M) {
        for (int i = 0; i < r; i++)
            for (int j = 0; j < c; j++)
                M[i][j] = value;
    } else {
        for (int i = 0; i < r; i++)
            for (int j = 0; j < c; j++)
                m[i][j] = (int) value;
    }
}

void CopyIntegerMatrix(int **orig, int **dest, int r, int c)
{
   for (int i = 0; i < r; i++)
       for (int j = 0; j < c; j++)
           dest[i][j] = orig[i][j];
}

void OrIntegerMatrix(int **orig, int **M, int r, int c)
{
    for (int i = 0; i < r; i++)
        for (int j = 0; j < c; j++)
            orig[i][j] = orig[i][j] || M[i][j];
}
