/**************************************************************************
*   Copyright (C) 2010 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#include <math.h>
#include "common.h"
#include "macros.h"
#include "maths.h"

void PrincipalComponentsAnalysis(scalar **data, int s, int v, int no_comps, scalar ***data_cp, scalar **coefficients)
{
    scalar sws;
    scalar **cor, **variance, **products;
    scalar *d, *e;
    int sw, tmi;

    // Memory allocation
    MALLOC(cor, scalar *, v, "PrincipalComponentsAnalysis: cor");
    MALLOC(variance, scalar *, v, "PrincipalComponentsAnalysis: variance");
    MALLOC(products, scalar *, v, "PrincipalComponentsAnalysis: products");
    for (int i = 0; i < v; i++) {
        MALLOC(cor[i], scalar, v, "PrincipalComponentsAnalysis: cor[i]");
        MALLOC(variance[i], scalar, v, "PrincipalComponentsAnalysis: variance[i]");
        MALLOC(products[i], scalar, v, "PrincipalComponentsAnalysis: products[i]");
    }
    MALLOC(d, scalar, v, "PrincipalComponentsAnalysis: d");
    MALLOC(e, scalar, v, "PrincipalComponentsAnalysis: e");

    // Data pre-processing
    MeanCorrect(data, data, s, v);
    Variance(data, variance, s, v);
    CrossProducts(data, products, s, v);
    TypifyData(data, data, variance, s, v);
    Correlation(variance, products, cor, s, v);

    // Eigenvalues and eigenvectors
    TridiagonalizeHH(cor, d, e, v);
    EigenValuesVectors(d, e, v, cor, &tmi);

    // Sort eigenvalues and eigenvectors
    do {
        sw = false;
        for (int i = 0; i < v - 1; i++)
            if (d[i] < d[i + 1]) {
                sws = d[i];
                d[i] = d[i + 1];
                d[i + 1] = sws;
                for (int j = 0; j < v; j++) {
                    sws = cor[j][i];
                    cor[j][i] = cor[j][i + 1];
                    cor[j][i + 1] = sws;
                }
                sw = true;
            }
    } while (sw == true);

    // Construct components
    if (data_cp != NULL) {
        // Matrix for components ...
        MALLOC(*data_cp, scalar *, s, "PrincipalComponentsAnalysis: *data_cp");
        for (int i = 0; i < s; i++)
            MALLOC((*data_cp) [i], scalar, no_comps,
                   "PrincipalComponentsAnalysis: (*data_cp)[i]");

        // Get components
        MatrixProduct(data, cor, *data_cp, s, v, no_comps);
    } else {
        // ... or return coefficients
        for (int i = 0; i < no_comps; i++)
            for (int j = 0; j < v; j++)
                coefficients[i][j] = cor[j][i];
    }

    // Free memory
    FREE(d);
    FREE(e);
    for (int i = 0; i < v; i++) {
        FREE(variance[i]);
        FREE(products[i]);
        FREE(cor[i]);
    }
    FREE(products);
    FREE(variance);
    FREE(cor);
}

void TypifyData(scalar **data_mc, scalar **data_t, scalar **variance, int s, int v)
{
    for (int i = 0; i < v; i++) {
        if (variance[i][i] < ZERO)
            variance[i][i] = 1.0;
        else
            variance[i][i] = 1.0 / sqrt(variance[i][i]);
    }
    // Typify
    for (int i = 0; i < s; i++)
        for (int j = 0; j < v; j++)
            data_t[i][j] = data_mc[i][j] * variance[j][j];
}

void Variance(scalar **data_mc, scalar **variance, int s, int v)
{

    for (int i = 0; i < v; i++)
        for (int j = 0; j < v; j++)
            variance[i][j] = 0.0;

    // Get variance
    for (int i = 0; i < v; i++) {
        for (int h = 0; h < s; h++)
            variance[i][i] += data_mc[h][i] * data_mc[h][i];
        variance[i][i] /= (s - 1);
    }
}


void MeanCorrect(scalar **data, scalar **data_mc, int s, int v)
{
    scalar *cent;

    MALLOC(cent, scalar, v, "MeanCorrect: cent");

    // Centroid
    Centroid(data, s, v, cent);

    for (int i = 0; i < s; i++)
        for (int j = 0; j < v; j++)
            data_mc[i][j] -= cent[j];

    FREE(cent);
}

void Centroid(scalar **data, int s, int v, scalar *centroid)
{
    for (int i = 0; i < v; i++) {
        centroid[i] = 0.0;
        for (int j = 0; j < s; j++)
            centroid[i] += data[j][i];
        centroid[i] /= s;
    }
}


void Correlation(scalar **variance, scalar **products, scalar **correlacion, int s,
                 int v)
{
    for (int i = 0; i < v; i++)
        for (int j = 0; j < v; j++)
            correlacion[i][j] = (products[i][j] * variance[i][i]
                                 * variance[j][j]) / (scalar)(s - 1);
}

void CrossProducts(scalar **data_mc, scalar **products, int s, int v)
{
    for (int i = 0; i < v; i++)
        for (int j = 0; j < v; j++) {
            products[i][j] = 0.0;
            for (int k = 0; k < s; k++)
                products[i][j] += data_mc[k][i] * data_mc[k][j];
        }
}

void MatrixProduct(scalar **m1, scalar **m2, scalar **res, int fil, int fc, int col)
{
    for (int i = 0; i < fil; i++)
        for (int j = 0; j < col; j++) {
            res[i][j] = 0.0;
            for (int h = 0; h < fc; h++)
                res[i][j] += m1[i][h] * m2[h][j];
        }
}

void EigenValuesVectors(scalar *d, scalar *e, int v, scalar **z, int *sw)
{
    int m, l, iter, i, k;
    scalar s, r, p, g, f, dd, c, b;

    *sw = 0;
    for (i = 1; i < v; i++)
        e[i - 1] = e[i];

    e[v - 1] = 0.0;
    for (l = 0; l < v; l++) {
        iter = 0;
label_1:
        for (m = l; m < v - 1; m++) {
            dd = fabs(d[m]) + fabs(d[m + 1]);
            if ((fabs(e[m]) + dd) == dd)
                goto label_2;
        }

        m = v - 1;
label_2:
        if (m != l) {
            if (iter == 30) {
                *sw = 1;
                return;
            }

            iter++;
            g = (d[l + 1] - d[l]) / (2.0 * e[l]);
            r = (sqrt(g * g + 1.0));
            g = d[m] - d[l] + e[l] / (g + Sign2(r, g));
            s = c = 1.0;
            p = 0;
            for (i = m - 1; i >= l; i--) {
                f = s * e[i];
                b = c * e[i];
                if (fabs(f) >= fabs(g)) {
                    c = g / f;
                    r = sqrt(c * c + 1.0);
                    e[i + 1] = f * r;
                    s = 1.0 / r;
                    c *= s;
                } else {
                    s = f / g;
                    r = sqrt(s * s + 1.0);
                    e[i + 1] = g * r;
                    c = 1.0 / r;
                    s *= c;
                }
                g = d[i + 1] - p;
                r = (d[i] - g) * s + 2.0 * c * b;
                p = s * r;
                d[i + 1] = g + p;
                g = c * r - b;
                for (k = 0; k < v; k++) {
                    f = z[k][i + 1];
                    z[k][i + 1] = s * z[k][i] + c * f;
                    z[k][i] = c * z[k][i] - s * f;
                }
            }
            d[l] -= p;
            e[l] = g;
            e[m] = 0.0;
            goto label_1;
        }
    }
}

void TridiagonalizeHH(scalar **a, scalar *d, scalar *e, int v)
{
    int i, j, k, l;
    scalar scale, hh, h, g, f;

    for (i = v - 1; i > 0; i--) {
        l = i - 1;
        scale = h = 0.0;
        if (l > 0) {
            for (k = 0; k <= l; k++)
                scale += fabs(a[i][k]);

            if (scale == 0.0)
                e[i] = a[i][l];
            else {
                for (k = 0; k <= l; k++) {
                    a[i][k] /= scale;
                    h += a[i][k] * a[i][k];
                }
                f = a[i][l];
                g = -Sign2(sqrt(h), f);
                e[i] = scale * g;
                h -= f * g;
                a[i][l] = f - g;
                f = 0.0;
                for (j = 0; j <= l; j++) {
                    a[j][i] = a[i][j] / h;
                    g = 0.0;
                    for (k = 0; k <= j; k++)
                        g += a[j][k] * a[i][k];

                    if (l > j)
                        for (k = j + 1; k <= l; k++)
                            g += a[k][j] * a[i][k];

                    e[j] = g / h;
                    f += e[j] * a[i][j];
                }
                hh = f / (h + h);
                for (j = 0; j <= l; j++) {
                    f = a[i][j];
                    g = e[j] - hh * f;
                    e[j] = g;
                    for (k = 0; k <= j; k++)
                        a[j][k] = a[j][k] - f * e[k] - g * a[i][k];
                }
            }
        } else
            e[i] = a[i][l];

        d[i] = h;
    }
    d[0] = e[0] = 0.0;
    for (i = 0; i < (int) v; i++) {
        l = i - 1;
        if (d[i] != 0.0) {
            for (j = 0; j <= l; j++) {
                g = 0.0;
                for (k = 0; k <= l; k++)
                    g += a[i][k] * a[k][j];

                for (k = 0; k <= l; k++)
                    a[k][j] -= g * a[k][i];
            }
        }
        d[i] = a[i][i];
        a[i][i] = 1.0;
        if (l >= 0)
            for (j = 0; j <= l; j++)
                a[i][j] = a[j][i] = 0.0;
    }
}

scalar Sign2(scalar a, scalar b)
{
    if (b < 0.0)
        return (-fabs(a));
    else
        return (fabs(a));
}
