/***************************************************************************
 *   Copyright (C) 2010 by César Garcia-Osorio                             *
 *   cgosorio@ubu.es                                                       *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*
 * EXPLANATION
 * This program illustrate the use of getVec that generates a sequence
 * of vector using the random curve technique.
 * The typical function call sequence is:
 * getVec(INIT,23,0.02); // Init the vec generation dim: 23, step: 0.02
 * getVec(SET,vec); // Set the initial direction to vec
 * getVec(NEXT,vec); // Get the next direction in vec
 * getVec(CLEAR); // Free memory
 *
 * AUTHOR: Cesar Garcia-Osorio
 * DATE: May 31st, 2008.
 */

#define SET     1
#define INIT    2
#define NEXT    3
#define CLEAR   4
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include "macros.h"
#include "maths.h"
#include <gsl/gsl_math.h>

scalar EuclideanDistanceSparseData(scalar *v, scalar *u, int n, int *v_index, int *u_index)
{
    scalar dist = 0.0;

    if (v_index) {
        int i = 0, j = 0;

        // Common inputs
        while (v_index[i] != -1 && u_index[j] != -1) {
            if (v_index[i] == u_index[j]) {
                dist += (v[i] - u[j]) * (v[i] - u[j]);
                i++;
                j++;
            } else {
                if (v_index[i] > u_index[j]) {
                    dist += u[j]*u[j];
                    j++;
                } else {
                    dist += v[i]*v[i];
                    i++;
                }
            }
        }

        // End of one of the vectors
        while (v_index[i] != -1) {
            dist += v[i]*v[i];
            i++;
        }

        while (u_index[j] != -1) {
            dist += u[j]*u[j];
            j++;
        }
    } else
        for (int i = 0; i < n; i++)
            dist += (v[i] - u[i]) * (v[i] - u[i]);

    return sqrt(dist);
}


scalar SqEuclideanDistance(scalar *v, scalar *u, int n)
{
    scalar dist = 0.0;

    for (int i = 0; i < n; i++)
        dist += (v[i] - u[i]) * (v[i] - u[i]);

    return dist;
}

//Get euclidean distance between vectors v and w
scalar EuclideanDistance(scalar *v, scalar *w, int n)
{
    scalar dist = 0.0;
    int i;

    for (i = 0; i < n; i++)
        dist += (v[i] - w[i]) * (v[i] - w[i]);

    return sqrt(dist);
}

scalar EuclideanDistanceSparseDataSubspace(scalar *v, scalar *u, int n, int *v_index, int *u_index, int *subspace)
{
    scalar dist = 0.0;

    if (v_index || u_index) {
        if (subspace) {
            int i = 0, j = 0;

            // Common inputs
            while (v_index[i] != -1 && u_index[j] != -1) {
                if (v_index[i] == u_index[j]) {
                    dist += (v[i] - u[j]) * (v[i] - u[j]) * subspace[v_index[i]-1];
                    i++;
                    j++;
                } else {
                    if (v_index[i] > u_index[j]) {
                        dist += u[j]*u[j] * subspace[u_index[j]-1];
                        j++;
                    } else {
                        dist += v[i]*v[i] * subspace[v_index[i]-1];
                        i++;
                    }
                }
            }

            // End of one of the vectors
            while (v_index[i] != -1) {
                dist += v[i]*v[i] * subspace[v_index[i]-1];
                i++;
            }

            while (u_index[j] != -1) {
                dist += u[j]*u[j] * subspace[u_index[j]-1];
                j++;
            }
        } else {
            int i = 0, j = 0;

            // Common inputs
            while (v_index[i] != -1 && u_index[j] != -1) {
                if (v_index[i] == u_index[j]) {
                    dist += (v[i] - u[j]) * (v[i] - u[j]);
                    i++;
                    j++;
                } else {
                    if (v_index[i] > u_index[j]) {
                        dist += u[j]*u[j];
                        j++;
                    } else {
                        dist += v[i]*v[i];
                        i++;
                    }
                }
            }

            // End of one of the vectors
            while (v_index[i] != -1) {
                dist += v[i]*v[i];
                i++;
            }

            while (u_index[j] != -1) {
                dist += u[j]*u[j];
                j++;
            }
        }
    } else if (subspace)
        for (int i = 0; i < n; i++)
            dist += (v[i] - u[i]) * (v[i] - u[i]) * subspace[i];
    else
        for (int i = 0; i < n; i++)
            dist += (v[i] - u[i]) * (v[i] - u[i]);

    return sqrt(dist);
}

// #define M_PI 3.14159265358979323846f

void getVec(int OP, ...)
{
    va_list ap;

    static int dim;
    static long int p;
    static scalar *last;
    static scalar *next;
    static scalar *alpha;
    static scalar *ivec;
    static scalar stepSize;
    static int initialized = 0;
    scalar *vec;

    va_start(ap, OP);

    switch (OP) {
        case INIT:                      // getVec(INIT, int dim, scalar step)
            if (initialized) {
                fprintf(stderr, "getVec error: already initialized\n");
                exit(-1);
            }
            dim = va_arg(ap, int);
            stepSize = va_arg(ap, double);

            p = ((long int) dim) * (dim - 1)/2;
            alpha = (scalar *) malloc(sizeof(scalar) * p);
            last = (scalar *) malloc(sizeof(scalar) * p);
            next = (scalar *) malloc(sizeof(scalar) * p);

            nextAlpha(dim, p, stepSize, last, next, NULL);          // INITIALIZATION

            // The unit vector that is going to be rotate if not initial direction is provided.
            //ivec=(scalar *)malloc(sizeof(scalar)*dim);for (i=1;i<dim;i++) ivec[i]=0; ivec[0]=1;
            ivec = (scalar *) calloc(dim, sizeof(scalar));
            ivec[0] = 1;

            initialized = 1;
            break;
        case SET:                       // getVec(SET, scalar * ivec)
            if (!initialized) {
                fprintf(stderr, "getVec error: no initialization\n");
                exit(-1);
            }
            vec = va_arg(ap, scalar *);
            memcpy(ivec, vec, dim * sizeof(scalar));        // for (i=0;i<dim;i++) ivec[i]=vec[i];
            normalize(dim, ivec);
            break;
        case NEXT:                      // getVec(NEXT, scalar * nvec)
            if (!initialized) {
                fprintf(stderr, "getVec error: no initialization\n");
                exit(-1);
            }
            vec = va_arg(ap, scalar *);
            nextAlpha(dim, p, stepSize, last, next, alpha);
            memcpy(vec, ivec, dim * sizeof(scalar));        // for (i=0;i<dim;i++) vec[i]=ivec[i];
            applyRotation(dim, alpha, vec);
            break;
        case CLEAR:             // getVec(CLEAR)
            free(alpha);
            free(ivec);
            free(last);
            free(next);
            initialized = 0;
            break;
    }
    va_end(ap);
}

void nextAlpha(int dim, int p, scalar stepSize, scalar *last, scalar *next, scalar *alpha)
// stepSize is in degrees [1,360]
{
    static scalar t;
    static int nsteps;
    static int cstep;
    static scalar step;
    int i;
    if (stepSize <= 0) {
        fprintf(stderr, "Error in nextAlpha: stepSize must be positive and non zero\n");
        exit(-1);
    }
    if (alpha == NULL) {    // INITIALIZATION ----------
        scalar *aux = (scalar *) malloc(sizeof(scalar) * p);
        // the generation of s_1 and s_2 -------------------
        //    randvecOfAngles(p,last); // using this there is a big jump from initial vector
        zeros(p, last);                 // but with this the first vectors are similar to the initial one
        randvecOfAngles(p, next);
        {
            // the next one for debuggin purposes
            for (i = 0; i < p; i++)
                aux[i] = next[i] - last[i];
            // How many steps between last and next
            nsteps = (int)(0.5+360.0 * norm(p, aux)/(2.0 * M_PI * stepSize));
            step = 1.0/nsteps;
        }
        // END OF the generation of s_1 and s_2 --------------
        cstep = 0;
        t = 1;
        free(aux);
    } else {
        for (i = 0; i < p; i++)
            alpha[i] = t * last[i]+(1 - t) * next[i];     // lineal interpolant

        cstep = cstep+1;
        if (cstep == (nsteps - 1)) {
            scalar *aux = (scalar *) malloc(sizeof(scalar) * p);
            // the generation of s_n and s_{n+1} -------------------
            memcpy(last, next, p * sizeof(scalar));         //for (i=0;i<p;i++) last[i]=next[i];
            randvecOfAngles(p, next);
            {
                for (i = 0; i < p; i++)
                    aux[i] = next[i] - last[i];
                // How many steps between last and next
                // sqrt(dim)*2*M_PI is the size of the larger diagonal of the hypercube
                nsteps =
                    (int)(0.5 +
                          (stepSize * 360.0/0.0175) * norm(p,
                                  aux)/(sqrt(dim) * 2.0 *
                                        M_PI));
                step = 1.0/nsteps;
            }
            // END OF the generation of s_n and s_{n+1} --------------
            cstep = 0;
            t = 1;
            free(aux);
        } else {
            t = t - step;
        }
    }
}


void applyRotation(int dim, scalar *alpha, scalar *vec)
{
    int e_i;
    int e_j;
    int a = 0;
    scalar theta;
    scalar cosangle;
    scalar sinangle;
    scalar new_vec_i;
    scalar new_vec_j;

    for (e_i = 0; e_i < dim; e_i++)
        for (e_j = e_i+1; e_j < dim; e_j++) {
            theta = alpha[a];
            a = a+1;
            cosangle = cos(theta);
            sinangle = sin(theta);
            new_vec_i = vec[e_i] * cosangle+vec[e_j] * sinangle;
            new_vec_j = -vec[e_i] * sinangle+vec[e_j] * cosangle;
            vec[e_i] = new_vec_i;
            vec[e_j] = new_vec_j;
        }
    // WARNING: I do not normalize since I am rotating a unit vector,
    // but if this is going to be over and over, the vector could
    // degenerate is one with different norm than 1.
    // To avoid that, use:
    // normalize(dim, vec);
}

scalar norm(int dim, scalar *vec)
{
    int i;
    scalar norma = 0.0;
    for (i = 0; i < dim; i++, vec++)
        norma += (*vec) * (*vec);       // norma+=vec[i]*vec[i];
    return sqrt(norma);
}

void normalize(int dim, scalar *vec)
{
    int i;
    scalar norma = norm(dim, vec);
    for (i = 0; i < dim; i++, vec++)
        (*vec) /= norma;        // vec[i]/=norma;
}



// ----- All the following code was just to play a little bit -----------------
// inline
void zeros(int dim, scalar *vec)
{
    //int i;for (i=0;i<dim;i++) vec[i]=0;
    memset(vec, 0, dim * sizeof(scalar));
}

void ones(int dim, scalar *vec)
{
    int i;
    for (i = 0; i < dim; i++)
        vec[i] = 1;
}

void printvec(int dim, scalar *vec)
{
    int i;
    printf("[");
    for (i = 0; i < dim; i++)
        printf("%6.4lf ", vec[i]);
    printf("]\n");
}

/*******************************************************

Note: Do NOT use
      y = rand()  %  M;
as this focuses on the lower bits of rand().
For linear congruential random number generators,
which rand() often is, the lower bytes are much
less random than the higher bytes. In fact the
lowest bit cycles between 0 and 1. Thus rand()
may cycle between even and odd (try it out).
Note rand() does not have to be a linear
congruential random number generator.
It's perfectly permissible for it to be
something better which does not have this problem.

[FROM: http://members.cox.net/srice1/random/crandom.html]

*********************************************************/


void randvecOfAngles(int dim, scalar *vec)
{
    int i;
    for (i = 0; i < dim; i++)
        vec[i] = RANDOM(0.0, 1.0) * 2.0 * M_PI;
}

void randvec(int dim, scalar *vec)
{
    int i;
    for (i = 0; i < dim; i++)
        vec[i] = RANDOM(0.0, 1.0);
}

scalar SumVector(scalar *v, int *w, int size)
{
    scalar sum = 0.0;

    if (v)
        for (int i = 0; i < size; i++)
            sum += v[i];
    else
        for (int i = 0; i < size; i++)
            sum += w[i];

    return sum;
}

int MaxVectorIndex(scalar *v, int *w, int size)
{
    int max = 0;

    if (v) {
        for (int i = 1; i < size; i++)
            if (v[i] > v[max])
                max = i;
    } else {
        for (int i = 1; i < size; i++)
            if (w[i] > w[max])
                max = i;
    }

    return max;
}

scalar AverageScalarVector(scalar *v, int n)
{
    scalar average = 0.0;
    for (int i = 0; i < n; i++)
        average += v[i];

    return average/n;
}

scalar AverageIntegerVector(int *v, int n)
{
    scalar average = 0.0;
    for (int i = 0; i < n; i++)
        average += v[i];

    return average/n;
}

scalar SumColumn(scalar **m, int rows, int column)
{
    scalar sum = 0.0;

    for (int i = 0; i < rows; i++)
        sum += m[i][column];

    return sum;
}

scalar AverageIntegerMatrix(int **m, int rows, int columns)
{
    scalar average = 0.0;
    for (int i = 0; i < rows; i++)
        for (int j = 0; j < columns; j++)
            average += m[i][j];

    return average/(rows*columns);
}

bool EmptyRow(int *row, int n)
{
    for (int i = 0; i < n; i++)
        if (row[i])
            return false;

    return true;
}

bool EmptyColumn(int **m, int rows, int column)
{
    for (int i = 0; i < rows; i++)
        if (m[i][column])
            return false;

    return true;
}

//Get euclidean distance between vectors v and w
scalar MinkowskiDistance(scalar *v, scalar *w, int n, scalar f, int *subspace)
{
    scalar dist = 0.0;
    int i;

    if (subspace) {
        for (i = 0; i < n; i++)
            if (subspace[i])
                dist += pow(fabs(v[i]-w[i]), f);
    } else
        for (i = 0; i < n; i++)
            dist += pow(fabs(v[i]-w[i]), f);

    return pow(dist, 1.0/f);
}

//Get euclidean distance between vectors v and w
scalar MinkowskiDistanceSparse(scalar *v, scalar *w, int n, scalar f, int *subspace, int *v_index, int *w_index)
{
    scalar dist = 0.0;
    int i;

    if (v_index) {
        int i = 0, j = 0;

        // Common inputs
        while (v_index[i] != -1 && w_index[j] != -1) {
            if (v_index[i] == w_index[j]) {
                dist += pow(fabs(v[i]-w[j]), f);
                i++;
                j++;
            } else {
                if (v_index[i] > w_index[j]) {
                    dist += pow(w[j]*w[j], f);
                    j++;
                } else {
                    dist += pow(v[i]*v[i], f);
                    i++;
                }
            }
        }

        // End of one of the vectors
        while (v_index[i] != -1) {
            dist += pow(v[i]*v[i], f);
            i++;
        }

        while (w_index[j] != -1) {
            dist += pow(w[j]*w[j], f);
            j++;
        }

        return pow(dist, 1.0/f);
    } else {
        if (subspace) {
            for (i = 0; i < n; i++)
                if (subspace[i])
                    dist += pow(fabs(v[i]-w[i]), f);
        } else
            for (i = 0; i < n; i++)
                dist += pow(fabs(v[i]-w[i]), f);

        return pow(dist, 1.0/f);
    }
}

//Get euclidean distance between vectors v, sparse, and w, dense
scalar MinkowskiDistanceSparseDense(scalar *v, scalar *w, int n, scalar f, int *subspace, int *v_index)
{
    scalar dist = 0.0;
    int i;

    if (v_index) {
        if (!subspace) {
            int j = 0, x = 0;
            for (; v_index[j] != -1; x++)
                if (v_index[j]-1 == x) {
                    dist += pow(v[j]-w[x], f);
                    j++;
                } else
                    dist += pow(w[x], f);

            for (; x < n; x++)
                dist += pow(w[x], f);

            return pow(dist, 1.0/f);
        } else {
            int j = 0, x = 0;
            for (; v_index[j] != -1; x++)
                if (v_index[j]-1 == x) {
                    dist += pow(v[j]-w[x], f)*subspace[x];
                    j++;
                } else
                    dist += pow(w[x], f)*subspace[x];

            for (; x < n; x++)
                dist += pow(w[x], f)*subspace[x];

            return pow(dist, 1.0/f);
        }
    } else {
        if (subspace) {
            for (i = 0; i < n; i++)
                if (subspace[i])
                    dist += pow(fabs(v[i]-w[i]), f);
        } else
            for (i = 0; i < n; i++)
                dist += pow(fabs(v[i]-w[i]), f);

        return pow(dist, 1.0/f);
    }
}

// Returns the maximum value of the given column

scalar MaximumOfMatrixColumn(scalar **M, int **m, int rows, int c)
{
    scalar max;

    if (M) {
        max = M[0][c];
        for (int i = 1; i < rows; i++)
            if (max < M[i][c])
                max = M[i][c];
    } else {
        max = m[0][c];
        for (int i = 1; i < rows; i++)
            if (max < m[i][c])
                max = m[i][c];
    }

    return max;
}
