#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 11:41:15 2021

@author: frodo
"""

import argparse
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.datasets import make_moons, make_circles, make_classification
from sklearn.neural_network import MLPClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import RBF
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from skmultilearn.problem_transform import ClassifierChain
import sklearn
import arff
from skmultilearn.dataset import load_from_arff
import warnings
import sklearn.metrics as metrics
import numpy as np
import sys
import random
import scipy

from parameters import rfparam

warnings.filterwarnings("ignore")

parser = argparse.ArgumentParser()
parser.add_argument("dataset", help="Dataset (must be a -.train and a -.test)", type=str)
parser.add_argument("nlabels", help="Number of labels", type=int)
parser.add_argument("method", help="Based method (as defined in 'typex.h'", type=int)

parser.add_argument("-sparse", help="File is in sparse format", action='store_true')
parser.add_argument("-cv", help="Cross-validate hyper-parameters", action='store_true')
parser.add_argument("-r", help="Use random order of labels", action='store_true')
parser.add_argument("-v", help="Verbose mode", action='store_true')
args = parser.parse_args()

classifiers = [
    LogisticRegression(),
    KNeighborsClassifier(3),
    RandomForestClassifier(n_estimators=rfparam['n_estimators'], max_depth=rfparam['max_depth'], min_samples_split=rfparam['min_samples_split']),
    DecisionTreeClassifier(max_depth=10),
    GaussianNB(),
    SVC(kernel="linear", C=1, probability=True),
    SVC(gamma=2.0, C=1, probability=True),
    AdaBoostClassifier(n_estimators=10),
    MLPClassifier()
    ]

# Hyper-parameters to optimize depending on the classifier
#parameters = [{},
              #{'classifier__n_neighbors': range(1,10)},
              #{'classifier__n_estimators': [10, 20, 30]},
              #{},
              #{},
              #{'classifier__C': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]},
              #{'classifier__C': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0], 'classifier__gamma': [0.1, 1.0, 10.0]},
              #{'classifier__n_estimators': [10, 30, 50]},
              #{}
    #]

# Hyper-parameters to optimize
parameters = {'classifier':[
    KNeighborsClassifier(3),
    DecisionTreeClassifier(max_depth=10),
    RandomForestClassifier(n_estimators=rfparam['n_estimators'], max_depth=rfparam['max_depth'], min_samples_split=rfparam['min_samples_split']),
    AdaBoostClassifier(n_estimators=10, base_estimator=DecisionTreeClassifier(max_depth=10))]
}

# SVM and validate use only SVM models
if args.method == 5 or args.method == 6:
    parameters = [
                  {'classifier__kernel': ['linear'], 'classifier__C': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0]},
                  {'classifier__kernel': ['rbf'], 'classifier__C': [0.001, 0.01, 0.1, 1.0, 10.0, 100.0], 'classifier__gamma': [0.01, 0.1, 1.0, 10.0]}
                 ]
    classifier = SVC(probability=True)

if args.v:
    print("Base classifier: ", classifiers[args.method])

X_train, Y_train = load_from_arff(args.dataset+".training", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)
X_test, Y_test = load_from_arff(args.dataset+".testing", label_count=args.nlabels, label_location='end', load_sparse=args.sparse)

# Test in train and test columns do not match. That could happen for sparse matrices
if X_train.shape[1] > X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_test.shape[0], X_train.shape[1] - X_test.shape[1]))
    X_test = scipy.sparse.hstack([X_test, toadd])
elif X_train.shape[1] < X_test.shape[1]:
    toadd = scipy.sparse.lil_matrix((X_train.shape[0], X_test.shape[1] - X_train.shape[1]))
    X_train = scipy.sparse.hstack([X_train, toadd])

order = list(range(args.nlabels))
if args.r:
    random.shuffle(order)
    print(order)

classifier = ClassifierChain(classifier=classifiers[args.method], order=order)

if args.cv:
    # Parameters to optimize: classifier.get_params().keys()
    # Valid scores: sorted(sklearn.metrics.SCORERS.keys())
    #clf = GridSearchCV(classifier, parameters[args.method], scoring='f1_micro', n_jobs=-1)
    clf = GridSearchCV(classifier, parameters, scoring='f1_micro', n_jobs=-1, cv=3, verbose=3)
    #clf = RandomizedSearchCV(classifier, parameters, scoring='f1_micro', n_jobs=-1, cv=3)
    clf.fit(X_train.A, Y_train.A)
    f = open(args.dataset+'.parameters', 'w')
    f.write(str(clf.best_params_) + ", Order: " + str(order))
    f.close()
else:
    clf = classifier.fit(X_train, Y_train)
    #f = open(args.dataset+'.parameters', 'w')
    #f.write("Order: " + str(order))
    #f.close()

# Test error
preds = clf.predict(X_test.todense()).toarray()
np.savetxt(args.dataset + ".testing.preds", preds, fmt='%d')
if args.v:
    print("F1: %f" % metrics.f1_score(Y_test, preds, average='samples', zero_division=0))

probs = clf.predict_proba(X_test.todense()).toarray()
np.savetxt(args.dataset + ".testing.probs", probs, fmt='%g')

sys.exit(1)
