/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "ml.h"
#include "macros.h"
#include "maths.h"

// Test all classifiers
void TestAllExternalMultiLabelClassifiers(const t_Parameters *parm, scalar **test_data, int **test_labels, int *test_n_labels, scalar **train_data, int train_instances, int **train_labels,
        int *train_n_labels, int n_labels, int inputs, int **test_index, int **train_index, int test_instances, int *selection, char *out_file_name, int **ss, int *subspace)
{
    scalar **f;
    int *n_h, **h;
    bool classification = false, ranks = false;
    t_Performance performance;
    extern char *ml_ext_method[];
    extern int n_ml_ext_method;

    char message[MAX_LENGTH];

    // Classify testing set
    AllocateScalarMatrix(&f, parm->evaluate_train_error ? train_instances : test_instances, n_labels);
    AllocateIntegerMatrix(&h, parm->evaluate_train_error ? train_instances : test_instances, n_labels);
    MALLOC(n_h, int, parm->evaluate_train_error ? train_instances : test_instances, "TestAllMultiLabelClassifiers: n_h");

    for (t_MLExternalClassifier method = 0; method < n_ml_ext_method; method++) {
        time_t init_time = time(NULL);

        // Use this evaluator?
        if (!parm->evaluate_ml_with_external[method]) {
            SHOWPROGRESS(parm->show_progress > 1, "NOT evaluating %s\n", ml_ext_method[method]);
            continue;
        }

        SHOWPROGRESS(parm->show_progress > 0, "Training and evaluating %s\n", ml_ext_method[method]);

        ClassifyExternalDataset(parm, method, test_data, h, f, train_data, train_instances, train_labels,  train_n_labels, n_labels, n_h, inputs, test_index, train_index, test_instances, selection,
                                &classification, &ranks, out_file_name, test_labels, test_n_labels, parm->cv_external_hyperparameters_for_testing, ss, subspace);

        fprintf(parm->out_file, "%s evaluation time: %ld\n", ml_ext_method[method], time(NULL) - init_time);

        // Message for output
        sprintf(message, "%s", ml_ext_method[method]);

        // Average number of predicted labels per instances
        scalar av_pred = 0.0, av_actual = 0.0;
        for (int i = 0; i < test_instances; i++) {
            av_pred += n_h[i];
            av_actual += test_n_labels[i];
        }
        fprintf(parm->out_file, "@%s average number of PREDICTED relevant labels: %.2lf\n", message, av_pred/test_instances);
        fprintf(parm->out_file, "@%s average number of ACTUAL relevant labels: %.2lf\n", message, av_actual/test_instances);

        // Evaluate results
        if (classification)
            PrintClassificationMetrics(&performance, test_labels, test_n_labels, h, n_h, test_instances, n_labels, true, parm->out_file, message, true);

        if (ranks)
            PrintRankingMetrics(&performance, test_labels, test_n_labels, f, test_instances, n_labels, parm->out_file, message, true);

        // Save actual and predicted outputs?
        if (parm->save_outputs) {
            char filename[MAX_FILENAME_LENGTH];
            sprintf(filename, "%s.testing.%s.outputs", out_file_name, ml_ext_method[method]);

            SaveMLOutputs(filename, test_instances, n_labels, h, n_h, test_labels, test_n_labels);
        }

        // Save results so far to file
        fprintf(parm->out_file, "@ok\n");
        fflush(parm->out_file);
    }

    FreeIntegerMatrix(&h, test_instances);
    FreeScalarMatrix(&f, test_instances);
    FREE(n_h);
}

#define MAX_COMMAND_LENGTH  10000
void ClassifyExternalDataset(const t_Parameters *parm, t_MLExternalClassifier method, scalar **test_data, int **h, scalar **f, scalar **train_data, int train_instances, int **train_labels,
                             int *train_n_labels, int n_labels, int *n_h, int inputs, int **test_index, int **train_index, int test_instances, int *selection, bool *classification, bool *ranks,
                             char *out_file_name, int **test_labels, int *test_n_labels, bool cv, int **ss, int *subspace)
{
    char *f_train_name, *f_test_name, *f_name, *command, *local_file;

    ALLOCA(f_train_name, char, MAX_FILENAME_LENGTH, "ClassifierChain: f_train_name");
    ALLOCA(f_test_name, char, MAX_FILENAME_LENGTH, "ClassifierChain: f_test_name");
    ALLOCA(f_name, char, MAX_FILENAME_LENGTH, "ClassifierChain: f_name");
    ALLOCA(local_file, char, MAX_FILENAME_LENGTH, "ClassifierChain: local_file");
    ALLOCA(command, char, MAX_COMMAND_LENGTH, "ClassifierChain: command");

    extern char *ml_ext_method[];
    int selected = 0;

    // Random string to avoid hammering files
    char rname[11];
    for (int i = 0; i < 10; i++)
        rname[i] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"[random()%62];
    rname[10] = '\0';

    sprintf(local_file, "%s.%s", out_file_name, rname);

    // Create training and test files
    sprintf(f_train_name, "%s.%s.training", local_file, ml_ext_method[method]);
    SaveDatasetAsMLARFF(f_train_name, inputs, train_data, train_instances, train_index, train_labels, train_n_labels, n_labels, selection, ss, subspace);

    sprintf(f_test_name, "%s.%s.testing", local_file, ml_ext_method[method]);
    SaveDatasetAsMLARFF(f_test_name, inputs, test_data, test_instances, test_index, test_labels, test_n_labels, n_labels, NULL, NULL, subspace);

    // By default all classifiers classification and ranks
    *classification = *ranks= true;

    switch (method) {
        case ExMLkNN:
            // k cannot be larger than the number of actual instances
            if (!selection)
                selected = train_instances;
            else
                for (int i = 0; i < train_instances; i++)
                    if (selection[i])
                        selected++;
            sprintf(command, "mlknn.py %s.%s %d %d %s -s %f %s", local_file, ml_ext_method[method], MIN(parm->mlknn_k, selected), n_labels, train_index ? "-sparse" : " ",
                    parm->mlknn_s, cv ? "-cv" : "");
            break;

        case RAkELd:
            sprintf(command, "rakel.py %s.%s %d %s %d %s", local_file, ml_ext_method[method], n_labels, train_index ? "-sparse" : " ", (int) parm->ml_ext_base_method, cv ? "-cv" : "");
            break;

        case RAkELo:
            *ranks = false;
            sprintf(command, "rakel.py -o %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, train_index ? "-sparse" : " ", cv ? "-cv" : "");
            break;

        case MLARAM:
            sprintf(command, "mlaram.py %s.%s %d %s %s", local_file, ml_ext_method[method], n_labels,  train_index ? "-sparse" : " ", cv ? "-cv" : "");
            break;

        case ClassifierChain:
            sprintf(command, "cc.py %s.%s %d %d %s %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, train_index ? "-sparse" : " ", cv ? "-cv" : "",
                    parm->ml_cc_random_order ? "-r" : " ");
            break;

        case BinaryRelevance:
            sprintf(command, "br.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, train_index ? "-sparse" : " ", cv ? "-cv" : "");
            break;

        case LabelPowerset:
            sprintf(command, "lp.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, train_index ? "-sparse" : " ", cv ? "-cv" : "");
            break;

        case BRkNNa:
        case BRkNNb:
            *ranks = false;
            sprintf(command, "brknn.py %s.%s %c %d %d %s %s", local_file, ml_ext_method[method], method == BRkNNa ? 'a' : 'b', parm->mlknn_k, n_labels, train_index ? "-sparse" : " ",
                    cv ? "-cv" : "");
            break;

        case MLTSVM:
            *ranks = false;
            sprintf(command, "mltsvm.py %s.%s %d %s %s", local_file, ml_ext_method[method], n_labels, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case MLMLP:
            sprintf(command, "mlp.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, parm->mlmlp_hidden_nodes, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case DL:
            sprintf(command, "dl.py %s.%s %d %s %s", local_file, ml_ext_method[method], n_labels, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case BOOMER:
            sprintf(command, "boomer.py %s.%s %d %s %s", local_file, ml_ext_method[method], n_labels, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case BCC:
            *ranks = false;
            sprintf(command, "bcc.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case CDN:
            *ranks = false;
            sprintf(command, "cdn.py %s.%s %d %d %s %s -I %d -Ic %d", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ",
                    parm->cdn_I, parm->cdn_Ic);
            break;

        case HASEL:
            // OnErrorExit(ERR_UNSPECIFIED, "Undocumented method");
            *ranks = false;
            sprintf(command, "hasel.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case PSt:
        case PS:
            *ranks = false;
            sprintf(command, "pst.py %s %s.%s %d %d %s %s", method == PSt ? "-t" : "", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "",
                    train_index ? "-sparse" : " ");
            break;

        case CT:
            *ranks = false;
            sprintf(command, "ct.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case MCC:
            *ranks = false;
            sprintf(command, "mcc.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case NSR:
            *ranks = false;
            sprintf(command, "nsr.py %s.%s %d %d %s %s -P %d -N %d", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ",
                    parm->nsr_p, parm->nsr_n);
            break;

        case LC:
            *ranks = false;
            sprintf(command, "lc.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case SCC:
            *ranks = false;
            sprintf(command, "scc.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case CDT:
            *ranks = false;
            sprintf(command, "cdt.py %s.%s %d %d %s %s -I %d -Ic %d", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ",
                    parm->cdn_I, parm->cdn_Ic);
            break;

        case DBPNN:
            OnErrorExit(ERR_UNSPECTED_RESULT, "It does not work");
            *ranks = false;
            sprintf(command, "dbpnn.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case FW:
            *ranks = false;
            sprintf(command, "fw.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case Maniac:
            *ranks = false;
            sprintf(command, "maniac.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case PCC:
            *ranks = false;
            sprintf(command, "pcc.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;

        case PMCC:
            // OnErrorExit(ERR_UNSPECTED_RESULT, "It does not work");
            *ranks = false;
            sprintf(command, "pmcc.py %s.%s %d %d %s %s", local_file, ml_ext_method[method], n_labels, (int) parm->ml_ext_base_method, cv ? "-cv" : "", train_index ? "-sparse" : " ");
            break;
    }

    int status = system(command);
    if (!WEXITSTATUS(status)) {
        fprintf(parm->out_file, "@ok\n");

        // Remove files
        if (!parm->keep_files) {
            remove(f_train_name);
            remove(f_test_name);
            // Prediction
            sprintf(f_name, "%s.preds", f_test_name);
            remove(f_name);
            // Probabilities
            sprintf(f_name, "%s.probs", f_test_name);
            remove(f_name);
        }

        printf("Failed command: %s\n", command);
        OnErrorExit(ERR_UNSPECIFIED, "Python learning method failed");
    }

    // Obtain predictions and probabilities
    // Get h: Read predictions file
    FILE *file;
    sprintf(f_name, "%s.preds", f_test_name);
    if ((file = fopen(f_name, "r")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open predictions file");

    // Read prediction for every instance
    for (int i = 0; i < test_instances; i++) {
        n_h[i] = 0;
        for (int j = 0; j < n_labels; j++) {
            int pred;
            fscanf(file, "%d", &pred);
            if (pred) {
                h[i][n_h[i]] = j;
                n_h[i]++;
            }
        }
    }
    fclose(file);

    // Get f: Read probabilities file
    if (*ranks) {
        sprintf(f_name, "%s.probs", f_test_name);
        if ((file = fopen(f_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open predictions file");

        // Read probabilities for every instance
        for (int i = 0; i < test_instances; i++)
            for (int j = 0; j < n_labels; j++)
                fscanf(file, "%lf", &f[i][j]);

        fclose(file);
    }

    // Remove files
    if (!parm->keep_files) {
        remove(f_train_name);
        remove(f_test_name);
        // Probabilities
        sprintf(f_name, "%s.probs", f_test_name);
        remove(f_name);
        if (*ranks) {
            // Prediction
            sprintf(f_name, "%s.preds", f_test_name);
            remove(f_name);
        }
    }

    // Save parameters if apply
    sprintf(f_name, "%s.%s.parameters", local_file, ml_ext_method[method]);
    if (access(f_name, F_OK) == 0) {
        char line[1000];
        FILE *fi;
        if ((fi = fopen(f_name, "r")) == NULL)
            OnErrorExit(ERR_FILE, "Cannot open parameters file");

        fgets(line, 1000, fi);
        fclose(fi);
        SAVEPROGRESS(true, "External %s parameters: %s\n", ml_ext_method[method], line);
        remove(f_name);
    }

}

// Save file using ARFF format
void SaveDatasetAsMLARFF(char *out_file_name, int inputs, scalar **data, int instances, int **index, int **labels, int *n_labels, int classes, int *selection, int **ss, int *subspace)
{
    int *lsubspace;
    FILE *f;

    if (subspace)
        lsubspace = subspace;
    else {
        MALLOC(lsubspace, int, inputs, "SaveDatasetAsMLARFF: lsubspace");
        for (int i = 0; i < inputs; i++)
            lsubspace[i] = 1;
    }

    if (ss && selection)
        OnErrorExit(ERR_UNSPECTED_RESULT, "Selection and per-label selection are incompatible");

    if ((f = fopen(out_file_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open dataset file");

    // Relation name
    fprintf(f, "@relation TRAINING\n\n");

    // Attributes
    for (int i = 0; i < inputs; i++)
        fprintf(f, "@attribute INPUT%d numeric\n", i+1);

    // Labels
    for (int i = 0; i < classes; i++)
        fprintf(f, "@attribute LABEL%d {0,1}\n", i+1);

    // Data
    fprintf(f, "\n@data\n");

    // Sparse or dense format?
    if (index) {
        // Variable index begins with 0
        for (int i = 0; i < instances; i++) {
            // Check selection or instance without non zero values
            int count = index[i][0] == -1 ? 0 : 1;
            if (selection && count)
                count = selection[i];
            else if (ss && count)
                count = EmptyColumn(ss, classes, i) ? 0 : 1;
            for (int l = 0; l < count; l++) {
                fprintf(f,"{");
                // Attributes
                int j = 0;
                while (index[i][j] != -1) {
                    if (lsubspace[index[i][j]-1])
                        fprintf(f, "%d %g%s", index[i][j]-1, data[i][j], index[i][j+1] == -1 ? " " : ", ");
                    j++;
                }

                // Labels might not be included for testing
                if (labels)
                    for (int j = 0; j < n_labels[i]; j++)
                        if (!ss || ss[labels[i][j]][i])
                            fprintf(f, ", %d 1", labels[i][j]+inputs);

                fprintf(f, "}\n");
            }
        }
    } else {
        for (int i = 0; i < instances; i++) {
            // Number of repetitions or presence
            int count = 1;
            if (selection)
                count = selection[i];
            else if (ss)
                count = EmptyColumn(ss, classes, i) ? 0 : 1;
            for (int l = 0; l < count; l++) {
                //  Attributes
                for (int j = 0; j < inputs; j++)
                    fprintf(f, "%g%s", data[i][j]*lsubspace[j], j+1 == inputs ? " " : ", ");

                for (int j = 0; j < classes; j++)
                    fprintf(f, ", %d", (int)(LabelInList(j, labels[i], n_labels[i]) && (!ss || ss[j][i])));

                fprintf(f, "\n");
            }
        }
    }

    fclose(f);

    if (!subspace)
        FREE(lsubspace);
}
