/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <math.h>
#include "macros.h"
#include "maths.h"
#include "typex.h"
#include "ml.h"
#include "ensemble.h"

// Adaboost.MH [Schapire and Singer, 2000]

#define SCALE(x)   (2.0*(x)-1.0)   // Scale [0, 1] to [-1, 1]

void TrainAdaBoostMH(const t_Parameters *parm, t_MLEnsemble *ensemble, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, int *subset, int **index, int **SS,
                     int *subspace, bool real, bool cv)
{
    if (SS || subset)
        OnErrorExit(ERR_PARAM, "Selection not valid for AdaBoost.MH/AdaBoost.MR");

    if (parm->base_learner == NN || parm->base_learner == kNN || parm->base_learner == OVO)
        OnErrorExit(ERR_PARAM, "Invalid base learner for AdaBoost.MH/AdaBoost.MR");

    if (IsEnsemble(parm->base_learner))
        OnErrorExit(ERR_PARAM, "Base learner for AdaBoost.MH/AdaBoost.MR cannot be an ensemble");

    scalar **D;
    int **b_labels;

    // Create binary labels
    AllocateIntegerMatrix(&b_labels, classes, instances);
    ZeroIntegerMatrix(b_labels, classes, instances);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            b_labels[labels[i][j]][i] = 1;

    AllocateScalarMatrix(&D, instances, classes);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < classes; j++)
            D[i][j] = 1.0/(instances*classes);

    MALLOC(ensemble->adaboost, t_MLAdaBoost, 1, "TrainAdaBoostMH: ensemble->adaboost");
    ensemble->adaboost->T = parm->ml_ensemble_size;
    ensemble->adaboost->base = parm->base_learner;
    MALLOC(ensemble->adaboost->alpha, scalar, parm->ml_ensemble_size, "TrainAdaBoostMH: ensemble->adaboost->alpha");
    MALLOC(ensemble->adaboost->bcs, t_Learner *, parm->ml_ensemble_size, "TrainAdaBoostMH: ensemble->adaboost->bcs");
    for (int t = 0; t < parm->ml_ensemble_size; t++)
        MALLOC(ensemble->adaboost->bcs[t], t_Learner, classes, "TrainAdaBoostMH: ensemble->adaboost->bcs[t]");

    ensemble->th = parm->ml_ensemble_t;

    // Learn models
    for (int t = 0; t < ensemble->adaboost->T; t++) {
        // Train BR models
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int l = 0; l < classes; l++) {
            int **lindex = NULL, *llabels;
            scalar **ldata;
            MALLOC(ldata, scalar *, instances, "TrainAdaBoostMH: ldata");
            MALLOC(llabels, int, instances, "TrainAdaBoostMH: llabels");
            if (index)
                MALLOC(lindex, int *, instances, "TrainAdaBoostMH: lindex");

            if (t == 0) {
                // Initial sample is all data
                memcpy(ldata, data, sizeof(scalar *)*instances);
                memcpy(llabels, b_labels[l], sizeof(int)*instances);
                if (index)
                    memcpy(lindex, index, sizeof(int *)*instances);
            } else { // Resample
                scalar *cdf;
                MALLOC(cdf, scalar, instances, "TrainAdaBoostMH: cdf");
                cdf[0] = D[0][l];
                for (int j = 1; j < instances; j++)
                    cdf[j] = D[j][l]+cdf[j-1];
                for (int j = 0; j < instances; j++) {
                    int next = RouletteSelection(cdf, instances);
                    ldata[j] = data[next];
                    llabels[j] = b_labels[l][next];
                    if (index)
                        lindex[j] = index[next];
                }
                FREE(cdf);
            }

            // Cross-validate hyper-parameters?
            t_Parameters lparm = *parm;
            if (cv)
                SetCrossValidatedParameters(&lparm, ldata, instances, llabels, subspace, inputs, 2, ensemble->adaboost->base, 1, NULL, lindex);

            // Train
            SHOWPROGRESS(parm->show_progress > 0, "BR training model %*d/%d for label %*d\r", WIDTH(classes), l+1, classes, WIDTH(classes), l+1);
            ClassifierTraining(&lparm, ensemble->adaboost->base, ldata, instances, llabels, inputs, 2, lindex, subspace, NULL, true, &(ensemble->adaboost->bcs[t][l]), NULL);
            SHOWPROGRESS(parm->show_progress > 0, "BR finished training model %*d/%d for label %*d\r", WIDTH(classes), l+1, classes, WIDTH(classes), l+1);

            FREE(ldata);
            FREE(llabels);
            if (index)
                FREE(lindex);
        }

        // Obtain alpha
        ensemble->adaboost->alpha[t] = 1.0;

        // Update weights
        AdaBoostMHUpdateWeights(parm, instances, ensemble->adaboost->bcs[t], data, labels, n_labels, classes, inputs, index, D, subspace, subset, SS, real, ensemble->adaboost->base,
                                ensemble->adaboost->alpha[t]);
    }

    if (parm->ml_ensemble_threshold_by_resubs && !parm->cv_ml_hyperparameters) {
        // Cross-validate the threshold if apply
        scalar fitness = -1.0, thresholds[] = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9}, best = 0.0;;
        int n_thresholds = 9, *mask;
        MALLOC(mask, int, instances, "TrainMultiLabelBagging: mask");
        if (subset)
            CopyIntegerVector(subset, mask, instances);
        else
            FillVector(NULL, mask, instances, 1.0);

        t_MLEnsemble learner;
        learner = *ensemble;
        for (int t = 0; t < n_thresholds; t++) {
            learner.th = thresholds[t];

            // Evaluate fold
            scalar l_fitness = EvaluateFoldPerformance(parm, instances, labels, n_labels, classes, inputs, data, subset, NULL, index, mask, NULL, real ? REALADABOOSTMH : DISCRETEADABOOSTMH, subspace,
                                                       &learner);

            if (l_fitness > fitness) {
                fitness = l_fitness;
                best = thresholds[t];
            }

            SHOWPROGRESS(parm->show_progress > 1, "Evaluated %3.1lf threshold with performance: %6.4lf\n", thresholds[t], l_fitness);
        }

        ensemble->th = best;

        SAVEPROGRESS(true, "@ML ensemble best cross-validated threshold: %3.1lf\n", best);
        FREE(mask);
    }

    FreeIntegerMatrix(&b_labels, classes);
    FreeScalarMatrix(&D, instances);
}

void AdaBoostMHUpdateWeights(const t_Parameters *parm, int instances, t_Learner *learner, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, scalar **D,
                             int *subspace, int *subset, int **SS, bool real, t_Classifier base, scalar alpha)
{
    scalar f[classes];

    if (SS || subset)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Instance selection for AdaBoost.MH");

    int category[classes];

    // Update weights
    for (int i = 0; i < instances; i++) {
        for (int l = 0; l < classes; l++)
            category[l] = Classifier2ClassOutput(parm, base, data[i], inputs, 2, index ? index[i] : NULL, subspace, data, instances, NULL, NULL, &(learner[l]), 1, &f[l], index, NULL);

        // Real or discrete version?
        if (real) {
            for (int l = 0; l < classes; l++)
                if (LabelInList(l, labels[i], n_labels[i]))
                    D[i][l] *= exp(-alpha*f[l]);
                else
                    D[i][l] *= exp(alpha*f[l]);
        } else {
            for (int l = 0; l < classes; l++)
                D[i][l] *= exp(-alpha*SCALE(category[l])*f[l]);
        }
    }

    // Normalize matrix
    NormalizeMatrix(D, instances, classes);
}

void FreeAdaBoostMHEnsemble(const t_Parameters *parm, t_MLEnsemble *ensemble, int n_labels, int inputs)
{
    for (int t = 0; t < ensemble->adaboost->T; t++) {
        for (int l = 0; l < n_labels; l++)
            ClassifierFree(parm, &(ensemble->adaboost->bcs[t][l]), ensemble->adaboost->base, 2, inputs, NULL);
        FREE(ensemble->adaboost->bcs[t]);
    }

    FREE(ensemble->adaboost->bcs);
    FREE(ensemble->adaboost->alpha);
}

void AdaBoostMHClassifyInstance(const t_Parameters *parm, scalar *query, int *query_index, t_MLEnsemble *ensemble, scalar **data, int classes, int inputs, int **index, int *subspace, int *subset,
                                int **SS, int *h, scalar *f, int *n_predicted, int instances, bool *classification, bool *ranks)
{
    if (SS || subset)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Instance selection for AdaBoost.MH");

    scalar votes[classes];
    ZeroVector(votes, NULL, classes);

    for (int t = 0; t < ensemble->adaboost->T; t++)
        for (int l = 0; l < classes; l++) {
            int category = Classifier2ClassOutput(parm, ensemble->adaboost->base, query, inputs, 2, query_index, subspace, data, instances, NULL, NULL, &(ensemble->adaboost->bcs[t][l]), 1, &f[l],
                                                  index, NULL);

            if (category)
                votes[l] += ensemble->adaboost->alpha[t];
        }

    // Majority voting
    *n_predicted = 0;
    for (int l = 0; l < classes; l++)
        if (votes[l] >= ensemble->adaboost->T * ensemble->th)
            h[(*n_predicted)++] = l;

    // Normalize probabilities
    memcpy(f, votes, sizeof(scalar)*classes);
    NormalizeVector(f, classes);

    *classification = *ranks = true;
}
