/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <math.h>
#include "macros.h"
#include "maths.h"
#include "typex.h"
#include "ml.h"
#include "ensemble.h"

#define SCALE(x)   (2.0*(x)-1.0)   // Scale [0, 1] to [-1, 1]

void TrainAdaBoostByResampling(const t_Parameters *parm, t_MLEnsemble *ensemble, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, int *subset, int **index,
                               int **SS, int *subspace, bool real)
{
    if (SS)
        OnErrorExit(ERR_PARAM, "Partial selection not valid for AdaBoost.RE");

    if (subset)
        for (int i = 0; i < instances; i++)
            if (!subset[i])
                OnErrorExit(ERR_PARAM, "Selection not valid for AdaBoost.RE");

    scalar **ldata, **D;
    int **llabels, *ln_labels, **lindex = NULL, *list;

    AllocateScalarMatrix(&D, instances, classes);
    MALLOC(list, int, instances, "TrainAdaBoostByResampling: list");
    MALLOC(ldata, scalar *, instances, "TrainAdaBoostByResampling: ldata");
    MALLOC(llabels, int *, instances, "TrainAdaBoostByResampling: llabels");
    MALLOC(ln_labels, int, instances, "TrainAdaBoostByResampling: ln_labels");
    if (index)
        MALLOC(lindex, int *, instances, "TrainAdaBoostByResampling: lindex");

    // Initial sample is all data
    memcpy(ldata, data, sizeof(scalar *)*instances);
    memcpy(llabels, labels, sizeof(int *)*instances);
    memcpy(ln_labels, n_labels, sizeof(int)*instances);
    if (index)
        memcpy(lindex, index, sizeof(int *)*instances);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < classes; j++)
            D[i][j] = 1.0/(instances*classes);

    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "PARIS for bagging ensembles");

    MALLOC(ensemble->boost, t_MLBoosting, 1, "TrainAdaBoostByResampling: ensemble->boost");
    ensemble->boost->T = parm->ml_ensemble_size;
    ensemble->boost->base = parm->ml_base_learner;
    MALLOC(ensemble->boost->alpha, scalar, parm->ml_ensemble_size, "TrainAdaBoostByResampling: ensemble->boost->alpha");
    MALLOC(ensemble->boost->bcs, t_MLLearner, parm->ml_ensemble_size, "TrainAdaBoostByResampling: ensemble->boost->bcs");

    // Learn models
    for (int t = 0; t < ensemble->boost->T; t++) {
        TrainMultiLabelClassifier(parm, ensemble->boost->base, ldata, instances, llabels, ln_labels, classes, inputs, NULL, NULL, lindex, NULL, SS, &(ensemble->boost->bcs[t]), parm->mlknn_k, NULL,
                                  NULL, -1, subspace, NULL);

        // Obtain alpha
        ensemble->boost->alpha[t] = 1.0;

        // Update weights
        AdaBoostREUpdateWeights(parm, instances, &(ensemble->boost->bcs[t]), data, labels, n_labels, classes, inputs, index, D, subspace, subset, SS, real, ensemble->boost->base,
                                ensemble->boost->alpha[t]);

        // New sample
        ResampleForML(D, list, instances, classes, parm->ml_resampling_method);

        for (int i = 0; i < instances; i++) {
            ldata[i] = data[list[i]];
            llabels[i] = labels[list[i]];
            ln_labels[i] = n_labels[list[i]];
            if (index)
                lindex[i] = index[list[i]];
        }
    }

    FREE(ldata);
    FREE(llabels);
    FREE(ln_labels);
    if (index)
        FREE(lindex);
    FreeScalarMatrix(&D, instances);
    FREE(list);
}

// Normalize matrix to sum 1
void NormalizeMatrix(scalar **m, int r, int c)
{
    scalar sum = 0.0;
    for (int i = 0; i < r; i++)
        for (int j = 0; j < c; j++)
            sum += m[i][j];

    for (int i = 0; i < r; i++)
        for (int j = 0; j < c; j++)
            m[i][j] /= sum;
}

void ResampleForML(scalar **D, int *list, int instances, int classes, t_MLResampling method)
{
    scalar *cdf;
    int selected;

    // Resampling method
    switch (method) {
        case ResamplingPerInstance:
            MALLOC(cdf, scalar, instances, "ResampleForML: cdf");

            // Resample each instance using all the labels weight
            selected = 0;
            cdf[0] = 0.0;
            for (int j = 0; j < classes; j++)
                cdf[0] += D[0][j];
            for (int i = 1; i < instances; i++) {
                cdf[i] = 0.0;
                for (int j = 0; j < classes; j++)
                    cdf[i] += D[i][j];
                cdf[i] += cdf[i-1];
            }

            selected = 0;
            for (int j = 0; j < instances; j++)
                list[selected++] = RouletteSelection(cdf, instances);

            FREE(cdf);
            break;

        case ResamplingPerLabel:
            MALLOC(cdf, scalar, instances, "ResampleForML: cdf");

            // Resample from each label a proportional set of instances
            selected = 0;
            for (int i = 0; i < classes; i++) {
                cdf[0] = D[0][i];
                for (int j = 1; j < instances; j++)
                    cdf[j] = cdf[j-1]+D[j][i];

                for (int j = 0; j < instances/classes; j++, selected++)
                    list[selected] = RouletteSelection(cdf, instances);
            }

            // Assign remaining randomly
            for (; selected < instances; selected++)
                list[selected] = RANDOM(0, instances);

            FREE(cdf);
            break;

        case ResamplingPartial:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial resampling for AdaBoost.RE");
            break;
    }
}

void AdaBoostREUpdateWeights(const t_Parameters *parm, int instances, t_MLLearner *learner, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, scalar **D,
                             int *subspace, int *subset, int **SS, bool real, t_MLClassifier base, scalar alpha)
{
    bool classification, ranks;
    scalar f[classes];

    // Real or discrete outputs?
    if (real) {
        int h[classes];

        // AdaBoost.MH method [Schapire and Singer, 2000]
        // Update weights
        for (int i = 0; i < instances; i++) {
            int n_predicted;
            ClassifyInstance(parm, base, data[i], h, f, &classification, &ranks, data, instances, labels, n_labels, classes, &n_predicted, inputs, index ? index[i] : NULL, index, subset, NULL, -1, SS,
                             learner, subspace, labels, n_labels, NULL);

            for (int j = 0; j < classes; j++)
                if (LabelInList(j, labels[i], n_labels[i]))
                    D[i][j] *= exp(-alpha*SCALE(f[j]));
                else
                    D[i][j] *= exp(alpha*SCALE(f[j]));
        }
    } else {
        int **h, *n_h;
        scalar *w;

        // Need to obtain e_i for every label
        AllocateIntegerMatrix(&h, instances, classes);
        MALLOC(n_h, int, instances, "AdaBoostREUpdateWeights: n_h");
        MALLOC(w, scalar, instances, "AdaBoostREUpdateWeights: w");

        // MultiBoost method [Webb, 2000]
        for (int i = 0; i < instances; i++)
            ClassifyInstance(parm, base, data[i], h[i], f, &classification, &ranks, data, instances, labels, n_labels, classes, &n_h[i], inputs, index ? index[i] : NULL, index, subset, NULL, -1, SS,
                             learner, subspace, labels, n_labels, NULL);

        for (int j = 0; j < classes; j++) {
            // Get normalized weights
            for (int i = 0; i < instances; i++)
                w[i] = D[i][j];
            NormalizeVector(w, instances);

            scalar e_i = 0;
            for (int i = 0; i < instances; i++)
                if (LabelInList(j, labels[i], n_labels[i]) != LabelInList(j, h[i], n_h[i]))
                    e_i += w[i];

            // Error above 0.5 or near 0 reinitialize
            if (e_i <= ZERO || e_i >= 0.5)
                for (int i = 0; i < instances; i++)
                    D[i][j] = Poisson();
            else
                for (int i = 0; i < instances; i++)
                    if (LabelInList(j, labels[i], n_labels[i]) != LabelInList(j, h[i], n_h[i]))
                        D[i][j] /= 2.0*(1.0-e_i);
                    else
                        D[i][j] /= 2.0*e_i;
        }

        FreeIntegerMatrix(&h, instances);
        FREE(n_h);
        FREE(w);
    }

    // Normalize matrix
    NormalizeMatrix(D, instances, classes);
}
