/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include "ml.h"
#include "macros.h"
#include "maths.h"
#include "ensemble.h"

void TrainMultiLabelBagging(const t_Parameters *parm, t_MLEnsemble *ensemble, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, int *subset, int **index, int **SS,
                            int *subspace, int size, t_MLClassifier ml_base_learner, bool replacement)
{
    scalar **ldata;
    int **llabels, *ln_labels, **lindex = NULL, *lsubset = NULL, *list = NULL;

    MALLOC(ldata, scalar *, instances, "TrainMultiLabelBagging: ldata");
    MALLOC(llabels, int *, instances, "TrainMultiLabelBagging: llabels");
    MALLOC(ln_labels, int, instances, "TrainMultiLabelBagging: ln_labels");
    if (subset) {
        MALLOC(lsubset, int, instances, "TrainMultiLabelBagging: lsubset");
        memcpy(lsubset, subset, sizeof(int)*instances);
    }
    if (index) {
        MALLOC(lindex, int *, instances, "TrainMultiLabelBagging: lindex");
        memcpy(lindex, index, sizeof(int *)*instances);
    }

    if (!replacement)
        MALLOC(list, int, instances, "TrainMultiLabelBagging: list");

    // Initial sample is all data
    memcpy(ldata, data, sizeof(scalar *)*instances);
    memcpy(llabels, labels, sizeof(int *)*instances);
    memcpy(ln_labels, n_labels, sizeof(int)*instances);

    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "PARIS for bagging ensembles");

    MALLOC(ensemble->boost, t_MLBoosting, 1, "TrainMultiLabelBagging: ensemble->boost");
    ensemble->boost->T = size;
    ensemble->boost->base = ml_base_learner;
    MALLOC(ensemble->boost->alpha, scalar, size, "TrainMultiLabelBagging: ensemble->boost->alpha");
    FillVector(ensemble->boost->alpha, NULL, size, 1.0);
    MALLOC(ensemble->boost->bcs, t_MLLearner, size, "TrainMultiLabelBagging: ensemble->boost->bcs");

    ensemble->th = ml_base_learner == EPS ? parm->eps_t : parm->ml_ensemble_t;

    // Learn models
    int t_instances = instances;
    for (int t = 0; t < ensemble->boost->T; t++) {
        TrainMultiLabelClassifier(parm, ensemble->boost->base, ldata, t_instances, llabels, ln_labels, classes, inputs, lsubset, NULL, lindex, NULL, SS, &(ensemble->boost->bcs[t]), parm->mlknn_k,
                                  NULL, NULL, -1, subspace, NULL);

        // New random sample with or without replacement
        if (replacement) {
            for (int i = 0; i < instances; i++) {
                int next = RANDOM(0, instances);

                ldata[i] = data[next];
                llabels[i] = labels[next];
                ln_labels[i] = n_labels[next];
                if (subset)
                    lsubset[i] = subset[next];
                if (index)
                    lindex[i] = index[next];
            }
        } else {
            FillVector(NULL, list, instances, 0.0);
            for (int i = 0; i < instances; i++)
                list[(int) RANDOM(0, instances)] = 1;

            t_instances = 0;
            for (int i = 0; i < instances; i++)
                if (list[i]) {
                    ldata[t_instances] = data[i];
                    llabels[t_instances] = labels[i];
                    ln_labels[t_instances] = n_labels[i];
                    if (subset)
                        lsubset[t_instances] = subset[i];
                    if (index)
                        lindex[t_instances] = index[i];

                    t_instances++;
                }
        }
    }

    if (parm->ml_ensemble_threshold_by_resubs && !parm->cv_ml_hyperparameters) {
        // Cross-validate the threshold if apply
        scalar fitness = -1.0, thresholds[] = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9}, best = 0.0;;
        int n_thresholds = 9, *mask;
        MALLOC(mask, int, instances, "TrainMultiLabelBagging: mask");
        if (subset)
            CopyIntegerVector(subset, mask, instances);
        else
            FillVector(NULL, mask, instances, 1.0);

        t_MLEnsemble learner;
        learner = *ensemble;
        for (int t = 0; t < n_thresholds; t++) {
            learner.th = thresholds[t];

            // Evaluate fold
            scalar l_fitness = EvaluateFoldPerformance(parm, instances, labels, n_labels, classes, inputs, data, subset, NULL, index, mask, NULL, BAGGING, subspace, &learner);

            if (l_fitness > fitness) {
                fitness = l_fitness;
                best = thresholds[t];
            }

            SHOWPROGRESS(parm->show_progress > 1, "Evaluated %3.1lf threshold with performance: %6.4lf\n", thresholds[t], l_fitness);
        }

        ensemble->th = best;

        SAVEPROGRESS(true, "@ML ensemble best cross-validated threshold: %3.1lf\n", best);
        FREE(mask);
    }

    FREE(ldata);
    FREE(llabels);
    FREE(ln_labels);
    if (subset)
        FREE(lsubset);
    if (index)
        FREE(lindex);
    if (!replacement)
        FREE(list);
}

void BaggingClassifyInstance(const t_Parameters *parm, t_MLEnsemble *ensemble, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances, int **relevant_labels,
                             int *relevant_classes, int classes, int *n_predicted, int inputs, int *query_index, int **index, int *S, int **SS,  int *subspace, int **labels, int *n_labels)
{
    scalar votes[classes];

    ZeroVector(votes, NULL, classes);
    for (int t = 0; t < ensemble->boost->T; t++) {
        ClassifyInstance(parm, ensemble->boost->base, query, h, f, classification, ranks, data, instances, relevant_labels, relevant_classes, classes, n_predicted, inputs, query_index, index, S, NULL,
                         -1, SS, &(ensemble->boost->bcs[t]), subspace, labels, n_labels, NULL);

        for (int i = 0; i < *n_predicted; i++)
            votes[h[i]] += ensemble->boost->alpha[t];
    }

    memcpy(f, votes, sizeof(scalar)*classes);
    NormalizeVector(f, classes);

    *n_predicted = 0;
    for (int i = 0; i < classes; i++)
        if (votes[i]/ensemble->boost->T >= ensemble->th)
            h[(*n_predicted)++] = i;

    *ranks = *classification = true;
}

void FreeBaggingEnsemble(const t_Parameters *parm, t_MLEnsemble *ensemble, int n_labels, int inputs, int instances)
{
    for (int t = 0; t < ensemble->boost->T; t++)
        FreeMLClassifier(parm, &(ensemble->boost->bcs[t]), ensemble->boost->base, n_labels, inputs, instances, NULL);

    FREE(ensemble->boost->bcs);
    FREE(ensemble->boost->alpha);
}
