/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include <stdbool.h>
#include <stdio.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"
#include "maths.h"

// Balanced k-means algorithm for HOMER multi-label model

void BalancedKMeans(int k, int *Ln, int n, int **labels, int *n_labels, int it, int instances, int **clusters, int *cluster_size, bool balanced)
{
    scalar **centroids, **distances, **sorted;

    // Construct binary version of labels to cluster
    int **w;
    AllocateIntegerMatrix(&w, instances, n);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n; j++)
            w[i][j] = LabelInList(Ln[j], labels[i], n_labels[i]);

    // Initialize clusters and centers
    AllocateScalarMatrix(&centroids, k, instances);
    AllocateScalarMatrix(&sorted, k, n);
    for (int i = 0; i < k; i++)
        cluster_size[i] = 0;

    // First three labels as random initial centroids
    for (int i = 0; i < k; i++)
        for (int j = 0; j < instances; j++)
            centroids[i][j] = w[j][i];

    // Distances to centers of clusters
    AllocateScalarMatrix(&distances, n, k);

    for (int i = 0; i < it; i++) {
        ZeroVector(NULL, cluster_size, k);

        // Distances of all labels to all clusters
        for (int m = 0; m < n; m++)
            for (int l = 0; l < k; l++)
                distances[m][l] = DistanceFromLabelToCluster(m, centroids[l], w, instances); // Distance(lambda, c_i, W_i);

        for (int m = 0; m < n; m++) {
            bool finished = false;

            int nu = m;

            while (!finished) {
                int j = IndexOfMinimumOfVector(distances[nu], NULL, k);

                // Insert sorted (v, d_v) to sorted list C_j
                int l = 0;
                while (distances[nu][j] > sorted[j][l] && l < cluster_size[j])
                    l++;

                for (int o = cluster_size[j]; o > l; o--) {
                    clusters[j][o] = clusters[j][o-1];
                    sorted[j][o] = sorted[j][o-1];
                }

                sorted[j][l] = distances[nu][j];
                clusters[j][l] = nu;
                cluster_size[j]++;

                if (balanced) {
                    // Remove last one if cluster too large
                    if (cluster_size[j] > (int) ceil(((scalar) n)/k)) {
                        // Remove last from C_j
                        cluster_size[j]--;
                        nu = clusters[j][cluster_size[j]];

                        distances[nu][j] = INFINITE;    // To avoid infinite loop
                    } else
                        finished = true;
                } else
                    finished = true;
            }
        }

        // Recalculate centers
        for (int m = 0; m < k; m++)
            if (cluster_size[m])
                for (int j = 0; j < instances; j++) {
                    centroids[m][j] = w[j][clusters[m][0]];
                    for (int l = 1; l < cluster_size[m]; l++)
                        centroids[m][j] += w[j][clusters[m][l]];
                    centroids[m][j] /= cluster_size[m];
                }
    }

    FreeIntegerMatrix(&w, instances);
    FreeScalarMatrix(&centroids, k);
    FreeScalarMatrix(&sorted, k);
    FreeScalarMatrix(&distances, n);

    // Convert to actual label
    for (int i = 0; i < k; i++)
        for (int j = 0; j < cluster_size[i]; j++)
            clusters[i][j] = Ln[clusters[i][j]];

#ifdef VERBOSE_MODE
    for (int i = 0; i < k; i++) {
        printf("Cluster %d:", i+1);
        for (int j = 0; j < cluster_size[i]; j++)
            printf(" %d", clusters[i][j]+1);
        printf("\n");
    }
#endif
}

// Distance is measure from label column to centroid
scalar DistanceFromLabelToCluster(int lambda, scalar *centroid, int **w, int n)
{
    scalar dist = 0.0;

    for (int i = 0; i < n; i++)
        dist += (w[i][lambda]-centroid[i])*(w[i][lambda]-centroid[i]);

    return sqrt(dist);
}
