/***************************************************************************
 *   Copyright (C) 20089 by Nicolas Garcia-Pedrajas                        *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the wpe that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; witwut even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You swuld have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include "macros.h"
#include "typex.h"
#include "classifiers.h"
#include "data.h"
#include "ml.h"
#include "maths.h"

// Multi-label BP learning algorithm
// M.-L. Zhang and Z.-H. Zwu (2006) "Multilabel neural networks with applications to functional genomics and text categorization," IEEE Trans. Knowledge and Data Engineering 18(10):1338-1351.

void TrainBPMLL(const t_Parameters *parm, t_BPMLL *bpmll, scalar **data, int instances, int Q, int d, int **labels, int *n_labels, int *subspace, int *subset)
{
    scalar **desired, **t_data;

    if (subspace)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for BP-MLL");

    // Training data
    AllocateScalarMatrix(&desired, instances, Q);
    MALLOC(t_data, scalar *, instances, "TrainBPMLL: t_data");
    int next = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            for (int j = 0; j < Q; j++)
                desired[next][j] = -1.0;
            for (int j = 0; j < n_labels[i]; j++)
                desired[next][labels[i][j]] = 1.0;
            t_data[next] = data[i];
            next++;
        }

    // Create classifier
    bpmll->M = parm->bpmll_nodes < 2.0 ? (int)(d*parm->bpmll_nodes) : (int) parm->bpmll_nodes;    // Less than 2 means percentage of inputs
    AllocateScalarMatrix(&(bpmll->v), d+1, bpmll->M);
    AllocateScalarMatrix(&(bpmll->w), bpmll->M+1, Q);

    // Random initial weights
    for (int i = 0; i < d+1; i++)
        for (int j = 0; j < bpmll->M; j++)
            bpmll->v[i][j] = RANDOM(-parm->init_weight, parm->init_weight);

    for (int i = 0; i < bpmll->M+1; i++)
        for (int j = 0; j < Q; j++)
            bpmll->w[i][j] = RANDOM(-parm->init_weight, parm->init_weight);

    // BP specific to ML version
    BackPropForBPMLL(parm, bpmll, parm->n_epochs, t_data, next, desired, d, Q, n_labels);
    SHOWPROGRESS(parm->show_progress > 0, "Finished training network%c", '\n');

    // Get weights for t threshold
    MALLOC(bpmll->w_t, scalar, Q+1, "TrainBPMLL: bpmll->w_t");
    GetThresholdWeights(bpmll, data, instances, labels, n_labels, Q, d, subspace);
    SHOWPROGRESS(parm->show_progress > 0, "Finished getting threshold weights%c", '\n');

    FREE(t_data);
    FreeScalarMatrix(&desired, instances);
}

void BPMLLClassifyInstance(t_BPMLL *bpmll, scalar *instance, scalar *f, int *h, int *n_relevant, int d, int Q, int *subspace, bool *classification, bool *ranks)
{
    scalar b[bpmll->M];

    GetBPMLLPOutput(bpmll, Q, subspace, instance, d, f, b);

    // Get threshold
    scalar t = bpmll->w_t[Q];   // Bias
    for (int i = 0; i < Q; i++)
        t += f[i]*bpmll->w_t[i];

    t = AssignWithinBoundsScalar(-0.9999, t, 0.9999);

    // Get relevant labels
    *n_relevant = 0;
    for (int i = 0; i < Q; i++)
        if (f[i] >= t)
            h[(*n_relevant)++] = i;

#ifdef VERBOSE_MODE
    PrintVector(f, NULL, Q, stdout);
#endif

    *classification = *ranks = true;
}

void BackPropForBPMLL(const t_Parameters *parm, t_BPMLL *bpmll, int n_epochs, scalar **data, int instances, scalar **desired, int inputs, int Q, int *n_labels)
{
    scalar *b, *c, *e, *d, **delta_v, **delta_w;

    // Outputs of layers
    MALLOC(b, scalar, bpmll->M, "BPMLL: b");
    MALLOC(c, scalar, Q, "BPMLL: c");
    MALLOC(e, scalar, bpmll->M, "BPMLL: e");
    MALLOC(d, scalar, Q, "BPMLL: d");
    AllocateScalarMatrix(&delta_v, inputs+1, bpmll->M);
    AllocateScalarMatrix(&delta_w, bpmll->M+1, Q);

    for (int i = 0; i < n_epochs; i++) {
        ZeroScalarMatrix(delta_v, inputs+1, bpmll->M);
        ZeroScalarMatrix(delta_w, bpmll->M+1, Q);

        SHOWPROGRESS(parm->show_progress > 1, "Training epoch %*d/%d\r", WIDTH(n_epochs), i+1, n_epochs);

        for (int k = 0; k < parm->epoch_size; k++) {
            int next = RANDOM(0, instances);    // Random pattern to train

            // Ignore if no relevant or all relevant
            if (!n_labels[next] || n_labels[next] == Q) {
                SHOWPROGRESS(parm->show_progress > 1, "Instance has %d relevant labels out of %d\n", n_labels[next], Q);
                continue;
            }

            GetBPMLLPOutput(bpmll, Q, NULL, data[next], inputs, c, b);

            // Get error for output layer
            for (int j = 0; j < Q; j++) {
                d[j] = 0.0;

                // Error depends of whether j is a relevant/irrelevant label
                if (desired[next][j] > 0.0) {
                    for (int l = 0; l < Q; l++)
                        if (desired[next][l] < 0.0)
                            d[j] += exp(-(c[j]-c[l]));
                    d[j] *= (1.0/(n_labels[next]*(Q-n_labels[next])))*(1.0+c[j])*(1.0-c[j]);
                } else {
                    for (int k = 0; k < Q; k++)
                        if (desired[next][k] > 0.0)
                            d[j] += exp(-(c[k]-c[j]));
                    d[j] *= -(1.0/(n_labels[next]*(Q-n_labels[next])))*(1.0+c[j])*(1.0-c[j]);
                }
            }

            // Get error for hidden layer
            for (int s = 0; s < bpmll->M; s++) {
                e[s] = 0.0;
                for (int j = 0; j < Q; j++)
                    e[s] += d[j] * bpmll->w[s][j];
                e[s]*= (1.0+b[s])*(1.0-b[s]);
            }

            // Store weight update
            // Hidden -> output
            for (int s = 0; s < bpmll->M; s++)
                for (int j = 0; j < Q; j++)
                    delta_w[s][j] += parm->alpha * d[j] * b[s];

            // Bias
            for (int j = 0; j < Q; j++)
                delta_w[bpmll->M][j] += parm->alpha * d[j];   // Bias

            // Input -> hidden
            for (int h = 0; h < inputs; h++)
                for (int s = 0; s < bpmll->M; s++)
                    delta_v[h][s] += parm->alpha * e[s] * data[next][h];

            // Bias
            for (int j = 0; j < bpmll->M; j++)
                delta_v[inputs][j] += parm->alpha * e[j];   // Bias
        }

        // Update weights
        // Hidden -> output
        for (int s = 0; s < bpmll->M+1; s++)
            for (int j = 0; j < Q; j++)
                bpmll->w[s][j] += delta_w[s][j];

        // Input -> hidden
        for (int h = 0; h < inputs+1; h++)
            for (int s = 0; s < bpmll->M; s++)
                bpmll->v[h][s] += delta_v[h][s];
    }

    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

    FREE(b);
    FREE(c);
    FREE(e);
    FREE(d);
    FreeScalarMatrix(&delta_v, inputs+1);
    FreeScalarMatrix(&delta_w, bpmll->M+1);
}

void FreeBPMLL(t_BPMLL *bpmll, int d)
{
    FreeScalarMatrix(&(bpmll->v), d);
    FreeScalarMatrix(&(bpmll->w), bpmll->M);
    FREE(bpmll->w_t);
}

void GetBPMLLPOutput(t_BPMLL *bpmll, int Q, int *subspace, scalar *instance, int d, scalar *c, scalar *b)
{
    int *l_subspace;

    if (subspace)
        l_subspace = subspace;
    else {
        MALLOC(l_subspace, int, d, "MLBasciBPClassifyInstance: l_subspace");
        for (int i = 0; i < d; i++)
            l_subspace[i] = 1;
    }

    // Output of hidden layer
    for (int s = 0; s < bpmll->M; s++) {
        b[s] = 0.0;
        for (int h = 0; h < d; h++)
            b[s] += bpmll->v[h][s]*instance[h]*l_subspace[h];
        b[s] += bpmll->v[d][s]; // Bias

        b[s] = tanh(b[s]);
    }

    // Output layer
    for (int j = 0; j < Q; j++) {
        c[j] = 0.0;
        for (int s = 0; s < bpmll->M; s++)
            c[j] += bpmll->w[s][j]*b[s];
        c[j] += bpmll->w[bpmll->M][j];  // Bias

        c[j] = tanh(c[j]);
    }
}

void GetThresholdWeights(t_BPMLL *bpmll, scalar **data, int instances, int **labels, int *n_labels, int Q, int d, int *subspace)
{
    scalar b[bpmll->M];

    gsl_vector *t = gsl_vector_alloc(instances);
    gsl_matrix *Phi = gsl_matrix_alloc(instances, Q+1);

    // Get Phi and t
    for (int i = 0; i < instances; i++) {
        GetBPMLLPOutput(bpmll, Q, subspace, data[i], d, Phi->data+(i*(Q+1)), b);
        Phi->data[i*(Q+1)+Q] = 1.0;    // Bias

        // Get optimum t value
        t->data[i] = GetOptimumThreshold(Phi->data+i*(Q+1), Q, labels[i], n_labels[i]);
    }

    // Obtain w solving the linear equations: Phi·w' = t
    gsl_vector *w = gsl_vector_alloc(Q+1);
    gsl_matrix *cov = gsl_matrix_alloc(instances, instances);
    double chisq;

    // RemoveRepeatedLinesWithVector(Phi, t);

    gsl_multifit_linear_workspace *work = gsl_multifit_linear_alloc(instances, Q+1);
    gsl_multifit_linear(Phi, t, w, cov, &chisq, work);

    for (int i = 0; i < Q+1; i++)
        bpmll->w_t[i] = w->data[i];

    gsl_vector_free(w);
    gsl_vector_free(t);
    gsl_matrix_free(Phi);
    gsl_matrix_free(cov);
    gsl_multifit_linear_free(work);
}

scalar GetOptimumThreshold(scalar *c, int Q, int *labels, int n_labels)
{
    int eval[Q];
    scalar values[Q];

    // Copy values and compact vector
    for (int i = 0; i < Q; i++)
        values[i] = c[i];
    qsort(values, Q, sizeof(scalar), CompareDouble);
    int n = Q;
    CompactVector(values, NULL, &n);

    // Evaluate every possible threshold
    for (int i = 0; i < n; i++) {
        eval[i] = 0;
        for (int j = 0; j < Q; j++) {
            bool present = LabelInList(j, labels, n_labels);
            if ((present && c[j] >= values[i]) || (!present && c[j] < values[i]))
                eval[i]++;
        }
    }

    // Get optimum or linear combination
    int index = 0, count = 1;
    scalar t = values[0];

    for (int i = 1; i < n; i++)
        if (eval[i] > eval[index]) {
            index = i;
            count = 1;
            t = values[index];
        } else if (eval[i] == eval[index]) {
            count++;
            t += values[i];
        }

    return t/count;
}

#define VALUE(m, i, j)  (m->data[(i)*m->size2+(j)])

// Remove repeated lines and the same values from vector
void RemoveRepeatedLinesWithVector(gsl_matrix *m, gsl_vector *v)
{
    for (int i = 0; i < m->size1; i++)
        for (int j = 0; j < i; j++) {
            bool repeated = true;

            for (int k = 0; k < m->size2 && repeated; k++)
                if (fabs(VALUE(m, i, k) - VALUE(m, j, k)) > ZERO)
                    repeated = false;

            // Substitute with last if repeated
            if (repeated) {
                for (int j = 0; j < m->size2; j++)
                    VALUE(m, i, j) = VALUE(m, m->size1-1, j);

                v->data[i] = v->data[m->size1-1];

                // Remove last one and revisit i-th
                m->size1--;
                v->size--;
                i--;

                // Cannot be repeated more than once
                break;
            }
        }
}
