/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Binary Relevance method

void TrainBinaryRelevanceModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_BR *br, int *subspace, bool cv,
                                t_Classifier classifier, int *subset, int **SS, int **index)
{
    int **b_labels, *l_subspace;

    // Test if valid BR classification model
    if (/*classifier == kNN || */classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "BR classification method");

    // If no subspace select all
    if (!subspace) {
        MALLOC(l_subspace, int, inputs, "TrainBinaryRelevanceModels: l_subspace");
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = 1;
    } else
        l_subspace = subspace;

    // Label dependent instance selection and standard instance selection are incompatible
    if (subset && SS)
        OnErrorExit(ERR_PARAM, "Label dependent instance selection and standard instance selection are incompatible");

    // Create binary labels
    AllocateIntegerMatrix(&b_labels, classes, instances);
    ZeroIntegerMatrix(b_labels, classes, instances);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            b_labels[labels[i][j]][i] = 1;

    br->method = classifier;
    br->models = NULL;
    br->ensembles = NULL;
    if (!IsEnsemble(classifier)) {
        MALLOC(br->models, t_Learner, classes, "TrainBinaryRelevanceModels: br->models");
    } else
        MALLOC(br->ensembles, t_Ensemble, classes, "TrainBinaryRelevanceModels: br->ensembles");

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    // Train a model for every label
    for (int i = 0; i < classes; i++) {
        // If undersample, use local subset
        int *S = NULL;
        MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

        if (SS)
            memcpy(S, SS[i], sizeof(int)*instances);
        else if (subset)
            memcpy(S, subset, sizeof(int)*instances);
        else
            FillVector(NULL, S, instances, 1.0);

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, b_labels[i], S, instances);

        // Cross-validate hyper-parameters?
        t_Parameters lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, data, instances, b_labels[i], l_subspace, inputs, 2, classifier, 1, S, index);

        // Train
        SHOWPROGRESS(parm->show_progress > 0, "BR training model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);
        ClassifierTraining(&lparm, classifier, data, instances, b_labels[i], inputs, 2, index, l_subspace, S, true, &(br->models[i]), &(br->ensembles[i]));
        SHOWPROGRESS(parm->show_progress > 0, "BR finished training model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);

        // For k-NN add labels
        if (classifier == kNN || classifier == NN) {
            MALLOC(br->models[i].knn->labels, int, instances, "TrainBinaryRelevanceModels: br->models[i].knn->labels");
            memcpy(br->models[i].knn->labels, b_labels[i], instances*sizeof(int));
        }

        if (parm->print_base_training_error)
            PrintClassifierPerformanceMetrics(parm, instances, inputs, data, 2, index, l_subspace, S, b_labels[i], br->method, &(br->models[i]), &(br->ensembles[i]), 0);

        FREE(S);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FreeIntegerMatrix(&b_labels, classes);
    if (!subspace)
        FREE(l_subspace);
}

void BinaryRelevanceClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                                     int inputs, int *query_index, t_BR *br, int *subspace, int *S, int **SS, int **index)
{
    *n_predicted = 0;

    // Output of all binary classifiers
    for (int i = 0; i < n_labels; i++) {
        int category = Classifier2ClassOutput(parm, br->method, query, inputs, 2, query_index, subspace, data, instances, (br->method == kNN || br->method == NN)? br->models[i].knn->labels :
                        NULL, SS ? SS[i] : S, &(br->models[i]), 1, &f[i], index, &(br->ensembles[i]));

        if (category) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }
    }

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *classification = *ranks = true;
}

// Undersample data using one label to ratio 1:1

#define MINIMUM_SAMPLES 100

void UndersampleDataWithLabel(const t_Parameters *parm, int *labels, int *S, int instances)
{
    int count[] = {0, 0};

    // Obtain minority class
    for (int i = 0; i < instances; i++)
        if (S[i])
            count[labels[i]]++;

    int minority = (count[0] < count[1] ? 0 : 1);

    scalar ratio = ((scalar) count[minority])/count[(minority+1)%2];

    // Sample all minority class instances and ratio% of majority class
    // Not exactly 1:1 but faster
    int undersample = 0;
    for (int i = 0; i < instances; i++)
        if (S[i]) {
            if (labels[i] == minority)
                undersample++;
            else {
                if (RANDOM(0.0, 1.0) < ratio)
                    undersample++;
                else
                    S[i] = 0;
            }
        }

    // At least MINIMUM_SAMPLES
    if (undersample < MINIMUM_SAMPLES)
        for ( ; undersample < MINIMUM_SAMPLES && undersample < instances; undersample++) {
            int next;
            do {
                next = RANDOM(0, instances);
            } while (S[next]);
            S[next] = 1;
        }

    SHOWPROGRESS(parm->show_progress > 1, "Undersampling from %d -> %d\n", instances, undersample);
}

void FreeBinaryRelevance(const t_Parameters *parm, t_BR *br, int classes, int inputs)
{
    for (int i = 0; i < classes; i++)
        // Free if not null
        if (br->models[i].svm) {    // Is an union, check any of them
            if (br->method == NN || br->method == kNN)
                FREE(br->models[i].knn->labels);
            ClassifierFree(parm, &(br->models[i]), br->method, 2, inputs, &(br->ensembles[i]));
        }

    FREE(br->models);
}
