/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Classifier chains method

// Also nested stacking:
// R. Senge, J.J. del Coz, E. Hüllermeier, Rectifying classifier chains for multi-label classification, in: A. Henrich, H.-C. Sperker (Eds.), Lernen, Wissen, Adaption
// (LWA) 2013, Proceedings, 2013, pp. 162–169.

void TrainClassifierChainModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_CC *cc, int *subspace, bool cv,
    t_Classifier classifier, int *subset, int *labels_chain, int **SS, int **index, bool ns)
{
    int **b_labels, *l_subspace, **t_index = NULL, *count = NULL;
    scalar **t_data;

    // Test if valid CC classification model
    if (classifier == kNN || classifier == IBT || classifier == OVO || classifier == NN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CC classification method");

    // Only used for NS with subset correction
    cc->diff = NULL;
    cc->n_diff = 0;

    // Set classifier
    cc->method = classifier;

    // If no subspace select all
    MALLOC(l_subspace, int, inputs+classes, "TrainDependentBinaryRelevanceModels: l_subspace");
    if (!subspace) {
        for (int i = 0; i < inputs+classes; i++)
            l_subspace[i] = 1;
    } else {
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = subspace[i];
        for (int i = inputs; i < inputs+classes; i++)
            l_subspace[i] = 1;
    }

    // Label dependent instance selection and standard instance selection are incompatible
    if (subset && SS)
        OnErrorExit(ERR_PARAM, "Label dependent instance selection and standard instance selection are incompatible");

    // Generate chain if given is not valid
    bool *test, valid = true;
    CALLOC(test, bool, classes, "TrainClassifierChainModels: test");
    for (int i = 0; i < classes && valid; i++)
        if (labels_chain[i] < 0 || labels_chain[i] > classes-1)
            valid = false;
        else
            test[labels_chain[i]] = true;

    for (int i = 0; i < classes && valid; i++)
        if (!test[i])
            valid = false;
    FREE(test);

    if (!valid) {
        for (int i = 0; i < classes; i++)
            labels_chain[i] = i;
        if (parm->ml_cc_random_order)
            RandomizeVector(labels_chain, classes);
    }

    // Copy chain to model
    MALLOC(cc->chain, int,  classes, "TrainClassifierChainModels: cc->chain");
    for (int i = 0; i < classes; i++)
        cc->chain[i] = labels_chain[i];

    // Dense version
    if (!index) {
        AllocateScalarMatrix(&t_data, instances, inputs+classes);
        CopyScalarMatrix(data, t_data, instances, inputs);
    } else {
        MALLOC(t_index, int *, instances, "TrainClassifierChainModels: t_index");
        MALLOC(t_data, scalar *, instances, "TrainClassifierChainModels: t_data");
        MALLOC(count, int, instances, "TrainClassifierChainModels: count")
        for (int i = 0; i < instances; i++) {
            count[i] = 0;
            while (index[i][count[i]] != -1)
                count[i]++;
            MALLOC(t_index[i], int, count[i]+classes+1, "TrainClassifierChainModels: t_index[i]");
            MALLOC(t_data[i], scalar, count[i]+classes, "TrainClassifierChainModels: t_data[i]");
            memcpy(t_data[i], data[i], sizeof(scalar)*count[i]);
            memcpy(t_index[i], index[i], sizeof(int)*(count[i]+1));
        }
    }

    // Create binary labels
    AllocateIntegerMatrix(&b_labels, classes, instances);
    ZeroIntegerMatrix(b_labels, classes, instances);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            b_labels[labels[i][j]][i] = 1;

    // Save chain
    fprintf(parm->out_file, "Classifier's chain:");
    for (int i = 0; i < classes; i++)
        fprintf(parm->out_file, " %d", cc->chain[i]);
    fprintf(parm->out_file, "\n");

    if (!IsEnsemble(cc->method)) {
        MALLOC(cc->models, t_Learner, classes, "TrainClassifierChainModels: cc->models");
    } else
        MALLOC(cc->ensembles, t_Ensemble, classes, "TrainClassifierChainModels: cc->ensembles");

    // CC can be parallelized, NS cannot
    if (ns) {
        // If undersample, use local subset
        int *S = NULL;
        if (parm->ml_undersample_binary_learners)
            MALLOC(S, int, instances, "TrainClassifierChainModels: S");

        // Train a model for every label
        scalar **outputs;
        AllocateScalarMatrix(&outputs, instances, classes);

        for (int i = 0; i < classes; i++) {
            // Undersample?
            if (parm->ml_undersample_binary_learners) {
                if (subset)
                    memcpy(S, subset, sizeof(int)*instances);
                else
                    for (int j = 0; j < instances; j++)
                        S[j] = 1;

                UndersampleDataWithLabel(parm, b_labels[cc->chain[i]], S, instances);
            }

            if (SS)
                subset = SS[cc->chain[i]];

            // Cross-validate hyper-parameters?
            SHOWPROGRESS(parm->show_progress > 0, "NS training model %*d/%d for label %*d CV parameters\r", WIDTH(classes), i+1, classes, WIDTH(classes), cc->chain[i]+1);
            if (cv)
                SetCrossValidatedParameters((t_Parameters *) parm, t_data, instances, b_labels[cc->chain[i]], l_subspace, inputs+i, 2, cc->method, 1,
                                            parm->ml_undersample_binary_learners ? S : subset, t_index);

            // Train
            SHOWPROGRESS(parm->show_progress > 0, "NS training model %*d/%d for label %*d              \r", WIDTH(classes), i+1, classes, WIDTH(classes), cc->chain[i]+1);
            ClassifierTraining(parm, cc->method, t_data, instances, b_labels[cc->chain[i]], inputs+i, 2, t_index, l_subspace, parm->ml_undersample_binary_learners ? S : subset, true,
                                     &(cc->models[i]), &(cc->ensembles[i]));
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            // Add next label to inputs depending on CC or NS method
            for (int j = 0; j < instances; j++)
                Classifier2ClassOutput(parm, cc->method, data[j], inputs+i, 2, index ? index[j] : NULL, subspace, data, instances, NULL, SS ? SS[i] : S, &(cc->models[i]), 1, &outputs[j][i],
                                                 index, &(cc->ensembles[i]));

            // Dense format
            if (!index)
                for (int j = 0; j < instances; j++)
                    t_data[j][inputs+i] = outputs[j][i];
            else
                for (int j = 0; j < instances; j++) {
                    if (fabs(outputs[j][i]) > ZERO) {
                        count[j]++;

                        t_data[j][count[j]-1] = outputs[j][i];
                        t_index[j][count[j]-1] = inputs+i+1;
                        t_index[j][count[j]] = -1;
                    }
                }
        }
        SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

        FreeScalarMatrix(&outputs, instances);
        if (S)
            FREE(S);
    } else {
        // Add next label to inputs
        for (int i = 0; i < classes; i++)
            // Dense format
            if (!index)
                for (int j = 0; j < instances; j++)
                    t_data[j][inputs+i] = b_labels[cc->chain[i]][j];

#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < classes; i++) {
            t_Parameters lparm;

            // If undersample, use local subset
            int *S = NULL;
            MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

            if (SS)
                memcpy(S, SS[cc->chain[i]], sizeof(int)*instances);
            else if (subset)
                memcpy(S, subset, sizeof(int)*instances);
            else
                for (int j = 0; j < instances; j++)
                    S[j] = 1;

            // Undersample?
            if (parm->ml_undersample_binary_learners)
                UndersampleDataWithLabel(parm, b_labels[i], S, instances);

            // Data must be local if sparse
            int **l_index = NULL;
            scalar **l_data = t_data;

            if (index) {
                AllocateScalarMatrix(&l_data, instances, inputs+i);
                AllocateIntegerMatrix(&l_index, instances, inputs+i+1);
                for (int j = 0; j < instances; j++) {
                    int c = 0;
                    for (; t_index[j][c] != -1; c++) {
                        l_data[j][c] = t_data[j][c];
                        l_index[j][c] = t_index[j][c];
                    }

                    for (int k = 0; k < i; k++)
                        if (b_labels[cc->chain[k]][j] > ZERO) {
                            l_data[j][c] = b_labels[cc->chain[k]][j];
                            l_index[j][c] = inputs+k+1;
                            c++;
                    }
                    l_index[j][c] = -1;
                }
            }

            // Cross-validate hyper-parameters?
            lparm = *parm;
            SHOWPROGRESS(parm->show_progress > 0, "CC training model %*d/%d for label %*d CV parameters\r", WIDTH(classes), i+1, classes, WIDTH(classes), cc->chain[i]+1);
            if (cv)
                SetCrossValidatedParameters(&lparm, l_data, instances, b_labels[cc->chain[i]], l_subspace, inputs+i, 2, cc->method, 1, S, l_index);

            // Train
            SHOWPROGRESS(parm->show_progress > 0, "CC training model %*d/%d for label %*d              \r", WIDTH(classes), i+1, classes, WIDTH(classes), cc->chain[i]+1);
            ClassifierTraining(&lparm, cc->method, l_data, instances, b_labels[cc->chain[i]], inputs+i, 2, l_index, l_subspace, S, true, &(cc->models[i]), &(cc->ensembles[i]));

            FREE(S);

            if (index) {
                FreeScalarMatrix(&l_data, instances);
                FreeIntegerMatrix(&l_index, instances);
            }
        }
        SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    }

    // If NS with subset correction get info of different sets of labels
    GetListOfTrainingPowersets(instances, labels, n_labels, &(cc->diff), &(cc->n_diff), classes);

    FreeIntegerMatrix(&b_labels, classes);
    FreeScalarMatrix(&t_data, instances);
    if (index) {
        FreeIntegerMatrix(&t_index, instances);
        FREE(count);
    }
    FREE(l_subspace);
}

void ClassifierChainClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                                     int inputs, int *query_index, t_CC *cc, int *subspace, int *S, int **SS, int **index, bool sc)
{
    scalar *q;
    int *q_index = NULL, count = 0;

    // Sparse
    if (query_index) {
        while (query_index[count] != -1)
            count++;
        MALLOC(q, scalar, count+n_labels, "ClassifierChainClassifyInstance: q");
        memcpy(q, query, sizeof(scalar)*count);
        MALLOC(q_index, int, count+1+n_labels, "ClassifierChainClassifyInstance: q_index");
        memcpy(q_index, query_index, sizeof(int)*(count+1));
    } else {    // Dense
        MALLOC(q, scalar, inputs+n_labels, "ClassifierChainClassifyInstance: q");
        memcpy(q, query, sizeof(scalar)*inputs);
    }

    // Output of all binary classifiers
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++) {
        int category = Classifier2ClassOutput(parm, cc->method, q, inputs+i, 2, query_index, subspace, data, instances, NULL, SS ? SS[i] : S, &(cc->models[i]), 1, &f[cc->chain[i]], index,
                                              &(cc->ensembles[i]));

        if (category) {
            h[*n_predicted] = cc->chain[i];
            (*n_predicted)++;
        }

        // Add for next round: Sparse or dense
        if (query_index) {
            if (fabs(f[cc->chain[i]]) > ZERO) {
                count++;

                q[count-1] = f[cc->chain[i]];
                q_index[count-1] = inputs+i+1;
                q_index[count] = -1;
            }
        } else
            q[inputs+i] = f[cc->chain[i]];
    }

    *classification = *ranks = true;

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    // Sort predicted labels
    qsort(h, *n_predicted, sizeof(int), CompareInt);

    // If subset correction only known label set can be predicted
    if (sc)
        GetClosestSetOfLabels(h, n_predicted, cc->diff, cc->n_diff, n_labels);

    FREE(q);
    if (query_index)
        FREE(q_index);
}

void FreeClassifierChain(const t_Parameters *parm, t_CC *cc, int classes, int inputs)
{
    for (int i = 0; i < classes; i++)
        if (cc->models[i].svm)  // Check any, it is an union
            ClassifierFree(parm, &(cc->models[i]), cc->method, 2, inputs, &(cc->ensembles[i]));

    FREE(cc->models);
    FREE(cc->chain);
}

// Get list of different powersets for subset corrrection of NS

void GetListOfTrainingPowersets(int instances, int **labels, int *n_labels, int ***list, int *n_different, int classes)
{
    // Label diversity
    scalar *decimal;
    MALLOC(decimal, scalar, instances, "GetListOfTrainingPowersets: decimal");

    // For label diversity, first convert to decimal
    for (int j = 0; j < instances; j++) {
        decimal[j] = 0.0;
        for (int i = 0; i < n_labels[j]; i++)
            decimal[j] += pow(2.0, labels[j][i]);
    }

    // Sort decimal values
    qsort(decimal, instances, sizeof(int), CompareInt);

    // Store different values in list
    *n_different = 1;
    MALLOC(*list, int *, instances, "GetListOfTrainingPowersets: list");
    MALLOC((*list)[0], int, classes, "GetListOfTrainingPowersets: (*list)[0]");
    DecodifyLabelPowerset(decimal[0], classes, (*list)[0]);

    scalar last = decimal[0];

    for (int i = 1; i < instances; i++)
        if (decimal[i] != last) {
            MALLOC((*list)[*n_different], int, classes, "GetListOfTrainingPowersets: (*list)[0]");
            DecodifyLabelPowerset(decimal[i], classes, (*list)[*n_different]);
            (*n_different)++;
            last = decimal[i];
        }

    FREE(decimal);

    REALLOC(*list, int *, *n_different, "GetListOfTrainingPowersets: *list");
}

// int HammingDistanceForIntegers(int n1, int n2)
// {
//     int x = n1 ^ n2, bits = 0;
//
//     while (x > 0) {
//         bits += x & 1;
//         x >>= 1;
//     }
//
//     return bits;
// }

void GetClosestSetOfLabels(int *labels, int *n_labels, int **list, int n, int classes)
{
   int min_distance, distance, closest = 0, *bits;

   MALLOC(bits, int, classes, "GetClosestSetOfLabels: bits");
   ZeroVector(NULL, bits, classes);
   for (int i = 0; i < *n_labels; i++)
       bits[labels[i]] = 1;

   min_distance = HammingDistance(bits, list[0], classes);

   for (int i = 1; i < n && min_distance; i++)
       if ((distance = HammingDistance(bits, list[i], classes)) < min_distance) {
           min_distance = distance;
           closest = i;
    }

    // If minimum distance is 0 nothing to do
    if (min_distance) {
        *n_labels = 0;
        for (int i = 0; i < classes; i++)
            if (list[closest][i])
                labels[(*n_labels)++] = i;
    }

    FREE(bits);
}
