/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

// CHC Feature Selection

int CHCFeatureSelectionForML(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int *selection, int **index, t_MLClassifier method,
                              int *generations_performed, int *chain)
{
    int **population, *list, selected = 0, loperk_count = 0, gs_without_improvement = 0, **ccs;
    scalar *fitness, *performance, loperk_inc = 0.0, old_best = -1.0;
    t_HashTable ht;

    if (method == LAMLksNN)
        OnErrorExit(ERR_PARAM, "LAML-k'sNN not suitable for feature selection");

    // Length of individuals
    int length = inputs;

    // Adjust minimum distance taking into account selection probability
    int minimum_distance = MAX(1, (int)((length*parm->chc_init_prob)/parm->chc_minimum_distance_factor));

    SAVEPROGRESS(true, "Effective minimum distance: %d\n", minimum_distance);
    SHOWPROGRESS(parm->show_progress > 0, "Effective minimum distance: %d\n", minimum_distance);

    // Population size must be even
    if (parm->population_size % 2)
        ((t_Parameters *)parm)->population_size++;

    // Create and initialize global population
    MALLOC(fitness, scalar, 2*parm->population_size, "CHCFeatureSelectionForML: fitness");
    MALLOC(performance, scalar, 2*parm->population_size, "CHCFeatureSelectionForML: performance");
    MALLOC(list, int, parm->population_size, "CHCFeatureSelectionForML: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;
    MALLOC(population, int *, 2*parm->population_size, "CHCFeatureSelectionForML: population");
    for (int i = 0; i < 2*parm->population_size; i++)
        MALLOC(population[i], int, length, "CHCFeatureSelectionForML: population[i]");

    for (int i = 0; i < parm->population_size; i++)
        for (int j = 0; j < length; j++)
            population[i][j] = (RANDOM(0.0, 1.0) < parm->chc_init_prob) ? 1 : 0;

    if (parm->ml_use_hash_table)
        CreateHashTable(&ht);

    // If cross-validate hyper-parameters k is fixed to best value
    if (method == MLkNN && parm->cv_ml_hyperparameters) {
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);
        ((t_Parameters *)parm)->ml_min_k = ((t_Parameters *)parm)->ml_max_k = parm->mlknn_k;
        printf("WARNING: k fixed to %d, not evolved\n", parm->mlknn_k);
    }

    // Evolution of k
    int *k = NULL;
    MALLOC(k, int, 2*parm->population_size, "CHCFeatureSelectionForML: k");
    for (int i = 0; i < parm->population_size; i++)
        k[i] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);

    // Evolution of classifier chain if needed
    AllocateIntegerMatrix(&ccs, 2*parm->population_size, classes);
    for (int i = 0; i < parm->population_size; i++) {
        // Each individual is a permutation of the labels
        for (int j = 0; j < classes; j++)
            ccs[i][j] = j;

        // Random chains make no sense with stratification
        if (parm->data_partition_method == NoDataPartition || !parm->subset_size)
            RandomizeVector(ccs[i], classes);
    }

    t_MLLearner learner[2*parm->population_size];
    t_MLEnsemble ensemble[2*parm->population_size];

    // Evaluate initial individuals
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < parm->population_size; i++)
        fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, NULL, &ht, NULL, index, &performance[i], NULL, NULL, method,
                                               &(learner[i]), NULL, 0, ccs[i], k[i], NULL, 0, population[i], &(ensemble[i]), instances);

    time_t init_time = time(NULL);
    scalar new = 0.0;

    *generations_performed = parm->generations;
    for (int g = 0; g < parm->generations; g++) {
        SHOWPROGRESS(parm->show_progress > 0, "Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);

        time_t init_generation = time(NULL);

        // Random selection without replacement
        RandomizeVector(list, parm->population_size);
        int actual_size = parm->population_size;

        // Generate new individuals using crossover only
        for (int i = 0; i < parm->population_size; i += 2) {
            int father = list[i];
            int mother = list[i+1];

            // If no minimum distance skip
            if (HammingDistance(population[mother], population[father], length) < minimum_distance)
                continue;

            // Perform HUX crossover creating new individuals
            HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], length);

            // Perform BLX-alpha crossover for k
            if (parm->ml_use_hux)
                HuxCrossover(k+mother, k+father, k+actual_size, k+actual_size+1, classes);
            else
                BLXAlphaCrossover(k+actual_size, k+actual_size+1, k+father, k+mother, classes, parm->ml_min_k, parm->ml_max_k);

            // Perform permutation aware crossover of classifier chain
            OnePointPermutationCrossover(ccs[father], ccs[mother], ccs[actual_size], ccs[actual_size+1], classes);

            actual_size += 2;
        }

        // If apply evaluate all individuals
        int start = parm->population_size;
        if (parm->ml_external_eval_interval && ((g+1) % parm->ml_external_eval_interval == 0 || (g+1) % parm->ml_external_eval_interval == 1))
            start = 0;

        // Evaluate NEW individuals
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = start; i < actual_size; i++)

            fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, NULL, &ht, NULL, index, &performance[i], NULL, NULL, method,
                                                   &(learner[i]), NULL, g, ccs[i], k[i], NULL, 0, population[i], &(ensemble[i]), instances);

        ///////////////////////////////////////////
        // Mutation operations ////////////////////
        ///////////////////////////////////////////
        // Random mutation
        //         if (parm->ml_chc_random_mut > 0.0)
        // #ifdef _OPENMP
        //             #pragma omp parallel for schedule(dynamic)
        // #endif
        //             // If multi-selection also mutate k
        //             for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
        //                 if (RANDOM(0.0, 1.0) < parm->ml_chc_random_mut) {
        //                     RandomMLMutation(parm, population[i], instances, parm->bit_mutation, g, parm->generations, parm->multiselection, max_oversampling_ratio);
        //
        //                     // Check if multi-selection is correct
        //                     if (parm->multiselection && parm->multiselect_only_minority_class)
        //                         for (int j = 0; j < instances; j++)
        //                             if (!min_samples[j])
        //                                 population[i][j] = (population[i][j] >= 1);
        //
        //                     RandomMLMutation(parm, NULL, 1, parm->bit_mutation, g, parm->generations, true, max_oversampling_ratio);
        //                     // Avoid k outside bounds
        //                     k[i] = AssignWithinBoundsInteger(parm->ml_min_k, k[i], parm->ml_max_k);
        //
        //                     // Random mutation for the permutation of classes
        //                     RandomMutationForPermutations(ccs[i], classes);
        //
        //                     fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL,
        //                                                            NULL, &ht, NULL, index, &performance[i], NULL, NULL, method, &(learner[i]), NULL,
        //                                                            g, ccs[i], k[i], NULL, 0);
        //                 }


        // Memetic mutation for k
        //         if (parm->ml_k_memetic_mutation_ratio > 0.0)
        // #ifdef _OPENMP
        //             #pragma omp parallel for schedule(dynamic)
        // #endif
        //             for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
        //                 if (RANDOM(0.0, 1.0) < parm->ml_k_memetic_mutation_ratio)
        //                     kMemeticMutation(parm, population[i], data, instances, labels, n_labels, classes, NULL, method, &fitness[i], k[i], 0, inputs, index, NULL, ccs[i]);


        // Reinitialize or elitism
        if (actual_size == parm->population_size) {
            minimum_distance--;
            SHOWPROGRESS(parm->show_progress > 0, "No individuals generated, decreasing minimum distance to: %d\n", minimum_distance);
        } else
            for (int i = 0; i < parm->population_size; i++) {
                int order = OrderedSelection(fitness+i, 0, actual_size-i)+i;

                int *aux_pt = population[order];
                population[order] = population[i];
                population[i] = aux_pt;

                scalar aux_f = fitness[order];
                fitness[order] = fitness[i];
                fitness[i] = aux_f;

                aux_f = performance[order];
                performance[order] = performance[i];
                performance[i] = aux_f;

                int aux_i = k[order];
                k[order] = k[i];
                k[i] = aux_i;

                aux_pt = ccs[order];
                ccs[order] = ccs[i];
                ccs[i] = aux_pt;
            }

        // Improvement?
        if (fitness[0] > old_best) {
            old_best = fitness[0];
            gs_without_improvement = 0;
        } else
            gs_without_improvement++;

        // If minimum distance 0 reinitialize
        if (minimum_distance == 0 || (parm->reinitalize_after_gs_no_improvement && gs_without_improvement > parm->reinitalize_after_gs_no_improvement)) {
            old_best = -1.0;
            // Reinitialize populations excepting best one
            int best = OrderedSelection(fitness, 0, parm->population_size);

            for (int i = 0; i < parm->population_size; i++)
                // Reseed changing 35%
                if (i != best) {
                    for (int k = 0; k < length; k++)
                        population[i][k] = RANDOM(0.0, 1.0) < 0.35 ? (population[best][k]+1)%2 : population[best][k];
                    for (int j = 0; j < classes; j++)
                        k[i] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);
                    for (int j = 0; j < classes; j++)
                        ccs[i][j] = j;
                    RandomizeVector(ccs[i], classes);

                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, NULL, &ht, NULL, index, &performance[i], NULL, NULL, method,
                                                           &(learner[i]), NULL, g, ccs[i], k[i], NULL, 0, population[i], &(ensemble[i]), instances);
                }

            minimum_distance = MAX(1, (int)((length*parm->chc_init_prob)/parm->chc_minimum_distance_factor));
            SHOWPROGRESS(parm->show_progress > 0, "\nPopulation reinitialized at generation %d\n", g+1);
        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            switch (method) {
                case MLkNN:
                    SHOWPROGRESS(parm->show_progress > 0, "%*d g's best P(MEAN), F(MEAN), k: %6.4lf(%6.4lf), %6.4lf(%6.4lf), %3d %s", WIDTH(parm->generations), g+1, performance[best],
                                 AverageScalarVector(performance, parm->population_size), fitness[best], AverageScalarVector(fitness, parm->population_size), k[best],
                                 parm->ml_external_eval_interval && ((g+1) % parm->ml_external_eval_interval) == 0 ? "*" : " ");

                    SHOWPROGRESS(parm->show_progress > 0, " (New = %*d) ", WIDTH(parm->population_size),  actual_size - parm->population_size);
                    SHOWPROGRESS(parm->show_progress > 0, "(Time = %4lds, %4lds)\n", time(NULL) - init_generation,  time(NULL) - init_time);
                    break;

                case LAMLksNN:
                case LAMLkNN:
                case MLksNN:
                    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "FS for this method");
                    break;

                default:
                    SHOWPROGRESS(parm->show_progress > 0, "%*d g's best P(MEAN), F(MEAN): %6.4lf(%6.4lf), %6.4lf(%6.4lf) %s", WIDTH(parm->generations), g+1, performance[best],
                                 AverageScalarVector(performance, parm->population_size), fitness[best], AverageScalarVector(fitness, parm->population_size),
                                 parm->ml_external_eval_interval && ((g+1) % parm->ml_external_eval_interval) == 0 ? "*" : " ");
                    SHOWPROGRESS(parm->show_progress > 0, " (New = %*d) ", WIDTH(parm->population_size),  actual_size - parm->population_size);
                    SHOWPROGRESS(parm->show_progress > 0, "(Time = %4lds, %4lds)\n", time(NULL) - init_generation,  time(NULL) - init_time);
                    break;
            }

            VERBOSE(parm->show_progress > 1, parm->out_file, "%*d generation best individual PERFORMANCE, FITNESS(MEAN): %6.4lf, %6.4lf(%6.4lf)\n", WIDTH(parm->generations), g+1, fitness[best],
                    performance[best], AverageScalarVector(fitness, parm->population_size));
        }

        // Percentage of new individuals per generation
        new += actual_size - parm->population_size;

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > (time_t) parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);
            *generations_performed = g+1;
            break;
        }
    }

    fprintf(parm->out_file, "@Percentage of new individuals per generation: %lf\n", new/(*generations_performed*parm->population_size));


    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    SHOWPROGRESS(parm->show_progress > 0 && k && loperk_count > 0, "Average increment for k per label local optimization: %6.4lf\n", loperk_inc/loperk_count);
    VERBOSE(parm->show_progress > 0 && k && loperk_count > 0, parm->out_file, "Average increment for k per label local optimization: %6.4lf\n", loperk_inc/loperk_count);


    int best = OrderedSelection(fitness, 0, parm->population_size);

    selected = 0;
    for (int i = 0; i < instances; i++)
        if ((selection[i] = population[best][i]))
            selected++;

    ((t_Parameters *) parm)->mlknn_k = k[best];

    if (chain) {
        fprintf(parm->out_file, "Best chain:");
        for (int i = 0; i < classes; i++) {
            chain[i] = ccs[best][i];
            fprintf(parm->out_file, " %d", chain[i]);
        }
        fprintf(parm->out_file, "\n");
    }

#ifdef VERBOSE_MODE
    SHOWPROGRESS(true, "Classifier chain%c", ':');
    for (int i = 0; i < classes; i++)
        SHOWPROGRESS(true, " %d", chain[i]);
    SHOWPROGRESS(true, "%c", '\n');
#endif

    // Free memory
    for (int i = 0; i < 2*parm->population_size; i++)
        FREE(population[i]);
    FREE(population);
    FREE(fitness);
    FREE(list);
    FREE(performance);
    if (parm->ml_use_hash_table)
        DestroyHashTable(&ht);
    FREE(k);

    return selected;
}
