/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

int CHCInstanceSelectionForML(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int *selection, int **index, int **sselection,
                              int **initial_population, int *ks, t_MLClassifier method, int **ksc, int *generations_performed, scalar *IRLbl, scalar MeanIR, int *chain)
{
    int **population, *list, * **ss = NULL, selected = 0, loperk_count = 0, gs_without_improvement = 0, **ccs;
    scalar *fitness, *performance, best_performance = -1.0, best_fitness = -1.0, loperk_inc = 0.0, old_best = -1.0;
    cache_t **neighbors_cache = NULL;
    t_HashTable ht;

    if (method == LAMLksNN)
        OnErrorExit(ERR_PARAM, "LAML-k'sNN not suitable for instance selection");

    // In case of multi-selection oversampling ratio could be given or could be depending on maximum k
    int max_oversampling_ratio = parm->ml_max_ioversampling_ratio != 0 ? parm->ml_max_ioversampling_ratio : parm->ml_max_k;
    if (!parm->multiselection)
        max_oversampling_ratio = 1;

    // 3 way must always be coupled with LDIS
    if (parm->three_way_is) {
        ((t_Parameters *) parm)->ml_ins_sel_per_label = true;
        ((t_Parameters *) parm)->symmetric_selection = false;
        ((t_Parameters *) parm)->ml_ins_sel_per_label_same_neighbors = false;
    }

    if (parm->symmetric_selection && parm->ml_ins_sel_per_label)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Symmetric evolution for instance selection per label");

    if (method != MLkNN && parm->multiselection && parm->ml_relative_multiselection)
        OnErrorExit(ERR_PARAM, "Relative multi-selection only available for ML-kNN");

    if (parm->multiselection &&  parm->ml_relative_multiselection)
        return RCGAMultiSelectionForML(parm, instances, data, labels, n_labels, classes, inputs, selection, index, ks, method, ksc, generations_performed);

    // Population size must be even
    if (parm->population_size % 2)
        ((t_Parameters *)parm)->population_size++;

    // Length of individuals
    int length = instances;
    if (parm->ml_ins_sel_per_label) {
        length = instances*classes;
        MALLOC(ss, int **, 2*parm->population_size, "CHCInstanceSelectionForML: ss");
        for (int i = 0; i < 2*parm->population_size; i++)
            MALLOC(ss[i], int *, classes, "CHCInstanceSelectionForML: ss[i]");

        if (parm->initialize_all_minority_class)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Minority class for per label selection");

        if (parm->ml_undersample_binary_learners)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Undersample binary learners for per label selection");
    }

    // Obtain majority/minority of each instance
    bool minority[classes];
    for (int i = 0; i < classes; i++)
        minority[i] = IRLbl[i] > MeanIR ? true : false;

    bool *min_samples;
    int min_count = 0;
    MALLOC(min_samples, bool, instances, "CHCInstanceSelectionForML: min_samples");
    for (int i = 0; i < instances; i++) {
        min_samples[i] = false;
        for (int j = 0; j < n_labels[i]; j++)
            if (minority[labels[i][j]]) {
                min_samples[i] = true;
                min_count++;
                break;
            }
    }

    fprintf(parm->out_file, "@Minority samples: %d\n", min_count);
    fprintf(parm->out_file, "@Minority samples ratio: %lf\n", ((scalar)min_count)/instances);

    // Cache of neighbors?
    scalar **distance_cache = NULL;
    int neighbors_cache_size = 0;
    SHOWPROGRESS(parm->show_progress > 0, "Cache size (maximum size): %.1lfGb (%.1lfGb)\n", NEIGHBORS_CACHE_SIZE(instances), MAXIMUM_CACHE_SIZE);
    if (method == MLkNN && parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(data, instances, inputs, &distance_cache, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
    else if (method == MLkNN && parm->ml_loo)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation with LOO without cache");

    // Performance with all instances, if needed
    t_Performance *perf;
    if ((parm->fitness_as_ratio || parm->ml_is_set_k_cv) && (method == MLkNN || method == LAMLkNN)) {
        int old_k;
        old_k = parm->mlknn_k;
        if (parm->cv_for_fitness_as_ratio || parm->ml_is_set_k_cv)
            CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);

        if (parm->ml_is_set_k_cv)
            ((t_Parameters *)parm)->ml_min_k = ((t_Parameters *)parm)->ml_max_k = parm->mlknn_k;

        ALLOCA(perf, t_Performance, 1, "CHCInstanceSelectionForML: perf");
        GetPerformanceWithAllInstances(parm, method, n_labels, instances, parm->mlknn_k, data, index, inputs, labels,  classes, neighbors_cache, perf, NULL, parm->lamlknn_n_clusters);
        ((t_Parameters *)parm)->mlknn_k = old_k;
    } else
        perf = NULL;

    // Adjust for multi-selection of instances
    scalar chc_init_prob = parm->chc_init_prob;
    if (parm->multiselection && parm->multiis_adjust_initial_percentage) {
        scalar sum = 0.0;
        int n = max_oversampling_ratio;

        for (int i = 1; i <= n; i++)
            sum += i;

        fprintf(parm->out_file, "Initial percentage adjusted: %lf", parm->chc_init_prob);

        // With this adjustment, multi-selection selects the same number of instances, although some of them repeated
        chc_init_prob /= sum/n;

        fprintf(parm->out_file, " => %lf\n", chc_init_prob);
    }

    // Adjust minimum distance taking into account selection probability
    int minimum_distance = MAX(1, (int)((length*chc_init_prob)/parm->chc_minimum_distance_factor));

    SAVEPROGRESS(true, "Effective minimum distance: %d\n", minimum_distance);
    SHOWPROGRESS(parm->show_progress > 0, "Effective minimum distance: %d\n", minimum_distance);

    // Create and initialize global population
    MALLOC(fitness, scalar, 2*parm->population_size, "CHCInstanceSelectionForML: fitness");
    MALLOC(performance, scalar, 2*parm->population_size, "CHCInstanceSelectionForML: performance");
    MALLOC(list, int, parm->population_size, "CHCInstanceSelectionForML: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;
    MALLOC(population, int *, 2*parm->population_size, "CHCInstanceSelectionForML: population");
    for (int i = 0; i < 2*parm->population_size; i++)
        MALLOC(population[i], int, length, "CHCInstanceSelectionForML: population[i]");

    if (parm->ml_ins_sel_per_label) {
        if (parm->three_way_is) {
            for (int i = 0; i < parm->population_size; i++)
                ThreeWayIndividualInitialization(parm, population[i], length, parm->ml_init_with_undersampling, instances, labels, n_labels, classes);
        } else if (parm->ml_init_with_undersampling)
            for (int i = 0; i < parm->population_size; i++)
                for (int j = 0; j < classes; j++)
                    MultiSelectionIndividulaWithRandomUndersampling(parm, population[i]+j*instances, labels, n_labels, j, chc_init_prob, instances, parm->multiselection);
        else if (parm->symmetric_selection)
            for (int i = 0; i < parm->population_size; i++)
                for (int j = 0; j < length; j++)
                    population[i][j] = (RANDOM(0.0, 1.0) < chc_init_prob) ? (RANDOM(0.0, 1.0) < 0.5 ? -1 : 1) : 0;
        else
            for (int i = 0; i < parm->population_size; i++)
                for (int j = 0; j < length; j++)
                    population[i][j] = (RANDOM(0.0, 1.0) < chc_init_prob) ? 1 : 0;
    } else if (parm->ml_init_with_undersampling && parm->multiselection)
        for (int i = 0; i < parm->population_size; i++)
            MultiSelectionIndividulaWithRandomUndersampling(parm, population[i], labels, n_labels, RANDOM(0, classes), chc_init_prob, instances, true);
    else if (parm->ml_init_with_roulette && parm->multiselection)
        for (int i = 0; i < parm->population_size; i++)
            MultiSelectionWithRoulette(parm, population[i], chc_init_prob, instances);
    else if (parm->ml_init_with_cardinality && parm->multiselection)
        for (int i = 0; i < parm->population_size; i++)
            MultiSelectionIndividualWithCardinality(parm, population[i], n_labels, classes, chc_init_prob, instances);
    else
        for (int i = 0; i < parm->population_size; i++) {
            if (initial_population)
                memcpy(population[i], initial_population[i], sizeof(int)*length);
            else
                for (int j = 0; j < length; j++)
                    population[i][j] = (RANDOM(0.0, 1.0) < chc_init_prob) || (parm->initialize_all_minority_class && min_samples[j]) ? (parm->multiselection ?
                                       RANDOM((parm->symmetric_selection ? -max_oversampling_ratio : 1), max_oversampling_ratio+1) : 1) : 0;
        }

    // Check if multi-selection is correct
    if (parm->multiselection && parm->multiselect_only_minority_class)
        for (int j = 0; j < instances; j++)
            if (!min_samples[j])
                for (int i = 0; i < parm->population_size; i++)
                    population[i][j] = (population[i][j] >= 1);

    if (parm->ml_use_hash_table)
        CreateHashTable(&ht);

    // If cross-validate hyper-parameters k is fixed to best value
    if (method == MLkNN && parm->cv_ml_hyperparameters) {
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);
        ((t_Parameters *)parm)->ml_min_k = ((t_Parameters *)parm)->ml_max_k = parm->mlknn_k;
        printf("WARNING: k fixed to %d, not evolved\n", parm->mlknn_k);
    }

    // Evolution of k
    int ***k = NULL;
    Allocate3DIntegerMatrix(&k, 2*parm->population_size, parm->lamlknn_max_n_clusters, classes);
    for (int i = 0; i < parm->population_size; i++)
        for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
            for (int j = 0; j < classes; j++)
                k[i][c][j] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);   // If only one k, k[i][0][0] is used

    // Evolution of classifier chain if needed
    AllocateIntegerMatrix(&ccs, 2*parm->population_size, classes);
    for (int i = 0; i < parm->population_size; i++) {
        // Each individual is a permutation of the labels
        for (int j = 0; j < classes; j++)
            ccs[i][j] = j;

        // Random chains make no sense with stratification
        if (parm->data_partition_method == NoDataPartition || !parm->subset_size)
            RandomizeVector(ccs[i], classes);
    }

    // Evolution of number of clusters for LAML-kNN and LAML-k'sNN
    int *clusters = NULL;
    MALLOC(clusters, int, 2*parm->population_size, "CHCInstanceSelectionForML: clusters");
    for (int i = 0; i < parm->population_size; i++)
        clusters[i] = (int) RANDOM(parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters+1);

    t_MLLearner learner[2*parm->population_size];
    t_MLEnsemble ensemble[2*parm->population_size];

    // Evaluate initial individuals
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < parm->population_size; i++) {
        if (parm->ml_ins_sel_per_label)
            for (int j = 0; j < classes; j++)
                ss[i][j] = population[i]+(j*instances);

        fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i], neighbors_cache, &ht, NULL,
                                               index, &performance[i], parm->ml_ins_sel_per_label ? ss[i] : NULL, NULL, method, &(learner[i]), perf, 0, ccs[i], k[i][0][0], k[i][0],
                                               clusters[i], NULL, &(ensemble[i]), instances);
    }

    time_t init_time = time(NULL);
    scalar new = 0.0;

    *generations_performed = parm->generations;
    for (int g = 0; g < parm->generations; g++) {
        SHOWPROGRESS(parm->show_progress > 0, "CHC ML IS Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);

        time_t init_generation = time(NULL);

        // Random selection without replacement
        RandomizeVector(list, parm->population_size);
        int actual_size = parm->population_size;

        // Generate new individuals using crossover only
        for (int i = 0; i < parm->population_size; i += 2) {
            int father = list[i];
            int mother = list[i+1];

            // If no minimum distance skip
            if (HammingDistance(population[mother], population[father], length) < minimum_distance)
                continue;

            // Perform HUX crossover creating new individuals
            if (parm->ml_use_hux)
                HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], length);
            else if (parm->ml_ins_sel_per_label && parm->biased_hux_ratio > 0.0)
                BiasedHuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], length, parm->biased_hux_ratio);
            else
                BLXAlphaCrossover(population[actual_size], population[actual_size+1], population[father], population[mother], length, parm->symmetric_selection ? -max_oversampling_ratio : 0,
                                  max_oversampling_ratio);

            // Check if multi-selection is correct
            if (parm->multiselection && parm->multiselect_only_minority_class)
                for (int j = 0; j < instances; j++)
                    if (!min_samples[j]) {
                        population[actual_size][j] = (population[actual_size][j] >= 1);
                        population[actual_size+1][j] = (population[actual_size+1][j] >= 1);
                    }

            // Perform BLX-alpha crossover for k
            for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
                if (parm->ml_use_hux)
                    HuxCrossover(k[mother][c], k[father][c], k[actual_size][c], k[actual_size+1][c], classes);
                else
                    BLXAlphaCrossover(k[actual_size][c], k[actual_size+1][c], k[father][c], k[mother][c], classes, parm->ml_min_k, parm->ml_max_k);

            // Perform BLX-alpha crossover for number of clusters
            BLXAlphaCrossover(&clusters[actual_size], &clusters[actual_size+1], &clusters[father], &clusters[mother], 1, parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters);

            // Perform permutation aware crossover of classifier chain
            OnePointPermutationCrossover(ccs[father], ccs[mother], ccs[actual_size], ccs[actual_size+1], classes);

            actual_size += 2;
        }

        // If apply evaluate all individuals
        int start = parm->population_size;
        if (parm->ml_external_eval_interval && ((g+1) % parm->ml_external_eval_interval == 0 || (g+1) % parm->ml_external_eval_interval == 1))
            start = 0;

        // Evaluate NEW individuals
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = start; i < actual_size; i++) {
            if (parm->ml_ins_sel_per_label)
                for (int j = 0; j < classes; j++)
                    ss[i][j] = population[i]+(j*instances);

            fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i], neighbors_cache, &ht,
                                                   NULL, index, &performance[i], parm->ml_ins_sel_per_label ? ss[i] : NULL, NULL, method, &(learner[i]), perf, g, ccs[i],
                                                   k[i][0][0], k[i][0], clusters[i], NULL, &(ensemble[i]), instances);
        }

        ///////////////////////////////////////////
        // Mutation operations ////////////////////
        ///////////////////////////////////////////
        // Instance selection for a single label for partial selection
        if (parm->ml_ins_sel_per_label && parm->partial_is_mutation > ZERO)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->partial_is_mutation) {
                    for (int j = 0; j < classes; j++)
                        ss[i][j] = population[i]+(j*instances);

                    PartialInstanceSelectionOneLabelOptimization(parm, parm->ml_ins_sel_per_label ? ss[i] : NULL, instances, data, inputs, classes, labels, n_labels, index, neighbors_cache, method,
                            &(learner[i]), ccs[i], k[i][0][0], k[i][0], clusters[i]);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i],
                                                           neighbors_cache, &ht, NULL, index, &performance[i], parm->ml_ins_sel_per_label ? ss[i] : NULL, NULL, method, &(learner[i]), perf,
                                                           g, ccs[i], k[i][0][0], k[i][0], clusters[i], NULL, &(ensemble[i]), instances);
                }

        // Mutation for partial selection
        if (parm->ml_ins_sel_per_label && parm->partial_selection_mutation > ZERO)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->partial_selection_mutation) {
                    for (int j = 0; j < classes; j++)
                        ss[i][j] = population[i]+(j*instances);

                    PartialSelectionMutation(parm, parm->ml_ins_sel_per_label ? ss[i] : NULL, labels, n_labels, classes, instances, neighbors_cache, &ht, index, inputs, data, &fitness[i], method,
                                             &(learner[i]), ccs[i], k[i][0][0], k[i][0], clusters[i]);
                }

        // Random mutation
        if (parm->ml_chc_random_mut > 0.0)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            // If multi-selection also mutate k
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_chc_random_mut) {
                    RandomMLMutation(parm, population[i], instances, parm->bit_mutation, g, parm->generations, parm->multiselection, max_oversampling_ratio);

                    // Check if multi-selection is correct
                    if (parm->multiselection && parm->multiselect_only_minority_class)
                        for (int j = 0; j < instances; j++)
                            if (!min_samples[j])
                                population[i][j] = (population[i][j] >= 1);

                    RandomMLMutation(parm, k[i][0], 1, parm->bit_mutation, g, parm->generations, true, max_oversampling_ratio);
                    // Avoid k outside bounds
                    k[i][0][0] = AssignWithinBoundsInteger(parm->ml_min_k, k[i][0][0], parm->ml_max_k);

                    // Random mutation for the permutation of classes
                    RandomMutationForPermutations(ccs[i], classes);

                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i],
                                                           neighbors_cache, &ht, NULL, index, &performance[i], parm->ml_ins_sel_per_label ? ss[i] : NULL, NULL, method, &(learner[i]), perf,
                                                           g, ccs[i], k[i][0][0], k[i][0], clusters[i], NULL, &(ensemble[i]), instances);
                }

        // Multi-selection mutation
        if (parm->ml_multiselection_mutation_ratio > 0.0 && parm->multiselection)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_multiselection_mutation_ratio) {
                    MultiSelectionMutation(parm, population[i], data, instances, labels, n_labels, classes, neighbors_cache, method, &fitness[i], k[i], clusters[i], inputs, index, perf, ccs[i]);
                    // Check if multi-selection is correct
                    if (parm->multiselection && parm->multiselect_only_minority_class)
                        for (int j = 0; j < instances; j++)
                            if (!min_samples[j])
                                population[i][j] = (population[i][j] >= 1);
                }

        // Memetic mutation for k
        if (parm->ml_k_memetic_mutation_ratio > 0.0)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_k_memetic_mutation_ratio)
                    kMemeticMutation(parm, population[i], data, instances, labels, n_labels, classes, neighbors_cache, method, &fitness[i], k[i], clusters[i], inputs, index, perf, ccs[i]);

        // Mutation for k per label evolution
        if (k && parm->kperlabel_local_opt_ratio > ZERO)
            // If multi-selection also mutate k
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->kperlabel_local_opt_ratio) {
                    for (int j = 0; j < classes; j++)
                        ss[i][j] = population[i]+(j*instances);
                    GetKPerLabel(parm, instances, data, labels, n_labels, classes, inputs, index, ks, population[i]);
                    scalar previous = fitness[i];

                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i],
                                                           neighbors_cache, &ht, NULL, index, &performance[i], parm->ml_ins_sel_per_label ? ss[i] : NULL, NULL, method, &(learner[i]), perf,
                                                           g, ccs[i], k[i][0][0], k[i][0], clusters[i], NULL, &(ensemble[i]), instances);
                    loperk_inc += fitness[i] - previous;
                    loperk_count++;
                }

        // If return best performing one obtain best
        if (parm->ml_return_best_performing) {
            int best_index = -1;
            for (int i = 0; i < actual_size; i++)
                if (performance[i] > best_performance) {
                    best_index = i;
                    best_performance = performance[i];
                    best_fitness = fitness[i];
                }

            if (best_index != -1) {
                selected = 0;
                if (parm->ml_ins_sel_per_label) {
                    int next = 0;
                    for (int j = 0; j < classes; j++)
                        for (int i = 0; i < instances; i++)
                            selected += (sselection[j][i] = population[best_index][next++]);
                } else
                    for (int i = 0; i < instances; i++)
                        if ((selection[i] = population[best_index][i]))
                            selected++;

                ((t_Parameters *) parm)->mlknn_k = k[best_index][0][0];
                if (ks)
                    for (int i = 0; i < classes; i++)
                        ks[i] = k[best_index][0][i];
                if (ksc)
                    for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
                        for (int i = 0; i < classes; i++)
                            ksc[c][i] = k[best_index][c][i];

                ((t_Parameters *) parm)->lamlknn_n_clusters = clusters[best_index];

                if (chain)
                    for (int i = 0; i < classes; i++)
                        chain[i] = ccs[best_index][i];
            }
        }

        if (parm->show_progress > 1)
            kInfo(parm, k, actual_size, fitness);

        // Reinitialize or elitism
        if (actual_size == parm->population_size) {
            minimum_distance--;
            SHOWPROGRESS(parm->show_progress > 0, "No individuals generated, decreasing minimum distance to: %d\n", minimum_distance);
        } else
            for (int i = 0; i < parm->population_size; i++) {
                int order = OrderedSelection(fitness+i, 0, actual_size-i)+i;

                int *aux_pt = population[order];
                population[order] = population[i];
                population[i] = aux_pt;

                scalar aux_f = fitness[order];
                fitness[order] = fitness[i];
                fitness[i] = aux_f;

                aux_f = performance[order];
                performance[order] = performance[i];
                performance[i] = aux_f;

                int **aux_dpt = k[order];
                k[order] = k[i];
                k[i] = aux_dpt;

                int aux_i = clusters[order];
                clusters[order] = clusters[i];
                clusters[i] = aux_i;

                aux_pt = ccs[order];
                ccs[order] = ccs[i];
                ccs[i] = aux_pt;
            }

        // Improvement?
        if (fitness[0] > old_best) {
            old_best = fitness[0];
            gs_without_improvement = 0;
        } else
            gs_without_improvement++;

        // If minimum distance 0 reinitialize
        if (minimum_distance == 0 || (parm->reinitalize_after_gs_no_improvement && gs_without_improvement > parm->reinitalize_after_gs_no_improvement)) {
            old_best = -1.0;
            // Reinitialize populations excepting best one
            int best = OrderedSelection(fitness, 0, parm->population_size);

            for (int i = 0; i < parm->population_size; i++)
                // Reseed changing 35%
                if (i != best) {
                    if (parm->multiselection)
                        for (int k = 0; k < length; k++)
                            if (RANDOM(0.0, 1.0) >= 0.35)
                                population[i][k] = population[best][k];
                            else
                                population[i][k] = RANDOM(0.0, 1.0) < chc_init_prob ? RANDOM(parm->symmetric_selection ? -max_oversampling_ratio : parm->ml_min_k, max_oversampling_ratio) : 0;
                    else
                        for (int k = 0; k < length; k++)
                            population[i][k] = RANDOM(0.0, 1.0) < 0.35 ? (population[best][k]+1)%2 : population[best][k];

                    for (int j = 0; j < classes; j++)
                        k[i][0][j] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);

                    for (int j = 0; j < classes; j++)
                        ccs[i][j] = j;
                    RandomizeVector(ccs[i], classes);

                    if (parm->ml_ins_sel_per_label)
                        for (int j = 0; j < classes; j++)
                            ss[i][j] = population[i]+(j*instances);

                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i],
                                                           neighbors_cache, &ht, NULL, index, &performance[i], parm->ml_ins_sel_per_label ? ss[i] : NULL, NULL, method, &(learner[i]), perf,
                                                           g, ccs[i], k[i][0][0], k[i][0], clusters[i], NULL, &(ensemble[i]), instances);
                }

            minimum_distance = MAX(1, (int)((length*chc_init_prob)/parm->chc_minimum_distance_factor));
            SHOWPROGRESS(parm->show_progress > 0, "\nPopulation reinitialized at generation %d\n", g+1);
        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            switch (method) {
                case MLkNN:
                case LAMLkNN:
                case DWKNND:
                case DWKNNM:
                case DWKNNZ:
                    SHOWPROGRESS(parm->show_progress > 0, "%*d g's best P(MEAN), F(MEAN), k: %6.4lf(%6.4lf), %6.4lf(%6.4lf), %3d %s", WIDTH(parm->generations), g+1, performance[best],
                                 AverageScalarVector(performance, parm->population_size), fitness[best], AverageScalarVector(fitness, parm->population_size), k[best][0][0],
                                 parm->ml_external_eval_interval && ((g+1) % parm->ml_external_eval_interval) == 0 ? "*" : " ");
                    SHOWPROGRESS(method == LAMLkNN, ", clusters = %d", clusters[best]);
                    if (parm->ml_return_best_performing) {
                        SHOWPROGRESS(parm->show_progress > 0, "  BEST PERF, FITNESS, k: %6.4lf, %6.4lf, %3d ", best_performance, best_fitness, parm->mlknn_k);
                    } else
                        SHOWPROGRESS(parm->show_progress > 0, " (New = %*d) ", WIDTH(parm->population_size),  actual_size - parm->population_size);
                    SHOWPROGRESS(parm->show_progress > 0, "(Time = %4lds, %4lds)\n", time(NULL) - init_generation,  time(NULL) - init_time);
                    break;

                case LAMLksNN:
                    SHOWPROGRESS(parm->show_progress > 0, "%*d g's best PERF, FITNESS, k's: %6.4lf, %6.4lf\n", WIDTH(parm->generations), g+1, performance[best], fitness[best]);
                    break;

                case MLksNN:
                    SHOWPROGRESS(parm->show_progress > 0, "%*d g's best PERF, FITNESS, k's: %6.4lf, %6.4lf, {", WIDTH(parm->generations), g+1, performance[best], fitness[best]);
                    for (int i = 0; i < classes-1; i++)
                        SHOWPROGRESS(parm->show_progress > 0, "%3d, ", k[best][0][i]);
                    SHOWPROGRESS(parm->show_progress > 0, "%3d}\n", k[best][0][classes-1]);
                    break;

                case CC:
                case BR:
                case RAkEL_d:
                case RAkEL_o:
                    SHOWPROGRESS(parm->show_progress > 0, "%*d g's best P(MEAN), F(MEAN): %6.4lf(%6.4lf), %6.4lf(%6.4lf) %s", WIDTH(parm->generations), g+1, performance[best],
                                 AverageScalarVector(performance, parm->population_size), fitness[best], AverageScalarVector(fitness, parm->population_size),
                                 parm->ml_external_eval_interval && ((g+1) % parm->ml_external_eval_interval) == 0 ? "*" : " ");
                    SHOWPROGRESS(parm->show_progress > 0, " (New = %*d) ", WIDTH(parm->population_size),  actual_size - parm->population_size);
                    SHOWPROGRESS(parm->show_progress > 0, "(Time = %4lds, %4lds)\n", time(NULL) - init_generation,  time(NULL) - init_time);
                    break;

                default:
                    break;
            }

            VERBOSE(parm->show_progress > 1, parm->out_file, "%*d generation best individual PERFORMANCE, FITNESS(MEAN): %6.4lf, %6.4lf(%6.4lf)\n", WIDTH(parm->generations), g+1, fitness[best],
                    performance[best], AverageScalarVector(fitness, parm->population_size));
        }

        // Percentage of new individuals per generation
        new += actual_size - parm->population_size;

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > (time_t) parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);
            *generations_performed = g+1;
            break;
        }
    }

    // If 3 way, assign new labels
    if (parm->three_way_is) {
        int best = OrderedSelection(fitness, 0, parm->population_size);

        for (int j = 0; j < classes; j++)
            ss[best][j] = population[best]+(j*instances);

        CheatLabels(labels, n_labels, labels, n_labels, instances, classes, ss[best], true, parm->cheat_neg_to_pos_only);

        // Save summary of results
        scalar removed = 0.0, changed = 0.0, kept = 0.0;
        for (int i = 0; i < classes; i++)
            for (int j = 0; j < instances; j++)
                if (!ss[best][i][j])
                    removed++;
                else if (ss[best][i][j] == -1)
                    changed++;
                else
                    kept++;

        fprintf(parm->out_file, "@3way removed labels: %f\n", removed/(instances*classes));
        fprintf(parm->out_file, "@3way changed labels: %f\n", changed/(instances*classes));
        fprintf(parm->out_file, "@3way kept labels: %f\n", kept/(instances*classes));
    }

    fprintf(parm->out_file, "@Percentage of new individuals per generation: %lf\n", new/(*generations_performed*parm->population_size));

    // If external fitness evaluator recalculate fitness
    if (parm->ml_external_eval_interval && !parm->ml_ins_sel_per_label) {
        // Average differential fitness per k value
        int *c;
        scalar *average;
        CALLOC(c, int, parm->ml_max_k+1, "CHCInstanceSelectionForML: c");
        CALLOC(average, scalar, parm->ml_max_k+1, "CHCInstanceSelectionForML: average");

        // Evaluate NEW individuals
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < parm->population_size; i++) {
            scalar f = fitness[i];
            fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : population[i], neighbors_cache, &ht,
                                                   NULL, index, &performance[i], NULL, NULL, method, NULL, perf, parm->ml_external_eval_interval-1, ccs[i], k[i][0][0], k[i][0], clusters[i], NULL,
                                                   &(ensemble[i]), instances);
            average[k[i][0][0]] += fabs(f - fitness[i]);
            c[k[i][0][0]]++;
        }

        fprintf(parm->out_file, "Different in fitness between ML-kNN and classifier. k, difference, count:\n");
        for (int i = parm->ml_min_k; i < parm->ml_max_k+1; i++)
            fprintf(parm->out_file, "[%2d, %6.4lf, %3d]\n", i, c[i] > 0 ? average[i]/c[i] : 0.0, c[i]);

        FREE(c);
        FREE(average);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
    SHOWPROGRESS(parm->show_progress > 0 && k && loperk_count > 0, "Average increment for k per label local optimization: %6.4lf\n", loperk_inc/loperk_count);
    VERBOSE(parm->show_progress > 0 && k && loperk_count > 0, parm->out_file, "Average increment for k per label local optimization: %6.4lf\n", loperk_inc/loperk_count);

    if (!parm->ml_return_best_performing) {
        int best = OrderedSelection(fitness, 0, parm->population_size);

        selected = 0;
        if (parm->ml_ins_sel_per_label) {
            for (int j = 0; j < classes; j++)
                for (int i = 0; i < instances; i++)
                    selected += (sselection[j][i] = population[best][j*instances+i]);
        } else
            for (int i = 0; i < instances; i++)
                if ((selection[i] = population[best][i]))
                    selected++;

        ((t_Parameters *) parm)->mlknn_k = k[best][0][0];
        if (ks)
            for (int i = 0; i < classes; i++)
                ks[i] = k[best][0][i];

        if (ksc)
            for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
                for (int i = 0; i < classes; i++)
                    ksc[c][i] = k[best][c][i];

        ((t_Parameters *) parm)->lamlknn_n_clusters = clusters[best];

        if (chain) {
            fprintf(parm->out_file, "Best chain:");
            for (int i = 0; i < classes; i++) {
                chain[i] = ccs[best][i];
                fprintf(parm->out_file, " %d", chain[i]);
            }
            fprintf(parm->out_file, "\n");
        }

        scalar perf;
        if (parm->show_progress > 1) {
            perf = EvaluateClassifierPerformance(parm, method, &(learner[0]), parm->ml_ins_sel_per_label ? NULL : selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data,
                                                 index, inputs, sselection, method == MLkNN, chain, parm->mlknn_k, ks, parm->lamlknn_n_clusters, NULL, &(ensemble[0]));
            SHOWPROGRESS(true, "FINAL BEST Performance (%s): %lf\n", parm->ml_loo ? "Leave-one-out" : "Resubstitution", perf);
            if (parm->ml_loo) {
                perf = EvaluateClassifierPerformance(parm, method, &(learner[0]), parm->ml_ins_sel_per_label ? NULL : selection, NULL, NULL, instances, labels, n_labels, classes, data, index,
                                                     inputs, sselection, false, chain, parm->mlknn_k, ks, parm->lamlknn_n_clusters, NULL, &(ensemble[0]));
                SHOWPROGRESS(true, "FINAL BEST Performance (Resubstitution): %lf\n", perf);
            }
        }
    }

    switch (method) {
        case MLksNN:
            fprintf(parm->out_file, "@CHC average best ML-k'sNN k's: %lf\nk's = {", AverageIntegerVector(ks, classes));
            for (int i = 0; i < classes-1; i++)
                fprintf(parm->out_file, "%3d, ", ks[i]);
            fprintf(parm->out_file, "%3d}\n", ks[classes-1]);
            break;

        case MLkNN:
            fprintf(parm->out_file, "@CHC best ML-kNN k: %d\n", parm->mlknn_k);
            break;

        case LAMLkNN:
            fprintf(parm->out_file, "@CHC best LAML-kNN k: %d\n", parm->mlknn_k);
            break;

        default:
            break;
    }

    // If per label selection show info if symmetric selection
    if (parm->ml_ins_sel_per_label && parm->symmetric_selection) {
        int over_selected = 0, s[classes];
        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < classes; j++)
                s[j] = sselection[j][i];

            for (int j = 0; j < n_labels[i]; j++)
                s[labels[i][j]] = 0;

            for (int j = 0; j < classes; j++)
                if (s[j] == -1)
                    over_selected++;
        }

        fprintf(parm->out_file, "@Over-selection percentage: %6.4lf\n", ((scalar) over_selected)/(instances*classes));
    }

    // If per label multi-selection show info
    if (parm->ml_ins_sel_per_label && parm->multiselection) {
        scalar count[parm->ml_max_k+1];
        ZeroVector(count, NULL, parm->ml_max_k+1);

        for (int i = 0; i < classes; i++)
            for (int j = 0; j < instances; j++)
                count[sselection[i][j]]++;

        count[0] /= instances*classes;
        fprintf(parm->out_file, "Selected %2d times percentage: %8.6lf (%8.6lf)\n", 0, count[0], count[0]);
        for (int i = 1; i <= MIN(parm->ml_max_k, parm->ml_max_ioversampling_ratio); i++) {
            count[i] /= instances*classes;
            fprintf(parm->out_file, "Selected %2d times percentage: %8.6lf (%8.6lf)\n", i, count[i], count[i-1]+count[i]);
            count[i] += count[i-1];
        }
    }

#ifdef VERBOSE_MODE
    SHOWPROGRESS(true, "Classifier chain%c", ':');
    for (int i = 0; i < classes; i++)
        SHOWPROGRESS(true, " %d", chain[i]);
    SHOWPROGRESS(true, "%c", '\n');
#endif

    // Free memory
    for (int i = 0; i < 2*parm->population_size; i++)
        FREE(population[i]);
    FREE(population);
    FREE(fitness);
    FREE(list);
    FREE(performance);
    if (parm->ml_use_hash_table)
        DestroyHashTable(&ht);
    Free3DIntegerMatrix(&k, 2*parm->population_size, parm->lamlknn_max_n_clusters);
    FREE(clusters);
    if (neighbors_cache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }
    if (distance_cache) {
        for (int i = 0; i < instances; i++)
            FREE(distance_cache[i]);
        FREE(distance_cache);
    }

    if (parm->ml_ins_sel_per_label) {
        for (int i = 0; i < 2*parm->population_size; i++)
            FREE(ss[i]);
        FREE(ss);
    }

    FREE(min_samples);

    // If symmetric multi-selection remove negative values
    if (parm->multiselection && parm->symmetric_selection)
        for (int i = 0; i < instances; i++)
            selection[i] = selection[i] > 0 ? selection[i] : 0;

    return selected;
}

bool RandomMLMutation(const t_Parameters *parm, int *individual, int size, scalar bit_mutation, int generation, int max_generations, bool multiselection, int max_oversampling_ratio)
{
    bool modified = false;

    if (!multiselection) {
        // Mutation just flips bit
        for (int j = 0; j < size; j++)
            if (RANDOM(0.0, 1.0) < bit_mutation) {
                modified = true;

                individual[j] = !individual[j];
            }
    } else {
        modified = true;
        NonUniformMutation(individual, size, generation,  max_generations, bit_mutation, parm->symmetric_selection ? -max_oversampling_ratio : 0, max_oversampling_ratio);
    }

    return modified;
}

// Random mutation for individuals representing permutations

void RandomMutationForPermutations(int *individual, int lenght)
{
    int f, s;

    f = RANDOM(0, lenght);
    s = RANDOM(0, lenght);

    int  swap = individual[f];
    individual[f] = individual[s];
    individual[s] = swap;
}


// Test the performance of every individual modifying N -> 1

void MultiSelectionMutation(const t_Parameters *parm, int *individual, scalar **data, int instances, int **labels, int *n_labels, int classes, cache_t **neighbors_cache, t_MLClassifier method,
                            scalar *c_fitness, int **k, int clusters, int inputs, int **index, t_Performance *perf, int *chain)
{
    scalar fitness = 0.0, performance;

    // For evaluations of individuals
    t_MLLearner learner;
    t_MLEnsemble ensemble;

    SHOWPROGRESS(parm->show_progress > 1, "Multi-selection mutation fitness: %6.4lf ---> ", *c_fitness);

    for (int i = 0; i < instances; i++)
        if (individual[i] > 1) {
            int old = individual[i];
            individual[i] = 1;

            fitness = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, individual, neighbors_cache, NULL, NULL, index, &performance, NULL, NULL,
                                                method, &learner, perf, 0, chain, k[0][0], k[0], clusters, NULL, &ensemble, instances);

            if (fitness > *c_fitness)
                *c_fitness = fitness;
            else
                individual[i] = old;
        }

    SHOWPROGRESS(parm->show_progress > 1, "%6.4lf\n", *c_fitness);

    FreeMLClassifier(parm, &learner, method, classes, inputs, instances, &ensemble);
}

void GetPerformanceWithAllInstances(const t_Parameters *parm, t_MLClassifier method, int *n_labels, int instances, int k, scalar **data, int **index, int inputs, int **labels,  int classes,
                                    cache_t **neighbors_cache, t_Performance *performance, int *ks, int n_clusters)
{
    bool classification, ranks;
    scalar **f;
    int **h, *n_h;

    t_MLLearner learner;
    t_MLEnsemble ensemble;

    AllocateScalarMatrix(&f, instances, classes);
    AllocateIntegerMatrix(&h, instances, classes);
    MALLOC(n_h, int, instances, "GetPerformanceWithAllInstances: n_h");

    TrainMultiLabelClassifier(parm, method, data, instances, labels, n_labels, classes, inputs, NULL, NULL, index, neighbors_cache, NULL, &learner, k, ks, NULL, n_clusters, NULL, &ensemble);
    ClassifyDataset(parm, method, data, h, f, &classification, &ranks, data, instances, labels, n_labels, classes, n_h, inputs, index, index, instances, NULL, neighbors_cache, NULL, &learner, NULL,
                    &ensemble);

    // Evaluate results
    PrintClassificationMetrics(performance, labels, n_labels, h, n_h, instances, classes, true, parm->out_file, "", false);
    PrintRankingMetrics(performance, labels, n_labels, f, instances, classes, parm->out_file, "", false);

    FreeScalarMatrix(&f, instances);
    FreeIntegerMatrix(&h, instances);
    FREE(n_h);
    FreeMLClassifier(parm, &learner, method, classes, inputs, instances, &ensemble);
}


void kMemeticMutation(const t_Parameters *parm, int *individual, scalar **data, int instances, int **labels, int *n_labels, int classes, cache_t **neighbors_cache, t_MLClassifier method,
                      scalar *c_fitness, int **ks, int clusters, int inputs, int **index, t_Performance *perf, int *chain)
{
    scalar fitness = 0.0, performance;
    int **ss = NULL;

    // For evaluations of individuals
    t_MLLearner learner;
    t_MLEnsemble ensemble;

    SHOWPROGRESS(parm->show_progress > 1, "k memetic mutation fitness: k = %d, %6.4lf ---> ", ks[0][0], *c_fitness);

    if (parm->ml_ins_sel_per_label)
        MALLOC(ss, int *, classes, "kMemeticMutation: ss");

    for (int k = parm->ml_min_k; k <= parm->ml_max_k; k++) {
        // Evaluate modified individual
        if (parm->ml_ins_sel_per_label)
            for (int j = 0; j < classes; j++)
                ss[j] = individual+(j*instances);
        fitness = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, parm->ml_ins_sel_per_label ? NULL : individual, neighbors_cache, NULL, NULL,
                                            index, &performance, ss, NULL, method, &learner, perf, 0, chain, ks[0][0], ks[0], clusters, NULL, &ensemble, instances);

        if (fitness > *c_fitness) {
            *c_fitness = fitness;
            ks[0][0] = k;
        } else if (fitness == *c_fitness)
            ks[0][0] = RANDOM(0.0, 1.0) < 0.5 ? ks[0][0] : k;
    }

    SHOWPROGRESS(parm->show_progress > 1, "k = %d, %6.4lf\n", ks[0][0], *c_fitness);

    FreeMLClassifier(parm, &learner, method, classes, inputs, instances, &ensemble);
    if (parm->ml_ins_sel_per_label)
        FREE(ss);
}

// Show k distribution
void kInfo(const t_Parameters *parm, int ***k, int size, scalar *fitness)
{
    scalar *c, *f;
    CALLOC(c, scalar, parm->ml_max_k+1, "kInfo: c");
    CALLOC(f, scalar, parm->ml_max_k+1, "kInfo: f");

    for (int i = 0; i < size; i++) {
        c[k[i][0][0]]++;
        f[k[i][0][0]] += fitness[i];
    }

    for (int i = parm->ml_min_k; i <= parm->ml_max_k; i++)
        SHOWPROGRESS(true, "[%2d, %4.2lf, %6.4lf] ", i, c[i]/size, c[i] > 0.0 ? f[i]/c[i] : 0.0);
    SHOWPROGRESS(true, "%c", '\n');

    FREE(c);
    FREE(f);
}

// One point crossover for permutations
// From: A. Andreica and C. Chira (2015) "Best-order crossover for permutation-base evolutionary algorithms," Applied Intelligence 42:751 - 776.

void OnePointPermutationCrossover(int *father, int *mother, int *son, int *daughter, int lenght)
{
    int q = RANDOM(0, lenght-1);

    // Copy first q to both siblings
    for (int i = 0; i < q; i++) {
        son[i] = father[i];
        daughter[i] = mother[i];
    }

    // Copy remaining not yet used for the other parent
    int next = 0;
    for (int i = q; i < lenght; i++) {
        while (LabelInListUnsorted(mother[next], son, i))
            next++;
        son[i] = mother[next++];
    }

    next = 0;
    for (int i = q; i < lenght; i++) {
        while (LabelInListUnsorted(father[next], daughter, i))
            next++;
        daughter[i] = father[next++];
    }
}
