/***************************************************************************
 *   Copyright (C) 2021 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "typex.h"
#include "ml.h"
#include "maths.h"

// Functions to deal with class-imbalance datasets for ML problems

// Stub for ML class-imbalanced methods

void ClassImbalanceMethodMultiLabelAlgorithm(const t_Parameters *parm, int *instances, scalar ***data, int ***labels, int **n_labels, int classes, int inputs, int ***index, scalar *IRLbl,
                                             scalar *MeanIR, int *lcount, scalar *CVIR, int *clabel, int *subspace, int ***ss, int *generations)
{
    extern char *ml_class_imbalance_method[];
    int evaluations;

    int previous_instances = *instances;

    switch (parm->ml_imbalance_method) {
        case MLNoneS:
            // Do nothing
            break;

        case MLRandomOS:
            MLRandomOversampling(parm, instances, data, labels, n_labels, classes, inputs, index, IRLbl, MeanIR, parm->ml_ros_p, lcount);
            break;

        case MLSmote:
            MultiLabelSmote(parm, instances, data, labels, n_labels, classes, inputs, index, IRLbl, MeanIR, CVIR, clabel, subspace);
            break;

        case MLESO:
            MultiLabelEvolutionarySyntheticOversampling(parm, data, instances, index, inputs, classes, labels, subspace, &evaluations, generations, n_labels, IRLbl, *MeanIR);
            break;

        case MLRandomUS:
            MLRandomUndersampling(parm, instances, data, labels, n_labels, classes, inputs, index, IRLbl, MeanIR, parm->ml_rus_p);
            break;

        case PRUS:
            AllocateIntegerMatrix(ss, classes, *instances);
            FillMatrix(NULL, *ss, classes, *instances, 1.0);
            PartialRandomUndersampling(parm, *labels, *n_labels, classes, *instances, *ss, parm->ml_rus_p, true);
            break;

        case PROS:
            AllocateIntegerMatrix(ss, classes, *instances);
            FillMatrix(NULL, *ss, classes, *instances, 1.0);
            PartialRandomOversampling(parm, labels, n_labels, classes, instances, ss, parm->ml_ros_p, data, index, inputs);
            break;

        case PRUOS:
            AllocateIntegerMatrix(ss, classes, *instances);
            FillMatrix(NULL, *ss, classes, *instances, 1.0);
            PartialRandomUndersampling(parm, *labels, *n_labels, classes, *instances, *ss, parm->ml_rus_p, true);
            PartialRandomOversampling(parm, labels, n_labels, classes, instances, ss, parm->ml_ros_p, data, index, inputs);
            break;

        case PROUS:
            AllocateIntegerMatrix(ss, classes, *instances);
            FillMatrix(NULL, *ss, classes, *instances, 1.0);
            PartialRandomOversampling(parm, labels, n_labels, classes, instances, ss, parm->ml_ros_p, data, index, inputs);
            PartialRandomUndersampling(parm, *labels, *n_labels, classes, *instances, *ss, parm->ml_rus_p, true);
            break;

        case PEVUS:
            AllocateIntegerMatrix(ss, classes, *instances);
            PartialEvolutionaryUndersampling(parm, labels, n_labels, classes, instances, *ss, data, index);
            break;

        case REPOPLABELS:
            EvolutionaryLabelRepopulation(parm, *data, *instances, *index, inputs, classes, *labels, subspace, &evaluations, generations, *n_labels);
            break;
    }

    fprintf(parm->out_file, "CI %s method: %d -> %d instances\n", ml_class_imbalance_method[(int) parm->ml_imbalance_method], previous_instances, *instances);
    fprintf(parm->out_file, "@CI reduction: %lf\n",  1.0 - ((scalar) *instances)/previous_instances);

    if (parm->ml_imbalance_method == PRUS || parm->ml_imbalance_method == PEVUS) {
        scalar per_label_red = 0.0;
        for (int i = 0; i < classes; i++)
            for (int j = 0; j < *instances; j++)
                if (!(*ss)[i][j])
                    per_label_red++;

        fprintf(parm->out_file, "@CI per label reduction: %lf\n", per_label_red/(*instances*classes));
    }
}

void ClassImbalanceMethodMultiLabel(const t_Parameters *parm, int *instances, scalar ***data, int ***labels, int **n_labels, int classes, int inputs, int ***index, scalar *IRLbl, scalar *MeanIR,
                                      int *lcount, scalar *CVIR, int *clabel, int *subspace, int ***ss)
{
    extern char *ml_class_imbalance_method[];
    int old_instances = *instances, generations;

    // Data partition
    if (parm->data_partition_method == NoDataPartition || !parm->subset_size || *instances < parm->subset_size) {
        ClassImbalanceMethodMultiLabelAlgorithm(parm, instances, data, labels, n_labels, classes, inputs, index, IRLbl, MeanIR, lcount, CVIR, clabel, subspace, ss, &generations);
        SAVEPROGRESS(parm->ml_imbalance_method == MLESO, "@Actual generations performed: %d\n", generations);
    } else {
        if (parm->ml_imbalance_method == PRUS || parm->ml_imbalance_method == PROS)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Data partition for PRUS/PROS");

        generations = 0;

        int n_subsets = *instances%parm->subset_size ? *instances/parm->subset_size+1 : *instances/parm->subset_size;

        SHOWPROGRESS(parm->show_progress > 0, "Data partition into %d subsets of size %d\n", n_subsets, parm->subset_size);
        SAVEPROGRESS(parm->show_progress > 0, "Data partition into %d subsets of size %d\n", n_subsets, parm->subset_size);

        scalar ***subset_data;
        int ***subset_index = NULL, *subset_instances, ***subset_labels, **subset_list, **subset_n_labels;

        MALLOC(subset_data, scalar **, n_subsets, "ClassImbalanceMethodMultiLabel: subset_data");
        MALLOC(subset_index, int **, n_subsets, "ClassImbalanceMethodMultiLabel: subset_index");
        MALLOC(subset_instances, int, n_subsets, "ClassImbalanceMethodMultiLabel: subset_list");
        MALLOC(subset_labels, int **, n_subsets, "ClassImbalanceMethodMultiLabel: subset_labels");
        MALLOC(subset_n_labels, int *, n_subsets, "ClassImbalanceMethodMultiLabel: subset_n_labels");
        MALLOC(subset_list, int *, n_subsets, "ClassImbalanceMethodMultiLabel: subset_list");
        for (int i = 0; i < n_subsets; i++) {
            MALLOC(subset_data[i], scalar *, parm->subset_size, "ClassImbalanceMethodMultiLabel: subset_data[i]");
            if (*index) {
                MALLOC(subset_index[i], int *, parm->subset_size, "ClassImbalanceMethodMultiLabel: subset_index[i]");
            } else
                subset_index[i] = NULL;
            MALLOC(subset_labels[i], int *, parm->subset_size, "ClassImbalanceMethodMultiLabel: subset_labels[i]");
            MALLOC(subset_n_labels[i], int, parm->subset_size, "ClassImbalanceMethodMultiLabel: subset_n_labels[i]");
            MALLOC(subset_list[i], int, parm->subset_size, "ClassImbalanceMethodMultiLabel: subset_labels[i]");
        }

        DataPartitionML(parm, *data, *instances, *labels, &n_subsets, &subset_data, &subset_instances, &subset_labels, &subset_list, *index, &subset_index, parm->subset_size, *n_labels,
                        &subset_n_labels, inputs);

#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < n_subsets; i++) {
            int subset_generations;
            ClassImbalanceMethodMultiLabelAlgorithm(parm, &(subset_instances[i]), &(subset_data[i]), &(subset_labels[i]), &(subset_n_labels[i]), classes, inputs, &(subset_index[i]), IRLbl, MeanIR,
                                                    lcount, CVIR, clabel, subspace, ss, &subset_generations);
#ifdef _OPENMP
            #pragma omp critical
#endif
            generations += subset_generations;
        }

        SAVEPROGRESS(parm->ml_imbalance_method == MLESO, "@Actual generations performed: %d\n", generations/n_subsets);

        // Aggregate selection
        int n_instances = 0;
        for (int i = 0; i < n_subsets; i++)
            n_instances += subset_instances[i];

        REALLOC(*data, scalar *, n_instances, "ClassImbalanceMethodMultiLabel: *data");
        REALLOC(*labels, int *, n_instances, "ClassImbalanceMethodMultiLabel: *labels");
        REALLOC(*n_labels, int, n_instances, "ClassImbalanceMethodMultiLabel: *n_labels");
        if (*index)
            REALLOC(*index, int *, n_instances, "ClassImbalanceMethodMultiLabel: *index");
        *instances = 0;
        for (int i = 0; i < n_subsets; i++) {
            for (int j = 0; j < subset_instances[i]; j++) {
                (*data)[*instances] = subset_data[i][j];
                (*labels)[*instances] = subset_labels[i][j];
                (*n_labels)[*instances] = subset_n_labels[i][j];
                if (*index)
                    (*index)[*instances] = subset_index[i][j];
                (*instances)++;
            }
        }

        for (int i = 0; i < n_subsets; i++) {
            FREE(subset_data[i]);
            FREE(subset_labels[i]);
            FREE(subset_list[i]);
            FREE(subset_n_labels[i]);
            if (*index)
                FREE(subset_index[i]);
        }

        FREE(subset_data);
        if (*index)
            FREE(subset_index);
        FREE(subset_instances);
        FREE(subset_labels);
        FREE(subset_n_labels);
        FREE(subset_list);
    }

    SAVEPROGRESS(true, "%s CI method %s %d instances\n", ml_class_imbalance_method[parm->ml_imbalance_method], *instances > old_instances ? "ADDED" : "REMOVED", abs(*instances -old_instances))
}
