/***************************************************************************
 *   Copyright (C) 2023 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdbool.h>
#include "typex.h"
#include "ml.h"
#include "maths.h"

// Partial undersampling: Every label is undersampled until ratio

void PartialRandomUndersampling(const t_Parameters *parm, int **labels, int *n_labels, int classes, int instances, int **prus, scalar ratio, bool info)
{
    int *maj_list, maj_count, *count_r, *count_i;
    bool *minority_relevant;

    MALLOC(maj_list, int, instances, "PartialRandomUndersampling: maj_list");
    MALLOC(minority_relevant, bool, classes, "PartialRandomUndersampling: minority_relevant");
    MALLOC(count_r, int, classes, "PartialRandomUndersampling: count_r");
    MALLOC(count_i, int, classes, "PartialRandomUndersampling: count_i");

    // Count labels to get minority for every label
    ZeroVector(NULL, count_r, classes);
    ZeroVector(NULL, count_i, classes);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < classes; j++)
            if (prus[j][i]) {
                if (LabelInList(j, labels[i], n_labels[i]))
                    count_r[j]++;
                else
                    count_i[j]++;
            }

    for (int i = 0; i < classes; i++)
        minority_relevant[i] = (count_r[i] < count_i[i]);

    for (int i = 0; i < classes; i++) {
        maj_count = 0;

        for (int j = 0; j < instances; j++)
            if (prus[i][j] && minority_relevant[i] == !LabelInList(i, labels[j], n_labels[j]))
                maj_list[maj_count++] = j;

        // Randomize majority list
        RandomizeVector(maj_list, maj_count);

        // Under-sample majority class
        int start;

        if (!parm->prus_relative_ratio)
            start = MIN(count_r[i], count_i[i])*ratio;
        else {
            start = MAX(count_r[i], count_i[i])*(1.0-ratio);
            start = MAX(start, MIN(count_r[i], count_i[i]));    // Limit is 1:1
        }

        for (int j = start; j < maj_count; j++)
            prus[i][maj_list[j]] = 0;
    }

    // Under-sampling info
    if (info) {
        fprintf(parm->out_file,"Majority class UNDER-sampling: ");
        for (int i = 0; i < classes; i++) {
            int relevant = 0, irrelevant = 0;
            for (int j = 0; j < instances; j++)
                if (prus[i][j]) {
                    if (LabelInList(i, labels[j], n_labels[j]))
                        relevant++;
                    else
                        irrelevant++;
                }

            fprintf(parm->out_file, "(%d,%d)->(%d,%d) ", count_r[i], count_i[i], relevant, irrelevant);
        }

        fprintf(parm->out_file, "\n");
    }

    FREE(minority_relevant);
    FREE(count_r);
    FREE(count_i);
    FREE(maj_list);
}

// Partial oversampling: Every label is oversampled until ratio

void PartialRandomOversampling(const t_Parameters *parm, int ***labels, int **n_labels, int classes, int *instances, int ***pros, scalar ratio, scalar ***data, int ***index, int inputs)
{
    int *min_list, min_count, *count_i, *count_r, **previous;
    bool *minority_relevant;

    MALLOC(min_list, int, *instances, "PartialRandomOversampling: min_list");
    MALLOC(minority_relevant, bool, classes, "PartialRandomOversampling: minority_relevant");
    MALLOC(count_r, int, classes, "PartialRandomUndersampling: count_r");
    MALLOC(count_i, int, classes, "PartialRandomUndersampling: count_i");

    AllocateIntegerMatrix(&previous, classes, *instances);
    CopyIntegerMatrix(*pros, previous, classes, *instances);

    // Count labels to get minority for every label
    ZeroVector(NULL, count_r, classes);
    ZeroVector(NULL, count_i, classes);
    for (int i = 0; i < *instances; i++)
        for (int j = 0; j < classes; j++)
            if ((*pros)[j][i]) {
                if (LabelInList(j, (*labels)[i], (*n_labels)[i]))
                    count_r[j]++;
                else
                    count_i[j]++;
            }
    for (int i = 0; i < classes; i++)
        minority_relevant[i] = (count_r[i] < count_i[i]);

    // Count number of times every instance must be oversampled per label
    for (int i = 0; i < classes; i++) {
        min_count = 0;

        for (int j = 0; j < *instances; j++)
            if ((*pros)[i][j] && minority_relevant[i] == LabelInList(i, (*labels)[j], (*n_labels)[j]))
                min_list[min_count++] = j;

        // Randomize majority list
        RandomizeVector(min_list, min_count);

        // Initially unselect all
        FillVector(NULL, (*pros)[i], *instances, 0.0);

        // Over-sample minority class
        int nover;
        if (!parm->pros_relative_ratio)
            nover = (count_i[i]+count_r[i]-min_count)*ratio-min_count;
        else {
            nover = (*instances-min_count)*ratio;
            nover = MIN(nover, (*instances-min_count)-min_count);   // Limit is 1:1
        }

        if (min_count)
            for (int j = 0, k = 0; j < nover; j++, k = (k+1)%min_count)
                (*pros)[i][min_list[k]]++;
    }

    // Partial oversampling once count is made
    int original_instances = *instances;
    for (int i = 0; i < original_instances; i++) {
        int copies = (int) MaximumOfMatrixColumn(NULL, *pros, classes, i), c = 0;

        if (!copies)
            continue;

        // Count non-zero inputs if needed
        if (*index)
            for (int j = 0; (*index)[i][j] != -1; j++)
                c++;

        // Reallocate data
        REALLOC(*labels, int *, *instances+copies, "PartialRandomOversampling: *labels");
        for (int j = 0; j < classes; j++)
            REALLOC((*pros)[j], int, *instances+copies, "PartialRandomOversampling: *pros");
        REALLOC(*n_labels, int, *instances+copies, "PartialRandomOversampling: *n_labels");
        REALLOC(*data, scalar *, *instances+copies, "PartialRandomOversampling: *data");
        if (*index)
            REALLOC(*index, int *, *instances+copies, "PartialRandomOversampling: *index");

        for (int j = 0; j < copies; j++) {
            (*n_labels)[*instances+j] = (*n_labels)[i];
            MALLOC((*labels)[*instances+j], int, (*n_labels)[i], "PartialRandomOversampling: (*labels)[*instances+j]");
            for (int k = 0; k < (*n_labels)[i]; k++)
                (*labels)[*instances+j][k] = (*labels)[i][k];

            for (int k = 0; k  < classes; k++)
                (*pros)[k][*instances+j] = ((*pros)[k][i]-- > 0);

            if (*index) {
                MALLOC((*data)[*instances+j], scalar, c, "PartialRandomOversampling: (*labels)[*instances+j]");
                MALLOC((*index)[*instances+j], int, c+1, "PartialRandomOversampling: (*labels)[*instances+j]");
                 for (int k = 0; (*index)[i][k] != -1; k++) {
                     (*index)[*instances+j][k] = (*index)[i][k];
                     (*data)[*instances+j][k] = (*data)[i][k];
                }
                (*index)[*instances+j][c] = -1;
            } else {
                MALLOC((*data)[*instances+j], scalar, inputs, "PartialRandomOversampling: (*labels)[*instances+j]");
                memcpy((*data)[*instances+j], (*data)[i], inputs*sizeof(scalar));
            }
        }

        SHOWPROGRESS(parm->show_progress > 1, "Adding %*d copies of instance %*d\r", WIDTH(original_instances), copies, WIDTH(original_instances), i+1);

        *instances += copies;
    }

    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

    // Restore original instances
    CopyIntegerMatrix(previous, *pros, classes, original_instances);
    FreeIntegerMatrix(&previous, classes);

    // Over-sampling info
    fprintf(parm->out_file,"Minority class OVER-sampling: ");
    for (int i = 0; i < classes; i++) {
        int relevant = 0, irrelevant = 0;
        for (int j = 0; j < *instances; j++)
            if ((*pros)[i][j]) {
                if (LabelInList(i, (*labels)[j], (*n_labels)[j]))
                    relevant++;
                else
                    irrelevant++;
            }

        fprintf(parm->out_file, "(%d,%d)->(%d,%d) ", count_r[i], count_i[i], relevant, irrelevant);
    }

    fprintf(parm->out_file, "\n");

    FREE(minority_relevant);
    FREE(count_i);
    FREE(count_r);
    FREE(min_list);
}
