/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include "ml.h"
#include "macros.h"
#include "maths.h"

// Test all classifiers
void TestAllMultiLabelClassifiers(const t_Parameters *parm, scalar **test_data, int **test_labels, int *test_n_labels,  scalar **train_data, int train_instances, int **train_labels,
                                  int *train_n_labels, int n_labels, int inputs, int **test_index, int **train_index, int test_instances, int *selection, cache_t **neighbors_cache, int **SS,
                                  char *file, int k, int *ks, int **ksc, int *chain, int *subspace)
{
    scalar **f;
    int *n_h, **h, *lchain = NULL;
    bool classification, ranks;
    t_Performance performance;
    extern char *ml_method[];
    char message[MAX_LENGTH];

    t_MLLearner learner;
    t_MLEnsemble ensemble;

    // To avoid recursive loop
    learner.mlksnn.ks = NULL;

    // Classify testing set
    AllocateScalarMatrix(&f, parm->evaluate_train_error ? train_instances : test_instances, n_labels);
    AllocateIntegerMatrix(&h, parm->evaluate_train_error ? train_instances : test_instances, n_labels);
    MALLOC(n_h, int, parm->evaluate_train_error ? train_instances : test_instances, "TestAllMultiLabelClassifiers: n_h");
    if (!chain) {
        MALLOC(lchain, int, n_labels, "TrainMultiLabelClassifier: lchain");
    } else
        lchain = chain;

    for (t_MLClassifier method = 0; method < N_ML_CLASSIFIERS; method++) {
        time_t init_time = time(NULL);

        // Use this evaluator?
        if (!parm->evaluate_ml_with[method]) {
            SHOWPROGRESS(parm->show_progress > 1, "NOT evaluating %s\n", ml_method[method]);
            continue;
        }

        SHOWPROGRESS(parm->show_progress > 0, "Training and evaluating %s\n", ml_method[method]);

        // Cross-validate hyper-parameters
        if (parm->cv_ml_hyperparameters) {
            CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, method, train_data, train_instances, inputs, train_n_labels, train_labels, n_labels, train_index, selection, subspace,
                                                      parm->cv_folds);
            k = parm->mlknn_k;
        }

        TrainMultiLabelClassifier(parm, method, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, selection, lchain, train_index, neighbors_cache, SS,
                                  &learner, k, ks, ksc, parm->lamlknn_n_clusters, subspace, &ensemble);

        if (parm->evaluate_train_error) {
            ClassifyDataset(parm, method, train_data, h, f, &classification, &ranks, train_data, train_instances, train_labels, train_n_labels, n_labels, n_h, inputs, train_index, train_index,
                            train_instances, selection, neighbors_cache, SS, &learner, subspace, &ensemble);

            // Message for output
            sprintf(message, "Training %s", ml_method[method]);
            // Average number of predicted labels per instances
            scalar av_pred = 0.0, av_actual = 0.0;
            for (int i = 0; i < train_instances; i++) {
                av_pred += n_h[i];
                av_actual += train_n_labels[i];
            }
            fprintf(parm->out_file, "@%s average number of PREDICTED relevant labels: %.2lf\n", message, av_pred/train_instances);
            fprintf(parm->out_file, "@%s average number of ACTUAL relevant labels: %.2lf\n", message, av_actual/train_instances);

            // Evaluate results
            if (classification)
                PrintClassificationMetrics(&performance, train_labels, train_n_labels, h, n_h, train_instances, n_labels, true, parm->out_file, message, true);

            if (ranks)
                PrintRankingMetrics(&performance, train_labels, train_n_labels, f, train_instances, n_labels, parm->out_file, message, true);

            // Save actual and predicted outputs?
            if (parm->save_outputs) {
                char filename[MAX_FILENAME_LENGTH];
                sprintf(filename, "%s.training.%s.outputs", file, ml_method[method]);

                SaveMLOutputs(filename, train_instances, n_labels, h, n_h, train_labels, train_n_labels);
            }
        }

        ClassifyDataset(parm, method, test_data, h, f, &classification, &ranks, train_data, train_instances, train_labels, train_n_labels, n_labels, n_h, inputs, test_index, train_index,
                        test_instances, selection, neighbors_cache, SS, &learner, subspace, &ensemble);

        fprintf(parm->out_file, "%s evaluation time: %ld\n", ml_method[method], time(NULL) - init_time);

        // Message for output
        sprintf(message, "%s", ml_method[method]);

        // Average number of predicted labels per instances
        scalar av_pred = 0.0, av_actual = 0.0;
        for (int i = 0; i < test_instances; i++) {
            av_pred += n_h[i];
            av_actual += test_n_labels[i];
        }
        fprintf(parm->out_file, "@%s average number of PREDICTED relevant labels: %.2lf\n", message, av_pred/test_instances);
        fprintf(parm->out_file, "@%s average number of ACTUAL relevant labels: %.2lf\n", message, av_actual/test_instances);

        // Evaluate results
        if (classification)
            PrintClassificationMetrics(&performance, test_labels, test_n_labels, h, n_h, test_instances, n_labels, true, parm->out_file, message, true);

        if (ranks)
            PrintRankingMetrics(&performance, test_labels, test_n_labels, f, test_instances, n_labels, parm->out_file, message, true);

        // Save actual and predicted outputs?
        if (parm->save_outputs) {
            char filename[MAX_FILENAME_LENGTH];
            sprintf(filename, "%s.testing.%s.outputs", file, ml_method[method]);

            SaveMLOutputs(filename, test_instances, n_labels, h, n_h, test_labels, test_n_labels);
        }

        // Info of classifier
        if (method == MLDT)
            SAVEPROGRESS(true, "@ML-DT nodes: %d\n", MLDTCountNodes(&(learner.mldt)));

        if (method == OPCT)
            SAVEPROGRESS(true, "@OPCT nodes: %d\n", OptionMLDTCountNodes(&(learner.otree)));

        // Info of classifier
        if (method == MLRF)
            SAVEPROGRESS(true, "@ML-RF trees: %d\n@ML-RF total nodes: %d\n", learner.mlrf.size, MLRFCountNodes(&(learner.mlrf)));

        FreeMLClassifier(parm, &learner, method, n_labels, inputs, train_instances, &ensemble);

        // Save results so far to file
        fprintf(parm->out_file, "@ok\n");
        fflush(parm->out_file);
    }

    FreeIntegerMatrix(&h, test_instances);
    FreeScalarMatrix(&f, test_instances);
    if (!chain)
        FREE(lchain);
    FREE(n_h);
}

// Classify a query instance

// h(·) Multi label classifier: List of relevant labels for the query instance
// f(·, ·)  Real valued function, f(x, y) measures of y being a proper label for x

void ClassifyInstance(const t_Parameters *parm, t_MLClassifier ml_method, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances,
                      int **relevant_labels, int *relevant_n_labels, int classes, int *n_predicted, int inputs, int *query_index, int **index, int *S, cache_t **neighbors_cache,
                      int query_pos, int **SS, t_MLLearner *learner, int *subspace, int **labels, int *n_labels, t_MLEnsemble *ensemble)
{
    t_MLkNN mlknn;

    switch (ml_method) {
        case MLkNN:
            MultiLabelkNearestNeighbors(parm, query, data, instances, relevant_labels, relevant_n_labels, classes, h, n_predicted, S, f, inputs, query_index, index, neighbors_cache, query_pos,
                                        &(learner->mlknn), SS, NULL, NULL, -1, subspace);
            *classification = *ranks = true;
            break;

        case SMLKNN:
             MultiLabelStackedkNearestNeighbors(parm, query, data, instances, relevant_labels, relevant_n_labels, classes, h, n_predicted, S, f, inputs, query_index, index, neighbors_cache,
                                                query_pos, &(learner->smlknn), SS, subspace);
            *classification = *ranks = true;
            break;

        case BR:
            BinaryRelevanceClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->br), subspace, S, SS, index);
            *classification = *ranks = true;
            break;

        case DBR:
            DependentBinaryRelevanceClassifyInstance(parm, query, h, f,  classification, ranks, classes, n_predicted, inputs, query_index, &(learner->dbr), subspace, S, SS,
                    index);
            *classification = *ranks = true;
            break;

        case STA:
            MLStackingClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->dbr), subspace, S, SS, index);
            *classification = *ranks = true;
            break;

        case CC:
        case NS:
        case NSSC:
            ClassifierChainClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->cc), subspace, S, SS, index,
                                            ml_method == NSSC);
            *classification = *ranks = true;
            break;

        case ECC:
            ECCClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index,  &(learner->ecc), subspace, S, SS, index);
            *classification = *ranks = true;
            break;

        case LP:
        case PRUNEDSET:
        case PPT:
        case PPTN:
            LabelPowersetClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->lp), subspace, S, index);
            *classification = *ranks = true;
            break;

        case PPTC:
            PPTcClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->lp), subspace, S, index);
            *classification = *ranks = true;
            break;
            break;

        case CLR:
            CalibratedLabelRankingClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->clr), subspace, index);
            *classification = *ranks = true;
            break;

        case TSVM:
        case TSCCM:
        case TSPCCM:
            TwoStageVotingMethodClassifyInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->clr), subspace, index,
                                                 ml_method == TSCCM, ml_method == TSPCCM);
            *classification = *ranks = true;
            break;

        case RAkEL_d:
        case RAkEL_o:
            ClassifierRAkELInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->rakel), subspace, S, index);
            *classification = *ranks = true;
            break;

        case LAMLkNN:
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for LAML-kNN");
            LocallyAdaptiveMultiLabelkNearestNeighbors(parm, query, data, instances, relevant_labels, relevant_n_labels, classes, h, n_predicted, S, f, inputs, query_index, index, neighbors_cache,
                    query_pos, SS, &(learner->lamlknn), NULL);
            *classification = *ranks = true;
            break;

        case LAMLksNN:
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for LAML-ksNN");
            LocallyAdaptiveMultiLabelkNearestsNeighbors(parm, query, data, instances, relevant_labels, relevant_n_labels, classes, h, n_predicted, S, f, inputs, query_index, index, neighbors_cache,
                    query_pos, SS, &(learner->lamlksnn));
            *classification = *ranks = true;
            break;

        case MLksNN:
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for ML-ksNN");
            mlknn.prior = learner->mlknn.prior;
            mlknn.post_0 = learner->mlknn.post_0;
            mlknn.post_1 = learner->mlknn.post_1;

            MultiLabelkNearestNeighbors(parm, query, data, instances, relevant_labels, relevant_n_labels, classes, h, n_predicted, S, f, inputs, query_index, index, neighbors_cache, query_pos,
                                        &mlknn, SS, learner->mlksnn.ks, NULL, -1, NULL);
            *classification = *ranks = true;
            break;

        case HOMER_B:
        case HOMER_R:
        case HOMER_K:
            ClassifierHOMERInstance(parm, query, h, f,  classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->homer), subspace, S, index);
            *classification = *ranks = true;
            break;

        case MLDT:
        case PCT:
            MLDecisionTreeClassify(learner->mldt.nodes, query, f, classes, query_index, h, n_predicted, classification, ranks, learner->mldt.w);
            *classification = *ranks = true;
            break;

        case MLBASICBP:
            MLBasciBPClassifyInstance(&(learner->mlmlp), query, f, h, n_predicted, inputs, classes, subspace, classification, ranks);
            *classification = *ranks = true;
            break;

        case BPMLL:
            BPMLLClassifyInstance(&(learner->bpmll), query, f, h, n_predicted, inputs, classes, subspace, classification, ranks);
            *classification = *ranks = true;
            break;

        case MLRF:
        case EPCT:
        case RFPCT:
        case EDT:
        case EXTRAPCT:
        case EXTRAMLDT:
            MLRandomForestClassify(&(learner->mlrf), query, f, classes, query_index, h, n_predicted, classification, ranks, parm->mlrf_real_votes);
            *classification = *ranks = true;
            break;

        case IBLRML:
        case IBLRMLplus:
            IBLRMLClassifyInstance(query, h, f, classification, ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->iblr), subspace, S, SS, index, relevant_labels,
                                   relevant_n_labels, ml_method == IBLRMLplus);
            *classification = *ranks = true;
            break;

        case MLTREE:
            MLTreeClassifyInstance(parm, query, h, f,  classification,ranks, data, instances, classes, n_predicted, inputs, query_index, &(learner->mltree), subspace, S, SS, index);
            *classification = *ranks = true;
            break;

        case MLFOREST:
            MLForestClassify(parm, &(learner->forest), query, f, classes, query_index, h, n_predicted, classification, ranks, data, instances, inputs, subspace, index);
            *classification = *ranks = true;
            break;

        case MLRS:
            MLRSClassify(&(learner->mlrs), query, f, classes, index, h, n_predicted, classification, ranks, data, instances, inputs, query_index, subspace, S, labels, n_labels);
            *classification = *ranks = true;
            break;

        case MLRSLC:
            MLRSLCClassify(&(learner->mlrslc), query, f, classes, index, h, n_predicted, classification, ranks, data, instances, inputs, query_index, subspace, S, labels, n_labels);
            *classification = *ranks = true;
            break;

        case OPCT:
            ZeroVector(f, NULL, classes);
            OptionMLDecisionTreeClassify(learner->otree.root, query, f, classes, query_index, h, n_predicted, classification, ranks);
            *classification = *ranks = true;
            break;

        case DMLKNN:
            DMultiLabelkNearestNeighbors(parm, query, data, instances, labels, n_labels, classes, h, n_predicted, S, f, inputs, query_index, index, neighbors_cache, &(learner->dmlknn), SS, subspace);
            *classification = *ranks = true;
            break;

        case MrKNN:
            MrKNNClassify(&(learner->mrknn), query, query_index, data, instances, index, classes, S, subspace, inputs, neighbors_cache, f, h, n_predicted);
            *classification = *ranks = true;
            break;

        case BAGGING:
        case REALADABOOSTRE:
        case DISCRETEADABOOSTRE:
        case EPS:
            BaggingClassifyInstance(parm, ensemble, query, h, f, classification, ranks, data, instances, relevant_labels, relevant_n_labels, classes, n_predicted, inputs, query_index, index, S, SS,
                                    subspace, labels, n_labels);
            break;

        case REALADABOOSTMH:
        case DISCRETEADABOOSTMH:
            AdaBoostMHClassifyInstance(parm, query, query_index, ensemble, data, classes, inputs, index, subspace, S, SS, h, f, n_predicted, instances, classification, ranks);
            break;

        case DWKNND:
        case DWKNNM:
        case DWKNNZ:
            DWkNNClassify(ml_method, &(learner->dwknn), query, data, index, inputs, query_index, NULL, SS, S, relevant_labels, relevant_n_labels, classes, subspace, n_predicted, h, f, instances);
            *classification = *ranks = true;
            break;

        case FRONEC1H:
        case FRONEC2H:
        case FRONEC3H:
        case FRONEC1LD:
        case FRONEC2LD:
        case FRONEC3LD:
            FRONECClassifyInstance(&(learner->fronec), query, data, instances, index, inputs, query_index, S, neighbors_cache, SS, classes,  subspace, relevant_labels, relevant_n_labels, ml_method,
                                   f, n_predicted, h);
            *classification = *ranks = true;
            break;

        case FVkNN:
            FVkNNClassify(&(learner->fvknn), query, data, instances, index, query_index, inputs, S, SS, neighbors_cache, classes, subspace, f, n_predicted, h);
            *classification = *ranks = true;
            break;

        case ZEROML:
            // Predict always no relevant label
            *classification = true;
            *ranks = false;
            *n_predicted = 0;
            break;
    }
}

// Classify whole dataset
void ClassifyDataset(const t_Parameters *parm, t_MLClassifier ml_method, scalar **query, int **h, scalar **f,  bool *classification, bool *ranks, scalar **data, int instances,
                     int **relevant_labels, int *relevant_n_labels, int n_labels, int *n_predicted, int inputs, int **query_index, int **index, int query_instances, int *S,
                     cache_t **neighbors_cache, int **SS, t_MLLearner *learner, int *subspace, t_MLEnsemble *ensemble)
{
    // Instance selection per label with symmetric selection must rebuild label data
    int **local_labels = NULL, *local_n_labels = NULL;
    if (SS && parm->symmetric_selection) {
        int *labels;

        MALLOC(local_labels, int *, instances, "TrainMultiLabelClassifier: local_labels");
        MALLOC(local_n_labels, int, instances, "TrainMultiLabelClassifier: local_n_labels");
        MALLOC(labels, int, n_labels, "TrainMultiLabelClassifier: labels");

        for (int i = 0; i < instances; i++) {
            // If selection == -1 labels is set
            for (int j = 0; j < n_labels; j++)
                labels[j] = (SS[j][i] == -1);

            // Old set of labels
            for (int j = 0; j < relevant_n_labels[i]; j++)
                labels[relevant_labels[i][j]] = 1;

            int n = 0;
            for (int j = 0; j < n_labels; j++)
                if (labels[j])
                    n++;

            local_n_labels[i] = n;
            MALLOC(local_labels[i], int, n, "TrainMultiLabelClassifier: local_labels[i]");
            n = 0;
            for (int j = 0; j < n_labels; j++)
                if (labels[j]) {
                    local_labels[i][n] = j;
                    n++;
                }
        }

        FREE(labels);

        // Substitute old labels
        relevant_labels = local_labels;
        relevant_n_labels = local_n_labels;
    }
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < query_instances; i++) {
        ClassifyInstance(parm, ml_method, query[i], h[i], f[i],  classification, ranks, data, instances, relevant_labels, relevant_n_labels, n_labels, &n_predicted[i], inputs,
                         query_index ? query_index[i] : NULL, index, S, neighbors_cache, i, SS, learner, subspace, relevant_labels, relevant_n_labels, ensemble);
        SHOWPROGRESS(parm->show_progress > 2, "Classified instance: %*d/%d\r", WIDTH(query_instances), i+1, query_instances);
        if (parm->show_progress > 2) {
            int p = 0;

            SHOWPROGRESS(true, "%*d: ", WIDTH(query_instances), i+1);
            for (int j = 0; j < n_labels; j++)
                if (h[i][p] == j) {
                    SHOWPROGRESS(true, "1/(%6.4lf) ", f[i][j]);
                    p++;
                } else
                    SHOWPROGRESS(true, "0/(%6.4lf) ", f[i][j]);
            SHOWPROGRESS(true, "%c", '\n');
        }
    }

    SHOWPROGRESS(parm->show_progress > 2, "%c", '\n');

    if (SS && parm->symmetric_selection) {
        FREE(local_n_labels);
        for (int i = 0; i < instances; i++)
            FREE(local_labels[i]);
        FREE(local_labels);
    }
}

// Train multi-label classifier
// Hyper-parameters must be fixed

void TrainMultiLabelClassifier(const t_Parameters *parm, t_MLClassifier method, scalar **train_data, int train_instances, int **train_labels, int *train_n_labels, int n_labels, int inputs,
                               int *selection, int *chain, int **train_index, cache_t **neighbors_cache, int **SS, t_MLLearner *learner, int k, int *ks, int **ksc, int n_clusters,
                               int *subspace, t_MLEnsemble *ensemble)
{
    bool clustering = false, balanced = false, overlapping = true;


    // Instance selection per label with symmetric selection must rebuild label data
    int **local_labels = NULL, *local_n_labels = NULL;
    if (SS && parm->symmetric_selection) {
        int *labels;

        MALLOC(local_labels, int *, train_instances, "TrainMultiLabelClassifier: local_labels");
        MALLOC(local_n_labels, int, train_instances, "TrainMultiLabelClassifier: local_n_labels");
        MALLOC(labels, int, n_labels, "TrainMultiLabelClassifier: labels");

        for (int i = 0; i < train_instances; i++) {
            // If selection == -1 labels is set
            for (int j = 0; j < n_labels; j++)
                labels[j] = (SS[j][i] == -1);

            // Old set of labels
            for (int j = 0; j < train_n_labels[i]; j++)
                labels[train_labels[i][j]] = 1;

            int n = 0;
            for (int j = 0; j < n_labels; j++)
                if (labels[j])
                    n++;

            local_n_labels[i] = n;
            MALLOC(local_labels[i], int, n, "TrainMultiLabelClassifier: local_labels[i]");
            n = 0;
            for (int j = 0; j < n_labels; j++)
                if (labels[j]) {
                    local_labels[i][n] = j;
                    n++;
                }
        }

        FREE(labels);

        // Substitute old labels
        train_labels = local_labels;
        train_n_labels = local_n_labels;
    }

    switch (method) {
        case MLkNN:
            TrainMultiLabelkNearestNeighbors(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, selection, inputs, train_index, neighbors_cache, &(learner->mlknn),
                                             SS, k, true, subspace);
            break;

        case DMLKNN:
            TrainDMultiLabelkNearestNeighbors(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, selection, inputs, train_index, neighbors_cache, &(learner->dmlknn), SS,
                                              k, true, subspace);
            break;

        case SMLKNN:
            TrainStackedMLkNN(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, selection, inputs, train_index, neighbors_cache, &(learner->smlknn), SS, k, true, subspace);
            break;

        case LAMLkNN:
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for LAML-kNN");
            TrainLocallyAdaptiveMultiLabelkNearestNeighbors(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, selection, inputs, train_index, SS, &(learner->lamlknn),
                    neighbors_cache, k, n_clusters, true);
            break;

        case BR:
            TrainBinaryRelevanceModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->br), subspace, parm->cv_hyperparameters, parm->br_model,
                                       selection, SS, train_index);
            break;

        case DBR:
            TrainDependentBinaryRelevanceModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->dbr), subspace, parm->cv_hyperparameters,
                                                parm->dbr_model, selection, SS, train_index);
            break;

        case STA:
            TrainMLStackingModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->dbr), subspace, parm->cv_hyperparameters, parm->sta_model,
                                  selection, SS, train_index);
            break;

        case CC:
        case NS:
        case NSSC:
            TrainClassifierChainModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->cc), subspace, parm->cv_hyperparameters,
                                       method == CC ? parm->cc_model : parm->ns_model, selection, chain, SS, train_index, method == NS || method == NSSC);
#ifdef VERBOSE_MODE
            SHOWPROGRESS(true, "Classifier chain%c", ':');
            for (int i = 0; i < n_labels; i++)
                SHOWPROGRESS(true, " %d", ((t_CC *)classifier)->chain[i]);
            SHOWPROGRESS(true, "%c", '\n');
#endif
            break;

        case ECC:
            TrainECCModel(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->ecc), subspace, parm->cv_hyperparameters, parm->cc_model, selection, SS,
                          train_index);
            break;

        case LP:
        case PRUNEDSET:
        case PPT:
        case PPTN:
        case PPTC:
            TrainLabelPowersetModel(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->lp), subspace, parm->cv_hyperparameters,
                                    method == LP ? parm->lp_model : parm->ps_model, selection, SS, train_index, method == PRUNEDSET, method == PPT || (method == PPTC && !parm->pptc_no_loss),
                                    method == PPTN || (method == PPTC && parm->pptc_no_loss));
            break;

        case CLR:
        case TSVM:
        case TSCCM:
        case TSPCCM:
            TrainCalibratedLabelRankingModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->clr), subspace, parm->cv_hyperparameters,
                                              parm->clr_model, selection, SS, train_index, method == TSCCM, method == TSPCCM);
            break;

        case RAkEL_d:
            overlapping = false;
        case RAkEL_o:
            TrainRAkELModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->rakel), subspace, parm->cv_hyperparameters, parm->rakel_model,
                             selection, SS, train_index, overlapping);
            break;

        case HOMER_B:
            balanced = true;
        case HOMER_K:
            clustering = true;
        case HOMER_R:
            TrainHOMERModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->homer), subspace, parm->cv_hyperparameters, parm->homer_model,
                             selection, SS, train_index, clustering, balanced);
            break;

        case MLDT:
        case PCT:
            MLDecisionTreeTrain(parm, train_data, train_instances, train_labels, inputs, &(learner->mldt), selection, subspace, n_labels, train_index, train_n_labels, false, method == PCT, SS, false);
            break;

        case MLksNN:
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for ML-ksNN");
            TrainMultiLabelKsNearestNeighbors(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, selection, inputs, train_index, neighbors_cache, &(learner->mlksnn), SS,
                                              ks, true);
            break;

        case MLBASICBP:
            if (train_index)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for MLBASICBP");
            TrainMLBasicBP(parm, &(learner->mlmlp), train_data, train_instances, n_labels, inputs, train_labels, train_n_labels, subspace, selection);
            break;

        case BPMLL:
            TrainBPMLL(parm, &(learner->bpmll), train_data, train_instances, n_labels, inputs, train_labels, train_n_labels, subspace, selection);
            break;

        case LAMLksNN:
            if (subspace)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for LAML-ksNN");
            CreateLAMLksNNClassifier(parm, &(learner->lamlksnn), n_labels, ksc, train_data, train_instances, train_n_labels, train_labels, inputs, train_index, parm->lamlknn_n_clusters);

            if (parm->ml_kperlabel_method == LAMLksNNEvolutionary || parm->lamlksnn_is)
                EvolveAndTrainLAMLksNN(parm, train_instances, train_data, train_labels, train_n_labels, n_labels, inputs, train_index, neighbors_cache, &(learner->lamlksnn), selection,
                                       parm->lamlksnn_is);
            else {
                for (int cluster = 0; cluster < learner->lamlknn.n_clusters; cluster++)
                    for (int i = 0; i < n_labels; i++)
                        learner->lamlksnn.ks[cluster][i] = -1;
                TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, selection, inputs, train_index, SS,
                        &(learner->lamlksnn), neighbors_cache);
            }

            // Report results
            if (parm->ml_kperlabel_method != LAMLksNNEvolutionary) {
                fprintf(parm->out_file, "@LAML-ksNN best number of clusters: %d\n", learner->lamlknn.n_clusters);
                fprintf(parm->out_file, "@LAML-ksNN average k's: %lf\n", AverageIntegerMatrix(learner->lamlksnn.ks, learner->lamlknn.n_clusters, n_labels));
                fprintf(parm->out_file, "Optimal LAML-k'sNN\n");
                for (int j = 0; j < learner->lamlknn.n_clusters; j++) {
                    fprintf(parm->out_file, "Cluster %d: (", j+1);
                    for (int i = 0; i < n_labels-1; i++)
                        fprintf(parm->out_file, "%*d, ", WIDTH(parm->mlksnn_max_k), learner->lamlksnn.ks[j][i]);
                    fprintf(parm->out_file, "%*d)\n", WIDTH(parm->mlksnn_max_k), learner->lamlksnn.ks[j][n_labels-1]);
                }
            }
            break;

        case EPCT:
        case RFPCT:
        case MLRF:
        case EDT:
        case EXTRAPCT:
        case EXTRAMLDT:
            MLRandomForestTrain(parm, train_data, train_instances, train_labels, inputs, &(learner->mlrf), selection, subspace, n_labels, train_index, train_n_labels, method == EPCT || method == RFPCT,
                                method == EPCT || method == EDT, SS, method == EXTRAPCT || method == EXTRAMLDT);
            break;

        case IBLRML:
        case IBLRMLplus:
            TrainIBLRMLModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->iblr), subspace, selection, SS, train_index, method == IBLRMLplus);
            break;

        case MLTREE:
            TrainMLTreeModels(parm, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, &(learner->mltree), subspace, parm->cv_hyperparameters, parm->mltree_model,
                       selection, SS, train_index);
            break;

        case MLFOREST:
            MLForestTrain(parm, train_data, train_instances, train_labels, inputs, &(learner->forest), selection, subspace, n_labels, train_index, train_n_labels);
            break;

        case MLRS:
            TrainMLBasedOnNeighborhoodRoughSets(parm, &(learner->mlrs), train_labels, train_n_labels, n_labels, train_instances, selection, train_data, train_index, subspace, inputs);
            break;

        case MLRSLC:
            TrainMLBasedOnNeighborhoodRoughSetsLocalCorrelation(parm, &(learner->mlrslc), train_labels, train_n_labels, n_labels, train_instances, selection, train_data, train_index, subspace,
                                                                inputs);
            break;

        case OPCT:
            OptionMLDecisionTreeTrain(parm, train_data, train_instances, train_labels, inputs, &(learner->otree), selection, subspace, n_labels, train_index, train_n_labels, false, true);
            break;

        case MrKNN:
            TrainMrKNN(parm, &(learner->mrknn), train_data, train_instances, train_labels, train_n_labels, train_index, n_labels, selection, subspace, inputs, neighbors_cache);
            break;

        case EPS:
        case BAGGING:
            TrainMultiLabelBagging(parm, ensemble, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, selection, train_index, SS, subspace,
                                   method == BAGGING ? parm->ml_ensemble_size : parm->eps_size, method == BAGGING ? parm->ml_base_learner : PRUNEDSET, method == BAGGING ? true : false);
            break;

        case REALADABOOSTRE:
        case DISCRETEADABOOSTRE:
            TrainAdaBoostByResampling(parm, ensemble, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, selection, train_index, SS, subspace, method == REALADABOOSTRE);
            break;

        case REALADABOOSTMH:
        case DISCRETEADABOOSTMH:
            TrainAdaBoostMH(parm, ensemble, train_data, train_instances, train_labels, train_n_labels, n_labels, inputs, selection, train_index, SS, subspace, method == REALADABOOSTMH,
                            parm->cv_hyperparameters);
            break;

        case DWKNND:
        case DWKNNM:
        case DWKNNZ:
            // Almost nothing to do, lazy learner
            TrainDWkNN(parm, &(learner->dwknn));
            break;

        case FRONEC1H:
        case FRONEC2H:
        case FRONEC3H:
        case FRONEC1LD:
        case FRONEC2LD:
        case FRONEC3LD:
            TrainFRONEC(parm, &(learner->fronec), n_labels, train_labels, train_n_labels, train_instances, inputs, train_data);
            break;

        case FVkNN:
            TrainFVkNN(&(learner->fvknn), parm, train_data, train_instances, train_labels, train_n_labels, n_labels, neighbors_cache, selection, SS, subspace, train_index, inputs);
            break;

        case ZEROML:
            // Nothing to do
            break;
    }

    if (SS && parm->symmetric_selection) {
        FREE(local_n_labels);
        for (int i = 0; i < train_instances; i++)
            FREE(local_labels[i]);
        FREE(local_labels);
    }
}

void FreeMLClassifier(const t_Parameters *parm, t_MLLearner *learner, t_MLClassifier method, int n_labels, int inputs, int instances, t_MLEnsemble *ensemble)
{
    switch (method) {
        case MLkNN:
            FreeMLkNN(&(learner->mlknn), n_labels);
            break;

        case SMLKNN:
            FreeStackedMLkNN(&(learner->smlknn), n_labels, instances);
            break;

        case LAMLkNN:
            FreeLAMLkNN(&(learner->lamlknn), n_labels);
            break;

        case CC:
        case NS:
        case NSSC:
            FreeClassifierChain(parm, &(learner->cc), n_labels, inputs);
            break;

        case ECC:
            FreeECC(parm, &(learner->ecc), n_labels, inputs);
            break;

        case LP:
        case PRUNEDSET:
        case PPT:
        case PPTN:
        case PPTC:
            FreeLabelPowerset(parm, &(learner->lp), inputs);
            break;

        case CLR:
        case TSVM:
        case TSCCM:
        case TSPCCM:
            FreeCalibratedLabelRanking(parm, &(learner->clr), n_labels, inputs);
            break;

        case BR:
            FreeBinaryRelevance(parm, &(learner->br), n_labels, inputs);
            break;

        case DBR:
            FreeDependentBinaryRelevance(parm, &(learner->dbr), n_labels, inputs);
            break;

        case STA:
            FreeMLStacking(parm, &(learner->dbr), n_labels, inputs);
            break;

        case RAkEL_d:
        case RAkEL_o:
            FreeRAkEL(parm, &(learner->rakel), inputs);
            break;

        case MLksNN:
            FreeMLksNN(&(learner->mlksnn), n_labels);
            break;

        case LAMLksNN:
            FreeLAMLksNN(&(learner->lamlksnn), n_labels);
            break;

        case HOMER_B:
        case HOMER_K:
        case HOMER_R:
            FreeHOMER(parm, &(learner->homer), inputs);
            break;

        case MLDT:
        case PCT:
            FreeMLDecisionTree(&(learner->mldt));
            break;

        case MLBASICBP:
            FreeMLBasicBP(&(learner->mlmlp), n_labels);
            break;

        case BPMLL:
            FreeBPMLL(&(learner->bpmll), inputs);
            break;

        case MLRF:
        case EPCT:
        case RFPCT:
        case EDT:
        case EXTRAPCT:
        case EXTRAMLDT:
            FreeMLRandomForest(&(learner->mlrf));
            break;

        case IBLRML:
        case IBLRMLplus:
            FreeIBLRML(&(learner->iblr), n_labels);
            break;

        case MLTREE:
            FreeMLTree(parm, &(learner->mltree), inputs);
            break;

        case MLFOREST:
            FreeMLForest(parm, &(learner->forest), inputs);
            break;

        case MLRS:
            MLRSFree(&(learner->mlrs), n_labels);
            break;

        case MLRSLC:
            MLRSLCFree(&(learner->mlrslc), n_labels);
            break;

        case OPCT:
            FreeOptionMLDecisionTree(&(learner->otree));
            break;

        case DMLKNN:
            FreeDMLkNN(&(learner->dmlknn), instances);
            break;

        case MrKNN:
            FreeMrKNN(&(learner->mrknn), n_labels);
            break;

        case BAGGING:
        case REALADABOOSTRE:
        case DISCRETEADABOOSTRE:
        case EPS:
            FreeBaggingEnsemble(parm, ensemble, n_labels, inputs, instances);
            break;

        case REALADABOOSTMH:
        case DISCRETEADABOOSTMH:
            FreeAdaBoostMHEnsemble(parm, ensemble, n_labels, inputs);
            break;

        case DWKNND:
        case DWKNNM:
        case DWKNNZ:
            // Nothing allocated
            break;

        case FRONEC1H:
        case FRONEC2H:
        case FRONEC3H:
        case FRONEC1LD:
        case FRONEC2LD:
        case FRONEC3LD:
            FreeFRONEC(&(learner->fronec));
            break;

        case FVkNN:
            FreeFVkNN(&(learner->fvknn), instances);
            break;

        case ZEROML:
            // Nothing to do
            break;
    }
}

void SaveMLOutputs(char *filename, int instances, int classes, int **h, int *n_h, int **labels, int *n_labels)
{
    FILE *f;

    if ((f = fopen(filename, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open file for outputs");

    fprintf(f, "Actual\tPredicted\n");
    for (int i = 0; i < instances; i++) {
        for (int j = 0; j < classes; j++)
            fprintf(f, "%d ", LabelInList(j, labels[i], n_labels[i]) ? 1 : 0);
        fprintf(f, "\t");
        for (int j = 0; j < classes; j++)
            fprintf(f, "%d ", LabelInList(j, h[i], n_h[i]) ? 1 : 0);
        fprintf(f, "\n");
    }

    fclose(f);
}
