/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Calibrated Label Ranking method
// J. Fürnkranz, E. Hüllermeier, E. Loza Mencía, and K. Brinker, “Multilabel classification via calibrated label ranking,” Mach. Learn., vol. 73, no. 2, pp. 133–153, 2008.

// QWeighted voting algorithm is just a modification for a faster testing time:
// E. Loza Mencı́a, S.-H. Park, J. Fürnkranz (2010) "Efficient voting prediction for pairwise multilabel classification," Neurocomputing 73:1164–1176.


void TrainCalibratedLabelRankingModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_CLR *clr, int *subspace, bool cv,
                                       t_Classifier classifier, int *subset, int **SS, int **index, bool tsa, bool pruned_cc)
{
    int *l_subspace;
    scalar **outputs = NULL;

    // Test if valid CLR classification model
    if (classifier == kNN || classifier == NN || classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_PARAM, "CLR classification method");

    if (SS && subset)
        OnErrorExit(ERR_PARAM, "Standard and partial instance selection are incompatible");

    // If no subspace select all
    MALLOC(l_subspace, int, inputs+classes, "TrainCalibratedLabelRankingModels: l_subspace");
    if (!subspace)
        for (int i = 0; i < inputs+classes; i++)
            l_subspace[i] = 1;
    else {
        memcpy(l_subspace, subspace, sizeof(int)*inputs);
        for (int i = inputs; i < inputs+classes; i++)
            l_subspace[i] = 1;
    }

    // Label dependent instance selection and standard instance selection are incompatible
    if (subset && SS)
        OnErrorExit(ERR_PARAM, "Label dependent instance selection and standard instance selection are incompatible");

    clr->method = classifier;

    MALLOC(clr->mask, bool, (classes*(classes-1))/2, "TrainCalibratedLabelRankingModels: clr->maks");
    MALLOC(clr->models, t_Learner *, classes, "TrainCalibratedLabelRankingModels: clr->models");
    MALLOC(clr->models[0], t_Learner, classes, "TrainCalibratedLabelRankingModels; clr->models[0]");
    MALLOC(clr->ensembles, t_Ensemble *, classes, "TrainCalibratedLabelRankingModels: clr->models");
    MALLOC(clr->ensembles[0], t_Ensemble, classes, "TrainCalibratedLabelRankingModels; clr->models[0]");

    // Train models with virtual labels
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < classes; i++) {
        int *b_labels;
        t_Parameters lparm;

        // Create training dataset for virtual labels
        MALLOC(b_labels, int, instances, "TrainCalibratedLabelRankingModels: b_labels");
        for (int k = 0; k < instances; k++)
            b_labels[k] = LabelInList(i, labels[k], n_labels[k]) ? 1 : 0;

        // If undersample, use local subset
        int *S = NULL;
        MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

        if (SS)
            memcpy(S, SS[i], sizeof(int)*instances);
        else if (subset)
            memcpy(S, subset, sizeof(int)*instances);
        else
            for (int j = 0; j < instances; j++)
                S[j] = 1;

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, b_labels, S, instances);


        // Cross-validate hyper-parameters?
        lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, data, instances, b_labels, l_subspace, inputs, 2, classifier, 1, S, index);

        // Train
        SHOWPROGRESS(parm->show_progress > 0, "CLR training VIRTUAL model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);
        ClassifierTraining(&lparm, classifier, data, instances, b_labels, inputs, 2, index, l_subspace, S, true, &(clr->models[0][i]), &(clr->ensembles[0][i]));

        FREE(b_labels);
        FREE(S);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // For TSCCM outputs of virtual models are fed to binary classifiers
    int *count = NULL;
    if (tsa) {
        scalar outputs[classes];
        MALLOC(count, int, instances, "TrainCalibratedLabelRankingModels: count");
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < classes; j++)
                Classifier2ClassOutput(parm, clr->method, data[i], inputs, 2, index ? index[i] : NULL, subspace, data, instances, NULL, NULL, &(clr->models[0][j]), 1, &outputs[j],
                                             index, &(clr->ensembles[0][j]));

            // Dense format
            if (!index) {
                REALLOC(data[i], scalar, inputs+classes, "TrainCalibratedLabelRankingModels: data[i]");
                for (int j = 0; j < classes; j++)
                    data[i][inputs+j] = outputs[j];
            } else {
                for (count[i] = 0; index[i][count[i]] != -1; count[i]++)
                    ;
                REALLOC(data[i], scalar, count[i]+classes, "TrainCalibratedLabelRankingModels: data[i]");
                REALLOC(index[i], int, count[i]+classes+1, "TrainCalibratedLabelRankingModels: index[i]");
                for (int j = 0; j < classes; j++) {
                    if (fabs(outputs[j]) > ZERO) {
                        count[i]++;

                        data[i][count[i]-1] = outputs[j];
                        index[i][count[i]-1] = inputs+j+1;
                    }
                }
                index[i][count[i]] = -1;
            }
        }
    } else if (pruned_cc) { // For pruned CC only labels from i,j classes are added to training set
        MALLOC(count, int, instances, "TrainCalibratedLabelRankingModels: count");

        AllocateScalarMatrix(&outputs, instances, classes);
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < classes; j++)
                Classifier2ClassOutput(parm, clr->method, data[i], inputs, 2, index ? index[i] : NULL, subspace, data, instances, NULL, NULL, &(clr->models[0][j]), 1, &outputs[i][j],
                                             index, &(clr->ensembles[0][j]));
            // Dense format
            if (!index) {
                REALLOC(data[i], scalar, inputs+2, "TrainCalibratedLabelRankingModels: data[i]");
            } else {
                for (count[i] = 0; index[i][count[i]] != -1; count[i]++)
                    ;
                REALLOC(data[i], scalar, count[i]+2, "TrainCalibratedLabelRankingModels: data[i]");
                REALLOC(index[i], int, count[i]+3, "TrainCalibratedLabelRankingModels: index[i]");
            }
        }
    }

    // Train a model for every label pair i,j
    int length = inputs;
    if (tsa)
        length += classes;
    else if (pruned_cc)
        length += 2;
    for (int i = 1; i < classes; i++) {
        MALLOC(clr->models[i], t_Learner, i, "TrainCalibratedLabelRankingModels: clr->models[i]");
        MALLOC(clr->ensembles[i], t_Ensemble, i, "TrainCalibratedLabelRankingModels: clr->ensembles[i]");
    }

    for (int i = 1; i < classes; i++)
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int j = 0; j < i; j++) {
            int t_instances = 0, **t_index = NULL, *b_labels;
            scalar **t_data;

            // Binary labels and training data
            MALLOC(b_labels, int, instances, "TrainCalibratedLabelRankingModels: b_labels");
            if (pruned_cc) {
                AllocateScalarMatrix(&t_data, instances, inputs+2);
                if (index)
                    AllocateIntegerMatrix(&t_index, instances, inputs+3);
            } else {
                MALLOC(t_data, scalar *, instances, "TrainCalibratedLabelRankingModels: t_data");
                if (index)
                    MALLOC(t_index, int *, instances, "TrainCalibratedLabelRankingModels: t_index");
            }

            bool relevant_i;

            // Create training dataset
            t_instances = 0;
            for (int k = 0; k < instances; k++) {
                bool consider = false;

                if ((!subset && !SS) || (subset && subset[k]) || (SS && (SS[i][k] || SS[j][k])))
                    consider = true;

                // Add instance if relevancy of label i and j is different
                if (consider && (relevant_i = LabelInList(i, labels[k], n_labels[k])) != LabelInList(j, labels[k], n_labels[k])) {
                    b_labels[t_instances] = relevant_i ? 1 : 0;

                    if (!pruned_cc) {
                        t_data[t_instances] = data[k];

                        if (index)
                            t_index[t_instances] = index[k];
                    } else {
                        // If pruned CC add outputs from virtual classifiers
                        if (!index) {
                            memcpy(t_data[t_instances], data[k], sizeof(scalar)*inputs);
                            t_data[t_instances][inputs] = outputs[k][j];
                            t_data[t_instances][inputs+1] = outputs[k][i];
                        } else {
                            int c = 0;
                            for (; index[k][c] != -1; c++) {
                                t_data[t_instances][c] = data[k][c];
                                t_index[t_instances][c] = index[k][c];
                            }

                            if (fabs(outputs[k][j]) > ZERO) {
                                t_data[t_instances][c] = outputs[k][j];
                                t_index[t_instances][c] = inputs+1;
                                c++;
                            }
                            if (fabs(outputs[k][i]) > ZERO) {
                                t_data[t_instances][c] = outputs[k][i];
                                t_index[t_instances][c] = inputs+2;
                                c++;
                            }
                            t_index[t_instances][c] = -1;
                        }
                    }

                    t_instances++;
                }
            }

            // For some pairs can be empty
            if (!t_instances) {
                clr->mask[TriangularOrder(i, j)] = false;
                continue;
            } else
                clr->mask[TriangularOrder(i, j)] = true;

            // Undersample?
            int *S = NULL;
            MALLOC(S, int, t_instances, "TrainBinaryRelevanceModels: S");

            FillVector(NULL, S, t_instances, 1.0);

            if (parm->ml_undersample_binary_learners)
                UndersampleDataWithLabel(parm, b_labels, S, t_instances);

            // Cross-validate hyper-parameters?
            t_Parameters lparm;
            lparm = *parm;
            if (cv)
                SetCrossValidatedParameters(&lparm, t_data, t_instances, b_labels, l_subspace, length, 2, classifier, 1, S, t_index);

            // Train
            SHOWPROGRESS(parm->show_progress > 0, "CLR training  binary model %*d/%d for label %*d vs %*d with %*d instances\r", WIDTH((classes*(classes+1))/2), TriangularOrder(i, j)+1,
                         (classes*(classes+1))/2, WIDTH(classes), j+1, WIDTH(classes), i+1, WIDTH(instances), t_instances);
            ClassifierTraining(&lparm, classifier, t_data, t_instances, b_labels, length, 2, t_index, l_subspace, S, true, &(clr->models[i][j]), &(clr->ensembles[i][j]));

            FREE(S);
            FREE(b_labels);
            if (pruned_cc) {
                FreeScalarMatrix(&t_data, instances);
                if (index)
                    FreeIntegerMatrix(&t_index, instances);
            } else {
                FREE(t_data);
                if (index)
                    FREE(t_index);
            }
        }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FREE(l_subspace);

    // Back to original length
    if (tsa || pruned_cc) {
        if (!index) {
            for (int i = 0; i < instances; i++)
                REALLOC(data[i], scalar, inputs, "TrainCalibratedLabelRankingModels: data[i]");
        } else {
            for (int i = 0; i < instances; i++) {
                REALLOC(data[i], scalar, count[i], "TrainCalibratedLabelRankingModels: data[i]");
                REALLOC(index[i], int, count[i]+1, "TrainCalibratedLabelRankingModels: data[i]");
            }
            FREE(count);
        }
        if (pruned_cc)
            FreeScalarMatrix(&outputs, instances);
    }
}

void CalibratedLabelRankingClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances, int n_labels,
        int *n_predicted, int inputs, int *query_index, t_CLR *clr, int *subspace, int **index)
{
    scalar f_virtual = 0.0, output;

    ZeroVector(f, NULL, n_labels);

    // Output for virtual labels
    for (int j = 0; j < n_labels; j++) {
        int category = Classifier2ClassOutput(parm, clr->method, query, inputs, 2, query_index, subspace, data, instances, NULL, NULL, &(clr->models[0][j]), 1, &output, index, &(clr->ensembles[0][j]));

        if (category)
            f[j]++;
        else
            f_virtual++;
    }

    // Output of all binary classifiers
    int n = 0;
    for (int i = 1; i < n_labels; i++)
        for (int j = 0; j < i; j++)
            if (clr->mask[n++]) {
                int category = Classifier2ClassOutput(parm, clr->method, query, inputs, 2, query_index, subspace, data, instances, NULL, NULL, &(clr->models[i][j]), 1, &output, index,
                                                      &(clr->ensembles[i][j]));

                if (category)
                    f[i]++;
                else
                    f[j]++;
            }

    // Relevant labels those above votes of virtual label
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++)
        if (f[i] > f_virtual) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }

    *classification = true;

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *ranks = true;
}

void FreeCalibratedLabelRanking(const t_Parameters *parm, t_CLR *clr, int classes, int inputs)
{
    for (int i = 0; i < classes; i++)
        ClassifierFree(parm, &(clr->models[0][i]), clr->method, 2, inputs, &(clr->ensembles[0][i]));

    int n = 0;
    for (int i = 1; i < classes; i++)
        for (int j = 0; j < i; j++)
            if (clr->mask[n++])
                ClassifierFree(parm, &(clr->models[i][j]), clr->method, 2, inputs, &(clr->ensembles[i][j]));

    for (int i = 0; i < classes; i++) {
        FREE(clr->models[i]);
        FREE(clr->ensembles[i]);
    }
    FREE(clr->models);
    FREE(clr->ensembles);
    FREE(clr->mask);
}

// Returns the linear index of a cell i,j in a triangular matrix

int TriangularOrder(int i, int j)
{
    int order = 0;
    for (int k = 1; k < i; k++)
        order += k;

    return order+j;
}
