/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include "maths.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "isfs.h"

/*
 * Cooperative co-evolution for multi-label instance selection
 */

scalar CooperativeCoevolutiveInstanceSelectionWithoutPopulation(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index,
        int *selection, int **ss, t_MLClassifier method)
{
    int ***subpops, neighbors_cache_size = 0;
    cache_t **neighbors_cache = NULL;

#ifdef _OPENMP
    omp_set_nested(1);  // Allow nested loops
#endif

    // Check parameters
    if (parm->subpop_fitness_w > ZERO || parm->subpop_difference_w > ZERO || parm->subpop_substitution_w > ZERO)
        OnErrorExit(ERR_PARAM, "Parameters incompatible with evolution without a population");

    // As many subpopulations as labels
    int n_subpops = classes;

    // Create and initialize subpopulations
    bool *first_time;
    t_HashTable *ht = NULL, html;
    if (parm->ml_use_hash_table) {
        MALLOC(ht, t_HashTable, n_subpops, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: ht");
        for (int i = 0; i < n_subpops; i++)
            CreateHashTable(&ht[i]);
        CreateHashTable(&html);
    }

    InitializeSubpopulations(parm, &subpops, n_subpops, &first_time, instances, labels, n_labels);

    // Cache of neighbors?
    SHOWPROGRESS(parm->show_progress > 0, "BEGIN Cache construction%c", '\n');
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE) {
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
        SHOWPROGRESS(parm->show_progress > 0, "DONE Cache construction%c", '\n');
    } else
        SHOWPROGRESS(parm->show_progress > 0, "Neighbors cache DISABLED%c", '\n');

    // For CHC subpopulation individuals
    int *minimum_distance;
    MALLOC(minimum_distance, int, n_subpops, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: minimum_distance");
    for (int i = 0; i < n_subpops; i++)
        minimum_distance[i] = instances/parm->chc_subpop_minimum_distance_factor;

    // Subpopulation fitness
    scalar **subpop_fitness, **subpop_accuracy;
    MALLOC(subpop_fitness, scalar *, n_subpops, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: subpop_fitness");
    MALLOC(subpop_accuracy, scalar *, n_subpops, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: subpop_accuracy");
    for (int i = 0; i < n_subpops; i++) {
        MALLOC(subpop_fitness[i], scalar, 2*parm->subpop_size, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: subpop_fitness[i]");
        MALLOC(subpop_accuracy[i], scalar, 2*parm->subpop_size, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: subpop_accuracy[i]");
    }

    // There is no elite
    int *elite;
    CALLOC(elite, int, parm->subpop_size, "CooperativeCoevolutiveInstanceSelectionWithoutPopulation: elite");

    time_t init_time = time(NULL);

    // Main loop
    bool ended = false;
    int g;
    for (g = 0; g < parm->generations && !ended; g++) {
        SHOWPROGRESS(parm->show_progress > 1, "Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);

        // CHC algorithm
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int j = 0; j < n_subpops; j++) {
            PerformGeneration(parm, subpops[j], parm->subpop_size, instances, minimum_distance, j, NULL, data, labels, n_labels, classes, inputs, n_subpops, NULL,
                              subpops, neighbors_cache, subpop_fitness[j], first_time, &ht[j], &html, NULL, NULL, index, elite, NULL, subpop_accuracy[j], method);
        }

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit)
            ended = true;
    }
    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

    // Select best individual from every subpopulation
    int best[n_subpops];
    for (int i = 0; i < n_subpops; i++)
        best[i] = OrderedSelection(subpop_fitness[i], 0, parm->subset_size);

    // Get optimal QS and k if apply
    scalar best_qs = -1.0;
    if (parm->ml_min_k != parm->ml_max_k || (parm->quorum_sensing_threshold_max - parm->quorum_sensing_threshold_min) > ZERO) {
        scalar inc = (parm->quorum_sensing_threshold_max - parm->quorum_sensing_threshold_min)/10;
        int best_k = parm->ml_min_k;
        scalar best_fitness = -1.0, fitness, performance;

        // For evaluations of individuals
        t_MLkNN mlknn;
        void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};

        int total = (parm->ml_max_k - parm->ml_min_k + 1)*11;
        int current = 0;
        for (int k = parm->ml_min_k; k <= parm->ml_max_k; k++)
            for (scalar qs = parm->quorum_sensing_threshold_min; qs <= parm->quorum_sensing_threshold_max; qs += inc) {
                if ((fitness = EvaluateIndividualFitness(parm, best, subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html, &qs,
                               index, &performance, NULL, NULL, method, classifier[method], NULL, 0, NULL, k, NULL, 0, NULL, NULL, instances)) > best_fitness) {
                    best_k = k;
                    best_qs = qs;
                    best_fitness = fitness;

                    SHOWPROGRESS(parm->show_progress > 1, "Best QS x k combination so far: %3.1lf x %3d\tFitness(Performance): %6.4lf(%6.4lf)\n", best_qs, best_k, best_fitness, performance);
                }

                SHOWPROGRESS(parm->show_progress > 1, "Evaluated QS x k combination: %*d/%d\r", WIDTH(total), current+1, total);
                current++;
            }
        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        if (parm->merge_method == QuorumSensing && !parm->ml_ins_sel_per_label)
            fprintf(parm->out_file, "@Quorum sensing threshold: %lf\n", best_qs);
        ((t_Parameters *) parm)->mlknn_k = best_k;
    } else {
        // Combine best individuals
        ((t_Parameters *) parm)->mlknn_k = parm->ml_min_k;

        if (parm->merge_method == QuorumSensing && !parm->ml_ins_sel_per_label) {
            fprintf(parm->out_file, "@Quorum sensing threshold: %lf\n", parm->quorum_sensing_threshold_min);
            best_qs = parm->quorum_sensing_threshold_min;
        }
    }

    if (parm->ml_ins_sel_per_label)
        for (int i = 0; i < n_subpops; i++)
            for (int j = 0; j < instances; j++)
                ss[i][j] = subpops[i][best[i]][j];
    else
        CreateSelectionFromSubpopulations(parm, best, subpops, n_subpops, selection, instances, best_qs, NULL);

    fprintf(parm->out_file, "@CC best MLkNN k: %d\n", parm->mlknn_k);

    // Check it time limit exceeded
    if (ended)
        SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);

    // Free memory
    for (int i = 0; i < n_subpops; i++) {
        for (int j = 0; j < 2*parm->subpop_size; j++)
            FREE(subpops[i][j]);
        FREE(subpop_fitness[i]);
        FREE(subpop_accuracy[i]);
        FREE(subpops[i]);
    }
    FREE(subpops);
    FREE(subpop_fitness);
    FREE(subpop_accuracy);
    FREE(elite);

    if (parm->ml_use_hash_table) {
        // Free hash tables
        for (int i = 0; i < n_subpops; i++) {
            fprintf(parm->out_file, "Subpopulation %*d hash table hits: %d\n", WIDTH(classes), i, ht[i].hits);
            DestroyHashTable(&ht[i]);
        }

        fprintf(parm->out_file, "Population hash table hits: %d\n", html.hits);
        DestroyHashTable(&html);
        FREE(ht);
    }

    scalar reduction = 0.0;
    if (parm->ml_ins_sel_per_label) {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < n_subpops; j++)
                if (!ss[j][i])
                    reduction++;
        reduction /= instances*n_subpops;
    } else {
        for (int i = 0; i < instances; i++)
            if (!selection[i])
                reduction++;
        reduction /= instances;
    }

    return reduction;
}
