/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>
#include "maths.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "isfs.h"

/*
 * Cooperative co-evolution for multi-label instance selection
 */

scalar CooperativeCoevolutiveInstanceSelection(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index, int *selection, int **ss,
        t_MLClassifier method)
{
    int ***subpops, **population, neighbors_cache_size = 0, best_k = -1;
    scalar *fitness, *cummulative, best_fitness = -1.0, best_qs = -1.0, *performance, best_negs = 0.0;
    cache_t **neighbors_cache = NULL;

    if (method != MLkNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CC IS for other method than ML-kNN");

#ifdef _OPENMP
    omp_set_nested(1);  // Allow nested loops
#endif

    // As many subpopulations as labels
    int n_subpops = classes;

    scalar chc_pop_init_prob = parm->chc_pop_init_prob;

    // Selection of posterior probabilities incompatible with allowing negative subpopulations
    if (parm->ml_ins_sel_per_label) {
        ((t_Parameters *)parm)->ml_allow_negative_subpops = false;
        chc_pop_init_prob = 1.0;
    }

    // Create and initialize global population
    MALLOC(fitness, scalar, parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1), "CooperativeCoevolutiveInstanceSelection: fitness");
    MALLOC(performance, scalar, parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1), "CooperativeCoevolutiveInstanceSelection: performance");
    MALLOC(cummulative, scalar, parm->population_size, "CooperativeCoevolutiveInstanceSelection: cummulative");
    MALLOC(population, int *, parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1), "CooperativeCoevolutiveInstanceSelection: population");
    for (int i = 0; i < parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1); i++) {
        MALLOC(population[i], int, n_subpops, "CooperativeCoevolutiveInstanceSelection: population[i]");
        // Allow negative values to ignore the population?
        if (parm->ml_allow_negative_subpops)
            for (int j = 0; j < n_subpops; j++)
                population[i][j] = (int) RANDOM(-parm->subpop_size, parm->subpop_size);
        else
            for (int j = 0; j < n_subpops; j++)
                population[i][j] = (RANDOM(0.0, 1.0) < chc_pop_init_prob) ? (int) RANDOM(0, parm->subpop_size) : -1;  // Allow -1 to not consider a subpopulation
    }

    // Create and initialize subpopulations
    bool *first_time;
    t_HashTable *ht = NULL, html;
    if (parm->ml_use_hash_table) {
        MALLOC(ht, t_HashTable, n_subpops, "CooperativeCoevolutiveInstanceSelection: ht");
        for (int i = 0; i < n_subpops; i++)
            CreateHashTable(&ht[i]);
        CreateHashTable(&html);
    }

    InitializeSubpopulations(parm, &subpops, n_subpops, &first_time, instances, labels, n_labels);

    // Cache of neighbors?
    SHOWPROGRESS(parm->show_progress > 0, "BEGIN Cache construction%c", '\n');
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE) {
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
        SHOWPROGRESS(parm->show_progress > 0, "DONE Cache construction%c", '\n');
    } else {
        if (parm->ml_loo)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation with LOO without cache");
        SHOWPROGRESS(parm->show_progress > 0, "Neighbors cache DISABLED%c", '\n');
    }

    // For CHC subpopulation individuals
    int *minimum_distance;
    MALLOC(minimum_distance, int, n_subpops, "CooperativeCoevolutiveInstanceSelection: minimum_distance");
    for (int i = 0; i < n_subpops; i++)
        minimum_distance[i] = instances/parm->chc_subpop_minimum_distance_factor;

    int *order;
    MALLOC(order, int, n_subpops, "CooperativeCoevolutiveInstanceSelection: order");
    for (int i = 0; i < n_subpops; i++)
        order[i] = i;

    // Subpopulation fitness
    scalar **subpop_fitness, **subpop_accuracy;
    MALLOC(subpop_fitness, scalar *, n_subpops, "CooperativeCoevolutiveInstanceSelection: subpop_fitness");
    MALLOC(subpop_accuracy, scalar *, n_subpops, "CooperativeCoevolutiveInstanceSelection: subpop_accuracy");
    for (int i = 0; i < n_subpops; i++) {
        MALLOC(subpop_fitness[i], scalar, 2*parm->subpop_size, "CooperativeCoevolutiveInstanceSelection: subpop_fitness[i]");
        MALLOC(subpop_accuracy[i], scalar, 2*parm->subpop_size, "CooperativeCoevolutiveInstanceSelection: subpop_accuracy[i]");
    }

    // Difference for sub-population selection
    scalar *difference, *cprobability;
    MALLOC(difference, scalar, n_subpops, "CooperativeCoevolutiveInstanceSelection: difference");
    MALLOC(cprobability, scalar, n_subpops, "CooperativeCoevolutiveInstanceSelection: cprobability");
    for (int i = 0; i < n_subpops; i++)
        difference[i] = 1.0;

    // Evolution of quorum sensing threshold if apply
    scalar *qs = NULL;
    MALLOC(qs, scalar, parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1), "CooperativeCoevolutiveInstanceSelection: qs");
    for (int i = 0; i < parm->population_size; i++)
        qs[i] = RANDOM(parm->quorum_sensing_threshold_min, parm->quorum_sensing_threshold_max);

    // Evolution of k
    int *k = NULL;
    MALLOC(k, int, parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1), "CooperativeCoevolutiveInstanceSelection: k");
    for (int i = 0; i < parm->population_size; i++)
        k[i] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);

    // For protection of best individuals
    int *elite, **elite_members;
    int *elite_list;
    MALLOC(elite, int, parm->population_size, "CooperativeCoevolutiveInstanceSelection: elite");
    MALLOC(elite_members, int *, n_subpops, "CooperativeCoevolutiveInstanceSelection: elite_members");
    MALLOC(elite_list, int, parm->population_size, "CooperativeCoevolutiveInstanceSelection: elite_list");
    for (int i = 0; i < n_subpops; i++)
        MALLOC(elite_members[i], int, parm->subpop_size, "CooperativeCoevolutiveInstanceSelection: elite_members[i]");

    time_t init_time = time(NULL);

    // Evaluate isolated fitness of subindividuals if needed
    if (parm->ml_cc_weighted_qs)
        for (int i = 0; i < n_subpops; i++)
#ifdef _OPENMP
            #pragma omp parallel for schedule(dynamic)
#endif
            for (int j = 0; j < parm->subpop_size; j++)
                subpop_accuracy[i][j] = EvaluateSelectionOneLabel(parm, data, instances, labels, inputs, parm->binary_classifier, subpops[i][j], neighbors_cache, index, parm->k, k, i, population,
                                        j, n_labels, parm->subpop_perf_measure, NULL, -1);

    // Evaluate the fitness of the current population
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < parm->population_size; i++) {
        t_MLkNN mlknn;
        void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};

        fitness[i] = EvaluateIndividualFitness(parm, population[i], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html, &qs[i], index,
                                               &performance[i], NULL, parm->ml_cc_weighted_qs ? subpop_accuracy : NULL, method, classifier[method], NULL, 0, NULL, k[i], NULL, 0, NULL, NULL,
                                               instances);
    }

    int *list = NULL;
    if (parm->ml_cc_pop_evol_scheme == schemeCHCGA) {
        MALLOC(list, int, parm->population_size, "CooperativeCoevolutiveInstanceSelection: list");
        for (int i = 0; i < parm->population_size; i++)
            list[i] = i;
    }

    // Main loop
    bool ended = false;
    int g;
    for (g = 0; g < parm->generations && !ended; g++) {
        SHOWPROGRESS(parm->show_progress > 1, "Generation: %*d/%d\n", WIDTH(parm->generations), g+1, parm->generations);

        // Np loops of population
        // Steady state genetic algorithm
        for (int np = 0; np < parm->np && !ended; np++) {
            SHOWPROGRESS(parm->show_progress > 1, "Np generation: %*d/%d\r", WIDTH(parm->np), np+1, parm->np);

            switch (parm->ml_cc_pop_evol_scheme) {
                case schemeSSGA:
                    cummulative[0] = fitness[0];
                    for (int k = 1; k < parm->population_size; k++)
                        cummulative[k] = cummulative[k-1]+fitness[k];

                    // Select worsts individuals
                    int worst = OrderedSelection(fitness, parm->population_size - 1, parm->population_size);
                    int sworst = OrderedSelection(fitness, parm->population_size - 2, parm->population_size);

                    // Select two parents
                    int father, mother;
                    do {
                        father = RouletteSelection(cummulative, parm->population_size);
                        mother = RouletteSelection(cummulative, parm->population_size);
                    } while (mother == worst || father == worst || mother == sworst || father == sworst);

                    // Perform crossover
                    TwoPointCrossover(population[worst], population[sworst], population[father], population[mother], n_subpops);

                    // Perform BLX-alpha crossover for quorum sensing threshold
                    scalar minimum = MIN(qs[father], qs[mother]);
                    scalar maximum = MAX(qs[father], qs[mother]);
                    scalar I = maximum - minimum;
                    qs[worst] = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, RANDOM(minimum - I * BLX_ALPHA, maximum + I * BLX_ALPHA), parm->quorum_sensing_threshold_max);
                    qs[sworst] = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, RANDOM(minimum - I * BLX_ALPHA, maximum + I * BLX_ALPHA), parm->quorum_sensing_threshold_max);

                    // Perform BLX-alpha crossover for k
                    int min = MIN(k[father], k[mother]);
                    int max = MAX(k[father], k[mother]);
                    I = max - min;
                    k[worst] = (int) AssignWithinBoundsScalar(parm->ml_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->ml_max_k);
                    k[sworst] = (int) AssignWithinBoundsScalar(parm->ml_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->ml_max_k);

                    // Evaluate new individuals
                    t_MLkNN mlknn;
                    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
                    fitness[worst] = EvaluateIndividualFitness(parm, population[worst], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html,
                                     &qs[worst], index, &performance[worst], NULL, subpop_accuracy, method, classifier[method], NULL, g, NULL, k[worst], NULL, 0, NULL, NULL, instances);
                    fitness[sworst] = EvaluateIndividualFitness(parm, population[sworst], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html,
                                      &qs[sworst], index, &performance[sworst], NULL, subpop_accuracy, method, classifier[method], NULL, g, NULL, k[sworst], NULL, 0, NULL, NULL, instances);
                    break;

                case schemeCHCGA:
                    // WARNING: This is an adaptation of CHC to non-binary individuals
                    RandomizeVector(list, parm->population_size);
                    for (int i = 0; i < parm->population_size; i += 2) {
                        int father = list[i];
                        int mother = list[i+1];

                        // Perform HUX crossover creating new individuals
                        HuxCrossover(population[mother], population[father], population[i+parm->population_size], population[i+parm->population_size+1], n_subpops);

                        // Perform BLX-alpha crossover for quorum sensing threshold
                        scalar minimum = MIN(qs[father], qs[mother]);
                        scalar maximum = MAX(qs[father], qs[mother]);
                        scalar I = maximum - minimum;
                        qs[i+parm->population_size] = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, RANDOM(minimum - I * BLX_ALPHA, maximum + I * BLX_ALPHA),
                                                      parm->quorum_sensing_threshold_max);
                        qs[i+parm->population_size+1] = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, RANDOM(minimum - I * BLX_ALPHA, maximum + I * BLX_ALPHA),
                                                        parm->quorum_sensing_threshold_max);

                        // Perform BLX-alpha crossover for k
                        int min = MIN(k[father], k[mother]);
                        int max = MAX(k[father], k[mother]);
                        I = max - min;
                        k[i+parm->population_size] = (int) AssignWithinBoundsScalar(parm->ml_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->ml_max_k);
                        k[i+parm->population_size+1] = (int) AssignWithinBoundsScalar(parm->ml_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->ml_max_k);
                    }

                    // Evaluate new individuals
#ifdef _OPENMP
                    #pragma omp parallel for schedule(dynamic)
#endif
                    for (int i = parm->population_size; i < 2*parm->population_size; i++) {
                        t_MLkNN mlknn;
                        void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
                        fitness[i] = EvaluateIndividualFitness(parm, population[i], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html, &qs[i],
                                                               index, &performance[i], NULL, subpop_accuracy, method, classifier[method], NULL, g, NULL, k[i], NULL, 0, NULL, NULL, instances);
                    }

                    // Select new population
                    for (int i = 0; i < parm->population_size; i++) {
                        int order = OrderedSelection(fitness+i, 0, 2*parm->population_size-i)+i;

                        int *aux_pt = population[order];
                        population[order] = population[i];
                        population[i] = aux_pt;

                        scalar aux_f = fitness[order];
                        fitness[order] = fitness[i];
                        fitness[i] = aux_f;

                        aux_f = qs[order];
                        qs[order] = qs[i];
                        qs[i] = aux_f;

                        int new_k = k[order];
                        k[order] = k[i];
                        k[i] = new_k;
                    }
                    break;

                case schemeGGA:
                case schemeSSMA:
                    OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for ML CC");
                    break;
            }

            // Protect best ones from mutation
            for (int i = 0; i < parm->population_size; i++)
                elite[i] = false;
            for (int i = 0; i < n_subpops; i++)
                for (int j = 0; j < parm->subpop_size; j++)
                    elite_members[i][j] = false;
            NBestOrderedSelection(fitness, parm->population_size, parm->population_size*parm->ml_elitism, elite_list);
            for (int i = 0; i < parm->population_size*parm->ml_elitism; i++) {
                elite[elite_list[i]] = true;
                for (int j = 0; j < n_subpops; j++)
                    if (population[elite_list[i]][j] >= 0)
                        elite_members[j][population[elite_list[i]][j]] = true;
            }

            bool mutated = false;   // To know if elite list must be updated

            // Mutation
            for (int j = 0; j < parm->population_size; j++)
                if ((parm->ml_cc_mutate_elite || !elite[j]) && RANDOM(0.0, 1.0) < parm->ssga_mutation) {
                    CCMutateIndividual(parm, population[j], n_subpops, parm->subpop_size, chc_pop_init_prob, &qs[j], g, &k[j]);
                    t_MLkNN mlknn;
                    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
                    fitness[j] = EvaluateIndividualFitness(parm, population[j], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html, &qs[j],
                                                           index, &performance[j], NULL, subpop_accuracy, method, classifier[method], NULL, g, NULL, k[j], NULL, 0, NULL, NULL, instances);
                    mutated = true;
                }

            // Memetic mutation
            for (int j = 0; j < parm->population_size; j++)
                if ((parm->ml_cc_mutate_elite || !elite[j]) && RANDOM(0.0, 1.0) < parm->cc_memetic_mutation) {
                    CCMemeticMutation(parm, population[j], &qs[j], k[j], data, instances, labels, n_labels, index, &fitness[j], &performance[j], subpops, neighbors_cache, classes, inputs,
                                      subpop_accuracy, method);
                    mutated = true;
                }

            // Memetic mutation for subpopulations
            for (int j = 0; j < parm->population_size; j++)
                if ((parm->ml_cc_mutate_elite || !elite[j]) && RANDOM(0.0, 1.0) < parm->cc_memetic_mutation_for_subpop) {
                    CCMemeticMutationForSubpopulations(parm, population[j], &qs[j], k[j], data, instances, labels, n_labels, index, &fitness[j], &performance[j], subpops, neighbors_cache, classes,
                                                       inputs, elite_members, false, parm->ml_cc_weighted_qs ? subpop_accuracy : NULL, method);
                    mutated = true;
                }

            // Memetic mutation of best individual
            if (parm->ml_cc_local_opt_best) {
                int best = OrderedSelection(fitness, 0, parm->population_size);
                CCMemeticMutation(parm, population[best], &qs[best], k[best], data, instances, labels, n_labels, index, &fitness[best], &performance[best], subpops, neighbors_cache, classes, inputs,
                                  subpop_accuracy, method);
                CCMemeticMutationForSubpopulations(parm, population[best], &qs[best], k[best], data, instances, labels, n_labels, index, &fitness[best], &performance[best], subpops, neighbors_cache,
                                                   classes, inputs, elite_members, true, parm->ml_cc_weighted_qs ? subpop_accuracy : NULL, method);
                mutated = true;
            }

            if (mutated) {
                // Protect best ones from mutation
                for (int i = 0; i < parm->population_size; i++)
                    elite[i] = false;
                for (int i = 0; i < n_subpops; i++)
                    for (int j = 0; j < parm->subpop_size; j++)
                        elite_members[i][j] = false;
                NBestOrderedSelection(fitness, parm->population_size, parm->population_size*parm->ml_elitism, elite_list);
                for (int i = 0; i < parm->population_size*parm->ml_elitism; i++) {
                    elite[elite_list[i]] = true;
                    for (int j = 0; j < n_subpops; j++)
                        if (population[elite_list[i]][j] >= 0)
                            elite_members[j][population[elite_list[i]][j]] = true;
                }
            }

            // Check it time limit exceeded
            if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit)
                ended = true;
        }

        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        // Nsp loops of subpopulations
        // CHC algorithm
        for (int nsp = 0; nsp < parm->nsp && !ended; nsp++) {
            SHOWPROGRESS(parm->show_progress > 1, "Nsp generation: %*d/%d\n", WIDTH(parm->nsp), nsp+1, parm->nsp);

            // Random order of subpopulations evolution
            RandomizeVector(order, n_subpops);

            if (parm->ml_onesubpop_per_cycle) {
                int next = order[0];
                if (parm->ml_select_subpop_by_relevance) {
                    cprobability[0] = difference[0];
                    for (int i = 1; i < n_subpops; i++)
                        cprobability[i] = cprobability[i-1] + difference[i];

                    next = RouletteSelection(cprobability, n_subpops);
                }

                PerformGeneration(parm, subpops[next], parm->subpop_size, instances, minimum_distance, next, population, data, labels, n_labels, classes, inputs, n_subpops, fitness, subpops,
                                  neighbors_cache, subpop_fitness[next], first_time, &ht[next], &html, &difference[next], qs, index, elite_members[next], k, subpop_accuracy[next], method);
            } else
#ifdef _OPENMP
                #pragma omp parallel for schedule(dynamic)
#endif
                for (int j = 0; j < n_subpops; j++) {
                    PerformGeneration(parm, subpops[order[j]], parm->subpop_size, instances, minimum_distance, order[j], population, data, labels, n_labels, classes, inputs, n_subpops, fitness,
                                      subpops, neighbors_cache, subpop_fitness[order[j]], first_time, &ht[order[j]], &html, &difference[order[j]], qs, index, elite_members[order[j]], k,
                                      subpop_accuracy[order[j]], method);
                }

            // Reevaluate individuals
            if (parm->reevaluate_individuals)
#ifdef _OPENMP
                #pragma omp parallel for schedule(dynamic)
#endif
                for (int j = 0; j < parm->population_size; j++) {
                    t_MLkNN mlknn;
                    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
                    fitness[j] = EvaluateIndividualFitness(parm, population[j], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html, &qs[j],
                                                           index, &performance[j], NULL, subpop_accuracy, method, classifier[method], NULL, g, NULL, k[j], NULL, 0, NULL, NULL, instances);
                }

            // Check it time limit exceeded
            if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit)
                ended = true;
        }

        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        // Reevaluate and get best individual
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int j = 0; j < parm->population_size; j++) {
            t_MLkNN mlknn;
            void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
            fitness[j] = EvaluateIndividualFitness(parm, population[j], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, &html, &qs[j], index,
                                                   &performance[j], NULL, subpop_accuracy, method, classifier[method], NULL, g, NULL, k[j], NULL, 0, NULL, NULL, instances);
        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            SHOWPROGRESS(true, "%*d generation best individual FITNESS(MEAN), QS, K: %6.4lf(%6.4lf) %4.2lf %2d\n", WIDTH(parm->generations), g+1, fitness[best],
                         AverageScalarVector(fitness, parm->population_size), qs[best], k[best]);
            VERBOSE(true, parm->out_file, "%*d generation best individual FITNESS(MEAN), QS, K: %6.4lf(%6.4lf) %4.2lf %2d\n", WIDTH(parm->generations), g+1, fitness[best],
                    AverageScalarVector(fitness, parm->population_size), qs[best], k[best]);
        }


        // Return best overall individual
        if (parm->ml_return_overall_best) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            if (fitness[best] > best_fitness) {
                if (parm->ml_ins_sel_per_label)
                    for (int i = 0; i < n_subpops; i++)
                        for (int j = 0; j < instances; j++)
                            ss[i][j] = subpops[i][population[best][i]][j];
                else {
                    best_qs = qs[best];
                    CreateSelectionFromSubpopulations(parm, population[best], subpops, n_subpops, selection, instances, qs[best], parm->ml_cc_weighted_qs ? subpop_accuracy : NULL);
                }

                best_k = k[best];
                best_fitness = fitness[best];

                if (parm->ml_allow_negative_subpops) {
                    best_negs = 0.0;
                    for (int i = 0; i < n_subpops; i++)
                        if (population[best][i] < 0)
                            best_negs++;
                    best_negs /= n_subpops;
                }
            }
        }

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit)
            ended = true;
    }

    // Check it time limit exceeded
    if (ended)
        SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);

    if (!parm->ml_return_overall_best) {
        int best = OrderedSelection(fitness, 0, parm->population_size);

        if (parm->ml_ins_sel_per_label)
            for (int i = 0; i < n_subpops; i++)
                for (int j = 0; j < instances; j++)
                    ss[i][j] = subpops[i][population[best][i]][j];
        else {
            if (parm->merge_method == QuorumSensing)
                fprintf(parm->out_file, "@Quorum sensing threshold: %lf\n", qs[best]);
            CreateSelectionFromSubpopulations(parm, population[best], subpops, n_subpops, selection, instances, qs[best], parm->ml_cc_weighted_qs ? subpop_accuracy : NULL);
        }

        ((t_Parameters *) parm)->mlknn_k = k[best];
        fprintf(parm->out_file, "@CC best MLkNN k: %d\n", k[best]);
    } else {
        ((t_Parameters *) parm)->mlknn_k = best_k;
        fprintf(parm->out_file, "@CC best MLkNN k: %d\n", best_k);

        if (parm->merge_method == QuorumSensing && !parm->ml_ins_sel_per_label)
            fprintf(parm->out_file, "@Quorum sensing threshold: %lf\n", best_qs);
    }

    if (parm->ml_allow_negative_subpops)
        fprintf(parm->out_file, "@Negative subpopulations percentage: %lf\n", best_negs);

    // Free memory
    for (int i = 0; i < n_subpops; i++) {
        for (int j = 0; j < 2*parm->subpop_size; j++)
            FREE(subpops[i][j]);
        FREE(subpop_fitness[i]);
        FREE(subpop_accuracy[i]);
        FREE(subpops[i]);
        FREE(elite_members[i]);
    }
    FREE(subpops);
    FREE(subpop_fitness);
    FREE(subpop_accuracy);
    for (int i = 0; i < parm->population_size * (parm->ml_cc_pop_evol_scheme == schemeCHCGA ? 2 : 1); i++)
        FREE(population[i]);
    FREE(population);
    FREE(fitness);
    FREE(performance);
    FREE(cummulative);
    FREE(difference);
    FREE(cprobability);
    FREE(qs);
    FREE(k);
    FREE(elite);
    FREE(elite_members);
    FREE(elite_list);
    FREE(order);
    if (list)
        FREE(list);

    if (parm->ml_use_hash_table) {
        // Free hash tables
        for (int i = 0; i < n_subpops; i++) {
            fprintf(parm->out_file, "Subpopulation %*d hash table hits: %d\n", WIDTH(classes), i, ht[i].hits);
            DestroyHashTable(&ht[i]);
        }

        fprintf(parm->out_file, "Population hash table hits: %d\n", html.hits);
        DestroyHashTable(&html);
        FREE(ht);
    }

    scalar reduction = 0.0;
    if (parm->ml_ins_sel_per_label) {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < n_subpops; j++)
                if (!ss[j][i])
                    reduction++;
        reduction /= instances*n_subpops;
    } else {
        for (int i = 0; i < instances; i++)
            if (!selection[i])
                reduction++;
        reduction /= instances;
    }

    return reduction;
}

#define NONUNIF_B       5.00

void CCMutateIndividual(const t_Parameters *parm, int *individual, int n_subpops, int subpop_size, scalar one_prob, scalar *qs, int g, int *k)
{
    // Select a random subpopulation and modify the individual
    if (parm->ml_allow_negative_subpops)
        individual[(int) RANDOM(0, n_subpops)] = (int) RANDOM(-subpop_size, subpop_size);
    else
        individual[(int) RANDOM(0, n_subpops)] = RANDOM(0, subpop_size);

    // Non-uniform mutation of quorum sensing threshold
    scalar delta = (RANDOM(0.0, 1.0) < 0.5) ? Delta((scalar) g, parm->quorum_sensing_threshold_max - *qs, parm->generations) :
                   -Delta((scalar) g, *qs - parm->quorum_sensing_threshold_min, parm->generations);
    *qs = AssignWithinBoundsScalar(parm->quorum_sensing_threshold_min, *qs + delta, parm->quorum_sensing_threshold_max);

    // Non-uniform mutation of k
    delta = (RANDOM(0.0, 1.0) < 0.5) ? Delta((scalar) g, (scalar) parm->ml_max_k - *k, parm->generations) : -Delta((scalar) g, (scalar) *k - parm->ml_min_k, parm->generations);
    *k = (int) AssignWithinBoundsScalar(parm->ml_min_k, *k + delta, parm->ml_max_k);
}

scalar Delta(scalar t, scalar y, int max_t)
{
    return y*RANDOM(0.0, 1.0)*pow(1.0-t/max_t, NONUNIF_B);
}

// Remove every subpopulation member and test if better or equal

void CCMemeticMutation(const t_Parameters *parm, int *individual, scalar *qs, int k, scalar **data, int instances, int **labels, int *n_labels, int **index, scalar *fitness, scalar *performance,
                       int ***subpops, cache_t **neighbors_cache, int classes, int inputs, scalar **subpop_accuracy, t_MLClassifier method)
{
    int list[classes], removed = 0, added = 0, old;
    scalar new_performance, previuos_performance = *performance, previous_fitness = *fitness;

    t_MLkNN mlknn;
    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
    if (method != MLkNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CC IS for other method than ML-kNN");

    // List of positive subpopulation members
    for (int i = 0; i < classes; i++)
        list[i] = i;

    RandomizeVector(list, classes);

    // Remove/add and test
    for (int i = 0; i < classes; i++) {
        individual[list[i]] = -(old = individual[list[i]]);

        scalar new_fitness = EvaluateIndividualFitness(parm, individual, subpops, classes, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, NULL, qs, index,
                             &new_performance, NULL, subpop_accuracy, method, classifier[method], NULL, 0, NULL, k, NULL, 0, NULL, NULL, instances);

        if (new_fitness > *fitness) {
            *fitness = new_fitness;
            *performance = new_performance;

            if (old >= 0)
                removed++;
            else
                added++;
        } else
            individual[list[i]] = old;
    }

    if (removed || added) {
        SHOWPROGRESS(parm->show_progress > 0, "Removed/added %3d/%-3d members.   Accuracy(Fitness): %6.4lf(%6.4lf) => %6.4lf(%6.4lf)\n", removed, added, previuos_performance, previous_fitness,
                     *performance, *fitness);
    } else
        SHOWPROGRESS(parm->show_progress > 1, "Memetic mutation removed/added 0 members%c", '\n');
}

void CCMemeticMutationForSubpopulations(const t_Parameters *parm, int *individual, scalar *qs, int k, scalar **data, int instances, int **labels, int *n_labels, int **index, scalar *fitness,
                                        scalar *performance, int ***subpops, cache_t **neighbors_cache, int classes, int inputs, int **elite_members, bool mutate_elite, scalar **subpop_accuracy,
                                        t_MLClassifier method)
{
    scalar new_performance, previuos_performance = *performance, previous_fitness = *fitness;
    int removed = 0, added = 0;

    t_MLkNN mlknn;
    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, NULL, NULL};
    if (method != MLkNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CC IS for other method than ML-kNN");

    // Get selection from individual
    int *selection, *list;
    MALLOC(selection, int, instances, "CCMemeticMutationForSubpopulations: selection");
    MALLOC(list, int, instances, "CCMemeticMutationForSubpopulations: list");
    CreateSelectionFromSubpopulations(parm, individual, subpops, classes, selection, instances, *qs, parm->ml_cc_weighted_qs ? subpop_accuracy : NULL);

    for (int i = 0; i < instances; i++)
        list[i] = i;
    RandomizeVector(list, instances);

    // Activate/deactivate bits in all subpopulation members
    for (int i = 0; i < instances; i++)
        if (RANDOM(0.0, 1.0) < parm->cc_memetic_mutation_for_subpop_bit_ratio) {
            // Change and evaluate
            selection[list[i]] = !selection[list[i]];

            scalar new_fitness = EvaluateIndividualFitness(parm, NULL, subpops, classes, instances, labels, n_labels, classes, inputs, data, selection, neighbors_cache, NULL, qs, index,
                                 &new_performance, NULL, subpop_accuracy, method, classifier[method], NULL, 0, NULL, k, NULL, 0, NULL, NULL, instances);

            // Keep or restore
            if (new_fitness > *fitness) {
                *fitness = new_fitness;
                *performance = new_performance;

                if (!selection[list[i]])
                    removed++;
                else
                    added++;

                for (int j = 0; j < classes; j++)
                    if (individual[j] >= 0 && (mutate_elite || !elite_members[j][individual[j]]))
                        subpops[j][individual[j]][list[i]] = selection[list[i]];
            } else
                // Restore
                selection[list[i]] = !selection[list[i]];
        }

    FREE(selection);
    FREE(list);

    if (removed || added) {
        SHOWPROGRESS(parm->show_progress > 0, "Removed/added %3d/%-3d instances. Accuracy(Fitness): %6.4lf(%6.4lf) => %6.4lf(%6.4lf)\n", removed, added, previuos_performance, previous_fitness,
                     *performance, *fitness);
    } else
        SHOWPROGRESS(parm->show_progress > 1, "Memetic mutation removed/added 0 instances%c", '\n');
}
