/**************************************************************************
*   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
*   npedrajas@uco.es                                                      *
*                                                                         *
*   This program is FREE software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
/*
 * Input/Output functions.
 */
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include "common.h"
#include "macros.h"
#include "ml.h"
#include "maths.h"

/*
 * Read a data file. The structure of the file must be:
 * [SPARSE] or nothing
 * $ Number of lines
 * $ Number of inputs
 * $ Number of classes
 * One pattern per line, variables separated by spaces, tabs or new lines.
 * Labels in binary list: [ 0 1 1 1 1 0 1 1 0 1 ]
 */

#define MAX_LINE_LENGTH 100000

void ReadMultiLabelData(const t_Parameters *parm, scalar ***data, int ***labels, int **n_labels, int *instances, int *inputs, int *classes, char *file, int ***in_index)
{
    FILE *f;
    char *line;
    bool sparse_file = false;

    MALLOC(line, char, MAX_LINE_LENGTH, "ReadMultiLabelData: line");

    // Opens the file and read number of elements
    if ((f = fopen(file, "rt")) == NULL) {
        printf("Dataset %s not found\n", file);
        OnErrorExit(ERR_FILE, "ReadMultiLabelData");
    }

    // Check multilabel label
    if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
        OnErrorExit(ERR_READ, "ReadMultiLabelData");

    if (strncasecmp(line, "[MULTILABEL", 11)) {
        fclose(f);
        OnErrorExit(ERR_FILE, "File not in multi-label format");
    }

    // Check sparse or dense label
    if (!strncasecmp(line+11, ", SPARSE]", 9)) {
        sparse_file = true;
        SHOWPROGRESS(parm->show_progress > 0, "Sparse file format%c", '\n');
    } else {
        if (!strncasecmp(line+11, ", DENSE]", 8)) {
            // Dense format
            sparse_file = false;
            SHOWPROGRESS(parm->show_progress > 0, "Dense file format%c", '\n');
        } else
            OnErrorExit(ERR_FILE, "Unknown format");
    }

    // Read number of lines and columns
    if (!fscanf(f, "$ %d\n", instances))
        OnErrorExit(ERR_READ, "ReadMultiLabelData");
    if (!fscanf(f, "$ %d\n", inputs))
        OnErrorExit(ERR_READ, "ReadMultiLabelData");
    if (!fscanf(f, "$ %d\n", classes))
        OnErrorExit(ERR_READ, "ReadMultiLabelData");

    MALLOC(*data, scalar *, *instances, "ReadMultiLabelData: *data");
    MALLOC(*labels, int *, *instances, "ReadMultiLabelData: *labels");
    MALLOC(*n_labels, int, *instances, "ReadMultiLabelData: *n_labels")

    if (parm->sparse_format) {
        MALLOC(*in_index, int *, *instances, "ReadMultiLabelData: *in_index");
        SHOWPROGRESS(true, "File will be read into SPARSE data%c", '\n');
    }

    // Read data
    for (int i = 0; i < *instances; i++) {
        if (fgets(line, MAX_LINE_LENGTH, f) == NULL)
            OnErrorExit(ERR_READ, "ReadMultiLabelData");

        if (sparse_file)
            ReadNextInstanceSparseFormat(line, &(*data)[i], &(*labels)[i], &(*n_labels)[i], *inputs, *classes);
        else
            ReadNextInstanceDenseFormat(line, &(*data)[i], &(*labels)[i], &(*n_labels)[i], *inputs, *classes);

        if (parm->sparse_format)
            ConvertToSparseFormat(&(*data)[i], &(*in_index)[i], *inputs);

        // Sort labels
        qsort((*labels)[i], (*n_labels)[i], sizeof(int), CompareInt);
    }

    fclose(f);

    RemoveInstancesWithNoLabelsOrNull(data, labels, n_labels, instances, in_index);

    FREE(line);
}

void ReadNextInstanceDenseFormat(char *line, scalar **data, int **labels, int *n_labels, int inputs, int classes)
{
    char *next;

    // Read inputs
    MALLOC(*data, scalar, inputs, "ReadNextInstanceDenseFormat: *data");
    next = strtok(line, " ,\t");
    for (int i = 0; i < inputs; i++) {
        (*data)[i] = strtod(next, NULL);
        next = strtok(NULL, " ,\t");
    }

    // Read labels
    MALLOC(*labels, int, classes, "ReadNextInstanceDenseFormat: *labels");
    *n_labels = 0;

    // Check correct format
    if (next[0] != '[')
        OnErrorExit(ERR_FILE, "Input file wrong format");
    next = strtok(NULL, " ,\t");

    for (int i = 0; i < classes; i++) {
        int d = (int) strtol(next, NULL, 0);

        if (d) {
            (*labels)[*n_labels] = i;
            (*n_labels)++;
        }
        next = strtok(NULL, " ,\t");
    }
    REALLOC(*labels, int, *n_labels, "ReadNextInstanceDenseFormat: *labels");

    // Check correct format
    if (next[0] != ']')
        OnErrorExit(ERR_FILE, "Input file wrong format");
}

void ReadNextInstanceSparseFormat(char *line, scalar **data, int **labels, int *n_labels, int inputs, int classes)
{
    char *next;

    // Read inputs
    CALLOC(*data, scalar, inputs, "ReadNextInstanceDenseFormat: *data");
    next = strtok(line, " ,\t");
    for (int i = 0; i < inputs; i++) {
        // Check if there is a ":" meaning a pair
        char *pair = strchr(next, ':');

        if (pair == NULL)
            // No more inputs
            break;
        else
            (*data)[strtol(next, NULL, 0)-1] = strtod(pair+1, NULL);

        next = strtok(NULL, " ,\t");
    }

    // Read labels
    MALLOC(*labels, int, classes, "ReadNextInstanceDenseFormat: *labels");
    *n_labels = 0;

    // Check correct format
    if (next[0] != '[')
        OnErrorExit(ERR_FILE, "Input file wrong format");

    next = strtok(NULL, " ,\t");

    for (int i = 0; i < classes; i++) {
        int d = (int) strtol(next, NULL, 0);

        if (d) {
            (*labels)[*n_labels] = i;
            (*n_labels)++;
        }
        next = strtok(NULL, " ,\t");
    }
    REALLOC(*labels, int, *n_labels, "ReadNextInstanceDenseFormat: *labels");

    // Check correct format
    if (next[0] != ']')
        OnErrorExit(ERR_FILE, "Input file wrong format");
}

void ConvertToSparseFormat(scalar **data, int **index, int inputs)
{
    int c = 0;

    // Count non zero values
    for (int i = 0; i < inputs; i++)
        if (fabs((*data)[i]) >= ZERO)
            c++;

    MALLOC(*index, int, c+1, "ConvertToSparseFormat: *index");

    c = 0;
    for (int i = 0; i < inputs; i++)
        if (fabs((*data)[i]) >= ZERO) {
            (*data)[c] = (*data)[i];
            (*index)[c] = i+1;
            c++;
        }
    (*index)[c] = -1;

    REALLOC(*data, scalar, c, "ConvertToSparseFormat: *data");
}

void RemoveInstancesWithNoLabelsOrNull(scalar ***data, int ***labels, int **n_labels, int *instances, int ***in_index)
{
    int removed = 0;

    for (int i = 0; i < *instances; i++)
        if ((*n_labels)[i] == 0 || (*data)[i] == NULL) {
            // Substitute by last one
            if ((*data)[i]) // If all zeros can be null
                FREE((*data)[i]);

            (*data)[i] = (*data)[*instances-1];
            (*labels)[i] = (*labels)[*instances-1];
            (*n_labels)[i] = (*n_labels)[*instances-1];

            if (in_index && *in_index) {
                FREE((*in_index)[i]);
                (*in_index)[i] = (*in_index)[*instances-1];
            }

            i--;
            (*instances)--;
            removed++;
        }

    if (removed) {
        REALLOC(*data, scalar *, *instances, "RemoveInstancesWithNoLabelsOrNull: *data");
        REALLOC(*labels, int *, *instances, "RemoveInstancesWithNoLabelsOrNull: *labels");
        REALLOC(*n_labels, int, *instances, "RemoveInstancesWithNoLabelsOrNull: *n_labels");

        if (in_index && *in_index)
            REALLOC(*in_index, int *, *instances, "RemoveInstancesWithNoLabelsOrNull: *in_index");

        printf("Removed %d instances with no labels or null\n", removed);
    }
}

void ConvertDatasetToSparseFormat(scalar **data, int ***index, int inputs, int instances)
{
    MALLOC(*index, int *, instances, "ConvertDatasetToSparseFormat: *index");

    for (int i = 0; i < instances; i++)
        ConvertToSparseFormat(&(data[i]), &((*index)[i]), inputs);
}

void ConvertToDenseFormat(scalar *instance, int *index, int inputs, scalar *dense)
{
    ZeroVector(dense, NULL, inputs);

    int i = 0;
    while (index[i] != -1) {
        dense[index[i]] = instance[i];
        i++;
    }
}

// Obtain some parameters of the dataset

void MLDatasetInfo(int tr_instances, int **tr_labels, int *tr_n_labels, int te_instances, int **te_labels, int *te_n_labels, int classes, scalar *lcard, scalar *lden, int *ldiv, scalar *pldiv,
                   FILE *out_file, scalar *MeanIR, scalar *IRLbl, scalar *CVIR, int *per_label_count, int features, scalar ml[])
{
    // Label cardinality and density
    *lcard = 0.0;

    for (int i = 0; i < tr_instances; i++)
        *lcard += tr_n_labels[i];

    for (int i = 0; i < te_instances; i++)
        *lcard += te_n_labels[i];

    *lcard /= (tr_instances+te_instances);
    *lden = *lcard/classes;

    // Label diversity
    scalar *decimal;
    MALLOC(decimal, scalar, tr_instances+te_instances, "MLDatasetInfo: decimal");

    // For label diversity, first convert to decimal
    for (int j = 0; j < tr_instances; j++) {
        decimal[j] = 0.0;
        for (int i = 0; i < tr_n_labels[j]; i++) {
            decimal[j] += pow(2.0, tr_labels[j][i]);
        }
    }

    // For label diversity, first convert to decimal
    for (int j = 0; j < te_instances; j++) {
        decimal[j+tr_instances] = 0.0;
        for (int i = 0; i < te_n_labels[j]; i++) {
            decimal[j+tr_instances] += pow(2.0, te_labels[j][i]);
        }
    }

    // Count distinct labels
    *ldiv = 1;
    for (int i = 1; i < tr_instances+te_instances; i++) {
        bool present = false;
        for (int j = 0; j < i && !present; j++)
            if (decimal[i] == decimal[j])
                present = true;

        if (!present)
            (*ldiv)++;
    }

    FREE(decimal);
    *pldiv = ((scalar)*ldiv)/(tr_instances+te_instances);

    // Imbalance ratio per label and mean Imbalance ratio
    ZeroVector(IRLbl, NULL, classes);
    for (int i = 0; i < tr_instances; i++)
        for (int j = 0; j < tr_n_labels[i]; j++)
            IRLbl[tr_labels[i][j]]++;

    for (int i = 0; i < te_instances; i++)
        for (int j = 0; j < te_n_labels[i]; j++)
            IRLbl[te_labels[i][j]]++;

    scalar maximum = MaximumOfVector(IRLbl, classes);
    *MeanIR = 0.0;

    // Single-label imbalance ratio
    if (out_file) {
        fprintf(out_file, "SL IR:");
        for (int i = 0; i < classes; i++)
            fprintf(out_file, " [%*d, %6.4lf]", WIDTH(classes), i+1, IRLbl[i]/(tr_instances+te_instances));
        fprintf(out_file, "\n");
    }

    for (int i = 0; i < classes; i++) {
        if (per_label_count)
            per_label_count[i] = IRLbl[i];
        *MeanIR += (IRLbl[i] = IRLbl[i] > ZERO ? maximum/IRLbl[i] : 0.0);
    }

    *MeanIR /= classes;

    // Pairwise imbalance ratio
    scalar **ci;
    AllocateScalarMatrix(&ci, classes, classes);
    for (int i = 0; i < classes; i++) {
        // CI of label
        ci[i][i] = 0;
        for (int k = 0; k < tr_instances; k++)
            if (LabelInList(i, tr_labels[k], tr_n_labels[k]))
                ci[i][i]++;

        for (int k = 0; k < te_instances; k++)
            if (LabelInList(i, te_labels[k], te_n_labels[k]))
                ci[i][i]++;

        ci[i][i] /= tr_instances+te_instances;

        for (int j = 0; j < i; j++) {
            // CI of pairwise labels
            scalar c_i = 0, c_j = 0;
            for (int k = 0; k < tr_instances; k++) {
                bool rel_i = LabelInList(i, tr_labels[k], tr_n_labels[k]);
                bool rel_j = LabelInList(j, tr_labels[k], tr_n_labels[k]);

                if (rel_i != rel_j) {
                    if (rel_i)
                        c_i++;
                    else
                        c_j++;
                }
            }

            for (int k = 0; k < te_instances; k++) {
                bool rel_i = LabelInList(i, te_labels[k], te_n_labels[k]);
                bool rel_j = LabelInList(j, te_labels[k], te_n_labels[k]);

                if (rel_i != rel_j) {
                    if (rel_i)
                        c_i++;
                    else
                        c_j++;
                }
            }

            ci[i][j] = c_i/c_j;
            ci[j][i] = c_j/c_i;
        }
    }

    if (out_file) {
        fprintf(out_file, "CI Pairwise: \n");
        PrintMatrix(ci, classes, classes, out_file, 3, 4);
    }
    FreeScalarMatrix(&ci, classes);

    // Coefficient of variation of IRLbl
    scalar IRLbl_sigma = 0.0;
    for (int i = 0; i < classes; i++)
        IRLbl_sigma += pow(IRLbl[i]-(*MeanIR), 2.0)/(classes-1);

    *CVIR = sqrt(IRLbl_sigma)/(*MeanIR);

    // Percentage of multi-label instances
    ZeroVector(ml, NULL, classes+1);

    for (int i = 0; i < tr_instances; i++)
        ml[tr_n_labels[i]]++;

    for (int i = 0; i < te_instances; i++)
        ml[te_n_labels[i]]++;

    for (int i = 0; i <= classes; i++)
        ml[i] /= tr_instances+te_instances;

    // If output file given, write values
    if (out_file) {
        fprintf(out_file, "@Instances: %d\n", tr_instances+te_instances);
        fprintf(out_file, "@Features: %d\n", features);
        fprintf(out_file, "@Labels: %d\n", classes);
        fprintf(out_file, "@Label cardinality: %lf\n", *lcard);
        fprintf(out_file, "@Label density: %lf\n", *lden);
        fprintf(out_file, "@Label diversity: %d\n", *ldiv);
        fprintf(out_file, "@Proportion of distinct labels: %lf\n", *pldiv);

        fprintf(out_file, "IRLbl:");
        for (int i = 0; i < classes; i++)
            fprintf(out_file, " [%*d, %6.4lf]", WIDTH(classes), i+1, IRLbl[i]);
        fprintf(out_file, "\n");
        fprintf(out_file, "@MeanIR: %lf\n", *MeanIR);
        fprintf(out_file, "@CVIR: %lf\n", *CVIR);
        fprintf(out_file, "@Multi-label instances: %lf\n", 1.0-ml[0]-ml[1]);
        fprintf(out_file, "Number of label distribution:           ");
        for (int i = 1; i <= classes; i++)
            fprintf(out_file, " [%*d, %6.4lf]", WIDTH(classes), i, ml[i]);
        fprintf(out_file, "\n");
        fprintf(out_file, "Number of label cumulative distribution: [%*d, %6.4lf]", WIDTH(classes), 1, ml[1]);
        for (int i = 2; i <= classes; i++) {
            ml[i] += ml[i-1];
            fprintf(out_file, " [%*d, %6.4lf]", WIDTH(classes), i, ml[i]);
        }
        fprintf(out_file, "\n");
    }
}

void SaveSelectedDataset(scalar **data, int **labels, int instances, int features, int *n_labels, int classes, int **index, int *S, char *name)
{
    FILE *f;

    if ((f = fopen(name, "wt")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open selected instances file");

    // Count instances
    int count = instances;
    if (S) {
        count = 0;
        for (int i = 0; i < instances; i++)
            count += MAX(0, S[i]);
    }

    fprintf(f, "[MULTILABEL, %s]\n", index ? "sparse" : "dense");
    fprintf(f, "$ %d\n", count);
    fprintf(f, "$ %d\n", features);
    fprintf(f, "$ %d\n", classes);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < (S ? S[i] > 0 : 1); j++) {
            // Sparse format
            if (index) {
                int next = 0;
                while (index[i][next] != -1) {
                    fprintf(f, "%d:%g ", index[i][next], data[i][next]);
                    next++;
                }
            } else { // Dense format
                for (int k = 0; k < features; k++)
                    fprintf(f, "%g ", data[i][k]);
            }

            // Labels
            fprintf(f, "[ ");
            for (int k = 0; k < classes; k++)
                fprintf(f, "%d ", LabelInList(k, labels[i], n_labels[i]) ? 1 : 0);
            fprintf(f, " ]\n");
        }
}
