/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Dependent Binary Relevance method
// E. Montañes, R. Senge, J. Barranquero, J. R. Quevedo, J. J. del Coz, E. Hüllermeier (2014) "Dependent binary relevance models for multi-label classification"
// Pattern Recognition 47:1494–1508.

void TrainDependentBinaryRelevanceModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_DBR *dbr, int *subspace, bool cv,
        t_Classifier classifier, int *subset, int **SS, int **index)
{
    int **b_labels, *l_subspace;

    // Test if valid BR classification model
    if (classifier == NN || classifier == kNN || classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "DBR classification method");

    // If no subspace select all
    MALLOC(l_subspace, int, inputs+classes, "TrainDependentBinaryRelevanceModels: l_subspace");
    if (!subspace)
        FillVector(NULL, l_subspace, inputs+classes, 1.0);
    else {
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = subspace[i];
        for (int i = inputs; i < inputs+classes; i++)
            l_subspace[i] = 1;
    }

    // Create binary labels
    AllocateIntegerMatrix(&b_labels, classes, instances);
    ZeroIntegerMatrix(b_labels, classes, instances);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            b_labels[labels[i][j]][i] = 1;

    dbr->method = classifier;
    if (!IsEnsemble(classifier)) {
        MALLOC(dbr->models, t_Learner, classes, "TrainDependentBinaryRelevanceModels: dbr->models");
        MALLOC(dbr->predictors, t_Learner, classes, "TrainDependentBinaryRelevanceModels: dbr->predictors");
    } else {
        MALLOC(dbr->mensembles, t_Ensemble, classes, "TrainDependentBinaryRelevanceModels: dbr->mensembles");
        MALLOC(dbr->pensembles, t_Ensemble, classes, "TrainDependentBinaryRelevanceModels: dbr->pensembles");
    }

    int *count = NULL;
    if (index) {
        MALLOC(count, int, instances, "TrainDependentBinaryRelevanceModels: count");
        for (int i = 0; i < instances; i++) {
            count[i] = 0;
            for (int j = 0; index[i][j] != -1; j++)
                count[i]++;
        }
    }

    // Train a model for every label
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < classes; i++) {
        t_Parameters lparm;

        // If undersample, use local subset
        int *S = NULL;
        MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

        if (SS)
            memcpy(S, SS[i], sizeof(int)*instances);
        else if (subset)
            memcpy(S, subset, sizeof(int)*instances);
        else
            for (int j = 0; j < instances; j++)
                S[j] = 1;

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, b_labels[i], S, instances);

        // Cross-validate hyper-parameters?
        lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, data, instances, b_labels[i], l_subspace, inputs, 2, classifier, 1, S, index);

        // Train BR model
        SHOWPROGRESS(parm->show_progress > 0, "DBR training 1st level model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);
        ClassifierTraining(&lparm, classifier, data, instances, b_labels[i], inputs, 2, index, l_subspace, S, true, &(dbr->models[i]), &(dbr->mensembles[i]));

        FREE(S);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // Train a model for every label using remaining labels as inputs
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < classes; i++) {
        // If undersample, use local subset
        int *S = NULL;
        MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

        if (SS)
            memcpy(S, SS[i], sizeof(int)*instances);
        else if (subset)
            memcpy(S, subset, sizeof(int)*instances);
        else
            FillVector(NULL, S, instances, 1.0);

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, b_labels[i], S, instances);

        // Train DBR model
        scalar **t_data;
        int **t_index = NULL;

        AllocateScalarMatrix(&t_data, instances, inputs+classes-1);
        if (index) {
            AllocateIntegerMatrix(&t_index, instances, inputs+classes);
            CopySparseData(index, t_index, data, t_data, instances);

            for (int j = 0; j < instances; j++) {
                int c = count[j];
                // Just in case all zero
                for (int k = 0; k < classes; k++)
                    if (b_labels[k][j] && k != i) {
                        t_data[j][c] = 1.0;
                        t_index[j][c] = inputs+k+1;
                        c++;
                    }
                t_index[j][c] = -1;
            }
        } else {
            CopyScalarMatrix(data, t_data, instances, inputs);
            for (int j = 0; j < instances; j++) {
                for (int k = 0; k < i; k++)
                    t_data[j][k+inputs] = b_labels[k][j];
                for (int k = i+1; k < classes; k++)
                    t_data[j][k+inputs-1] = b_labels[k][j];
            }
        }

        // Cross-validate hyper-parameters?
        t_Parameters lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, t_data, instances, b_labels[i], l_subspace, inputs+classes-1, 2, classifier, 1, S, t_index);

        // Train BR model
        SHOWPROGRESS(parm->show_progress > 0, "DBR training 2nd level model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);
        ClassifierTraining(&lparm, classifier, t_data, instances, b_labels[i], inputs+classes-1, 2, t_index, l_subspace, S, true, &(dbr->predictors[i]), &(dbr->pensembles[i]));

        FREE(S);
        FreeScalarMatrix(&t_data, instances);
        if (index)
            FreeIntegerMatrix(&t_index, instances);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FreeIntegerMatrix(&b_labels, classes);
    FREE(l_subspace);
    if (index)
        FREE(count);
}

void DependentBinaryRelevanceClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, int n_labels,  int *n_predicted, int inputs,
                                              int *query_index, t_DBR *dbr, int *subspace, int *S, int **SS, int **index)
{
    *n_predicted = 0;
    scalar output[n_labels], *iquery;
    int *iquery_index = NULL, count = 0;

    // If no subspace select all
    int *l_subspace;
    MALLOC(l_subspace, int, inputs+n_labels, "TrainDependentBinaryRelevanceModels: l_subspace");
    if (!subspace)
        FillVector(NULL, l_subspace, inputs+n_labels, 1.0);
    else {
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = subspace[i];
        for (int i = inputs; i < inputs+n_labels; i++)
            l_subspace[i] = 1;
    }

    MALLOC(iquery, scalar, inputs+n_labels, "DependentBinaryRelevanceClassifyInstance: data");
    if (query_index) {
        MALLOC(iquery_index, int, inputs+n_labels+1, "DependentBinaryRelevanceClassifyInstance: iquery_index");
        for (int i = 0; query_index[i] != -1; i++) {
            iquery_index[i] = query_index[i];
            iquery[i] = query[i];
            count++;
        }
        iquery_index[count] = -1;
    } else
        memcpy(iquery, query, sizeof(scalar)*inputs);

    // Output of all binary classifiers
    for (int i = 0; i < n_labels; i++)
        Classifier2ClassOutput(parm, dbr->method, query, inputs, 2, query_index, l_subspace, NULL, -1, NULL, SS ? SS[i] : S, &(dbr->models[i]), 1, &output[i], index, &(dbr->mensembles[i]));

    // Output of all dependent binary classifiers
    for (int i = 0; i < n_labels; i++) {
        // Add outputs of previous classifiers
        if (!query_index) {
            for (int k = 0; k < i; k++)
                iquery[k+inputs] = output[k];
            for (int k = i+1; k < n_labels; k++)
                iquery[k+inputs-1] = output[k];

        } else {
            int c = count;
            for (int k = 0; k < n_labels; k++)
                if (fabs(output[k]) > ZERO && k != i) {
                    iquery[c] = 1.0;
                    iquery_index[c] = inputs+k+1;
                    c++;
                }
            iquery_index[c] = -1;
        }

        int category = Classifier2ClassOutput(parm, dbr->method, iquery, inputs+n_labels-1, 2, iquery_index, l_subspace, NULL, -1, NULL, SS ? SS[i] : S, &(dbr->predictors[i]), 1, &f[i],
                                                    index, &(dbr->pensembles[i]));

        if (category) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }
    }

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *classification = *ranks = true;

    FREE(iquery);
    FREE(l_subspace);
    if (query_index)
        FREE(iquery_index);
}

void FreeDependentBinaryRelevance(const t_Parameters *parm, t_DBR *dbr, int classes, int inputs)
{
    for (int i = 0; i < classes; i++) {
        ClassifierFree(parm, &(dbr->models[i]), dbr->method, 2, inputs, &(dbr->mensembles[i]));
        ClassifierFree(parm, &(dbr->predictors[i]), dbr->method, 2, inputs, &(dbr->pensembles[i]));
    }
    FREE(dbr->models);
    FREE(dbr->predictors);
}

void CopySparseIndex(int **or, int **dest, int n)
{
    for (int i = 0; i < n; i++)  {
        int c = 0;
        for (; or[i][c] != -1; c++)
            dest[i][c] = or[i][c];
        dest[i][c] = -1;
    }
}

void CopySparseData(int **i_or, int **i_dest, scalar **d_or, scalar **d_dest, int n)
{
    for (int i = 0; i < n; i++)  {
        int c = 0;
        for (; i_or[i][c] != -1; c++) {
            i_dest[i][c] = i_or[i][c];
            d_dest[i][c] = d_or[i][c];
        }
        i_dest[i][c] = -1;
    }
}
