/***************************************************************************
 *   Copyright (C) 2012 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <time.h>
#include <math.h>
#include <stdbool.h>
#include <omp.h>
#include "ml.h"
#include "maths.h"
#include "macros.h"

// Differential evolution for ML-k'sNN

scalar DifferentialEvolutionInstanceSelection(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *selection,
        t_MLClassifier method)
{
    scalar **x, **x_new, *f, **u, *performance;
    int neighbors_cache_size = 0;
    cache_t **neighbors_cache = NULL;

    // Available methods
    if (method == CC || method == BR)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CHC IS for BR/CC methods");

    if (method == LAMLksNN || method == MLksNN || method == LAMLkNN)
        OnErrorExit(ERR_PARAM, "LAML-kNN/LAML-k'sNN/ML-k'sNN not suitable for differential evolution instance selection");

    if (parm->ml_ins_sel_per_label)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Differential evolution for instance selection per label");

    if (parm->multiselection &&  parm->ml_relative_multiselection)
        OnErrorExit(ERR_PARAM, "Relative multi-selection should not be here");

    if (parm->multiselection && parm->multiis_adjust_initial_percentage)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Adjust of initial percentage for DE MS");

    // Allocate memory for weights
    AllocateScalarMatrix(&x, parm->de_np, instances+1);
    AllocateScalarMatrix(&x_new, parm->de_np, instances+1);
    AllocateScalarMatrix(&u, parm->de_np, instances+1);

    // If CV get folds
    int *count = NULL, *partition = NULL;
    if (parm->cv_folds) {
        MALLOC(count, int, parm->cv_folds, "DifferentialEvolutionInstanceSelection: count");
        MALLOC(partition, int, instances, "DifferentialEvolutionInstanceSelection: partition");

        RandomPartition(partition, count, instances, parm->cv_folds);
    }

    // Create initial population
    for (int i = 0; i < parm->de_np; i++)
        for (int j = 0; j < instances; j++)
            x[i][j] = RANDOM(0.0, 1.0) < parm->chc_init_prob ? (parm->multiselection ? RANDOM(1, parm->ml_max_k/2) : 1) : 0;

    // k is last value of individual
    for (int i = 0; i < parm->de_np; i++)
        x[i][instances] = RANDOM(parm->ml_min_k, parm->ml_max_k+1);

    // Cache of neighbors?
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
    else if (parm->ml_loo)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation with LOO without cache");

    // Initial evaluation
    int np = parm->de_np;   // Needed for OpenMP parallel for
    MALLOC(f, scalar, parm->de_np, "DifferentialEvolutionInstanceSelection: f");
    MALLOC(performance, scalar, parm->de_np, "DifferentialEvolutionInstanceSelection: f");

    t_MLkNN *mlknn;
    int **lsel;
    MALLOC(mlknn, t_MLkNN, np, "DifferentialEvolutionInstanceSelection: mlknn");
    MALLOC(lsel, int *, np, "DifferentialEvolutionInstanceSelection: lsel");
    for (int i = 0; i < np; i++)
        MALLOC(lsel[i], int, instances, "DifferentialEvolutionInstanceSelection: lsel[i]");

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < np; i++) {
        for (int j = 0; j < instances; j++)
            lsel[i][j] = (int) x[i][j];
        f[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, lsel[i], neighbors_cache, NULL, NULL, index, &performance[i], NULL, NULL, MLkNN,
                                         (void *) &mlknn[i], NULL, 0, NULL, (int) x[i][instances], NULL, 0, NULL, NULL, instances);
    }

    // Time
    time_t init_time = time(NULL);

    for (int g = 0; g < parm->generations ; g++) {
        // If time limit end
        if (parm->time_limit && ((time(NULL) - init_time) > parm->time_limit)) {
            fprintf(parm->out_file, "Reached time limit at generation: %d\n", g+1);
            break;
        }

        // Generate new population by differential mutation, recombination and selection
        // Differential mutation
        for (int i = 0; i < parm->de_np; i++) {
            int one, two, three;
            do {
                one = RANDOM(0, parm->de_np);
            } while (one == i);
            do {
                two = RANDOM(0, parm->de_np);
            } while (two == i || two == one);
            do {
                three = RANDOM(0, parm->de_np);
            } while (three == i || three == one || three == two);

            // Differential mutation
            DifferentialMutation(x_new[i], x[one], x[two], x[three], instances, parm->de_f, 0, (parm->multiselection ? RANDOM(1, parm->ml_max_k/2) : 1));
            DifferentialMutation(x_new[i]+instances, x[one]+instances, x[two]+instances, x[three]+instances, 1, parm->de_f, parm->ml_min_k, parm->ml_max_k/2);
        }

        // Recombination and selection
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < np; i++) {
            scalar f_u, perf;

            Recombination(u[i], x[i], x_new[i], parm->de_cr, instances, 0, (parm->multiselection ? RANDOM(1, parm->ml_max_k/2) : 1));
            Recombination(u[i]+instances, x[i]+instances, x_new[i]+instances, parm->de_cr, 1, parm->ml_min_k, parm->ml_max_k/2);

            // Evaluate target and add to the population if better or at least equal
            for (int j = 0; j < instances; j++)
                lsel[i][j] = (int) u[i][j];
            f_u = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, lsel[i], neighbors_cache, NULL, NULL, index, &perf, NULL, NULL, MLkNN,
                                            (void *) &mlknn[i], NULL, g, NULL, (int) u[i][instances], NULL, 0, NULL, NULL, instances);

            if (f_u >= f[i]) {
                f[i] = f_u;
                performance[i] = perf;

                // Exchange, it is faster
                scalar *sw = u[i];
                u[i] = x[i];
                x[i] = sw;
            }
        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(f, 0, np);

            for (int j = 0; j < instances; j++)
                lsel[best][j] = (int) x[best][j];

            SHOWPROGRESS(true, "%*d g's best FITNESS, PERFOMANCE, REDUCTION, k: %6.4lf, %6.4lf, %6.4lf, %2d\n", WIDTH(parm->generations), g+1, f[best], performance[best],
                         Reduction(lsel[best], x[best], instances), (int) x[best][instances]);
            VERBOSE(parm->show_progress > 1, parm->out_file, "%*d generation best individual FITNESS (MEAN): %6.4lf (%6.4lf)\n", WIDTH(parm->generations), g+1, f[best],
                    AverageScalarVector(f, parm->de_np));
        }
    }

    // Get best individual and copy to results
    int best = OrderedSelection(f, 0, np);

    // Convert to K's
    scalar reduction = 0.0;
    for (int j = 0; j < instances; j++)
        if (!(selection[j] = (int) x[best][j]))
            reduction++;
    ((t_Parameters *) parm)->mlknn_k = x[best][instances];

    SHOWPROGRESS(parm->show_progress > 0, "FINAL BEST f, performance, reduction: %6.4lf, %6.4lf, %6.4lf\n", f[best], performance[best], reduction/instances);

    // Free memory
    FREE(f);
    FreeScalarMatrix(&x, parm->de_np);
    FreeScalarMatrix(&x_new, parm->de_np);
    FreeScalarMatrix(&u, parm->de_np);
    if (parm->cv_folds) {
        FREE(count);
        FREE(partition);
    }
    if (neighbors_cache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }

    for (int i = 0; i < np; i++) {
        FreeMLClassifier(parm, (void *) &mlknn[i], MLkNN, classes, inputs, instances, NULL);
        FREE(lsel[i]);
    }
    FREE(mlknn);
    FREE(lsel);

    return reduction/instances;
}

scalar Reduction(int *selection, scalar *x, int instances)
{
    scalar reduction = 0.0;

    if (selection) {
        for (int i = 0; i < instances; i++)
            if (!selection[i])
                reduction++;
    } else {
        for (int i = 0; i < instances; i++)
            if (x[i] <= ZERO)
                reduction++;
    }

    return reduction/instances;
}
