/***************************************************************************
 *   Copyright (C) 2012 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <time.h>
#include <math.h>
#include <stdbool.h>
#include <omp.h>
#include "ml.h"
#include "maths.h"
#include "macros.h"

// Differential evolution for ML-k'sNN

void DifferentialEvolutionMLksNN(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *ks)
{
    scalar **x, **x_new, *f, **u;
    int *selection, neighbors_cache_size = 0;
    cache_t **neighbors_cache = NULL;

    if (parm->mlksnn_use_window)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Window for differential evolution ML-ksNN");

    // Allocate memory for weights
    AllocateScalarMatrix(&x, parm->de_np, classes);
    AllocateScalarMatrix(&x_new, parm->de_np, classes);
    AllocateScalarMatrix(&u, parm->de_np, classes);

    // No selection
    MALLOC(selection, int, instances, "EvolveKPerLabel: selection");
    for (int i = 0; i < instances; i++)
        selection[i] = 1;

    // If CV get folds
    int *count = NULL, *partition = NULL;
    if (parm->cv_folds) {
        MALLOC(count, int, parm->cv_folds, "EvolveKPerLabel: count");
        MALLOC(partition, int, instances, "EvolveKPerLabel: partition");

        RandomPartition(partition, count, instances, parm->cv_folds);
    }

    // Create initial population
    for (int i = 0; i < parm->de_np; i++)
        for (int j = 0; j < classes; j++)
            x[i][j] = RANDOM(parm->ml_min_k, parm->ml_max_k+1);

    // Cache of neighbors?
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
    else if (parm->ml_loo)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation with LOO without cache");

    // Initial evaluation
    int np = parm->de_np;   // Needed for OpenMP parallel for
    MALLOC(f, scalar, parm->de_np, "DifferentialEvolutionMLksNN: f");
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < np; i++) {
        t_MLksNN mlksnn;
        int ks[classes];
        for (int j = 0; j < classes; j++)
            ks[j] = (int) x[i][j];
        f[i] = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, 0, ks, 0,
                                             NULL, NULL)
               - RegularizationTerm(parm, ks, classes);
        FreeMLksNN(&mlksnn, classes);
    }

    // Time
    time_t init_time = time(NULL);

    for (int g = 0; g < parm->generations ; g++) {
        // If time limit end
        if (parm->time_limit && ((time(NULL) - init_time) > parm->time_limit)) {
            fprintf(parm->out_file, "Reached time limit at generation: %d\n", g+1);
            break;
        }

        // Generate new population by differential mutation, recombination and selection
        // Differential mutation
        for (int i = 0; i < parm->de_np; i++) {
            int one, two, three;
            do {
                one = RANDOM(0, parm->de_np);
            } while (one == i);
            do {
                two = RANDOM(0, parm->de_np);
            } while (two == i || two == one);
            do {
                three = RANDOM(0, parm->de_np);
            } while (three == i || three == one || three == two);

            // Differential mutation
            DifferentialMutation(x_new[i], x[one], x[two], x[three], classes, parm->de_f, parm->ml_min_k, parm->ml_max_k);
        }

        // Recombination and selection
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < np; i++) {
            t_MLksNN mlksnn;
            scalar f_u;

            Recombination(u[i], x[i], x_new[i], parm->de_cr, classes, (scalar) parm->ml_min_k, (scalar) parm->ml_max_k);

            // Evaluate target and add to the population if better or at least equal
            int ks[classes];
            for (int j = 0; j < classes; j++)
                ks[j] = (int) u[i][j];
            f_u = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, 0, ks,
                                                0, NULL, NULL) - RegularizationTerm(parm, ks, classes);
            FreeMLksNN(&mlksnn, classes);

            if (f_u >= f[i]) {
                f[i] = f_u;

                // Exchange, it is faster
                scalar *sw = u[i];
                u[i] = x[i];
                x[i] = sw;
            }
        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(f, 0, parm->de_np);

            SHOWPROGRESS(true, "%*d g's best FITNESS: %6.4lf\n", WIDTH(parm->generations), g+1, f[best]);
            VERBOSE(parm->show_progress > 1, parm->out_file, "%*d generation best individual FITNESS (MEAN): %6.4lf (%6.4lf)\n", WIDTH(parm->generations), g+1, f[best],
                    AverageScalarVector(f, parm->de_np));
        }
    }

    // Get best individual and copy to results
    int best = 0;
    for (int i = 1; i < parm->de_np; i++)
        if (f[best] < f[i])
            best = i;

    // Convert to K's
    for (int i = 0; i < classes; i++)
        ks[i] = (int) x[best][i];

    SHOWPROGRESS(parm->show_progress > 0, "FINAL BEST f: %6.4lf\n", f[best]);

    // Free memory
    FREE(selection);
    FREE(f);
    FreeScalarMatrix(&x, parm->de_np);
    FreeScalarMatrix(&x_new, parm->de_np);
    FreeScalarMatrix(&u, parm->de_np);
    if (parm->cv_folds) {
        FREE(count);
        FREE(partition);
    }
    if (neighbors_cache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }
}

// Perform differential mutation
void DifferentialMutation(scalar *v, scalar *x_1, scalar *x_2, scalar *x_3, int length, scalar F, scalar min, scalar max)
{
    for (int i = 0; i < length; i++)
        v[i] = x_1[i]+F*(x_2[i] - x_3[i]);

    // Check limits
    for (int j = 0; j < length; j++)
        if (v[j] < min)
            v[j] = min;
        else if (v[j] > max)
            v[j] = max;
}

// Perform recombination
void Recombination(scalar *u, scalar *x, scalar *v, scalar CR, int length, scalar min, scalar max)
{
    int I_rand = RANDOM(0, length);

    for (int i = 0; i < length; i++)
        if (RANDOM(0.0, 1.0) < CR || i == I_rand)
            u[i] = v[i];
        else
            u[i] = x[i];

    // Check limits
    for (int j = 0; j < length; j++)
        if (u[j] < min)
            u[j] = min;
        else if (u[j] > max)
            u[j] = max;
}
