/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "common.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "maths.h"
#include "ensemble.h"

// Label Dependent Multi-label k-NN
// 6th European Signal Processing Conference (EUSIPCO 2008), Lausanne, Switzerland, August 25-29, 2008, copyright by EURASIP
// MULTI-LABEL CLASSIFICATION ALGORITHM DERIVED FROM K-NEAREST NEIGHBOR RULE WITH LABEL DEPENDENCIES
// Zoulficar Younes, Fahed Abdallah, and Thierry Denœux

void DMultiLabelkNearestNeighbors(const t_Parameters *parm, scalar *query, scalar **data, int instances, int **labels, int *n_labels, int classes, int *predicted_labels, int *predicted_n_labels,
                                  int *S, scalar *f, int inputs, int *query_index, int **index, cache_t **neighbors_cache, t_DMLkNN *dmlknn, int **SS, int *subspace)
{
    // Classify query instance
    *predicted_n_labels = 0;
    int c[classes];

    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial instance selection for DML-kNN");

    int neighbors[dmlknn->k];
    GetkNeighborsForML(query, data, instances, index, inputs, query_index, dmlknn->k, neighbors, S, neighbors_cache, -1, false, NULL, classes, NULL, -1, subspace);

    ZeroVector(NULL, c, classes);
    for (int j = 0; j < dmlknn->k; j++)
        if (neighbors[j] != -1)
            for (int i = 0; i < n_labels[neighbors[j]]; i++)
                c[labels[neighbors[j]][i]]++;

    scalar V[classes], Vprime[classes];
    ZeroVector(V, NULL, classes);
    ZeroVector(Vprime, NULL, classes);

    for (int i = 0; i < instances; i++)
        if (!S || S[i]) {
            // Check restriction
            bool restriction = true;
            for (int j = 0; j < classes && restriction; j++)
                if (!(c[j]-parm->dmlknn_delta <= dmlknn->c[i][j] && dmlknn->c[i][j] <= c[j]+parm->dmlknn_delta))
                    restriction = false;

            if (restriction)
                for (int j = 0; j < classes; j++)
                    if (c[j] == dmlknn->c[i][j]) {
                        if (LabelInList(j, labels[i], n_labels[i]))
                            V[j]++;
                        else
                            Vprime[j]++;
                    }
        }

    for (int l = 0; l < classes; l++) {
        scalar P1 = dmlknn->prior[l]*((parm->mlknn_s + V[l])/(parm->mlknn_s*classes+dmlknn->C[l]));
        scalar P0 = (1.0-dmlknn->prior[l])*((parm->mlknn_s + Vprime[l])/(parm->mlknn_s*classes+dmlknn->Cprime[l]));

        // Check if label relevant
        if (P1 > P0)
            predicted_labels[(*predicted_n_labels)++] = l;

        // Get confidence
        f[l] = P1/(P1+P0);
    }

    NormalizeVector(f, classes);
}

void TrainDMultiLabelkNearestNeighbors(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int *S, int inputs, int **index, cache_t **neighbors_cache,
                                       t_DMLkNN *dmlknn, int **SS, int k, bool create, int *subspace)
{
    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial instance selection for DML-kNN");

    // First create classifier
    if (create)
        CreateDMLkNNClassifier(dmlknn, classes, k, instances);

    // Compute prior probabilities
    for (int i = 0; i < classes; i++)
        dmlknn->prior[i] = 0.0;

    // Selection?
    if (S) {
        int actual_instances = 0;
        for (int i = 0; i < instances; i++)
            if (S[i] > 0) {
                actual_instances += S[i];
                for (int j = 0; j < n_labels[i]; j++)
                    dmlknn->prior[labels[i][j]] += S[i];
            }

        for (int i = 0; i < classes; i++) {
            dmlknn->C[i] = dmlknn->prior[i];
            dmlknn->Cprime[i] = actual_instances - dmlknn->C[i];
            dmlknn->prior[i] = (parm->mlknn_s+dmlknn->prior[i])/(2.0*parm->mlknn_s+actual_instances);
        }
    } else {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < n_labels[i]; j++)
                dmlknn->prior[labels[i][j]]++;

        for (int i = 0; i < classes; i++) {
            dmlknn->C[i] = dmlknn->prior[i];
            dmlknn->Cprime[i] = instances - dmlknn->C[i];
            dmlknn->prior[i] = (parm->mlknn_s+dmlknn->prior[i])/(2.0*parm->mlknn_s+instances);
        }
    }

    // Identify the neighbors and c() of every training instance
    int neighbors[dmlknn->k];
    ZeroIntegerMatrix(dmlknn->c, instances, classes);
    for (int i = 0; i < instances; i++)
        if (!S || S[i]) {
            GetkNeighborsForML(data[i], data, instances, index, inputs, index ? index[i] : NULL, dmlknn->k, neighbors, S, neighbors_cache, -1, false, NULL, classes, NULL, -1, subspace);
            for (int j = 0; j < dmlknn->k; j++)
                if (neighbors[j] != -1)
                    for (int l = 0; l < n_labels[neighbors[j]]; l++)
                        dmlknn->c[i][labels[neighbors[j]][l]]++;
        }
}

void FreeDMLkNN(t_DMLkNN *dmlknn, int instances)
{
    FREE(dmlknn->prior);
    FREE(dmlknn->C);
    FREE(dmlknn->Cprime);
    FreeIntegerMatrix(&(dmlknn->c), instances);
}

void CreateDMLkNNClassifier(t_DMLkNN *dmlknn, int n_labels, int k, int instances)
{
    dmlknn->k = k;

    MALLOC(dmlknn->prior, scalar, n_labels, "CreateDMLkNNClassifier: dmlknn->prior");
    MALLOC(dmlknn->C, scalar, n_labels, "CreateDMLkNNClassifier: dmlknn->C");
    MALLOC(dmlknn->Cprime, scalar, n_labels, "CreateDMLkNNClassifier: dmlknn->Cprimer");
    AllocateIntegerMatrix(&(dmlknn->c), instances, n_labels);
}
