/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "maths.h"
#include "ensemble.h"

// Implements  distance-weighted k-nearest neighbour (DWkNN)

// AN EMPIRICAL COMPARISON OF WEIGHTING FUNCTIONS FOR MULTI-LABEL D ISTANCE-WEIGHTED K-NEAREST NEIGHBOUR METHOD, Jianhua Xu
// David Bracewell, et al. (Eds): AIAA 2011,CS & IT 03, pp. 13–20, 2011.

// No training is needed
void TrainDWkNN(const t_Parameters *parm, t_DWkNN *dwknn)
{
    dwknn->k = parm->mlknn_k;
}

void DWkNNClassify(t_MLClassifier method, t_DWkNN *dwknn, scalar *query, scalar **data, int **index, int inputs, int *query_index, cache_t **neighbors_cache, int **SS, int *S, int **labels,
                   int *n_labels, int classes, int *subspace, int *n_predicted, int *h, scalar *f, int instances)
{
    // Get neighbours
    int neighbors[dwknn->k];
    GetkNeighborsForML(query, data, instances, index, inputs, query_index, dwknn->k, neighbors, S, neighbors_cache, -1, false, SS, classes, NULL, -1, subspace);

    // Select last in case too few selected
    int k = 0;
    while (k < dwknn->k && neighbors[k] != -1)
        k++;

    // Weight of votes
    scalar w[k], d[k];
    for (int i = 0; i < k; i++)
        d[i] = EuclideanDistanceSparseDataSubspace(query, data[neighbors[i]], inputs, query_index, index ? index[neighbors[i]] : NULL, subspace);

    switch(method) {
        case DWKNND:
            for (int i = 0; i < k; i++)
                w[i] = Dudani(d[i], d[0], d[k-1]);
            break;

        case DWKNNM:
            for (int i = 0; i < k; i++)
                w[i] = Macleod(d[i], d[0], d[k-1]);
            break;

        case DWKNNZ:
            for (int i = 0; i < k; i++)
                w[i] = Zavrel(d[i]);
            break;

        default:
            OnErrorExit(ERR_PARAM, "Should not be in DWKNN function");
    }

    // Weight votes
    ZeroVector(f, NULL, classes);
    for (int i = 0; i < k; i++)
        for (int l = 0; l < classes; l++)
            if (LabelInList(l, labels[neighbors[i]], n_labels[neighbors[i]]))
                f[l] += w[i];
            else
                f[l] -= w[i];

    // PrintVector(f, NULL, classes, stdout);

    *n_predicted = 0;
    for (int i = 0; i < classes; i++)
        if (f[i] >= 0.0)
            h[(*n_predicted)++] = i;

    NormalizeVector(f, classes);
}

// Different weighting functions
scalar Dudani(scalar d, scalar d_nearest, scalar d_furthest)
{
    if (fabs(d_furthest-d_nearest) <= ZERO)
        return 1.0;
    else
        return (d_furthest-d)/(d_furthest-d_nearest);
}

// s is k, k+1, ... neighbor, in this implementation s=k
#define ALPHA   1.0
scalar Macleod(scalar d, scalar d_nearest, scalar d_s)
{
    if (fabs(d_s-d_nearest) <= ZERO)
        return 1.0;
    else
        return ((d_s-d)+ALPHA*(d_s-d_nearest))/((1+ALPHA)*(d_s-d_nearest));
}

#define BETA    1.0
scalar Zavrel(scalar d)
{
    return exp(-ALPHA*pow(d, BETA));
}
