/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Classifier chains method

void TrainECCModel(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_ECC *ecc, int *subspace, bool cv,  t_Classifier classifier,
                   int *subset, int **SS, int **index)
{
    // Test if valid CC classification model
    if (classifier == kNN || classifier == IBT || classifier == OVO || classifier == NN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ECC classification method");

    // Set classifier
    ecc->T = parm->ecc_T;
    MALLOC(ecc->ccs, t_CC, ecc->T, "TrainECCModel: ecc->ccs");

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < ecc->T; i++) {
        int chain[classes];
        int **l_labels, *l_n_labels, **l_index = NULL, *l_subset = NULL;
        scalar **l_data;

        // Generate CC's'
        for (int i = 0; i < classes; i++)
            chain[i] = i;

        if (parm->ecc_with_bagging) {
            if (SS)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial instance selection for ECC wiht bagging");

            MALLOC(l_data, scalar *, instances, "TrainECCModel: l_data");
            MALLOC(l_labels, int *, instances, "TrainECCModel: l_labels");
            MALLOC(l_n_labels, int, instances, "TrainECCModel: l_n_labels");
            if (subset)
                MALLOC(l_subset, int, instances, "TrainECCModel: l_subset");
            if (index)
                MALLOC(l_index, int *, instances, "TrainECCModel: l_index");

            // Random bagging sample
            for (int i = 0; i < instances; i++) {
                int next = RANDOM(0, instances);
                l_data[i] = data[next];
                l_labels[i] = labels[next];
                l_n_labels[i] = n_labels[next];
                if (subset)
                    l_subset[i] = subset[next];
                if (index)
                    l_index[i] = index[next];
            }
        } else {
            l_data = data;
            l_index = index;
            l_labels = labels;
            l_n_labels = n_labels;
            l_subset = subset;
        }

        // New random chain
        RandomizeVector(chain, classes);

        SHOWPROGRESS(parm->show_progress > 0, "ECC training member %*d/%d\r", WIDTH(ecc->T), i+1, ecc->T);
        TrainClassifierChainModels(parm, l_data, instances, l_labels, l_n_labels, classes, inputs, &ecc->ccs[i], subspace, cv, classifier, l_subset, chain, SS, l_index, false);

        if (parm->ecc_with_bagging) {
            FREE(l_data);
            FREE(l_labels);
            FREE(l_n_labels);
            if (subset)
                FREE(l_subset);
            if (index)
                FREE(l_index);
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
}

void ECCClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                         int inputs, int *query_index, t_ECC *ecc, int *subspace, int *S, int **SS, int **index)
{
    scalar F[n_labels];
    int H[n_labels];

    ZeroVector(f, NULL, n_labels);

    for (int i = 0; i < ecc->T; i++) {
        ClassifierChainClassifyInstance(parm, query, H, F,  classification, ranks, data, instances, n_labels, n_predicted, inputs, query_index, &(ecc->ccs[i]), subspace, S, SS, index, false);

        for (int j = 0; j < *n_predicted; j++)
            f[H[j]]++;
    }

    // For predicted labels majority voting
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++)
        if (f[i] > ecc->T/2.0) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }

    NormalizeVector(f, n_labels);
}

void FreeECC(const t_Parameters *parm, t_ECC *ecc, int classes, int inputs)
{
    for (int i = 0; i < ecc->T; i++)
        FreeClassifierChain(parm, &(ecc->ccs[i]), classes, inputs);

    FREE(ecc->ccs);
}
