/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "typex.h"
#include "maths.h"

scalar EvaluateSelectionOneLabel(const t_Parameters *parm, scalar **data, int instances, int **labels, int inputs, t_Classifier evaluator, int *S, cache_t **neighbors_cache, int **index,
                                 int one_label_k, int *k, int pop_index, int **population, int individual_index, int *n_labels, t_PerformanceMeasure measure, int *list, int list_size)
{
    scalar ok = 0.0;

    // Get labels info
    int *one_labels;
    MALLOC(one_labels, int, instances, "EvaluateSelectionOneLabel: one_labels");
    for (int i = 0; i < instances; i++)
        one_labels[i] = LabelInList(pop_index, labels[i], n_labels[i]) ? 1 : 0;

    // If k == 0, get k as average value where member is present
    if (!one_label_k) {
        int count = 1;  // Use pseudo-counts to avoid '0'
        scalar average_k = 1.0;
        for (int i = 0; i < parm->population_size; i++)
            if (population[i][pop_index] == individual_index) {
                average_k += k[i];
                count++;
            }

        one_label_k = ceil(average_k/count);
    }

    switch (evaluator) {
        // Evaluate using appropriate classifier
        case kNN:
            ok = EvaluatekNNCache(instances, one_labels, S, data, instances, one_labels, inputs, 2, one_label_k, false, NULL, NULL, index, neighbors_cache, measure, list, list_size);
            break;

        default:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Selected classifier");
            break;
    }

    FREE(one_labels);

    return ok;
}

scalar EvaluatekNNCache(int instances, int *labels, int *S, scalar **protos, int n_protos, int *p_labels, int inputs, int classes, int k, bool knn_with_weighted_votes, scalar *r,
                        scalar *distance_to_furthest_neighbor, int **protos_index, cache_t **neighbors_cache, t_PerformanceMeasure measure, int *list, int list_size)
{
    scalar M[2][2] = {{0.0, 0.0}, {0.0, 0.0}}, sn, sp, ok = 0.0, prec, p0, pe, pYes, pNo, margin = 0.0;
    int nearest[k], class;

    // If a list is given evaluate member of list only
    if (measure == pfMargin)
        for (int i = 0; i < (list ? list_size : instances); i++)
            margin += kNNSingleLabelCategoryAndMargin(protos[list ? list[i] : i], protos, NULL, n_protos, k, inputs, p_labels, 2, nearest,   S, neighbors_cache, list ? list[i] : i, protos_index,
                      protos_index ? protos_index[list ? list[i] : i] : NULL, &class);
    else
        for (int i = 0; i < (list ? list_size : instances); i++)
            M[labels[list ? list[i] : i]][kNearestNeighbourCache(list ? list[i] : i, S, n_protos, nearest, p_labels, k, neighbors_cache, protos, inputs, classes, knn_with_weighted_votes, r,
                                          distance_to_furthest_neighbor, protos_index, NULL)]++;

    int positive = 1;
    int negative = 0;

    scalar tp = M[positive][positive];
    scalar tn = M[negative][negative];
    scalar fp = M[negative][positive];
    scalar fn = M[positive][negative];

    switch (measure) {
        case pfAccuracy:
        case pfAverageAccuracy:
            ok = (tp+tn)/(tp+tn+fp+fn);
            break;

        case pfGMean:
            sn = (tp+fn > ZERO) ? tp/(tp+fn) : 0.0;
            sp = (tn+fp > ZERO) ? tn/(tn+fp) : 0.0;

            ok = sqrt(sn*sp);
            break;

        case pfPrecisionMacro:
            ok = (tp+fp > ZERO) ? tp/(tp+fp) : 0.0;
            break;

        case pfCC:
            ok = sqrt((tp+fn)*(tn+fp)*(tp+fp)*(tn+fn)) > 0.0 ? ((tp*tn)-(fn*fp))/sqrt((tp+fn)*(tn+fp)*(tp+fp)*(tn+fn)) : 0.0;
            break;

        case pfRecallMacro:
        case pfRecallMacroGMean:
            ok = (tp+fn > ZERO) ? tp/(tp+fn) : 0.0;
            break;

        case pfF1Macro:
            prec = (tp+fp > ZERO) ? tp/(tp+fp) : 0.0;
            sn = (tp+fn > ZERO) ? tp/(tp+fn) : 0.0;

            ok = prec+sn > ZERO ? (2.0*prec*sn)/(prec+sn) : 0.0;
            break;

        case pfKappa:
            // Sum of the diagonal
            p0 = (tp+tn)/(tp+tn+fp+fn);
            pYes = ((tp+fn)/(tp+tn+fp+fn))*((tp+fp)/(tp+tn+fp+fn));
            pNo = ((fp+tn)/(tp+tn+fp+fn))*((fn+tn)/(tp+tn+fp+fn));
            pe = pYes + pNo;

            ok = (p0 < 1.0) ? (p0-pe)/(1.0-pe) : 1.0;
            break;

        case pfMargin:
            ok = margin/list_size;
            break;

        case pfauROC:
        case pfauPRC:
        case pfF1:
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Performance measure for subpopulation evaluation");
            break;
    }

    return ok;
}

// Classify a pattern using k nearest neighbors rule
scalar kNNSingleLabelCategoryAndMargin(scalar *instance, scalar **data, int *subspace, int instances, int k, int inputs, int *labels, int classes, int *nearest, int *subset, cache_t **neighbors_cache,
                                       int inst_order, int **index, int *instance_index, int *aclass)
{
    scalar votes[classes], margin;

    ZeroVector(votes, NULL, classes);

    for (int i = 0; i < k; i++)
        nearest[i] = -1;

    // Use cache?
    if (neighbors_cache) {
        int c = 0;
        for (int i = 0; i < instances && c < k; i++)
            if (!subset || subset[neighbors_cache[inst_order][i]]) {
                votes[labels[neighbors_cache[inst_order][i]]]++;
                nearest[c] = neighbors_cache[inst_order][i];
                c++;
            }
    } else {
        scalar dmin = INFINITE, dist[k], value;
        for (int i = 0; i < k; i++)
            dist[i] = INFINITE;

        // Get k nearest neighbors
        for (int i = 0; i < instances && dmin > 0.0; i++)
            if (!subset || subset[i]) {
                value = EuclideanDistanceSparseDataSubspace(instance, data[i], inputs, instance_index, index ? index[i] : NULL, subspace);

                // Test if is one of the nearest
                if (value < dmin /*&& instance != data[i]*/) {
                    // Look for its place
                    int j = 0;
                    while (value > dist[j])
                        j++;

                    // Sets in j-th place
                    for (int l = k - 1; l > j; l--) {
                        nearest[l] = nearest[l - 1];
                        dist[l] = dist[l - 1];
                    }
                    nearest[j] = i;
                    dist[j] = value;
                    dmin = dist[k - 1];
                }
            }

        // Vote neighbors
        for (int i = 0; i < k; i++)
            if (nearest[i] != -1)
                votes[labels[nearest[i]]]++;
    }

    // Get maximum votes
    *aclass = 0;
    for (int i = 1; i < classes; i++)
        if (votes[*aclass] < votes[i])
            *aclass = i;

    // Votes of the correct class
    int ok_votes = votes[(*aclass)];

    // Get maximum of votes of wrong class
    int wrong_votes = 0;
    for (int i = 0; i < classes; i++)
        if (wrong_votes < votes[i] && i != *aclass)
            wrong_votes = votes[i];

    margin = ((scalar)(ok_votes-wrong_votes))/k;

    return margin;
}
