/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <math.h>
#include "macros.h"
#include "maths.h"
#include "typex.h"
#include "ml.h"
#include "ensemble.h"

// Implementes FRONEC: Fuzzy ROugh NEighborhood Consensus
// Information Sciences 433–434 (2018) 96–114
// Multi-label classification using a fuzzy rough neighborhood consensus, Sarah Vluymans, Chris Cornelis, Francisco Herrera, Yvan Saeys

// Not much to do, lazy learner
void TrainFRONEC(const t_Parameters *parm, t_FRONEC *fronec, int classes, int **labels, int *n_labels, int instances, int inputs, scalar **data)
{
    fronec->k = parm->mlknn_k;

    // Prior probabilities of the labels
    MALLOC(fronec->prior, scalar, classes, "TrainFRONEC: fronec->prior");

    ZeroVector(fronec->prior, NULL, classes);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            fronec->prior[labels[i][j]]++;

    for (int i = 0; i < classes; i++)
        fronec->prior[i] /= instances;

    // Ranges of variables
    scalar *max, *min;
    MALLOC(max, scalar, inputs, "TrainFRONEC: max");
    MALLOC(min, scalar, inputs, "TrainFRONEC: min");
    MALLOC(fronec->ranges, scalar, inputs, "TrainFRONEC: frone->ranges");

    for (int i = 0; i < inputs; i++)
        max[i] = min[i] = data[0][i];

    for (int i = 1; i < instances; i++)
        for (int j = 0; j < inputs; j++)
            if (max[j] < data[i][j])
                max[j] = data[i][j];
            else if (min[j] > data[i][j])
                min[j] = data[i][j];

    for (int i = 0; i < inputs; i++)
        fronec->ranges[i] = fabs(min[i]-max[i]) > ZERO ? max[i]-min[i] : 1.0;

    FREE(max);
    FREE(min);
}

void FRONECClassifyInstance(t_FRONEC *fronec, scalar *query, scalar **data, int instances, int **index, int inputs, int *query_index, int *S, cache_t **neighbors_cache, int **SS, int classes,
                            int *subspace, int **labels, int *n_labels, t_MLClassifier method, scalar *f, int *n_h, int *h)
{
    if (query_index)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Sparse format for FRONEC method");

    // Get neighbours
    int neighbors[fronec->k];
    GetkNeighborsForML(query, data, instances, index, inputs, query_index, fronec->k, neighbors, S, neighbors_cache, -1, false, SS, classes, NULL, -1, subspace);

    // Get Q_x(y)
    // Get effective k in case too few selected
    int k = 0;
    while (k < fronec->k && neighbors[k] != -1)
        k++;

    // Weight vectors: Additive weights
    scalar WL[k], WU[k];
    for (int i = 0; i < k; i++)
        WL[i] = WU[k-i-1] = (2.0*(i+1))/(k*(k+1.0));

    scalar Qx[k], Rz[k];

    for (int z = 0; z < k; z++)
        Rz[z] = R(data[neighbors[z]], query, inputs, fronec->ranges);

    scalar m[k], Q1, Q2, m1[k], m2[k];

    switch (method) {
        case FRONEC1H:
            for (int y = 0; y < k; y++) {
                for (int z = 0; z < k; z++)
                    m[z] = I(Rz[z], RdHammingBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes));

                // Sort m
                qsort(m, k, sizeof(scalar), CompareDoubleReverse);

                Qx[y] = 0.0;
                for (int l = 0; l < k; l++)
                    Qx[y] += m[l]*WL[l];
            }
            break;

        case FRONEC2H:
            for (int y = 0; y < k; y++) {
                for (int z = 0; z < k; z++)
                    m[z] = T(Rz[z], RdHammingBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes));

                // Sort m
                qsort(m, k, sizeof(scalar), CompareDoubleReverse);

                Qx[y] = 0.0;
                for (int l = 0; l < k; l++)
                    Qx[y] += m[l]*WU[l];
            }
            break;

        case FRONEC3H:
            for (int y = 0; y < k; y++) {
                for (int z = 0; z < k; z++) {
                    m1[z] = I(Rz[z], RdHammingBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes));
                    m2[z] = T(Rz[z], RdHammingBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes));
                }

                // Sort m
                qsort(m1, k, sizeof(scalar), CompareDoubleReverse);
                qsort(m2, k, sizeof(scalar), CompareDoubleReverse);

                Q1 = Q2 = 0.0;
                for (int l = 0; l < k; l++) {
                    Q1 += m1[l]*WL[l];
                    Q2 += m2[l]*WU[l];
                }

                Qx[y] = (Q1+Q2)/2;
            }
            break;

        case FRONEC1LD:
            for (int y = 0; y < k; y++) {
                for (int z = 0; z < k; z++)
                    m[z] = I(Rz[z], RdLabelDistributionBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes, fronec->prior));

                // Sort m
                qsort(m, k, sizeof(scalar), CompareDoubleReverse);

                Qx[y] = 0.0;
                for (int l = 0; l < k; l++)
                    Qx[y] += m[l]*WL[l];
            }
            break;

        case FRONEC2LD:
            for (int y = 0; y < k; y++) {
                for (int z = 0; z < k; z++)
                    m[z] = T(Rz[z], RdLabelDistributionBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes, fronec->prior));

                // Sort m
                qsort(m, k, sizeof(scalar), CompareDoubleReverse);

                Qx[y] = 0.0;
                for (int l = 0; l < k; l++)
                    Qx[y] += m[l]*WU[l];
            }
            break;

        case FRONEC3LD:
            for (int y = 0; y < k; y++) {
                for (int z = 0; z < k; z++) {
                    m1[z] = I(Rz[z], RdLabelDistributionBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes, fronec->prior));
                    m2[z] = T(Rz[z], RdLabelDistributionBased(labels[neighbors[y]], n_labels[neighbors[y]], labels[neighbors[z]], n_labels[neighbors[z]], classes, fronec->prior));
                }

                // Sort m
                qsort(m1, k, sizeof(scalar), CompareDoubleReverse);
                qsort(m2, k, sizeof(scalar), CompareDoubleReverse);

                Q1 = Q2 = 0.0;
                for (int l = 0; l < k; l++) {
                    Q1 += m1[l]*WL[l];
                    Q2 += m2[l]*WU[l];
                }

                Qx[y] = (Q1+Q2)/2;
            }
            break;

        default:
            OnErrorExit(ERR_UNSPECTED_RESULT, "Should not be in FRONEC classification function");
            break;
    }

    // Get maximum of Qx
    int Y[k], n = 0;
    scalar max = -1.0;
    for (int i = 0; i < k; i++)
        if (Qx[i] > max) {
            n = 1;
            Y[0] = i;
            max = Qx[i];
        } else if (fabs(Qx[i]-max) <= ZERO)
            Y[n++] = i;

    // Voting
    ZeroVector(f, NULL, classes);
    for (int i = 0; i < n; i++)
        for (int j = 0; j < n_labels[Y[i]]; j++)
            f[labels[Y[i]][j]]++;

    // Relevant labels
    *n_h = 0;
    for (int i = 0; i < classes; i++)
        if (f[i] >= n/2)
            h[(*n_h)++] = i;

    NormalizeVector(f, classes);
}

scalar RdHammingBased(int *labels_x, int n_labels_x, int *labels_y, int n_labels_y, int labels)
{
    scalar R = 0.0;

    for (int i = 0; i < labels; i++)
        if (LabelInList(i, labels_x, n_labels_x) != LabelInList(i, labels_y, n_labels_y))
            R++;

    return 1.0 - R/labels;
}

scalar RdLabelDistributionBased(int *labels_x, int n_labels_x, int *labels_y, int n_labels_y, int labels, scalar *prior)
{
    scalar a, b, c, d;

    a = b = c = d = 0.0;

    for (int i = 0; i < labels; i++) {
        bool rel_x = LabelInList(i, labels_x, n_labels_x);
        bool rel_y = LabelInList(i, labels_y, n_labels_y);

        if (rel_x == rel_y) {
            if (rel_x == true)
                a += 1-prior[i];
            else
                d += prior[i];
        } else {
            if (rel_x)
                b += 0.5;
            else
                c += 0.5;
        }
    }

    return (a+d)/(a+b+c+d);
}

scalar R(scalar *x, scalar *y, int inputs, scalar *ranges)
{
    scalar d = 0.0;

    for (int i = 0; i < inputs; i++)
        d += 1.0 - fabs(x[i]-y[i])/ranges[i];

    return d/inputs;
}

void FreeFRONEC(t_FRONEC *fronec)
{
    FREE(fronec->prior);
    FREE(fronec->ranges);
}


// Fuzzy operators
scalar I(scalar a, scalar b)
{
    return MIN(1.0, 1.0-a+b);
}

scalar T(scalar a, scalar b)
{
    return MAX(a+b-1.0, 0.0);
}

int CompareDoubleReverse(const void *x, const void *y)
{
    scalar diff;

    if ((diff = * (scalar *) x - * (scalar *) y) < 0.0)
        return +1;
    else {
        if (diff > 0.0)
            return -1;
        else
            return 0;
    }
}
