/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "ml.h"

// Perform instance considering possible data partition

scalar FeatureSelectionML(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index, int **subspace, int **ssubspace,
                        t_MLClassifier method, scalar **test_data, int **test_labels, int *test_n_labels, int test_instances, int **test_index, int **chain)
{
    scalar reduction = 0.0;
    int generations = 0;

    if (parm->ml_feature_selection == NoFeatureSelection)
        return reduction;

    if (ssubspace)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Per label subspace selection");

    // Data partition
    if (parm->data_partition_method == NoDataPartition || !parm->subset_size) {
        scalar reduction = FeatureSelectionAlgorithmML(parm, data, labels, n_labels, instances, inputs, classes, index, subspace, ssubspace, method, test_data, test_labels, test_n_labels,
test_instances,
                                                       test_index, &generations, chain);
        // SAVEPROGRESS(parm->ml_feature_selection == CHCIFS, "@Actual generations performed: %d\n", generations);
        return reduction;
    } else {
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partition for ML feature selection");
        int n_subsets = instances%parm->subset_size ? instances/parm->subset_size+1 : instances/parm->subset_size, subset_generations;

        SHOWPROGRESS(parm->show_progress > 0, "Data partition into %d subsets of size %d\n", n_subsets, parm->subset_size);
        SAVEPROGRESS(parm->show_progress > 0, "Data partition into %d subsets of size %d\n", n_subsets, parm->subset_size);

        scalar ***subset_data;
        int ***subset_index = NULL, *subset_instances, * **subset_labels, **subset_list, **subset_selection = NULL, **subset_n_labels, ***subset_ss = NULL;

        MALLOC(subset_data, scalar **, n_subsets, "FeatureSelection: subset_data");
        if (index)
            MALLOC(subset_index, int **, n_subsets, "FeatureSelection: subset_index");
        MALLOC(subset_instances, int, n_subsets, "FeatureSelection: subset_list");
        MALLOC(subset_labels, int **, n_subsets, "FeatureSelection: subset_labels");
        MALLOC(subset_n_labels, int *, n_subsets, "FeatureSelection: subset_n_labels");
        MALLOC(subset_list, int *, n_subsets, "FeatureSelection: subset_list");
        for (int i = 0; i < n_subsets; i++) {
            MALLOC(subset_data[i], scalar *, parm->subset_size, "FeatureSelection: subset_data[i]");
            if (index) {
                MALLOC(subset_index[i], int *, parm->subset_size, "FeatureSelection: subset_index[i]");
            }
            MALLOC(subset_labels[i], int *, parm->subset_size, "FeatureSelection: subset_labels[i]");
            MALLOC(subset_n_labels[i], int, parm->subset_size, "FeatureSelection: subset_n_labels[i]");
            MALLOC(subset_list[i], int, parm->subset_size, "FeatureSelection: subset_labels[i]");
        }

        // Standard or per label instance subspace?
        MALLOC(subset_selection, int *, n_subsets, "FeatureSelection: subset_selection");
        MALLOC(subset_ss, int **, n_subsets, "FeatureSelection: subset_ss");

        DataPartitionML(parm, data, instances, labels, &n_subsets, &subset_data, &subset_instances, &subset_labels, &subset_list, index, &subset_index, parm->subset_size, n_labels, &subset_n_labels,
                        inputs);

#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < n_subsets; i++) {
            reduction += FeatureSelectionAlgorithmML(parm, subset_data[i], subset_labels[i], subset_n_labels[i], subset_instances[i], inputs, classes, subset_index ? subset_index[i] : NULL,
                         &(subset_selection[i]), NULL, method, test_data, test_labels, test_n_labels, test_instances, test_index, &subset_generations, chain);
#ifdef _OPENMP
            #pragma omp critical
#endif
            generations += subset_generations;

            // Aggregate subspace
            for (int j = 0; j < subset_instances[i]; j++)
                (*subspace)[subset_list[i][j]] = subset_selection[i][j];
            FREE(subset_selection[i]);
        }

        // SAVEPROGRESS(parm->ml_feature_selection == CHCFS, "@Actual generations performed: %d\n", generations/n_subsets);

        reduction /= n_subsets;

        for (int i = 0; i < n_subsets; i++) {
            FREE(subset_data[i]);
            FREE(subset_labels[i]);
            FREE(subset_list[i]);
            FREE(subset_n_labels[i]);
            if (subset_index)
                FREE(subset_index[i]);
        }

        FREE(subset_data);
        if (index)
            FREE(subset_index);
        FREE(subset_instances);
        FREE(subset_labels);
        FREE(subset_n_labels);
        FREE(subset_list);
        FREE(subset_selection);
        FREE(subset_ss);
    }

    // If none select one randomly
    if (reduction >= 1.0) {
        reduction = (inputs-1.0)/inputs;
        (*subspace)[(int) RANDOM(0, instances)] = 1;
    }

    return reduction;
}


scalar FeatureSelectionAlgorithmML(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index, int **subspace, int **ssubspace,
                                    t_MLClassifier method, scalar **test_data, int **test_labels, int *test_n_labels, int test_instances, int **test_index, int *generations, int **chain)
{
    scalar reduction = 0;

    // Instance subspace algorithms
    switch (parm->ml_feature_selection) {
        case NoFeatureSelection:
            reduction = 0.0;
            FREE(*subspace);
            break;

        case CHCFS:
            if (method == CC)
                MALLOC(*chain, int, classes, "FeatureSelectionAlgorithmML: chain");
            reduction = (scalar) CHCFeatureSelectionForML(parm, instances, data, labels, n_labels, classes, inputs, *subspace, index, method, generations, *chain);
            break;
    }

    return reduction;
}
