/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "maths.h"
#include "ensemble.h"

// Implements FV-kNN
// Z. Younes, F. Abdallah and T. Denœux, "Fuzzy multi-label learning under veristic variables," International Conference on Fuzzy Systems, 2010, pp. 1-8.

// Weighted conjunctive version: Best in paper

void TrainFVkNN(t_FVkNN *fvknn, const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, cache_t **neighbors_cache, int *S, int **SS, int *subspace,
                int **index, int inputs)
{
    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial instance selection for FV-kNN");

    fvknn->k = parm->mlknn_k;
    AllocateScalarMatrix(&(fvknn->fuzzy_labels), instances, classes);

    // Fuzzify labels
    int neighbors[fvknn->k];
    for (int i = 0; i < instances; i++) {
        // Get neighbours
        GetkNeighborsForML(data[i], data, instances, index, inputs, index ? index[i] : NULL, fvknn->k, neighbors, S, neighbors_cache, i, false, SS, classes, NULL, -1, subspace);

        // Select last in case too few selected
        int k = 0;
        while (k < fvknn->k && neighbors[k] != -1)
            k++;

        // Construct fuzzy version of labels
        for (int j = 0; j < classes; j++) {
            scalar n_q = 0.0;
            for (int l = 0; l < k; l++)
                if (LabelInList(j, labels[neighbors[l]], n_labels[neighbors[l]]))
                    n_q++;

            fvknn->fuzzy_labels[i][j] = 0.49*(n_q/k);
            if (LabelInList(j, labels[i], n_labels[i]))
                fvknn->fuzzy_labels[i][j] += 0.51;
        }
    }
}

void FVkNNClassify(t_FVkNN *fvknn, scalar *query, scalar **data, int instances, int **index, int *query_index, int inputs, int *S, int **SS, cache_t **neighbors_cache, int classes, int *subspace,
                   scalar *f, int *n_h, int *h)
{
    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial instance selection for FV-kNN");

    if (neighbors_cache)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Neighbors cache for FV-kNN");

    // Get neighbours
    int neighbors[fvknn->k+1];
    GetkNeighborsForML(query, data, instances, index, inputs, query_index, fvknn->k+1, neighbors, S, NULL, -1, false, SS, classes, NULL, -1, subspace);

    // Select last in case too few selected
    int k = 0;
    while (k < fvknn->k+1 && neighbors[k] != -1)
        k++;

    // Distances for weighting votes
    scalar d[k];
    for (int i = 0; i < k; i++)
        d[i] = EuclideanDistanceSparseDataSubspace(query, data[neighbors[i]], inputs, query_index, index ? index[neighbors[i]] : NULL, subspace);

    // Membership of every label to every neighbor
    scalar **m;
    AllocateScalarMatrix(&m, MAX(1, k-1), classes);
    for (int i = 0; i < k-1; i++)
        for (int j = 0; j < classes; j++) {
            scalar w_i = (d[k-1]-d[i])/(d[k-1]-d[0]);
            m[i][j] = fvknn->fuzzy_labels[neighbors[i]][j] * w_i;
        }

    // Predict label if minimum confidence is above 0.5
    scalar t = 0.5;
    *n_h = 0;
    for (int i = 0; i < classes; i++) {
        f[i] = m[0][i];
        for (int j = 1; j < k-1; j++)
            if (m[j][i] > f[i])
                f[i] = m[j][i];

        if (f[i] > t)
            h[(*n_h)++] = i;
    }

    FreeScalarMatrix(&m, k-1);

    NormalizeVector(f, classes);
}


void FreeFVkNN(t_FVkNN *fvknn, int instances)
{
    FreeScalarMatrix(&(fvknn->fuzzy_labels), instances);
}
