/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "maths.h"
#include "ensemble.h"

// Implements Hierarchy Of Multi-label classifiERs (HOMER)
// G. Tsoumakas, I. Katakis, I. Vlahavas, "Effective and efficient multilabel classification in domains with large number of labels"
// Proceedings of the ECML/PKDD Workshop on Mining Multidimensional Data, 2008, pp. 30–44.

void TrainHOMERModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_HOMER *homer, int *subspace, bool cv, t_Classifier classifier,
                      int *subset, int **SS, int **index, bool clustering, bool balanced)
{
    // Test if valid CC classification model
    if (classifier == NN || classifier == kNN || classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "HOMER classification method");

    // Pending implementation
    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "HOMER for partial instance selection");

    // Set model parameters
    homer->method = classifier;
    homer->k = parm->homer_k;

    // Label dependent instance selection and standard instance selection are incompatible
    if (subset && SS)
        OnErrorExit(ERR_PARAM, "Label dependent instance selection and standard instance selection are incompatible");

    // Generate label hierarchy
    CreateHierarchy(&(homer->hierarchy), classes, homer->k, parm, data, instances, labels, n_labels, inputs, subspace, cv, homer->method, subset, index, clustering, balanced);
}

void ClassifierHOMERInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                             int inputs, int *query_index, t_HOMER *homer, int *subspace, int *S, int **index)
{
    int *relevant;

    CALLOC(relevant, int, n_labels, "ClassifierHOMERInstance: relevant");

    // Output of all powersets classifiers
    ZeroVector(f, NULL, n_labels);
    RecursiveClassifierHOMERInstance(parm, query, data, instances, n_labels, inputs, query_index, homer->hierarchy, subspace, S, index, f, relevant, homer->method);

    // Predicted labels
    *n_predicted = 0;

    for (int i = 0; i < n_labels; i++)
        if (relevant[i] > 0)
            h[(*n_predicted)++] = i;

    *classification = true;

    // Normalize probabilities
    NormalizeVector(f, n_labels);
    *ranks = true;

    FREE(relevant);
}

void RecursiveClassifierHOMERInstance(const t_Parameters *parm, scalar *query, scalar **data, int instances, int n_labels, int inputs, int *query_index, struct t_Leaf *hierarchy, int *subspace,
                                      int *S, int **index, scalar *f, int *relevant, t_Classifier method)
{
    // If leaf it is a label
    if (hierarchy->n_children == 0) {
        scalar output[(int) pow(2.0, hierarchy->n)];

        // Get labels
        int category = ClassifierProbabilities(parm, method, query, inputs, (int) pow(2.0, hierarchy->n), query_index, subspace, data, instances, NULL, S, &(hierarchy->model), -1, NULL, -1,
                       index, output, &(hierarchy->ensemble));

        // Decodify output
        for (int i = 0; i < hierarchy->n; i++) {
            if (category % 2)
                relevant[hierarchy->labels[i]] = 1;

            category /= 2;
        }

        // For rankings scikit method for LP is used
        int binary[hierarchy->n];
        for (int i = 0; i < pow(2.0, hierarchy->n); i++) {
            IntegerToBinary(i, binary, hierarchy->n);
            for (int j = 0; j < hierarchy->n; j++)
                f[hierarchy->labels[j]] += output[i]*binary[j];
        }

        return;
    }

    // If not a leaf node get meta-labels
    scalar output[(int) pow(2.0, hierarchy->n_children)];
    int category = ClassifierProbabilities(parm, method, query, inputs, pow(2.0, hierarchy->n_children), query_index, subspace, data, instances, NULL, S, &(hierarchy->model), -1, NULL, -1,
                   index, output, &(hierarchy->ensemble));
    int meta[hierarchy->n_children];
    for (int i = 0; i < hierarchy->n_children; i++) {
        meta[i] = category % 2;
        category /= 2;
    }

    // Visit selected meta-labels
    for (int i = 0; i < hierarchy->n_children; i++)
        if (meta[i])
            RecursiveClassifierHOMERInstance(parm, query, data, instances, n_labels, inputs, query_index, &(hierarchy->children[i]), subspace, S, index, f, relevant, method);
}

void FreeHOMER(const t_Parameters *parm, t_HOMER *homer, int inputs)
{
    RecursiveFreeHiearchy(parm, homer->hierarchy, inputs, homer->method, 2);

    FREE(homer->hierarchy);
}

void RecursiveFreeHiearchy(const t_Parameters *parm, struct t_Leaf *hierarchy, int inputs, t_Classifier method, int classes)
{
    for (int i = 0; i < hierarchy->n_children; i++)
        RecursiveFreeHiearchy(parm, &(hierarchy->children[i]), inputs, method, classes);

    FREE(hierarchy->labels);
    if (hierarchy->children)
        FREE(hierarchy->children);
    ClassifierFree(parm, &(hierarchy->model), method, classes, inputs, &(hierarchy->ensemble));
}

void CreateHierarchy(struct t_Leaf **hierarchy, int classes, int k, const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int inputs,int *subspace, bool cv,
                     t_Classifier classifier, int *subset, int **index, bool clustering, bool balanced)
{
    // Create level 0
    MALLOC(*hierarchy, struct t_Leaf, 1, "CreateHierarchy: *hierarchy");
    MALLOC((*hierarchy)->labels, int, classes, "CreateHierarchy: *hierarchy->labels");

    (*hierarchy)->n = classes;
    (*hierarchy)->n_children = 0;
    (*hierarchy)->children = NULL;
    (*hierarchy)->level = 0;    // To control maximum levels
    for (int i = 0; i < classes; i++)
        (*hierarchy)->labels[i] = i;

    RecursiveHierarchy(*hierarchy, k, parm, data, instances, labels, n_labels, inputs, subspace, cv, classifier, subset, index, clustering, balanced);
}

// Create random hierarchy
void RecursiveHierarchy(struct t_Leaf *hierarchy, int k, const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int inputs, int *subspace, bool cv,
                        t_Classifier classifier, int *subset, int **index, bool clustering, bool balanced)
{
    int *list;

    // If number of labels <= k do not continue/*, or maximum depth*/
    if (hierarchy->n <= k/* || (parm->homer_max_depth && hierarchy->level >= parm->homer_max_depth)*/) {
        hierarchy->n_children = 0;
        hierarchy->children = NULL;

        // Train model
        TrainModelForLevel(parm, data, instances, labels, n_labels, inputs, hierarchy, subspace, cv, classifier, subset, index);
        return;
    }

    // Partition among the children
    MALLOC(list, int, hierarchy->n, "RecursiveHierarchy");
    for (int i = 0; i < hierarchy->n; i++)
        list[i] = hierarchy->labels[i];

    // Create children

    // Partition of labels can be random or using balanced k-means
    // BALANCED K-MEANS PARTITION
    if (clustering) {
        int **clusters, cluster_size[k];
        AllocateIntegerMatrix(&clusters, k, hierarchy->n);
        BalancedKMeans(k, list, hierarchy->n, labels, n_labels, parm->bkmeans_iter, instances, clusters, cluster_size, balanced);

        // Some clusters might be empty
        int actual_size = k;
        for (int i = 0; i < actual_size; i++)
            if (cluster_size[i] == 0) {
                actual_size--;
                cluster_size[i] = cluster_size[actual_size];
                int *sw = clusters[i];
                clusters[i] = clusters[actual_size];
                clusters[actual_size] = sw;
                i--;
            }

        // PrintVector(NULL, cluster_size, k, stdout);

        if (actual_size == 0)
            OnErrorExit(ERR_UNSPECTED_RESULT, "All HOMER clusters are empty");
        else if (actual_size == 1) {    // Break randomly in two parts
            cluster_size[1] = cluster_size[0]/2;
            cluster_size[0] -= cluster_size[1];
            for (int i = 0; i < cluster_size[1]; i++)
                clusters[1][i] = clusters[0][i+cluster_size[0]];

            actual_size = 2;
        }

        // PrintVector(NULL, cluster_size, k, stdout);

        hierarchy->n_children = actual_size;
        MALLOC(hierarchy->children, struct t_Leaf, actual_size, "RecursiveHierarchy: hierarchy->children");

        for (int i = 0; i < actual_size; i++) {
            hierarchy->children[i].n_children = 0;
            hierarchy->children[i].children = NULL;
            hierarchy->children[i].n = cluster_size[i];
            MALLOC(hierarchy->children[i].labels, int, cluster_size[i], "RecursiveHierarchy: hierarchy->children[i].labels");
            for (int j = 0; j < cluster_size[i]; j++)
                hierarchy->children[i].labels[j] = clusters[i][j];
        }

        FreeIntegerMatrix(&clusters, k);
    } else {
        hierarchy->n_children = k;
        MALLOC(hierarchy->children, struct t_Leaf, k, "RecursiveHierarchy: hierarchy->children");

        // RANDOM PARTITION
        RandomizeVector(list, hierarchy->n);
        int next = 0;
        int length = hierarchy->n/k;
        for (int i = 0; i < k; i++) {
            hierarchy->children[i].n_children = 0;
            hierarchy->children[i].children = NULL;
            hierarchy->children[i].n = length;
            MALLOC(hierarchy->children[i].labels, int, length, "RecursiveHierarchy: hierarchy->children[i].labels");
            for (int j = 0; j < length; j++, next++)
                hierarchy->children[i].labels[j] = list[next];
        }

        // Assign remaining randomly
        for (; next < hierarchy->n; next++) {
            int sibling = RANDOM(0, k);
            REALLOC(hierarchy->children[sibling].labels, int, hierarchy->children[sibling].n+1, "RecursiveHierarchy: hierarchy->children[sibling].labels");
            hierarchy->children[sibling].labels[hierarchy->children[sibling].n] = list[next];
            hierarchy->children[sibling].n++;
        }
    }

    FREE(list);

    // Train models for created partition
    TrainModelForLevel(parm, data, instances, labels, n_labels, inputs, hierarchy, subspace, cv, classifier, subset, index);

    SHOWPROGRESS(parm->show_progress > 1, "HOMER adding level %d to hierarchy\n", hierarchy->level+1);

    // Recursively partition all children
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < hierarchy->n_children; i++) {
        hierarchy->children[i].level = hierarchy->level+1;
        RecursiveHierarchy(&(hierarchy->children[i]), k, parm, data, instances, labels, n_labels, inputs, subspace, cv, classifier, subset, index, clustering, balanced);
    }
}

void TrainModelForLevel(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int inputs, struct t_Leaf *level, int *subspace, bool cv, t_Classifier classifier,
                        int *subset, int **index)
{
    int *l_subspace = NULL;
    int *meta_labels = NULL, **decod_labels = NULL;

    // Create subset, needed for filtering out instances with no relevant labels
    int *S;
    MALLOC(S, int, instances, "TrainModelForLevel: S");
    if (subset)
        memcpy(S, subset, sizeof(int)*instances);
    else
        for (int i = 0; i < instances; i++)
            S[i] = 1;

    MALLOC(meta_labels, int, instances, "TrainModelForLevel: meta_labels");

    // If no subspace select all
    if (!subspace) {
        MALLOC(l_subspace, int, inputs, "TrainHOMERModels: l_subspace");
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = 1;
    } else
        l_subspace = subspace;

    // Create meta-labels
    int classes = -1;
    // if no children use leaf labels
    if (!level->n_children) {
        classes = (int) pow(2.0, level->n);
        AllocateIntegerMatrix(&decod_labels, instances, level->n);
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < level->n; j++) {
                decod_labels[i][j] = 0;
                for (int k = 0; k < level->n && !decod_labels[i][j]; k++)
                    if (LabelInList(level->labels[k], labels[i], n_labels[i]))
                        decod_labels[i][j] = 1;
            }
    } else {
        classes = (int) pow(2.0, level->n_children);
        AllocateIntegerMatrix(&decod_labels, instances, level->n_children);
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < level->n_children; j++) {
                decod_labels[i][j] = 0;
                for (int k = 0; k < level->children[j].n && !decod_labels[i][j]; k++)
                    if (LabelInList(level->children[j].labels[k], labels[i], n_labels[i]))
                        decod_labels[i][j] = 1;
            }
    }

    int count = level->n_children ? level->n_children : level->n;
    for (int i = 0; i < instances; i++) {
        meta_labels[i] = 0;
        for (int j = 0; j < count; j++)
            meta_labels[i] += (int) decod_labels[i][j]*pow(2.0, j);
    }

    // Cross-validate hyper-parameters?
    if (cv)
        SetCrossValidatedParameters((t_Parameters *) parm, data, instances, meta_labels, l_subspace, inputs, classes, classifier, 1, S, index);

    // Train
    ClassifierTraining(parm, classifier, data, instances, meta_labels, inputs, classes, index, l_subspace, S, true, &(level->model), &(level->ensemble));

    FreeIntegerMatrix(&decod_labels, instances);
    FREE(meta_labels);
    FREE(S);
    if (!subspace)
        FREE(l_subspace);
}

void IntegerToBinary(int integer, int *decoded, int length)
{
    for (int i = 0; i < length; i++) {
        decoded[i] = integer % 2;

        integer /= 2;
    }
}
