/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "maths.h"
#include "ensemble.h"

// Implements IBLR-ML
// W. Cheng and E. Hüllermeier (2009) "Combining instance-based learning and logistic regression for multilabel classification", Machine Learning 76: 211–225

void TrainIBLRMLModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_IBLR *iblr, int *subspace, int *subset, int **SS,
                       int **train_index, bool plus)
{
    scalar **meta_data;
    int **meta_index = NULL, *count;

    if (SS && subset)
        OnErrorExit(ERR_PARAM, "Partial and standard instance selection are incompatible");

    MALLOC(iblr->lrs, t_LR, classes, "TrainIBLRMLModels: iblr->lrs");
    iblr->k = parm->iblr_k;

    int features = plus ? inputs+classes : classes;
    int shift = plus ? inputs : 0;

    AllocateScalarMatrix(&meta_data, instances, features);
    MALLOC(count, int, instances, "TrainIBLRMLModels: count");
    ZeroVector(NULL, count, instances);
    if (train_index)
        AllocateIntegerMatrix(&meta_index, instances, features+1);

    // Create the new training data with label info as features
    // New attributes can be labels only or inputs also
    int neighbors[iblr->k];
    scalar freq[classes];

    // Effective selection for choosing neighbors depends on parameters
    int *S = NULL;
    if (subset)
        S = subset;
    else if (SS) {
        MALLOC(S, int, instances, "TrainIBLRMLModels: S");
        for (int i = 0; i < instances; i++)
            S[i] = !EmptyColumn(SS, classes, i);
    }

    for (int i = 0; i < instances; i++) {
        // Add inputs?
        if (plus) {
            if (train_index) {
                for (; train_index[i][count[i]] != -1; count[i]++) {
                    meta_data[i][count[i]] = data[i][count[i]];
                    meta_index[i][count[i]] = train_index[i][count[i]];
                }
                meta_index[i][count[i]] = -1;
            } else
                for (int j = 0; j < inputs; j++)
                    meta_data[i][j] = data[i][j];
        }

        // Get k-neighbors of i-th instance
        GetkNeighborsForML(data[i], data, instances, train_index, inputs, train_index ? train_index[i] : NULL, iblr->k, neighbors, S, NULL, -1, false, NULL, classes, NULL, -1, subspace);

        // Add freq of label j in neighbors of k as attribute
        ZeroVector(freq, NULL, classes);
        for (int j = 0; j < iblr->k; j++)
            for (int k = 0; k < n_labels[neighbors[j]]; k++)
                freq[labels[neighbors[j]][k]]++;

        if (train_index) {
            int c = 0;
            for (int j = 0; j < classes; j++)
                if (freq[j] > 0.0) {
                    meta_data[i][count[i]+c] = freq[j]/iblr->k;
                    meta_index[i][count[i]+c] = shift+j+1;
                    c++;
                }
            meta_index[i][count[i]+c] = -1;
        } else
            for (int j = 0; j < classes; j++)
                meta_data[i][shift+j] = freq[j]/iblr->k;
    }

    // Create a logistic regression for every label
    int **b_labels;
    AllocateIntegerMatrix(&b_labels, classes, instances);
    ZeroIntegerMatrix(b_labels, classes, instances);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            b_labels[labels[i][j]][i] = 1;

    for (int i = 0; i < classes; i++) {
        // Create binary labels
        SHOWPROGRESS(parm->show_progress > 0, "IBLR-ML%s training LR %*d/%d\r", plus ? "+" : "", WIDTH(classes), i+1, classes);
        TrainLogisticRegressionClassifier(parm, &(iblr->lrs[i]), instances, b_labels[i], features, meta_data, 2, SS ? SS[i] : S, subspace, meta_index);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FreeIntegerMatrix(&b_labels, classes);
    FreeScalarMatrix(&meta_data, instances);
    if (train_index)
        FreeIntegerMatrix(&meta_index, instances);
    FREE(count);
}

void IBLRMLClassifyInstance(scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int classes, int *n_predicted, int inputs, int *query_index,
                            t_IBLR *iblr, int *subspace, int *S, int **SS, int **index, int **labels, int *n_labels, bool plus)
{
    *n_predicted = 0;

    // Get k-neighbors of query
    int neighbors[iblr->k];
    GetkNeighborsForML(query, data, instances, index, inputs, query_index, iblr->k, neighbors, S, NULL, -1, false, NULL, classes, NULL, -1, subspace);

    int features = plus ? inputs+classes : classes;
    int shift = plus ? inputs : 0;

    // Construct meta data
    scalar *meta_data, freq[classes];
    int *meta_index = NULL;
    MALLOC(meta_data, scalar, features, "IBLRMLClassifyInstance: meta_data");
    if (index)
        MALLOC(meta_index, int, features+1, "IBLRMLClassifyInstance: meta_index");
    ZeroVector(freq, NULL, classes);
    for (int j = 0; j < iblr->k; j++)
        for (int k = 0; k < n_labels[neighbors[j]]; k++)
            freq[labels[neighbors[j]][k]]++;

    if (index) {
        int count = 0;
        if (plus) {
            for (; query_index[count] != -1; count++) {
                    meta_data[count] = query[count];
                    meta_index[count] = query_index[count];
                }
            meta_index[count] = -1;
        }
        int c = 0;
        for (int j = 0; j < classes; j++)
            if (freq[j] > 0.0) {
                meta_data[count+c] = freq[j]/iblr->k;
                meta_index[count+c] = j+shift+1;
                c++;
            }
        meta_index[count+c] = -1;
    } else {
        if (plus)
            for (int j = 0; j < inputs; j++)
                meta_data[j] = query[j];
        for (int j = 0; j < classes; j++)
            meta_data[shift+j] = freq[j]/iblr->k;
    }

    // Output of logistic regressors
    scalar probabilities[2];

    (*n_predicted) = 0;
    for (int i = 0; i < classes; i++) {
        LROutput(&(iblr->lrs[i]), meta_data, features, 2, probabilities, subspace, meta_index);
        f[i] = probabilities[1];
        if (f[i] >= 0.5)
            h[(*n_predicted)++] = i;
    }

    // PrintVector(f, NULL, classes, stdout);

    // Normalize probabilities
    NormalizeVector(f, classes);

    *classification = *ranks = true;

    FREE(meta_data);
    if (index)
        FREE(meta_index);
}

void FreeIBLRML(t_IBLR *iblr, int classes)
{
    for (int i = 0; i < classes; i++)
        FreeLR(&iblr->lrs[i]);
}
