/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"
#include "classifiers.h"

// Instance selection for LAML-kNN

scalar LAMLkNNInstanceSelection(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *S, scalar **test_data,
                                int **test_labels, int *test_n_labels, int test_instances, int **test_index, int *generations_performed)
{
    int *list, **selection, evaluations = 0;
    scalar *fitness;

    // Create and initialize global population
    MALLOC(fitness, scalar, 2*parm->population_size, "LAMLkNNInstanceSelection: fitness");
    MALLOC(list, int, parm->population_size, "LAMLkNNInstanceSelection: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;

    MALLOC(selection, int *, 2*parm->population_size, "LAMLkNNInstanceSelection: selection");

    // Cache of neighbors?
    scalar **distance_cache = NULL;
    cache_t **neighbors_cache = NULL;
    int neighbors_cache_size = 0;
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE) {
        SHOWPROGRESS(parm->show_progress > 0, "Constructing cache of size: %gG\n", NEIGHBORS_CACHE_SIZE(instances));
        ConstructAllNeighborsCache(data, instances, inputs, &distance_cache, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
    } else if (parm->ml_loo)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation with LOO without cache");

    // Adjust for multi-selection of instances
    scalar chc_init_prob = parm->chc_init_prob;
    if (parm->multiselection && parm->multiis_adjust_initial_percentage) {
        scalar sum = 0.0;
        int n = parm->ml_max_k/2+1;

        for (int i = 1; i <= n; i++)
            sum += i;

        fprintf(parm->out_file, "Initial percentage adjusted: %lf", parm->chc_init_prob);

        // With this adjustment, multi-selection selects the same number of instances, although some of them repeated
        chc_init_prob /= sum/n;

        fprintf(parm->out_file, " => %lf\n", chc_init_prob);
    }

    // If cross-validate hyper-parameters k is fixed to best value
    if (parm->cv_ml_hyperparameters) {
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);
        ((t_Parameters *)parm)->ml_min_k = ((t_Parameters *)parm)->ml_max_k = parm->mlknn_k;
        printf("WARNING: k fixed to %d, not evolved\n", parm->mlknn_k);
    }

    // Evolution of LAML-kNN with selection
    t_LAMLkNN *individual;
    MALLOC(individual, t_LAMLkNN, 2*parm->population_size, "LAMLkNNInstanceSelection: individual");
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < parm->population_size; i++) {
        CreateLAMLkNNClassifier(&individual[i], instances, index, inputs, parm->ml_max_k, parm->lamlknn_max_n_clusters, classes);
        individual[i].k = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);
        individual[i].n_clusters = (int) RANDOM(parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters+1);

        MALLOC(selection[i], int, instances, "LAMLkNNInstanceSelection");
        for (int j = 0; j < instances; j++)
            selection[i][j] = RANDOM(0.0, 1.0) < chc_init_prob ? (parm->multiselection ? RANDOM(1, parm->ml_max_k/2) : 1) : 0;

        // Carry out clustering
        kMeansSparse(data, instances, individual[i].n_clusters, inputs, parm->kmeans_max_iterations, index, individual[i].centroids, individual[i].centroids_index, individual[i].partition,
                     selection[i]);
        individual[i].clustered = true;

        TrainLocallyAdaptiveMultiLabelkNearestNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[i], neighbors_cache, 0, 0, false);
    }

    for (int i = parm->population_size; i < 2*parm->population_size; i++) {
        CreateLAMLkNNClassifier(&individual[i], instances, index, inputs, parm->ml_max_k, parm->lamlknn_max_n_clusters, classes);
        MALLOC(selection[i], int, instances, "LAMLkNNInstanceSelection");
    }

    // Evaluate initial individuals
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < parm->population_size; i++) {
        time_t init_time = time(NULL);
        fitness[i] = EvaluateLAMLkNNFitness(parm, &individual[i], selection[i], neighbors_cache, instances, labels, n_labels, classes, data, index, inputs);
        SHOWPROGRESS(parm->time_all, "Evaluation of individual: %ld s\n", time(NULL) - init_time);
    }

    evaluations += parm->population_size;

    time_t init_time = time(NULL);

    *generations_performed = parm->generations;
    for (int g = 0; g < parm->generations; g++) {
        SHOWPROGRESS(parm->show_progress > 0, "Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);
        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        // Random selection[i] without replacement
        RandomizeVector(list, parm->population_size);

        // Generate new individuals using crossover only
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = parm->population_size; i < 2*parm->population_size; i += 2) {
            int father = list[i-parm->population_size];
            int mother = list[i-parm->population_size+1];

            // Perform HUX crossover creating new individuals
            if (parm->multiselection && !parm->ml_use_hux)
                BLXAlphaCrossover(selection[i], selection[i+1], selection[father], selection[mother], instances, 0, parm->ml_max_k/2+1);
            else
                HuxCrossover(selection[mother], selection[father], selection[i], selection[i+1], instances);

            // Perform BLX-alpha crossover for k
            BLXAlphaCrossover(&individual[i].k, &individual[i+1].k, &individual[father].k, &individual[mother].k, 1, parm->ml_min_k, parm->ml_max_k);

            // Perform BLX-alpha crossover for number of clusters
            BLXAlphaCrossover(&individual[i].n_clusters, &individual[i+1].n_clusters, &individual[father].n_clusters, &individual[mother].n_clusters, 1, parm->lamlknn_min_n_clusters,
                              parm->lamlknn_max_n_clusters);

            // Carry out clustering
            kMeansSparse(data, instances, individual[i].n_clusters, inputs, parm->kmeans_max_iterations, index, individual[i].centroids, individual[i].centroids_index, individual[i].partition,
                         selection[i]);
            kMeansSparse(data, instances, individual[i+1].n_clusters, inputs, parm->kmeans_max_iterations, index, individual[i+1].centroids, individual[i+1].centroids_index,
                         individual[i+1].partition, selection[i]);
            individual[i].clustered = individual[i+1].clustered = true;
            TrainLocallyAdaptiveMultiLabelkNearestNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[i], neighbors_cache, 0, 0, false);
            TrainLocallyAdaptiveMultiLabelkNearestNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[i+1], neighbors_cache, 0, 0, false);

#ifdef _OPENMP
            #pragma omp critical
#endif
            SHOWPROGRESS(parm->show_progress > 1, "Generation %*d added individuals %d, %d\r", WIDTH(parm->generations), g+1, i+1, i+2);
        }

        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        // Evaluate NEW individuals
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = parm->population_size; i < 2*parm->population_size; i++) {
            fitness[i] = EvaluateLAMLkNNFitness(parm, &individual[i], selection[i], neighbors_cache, instances, labels, n_labels, classes, data, index, inputs);
            SHOWPROGRESS(parm->show_progress > 1, "Generation %*d evaluated individual %d\r", WIDTH(parm->generations), g+1, i+1);
        }

        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

        evaluations += parm->population_size;

        ///////////////////////////////////////////
        // Mutation operations ////////////////////
        ///////////////////////////////////////////
        if (parm->ml_chc_random_mut > 0.0)
            for (int i = parm->population_size*parm->ml_elitism; i < 2*parm->population_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_chc_random_mut) {
                    NonUniformMutation(&individual[i].k, 1, g,  parm->generations, parm->bit_mutation, parm->ml_min_k, parm->ml_max_k);
                    NonUniformMutation(&(individual[i].n_clusters), 1, g,  parm->generations, parm->bit_mutation, parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters);

                    kMeansSparse(data, instances, individual[i].n_clusters, inputs, parm->kmeans_max_iterations, index, individual[i].centroids, individual[i].centroids_index,
                                 individual[i].partition, selection[i]);
                    individual[i].clustered = true;
                    TrainLocallyAdaptiveMultiLabelkNearestNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[i], neighbors_cache, 0, 0, false);

                    fitness[i] = EvaluateLAMLkNNFitness(parm, &individual[i], selection[i], neighbors_cache, instances, labels, n_labels, classes, data, index, inputs);

                    evaluations++;
                }

        for (int i = 0; i < parm->population_size; i++) {
            int order = OrderedSelection(fitness+i, 0, 2*parm->population_size-i)+i;

            t_LAMLkNN aux = individual[i];
            individual[i] = individual[order];
            individual[order] = aux;

            scalar f = fitness[i];
            fitness[i] = fitness[order];
            fitness[order] = f;

            int *s = selection[i];
            selection[i] = selection[order];
            selection[order] = s;

        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            SHOWPROGRESS(true, "%*d g's best FITNESS, REDUCTION: %6.4lf, %6.4lf\n", WIDTH(parm->generations), g+1, fitness[best], Reduction(selection[best], NULL, instances));
            VERBOSE(parm->show_progress > 2, parm->out_file, "%*d generation best individual FITNESS (MEAN): %6.4lf (%6.4lf)\n", WIDTH(parm->generations), g+1, fitness[best],
                    AverageScalarVector(fitness, parm->population_size));
        }

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);
            *generations_performed = g+1;
            break;
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    int best = OrderedSelection(fitness, 0, parm->population_size);

    // Report best results
    fprintf(parm->out_file, "@LAML-kNN best number of clusters: %d\n", individual[best].n_clusters);
    fprintf(parm->out_file, "@LAML-kNN best k: %d\n", individual[best].k);

    // Evaluate selection and classifier
    bool classification, ranks;
    scalar **f;
    int *n_h, **h;
    AllocateScalarMatrix(&f, test_instances, classes);
    AllocateIntegerMatrix(&h, test_instances, classes);
    MALLOC(n_h, int, test_instances, "LAMLkNNInstanceSelection: n_h");
    t_MLLearner learner;
    learner.lamlknn = individual[best];
    ClassifyDataset(parm, LAMLkNN, test_data, h, f, &classification, &ranks, data, instances, labels, n_labels, classes, n_h, inputs, test_index, index, test_instances, selection[best],
                    neighbors_cache, NULL, &learner, NULL, NULL);

    // Message for output
    char message[] = "LAML-kNN IS";

    // Average number of predicted labels per instances
    scalar av_pred = 0.0, av_actual = 0.0;
    for (int i = 0; i < test_instances; i++) {
        av_pred += n_h[i];
        av_actual += test_n_labels[i];
    }
    fprintf(parm->out_file, "@%s average number of PREDICTED relevant labels: %.2lf\n", message, av_pred/test_instances);
    fprintf(parm->out_file, "@%s average number of ACTUAL relevant labels: %.2lf\n", message, av_actual/test_instances);

    // Evaluate results
    t_Performance performance;
    if (classification)
        PrintClassificationMetrics(&performance, test_labels, test_n_labels, h, n_h, test_instances, classes, true, parm->out_file, message, true);

    if (ranks)
        PrintRankingMetrics(&performance, test_labels, test_n_labels, f, test_instances, classes, parm->out_file, message, true);

    scalar reduction = 0.0;
    for (int i = 0; i < instances; i++)
        if ((S[i] = selection[best][i]) <= 0)
            reduction++;

    // Free memory
    FREE(fitness);
    FREE(list);
    for (int i = 0; i < 2*parm->population_size; i++) {
        FreeLAMLkNN(&individual[i], classes);
        FREE(selection[i]);
    }
    FREE(selection);

    if (neighbors_cache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }
    if (distance_cache) {
        for (int i = 0; i < instances; i++)
            FREE(distance_cache[i]);
        FREE(distance_cache);
    }

    return reduction/instances;
}

scalar EvaluateLAMLkNNFitness(const t_Parameters *parm, t_LAMLkNN *lamlknn, int *S, cache_t **neighbors_cache, int instances, int **labels, int *n_labels, int classes, scalar **data, int **index,
                              int inputs)
{
    // Performance
    scalar fitness = EvaluateClassifierPerformance(parm, LAMLkNN, (void *) lamlknn, S, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, true, NULL, 0, NULL,
                                                   0, NULL, NULL);

    // Reduction
    scalar reduction = 0.0, ms_reduction = 0.0;
    if (parm->reduction_ml_w > 0.0 || (parm->multiselection && parm->reduction_ml_ms_w > 0.0))
        for (int i = 0; i < instances; i++)
            if (S[i] > 0) {
                reduction++;
                ms_reduction += S[i];
            }
    reduction = 1.0 - reduction/instances;
    ms_reduction = 1.0 - ms_reduction/instances;

    if (!parm->combine_metrics_gmean) {
        // Reduction
        fitness += parm->reduction_ml_w * reduction;
        fitness += parm->reduction_ml_ms_w * ms_reduction;
    } else {
        fitness *= pow(reduction, parm->reduction_ml_w);
        fitness *= pow(ms_reduction, parm->reduction_ml_ms_w);
    }

    return fitness;
}
