/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "ml.h"

// Perform instance considering possible data partition

scalar InstanceSelectionML(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index, int **selection, int ***ss, int *ks,
                         t_MLClassifier method, int **ksc, scalar **test_data, int **test_labels, int *test_n_labels, int test_instances, int **test_index, scalar *IRLbl, scalar MeanIR,
                         int **chain)
{
    scalar reduction = 0.0;
    int generations = 0;

    // Data partition
    if (parm->data_partition_method == NoDataPartition || !parm->subset_size || instances < parm->subset_size) {
        scalar reduction = InstanceSelectionAlgorithmML(parm, data, labels, n_labels, instances, inputs, classes, index, selection, ss, ks, method, ksc, test_data, test_labels, test_n_labels,
                           test_instances, test_index, &generations, IRLbl, MeanIR, chain);
        SAVEPROGRESS(parm->ml_instance_selection == CHCIS || parm->ml_instance_selection == LAMLkNNIS, "@Actual generations performed: %d\n", generations);
        return reduction;
    } else {
        int n_subsets = instances%parm->subset_size ? instances/parm->subset_size+1 : instances/parm->subset_size, subset_generations;

        SHOWPROGRESS(parm->show_progress > 0, "Data partition into %d subsets of size %d\n", n_subsets, parm->subset_size);
        SAVEPROGRESS(parm->show_progress > 0, "Data partition into %d subsets of size %d\n", n_subsets, parm->subset_size);

        scalar ***subset_data;
        int ***subset_index = NULL, *subset_instances, * **subset_labels, **subset_list, **subset_selection = NULL, **subset_n_labels, ***subset_ss = NULL;

        MALLOC(subset_data, scalar **, n_subsets, "InstanceSelectionML: subset_data");
        if (index)
            MALLOC(subset_index, int **, n_subsets, "InstanceSelectionML: subset_index");
        MALLOC(subset_instances, int, n_subsets, "InstanceSelectionML: subset_list");
        MALLOC(subset_labels, int **, n_subsets, "InstanceSelectionML: subset_labels");
        MALLOC(subset_n_labels, int *, n_subsets, "InstanceSelectionML: subset_n_labels");
        MALLOC(subset_list, int *, n_subsets, "InstanceSelectionML: subset_list");
        for (int i = 0; i < n_subsets; i++) {
            MALLOC(subset_data[i], scalar *, parm->subset_size, "InstanceSelectionML: subset_data[i]");
            if (index) {
                MALLOC(subset_index[i], int *, parm->subset_size, "InstanceSelectionML: subset_index[i]");
            }
            MALLOC(subset_labels[i], int *, parm->subset_size, "InstanceSelectionML: subset_labels[i]");
            MALLOC(subset_n_labels[i], int, parm->subset_size, "InstanceSelectionML: subset_n_labels[i]");
            MALLOC(subset_list[i], int, parm->subset_size, "InstanceSelectionML: subset_labels[i]");
        }

        // Standard or per label instance selection?
        MALLOC(subset_selection, int *, n_subsets, "InstanceSelectionML: subset_selection");
        MALLOC(subset_ss, int **, n_subsets, "InstanceSelectionML: subset_ss");
        if (!parm->ml_ins_sel_per_label) {
            MALLOC(*selection, int, instances, "InstanceSelectionML: *selection");
        } else
            AllocateIntegerMatrix(ss, classes, instances);

        DataPartitionML(parm, data, instances, labels, &n_subsets, &subset_data, &subset_instances, &subset_labels, &subset_list, index, &subset_index, parm->subset_size, n_labels, &subset_n_labels,
                        inputs);

#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < n_subsets; i++) {
            reduction += InstanceSelectionAlgorithmML(parm, subset_data[i], subset_labels[i], subset_n_labels[i], subset_instances[i], inputs, classes, subset_index ? subset_index[i] : NULL,
                         &subset_selection[i], parm->ml_ins_sel_per_label ? &subset_ss[i] : NULL, ks, method, ksc, test_data, test_labels, test_n_labels, test_instances, test_index,
                         &subset_generations, IRLbl, MeanIR, chain);
#ifdef _OPENMP
            #pragma omp critical
#endif
            generations += subset_generations;

            // Aggregate selection
            if (!parm->ml_ins_sel_per_label) {
                for (int j = 0; j < subset_instances[i]; j++)
                    (*selection)[subset_list[i][j]] = subset_selection[i][j];
                FREE(subset_selection[i]);
            } else {
                for (int j = 0; j < subset_instances[i]; j++)
                    for (int l = 0; l < classes; l++)
                        (*ss)[l][subset_list[i][j]] = subset_ss[i][l][j];
                FreeIntegerMatrix(&subset_ss[i], classes);
            }
        }

        SAVEPROGRESS(parm->ml_instance_selection == CHCIS, "@Actual generations performed: %d\n", generations/n_subsets);

        reduction /= n_subsets;

        for (int i = 0; i < n_subsets; i++) {
            FREE(subset_data[i]);
            FREE(subset_labels[i]);
            FREE(subset_list[i]);
            FREE(subset_n_labels[i]);
            if (subset_index)
                FREE(subset_index[i]);
        }

        FREE(subset_data);
        if (index)
            FREE(subset_index);
        FREE(subset_instances);
        FREE(subset_labels);
        FREE(subset_n_labels);
        FREE(subset_list);
        FREE(subset_selection);
        FREE(subset_ss);
    }

    // If none select one randomly
    if (reduction >= 1.0 && selection) {
        reduction = (instances-1.0)/instances;
        (*selection)[(int) RANDOM(0, instances)] = 1;
    }

    return reduction;
}


scalar InstanceSelectionAlgorithmML(const t_Parameters *parm, scalar **data, int **labels, int *n_labels, int instances, int inputs, int classes, int **index, int **selection, int ***ss, int *ks,
                                    t_MLClassifier method, int **ksc, scalar **test_data, int **test_labels, int *test_n_labels, int test_instances, int **test_index, int *generations,
                                    scalar *IRLbl, scalar MeanIR, int **chain)
{
    scalar reduction = 0;

    // Instance selection algorithms
    switch (parm->ml_instance_selection) {
        case NoInstanceSelection:
            reduction = 0.0;
            break;

        case CooperativeCoevolution:
            if (parm->ml_ins_sel_per_label)
                AllocateIntegerMatrix(ss, classes, instances);
            else
                MALLOC(*selection, int, instances, "InstanceSelectionAlgorithmML: selection");
#ifdef MPI
            if (parm->ml_cc_combine_best_subinviduals)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "MPI version for CC without a population");
            reduction = MPICooperativeCoevolutiveInstanceSelection(parm, data, labels, n_labels, instances, inputs, classes, index, *selection);
#else
            if (parm->ml_cc_combine_best_subinviduals)
                reduction = CooperativeCoevolutiveInstanceSelectionWithoutPopulation(parm, data, labels, n_labels, instances, inputs, classes, index, *selection, *ss, method);
            else
                reduction = CooperativeCoevolutiveInstanceSelection(parm, data, labels, n_labels, instances, inputs, classes, index, *selection, *ss, method);
#endif
            break;

        case CHCIS:
            if (parm->ml_ins_sel_per_label)
                AllocateIntegerMatrix(ss, classes, instances);
            else
                MALLOC(*selection, int, instances, "InstanceSelectionAlgorithmML: selection");
            // If apply allocate chain
            if (method == CC)
                MALLOC(*chain, int, classes, "InstanceSelectionAlgorithmML: chain");
            int selected = CHCInstanceSelectionForML(parm, instances, data, labels, n_labels, classes, inputs, selection ? *selection : NULL, index, ss ? *ss : NULL, NULL, ks, method, ksc, generations,
                                                     IRLbl, MeanIR, method == CC ? *chain : NULL);

            reduction = 1.0 - ((scalar) selected)/(parm->ml_ins_sel_per_label ? instances *classes : instances);
            break;

        case LAMLkNNIS:
            if (parm->ml_ins_sel_per_label)
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Instance selection per label for LAML-kNN instance selection");
            else
                MALLOC(*selection, int, instances, "InstanceSelectionAlgorithmML: selection");
            reduction = LAMLkNNInstanceSelection(parm, instances, data, labels, n_labels, classes, inputs, index, *selection, test_data, test_labels, test_n_labels, test_instances, test_index,
                                                 generations);
            break;

        case DEIS:
            MALLOC(*selection, int, instances, "InstanceSelectionAlgorithmML: selection");
            reduction = DifferentialEvolutionInstanceSelection(parm, instances, data, labels, n_labels, classes, inputs, index, *selection, method);
            break;
    }

    return reduction;
}

void ShowMLInstanceSelectionInfo(const t_Parameters *parm, int train_instances, int classes, int *selection, int **ss, scalar reduction)
{
    if (parm->ml_instance_selection == NoInstanceSelection) {
        fprintf(parm->out_file, "@Reduction: %lf\n", 0.0);
        return;
    }

    if (parm->ml_ins_sel_per_label && (parm->ml_instance_selection == CooperativeCoevolution || parm->ml_instance_selection == CHCIS)) {
        scalar reduction = 0.0, per_label_reduction = 0.0;

        for (int i = 0; i < train_instances; i++) {
            bool none = true;
            for (int j = 0; j < classes; j++)
                if (ss[j][i]) {
                    none = false;
                    per_label_reduction++;
                }

            if (none)
                reduction++;
        }

        reduction /= train_instances;
        per_label_reduction = 1.0 - per_label_reduction/(train_instances*classes);

        fprintf(parm->out_file, "@Reduction: %lf\n", reduction);
        fprintf(parm->out_file, "@Per label reduction: %lf\n", per_label_reduction);
    } else {
        fprintf(parm->out_file, "@Reduction: %lf\n", reduction);

        if (selection)
            fprintf(parm->out_file, "Selected instances: %d/%d\n", (int)(train_instances*(1.0-reduction)), train_instances);

        // If multi-selection summarize results
        if (parm->multiselection) {
            int *count, c = 0;
            scalar selected = 0.0;

            int max_ms = parm->ml_max_ioversampling_ratio != 0 ? parm->ml_max_ioversampling_ratio : parm->ml_max_k;

            CALLOC(count, int, max_ms+1, "main: count");
            for (int i = 0; i < train_instances; i++)
                if (selection[i] > 0) {
                    count[selection[i]]++;
                    c++;
                    selected += selection[i];
                } else
                    count[0]++;

            fprintf(parm->out_file, "Results for multi-selection of instances:\n");
            fprintf(parm->out_file, "[Unselected: %d]", count[0]);
            for (int i = 1; i < max_ms+1; i++)
                if (count[i])
                    fprintf(parm->out_file, " [%d time(s): %d]", i, count[i]);

            fprintf(parm->out_file, "\n@Multi-selection average number of times each instance is selected: %lf\n", c ? selected/c : 0.0);
        }
    }
}
