/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"
#include "data.h"
#include "abandoned.h"

// Stub for MLk'sNN

void GetKPerLabel(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *ks, int *selection)
{

    int **s_labels, *s_n_labels, s_instances, **s_index = NULL, *s_selection, neighbors_cache_size = 0, cv_k = 0;
    scalar **s_data;
    bool *selected = NULL;
    cache_t **neighbors_cache = NULL;
    t_MLksNN mlksnn;

    // Cache of neighbors?
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->mlksnn_loo);

    // Used to break ties
    if (parm->mlksnn_ties_closer_to_cv_k || parm->regularize_mlksnn) {
        int old_k;
        old_k = parm->mlknn_k;
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, selection, NULL, parm->cv_folds);
        cv_k = parm->mlknn_k;
        ((t_Parameters *)parm)->mlknn_k = old_k;
    }

    scalar *KS;
    MALLOC(s_labels, int *, instances, "GetKPerLabel: s_labels");
    if (index)
        MALLOC(s_index, int *, instances, "GetKPerLabel: s_index");
    MALLOC(s_n_labels, int, instances, "GetKPerLabel: s_n_labels");
    MALLOC(s_data, scalar *, instances, "GetKPerLabel: s_data");
    MALLOC(s_selection, int, instances, "GetKPerLabel: s_selection");
    MALLOC(KS, scalar, classes, "GetKPerLabel: KS");
    if (!parm->with_replacement)
        MALLOC(selected, bool, instances, "GetKPerLabel: selected");
    ZeroVector(KS, NULL, classes);

    // First rounds with whole dataset
    s_instances = 0;
    for (int i = 0; i < instances; i++) {
        s_data[s_instances] = data[i];
        if (index)
            s_index[s_instances] = index[i];
        s_labels[s_instances] = labels[i];
        s_n_labels[s_instances] = n_labels[i];
        s_selection[s_instances] = selection ? selection[i] : 1;
        s_instances++;
    }

    // Matrix of weights
    scalar **W = NULL;
    if (parm->ml_ensemble_method == MLAdaBoostMH) {
        AllocateScalarMatrix(&W, classes, instances);

        for (int i = 0; i < instances; i++)
            for (int l = 0; l < classes; l++)
                W[l][i] = 1.0/(instances*classes);
    }

    for (int round = 0; round < parm->mlksnn_rounds; round++) {
        switch (parm->ml_kperlabel_method) {
            case Evolutionary:
                EvolveKPerLabel(parm, s_instances, s_data, s_labels, s_n_labels, classes, inputs, s_index, ks, neighbors_cache);
                break;

            case SingleLabel:
                CrossValidateKPerLabel(parm, s_instances, s_data, s_labels, s_n_labels, classes, inputs, s_index, ks, s_selection, &neighbors_cache, round != 0 ? W : NULL, cv_k);
                break;

            case GreedyPerLabel:
                GreedyKPerLabel(parm, s_instances, s_data, s_labels, s_n_labels, classes, inputs, s_index, ks, s_selection, neighbors_cache, cv_k);
                break;

            case PerpendicularSearch:
                PerpendicularSearchPerLabel(parm, s_instances, s_data, s_labels, s_n_labels, classes, inputs, s_index, ks, s_selection, round != 0 ? W : NULL, neighbors_cache, cv_k);
                break;

            case DifferentialEvolution:
                DifferentialEvolutionMLksNN(parm, s_instances, s_data, s_labels, s_n_labels, classes, inputs, s_index, ks);
                break;

            case LAMLksNNEvolutionary:
                OnErrorExit(ERR_UNSPECTED_RESULT, "With LAML-k'sNN Evolutionary method you should not be here");
                break;
        }

        for (int i = 0; i < classes; i++)
            KS[i] += ks[i];

        fprintf(parm->out_file, "LAST iteration ML-ksNN: (");
        for (int i = 0; i < classes-1; i++)
            fprintf(parm->out_file, "%*d, ", WIDTH(parm->mlksnn_max_k), ks[i]);
        fprintf(parm->out_file, "%*d)\n", WIDTH(parm->mlksnn_max_k), ks[classes-1]);

        SAVEPROGRESS(parm->show_progress > 0, "Last iteration performance: %lf\n",
            EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, NULL, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, ks[0], ks, 0, NULL,
                                          NULL));

        PRINTCHARAFTERSPACES(24, '-', (WIDTH(parm->mlksnn_max_k)+2)*classes, parm->out_file);

        fprintf(parm->out_file, "Current BEST ML-ksNN:   (");
        for (int i = 0; i < classes-1; i++)
            fprintf(parm->out_file, "%*d, ", WIDTH(parm->mlksnn_max_k), (int) KS[i]/(round+1));
        fprintf(parm->out_file, "%*d)\n", WIDTH(parm->mlksnn_max_k), (int) KS[classes-1]/(round+1));

        for (int i = 0; i < classes; i++)
            ks[i] = KS[i]/(round+1);
        SAVEPROGRESS(parm->show_progress > 0, "Current best performance: %lf\n",
            EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, NULL, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, ks[0], ks, 0, NULL,
                                          NULL));

        PRINTCHARAFTERSPACES(24, '-', (WIDTH(parm->mlksnn_max_k)+2)*classes, parm->out_file);

        if (parm->mlksnn_rounds > 1) {
            for (int i = 0; i < classes; i++)
                mlksnn.ks[i] = parm->mlksnn_update_ada_weight_with_current_best ? KS[i]/(round+1) : ks[i];

            switch (parm->ml_ensemble_method) {        // Sample for next round
                case MLNone:
                    // Do nothing
                    break;

                case MLeBagging:
                    ((t_Parameters *) parm)->mlksnn_per_label_sampling = false;
                    FreeCacheMatrix(&neighbors_cache, s_instances);
                    MLksNNSampling(parm, data, instances, labels, n_labels, index, selection, s_data, &s_instances, s_labels, s_n_labels, s_index, s_selection, NULL, classes, -1, &neighbors_cache,
                                   inputs);
                    break;

                case MLAdaBoostMH:
                    UpdateAdaBoostMHMLWeights(parm, W,  data, instances, n_labels, labels, index, classes, inputs, selection, NULL, NULL, MLksNN, (void *) &mlksnn, NULL);
                    if (!parm->mlksnn_per_label_sampling) {
                        FreeCacheMatrix(&neighbors_cache, s_instances);
                        MLksNNSampling(parm, data, instances, labels, n_labels, index, selection, s_data, &s_instances, s_labels, s_n_labels, s_index, s_selection, W, classes, -1, &neighbors_cache,
                                       inputs);
                    }
                    break;
            }
        }
    }

    for (int i = 0; i < classes; i++)
        ks[i] = (int) KS[i]/parm->mlksnn_rounds;

    if (parm->regularize_mlksnn) {
        // IMPORTANT: Do not use neighbors cache, otherwise redundant for most cases
        int regularized = RegularizeMLksNN(parm, instances, data, labels, n_labels, classes, inputs, index, ks, selection, NULL, cv_k);
        fprintf(parm->out_file, "@Regularized MLksNN k's: %d\n", regularized);
    }

    fprintf(parm->out_file, "ML-kNN k per label:");
    for (int i = 0; i < classes; i++)
        fprintf(parm->out_file, " [%d, k=%d]", i+1, ks[i]);
    fprintf(parm->out_file, "\n");
    fprintf(parm->out_file, "@Average ML-ksNN k's: %lf\n", AverageIntegerVector(ks, classes));

    FREE(s_labels);
    if (index)
        FREE(s_index);
    FREE(s_data);
    FREE(s_n_labels);
    FREE(s_selection);
    if (selected)
        FREE(selected);
    FREE(KS);
    if (W)
        FreeScalarMatrix(&W, classes);

    if (neighbors_cache)
        FreeCacheMatrix(&neighbors_cache, s_instances);
}


// Evolutionary method

void EvolveKPerLabel(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *ks, cache_t **neighbors_cache)
{
    int *list, *selection, evaluations = 0;
    scalar *fitness, penalty, average_penalty = 0.0;

    // Create and initialize global population
    MALLOC(fitness, scalar, 2*parm->population_size, "EvolveKPerLabel: fitness");
    MALLOC(list, int, parm->population_size, "EvolveKPerLabel: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;

    // No selection
    MALLOC(selection, int, instances, "EvolveKPerLabel: selection");
    for (int i = 0; i < instances; i++)
        selection[i] = 1;

    // Evolution of k
    int **k = NULL;
    AllocateIntegerMatrix(&k, 2*parm->population_size, classes);
    for (int i = 0; i < parm->population_size; i++)
        for (int j = 0; j < classes; j++)
            k[i][j] = (int) RANDOM(parm->mlksnn_min_k, parm->mlksnn_max_k+1);

    // Evaluate initial individuals
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic) private(penalty)
#endif
    for (int i = 0; i < parm->population_size; i++) {
        t_MLksNN mlksnn;
        fitness[i] = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL,
                     k[i][0], k[i], 0, NULL, NULL);

        fitness[i] -= (penalty = RegularizationTerm(parm, k[i], classes));
        average_penalty += penalty;

    }

    evaluations += parm->population_size;

    time_t init_time = time(NULL);

    for (int g = 0; g < parm->generations; g++) {
        SHOWPROGRESS(parm->show_progress > 0, "Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);

        // Random selection without replacement
        RandomizeVector(list, parm->population_size);
        int actual_size = parm->population_size;

        // Generate new individuals using crossover only
        for (int i = 0; i < parm->population_size; i += 2) {
            int father = list[i];
            int mother = list[i+1];

            // Perform BLX-alpha crossover for k
            for (int j = 0; j < classes; j++) {
                int min = MIN(k[father][j], k[mother][j]);
                int max = MAX(k[father][j], k[mother][j]);
                scalar I = max - min;
                k[actual_size][j] = (int) AssignWithinBoundsScalar(parm->mlksnn_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->mlksnn_max_k);
                k[actual_size+1][j] = (int) AssignWithinBoundsScalar(parm->mlksnn_min_k, RANDOM(min - I * BLX_ALPHA, max + I * BLX_ALPHA), parm->mlksnn_max_k);
            }

            actual_size += 2;
        }

        // Evaluate NEW individuals
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = parm->population_size; i < actual_size; i++) {
            t_MLksNN mlksnn;
            fitness[i] = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL,
                         k[i][0], k[i], 0, NULL, NULL);

            fitness[i] -= (penalty = RegularizationTerm(parm, k[i], classes));
            average_penalty += penalty;
        }

        evaluations += parm->population_size;

        ///////////////////////////////////////////
        // Mutation operations ////////////////////
        ///////////////////////////////////////////

        // Random mutation
        if (parm->ml_chc_random_mut > 0.0)
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_chc_random_mut) {
                    NonUniformMutation(k[i], classes, g,  parm->generations, parm->bit_mutation, parm->mlksnn_min_k, parm->mlksnn_max_k);

                    t_MLksNN mlksnn;
                    fitness[i] = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL,
                                 false, NULL, k[i][0], k[i], 0, NULL, NULL);

                    fitness[i] -= (penalty = RegularizationTerm(parm, k[i], classes));
                    average_penalty += penalty;

                    evaluations++;
                }

        // Memetic mutation
        if (parm->mlksnn_memetic_mut > 0.0)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->mlksnn_memetic_mut) {
                    scalar old_fitness = fitness[i];
                    fitness[i] = MLksNNMemeticMutation(parm, k[i], data, instances, labels, n_labels, index, selection, inputs, classes, neighbors_cache);

                    SHOWPROGRESS(parm->show_progress > 1, "Memetic mutation fitness: %6.4lf --> %6.4lf\n", old_fitness, fitness[i]);
                }

        // Local search mutation
        if (parm->mlksnn_ls_mut > 0.0)
            for (int i = 0; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->mlksnn_ls_mut) {
                    scalar old_fitness = fitness[i];
                    fitness[i] = MLksNNLocalSearchMutation(parm, k[i], data, instances, labels, n_labels, index, selection, inputs, classes, neighbors_cache);

                    SHOWPROGRESS(parm->show_progress > 1, "Local search mutation fitness: %6.4lf --> %6.4lf\n", old_fitness, fitness[i]);
                }

        // Reinitialize or elitism
        SHOWPROGRESS(parm->show_progress > 2, "New individuals at generation %d: %d\n", g+1, actual_size - parm->population_size);

        for (int i = 0; i < parm->population_size; i++) {
            int order = OrderedSelection(fitness+i, 0, actual_size-i)+i;

            scalar aux_f = fitness[order];
            fitness[order] = fitness[i];
            fitness[i] = aux_f;

            for (int j = 0; j < classes; j++) {
                int new_k = k[order][j];
                k[order][j] = k[i][j];
                k[i][j] = new_k;
            }
        }

        if (parm->show_progress > 1) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            SHOWPROGRESS(true, "%*d g's best FITNESS, PENALTY: %6.4lf, %6.4lf\n", WIDTH(parm->generations), g+1, fitness[best] + RegularizationTerm(parm, k[best], classes),
                         -RegularizationTerm(parm, k[best], classes));

            VERBOSE(parm->show_progress > 2, parm->out_file, "%*d generation best individual FITNESS (MEAN), PENALTY: %6.4lf (%6.4lf), %6.4lf\n", WIDTH(parm->generations), g+1,
                    fitness[best] + RegularizationTerm(parm, k[best], classes), AverageScalarVector(fitness, parm->population_size), -RegularizationTerm(parm, k[best], classes));
        }

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);
            break;
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    int best = OrderedSelection(fitness, 0, parm->population_size);
    for (int i = 0; i < classes; i++)
        ks[i] = k[best][i];

    t_MLksNN mlksnn;
    scalar perf = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, ks[0],
                  ks, 0, NULL, NULL);
    SHOWPROGRESS(parm->show_progress > 0, "FINAL BEST performance: %6.4lf\n", perf);
    if (average_penalty > ZERO)
        SAVEPROGRESS(true, "Average penalty per evaluation for %d evaluations: %lf\n", evaluations, average_penalty/evaluations);

    perf = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, NULL, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, ks[0], ks, 0, NULL, NULL);
    SHOWPROGRESS(parm->show_progress > 0, "FINAL BEST (no LOO) performance: %6.4lf\n", perf);

    // Free memory
    FREE(fitness);
    FREE(list);
    FREE(selection);

    FreeIntegerMatrix(&k, 2*parm->population_size);
}

// Obtain a k per label using cross-validation of each label separately
// SINGLE-LABEL APPROACH

void CrossValidateKPerLabel(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *ks, int *S, cache_t ***neighbors_cache,
                            scalar **W, int cv_k)
{
    int *partition = NULL, *single_labels = NULL, *list;

    MALLOC(list, int, instances, "CrossValidateKPerLabel: list");
    if (parm->cv_folds) {
#ifdef _OPENMP
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "MLksNN.SL using cross-validation and OpenMP");
#endif
        MALLOC(single_labels, int, instances, "CrossValidateKPerLabel: single_labels");
        MALLOC(partition, int, instances, "CrossValidateKPerLabel: partition");
    }

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int l = 0; l < classes; l++) {
        int *partition = NULL, *single_labels = NULL, **slabels, sinstances, *sn_labels, **sindex = NULL, *sS;  // Local for OPENMP
        scalar best_fitness = -INFINITE, *fitness, **sdata;
        cache_t **lneighbors_cache =  NULL;

        ks[l] = -1;  // Initial best k

        if (parm->mlksnn_per_label_sampling && W) {
            MALLOC(sdata, scalar *, instances, "CrossValidateKPerLabel: sdata");
            MALLOC(slabels, int *, instances, "CrossValidateKPerLabel: slabels");
            MALLOC(sn_labels, int, instances, "CrossValidateKPerLabel: sn_labels");
            if (index)
                MALLOC(sindex, int *, instances, "CrossValidateKPerLabel: sindex");
            MALLOC(sS, int, instances, "CrossValidateKPerLabel: sS");

            MLksNNSampling(parm, data, instances, labels, n_labels, index, S, sdata, &sinstances, slabels, sn_labels, sindex, sS, W, classes, l, &lneighbors_cache, inputs);
        } else {
            sdata = data;
            sinstances = instances;
            slabels = labels;
            sn_labels = n_labels;
            sindex = index;
            sS = S;

            lneighbors_cache = *neighbors_cache;
        }

        MALLOC(fitness, scalar, parm->mlksnn_max_k+1, "CrossValidateKPerLabel: fitness");

        if (parm->cv_folds) {
            for (int i = 0; i < sinstances; i++)
                single_labels[i] = LabelInList(l, slabels[i], sn_labels[i]);

            // Stratified partition
            StratifiedPartition(sinstances, single_labels, parm->cv_folds, 2, partition);
        }

        for (int k = parm->mlksnn_min_k; k <= parm->mlksnn_max_k; k++) {
            if (parm->cv_folds) {
                // Evaluate each fold
                fitness[k] = 0.0;
                for (int fold = 0; fold < parm->cv_folds; fold++) {
                    // Get fold and selection
                    int next = 0;
                    for (int i = 0; i < instances; i++)
                        if (partition[i] == fold) {
                            list[next++] = i;
                            sS[i] = 0;
                        } else
                            sS[i] = (!S || S[i]);

                    fitness[k] += EvaluateSelectionOneLabel(parm, data, instances, labels, inputs, kNN, sS, lneighbors_cache, index, k, NULL, l, NULL, -1, n_labels, parm->cv_performance_measure,
                                                            list,
                                                            next);
                }
                fitness[k] /= parm->cv_folds;
            } else
                fitness[k] = EvaluateSelectionOneLabel(parm, sdata, sinstances, slabels, inputs, kNN, sS, lneighbors_cache, index, k, NULL, l, NULL, -1, sn_labels, parm->cv_performance_measure,
                                                       NULL, -1);

            if (fitness[k] > best_fitness || (parm->mlksnn_ties_closer_to_cv_k && fabs(fitness[k] - best_fitness) < ZERO && abs(k-cv_k) < abs(ks[l]-cv_k))) {
                best_fitness = fitness[k];
                ks[l] = k;
            }

            SHOWPROGRESS(parm->show_progress > 1, "Label %*d PERFORMANCE for k = %*d: %6.4lf\n", WIDTH(classes), l+1, WIDTH(parm->mlksnn_max_k), k, fitness[k]);
        }

        SHOWPROGRESS(parm->show_progress > 0, "Label %*d PERFORMANCE for BEST k = %*d: %6.4lf\n", WIDTH(classes), l+1, WIDTH(parm->mlksnn_max_k), ks[l], best_fitness);

        if (parm->mlksnn_use_window)
            ks[l] = BestWindowedValue(parm->mlksnn_min_k, parm->mlksnn_max_k, parm->mlksnn_window_size, fitness);

        FREE(fitness);

        if (parm->mlksnn_per_label_sampling && W) {
            FREE(sdata);
            FREE(slabels);
            FREE(sn_labels);
            if (index)
                FREE(sindex);
            FREE(sS);

            FreeCacheMatrix(&lneighbors_cache, sinstances);
        }
    }

    if (parm->cv_folds) {
        FREE(single_labels);
        FREE(partition);
    }

    FREE(list);
}

// Obtain a k per label using a stepwise method per randomly chosen label
// If a classification metrics with no penalty it does not make sense

void GreedyKPerLabel(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *KS, int *S, cache_t **neighbors_cache, int cv_k)
{
    scalar fitness, overall_best = -1.0;
    int *list, *mask, *subset_ks, *selection, *ks;
    t_MLksNN mlksnn;

    if (!parm->ml_loo)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Greedy approach without LOO");

    if (parm->mlksnn_use_window)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Window for greedy ML-ksNN");

    // Random order of labels
    MALLOC(list, int, classes, "GreedyKPerLabel: list");
    for (int i = 0; i < classes; i++)
        list[i] = i;

    // Selected all instances
    if (S)
        selection = S;
    else {
        MALLOC(selection, int, instances, "GreedyKPerLabel: selection");
        for (int i = 0; i < instances; i++)
            selection[i] = 1;
    }

    // Initial mask all 0's
    MALLOC(mask, int, classes, "GreedyKPerLabel: mask");

    // K's relative to the subset
    MALLOC(ks, int, classes, "GreedyKPerLabel: ks");
    MALLOC(subset_ks, int, classes, "GreedyKPerLabel: subset_ks");

    // Data for evaluation of subsets of labels
    int **subset_labels, *subset_n_labels;
    MALLOC(subset_n_labels, int, instances, "GreedyKPerLabel: subset_n_labels");
    AllocateIntegerMatrix(&subset_labels, instances, classes);

    time_t init_time = time(NULL);

    for (int attempt = 0; attempt < parm->mlknn_k_per_label_stepwise_attempts; attempt++) {
        RandomizeVector(list, classes);
        for (int i = 0; i < classes; i++)
            mask[i] = ks[i] = 0;

        scalar best_fitness = -INFINITE;
        for (int l = 0; l < classes; l++) {
            // Add to mask
            mask[list[l]] = 1;

            // Get current K's
            int current = -1;
            for (int i = 0, next = 0; i < classes; i++)
                if (mask[i]) {
                    subset_ks[next++] = ks[i];
                    if (i == list[l])
                        current = next-1;
                }

            // Get subset of labels for evaluation
            MaskLabels(n_labels, labels, subset_n_labels, subset_labels, instances, mask, classes);

            best_fitness = -INFINITE;
            int best_k = -1;
            for (int k = parm->mlksnn_min_k; k <= parm->mlksnn_max_k; k++) {
                subset_ks[current] = k;
                fitness = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, subset_labels, subset_n_labels, l+1, data, index, inputs, NULL,
                                                        false, NULL, subset_ks[0], subset_ks, 0, NULL, NULL);

                fitness -= RegularizationTerm(parm, subset_ks, l+1);
                if (fitness > best_fitness || (parm->mlksnn_ties_closer_to_cv_k && fabs(fitness - best_fitness) < ZERO && abs(k-cv_k) < abs(ks[l]-cv_k))) {
                    best_fitness = fitness;
                    best_k = k;
                }
            }

            SHOWPROGRESS(parm->show_progress > 1, "Subset of %*d labels PERFORMANCE for k = %*d: %6.4lf\n", WIDTH(classes), l+1, WIDTH(parm->mlksnn_max_k), best_k, best_fitness +
                         RegularizationTerm(parm, subset_ks, l+1));

            ks[list[l]] = best_k;
        }

        if (overall_best < best_fitness) {
            for (int i = 0; i < classes; i++)
                KS[i] = ks[i];

            overall_best = best_fitness;
        }

        SHOWPROGRESS(parm->show_progress > 0, "Best stepwise PERFORMANCE after %*d attempts: %6.4lf\n", WIDTH(parm->mlknn_k_per_label_stepwise_attempts), attempt+1, overall_best);

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d attempts\n", attempt+1);
            break;
        }
    }

    FREE(subset_n_labels);
    FreeIntegerMatrix(&subset_labels, instances);
    FREE(list);
    FREE(mask);
    FREE(subset_ks);
    if (!S)
        FREE(selection);
    FREE(ks);
}

// Create a new set of labels using a mask
// Only labels masked are kept

void MaskLabels(int *n_labels, int **labels, int *new_n_labels, int **new_labels, int instances, int *mask, int classes)
{
    int *translate;

    // Create conversion array from mask
    MALLOC(translate, int, classes, "MaskLabels: translate");

    int next = 0;
    for (int i = 0; i < classes; i++)
        translate[i] = mask[i] ? next++ : -1;

    for (int i = 0; i < instances; i++) {
        new_n_labels[i] = 0;

        for (int j = 0; j < n_labels[i]; j++)
            if (translate[labels[i][j]] != -1)
                new_labels[i][new_n_labels[i]++] = translate[labels[i][j]];
    }

    FREE(translate);
}

// Search for a label at a time while the remaining k's are fixed to the given label
// For classification metrics with no penalty obtains the optimum training value
// For ranking metrics a k = 0 is advisable to get best value

void PerpendicularSearchPerLabel(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *KS, int *S, scalar **W,
                                 cache_t **neighbors_cache, int cv_k)
{
    int **ks;

    // K's for the search
    AllocateIntegerMatrix(&ks, classes, classes);

    for (int i = 0; i < classes; i++)
        for (int j = 0; j < classes; j++)
            ks[i][j] = parm->mlknn_k;

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int l = 0; l < classes; l++) {
        int *partition = NULL, *single_labels = NULL, best_k = -1, **slabels, sinstances, *sn_labels, **sindex = NULL, *sS;  // Local for OPENMP
        scalar best_fitness = -INFINITE, *fitness, **sdata;
        cache_t **lneighbors_cache =  NULL;
        t_MLksNN mlksnn;

        if (parm->mlksnn_per_label_sampling && W) {
            MALLOC(sdata, scalar *, instances, "PerpendicularSearchPerLabel: sdata");
            MALLOC(slabels, int *, instances, "PerpendicularSearchPerLabel: slabels");
            MALLOC(sn_labels, int, instances, "PerpendicularSearchPerLabel: sn_labels");
            if (index)
                MALLOC(sindex, int *, instances, "PerpendicularSearchPerLabel: sindex");
            MALLOC(sS, int, instances, "PerpendicularSearchPerLabel: sS");

            MLksNNSampling(parm, data, instances, labels, n_labels, index, S, sdata, &sinstances, slabels, sn_labels, sindex, sS, W, classes, l, &lneighbors_cache, inputs);
        } else {
            sdata = data;
            sinstances = instances;
            slabels = labels;
            sn_labels = n_labels;
            sindex = index;
            sS = S;

            lneighbors_cache = neighbors_cache;
        }

        MALLOC(fitness, scalar, parm->mlksnn_max_k+1, "PerpendicularSearchPerLabel: fitness");

        if (parm->cv_folds) {
            MALLOC(single_labels, int, sinstances, "PerpendicularSearchPerLabel: single_labels");
            MALLOC(partition, int, sinstances, "PerpendicularSearchPerLabel: partition");

            for (int i = 0; i < sinstances; i++)
                single_labels[i] = LabelInList(l, slabels[i], sn_labels[i]);

            // Stratified partition
            StratifiedPartition(sinstances, single_labels, parm->cv_folds, 2, partition);
        }

        for (int k = parm->mlksnn_min_k; k <= parm->mlksnn_max_k; k++) {
            ks[l][l] = k;
            fitness[k] = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, sS, partition, lneighbors_cache, sinstances, slabels, sn_labels, classes, sdata, sindex, inputs, NULL, false,
                         NULL, ks[l][l], ks[l], 0, NULL, NULL);

            SHOWPROGRESS(parm->show_progress > 1, "Performance for %*d label for k = %*d: %6.4lf\n", WIDTH(classes), l+1, WIDTH(parm->mlksnn_max_k), k, fitness[k]);

            // If better or...
            // if equal chose the one nearest the global best
            if (fitness[k] > best_fitness || (parm->mlksnn_ties_closer_to_cv_k && fabs(fitness[k] - best_fitness) < ZERO && abs(k-cv_k) < abs(best_k-cv_k))) {
                best_fitness = fitness[k];
                best_k = k;
            }
        }

        if (parm->cv_folds) {
            FREE(single_labels);
            FREE(partition);
        }

        SHOWPROGRESS(parm->show_progress > 1, "Global PERFORMANCE for %*d label BEST k = %*d: %6.4lf\n", WIDTH(classes), l+1, WIDTH(parm->mlksnn_max_k), best_k, best_fitness);

        if (parm->mlksnn_use_window)
            KS[l] = BestWindowedValue(parm->mlksnn_min_k, parm->mlksnn_max_k, parm->mlksnn_window_size, fitness);
        else
            KS[l] = best_k;

        FREE(fitness);

        if (parm->mlksnn_per_label_sampling && W) {
            FREE(sdata);
            FREE(slabels);
            FREE(sn_labels);
            if (index)
                FREE(sindex);
            FREE(sS);

            FreeCacheMatrix(&lneighbors_cache, sinstances);
        }
    }

    FreeIntegerMatrix(&ks, classes);
}

scalar RegularizationTerm(const t_Parameters *parm, int *ks, int labels)
{
    scalar difference = 0.0, average, penalty;
    int *used;

    // Used k's
    CALLOC(used, int, parm->mlksnn_max_k+1, "RegularizationTerm: used");

    // Rubber band penalty
    average = AverageIntegerVector(ks, labels);

    for (int i = 0; i < labels; i++) {
        difference += fabs((scalar) ks[i] - average);
        used[ks[i]] = 1;
    }

    penalty = parm->kperlabel_rubber_band_penalty*(difference/labels);

    // Number of different k's penalty
    penalty += parm->kperlabel_different_ks_penalty*(VectorSum(NULL, used, parm->mlksnn_max_k+1)/labels);

    FREE(used);

    return penalty;
}

scalar EvaluateNFoldsCrossValidation(const t_Parameters *parm, int folds, int *partition, int *S, scalar **data, int instances, cache_t **neighbors_cache, int **index, int *n_labels, int **labels,
                                     int classes, int inputs, t_MLLearner *learner, t_MLClassifier method)
{
    scalar fitness = 0.0;
    int *mask, *selection;

    MALLOC(selection, int, instances, "EvaluateNFoldsCrossValidation: selection");
    MALLOC(mask, int, instances, "EvaluateNFoldsCrossValidation: mask");

    // Evaluate each fold
    fitness = 0.0;
    for (int fold = 0; fold < folds; fold++) {
        // Get fold and selection
        for (int i = 0; i < instances; i++) {
            selection[i] = (partition[i] != fold) && (!S || S[i] > 0);
            mask[i] = (partition[i] == fold);
        }

        fitness +=  EvaluateFoldPerformance(parm, instances, labels, n_labels, classes, inputs, data, selection, neighbors_cache, index, mask, learner, method, NULL, NULL);
    }
    fitness /= folds;

    FREE(selection);
    FREE(mask);

    return fitness;
}

// Replace k's with global k if unable to improve global performance

int RegularizeMLksNN(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, int *ks, int *selection, cache_t **neighbors_cache,
                     int cv_k)
{
    int removed = 0;
    t_MLksNN mlksnn;

    for (int l = 0; l < classes; l++)
        if (ks[l] != cv_k) {
            int *partition = NULL, *single_labels = NULL;  // Local for OPENMP

            if (parm->cv_folds) {
                MALLOC(single_labels, int, instances, "RegularizeMLksNN: single_labels");
                MALLOC(partition, int, instances, "RegularizeMLksNN: partition");

                for (int i = 0; i < instances; i++)
                    single_labels[i] = LabelInList(l, labels[i], n_labels[i]);

                // Stratified partition
                StratifiedPartition(instances, single_labels, parm->cv_folds, 2, partition);
            }

            scalar mlksnn_perf, mlknn_perf;

            // Evaluate with current ML-ksNN
            mlksnn_perf = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, partition, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false,
                          NULL, ks[0], ks, 0, NULL, NULL);

            // Set current k to global best
            int old = ks[l];
            ks[l] = cv_k;

            // Evaluate with k replaced
            for (int j = 0; j < classes; j++)
                mlksnn.ks[j] = ks[j];
            mlknn_perf = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, partition, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false,
                         NULL, ks[0], ks, 0, NULL, NULL);

            // Must improve
            if (mlksnn_perf >= mlknn_perf) {
                ks[l] = old;
                SHOWPROGRESS(parm->show_progress > 0, "ML-ksNN k = %*d for label %*d KEPT: %6.4lf >= %6.4lf\n", WIDTH(parm->mlksnn_max_k), ks[l], WIDTH(classes), l+1, mlksnn_perf, mlknn_perf);
            } else {
                SHOWPROGRESS(parm->show_progress > 0, "ML-ksNN k = %*d for label %*d SUBSTITUTED by global k value %d: %6.4lf < %6.4lf\n", WIDTH(parm->mlksnn_max_k), old, WIDTH(classes), l+1,
                             cv_k, mlksnn_perf, mlknn_perf);
                removed++;
            }

            if (parm->cv_folds) {
                FREE(single_labels);
                FREE(partition);
            }
        } else
            SHOWPROGRESS(parm->show_progress > 0, "ML-ksNN k = %*d for label %*d EQUAL to global k value %d\n", WIDTH(parm->mlksnn_max_k), ks[l], WIDTH(classes), l+1, cv_k);

    SHOWPROGRESS(true, "Regularization for ML-ksNN removed %d k's\n", removed);

    return removed;
}

// Optimize the k for one randomly chose label

scalar MLksNNMemeticMutation(const t_Parameters *parm, int *ks, scalar **data, int instances, int **labels, int *n_labels, int **index, int *selection, int inputs, int classes,
                             cache_t **neighbors_cache)
{
    scalar fitness, best_fitness = -INFINITE;
    t_MLksNN mlksnn;

    int label = RANDOM(0, classes), best_k = -1;

    for (int j = 0; j < classes; j++)
        mlksnn.ks[j] = ks[j];

    for (int k = parm->mlksnn_min_k; k <= parm->mlksnn_max_k; k++) {
        fitness = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, k, ks,
                                                0, NULL, NULL) - RegularizationTerm(parm, mlksnn.ks, classes);

        if (fitness > best_fitness) {
            best_fitness = fitness;
            best_k = k;
        }
    }

    ks[label] = best_k;
    return best_fitness;
}

scalar MLksNNLocalSearchMutation(const t_Parameters *parm, int *ks, scalar **data, int instances, int **labels, int *n_labels, int **index, int *selection, int inputs, int classes,
                                 cache_t **neighbors_cache)
{
    scalar before, after;
    t_MLksNN mlksnn;
    int label, step;

    before = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, ks[0], ks,
                                           0, NULL, NULL) - RegularizationTerm(parm, ks, classes);

    for (int i = 0; i < parm->mlksnn_ls_steps; i++) {
        label = RANDOM(0, classes);
        step = RANDOM(0.0, 1.0) < 0.5 ? +1 : -1;

        if ((step == -1 && mlksnn.ks[label] == 1) || (step == +1 && mlksnn.ks[label] == parm->mlksnn_max_k))
            continue;

        mlksnn.ks[label] += step;

        after = EvaluateClassifierPerformance(parm, MLksNN, (void *) &mlksnn, selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL, ks[0], ks,
                                              0, NULL, NULL) - RegularizationTerm(parm, mlksnn.ks, classes);

        if (after >= before)
            before = after;
        else
            mlksnn.ks[label] -= step;
    }

    for (int j = 0; j < classes; j++)
        ks[j] = mlksnn.ks[j];

    return before;
}


#define ZEROONE2MINUSPLUSONE(x) (2.0*(x)-1.0)

// #define STANDARD_ADABOOST_ALPHA
#define MIN_ALPHA   0.001
#define MAX_ALPHA   100.0

void UpdateAdaBoostMHMLWeights(const t_Parameters *parm, scalar **W, scalar **data, int instances, int *n_labels, int **labels, int **index, int classes, int inputs, int *selection,
                               cache_t **neighbors_cache, int **SS, t_MLClassifier method, void *classifier, int *subspace)
{
    bool classification, ranks;
    int *n_h, **h, *chain = NULL;
    scalar **f, *alpha;

    if (method != MLksNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "AdaBoost.MH for a classifier other than ML-ksNN");

    // Classify testing set
    AllocateScalarMatrix(&f, instances, classes);
    AllocateIntegerMatrix(&h, instances, classes);
    MALLOC(n_h, int, instances, "UpdateAdaBoostMHMLWeights: n_h");
    MALLOC(chain, int, classes, "UpdateAdaBoostMHMLWeights: chain");
    MALLOC(alpha, scalar, classes, "UpdateAdaBoostMHMLWeights: alpha");


    ClassifyDataset(parm, method, data, h, f, &classification, &ranks, data, instances, labels, n_labels, classes, n_h, inputs, index, index, instances, selection, neighbors_cache, SS,
                    classifier, subspace, NULL);

#ifdef STANDARD_ADABOOST_ALPHA
    // Obtain alpha for every label using standard AdaBoost equation
    scalar *e_i;
    CALLOC(e_i, scalar, classes, "UpdateAdaBoostMHMLWeights: e_i");
    for (int i = 0; i < instances; i++)
        for (int l = 0; l < classes; l++) {
            bool relevant = LabelInList(l, labels[i], n_labels[i]);

            if ((relevant && h[i][l] == 0) || (!relevant && h[i][l] == 1))
                e_i[l] += W[l][i];
        }

    for (int l = 0; l < classes; l++) {
        e_i[l] /= instances;

        if (isnan(e_i[l]) || isinf(e_i[l]))
            OnErrorExit(ERR_NUMERICAL, "Error NaN");

        if (e_i[l] <= ZERO)
            alpha[l] = MAX_ALPHA;
        else if (e_i[l] >= 0.5)
            alpha[l] = MIN_ALPHA;
        else
            alpha[l] = 0.5*log((1.0-e_i[l])/e_i[l]);
    }

    FREE(e_i);
#else
    // Alpha constant
    for (int i = 0; i < classes; i++)
        alpha[i] = 1.0;
#endif

    // Update W following AdaBoost.MH
    for (int l = 0; l < classes; l++) {
        scalar Z = 0.0;
        for (int i = 0; i < instances; i++)
            Z += (W[l][i] += exp(-alpha[l]*ZEROONE2MINUSPLUSONE((int) LabelInList(l, labels[i], n_labels[i]))*ZEROONE2MINUSPLUSONE((int) LabelInList(l, h[i], n_h[i]))));

        //  Normalize
        if (Z > ZERO)
            for (int i = 0; i < instances; i++)
                W[l][i] /= Z;
    }

    FreeIntegerMatrix(&h, instances);
    FreeScalarMatrix(&f, classes);
    FREE(chain);
    FREE(n_h);
    FREE(alpha);
}

void MLksNNSampling(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int **index, int *selection, scalar **s_data, int *s_instances, int **s_labels,
                    int *s_n_labels, int **s_index, int *s_selection, scalar **W, int classes, int sampling_label, cache_t ***neighbors_cache, int inputs)
{
    scalar *cprobs = NULL;

    // Create vector of probabilities if apply
    if (parm->ml_ensemble_method == MLAdaBoostMH) {
        MALLOC(cprobs, scalar, instances, "MLksNNSampling: v");

        if (sampling_label != -1) {
            cprobs[0] = W[sampling_label][0];
            for (int i = 1; i < instances; i++)
                cprobs[i] = W[sampling_label][i] + cprobs[i-1];
        } else {
            cprobs[0] = SumColumn(W, classes, 0);
            for (int i = 1; i < instances; i++)
                cprobs[i] = SumColumn(W, classes, i) + cprobs[i-1];
        }
    }

    if (parm->with_replacement) {
        // Bagging/Boosting sample for next round with replacement
        for (*s_instances = 0; *s_instances < instances; (*s_instances)++) {
            int next = (parm->ml_ensemble_method == MLAdaBoostMH) ? RouletteSelection(cprobs, instances) : RANDOM(0, instances);
            s_data[*s_instances] = data[next];
            if (index)
                s_index[*s_instances] = index[next];
            s_labels[*s_instances] = labels[next];
            s_n_labels[*s_instances] = n_labels[next];
            s_selection[*s_instances] = selection[next];
        }
    } else {
        bool *selected;
        MALLOC(selected, bool, instances, "MLksNNSampling: selected");
        for (int i = 0; i < instances; i++)
            selected[i] = false;

        // Bagging/Boosting sample for next round without replacement
        for (int i = 0; i < instances; i++)
            selected[(int)(parm->ml_ensemble_method == MLAdaBoostMH ? RouletteSelection(cprobs, instances) : RANDOM(0, instances))] = true;

        *s_instances = 0;
        for (int i = 0; i < instances; i++)
            if (selected[i]) {
                s_data[*s_instances] = data[i];
                if (index)
                    s_index[*s_instances] = index[i];
                s_labels[*s_instances] = labels[i];
                s_n_labels[*s_instances] = n_labels[i];
                s_selection[*s_instances] = selection[i];
                (*s_instances)++;
            }

        FREE(selected);
    }

    if (parm->use_neighbors_cache) {
        int neighbors_cache_size = 0;
        ConstructAllNeighborsCache(s_data, *s_instances, inputs, NULL, neighbors_cache, NULL, s_index, &neighbors_cache_size, NULL, parm->ml_loo);
    }

    if (cprobs)
        FREE(cprobs);
}
