/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "maths.h"

// Locally adaptive ML-kNN
// D. Wang, J. Wang, F. Hu, L. Li, and X. Zhang (2018) "A Locally Adaptive Multi-Label k-Nearest Neighbor Algorithm", in D. Phung et al. (Eds.): PAKDD 2018, LNAI 10937, pp. 81–93, 2018.

void LocallyAdaptiveMultiLabelkNearestNeighbors(const t_Parameters *parm, scalar *query, scalar **prototype_data, int prototype_instances, int **prototype_labels, int *prototype_n_labels,
        int n_labels, int *predicted_labels, int *predicted_n_labels, int *S, scalar *f, int inputs, int *index, int **prototype_index, cache_t **neighbors_cache, int query_index,
        int **SS, t_LAMLkNN *lamlknn, int **ks)
{
    t_MLkNN mlknn;

    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Per label prototype selection for LAML-kNN");

    // Obtain nearest centroid
    scalar nearest_distance = EuclideanDistanceSparseData(query, lamlknn->centroids[0], inputs, index, lamlknn->centroids_index ? lamlknn->centroids_index[0] : NULL);
    int nearest = 0;
    for (int i = 1; i < lamlknn->n_clusters; i++) {
        scalar distance = EuclideanDistanceSparseData(query, lamlknn->centroids[i], inputs, index, lamlknn->centroids_index ? lamlknn->centroids_index[i] : NULL);
        if (distance < nearest_distance) {
            nearest_distance = distance;
            nearest = i;
        }
    }

    // If nearest had no instances choose another one randomly
    while (lamlknn->prior[nearest] == NULL)
        nearest = RANDOM(0, lamlknn->n_clusters);

    mlknn.k = ks ? MaximumOfIntegerVector(ks[nearest], n_labels) : lamlknn->k;
    mlknn.prior = lamlknn->prior[nearest];
    mlknn.post_0 = lamlknn->post_0[nearest];
    mlknn.post_1 = lamlknn->post_1[nearest];
    MultiLabelkNearestNeighbors(parm, query, prototype_data, prototype_instances, prototype_labels, prototype_n_labels, n_labels, predicted_labels, predicted_n_labels, S, f, inputs, index,
                                prototype_index, neighbors_cache, query_index, &mlknn, SS, ks ? ks[nearest] : NULL, lamlknn->partition, nearest, NULL);
}

void kMeansSparse(scalar **data, int instances, int n_clusters, int inputs, int kmeans_max_iterations, int **index, scalar **centroids, int **centroids_index, int *partition, int *S)
{
    scalar dist_to_nearest, dist;
    int *old_belong, count[n_clusters];
    bool finished;

    // Allocate memory
    MALLOC(old_belong, int, instances, "kMeansSparse: old_belong");

    // Assign random instances as initial centers
    for (int i = 0; i < instances; i++)
        partition[i] = i;
    RandomizeVector(partition, instances);
    for (int i = 0; i < n_clusters; i++)
        if (!index)
            for (int j = 0; j < inputs; j++)
                centroids[i][j] = data[partition[i]][j];
        else {
            int j = 0;
            for (; index[partition[i]][j] != -1; j++) {
                centroids[i][j] = data[partition[i]][j];
                centroids_index[i][j] = index[partition[i]][j];
            }
            centroids_index[i][j] = -1;
        }

    for (int i = 0; i < instances; i++)
        partition[i] = -1;

    // Do while not converged
    int iteration = 0;
    do {
        // Assign instance to cluster
        for (int i = 0; i < instances; i++)
            if (!S || S[i] > 0) {
                old_belong[i] = partition[i];
                partition[i] = 0;
                dist_to_nearest = EuclideanDistanceSparseData(data[i], centroids[0], inputs, index ? index[i] : NULL, centroids_index ? centroids_index[0] : NULL);
                for (int j = 1; j < n_clusters; j++)
                    if ((dist = EuclideanDistanceSparseData(data[i], centroids[j], inputs, index ? index[i] : NULL, centroids_index ? centroids_index[j] : NULL)) < dist_to_nearest) {
                        dist_to_nearest = dist;
                        partition[i] = j;
                    }
            }

        // Recalculate centroids
        for (int i = 0; i < n_clusters; i++) {
            count[i] = 0;
            for (int j = 0; j < inputs; j++)
                centroids[i][j] = 0.0;
        }

        for (int i = 0; i < instances; i++)
            if (!S || S[i] > 0) {
                count[partition[i]] += (S ? S[i] : 1);
                if (!index)
                    for (int j = 0; j < inputs; j++)
                        centroids[partition[i]][j] += data[i][j] * (S ? S[i] : 1);
                else {
                    for (int j = 0; index[i][j] != -1; j++)
                        centroids[partition[i]][index[i][j]] += data[i][j] * (S ? S[i] : 1);
                }
            }

        for (int i = 0; i < n_clusters; i++)
            if (count[i])
                for (int j = 0; j < inputs; j++)
                    centroids[i][j] /= count[i];

        // Convert to sparse if needed
        if (index) {
            for (int i = 0; i < n_clusters; i++) {
                int c = 0;
                for (int j = 0; j < inputs; j++)
                    if (fabs(centroids[i][j]) > ZERO) {
                        centroids[i][c] = centroids[i][j];
                        centroids_index[i][c] = j;
                        c++;
                    }
                centroids_index[i][c] = -1;
            }
        }

        // Check if something changed
        finished = true;
        for (int i = 0; i < instances && finished; i++)
            if ((!S || S[i] > 0) && partition[i] != old_belong[i])
                finished = false;

        iteration++;

    } while (!finished && iteration < kmeans_max_iterations);


    FREE(old_belong);
}

void TrainLocallyAdaptiveMultiLabelkNearestNeighbors(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int *S, int inputs, int **index, int **SS,
        t_LAMLkNN *lamlknn, cache_t **neighbors_cache, int k, int n_clusters, bool create)
{
    scalar **cl_data = NULL;
    int cl_instances, **cl_index = NULL, **cl_labels, *cl_n_labels, *cl_S, *translation;
    t_MLkNN mlknn;
    cache_t **cl_neighbors_cache = NULL;

    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Per label prototype selection for LAML-kNN");

    // Create classifier
    if (create)
        CreateLAMLkNNClassifier(lamlknn, instances, index, inputs, k, n_clusters, classes);
    lamlknn->clustered = false;

    if (!lamlknn->clustered) {
        time_t init_time = time(NULL);
        kMeansSparse(data, instances, lamlknn->n_clusters, inputs, parm->kmeans_max_iterations, index, lamlknn->centroids, lamlknn->centroids_index, lamlknn->partition, S);
        lamlknn->clustered = true;
        SHOWPROGRESS(parm->time_all, "k-Means sparse algorithm: %ld s\n", time(NULL) - init_time);
    }

    // Get data belonging to the cluster
    MALLOC(cl_data, scalar *, instances, "TrainLocallyAdaptiveMultiLabelkNearestNeighbors: cl_data");
    if (index)
        MALLOC(cl_index, int *, instances, "TrainLocallyAdaptiveMultiLabelkNearestNeighbors: cl_index");
    MALLOC(cl_labels, int *, instances, "TrainLocallyAdaptiveMultiLabelkNearestNeighbors: cl_labels");
    MALLOC(cl_n_labels, int, instances, "TrainLocallyAdaptiveMultiLabelkNearestNeighbors: cl_n_labels");
    MALLOC(cl_S, int, instances, "TrainLocallyAdaptiveMultiLabelkNearestNeighbors: cl_S");
    MALLOC(translation, int, instances, "TrainLocallyAdaptiveMultiLabelkNearestNeighbors: translation");

    if (neighbors_cache)
        AllocateCacheMatrix(&cl_neighbors_cache, instances, instances);

    // Train all ML-kNN's
    for (int cluster = 0; cluster < lamlknn->n_clusters; cluster++) {
        cl_instances = 0;

        for (int i = 0; i < instances; i++)
            translation[i] = -1;

        for (int i = 0; i < instances; i++)
            if (lamlknn->partition[i] == cluster) {
                cl_data[cl_instances] = data[i];
                cl_labels[cl_instances] = labels[i];
                cl_n_labels[cl_instances] = n_labels[i];
                if (index)
                    cl_index[cl_instances] = index[i];
                cl_S[cl_instances] = !S ? 1 : S[i];

                // For translating cache
                translation[i] = cl_instances;

                cl_instances++;
            }

        // Instances in the cluster?
        if (!cl_instances)
            continue;

        if (neighbors_cache) {
            time_t init_time = time(NULL);
            TranslateCache(translation, neighbors_cache, cl_neighbors_cache, instances);
            SHOWPROGRESS(parm->time_all, "Translate cache: %ld s\n", time(NULL) - init_time);
        }

        mlknn.k = lamlknn->k;
        mlknn.prior = lamlknn->prior[cluster];
        mlknn.post_0 = lamlknn->post_0[cluster];
        mlknn.post_1 = lamlknn->post_1[cluster];

        time_t init_time = time(NULL);
        TrainMultiLabelkNearestNeighbors(parm, cl_data, cl_instances, cl_labels, cl_n_labels, classes, cl_S, inputs, cl_index, cl_neighbors_cache, &mlknn, SS, lamlknn->k, false, NULL);
        SHOWPROGRESS(parm->time_all, "Training ML-kNN cache: %ld s\n", time(NULL) - init_time);
    }

    FREE(cl_data);
    if (index)
        FREE(cl_index);
    FREE(cl_labels);
    FREE(cl_n_labels);
    FREE(cl_S);
    FREE(translation);

    if (neighbors_cache)
        FreeCacheMatrix(&cl_neighbors_cache, instances);
}

void FreeLAMLkNN(t_LAMLkNN *lamlknn, int n_labels)
{
    FREE(lamlknn->partition);
    for (int i = 0; i < lamlknn->n_clusters; i++) {
        if (lamlknn->prior[i]) {
            FREE(lamlknn->prior[i]);
            for (int j = 0; j < n_labels; j++) {
                FREE(lamlknn->post_0[i][j]);
                FREE(lamlknn->post_1[i][j]);
            }
            FREE(lamlknn->post_0[i]);
            FREE(lamlknn->post_1[i]);
        }
    }
    FREE(lamlknn->prior);
    FREE(lamlknn->post_0);
    FREE(lamlknn->post_1);
    for (int i = 0; i < lamlknn->n_clusters; i++) {
        FREE(lamlknn->centroids[i]);
        if (lamlknn->centroids_index)
            FREE(lamlknn->centroids_index[i]);
    }
    FREE(lamlknn->centroids);
    if (lamlknn->centroids_index)
        FREE(lamlknn->centroids_index);
}

// Adapt a cache for a sample without repeated entries

void TranslateCache(int *translation, cache_t **original, cache_t **sampled, int instances)
{
    int next = 0;
    for (int i = 0; i < instances; i++) {
        if (translation[i] == -1)
            continue;

        for (int j = 0; j < instances; j++)
            sampled[next][j] = (original[i][j] == -1) ? -1 : translation[original[i][j]];

        next++;
    }
}

void CreateLAMLkNNClassifier(t_LAMLkNN *lamlknn, int instances, int **index, int inputs, int k, int n_clusters, int n_labels)
{
    lamlknn->n_clusters = n_clusters;
    lamlknn->k = k;
    lamlknn->clustered = false; // K-means not yet carried out

    MALLOC(lamlknn->partition, int, instances, "CreateLAMLkNNClassifier: partition");

    MALLOC(lamlknn->centroids, scalar *, n_clusters, "CreateLAMLkNNClassifier: lamlknn->centroids");
    if (index) {
        MALLOC(lamlknn->centroids_index, int *, n_clusters, "CreateLAMLkNNClassifier: lamlknn->centroids_index");
    } else
        lamlknn->centroids_index = NULL;

    for (int i = 0; i < n_clusters; i++) {
        MALLOC(lamlknn->centroids[i], scalar, inputs+1, "CreateLAMLkNNClassifier: lamlknn->centroids[i]");
        if (index)
            MALLOC(lamlknn->centroids_index[i], int, inputs+1, "CreateLAMLkNNClassifier: lamlknn->centroids_index[i]");
    }

    MALLOC(lamlknn->prior, scalar *, n_clusters, "CreateLAMLkNNClassifier: lamlknn->prior");
    MALLOC(lamlknn->post_0, scalar **, n_clusters, "CreateLAMLkNNClassifier: lamlknn->post_0");
    MALLOC(lamlknn->post_1, scalar **, n_clusters, "CreateLAMLkNNClassifier: lamlknn->post_1");
    for (int i = 0; i < n_clusters; i++) {
        MALLOC(lamlknn->prior[i], scalar, n_labels, "CreateLAMLkNNClassifier: lamlknn->prior[i]");
        MALLOC(lamlknn->post_0[i], scalar *, n_labels, "CreateLAMLkNNClassifier: lamlknn->post_0[i]");
        MALLOC(lamlknn->post_1[i], scalar *, n_labels, "CreateLAMLkNNClassifier: lamlknn->post_1[i]");
        for (int j = 0; j < n_labels; j++) {
            MALLOC(lamlknn->post_0[i][j], scalar, k+1, "CreateLAMLkNNClassifier: lamlknn->post_0[i][j]");
            MALLOC(lamlknn->post_1[i][j], scalar, k+1, "CreateLAMLkNNClassifier: lamlknn->post_1[i][j]");
        }
    }
}


void GetkNeighborsForMLWithPartition(scalar *query, scalar **prototype_data, int prototype_instances, int **prototype_index, int inputs, int *index, int k, int *neighbors, int *S,
                                     cache_t **neighbors_cache, int instance, bool multiselection_as_local_k, int **SS, int classes, int *part_info, int partition, int *subspace)
{
    for (int i = 0; i < k; i++)
        neighbors[i] = -1;

    // Cache?
    if (neighbors_cache) {
        if (!S && !SS) {
            for (int i = 0, current = 0; current < k && i < prototype_instances; i++)
                if (part_info[neighbors_cache[instance][i]] == partition)
                    neighbors[current++] = neighbors_cache[instance][i];
        } else if (SS) {
            for (int i = 0, current = 0; i < prototype_instances && current < k; i++)
                if (part_info[neighbors_cache[instance][i]] == partition && !NullColumn(SS, classes, neighbors_cache[instance][i]))
                    neighbors[current++] = neighbors_cache[instance][i];
        } else {
            if (multiselection_as_local_k) {
                int count;
                for (int i = 0, current = 0, local_k = 0; i < prototype_instances && local_k < k; i++) {
                    if (part_info[neighbors_cache[instance][i]] == partition && (count = neighbors_cache[instance][i] != -1 ? MAX(0, S[neighbors_cache[instance][i]]) : 0)) {
                        neighbors[current++] = neighbors_cache[instance][i];
                        local_k += count;
                    }
                }
            } else {
                for (int i = 0, current = 0; i < prototype_instances && current < k; i++) {
                    int count = (part_info[neighbors_cache[instance][i]] == partition && neighbors_cache[instance][i] != -1) ? MAX(0, S[neighbors_cache[instance][i]]) : 0;
                    while (count-- && current < k)
                        neighbors[current++] = neighbors_cache[instance][i];
                }
            }
        }

        return;
    }

    RANK *rank;
    MALLOC(rank, RANK, prototype_instances, "main: r");

    int next = 0;
    for (int j = 0; j < prototype_instances; j++)
        if (part_info[j] == partition && ((!S && !SS) || (S && S[j] > 0) || (SS && !NullColumn(SS, classes, j)))) {
            rank[next].fitness = -EuclideanDistanceSparseDataSubspace(query, prototype_data[j], inputs, index, prototype_index ? prototype_index[j] : NULL, subspace);
            rank[next].being = j;
            next++;
        }

    // Sort
    qsort(rank, next, sizeof(RANK), Compare);

    // Store neighbors in ascending order of distance
    int current = 0;
    if (multiselection_as_local_k) {
        int local_k = 0;
        for (int j = 0; local_k < k && j < next; j++) {
            int count = S ? MAX(0, S[rank[j].being]) : 1;
            neighbors[current++] = rank[j].being;
            local_k += count;
        }
    } else {
        for (int j = 0; current < k && j < next; j++) {
            int count = S ? MAX(0, S[rank[j].being]) : 1;
            while (count-- && current < k)
                neighbors[current++] = rank[j].being;
        }
    }

    // Not enough?
    for (int j = current; j < k; j++)
        neighbors[j] = -1;

    FREE(rank);
}
