/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "common.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "maths.h"

void CreateLAMLksNNClassifier(const t_Parameters *parm, t_LAMLksNN *lamlksnn, int classes, int **ks, scalar **data, int instances, int *n_labels, int **labels, int inputs, int **index,
                              int n_clusters)
{
    // If minimum k is less than zero consider an interval around the optimum global value
    if (parm->mlksnn_min_k < 0) {
        int old_k;
        old_k = parm->mlknn_k;
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);
        int cv_k = parm->mlknn_k;
        ((t_Parameters *)parm)->mlknn_k = old_k;

        int increase = MAX(0, -(cv_k+parm->mlksnn_min_k));

        ((t_Parameters *) parm)->mlksnn_min_k = MAX(1, cv_k+parm->mlksnn_min_k);
        ((t_Parameters *) parm)->mlksnn_max_k = MIN(instances-1, cv_k+parm->mlksnn_max_k);

        // Increase if left bound bellow 0
        ((t_Parameters *) parm)->mlksnn_max_k = MIN(instances, parm->mlksnn_max_k+increase);

        SHOWPROGRESS(true, "Interval for multiple k's adjusted to [%d, %d]\n", parm->mlksnn_min_k, parm->mlksnn_max_k);
        SAVEPROGRESS(true, "Interval for multiple k's adjusted to [%d, %d]\n", parm->mlksnn_min_k, parm->mlksnn_max_k);
    }

    MALLOC(lamlksnn->ks, int *, n_clusters, "main: ks");
    for (int c = 0; c < n_clusters; c++) {
        MALLOC(lamlksnn->ks[c], int, classes, "CreateLAMLksNNClassifier: lamlksnn->ks[c]");
        for (int i = 0; i < classes; i++)
            lamlksnn->ks[c][i] = ks ? ks[c][i] : -1;
    }

    MALLOC(lamlksnn->lamlknn, t_LAMLkNN, 1, "CreateLAMLksNNClassifier: lamlksnn->lamlknn");
    CreateLAMLkNNClassifier(lamlksnn->lamlknn, instances, index, inputs, parm->mlksnn_max_k, n_clusters, classes);
}

void FreeLAMLksNN(t_LAMLksNN *lamlksnn, int n_labels)
{
    for (int i = 0; i < lamlksnn->lamlknn->n_clusters; i++)
        FREE(lamlksnn->ks[i]);
    FREE(lamlksnn->ks);

    FreeLAMLkNN(lamlksnn->lamlknn, n_labels);
    FREE(lamlksnn->lamlknn);
}


void TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int *S, int inputs, int **index, int **SS,
        t_LAMLksNN *lamlksnn, cache_t **neighbors_cache)
{
    scalar **cl_data = NULL;
    int cl_instances, **cl_index = NULL, **cl_labels, *cl_n_labels, *cl_S, *translation;
    t_MLksNN mlksnn;
    cache_t **cl_neighbors_cache = NULL;

    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Per label prototype selection for LAML-kNN");

    if (!lamlksnn->lamlknn->clustered)
        kMeansSparse(data, instances, lamlksnn->lamlknn->n_clusters, inputs, parm->kmeans_max_iterations, index, lamlksnn->lamlknn->centroids, lamlksnn->lamlknn->centroids_index,
                     lamlksnn->lamlknn->partition, S);

    // Get data belonging to the cluster
    MALLOC(cl_data, scalar *, instances, "TrainLocallyAdaptiveMultiLabelkNearestsNeighbors: cl_data");
    if (index)
        MALLOC(cl_index, int *, instances, "TrainLocallyAdaptiveMultiLabelkNearestsNeighbors: cl_index");
    MALLOC(cl_labels, int *, instances, "TrainLocallyAdaptiveMultiLabelkNearestsNeighbors: cl_labels");
    MALLOC(cl_n_labels, int, instances, "TrainLocallyAdaptiveMultiLabelkNearestsNeighbors: cl_n_labels");
    MALLOC(cl_S, int, instances, "TrainLocallyAdaptiveMultiLabelkNearestsNeighbors: cl_S");
    MALLOC(translation, int, instances, "TrainLocallyAdaptiveMultiLabelkNearestsNeighbors: translation");

    if (neighbors_cache)
        AllocateCacheMatrix(&cl_neighbors_cache, instances, instances);

    // Train all ML-kNN's
    for (int cluster = 0; cluster < lamlksnn->lamlknn->n_clusters; cluster++) {
        cl_instances = 0;

        for (int i = 0; i < instances; i++)
            translation[i] = -1;

        for (int i = 0; i < instances; i++)
            if (lamlksnn->lamlknn->partition[i] == cluster) {
                cl_data[cl_instances] = data[i];
                cl_labels[cl_instances] = labels[i];
                cl_n_labels[cl_instances] = n_labels[i];
                if (index)
                    cl_index[cl_instances] = index[i];
                cl_S[cl_instances] = !S ? 1 : S[i];

                // For translating cache
                translation[i] = cl_instances;

                cl_instances++;
            }

        // Instances in the cluster?
        if (!cl_instances)
            continue;

        if (neighbors_cache)
            TranslateCache(translation, neighbors_cache, cl_neighbors_cache, instances);

        mlksnn.ks = lamlksnn->ks[cluster];
        mlksnn.prior = lamlksnn->lamlknn->prior[cluster];
        mlksnn.post_0 = lamlksnn->lamlknn->post_0[cluster];
        mlksnn.post_1 = lamlksnn->lamlknn->post_1[cluster];

        TrainMultiLabelKsNearestNeighbors(parm, cl_data, cl_instances, cl_labels, cl_n_labels, classes, S, inputs, cl_index, cl_neighbors_cache, &mlksnn, SS, NULL, false);
    }

    FREE(cl_data);
    if (index)
        FREE(cl_index);
    FREE(cl_labels);
    FREE(cl_n_labels);
    FREE(cl_S);
    FREE(translation);

    if (neighbors_cache)
        FreeCacheMatrix(&cl_neighbors_cache, instances);
}

void LocallyAdaptiveMultiLabelkNearestsNeighbors(const t_Parameters *parm, scalar *query, scalar **prototype_data, int prototype_instances, int **prototype_labels, int *prototype_n_labels,
        int n_labels, int *predicted_labels, int *predicted_n_labels, int *S, scalar *f, int inputs, int *index, int **prototype_index, cache_t **neighbors_cache, int query_index,
        int **SS, t_LAMLksNN *lamlksnn)
{
    LocallyAdaptiveMultiLabelkNearestNeighbors(parm, query, prototype_data, prototype_instances, prototype_labels, prototype_n_labels, n_labels, predicted_labels, predicted_n_labels, S, f, inputs,
            index, prototype_index, neighbors_cache, query_index, SS, lamlksnn->lamlknn, lamlksnn->ks);
}


// Evolutionary method for LAML-ksNN with all parameters at the same time
// Can also evolve a selection

void EvolveAndTrainLAMLksNN(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int **index, cache_t **neighbors_cache, t_LAMLksNN *lamlksnn,
                            int *S, bool is)
{
    int *list, **selection, evaluations = 0;
    scalar *fitness;

    if (parm->cv_folds)
        OnErrorExit(ERR_PARAM, "Cannot evaluate LAML-k'sNN with n-fold cross-validation");

    // Create and initialize global population
    MALLOC(fitness, scalar, 2*parm->population_size, "EvolveAndTrainLAMLksNN: fitness");
    MALLOC(list, int, parm->population_size, "EvolveAndTrainLAMLksNN: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;

    // No selection
    MALLOC(selection, int *, 2*parm->population_size, "EvolveAndTrainLAMLksNN: selection");

    // Evolution of LAML-ksNN
    t_LAMLksNN *individual;
    MALLOC(individual, t_LAMLksNN, 2*parm->population_size, "EvolveAndTrainLAMLksNN: individual");
    for (int i = 0; i < parm->population_size; i++) {
        CreateLAMLksNNClassifier(parm, &individual[i], classes, NULL, data, instances, n_labels, labels, inputs, index, parm->lamlknn_max_n_clusters);
        for (int l = 0; l < parm->lamlknn_max_n_clusters; l++)
            for (int j = 0; j < classes; j++)
                individual[i].ks[l][j] = (int) RANDOM(parm->mlksnn_min_k, parm->mlksnn_max_k+1);
        individual[i].lamlknn->n_clusters = (int) RANDOM(parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters+1);
        individual[i].lamlknn->clustered = true;

        MALLOC(selection[i], int, instances, "EvolveAndTrainLAMLksNN");
        for (int j = 0; j < instances; j++)
            selection[i][j] = is ? RANDOM(0.0, 1.0) < parm->chc_init_prob : (S ? S[i] : 1);

        // Carry out clustering
        kMeansSparse(data, instances, individual[i].lamlknn->n_clusters, inputs, parm->kmeans_max_iterations, index, individual[i].lamlknn->centroids, individual[i].lamlknn->centroids_index,
                     individual[i].lamlknn->partition, selection[i]);
        TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[i], neighbors_cache);
    }

    for (int i = parm->population_size; i < 2*parm->population_size; i++) {
        CreateLAMLksNNClassifier(parm, &individual[i], classes, NULL, data, instances, n_labels, labels, inputs, index, parm->lamlknn_max_n_clusters);
        MALLOC(selection[i], int, instances, "EvolveAndTrainLAMLksNN");
    }

    // Evaluate initial individuals
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < parm->population_size; i++)
        fitness[i] = EvaluateClassifierPerformance(parm, LAMLksNN, (void *) &individual[i], selection[i], NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL,
                     true, NULL, 0, NULL, 0, NULL, NULL);

    evaluations += parm->population_size;

    time_t init_time = time(NULL);

    for (int g = 0; g < parm->generations; g++) {
        SHOWPROGRESS(parm->show_progress > 0, "Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);

        // Random selection[i] without replacement
        RandomizeVector(list, parm->population_size);
        int actual_size = parm->population_size;

        // Generate new individuals using crossover only
        for (int i = 0; i < parm->population_size; i += 2) {
            int father = list[i];
            int mother = list[i+1];

            // Perform HUX crossover creating new individuals
            HuxCrossover(selection[mother], selection[father], selection[actual_size], selection[actual_size+1], instances);

            // Perform BLX-alpha crossover for k
            for (int l = 0; l < parm->lamlknn_max_n_clusters; l++)
                BLXAlphaCrossover(individual[actual_size].ks[l], individual[actual_size+1].ks[l], individual[father].ks[l], individual[mother].ks[l], classes, parm->mlksnn_min_k, parm->mlksnn_max_k);

            // Perform BLX-alpha crossover for number of clusters
            BLXAlphaCrossover(&individual[actual_size].lamlknn->n_clusters, &individual[actual_size+1].lamlknn->n_clusters, &individual[father].lamlknn->n_clusters,
                              &individual[mother].lamlknn->n_clusters, 1, parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters);

            // Carry out clustering
            kMeansSparse(data, instances, individual[actual_size].lamlknn->n_clusters, inputs, parm->kmeans_max_iterations, index, individual[actual_size].lamlknn->centroids,
                         individual[actual_size].lamlknn->centroids_index, individual[actual_size].lamlknn->partition, selection[i]);
            kMeansSparse(data, instances, individual[actual_size+1].lamlknn->n_clusters, inputs, parm->kmeans_max_iterations, index, individual[actual_size+1].lamlknn->centroids,
                         individual[actual_size+1].lamlknn->centroids_index, individual[actual_size+1].lamlknn->partition, selection[i]);
            individual[actual_size].lamlknn->clustered = individual[actual_size+1].lamlknn->clustered = true;
            TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[actual_size], neighbors_cache);
            TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[actual_size+1], neighbors_cache);

            actual_size += 2;
        }

        // Evaluate NEW individuals
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = parm->population_size; i < actual_size; i++)
            fitness[i] = EvaluateClassifierPerformance(parm, LAMLksNN, (void *) &individual[i], selection[i], NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL,
                         true, NULL, 0, NULL, 0, NULL, NULL);

        evaluations += parm->population_size;

        ///////////////////////////////////////////
        // Mutation operations ////////////////////
        ///////////////////////////////////////////
        if (parm->ml_chc_random_mut > 0.0)
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_chc_random_mut) {
                    for (int j = 0; j < parm->lamlknn_max_n_clusters; j++)
                        NonUniformMutation(individual[i].ks[j], classes, g,  parm->generations, parm->bit_mutation, parm->mlksnn_min_k, parm->mlksnn_max_k);
                    NonUniformMutation(&(individual[i].lamlknn->n_clusters), 1, g,  parm->generations, parm->bit_mutation, parm->mlksnn_min_k, parm->mlksnn_max_k);

                    kMeansSparse(data, instances, individual[i].lamlknn->n_clusters, inputs, parm->kmeans_max_iterations, index, individual[i].lamlknn->centroids,
                                 individual[i].lamlknn->centroids_index, individual[i].lamlknn->partition, selection[i]);
                    individual[i].lamlknn->clustered = true;
                    TrainLocallyAdaptiveMultiLabelkNearestsNeighbors(parm, data, instances, labels, n_labels, classes, selection[i], inputs, index, NULL, &individual[i], neighbors_cache);

                    fitness[i] = EvaluateClassifierPerformance(parm, LAMLksNN, (void *) &individual[i], selection[i], NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs,
                                 NULL, true, NULL, 0, NULL, 0, NULL, NULL);

                    evaluations++;
                }

        // Reinitialize or elitism
        SHOWPROGRESS(parm->show_progress > 2, "New individuals at generation %d: %d\n", g+1, actual_size - parm->population_size);

        for (int i = 0; i < parm->population_size; i++) {
            int order = OrderedSelection(fitness+i, 0, actual_size-i)+i;

            t_LAMLksNN aux = individual[i];
            individual[i] = individual[order];
            individual[order] = aux;

            scalar f = fitness[i];
            fitness[i] = fitness[order];
            fitness[order] = f;

            int *s = selection[i];
            selection[i] = selection[order];
            selection[order] = s;

        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            SHOWPROGRESS(true, "%*d g's best FITNESS: %6.4lf\n", WIDTH(parm->generations), g+1, fitness[best]);
            VERBOSE(parm->show_progress > 2, parm->out_file, "%*d generation best individual FITNESS (MEAN): %6.4lf (%6.4lf)\n", WIDTH(parm->generations), g+1, fitness[best],
                    AverageScalarVector(fitness, parm->population_size));
        }

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);
            break;
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    int best = OrderedSelection(fitness, 0, parm->population_size);
    ReallocateCopyLAMLksNNClassifier(parm, lamlksnn, &individual[best], classes, instances, inputs);
    scalar perf = EvaluateClassifierPerformance(parm, LAMLksNN, (void *) lamlksnn, selection[best], NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, true, NULL,
                  0, NULL, 0, NULL, NULL);

    SHOWPROGRESS(parm->show_progress > 0, "FINAL BEST performance: %6.4lf\n", perf);
    SAVEPROGRESS(parm->show_progress > 0, "FINAL BEST performance: %6.4lf\n", perf);

    // Report best results
    fprintf(parm->out_file, "@LAML-ksNN best number of clusters: %d\n", lamlksnn->lamlknn->n_clusters);
    fprintf(parm->out_file, "@LAML-ksNN average k's: %lf\n", AverageIntegerMatrix(lamlksnn->ks, lamlksnn->lamlknn->n_clusters, classes));
    fprintf(parm->out_file, "Optimal evolutionary LAML-ksNN\n");
    for (int j = 0; j < lamlksnn->lamlknn->n_clusters; j++) {
        fprintf(parm->out_file, "Cluster %d: (", j+1);
        for (int i = 0; i < classes-1; i++)
            fprintf(parm->out_file, "%*d, ", WIDTH(parm->mlksnn_max_k), lamlksnn->ks[j][i]);
        fprintf(parm->out_file, "%*d)\n", WIDTH(parm->mlksnn_max_k), lamlksnn->ks[j][classes-1]);
    }

    if (is) {
        scalar reduction = 0.0;
        for (int i = 0; i < instances; i++) {
            if (!selection[best][i])
                reduction++;
            if (S)
                S[i] = selection[best][i];
        }

        fprintf(parm->out_file, "@Reduction: %lf\n", reduction/instances);
    }

    // Free memory
    FREE(fitness);
    FREE(list);
    for (int i = 0; i < 2*parm->population_size; i++) {
        FreeLAMLksNN(&individual[i], classes);
        FREE(selection[i]);
    }
    FREE(selection);
}

// Copy a new LAML-k'sNN classifier on an old one reallocating memory

void ReallocateCopyLAMLksNNClassifier(const t_Parameters *parm, t_LAMLksNN *dest, t_LAMLksNN *or, int classes, int instances, int inputs)
{
    // Free and create a new one
    FreeLAMLksNN(dest, classes);
    CreateLAMLksNNClassifier(parm, dest, classes, NULL, NULL, instances, NULL, NULL, inputs, or->lamlknn->centroids_index, or->lamlknn->n_clusters);

    dest->lamlknn->n_clusters = or->lamlknn->n_clusters;
    dest->lamlknn->clustered = true;
    dest->lamlknn->k = or->lamlknn->k;
    memcpy(dest->lamlknn->partition, or->lamlknn->partition, sizeof(int)*instances);

    for (int j = 0; j < dest->lamlknn->n_clusters; j++) {
        for (int i = 0; i < classes; i++)
            dest->ks[j][i] = or->ks[j][i];

        memcpy(dest->lamlknn->centroids[j], or->lamlknn->centroids[j], sizeof(scalar)*inputs);
        if (or->lamlknn->centroids_index)
            memcpy(dest->lamlknn->centroids_index[j], or->lamlknn->centroids_index[j], sizeof(int)*(inputs+1));
        memcpy(dest->lamlknn->prior[j], or->lamlknn->prior[j], sizeof(scalar)*classes);
        for (int i = 0; i < classes; i++) {
            memcpy(dest->lamlknn->post_0[j][i], or->lamlknn->post_0[j][i], sizeof(scalar)*(dest->lamlknn->k+1));
            memcpy(dest->lamlknn->post_1[j][i], or->lamlknn->post_1[j][i], sizeof(scalar)*(dest->lamlknn->k+1));
        }
    }
}
