/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Label Powerset and Pruned set methods
// Jesse Read, Bernhard Pfahringer, Geoff Holmes: Multi-label Classification Using Ensembles of Pruned Sets. In: ICDM'08: International Conference on Data Mining (ICDM 2008). Pisa, Italy., 2008.

void TrainLabelPowersetModel(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_LP *lp, int *subspace, bool cv, t_Classifier classifier,
                             int *subset, int **SS, int **index, bool ps, bool ppt, bool ppt_n)
{
    int *l_subspace;

    // Test if valid LP classification model
    if (classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "LP classification method");

    // If no subspace select all
    if (!subspace) {
        MALLOC(l_subspace, int, inputs, "TrainLabelPowersetModel: l_subspace");
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = 1;
    } else
        l_subspace = subspace;

    // Label dependent instance selection and standard instance selection are incompatible
    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "LP for partial instance selection");

    // Local data in case it must be modified
    scalar **l_data = data;
    int **l_index = index, **l_labels = labels, *l_n_labels = n_labels;

    // Prune subsets if apply
    int added = 0, *s = subset;
    if (ps || ppt || ppt_n) {
        MALLOC(s, int, instances, "TrainLabelPowersetModel: s");
        if (subset)
            memcpy(s, subset, sizeof(int)*instances);
        else
            FillVector(NULL, s, instances, 1.0);

        AllocateScalarMatrix(&l_data, instances, inputs);
        AllocateIntegerMatrix(&l_labels, instances, classes);
        MALLOC(l_n_labels, int, instances, "TrainLabelPowersetModel: l_n_labels");
        if (index)
            AllocateIntegerMatrix(&l_index, instances, inputs+1);

        CopyMultiLabelDataset(data, labels, n_labels, index, l_data, l_labels, l_n_labels, l_index, instances, inputs);

        added = PruneSubsets(parm->ps_p, parm->ps_b, &l_labels, &l_n_labels, instances, &s, &l_data, &l_index, parm->ps_strategy, inputs, parm->ps_labels_bound, ppt, ppt_n);

        SAVEPROGRESS(parm->show_progress > 0, "PS removed, added instances: %d, %d\n", instances - (int) VectorSum(NULL, s, instances), added);
    }

    // Create single-label classes and translation array
    scalar *sc_labels;
    int *train_labels;
    MALLOC(sc_labels, scalar, instances+added, "TrainLabelPowersetModel: sc_labels");
    MALLOC(train_labels, int, instances+added, "TrainLabelPowersetModel: train_labels");
    for (int i = 0; i < instances+added; i++) {
        scalar label = 0.0;
        for (int j = 0; j < l_n_labels[i]; j++)
            label += (int) pow(2.0, l_labels[i][j]);

        sc_labels[i] = label;
    }

    // Create classes and decodification
    lp->n_classes = 1;
    MALLOC(lp->decoded, int *, 1, "TrainLabelPowersetModel: lp->decoded");
    MALLOC(lp->decoded[0], int, classes, "TrainLabelPowersetModel: lp->decoded[0]");
    DecodifyLabelPowerset(sc_labels[0], classes, lp->decoded[0]);
    train_labels[0] = 0;
    for (int i = 1; i < instances+added; i++) {
            bool present = false;
            for (int j = 0; j < i && !present; j++)
                if (sc_labels[i] == sc_labels[j]) {
                    // Already present
                    present = true;
                    train_labels[i] = train_labels[j];
                }

            if (!present) {
                // Add new one
                lp->n_classes++;
                REALLOC(lp->decoded, int *, lp->n_classes, "TrainLabelPowersetModel: lp->decoded");
                MALLOC(lp->decoded[lp->n_classes-1], int, classes, "TrainLabelPowersetModel: lp->decoded[0]");
                DecodifyLabelPowerset(sc_labels[i], classes, lp->decoded[lp->n_classes-1]);
                train_labels[i] = lp->n_classes-1;
            }
        }

    lp->method = classifier;
    SAVEPROGRESS(parm->show_progress > 1, "LP/PS number of classes: %d\n", lp->n_classes);

    // Cross-validate hyper-parameters?
    if (cv)
        SetCrossValidatedParameters((t_Parameters *) parm, l_data, instances+added, train_labels, l_subspace, inputs, lp->n_classes, classifier, -1, s, l_index);

    // Train
    ClassifierTraining(parm, classifier, l_data, instances+added, train_labels, inputs, lp->n_classes, l_index, l_subspace, s, true, &(lp->model), &(lp->ensemble));

    // For k-NN add labels
    if (classifier == kNN || classifier == NN) {
        MALLOC(lp->model.knn->labels, int, instances+added, "TrainLabelPowersetModel: lp->model.knn->labels");
        memcpy(lp->model.knn->labels, train_labels, instances+added*sizeof(int));
    }

    FREE(sc_labels);
    FREE(train_labels);
    if (!subspace)
        FREE(l_subspace);

    // If pruned sets restore data
    if (ps) {
        FREE(s);
        FreeScalarMatrix(&l_data, instances);
        FreeIntegerMatrix(&l_labels, instances);
        FREE(l_n_labels);
        if (index)
            FreeIntegerMatrix(&l_index, instances);
    }
}

void LabelPowersetClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                                   int inputs, int *query_index, t_LP *lp, int *subspace, int *S, int **index)
{
    scalar *output;

    MALLOC(output, scalar, lp->n_classes, "LabelPowersetClassifyInstance: output");

    // Output of classifier
    int category = ClassifierProbabilities(parm, lp->method, query, inputs, lp->n_classes, query_index, subspace, data, instances, (lp->method == kNN || lp->method == NN) ?
                                           lp->model.knn->labels : NULL, S, &(lp->model), -1, NULL, -1, index, output, &(lp->ensemble));

    // Relevant labels
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++)
        if (lp->decoded[category][i]) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }

    *classification = true;

    // For rankings scikit method is used
    ZeroVector(f, NULL, n_labels);
    for (int i = 0; i < lp->n_classes; i++)
        for (int j = 0; j < n_labels; j++)
            f[j] += output[i]*lp->decoded[i][j];

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *ranks = true;

    FREE(output);
}

void FreeLabelPowerset(const t_Parameters *parm, t_LP *lp, int inputs)
{
    if (lp->method == NN || lp->method == kNN)
        FREE(lp->model.knn->labels);
    FreeIntegerMatrix(&(lp->decoded), lp->n_classes);
    ClassifierFree(parm, &(lp->model), lp->method, lp->n_classes, inputs, &(lp->ensemble));
}

void DecodifyLabelPowerset(scalar category, int n_classes, int *decoded)
{
    for (int i = 0; i < n_classes; i++) {
        decoded[i] = ((int) category) % 2;

        category /= 2;
    }
}

typedef struct {
    int *labels, l, c;
} t_ps_info;

int PruneSubsets(int p, int b, int ***labels, int **n_labels, int instances, int **subset, scalar ***data, int ***index, t_PS_Strategy strategy, int inputs, int upper_bound, bool ppt,
                 bool ppt_n)
{
    int distinct = 0, *c;
    scalar *pss;
    bool *pruned;

    MALLOC(pss, scalar, instances, "PruneSubsets: pss");
    MALLOC(c, int, instances, "PruneSubsets: c");

    for (int i = 0; i < instances; i++)
        if ((*subset)[i]) {
            scalar ps = LabelsToPowerset((*labels)[i], (*n_labels)[i]);

            bool present = false;
            for (int j = 0; j < distinct && !present; j++)
                if (pss[j] == ps) {
                    present = true;
                    c[j]++;
                }

            // If not present add
            if (!present) {
                pss[distinct] = ps;
                c[distinct] = 1;
                distinct++;
            }
        }

    MALLOC(pruned, bool, instances, "PruneSubsets: pruned");

    int kept = instances;

    // If all removed decrease p parameter
    do {
        for (int i = 0; i < instances; i++)
            if ((*subset)[i])  {
                scalar ps = LabelsToPowerset((*labels)[i], (*n_labels)[i]);

                for (int j = 0; j < distinct; j++)
                    if (pss[j] == ps) {
                        pruned[i] = c[j] > p ? false : true;
                        break;
                    }
            } else
                pruned[i] = false;

        kept = 0;
        for (int i = 0; i < instances; i++)
            if ((*subset)[i] && !pruned[i])
                kept++;

        if (!kept)
            p--;
    } while (!kept);

    // Recover pruned subsets
    int **combs, *n_combs;
    t_ps_info *info;

    // PPT method does not add any instances
    if (ppt) {
        for (int i = 0; i < instances; i++)
            (*subset)[i] = !pruned[i];

        FREE(pss);
        FREE(c);
        FREE(pruned);

        return 0;
    }

    int added = 0;

    for (int i = 0; i < instances; i++)
        if (pruned[i]) {
            (*subset)[i] = 0;

            if ((*n_labels)[i] == 1)
                continue;

            int count = -1;
            scalar value;

            // Recover subsets?
            // Number of labels must be bounded to avoid exponential growth of subsets
            if ((*n_labels)[i] > upper_bound) {
                count = pow(2, upper_bound)-2;
                AllocateIntegerMatrix(&combs, count, upper_bound-1);
                MALLOC(n_combs, int, count, "PruneSubsets: n_combs");
                MALLOC(info, t_ps_info, count, "PruneSubsets: info");

                // Random subset of all labels of instance
                int l_labels[(*n_labels)[i]];
                CopyIntegerVector((*labels)[i], l_labels, (*n_labels)[i]);
                RandomizeVector(l_labels, (*n_labels)[i]);
                GetSubCombinations(l_labels, upper_bound, combs, n_combs);
            } else {
                count = pow(2, (*n_labels)[i])-2;
                AllocateIntegerMatrix(&combs, count, (*n_labels)[i]-1);
                MALLOC(n_combs, int, count, "PruneSubsets: n_combs");
                MALLOC(info, t_ps_info, count, "PruneSubsets: info");

                // Get all subsets of labels
                GetSubCombinations((*labels)[i], (*n_labels)[i], combs, n_combs);
            }

            int above = 0;
            for (int j = 0; j < count; j++) {
                value = LabelsToPowerset(combs[j], n_combs[j]);
                int freq = 0;

                for (int l = 0; l < distinct; l++)
                    if (value == pss[l]) {
                        freq = c[l];
                        break;
                    }

                if (freq > p) {
                    info[above].labels = combs[j];
                    info[above].l = n_combs[j];
                    info[above].c = freq;
                    above++;
                }
            }

            if (above) {
                // Sort subsets
                qsort(info, above, sizeof(t_ps_info), CompareInfos);

                // Add instances depending on method
                if (ppt_n) {
                    int next = 0;
                    while (above > next) {
                        // Add maximum subset
                        added++;
                        AddInstance(data, instances+added, labels, n_labels, index, info[next].labels, info[next].l, inputs, i, subset);

                        // Remove all subsets with non null intersection
                        for (int j = next+1; j < above; j++)
                            if (IntersectionCardinality(info[next].labels, info[j].labels, info[next].l, info[j].l)) {
                                // Remove
                                for (int l = j+1; l < above; l++)
                                    info[l-1] = info[l];

                                above--;
                                j--;
                            }
                        next++;
                    }
                } else if (strategy == A) {
                    for (int j = 0; j < b && j < above; j++) {
                            // Add
                            added++;
                            AddInstance(data, instances+added, labels, n_labels, index, info[j].labels, info[j].l, inputs, i, subset);
                        }
                } else {
                    for (int j = 0; j < above; j++)
                        if (info[j].l > b) {
                            // Add
                            added++;
                            AddInstance(data, instances+added, labels, n_labels, index, info[j].labels, info[j].l, inputs, i, subset);
                        }
                }
            }

            FreeIntegerMatrix(&combs, count);
            FREE(n_combs);
            FREE(info);
        }

    FREE(pss);
    FREE(c);
    FREE(pruned);

    return added;
}

scalar LabelsToPowerset(int *labels, int n_labels)
{
    scalar ps = 0.0;
    for (int j = 0; j < n_labels; j++)
        ps += (int) pow(2.0, labels[j]);

    return ps;
}

void GetSubCombinations(int *labels, int n_labels, int **cs, int *c_n_labels)
{
    double count = pow(2, n_labels);

    for (int i = 1; i < count-1; i++) {
        c_n_labels[i-1] = 0;
        for (int j = 0; j < n_labels; j++) {
            int b = i & (1 << j);
            if (b > 0) {
                cs[i-1][c_n_labels[i-1]++] = labels[j];
            }
        }
    }
}

// Auxiliary function for qsort()
// Descending order

int CompareInfos(const void *x, const void *y)
{
    if ((((t_ps_info *) x)->l < ((t_ps_info *) y)->l) || ((((t_ps_info *) x)->l == ((t_ps_info *) y)->l) && (((t_ps_info *) x)->c < ((t_ps_info *) y)->c)))
        return +1;
    else {
        if ((((t_ps_info *) x)->l > ((t_ps_info *) y)->l) || ((((t_ps_info *) x)->l == ((t_ps_info *) y)->l) && (((t_ps_info *) x)->c > ((t_ps_info *) y)->c)))
            return -1;
        else
            return 0;
    }
}

void AddInstance(scalar ***data, int instances, int ***labels, int **n_labels, int ***index, int *new_labels, int new_n_labels, int inputs, int pos, int **subset)
{
    REALLOC(*data, scalar *, instances, "AddInstance: *data");
    REALLOC(*labels, int *, instances, "AddInstance: *labels");
    REALLOC(*n_labels, int, instances, "AddInstance: *n_labels");
    if (*index)
        REALLOC(*index, int *, instances, "AddInstance: *index");
    if (*subset)
        REALLOC(*subset, int, instances, "AddInstance: *subset");

    MALLOC((*data)[instances-1], scalar, inputs, "AddInstance: (*data)[instances]");
    memcpy((*data)[instances-1], (*data)[pos], sizeof(scalar)*inputs);

    if (*index) {
        MALLOC((*index)[instances-1], int, inputs+1, "AddInstance: (*data)[instances]");
        memcpy((*index)[instances-1], (*index)[pos], sizeof(int)*(inputs+1));
    }

    if (*subset)
        (*subset)[instances-1] = 1;

    MALLOC((*labels)[instances-1], int, new_n_labels, "AddInstance: (*labels)[instances]");
    memcpy((*labels)[instances-1], new_labels, sizeof(int)*new_n_labels);

    (*n_labels)[instances-1] = new_n_labels;
}

void CopyMultiLabelDataset(scalar **data, int **labels, int *n_labels, int **index, scalar **c_data, int **c_labels, int *c_n_labels, int **c_index, int instances, int inputs)
{
    memcpy(c_n_labels, n_labels, sizeof(int)*instances);

    if (index) {
        for (int i = 0; i < instances; i++) {
            memcpy(c_labels[i], labels[i], sizeof(int)*n_labels[i]);
            int j = 0;
            while (index[i][j] != -1) {
                c_index[i][j] = index[i][j];
                c_data[i][j] = data[i][j];
                j++;
            }
            c_index[i][j] = -1;
        }
    } else
        for (int i = 0; i < instances; i++) {
            memcpy(c_data[i], data[i], sizeof(scalar)*inputs);
            memcpy(c_labels[i], labels[i], sizeof(int)*n_labels[i]);
        }
}
