/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <stdlib.h>
#include "common.h"
#include "macros.h"
#include "ml.h"

void PrintClassificationMetrics(t_Performance *perf, int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances, int classes, bool sorted,
                                FILE *out_file, char *text, bool print)
{
    // Sort all labels?
    if (!sorted)
        for (int i = 0; i < instances; i++)
            qsort(predicted_labels[i], predicted_n_labels[i], sizeof(int), CompareInt);
    else // Test actually sorted
        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < predicted_n_labels[i]-1; j++)
                if (predicted_labels[i][j] >= predicted_labels[i][j+1])
                    OnErrorExit(ERR_UNSPECTED_RESULT, "Predicted labels unsorted");

            for (int j = 0; j < actual_n_labels[i]-1; j++)
                if (actual_labels[i][j] >= actual_labels[i][j+1])
                    OnErrorExit(ERR_UNSPECTED_RESULT, "Actual labels unsorted");

        }

    // EXAMPLE BASED METRICS
    perf->eb_subset_accuracy = SubsetAccuracy(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances);
    perf->eb_hamming_loss = HammingLoss(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances, classes);
    perf->eb_jaccard = JaccardCoefficient(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances);

    ExampleBased(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances, &(perf->eb_precision), &(perf->eb_recall), &(perf->eb_accuracy), &(perf->eb_f1));
    perf->eb_f1_alt = F1Instance(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances);

    if (print) {
        fprintf(out_file, "CLASSIFICATION PERFORMANCE METRICS\n");
        fprintf(out_file, "@%s Subset Accuracy (+): %lf\n", text, perf->eb_subset_accuracy);
        fprintf(out_file, "@%s Hamming Loss (-): %lf\n", text, perf->eb_hamming_loss);
        fprintf(out_file, "@%s Jaccard coefficient (+): %lf\n", text, perf->eb_jaccard);
        fprintf(out_file, "@%s Accuracy (example based, +): %lf\n", text, perf->eb_accuracy);
        fprintf(out_file, "@%s Precision (example based, +): %lf\n", text, perf->eb_precision);
        fprintf(out_file, "@%s Recall (example based, +): %lf\n", text, perf->eb_recall);
        fprintf(out_file, "@%s F1 (example based, +): %lf\n", text, perf->eb_f1);
        fprintf(out_file, "@%s F1 alternative (example based, +): %lf\n", text, perf->eb_f1_alt);
    }

    // LABEL BASED METRICS
    scalar *TP, *FP, *TN, *FN;
    MALLOC(TP, scalar, classes, "PrintClassificationMetrics: TP");
    MALLOC(FP, scalar, classes, "PrintClassificationMetrics: FP");
    MALLOC(TN, scalar, classes, "PrintClassificationMetrics: TN");
    MALLOC(FN, scalar, classes, "PrintClassificationMetrics: FN");

    BinaryPerformancePerLabel(TP, FP, TN, FN, actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances, classes);

    // MACRO AVERAGING
    perf->lb_accuracy_macro = AccuracyMacro(TP, TN, classes, instances);
    perf->lb_precision_macro = PrecisionMacro(TP, FP, classes);
    perf->lb_recall_macro = RecallMacro(TP, FN, classes);
    perf->lb_f1_macro = F1Macro(TP, FN, FP, classes);

    // MICRO AVERAGING
    perf->lb_accuracy_micro = AccuracyMicro(TP, TN, FP, FN, instances, classes);
    perf->lb_precision_micro = PrecisionMicro(TP, FP, classes);
    perf->lb_recall_micro = RecallMicro(TP, FN, classes);
    perf->lb_f1_micro = F1Micro(TP, FN, FP, classes);

    if (print) {
        fprintf(out_file, "@%s Accuracy macro average (label based, +): %lf\n", text, perf->lb_accuracy_macro);
        fprintf(out_file, "@%s Precision macro average (label based, +): %lf\n", text, perf->lb_precision_macro);
        fprintf(out_file, "@%s Recall macro average (label based, +): %lf\n", text, perf->lb_recall_macro);
        fprintf(out_file, "@%s F1 macro average (label based, +): %lf\n", text, perf->lb_f1_macro);
        fprintf(out_file, "@%s Accuracy micro average (label based, +): %lf\n", text, perf->lb_accuracy_micro);
        fprintf(out_file, "@%s Precision micro average (label based, +): %lf\n", text, perf->lb_precision_micro);
        fprintf(out_file, "@%s Recall micro average (label based, +): %lf\n", text, perf->lb_recall_micro);
        fprintf(out_file, "@%s F1 micro average (label based, +): %lf\n", text, perf->lb_f1_micro);
    }

    FREE(FP);
    FREE(TN);
    FREE(TP);
    FREE(FN);
}

scalar HammingDistanceListOfIntegers(int *x, int *y, int x_length, int y_length)
{
    scalar hamming = 0;
    int x_index = 0, y_index = 0;

    if (!x_length || !y_length)
        hamming = x_length+y_length;
    else {
        do {
            if (x[x_index] == y[y_index]) {
                x_index++;
                y_index++;
            } else {
                hamming++;
                if (x[x_index] < y[y_index])
                    x_index++;
                else
                    y_index++;
            }
        } while (x_index < x_length && y_index < y_length);

        hamming += x_length-x_index + y_length-y_index;
    }

    return hamming;
}

scalar IntersectionCardinality(int *x, int *y, int x_length, int y_length)
{
    scalar int_cardinality = 0.0;
    int x_index = 0, y_index = 0;

    if (!x_length || !y_length)
        return 0.0;
    do {
        if (x[x_index] == y[y_index]) {
            x_index++;
            y_index++;
            int_cardinality++;
        } else {
            if (x[x_index] < y[y_index])
                x_index++;
            else
                y_index++;
        }
    } while (x_index < x_length && y_index < y_length);

    return int_cardinality;
}

scalar UnionCardinality(int *x, int *y, int x_length, int y_length)
{
    scalar union_cardinality = 0;
    int x_index = 0, y_index = 0;

    if (!x_length || !y_length)
        return x_length+y_length;
    do {
        union_cardinality++;
        if (x[x_index] == y[y_index]) {
            x_index++;
            y_index++;
        } else {
            if (x[x_index] < y[y_index])
                x_index++;
            else
                y_index++;
        }
    } while (x_index < x_length && y_index < y_length);

    union_cardinality += x_length-x_index + y_length-y_index;

    return union_cardinality;
}

void BinaryPerformancePerLabel(scalar *TP, scalar *FP, scalar *TN, scalar *FN, int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances, int classes)
{
    bool pred_binary[classes], actual_binary[classes];

    for (int i = 0; i < classes; i++)
        TP[i] = FP[i] = TN[i] = FN[i] = 0.0;

    for (int i = 0; i < instances; i++) {
        // Convert to binary labels
        for (int j = 0; j < classes; j++)
            pred_binary[j] = actual_binary[j] = false;

        for (int j = 0; j < predicted_n_labels[i]; j++)
            pred_binary[predicted_labels[i][j]] = true;
        for (int j = 0; j < actual_n_labels[i]; j++)
            actual_binary[actual_labels[i][j]] = true;

        for (int j = 0; j < classes; j++)
            if (actual_binary[j] && pred_binary[j])
                TP[j]++;
            else if (actual_binary[j] && !pred_binary[j])
                FN[j]++;
            else if (!actual_binary[j] && pred_binary[j])
                FP[j]++;
            else
                TN[j]++;
    }
}

scalar Average(scalar *v, int n)
{
    scalar average = 0.0;

    for (int i = 0; i < n; i++)
        average += v[i];

    return average/n;
}

scalar VectorSum(scalar *v, int *w, int n)
{
    scalar sum = 0.0;

    if (v)
        for (int i = 0; i < n; i++)
            sum += v[i];
    else
        for (int i = 0; i < n; i++)
            sum += w[i];

    return sum;
}


scalar JaccardDistanceListOfIntegers(int *x, int *y, int x_length, int y_length)
{
    scalar m11 = 0.0, m01 = 0.0, m10 = 0.0, jaccard = 0.0;
    int x_index = 0, y_index = 0;

    if (x_length && y_length) { // All are not null
        do {
            if (x[x_index] == y[y_index]) {
                m11++;
                x_index++;
                y_index++;
            } else {
                if (x[x_index] < y[y_index]) {
                    m10++;
                    x_index++;
                } else {
                    m01++;
                    y_index++;
                }
            }
        } while (x_index < x_length && y_index < y_length);
    }

    m10 += x_length - x_index;
    m01 += y_length - y_index;

    if (m11+m01+m10)
        jaccard = m11/(m11+m01+m10);

    return jaccard;
}
