/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <stdlib.h>
#include "common.h"
#include "ml.h"
#include "macros.h"
#include "classifiers.h"

void PrintRankingMetrics(t_Performance *perf, int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels, FILE *out_file, char *text, bool print)
{
    scalar optimal_coverage;

    // EXAMPLE BASED METRICS
    perf->eb_one_error = OneError(relevant_labels, relevant_n_labels, f, instances, n_labels);
    perf->eb_coverage = Coverage(relevant_labels, relevant_n_labels, f, instances, n_labels, &optimal_coverage);
    perf->eb_ranking_loss = RankingLoss(relevant_labels, relevant_n_labels, f, instances, n_labels);
    perf->eb_average_precision = AveragePrecision(relevant_labels, relevant_n_labels, f, instances, n_labels);

    // LABEL BASED METRICS
    perf->lb_auc_macro = AUCMacro(relevant_labels, relevant_n_labels, f, instances, n_labels);
    perf->eb_auc = AUCInstance(relevant_labels, relevant_n_labels, f, instances, n_labels);
    perf->lb_auc_micro = AUCMicro(relevant_labels, relevant_n_labels, f, instances, n_labels);

    // auROC and auPRC macro-averaged
    scalar *sc_pos, *sc_neg;
    int *truth;

    MALLOC(sc_pos, scalar, instances, "PrintRankingMetrics: sc_pos");
    MALLOC(sc_neg, scalar, instances, "PrintRankingMetrics: sc_neg");
    MALLOC(truth, int, instances, "PrintRankingMetrics: truth");

    perf->auROC_macro = perf->auPRC_macro = 0.0;

    for (int i = 0; i < n_labels; i++) {
        int rel = 0, irrel = 0;

        for (int j = 0; j < instances; j++)
            if ((truth[j] = (int) LabelInList(i, relevant_labels[j], relevant_n_labels[j])))
                sc_pos[rel++] = f[j][i];
            else
                sc_neg[irrel++] = f[j][i];


        perf->auROC_macro += ClassifierAUROC(NULL, -1, NULL, instances, truth, NULL, -1, NULL, NULL, NULL, -1, 1, false, NULL, false, NULL, 2, NULL, NULL, sc_pos, sc_neg, NULL, false);
        perf->auPRC_macro += ClassifierAUPRC(NULL, -1, NULL, instances, truth, NULL, -1, NULL, -1, 1, false, NULL, NULL, 2, NULL, NULL, NULL, NULL, sc_pos, sc_neg, NULL, false);
    }

    FREE(sc_neg);
    FREE(sc_pos);
    FREE(truth);

    perf->auPRC_macro /= n_labels;
    perf->auROC_macro /= n_labels;

    if (print) {
        fprintf(out_file, "RANKING PERFORMANCE METRICS\n");
        fprintf(out_file, "@%s One-error (-): %lf\n", text, perf->eb_one_error);
        fprintf(out_file, "@%s Coverage (-): %lf\n", text, perf->eb_coverage);
        fprintf(out_file, "@%s Optimal coverage: %lf\n", text, optimal_coverage);
        fprintf(out_file, "@%s Relative coverage (+): %lf\n", text, perf->eb_coverage > 0.0 ? optimal_coverage/perf->eb_coverage : 0.0);
        fprintf(out_file, "@%s Ranking loss (-): %lf\n", text, perf->eb_ranking_loss);
        fprintf(out_file, "@%s Average precision (+): %lf\n", text, perf->eb_average_precision);
        fprintf(out_file, "@%s AUC macro (+): %lf\n", text, perf->lb_auc_macro);
        fprintf(out_file, "@%s AUC (example based, +): %lf\n", text, perf->eb_auc);
        fprintf(out_file, "@%s AUC micro (+): %lf\n", text, perf->lb_auc_micro);
        fprintf(out_file, "@%s auROC macro (+): %lf\n", text, perf->auROC_macro);
        fprintf(out_file, "@%s auPRC macro (+): %lf\n", text, perf->auPRC_macro);
    }
}

scalar Maximum(scalar *v, int n, int *index)
{
    scalar max = v[0];

    *index = 0;

    for (int i = 1; i < n; i++)
        if (max < v[i]) {
            max = v[i];
            *index = i;
        }

    return max;
}

bool ValueInVector(int value, int *v, int n)
{
    for (int i = 0; i < n; i++)
        if (v[i] == value)
            return true;

    return false;
}

// In case of a tie, lower index is considered
int OrderOfValue(scalar value, scalar *v, int n, int pos)
{
    int order = 0;

    for (int i = 0; i < n; i++)
        if (v[i] > value || (v[i] == value && i < pos))
            order++;

    return order+1;
}

int ComplementVector(int *v, int *bar_v, int n)
{
    int size = 0, next = 0;

    for (int i = 0; i < n; i++)
        if (i < v[next]) {
            bar_v[size] = i;
            size++;
        } else
            next++;

    return size;
}

scalar Minimum(scalar *v, int n, int *index)
{
    scalar min = v[0];

    *index = 0;

    for (int i = 1; i < n; i++)
        if (min > v[i]) {
            min = v[i];
            *index = i;
        }

    return min;
}
