/***************************************************************************
 *   Copyright (C) 2019 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <stdlib.h>
#include "common.h"
#include "macros.h"
#include "ml.h"

// Subset accuracy: Fraction of instances with all labels predicted accurately

scalar SubsetAccuracy(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances)
{
    scalar subset_accuracy = 0.0;
    for (int i = 0; i < instances; i++) {
        bool all_ok = (predicted_n_labels[i] == actual_n_labels[i]);
        for (int j = 0; j < predicted_n_labels[i] && all_ok; j++)
            if (predicted_labels[i][j] != actual_labels[i][j])
                all_ok = false;

        subset_accuracy += all_ok ? 1.0 : 0.0;
    }

    subset_accuracy /= instances;

    return subset_accuracy;
}

// Hamming loss: Fraction of miss-classified instance labels pairs

scalar HammingLoss(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances, int classes)
{
    scalar hamming_loss = 0.0;
    for (int i = 0; i < instances; i++)
        hamming_loss += HammingDistanceListOfIntegers(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i])/classes;

    hamming_loss /= instances;

    return hamming_loss;
}

// Accuracy (example based): Intersection of labels/Union of labels

scalar ExampleBasedAccuracy(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances)
{
    scalar eb_accuracy = 0.0;
    for (int i = 0; i < instances; i++) {
        scalar intersection_card = IntersectionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);
        scalar union_card = UnionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);
        eb_accuracy += union_card > ZERO ? intersection_card/union_card : 0.0;

    }

    eb_accuracy /= instances;

    return eb_accuracy;
}

// Precision (example based): Intersection of labels/number of predicted labels

scalar ExampleBasedPrecision(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances)
{
    scalar eb_precision = 0.0;
    for (int i = 0; i < instances; i++) {
        scalar intersection_card = IntersectionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);
        eb_precision += predicted_n_labels[i] ? intersection_card/predicted_n_labels[i] : 0.0;
    }

    eb_precision /= instances;

    return eb_precision;
}


// Recall (example based): Intersection of labels/number of actual labels

scalar ExampleBasedRecall(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances)
{
    scalar eb_recall = 0.0;
    for (int i = 0; i < instances; i++) {
        scalar intersection_card = IntersectionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);
        eb_recall += actual_n_labels[i] ? intersection_card/actual_n_labels[i] : 0.0;
    }

    eb_recall /= instances;

    return eb_recall;
}

// All four example based metrics
void ExampleBased(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances, scalar *precision, scalar *recall, scalar *accuracy, scalar *f1)
{
    *accuracy = *recall = *precision = 0.0;
    for (int i = 0; i < instances; i++) {
        scalar intersection_card = IntersectionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);
        scalar union_card = UnionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);

        *accuracy += union_card > ZERO ? intersection_card/union_card : 0.0;
        *recall += actual_n_labels[i] ? intersection_card/actual_n_labels[i] : 0.0;
        *precision += predicted_n_labels[i] ? intersection_card/predicted_n_labels[i] : 0.0;
    }

    *accuracy /= instances;
    *precision /= instances;
    *recall /= instances;

    *f1 = *precision+*recall > ZERO ? (2.0*(*precision)*(*recall))/(*precision+*recall) : 0.0;
}

// Precision macro
scalar PrecisionMacro(scalar *TP, scalar *FP, int classes)
{
    scalar lb_precision_macro = 0.0;
    for (int i = 0; i < classes; i++)
        lb_precision_macro += (TP[i]+FP[i] > ZERO) ? TP[i]/(TP[i]+FP[i]) : 0.0;
    lb_precision_macro /= classes;

    return lb_precision_macro;
}

// Recall macro

scalar RecallMacro(scalar *TP, scalar *FN, int classes)
{
    scalar lb_recall_macro = 0.0;
    for (int i = 0; i < classes; i++)
        lb_recall_macro += (TP[i]+FN[i] > ZERO) ? TP[i]/(TP[i]+FN[i]) : 0.0;
    lb_recall_macro /= classes;

    return lb_recall_macro;
}

// F1 macro

scalar F1Macro(scalar *TP, scalar *FN, scalar *FP, int classes)
{
    scalar lb_f1_macro = 0.0;
    for (int i = 0; i < classes; i++) {
        scalar precision = (TP[i]+FP[i] > ZERO) ? TP[i]/(TP[i]+FP[i]) : 0.0;
        scalar recall = (TP[i]+FN[i] > ZERO) ? TP[i]/(TP[i]+FN[i]) : 0.0;

        lb_f1_macro += precision+recall > ZERO ? (2.0*precision*recall)/(precision+recall) : 0.0;
    }
    lb_f1_macro /= classes;

    return lb_f1_macro;
}

// F1 Example-based alternative definition
// G. Madjarov, D. Kocev, D. Gjorgjevikj and S. Džeroski (2012) "An extensive experimental comparison of methods for multi-label learning", Pattern Recognition 45:3084 - 3104.

scalar F1Instance(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances)
{

    scalar eb_f1_alt = 0.0;
    for (int i = 0; i < instances; i++)
        eb_f1_alt += (2.0*IntersectionCardinality(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]))/(actual_n_labels[i]+predicted_n_labels[i]);

    eb_f1_alt /= instances;

    return eb_f1_alt;
}

// F (example based) with beta=1
// If precision or recall are negative they are calculated

scalar ExampleBasedF1(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances, scalar precision, scalar recall)
{
    if (precision < 0.0)
        precision = ExampleBasedPrecision(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances);

    if (recall < 0.0)
        recall = ExampleBasedRecall(actual_labels, actual_n_labels, predicted_labels, predicted_n_labels, instances);

    scalar eb_f1 = precision+recall > ZERO ? (2.0*precision*recall)/(precision+recall) : 0.0;

    return eb_f1;
}

// Accuracy micro
scalar AccuracyMicro(scalar *TP, scalar *TN, scalar *FP, scalar *FN, int instances, int classes)
{
    scalar tp = VectorSum(TP, NULL, classes);
    scalar tn = VectorSum(TN, NULL, classes);
    scalar fp = VectorSum(FP, NULL, classes);
    scalar fn = VectorSum(FN, NULL, classes);

    scalar lb_accuracy_micro = (tp+tn)/(tp+tn+fp+fn);

    return lb_accuracy_micro;
}

// Accuracy macro

scalar AccuracyMacro(scalar *TP, scalar *TN, int classes, int instances)
{
    // Accuracy macro
    scalar lb_accuracy_macro = 0.0;
    for (int i = 0; i < classes; i++)
        lb_accuracy_macro += (TP[i]+TN[i])/instances;
    lb_accuracy_macro /= classes;

    return lb_accuracy_macro;
}

// Precision micro

scalar PrecisionMicro(scalar *TP, scalar *FP, int classes)
{
    scalar tp = VectorSum(TP, NULL, classes);
    scalar fp = VectorSum(FP, NULL, classes);

    // Precision micro
    scalar lb_precision_micro = (tp+fp > ZERO) ? tp/(tp+fp) : 0.0;

    return lb_precision_micro;
}

// Recall micro

scalar RecallMicro(scalar *TP, scalar *FN, int classes)
{
    scalar tp = VectorSum(TP, NULL, classes);
    scalar fn = VectorSum(FN, NULL, classes);

    scalar lb_recall_micro = (tp+fn > ZERO) ? tp/(tp+fn) : 0.0;

    return lb_recall_micro;
}

// F1 micro

scalar F1Micro(scalar *TP, scalar *FN, scalar *FP, int classes)
{
    scalar tp = VectorSum(TP, NULL, classes);
    scalar fp = VectorSum(FP, NULL, classes);
    scalar fn = VectorSum(FN, NULL, classes);

    scalar precision = (tp+fp > ZERO) ? tp/(tp+fp) : 0.0;
    scalar recall = (tp+fn > ZERO) ? tp/(tp+fn) : 0.0;

    scalar lb_f1_micro = precision+recall > ZERO ? (2.0*precision*recall)/(precision+recall) : 0.0;

    return lb_f1_micro;
}

// One-error: Fraction of instances whose top ranking label is not in the relevant label set

scalar OneError(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
{
    int top;
    scalar eb_one_error = 0.0;

    for (int i = 0; i < instances; i++) {
        Maximum(f[i], n_labels, &top);
        if (!ValueInVector(top, relevant_labels[i], relevant_n_labels[i]))
            eb_one_error++;
    }
    eb_one_error /= instances;

    return eb_one_error;
}

// Coverage: Steps needed to cover all relevant labels

scalar Coverage(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels, scalar *optimal_coverage)
{
    scalar eb_coverage = 0.0, with_relevant = 0.0;
    *optimal_coverage = 0.0;

    for (int i = 0; i < instances; i++)
        if (relevant_n_labels[i]) {
            int lowest = OrderOfValue(f[i][relevant_labels[i][0]], f[i], n_labels, relevant_labels[i][0]);

            *optimal_coverage += relevant_n_labels[i];
            with_relevant++;

            for (int j = 1; j < relevant_n_labels[i]; j++) {
                int order = OrderOfValue(f[i][relevant_labels[i][j]], f[i], n_labels, relevant_labels[i][j]);
                if (order > lowest)
                    lowest = order;
            }

            eb_coverage += lowest;
        }
    eb_coverage = eb_coverage/instances - 1;
    *optimal_coverage = *optimal_coverage/instances - 1;

    return eb_coverage;
}

// Ranking loss: Fraction of reversely ordered label pairs

scalar RankingLoss(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
{
    int irrelevant[n_labels];
    scalar eb_ranking_loss = 0.0;

    for (int i = 0; i < instances; i++) {
        int n_irrelevant;
        scalar loss;

        // Get irrelevant labels of instance
        n_irrelevant = ComplementVector(relevant_labels[i], irrelevant, relevant_n_labels[i]);

        // Compare all pairs
        loss = 0.0;
        for (int j = 0; j < relevant_n_labels[i]; j++)
            for (int k = 0; k < n_irrelevant; k++)
                if (f[i][relevant_labels[i][j]] <= f[i][irrelevant[k]])
                    loss++;

        eb_ranking_loss += relevant_n_labels[i]*n_irrelevant > ZERO ? loss/(relevant_n_labels[i]*n_irrelevant) : 0.0;
    }
    eb_ranking_loss /= instances;

    return eb_ranking_loss;
}

// // Average precision: Average fraction of relevant labels ranked higher that a particular relevant label
// scalar AveragePrecision(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
// {
//     scalar eb_average_precision = 0.0;
//     for (int i = 0; i < instances; i++) {
//         scalar average = 0.0;
//
//         for (int j = 0; j < relevant_n_labels[i]; j++) {
//             // Fraction of relevant labels higher than y
//             int c = 0;
//             scalar fraction = 0.0;
//             for (int k = 0; k < n_labels; k++)
//                 if (f[i][k] >= f[i][relevant_labels[i][j]]) {
//                     c++;
//                     if (ValueInVector(k, relevant_labels[i], relevant_n_labels[i]))
//                         fraction++;
//                 }
//             average += (c > 0) ? fraction/c : 0.0;
//         }
//
//         eb_average_precision += (relevant_n_labels && relevant_n_labels[i] > 0)  ? average/relevant_n_labels[i] : 0.0;
//     }
//     eb_average_precision /= instances;
//
//     return eb_average_precision;
// }

// Average precision: Average fraction of relevant labels ranked higher that a particular relevant label
scalar AveragePrecision(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
{
    scalar eb_average_precision = 0.0;

    for (int i = 0; i < instances; i++) {
        scalar fraction = 0.0;

        for (int j = 0; j < relevant_n_labels[i]; j++) {
            scalar count = 0.0;
            int rank = OrderOfValue(f[i][relevant_labels[i][j]], f[i], n_labels, relevant_labels[i][j]);

            // Relevant labels higher than y
            for (int k = 0; k < relevant_n_labels[i]; k++)
                if (OrderOfValue(f[i][relevant_labels[i][k]], f[i], n_labels, relevant_labels[i][k]) <= rank)
                    count++;

            fraction += count/rank;
        }

        eb_average_precision += fraction/relevant_n_labels[i];
    }

    eb_average_precision /= instances;

    return eb_average_precision;
}

// AUC Macro

scalar AUCMacro(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
{
    scalar lb_auc_macro = 0.0;
    int *z, *bar_z;
    MALLOC(z, int, instances, "AUCMacro: z");
    MALLOC(bar_z, int, instances, "AUCMacro: bar_z");

    for (int i = 0; i < n_labels; i++) {
        int with = 0, without = 0;

        // List if instances with y and without y as relevant label
        for (int j = 0; j < instances; j++)
            if (ValueInVector(i, relevant_labels[j], relevant_n_labels[j])) {
                z[with] = j;
                with++;
            } else {
                bar_z[without] = j;
                without++;
            }

        scalar auc_i = 0.0;
        for (int j = 0; j < with; j++)
            for (int k = 0; k < without; k++)
                if (f[z[j]][i] >= f[bar_z[k]][i])
                    auc_i++;

        lb_auc_macro += with*without > ZERO ? auc_i/(with*without) : 0.0;
    }
    FREE(z);
    FREE(bar_z);

    lb_auc_macro /= n_labels;

    return lb_auc_macro;
}

// AUC instance

scalar AUCInstance(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
{
    scalar eb_auc = 0.0;

    int *list;
    MALLOC(list, int, n_labels, "AUCInstance: list");

    for (int i = 0; i < instances; i++) {
        scalar sum = 0.0;
        // Irrelevant labels of i-th
        int c = 0, next = 0;
        for (int ii = 0; ii < relevant_n_labels[i];)
            if (relevant_labels[i][ii] == next) {
                ii++;
                next++;
            } else {
                list[c] = next;
                next++;
                c++;
            }

        for (; next < n_labels; next++, c++)
            list[c] = next;

        // Relevant labels of i-th
        for (int a = 0; a < relevant_n_labels[i]; a++)
            for (int b = 0; b < c; b++)
                if (f[i][relevant_labels[i][a]] >= f[i][list[b]])
                    sum++;

        eb_auc += relevant_n_labels[i]*(n_labels-relevant_n_labels[i]) > 0.0 ? sum/(relevant_n_labels[i]*(n_labels-relevant_n_labels[i])) : 0.0;
    }

    eb_auc /= instances;
    FREE(list);

    return eb_auc;
}

// AUC micro

scalar AUCMicro(int **relevant_labels, int *relevant_n_labels, scalar **f, int instances, int n_labels)
{
    int *list;
    MALLOC(list, int, n_labels, "AUCInstance: list");

    // AUC micro
    scalar lb_auc_micro = 0.0;
    scalar rel_x_irel = 0.0;

    for (int i = 0; i < instances; i++)
        rel_x_irel += relevant_n_labels[i];

    rel_x_irel = rel_x_irel * (instances*n_labels - rel_x_irel);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < instances; j++) {
            // Irrelevant labels of j-th
            int c = 0, next = 0;
            for (int ii = 0; ii < relevant_n_labels[j];)
                if (relevant_labels[j][ii] == next) {
                    ii++;
                    next++;
                } else {
                    list[c] = next;
                    next++;
                    c++;
                }

            for (; next < n_labels; next++, c++)
                list[c] = next;

            // Relevant labels of i-th
            for (int a = 0; a < relevant_n_labels[i]; a++)
                for (int b = 0; b < c; b++)
                    if (f[i][relevant_labels[i][a]] >= f[j][list[b]])
                        lb_auc_micro++;
        }
    FREE(list);

    lb_auc_micro /= rel_x_irel;

    return lb_auc_micro;
}

// Jaccard coefficient

scalar JaccardCoefficient(int **actual_labels, int *actual_n_labels, int **predicted_labels, int *predicted_n_labels, int instances)
{
    scalar jaccard_coefficient = 0.0;
    for (int i = 0; i < instances; i++)
        jaccard_coefficient += JaccardDistanceListOfIntegers(actual_labels[i], predicted_labels[i], actual_n_labels[i], predicted_n_labels[i]);

    jaccard_coefficient /= instances;

    return jaccard_coefficient;
}
