/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "imbalance.h"
#include "typex.h"
#include "maths.h"
#include "is.h"
#include "ml.h"
#include "ensemble.h"

#define DO_NOT_USE_SUBPACES // If define the implementation does nor accept subspaces

// ML-ESO: Multi-Label Evolutionary Synthetic Oversampling

// Access one-dimensional arrays as two-dimensional array
#define INDEX(k, j) ((k)*max_N+(j))

void MultiLabelEvolutionarySyntheticOversampling(const t_Parameters *parm, scalar ***data, int *instances, int ***index, int inputs, int classes, int ***labels, int *subspace, int *evaluations,
        int *generations_performed,int **n_labels, scalar *IRLbl, scalar MeanIR)
{
    scalar *fitness, *reduction, *performance, delta_rm = 0.0;
    int c_rm = 0, reinit_attempts = 0;
    bool *elite;

#ifdef DO_NOT_USE_SUBPACES
    if (subspace)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ML-ESO for subspaces");
#endif

    // Set temporarily maximum depth to 10 just in case
    int old_max_dept = parm->dt_max_depth;
    ((t_Parameters *)parm)->dt_max_depth = 10;

    int min_N = parm->evosmote_min_N, max_N = parm->evosmote_max_N, plength, wlength;

    plength = *instances * max_N;
    wlength = parm->mleso_one_w ? 1 : inputs;

    REALLOC(*labels, int *, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: *labels");
    REALLOC(*n_labels, int, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: *n_labels");
    REALLOC(*data, scalar *, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: *data");
    if (*index)
        REALLOC(*index, int *, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: *index");
    for (int i = *instances; i < *instances+plength; i++) {
        MALLOC((*data)[i], scalar, inputs, "MultiLabelEvolutionarySyntheticOversampling: (*data)[i]");
        MALLOC((*labels)[i], int, classes, "MultiLabelEvolutionarySyntheticOversampling: (*labels)[i]");
        if (*index)
            MALLOC((*index)[i], int, inputs+1, "MultiLabelEvolutionarySyntheticOversampling: (*index)[i]");
    }

    // Neighbors to consider: All, as there is no "minority" class
    int n_neighbors, **neighbors;
    n_neighbors = (parm->smote_k == 0) ? *instances-1 : MIN(parm->smote_k, *instances-1);
    MALLOC(neighbors, int *, *instances, "MultiLabelEvolutionarySyntheticOversampling: neighbors");

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < *instances; i++) {
        MALLOC(neighbors[i], int, n_neighbors, "MultiLabelEvolutionarySyntheticOversampling: neighbors[i]");
        GetkNeighborsLOOForML(parm, i, *data, *instances, *index, inputs, n_neighbors, neighbors[i], subspace);
    }

    // Create initial population
    // Individual is composed of five parts:
    //      pair: Neighbors of instance to mate
    //      weight: Weight of the combination of each mating
    //      labels: source of every label, 0/1
    //      subset: Undersampling following ratios
    //      N: number of copies of each instance
    int size = parm->population_size;
    t_Individual *population;
    MALLOC(population, t_Individual, 2*size, "MultiLabelEvolutionarySyntheticOversampling: population");

    MALLOC(fitness, scalar, 2*size, "MultiLabelEvolutionarySyntheticOversampling: fitness");
    MALLOC(performance, scalar, 2*size, "MultiLabelEvolutionarySyntheticOversampling: performance");
    MALLOC(reduction, scalar, 2*size, "MultiLabelEvolutionarySyntheticOversampling: reduction");
    MALLOC(elite, bool, 2*size, "MultiLabelEvolutionarySyntheticOversampling: elite");

    // Initial population
    for (int i = 0; i < 2*size; i++) {
        MALLOC(population[i].N, int, *instances, "MultiLabelEvolutionarySyntheticOversampling: population[i].N");
        MALLOC(population[i].subset, int, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: population[i].subset");
        MALLOC(population[i].pairs, int, plength, "MultiLabelEvolutionarySyntheticOversampling: population[i].pairs");
        AllocateIntegerMatrix(&(population[i].labels), plength, classes);
        AllocateScalarMatrix(&(population[i].w), plength, wlength);
        elite[i] = false;
    }

    InitializeMLPopulation(parm, population, *instances, n_neighbors, inputs, subspace, neighbors, *data, size, min_N, max_N, classes, *index, wlength);

    // Cache of neighbors?
    scalar **distance_cache = NULL;
    cache_t **neighbors_cache = NULL;

    SHOWPROGRESS(parm->show_progress > 0 && parm->ml_method == MLkNN && parm->use_neighbors_cache, "Cache size (maximum size): %.1lfGb (%.1lfGb)\n", NEIGHBORS_CACHE_SIZE(*instances),
                 MAXIMUM_CACHE_SIZE);
    if (parm->ml_method == MLkNN && parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(*instances) < MAXIMUM_CACHE_SIZE) {
        AllocateScalarMatrix(&distance_cache, *instances+plength, *instances+plength);
        MALLOC(neighbors_cache, cache_t *, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: neighbors_cache");
        for (int i = 0; i < *instances+plength; i++)
            MALLOC(neighbors_cache[i], cache_t, *instances+plength, "MultiLabelEvolutionarySyntheticOversampling: neighbors_cache[i]");

        for (int i = 0; i < *instances; i++) {
            distance_cache[i][i] = parm->evaluate_with_loo ? INFINITE: 0.0;
            for (int j = 0; j < i; j++)
                distance_cache[i][j] = distance_cache[j][i] =
#ifdef DO_NOT_USE_SUBPACES
                                           EuclideanDistanceSparseData((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL);
#else
                                           EuclideanDistanceSparseDataSubspace((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL, subspace);
#endif
        }
    }

    // Needed for evaluation
    t_MLLearner learner[2*size];
    t_MLEnsemble ensemble[2*size];

    // Evaluate initial population
    time_t init_time  = time(NULL);
    t_Performance perf;
    for (int k = 0; k < size; k++) {
        SHOWPROGRESS(parm->show_progress > 1, "Evaluating initial population: %*d/%d\r", WIDTH(size), k+1, size);

        // If using cache add new instances
        int added = AddMLInstances(parm, &population[k], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
        if (distance_cache) {
            for (int i = 0; i < *instances; i++)
                for (int j = *instances; j < *instances+added; j++)
                    distance_cache[i][j] =
#ifdef DO_NOT_USE_SUBPACES
                        EuclideanDistanceSparseData((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL);
#else
                        EuclideanDistanceSparseDataSubspace((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL, subspace);
#endif

            for (int i = *instances; i < *instances+added; i++)
                for (int j = 0; j < *instances+added; j++)
                    distance_cache[i][j] =
#ifdef DO_NOT_USE_SUBPACES
                        EuclideanDistanceSparseData((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL);
#else
                        EuclideanDistanceSparseDataSubspace((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL, subspace);
#endif

            // Add distances to new instances
            CacheOfNeghborsFromDistances(*instances+added, distance_cache, neighbors_cache);
        }
        fitness[k] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances+added, *labels, *n_labels, classes, inputs, *data, population[k].subset, neighbors_cache, NULL, NULL, *index,
                                               &performance[k], NULL, NULL, parm->ml_method, &(learner[k]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[k]), parm->mleso_donot_eval_copies ?
                                               *instances : *instances+added);
    }
    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
    *evaluations = size;

    // Main loop of CHC algorithm
    SHOWPROGRESS(parm->show_progress > 1, "Generations: %d/%0*d/%0*d", parm->generations, WIDTH(parm->generations), 0, WIDTH(parm->generations), parm->generations);
    int *list;
    MALLOC(list, int, size, "MultiLabelEvolutionarySyntheticOversampling: list");
    for (int i = 0; i < size; i++)
        list[i] = i;

    scalar best_fitness = 0.0;
    int best_fitness_gs = 0;

    for (*generations_performed = 0; *generations_performed < parm->generations; (*generations_performed)++) {
        // CHC next generation
        int actual_size = size, order, father, mother;

        // Check evolutionary scheme
        switch (parm->evolutionary_scheme) {
            case schemeCHCGA:
                RandomizeVector(list, size);
                for (int i = 0; i < size; i += 2) {
                    // Random selection without replacement
                    father = list[i];
                    mother = list[i+1];

                    // Two crossovers are used, one more conservative
                    if (RANDOM(0.0, 1.0) < 0.5) {
                        HuxCrossover(population[mother].pairs, population[father].pairs, population[actual_size].pairs, population[actual_size+1].pairs, plength);
                        for (int j = 0; j < plength; j++) {
                            BLXAlphaCrossoverReal(population[actual_size].w[j], population[actual_size+1].w[j], population[father].w[j], population[mother].w[j], wlength, 0.0, 1.0);
                            HuxCrossover(population[mother].labels[j], population[father].labels[j], population[actual_size].labels[j], population[actual_size+1].labels[j], classes);
                        }
                    } else
                        MLESOCrossover(&population[father], &population[mother], &population[actual_size], &population[actual_size+1], wlength, classes, plength);

                    HuxCrossover(population[mother].subset, population[father].subset, population[actual_size].subset, population[actual_size+1].subset, *instances+plength);

                    if (RANDOM(0.0, 1.0) < 0.5)
                        BLXAlphaCrossover(population[actual_size].N, population[actual_size+1].N, population[father].N, population[mother].N, *instances, min_N, max_N);
                    else
                        HuxCrossover(population[father].N, population[mother].N, population[actual_size].N, population[actual_size+1].N, *instances);

                    // Assign average parents fitness just in case some of the two are not actually evaluated
                    fitness[actual_size] = fitness[actual_size+1] = (fitness[father]+fitness[mother])/2.0;

                    // Two new individuals in the population.
                    actual_size += 2;
                }

                break;

            case schemeSSGA:
            case schemeSSMA:
            case schemeGGA:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for Evolutionary SMOTE");
        }

        // Evaluate new individuals
        for (int k = size; k < actual_size; k++) {
            SHOWPROGRESS(parm->show_progress > 1, "Evaluating new population: %*d/%d\r", WIDTH(actual_size), k+1, actual_size);
            int added = AddMLInstances(parm, &population[k], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
            if (distance_cache) {
                for (int i = 0; i < *instances; i++)
                    for (int j = *instances; j < *instances+added; j++)
                        distance_cache[i][j] =
#ifdef DO_NOT_USE_SUBPACES
                            EuclideanDistanceSparseData((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL);
#else
                            EuclideanDistanceSparseDataSubspace((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL, subspace);
#endif

                for (int i = *instances; i < *instances+added; i++)
                    for (int j = 0; j < *instances+added; j++)
                        distance_cache[i][j] =
#ifdef DO_NOT_USE_SUBPACES
                            EuclideanDistanceSparseData((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL);
#else
                            EuclideanDistanceSparseDataSubspace((*data)[i], (*data)[j], inputs, *index ? (*index)[i] : NULL, *index ? (*index)[j] : NULL, subspace);
#endif

                // Add distances to new instances
                CacheOfNeghborsFromDistances(*instances+added, distance_cache, neighbors_cache);
            }

            // Evaluate new individuals if ratio applies
            if (RANDOM(0.0, 1.0) < parm->chc_random_evaluation_ratio)
                fitness[k] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances+added, *labels, *n_labels, classes, inputs, *data, population[k].subset, neighbors_cache, NULL, NULL, *index,
                                                   &performance[k], NULL, NULL, parm->ml_method, &(learner[k]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[k]), parm->mleso_donot_eval_copies
                                                   ? *instances : *instances+added);
        }
        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
        *evaluations += actual_size - size;

        // Elitism?
        if (parm->elite > 0.0) {
            for (int i = 0; i < actual_size; i++)
                elite[i] = false;
            for (int i = 0; i < parm->elite * size; i++)
                elite[OrderedSelection(fitness, i, actual_size)] = true;
        }

        // // Random mutations
        if (parm->mutation > ZERO)
            for (int i = 0; i < actual_size; i++) {
                scalar p_fitness;
                if (elite[i])
                    continue;

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    NonUniformMutation(population[i].pairs, plength, *generations_performed, parm->generations, parm->bit_mutation, parm->symmetric_selection ? -n_neighbors : 0, n_neighbors);
                    p_fitness = fitness[i];
                    int added = AddMLInstances(parm, &population[i], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances, *labels, *n_labels, classes, inputs, *data, population[i].subset, NULL, NULL, NULL, *index,
                                                           &performance[i], NULL, NULL, parm->ml_method, &(learner[i]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[i]),
                                                           parm->mleso_donot_eval_copies ? *instances : *instances+added);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation && parm->smote_undersampling_ratio > 0) {
                    NonUniformMutation(population[i].subset, *instances+plength, *generations_performed, parm->generations, parm->bit_mutation, 0, 1);
                    p_fitness = fitness[i];
                    int added = AddMLInstances(parm, &population[i], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances, *labels, *n_labels, classes, inputs, *data, population[i].subset, NULL, NULL, NULL, *index,
                                                           &performance[i], NULL, NULL, parm->ml_method, &(learner[i]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[i]),
                                                           parm->mleso_donot_eval_copies ? *instances : *instances+added);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    for (int j = 0; j < plength; j++)
                        NonUniformMutationReal(population[i].w[j], wlength, *generations_performed, parm->generations, parm->bit_mutation, 0.0, 1.0);
                    p_fitness = fitness[i];
                    int added = AddMLInstances(parm, &population[i], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances, *labels, *n_labels, classes, inputs, *data, population[i].subset, NULL, NULL, NULL, *index,
                                                           &performance[i], NULL, NULL, parm->ml_method, &(learner[i]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[i]),
                                                           parm->mleso_donot_eval_copies ? *instances : *instances+added);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    for (int j = 0; j < plength; j++)
                        NonUniformMutation(population[i].labels[j], classes, *generations_performed, parm->generations, parm->bit_mutation, 0, 1);
                    p_fitness = fitness[i];
                    int added = AddMLInstances(parm, &population[i], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances, *labels, *n_labels, classes, inputs, *data, population[i].subset, NULL, NULL, NULL, *index,
                                                           &performance[i], NULL, NULL, parm->ml_method, &(learner[i]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[i]),
                                                           parm->mleso_donot_eval_copies ? *instances : *instances+added);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                    continue;
                }

                if (RANDOM(0.0, 1.0) < parm->mutation) {
                    NonUniformMutation(population[i].N, *instances, *generations_performed, parm->generations, 1.0, min_N, max_N);
                    p_fitness = fitness[i];
                    int added = AddMLInstances(parm, &population[i], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances, *labels, *n_labels, classes, inputs, *data, population[i].subset, NULL, NULL, NULL, *index,
                                                           &performance[i], NULL, NULL, parm->ml_method, &(learner[i]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[i]),
                                                           parm->mleso_donot_eval_copies ? *instances : *instances+added);
                    delta_rm += fitness[i] - p_fitness;
                    c_rm++;
                    elite[i] = true;    // Avoid new mutations on the same individual
                }
            }

        // Obtain new population selecting best N individuals
        switch (parm->evolutionary_scheme) {
            case schemeCHCGA:
                for (int i = 0; i < size; i++) {
                    order = OrderedSelection(fitness+i, 0, actual_size - i)+i;
                    t_Individual indi = population[order];
                    population[order] = population[i];
                    population[i] = indi;
                    scalar aux_f = fitness[order];
                    fitness[order] = fitness[i];
                    fitness[i] = aux_f;
                    aux_f = reduction[order];
                    reduction[order] = reduction[i];
                    reduction[i] = aux_f;
                    aux_f = performance[order];
                    performance[order] = performance[i];
                    performance[i] = aux_f;
                }
                break;

            case schemeSSGA:
            case schemeSSMA:
            case schemeGGA:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for ML-ESO");
        }

        SHOWPROGRESS(parm->show_progress > 0, "%*d g's best P(MEAN), F(MEAN): %6.4lf(%6.4lf), %6.4lf(%6.4lf)\n", WIDTH(parm->generations), *generations_performed+1, performance[0],
                     AverageScalarVector(performance, parm->population_size), fitness[0], AverageScalarVector(fitness, parm->population_size));

        // Finished?
        if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed)) {
            fprintf(parm->out_file, "Reached limit at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
            fprintf(parm->out_file, "@Actual number of evaluations: %d\n", *evaluations);
            break;
        }

        int best = OrderedSelection(fitness, 0, size);

        // Show mutation usefulness
        SHOWPROGRESS(parm->show_progress > 1, "Random mutation fitness increment: %7.4lf\n", c_rm ? delta_rm/c_rm: 0.0)

        // Number of generations best is current
        if (fitness[best] == best_fitness) {
            best_fitness_gs++;

            // Reinitialize worst individuals?
            if (parm->best_individual_fitness_improvement_generations && best_fitness_gs >= parm->best_individual_fitness_improvement_generations) {
                int shift = size * (1.0-parm->reinitalize_percent_addition);
                InitializeMLPopulation(parm, population+shift, *instances, n_neighbors, inputs, subspace, neighbors, *data, size-shift, min_N, max_N, classes, *index, wlength);
                SHOWPROGRESS(parm->show_progress > 0, "%d worst individuals substituted\n", size-shift);
                for (int k = shift; k < size; k++) {
                    SHOWPROGRESS(parm->show_progress > 0, "Evaluating NEW population: %*d/%d\r", WIDTH(size), k+1, size);
                    int added = AddMLInstances(parm, &population[k], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
                    fitness[k] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances+added, *labels, *n_labels, classes, inputs, *data, population[k].subset, NULL, NULL, NULL, *index,
                                                           &performance[k], NULL, NULL, parm->ml_method, &(learner[k]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[k]),
                                                           parm->mleso_donot_eval_copies ? *instances : *instances+added);
                }
                SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
                reinit_attempts++;

                if (parm->chc_max_reinit_attempts && (reinit_attempts > parm->chc_max_reinit_attempts)) {
                    fprintf(parm->out_file, "Reached limit of REINITILIZATIONS at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
                    fprintf(parm->out_file, "@Actual number of evaluations: %d\n", *evaluations);
                    break;
                }
            }
        } else {
            best_fitness = fitness[best];
            best_fitness_gs = 0;
        }
    }

    // Reevaluate if necessary
    if (parm->chc_random_evaluation_ratio < 1.0)
        for (int k = 0; k < size; k++) {
            int added = AddMLInstances(parm, &population[k], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);
            fitness[k] = EvaluateIndividualFitness(parm, NULL, NULL, -1, *instances+added, *labels, *n_labels, classes, inputs, *data, population[k].subset, neighbors_cache, NULL, NULL, *index,
                                                   &performance[k], NULL, NULL, parm->ml_method, &(learner[k]), &perf, 0, NULL, parm->mlknn_k, 0, 0, NULL, &(ensemble[k]), parm->mleso_donot_eval_copies
                                                   ? *instances : *instances+added);
        }

    // Create new dataset with best individual
    int original = *instances;
    int best = OrderedSelection(fitness, 0, size);
    int added = AddMLInstances(parm, &population[best], *data, *instances, *labels, *n_labels, inputs, neighbors, max_N, classes, *index, IRLbl, MeanIR);

    // Mutation benefits
    fprintf(parm->out_file, "Random mutation average fitness increment: %7.4lf\n", c_rm ? delta_rm/c_rm: 0.0);

    fprintf(parm->out_file, "@ML-ESO added instances: %d\n", added);
    fprintf(parm->out_file, "@ML-ESO oversampling: %lf\n", ((scalar) added)/(*instances));
    fprintf(parm->out_file, "@ML-ESO average N: %.1lf\n", AverageIntegerVector(population[best].N, *instances));

    // Remove undersampled instances
    *instances += added;
    int removed = RemoveUnselectedMLInstances(data, instances, labels, n_labels, index, population[best].subset);
    fprintf(parm->out_file, "@Number of undersampled instances: %d\n", removed);
    fprintf(parm->out_file, "@ML-ESO undersampling: %lf\n", ((scalar) removed)/(*instances-added));

    fprintf(parm->out_file, "@ML-ESO reduction: %lf\n", 1.0 - ((scalar)(*instances))/original);

    FREE(list);
    for (int i = 0; i < 2*size; i++) {
        FREE(population[i].pairs);
        FREE(population[i].subset);
        FreeIntegerMatrix(&(population[i].labels), plength);
        FreeScalarMatrix(&(population[i].w), plength);

    }
    for (int i = 0; i < original; i++)
        FREE(neighbors[i]);
    FREE(neighbors);
    FREE(population);
    FREE(fitness);
    FREE(performance);
    FREE(reduction);
    FREE(elite);
    if (distance_cache) {
        for (int i = 0; i < original+plength; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
        FreeScalarMatrix(&distance_cache, original+plength);
    }

    ((t_Parameters *)parm)->dt_max_depth = old_max_dept;
}


// Initialize population
// Initialization using distance to favor nearer neighbors
void InitializeMLPopulation(const t_Parameters *parm, t_Individual *population, int instances, int n_neighbors, int inputs, int *subspace, int **neighbors, scalar **data, int size, int min_N,
                            int max_N, int classes, int **index, int wlength)
{
    scalar *roulette = NULL;
    if (parm->evosmote_roulette_initialization)
        MALLOC(roulette, scalar, n_neighbors, "InitializeMLPopulation: roulette");

    // Initial population
    for (int i = 0; i < size; i++) {
        // Roulette initialization
        if (parm->evosmote_roulette_initialization) {
            // Prepare roulette for each minority instance
            for (int k = 0; k < instances; k++) {
                population[i].N[k] = RANDOM(min_N, max_N+1);

                scalar farthest =
#ifdef DO_NOT_USE_SUBPACES
                    EuclideanDistanceSparseData(data[k], data[neighbors[k][n_neighbors-1]], inputs, index ? index[k] : NULL, index ? index[neighbors[k][n_neighbors-1]] : NULL);
#else
                    EuclideanDistanceSparseDataSubspace(data[k], data[neighbors[k][n_neighbors-1]], inputs, NULL, NULL, subspace);
#endif
                roulette[0] = farthest -
#ifdef DO_NOT_USE_SUBPACES
                              EuclideanDistanceSparseData(data[k], data[neighbors[k][0]], inputs, index ? index[k] : NULL, index ? index[neighbors[k][0]] : NULL);
#else
                              EuclideanDistanceSparseDataSubspace(data[k], data[neighbors[k][0]], inputs, NULL, NULL, subspace);
#endif
                for (int j = 1; j < n_neighbors; j++)
                    roulette[j] = roulette[j-1] + (farthest -
#ifdef DO_NOT_USE_SUBPACES
                                                   EuclideanDistanceSparseData(data[k], data[neighbors[k][j]], inputs, index ? index[k] : NULL, index ? index[neighbors[k][j]] : NULL));
#else
                                                   EuclideanDistanceSparseDataSubspace(data[k], data[neighbors[k][j]], inputs, NULL, NULL, subspace));
#endif

                for (int j = 0; j < max_N; j++)
                    population[i].pairs[INDEX(k, j)] = !parm->symmetric_selection ? RouletteSelection(roulette, n_neighbors) : (RANDOM(0.0, 1.0) < 0.5 ?
                                                       RouletteSelection(roulette, n_neighbors) : -RouletteSelection(roulette, n_neighbors));
            }
        } else
            for (int k = 0; k < instances; k++) {
                population[i].N[k] = RANDOM(min_N, max_N+1);
                for (int j = 0; j < max_N; j++)
                    population[i].pairs[INDEX(k, j)] = RANDOM(parm->symmetric_selection ? -n_neighbors : 0, n_neighbors);
            }

        // Select initial instances depending on ratios
        for (int j = 0; j < instances+instances*max_N; j++) // Undersampling of current instances
            population[i].subset[j] = RANDOM(0.0, 1.0) < parm->ml_eso_us_ratio ? 1 : 0;
        for (int j = instances; j < instances+instances*max_N; j++) // Oversampling of new instances
            population[i].subset[j] = RANDOM(0.0, 1.0) < parm->ml_eso_os_ratio ? 1 : 0;

        for (int j = 0; j < instances; j++) {
            // Combination weights and label selection
            for (int n = 0; n < max_N; n++) {
                for (int k = 0; k < wlength; k++)
                    population[i].w[INDEX(j, n)][k] = RANDOM(0.0, 1.0);

                for (int k = 0; k < classes; k++)
                    population[i].labels[INDEX(j, n)][k] = RANDOM(0.0, 1.0) < 0.5 ? 1 : 0;
            }
        }
    }

    if (roulette)
        FREE(roulette);
}

// Add instances using an individual

int AddMLInstances(const t_Parameters *parm, t_Individual *individual, scalar **data, int instances, int **labels, int *n_labels, int inputs, int **neighbors, int max_N, int classes, int **index,
                   scalar *IRLbl, scalar MeanIR)
{
    int added = 0;

    // Add instances SMOTE way
    for (int i = 0; i < instances; i++)
        if (!parm->mleso_unsel_nocopy || individual->subset[i])  // If instance not selected do not add copies
            for (int n = 0; n < individual->N[i]; n++)
                // If negative do not add instance
                if (individual->pairs[n] >= 0) {
                    // Dense or sparse?
                    if (index)
                        CombineSparseVector(data[instances+added], index[instances+added], individual->w[INDEX(i,n)], inputs, data[i], index[i], data[neighbors[i][individual->pairs[INDEX(i,n)]]],
                                            index[neighbors[i][individual->pairs[INDEX(i,n)]]], parm->mleso_one_w);
                    else
                        for (int l = 0; l < inputs; l++)
                            data[instances+added][l] = data[i][l] + individual->w[INDEX(i,n)][parm->mleso_one_w ? 0 : l]*(data[neighbors[i][individual->pairs[INDEX(i,n)]]][l] - data[i][l]);

                    // Label propagation
                    switch (parm->label_propagation) {
                        case COPY:
                            n_labels[instances+added] = n_labels[i];
                            for (int l = 0; l < n_labels[i]; l++)
                                labels[instances+added][l] = labels[i][l];
                            break;

                        case EVOL:
                            n_labels[instances+added] = 0;
                            for (int l = 0; l < classes; l++) {
                                bool label = !individual->labels[INDEX(i,n)][l] ? LabelInList(l, labels[i], n_labels[i]) :
                                            LabelInList(l, labels[neighbors[i][individual->pairs[INDEX(i,n)]]], n_labels[neighbors[i][individual->pairs[INDEX(i,n)]]]);
                                if (label)
                                    labels[instances+added][n_labels[instances+added]++] = l;
                            }
                            break;

                        case AND:
                            n_labels[instances+added] = 0;
                            for (int l = 0; l < classes; l++)
                                if (LabelInList(l, labels[i], n_labels[i]) &&
                                            LabelInList(l, labels[neighbors[i][individual->pairs[INDEX(i,n)]]], n_labels[neighbors[i][individual->pairs[INDEX(i,n)]]]))
                                    labels[instances+added][n_labels[instances+added]++] = l;
                            break;

                        case OR:
                            n_labels[instances+added] = 0;
                            for (int l = 0; l < classes; l++)
                                if (LabelInList(l, labels[i], n_labels[i]) ||
                                            LabelInList(l, labels[neighbors[i][individual->pairs[INDEX(i,n)]]], n_labels[neighbors[i][individual->pairs[INDEX(i,n)]]]))
                                    labels[instances+added][n_labels[instances+added]++] = l;
                            break;

                        case MINORITY:
                            // Copy common and either of them in minority
                            n_labels[instances+added] = 0;
                            for (int l = 0; l < classes; l++) {
                                bool present = LabelInList(l, labels[i], n_labels[i]);
                                bool present_neighbor = LabelInList(l, labels[neighbors[i][individual->pairs[INDEX(i,n)]]], n_labels[neighbors[i][individual->pairs[INDEX(i,n)]]]);
                                if ((present && present_neighbor) || (IRLbl[l] > MeanIR && (present || present_neighbor)))
                                    labels[instances+added][n_labels[instances+added]++] = l;
                            }
                            break;

                        case MINORITYOR:
                            // Copy from both if in minority
                            n_labels[instances+added] = 0;
                            for (int l = 0; l < classes; l++) {
                                bool present = LabelInList(l, labels[i], n_labels[i]);
                                bool present_neighbor = LabelInList(l, labels[neighbors[i][individual->pairs[INDEX(i,n)]]], n_labels[neighbors[i][individual->pairs[INDEX(i,n)]]]);
                                if ((present || present_neighbor) && IRLbl[l] > MeanIR)
                                    labels[instances+added][n_labels[instances+added]++] = l;
                            }
                            break;
                    }

                    // If no labels add one randomly
                    if (!n_labels[instances+added]) {
                        labels[instances+added][0] = RANDOM(0, classes);
                        n_labels[instances+added] = 1;
                    }

                    // Instance actually added
                    added++;
                }

    return added;
}

// Remove unselected instances if undersampling was used
int RemoveUnselectedMLInstances(scalar ***data, int *instances, int ***labels, int **n_labels, int ***index, int *subset)
{
    int removed = 0;
    // Instances must be removed from last to first
    for (int i = 0; i < *instances; i++)
        // Instance must be removed
        if (!subset[i]) {
            (*instances)--;
            (*data)[i] = (*data)[*instances];
            (*labels)[i] = (*labels)[*instances];
            (*n_labels)[i] = (*n_labels)[*instances];
            if (*index)
                (*index)[i] = (*index)[*instances];
            subset[i] = subset[*instances];
            i--;
            removed++;
        }

    REALLOC(*data, scalar *, *instances, "RemoveUnselectedInstances: *data");
    REALLOC(*labels, int *, *instances, "RemoveUnselectedInstances: *labels");
    REALLOC(*n_labels, int, *instances, "RemoveUnselectedInstances: *labels");
    if (*index)
        REALLOC(*index, int *, *instances, "RemoveUnselectedInstances: *index");

    return removed;
}

void GetkNeighborsLOOForML(const t_Parameters *parm, int instance, scalar **data, int instances, int **index, int inputs, int k, int *neighbors, int *subspace)
{
    RANK *rank;

    FillVector(NULL, neighbors, k, -1.0);
    MALLOC(rank, RANK, instances, "main: rank");

    for (int j = 0; j < instances; j++) {
        rank[j].fitness =
#ifdef DO_NOT_USE_SUBPACES
            -EuclideanDistanceSparseData(data[instance], data[j], inputs, index ? index[instance] : NULL, index ? index[j] : NULL);
#else
            -EuclideanDistanceSparseDataSubspace(data[instance], data[j], inputs, index ? index[instance] : NULL, index ? index[j] : NULL, subspace);
#endif
        rank[j].being = j;
    }

    // LOO?
    if (parm->mleso_loo)
        for (int j = 0; j < instances; j++)
            if (fabs(rank[j].fitness) <= ZERO)
                rank[j].fitness = -INFINITE;

    // Sort
    qsort(rank, instances, sizeof(RANK), Compare);

    // Store neighbors in ascending order of distance
    for (int j = 0; j < k; j++)
        neighbors[j] = rank[j].being;

    FREE(rank);
}

// Conservative crossover: pairs and weights and labels are kept together

void MLESOCrossover(t_Individual *father, t_Individual *mother, t_Individual *son, t_Individual *daughter, int wlength, int classes, int plength)
{
    for (int i = 0; i < plength; i++)
        if (i%2) {
            son->pairs[i] = father->pairs[i];
            daughter->pairs[i] = mother->pairs[i];
            for (int j = 0; j < wlength; j++) {
                son->w[i][j] = father->w[i][j];
                daughter->w[i][j] = mother->w[i][j];
            }
            for (int j = 0; j < classes; j++) {
                son->labels[i][j] = father->labels[i][j];
                daughter->labels[i][j] = mother->labels[i][j];
            }
        } else {
            son->pairs[i] = mother->pairs[i];
            daughter->pairs[i] = father->pairs[i];
            for (int j = 0; j < wlength; j++) {
                son->w[i][j] = mother->w[i][j];
                daughter->w[i][j] = father->w[i][j];
            }
            for (int j = 0; j < classes; j++) {
                son->labels[i][j] = mother->labels[i][j];
                daughter->labels[i][j] = father->labels[i][j];
            }
        }
}

void CacheOfNeghborsFromDistances(int instances, scalar **distance_cache, cache_t **neighbors_cache)
{
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < instances; i++) {
        RANK *rank;

        MALLOC(rank, RANK, instances, "main: r");

        for (int j = 0; j < instances; j++) {
            rank[j].fitness = distance_cache[i][j];
            rank[j].being = j;
        }

        // Sort
        qsort(rank, instances, sizeof(RANK), Compare);

        // Store neighbors in ascending order of distance
        if (neighbors_cache)
            for (int j = 0; j < instances; j++)
                neighbors_cache[i][j] = rank[j].being;

        FREE(rank);
    }
}

// Combine sparse vector: x = a + w(b-a)

void CombineSparseVector(scalar *x, int *index_x, scalar *w, int n, scalar *a, int *index_a, scalar *b, int *index_b, bool use_one_w)
{
    ZeroVector(x, NULL, n);

    int i = 0, j = 0;

    // Common inputs
    while (index_a[i] != -1 && index_b[j] != -1) {
        if (index_a[i] == index_b[j]) {
            x[index_a[i]-1] = a[i] + w[use_one_w ? 0 : index_a[i]-1]*(b[j] - a[i]);
            i++;
            j++;
        } else {
            if (index_a[i] > index_b[j]) {
                x[index_b[j]-1] = w[use_one_w ? 0 : index_b[j]-1]*b[j];
                j++;
            } else {
                x[index_a[i]-1] = a[i]-w[use_one_w ? 0 : index_a[i]-1]*a[i];
                i++;
            }
        }
    }

    // End of one of the vectors
    while (index_a[i] != -1) {
        x[index_a[i]-1] = a[i]-w[use_one_w ? 0 : index_a[i]-1]*a[i];
        i++;
    }

    while (index_b[j] != -1) {
        x[index_b[j]-1] = w[use_one_w ? 0 : index_b[j]-1]*b[j];
        j++;
    }

    // Convert to sparse
    i = 0;
    for (int l = 0; l < n; l++)
        if (fabs(x[l]) > ZERO) {
            x[i] = x[l];
            index_x[i] = l+1;
            i++;
        }
    index_x[i] = -1;
}
