/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"

// Implement MULTI-LABEL decision trees

// For PCTs: H. Blockeel, S. Džeroski2 and J. Grbović (1999) "Simultaneous Prediction of Multiple Chemical Parameters of River Water Quality with TILDE".

void MLDecisionTreeClassify(struct mldt_node *tree, scalar *instance, scalar *f, int classes, int *index, int *h, int *n_predicted, bool *classification, bool *ranks, scalar *w)
{
    // Leaf node?
    if (tree->leaf) {
        memcpy(f, tree->probabilities, sizeof(scalar)*classes);
        *n_predicted = 0;

        scalar t = 0.5;
        if (w) {
            // Get threshold
            t = w[classes];   // Bias
            for (int i = 0; i < classes; i++)
                t += f[i]*w[i];

            t = AssignWithinBoundsScalar(0.0001, t, 0.9999);
        }

#ifdef VERBOSE_MODE
        PrintVector(f, NULL, classes, stdout);
        printf("Threshold: %lf\n", t);
#endif

        for (int i = 0; i < classes; i++)
            if (f[i] > t)
                h[(*n_predicted)++] = i;

        *classification = *ranks = true;
    } else {
        if ((index ? SparseFeature(instance, index, tree->feat) : instance[tree->feat]) <= tree->cut_point)
            MLDecisionTreeClassify(tree->left, instance, f, classes, index, h, n_predicted, classification, ranks, w);
        else
            MLDecisionTreeClassify(tree->right, instance, f, classes, index, h, n_predicted, classification, ranks, w);
    }
}

void MLDecisionTreeTrain(const t_Parameters *parm, scalar **data, int instances, int **labels, int features, t_MLDTree *tree, int *subset, int *subspace, int classes, int **index, int *n_labels,
                         bool for_rf, bool pct, int **ss, bool extra_tree)
{
    // Check compatibility
    if (subset && ss)
        OnErrorExit(ERR_PARAM, "Selection and partial selection are incompatible");

    if (pct && ss)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Partial instance selection for PCT");

    int *f_list, fs = 0;
    MALLOC(f_list, int, features, "MLDecisionTreeTrain: f_list");
    for (int i = 0; i < features; i++)
        if (!subspace || subspace[i]) {
            f_list[fs] = i;
            fs++;
        }

    int *i_list, is = 0;
    MALLOC(i_list, int, instances, "MLDecisionTreeTrain: i_list");

    MALLOC(tree->nodes, struct mldt_node, 1, "MLDecisionTreeTrain: tree");
    CALLOC(tree->nodes->probabilities, scalar, classes, "MLDecisionTreeTrain: tree->probabilities");

    // Class distribution
    int *iss = NULL;
    if (ss) {
        MALLOC(iss, int, classes, "MLDecisionTreeTrain: iss");
        ZeroVector(NULL, iss, classes);

        for (int i = 0; i < instances; i++) {
            if (!EmptyColumn(ss, classes, i))
                i_list[is++] = i;
            else
                continue;

            for (int j = 0; j < classes; j++)
                iss[j] += ss[j][i];

            for (int j = 0; j < n_labels[i]; j++)
                if (ss[labels[i][j]][i])
                    tree->nodes->probabilities[labels[i][j]]++;
        }
    } else {
        for (int i = 0; i < instances; i++)
            if (!subset || subset[i]) {
                i_list[is++] = i;

                for (int j = 0; j < n_labels[i]; j++)
                    tree->nodes->probabilities[labels[i][j]]++;
            }
    }

    // Build root node with no children and label global probabilities
    tree->nodes->leaf = true;
    tree->nodes->depth = 0;
    tree->nodes->cut_point = 0.0;
    tree->nodes->feat = -1;
    tree->nodes->left = tree->nodes->right = NULL;

    // Split measure is entropy
    tree->nodes->score = pct ? ClassLabelVariance(classes, is, labels, n_labels, i_list) : MLDTClassEntropy(classes, tree->nodes->probabilities, is, iss);

    // Get probabilities after entropy
    if (ss) {
        for (int i = 0; i < classes; i++)
            if (iss[i] > 0)
                tree->nodes->probabilities[i] /= iss[i];
    } else
        for (int i = 0; i < classes; i++)
            tree->nodes->probabilities[i] /= is;

    MLDecisionTreeBuild(parm, tree->nodes, data, labels, is, i_list, fs, f_list, classes, 0, index, n_labels, for_rf, pct, ss, extra_tree);

    // Get adaptive threshold following BP-MLL if apply
    if (parm->mldt_adaptive_threshold) {
        MALLOC(tree->w, scalar, classes+1, "MLDecisionTreeTrain: tree->w");
        MLDTGetThresholdWeights(tree, data, is, i_list, labels, n_labels, classes, index);
    } else
        tree->w = NULL;

    FREE(i_list);
    FREE(f_list);
    if (iss)
        FREE(iss);
}

typedef struct {
    int feature;
    scalar cut_point, score;
} t_BestSplitInfo;

void MLDecisionTreeBuild(const t_Parameters *parm, struct mldt_node *tree, scalar **data, int **labels, int instances, int *i_list, int features, int *f_list, int classes, int depth, int **index,
                         int *n_labels, bool for_rf, bool pct, int **ss, bool extra_tree)
{
    t_BestSplitInfo *cutpoint;
    bool improves;

    //  No minimum samples or maximum depth exceeded?
    if (instances < parm->dt_min_samples_split || (parm->dt_max_depth && depth >= parm->dt_max_depth)) {
        // Leaf node
        tree->cut_point = 0.0;
        tree->feat = -1;
        tree->leaf = true;
        tree->depth = depth;

        // Get label frequency
        MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->probabilities, ss);

        tree->left = tree->right = NULL;
        return;
    }

    MALLOC(cutpoint, t_BestSplitInfo, features, "MLDecisionTreeBuild: cutpoint");

    // If random forest get random subset of features
    int efeatures = features;
    if (for_rf) {
        efeatures = parm->rf_percent > 0 ? parm->rf_percent*features : ceil(sqrt((scalar) features));

        RandomizeVector(f_list, features);
    }

    // Evaluate every remaining feature
    for (int i = 0; i < efeatures; i++) {
        cutpoint[i].feature = f_list[i];
        MLGetBestCupPoint(data, labels, instances, i_list, f_list[i], &(cutpoint[i].cut_point), &(cutpoint[i].score), classes, index, n_labels, pct, ss, extra_tree);
    }

    // Test if best improves
    int best_index = 0;
    for (int i = 1; i < efeatures; i++)
        if (cutpoint[i].score > cutpoint[best_index].score)
            best_index = i;

    // If best improves create children and follow up, else return
    // Improve criterion depends on measure
    improves = cutpoint[best_index].score > 0.0;

    if (improves) {
        tree->cut_point = cutpoint[best_index].cut_point;
        tree->feat = cutpoint[best_index].feature;
        MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->probabilities, ss);

        tree->leaf = false;
        tree->depth = depth;
        tree->score = cutpoint[best_index].score;

        MALLOC(tree->left, struct mldt_node, 1, "MLDecisionTreeBuild: tree->left");
        MALLOC(tree->left->probabilities, scalar, classes, "MLDecisionTreeBuild: tree->left->probabilities");
        tree->left->left = tree->left->right = NULL;

        int *new_i_list, is;
        MALLOC(new_i_list, int, instances, "MLDecisionTreeBuild: new_i_list");

        // Instances for left node
        is = 0;
        if (index) {
            for (int i = 0; i < instances; i++)
                if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->feat) <= tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
        } else {
            for (int i = 0; i < instances; i++)
                if (data[i_list[i]][tree->feat] <= tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
        }

        if (is)
            MLDecisionTreeBuild(parm, tree->left, data, labels, is, new_i_list, features, f_list, classes, depth+1, index, n_labels, for_rf, pct, ss, extra_tree);
        else {    // If no instances treat as leaf node
            // Leaf node
            tree->left->cut_point = 0.0;
            tree->left->feat = -1;
            tree->left->leaf = true;
            tree->left->depth = depth;

            // Get label frequency
            MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->left->probabilities, ss);

            tree->left->left = tree->left->right = NULL;
        }

        // Instances for right node
        is = 0;
        if (index) {
            for (int i = 0; i < instances; i++) {
                if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->feat) > tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
            }
        } else {
            for (int i = 0; i < instances; i++)
                if (data[i_list[i]][tree->feat] > tree->cut_point) {
                    new_i_list[is] = i_list[i];
                    is++;
                }
        }
        MALLOC(tree->right, struct mldt_node, 1, "MLDecisionTreeBuild: tree->right");
        MALLOC(tree->right->probabilities, scalar, classes, "MLDecisionTreeBuild: tree->right->probabilities");
        tree->right->left = tree->right->right = NULL;

        if (is)
            MLDecisionTreeBuild(parm, tree->right, data, labels, is, new_i_list, features, f_list, classes, depth+1, index, n_labels, for_rf, pct, ss, extra_tree);
        else {    // If no instances treat as leaf node
            // Leaf node
            tree->right->cut_point = 0.0;
            tree->right->feat = -1;
            tree->right->leaf = true;
            tree->right->depth = depth;

            // Get label frequency
            MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->right->probabilities, ss);

            tree->right->left = tree->right->right = NULL;
        }

        FREE(new_i_list);
    } else {
        // Leaf node
        tree->cut_point = 0.0;
        tree->feat = -1;
        tree->leaf = true;
        tree->depth = depth;

        // Get label frequency
        MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->probabilities, ss);

        tree->left = tree->right = NULL;
    }

    FREE(cutpoint);
}

void MLGetLabelsFrequency(int **labels, int *n_labels, int *list, int instances, int classes, scalar *freq, int **ss)
{
    ZeroVector(freq, NULL, classes);

    if (ss) {
        int *iss;
        MALLOC(iss, int, classes, "MLDecisionTreeTrain: iss");
        ZeroVector(NULL, iss, classes);

        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < classes; j++)
                iss[j] += ss[j][list[i]];

            for (int j = 0; j < n_labels[list[i]]; j++)
                if (ss[labels[list[i]][j]][list[i]])
                    freq[labels[list[i]][j]]++;
        }

        for (int i = 0; i < classes; i++)
            if (iss[i] > 0)
                freq[i] /= iss[i];

        FREE(iss);
    } else {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < n_labels[list[i]]; j++)
                freq[labels[list[i]][j]]++;

        for (int i = 0; i < classes; i++)
            freq[i] /= instances;
    }
}

void FreeMLDecisionTree(t_MLDTree *tree)
{
    FreeMLDecisionTreeNodes(tree->nodes);
    if (tree->w)
        FREE(tree->w);
}

void FreeMLDecisionTreeNodes(struct mldt_node *tree)
{
    if (tree->left) {
        FreeMLDecisionTreeNodes(tree->left);
        FREE(tree->left);
    }
    if (tree->right) {
        FreeMLDecisionTreeNodes(tree->right);
        FREE(tree->right);
    }
    if (tree)
        FREE(tree->probabilities);
}


int MLDTCountNodes(t_MLDTree *tree)
{
    int nodes = 0;

    MLDTCountAllNodes(tree->nodes, &nodes);

    return nodes;
}

void MLDTCountAllNodes(struct mldt_node *tree, int *nodes)
{
    if (tree) {
        (*nodes)++;
        MLDTCountAllNodes(tree->left, nodes);
        MLDTCountAllNodes(tree->right, nodes);
    }
}

scalar MLDTClassEntropy(int classes, scalar *freq, int n, int *ns)
{
    scalar entropy = 0.0;

    // Use weights? freq must include the weights
    for (int i = 0; i < classes; i++)
        if (freq[i] > ZERO && freq[i] != (ns ? ns[i]: n))
            entropy += -(freq[i]/(ns ? ns[i] : n))*log2(freq[i]/(ns ? ns[i] : n)) -(1.0-freq[i]/(ns ? ns[i] : n))*log2(1.0-freq[i]/(ns ? ns[i] : n));

    if (isnan(entropy) || isinf(entropy))
        OnErrorExit(ERR_NUMERICAL, "Entropy is NaN of infinite");

    return entropy;
}

void MLDTGetThresholdWeights(t_MLDTree *tree, scalar **data, int instances, int *list, int **labels, int *n_labels, int Q, int **index)
{
    bool ranks, classification;
    int n_predicted, h[Q];

    gsl_vector *t = gsl_vector_alloc(instances);
    gsl_matrix *Phi = gsl_matrix_alloc(instances, Q+1);

    // Get Phi and t
    for (int i = 0; i < instances; i++) {
        MLDecisionTreeClassify(tree->nodes, data[list[i]], Phi->data+(i*(Q+1)), Q, index ? index[list[i]] : NULL, h, &n_predicted, &classification, &ranks, NULL);
        Phi->data[i*(Q+1)+Q] = 1.0;    // Bias

        // Get optimum t value
        t->data[i] = GetOptimumThreshold(Phi->data+i*(Q+1), Q, labels[list[i]], n_labels[list[i]]);
    }

#ifdef VERBOSE_MODE
    PrintVector(t->data, NULL, Q, stdout);
    for (int i = 0; i < instances; i++)
        PrintVector(Phi->data+(i*(Q+1)), NULL, Q+1, stdout);
#endif

    // Obtain w solving the linear equations: Phi·w' = t
    gsl_vector *w = gsl_vector_alloc(Q+1);
    gsl_matrix *cov = gsl_matrix_alloc(instances, instances);
    double chisq;

    gsl_multifit_linear_workspace *work = gsl_multifit_linear_alloc(instances, Q+1);
    gsl_multifit_linear(Phi, t, w, cov, &chisq, work);

    for (int i = 0; i < Q+1; i++)
        tree->w[i] = w->data[i];

    gsl_vector_free(w);
    gsl_vector_free(t);
    gsl_matrix_free(Phi);
    gsl_matrix_free(cov);
    gsl_multifit_linear_free(work);
}
