/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"
#include "ensemble.h"

// Implements ML-FOREST method
// From: IEEE TRANSACTIONS ON KNOWLEDGE AND DATA ENGINEERING, VOL. 28, NO. 10, OCTOBER 2016
// Qingyao Wu, Mingkui Tan, Hengjie Song, Jian Chen, and Michael K. Ng "ML-FOREST: A Multi-Label Tree Ensemble Method for Multi-Label Classification"

void MLForestClassify(const t_Parameters *parm, t_MLForest *forest, scalar *instance, scalar *f, int classes, int *index, int *h, int *n_predicted, bool *classification, bool *ranks, scalar **data,
                      int instances, int inputs, int *subspace, int **train_index)
{
    scalar F[classes];
    int n_relevant;

    ZeroVector(f, NULL, classes);

    for (int i = 0; i < forest->size; i++) {
        MLTreeClassifyInstance(parm, instance, h, F, classification, ranks, data, instances, classes, &n_relevant, inputs, index, &(forest->trees[i]), subspace, NULL, NULL, train_index);

        for (int j = 0; j < n_relevant; j++)
            f[h[j]]++;
    }

    *n_predicted = 0;
    if (parm->use_mlcut) {
        // Use MLCut method for the threshold
        NormalizeVector(f, classes);
        scalar th = MLCut(f, classes);

        for (int i = 0; i < classes; i++) {
            if (f[i] >= th)
                h[(*n_predicted)++] = i;
        }
    } else {
        // Majority voting
        for (int i = 0; i < classes; i++) {
            if (f[i] >= forest->size/2)
                h[(*n_predicted)++] = i;
        }

        NormalizeVector(f, classes);
    }

#ifdef VERBOSE_MODE
    PrintVector(f, NULL, classes, stdout);
    PrintVector(NULL, H, classes, stdout);
#endif
}

void MLForestTrain(const t_Parameters *parm, scalar **data, int instances, int **labels, int features, t_MLForest *forest, int *subset, int *subspace, int classes, int **index, int *n_labels)
{
    // Number of trees
    forest->size = parm->mlforest_ntrees;

    // Allocate trees
    MALLOC(forest->trees, t_MLTree, forest->size, "MLForestTrain: forest->trees");

    // Train trees
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < forest->size; i++) {
        int *l_n_labels, **l_labels, *l_subset, **l_index = NULL;
        scalar **l_data;

        // Memory for bagging sample
        MALLOC(l_data, scalar *, instances, "MLForestTrain: l_data");
        MALLOC(l_n_labels, int, instances, "MLForestTrain: l_n_labels");
        MALLOC(l_labels, int *, instances, "MLForestTrain: l_labels");
        MALLOC(l_subset, int, instances, "MLForestTrain: l_subset");
        if (index)
            MALLOC(l_index, int *, instances, "MLForestTrain: l_index");

        // Bagging sample
        for (int j = 0; j < instances; j++) {
            int next = RANDOM(0, instances);
            l_data[j] = data[next];
            l_labels[j] = labels[next];
            l_n_labels[j] = n_labels[next];
            l_subset[j] = subset ? subset[next] : 1;
            if (index)
                l_index[j] = index[next];
        }

        SHOWPROGRESS(parm->show_progress > 1, "Adding tree %*d/%d ", WIDTH(forest->size), i+1, forest->size);
        TrainMLTreeModels(parm, l_data, instances, l_labels, l_n_labels, classes, features, &(forest->trees[i]), subspace, parm->cv_hyperparameters, parm->mltree_model, l_subset, NULL, l_index);

        SHOWPROGRESS(parm->show_progress > 1, "with %d nodes\n", MLTreeCountNodes(&(forest->trees[i])));

        FREE(l_data);
        FREE(l_n_labels);
        FREE(l_labels);
        FREE(l_subset);
        if (index)
            FREE(l_index);
    }

    fprintf(parm->out_file, "@ML-FOREST number of nodes: %d\n", MLForestCountNodes(forest));
    fprintf(parm->out_file, "@ML-FOREST number of trees: %d\n", forest->size);
}

void FreeMLForest(const t_Parameters *parm, t_MLForest *forest, int inputs)
{
    for (int i = 0; i < forest->size; i++)
        FreeMLTree(parm, &(forest->trees[i]), inputs);
    FREE(forest->trees);
}

int MLForestCountNodes(t_MLForest *forest) {
    int count = 0;
    for (int i = 0; i < forest->size; i++)
        count += MLTreeCountNodes((t_MLTree *) &forest->trees[i]);

    return count;
}

// Largeron, C., Moulin, C., Géry, M. (2012). MCut: A Thresholding Strategy for Multi-label Classification.
// In: Hollmén, J., Klawonn, F., Tucker, A. (eds) Advances in Intelligent Data Analysis XI. IDA 2012. Lecture Notes in Computer Science, vol 7619. Springer, Berlin, Heidelberg.

scalar MLCut(scalar *f, int n_labels)
{
    scalar sorted[n_labels];

    // Threshold is maximum gap
    memcpy(sorted, f, sizeof(scalar)*n_labels);
    qsort(sorted, n_labels, sizeof(scalar), CompareDouble);

    scalar max = -1.0, th = 0.0;
    for (int i = 0; i < n_labels-1; i++)
        if (sorted[i+1] - sorted[i] > max) {
            max = sorted[i+1]-sorted[i];
            th = (sorted[i+1]+sorted[i])/2;
        }

    return th;
}
