/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "common.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "maths.h"
#include "ensemble.h"

// Multi-label k-NN [Zhang and Zhou 2007, Pattern Recognition 40:2038-2048]

void MultiLabelkNearestNeighbors(const t_Parameters *parm, scalar *query, scalar **prototype_data, int prototype_instances, int **prototype_labels, int *prototype_n_labels, int n_labels,
                                 int *predicted_labels, int *predicted_n_labels, int *S, scalar *f, int inputs, int *index, int **prototype_index, cache_t **neighbors_cache,
                                 int query_index, t_MLkNN *mlknn, int **SS, int *ks, int *part_info, int partition, int *subspace)
{
    // Classify query instance
    *predicted_n_labels = 0;
    int C[n_labels];
    ZeroVector(NULL, C, n_labels);

    if (SS && parm->ml_ins_sel_per_label_same_neighbors)  {
        if (subspace)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for partial instance selection");
        int **neighbors;
        if (ks)
            mlknn->k = MaximumOfIntegerVector(ks, n_labels);
        AllocateIntegerMatrix(&neighbors, n_labels, mlknn->k);
        GetkNeighborsForMLWithSelectionPerLabel(query, prototype_data, prototype_instances, prototype_index, inputs, index, mlknn->k, neighbors, n_labels, neighbors_cache, query_index, SS, ks,
                                                subspace);

        for (int l = 0; l < n_labels; l++)
            for (int j = 0; j < (ks ? ks[l] : mlknn->k); j++)
                if (neighbors[l][j] != -1 && SS[l][neighbors[l][j]] && LabelInList(l, prototype_labels[neighbors[l][j]], prototype_n_labels[neighbors[l][j]]))
                    C[l]++;

        FreeIntegerMatrix(&neighbors, n_labels);
    } else if (SS && !parm->ml_ins_sel_per_label_same_neighbors) {
        if (subspace)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Subspace for partial instance selection");
        if (ks)
            mlknn->k = MaximumOfIntegerVector(ks, n_labels);

        int neighbors[mlknn->k];
        GetkNeighborsForML(query, prototype_data, prototype_instances, prototype_index, inputs, index, mlknn->k, neighbors, NULL, neighbors_cache, query_index,
                           parm->multiselection && parm->multiselection_as_local_k, SS, n_labels, part_info, partition, subspace);

        // Count the number of neighbors considered so far for each label
        int *count;

        CALLOC(count, int, n_labels, "MultiLabelkNearestNeighbors: count");

        for (int j = 0; j < mlknn->k; j++)
            if (neighbors[j] != -1)
                for (int i = 0; i < prototype_n_labels[neighbors[j]]; i++) {
                    int c = (!ks || j < ks[prototype_labels[neighbors[j]][i]]) ? abs(SS[prototype_labels[neighbors[j]][i]][neighbors[j]]) : 0;
                    while (c-- && count[prototype_labels[neighbors[j]][i]]++ < (ks ? ks[prototype_labels[neighbors[j]][i]] : mlknn->k))
                        C[prototype_labels[neighbors[j]][i]]++;
                }

        FREE(count);
    } else {
        if (ks)
            mlknn->k = MaximumOfIntegerVector(ks, n_labels);

        int neighbors[mlknn->k];
        GetkNeighborsForML(query, prototype_data, prototype_instances, prototype_index, inputs, index, mlknn->k, neighbors, S, neighbors_cache, query_index, parm->multiselection &&
            parm->multiselection_as_local_k, NULL, n_labels, part_info, partition, subspace);

        for (int j = 0; j < mlknn->k; j++)
            if (neighbors[j] != -1)
                for (int i = 0; i < prototype_n_labels[neighbors[j]]; i++)
                    if (!ks || j < ks[prototype_labels[neighbors[j]][i]])
                        C[prototype_labels[neighbors[j]][i]]++;
    }

    for (int l = 0; l < n_labels; l++) {
        scalar P1 = mlknn->prior[l]*mlknn->post_1[l][C[l]];
        scalar P0 = (1.0-mlknn->prior[l])*mlknn->post_0[l][C[l]];
        // Check if label relevant
        if (P1 > P0)
            predicted_labels[(*predicted_n_labels)++] = l;

        // Get confidence
        f[l] = P1/(P1+P0);
    }

    NormalizeVector(f, n_labels);
}

// Compute posterior probabilities for using the training set
void ComputePosteriorProbabilities(const t_Parameters *parm, scalar **data, int m, int **labels, int *n_labels, int K, int Y, int *S, int inputs, int **index, scalar **posterior_0,
                                   scalar **posterior_1, scalar s, cache_t **neighbors_cache, int **SS, int *KS, int *part_info, int partition, int *subspace)
{
    // If different k's  get maximum
    if (KS)
        K = MaximumOfIntegerVector(KS, Y);

    int c[K+1], c_prime[K+1];

    if (SS && S)
        OnErrorExit(ERR_UNSPECTED_RESULT, "ML instance selection and per-label instance selection are incompatible");

    if (SS && parm->ml_ins_sel_per_label_same_neighbors) {
        if (part_info)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Instance selection per label same neighbors with LAMNL-kNN");

        // Identify all neighbors
        if (KS)
            K = MaximumOfIntegerVector(KS, Y);
        int ***neighbors;
        Allocate3DIntegerMatrix(&neighbors, m, Y, K);

        for (int i = 0; i < m; i++)
            GetkNeighborsForMLWithSelectionPerLabel(data[i], data, m, index, inputs, index ? index[i] : NULL, K, neighbors[i], Y, neighbors_cache, i, SS, KS, subspace);

        for (int l = 0; l < Y; l++) {
            ZeroVector(NULL, c, (KS ? KS[l] : K)+1);
            ZeroVector(NULL, c_prime, (KS ? KS[l] : K)+1);

            for (int i = 0; i < m; i++) {
                int delta = 0;
                for (int j = 0; j < (KS ? KS[l] : K); j++)
                    if (neighbors[i][l][j] != -1 && SS[l][neighbors[i][l][j]] && LabelInList(l, labels[neighbors[i][l][j]], n_labels[neighbors[i][l][j]]))
                        delta++;

                if (LabelInList(l, labels[i], n_labels[i]))
                    c[delta]++;
                else
                    c_prime[delta]++;
            }

            for (int j = 0; j <= (KS ? KS[l] : K); j++) {
                posterior_0[l][j] = posterior_1[l][j] = 0.0;
                for (int p = 0; p <= (KS ? KS[l] : K); p++) {
                    posterior_1[l][j] += c[p];
                    posterior_0[l][j] += c_prime[p];
                }
                posterior_1[l][j] = (s+c[j])/(s*((KS ? KS[l] : K)+1)+posterior_1[l][j]);
                posterior_0[l][j] = (s+c_prime[j])/(s*((KS ? KS[l] : K)+1)+posterior_0[l][j]);
            }
        }

        Free3DIntegerMatrix(&neighbors, m, Y);
    } else {
        // Identify all neighbors
        int **neighbors;
        AllocateIntegerMatrix(&neighbors, m, K);

        for (int i = 0; i < m; i++)
            GetkNeighborsForML(data[i], data, m, index, inputs, index ? index[i] : NULL, K, neighbors[i], S, neighbors_cache, i, parm->multiselection && parm->multiselection_as_local_k, SS, Y,
                               part_info, partition, subspace);

        for (int l = 0; l < Y; l++) {
            ZeroVector(NULL, c, (KS ? KS[l] : K)+1);
            ZeroVector(NULL, c_prime, (KS ? KS[l] : K)+1);

            int count[Y];
            for (int i = 0; i < m; i++)
                if (!S || S[i] > 0) {
                    int delta = 0;

                    // Different for partial selection
                    if (SS) {
                        count[l] = 0;
                        for (int j = 0; count[l] < (KS ? KS[l] : K) && j < (KS ? KS[l] : K); j++)
                            if (neighbors[i][j] != -1 && SS[l][neighbors[i][j]]) {
                                if (LabelInList(l, labels[neighbors[i][j]], n_labels[neighbors[i][j]])) {
                                    int c = abs(SS[l][neighbors[i][j]]);
                                    while (c-- && count[l] < (KS ? KS[l] : K)) {
                                        delta++;
                                        count[l]++;
                                    }
                                } else
                                    count[l] += abs(SS[l][neighbors[i][j]]);
                            }
                    } else
                        for (int j = 0; j < (KS ? KS[l] : K); j++)
                            if (neighbors[i][j] != -1 && LabelInList(l, labels[neighbors[i][j]], n_labels[neighbors[i][j]]))
                                delta++;

                    if (LabelInList(l, labels[i], n_labels[i]))
                        c[delta]++;
                    else
                        c_prime[delta]++;
                }

            scalar effective_k = SS ? count[l] : (KS ? KS[l] : K);

            for (int j = 0; j <= (KS ? KS[l] : K); j++) {
                posterior_0[l][j] = posterior_1[l][j] = 0.0;
                for (int p = 0; p <= (KS ? KS[l] : K); p++) {
                    posterior_1[l][j] += c[p];
                    posterior_0[l][j] += c_prime[p];
                }

                posterior_1[l][j] = (s+c[j])/(s*(effective_k+1)+posterior_1[l][j]);
                posterior_0[l][j] = (s+c_prime[j])/(s*(effective_k+1)+posterior_0[l][j]);
            }
        }

        FreeIntegerMatrix(&neighbors, m);
    }

}

void GetkNeighborsForML(scalar *query, scalar **prototype_data, int prototype_instances, int **prototype_index, int inputs, int *index, int k, int *neighbors, int *S, cache_t **neighbors_cache,
                        int instance, bool multiselection_as_local_k, int **SS, int classes, int *part_info, int partition, int *subspace)
{
    if (part_info) {
        GetkNeighborsForMLWithPartition(query, prototype_data, prototype_instances, prototype_index, inputs, index, k, neighbors, S, neighbors_cache,  instance, multiselection_as_local_k, SS,
                                        classes, part_info, partition, subspace);
        return;
    }

    for (int i = 0; i < k; i++)
        neighbors[i] = -1;

    // Cache?
    if (neighbors_cache) {
        if (!S && !SS)
            for (int i = 0; i < k && i < prototype_instances; i++)
                neighbors[i] = neighbors_cache[instance][i];
        else if (SS) {
            for (int i = 0, current = 0; i < prototype_instances && current < k; i++)
                if (!NullColumn(SS, classes, neighbors_cache[instance][i]))
                    neighbors[current++] = neighbors_cache[instance][i];
        } else {
            if (multiselection_as_local_k) {
                int count;
                for (int i = 0, current = 0, local_k = 0; i < prototype_instances && local_k < k; i++) {
                    if ((count = neighbors_cache[instance][i] != -1 ? MAX(0, S[neighbors_cache[instance][i]]) : 0)) {
                        neighbors[current++] = neighbors_cache[instance][i];
                        local_k += count;
                    }
                }
            } else {
                for (int i = 0, current = 0; i < prototype_instances && current < k; i++) {
                    int count = neighbors_cache[instance][i] != -1 ? MAX(0, S[neighbors_cache[instance][i]]) : 0;
                    while (count-- && current < k)
                        neighbors[current++] = neighbors_cache[instance][i];
                }
            }
        }

        return;
    }

    RANK *rank;
    MALLOC(rank, RANK, prototype_instances, "main: r");

    int next = 0;
    for (int j = 0; j < prototype_instances; j++)
        if ((!S && !SS) || (S && S[j] > 0) || (SS && !NullColumn(SS, classes, j))) {
            rank[next].fitness = -EuclideanDistanceSparseDataSubspace(query, prototype_data[j], inputs, index, prototype_index ? prototype_index[j] : NULL, subspace);
            rank[next].being = j;
            next++;
        }

    // Sort
    qsort(rank, next, sizeof(RANK), Compare);

    // Store neighbors in ascending order of distance
    int current = 0;
    if (multiselection_as_local_k) {
        int local_k = 0;
        for (int j = 0; local_k < k && j < next; j++) {
            int count = S ? MAX(0, S[rank[j].being]) : 1;
            neighbors[current++] = rank[j].being;
            local_k += count;
        }
    } else {
        for (int j = 0; current < k && j < next; j++) {
            int count = S ? MAX(0, S[rank[j].being]) : 1;
            while (count-- && current < k)
                neighbors[current++] = rank[j].being;
        }
    }

    // Not enough?
    for (int j = current; j < k; j++)
        neighbors[j] = -1;

    FREE(rank);
}

void ComputePriorProbabilities(const t_Parameters *parm, scalar *prior, int instances, int *S, int **SS, int *n_labels, int **labels, int classes)
{
    for (int i = 0; i < classes; i++)
        prior[i] = 0.0;

    // What type of selection
    if (SS) {
        int *actual_instances;
        CALLOC(actual_instances, int, classes, "ComputePriorProbabilities: actual_instances");

        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < n_labels[i]; j++)
                prior[labels[i][j]] += abs(SS[labels[i][j]][i]);

            for (int j = 0; j < classes; j++)
                actual_instances[j] += abs(SS[j][i]);
        }

        for (int i = 0; i < classes; i++)
            prior[i] = (parm->mlknn_s+prior[i])/(2.0*parm->mlknn_s+actual_instances[i]);

        FREE(actual_instances);
    } else if (S) {
        int actual_instances = 0;
        for (int i = 0; i < instances; i++)
            if (S[i] > 0) {
                actual_instances += S[i];
                for (int j = 0; j < n_labels[i]; j++)
                    prior[labels[i][j]] += S[i];
            }

        for (int i = 0; i < classes; i++)
            prior[i] = (parm->mlknn_s+prior[i])/(2.0*parm->mlknn_s+actual_instances);
    } else {
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < n_labels[i]; j++)
                prior[labels[i][j]]++;

        for (int i = 0; i < classes; i++)
            prior[i] = (parm->mlknn_s+prior[i])/(2.0*parm->mlknn_s+instances);
    }
}

// Select k neighbors per label
void GetkNeighborsForMLWithSelectionPerLabel(scalar *query, scalar **prototype_data, int prototype_instances, int **prototype_index, int inputs, int *index, int k, int **neighbors, int labels,
        cache_t **neighbors_cache, int instance, int **SS, int *ks, int *subspace)
{
    int count;

    for (int j = 0; j < labels; j++)
        for (int i = 0; i < (ks ? ks[j] : k); i++)
            neighbors[j][i] = -1;

    // Cache?
    if (neighbors_cache) {
        for (int j = 0; j < labels; j++)
            for (int i = 0, current = 0; i < prototype_instances && current < (ks ? ks[j] : k); i++)
                if ((count = abs(SS[j][neighbors_cache[instance][i]])))
                    while (count-- && current < (ks ? ks[j] : k))
                        neighbors[j][current++] = neighbors_cache[instance][i];

        return;
    }

    // Sort neighbors
    RANK *rank;
    MALLOC(rank, RANK, prototype_instances, "main: r");

    int next = 0;
    for (int j = 0; j < prototype_instances; j++)
        if (!NullColumn(SS, labels, j)) {
            rank[next].fitness = -EuclideanDistanceSparseDataSubspace(query, prototype_data[j], inputs, index, prototype_index ? prototype_index[j] : NULL, subspace);
            rank[next].being = j;
            next++;
        }

    // Sort
    qsort(rank, next, sizeof(RANK), Compare);

    // Store neighbors in ascending order of distance
    for (int l = 0; l < labels; l++)
        for (int j = 0, current = 0; j < next && current < (ks ? ks[l] : k); j++) {
            if ((count = abs(SS[l][rank[j].being])))
                while (count-- && current < (ks ? ks[l] : k))
                    neighbors[l][current++] = rank[j].being;
        }

    FREE(rank);
}

// Check if vector is null
bool NullColumn(int **m, int rows, int column)
{
    for (int i = 0; i < rows; i++)
        if (m[i][column])
            return false;

    return true;
}

void TrainMultiLabelkNearestNeighbors(const t_Parameters *parm, scalar **prototype_data, int prototype_instances, int **prototype_labels, int *prototype_n_labels, int n_labels, int *S,
                                      int inputs, int **prototype_index, cache_t **neighbors_cache, t_MLkNN *mlknn, int **SS, int k, bool create, int *subspace)
{
    // First create classifier
    if (create)
        CreateMLkNNClassifier(mlknn, n_labels, k);

    // Compute prior probabilities per class
    ComputePriorProbabilities(parm, mlknn->prior, prototype_instances, S, SS, prototype_n_labels, prototype_labels, n_labels);

    // Compute posterior probabilities
    ComputePosteriorProbabilities(parm, prototype_data, prototype_instances, prototype_labels, prototype_n_labels, mlknn->k, n_labels, S, inputs, prototype_index, mlknn->post_0, mlknn->post_1,
                                  parm->mlknn_s, neighbors_cache, SS, NULL, NULL, -1, subspace);
}

void FreeMLkNN(t_MLkNN *mlknn, int n_labels)
{
    FREE(mlknn->prior);
    for (int i = 0; i < n_labels; i++) {
        FREE(mlknn->post_0[i]);
        FREE(mlknn->post_1[i]);
    }
    FREE(mlknn->post_0);
    FREE(mlknn->post_1);
}

void CreateMLkNNClassifier(t_MLkNN *mlknn, int n_labels, int k)
{
    mlknn->k = k;

    MALLOC(mlknn->prior, scalar, n_labels, "CreateMLkNNClassifier: mlknn->prior");

    MALLOC(mlknn->post_0, scalar *, n_labels, "CreateMLkNNClassifier: mlknn->post_0");
    MALLOC(mlknn->post_1, scalar *, n_labels, "CreateMLkNNClassifier: mlknn->post_1");
    for (int i = 0; i < n_labels; i++) {
        MALLOC(mlknn->post_0[i], scalar, k+1, "CreateMLkNNClassifier: mlknn->post_0[i]");
        MALLOC(mlknn->post_1[i], scalar, k+1, "CreateMLkNNClassifier: mlknn->post_1[i]");
    }
}
