/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "common.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "maths.h"

void TrainMultiLabelKsNearestNeighbors(const t_Parameters *parm, scalar **train_data, int train_instances, int **train_labels, int *train_n_labels, int n_labels, int *S, int inputs,
                                       int **train_index, cache_t **neighbors_cache, t_MLksNN *mlksnn, int **SS, int *ks, bool create)
{
    // Create classifier
    if (create)
        CreateMLksNNClassifier(parm, mlksnn, n_labels, ks, train_data, train_instances, train_n_labels, train_labels, inputs, train_index);

    // Obtain k's if not given already
    if (mlksnn->ks[0] == -1)
        GetKPerLabel(parm, train_instances, train_data, train_labels, train_n_labels, n_labels, inputs, train_index, mlksnn->ks, S);

    // Compute prior probabilities per class
    ComputePriorProbabilities(parm, mlksnn->prior, train_instances, S, SS, train_n_labels, train_labels, n_labels);

    int k =  MaximumOfIntegerVector(mlksnn->ks, n_labels);
    ComputePosteriorProbabilities(parm, train_data, train_instances, train_labels, train_n_labels, k, n_labels, S, inputs, train_index, mlksnn->post_0, mlksnn->post_1, parm->mlknn_s, neighbors_cache,
                                  SS, mlksnn->ks, NULL, -1, NULL);
}

void FreeMLksNN(t_MLksNN *mlksnn, int n_labels)
{
    FREE(mlksnn->ks);
    FREE(mlksnn->prior);
    for (int i = 0; i < n_labels; i++) {
        FREE(mlksnn->post_0[i]);
        FREE(mlksnn->post_1[i]);
    }
    FREE(mlksnn->post_0);
    FREE(mlksnn->post_1);
}

void CreateMLksNNClassifier(const t_Parameters *parm, t_MLksNN *mlksnn, int classes, int *ks, scalar **data, int instances, int *n_labels, int **labels, int inputs, int **index)
{
    // If minimum k is less than zero consider an interval around the optimum global value
    if (parm->mlksnn_min_k < 0) {
        int old_k;
        old_k = parm->mlknn_k;
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);
        int cv_k = parm->mlknn_k;
        ((t_Parameters *)parm)->mlknn_k = old_k;

        int increase = MAX(0, -(cv_k+parm->mlksnn_min_k));

        ((t_Parameters *) parm)->mlksnn_min_k = MAX(1, cv_k+parm->mlksnn_min_k);
        ((t_Parameters *) parm)->mlksnn_max_k = MIN(instances, cv_k+parm->mlksnn_max_k);

        // Increase if left bound bellow 0
        ((t_Parameters *) parm)->mlksnn_max_k = MIN(instances, parm->mlksnn_max_k+increase);

        SHOWPROGRESS(true, "Interval for multiple k's adjusted to [%d, %d]\n", parm->mlksnn_min_k, parm->mlksnn_max_k);
        SAVEPROGRESS(true, "Interval for multiple k's adjusted to [%d, %d]\n", parm->mlksnn_min_k, parm->mlksnn_max_k);
    }

    MALLOC(mlksnn->ks, int, classes, "main: ks");
    for (int i = 0; i < classes; i++)
        mlksnn->ks[i] = ks ? ks[i] : -1;
    MALLOC(mlksnn->prior, scalar, classes, "CreateMLksNNClassifier: mlksnn->prior");

    // Compute posterior probabilities
    int k = parm->mlksnn_max_k;

    MALLOC(mlksnn->post_0, scalar *, classes, "CreateMLksNNClassifier: mlksnn->post_0");
    MALLOC(mlksnn->post_1, scalar *, classes, "CreateMLksNNClassifier: mlksnn->post_1");
    for (int i = 0; i < classes; i++) {
        MALLOC(mlksnn->post_0[i], scalar, k+1, "CreateMLksNNClassifier: mlksnn->post_0[i]");
        MALLOC(mlksnn->post_1[i], scalar, k+1, "CreateMLksNNClassifier: mlksnn->post_1[i]");
    }
}
