/***************************************************************************
 *   Copyright (C) 20089 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "macros.h"
#include "typex.h"
#include "classifiers.h"
#include "data.h"
#include "ensemble.h"
#include "ml.h"

void TrainMLBasicBP(const t_Parameters *parm, t_MLMLP *mlmlp, scalar **data, int instances, int classes, int inputs, int **labels, int *n_labels, int *subspace, int *subset)
{
    t_ClassModel *classifier;
    scalar **desired, **t_data;

    MALLOC(classifier, t_ClassModel, 1, "TrainMLBasicBP: classifier");

    AllocateScalarMatrix(&desired, instances, classes);

    // Training data
    MALLOC(t_data, scalar *, instances, "TrainMLBasicBP: t_data");
    int next = 0;
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            ZeroVector(desired[next], NULL, classes);
            for (int j = 0; j < n_labels[i]; j++)
                desired[next][labels[i][j]] = 1.0;
            t_data[next] = data[i];
            next++;
        }

    AllocateClassifier(parm, classifier, inputs, parm->nodes, true, instances, classes, 0);
    GenerateNewClassifier(parm, classifier, true, next, inputs, classes);
    MALLOC(classifier->subspace, scalar, inputs, "TrainMLBasicBP: classifier->subspace");
    for (int i = 0; i < inputs; i++)
        classifier->subspace[i] = (scalar) (subspace ? subspace[i] : 1);
    BP(parm, classifier, 0, parm->n_epochs, parm->epoch_size, t_data, next, desired, inputs, classes, NULL);

    FreeScalarMatrix(&desired, instances);
    FREE(t_data);

    // Copy to ML MLP
    mlmlp->n_hidden = parm->nodes;
    mlmlp->transfer_h = parm->transfer_h;
    mlmlp->transfer_o = parm->transfer_o;
    mlmlp->a = parm->a;
    mlmlp->b = parm->b;
    mlmlp->a_output = parm->a_output;
    mlmlp->b_output = parm->b_output;
    AllocateScalarMatrix(&(mlmlp->ih), parm->nodes, inputs);
    AllocateScalarMatrix(&(mlmlp->ho), classes, parm->nodes);
    CopyScalarMatrix(classifier->ih, mlmlp->ih, parm->nodes, inputs);
    CopyScalarMatrix(classifier->ho, mlmlp->ho, classes, parm->nodes);

    FreeScalarMatrix(&(classifier->ih), parm->nodes);
    FreeScalarMatrix(&(classifier->ho), classes);
    FREE(classifier->subspace);
    FREE(classifier);
}

void FreeMLBasicBP(t_MLMLP *mlmlp, int classes)
{
    FreeScalarMatrix(&(mlmlp->ih), mlmlp->n_hidden);
    FreeScalarMatrix(&(mlmlp->ho), classes);
}


void MLBasciBPClassifyInstance(t_MLMLP *mlmlp, scalar *instance, scalar *f, int *h, int *n_relevant, int inputs, int classes, int *subspace, bool *classification, bool *ranks)
{
    scalar x[mlmlp->n_hidden];
    int *l_subspace;

    if (subspace)
        l_subspace = subspace;
    else {
        MALLOC(l_subspace, int, inputs, "MLBasciBPClassifyInstance: l_subspace");
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = 1;
    }

    // Weighted output
    for (int i = 0; i < mlmlp->n_hidden-1; i++) {
        x[i] = 0.0;
        for (int j = 0; j < inputs; j++)
            x[i] += mlmlp->ih[i][j]*instance[j]*l_subspace[j];

        x[i] = TransferFunction(x[i], mlmlp->transfer_h, mlmlp->a, mlmlp->b);
    }

    // Bias input to output nodes
    x[mlmlp->n_hidden-1] = 0.0;

    // Transfer to output vector.
    for (int i = 0; i < classes; i++) {
        f[i] = 0.0;
        for (int j = 0; j < mlmlp->n_hidden; j++)
            f[i] += mlmlp->ho[i][j]*x[j];

        f[i] = TransferFunction(f[i], mlmlp->transfer_o, mlmlp->a_output, mlmlp->b_output);
    }

    // Get relevant labels
    *n_relevant = 0;
    for (int i = 0; i < classes; i++)
        if (f[i] > 0.5)
            h[(*n_relevant)++] = i;

    *classification = *ranks = true;
}
