/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"
#include "ensemble.h"

// Implement MULTI-LABEL random forests based on ML-DT

// PCTs from: D. Kocev, C. Vens, J. Struyf and S. Džeroski (2013) "Tree ensembles for predicting structured outputs", Pattern Recognition 46:817–833.

void MLRandomForestClassify(t_MLRandomForest *forest, scalar *instance, scalar *f, int classes, int *index, int *h, int *n_predicted, bool *classification, bool *ranks, bool mlrf_real_votes)
{
    scalar F[classes];
    int H[classes], n_relevant;

    ZeroVector(f, NULL, classes);
    ZeroVector(NULL, H, classes);

    for (int i = 0; i < forest->size; i++) {
        MLDecisionTreeClassify(forest->trees[i].nodes, instance, F, classes, index, h, &n_relevant, classification, ranks, forest->trees[i].w);

        for (int j = 0; j < classes; j++)
            f[j] += F[j];

        for (int j = 0; j < n_relevant; j++)
            H[h[j]]++;
    }

    // Majority voting
    *n_predicted = 0;
    if (!mlrf_real_votes) {
        // Binary votes
        for (int i = 0; i < classes; i++)
            if (H[i] >= forest->size*forest->threshold)
                h[(*n_predicted)++] = i;
    } else {
        // Actual output votes
        for (int i = 0; i < classes; i++)
            if (f[i] >= forest->size*forest->threshold)
                h[(*n_predicted)++] = i;
    }

    NormalizeVector(f, classes);

#ifdef VERBOSE_MODE
    PrintVector(f, NULL, classes, stdout);
    PrintVector(NULL, H, classes, stdout);
#endif
}

void MLRandomForestTrain(const t_Parameters *parm, scalar **data, int instances, int **labels, int features, t_MLRandomForest *forest, int *subset, int *subspace, int classes, int **index,
                         int *n_labels, bool pct, bool use_all_feats, int **ss, bool extra_tree)
{
    // Number of trees
    forest->size = parm->rf_ntree;
    forest->threshold = parm->mlrf_t;

    // Allocate trees
    MALLOC(forest->trees, t_MLDTree, forest->size, "MLRandomForestTrain: forest->trees");

    if (subset && ss)
        OnErrorExit(ERR_PARAM, "Partial and standard instance selection are incompatible");

    // Train trees
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < forest->size; i++) {
        int *l_n_labels, **l_labels, *l_subset = NULL, **l_index = NULL, **l_ss = NULL, l_instances = instances;
        scalar **l_data;

        // Memory for bagging sample
        AllocateScalarMatrix(&l_data, instances, features);
        AllocateIntegerMatrix(&l_labels, instances, classes);
        MALLOC(l_n_labels, int, instances, "MLRandomForestTrain: l_n_labels");

        if (ss)
            AllocateIntegerMatrix(&l_ss, classes, instances);
        else if (subset)
            MALLOC(l_subset, int, instances, "MLRandomForestTrain: l_subset");

        if (index)
           AllocateIntegerMatrix(&l_index, instances, features+1);

        // Bagging sample
        for (int j = 0; j < instances; j++) {
            int next = RANDOM(0, instances);

            // Data
            if (index) {
                int l;
                for (l = 0; index[next][l] != -1; l++) {
                    l_index[j][l] = index[next][l];
                    l_data[j][l] = data[next][l];
                }
                l_index[j][l] = -1;
            } else
                memcpy(l_data[j], data[next], features*sizeof(scalar));

            // Labels
            l_n_labels[j] = n_labels[next];
            for (int l = 0; l < n_labels[next]; l++)
                l_labels[j][l] = labels[next][l];

            // Selection or partial selection
            if (subset)
                l_subset[j] = subset[next];
            else if (ss)
                for (int k = 0; k < classes; k++)
                    l_ss[k][j] = ss[k][next];
        }

        // If SS another sampling process can be carried out
        if (ss && parm->partial_ci_samples) {
            int generations;
            FillMatrix(NULL, l_ss, classes, instances, 1.0);
            SHOWPROGRESS(true, "Sampling for %dth tree\n", i+1);
            ClassImbalanceMethodMultiLabelAlgorithm(parm, &l_instances, &l_data, &l_labels, &l_n_labels, classes, features, &l_index, NULL, NULL, NULL, NULL, NULL, subspace, &l_ss, &generations);
        }

        MLDecisionTreeTrain(parm, l_data, l_instances, l_labels, features, &(forest->trees[i]), l_subset, subspace, classes, l_index, l_n_labels, !use_all_feats, pct, l_ss, extra_tree);
        SHOWPROGRESS(parm->show_progress > 1, "Added tree %*d/%d with %d nodes\n", WIDTH(forest->size), i+1, forest->size, MLDTCountNodes(&(forest->trees[i])));

        FreeScalarMatrix(&l_data, l_instances);
        FreeIntegerMatrix(&l_labels, l_instances);
        FREE(l_n_labels);
        if (ss)
            FreeIntegerMatrix(&l_ss, classes);
        else if (subset)
            FREE(l_subset);

        if (index)
            FreeIntegerMatrix(&l_index, l_instances);
    }

    if (parm->ml_ensemble_threshold_by_resubs && !parm->cv_ml_hyperparameters) {
        // Cross-validate the threshold if apply
        scalar fitness = -1.0, thresholds[] = {0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9}, best = 0.0;;
        int n_thresholds = 9, *mask;
        MALLOC(mask, int, instances, "MLRandomForestTrain: mask");
        if (subset)
            CopyIntegerVector(subset, mask, instances);
        else
            FillVector(NULL, mask, instances, 1.0);

        t_MLLearner learner;
        learner.mlrf = *forest;
        for (int t = 0; t < n_thresholds; t++) {
            learner.mlrf.threshold = thresholds[t];

            // Evaluate fold
            scalar l_fitness = EvaluateFoldPerformance(parm, instances, labels, n_labels, classes, features, data, subset, NULL, index, mask, &learner, MLRF, subspace, NULL);

            if (l_fitness > fitness) {
                fitness = l_fitness;
                best = thresholds[t];
            }

            SHOWPROGRESS(parm->show_progress > 1, "Evaluated %3.1lf threshold with performance: %6.4lf\n", thresholds[t], l_fitness);
        }

        forest->threshold = best;

        SAVEPROGRESS(true, "@DT ensemble best cross-validated threshold: %3.1lf\n", best);
        FREE(mask);
    }
}

void FreeMLRandomForest(t_MLRandomForest *forest)
{
    for (int i = 0; i < forest->size; i++)
        FreeMLDecisionTree(&(forest->trees[i]));
    FREE(forest->trees);
}

int MLRFCountNodes(t_MLRandomForest *forest)
{
    int count = 0;
    for (int i = 0; i < forest->size; i++)
        count += MLDTCountNodes((t_MLDTree *) &forest->trees[i]);

    return count;
}
