/***************************************************************************
 *   Copyright (C) 2023 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "typex.h"
#include "ml.h"


// Source: F. Charte, A. J. Rivera, M J. del Jesus and F. Herrera, "Addressing imbalance in multi-label classification: Measures and random resampling algorithms",
// Neurocomputing, vol. 163, 3-16, 2015.

void MLRandomOversampling(const t_Parameters *parm, int *instances, scalar ***data, int ***labels, int **n_labels, int classes, int inputs, int ***index, scalar *IRLbl, scalar *MeanIR, scalar p,
                          int *lcount)
{
    int clone = *instances*p;

#ifdef AUTOTEST
    // Check counts
    int sum = 0;
    for (int i = 0; i < classes; i++)
        sum += lcount[i];
    for (int i = 0; i < *instances; i++)
        sum -= (*n_labels)[i];

    if (sum)
        OnErrorExit(ERR_UNSPECTED_RESULT, "Label counts do not match");
#endif

    // List of minority labels
    int *minority, n_minority = 0;
    MALLOC(minority, int, classes, "MLRandomOversampling: minority");
    for (int i = 0; i < classes; i++)
        if (IRLbl[i] > *MeanIR) {
            minority[i] = true;
            n_minority++;
        } else
            minority[i] = false;

    // No minority class? return
    if (!n_minority)
        return;

    // Bags of minority labels
    int **bags, *bag_size;
    CALLOC(bags, int *, classes, "MLRandomOversampling: bags");
    CALLOC(bag_size, int, classes, "MLRandomOversampling: bag_size");
    for (int i = 0; i < classes; i++)
        if (minority[i])
            MALLOC(bags[i], int, (int) ceil(*instances/IRLbl[i]), "MLRandomOversampling: bags[i]");    // For size, worst case scenario majority label is in all instances

    int minority_instances = 0;
    for (int i = 0; i < *instances; i++) {
        bool in_minority = false;
        for (int j = 0; j < (*n_labels)[i]; j++)
            if (minority[(*labels)[i][j]]) {
                bags[(*labels)[i][j]][bag_size[(*labels)[i][j]]] = i;
                bag_size[(*labels)[i][j]]++;

                in_minority = true;
            }

        if (in_minority)
            minority_instances++;
    }

    fprintf(parm->out_file, "@Minority samples: %d\n", minority_instances);
    fprintf(parm->out_file, "@Minority samples ratio: %lf\n", ((scalar) minority_instances)/(*instances));

    // Needed for recalculating the IRLbl
    int max_count = lcount[0];
    for (int i = 1; i < classes; i++)
        if (IRLbl[i] > max_count)
            max_count = IRLbl[i];

    // Oversample minority labels
    for (int i = 0; i < clone && n_minority; i++) {
        for (int j = 0; j < classes && n_minority; j++)
            if (minority[j]) {
                int os = bags[j][(int) RANDOM(0, bag_size[j])];

                // Clone
                REALLOC(*data, scalar *, *instances+1, "MLRandomOversampling: *data");
                REALLOC(*n_labels, int, *instances+1, "MLRandomOversampling: *n_labels");
                REALLOC(*labels, int *, *instances+1, "MLRandomOversampling: *labels");

                MALLOC((*labels)[*instances], int, classes, "MLRandomOversampling: (*labels)[*instances]");

                if (*index) {
                    int c = 0;
                    while ((*index)[os][c++] != -1)
                        ;

                    REALLOC(*index, int *, *instances+1, "MLRandomOversampling: *index");
                    MALLOC((*index)[*instances], int, c+1, "MLRandomOversampling: (*index)[*instances]");
                    MALLOC((*data)[*instances], scalar, c, "MLRandomOversampling: (*data)[*instances]");

                    CopyScalarVector((*data)[os], (*data)[*instances], c);
                    CopyIntegerVector((*index)[os], (*index)[*instances], c+1);
                } else {
                    MALLOC((*data)[*instances], scalar, inputs, "MLRandomOversampling: (*data)[*instances]");
                    CopyScalarVector((*data)[os], (*data)[*instances], inputs);
                }

                (*n_labels)[*instances] = (*n_labels)[os];
                CopyIntegerVector((*labels)[os], (*labels)[*instances], (*n_labels)[os]);

                (*instances)++;

                // Recalculate
                IRLbl[j] = ((scalar) max_count)/(++lcount[j]);

                // Check if still minority class
                if (IRLbl[j] <= *MeanIR) {
                    minority[j] = false;
                    n_minority--;
                }
            }
    }

    // Recalculate MeanIR
    *MeanIR = 0.0;
    for (int i = 0; i < classes; i++)
        *MeanIR += IRLbl[i];
    *MeanIR /= classes;

    FREE(minority);
    FREE(bag_size);
    for (int i = 0; i < classes; i++)
        if (bags[i])
            FREE(bags[i]);
    FREE(bags);
}

