/***************************************************************************
 *   Copyright (C) 2023 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include "typex.h"
#include "ml.h"


// Source: F. Charte, A. J. Rivera, M J. del Jesus and F. Herrera, "Addressing imbalance in multi-label classification: Measures and random resampling algorithms",
// Neurocomputing, vol. 163, 3-16, 2015.

#define EPOCH 25

void MLRandomUndersampling(const t_Parameters *parm, int *instances, scalar ***data, int ***labels, int **n_labels, int classes, int inputs, int ***index, scalar *IRLbl, scalar *MeanIR, scalar p)
{
    int ldiv;
    scalar CVIR, ml[classes+1], lcard, lden, pldiv;

    int to_delete = *instances*p;

    // List of minority labels
    bool *minority;
    int majority_instances;

    MALLOC(minority, bool, *instances, "MLRandomUndersampling: minority");

    majority_instances = CalculateMinorityInstances(classes, minority, *labels, *n_labels, IRLbl, *MeanIR, *instances);

    // No majority? Nothing to do
    if (!majority_instances) {
        FREE(minority);

        return;
    }

    fprintf(parm->out_file, "@Majority samples: %d\n", majority_instances);
    fprintf(parm->out_file, "@Majority samples ratio: %lf\n", ((scalar) majority_instances)/(*instances));

    // Undersample minority labels
    int to_delete_width = WIDTH(to_delete);
    for (; to_delete > 0 && majority_instances > 0; ) {
        for (int i = 0; i < EPOCH && to_delete > 0 && majority_instances > 0; i++) {
            int us = RANDOM(0, *instances);

            if (!minority[us]) {
                to_delete--;
                (*instances)--;
                majority_instances--;

                SHOWPROGRESS(parm->show_progress > 1, "Removed instance %*d (%*d remaining)\r", WIDTH(*instances), us, WIDTH(to_delete), to_delete_width);

                // Remove
                FREE((*labels)[us]);
                FREE((*data)[us]);

                (*data)[us] = (*data)[*instances];
                (*labels)[us] = (*labels)[*instances];
                (*n_labels)[us] = (*n_labels)[*instances];

                if (*index) {
                    FREE((*index)[us]);
                    (*index)[us] = (*index)[*instances];
                }
            }
        }

        // Recalculate IRLbl and MeanIR
        MLDatasetInfo(*instances, *labels, *n_labels, 0, NULL, NULL, classes, &lcard, &lden, &ldiv, &pldiv, NULL, MeanIR, IRLbl, &CVIR, NULL, inputs, ml);

        // Recalculate lists
        majority_instances = CalculateMinorityInstances(classes, minority, *labels, *n_labels, IRLbl, *MeanIR, *instances);
    }

    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');

    FREE(minority);

    REALLOC(*data, scalar *, *instances, "MLRandomOversampling: *data");
    REALLOC(*n_labels, int, *instances, "MLRandomOversampling: *n_labels");
    REALLOC(*labels, int *, *instances, "MLRandomOversampling: *labels");
    if (*index)
        REALLOC(*index, int *, *instances, "MLRandomOversampling: *index");
}

int CalculateMinorityInstances(int classes, bool *minority, int **labels, int *n_labels, scalar *IRLbl, scalar MeanIR, int instances)
{
    bool *minority_labels;

    MALLOC(minority_labels, bool, classes, "CalculateMinorityInstances: minority_labels");

    for (int i = 0; i < classes; i++)
        minority_labels[i] = (IRLbl[i] >= MeanIR);

    // Bags of minority labels
    int majority_instances = 0;
    for (int i = 0; i < instances; i++) {
        bool in_minority = false;
        for (int j = 0; j < n_labels[i]; j++)
            if (minority_labels[labels[i][j]]) {
                minority[i] = in_minority = true;

                break;
            }

        if (!in_minority) {
            minority[i] = false;
            majority_instances++;
        }
    }

    FREE(minority_labels);

    return majority_instances;
}
