/***************************************************************************
 *   Copyright (C) 2021 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"

/* Implementation of MLSMOTE
 * F. Charte, A. J. Rivera, M. J. del Jesus, F. Herrera, "MLSMOTE: Approaching imbalanced multilabel learning through synthetic instance generation",
 * Knowledge-Based Systems 89 (2015) 385–397
*/

void MultiLabelSmote(const t_Parameters *parm, int *instances, scalar ***data, int ***labels, int **n_labels, int classes, int inputs, int ***index, scalar *IRLbl, scalar *MeanIR, scalar *CVIR,
                     int *clabel, int *subspace)
{
    int *bag, added = 0, *neighbors;

    MALLOC(bag, int, *instances, "MultiLabelSmote: bag");
    MALLOC(neighbors, int, parm->mlknn_k, "MultiLabelSmote: neighbors");

    // Iterate for all classes
    cache_t **neighbors_cache = NULL;
    int neighbors_cache_size = 0;
    scalar **distance_cache = NULL;
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(*instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(*data, *instances, inputs, &distance_cache, &neighbors_cache, NULL, index ? *index : NULL, &neighbors_cache_size, NULL, parm->ml_loo);

    for (int i = 0; i < classes; i++)
        if (IRLbl[i] > *MeanIR) {
            // This label is a minority label
            // Get all instances with this label
            int bag_size = 0;
            for (int j = 0; j < *instances; j++)
                if (LabelInList(i, (*labels)[j], (*n_labels)[j])) {
                    bag[bag_size] = j;
                    bag_size++;
                }

            // For all minority samples create a new one
            for (int j = 0; j < bag_size; j++) {
                // Get k neighbors
                GetkNeighborsForML((*data)[bag[j]], *data, *instances, *index, inputs, *index ? (*index)[j] : NULL, parm->mlknn_k, neighbors, NULL, neighbors_cache, bag[j], false, NULL, classes, NULL,
                                   -1, subspace);
                int rneighbor = RANDOM(0, parm->mlknn_k);

                REALLOC(*data, scalar *, *instances+added+1, "MultiLabelSmote: *data");
                REALLOC(*labels, int *, *instances+added+1, "MultiLabelSmote: *labels");
                REALLOC(*n_labels, int, *instances+added+1, "MultiLabelSmote: *n_labels");
                if (*index)
                    REALLOC(*index, int *, *instances+added+1, "MultiLabelSmote: *index");

                // Add new instance
                AddNewInstance(*data, index ? *index : NULL, *labels, *n_labels, *instances+added, bag[j], rneighbor, neighbors, inputs, classes, parm->mlknn_k);

                added++;
            }
        }

    if (neighbors_cache) {
        for (int i = 0; i < *instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }
    if (distance_cache) {
        for (int i = 0; i < *instances; i++)
            FREE(distance_cache[i]);
        FREE(distance_cache);
    }

    // Add new instances
    *instances += added;

    // Recalculate CI metrics
    scalar lcard, lden, pldiv, ml[classes+1];
    int ldiv;
    MLDatasetInfo(*instances, *labels, *n_labels, 0, NULL, NULL, classes, &lcard, &lden, &ldiv, &pldiv, NULL, MeanIR, IRLbl, CVIR, clabel, inputs, ml);

    FREE(bag);
    FREE(neighbors);
}

void AddNewInstance(scalar **data, int **index, int **labels, int *n_labels, int new, int old, int ref, int *neighbors, int inputs, int classes, int k)
{
    int *counts;

    CALLOC(counts, int, classes, "AddNewInstance: counts");

    // Generate new instance
    CALLOC(data[new], scalar, inputs, "AddNewInstance: data[new]");
    MALLOC(labels[new], int, classes, "AddNewInstance: labels[new]");

    // Dense or sparse format
    if (index) {
        int i = 0, j = 0;
        // Common inputs
        while (index[old][i] != -1 && index[ref][j] != -1) {
            if (index[old][i] == index[ref][j]) {
                data[new][index[old][i]-1] = data[old][i] + (data[ref][j] - data[old][i])*RANDOM(0.0, 1.0);

                i++;
                j++;
            } else {
                if (index[old][i] > index[ref][j]) {
                    data[new][index[ref][j]-1] = 0.0 + (data[ref][j] - 0.0)*RANDOM(0.0, 1.0);
                    j++;
                } else {
                    data[new][index[old][i]-1] = data[old][i] + (0.0 - data[old][i])*RANDOM(0.0, 1.0);
                    i++;
                }
            }
        }

        // End of one of the vectors
        while (index[old][i] != -1) {
            data[new][index[old][i]-1] = data[old][i] + (0.0 - data[old][i])*RANDOM(0.0, 1.0);
            i++;
        }

        while (index[ref][j] != -1) {
            data[new][index[ref][j]-1] = 0.0 + (data[ref][j] - 0.0)*RANDOM(0.0, 1.0);
            j++;
        }

        ConvertToSparseFormat(data+new, index+new, inputs);
    } else
        for (int i = 0; i < inputs; i++)
            data[new][i] = data[old][i] + (data[ref][i] - data[old][i])*RANDOM(0.0, 1.0);

    // Generate new labelset
    for (int i = 0; i < n_labels[old]; i++)
        counts[labels[old][i]]++;

    for (int j = 0; j < k; j++)
        for (int i = 0; i < n_labels[neighbors[j]]; i++)
            counts[labels[neighbors[j]][i]]++;

    // Assign labels that are in a majority of neighbors
    n_labels[new] = 0;
    for (int i = 0; i < classes; i++)
        if (counts[i] > (k+1)/2) {
            labels[new][n_labels[new]] = i;
            n_labels[new]++;
        }

    // Reassign label memory
    REALLOC(labels[new], int, n_labels[new], "AddNewInstance: labels[new]");

    FREE(counts);
}
