/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"

// Implement ML decision trees split measures

typedef struct {
    int instance, *labels, n_labels;
    scalar value;
} t_MLFeatureInfo;


#define EPSILON 0.1

void MLGetBestCupPoint(scalar **data, int **labels, int instances, int *i_list, int feature, scalar *cut_point, scalar *score, int classes, int **index, int *n_labels, bool pct, int **ss,
                       bool extra_tree)
{
    t_MLFeatureInfo *info;
    int n_cuts, next, *llist = NULL, *rlist = NULL, *count;
    scalar *cuts, pentropy = 0.0, new_score = 0.0, fleft[classes], fright[classes];

    MALLOC(info, t_MLFeatureInfo, instances, "GetBestCupPoint: info");

    if (pct) {
        MALLOC(rlist, int, instances, "MLGetBestCupPoint: rlist");
        MALLOC(llist, int, instances, "MLGetBestCupPoint: llist");
    }

    // Sort values
    for (int i = 0; i < instances; i++) {
        info[i].instance = i;
        MALLOC(info[i].labels, int, MAX(1, n_labels[i_list[i]]), "MLGetBestCupPoint: info[i].labels");
        info[i].n_labels = n_labels[i_list[i]];
        CopyIntegerVector(labels[i_list[i]], info[i].labels, n_labels[i_list[i]]);
        info[i].value = index ? SparseFeature(data[i_list[i]], index[i_list[i]], feature) : data[i_list[i]][feature];
    }

    qsort(info, instances, sizeof(t_MLFeatureInfo), CompareMLFeature);

    // If constant values return 0 score
    if (info[0].value == info[instances-1].value) {
        *score = *cut_point = 0.0;
        for (int i = 0; i < instances; i++)
            FREE(info[i].labels);
        FREE(info);
        return;
    }

    // Get cut point scores
    // If extremely randomize (extra) tree only a random cut point is considered
    MALLOC(cuts, scalar, instances+1, "GetBestCupPoint: cuts");
    if (extra_tree) {
        n_cuts = 1;
        cuts[0] = RANDOM(info[0].value, info[instances-1].value);
    } else {
        cuts[0] = info[0].value - EPSILON;
        n_cuts = 1;
        for (int i = 0; i < instances-1; i++)
            if (info[i].value != info[i+1].value) {
                cuts[n_cuts] = (info[i].value+info[i+1].value)/2.0;
                n_cuts++;
            }

        cuts[n_cuts] = info[instances-1].value + EPSILON;
        n_cuts++;
    }

    // First cut point, all in the right child
    MALLOC(count, int, classes, "MLGetBestCupPoint: count");
    ZeroVector(NULL, count, classes);

    int linstances = 0;
    if (pct) {
        for (int i = 0; i < instances; i++)
            rlist[i] = info[i].instance;
        pentropy = ClassLabelVariance(classes, instances, labels, n_labels, rlist);
    } else {
        ZeroVector(fleft, NULL, classes);
        ZeroVector(fright, NULL, classes);
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < n_labels[i_list[i]]; j++)
                if (!ss || ss[labels[i_list[i]][j]][i_list[i]])
                    fright[labels[i_list[i]][j]]++;

        if (ss) {
            for (int i = 0; i < classes; i++)
                for (int j = 0; j < instances; j++)
                    if (ss[i][i_list[j]])
                        count[i]++;
        } else
            FillVector(NULL, count, classes, (scalar) instances);

        pentropy = MLDTClassEntropy(classes, fright, instances, count);
    }

    // Maximum best
    *score = -INFINITE;

    // All possible thresholds
    next = 0;
    for (int i = 0; i < n_cuts; i++) {
        // Predicting clustering trees
        if (pct) {
            // Standard decision trees
            while (next < instances && info[next].value <= cuts[i]) {
                llist[linstances] = info[next].instance;
                linstances++;
                next++;
            }

            // SPLIT MEASURE: Variance gain
            new_score = VarianceGain(classes, pentropy, llist, rlist+linstances, linstances, instances-linstances, labels, n_labels);
        } else {
            // Standard decision trees
            while (next < instances && info[next].value <= cuts[i]) {
                for (int j = 0; j < info[next].n_labels; j++)
                    if (!ss || ss[info[next].labels[j]][i_list[info[next].instance]]) {
                        fleft[info[next].labels[j]]++;
                        fright[info[next].labels[j]]--;
                    }
                linstances++;
                next++;
            }

            // SPLIT MEASURE: Information gain
            new_score = MLInformationGain(instances, classes, pentropy, fleft, fright, linstances, count);
        }

        if (new_score > *score) {
            *score = new_score;
            *cut_point = cuts[i];
        }
    }

    FREE(cuts);
    for (int i = 0; i < instances; i++)
        FREE(info[i].labels);
    FREE(info);
    if (pct) {
        FREE(llist);
        FREE(rlist);
    }
    FREE(count);
}

int CompareMLFeature(const void *a, const void *b)
{
    if (((t_MLFeatureInfo *)a)->value < ((t_MLFeatureInfo *)b)->value)
        return -1;
    else if (((t_MLFeatureInfo *)a)->value > ((t_MLFeatureInfo *)b)->value)
        return +1;
    else
        return 0;
}

scalar MLInformationGain(int instances, int classes, scalar pentropy, scalar *lfreq, scalar *rfreq, int linstances, int *count)
{
    scalar gain = pentropy;

    gain -= (((scalar)linstances)/instances)*MLDTClassEntropy(classes, lfreq, linstances, count);
    gain -= (((scalar)instances-linstances)/instances)*MLDTClassEntropy(classes, rfreq, instances-linstances, count);

    return gain;
}

scalar VarianceGain(int classes, scalar pvar, int *llist, int *rlist, int linstances, int rinstances, int **labels, int *n_labels)
{
    scalar gain = pvar;

    // Variance of left hand side
    scalar vleft = ClassLabelVariance(classes, linstances, labels, n_labels, llist);
    scalar vright = ClassLabelVariance(classes, rinstances, labels, n_labels, rlist);

    return gain - ((((scalar) linstances)/(linstances+rinstances))*vleft + (((scalar) rinstances)/(linstances+rinstances))*vright);
}

// Class label variance is obtained using Gini
// D. Kocev, C. Vens, J. Struyf and S. Džeroski (2013) "Tree ensembles for predicting structured outputs," Pattern Recognition 46:817–833


scalar ClassLabelVariance(int classes, int instances, int **labels, int *n_labels, int *list)
{
    // scalar var = 0.0;
    // // Variance of class labels in the subset
    // // Class vector
    // scalar mean[classes];
    // ZeroVector(mean, NULL, classes);
    // for (int i = 0; i < instances; i++)
    //     for (int j = 0; j < n_labels[list[i]]; j++)
    //         mean[labels[list[i]][j]]++;
    // for (int i = 0; i < classes; i++)
    //     mean[i] /= instances;
    //
    // scalar point[classes];
    // for (int i = 0; i < instances; i++) {
    //     ZeroVector(point, NULL, classes);
    //     for (int j = 0; j < n_labels[list[i]]; j++)
    //         point[labels[list[i]][j]] = 1.0;
    //     var += SqEuclideanDistance(point, mean, classes);
    // }
    //
    // return var/instances;

    scalar gini = 0.0, *p;
    MALLOC(p, scalar, classes, "ClassLabelVariance: p");
    ZeroVector(p, NULL, classes);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[list[i]]; j++)
            p[labels[list[i]][j]]++;

    for (int i = 0; i < classes; i++)
        p[i] /= instances;

    for (int i = 0; i < classes; i++)
        gini += 1.0 - (pow(p[i], 2.0) + pow(1.0-p[i], 2.0));

    FREE(p);
    return gini;
}
