/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "maths.h"
#include "ensemble.h"

// Implements ML-TREE method
// From: IEEE TRANSACTIONS ON KNOWLEDGE AND DATA ENGINEERING, VOL. 28, NO. 10, OCTOBER 2016
// Qingyao Wu, Mingkui Tan, Hengjie Song, Jian Chen, and Michael K. Ng "ML-FOREST: A Multi-Label Tree Ensemble Method for Multi-Label Classification"

void TrainMLTreeModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_MLTree *mltree, int *subspace, bool cv, t_Classifier classifier,
                       int *subset, int **SS, int **index)
{
    int *l_subspace;

    // Test if valid classification model
    if (classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ML-TREE classification method");

    // If no subspace select all
    if (!subspace) {
        MALLOC(l_subspace, int, inputs, "TrainMLTreeModels: l_subspace");
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = 1;
    } else
        l_subspace = subspace;

    // Label dependent instance selection and standard instance selection are incompatible
    if (subset && SS)
        OnErrorExit(ERR_PARAM, "Label dependent instance selection and standard instance selection are incompatible");

    // Initially all selected instances are used
    int *list, linstances = 0;
    MALLOC(list, int, instances, "TrainMLTreeModels: list");
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            list[linstances++] = i;

    mltree->method = classifier;
    MALLOC(mltree->root, struct MLTreeNode, 1, "TrainMLTreeModels: mltree->root");

    int b[classes];
    ZeroVector(NULL, b, classes);
    MLTreeBuild(parm, mltree->root, data, linstances, labels, n_labels, classes, index, b, list, cv, inputs, subspace, 0);

    if (!mltree->root->n)
        OnErrorExit(ERR_UNSPECTED_RESULT, "ML-TREE empty");

    SHOWPROGRESS(parm->show_progress > 0, "ML-TREE number of nodes: %d\n", MLTreeCountNodes(mltree));
    SAVEPROGRESS(parm->show_progress > 0, "@ML-TREE number of nodes: %d\n", MLTreeCountNodes(mltree));

    FREE(list);
    if (!subspace)
        FREE(l_subspace);
}

void MLTreeBuild(const t_Parameters *parm, struct MLTreeNode *node, scalar **data, int instances, int **labels, int *n_labels, int classes, int **index, int *b, int *list, bool cv, int inputs,
                 int *subspace, int depth)
{
    int **next_list, next_instances[classes];
    scalar p[classes];

    MALLOC(next_list, int *, classes, "MLTreeBuild: next_list");

    int n_children = SplitTest(parm, node, data, instances, labels, n_labels, classes, index, b, list, cv, inputs, subspace, next_list, p, next_instances);

    // No children, no further partition
    if (n_children <= 1 || (parm->mltree_depth && depth >= parm->mltree_depth)) {
        SHOWPROGRESS(parm->show_progress > 0, "MLTREE adding LEAF node at level %d\n", depth);
        node->n = 0;
        node->children = NULL;
        if (n_children) {
            ClassifierFree(parm, &(node->models[0]), parm->mltree_model, 2, inputs, &(node->ensembles[0]));
            FREE(node->models);
        } else
            node->models = NULL;
        MALLOC(node->b, int, classes, "MLTreeBuild: node->b");
        memcpy(node->b, b, sizeof(int)*classes);
        MALLOC(node->p, scalar, classes, "MLTreeBuild: node->p");
        memcpy(node->p, p, sizeof(scalar)*classes);

        FREE(next_list);
        return;
    }

    // Recursively create children
    MALLOC(node->children, struct MLTreeNode, n_children, "MLTreeBuild: node->children");
    node->n = n_children;
    node->p = NULL;
    node->b = NULL;

    SHOWPROGRESS(parm->show_progress > 0, "MLTREE adding %d INTERNAL nodes at level %d\n", n_children, depth);

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < n_children; i++) {
        // List of instances for each children
        int local_b[classes];
        CopyIntegerVector(b, local_b, classes);

        MLTreeBuild(parm, &(node->children[i]), data, next_instances[i], labels, n_labels, classes, index, local_b, next_list[i], cv, inputs, subspace, depth+1);

        FREE(next_list[i]);
    }

    FREE(next_list);
}

int SplitTest(const t_Parameters *parm, struct MLTreeNode *node, scalar **data, int instances, int **labels, int *n_labels, int classes, int **index, int *b, int *list, bool cv, int inputs,
              int *subspace, int **new_list, scalar *p, int *new_instances)
{
    // Construct b and p
    ZeroVector(p, NULL, classes);

    int c = 0;
    for (int i = 0; i < instances; i++) {
        c++;
        for (int j = 0; j < n_labels[list[i]]; j++)
            p[labels[list[i]][j]]++;
    }

    int not_relevant = 0;
    for (int i = 0; i < classes; i++) {
        p[i] /= c;
        b[i] = (p[i] >= parm->mltree_lambda || b[i]) ? 1 : 0;
        if (!b[i])
            not_relevant++;
    }

    // Build OVA classifiers for b[i] = 0
    if (!not_relevant)
        return 0;

    node->n = not_relevant;
    MALLOC(node->models, t_Learner, not_relevant, "SplitTest: node->models");

    // List of labels to train
    int train[classes];
    not_relevant = 0;
    for (int i = 0; i < classes; i++)
        if (!b[i])
            train[not_relevant++] = i;

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    // Construct OVA classifiers
    for (int i = 0; i < not_relevant; i++) {
        int *blabels, *S, **bindex = NULL;
        scalar **bdata;
        MALLOC(blabels, int, instances, "SplitTest: blabels");
        MALLOC(S, int, instances, "SplitTest: blabels");
        MALLOC(bdata, scalar *, instances, "SplitTest: bdata");
        if (index)
            MALLOC(bindex, int *, instances, "SplitTest: bindex");
        FillVector(NULL, S, instances, 1);

        for (int j = 0; j < instances; j++) {
            blabels[j] = LabelInList(train[i], labels[list[j]], n_labels[list[j]]) ? 1 :0;
            bdata[j] = data[list[j]];
            if (index)
                bindex[j] = index[list[j]];
        }

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, blabels, S, instances);

        // Cross-validate hyper-parameters?
        t_Parameters lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, bdata, instances, blabels, subspace, inputs, 2, lparm.mltree_model, 1, S, bindex);

        // Train
        SHOWPROGRESS(parm->show_progress > 0, "ML-TREE training binary model %*d/%d for label %*d\r", WIDTH(not_relevant), train[i]+1, not_relevant, WIDTH(not_relevant), i+1);
        ClassifierTraining(&lparm, lparm.mltree_model, bdata, instances, blabels, inputs, 2, bindex, subspace, S, true, &(node->models[i]),  &(node->ensembles[i]));
        SHOWPROGRESS(parm->show_progress > 0, "ML-TREE finished binary model %*d/%d for label %*d\r", WIDTH(not_relevant), train[i]+1, not_relevant, WIDTH(not_relevant), i+1);

        // For k-NN add labels
        if (lparm.mltree_model == kNN || lparm.mltree_model == NN) {
            MALLOC(node->models[i].knn->labels, int, instances, "TrainBinaryRelevanceModels: mode->models[i].knn->labels");
            memcpy(node->models[i].knn->labels, blabels, instances*sizeof(int));
        }

        FREE(S);
        FREE(blabels)
        FREE(bdata);
        if (index)
            FREE(bindex);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // Perform partition depending on assigned class
    for (int i = 0; i < not_relevant; i++) {
        MALLOC(new_list[i], int, instances, "SplitTest: next_list[i]");
        new_instances[i] = 0;
    }

    scalar outputs[not_relevant];
    for (int i = 0; i < instances; i++) {
        for (int j = 0; j < not_relevant; j++)
            Classifier2ClassOutput(parm, parm->mltree_model, data[list[i]], inputs, 2, index ? index[list[i]] : NULL, subspace, data, instances,
                                   (parm->mltree_model == kNN || parm->mltree_model == NN)? node->models[j].knn->labels : NULL, NULL, &(node->models[j]), 1, &outputs[j], index,
                                   &(node->ensembles[j]));

        int next = IndexOfMaximumOfScalarVector(outputs, not_relevant);
        new_list[next][new_instances[next]++] = list[i];
    }

    // Check that all classes have individuals
    for (int i = 0; i < not_relevant; i++)
        if (!new_instances[i]) {
            new_instances[i] = new_instances[--not_relevant];
            FREE(new_list[i]);
            new_list[i] = new_list[not_relevant];
            i--;
        }

    return not_relevant;
}

void MLTreeClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                            int inputs, int *query_index, t_MLTree *mltree, int *subspace, int *S, int **SS, int **index)
{
    int b[n_labels];

    RecursivelyMLTreeClassifyInstance(parm, query, f, classification, ranks, data, instances, n_labels, b, inputs, query_index, mltree->root, subspace, S, SS, index, mltree->method);

    // Relevant labels
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++)
        if (b[i])
            h[(*n_predicted)++] = i;

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *classification = *ranks = true;
}

void RecursivelyMLTreeClassifyInstance(const t_Parameters *parm, scalar *query, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *b, int inputs,
                                       int *query_index, struct MLTreeNode *node, int *subspace, int *S, int **SS, int **index, t_Classifier method)
{
    // Output of classifiers at node or node leaf
    if (node->n == 0) {
        memcpy(f, node->p, sizeof(scalar)*n_labels);
        memcpy(b, node->b, sizeof(int)*n_labels);
        return;
    }

    scalar outputs[node->n];
    for (int i = 0; i < node->n; i++)
        Classifier2ClassOutput(parm, method, query, inputs, 2, query_index, subspace, data, instances, NULL, S, &(node->models[i]), -1, &outputs[i], index, &(node->ensembles[i]));

    int child = IndexOfMaximumOfScalarVector(outputs, node->n);
    RecursivelyMLTreeClassifyInstance(parm, query, f, classification, ranks, data, instances, n_labels, b, inputs, query_index, &(node->children[child]), subspace, S, SS, index, method);
}

void FreeMLTree(const t_Parameters *parm, t_MLTree *tree, int inputs)
{
    RecursiveFreeMLTree(parm, tree->root, inputs, tree->method);

    FREE(tree->root);
}

void RecursiveFreeMLTree(const t_Parameters *parm, struct MLTreeNode *node, int inputs, t_Classifier method)
{
    for (int i = 0; i < node->n; i++) {
        RecursiveFreeMLTree(parm, &(node->children[i]), inputs, method);
        ClassifierFree(parm, &(node->models[i]), method, 2, inputs, &(node->ensembles[i]));
    }

    if (node->n) {
        FREE(node->children);
        FREE(node->models);
    } else {
        FREE(node->b);
        FREE(node->p);
    }
}

int MLTreeCountNodes(t_MLTree *tree)
{
    int nodes = 0;

    MLTreeCountAllNodes(tree->root, &nodes);

    return nodes;
}

void MLTreeCountAllNodes(struct MLTreeNode *tree, int *nodes)
{
    if (tree) {
        (*nodes) += tree->n;
        for (int i = 0; i < tree->n; i++)
            MLTreeCountAllNodes(&(tree->children[i]), nodes);
    }
}
