/***************************************************************************
 *   Copyright (C) 2011 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <mpi.h>
#include "macros.h"
#include "typex.h"
#include "common.h"
#include "ml.h"

// Functions to communicate master and slaves for cooperative co-evolutionary algorithm
void SendAndReceiveGenerationInfo(MPI_Comm communicator, const t_Parameters *parm, int root, int instances, int **population, int ***subpops, int n_subpops, scalar *pop_fitness, scalar *difference,
                                  int *k, scalar *qs, int **elite_members)
{
    // Population
    for (int i = 0; i < parm->population_size; i++)
        MPI_Bcast(population[i], instances, MPI_INT, root, communicator);

    // Subpopulations
    for (int i = 0; i < n_subpops; i++)
        for (int j = 0; j < parm->subpop_size; j++)
            MPI_Bcast(subpops[i][j], instances, MPI_INT, root, communicator);

#ifdef SCALAR_IS_DOUBLE
    MPI_Bcast(pop_fitness, parm->population_size, MPI_DOUBLE, root, communicator);
    MPI_Bcast(qs, parm->population_size, MPI_DOUBLE, root, communicator);
    MPI_Bcast(difference, n_subpops, MPI_DOUBLE, root, communicator);
#else
    MPI_Bcast(pop_fitness, parm->population_size, MPI_FLOAT, root, communicator);
    MPI_Bcast(qs, parm->population_size, MPI_FLOAT, root, communicator);
    MPI_Bcast(difference, n_subpops, MPI_FLOAT, root, communicator);
#endif

    MPI_Bcast(k, parm->population_size, MPI_INT, root, communicator);
    for (int i = 0; i < n_subpops; i++)
        MPI_Bcast(elite_members[i], parm->subpop_size, MPI_INT, root, communicator);
}


// With broadcast send/receive it's done with the same function
void SendData(MPI_Comm communicator, int instances, scalar **data, int **labels, int *n_labels, int inputs, int **index, int root, int n_subpops, int classes, int sparse)
{
    int *length = NULL;

    // If not root is receiving info
    MPI_Bcast(&instances, 1, MPI_INT, root, communicator);
    MPI_Bcast(&classes, 1, MPI_INT, root, communicator);
    MPI_Bcast(&inputs, 1, MPI_INT, root, communicator);
    MPI_Bcast(&n_subpops, 1, MPI_INT, root, communicator);
    MPI_Bcast(&sparse, 1, MPI_INT, root, communicator);

    if (sparse) {
        // Length of every instance
        MALLOC(length, int, instances, "SendData: length");

        // If root obtain count and send
        for (int i = 0; i < instances; i++) {
            int count = 0;
            while (index[i][count] != -1)
                count++;
            length[i] = count;
        }
        MPI_Bcast(length, instances, MPI_INT, root, communicator);
    }

    // Data info
    MPI_Bcast(n_labels, instances, MPI_INT, root, communicator);
    for (int i = 0; i < instances; i++) {

#ifdef SCALAR_IS_FLOAT
        MPI_Bcast(data[i], sparse ? length[i] : inputs, MPI_FLOAT, root, communicator);
#else
        MPI_Bcast(data[i], sparse ? length[i] : inputs, MPI_DOUBLE, root, communicator);
#endif
        if (sparse)
            MPI_Bcast(index[i], length[i]+1, MPI_INT, root, communicator);
        MPI_Bcast(labels[i], n_labels[i], MPI_INT, root, communicator);
    }

    if (sparse)
        FREE(length);
}

// With broadcast send/receive it's done with the same function
void ReceiveData(MPI_Comm communicator, int *instances, scalar ***data, int ***labels, int **n_labels, int *inputs, int ***index, int root, int *n_subpops, int *classes, int *sparse)
{
    int *length = NULL;

    MPI_Bcast(instances, 1, MPI_INT, root, communicator);
    MPI_Bcast(classes, 1, MPI_INT, root, communicator);
    MPI_Bcast(inputs, 1, MPI_INT, root, communicator);
    MPI_Bcast(n_subpops, 1, MPI_INT, root, communicator);
    MPI_Bcast(sparse, 1, MPI_INT, root, communicator);

    MALLOC(*data, scalar *, *instances, "main: data");
    MALLOC(*labels, int *, *instances, "main: labels");
    MALLOC(*n_labels, int, *instances, "main: n_labels");

    if (*sparse) {
        // Length of every instance
        MALLOC(length, int, *instances, "ReceiveData: length");
        MPI_Bcast(length, *instances, MPI_INT, root, communicator);
    }

    // Data info
    MPI_Bcast(*n_labels, *instances, MPI_INT, root, communicator);
    for (int i = 0; i < *instances; i++) {
        // Sparse format
        if (*sparse) {
            MALLOC((*data)[i], scalar, length[i], "ReceiveData: data[i]");
            MALLOC((*index)[i], int, length[i]+1, "ReceiveData: data[i]");
        } else
            MALLOC((*data)[i], scalar, *inputs, "ReceiveData: data[i]");

        MALLOC((*labels)[i], int, (*n_labels)[i], "ReceiveData: data[i]");

#ifdef SCALAR_IS_FLOAT
        MPI_Bcast((*data)[i], *sparse ? length[i] : *inputs, MPI_FLOAT, root, communicator);
#else
        MPI_Bcast((*data)[i], *sparse ? length[i] : *inputs, MPI_DOUBLE, root, communicator);
#endif
        if (*sparse)
            MPI_Bcast((*index)[i], length[i]+1, MPI_INT, root, communicator);
        MPI_Bcast((*labels)[i], (*n_labels)[i], MPI_INT, root, communicator);
    }

    if (*sparse)
        FREE(length);
}
