/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <mpi.h>
#include <string.h>
#include <unistd.h>
#include "macros.h"
#include "typex.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

// Global variables for C4.5 algorithm
unsigned short MaxClass, MaxDiscrVal = 2;
ItemNo MaxItem;
int MaxAtt;
Description *Item;
DiscrValue *MaxAttVal;
char *SpecialStatus;
String *ClassName, *AttName, * *AttValName, FileName;
short VERBOSITY, TRIALS;
Boolean GAINRATIO = true, SUBSET = false;
ItemNo MINOBJS = 2, WINDOW, INCREMENT;
float ConfLimit = 0.25;
Tree *Pruned;
Boolean AllKnown = true;

// MPI slave to evolve a subpopulation using CHC algorithm

int main(int argc, char *argv[])
{
    t_Parameters parm;
    int instances, **labels, *n_labels, classes, inputs, n_subpops, **index = NULL;
    scalar **data;
    cache_t **neighbors_cache = NULL;

    // Get communicator with parent
    MPI_Comm communicator, parent;

    if (MPI_Init(&argc, &argv) != MPI_SUCCESS)
        OnErrorExit(ERR_MPI, "MPI Initialize failed on slave");
    else
        printf("MPI Initialize successfully on slave\n");
    MPI_Comm_get_parent(&parent);
    if (MPI_Intercomm_merge(parent, 1, &communicator) != MPI_SUCCESS)
        OnErrorExit(ERR_MPI, "Cannot merge inter communicators");

    // subpopulation index corresponds to rank
    int p_index;
    MPI_Comm_rank(MPI_COMM_WORLD, &p_index);

    char name[MAX_FILENAME_LENGTH];
    gethostname(name, MAX_FILENAME_LENGTH);
    SHOWPROGRESS(true, "%d slave started execution in host: %s\n", p_index+1, name);

    // Receive root rank
    int root;
    MPI_Status status;
    MPI_Recv(&root, 1, MPI_INT, MPI_ANY_SOURCE, 0, communicator, &status);
    SHOWPROGRESS(TEST, "%d slave received root rank %d\n", p_index+1, root);

    // Receive parameters
    MPI_Bcast((void *) &parm, sizeof(t_Parameters), MPI_BYTE, root, communicator);
    SHOWPROGRESS(TEST, "%d slave received parameters\n", p_index+1);

    // Initialize random number generator
    char state[256];
    if (parm.seed == 0) {
        FILE *randDev;

        if ((randDev = fopen("/dev/urandom", "r")) == NULL)
            OnErrorExit(1, "main: Cannot open random device\n");

        if (!fread(&parm.seed, 1, sizeof(int), randDev))
            OnErrorExit(ERR_READ, "main");
        fclose(randDev);
    }

    char file_name[100];
    strcpy(file_name, "/dev/null");
    if ((parm.out_file = fopen(file_name, "w")) == NULL)
        OnErrorExit(ERR_FILE, "Cannot open output file using /dev/null");

    srandom(parm.seed);
    srand(parm.seed);
    for (int i = 0; i < 256; i++)
        state[i] = random();
    initstate(parm.seed, state, 256);
    setstate(state);

    // Check parameters
    if (parm.reevaluate_individuals)
        OnErrorExit(ERR_PARAM, "Reevaluate incompatible with MPI");

    // Allocate memory
    int sparse = 0;
    ReceiveData(communicator, &instances, &data, &labels, &n_labels, &inputs, &index, root, &n_subpops, &classes, &sparse);
    if (!sparse)
        index = NULL;
    SHOWPROGRESS(TEST, "%d slave received data correctly\n", p_index+1);

    // Cache of neighbors?
    SHOWPROGRESS(TEST, "BEGIN Cache construction finished%c", '\n');
    if (parm.use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE) {
        int neighbors_cache_size = 0;
        ConstructAllNeighborsCache(data, instances, inputs, NULL, &neighbors_cache, NULL, NULL, &neighbors_cache_size, NULL, parm.ml_loo);
        SHOWPROGRESS(TEST, "DONE Cache construction finished%c", '\n');
    } else
        SHOWPROGRESS(TEST, "Neighbors cache DISABLED%c", '\n');

    // Populations information
    int **population, * **subpops, *k, **elite_members, minimum_distance;
    scalar *pop_fitness, *subpop_fitness, *qs, *difference;
    bool first_time = true;

    MALLOC(population, int *, parm.population_size, "main: population");
    MALLOC(elite_members, int *, n_subpops, "main: elite_members");
    MALLOC(pop_fitness, scalar, parm.population_size, "main: population");
    MALLOC(qs, scalar, parm.population_size, "main: qs");
    MALLOC(k, int, parm.population_size, "main: k");
    for (int i = 0; i < parm.population_size; i++)
        MALLOC(population[i], int, instances, "main: population[i]");
    MALLOC(subpops, int **, n_subpops, "main: subpops");
    MALLOC(subpop_fitness, scalar, 2*parm.subpop_size, "main: subpop_fitness");
    MALLOC(difference, scalar, n_subpops, "main: difference");
    for (int i = 0; i < n_subpops; i++) {
        MALLOC(subpops[i], int *, 2*parm.subpop_size, "main: subpops[i]");
        MALLOC(elite_members[i], int, parm.subpop_size, "main: elite_members[i]");
        for (int j = 0; j < 2*parm.subpop_size; j++)
            MALLOC(subpops[i][j], int, instances, "main: subpops[i][j]");
    }

    minimum_distance = instances/parm.chc_subpop_minimum_distance_factor;

    SHOWPROGRESS(parm.show_progress > 2, "%d slave starting evolution\n", p_index+1);

    int end = 0;

    while (!end) {
        // Receive information of populations
        SendAndReceiveGenerationInfo(communicator, (const t_Parameters *) &parm, root, instances, population, subpops, n_subpops, pop_fitness, difference, k, qs, elite_members);
        SHOWPROGRESS(parm.show_progress > 2, "%d slave received next generation info\n", p_index+1);

        // Nsp loops of subpopulations
        for (int nsp = 0; nsp < parm.nsp; nsp++) {
            SHOWPROGRESS(parm.show_progress > 2, "Nsp generation of subpopulation %d: %*d/%d\n", p_index+1, WIDTH(parm.nsp), nsp+1, parm.nsp);

            PerformGeneration((const t_Parameters *) &parm, subpops[p_index], parm.subpop_size, instances, &minimum_distance, p_index, population, data, labels, n_labels, classes, inputs, n_subpops,
                              pop_fitness, subpops, neighbors_cache, subpop_fitness, &first_time, NULL, NULL, &difference[p_index], qs, index, elite_members[p_index], k, NULL, parm.ml_method);
        }

        first_time = false;

        // Send information of populations to parent
        MPI_Send(&p_index, 1, MPI_INT, root, 0, communicator);
        for (int i = 0; i < parm.subpop_size; i++)
            MPI_Send(subpops[p_index][i], instances, MPI_INT, root, 0, communicator);

        MPI_Bcast(&end, 1, MPI_INT, root, communicator);
    }

    MPI_Barrier(communicator);

    SHOWPROGRESS(TEST, "%d SUBPOPULATION slave ended correctly\n", p_index+1);

    MPI_Finalize();
}
