/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include "common.h"
#include "ml.h"
#include "macros.h"
#include "is.h"
#include "maths.h"
#include "ensemble.h"

// Mr.KNN
// X. Lin and X.-W. Chen (2010) "Mr.KNN: soft relevance for multi-label classification", CIKM '10: Proceedings of the 19th ACM international conference on Information and knowledge management
// October 2010 Pages 349–358

#define KVALUES 8
#define FVALUES 4

void TrainMrKNN(const t_Parameters *parm, t_MrKNN *mrknn, scalar **data, int instances, int **labels, int *n_labels, int **index, int classes, int *subset, int *subspace, int inputs,
                cache_t **neighbors_cache)
{
    // Following paper
    AllocateScalarMatrix(&(mrknn->u), classes, instances);

    // Validation of f and k?
    if (!parm->cv_hyperparameters) {
        mrknn->f = parm->mrknn_f;
        mrknn->k = parm->mlknn_k;

        // Obtain class memberships
        SHOWPROGRESS(parm->show_progress > 1, "Mr.KNN clustering for k = %d, f = %g\n", mrknn->k, mrknn->f);
        MultiLabelFuzzyCMeans(data, instances, labels, n_labels, classes, inputs, mrknn->u, subset, parm->kmeans_max_iterations, mrknn->f, subspace, index);
    } else {
        scalar largest = -INFINITE, f_values[FVALUES] = {1.0, 2.0, 4.0, 6.0};
        int k_values[KVALUES] = {5, 10, 15, 20, 25, 30, 35, 40};

#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic) collapse(2)
#endif
        for (int i = 0; i < KVALUES; i++)
            for (int j = 0; j < FVALUES; j++) {
                scalar **u, average_margin = 0.0, votes[classes];
                int k = k_values[i];
                scalar f = f_values[j];

                AllocateScalarMatrix(&u, classes, instances);
                SHOWPROGRESS(parm->show_progress > 1, "Mr.KNN clustering for k = %d, f = %g\n", k, f);
                MultiLabelFuzzyCMeans(data, instances, labels, n_labels, classes, inputs, u, subset, parm->kmeans_max_iterations, f, subspace, index);

                for (int m = 0; m < instances; m++)
                    if (!subset || subset[m]) {
                        int neighbors[k];
                        GetkNeighborsForML(data[m], data, instances, index, inputs, index ? index[m] : NULL, k, neighbors, subset, neighbors_cache, -1, false, NULL, classes, NULL, -1, subspace);

                        // Compute votes
                        ZeroVector(votes, NULL, classes);
                        for (int l = 0; l < k; l++)
                            if (neighbors[l] != -1) {
                                scalar dist = MinkowskiDistanceSparse(data[m], data[neighbors[l]], inputs, f, subspace, index ? index[m] : NULL, index ? index[neighbors[l]] : NULL);
                                for (int n = 0; n < classes; n++)
                                    votes[n] += exp(-dist)*u[n][neighbors[l]];
                            }

                        // Min votes for relevant label
                        scalar min = votes[labels[m][0]];
                        for (int l = 1; l < n_labels[m]; l++)
                            if (votes[labels[m][l]] < min)
                                min = votes[labels[m][l]];

                        // Max votes for irrelevant label
                        scalar max = 0.0;
                        for (int l = 0; l < classes; l++)
                            if (votes[l] > max && !LabelInList(l, labels[m], n_labels[m]))
                                max = votes[l];

                        average_margin += min-max;
                    }

                average_margin /= instances;
                SHOWPROGRESS(parm->show_progress > 1, "Average margin for k = %2d, f %g: %lf\n", k, f, average_margin);

#ifdef _OPENMP
                #pragma omp critical
#endif
                if (RANDOM(0.0, 1.0) < 0.5) {
                    largest = average_margin;
                    CopyScalarMatrix(u, mrknn->u, classes, instances);
                    mrknn->f = f;
                    mrknn->k = k;
                }

                FreeScalarMatrix(&u, classes);
            }

        SAVEPROGRESS(true, "BEST k = %d for Mr.KNN with average margin %lf\n", mrknn->k, largest);
        SAVEPROGRESS(true, "BEST f = %g for Mr.KNN\n", mrknn->f);
    }

    // Obtain ALL f's
    scalar **f;
    int **h, *n_h, *S;

    AllocateIntegerMatrix(&h, instances, classes);
    AllocateScalarMatrix(&f, instances, classes);
    MALLOC(n_h, int, instances, "TrainMrKNN: n_h");
    MALLOC(S, int, instances, "TrainMrKNN: S");
    mrknn->th = 0.0;    // Actually not used at this point

    if (subset)
        memcpy(S, subset, sizeof(int)*instances);
    else
        FillVector(NULL, S, instances, 1.0);

    for (int i = 0; i < instances; i++) {
        // Leave-one-out
        int old = S[i];
        S[i] = parm->mrknn_loo ? 0 : S[i];
        MrKNNClassify(mrknn, data[i], index ? index[i] : NULL, data, instances, index, classes, S, subspace, inputs, neighbors_cache, f[i], h[i], &n_h[i]);
        S[i] = old;
    }

    // Evaluate thresholds using a metric
    // scalar best = 1.0;
    scalar best = -1.0;
    t_Performance performance;
    for (scalar th = 0.00; th <= 0.5; th += 0.01) {
        SHOWPROGRESS(parm->show_progress > 1, "Mr.KNN evaluating threshold: %lf ", th);
        // // Best using Hamming Loss
        // scalar hl = 0.0;
        // for (int i = 0; i < instances; i++)
        //     for (int j = 0; j < classes; j++) {
        //         // PrintVector(f[i], NULL, classes, stdout);
        //         bool rel = LabelInList(j, labels[i], n_labels[i]);
        //         if ((f[i][j] > th && !rel) || (f[i][j] <= th && rel))
        //             hl++;
        //     }
        //
        // hl /= instances*classes;
        //
        // SHOWPROGRESS(parm->show_progress > 0, "with Hamming Loss: %lf\n", hl);
        //
        // if (hl < best) {
        //     mrknn->th = th;
        //     best = hl;
        // }

        // Best using F1
        for (int i = 0; i < instances; i++) {
            n_h[i] = 0;
            for (int j = 0; j < classes; j++)
                if (f[i][j] > th)
                    h[i][n_h[i]++] = j;
        }

        PrintClassificationMetrics(&performance, labels, n_labels, h, n_h, instances, classes, true, NULL, NULL, false);
        SHOWPROGRESS(parm->show_progress > 1, "with F1: %lf\n", performance.eb_f1);

        if (performance.eb_f1 > best) {
            mrknn->th = th;
            best = performance.eb_f1;
        }
    }

    // SAVEPROGRESS(true, "Mr.KNN optimum threshold: %lf with Hamming Loss %lf\n", mrknn->th, best);
    SAVEPROGRESS(true, "Mr.KNN optimum threshold: %lf with F1 %lf\n", mrknn->th, best);

    FreeIntegerMatrix(&h, instances);
    FreeScalarMatrix(&f, instances);
    FREE(n_h);
    FREE(S);
}

void MultiLabelFuzzyCMeans(scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, scalar **u, int *subset, int max_iterations, scalar f, int *subspace, int **index)
{
    // Initialize centroids as class centers
    scalar **centroids, m = 2.0;    // Fuzzifier fixed to 2 as suggested by authors
    int n[classes];

    AllocateScalarMatrix(&centroids, classes, inputs);
    ZeroScalarMatrix(centroids, classes, inputs);
    ZeroVector(NULL, n, classes);

    if (!index) {
        for (int i = 0; i < instances; i++)
            if (!subset || subset[i])
                for (int j = 0; j < n_labels[i]; j++) {
                    n[labels[i][j]]++;
                    for (int k = 0; k < inputs; k++)
                        centroids[labels[i][j]][k] += data[i][k];
                }
    } else {
        for (int i = 0; i < instances; i++)
            if (!subset || subset[i])
                for (int j = 0; j < n_labels[i]; j++) {
                    n[labels[i][j]]++;
                    for (int k = 0; index[i][k] != -1; k++)
                        centroids[labels[i][j]][index[i][k]-1] += data[i][k];
                }
    }

    for (int i = 0; i < classes; i++)
        if (n[i])
            for (int j = 0; j < inputs; j++)
                centroids[i][j] /= n[i];

    for (int iter = 0; iter < max_iterations; iter++) {
        // Update u membership function
        for (int i = 0; i < instances; i++) {
            scalar sum = 0.0;
            for (int j = 0; j < classes; j++)
                sum += (u[j][i] = ((int) LabelInList(j, labels[i], n_labels[i]))/MinkowskiDistanceSparseDense(data[i], centroids[j], inputs, f, subspace, index ? index[i] : NULL));

            if (sum > ZERO)
                for (int j = 0; j < classes; j++)
                    u[j][i] /= sum;
        }

        ZeroScalarMatrix(centroids, classes, inputs);
        for (int k = 0; k < classes; k++) {
            scalar sum = 0.0;
            if (index)
                for (int i = 0; i < instances; i++) {
                    sum += pow(u[k][i], m);
                    for (int l = 0; index[i][l] != -1; l++)
                        centroids[k][index[i][l]-1] += data[i][l] * pow(u[k][i], m);
                }
            else
                for (int i = 0; i < instances; i++) {
                    sum += pow(u[k][i], m);
                    for (int l = 0; l < inputs; l++)
                        centroids[k][l] += data[i][l] * pow(u[k][i], m);
                }

            if (sum > ZERO)
                for (int l = 0; l < inputs; l++)
                    centroids[k][l] /= sum;
        }
    }

    FreeScalarMatrix(&centroids, classes);
}

void MrKNNClassify(t_MrKNN *mrknn, scalar *query, int *query_index, scalar **data, int instances, int **index, int classes, int *subset, int *subspace, int inputs, cache_t **neighbors_cache,
                   scalar *f, int *h, int *n_predicted)
{
    // Compute neighbors
    int neighbors[mrknn->k];
    GetkNeighborsForML(query, data, instances, index, inputs, query_index, mrknn->k, neighbors, subset, neighbors_cache, -1, false, NULL, classes, NULL, -1, subspace);

    // Compute votes
    ZeroVector(f, NULL, classes);
    for (int l = 0; l < mrknn->k; l++)
        if (neighbors[l] != -1) {
            scalar dist = MinkowskiDistanceSparse(query, data[neighbors[l]], inputs, mrknn->f, subspace, query_index, index ? index[neighbors[l]] : NULL);
            for (int n = 0; n < classes; n++)
                f[n] += exp(-dist)*mrknn->u[n][neighbors[l]];
        }

    // PrintVector(f, NULL, classes, stdout);

    NormalizeVector(f, classes);

    *n_predicted = 0;
    for (int i = 0; i < classes; i++)
        if (f[i] > mrknn->th)
            h[(*n_predicted)++] = i;
}

void FreeMrKNN(t_MrKNN *mrknn, int classes)
{
    FreeScalarMatrix(&(mrknn->u), classes);
}
