/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements Boosting Instance+Feature Selection

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <stdbool.h>
#include "ml.h"
#include "macros.h"
#include "common.h"
#include "maths.h"

#define NAME_LENGTH 4096

// Global variables for C4.5 algorithm
unsigned short MaxClass, MaxDiscrVal = 2;
ItemNo MaxItem;
int MaxAtt;
Description *Item;
DiscrValue *MaxAttVal;
char *SpecialStatus;
String *ClassName, *AttName, * *AttValName, FileName;
short VERBOSITY, TRIALS;
Boolean GAINRATIO = true, SUBSET = false;
ItemNo MINOBJS = 2, WINDOW, INCREMENT;
float ConfLimit = 0.25;
Tree *Pruned;
Boolean AllKnown = true;

int main(int argc, char *argv[])
{
    time_t init_time;
    char *home, path[FILENAME_LENGTH], data_root[FILENAME_LENGTH], dataset_name[FILENAME_LENGTH];
    t_Parameters parm;
    scalar **train_data, **test_data;
    int **train_labels, **test_labels, train_instances, test_instances, **train_index = NULL, **test_index = NULL, inputs, classes, *train_n_labels, *test_n_labels;

    // Continue?
    int opt, shitf = 0;
    bool remove_output = true;
    while ((opt = getopt(argc, argv, "n")) != EOF)
        switch (opt) {
            case 'n':
                // Remove output
                remove_output = false;
                shitf++;
                break;
        }

    argv += shitf;
    argc -= shitf;


    // Test enough parameters
    if (argc >= 4) {
        // File for output messages.
        if ((parm.out_file = fopen(argv[3], remove_output ? "w" : "a")) == NULL)
            OnErrorExit(1, "main: Cannot open output file\n");

        LoadGlobalParameters(&parm, argv[2], "multi-label");
    } else {
        printf("multi-label problem parameters output\n");
        exit(1);
    }

    // Initialize random number generator
    char state[256];
    if (parm.seed == 0) {
        FILE *randDev;

        if ((randDev = fopen("/dev/urandom", "r")) == NULL)
            OnErrorExit(1, "main: Cannot open random device\n");

        if (!fread(&parm.seed, 1, sizeof(int), randDev))
            OnErrorExit(ERR_READ, "main");
        fclose(randDev);
        fprintf(parm.out_file, "Random seed: %d\n", parm.seed);
        fflush(parm.out_file);
        fdatasync(fileno(parm.out_file));
    } else {
        // Fixed seed to make the evolution reproducible.
        fprintf(parm.out_file, "Random seed: %d (fixed)\n", parm.seed);
    }

    srandom(parm.seed);
    srand(parm.seed);
    for (int i = 0; i < 256; i++)
        state[i] = random();
    initstate(parm.seed, state, 256);
    setstate(state);

    // Time
    init_time = time(NULL);

    // Load data
    home = getenv("HOME");
    sprintf(data_root, "%s/%s", home, parm.datad);

    // Training file
    sprintf(dataset_name, "%s.train", argv[1]);
    if (!FindFile(dataset_name, data_root, path))
        OnErrorExit(ERR_FILE, "Training file not found");

    if (parm.sparse_format) {
        // Read sparse data and check whether it uses less space than dense format
        ReadMultiLabelData((const t_Parameters *) &parm, &train_data, &train_labels, &train_n_labels, &train_instances, &inputs, &classes, path, &train_index);
        scalar sparse_size = ((scalar) train_instances)*(sizeof(scalar *)+sizeof(int *));
        for (int i = 0; i < train_instances; i++) {
            int c = 0;
            for (c = 0; train_index[i][c] != -1; c++);
            sparse_size += (sizeof(scalar) + sizeof(int)) * c;
        }
        sparse_size /= pow(2.0, 30.0);

        scalar dense_size = (((scalar) train_instances)*inputs*sizeof(scalar) + ((scalar) train_instances)*sizeof(scalar *))/pow(2.0, 30.0);

        // Switch to dense?
        SHOWPROGRESS(parm.show_progress > 0, "Sparse size: %.5lf GB\nDense size:  %.5lf GB\n", sparse_size, dense_size);
        if (sparse_size >= dense_size) {
            SHOWPROGRESS(parm.show_progress > 0, "Switching to dense format%c", '\n');
            parm.sparse_format = false;

            // Free memory and load file in standard format
            for (int i = 0; i < train_instances; i++) {
                FREE(train_data[i]);
                FREE(train_index[i]);
            }
            FREE(train_labels);
            FREE(train_data);
            FREE(train_index);

            ReadMultiLabelData((const t_Parameters *) &parm, &train_data, &train_labels, &train_n_labels, &train_instances, &inputs, &classes, path, &train_index);
        }
    } else
        ReadMultiLabelData((const t_Parameters *) &parm, &train_data, &train_labels, &train_n_labels, &train_instances, &inputs, &classes, path, &train_index);

    // Set maximum k for ML-kNN if too large
    if (parm.ml_max_k > train_instances)
        parm.ml_max_k = train_instances;

    // Maximum k for ML-k'sNN cannot be smaller
    parm.mlksnn_max_k = MAX(parm.mlksnn_max_k, parm.ml_max_k);

    // Testing file
    sprintf(dataset_name, "%s.gen", argv[1]);
    if (!FindFile(dataset_name, data_root, path))
        OnErrorExit(ERR_FILE, "Training file not found");

    ReadMultiLabelData((const t_Parameters *) &parm, &test_data, &test_labels, &test_n_labels, &test_instances, &inputs, &classes, path, &test_index);

    // Typify input training and testing data
    if (parm.typify) {
        if (!parm.sparse_format)
            TypifyInputData(train_data, test_data, train_instances, test_instances, inputs);
        else
            TypifySparseData(train_data, test_data, train_instances, test_instances, inputs, train_index, test_index);
    }

    // Preprocessing: Instance selection
    // Hash table only available for not evolving k
    if (parm.ml_use_hash_table && parm.ml_max_k != parm.ml_min_k)
        parm.ml_use_hash_table = false;

    // Print label info
    scalar lcard, lden, pldiv, MeanIR, IRLbl[classes], CVIR, ml[classes+1];
    int ldiv;
    MLDatasetInfo(train_instances, train_labels, train_n_labels, test_instances, test_labels, test_n_labels, classes, &lcard, &lden, &ldiv, &pldiv, parm.out_file, &MeanIR, IRLbl, &CVIR, NULL,
                  inputs, ml);

    // Training only label info for CI aware methods
    int *clabel;
    MALLOC(clabel, int, classes, "main: clabel");
    MLDatasetInfo(train_instances, train_labels, train_n_labels, 0, NULL, NULL, classes, &lcard, &lden, &ldiv, &pldiv, NULL, &MeanIR, IRLbl, &CVIR, clabel, inputs, ml);

    int *selection = NULL, **ss = NULL, *ks = NULL, **la_ks = NULL;

    // Class-imbalanced methods
    ClassImbalanceMethodMultiLabel((const t_Parameters *) &parm, &train_instances, &train_data, &train_labels, &train_n_labels, classes, inputs, &train_index, IRLbl, &MeanIR, clabel, &CVIR, clabel,
                                   NULL, &ss);

    MALLOC(la_ks, int *, parm.lamlknn_max_n_clusters, "main: la_ks");
    MALLOC(ks, int, classes, "main: ks");
    for (int i = 0; i < parm.lamlknn_max_n_clusters; i++)
        MALLOC(la_ks[i], int, classes, "main. la_ks[i]");
    for (int i = 0; i < classes; i++)
        ks[i] = -1;

    // If no instance selection algorithm nothing is done
    int *chain = NULL;
    if (parm.ml_instance_selection != NoInstanceSelection) {
        scalar is_reduction = InstanceSelectionML((const t_Parameters *) &parm, train_data, train_labels, train_n_labels, train_instances, inputs, classes, train_index, &selection, &ss, ks,
                                                parm.ml_method, la_ks, test_data, test_labels, test_n_labels, test_instances, test_index, IRLbl, MeanIR, &chain);
        ShowMLInstanceSelectionInfo((const t_Parameters *) &parm, train_instances, classes, selection, ss, is_reduction);
    }

    if (parm.save_selection) {
        sprintf(dataset_name, "%s.selected", argv[1]);
        SaveSelectedDataset(train_data, train_labels, train_instances,inputs, train_n_labels, classes, train_index, selection, dataset_name);
    }

    // If no feature selection algorithm nothing is done
    int *subspace = NULL, **ssubspace = NULL;
    if (parm.ml_feature_selection != NoFeatureSelection) {
        CALLOC(subspace, int, inputs, "main: subspace");
        scalar fs_reduction = FeatureSelectionML((const t_Parameters *) &parm, train_data, train_labels, train_n_labels, train_instances, inputs, classes, train_index, &subspace, ssubspace,
                                                parm.ml_method, test_data, test_labels, test_n_labels, test_instances, test_index, &chain);
        fprintf(parm.out_file, "Feature reduction: %lf\n", fs_reduction);
    }

    // Test classifiers
    char *outputs = NULL;
    if (parm.save_outputs)
        REALLOC(outputs, char, strlen(argv[1])+9, "main: outputs");

    TestAllMultiLabelClassifiers((const t_Parameters *) &parm, test_data, test_labels, test_n_labels, train_data, train_instances, train_labels, train_n_labels, classes, inputs, test_index,
                                 train_index, test_instances, selection, NULL, ss, outputs, ks[0] == -1 ? parm.mlknn_k : ks[0], ks, la_ks, chain, subspace);

    TestAllExternalMultiLabelClassifiers((const t_Parameters *) &parm, test_data, test_labels, test_n_labels, train_data, train_instances, train_labels, train_n_labels, classes, inputs, test_index,
                                         train_index, test_instances, selection, argv[1], ss, subspace);

    // Used time so far
    time_t total_time = time(NULL) - init_time;
    fprintf(parm.out_file, "@Total time: %ld\n", total_time);
    printf("Elapsed time: %02ld:%02ld:%02ld\n", total_time/3600, (total_time % 3600)/60, total_time % 60);

    // All finished well
    fprintf(parm.out_file, "@Ok\n");
    fclose(parm.out_file);

#ifdef MPI
    if (!ISSKYNET())    // Bug in skynet OPENMPI version
        MPI_Finalize();
#endif

    return 0;
}
