/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"
#include "ensemble.h"

// Implement MULTI-LABEL option decision trees
// T. Stepišnik, D. Kocev, S. Džeroski (2020) "Option Predictive Clustering Trees for Multi-label Classification," Acta Polytechnica Hungarica 17(10):109-128.
// Use only for PCT as in the paper

void OptionMLDecisionTreeClassify(struct omldt_node *tree, scalar *instance, scalar *f, int classes, int *index, int *h, int *n_predicted, bool *classification, bool *ranks)
{
    // Leaf node?
    if (tree->leaf)
        memcpy(f, tree->probabilities, sizeof(scalar)*classes);
    else if (tree->n_options) {
        // Aggregate all options
        scalar probs[classes];
        for (int i = 0; i < tree->n_options; i++) {
            OptionMLDecisionTreeClassify(&(tree->options[i]), instance, probs, classes, index, h, n_predicted, classification, ranks);
            for (int j = 0; j < classes; j++)
                f[j] += probs[j];
        }
    } else {
        if ((index ? SparseFeature(instance, index, tree->feat) : instance[tree->feat]) <= tree->cut_point)
            OptionMLDecisionTreeClassify(tree->left, instance, f, classes, index, h, n_predicted, classification, ranks);
        else
            OptionMLDecisionTreeClassify(tree->right, instance, f, classes, index, h, n_predicted, classification, ranks);
    }

    *n_predicted = 0;
    scalar t = 0.5;

    NormalizeVector(f, classes);

    for (int i = 0; i < classes; i++)
        if (f[i] > t)
            h[(*n_predicted)++] = i;

    *classification = *ranks = true;
}

void OptionMLDecisionTreeTrain(const t_Parameters *parm, scalar **data, int instances, int **labels, int features, t_OMLDTree *tree, int *subset, int *subspace, int classes, int **index,
                               int *n_labels, bool for_rf, bool pct)
{
    int *f_list, fs = 0;
    MALLOC(f_list, int, features, "OptionMLDecisionTreeTrain: f_list");
    for (int i = 0; i < features; i++)
        if (!subspace || subspace[i]) {
            f_list[fs] = i;
            fs++;
        }

    int *i_list, is = 0;
    MALLOC(i_list, int, instances, "OptionMLDecisionTreeTrain: i_list");

    MALLOC(tree->root, struct omldt_node, 1, "OptionMLDecisionTreeTrain: tree");
    CALLOC(tree->root->probabilities, scalar, classes, "OptionMLDecisionTreeTrain: tree->probabilities");

    for (int i = 0; i < instances; i++)
        if (!subset || subset[i]) {
            i_list[is] = i;
            is++;

            for (int j = 0; j < n_labels[i]; j++)
                tree->root->probabilities[labels[i][j]]++;
        }

    // Build root node with no children and label global probabilities
    tree->root->leaf = true;
    tree->root->depth = 0;
    tree->root->cut_point = 0.0;
    tree->root->feat = -1;
    tree->root->n_options = 0;
    tree->root->left = tree->root->right = tree->root->options = NULL;

    // Split measure is entropy
    tree->root->score = pct ? ClassLabelVariance(classes, instances, labels, n_labels, i_list) : MLDTClassEntropy(classes, tree->root->probabilities, is, NULL);

    // Class distribution
    for (int i = 0; i < classes; i++)
        tree->root->probabilities[i] /= is;

    OptionMLDecisionTreeBuild(parm, tree->root, data, labels, is, i_list, fs, f_list, classes, 0, index, n_labels, for_rf, pct);

    FREE(i_list);
    FREE(f_list);
}

typedef struct {
    int feature;
    scalar cut_point, score;
} t_BestSplitInfo;

void OptionMLDecisionTreeBuild(const t_Parameters *parm, struct omldt_node *tree, scalar **data, int **labels, int instances, int *i_list, int features, int *f_list, int classes, int depth,
                               int **index, int *n_labels, bool for_rf, bool pct)
{
    t_BestSplitInfo *cutpoint;
    bool improves;

    //  No minimum samples or maximum depth exceeded?
    if (instances < parm->dt_min_samples_split || (parm->dt_max_depth && depth >= parm->dt_max_depth)) {
        // Leaf node
        tree->cut_point = 0.0;
        tree->feat = -1;
        tree->leaf = true;
        tree->depth = depth;
        tree->n_options = 0;

        // Get label frequency
        MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->probabilities, NULL);

        tree->left = tree->right = tree->options = NULL;
        return;
    }

    MALLOC(cutpoint, t_BestSplitInfo, features, "OptionMLDecisionTreeBuild: cutpoint");

    // If random forest get random subset of features
    int efeatures = features;
    if (for_rf) {
        efeatures = parm->rf_percent > 0 ? parm->rf_percent*features : ceil(sqrt((scalar) features));

        RandomizeVector(f_list, features);
    }

    // Evaluate every remaining feature
    for (int i = 0; i < efeatures; i++) {
        cutpoint[i].feature = f_list[i];
        MLGetBestCupPoint(data, labels, instances, i_list, f_list[i], &(cutpoint[i].cut_point), &(cutpoint[i].score), classes, index, n_labels, pct, NULL, false);
    }

    // Test if best improves
    int best_index = 0;
    for (int i = 1; i < efeatures; i++)
        if (cutpoint[i].score > cutpoint[best_index].score)
            best_index = i;

    // If best improves create children and follow up, else return
    // Improve criterion depends on measure
    improves = cutpoint[best_index].score > 0.0;

    // Check if options
    tree->n_options = 0;
    if (improves && depth <= parm->optiondt_l) {
        // Sort scores
        qsort(cutpoint, efeatures, sizeof(t_BestSplitInfo), CompareSplitInfo);

        // New best is first
        best_index = 0;

        int opts = 0;
        while (cutpoint[opts+1].score/cutpoint[0].score >= 1.0 - parm->optiondt_epsilon && opts < parm->optiondt_O && opts < efeatures-1)
            opts++;

        tree->n_options = opts;
        SHOWPROGRESS(parm->show_progress > 0 && opts, "Added %d OPTION nodes at level %d\n", opts, depth);
    }

    if (improves) {
        // Normal node or option node
        if (tree->n_options) {
            MALLOC(tree->options, struct omldt_node, tree->n_options, "OptionMLDecisionTreeBuild: tree->options");
            tree->depth = depth;
            tree->feat = -1;
            tree->cut_point = 0;
            MALLOC(tree->probabilities, scalar, classes, "OptionMLDecisionTreeBuild: tree->probabilities");
            tree->leaf = false;
            for (int o = 0; o < tree->n_options; o++) {
                tree->options[o].cut_point = cutpoint[o].cut_point;
                tree->options[o].feat = cutpoint[o].feature;
                MALLOC(tree->options[o].probabilities, scalar, classes, "OptionMLDecisionTreeBuild: tree->options[o].probabilities");
                MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->options[o].probabilities, NULL);

                tree->options[o].leaf = false;
                tree->options[o].depth = depth;
                tree->options[o].score = cutpoint[o].score;
                tree->options[o].n_options = 0;
                tree->options[o].options = NULL;

                MALLOC(tree->options[o].left, struct omldt_node, 1, "OptionMLDecisionTreeBuild: tree->options[o].left");
                MALLOC(tree->options[o].left->probabilities, scalar, classes, "OptionMLDecisionTreeBuild: tree->options[o].left->probabilities");
                tree->options[o].left->left = tree->options[o].left->right = NULL;
                tree->options[o].left->n_options = 0;

                int *new_i_list, is;
                MALLOC(new_i_list, int, instances, "OptionMLDecisionTreeBuild: new_i_list");

                // Instances for left node
                is = 0;
                if (index) {
                    for (int i = 0; i < instances; i++)
                        if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->options[o].feat) <= tree->options[o].cut_point) {
                            new_i_list[is] = i_list[i];
                            is++;
                        }
                } else {
                    for (int i = 0; i < instances; i++)
                        if (data[i_list[i]][tree->options[o].feat] <= tree->options[o].cut_point) {
                            new_i_list[is] = i_list[i];
                            is++;
                        }
                }

                if (is)
                    OptionMLDecisionTreeBuild(parm, tree->options[o].left, data, labels, is, new_i_list, features, f_list, classes, depth+1, index, n_labels, for_rf, pct);
                else {    // If no instances treat as leaf node
                    // Leaf node
                    tree->options[o].left->cut_point = 0.0;
                    tree->options[o].left->feat = -1;
                    tree->options[o].left->leaf = true;
                    tree->options[o].left->depth = depth;
                    tree->options[o].left->n_options = 0;
                    tree->options[o].left->options = NULL;

                    // Get label frequency
                    MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->options[o].left->probabilities, NULL);

                    tree->options[o].left->left = tree->options[o].left->right = NULL;
                }

                // Instances for right node
                is = 0;
                if (index) {
                    for (int i = 0; i < instances; i++) {
                        if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->options[o].feat) > tree->options[o].cut_point) {
                            new_i_list[is] = i_list[i];
                            is++;
                        }
                    }
                } else {
                    for (int i = 0; i < instances; i++)
                        if (data[i_list[i]][tree->options[o].feat] > tree->options[o].cut_point) {
                            new_i_list[is] = i_list[i];
                            is++;
                        }
                }
                MALLOC(tree->options[o].right, struct omldt_node, 1, "OptionMLDecisionTreeBuild: tree->options[o].right");
                MALLOC(tree->options[o].right->probabilities, scalar, classes, "OptionMLDecisionTreeBuild: tree->options[o].right->probabilities");
                tree->options[o].right->left = tree->options[o].right->right = NULL;
                tree->options[o].right->n_options = 0;

                if (is)
                    OptionMLDecisionTreeBuild(parm, tree->options[o].right, data, labels, is, new_i_list, features, f_list, classes, depth+1, index, n_labels, for_rf, pct);
                else {    // If no instances treat as leaf node
                    // Leaf node
                    tree->options[o].right->cut_point = 0.0;
                    tree->options[o].right->feat = -1;
                    tree->options[o].right->leaf = true;
                    tree->options[o].right->depth = depth;
                    tree->options[o].right->n_options = 0;
                    tree->options[o].right->options = NULL;

                    // Get label frequency
                    MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->options[o].right->probabilities, NULL);

                    tree->options[o].right->left = tree->options[o].right->right = NULL;
                }

                FREE(new_i_list);
            }
        } else {
            SHOWPROGRESS(parm->show_progress > 0, "Added node at level %d\n", depth);

            tree->cut_point = cutpoint[best_index].cut_point;
            tree->feat = cutpoint[best_index].feature;
            MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->probabilities, NULL);

            tree->leaf = false;
            tree->depth = depth;
            tree->score = cutpoint[best_index].score;
            tree->n_options = 0;
            tree->options = NULL;

            MALLOC(tree->left, struct omldt_node, 1, "OptionMLDecisionTreeBuild: tree->left");
            MALLOC(tree->left->probabilities, scalar, classes, "OptionMLDecisionTreeBuild: tree->left->probabilities");
            tree->left->left = tree->left->right = NULL;

            int *new_i_list, is;
            MALLOC(new_i_list, int, instances, "OptionMLDecisionTreeBuild: new_i_list");

            // Instances for left node
            is = 0;
            if (index) {
                for (int i = 0; i < instances; i++)
                    if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->feat) <= tree->cut_point) {
                        new_i_list[is] = i_list[i];
                        is++;
                    }
            } else {
                for (int i = 0; i < instances; i++)
                    if (data[i_list[i]][tree->feat] <= tree->cut_point) {
                        new_i_list[is] = i_list[i];
                        is++;
                    }
            }

            if (is)
                OptionMLDecisionTreeBuild(parm, tree->left, data, labels, is, new_i_list, features, f_list, classes, depth+1, index, n_labels, for_rf, pct);
            else {    // If no instances treat as leaf node
                // Leaf node
                tree->left->cut_point = 0.0;
                tree->left->feat = -1;
                tree->left->leaf = true;
                tree->left->depth = depth;
                tree->left->n_options = 0;
                tree->left->options = NULL;
                tree->left = NULL;

                // Get label frequency
                MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->left->probabilities, NULL);

                tree->left->left = tree->left->right = NULL;
            }

            // Instances for right node
            is = 0;
            if (index) {
                for (int i = 0; i < instances; i++) {
                    if (SparseFeature(data[i_list[i]], index[i_list[i]], tree->feat) > tree->cut_point) {
                        new_i_list[is] = i_list[i];
                        is++;
                    }
                }
            } else {
                for (int i = 0; i < instances; i++)
                    if (data[i_list[i]][tree->feat] > tree->cut_point) {
                        new_i_list[is] = i_list[i];
                        is++;
                    }
            }
            MALLOC(tree->right, struct omldt_node, 1, "OptionMLDecisionTreeBuild: tree->right");
            MALLOC(tree->right->probabilities, scalar, classes, "OptionMLDecisionTreeBuild: tree->right->probabilities");
            tree->right->left = tree->right->right = NULL;

            if (is)
                OptionMLDecisionTreeBuild(parm, tree->right, data, labels, is, new_i_list, features, f_list, classes, depth+1, index, n_labels, for_rf, pct);
            else {    // If no instances treat as leaf node
                // Leaf node
                tree->right->cut_point = 0.0;
                tree->right->feat = -1;
                tree->right->leaf = true;
                tree->right->depth = depth;
                tree->right->n_options = 0;
                tree->right->options = NULL;

                // Get label frequency
                MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->right->probabilities, NULL);

                tree->right->left = tree->right->right = NULL;
            }

            FREE(new_i_list);
        }
    } else {
        if (tree->n_options)
            OnErrorExit(ERR_UNSPECTED_RESULT, "Should not be here");

        // Leaf node
        tree->cut_point = 0.0;
        tree->feat = -1;
        tree->leaf = true;
        tree->depth = depth;

        // Get label frequency
        MLGetLabelsFrequency(labels, n_labels, i_list, instances, classes, tree->probabilities, NULL);

        tree->left = tree->right = tree->options = NULL;
    }

    FREE(cutpoint);
}

void FreeOptionMLDecisionTree(t_OMLDTree *tree)
{
    FreeOptionMLDecisionTreeNodes(tree->root);
}

void FreeOptionMLDecisionTreeNodes(struct omldt_node *tree)
{
    if (tree->left) {
        FreeOptionMLDecisionTreeNodes(tree->left);
        FREE(tree->left);
    }
    if (tree->right) {
        FreeOptionMLDecisionTreeNodes(tree->right);
        FREE(tree->right);
    }

    if (tree->n_options) {
        for (int i = 0; i < tree->n_options; i++)
            FreeOptionMLDecisionTreeNodes(&(tree->options[i]));
        FREE(tree->options);
    }

    if (tree)
        FREE(tree->probabilities);
}

int OptionMLDTCountNodes(t_OMLDTree *tree)
{
    int nodes = 0;

    OptionMLDTCountAllNodes(tree->root, &nodes);

    return nodes;
}

void OptionMLDTCountAllNodes(struct omldt_node *tree, int *nodes)
{
    if (tree) {
        (*nodes)++;
        OptionMLDTCountAllNodes(tree->left, nodes);
        OptionMLDTCountAllNodes(tree->right, nodes);
        if (tree->n_options)
            for (int i = 0; i < tree->n_options; i++)
                OptionMLDTCountAllNodes(&(tree->options[i]), nodes);
    }
}

// Auxiliary function for qsort
int CompareSplitInfo(const void *x, const void *y)
{
    scalar diff;

    if ((diff = ((t_BestSplitInfo *) y)->score - ((t_BestSplitInfo *) x)->score) > 0)
        return +1;
    else {
        if (diff < 0)
            return -1;
        else
            return 0;
    }
}
