/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "is.h"

// Removes/adds labels randomly and keep them removed if performance is not worse
void PartialSelectionMutation(const t_Parameters *parm, int **individual, int **labels, int *n_labels, int classes, int instances, cache_t **neighbors_cache, t_HashTable *ht, int **index,
                              int inputs, scalar **data, scalar *fitness, t_MLClassifier method, void *classifier, int *chain, int k, int *ks, int n_clusters)
{
    scalar performance, new_performance, old_fitness;
    int *copy;

    MALLOC(copy, int, instances, "PartialSelectionMutation: copy");

    // Current performance
    *fitness = old_fitness = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, ht, NULL, index, &performance, individual,
                             NULL, method, classifier, NULL, 0, chain, k, ks, n_clusters, NULL, NULL, instances);

    for (int i = 0; i < classes; i++) {
        // Copy previous values
        for (int j = 0; j < instances; j++) {
            copy[j] = individual[i][j];
            if (RANDOM(0.0, 1.0) < parm->bit_mutation) {
                individual[i][j] = !individual[i][j];
                if (individual[i][j] && parm->three_way_is && RANDOM(0.0, 1.0) < parm->three_way_minus1_prob)
                    individual[i][j] = -1;
            }
        }

        EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, ht, NULL, index, &new_performance, individual, NULL, method,
                                  classifier, NULL, 0, chain, k, ks, n_clusters, NULL, NULL, instances);

        // Keep?
        if (new_performance >= performance)
            performance = new_performance;
        else
            for (int j = 0; j < instances; j++)
                individual[i][j] = copy[j];
    }

    *fitness = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, ht, NULL, index, &performance, individual, NULL, method,
                                         classifier, NULL, 0, chain, k, ks, n_clusters, NULL, NULL, instances);

    SHOWPROGRESS(parm->show_progress > 1, "Partial selection mutation: %lf -> %lf %s\n", old_fitness, *fitness, *fitness > old_fitness ? "(*)" : "");

    FREE(copy);
}

void PartialInstanceSelectionOneLabelOptimization(const t_Parameters *parm, int **individual, int instances, scalar **data, int inputs, int classes, int **labels, int *n_labels, int **index,
        cache_t **neighbors_cache, t_MLClassifier method, void *classifier, int *chain, int k, int *ks, int n_clusters)
{
    int count[] = {0, 0}, n_is_algorithms = 15;
    scalar performance;
    extern char *isalgorithm[];
    t_ISAlgorithm is_algorithms[] = {DROP2, DROP3, ICF, MSS, RNN, IB3, CNN, CNNTL, OSS, Undersampling, GCNN, RMHC, HMNEI, RNGE, MCS};

    t_ISAlgorithm isalg = parm->random_isalgorithm ? is_algorithms[(int) RANDOM(0, n_is_algorithms)] : parm->isalgorithm;

    // Current performance
    scalar old_fitness = 0.0;
    if (parm->show_progress > 1)
        old_fitness = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, NULL, NULL, index, &performance, individual, NULL,
                                                method, classifier, NULL, 0, chain, k, ks, n_clusters, NULL, NULL, instances);

    // Construct ONE LABEL LOCAL data
    int *l_labels, l_instances = 0, *S, **l_index = NULL;
    scalar **l_data;

    MALLOC(l_labels, int, instances, "PartialInstanceSelectionOneLabelOptimization: l_labels");
    MALLOC(S, int, instances, "PartialInstanceSelectionOneLabelOptimization: S");
    MALLOC(l_data, scalar *, instances, "PartialInstanceSelectionOneLabelOptimization: l_data");
    if (index)
        MALLOC(l_index, int *, instances, "PartialInstanceSelectionOneLabelOptimization: l_index");

    // One label data
    int rlabel = RANDOM(0, classes);

    for (int i = 0; i < instances; i++)
        if (individual[rlabel][i]) {
            // Check if in list
            l_labels[l_instances] = LabelInList(rlabel, labels[i], n_labels[i]) ? 1 : 0;

            // False flag?
            if (individual[rlabel][i] == -1)
                l_labels[l_instances] = !l_labels[l_instances];

            // Data
            l_data[l_instances] = data[i];
            if (index)
                l_index[l_instances] = index[i];

            count[l_labels[l_instances]]++;
            l_instances++;
        }

    // Perform instance selection in the corresponding label
    if (count[0] && count[1]) {  // If a class is empty do nothing
        int minority = (count[0] < count[1] ? 0 : 1);

        ((t_Parameters *)parm)->k = 1;
        InstanceSelectionAlgorithm(parm, isalg, &S, &l_data, &l_instances, inputs, &l_labels, 2, l_index, NULL, NULL, minority, NULL, 0.0, NULL, NULL, NULL, NULL, true, NULL, false, NULL);

        // Update individual with performed selection
        int j = 0;
        for (int i = 0; i < instances; i++)
            if (individual[rlabel][i]) {
                if (!S[j])
                    individual[rlabel][i] = 0;
                j++;
            }

        if (j != count[0]+count[1])
            OnErrorExit(ERR_UNSPECTED_RESULT, "Selected instances do not match");
    }

    // Current performance
    scalar new_fitness = 0.0;
    if (parm->show_progress > 1)
        new_fitness = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, NULL, NULL, index, &performance, individual,
                                                NULL, method, classifier, NULL, 0, chain, k, ks, n_clusters, NULL, NULL, instances);
    SHOWPROGRESS(parm->show_progress > 1, "Partial selection single-label instance selection %s mutation: %lf -> %lf %s\n", isalgorithm[isalg], old_fitness, new_fitness,
                 old_fitness < new_fitness ? "(*)" : "");

    FREE(l_labels);
    FREE(l_data);
    FREE(S);
    if (index)
        FREE(l_index);
}
