/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "ml.h"
#include "macros.h"
#include "common.h"

// Perform a data partition

void DataPartitionML(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_subsets, scalar ****subset_data, int **subset_instances, int ****subset_labels,
                     int ***subset_list, int **index, int ****subset_index, int subset_size, int *n_labels, int ***subset_n_labels, int inputs)
{
    int *list, *partition = NULL, *count = NULL, previous_n_subsets = 0;

    // Only one stratum? Do nothing
    if (*n_subsets == 1 || parm->data_partition_method == NoDataPartition) {
        for (int j = 0; j < instances; j++) {
            (*subset_data)[0][j] = data[j];
            (*subset_labels)[0][j] = labels[j];
            (*subset_n_labels)[0][j] = n_labels[j];
            (*subset_list)[0][j] = j;

            // Index?
            if (index)
                (*subset_index)[0][j] = index[j];
        }

        (*subset_instances)[0] = instances;

        return;
    }

    switch (parm->data_partition_method) {
        case RandomDataPartition:
            // Initialize instance count per stratum
            for (int i = 0; i < *n_subsets; i++)
                (*subset_instances)[i] = 0;

            int size, c;

            MALLOC(list, int, instances, "RandomPartition: list");

            // Random partition
            for (int i = 0; i < instances; i++)
                list[i] = i;
            RandomizeVector(list, instances);

            size = instances/(*n_subsets);
            if (size > subset_size)
                OnErrorExit(ERR_UNSPECTED_RESULT, "Stratum size exceeded");
            c = 0;
            for (int i = 0; i < *n_subsets; i++) {
                for (int j = 0; j < size; j++) {
                    (*subset_data)[i][j] = data[list[c]];
                    (*subset_labels)[i][j] = labels[list[c]];
                    (*subset_n_labels)[i][j] = n_labels[list[c]];
                    (*subset_list)[i][j] = list[c];

                    // Index?
                    if (index)
                        (*subset_index)[i][j] = index[list[c]];

                    c++;
                }
                (*subset_instances)[i] = size;
            }

            // Assign remaining randomly
            for (; c < instances; c++) {
                int subset;

                do {
                    subset = RANDOM(0, *n_subsets);
                } while ((*subset_instances)[subset] == subset_size);

                (*subset_data)[subset][(*subset_instances)[subset]] = data[list[c]];
                (*subset_labels)[subset][(*subset_instances)[subset]] = labels[list[c]];
                (*subset_n_labels)[subset][(*subset_instances)[subset]] = n_labels[list[c]];
                (*subset_list)[subset][(*subset_instances)[subset]] = list[c];

                // Index?
                if (index)
                    (*subset_index)[subset][(*subset_instances)[subset]] = index[list[c]];

                (*subset_instances)[subset]++;
            }

            FREE(list);

            break;

        case Clustering:
            MALLOC(partition, int, instances, "DataPartition: partition");
            MALLOC(count, int, *n_subsets, "DataPartition: count");

            previous_n_subsets = *n_subsets;
            DataDivisionkMeans(parm, partition, &count, data, instances, &*n_subsets, inputs, parm->kmeans_max_iterations, index, subset_size);

            // Resize if different for previous
            if (previous_n_subsets != *n_subsets) {
                // Free?
                for (int i = *n_subsets; i < previous_n_subsets; i++) {
                    FREE((*subset_labels)[i]);
                    FREE((*subset_n_labels)[i]);
                    FREE((*subset_list)[i]);
                    FREE((*subset_data)[i]);
                    if (index)
                        FREE((*subset_index)[i]);
                }

                REALLOC(*subset_instances, int, *n_subsets, "DataPartition: *subset_instances");
                REALLOC(*subset_labels, int **, *n_subsets, "DataPartition: *subset_labels");
                REALLOC(*subset_n_labels, int *, *n_subsets, "DataPartition: *subset_n_labels");
                REALLOC(*subset_list, int *, *n_subsets, "DataPartition: *subset_list");
                REALLOC(*subset_data, scalar **, *n_subsets, "DataPartition: *subset_data");
                if (index)
                    REALLOC(*subset_index, int **, *n_subsets, "DataPartition: *subset_index");


                // New?
                for (int i = previous_n_subsets; i < *n_subsets; i++) {
                    MALLOC((*subset_labels)[i], int *, subset_size, "DataPartition: (*subset_labels)[i]");
                    MALLOC((*subset_n_labels)[i], int, subset_size, "DataPartition: (*subset_n_labels)[i]");
                    MALLOC((*subset_list)[i], int, subset_size, "DataPartition: (*subset_list)[i]");
                    MALLOC((*subset_data)[i], scalar *, subset_size, "DataPartition: (*subset_data)[i]");
                    if (index) {
                        MALLOC((*subset_index)[i], int *, subset_size, "DataPartition: (*subset_index)[i]");
                    } else
                        (*subset_index)[i] = NULL;
                }
            }

            // Initialize instance count per stratum
            for (int i = 0; i < *n_subsets; i++)
                (*subset_instances)[i] = 0;

            for (int i = 0; i < instances; i++) {
                int next = partition[i];

                // If stratum already full assign to another one randomly
                while ((*subset_instances)[next] >= subset_size)
                    next = RANDOM(0, *n_subsets);

                (*subset_data)[next][(*subset_instances)[next]] = data[i];
                (*subset_labels)[next][(*subset_instances)[next]] = labels[i];
                (*subset_n_labels)[next][(*subset_instances)[next]] = n_labels[i];
                (*subset_list)[next][(*subset_instances)[next]] = i;

                // Index?
                if (index)
                    (*subset_index)[next][(*subset_instances)[next]] = index[i];

                (*subset_instances)[next]++;
            }

            FREE(partition);
            FREE(count);
            break;

        case NoDataPartition:
            OnErrorExit(ERR_UNSPECIFIED, "Should not be here");
            break;
    }
}
