/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place-Suite 330, Boston, MA  02111-1307, USA.               *
 ***************************************************************************/
#include "typex.h"
#include "ml.h"

// Evaluates the performance of a classifier using a weighting voting of all ML metrics
// Train the classifier which hyper-parameters are already fixed

#undef SHOW_INFO_PER_LABEL

scalar EvaluateClassifierPerformance(const t_Parameters *parm, t_MLClassifier method, t_MLLearner *learner, int *selection, int *partition, cache_t **neighbors_cache, int instances,
                                     int **labels, int *n_labels, int classes, scalar **data, int **index, int inputs, int **sselection, bool trained, int *chain, int k, int *ks, int n_clusters,
                                     int *subspace, t_MLEnsemble *ensemble)
{
    scalar performance;

    if (parm->cv_folds && trained) {
        int *lpartition = NULL, *count = NULL;

        // Can be given a certain prearranged partition
        if (partition)
            lpartition = partition;
        else {
            MALLOC(count, int, parm->cv_folds, "EvolveKPerLabel: count");
            MALLOC(lpartition, int, instances, "EvolveKPerLabel: partition");

            RandomPartition(lpartition, count, instances, parm->cv_folds);
        }

        performance = EvaluateNFoldsCrossValidation(parm, parm->cv_folds, lpartition, selection, data, instances, neighbors_cache, index, n_labels, labels, classes, inputs, learner, method);

        if (!partition) {
            FREE(count);
            FREE(lpartition);
        }
    } else {
        // Classify testing set
        scalar **f  = NULL;
        int **h = NULL, *n_h = NULL;
        bool classification, ranks;
        int *lchain = NULL;

        AllocateScalarMatrix(&f, instances, classes);
        AllocateIntegerMatrix(&h, instances, classes);
        MALLOC(n_h, int, instances, "main: n_h");


        if (method == CC) {
            if (!chain) {
                CALLOC(lchain, int, classes, "EvaluateIndividualFitness: chain");
            } else
                lchain = chain;
        }

        if (!trained)
            TrainMultiLabelClassifier(parm, method, data, instances, labels, n_labels, classes, inputs, selection, lchain, index, neighbors_cache, sselection, learner, k, ks, NULL, n_clusters,
                subspace, ensemble);
        ClassifyDataset(parm, method, data, h, f, &classification, &ranks, data, instances, labels, n_labels, classes, n_h, inputs, index, index, instances, selection, neighbors_cache, sselection,
                        learner, subspace, ensemble);

        if (method == CC && !chain)
            FREE(lchain);

        // METRICS
        if (!parm->combine_metrics_gmean) {
            performance = parm->eb_subset_accuracy_w > ZERO ? parm->eb_subset_accuracy_w * SubsetAccuracy(labels, n_labels, h, n_h, instances) : 0.0;
            performance += parm->eb_hamming_loss_w  > ZERO ? parm->eb_hamming_loss_w * (1.0 - HammingLoss(labels, n_labels, h, n_h, instances, classes)) : 0.0;
            performance += parm->eb_accuracy_w > ZERO ? parm->eb_accuracy_w * ExampleBasedAccuracy(labels, n_labels, h, n_h, instances) : 0.0;
            performance += parm->eb_precision_w > ZERO ? parm->eb_precision_w * ExampleBasedPrecision(labels, n_labels, h, n_h, instances) : 0.0;
            performance += parm->eb_recall_w > ZERO ? parm->eb_recall_w * ExampleBasedRecall(labels, n_labels, h, n_h, instances) : 0.0;
            performance += parm->eb_f1_w > ZERO ? parm->eb_f1_w * ExampleBasedF1(labels, n_labels, h, n_h, instances, -1.0, -1.0) : 0.0;

            if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
                parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {
                scalar *TP, *FP, *TN, *FN;

                ALLOCA(TP, scalar, classes, "EvaluateIndividualFitness: TP");
                ALLOCA(FP, scalar, classes, "EvaluateIndividualFitness: FP");
                ALLOCA(TN, scalar, classes, "EvaluateIndividualFitness: TN");
                ALLOCA(FN, scalar, classes, "EvaluateIndividualFitness: FN");

                BinaryPerformancePerLabel(TP, FP, TN, FN, labels, n_labels, h, n_h, instances, classes);

                performance += parm->lb_accuracy_macro_w > ZERO ? parm->lb_accuracy_macro_w * AccuracyMacro(TP, TN, classes, instances) : 0.0;
                performance += parm->lb_precision_macro_w > ZERO ? parm->lb_precision_macro_w * PrecisionMacro(TP, FP, classes) : 0.0;
                performance += parm->lb_recall_macro_w > ZERO ? parm->lb_recall_macro_w * RecallMacro(TP, FN, classes) : 0.0;
                performance += parm->lb_f1_macro_w > ZERO ? parm->lb_f1_macro_w * F1Macro(TP, FN, FP, classes) : 0.0;
                performance += parm->lb_accuracy_micro_w > ZERO ? parm->lb_accuracy_micro_w * AccuracyMicro(TP, TN, FP, FN, instances, classes) : 0.0;
                performance += parm->lb_precision_micro_w > ZERO ? parm->lb_precision_micro_w * PrecisionMicro(TP, FP, classes) : 0.0;
                performance += parm->lb_recall_micro_w > ZERO ? parm->lb_recall_micro_w * RecallMicro(TP, FN, classes) : 0.0;
                performance += parm->lb_f1_micro_w > ZERO ?  parm->lb_f1_micro_w * F1Micro(TP, FN, FP, classes) : 0.0;
            }

            performance += parm->eb_f1_alt_w > ZERO ? parm->eb_f1_alt_w * F1Instance(labels, n_labels, h, n_h, instances) : 0.0;
            performance += parm->eb_one_error_w > ZERO ? parm->eb_one_error_w * (1.0 - OneError(labels, n_labels, f, instances, classes)) : 0.0;
            performance += parm->eb_ranking_loss_w > ZERO ? parm->eb_ranking_loss_w * (1.0 - RankingLoss(labels, n_labels, f, instances, classes)) : 0.0;
            performance += parm->eb_average_precision_w > ZERO ? parm->eb_average_precision_w * AveragePrecision(labels, n_labels, f, instances, classes) : 0.0;
            performance += parm->lb_auc_macro_w > ZERO ? parm->lb_auc_macro_w * AUCMacro(labels, n_labels, f, instances, classes) : 0.0;
            performance += parm->lb_auc_micro_w > ZERO ? parm->lb_auc_micro_w * AUCMicro(labels, n_labels, f, instances, classes) : 0.0;
            performance += parm->eb_auc_w > ZERO ? parm->eb_auc_w * AUCInstance(labels, n_labels, f, instances, classes) : 0.0;

            if (parm->eb_coverage_w > ZERO) {
                scalar optimal;

                scalar coverage = Coverage(labels, n_labels, f, instances, classes, &optimal);
                performance += coverage > ZERO ? optimal/coverage : 0.0;
            }

            scalar diff = 0.0;
            performance += parm->n_predicted_ml_w > ZERO ? parm->n_predicted_ml_w * (diff = AverageNumberOfPredictedLabelsDifference(n_h, n_labels, instances)) : 0.0;
        } else {
            performance = parm->eb_subset_accuracy_w ? pow(SubsetAccuracy(labels, n_labels, h, n_h, instances), parm->eb_subset_accuracy_w) : 1.0;
            performance *= parm->eb_hamming_loss_w > ZERO ? pow(1.0 - HammingLoss(labels, n_labels, h, n_h, instances, classes), parm->eb_hamming_loss_w) : 1.0;
            performance *= parm->eb_accuracy_w > ZERO ? pow(ExampleBasedAccuracy(labels, n_labels, h, n_h, instances), parm->eb_accuracy_w) : 1.0;
            performance *= parm->eb_precision_w > ZERO ? pow(ExampleBasedPrecision(labels, n_labels, h, n_h, instances), parm->eb_precision_w) : 1.0;
            performance *= parm->eb_recall_w > ZERO ? pow(ExampleBasedRecall(labels, n_labels, h, n_h, instances), parm->eb_recall_w) : 1.0;
            performance *= parm->eb_f1_w > ZERO ? pow(ExampleBasedF1(labels, n_labels, h, n_h, instances, -1.0, -1.0), parm->eb_f1_w) : 1.0;

            if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
                parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {
                scalar *TP, *FP, *TN, *FN;

                ALLOCA(TP, scalar, classes, "EvaluateIndividualFitness: TP");
                ALLOCA(FP, scalar, classes, "EvaluateIndividualFitness: FP");
                ALLOCA(TN, scalar, classes, "EvaluateIndividualFitness: TN");
                ALLOCA(FN, scalar, classes, "EvaluateIndividualFitness: FN");

                BinaryPerformancePerLabel(TP, FP, TN, FN, labels, n_labels, h, n_h, instances, classes);

                performance *= parm->lb_accuracy_macro_w > ZERO ? pow(AccuracyMacro(TP, TN, classes, instances), parm->lb_accuracy_macro_w) : 1.0;
                performance *= parm->lb_precision_macro_w > ZERO ? pow(PrecisionMacro(TP, FP, classes),  parm->lb_precision_macro_w) : 1.0;
                performance *= parm->lb_recall_macro_w > ZERO ? pow(RecallMacro(TP, FN, classes), parm->lb_recall_macro_w) : 1.0;
                performance *= parm->lb_f1_macro_w > ZERO ? pow(F1Macro(TP, FN, FP, classes), parm->lb_f1_macro_w) : 1.0;
                performance *= parm->lb_accuracy_micro_w > ZERO ? pow(AccuracyMicro(TP, TN, FP, FN, instances, classes), parm->lb_accuracy_micro_w) : 1.0;
                performance *= parm->lb_precision_micro_w > ZERO ? pow(PrecisionMicro(TP, FP, classes), parm->lb_precision_micro_w) : 1.0;
                performance *= parm->lb_recall_micro_w > ZERO ? pow(RecallMicro(TP, FN, classes), parm->lb_recall_micro_w) : 1.0;
                performance *= parm->lb_f1_micro_w > ZERO ? pow(F1Micro(TP, FN, FP, classes), parm->lb_f1_micro_w) : 1.0;
            }

            performance *= parm->eb_f1_alt_w > ZERO ? pow(F1Instance(labels, n_labels, h, n_h, instances), parm->eb_f1_alt_w) : 1.0;
            performance *= parm->eb_one_error_w > ZERO ? pow(1.0 - OneError(labels, n_labels, f, instances, classes), parm->eb_one_error_w) : 1.0;
            performance *= parm->eb_ranking_loss_w > ZERO ? pow(1.0 - RankingLoss(labels, n_labels, f, instances, classes), parm->eb_ranking_loss_w) : 1.0;
            performance *= parm->eb_average_precision_w > ZERO ? pow(AveragePrecision(labels, n_labels, f, instances, classes), parm->eb_average_precision_w) : 1.0;
            performance *= parm->lb_auc_macro_w > ZERO ? pow(AUCMacro(labels, n_labels, f, instances, classes), parm->lb_auc_macro_w) : 1.0;
            performance *= parm->lb_auc_micro_w > ZERO ? pow(AUCMicro(labels, n_labels, f, instances, classes), parm->lb_auc_micro_w) : 1.0;
            performance *= parm->eb_auc_w > ZERO ?pow(AUCInstance(labels, n_labels, f, instances, classes),  parm->eb_auc_w) : 1.0;

            if (parm->eb_coverage_w > ZERO) {
                scalar optimal;

                scalar coverage = Coverage(labels, n_labels, f, instances, classes, &optimal);
                performance *= coverage > ZERO ? pow(optimal/coverage, parm->eb_coverage_w) : 1.0;
            }

            scalar diff = 0.0;
            performance *= parm->n_predicted_ml_w > ZERO ? parm->n_predicted_ml_w * (diff = AverageNumberOfPredictedLabelsDifference(n_h, n_labels, instances)) : 1.0;

            scalar root = parm->eb_subset_accuracy_w + parm->eb_hamming_loss_w + parm->eb_accuracy_w + parm->eb_precision_w + parm->eb_recall_w + parm->eb_f1_w + parm->lb_accuracy_macro_w +
                          parm->lb_precision_macro_w + parm->lb_recall_macro_w + parm->lb_f1_macro_w + parm->lb_accuracy_micro_w + parm->lb_precision_micro_w + parm->lb_recall_micro_w +
                          parm->lb_f1_micro_w + parm->eb_coverage_w + parm->n_predicted_ml_w;

            performance = pow(performance, root);
        }

        SHOWPROGRESS(parm->show_progress > 2, "Evaluated classifier performance: %6.4lf\n", performance);

#ifdef SHOW_INFO_PER_LABEL
        scalar *tp, *fp, *tn, *fn;

        ALLOCA(tp, scalar, classes, "EvaluateIndividualFitness: tp");
        ALLOCA(fp, scalar, classes, "EvaluateIndividualFitness: fp");
        ALLOCA(tn, scalar, classes, "EvaluateIndividualFitness: tn");
        ALLOCA(fn, scalar, classes, "EvaluateIndividualFitness: fn");

        BinaryPerformancePerLabel(tp, fp, tn, fn, labels, n_labels, h, n_h, instances, classes);
        printf("Label\tTP\tFP\tTN\tFN\n");
        for (int i = 0; i < classes; i++)
            printf("%3d\t%5d\t%5d\t%5d\t%5d\n", i+1, (int) fp[i], (int) fp[i], (int) tn[i], (int) fn[i]);
#endif

        FreeScalarMatrix(&f, instances);
        FreeIntegerMatrix(&h, instances);
        FREE(n_h);
    }

    if (isnan(performance) || isinf(performance))
        OnErrorExit(ERR_NUMERICAL, "Fitness invalid value (NaN or infinite)");

    return performance;
}
