/***************************************************************************
 *   Copyright (C) 2023 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>
#include <stdbool.h>
#include "typex.h"
#include "ml.h"
#include "maths.h"

void PartialEvolutionaryUndersampling(const t_Parameters *parm, int ***labels, int **n_labels, int classes, int *instances, int **prus, scalar ***data, int ***index)
{
    int ***ss, *list;
    scalar *fitness;

    // Initialize population
    Allocate3DIntegerMatrix(&ss, 2*parm->population_size, classes, *instances);
    MALLOC(fitness, scalar, 2*parm->population_size, "PartialEvolutionaryUndersampling: fitness");
    MALLOC(list, int, parm->population_size, "PartialEvolutionaryUndersampling: list");

    // Template to avoid removing minority class samples
    int **template;
    AllocateIntegerMatrix(&template, classes, *instances);
    GetMinorityTemplate(*labels, *n_labels, classes, *instances, template);

    for (int i = 0; i < parm->population_size; i++) {
        FillMatrix(NULL, ss[i], classes, *instances, 1.0);
        if (parm->pevus_random_init) {
            for (int j = 0; j < classes; j++)
                for (int k = 0; k < *instances; k++)
                    ss[i][j][k] = RANDOM(0.0, 1.0) < parm->chc_init_prob ? 1 : 0;
            OrIntegerMatrix(ss[i], template, classes, *instances);
        } else
            PartialRandomUndersampling(parm, (*labels), (*n_labels), classes, *instances, ss[i], parm->ml_rus_p, false);
        list[i] = i;
    }


    int minimum_distance = *instances/4;

    // Initial fitness value
    for (int i = 0; i < parm->population_size; i++)
        fitness[i] = UndersamplingFitness(ss[i], classes, *instances, (*labels), (*n_labels), parm->ml_rus_p);

    for (int g = 0; g < parm->generations; g++) {
        int actual_size = parm->population_size;

        // HUX crossover
        RandomizeVector(list, parm->population_size);

        for (int i = 0; i < parm->population_size; i += 2) {
            int father = list[i];
            int mother = list[i+1];
            bool at_least_one = false;

            for (int j = 0; j < classes; j++)
                if (HammingDistance(ss[mother][j], ss[father][j], *instances) > minimum_distance) {
                    at_least_one = true;
                    HuxCrossover(ss[mother][j], ss[father][j], ss[actual_size][j], ss[actual_size+1][j], *instances);
                }

            if (at_least_one) {
                // OR with template to always keep minority class labels
                OrIntegerMatrix(ss[actual_size], template, classes, *instances);
                OrIntegerMatrix(ss[actual_size+1], template, classes, *instances);

                actual_size += 2;
            }
        }

        // Fitness of new individuals
        for (int i = parm->population_size; i < actual_size; i++)
            fitness[i] = UndersamplingFitness(ss[i], classes, *instances, (*labels), (*n_labels), parm->ml_rus_p);

        // Reinitialize or elitism
        if (actual_size == parm->population_size)
            minimum_distance--;
        else
            for (int i = 0; i < parm->population_size; i++) {
                int order = OrderedSelection(fitness+i, 0, actual_size-i)+i;

                int **aux_pt = ss[order];
                ss[order] = ss[i];
                ss[i] = aux_pt;

                scalar aux_f = fitness[order];
                fitness[order] = fitness[i];
                fitness[i] = aux_f;
            }

        if (minimum_distance == 0) {
            // Reinitialize populations excepting best one
            int best = OrderedSelection(fitness, 0, parm->population_size);

            for (int i = 0; i < parm->population_size; i++)
                // Reseed changing 35%
                if (i != best) {
                    for (int k = 0; k < classes; k++)
                        for (int j = 0; j< *instances; j++)
                            ss[i][k][j] = RANDOM(0.0, 1.0) < 0.35 ? ss[best][k][j] : (ss[best][k][j]+1)%2;

                    fitness[i] = UndersamplingFitness(ss[i], classes, *instances, (*labels), (*n_labels), parm->ml_rus_p);
                }

            minimum_distance = *instances/4;
            SHOWPROGRESS(parm->show_progress > 0, "\nPopulation reinitialized at generation %d\n", g+1);
        }

        SHOWPROGRESS(parm->show_progress > 0, "PEVUS %*d g's best P(MEAN): %6.4lf(%6.4lf) New individuals: %*d\r", WIDTH(parm->generations), g+1, fitness[0],
                     AverageScalarVector(fitness, parm->population_size), WIDTH(parm->population_size), actual_size-parm->population_size);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // Best is first
    for (int i = 0; i < classes; i++)
        for (int j = 0; j < *instances; j++)
            prus[i][j] = ss[0][i][j];

    // Undersampling info
    fprintf(parm->out_file,"Majority class UNDER-sampling: ");
    for (int i = 0; i < classes; i++) {
        int relevant = 0, irrelevant = 0, count_r = 0;

        for (int j = 0; j < *instances; j++) {
            bool present = LabelInList(i, (*labels)[j], (*n_labels)[j]);

            if (present)
                count_r++;

            if (prus[i][j]) {
                if (present)
                    relevant++;
                else
                    irrelevant++;
            }
        }

        fprintf(parm->out_file, "(%d,%d) --> (%d,%d) ", count_r, *instances-count_r, relevant, irrelevant);
    }
    fprintf(parm->out_file, "\n");

    // Remove instances with no relevant label
    int original_instances = *instances;
    for (int i = 0; i < *instances; i++) {
        bool remove = true;

        if (!parm->pevus_remove_null)
            remove = EmptyColumn(prus, classes, i);
        else
            for (int j = 0; j < (*n_labels)[i] && remove; j++)
                if (prus[(*labels)[i][j]][i])
                    remove = false;

        if (remove) {
            (*instances)--; // One less instance

            // Remove
            FREE((*labels)[i]);
            FREE((*data)[i]);

            (*data)[i] = (*data)[*instances];
            (*labels)[i] = (*labels)[*instances];
            (*n_labels)[i] = (*n_labels)[*instances];

            if (*index) {
                FREE((*index)[i]);
                (*index)[i] = (*index)[*instances];
            }

            for (int j = 0; j < classes; j++)
                prus[j][i] = prus[j][*instances];

            i--;            // Revisit i
        }
    }

    // Undersampling info
    fprintf(parm->out_file,"Majority class UNDER-sampling AFTER removing instances: ");
    for (int i = 0; i < classes; i++) {
        int relevant = 0, irrelevant = 0, count_r = 0;

        for (int j = 0; j < *instances; j++) {
            bool present = LabelInList(i, (*labels)[j], (*n_labels)[j]);

            if (present)
                count_r++;

            if (prus[i][j]) {
                if (present)
                    relevant++;
                else
                    irrelevant++;
            }
        }

        fprintf(parm->out_file, "(%d,%d) --> (%d,%d) ", count_r, *instances-count_r, relevant, irrelevant);
    }
    fprintf(parm->out_file, "\n");

    // Reduction
    fprintf(parm->out_file, "@PEVUS instance reduction: %lf\n", ((scalar)(original_instances-*instances))/original_instances);

    Free3DIntegerMatrix(&ss, 2*parm->population_size, classes);
    FreeIntegerMatrix(&template, classes);
    FREE(fitness);
    FREE(list);
}

scalar UndersamplingFitness(int **ss, int classes, int instances, int **labels, int *n_labels, scalar ratio)
{
    scalar fitness, w[] = {0.5, 0.0, 0.5};

    // Get selected instances
    int selected = 0;
    for (int i = 0; i < instances; i++)
        if (!EmptyColumn(ss, classes, i))
            selected++;

    fitness = w[0]*(1.0 - ((scalar) selected)/instances);

    // Count instances that do not have relevant labels selected
    int all = 0;
    for (int i = 0; i < instances; i++) {
        bool none = true;
        for (int j = 0; j < n_labels[i] && none; j++)
            if (ss[labels[i][j]][i])
                none = false;

        if (none)
            all++;
    }

    fitness += w[1]*(((scalar) all)/instances);

    // Get IR
    scalar meanir = 0.0;
    for (int i = 0; i < classes; i++) {
        scalar relevant = 0.0, irrelevant = 0.0;
        for (int j = 0; j < instances; j++)
            if (ss[i][j]) {
                if (LabelInList(i, labels[j], n_labels[j]))
                    relevant++;
                else
                    irrelevant++;
            }

        if (relevant > ZERO && irrelevant > ZERO)  {
            scalar r = MAX(relevant, irrelevant)/MIN(relevant, irrelevant);
            meanir += r > ratio ? r : ratio;
        } else
            return 0.0;
    }

    fitness += w[2]*exp(-0.25*fabs(meanir/classes-ratio));

    return fitness;
}

void GetMinorityTemplate(int **labels, int *n_labels, int classes, int instances, int **template)
{
    int *maj_list, maj_count, *count_r, *count_i;
    bool *minority_relevant;

    MALLOC(maj_list, int, instances, "GetMinorityTemplate: maj_list");
    MALLOC(minority_relevant, bool, classes, "GetMinorityTemplate: minority_relevant");
    MALLOC(count_r, int, classes, "GetMinorityTemplate: count_r");
    MALLOC(count_i, int, classes, "GetMinorityTemplate: count_i");

    // Count labels to get minority for every label
    ZeroVector(NULL, count_r, classes);
    ZeroVector(NULL, count_i, classes);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < classes; j++) {
            if (LabelInList(j, labels[i], n_labels[i]))
                count_r[j]++;
            else
                count_i[j]++;
        }

    for (int i = 0; i < classes; i++)
        minority_relevant[i] = (count_r[i] < count_i[i]);

    FillMatrix(NULL, template, classes, instances, 1.0);

    for (int i = 0; i < classes; i++) {
        maj_count = 0;

        for (int j = 0; j < instances; j++)
            if (minority_relevant[i] == !LabelInList(i, labels[j], n_labels[j]))
                maj_list[maj_count++] = j;

        // Remove majority class samples from template
        for (int j = 0; j < maj_count; j++)
            template[i][maj_list[j]] = 0;
    }
}
