/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <omp.h>
#include <stdbool.h>
#include <stdio.h>
#include <math.h>
#include "macros.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

#define PI 3.14159265358979323846
#define GAUSSIAN(x, mu, sigma) (1.0/(sigma*sqrt(2.0*PI)))*exp(-pow(x-mu, 2.0)/(2.0*pow(sigma, 2.0)))

scalar EvaluateIndividualFitness(const t_Parameters *parm, int *individual, int ***subpops, int n_subpops, int instances, int **labels, int *n_labels, int classes, int inputs, scalar **data, int *s,
                                 cache_t **neighbors_cache, t_HashTable *ht, scalar *qs, int **index, scalar *performance, int **ss, scalar **weights, t_MLClassifier method, t_MLLearner *learner,
                                 t_Performance *perf, int generation, int *chain, int k, int *ks, int n_clusters, int *subspace, t_MLEnsemble *ensemble, int test_instances)
{
    char *tempnam(const char *dir, const char *pfx);

    int *selection = NULL, **sselection = NULL;
    scalar optimal, *TP, *FP, *TN, *FN, fitness = -1.0;
    bool classification, ranks;

    t_HashEntry new, *old;

    // Selection already given?
    if (!subspace) {
        if (s)
            selection = s;
        else if (ss)
            sselection = ss;
        else if (parm->ml_ins_sel_per_label) {
            MALLOC(sselection, int *, n_subpops, "EvaluateIndividualFitness: sselection");
            for (int i = 0; i < n_subpops; i++)
                sselection[i] = subpops[i][individual[i]];
        } else {
            // Get selection from individual
            MALLOC(selection, int, instances, "EvaluateIndividualFitness: selection");
            CreateSelectionFromSubpopulations(parm, individual, subpops, n_subpops, selection, instances, *qs, parm->ml_cc_weighted_qs ? weights : NULL);
        }
    }

    if (parm->ml_use_hash_table)
        MALLOC(new.key, char, (subspace ? inputs+1 : instances+1), "EvaluateIndividualFitness: new.key");

    if (parm->ml_use_hash_table) {
        // Check if already evaluated
        for (int i = 0; i < (subspace ? inputs : instances); i++)
            new.key[i] = selection[i] ? '1' : '0';
        new.key[subspace ? inputs : instances] = '\0';

#ifdef _OPENMP
        #pragma omp critical
#endif
        if ((old = HashSearchEntry(*ht, new.key)) != NULL)
            fitness = old->ok;
    }

    // Reduction
    scalar reduction = 0.0, per_label_reduction = 0.0, ms_reduction = 0.0;
    if (subspace) {
        for (int i = 0; i < inputs; i++)
            if (!subspace[i])
                reduction++;
        reduction /= inputs;
    } else {
        if (parm->ml_ins_sel_per_label) {
            for (int i = 0; i < instances; i++) {
                bool none = true;
                for (int j = 0; j < classes; j++)
                    if (sselection[j][i]) {
                        none = false;
                        per_label_reduction++;
                    }

                if (none)
                    reduction++;
            }

            reduction /= instances;
            per_label_reduction = 1.0 - per_label_reduction/(instances*classes);
        } else {
            for (int i = 0; i < test_instances; i++)
                if (!selection[i])
                    reduction++;
            reduction /= test_instances;
        }
    }

    // Oversampling ratio
    scalar oversampling = 0.0;
    if (parm->multiselection && fabs(parm->oversampling_ml_ms_w) > 0.0) {
        for (int i = 0; i < instances; i++)
            oversampling += selection[i];

        oversampling /= instances;
    }

    // Multi-selection specific reduction
    if (parm->multiselection && fabs(parm->reduction_ml_ms_w) > 0.0) {
        for (int i = 0; i < instances; i++)
            ms_reduction += selection[i];

        ms_reduction = 1.0 - ms_reduction/instances;
    }

    if (parm->gauss_reduction) {
        reduction = GAUSSIAN(reduction, 0.9, 0.2);
        per_label_reduction = GAUSSIAN(per_label_reduction, 0.9, 0.2);
        ms_reduction = GAUSSIAN(ms_reduction, 0.9, 0.2);
    }

    // Less subpopulations considered
    scalar negs = 0.0;
    if (parm->negatives_ml_w > ZERO) {
        for (int i = 0; i < n_subpops; i++)
            if (individual[i] < 0)
                negs++;
        negs /= n_subpops;
    }

    // Fitness already obtained?
    if (fitness >= 0.0) {
        if (!s)
            FREE(selection);
        FREE(new.key);

        *performance = fitness - parm->reduction_ml_w * reduction - parm->negatives_ml_w * negs - parm->reduction_perlabel_w * per_label_reduction;

        return fitness;
    }

    // If three way partial instance selection create local label sets
    int *local_n_labels = NULL, **local_labels = NULL;

    if (!parm->three_way_is) {
        local_n_labels = n_labels;
        local_labels = labels;
    } else {
        MALLOC(local_n_labels, int, instances, "EvaluateIndividualFitness: local_n_labels");
        AllocateIntegerMatrix(&local_labels, instances, classes);

        CheatLabels(labels, n_labels, local_labels, local_n_labels, instances, classes, sselection, false, parm->cheat_neg_to_pos_only);
    }

    // Classify testing set
    scalar **f  = NULL, metric = 0.0;
    int **h = NULL, *n_h = NULL;

    AllocateScalarMatrix(&f, instances, classes);
    AllocateIntegerMatrix(&h, instances, classes);
    MALLOC(n_h, int, instances, "main: n_h");

    // Test if evaluation must be carried out with external classifier
    if (parm->ml_external_eval_interval && ((generation+1) % parm->ml_external_eval_interval) == 0) {
        extern char *ml_ext_method[];
        SHOWPROGRESS(parm->show_progress > 1, "External evaluation of fitness with method %s at generation %d\n", ml_ext_method[parm->evolution_external_classifier], generation);
        char *file_template = tempnam(".", NULL);
        ClassifyExternalDataset(parm, parm->evolution_external_classifier, data, h, f, data, instances, local_labels, local_n_labels, classes, n_h, inputs, index, index, test_instances, selection,
                                &classification, &ranks, file_template, local_labels, local_n_labels, false, sselection, subspace);
    } else {
        int *lchain = NULL;
        t_Learner *learners = NULL;
        if (method == CC || method == BR) {
            if (method == CC) {
                if (!chain) {
                    CALLOC(lchain, int, classes, "EvaluateIndividualFitness: lchain");
                } else
                    lchain = chain;
            }
            CALLOC(learners, t_Learner, classes, "EvaluateIndividualFitness: learners");
        }

        TrainMultiLabelClassifier(parm, method, data, instances, local_labels, local_n_labels, classes, inputs, selection, lchain, index, neighbors_cache, sselection, learner, k, ks, NULL,
                                  n_clusters, subspace, ensemble);
        ClassifyDataset(parm, method, data, h, f, &classification, &ranks, data, instances, local_labels, local_n_labels, classes, n_h, inputs, index, index, test_instances, selection,
                        neighbors_cache, sselection, learner, subspace, ensemble);

        if (method == CC && !chain)
            FREE(lchain);

        FreeMLClassifier(parm, learner, method, classes, inputs, instances, ensemble);
    }

    // Get fitness
    fitness = parm->combine_metrics_gmean ? 1.0 : 0.0;

    // Performance METRICS
    scalar diff = 0.0;
    if (!parm->combine_metrics_gmean) {
        // Reduction
        fitness += parm->reduction_ml_w * reduction;
        fitness += parm->reduction_perlabel_w * per_label_reduction;
        fitness += parm->reduction_ml_ms_w * ms_reduction;

        // Oversampling
        fitness += parm->oversampling_ml_ms_w * oversampling;

        // Subpopulations used
        fitness += parm->negatives_ml_w * negs;

        fitness += parm->eb_subset_accuracy_w > ZERO ? parm->eb_subset_accuracy_w * SubsetAccuracy(local_labels, local_n_labels, h, n_h, test_instances)/
                   (perf && perf->eb_subset_accuracy > ZERO ? perf->eb_subset_accuracy : 1.0) : 0.0;
        fitness += parm->eb_hamming_loss_w  > ZERO ? parm->eb_hamming_loss_w *
                   (!perf ? (1.0 - HammingLoss(local_labels, local_n_labels, h, n_h, test_instances, classes)) :
                    perf->eb_hamming_loss/((metric = HammingLoss(local_labels, local_n_labels, h, n_h, test_instances, classes)) > ZERO ? metric : 1.0)) : 0.0;
        fitness += parm->eb_accuracy_w > ZERO ? parm->eb_accuracy_w * ExampleBasedAccuracy(local_labels, local_n_labels, h, n_h, test_instances)/(perf &&  perf->eb_accuracy > ZERO ? perf->eb_accuracy
:
                   1.0) : 0.0;
        fitness += parm->eb_precision_w > ZERO ? parm->eb_precision_w * ExampleBasedPrecision(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_precision > ZERO ?
perf->eb_precision
                   : 1.0) : 0.0;
        fitness += parm->eb_recall_w > ZERO ? parm->eb_recall_w * ExampleBasedRecall(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_recall > ZERO ? perf->eb_recall : 1.0) :
0.0;
        fitness += parm->eb_f1_w > ZERO ? parm->eb_f1_w * ExampleBasedF1(local_labels, local_n_labels, h, n_h, test_instances, -1.0, -1.0)/(perf && perf->eb_f1 > ZERO ? perf->eb_f1 : 1.0) : 0.0;

        if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
            parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {

            ALLOCA(TP, scalar, classes, "EvaluateIndividualFitness: TP");
            ALLOCA(FP, scalar, classes, "EvaluateIndividualFitness: FP");
            ALLOCA(TN, scalar, classes, "EvaluateIndividualFitness: TN");
            ALLOCA(FN, scalar, classes, "EvaluateIndividualFitness: FN");

            BinaryPerformancePerLabel(TP, FP, TN, FN, local_labels, local_n_labels, h, n_h, test_instances, classes);

            fitness += parm->lb_accuracy_macro_w > ZERO ? parm->lb_accuracy_macro_w * AccuracyMacro(TP, TN, classes, test_instances)/(perf && perf->lb_accuracy_macro > ZERO ? perf->lb_accuracy_macro :
                       1.0) : 0.0;
            fitness += parm->lb_precision_macro_w > ZERO ? parm->lb_precision_macro_w * PrecisionMacro(TP, FP, classes)/(perf && perf->lb_precision_macro > ZERO ? perf->lb_precision_macro :
                       1.0) : 0.0;
            fitness += parm->lb_recall_macro_w > ZERO ? parm->lb_recall_macro_w * RecallMacro(TP, FN, classes)/(perf && perf->lb_recall_macro > ZERO ? perf->lb_recall_macro : 1.0) : 0.0;
            fitness += parm->lb_f1_macro_w > ZERO ? parm->lb_f1_macro_w * F1Macro(TP, FN, FP, classes)/(perf && perf->lb_f1_macro > ZERO ? perf->lb_f1_macro : 1.0) : 0.0;
            fitness += parm->lb_accuracy_micro_w > ZERO ? parm->lb_accuracy_micro_w * AccuracyMicro(TP, TN, FP, FN, test_instances, classes)/(perf && perf->lb_f1_macro > ZERO ?
                       perf->lb_accuracy_micro : 1.0) : 0.0;
            fitness += parm->lb_precision_micro_w > ZERO ? parm->lb_precision_micro_w * PrecisionMicro(TP, FP, classes)/(perf && perf->lb_precision_micro > ZERO ? perf->lb_precision_micro :
                       1.0) : 0.0;
            fitness += parm->lb_recall_micro_w > ZERO ? parm->lb_recall_micro_w * RecallMicro(TP, FN, classes)/(perf && perf->lb_recall_micro > ZERO ? perf->lb_recall_micro : 1.0) : 0.0;
            fitness += parm->lb_f1_micro_w > ZERO ?  parm->lb_f1_micro_w * F1Micro(TP, FN, FP, classes)/(perf && perf->lb_f1_micro > ZERO ? perf->lb_f1_micro : 1.0) : 0.0;
        }

        fitness += parm->eb_f1_alt_w > ZERO ? parm->eb_f1_alt_w * F1Instance(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_f1_alt > ZERO ? perf->eb_f1_alt : 1.0) : 0.0;

        if (ranks) {
            fitness += parm->eb_one_error_w > ZERO ? parm->eb_one_error_w * (!perf ? 1.0 - OneError(local_labels, local_n_labels, f, test_instances, classes) :
                       perf->eb_one_error/((metric = OneError(local_labels, local_n_labels, f, test_instances, classes)) > ZERO ? metric : 1.0)) : 0.0;
            fitness += parm->eb_ranking_loss_w > ZERO ? parm->eb_ranking_loss_w *
                       (!perf ? (1.0 - RankingLoss(local_labels, local_n_labels, f, test_instances, classes)) :
                        perf->eb_ranking_loss/((metric = RankingLoss(local_labels, local_n_labels, f, test_instances, classes)) > ZERO ? metric : 1.0)) : 0.0;
            fitness += parm->eb_average_precision_w > ZERO ?
                       parm->eb_average_precision_w * AveragePrecision(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->eb_average_precision > ZERO ?
                               perf->eb_average_precision : 1.0) : 0.0;
            fitness += parm->lb_auc_macro_w > ZERO ?
                       parm->lb_auc_macro_w * AUCMacro(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->lb_auc_macro > ZERO ? perf->lb_auc_macro : 1.0) : 0.0;
            fitness += parm->lb_auc_micro_w > ZERO ?
                       parm->lb_auc_micro_w * AUCMicro(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->lb_auc_micro > ZERO ? perf->lb_auc_micro : 1.0) : 0.0;
            fitness += parm->eb_auc_w > ZERO ? parm->eb_auc_w * AUCInstance(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->eb_auc > ZERO ? perf->eb_auc : 1.0) : 0.0;

            if (parm->eb_coverage_w > ZERO) {
                scalar coverage = Coverage(local_labels, local_n_labels, f, test_instances, classes, &optimal);
                fitness += coverage > ZERO ? (!perf ? optimal/coverage : perf->eb_coverage/coverage) : 0.0;
            }
        }

        fitness += parm->n_predicted_ml_w > ZERO ? parm->n_predicted_ml_w * (diff = AverageNumberOfPredictedLabelsDifference(n_h, local_n_labels, test_instances/*, classes*/)) : 0.0;

        fitness += parm->eb_jaccard_coeff_w > ZERO ? parm->eb_jaccard_coeff_w * JaccardCoefficient(local_labels, local_n_labels, h, n_h, test_instances) : 0.0;

        // Classification performance
        *performance = fitness - parm->reduction_ml_w * reduction - parm->negatives_ml_w * negs - parm->n_predicted_ml_w * diff - parm->reduction_perlabel_w * per_label_reduction -
                       parm->reduction_ml_ms_w * ms_reduction;

        // k regularization term
        fitness += (method == MLkNN && parm->regularize_k_w > ZERO) ? parm->regularize_k_w * ((learner->mlknn.k)/parm->ml_max_k) : 0.0;
    } else {
        // Subpopulations used
        fitness *= pow(negs, parm->negatives_ml_w);

        fitness *= parm->eb_subset_accuracy_w ? pow(SubsetAccuracy(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_subset_accuracy > ZERO ? perf->eb_subset_accuracy : 1.0),
                   parm->eb_subset_accuracy_w) : 1.0;
        fitness *= parm->eb_hamming_loss_w > ZERO ?
                   pow(!perf ? 1.0 - HammingLoss(local_labels, local_n_labels, h, n_h, test_instances, classes) :
                       perf->eb_hamming_loss/((metric = HammingLoss(local_labels, local_n_labels, h, n_h, test_instances, classes)) > ZERO ? metric : 1.0), parm->eb_hamming_loss_w) : 1.0;
        fitness *= parm->eb_accuracy_w > ZERO ? pow(ExampleBasedAccuracy(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_accuracy > ZERO ? perf->eb_accuracy : 1.0),
                   parm->eb_accuracy_w) : 1.0;
        fitness *= parm->eb_precision_w > ZERO ? pow(ExampleBasedPrecision(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_precision > ZERO ? perf->eb_precision : 1.0),
                   parm->eb_precision_w) : 1.0;
        fitness *= parm->eb_recall_w > ZERO ? pow(ExampleBasedRecall(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_recall > ZERO ? perf->eb_recall : 1.0), parm->eb_recall_w)
:
                   1.0;
        fitness *= parm->eb_f1_w > ZERO ? pow(ExampleBasedF1(local_labels, local_n_labels, h, n_h, test_instances, -1.0, -1.0)/(perf && perf->eb_f1 > ZERO ? perf->eb_f1 : 1.0), parm->eb_f1_w) : 1.0;

        if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
            parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {

            ALLOCA(TP, scalar, classes, "EvaluateIndividualFitness: TP");
            ALLOCA(FP, scalar, classes, "EvaluateIndividualFitness: FP");
            ALLOCA(TN, scalar, classes, "EvaluateIndividualFitness: TN");
            ALLOCA(FN, scalar, classes, "EvaluateIndividualFitness: FN");

            BinaryPerformancePerLabel(TP, FP, TN, FN, local_labels, local_n_labels, h, n_h, test_instances, classes);

            fitness *= parm->lb_accuracy_macro_w > ZERO ? pow(AccuracyMacro(TP, TN, classes, test_instances)/(perf && perf->lb_accuracy_macro > ZERO ? perf->lb_accuracy_macro : 1.0),
                       parm->lb_accuracy_macro_w) : 1.0;
            fitness *= parm->lb_precision_macro_w > ZERO ? pow(PrecisionMacro(TP, FP, classes)/(perf && perf->lb_precision_macro > ZERO ? perf->lb_precision_macro : 1.0),  parm->lb_precision_macro_w)
                       : 1.0;
            fitness *= parm->lb_recall_macro_w > ZERO ? pow(RecallMacro(TP, FN, classes)/(perf && perf->lb_recall_macro > ZERO ? perf->lb_recall_macro : 1.0), parm->lb_recall_macro_w) : 1.0;
            fitness *= parm->lb_f1_macro_w > ZERO ? pow(F1Macro(TP, FN, FP, classes)/(perf && perf->lb_f1_macro > ZERO ? perf->lb_f1_macro : 1.0), parm->lb_f1_macro_w) : 1.0;
            fitness *= parm->lb_accuracy_micro_w > ZERO ? pow(AccuracyMicro(TP, TN, FP, FN, test_instances, classes)/(perf && perf->lb_accuracy_micro > ZERO ? perf->lb_accuracy_micro : 1.0),
                       parm->lb_accuracy_micro_w) : 1.0;
            fitness *= parm->lb_precision_micro_w > ZERO ? pow(PrecisionMicro(TP, FP, classes)/(perf && perf->lb_precision_micro > ZERO ? perf->lb_precision_micro : 1.0),
                       parm->lb_precision_micro_w) : 1.0;
            fitness *= parm->lb_recall_micro_w > ZERO ? pow(RecallMicro(TP, FN, classes)/(perf && perf->lb_recall_micro > ZERO ? perf->lb_recall_micro : 1.0), parm->lb_recall_micro_w) : 1.0;
            fitness *= parm->lb_f1_micro_w > ZERO ? pow(F1Micro(TP, FN, FP, classes)/(perf && perf->lb_f1_micro > ZERO ? perf->lb_f1_micro : 1.0), parm->lb_f1_micro_w) : 1.0;
        }

        fitness *= parm->eb_f1_alt_w > ZERO ? pow(F1Instance(local_labels, local_n_labels, h, n_h, test_instances)/(perf && perf->eb_f1_alt > ZERO ? perf->eb_f1_alt : 1.0), parm->eb_f1_alt_w) : 1.0;

        if (ranks) {
            fitness *= parm->eb_one_error_w > ZERO ? pow(!perf ? 1.0 - OneError(local_labels, local_n_labels, f, test_instances, classes)
                       : perf->eb_one_error/((metric = OneError(local_labels, local_n_labels, f, test_instances, classes)) > ZERO ? metric : 1.0), parm->eb_one_error_w) : 1.0;
            fitness *= parm->eb_ranking_loss_w > ZERO ?
                       pow(!perf ? 1.0 - RankingLoss(local_labels, local_n_labels, f, test_instances, classes) :
                           perf->eb_ranking_loss/((metric = RankingLoss(local_labels, local_n_labels, f, test_instances, classes)) > ZERO ? metric : 1.0), parm->eb_ranking_loss_w) : 1.0;
            fitness *= parm->eb_average_precision_w > ZERO ? pow(AveragePrecision(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->eb_average_precision > ZERO ?
                       perf->eb_average_precision : 1.0), parm->eb_average_precision_w) : 1.0;
            fitness *= parm->lb_auc_macro_w > ZERO ? pow(AUCMacro(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->lb_auc_macro > ZERO ? perf->lb_auc_macro : 1.0),
                       parm->lb_auc_macro_w) : 1.0;
            fitness *= parm->lb_auc_micro_w > ZERO ? pow(AUCMicro(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->lb_auc_micro > ZERO ?
                       perf->lb_auc_micro : 1.0), parm->lb_auc_micro_w) : 1.0;
            fitness *= parm->eb_auc_w > ZERO ?pow(AUCInstance(local_labels, local_n_labels, f, test_instances, classes)/(perf && perf->eb_accuracy > ZERO ? perf->eb_auc : 1.0),  parm->eb_auc_w) : 1.0;

            if (parm->eb_coverage_w > ZERO) {
                scalar coverage = Coverage(local_labels, local_n_labels, f, test_instances, classes, &optimal);
                fitness *= coverage > ZERO ? pow(!perf ? optimal/coverage : perf->eb_coverage/coverage, parm->eb_coverage_w) : 1.0;
            }
        }

        scalar root = parm->eb_subset_accuracy_w + parm->eb_hamming_loss_w + parm->eb_accuracy_w + parm->eb_precision_w + parm->eb_recall_w + parm->eb_f1_w + parm->lb_accuracy_macro_w +
                      parm->lb_precision_macro_w + parm->lb_recall_macro_w + parm->lb_f1_macro_w + parm->lb_accuracy_micro_w + parm->lb_precision_micro_w + parm->lb_recall_micro_w +
                      parm->lb_f1_micro_w + parm->eb_coverage_w + parm->reduction_ml_w + parm->reduction_perlabel_w + parm->negatives_ml_w + parm->n_predicted_ml_w + parm->reduction_ml_ms_w +
                      parm->oversampling_ml_ms_w;

        fitness *= parm->n_predicted_ml_w > ZERO ? pow(diff = AverageNumberOfPredictedLabelsDifference(n_h, local_n_labels, test_instances/*, classes*/), parm->n_predicted_ml_w) : 1.0;

        fitness *= parm->eb_jaccard_coeff_w > ZERO ? parm->eb_jaccard_coeff_w * JaccardCoefficient(local_labels, local_n_labels, h, n_h, test_instances) : 1.0;

        // Performance does not include reduction
        *performance = fitness;

        // k regularization term
        fitness *= (method == MLkNN && parm->regularize_k_w > ZERO) ? parm->regularize_k_w * (((scalar) learner->mlknn.k)/parm->ml_max_k) : 1.0;

        // Reduction
        fitness *= parm->reduction_ml_w > 0.0 ? pow(reduction, parm->reduction_ml_w) : 1.0;
        fitness *= parm->reduction_perlabel_w > 0.0 ? pow(per_label_reduction, parm->reduction_perlabel_w) : 1.0;
        fitness *= parm->reduction_ml_ms_w > 0.0 ? pow(ms_reduction, parm->reduction_ml_ms_w) : 1.0;

        // Oversampling
        fitness *= parm->oversampling_ml_ms_w > 0.0 ? pow(oversampling, parm->oversampling_ml_ms_w) : 0.0;

        // For geometric mean classification performance equals fitness
        fitness = pow(fitness, 1.0/root);
    }

    // Hast table?
    if (parm->ml_use_hash_table) {
        new.ok = fitness;
#ifdef _OPENMP
        #pragma omp critical
#endif
        HashAddEntry(ht, &new);

        FREE(new.key);
    }

    if (!ss && sselection) {
        FREE(sselection);
    } else if (!s && selection)
        FREE(selection);

    SHOWPROGRESS(parm->show_progress > 2, "Evaluated individual fitness, reduction: %6.4lf, %6.4lf\n", fitness, reduction);

    FreeScalarMatrix(&f, instances);
    FreeIntegerMatrix(&h, instances);
    FREE(n_h);

    if (isnan(fitness) || isinf(fitness))
        OnErrorExit(ERR_NUMERICAL, "Fitness invalid value (NaN or infinite)");

    if (isnan(fitness) || isinf(fitness))
        OnErrorExit(ERR_NUMERICAL, "Infinite of NaN fitness value");

    // Free memory if necessary
    if (parm->three_way_is) {
        FREE(local_n_labels);
        FreeIntegerMatrix(&local_labels, instances);
    }

    return fitness;
}

scalar AverageNumberOfPredictedLabelsDifference(int *n_h, int *n_labels, int instances/*, int classes*/)
{
    // Average number of predicted labels per instances
    scalar av_diff = 0.0;
    for (int i = 0; i < instances; i++)
        av_diff += ((scalar) abs(n_h[i] - n_labels[i]))/MAX(n_h[i], n_labels[i]); // classes

    return 1.0 - av_diff/instances;
}
