/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements Random k-Labelsets (RAkEL) method
// G. Tsoumakas, I. Katakis and I. Vlahavas, "Random k-Labelsets for Multilabel Classification," in IEEE Transactions on Knowledge and Data Engineering, vol. 23, no. 7, pp. 1079-1089, July 2011.

void TrainRAkELModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_RAkEL *rakel, int *subspace, bool cv, t_Classifier classifier,
                      int *subset, int **SS, int **index, bool overlapping)
{
    int *l_subspace;

    // Test if valid CC classification model
    if (classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RAkEL classification method");

    // Pending implementation
    if (SS)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RAkEL for partial instance selection");

    // If m = 0, set to m = 2·classes
    if (parm->rakel_m <= ZERO) {
        ((t_Parameters *) parm)->rakel_m = 2*classes;
        SAVEPROGRESS(parm->show_progress > 0, "@RAkEL m set to: %d\n", 2*classes);
    }

    // If not-overlapping powersets m is fixed by k
    if (!overlapping) {
        ((t_Parameters *) parm)->rakel_m = (int) ceil(((scalar) classes)/parm->rakel_k);
        SAVEPROGRESS(parm->show_progress > 0, "@RAkEL m set to: %d\n", parm->rakel_m);
    }

    if (parm->rakel_k >= classes) {
        printf("WARNING: k greater than number of labels for RAkEL, adjusting");
        ((t_Parameters *)parm)->rakel_k = classes-1;
    }

    // Set model parameters
    rakel->method = classifier;
    rakel->m = parm->rakel_m;
    MALLOC(rakel->k, int, rakel->m, "TrainRAkELModels: rakel->k");

    // If no subspace select all
    if (!subspace) {
        MALLOC(l_subspace, int, inputs, "TrainRAkELModels: l_subspace");
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = 1;
    } else
        l_subspace = subspace;

    // Label dependent instance selection and standard instance selection are incompatible
    if (subset && SS)
        OnErrorExit(ERR_PARAM, "Label dependent instance selection and standard instance selection are incompatible");

    // Generate powersets if none given
    // if (!rakel->h)
    GeneratePowersetsForRakel(parm, rakel, classes, overlapping);

    SHOWPROGRESS(parm->show_progress > 0, "RAkEL%c m: %d, k: %d\n", overlapping ? 'o' : 'd', rakel->m, parm->rakel_k);

    // Train a model for every label
    MALLOC(rakel->models, t_Learner, rakel->m, "TrainRAkELModels: rakel->models");
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < rakel->m; i++) {
        int *l_labels;

        // Create binary labels
        MALLOC(l_labels, int, instances, "TrainRAkELModels: l_labels");

        // Get label powersets
        GetSingleLabelFromPowerset(labels, n_labels, instances, rakel->k[i], rakel->h[i], l_labels);

        // Cross-validate hyper-parameters?
        t_Parameters lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, data, instances, l_labels, l_subspace, inputs, (int) pow(2.0, rakel->k[i]), rakel->method, 1, subset, index);

        // Train
        SHOWPROGRESS(parm->show_progress > 0, "Training RAkEL classifier %*d/%d\r", WIDTH(rakel->m), i+1, rakel->m);
        ClassifierTraining(&lparm, rakel->method, data, instances, l_labels, inputs, (int) pow(2.0, rakel->k[i]), index, l_subspace, subset, true, &(rakel->models[i]), &(rakel->ensembles[i]));

        if (parm->print_base_training_error)
            PrintClassifierPerformanceMetrics(parm, instances, inputs, data, (int) pow(2.0, rakel->k[i]), index, l_subspace, subset, l_labels, rakel->method, &(rakel->models[i]),
                                              &(rakel->ensembles[i]), -1);

        // For k-NN add labels
        if (rakel->method == kNN || rakel->method == NN) {
            MALLOC(rakel->models[i].knn->labels, int, instances, "TrainRAkELModels: rakel->models[i].knn->labels");
            memcpy(rakel->models[i].knn->labels, l_labels, instances*sizeof(int));
        }

        FREE(l_labels);
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    if (!subspace)
        FREE(l_subspace);
}

void ClassifierRAkELInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                             int inputs, int *query_index, t_RAkEL *rakel, int *subspace, int *S, int **index)
{
    int *votes;

    CALLOC(votes, int, n_labels, "ClassifierRAkELInstance: votes");

    // Output of all powersets classifiers
    *n_predicted = 0;
    for (int i = 0; i < rakel->m; i++) {
        int category = ClassifierProbabilities(parm, rakel->method, query, inputs, pow(2.0, rakel->k[i]), query_index, subspace, data, instances, (rakel->method == kNN || rakel->method == NN) ?
                       rakel->models[i].knn->labels : NULL, S, &(rakel->models[i]), -1, NULL, -1, index, NULL, &(rakel->ensembles[i]));

        // Decodify category
        DecodifyAndAddVotes(category, rakel->k[i], rakel->h[i], votes);
    }

    *classification = *ranks = true;

    // Predicted labels
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++) {
        if (votes[i] >= rakel->threshold[i])
            h[(*n_predicted)++] = i;

        f[i] = votes[i]/rakel->threshold[i];
    }

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    FREE(votes);
}

void GetSingleLabelFromPowerset(int **labels, int *n_labels, int instances, int k, int *h, int *l_labels)
{
    for (int i = 0; i < instances; i++) {
        int label = 0;
        for (int j = 0; j < k; j++)
            if (LabelInList(h[j], labels[i], n_labels[i]))
                label += (int) pow(2.0, j);

        l_labels[i] = label;
    }
}

void DecodifyAndAddVotes(int category, int k, int *lp, int *votes)
{
    for (int i = 0; i < k; i++) {
        if (category % 2)
            votes[lp[i]]++;

        category /= 2;
    }
}

void FreeRAkEL(const t_Parameters *parm, t_RAkEL *rakel, int inputs)
{
    for (int i = 0; i < rakel->m; i++)
        FREE(rakel->h[i]);
    FREE(rakel->h);
    FREE(rakel->threshold);
    FREE(rakel->k);
    for (int i = 0; i < rakel->m; i++)
        ClassifierFree(parm, &(rakel->models[i]), rakel->method, 2, inputs, &(rakel->ensembles[i]));
    FREE(rakel->models);
}

void GeneratePowersetsForRakel(const t_Parameters *parm, t_RAkEL *rakel, int classes, bool overlapping)
{
    // Allocate space for powersets
    MALLOC(rakel->h, int *, rakel->m, "GeneratePowersetsForRakel: rakel->h");
    for (int i = 0; i < rakel->m; i++)
        MALLOC(rakel->h[i], int, parm->rakel_k, "GeneratePowersetsForRakel: rakel->h[i]");
    MALLOC(rakel->threshold, scalar, classes, "GeneratePowersetsForRakel: rakel->threshold");
    ZeroVector(rakel->threshold, NULL, classes);

    for (int i = 0; i < rakel->m; i++)
        rakel->k[i] = overlapping ? parm->rakel_k : classes/rakel->m;

    if (overlapping) {
        int *binary;
        MALLOC(binary, int, classes, "GeneratePowersetsForRakel: binary");
        MALLOC(rakel->h, int *, rakel->m, "GeneratePowersetsForRakel: rakel->h");

        for (int i = 0; i < rakel->m; i++)
            MALLOC(rakel->h[i], int, rakel->k[i], "GeneratePowersetsForRakel: rakel->h[i]");
        for (int i = 0; i < rakel->m; i++) {
            ZeroVector(NULL, binary, classes);
            for (int j = 0; j < rakel->k[i]; j++) {
                int next;
                do {
                    next = RANDOM(0, classes);
                } while (binary[next]);
                binary[next] = 1;
            }

            for (int j = 0, l = 0; j < classes; j++)
                if (binary[j]) {
                    rakel->h[i][l] = j;
                    rakel->threshold[j]++;
                    l++;
                }
        }

        // Threshold for relevance classification
        for (int j = 0; j < classes; j++)
            rakel->threshold[j] /= 2;

        FREE(binary);
    } else {
        int *list, next = 0;
        MALLOC(list, int, classes, "GeneratePowersetsForRakel: list");
        for (int i = 0; i < classes; i++)
            list[i] = i;
        RandomizeVector(list, classes);

        // Assign randomly
        for (int i = 0; i < rakel->m; i++)
            for (int j = 0; j < rakel->k[i]; j++)
                rakel->h[i][j] = list[next++];

        // Assign RR randomly
        int rlp = RANDOM(0, rakel->m);
        for (; next < classes; next++) {
            rakel->h[rlp][rakel->k[rlp]] = list[next];
            rakel->k[rlp]++;

            rlp = (rlp+1) % rakel->m;
        }

        for (int j = 0; j < classes; j++)
            rakel->threshold[j] = 0.5;

        FREE(list);
    }
}
