/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

int RCGAMultiSelectionForML(const t_Parameters *parm, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs, int *selection, int **index, int *ks,
                            t_MLClassifier method, int **ksc, int *generations_performed)
{
    int *list, neighbors_cache_size = 0, **ss = NULL, selected = 0, loperk_count = 0, *individual;
    scalar **population, *fitness, *performance, best_performance = -1.0, best_fitness = -1.0, loperk_inc = 0.0;
    cache_t **neighbors_cache = NULL;

    if (!parm->multiselection)
        OnErrorExit(ERR_PARAM, "Function only available for multi-selection");

    // Available methods
    if (method == CC || method == BR)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "RCGA MS for BR/CC/RAkEL methods");

    if (method == LAMLksNN)
        OnErrorExit(ERR_PARAM, "LAML-k'sNN not suitable for instance selection");

    if (parm->ml_ins_sel_per_label)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Multi-selection for instance selection per label");

    if (method != MLkNN)
        OnErrorExit(ERR_PARAM, "Relative multi-selection only available for ML-kNN");

    // Adjust for multi-selection of instances
    scalar chc_init_prob = parm->chc_init_prob;
    if (parm->multiis_adjust_initial_percentage) {
        scalar sum = 0.0;
        int n = parm->ml_max_k/2+1;

        for (int i = 1; i <= n; i++)
            sum += i;

        fprintf(parm->out_file, "Initial percentage adjusted: %lf", parm->chc_init_prob);

        // With this adjustment, multi-selection selects the same number of instances, although some of them repeated
        chc_init_prob /= sum/n;

        fprintf(parm->out_file, " => %lf\n", chc_init_prob);
    }

    // Create and initialize global population
    MALLOC(fitness, scalar, 2*parm->population_size, "RCGAMultiSelectionForML: fitness");
    MALLOC(performance, scalar, 2*parm->population_size, "RCGAMultiSelectionForML: performance");
    MALLOC(list, int, parm->population_size, "RCGAMultiSelectionForML: list");
    MALLOC(individual, int, instances, "RCGAMultiSelectionForML: individual");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;
    AllocateScalarMatrix(&population, 2*parm->population_size, instances);

    if (parm->ml_init_with_undersampling)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Undersampling initialization wiht RCGA algorithm");
    else
        for (int i = 0; i < parm->population_size; i++)
            for (int j = 0; j < instances; j++)
                population[i][j] = RANDOM(0.0, 1.0) < chc_init_prob ? RANDOM(0.0, 1.0) : 0.0;

    // If cross-validate hyper-parameters k is fixed to best value
    if (parm->cv_ml_hyperparameters) {
        CrossValidateHyperParametersForMultiLabel((t_Parameters *) parm, MLkNN, data, instances, inputs, n_labels, labels, classes, index, NULL, NULL, parm->cv_folds);
        ((t_Parameters *)parm)->ml_min_k = ((t_Parameters *)parm)->ml_max_k = parm->mlknn_k;
        printf("WARNING: k fixed to %d, not evolved\n", parm->mlknn_k);
    }

    // Evolution of k
    int ***k = NULL;
    Allocate3DIntegerMatrix(&k, 2*parm->population_size, parm->lamlknn_max_n_clusters, classes);
    for (int i = 0; i < parm->population_size; i++)
        for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
            for (int j = 0; j < classes; j++)
                k[i][c][j] = (int) RANDOM(parm->ml_min_k, parm->ml_max_k+1);   // If only one k, k[i][0][0] is used

    // Evolution of number of clusters for LAML-kNN and LAML-k'sNN
    int *clusters = NULL;
    MALLOC(clusters, int, 2*parm->population_size, "RCGAMultiSelectionForML: clusters");
    for (int i = 0; i < parm->population_size; i++)
        clusters[i] = (int) RANDOM(parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters+1);

    // Cache of neighbors?
    scalar **distance_cache = NULL;
    SHOWPROGRESS(parm->show_progress > 0, "Cache size (maximum size): %.1lfGb (%.1lfGb)\n", NEIGHBORS_CACHE_SIZE(instances), MAXIMUM_CACHE_SIZE);
    if (parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(data, instances, inputs, &distance_cache, &neighbors_cache, NULL, index, &neighbors_cache_size, NULL, parm->ml_loo);
    else if (parm->ml_loo)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evaluation with LOO without cache");

    // For evaluations of individuals
    t_MLkNN mlknn;
    t_MLksNN mlksnn;
    t_LAMLkNN lamlknn;
    t_BR br;
    t_CC cc;
    t_RAkEL rakel;
    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, (void *) &br, (void *) &cc, (void *) &lamlknn, (void *) &mlksnn, NULL, (void *) &rakel};

    // Evaluate initial individuals
    for (int i = 0; i < parm->population_size; i++) {
        for (int j = 0; j < instances; j++)
            individual[j] = (int) ceil(population[i][j] * k[i][0][0]);
        fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, individual, neighbors_cache, NULL,  NULL, index, &performance[i], ss, NULL,
                                               method, classifier[method], NULL, 0, NULL, k[i][0][0], k[i][0], clusters[i], NULL, NULL, instances);
    }

    time_t init_time = time(NULL);

    for (int g = 0; g < parm->generations; g++) {
        SHOWPROGRESS(parm->show_progress > 0, "Generation: %*d/%d\r", WIDTH(parm->generations), g+1, parm->generations);
        *generations_performed = g+1;

        // Random selection without replacement
        RandomizeVector(list, parm->population_size);
        int actual_size = parm->population_size;

        // Generate new individuals using crossover only
        for (int i = 0; i < parm->population_size; i += 2) {
            int father = list[i];
            int mother = list[i+1];

            // Perform SBC crossover creating new individuals
            SimulatedBinaryCrossover(population[actual_size], population[actual_size+1], population[father], population[mother], instances, parm->sbc_nc, 0.0, 1.0);

            // Perform BLX-alpha crossover for k
            for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
                BLXAlphaCrossover(k[actual_size][c], k[actual_size+1][c], k[father][c], k[mother][c], classes, parm->ml_min_k, parm->ml_max_k);

            // Perform BLX-alpha crossover for number of clusters
            BLXAlphaCrossover(&clusters[actual_size], &clusters[actual_size+1], &clusters[father], &clusters[mother], 1, parm->lamlknn_min_n_clusters, parm->lamlknn_max_n_clusters);

            actual_size += 2;
        }

        // Evaluate NEW individuals
        for (int i = parm->population_size; i < actual_size; i++) {
            for (int j = 0; j < instances; j++)
                individual[j] = (int) ceil(population[i][j] * k[i][0][0]);
            fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, individual, neighbors_cache, NULL, NULL, index, &performance[i], ss, NULL,
                                                   method, classifier[method], NULL, g, NULL, k[i][0][0], k[i][0], clusters[i], NULL, NULL, instances);
        }

        ///////////////////////////////////////////
        // Mutation operations ////////////////////
        ///////////////////////////////////////////

        // Random mutation
        if (parm->ml_chc_random_mut > 0.0)
            // If multi-selection also mutate k
            for (int i = parm->population_size*parm->ml_elitism; i < actual_size; i++)
                if (RANDOM(0.0, 1.0) < parm->ml_chc_random_mut) {
                    RCGANonUniformMutation(population[i], instances, g, parm->generations, parm->bit_mutation, 0.0, 1.0);
                    RandomMLMutation(parm, k[i][0], 1, parm->bit_mutation, g, parm->generations, true, 1);
                    // Avoid k outside bounds
                    k[i][0][0] = AssignWithinBoundsInteger(parm->ml_min_k, k[i][0][0], parm->ml_max_k);
                    for (int j = 0; j < instances; j++)
                        individual[j] = (int) ceil(population[i][j] * k[i][0][0]);
                    fitness[i] = EvaluateIndividualFitness(parm, NULL, NULL, -1, instances, labels, n_labels, classes, inputs, data, individual, neighbors_cache, NULL, NULL, index, &performance[i],
                                                           ss, NULL, method, classifier[method], NULL, g, NULL, k[i][0][0], k[i][0], clusters[i], NULL, NULL, instances);
                }

        // Select subset of best individuals
        for (int i = 0; i < parm->population_size; i++) {
            int order = OrderedSelection(fitness+i, 0, actual_size-i)+i;

            scalar *aux_pt = population[order];
            population[order] = population[i];
            population[i] = aux_pt;

            scalar aux_f = fitness[order];
            fitness[order] = fitness[i];
            fitness[i] = aux_f;

            aux_f = performance[order];
            performance[order] = performance[i];
            performance[i] = aux_f;

            int **aux_dpt = k[order];
            k[order] = k[i];
            k[i] = aux_dpt;

            int aux_i = clusters[order];
            clusters[order] = clusters[i];
            clusters[i] = aux_i;
        }

        if (parm->show_progress > 0) {
            int best = OrderedSelection(fitness, 0, parm->population_size);

            switch (method) {
                case MLkNN:
                case LAMLkNN:
                    SHOWPROGRESS(true, "%*d g's best PERF, FITNESS, k: %6.4lf, %6.4lf, %3d", WIDTH(parm->generations), g+1, performance[best], fitness[best], k[best][0][0]);
                    SHOWPROGRESS(method == LAMLkNN, ", clusters = %d", clusters[best]);
                    SHOWPROGRESS(true, "%c", '\n');
                    break;

                case LAMLksNN:
                    SHOWPROGRESS(true, "%*d g's best PERF, FITNESS, k's: %6.4lf, %6.4lf\n", WIDTH(parm->generations), g+1, performance[best], fitness[best]);
                    break;

                case MLksNN:
                    SHOWPROGRESS(true, "%*d g's best PERF, FITNESS, k's: %6.4lf, %6.4lf, {", WIDTH(parm->generations), g+1, performance[best], fitness[best]);
                    for (int i = 0; i < classes-1; i++)
                        SHOWPROGRESS(true, "%3d, ", k[best][0][i]);
                    SHOWPROGRESS(true, "%3d}\n", k[best][0][classes-1]);
                    break;

                default:
                    break;
            }

            VERBOSE(parm->show_progress > 1, parm->out_file, "%*d generation best individual PERFORMANCE, FITNESS(MEAN): %6.4lf, %6.4lf(%6.4lf)\n", WIDTH(parm->generations), g+1, fitness[best],
                    performance[best], AverageScalarVector(fitness, parm->population_size));
            VERBOSE(parm->show_progress > 1, parm->out_file, "CURRENT best individual PERFORMANCE, FITNESS: %6.4lf, %6.4lf\n", best_performance, best_fitness);
        }

        // Check it time limit exceeded
        if (parm->time_limit && (time(NULL) - init_time) > parm->time_limit) {
            SAVEPROGRESS(true, "Time limit exceeded after %d generations\n", g+1);
            break;
        }
    }

    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    SHOWPROGRESS(parm->show_progress > 0 && k && loperk_count > 0, "Average increment for k per label local optimization: %6.4lf\n", loperk_inc/loperk_count);
    VERBOSE(parm->show_progress > 0 && k && loperk_count > 0, parm->out_file, "Average increment for k per label local optimization: %6.4lf\n", loperk_inc/loperk_count);

    int best = OrderedSelection(fitness, 0, parm->population_size);

    selected = 0;
    for (int i = 0; i < instances; i++)
        if ((selection[i] = (int) ceil(population[best][i] * k[best][0][0])))
            selected++;

    ((t_Parameters *) parm)->mlknn_k = k[best][0][0];
    if (ks)
        for (int i = 0; i < classes; i++)
            ks[i] = k[best][0][i];

    if (ksc)
        for (int c = 0; c < parm->lamlknn_max_n_clusters; c++)
            for (int i = 0; i < classes; i++)
                ksc[c][i] = k[best][c][i];

    ((t_Parameters *) parm)->lamlknn_n_clusters = clusters[best];

    scalar perf;
    perf = EvaluateClassifierPerformance(parm, method, classifier[method], selection, NULL, neighbors_cache, instances, labels, n_labels, classes, data, index, inputs, NULL, false, NULL,
                                         parm->mlknn_k, ks, parm->lamlknn_n_clusters, NULL, NULL);
    SHOWPROGRESS(parm->show_progress > 0, "FINAL BEST Performance, Reduction, k: %lf, %lf, %d\n", perf, 1.0 - ((scalar)selected)/instances, k[best][0][0]);

    switch (method) {
        case MLksNN:
            fprintf(parm->out_file, "@RCGA average best ML-k'sNN k's: %lf\nk's = {", AverageIntegerVector(ks, classes));
            for (int i = 0; i < classes-1; i++)
                fprintf(parm->out_file, "%3d, ", ks[i]);
            fprintf(parm->out_file, "%3d}\n", ks[classes-1]);
            break;

        case MLkNN:
            fprintf(parm->out_file, "@RCGA best ML-kNN k: %d\n", parm->mlknn_k);
            break;

        case LAMLkNN:
            fprintf(parm->out_file, "@RCGA best LAML-kNN k: %d\n", parm->mlknn_k);
            break;

        default:
            break;
    }

    // Free memory
    for (int i = 0; i < 2*parm->population_size; i++)
        FREE(population[i]);
    FREE(population);
    FREE(fitness);
    FREE(list);
    FREE(performance);
    FREE(individual);
    Free3DIntegerMatrix(&k, 2*parm->population_size, parm->lamlknn_max_n_clusters);
    FREE(clusters);
    if (neighbors_cache) {
        for (int i = 0; i < instances; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
    }
    if (distance_cache) {
        for (int i = 0; i < instances; i++)
            FREE(distance_cache[i]);
        FREE(distance_cache);
    }

    FreeMLClassifier(parm, classifier[method], method, classes, inputs, instances, NULL);

    return selected;
}

void SimulatedBinaryCrossover(scalar *son, scalar *daughter, scalar *father, scalar *mother, int length, int nc, scalar min_value, scalar max_value)
{
    scalar u = RANDOM(0.0, 1.0);
    scalar beta = (u <= 0.5) ? pow(2.0*u, 1.0/(nc+1.0)) : pow(1.0/(2.0*(1.0-u)), 1.0/(nc+1.0));

    for (int i = 0; i < length; i++) {
        son[i] = AssignWithinBoundsScalar(min_value, 0.5*((1.0+beta)*father[i] + (1.0-beta)*mother[i]), max_value);
        daughter[i] = AssignWithinBoundsScalar(min_value, 0.5*((1.0-beta)*father[i] + (1.0+beta)*mother[i]), max_value);
    }
}

void RCGANonUniformMutation(scalar *individual, int n, int generation, int max_generations, double bit_mutation_rate, int min_value, int max_value)
{
    scalar delta;

    for (int i = 0; i < n; i++)
        if (RANDOM(0.0, 1.0) < bit_mutation_rate) {
            if (RANDOM(0.0, 1.0) < 0.5)
                individual[i] += (delta = DeltaNUM(generation, (double) max_value - individual[i], max_generations));
            else
                individual[i] -= (delta = DeltaNUM(generation, (double) individual[i] - min_value, max_generations));

            if (individual[i] < min_value)
                individual[i] = min_value;
            else if (individual[i] > max_value)
                individual[i] = max_value;
        }
}
