/***************************************************************************
 *   Copyright (C) 2008 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "common.h"
#include "macros.h"
#include "imbalance.h"
#include "typex.h"
#include "maths.h"
#include "is.h"
#include "ml.h"
#include "ensemble.h"

// Repopulate labels for dealing with incomplete labeling

// Access one-dimensional array as two-dimensional array
#define INDEX(row, col, width) ((row)*(width)+(col))

void EvolutionaryLabelRepopulation(const t_Parameters *parm, scalar **data, int instances, int **index, int inputs, int classes, int **labels, int *subspace, int *evaluations,
                                   int *generations_performed, int *n_labels)
{
    scalar *fitness, delta_rm = 0.0, *performance;
    int c_rm = 0, reinit_attempts = 0;
    bool *elite;

    // Set temporarily maximum depth to 10 just in case
    int old_max_dept = parm->dt_max_depth;
    ((t_Parameters *)parm)->dt_max_depth = 10;

    // Create initial population
    int **population;
    MALLOC(population, int *, 2*parm->population_size, "EvolutionaryLabelRepopulation: population");

    MALLOC(fitness, scalar, 2*parm->population_size, "EvolutionaryLabelRepopulation: fitness");
    MALLOC(performance, scalar, 2*parm->population_size, "EvolutionaryLabelRepopulation: fitness");
    MALLOC(elite, bool, 2*parm->population_size, "EvolutionaryLabelRepopulation: elite");
    if (parm->elite <= 0.0)
        for (int i = 0; i < 2*parm->population_size; i++)
            elite[i] = false;

    // Initial population
    for (int i = 0; i < 2*parm->population_size; i++)
        MALLOC(population[i], int, instances*classes, "EvolutionaryLabelRepopulation: population[i].N");

    for (int i = 0; i < parm->population_size; i++)
        for (int j = 0; j < instances; j++)
            for (int k = 0; k < classes; k++)
                population[i][INDEX(j, k, classes)] = RANDOM(0.0, 1.0) < parm->chc_init_prob ? 1 : 0;

    // Cache of neighbors?
    scalar **distance_cache = NULL;
    cache_t **neighbors_cache = NULL;

    SHOWPROGRESS(parm->show_progress > 0 && (parm->ml_method == MLkNN || parm->ml_method == DWKNND || parm->ml_method == DWKNNM || parm->ml_method == DWKNNZ) && parm->use_neighbors_cache,
                 "Cache size (maximum size): %.1lfGb (%.1lfGb)\n", NEIGHBORS_CACHE_SIZE(instances), MAXIMUM_CACHE_SIZE);
    if ((parm->ml_method == MLkNN || parm->ml_method == DWKNND || parm->ml_method == DWKNNM || parm->ml_method == DWKNNZ) && parm->use_neighbors_cache && NEIGHBORS_CACHE_SIZE(instances)
        < MAXIMUM_CACHE_SIZE) {
        SHOWPROGRESS(parm->show_progress > 0, "Using neighbors cache%c", '\n');
        AllocateScalarMatrix(&distance_cache, instances, instances);

        int neighbors_cache_size = 0;
        ConstructAllNeighborsCache(data, instances, inputs, &distance_cache, &neighbors_cache, NULL, index, &neighbors_cache_size, subspace, parm->ml_loo);
    }

    // Needed for evaluation
    t_MLLearner learner[2*parm->population_size];
    t_MLEnsemble ensemble[2*parm->population_size];

    // Evaluate initial population
    time_t init_time  = time(NULL);
    t_Performance perf;
    for (int k = 0; k < parm->population_size; k++) {
        SHOWPROGRESS(parm->show_progress > 1, "Evaluating initial population: %*d/%d\r", WIDTH(parm->population_size), k+1, parm->population_size);

        fitness[k] = EvaluateRepopulatedLabels(parm, population[k], instances, labels, n_labels, classes, inputs, data, neighbors_cache, index, parm->ml_method, &(learner[k]), &perf, 0, NULL,
                                               parm->mlknn_k, NULL, 0, subspace, &(ensemble[k]), &performance[k]);
    }
    SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
    *evaluations = parm->population_size;

    // Main loop of CHC algorithm
    SHOWPROGRESS(parm->show_progress > 1, "Generations: %d/%0*d/%0*d", parm->generations, WIDTH(parm->generations), 0, WIDTH(parm->generations), parm->generations);
    int *list;
    MALLOC(list, int, parm->population_size, "EvolutionaryLabelRepopulation: list");
    for (int i = 0; i < parm->population_size; i++)
        list[i] = i;

    scalar best_fitness = 0.0;
    int best_fitness_gs = 0;

    // Adjust minimum distance taking into account selection probability
    int minimum_distance = MAX(1, (int)((instances*classes*parm->chc_init_prob)/parm->chc_minimum_distance_factor));

    SAVEPROGRESS(true, "Effective minimum distance: %d\n", minimum_distance);
    SHOWPROGRESS(parm->show_progress > 0, "Effective minimum distance: %d\n", minimum_distance);

    for (*generations_performed = 0; *generations_performed < parm->generations; (*generations_performed)++) {
        // CHC next generation
        int actual_size = parm->population_size, order, father, mother;

        // Check evolutionary scheme
        switch (parm->evolutionary_scheme) {
            case schemeCHCGA:
                RandomizeVector(list, parm->population_size);
                for (int i = 0; i < parm->population_size; i += 2) {
                    // Random selection without replacement
                    father = list[i];
                    mother = list[i+1];

                    // If no minimum distance skip
                    if (HammingDistance(population[mother], population[father], instances*classes) < minimum_distance)
                        continue;

                    HuxCrossover(population[mother], population[father], population[actual_size], population[actual_size+1], instances*classes);

                    // Two new individuals in the population.
                    actual_size += 2;
                }

                break;

            case schemeSSGA:
            case schemeSSMA:
            case schemeGGA:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for Evolutionary SMOTE");
        }

        if (actual_size == parm->population_size) {
            minimum_distance--;
            SHOWPROGRESS(parm->show_progress > 0, "No individuals generated, decreasing minimum distance to: %d\n", minimum_distance);
        }

        // Evaluate new individuals
        for (int k = parm->population_size; k < actual_size; k++) {
            SHOWPROGRESS(parm->show_progress > 1, "Evaluating new population: %*d/%d\r", WIDTH(actual_size), k+1, actual_size);

            // Evaluate new individuals if ratio applies
            fitness[k] = EvaluateRepopulatedLabels(parm, population[k], instances, labels, n_labels, classes, inputs, data, neighbors_cache, index, parm->ml_method, &(learner[k]), &perf, 0, NULL,
                                                   parm->mlknn_k, NULL, 0, subspace, &(ensemble[k]), &performance[k]);
        }
        SHOWPROGRESS(parm->show_progress > 1, "%c", '\n');
        *evaluations += actual_size - parm->population_size;

        // Elitism?
        if (parm->elite > 0.0) {
            for (int i = 0; i < actual_size; i++)
                elite[i] = false;
            for (int i = 0; i < parm->elite * parm->population_size; i++)
                elite[OrderedSelection(fitness, i, actual_size)] = true;
        }

        // // Random mutations
        if (parm->mutation > ZERO)
            for (int k = 0; k < actual_size; k++)
                if (!elite[k] && RANDOM(0.0, 1.0) < parm->mutation) {
                    RandomMLMutation(parm, population[k], instances*classes, parm->bit_mutation, *generations_performed, parm->generations, false, 0.0);
                    fitness[k] = EvaluateRepopulatedLabels(parm, population[k], instances, labels, n_labels, classes, inputs, data, neighbors_cache, index, parm->ml_method, &(learner[k]), &perf, 0,
                                                           NULL, parm->mlknn_k, NULL, 0, subspace, &(ensemble[k]), &performance[k]);
                }

        // Obtain new population selecting best N individuals
        switch (parm->evolutionary_scheme) {
            case schemeCHCGA:
                for (int i = 0; i < parm->population_size; i++) {
                    order = OrderedSelection(fitness+i, 0, actual_size - i)+i;
                    int *indi = population[order];
                    population[order] = population[i];
                    population[i] = indi;
                    scalar aux_f = fitness[order];
                    fitness[order] = fitness[i];
                    fitness[i] = aux_f;
                    aux_f = performance[order];
                    performance[order] = performance[i];
                    performance[i] = aux_f;
                }
                break;

            case schemeSSGA:
            case schemeSSMA:
            case schemeGGA:
                OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Evolutionary scheme for ML-ESO");
        }

        SHOWPROGRESS(parm->show_progress > 0, "%*d g's best F(MEAN), P(MEAN): %6.4lf(%6.4lf), %6.4lf(%6.4lf)\n", WIDTH(parm->generations), *generations_performed+1, fitness[0],
                     AverageScalarVector(fitness, parm->population_size), performance[0], AverageScalarVector(performance, parm->population_size));

        // Finished?
        if (StopCriterionMet(parm, time(NULL) - init_time, *evaluations, *generations_performed)) {
            fprintf(parm->out_file, "Reached limit at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
            fprintf(parm->out_file, "@Actual number of evaluations: %d\n", *evaluations);
            break;
        }

        int best = OrderedSelection(fitness, 0, parm->population_size);

        // Show mutation usefulness
        SHOWPROGRESS(parm->show_progress > 1, "Random mutation fitness increment: %7.4lf\n", c_rm ? delta_rm/c_rm: 0.0)

        // Number of generations best is current
        if (fitness[best] == best_fitness)
            best_fitness_gs++;
        else {
            best_fitness = fitness[best];
            best_fitness_gs = 0;
        }

        // Reinitialize worst individuals?
        if (minimum_distance == 0 || (parm->best_individual_fitness_improvement_generations && best_fitness_gs >= parm->best_individual_fitness_improvement_generations)) {
            minimum_distance = (instances*classes)/4;
            int shift = parm->population_size * (1.0-parm->reinitalize_percent_addition);
            for (int i = shift; i < parm->population_size; i++)
                for (int j = 0; j < instances; j++)
                    for (int k = 0; k < classes; k++)
                        population[i][INDEX(j, k, classes)] = RANDOM(0.0, 1.0) < parm->chc_init_prob ? 1 : 0;

            SHOWPROGRESS(parm->show_progress > 0, "%d worst individuals substituted\n", parm->population_size-shift);
            for (int k = shift; k < parm->population_size; k++) {
                SHOWPROGRESS(parm->show_progress > 0, "Evaluating NEW population: %*d/%d\r", WIDTH(parm->population_size), k+1, parm->population_size);

                fitness[k] = EvaluateRepopulatedLabels(parm, population[k], instances, labels, n_labels, classes, inputs, data, neighbors_cache, index, parm->ml_method, &(learner[k]), &perf, 0,
                                                       NULL, parm->mlknn_k, NULL, 0, subspace, &(ensemble[k]), &performance[k]);
            }

            SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');
            reinit_attempts++;

            if (parm->chc_max_reinit_attempts && (reinit_attempts > parm->chc_max_reinit_attempts)) {
                fprintf(parm->out_file, "Reached limit of REINITILIZATIONS at generation %d and %d evaluations\n", *generations_performed+1, (int) *evaluations);
                fprintf(parm->out_file, "@Actual number of evaluations: %d\n", *evaluations);
                break;
            }
        }
    }

    // Create new labels with best individual
    int original = instances;
    int best = OrderedSelection(fitness, 0, parm->population_size);

    int *local_n_labels = NULL, **local_labels = NULL;
    MALLOC(local_n_labels, int, instances, "EvaluateIndividualFitness: local_n_labels");
    AllocateIntegerMatrix(&local_labels, instances, classes);
    PopulateLabels(labels, n_labels, local_labels, local_n_labels, population[best], classes, instances);

    SAVEPROGRESS(true, "@Repopulation ratio: %6.4lf\n", LabelRepopulationRatio(n_labels, local_n_labels, instances));

    for (int i = 0; i < instances; i++) {
        REALLOC(labels[i], int, local_n_labels[i], "EvolutionaryLabelRepopulation: labels[i]");
        CopyIntegerVector(local_labels[i], labels[i], local_n_labels[i]);
        n_labels[i] = local_n_labels[i];
    }

    FREE(local_n_labels);
    FreeIntegerMatrix(&local_labels, instances);

    // Mutation benefits
    // fprintf(parm->out_file, "Random mutation average fitness increment: %7.4lf\n", c_rm ? delta_rm/c_rm: 0.0);

    FREE(list);
    for (int i = 0; i < 2*parm->population_size; i++)
        FREE(population[i]);

    FREE(population);
    FREE(fitness);
    FREE(performance);
    FREE(elite);
    if (distance_cache) {
        for (int i = 0; i < original; i++)
            FREE(neighbors_cache[i]);
        FREE(neighbors_cache);
        FreeScalarMatrix(&distance_cache, original);
    }

    ((t_Parameters *)parm)->dt_max_depth = old_max_dept;
}

void PopulateLabels(int **labels, int *n_labels, int **repop_labels, int *repop_n_labels, int *s, int classes, int instances)
{
    int *list;

    MALLOC(list, int, classes, "PopulateLabels: list");

    for (int i = 0; i < instances; i++) {
        // Convert to binary
        ZeroVector(NULL, list, classes);
        for (int j = 0; j < n_labels[i]; j++)
            list[labels[i][j]] = 1;

        // Set selected labels
        for (int j = 0; j < classes; j++)
            if (s[INDEX(i, j, classes)])
                list[j] = 1;

        // Assign to new labels
        repop_n_labels[i] = 0;
        for (int j = 0; j < classes; j++)
            if (list[j])
                repop_labels[i][repop_n_labels[i]++] = j;
    }

    FREE(list);
}

scalar LabelRepopulationRatio(int *n_labels, int *repop_n_labels, int instances)
{
    scalar added = 0.0, present = 0.0;

    for (int i = 0; i < instances; i++) {
        added += repop_n_labels[i] - n_labels[i];
        present += n_labels[i];
    }

    return added/present;
}

scalar EvaluateRepopulatedLabels(const t_Parameters *parm, int *individual, int instances, int **labels, int *n_labels, int classes, int inputs, scalar **data, cache_t **neighbors_cache, int **index,
                                 t_MLClassifier method, t_MLLearner *learner, t_Performance *perf, int generation, int *chain, int k, int *ks, int n_clusters, int *subspace, t_MLEnsemble *ensemble,
                                 scalar *performance)
{
    char *tempnam(const char *dir, const char *pfx);

    scalar optimal, *TP, *FP, *TN, *FN, fitness = -1.0;
    bool classification, ranks;

    // Create local label sets
    int *local_n_labels = NULL, **local_labels = NULL;
    MALLOC(local_n_labels, int, instances, "EvaluateIndividualFitness: local_n_labels");
    AllocateIntegerMatrix(&local_labels, instances, classes);
    PopulateLabels(labels, n_labels, local_labels, local_n_labels, individual, classes, instances);
    scalar ratio = LabelRepopulationRatio(n_labels, local_n_labels, instances);
    SHOWPROGRESS(parm->show_progress > 1, "\nRepopulation ratio: %6.4lf\n", ratio);
    // Classify testing set
    scalar **f  = NULL, metric = 0.0;
    int **h = NULL, *n_h = NULL;

    AllocateScalarMatrix(&f, instances, classes);
    AllocateIntegerMatrix(&h, instances, classes);
    MALLOC(n_h, int, instances, "main: n_h");

    // Test if evaluation must be carried out with external classifier
    if (parm->ml_external_eval_interval && ((generation+1) % parm->ml_external_eval_interval) == 0) {
        extern char *ml_ext_method[];
        SHOWPROGRESS(parm->show_progress > 1, "External evaluation of fitness with method %s at generation %d\n", ml_ext_method[parm->evolution_external_classifier], generation);
        char *file_template = tempnam(".", NULL);
        ClassifyExternalDataset(parm, parm->evolution_external_classifier, data, h, f, data, instances, local_labels, local_n_labels, classes, n_h, inputs, index, index, instances, NULL,
                                &classification, &ranks, file_template, labels, n_labels, false, NULL, subspace);
    } else {
        int *lchain = NULL;
        t_Learner *learners = NULL;
        if (method == CC || method == BR) {
            if (method == CC) {
                if (!chain) {
                    CALLOC(lchain, int, classes, "EvaluateIndividualFitness: lchain");
                } else
                    lchain = chain;
            }
            CALLOC(learners, t_Learner, classes, "EvaluateIndividualFitness: learners");
        }

        TrainMultiLabelClassifier(parm, method, data, instances, local_labels, local_n_labels, classes, inputs, NULL, lchain, index, neighbors_cache, NULL, learner, k, ks, NULL, n_clusters, subspace,
                                  ensemble);
        ClassifyDataset(parm, method, data, h, f, &classification, &ranks, data, instances, local_labels, local_n_labels, classes, n_h, inputs, index, index, instances, NULL, neighbors_cache,
                        NULL, learner, subspace, ensemble);

        if (method == CC && !chain)
            FREE(lchain);

        FreeMLClassifier(parm, learner, method, classes, inputs, instances, ensemble);
    }

    // Get fitness
    fitness = parm->combine_metrics_gmean ? 1.0 : 0.0;

    // Performance METRICS
    scalar diff = 0.0;
    if (!parm->combine_metrics_gmean) {
        fitness -= parm->reduction_ml_w * ratio;

        fitness += parm->eb_subset_accuracy_w > ZERO ? parm->eb_subset_accuracy_w * SubsetAccuracy(labels, n_labels, h, n_h, instances)/(perf && perf->eb_subset_accuracy > ZERO ?
                   perf->eb_subset_accuracy : 1.0) : 0.0;
        fitness += parm->eb_hamming_loss_w  > ZERO ? parm->eb_hamming_loss_w * (!perf ? (1.0 - HammingLoss(labels, n_labels, h, n_h, instances, classes)) :
                   perf->eb_hamming_loss/((metric = HammingLoss(labels, n_labels, h, n_h, instances, classes)) > ZERO ? metric : 1.0)) : 0.0;
        fitness += parm->eb_accuracy_w > ZERO ? parm->eb_accuracy_w * ExampleBasedAccuracy(labels, n_labels, h, n_h, instances)/(perf &&  perf->eb_accuracy > ZERO ? perf->eb_accuracy
                   : 1.0) : 0.0;
        fitness += parm->eb_precision_w > ZERO ? parm->eb_precision_w * ExampleBasedPrecision(labels, n_labels, h, n_h, instances)/(perf && perf->eb_precision > ZERO ?
                   perf->eb_precision : 1.0) : 0.0;
        fitness += parm->eb_recall_w > ZERO ? parm->eb_recall_w * ExampleBasedRecall(labels, n_labels, h, n_h, instances)/(perf && perf->eb_recall > ZERO ? perf->eb_recall : 1.0) : 0.0;
        fitness += parm->eb_f1_w > ZERO ? parm->eb_f1_w * ExampleBasedF1(labels, n_labels, h, n_h, instances, -1.0, -1.0)/(perf && perf->eb_f1 > ZERO ? perf->eb_f1 : 1.0) : 0.0;

        if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
            parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {

            ALLOCA(TP, scalar, classes, "EvaluateIndividualFitness: TP");
            ALLOCA(FP, scalar, classes, "EvaluateIndividualFitness: FP");
            ALLOCA(TN, scalar, classes, "EvaluateIndividualFitness: TN");
            ALLOCA(FN, scalar, classes, "EvaluateIndividualFitness: FN");

            BinaryPerformancePerLabel(TP, FP, TN, FN, labels, n_labels, h, n_h, instances, classes);

            fitness += parm->lb_accuracy_macro_w > ZERO ? parm->lb_accuracy_macro_w * AccuracyMacro(TP, TN, classes, instances)/(perf && perf->lb_accuracy_macro > ZERO ? perf->lb_accuracy_macro :
                       1.0) : 0.0;
            fitness += parm->lb_precision_macro_w > ZERO ? parm->lb_precision_macro_w * PrecisionMacro(TP, FP, classes)/(perf && perf->lb_precision_macro > ZERO ? perf->lb_precision_macro :
                       1.0) : 0.0;
            fitness += parm->lb_recall_macro_w > ZERO ? parm->lb_recall_macro_w * RecallMacro(TP, FN, classes)/(perf && perf->lb_recall_macro > ZERO ? perf->lb_recall_macro : 1.0) : 0.0;
            fitness += parm->lb_f1_macro_w > ZERO ? parm->lb_f1_macro_w * F1Macro(TP, FN, FP, classes)/(perf && perf->lb_f1_macro > ZERO ? perf->lb_f1_macro : 1.0) : 0.0;
            fitness += parm->lb_accuracy_micro_w > ZERO ? parm->lb_accuracy_micro_w * AccuracyMicro(TP, TN, FP, FN, instances, classes)/(perf && perf->lb_f1_macro > ZERO ?
                       perf->lb_accuracy_micro : 1.0) : 0.0;
            fitness += parm->lb_precision_micro_w > ZERO ? parm->lb_precision_micro_w * PrecisionMicro(TP, FP, classes)/(perf && perf->lb_precision_micro > ZERO ? perf->lb_precision_micro :
                       1.0) : 0.0;
            fitness += parm->lb_recall_micro_w > ZERO ? parm->lb_recall_micro_w * RecallMicro(TP, FN, classes)/(perf && perf->lb_recall_micro > ZERO ? perf->lb_recall_micro : 1.0) : 0.0;
            fitness += parm->lb_f1_micro_w > ZERO ?  parm->lb_f1_micro_w * F1Micro(TP, FN, FP, classes)/(perf && perf->lb_f1_micro > ZERO ? perf->lb_f1_micro : 1.0) : 0.0;
        }

        fitness += parm->eb_f1_alt_w > ZERO ? parm->eb_f1_alt_w * F1Instance(labels, n_labels, h, n_h, instances)/(perf && perf->eb_f1_alt > ZERO ? perf->eb_f1_alt : 1.0) : 0.0;

        if (ranks) {
            fitness += parm->eb_one_error_w > ZERO ? parm->eb_one_error_w * (!perf ? 1.0 - OneError(labels, n_labels, f, instances, classes) :
                       perf->eb_one_error/((metric = OneError(labels, n_labels, f, instances, classes)) > ZERO ? metric : 1.0)) : 0.0;
            fitness += parm->eb_ranking_loss_w > ZERO ? parm->eb_ranking_loss_w * (!perf ? (1.0 - RankingLoss(labels, n_labels, f, instances, classes)) :
                       perf->eb_ranking_loss/((metric = RankingLoss(labels, n_labels, f, instances, classes)) > ZERO ? metric : 1.0)) : 0.0;
            fitness += parm->eb_average_precision_w > ZERO ?
                       parm->eb_average_precision_w * AveragePrecision(labels, n_labels, f, instances, classes)/(perf && perf->eb_average_precision > ZERO ? perf->eb_average_precision : 1.0) : 0.0;
            fitness += parm->lb_auc_macro_w > ZERO ?
                       parm->lb_auc_macro_w * AUCMacro(labels, n_labels, f, instances, classes)/(perf && perf->lb_auc_macro > ZERO ? perf->lb_auc_macro : 1.0) : 0.0;
            fitness += parm->lb_auc_micro_w > ZERO ?
                       parm->lb_auc_micro_w * AUCMicro(labels, n_labels, f, instances, classes)/(perf && perf->lb_auc_micro > ZERO ? perf->lb_auc_micro : 1.0) : 0.0;
            fitness += parm->eb_auc_w > ZERO ? parm->eb_auc_w * AUCInstance(labels, n_labels, f, instances, classes)/(perf && perf->eb_auc > ZERO ? perf->eb_auc : 1.0) : 0.0;

            if (parm->eb_coverage_w > ZERO) {
                scalar coverage = Coverage(labels, n_labels, f, instances, classes, &optimal);
                fitness += coverage > ZERO ? (!perf ? optimal/coverage : perf->eb_coverage/coverage) : 0.0;
            }
        }

        fitness += parm->n_predicted_ml_w > ZERO ? parm->n_predicted_ml_w * (diff = AverageNumberOfPredictedLabelsDifference(n_h, n_labels, instances/*, classes*/)) : 0.0;

        fitness += parm->eb_jaccard_coeff_w > ZERO ? parm->eb_jaccard_coeff_w * JaccardCoefficient(labels, n_labels, h, n_h, instances) : 0.0;

        *performance = fitness + parm->reduction_ml_w * ratio;
    } else {
        if (parm->reduction_ml_w > ZERO)
            OnErrorExit(ERR_PARAM, "With geometric mean combination reduction cannot be used");

        fitness *= parm->eb_subset_accuracy_w ? pow(SubsetAccuracy(labels, n_labels, h, n_h, instances)/(perf && perf->eb_subset_accuracy > ZERO ? perf->eb_subset_accuracy : 1.0),
                   parm->eb_subset_accuracy_w) : 1.0;
        fitness *= parm->eb_hamming_loss_w > ZERO ?
                   pow(!perf ? 1.0 - HammingLoss(labels, n_labels, h, n_h, instances, classes) :
                       perf->eb_hamming_loss/((metric = HammingLoss(labels, n_labels, h, n_h, instances, classes)) > ZERO ? metric : 1.0), parm->eb_hamming_loss_w) : 1.0;
        fitness *= parm->eb_accuracy_w > ZERO ? pow(ExampleBasedAccuracy(labels, n_labels, h, n_h, instances)/(perf && perf->eb_accuracy > ZERO ? perf->eb_accuracy : 1.0),
                   parm->eb_accuracy_w) : 1.0;
        fitness *= parm->eb_precision_w > ZERO ? pow(ExampleBasedPrecision(labels, n_labels, h, n_h, instances)/(perf && perf->eb_precision > ZERO ? perf->eb_precision : 1.0),
                   parm->eb_precision_w) : 1.0;
        fitness *= parm->eb_recall_w > ZERO ? pow(ExampleBasedRecall(labels, n_labels, h, n_h, instances)/(perf && perf->eb_recall > ZERO ? perf->eb_recall : 1.0), parm->eb_recall_w)
                   :  1.0;
        fitness *= parm->eb_f1_w > ZERO ? pow(ExampleBasedF1(labels, n_labels, h, n_h, instances, -1.0, -1.0)/(perf && perf->eb_f1 > ZERO ? perf->eb_f1 : 1.0), parm->eb_f1_w) : 1.0;

        if (parm->lb_accuracy_macro_w > ZERO || parm->lb_precision_macro_w > ZERO || parm->lb_recall_macro_w > ZERO || parm->lb_f1_macro_w > ZERO || parm->lb_accuracy_micro_w > ZERO ||
            parm->lb_precision_micro_w > ZERO || parm->lb_recall_micro_w > ZERO || parm->lb_f1_micro_w > ZERO) {

            ALLOCA(TP, scalar, classes, "EvaluateIndividualFitness: TP");
            ALLOCA(FP, scalar, classes, "EvaluateIndividualFitness: FP");
            ALLOCA(TN, scalar, classes, "EvaluateIndividualFitness: TN");
            ALLOCA(FN, scalar, classes, "EvaluateIndividualFitness: FN");

            BinaryPerformancePerLabel(TP, FP, TN, FN, labels, n_labels, h, n_h, instances, classes);

            fitness *= parm->lb_accuracy_macro_w > ZERO ? pow(AccuracyMacro(TP, TN, classes, instances)/(perf && perf->lb_accuracy_macro > ZERO ? perf->lb_accuracy_macro : 1.0),
                       parm->lb_accuracy_macro_w) : 1.0;
            fitness *= parm->lb_precision_macro_w > ZERO ? pow(PrecisionMacro(TP, FP, classes)/(perf && perf->lb_precision_macro > ZERO ? perf->lb_precision_macro : 1.0),  parm->lb_precision_macro_w)
                       : 1.0;
            fitness *= parm->lb_recall_macro_w > ZERO ? pow(RecallMacro(TP, FN, classes)/(perf && perf->lb_recall_macro > ZERO ? perf->lb_recall_macro : 1.0), parm->lb_recall_macro_w) : 1.0;
            fitness *= parm->lb_f1_macro_w > ZERO ? pow(F1Macro(TP, FN, FP, classes)/(perf && perf->lb_f1_macro > ZERO ? perf->lb_f1_macro : 1.0), parm->lb_f1_macro_w) : 1.0;
            fitness *= parm->lb_accuracy_micro_w > ZERO ? pow(AccuracyMicro(TP, TN, FP, FN, instances, classes)/(perf && perf->lb_accuracy_micro > ZERO ? perf->lb_accuracy_micro : 1.0),
                       parm->lb_accuracy_micro_w) : 1.0;
            fitness *= parm->lb_precision_micro_w > ZERO ? pow(PrecisionMicro(TP, FP, classes)/(perf && perf->lb_precision_micro > ZERO ? perf->lb_precision_micro : 1.0),
                       parm->lb_precision_micro_w) : 1.0;
            fitness *= parm->lb_recall_micro_w > ZERO ? pow(RecallMicro(TP, FN, classes)/(perf && perf->lb_recall_micro > ZERO ? perf->lb_recall_micro : 1.0), parm->lb_recall_micro_w) : 1.0;
            fitness *= parm->lb_f1_micro_w > ZERO ? pow(F1Micro(TP, FN, FP, classes)/(perf && perf->lb_f1_micro > ZERO ? perf->lb_f1_micro : 1.0), parm->lb_f1_micro_w) : 1.0;
        }

        fitness *= parm->eb_f1_alt_w > ZERO ? pow(F1Instance(labels, n_labels, h, n_h, instances)/(perf && perf->eb_f1_alt > ZERO ? perf->eb_f1_alt : 1.0), parm->eb_f1_alt_w) : 1.0;

        if (ranks) {
            fitness *= parm->eb_one_error_w > ZERO ? pow(!perf ? 1.0 - OneError(labels, n_labels, f, instances, classes)
                       : perf->eb_one_error/((metric = OneError(labels, n_labels, f, instances, classes)) > ZERO ? metric : 1.0), parm->eb_one_error_w) : 1.0;
            fitness *= parm->eb_ranking_loss_w > ZERO ?
                       pow(!perf ? 1.0 - RankingLoss(labels, n_labels, f, instances, classes) :
                           perf->eb_ranking_loss/((metric = RankingLoss(labels, n_labels, f, instances, classes)) > ZERO ? metric : 1.0), parm->eb_ranking_loss_w) : 1.0;
            fitness *= parm->eb_average_precision_w > ZERO ? pow(AveragePrecision(labels, n_labels, f, instances, classes)/(perf && perf->eb_average_precision > ZERO ?
                       perf->eb_average_precision : 1.0), parm->eb_average_precision_w) : 1.0;
            fitness *= parm->lb_auc_macro_w > ZERO ? pow(AUCMacro(labels, n_labels, f, instances, classes)/(perf && perf->lb_auc_macro > ZERO ? perf->lb_auc_macro : 1.0),
                       parm->lb_auc_macro_w) : 1.0;
            fitness *= parm->lb_auc_micro_w > ZERO ? pow(AUCMicro(labels, n_labels, f, instances, classes)/(perf && perf->lb_auc_micro > ZERO ?
                       perf->lb_auc_micro : 1.0), parm->lb_auc_micro_w) : 1.0;
            fitness *= parm->eb_auc_w > ZERO ?pow(AUCInstance(labels, n_labels, f, instances, classes)/(perf && perf->eb_accuracy > ZERO ? perf->eb_auc : 1.0),  parm->eb_auc_w) : 1.0;

            if (parm->eb_coverage_w > ZERO) {
                scalar coverage = Coverage(labels, n_labels, f, instances, classes, &optimal);
                fitness *= coverage > ZERO ? pow(!perf ? optimal/coverage : perf->eb_coverage/coverage, parm->eb_coverage_w) : 1.0;
            }
        }

        scalar root = parm->eb_subset_accuracy_w + parm->eb_hamming_loss_w + parm->eb_accuracy_w + parm->eb_precision_w + parm->eb_recall_w + parm->eb_f1_w + parm->lb_accuracy_macro_w +
                      parm->lb_precision_macro_w + parm->lb_recall_macro_w + parm->lb_f1_macro_w + parm->lb_accuracy_micro_w + parm->lb_precision_micro_w + parm->lb_recall_micro_w +
                      parm->lb_f1_micro_w + parm->eb_coverage_w + parm->reduction_ml_w + parm->reduction_perlabel_w + parm->negatives_ml_w + parm->n_predicted_ml_w + parm->reduction_ml_ms_w +
                      parm->oversampling_ml_ms_w;

        fitness *= parm->n_predicted_ml_w > ZERO ? pow(diff = AverageNumberOfPredictedLabelsDifference(n_h, n_labels, instances/*, classes*/), parm->n_predicted_ml_w) : 1.0;

        fitness *= parm->eb_jaccard_coeff_w > ZERO ? parm->eb_jaccard_coeff_w * JaccardCoefficient(labels, n_labels, h, n_h, instances) : 1.0;

        *performance = fitness;

        // For geometric mean classification performance equals fitness
        fitness = pow(fitness, 1.0/root);
    }

    SHOWPROGRESS(parm->show_progress > 2, "Evaluated individual fitness: %6.4lf\n", fitness);

    FreeScalarMatrix(&f, instances);
    FreeIntegerMatrix(&h, instances);
    FREE(n_h);

    if (isnan(fitness) || isinf(fitness))
        OnErrorExit(ERR_NUMERICAL, "Fitness invalid value (NaN or infinite)");

    if (isnan(fitness) || isinf(fitness))
        OnErrorExit(ERR_NUMERICAL, "Infinite of NaN fitness value");

    // Free memory if necessary
    FREE(local_n_labels);
    FreeIntegerMatrix(&local_labels, instances);

    return fitness;
}
