/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "macros.h"
#include "maths.h"
#include "common.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

// Functions for sampling of instances


/*
 * The individual is initialized using random undersampling for a given class
 * Minority instances can be oversampled if multi-selection set to True
 */

void MultiSelectionIndividulaWithRandomUndersampling(const t_Parameters *parm, int *individual, int **labels, int *n_labels, int class, scalar chc_init_prob, int instances, bool multiselection)
{
    int minority, majority;

    // Get Minority and Majority instances
    int *list[2], count[] = {0, 0};
    MALLOC(list[0], int, instances, "MultiSelectionIndividulaWithRandomUndersampling: z");
    MALLOC(list[1], int, instances, "MultiSelectionIndividulaWithRandomUndersampling: bar_z");

    // List of instances with class and without class as relevant label
    for (int j = 0; j < instances; j++)
        if (ValueInVector(class, labels[j], n_labels[j])) {
            list[1][count[1]] = j;
            count[1]++;
        } else {
            list[0][count[0]] = j;
            count[0]++;
        }

    minority = count[1] > count[0] ? 0 : 1;
    majority = count[1] <= count[0] ? 0 : 1;

    // Undersample majority class
    RandomizeVector(list[majority], count[majority]);

    // Unselect all
    for (int i = 0; i < instances; i++)
        individual[i] = 0;

    // Select from minority/majority class
    // In case of multi-selection oversampling ratio could be given or could be depending on maximum k
    int max_oversampling_ratio = parm->ml_max_ioversampling_ratio != 0 ? parm->ml_max_ioversampling_ratio : parm->ml_max_k;

    for (int j = 0; j < count[minority]; j++) {
        individual[list[minority][j]] = RANDOM(0.0, 1.0) < chc_init_prob ? (multiselection ? RANDOM(1, max_oversampling_ratio) : 1) : 0;
        individual[list[majority][j]] = RANDOM(0.0, 1.0) < chc_init_prob ? 1 : 0;
    }

    FREE(list[0]);
    FREE(list[1]);
}

// Instances with higher cardinality are selected more often and more times

void MultiSelectionIndividualWithCardinality(const t_Parameters *parm, int *individual, int *n_labels, int classes, scalar chc_init_prob, int instances)
{
    for (int j = 0; j < instances; j++) {
        scalar prob = chc_init_prob * (n_labels[j]/((scalar) classes));
        individual[j] = RANDOM(0.0, 1.0) < prob ? RANDOM(1, parm->ml_max_k+1) : 0;
    }
}

void MultiSelectionWithRoulette(const t_Parameters *parm, int *individual, scalar chc_init_prob, int instances)
{
    scalar *probs;

    MALLOC(probs, scalar, parm->ml_max_k+1, "MultiSelectionWithRoulette: probs");
    probs[0] = 0.0;
    for (int i = 1; i < parm->ml_max_k+1; i++)
        probs[i] = probs[i-1] + (parm->ml_max_k+1)/((scalar) i);

    for (int j = 0; j < instances; j++)
        individual[j] = RANDOM(0.0, 1.0) < chc_init_prob ? RouletteSelection(probs, parm->ml_max_k+1) : 0;

    FREE(probs);
}
