/***************************************************************************
 *   Copyright (C) 2023 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"

// Stacked ML-kNN method
// A. Pakrashi, B. Mac Namee (2017) "Stacked-MLkNN: A stacking based improvement to Multi-Label k-Nearest Neighbours", Proceedings of Machine Learning Research 74:51–63

void TrainStackedMLkNN(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int *S, int inputs, int **index, cache_t **neighbors_cache,
                       t_StackedMLkNN *smlknn, int **SS, int k, bool create, int *subspace)
{
    // Train first layer as in standard ML-kNN
    TrainMultiLabelkNearestNeighbors(parm, data, instances, labels, n_labels, classes, S, inputs, index, neighbors_cache, &(smlknn->first_layer), SS, k, create, subspace);

    // Create inputs to second layer
    int **h, *n_predicted;
    scalar **f;
    bool classification, ranks;
    t_MLLearner learner;

    AllocateIntegerMatrix(&h, instances, classes);
    AllocateScalarMatrix(&f, instances, classes);
    MALLOC(n_predicted, int, instances, "TrainStackedMLkNN: n_predicted");

    learner.mlknn = smlknn->first_layer;
    ClassifyDataset(parm, MLkNN, data, h, f, &classification, &ranks, data, instances, labels, n_labels, classes, n_predicted, inputs, index, index, instances, S, neighbors_cache, SS, &learner,
                    subspace, NULL);

    // Train second layer using outputs of first layer as inputs
    TrainMultiLabelkNearestNeighbors(parm, f, instances, labels, n_labels, classes, S, classes, NULL, NULL, &(smlknn->second_layer), NULL, parm->smlknn_k, create, subspace);
    smlknn->data = f;

    FreeIntegerMatrix(&h, instances);
    FREE(n_predicted);
}

void MultiLabelStackedkNearestNeighbors(const t_Parameters *parm, scalar *query, scalar **data, int instances, int **labels, int *n_labels, int classes,
                                 int *predicted_labels, int *predicted_n_labels, int *S, scalar *f, int inputs, int *query_index, int **index, cache_t **neighbors_cache,
                                 int query_order, t_StackedMLkNN *smlknn, int **SS, int *subspace)
{
    // Output of first layer
    MultiLabelkNearestNeighbors(parm, query, data, instances, labels, n_labels, classes, predicted_labels, predicted_n_labels, S, f, inputs, query_index, index, neighbors_cache, query_order,
                                &(smlknn->first_layer), SS, NULL, NULL, -1, subspace);

    // Output of second layer
    MultiLabelkNearestNeighbors(parm, f, smlknn->data, instances, labels, n_labels, classes, predicted_labels, predicted_n_labels, S, f, classes, NULL, NULL, NULL, -1, &(smlknn->second_layer), NULL,
                                NULL, NULL, -1, subspace);
}

void FreeStackedMLkNN(t_StackedMLkNN *smlknn, int n_labels, int instances)
{
    FreeMLkNN(&(smlknn->first_layer), n_labels);
    FreeMLkNN(&(smlknn->second_layer), n_labels);
    FreeScalarMatrix(&(smlknn->data), instances);
}
