/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Implements ML Stacking model as described in:
// E. Montañes, R. Senge, J. Barranquero, J. R. Quevedo, J. J. del Coz, E. Hüllermeier (2014) "Dependent binary relevance models for multi-label classification"
// Pattern Recognition 47:1494–1508.

void TrainMLStackingModels(const t_Parameters *parm, scalar **data, int instances, int **labels, int *n_labels, int classes, int inputs, t_DBR *sta, int *subspace, bool cv, t_Classifier classifier,
                           int *subset, int **SS, int **index)
{
    int **b_labels, *l_subspace;

    // Test if valid BR classification model
    if (classifier == NN || classifier == kNN || classifier == IBT || classifier == OVO)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "ML MLStacking classification method");

    // If no subspace select all
    MALLOC(l_subspace, int, inputs+classes, "TrainMLStackingModels: l_subspace");
    if (!subspace)
        FillVector(NULL, l_subspace, inputs+classes, 1.0);
    else {
        for (int i = 0; i < inputs; i++)
            l_subspace[i] = subspace[i];
        for (int i = inputs; i < inputs+classes; i++)
            l_subspace[i] = 1;
    }

    // Create binary labels
    AllocateIntegerMatrix(&b_labels, classes, instances);
    ZeroIntegerMatrix(b_labels, classes, instances);

    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            b_labels[labels[i][j]][i] = 1;

    sta->method = classifier;
    if (!IsEnsemble(classifier)) {
        MALLOC(sta->models, t_Learner, classes, "TrainMLStackingModels: sta->models");
        MALLOC(sta->predictors, t_Learner, classes, "TrainMLStackingModels: sta->predictors");
    } else {
        MALLOC(sta->mensembles, t_Ensemble, classes, "TrainMLStackingModels: sta->mensembles");
        MALLOC(sta->pensembles, t_Ensemble, classes, "TrainMLStackingModels: sta->pensembles");
    }

    // Reallocate data to accommodate new inputs
    int *count = NULL;
    if (index)
        MALLOC(count, int, instances, "TrainMLStackingModels: count");
    for (int i = 0; i < instances; i++) {
        REALLOC(data[i], scalar, inputs+classes, "TrainMLStackingModels: data[i]");
        if (index) {
            count[i] = 0;
            for (int j = 0; index[i][j] != -1; j++)
                count[i]++;
            REALLOC(index[i], int, count[i]+classes+1, "TrainMLStackingModels: index[i]");
        }
    }

#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    // Train a model for every label
    for (int i = 0; i < classes; i++) {
        t_Parameters lparm;

        // If undersample, use local subset
        int *S = NULL;
        MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

        if (SS)
            memcpy(S, SS[i], sizeof(int)*instances);
        else if (subset)
            memcpy(S, subset, sizeof(int)*instances);
        else
            for (int j = 0; j < instances; j++)
                S[j] = 1;

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, b_labels[i], S, instances);

        // Cross-validate hyper-parameters?
        lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, data, instances, b_labels[i], l_subspace, inputs, 2, classifier, 1, S, index);

        // Train BR model
        SHOWPROGRESS(parm->show_progress > 0, "STA training 1st level model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);
        ClassifierTraining(&lparm, classifier, data, instances, b_labels[i], inputs, 2, index, l_subspace, S, true, &(sta->models[i]), &(sta->mensembles[i]));

        FREE(S);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    // New data with the outputs of the previous models
    if (!index) {
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic) collapse(2)
#endif
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < classes; j++)
                Classifier2ClassOutput(parm, sta->method, data[i], inputs, 2, index ? index[i] : NULL, subspace, data, instances, NULL, NULL, &(sta->models[j]), 1,
                                             &data[i][inputs+j], index, &(sta->mensembles[j]));
    } else {
        scalar output[classes];
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < instances; i++) {
            for (int j = 0; j < classes; j++)
                Classifier2ClassOutput(parm, sta->method, data[i], inputs, 2, index ? index[i] : NULL, subspace, data, instances, NULL, NULL, &(sta->models[j]), 1,
                                             &output[j], index, &(sta->mensembles[j]));
            int c = count[i];
            // Just in case all zero
            index[i][c] = -1;
            for (int k = 0; k < classes; k++)
                if (fabs(output[k]) > ZERO) {
                    data[i][c] = 1.0;
                    index[i][c] = inputs+k+1;
                    c++;
                }
            index[i][c] = -1;
        }
    }

    // Train a model for every label using remaining labels as inputs
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = 0; i < classes; i++) {
        t_Parameters lparm;

        // If undersample, use local subset
        int *S = NULL;
        MALLOC(S, int, instances, "TrainBinaryRelevanceModels: S");

        if (SS)
            memcpy(S, SS[i], sizeof(int)*instances);
        else if (subset)
            memcpy(S, subset, sizeof(int)*instances);
        else
            FillVector(NULL, S, instances, 1.0);

        // Undersample?
        if (parm->ml_undersample_binary_learners)
            UndersampleDataWithLabel(parm, b_labels[i], S, instances);

        // Cross-validate hyper-parameters?
        lparm = *parm;
        if (cv)
            SetCrossValidatedParameters(&lparm, data, instances, b_labels[i], l_subspace, inputs+classes, 2, classifier, 1, S, index);

        // Train BR model
        SHOWPROGRESS(parm->show_progress > 0, "STA training 2nd level model %*d/%d for label %*d\r", WIDTH(classes), i+1, classes, WIDTH(classes), i+1);
        ClassifierTraining(&lparm, classifier, data, instances, b_labels[i], inputs+classes, 2, index, l_subspace, S, true, &(sta->predictors[i]), &(sta->pensembles[i]));

        FREE(S);
    }
    SHOWPROGRESS(parm->show_progress > 0, "%c", '\n');

    FreeIntegerMatrix(&b_labels, classes);
    FREE(l_subspace);
    for (int i = 0; i < instances; i++) {
        REALLOC(data[i], scalar, inputs, "TrainMLStackingModels: data[i]");
        if (index) {
            REALLOC(index[i], int, count[i]+1, "TrainMLStackingModels: index[i]");
            index[i][count[i]] = -1;
        }
    }

    if (index)
        FREE(count);
}

void MLStackingClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f,  bool *classification, bool *ranks, scalar **data, int instances, int n_labels, int *n_predicted,
                                int inputs, int *query_index, t_DBR *sta, int *subspace, int *S, int **SS, int **index)
{
    *n_predicted = 0;
    scalar output[n_labels], *iquery;
    int *iquery_index = NULL, count = 0;

    MALLOC(iquery, scalar, inputs+n_labels, "MLStackingClassifyInstance: data");
    if (query_index) {
        MALLOC(iquery_index, int, inputs+n_labels+1, "MLStackingClassifyInstance: iquery_index");
        for (int i = 0; query_index[i] != -1; i++) {
            iquery_index[i] = query_index[i];
            iquery[i] = query[i];
            count++;
        }
        iquery_index[count] = -1;
    } else
        memcpy(iquery, query, sizeof(scalar)*inputs);

    // Output of all binary classifiers
    for (int i = 0; i < n_labels; i++)
        Classifier2ClassOutput(parm, sta->method, query, inputs, 2, query_index, subspace, data, instances, NULL, SS ? SS[i] : S, &(sta->models[i]), 1, &output[i], index, &(sta->mensembles[i]));

    // Add outputs of previous classifiers
    if (!query_index)
        for (int k = 0; k < n_labels; k++)
            iquery[k+inputs] = output[k];
    else {
        int c = count;
        // Just in case all zero
        iquery_index[c] = -1;
        for (int k = 0; k < n_labels; k++)
            if (fabs(output[k]) > ZERO) {
                iquery[c] = 1.0;
                iquery_index[c] = inputs+k+1;
                c++;
            }
        iquery_index[c] = -1;
    }

    // Output of all dependent binary classifiers
    for (int i = 0; i < n_labels; i++) {
        int category = Classifier2ClassOutput(parm, sta->method, iquery, inputs+n_labels, 2, iquery_index, subspace, data, instances, NULL, SS ? SS[i] : S, &(sta->predictors[i]), 1, &f[i],
                       index, &(sta->pensembles[i]));

        if (category) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }
    }

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *classification = *ranks = true;

    FREE(iquery);
    if (query_index)
        FREE(iquery_index);
}

void FreeMLStacking(const t_Parameters *parm, t_DBR *dbr, int classes, int inputs)
{
    for (int i = 0; i < classes; i++) {
        if (dbr->models[i].svm)  // Check any, it is an union
            ClassifierFree(parm, &(dbr->models[i]), dbr->method, 2, inputs, &(dbr->mensembles[i]));
        if (dbr->predictors[i].svm)  // Check any, it is an union
            ClassifierFree(parm, &(dbr->predictors[i]), dbr->method, 2, inputs+classes, &(dbr->pensembles[i]));
    }
    FREE(dbr->models);
    FREE(dbr->predictors);
}
