/***************************************************************************
 *   Copyright (C) 2018 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdbool.h>
#include <omp.h>
#include "macros.h"
#include "ml.h"
#include "is.h"
#include "isfs.h"

// Evolve a subpopulation using CHC algorithm

void PerformGeneration(const t_Parameters *parm, int **subpop, int subpop_size, int instances, int *minimum_distance, int subpop_index, int **population, scalar **data, int **labels, int *n_labels,
                       int classes, int inputs, int n_subpops, scalar *pop_fitness, int ***subpops, cache_t **neighbors_cache, scalar *fitness, bool *first_time, t_HashTable *ht, t_HashTable *html,
                       scalar *avereage_difference, scalar *qs, int **index, int *elite, int *k, scalar *accuracy, t_MLClassifier method)
{
    int *list;
    bool *overwrite;
    scalar *difference;

    if (method != MLkNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CC IS for other method than ML-kNN");

    MALLOC(list, int, subpop_size, "PerformGeneration: list");
    MALLOC(overwrite, bool, subpop_size, "PerformGeneration: selected");
    MALLOC(difference, scalar, 2*subpop_size, "PerformGeneration: difference");
    for (int i = 0; i < subpop_size; i++)
        list[i] = i;

    // Are the factors of the fitness randomly chosen?
    scalar fitness_w, isolated_w, difference_w, substitution_w, reduction_w, onelabel_w;

    if (parm->ml_cc_subpop_random_fitness) {
        scalar w[] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
        int sum = RANDOM(1, 7);

        for (int i = 0; i < sum; i++)
            w[i] = 1.0;
        RandomizeScalarVector(w, 6);

        fitness_w = w[0]/sum;
        isolated_w = w[1]/sum;
        difference_w = w[2]/sum;
        substitution_w = w[3]/sum;
        reduction_w = w[4]/sum;
        onelabel_w = w[5]/sum;
    } else {
        fitness_w = parm->subpop_fitness_w;
        isolated_w = parm->subpop_isolated_w;
        difference_w = parm->subpop_difference_w;
        substitution_w = parm->subpop_substitution_w;
        reduction_w = parm->subpop_reduction_w;
        onelabel_w = parm->subpop_onelabel_w;
    }

    // Random selection without replacement
    RandomizeVector(list, subpop_size);
    int actual_size = subpop_size;

    for (int i = 0; i < subpop_size; i += 2) {
        int father = list[i];
        int mother = list[i+1];

        // If no minimum distance skip
        if (HammingDistance(subpop[mother], subpop[father], instances) < *minimum_distance)
            continue;

        // Perform HUX crossover creating new individuals
        HuxCrossover(subpop[mother], subpop[father], subpop[actual_size], subpop[actual_size+1], instances);

        actual_size += 2;
    }

    // Evaluate new individuals
#ifdef _OPENMP
    #pragma omp parallel for schedule(dynamic)
#endif
    for (int i = subpop_size; i < actual_size; i++)
        fitness[i] = EvaluateSubindividualFitness(parm, subpop[i], subpop_index, population, instances, data, labels, n_labels, classes, inputs, n_subpops, i, pop_fitness, subpops, neighbors_cache, ht,
                     html, &difference[i], qs, index, k, fitness_w, isolated_w, difference_w, substitution_w, reduction_w, onelabel_w, accuracy ? &accuracy[i] : NULL, method);

    // Evaluate old individuals if first time or fitness depend on population individual
    if (first_time || fitness_w > 0.0 || difference_w > 0.0 || substitution_w > 0.0)
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < subpop_size; i++)
            fitness[i] = EvaluateSubindividualFitness(parm, subpop[i], subpop_index, population, instances, data, labels, n_labels, classes, inputs, n_subpops, i, pop_fitness, subpops, neighbors_cache,
                         ht, html, &difference[i], qs, index, k, fitness_w, isolated_w, difference_w, substitution_w, reduction_w, onelabel_w, accuracy ? &accuracy[i] : NULL, method);

    // Memetic mutation?
    if (parm->subpop_drop3_mut > ZERO || parm->subpop_oss_mut > ZERO) {
        // Construct one-label labels
        int *l_labels, count[] = {0, 0};
        MALLOC(l_labels, int, instances, "PerformGeneration: l_labels");
        for (int i = 0; i < instances; i++) {
            l_labels[i] = 0;

            // Check if in list
            for (int j = 0; j < n_labels[i]; j++)
                if (subpop_index == labels[i][j]) {
                    l_labels[i] = 1;
                    break;
                }

            count[l_labels[i]]++;
        }

        int minority = (count[0] < count[1] ? 0 : 1);
#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < actual_size; i++)
            if (i >= subpop_size || !elite[i]) {  // Do not mutate elite members
                if (RANDOM(0.0, 1.0) < parm->subpop_drop3_mut) {
                    Drop3(parm, subpop[i], data, instances, l_labels, parm->filter_noise, NULL, inputs, 2, neighbors_cache, NULL, NULL, index, NULL, false);
                    fitness[i] = EvaluateSubindividualFitness(parm, subpop[i], subpop_index, population, instances, data, labels, n_labels, classes, inputs, n_subpops, i, pop_fitness, subpops,
                                 neighbors_cache, ht, html, &difference[i], qs, index, k, fitness_w, isolated_w, difference_w, substitution_w, reduction_w, onelabel_w, accuracy ? &accuracy[i] : NULL,
                                 method);
                }

                if (RANDOM(0.0, 1.0) < parm->subpop_oss_mut) {
                    OneSidedSelection(subpop[i], data, instances, l_labels, inputs, minority, 1, 2, index, neighbors_cache, true);
                    fitness[i] = EvaluateSubindividualFitness(parm, subpop[i], subpop_index, population, instances, data, labels, n_labels, classes, inputs, n_subpops, i, pop_fitness, subpops,
                                 neighbors_cache, ht, html, &difference[i], qs, index, k, fitness_w, isolated_w, difference_w, substitution_w, reduction_w, onelabel_w, accuracy ? &accuracy[i] : NULL,
                                 method);
                }
            }

        FREE(l_labels);
    }

    // Reinitialize or elitism
    if (actual_size == subpop_size)
        (*minimum_distance)--;
    else {
        // Keep index of selected individuals from previous generation
        for (int i = 0; i < subpop_size; i++)
            overwrite[i] = !elite[i];

        // Complete selected with the best not in elite
        int n_new = 0, n_old = 0;
        for (int current = 0; n_old + n_new < subpop_size; current++) {
            int order = OrderedSelection(fitness, current, actual_size);
            if (order < subpop_size) {
                n_old++;

                if (!elite[order])
                    overwrite[order] = false;
            } else {
                list[n_new] = order;
                n_new++;
            }
        }

        for (int i = 0, next = 0; i < n_new; i++) {
            // Get next to overwrite
            while (!overwrite[next])
                next++;

            int *aux_pt = subpop[list[i]];
            subpop[list[i]] = subpop[next];
            subpop[next] = aux_pt;

            scalar aux_f = fitness[list[i]];
            fitness[list[i]] = fitness[next];
            fitness[next] = aux_f;

            aux_f = accuracy[list[i]];
            accuracy[list[i]] = accuracy[next];
            accuracy[next] = aux_f;
        }
    }

    // If minimum distance 0 reinitialize
    if (*minimum_distance == 0) {
        // Reinitialize populations excepting best one
        int best = OrderedSelection(fitness, 0, subpop_size);
        SHOWPROGRESS(parm->show_progress > 0, "Reinitialize %d subpopulation due to stagnation\n", subpop_index+1);

#ifdef _OPENMP
        #pragma omp parallel for schedule(dynamic)
#endif
        for (int i = 0; i < subpop_size; i++)
            // Reseed changing 35%
            if (i != best && !elite[i]) {
                for (int l = 0; l < instances; l++)
                    subpop[i][l] = RANDOM(0.0, 1.0) < 0.35 ? (subpop[best][l]+1)%2 : subpop[best][l];

                fitness[i] = EvaluateSubindividualFitness(parm, subpop[i], subpop_index, population, instances, data, labels, n_labels, classes, inputs, n_subpops, i, pop_fitness, subpops,
                             neighbors_cache, ht, html, &difference[i], qs, index, k, fitness_w, isolated_w, difference_w, substitution_w, reduction_w, onelabel_w, accuracy ? &accuracy[i] : NULL, method);
            }

        *minimum_distance = instances/4;
    }

    // Average difference
    if (avereage_difference) {
        *avereage_difference = 0.0;
        for (int i = 0; i < subpop_size; i++)
            *avereage_difference += difference[i];
        *avereage_difference /= subpop_size;
    }

    // Free memory
    FREE(difference);
    FREE(list);
    FREE(overwrite);
}

scalar EvaluateSubindividualFitness(const t_Parameters *parm, int *individual, int pop_index, int **population, int instances, scalar **data, int **labels, int *n_labels, int classes, int inputs,
                                    int n_subpops, int individual_index, scalar *pop_fitness, int ***subpops, cache_t **neighbors_cache, t_HashTable *ht, t_HashTable *html, scalar *difference,
                                    scalar *qs, int **index, int *k, scalar fitness_w, scalar isolated_w, scalar difference_w, scalar substitution_w, scalar reduction_w, scalar onelabel_w,
                                    scalar *accuracy, t_MLClassifier method)
{
    scalar fitness = 0.0, isolated = 0.0, substitution = 0.0, reduction = 0.0, one_label = 0.0;

    t_HashEntry new, *old;

    t_MLkNN mlknn;
    t_MLksNN mlksnn;
    void *classifier[N_ML_CLASSIFIERS] = {(void *) &mlknn, NULL, NULL, NULL, (void *) &mlksnn, NULL};

    if (method != MLkNN && method != MLksNN)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "CC instance selection for other than ML-kNN ML-ksNN");

    // Get list of individuals of the population where the subpopulation member appears
    int appearances = 0, *list = NULL;

    if (difference_w > 0.0 || substitution_w > 0.0 || fitness_w > 0.0) {
        MALLOC(list, int, parm->population_size, "EvaluateSubindividualFitness: list");
        for (int i = 0; i < parm->population_size; i++)
            if (population[i][pop_index] == individual_index) {
                list[appearances] = i;
                appearances++;
            }
    }

    if (fitness_w > 0.0) {
        for (int i = 0; i < appearances; i++)
            fitness += pop_fitness[list[i]];

        if (appearances)
            fitness /= appearances;
    }

    // Create key for hash table?
    if (parm->ml_use_hash_table && onelabel_w > 0.0) {
        MALLOC(new.key, char, instances+1, "EvaluateSubindividualFitness: new.key");

        // Check if already evaluated
        for (int i = 0; i < instances; i++)
            new.key[i] = individual[i] ? '1' : '0';
        new.key[instances] = '\0';
    }

    // Evaluation of the accuracy of the individual on its corresponding label
    if (onelabel_w > 0.0 || accuracy) {
        old = NULL;

        if (parm->ml_use_hash_table) {
#ifdef _OPENMP
            #pragma omp critical
#endif
            if ((old = HashSearchEntry(*ht, new.key)) != NULL) {
                one_label = old->ok;
                if (accuracy)
                    *accuracy = old->ok;
            }
        }

        if (!old) {
            one_label = EvaluateSelectionOneLabel(parm, data, instances, labels, inputs, parm->binary_classifier, individual, neighbors_cache, index, parm->k, k, pop_index, population,
                                                  individual_index, n_labels, parm->subpop_perf_measure, NULL, -1);

            if (accuracy)
                *accuracy = one_label;

            if (parm->ml_use_hash_table) {
                new.ok = one_label;
#ifdef _OPENMP
                #pragma omp critical
#endif
                HashAddEntry(ht, &new);
            }
        }
    }

    // Evaluation of the accuracy of the individual as an isolated selection
    if (isolated_w > 0.0) {
        scalar performance;
        isolated = EvaluateIndividualFitness(parm, NULL, NULL, 0, instances, labels, n_labels, classes, inputs, data, individual, neighbors_cache, html, NULL, index, &performance, NULL, NULL, method,
                                             classifier[method], NULL, 0, NULL, parm->mlknn_k, NULL, 0, NULL, NULL, instances);
    }

    // Average difference when individual is removed
    if (difference_w > 0.0 || substitution_w > 0.0) {
        scalar with, without, substituted;
        for (int i = 0; i < appearances; i++) {
            EvaluateIndividualFitness(parm, population[list[i]], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, html, &qs[list[i]], index,
                                      &with, NULL, NULL, method, classifier[method], NULL, 0, NULL, k[list[i]], NULL, 0, NULL, NULL, instances);

            if (difference_w > 0.0) {
                population[list[i]][pop_index] = -1;
                EvaluateIndividualFitness(parm, population[list[i]], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, html, &qs[list[i]],
                                          index, &without, NULL, NULL, method, classifier[method], NULL, 0, NULL, k[list[i]], NULL, 0, NULL, NULL, instances);
                population[list[i]][pop_index] = individual_index;

                *difference += (with - without);
            }

            // Average difference when individual is substituted by another randomly chosen
            if (substitution_w > 0.0) {
                population[list[i]][pop_index] = RANDOM(0, parm->subpop_size);
                EvaluateIndividualFitness(parm, population[list[i]], subpops, n_subpops, instances, labels, n_labels, classes, inputs, data, NULL, neighbors_cache, html, &qs[list[i]],
                                          index, &substituted, NULL, NULL, method, classifier[method], NULL, 0, NULL, k[list[i]], NULL, 0, NULL, NULL, instances);
                population[list[i]][pop_index] = individual_index;

                substitution += (with - substituted);
            }
        }

        if (appearances) {
            *difference /= appearances;
            substitution /= appearances;
        }
    } else
        *difference = 1.0; // For selection of sub-population to evolve

    if (reduction_w) {
        for (int i = 0; i < instances; i++)
            if (individual[i])
                reduction++;

        reduction = 1.0 - reduction/instances;
    }

    if (list)
        FREE(list);

    return fitness_w*fitness + isolated_w*isolated + difference_w*(*difference) + substitution_w*substitution + reduction_w*reduction + onelabel_w*one_label;
}

int CreateSelectionFromSubpopulations(const t_Parameters *parm, int *individual, int ***subpops, int n_subpops, int *selection, int instances, scalar quorum_sensing_threshold, scalar **w)
{
    int selected = 0;
    scalar qs = 0.5, total = 0.0;    // By default, 50% for majority voting

    ZeroVector(NULL, selection, instances);

    // Different alternatives: AND, OR and Majority voting
    switch (parm->merge_method) {
        case And:
            for (int i = 0; i < instances; i++) {
                bool all = true;
                for (int j = 0; j < n_subpops; j++)
                    if (individual[j] >= 0 && !subpops[j][individual[j]][i]) {
                        all = false;
                        break;
                    }

                selected += (selection[i] == all ? 1 : 0);
            }
            break;

        case Or:
            for (int i = 0; i < instances; i++)
                for (int j = 0; j < n_subpops; j++)
                    if (individual[j] >= 0 && subpops[j][individual[j]][i]) {
                        selection[i] = 1;
                        selected++;
                        break;
                    }
            break;

        case QuorumSensing:
            qs = quorum_sensing_threshold;

        case MajorityVoting:
            if (!w)
                total = n_subpops;
            else
                for (int i = 0; i < n_subpops; i++)
                    total += individual[i] >= 0 ? w[i][individual[i]] : 0.0;
            for (int i = 0; i < instances; i++) {
                scalar count = 0.0;
                for (int j = 0; j < n_subpops && count < total*qs; j++)
                    if (individual[j] >= 0 && subpops[j][individual[j]][i])
                        count += w ? w[j][individual[j]] : 1.0;

                selected += (selection[i] = (count >= total*qs ? 1 : 0));
            }
            break;
    }

    return selected;
}


void InitializeSubpopulations(const t_Parameters *parm, int ****subpops, int n_subpops, bool **first_time, int instances, int **labels, int *n_labels)
{
    MALLOC(*subpops, int **, n_subpops, "InitializeSubpopulations: *subpops");
    MALLOC(*first_time, bool, n_subpops, "InitializeSubpopulations: *first_time");

    // Initialize subpopulations using undersampling?
    if (parm->ml_init_with_undersampling) {
        // Construct one-label labels
        int *l_labels;
        MALLOC(l_labels, int, instances, "InitializeSubpopulations: l_labels");

        for (int i = 0; i < n_subpops; i++) {
            int count[] = {0, 0};

            for (int j = 0; j < instances; j++) {
                l_labels[j] = LabelInList(i, labels[j], n_labels[j]) ? 1 : 0;
                count[l_labels[j]]++;
            }

            int minority = count[0] < count[1] ? 0 : 1;
            scalar ratio = ((scalar) count[minority])/count[(minority+1)%2];

            (*first_time)[i] = true;   // First time all individuals must be evaluated
            MALLOC((*subpops)[i], int *, 2*parm->subpop_size, "InitializeSubpopulations: (*subpops)[i]");  // Space for duplication of population during evolution

            for (int j = 0; j < parm->subpop_size; j++) {
                MALLOC((*subpops)[i][j], int, instances, "InitializeSubpopulations: (*subpops)[i][j]");
                for (int k = 0; k < instances; k++)
                    (*subpops)[i][j][k] = l_labels[k] == minority ? 1 : (RANDOM(0.0, 1.0) < ratio ? 1 : 0);
            }

            for (int j = parm->subpop_size; j < 2*parm->subpop_size; j++)
                MALLOC((*subpops)[i][j], int, instances, "InitializeSubpopulations: (*subpops)[i][j]");
        }

        FREE(l_labels);
    } else {
        // Undersampled subpopulations?
        if (parm->ml_evolve_undersampled_subpops) {
            // Construct one-label labels
            int *l_labels, *list_0, *list_1;
            MALLOC(l_labels, int, instances, "InitializeSubpopulations: l_labels");
            MALLOC(list_0, int, instances, "InitializeSubpopulations: list_0");
            MALLOC(list_1, int, instances, "InitializeSubpopulations: list_1");

            for (int i = 0; i < n_subpops; i++) {
                int count[] = {0, 0};

                for (int j = 0; j < instances; j++) {
                    l_labels[j] = LabelInList(i, labels[j], n_labels[j]) ? 1 : 0;
                    if (l_labels[j])
                        list_1[count[1]] = j;
                    else
                        list_0[count[0]] = j;
                    count[l_labels[j]]++;
                }

                int minority = count[0] < count[1] ? 0 : 1;

                // Randomize majority class
                RandomizeVector(minority == 0 ? list_1 : list_0, count[(minority+1)%2]);

                (*first_time)[i] = true;   // First time all individuals must be evaluated
                MALLOC((*subpops)[i], int *, 2*parm->subpop_size, "InitializeSubpopulations: (*subpops)[i]");  // Space for duplication of population during evolution

                for (int j = 0; j < parm->subpop_size; j++) {
                    CALLOC((*subpops)[i][j], int, instances, "InitializeSubpopulations: (*subpops)[i][j]");
                    for (int k = 0; k < count[minority]; k++) {
                        // Minority and majority classes class
                        (*subpops)[i][j][list_0[k]] = (RANDOM(0.0, 1.0) < parm->chc_subpop_init_prob ? 1 : 0);
                        (*subpops)[i][j][list_1[k]] = (RANDOM(0.0, 1.0) < parm->chc_subpop_init_prob ? 1 : 0);
                    }
                }

                for (int j = parm->subpop_size; j < 2*parm->subpop_size; j++)
                    MALLOC((*subpops)[i][j], int, instances, "InitializeSubpopulations: (*subpops)[i][j]");
            }

            FREE(l_labels);
            FREE(list_0);
            FREE(list_1);
        } else {
            for (int i = 0; i < n_subpops; i++) {
                (*first_time)[i] = true;   // First time all individuals must be evaluated
                MALLOC((*subpops)[i], int *, 2*parm->subpop_size, "InitializeSubpopulations: (*subpops)[i]");  // Space for duplication of population during evolution

                for (int j = 0; j < parm->subpop_size; j++) {
                    MALLOC((*subpops)[i][j], int, instances, "InitializeSubpopulations: (*subpops)[i][j]");
                    for (int k = 0; k < instances; k++)
                        (*subpops)[i][j][k] = RANDOM(0.0, 1.0) < parm->chc_subpop_init_prob ? 1 : 0;
                }

                for (int j = parm->subpop_size; j < 2*parm->subpop_size; j++)
                    MALLOC((*subpops)[i][j], int, instances, "InitializeSubpopulations: (*subpops)[i][j]");
            }
        }
    }
}
