/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <string.h>
#include "ml.h"
#include "macros.h"
#include "classifiers.h"
#include "typex.h"
#include "ensemble.h"

// Two-Stage voting method
// From: G. Madjarov, D. Gjorgjevikj, S. Džeroski (2012) "Two stage architecture for multi-label learning," Pattern Recognition 45:1019–1034.

void TwoStageVotingMethodClassifyInstance(const t_Parameters *parm, scalar *query, int *h, scalar *f, bool *classification, bool *ranks, scalar **data, int instances, int n_labels,
        int *n_predicted, int inputs, int *query_index, t_CLR *clr, int *subspace, int **index, bool cc, bool pruned_cc)
{
    scalar f_virtual = 0.0, output, probabilities[n_labels], *q = NULL;
    int *q_index = NULL, count = -1, *s = NULL;

    ZeroVector(f, NULL, n_labels);

    if (cc || pruned_cc) {
        if (subspace) {
            MALLOC(s, int, inputs+n_labels, "TwoStageVotingMethodClassifyInstance: s");
            memcpy(s, subspace, sizeof(int)*inputs);
            for (int i = inputs; i < inputs+n_labels; i++)
                s[i] = 1;
        }
        // Sparse
        if (query_index) {
            for (count = 0; query_index[count] != -1; count++)
                ;
            MALLOC(q, scalar, count+n_labels, "ClassifierChainClassifyInstance: q");
            memcpy(q, query, sizeof(scalar)*count);
            MALLOC(q_index, int, count+1+n_labels, "ClassifierChainClassifyInstance: q_index");
            memcpy(q_index, query_index, sizeof(int)*(count+1));
        } else {    // Dense
            MALLOC(q, scalar, inputs+n_labels, "ClassifierChainClassifyInstance: q");
            memcpy(q, query, sizeof(scalar)*inputs);
        }
    } else {
        q_index = query_index;
        q = query;
        s = subspace;
    }

    // Output for virtual labels
    for (int j = 0; j < n_labels; j++) {
        int category = Classifier2ClassOutput(parm, clr->method, q, inputs, 2, q_index, subspace, NULL, instances, NULL, NULL, &(clr->models[0][j]), 1, &probabilities[j], index,
                                              &(clr->ensembles[0][j]));

        if (category)
            f[j]++;
        else
            f_virtual++;

        if (cc) {
            if (!query_index)
                q[inputs+j]= probabilities[j];
            else {
                if (fabs(probabilities[j]) > ZERO) {
                    q[count] = probabilities[j];
                    q_index[count] = inputs+j+1;
                    q_index[++count] = -1;
                }
            }
        }
    }

    // Output of all binary classifiers
    int n = 0;
    int length = inputs;
    if (cc)
        length += n_labels;
    else if (pruned_cc)
        length += 2;
    for (int i = 1; i < n_labels; i++)
        for (int j = 0; j < i; j++)
            if (clr->mask[n++] && probabilities[i] > parm->tsa_t && probabilities[j] > parm->tsa_t) {
                if (pruned_cc) {
                    if (!query_index) {
                        q[inputs]= probabilities[j];
                        q[inputs+1] = probabilities[i];
                    } else {
                        int n_count = count;
                        if (fabs(probabilities[j]) > ZERO) {
                            q[n_count] = probabilities[j];
                            q_index[n_count] = inputs+1;
                            q_index[++n_count] = -1;
                        }
                        if (fabs(probabilities[i]) > ZERO) {
                            q[n_count] = probabilities[i];
                            q_index[n_count] = inputs+2;
                            q_index[++n_count] = -1;
                        }
                    }
                }

                int category = Classifier2ClassOutput(parm, clr->method, q, length, 2, q_index, s, NULL, instances, NULL, NULL, &(clr->models[i][j]), 1, &output, index, &(clr->ensembles[i][j]));

                if (category)
                    f[i]++;
                else
                    f[j]++;
            } else if (probabilities[i] > parm->tsa_t)
                f[i]++;
            else if (probabilities[j] > parm->tsa_t)
                f[j]++;

    // Relevant labels those above votes of virtual label
    *n_predicted = 0;
    for (int i = 0; i < n_labels; i++)
        if (f[i] > f_virtual) {
            h[*n_predicted] = i;
            (*n_predicted)++;
        }

    *classification = true;

    // Normalize probabilities
    NormalizeVector(f, n_labels);

    *ranks = true;

    if (cc) {
        FREE(q);
        if (query_index)
            FREE(q_index);
        if (subspace)
            FREE(s);
    }
}
