/***************************************************************************
 *   Copyright (C) 2009 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*
*Bias/Variance main program
 */
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "macros.h"
#include "common.h"

/*
*Arguments to main function:
 *+#1+Problem
 *+#2+Parameters file
 */

int main(int argc, char *argv[])
{
    int *order, instances, inputs, classes, *label, class, classifier, **L, *C0, g_instances, t_instances;
    scalar **prob, **data, bias, variance, bias_x, var_x, C, *it_bias, *it_variance, max_value, average_g_error;
    time_t init_time, end_time;
    char file_name[FILENAME_LENGTH], *home, gl_datad[MAX_FILENAME_LENGTH];
    FILE *out_file;

    // Check arguments
    if (argc < 8) {
        printf("biasvar dataset data-directory parameters iterations partitions output program\n");
        exit(1);
    }

    strcpy(gl_datad, argv[2]);

    // File for output messages.
    if ((out_file = fopen(argv[6], "w")) == NULL)
        OnErrorExit(1, "main: Cannot create output file\n");

    int iterations = atoi(argv[4]);
    int partitions = atoi(argv[5]);

    // Time
    init_time = time(NULL);

    srandom(-time(NULL));

    // Load+data
    home = getenv("HOME");
    sprintf(file_name, "%s.all", argv[1]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s.all", home, gl_datad, argv[1]);
    ReadDataAndLabels(&data, &label, &instances, &inputs, &classes, file_name, false);

    // Initialize data needed for bias/variance
    MALLOC(prob, scalar *, instances, "main: prob");
    MALLOC(L, int *, instances, "main: L");
    for (int i = 0; i < instances; i++) {
        MALLOC(prob[i], scalar, classes, "main: prob[i]");
        MALLOC(L[i], int, partitions*iterations, "main: L[i]");
        for (int j = 0; j < classes; j++)
            prob[i][j] = 0.0;
        for (int j = 0; j < partitions*iterations; j++)
            L[i][j] = -1;
    }

    // Allocate memory for generalization data
    g_instances = instances/partitions;
    t_instances = instances-g_instances;

    // Main loop of bias/variance obtention
    classifier = 0;
    average_g_error = 0.0;
    MALLOC(order, int, instances, "main: order");

    // Get random file name
    char *template = tempnam(".", NULL);

    for (int iter = 0; iter < iterations; iter++) {
        // Make partition
        GetRandomisedVector(order, instances);

        // Try all partitions
        for (int part = 0; part < partitions; part++) {
            FILE *f;

            // Set current partition
            // Write data files
            // Training data
            int index = part*g_instances;

            sprintf(file_name, "%s.train", template);
            if ((f = fopen(file_name, "w")) == NULL)
                OnErrorExit(ERR_FILE, "Cannot open temporal file for training");
            fprintf(f, "[ONE LABEL]\n$ %d\n$ %d\n$ %d\n", t_instances, inputs, classes);

            for (int i = 0; i < t_instances; i++) {

                for (int j = 0; j < inputs; j++)
                    fprintf(f, "%lf ", data[order[index]][j]);
                fprintf(f, "%d\n", label[order[index]]);

                index = (index + 1) % instances;
            }

            fclose(f);

            sprintf(file_name, "%s.gen", template);
            if ((f = fopen(file_name, "w")) == NULL)
                OnErrorExit(ERR_FILE, "Cannot open temporal file for testing");
            fprintf(f, "[ONE LABEL]\n$ %d\n$ %d\n$ %d\n", g_instances, inputs, classes);

            // Test data
            for (int i = 0; i < g_instances; i++) {
                for (int j = 0; j < inputs; j++)
                    fprintf(f, "%lf ", data[order[index]][j]);
                fprintf(f, "%d\n", label[order[index]]);

                index = (index + 1) % instances;
            }

            fclose(f);

            // Run ensemble
            char command[1000];
            sprintf(command, "%s/%s/%s %s %s %s %s.out", home, "Run/bin", argv[7],  template, argv[3], "/dev/null", template);
            SHOWPROGRESS(false, "Running program: %s\n", command);
            int status = WEXITSTATUS(system(command));

            // If any error ignore this run and reshuffle
            if (status != 0) {
                GetRandomisedVector(order, instances);
                part--;
                continue;
            }

            // Get results
            sprintf(file_name, "%s.out", template);
            if ((f = fopen(file_name, "r")) == NULL)
                OnErrorExit(ERR_FILE, "Cannot open temporal output file");

            // Get info for calculating bias/variance
            // Test ensemble in generalization and store results
            char line[100];
            scalar error = 0.0;

            index = (part*g_instances + t_instances) % instances;
            for (int i = 0; i < g_instances; i++) {
                // Classify pattern
                if (fgets(line, 100, f) == NULL)
                    OnErrorExit(1, "Class info file error");
                class = atoi(line);
                prob[order[index]][class] += 1.0/iterations;
                L[order[index]][classifier] = class;

                if (class != label[order[index]])
                    error ++;

                index = (index + 1) % instances;
            }

            classifier++;

            fclose(f);

            // Average generalization error
            error /= g_instances;
            average_g_error += error;

            SHOWPROGRESS(true, "FINISHED %*d/%d of %*d/%d\r", WIDTH(partitions), part+1, partitions, WIDTH(iterations), iter+1, iterations);
            fflush(stdout);
        }
    }

    printf("\n");

    // Average generalization error
    fprintf(out_file, "Average generalization error: %8.6lf\n", average_g_error/classifier);

    // Bias/variance calculation
    bias = variance = 0.0;
    // [Kohavi, 96]
    for (int i = 0; i < instances; i++) {
        bias_x = var_x = 0.0;

        for (int j = 0; j < classes; j++) {
            // Bias
            C = (label[i] == j) ? 1.0 : 0.0;
            bias_x += (C-prob[i][j]) * (C-prob[i][j]);

            // Variance
            var_x += prob[i][j]*prob[i][j];
        }

        bias += 0.5*bias_x;
        variance += 0.5* (1.0-var_x);
    }

    bias /= instances;
    variance /= instances;
    fprintf(out_file, "KOHAVI\tBias: %8.6lf\tVariance: %8.6lf\t(%8.6lf)\n", bias, variance, bias + variance);

    // Estimate central tendency
    MALLOC(C0, int, instances, "main: C0");
    bias = variance = 0.0;
    for (int i = 0; i < instances; i++) {
        // Get central tendency of pattern i-th
        C0[i] = 0;
        max_value = prob[i][0];
        for (int j = 1; j < classes; j++)
            if (prob[i][j] > max_value) {
                max_value = prob[i][j];
                C0[i] = j;
            }
    }

    // [Breiman, ]
    MALLOC(it_bias, scalar, iterations*partitions, "main: it_bias");
    MALLOC(it_variance, scalar, iterations*partitions, "main: it_variance");
    for (int i = 0; i < iterations*partitions; i++) {
        it_bias[i] = it_variance[i] = 0.0;
        for (int j = 0; j < instances; j++)
            if (L[j][i] != -1) {
                if (label[j] != L[j][i] && L[j][i] == C0[j])
                    it_bias[i]++;
                if (label[j] != L[j][i] && L[j][i] != C0[j])
                    it_variance[i]++;
            }
        variance += (it_variance[i] /= g_instances);
        bias += (it_bias[i] /= g_instances);
    }
    bias /= iterations*partitions;
    variance /= iterations*partitions;
    fprintf(out_file, "BREIMAN\tBias: %8.6lf\tVariance: %8.6lf\t(%8.6lf)\n", bias, variance, bias + variance);

    // [Kong, ]
    // Bias
    bias = variance = 0.0;
    for (int j = 0; j < instances; j++)
        if (label[j] != C0[j])
            bias++;
    bias /= instances;

    // Variance
    for (int i = 0; i < iterations*partitions; i++) {
        it_variance[i] = 0.0;
        for (int j = 0; j < instances; j++)
            if (L[j][i] != -1 && label[j] != L[j][i])
                it_variance[i]++;

        it_variance[i] = it_variance[i]/g_instances-bias;
        variance += it_variance[i];
    }
    variance /= iterations*partitions;
    fprintf(out_file, "KONG\tBias: %8.6lf\tVariance: %8.6lf\t(%8.6lf)\n", bias, variance, bias + variance);

    end_time = time(NULL)-init_time;
    fprintf(stdout, "Elapsed time: %02ld:%02ld:%02ld\n", end_time/3600,
            (end_time % 3600) /60, end_time % 60);
    fprintf(out_file, "Elapsed time: %02ld:%02ld:%02ld\n", end_time/3600,
            (end_time % 3600) /60, end_time % 60);

    fclose(out_file);

    // Remove temporal files
    sprintf(file_name, "%s.train", template);
    remove(file_name);
    sprintf(file_name, "%s.gen", template);
    remove(file_name);
    sprintf(file_name, "%s.out", template);
    remove(file_name);

    return 0;
}
