/***************************************************************************
 *   Copyright (C) 2020 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// Implements most algorithms for single-label datasets

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <stdbool.h>
#include <string.h>
#include <omp.h>
#include "macros.h"
#include "common.h"
#include "typex.h"
#include "isfs.h"
#include "data.h"
#include "external.h"
#include "classifiers.h"
#include "maths.h"
#include "imbalance.h"
#include "fs.h"

// Global variables for C4.5 algorithm
unsigned short MaxClass, MaxDiscrVal = 2;
ItemNo MaxItem;
int MaxAtt;
Description *Item;
DiscrValue *MaxAttVal;
char *SpecialStatus;
String *ClassName, *AttName, * *AttValName, FileName;
short VERBOSITY, TRIALS;
Boolean GAINRATIO = true, SUBSET = false;
ItemNo MINOBJS = 2, WINDOW, INCREMENT;
float ConfLimit = 0.25;
Tree *Pruned;
Boolean AllKnown = true;

#define NEIGHBORS_CACHE_SIZE(instances) ((sizeof(cache_t) * (instances)*(instances)+sizeof(cache_t *) * (instances))/pow(2.0, 30.0))
#define k_NEIGHBORS_CACHE_SIZE(k, instances) ((sizeof(cache_t) * (k)*(instances)+sizeof(cache_t *) * (instances))/pow(2.0, 30.0))
#define DISTANCE_CACHE_SIZE(instances) ((sizeof(scalar) * (instances)*(instances)+sizeof(scalar *) * (instances))/pow(2.0, 30.0))

#define NODES(f, i) (parm.feature_selection*(f) + parm.instance_selection*(i))

#define MINIMUM_STRATA_TO_SPAWN 5

int main(int argc, char *argv[])
{
    time_t init_time;
    char file_name[FILENAME_LENGTH], *home;
    scalar reduction = 0.0, **train_data, **test_data;
    int *S = NULL, inputs, classes, train_instances, *train_labels, *test_labels, test_instances, **train_index = NULL, selected_i = 0, **test_index = NULL, n_slaves = 0, selected_f = 0,
         minority = -1;
    t_Parameters parm;

    // Continue?
    int opt, shitf = 0;
    bool continue_from_selection = false, remove_output = true;
    while ((opt = getopt(argc, argv, "cn")) != EOF)
        switch (opt) {
            case 'c':
                // Show progress
                continue_from_selection = true;
                shitf++;
                break;

            case 'n':
                // Remove output
                remove_output = false;
                shitf++;
                break;
        }

    argv += shitf;
    argc -= shitf;

    // Test enough parameters
    char f_output[FILENAME_LENGTH];

    if (argc >= 4) {
        strcpy(f_output, argv[3]);
        if (continue_from_selection)
            strcat(f_output, ".temp");

        // File for output messages.
        if ((parm.out_file = fopen(f_output, remove_output ? "w" : "a")) == NULL)
            OnErrorExit(ERR_FILE, "single-label: Cannot open output file\n");

        LoadGlobalParameters(&parm, argv[2], "single-label");
    } else {
        printf("single-label [-c] [-k] problem parameters output [slaves] \n");
        exit(1);
    }

    if (argc == 5) {
        n_slaves = atoi(argv[4]);
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Parallel execution for single-label");
        SHOWPROGRESS(true, "Parallel execution IS/FS and k-NN evaluation with %d slaves\n", n_slaves);
#ifdef _OPENMP
        OnErrorExit(ERR_PARAM, "OPEN MP and PVM are incompatible, exiting\n");
#endif
    }

    // Initialize random number generator
    char state[256];
    if (parm.seed == 0) {
        FILE *randDev;

        if ((randDev = fopen("/dev/urandom", "r")) == NULL)
            OnErrorExit(ERR_FILE, "main: Cannot open random device\n");

        if (!fread(&parm.seed, 1, sizeof(int), randDev))
            OnErrorExit(ERR_READ, "main");
        fclose(randDev);
        fprintf(parm.out_file, "Random seed: %d\n", parm.seed);
        fflush(parm.out_file);
    } else {
        // Fixed seed to make the evolution reproducible.
        fprintf(parm.out_file, "Random seed: %d (fixed)\n", parm.seed);
    }

    srandom(parm.seed);
    srand(parm.seed);
    for (int i = 0; i < 256; i++)
        state[i] = random();
    initstate(parm.seed, state, 256);
    setstate(state);

    // Time
    init_time = time(NULL);

    // Load data
    train_data = test_data = NULL;
    home = getenv("HOME");
    sprintf(file_name, "%s.train", argv[1]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s.train", home, parm.datad, argv[1]);
    if (parm.sparse_format) {
        // Read sparse data and check whether it uses less space than dense format
        scalar sparse_size = ReadDataAndLabelsInSparseFormat(&train_data, &train_labels, &train_instances, &inputs, &classes, file_name, &train_index, parm.show_progress);
        scalar dense_size = (((scalar) train_instances)*inputs*sizeof(scalar) + ((scalar) train_instances)*sizeof(scalar *))/pow(2.0, 30.0);

        // Switch to dense?
        SHOWPROGRESS(parm.show_progress > 0, "Sparse size: %.5lf GB\nDense size:  %.5lf GB\n", sparse_size, dense_size);
        if (sparse_size >= dense_size && !parm.force_sparse_format) {
            SHOWPROGRESS(parm.show_progress > 0, "Switching to dense format%c", '\n');
            parm.sparse_format = false;

            // Free memory and load file in standard format
            for (int i = 0; i < train_instances; i++) {
                if (train_data[i])
                    FREE(train_data[i]);
                FREE(train_index[i]);
            }
            FREE(train_labels);
            FREE(train_data);
            FREE(train_index);

            ReadDataAndLabels(&train_data, &train_labels, &train_instances, &inputs, &classes, file_name, parm.huge_dataset);
        }
    } else
        ReadDataAndLabels(&train_data, &train_labels, &train_instances, &inputs, &classes, file_name, parm.huge_dataset);
    SortClasses(train_data, train_instances, train_labels, classes, parm.eval_folds, train_index);

    sprintf(file_name, "%s.gen", argv[1]);
    if (access(file_name, R_OK))
        sprintf(file_name, "%s/%s/%s.gen", home, parm.datad, argv[1]);
    if (parm.sparse_format)
        ReadDataAndLabelsInSparseFormat(&test_data, &test_labels, &test_instances, &inputs, &classes, file_name, &test_index, parm.show_progress);
    else {
        ReadDataAndLabels(&test_data, &test_labels, &test_instances, &inputs, &classes, file_name, parm.huge_dataset);
        RemoveConstantFeatures(train_data, train_instances, &inputs, test_data, test_instances, parm.out_file);
    }

    // Sort instances by label to allow a better distribution for some tasks
    StratifyDataset(train_data, train_instances, train_labels, train_index, classes);

    // Duplicate inputs for redundancy testing?
    if (parm.duplicate_features) {
        if (train_index)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Duplicate inputs for sparse format");
        DuplicateFeatures(train_data, train_instances, inputs, parm.duplicate_features_ratio);
        inputs = DuplicateFeatures(test_data, test_instances, inputs, parm.duplicate_features_ratio);
    }

    // Duplicate instances for redundancy testing?
    if (parm.duplicate_instances) {
        if (train_index)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Duplicate instances for sparse format");

        train_instances *= 2;

        REALLOC(train_data, scalar *, train_instances, "main: train_data");
        REALLOC(train_labels, int, train_instances, "main: train_labels");
        for (int i = train_instances/2; i < train_instances; i++) {
            MALLOC(train_data[i], scalar, inputs, "main: train_data[i]");
            for (int j = 0; j < inputs; j++)
                train_data[i][j] = train_data[i-train_instances/2][j];
            train_labels[i] = train_labels[i-train_instances/2];
        }
    }

    // Typify input training and testing data
    if (parm.typify && !parm.sparse_format)
        TypifyInputData(train_data, test_data, train_instances, test_instances, inputs);
    else if (parm.typify && parm.sparse_format)
        TypifySparseData(train_data, test_data, train_instances, test_instances, inputs, train_index, test_index);

    // Add noise to duplicated instances?
    if (parm.duplicate_instances_with_noise) {
        for (int i = train_instances; i < train_instances/2; i++)
            for (int j = 0; j < inputs; j++)
                train_data[i][j] += Normal(0.0, 1);
    }

    // Two classes?
    minority = -1;
    if (classes == 2) {
        int ccount[2] = {0, 0};

        // Count classes
        for (int i = 0; i < train_instances; i++)
            ccount[train_labels[i]]++;

        minority = (ccount[0] > ccount[1]) ? 1 : 0;
    }

    // Continue from previous feature selection not completely tested?
    if (continue_from_selection)
        OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Continue from selection for single-label");

    // Class-imbalance method?
    if (classes == 2 && parm.imbalance_method != iNone) {
        if (train_index)
            OnErrorExit(ERR_NOT_YET_IMPLEMENTED, "Imbalance methods for sparse format");
        ClassImbalanceMethodFor2Class((const t_Parameters *) &parm, &train_data, &train_instances, inputs, classes, &train_labels);
    }

    MALLOC(S, int, train_instances+inputs, "main: S");
    for (int i = 0; i < train_instances+inputs; i++)
        S[i] = 1;

    // If feature ranking nothing else is done
    if (parm.feature_ranking) {
        //////////////////
        // Feature ranking
        //////////////////
        t_RankInfo *ranking;
        MALLOC(ranking, t_RankInfo, inputs, "main: ranking");
        time_t init_time = time(NULL);

        FeatureRankingAlgorithm((const t_Parameters *) &parm, parm.fralgorithm, inputs, train_data, train_labels, classes, train_instances, minority, ranking, NULL, NULL, train_index);

        fprintf(parm.out_file, "@Feature ranking time: %ld\n", time(NULL) - init_time);

        // Show sorted ranking list
        qsort(ranking, inputs, sizeof(t_RankInfo), CompareRankingScore);

        fprintf(parm.out_file, "Feature ranks and scores\n");
        fprintf(parm.out_file, "Rank\tFeature\tScore\n");
        for (int i = 0; i < inputs; i++)
            fprintf(parm.out_file, "%*d\t%*d\t%lf\n", WIDTH(inputs), i+1, WIDTH(inputs), ranking[i].feature+1, ranking[i].score);

        // Used time so far
        time_t end_time = time(NULL) - init_time;
        SHOWPROGRESS(parm.show_progress > 0, "Elapsed time: %02ld:%02ld:%02ld\n", end_time/3600, (end_time % 3600)/60, end_time % 60);

        // Random undersampling for training?
        if (parm.random_undersampling_for_training && !parm.sparse_format)
            UndersampleData(&train_data, &train_instances, inputs, classes, &train_labels, parm.random_undersampling_ratio);

        int *M_test[classes], *subspace;
        MALLOC(subspace, int, inputs, "main: subspace");

        for (int i = 0; i < classes; i++)
            MALLOC(M_test[i], int, classes, "main: M_test[i]");

        // Performance info
        t_PerformanceInfo **perf;
        int *size;
        MALLOC(size, int, inputs, "main: size");
        MALLOC(perf, t_PerformanceInfo *, inputs, "main: perf");
        for (int i = 0; i < inputs; i++)
            MALLOC(perf[i], t_PerformanceInfo, N_CLASSIFIERS+N_EXTERNAL, "main: perf");

        ZeroVector(NULL, subspace, inputs);
        for (int i = 0; i < inputs; i++) {
            subspace[ranking[i].feature] = 1;

            // Show list of selected inputs
            fprintf(parm.out_file, "Selected inputs (%*d): ", WIDTH(inputs), i+1);
            for (int j = 0; j <= i; j++)
                fprintf(parm.out_file, "%d ", ranking[j].feature+1);
            fprintf(parm.out_file, "\n");

            SHOWPROGRESS(parm.show_progress > 0, "Features %*d/%d\r", WIDTH(inputs), i+1, inputs);

            TestAllClassifiers((const t_Parameters *) &parm, train_data, train_instances, train_labels, classes, test_data, test_instances, test_labels, inputs, subspace, NULL, argv[3], minority,
                               parm.evaluate_train_error, perf[i], NULL, NULL, train_index, test_index, NULL);
        }

        // Do nothing else
        exit(1);
    }

    // Instance selection?
    int generations, neighbors_cache_size = 0;
    scalar evaluations;
    cache_t **neighbors_cache = NULL;
    scalar **distance_cache = NULL;

    if (parm.instance_selection && parm.use_neighbors_cache && NEIGHBORS_CACHE_SIZE(train_instances) + DISTANCE_CACHE_SIZE(train_instances) < MAXIMUM_CACHE_SIZE)
        ConstructAllNeighborsCache(train_data, train_instances, inputs, &distance_cache, &neighbors_cache, NULL, train_index, &neighbors_cache_size, NULL, parm.evaluate_with_loo);

    if (parm.instance_selection && !parm.feature_selection)
        InstanceSelectionAlgorithm((const t_Parameters *) &parm, parm.isalgorithm, &S, &train_data, &train_instances, inputs, &train_labels, classes, train_index, neighbors_cache, distance_cache,
                                   minority, &generations, parm.randomsel_ratio, NULL, NULL, &evaluations, NULL, false, NULL, false, NULL);

    // Feature selection?
    if (!parm.instance_selection && parm.feature_selection)
        FeatureSelection((const t_Parameters *) &parm, S+train_instances, inputs, train_data, train_labels, classes, train_instances, minority, train_index);

    // Simultaneous instance and feature selection?
    scalar ratio = 1.0;
    if (parm.instance_selection && parm.feature_selection)
        InstanceAndFeatureSelectionAlgorithm((const t_Parameters *) &parm, &S, &train_data, &train_instances, inputs, &train_labels, &generations, &evaluations, classes, minority, S+train_instances,
                                             train_index, &ratio);

    selected_f = selected_i = 0;
    for (int i = 0; i < train_instances; i++)
        if (S[i])
            selected_i++;

    if (!selected_i) {
        S[(int) RANDOM(0, train_instances)] = 1;
        selected_i = 1;
    }

    for (int i = 0; i < inputs; i++)
        if (S[i+train_instances])
            selected_f++;

    if (!selected_f) {
        S[((int) RANDOM(0, inputs)) + train_instances] = 1;
        selected_f = 1;
    }

    fprintf(parm.out_file, "@Number of selected instances: %d\n", selected_i);
    fprintf(parm.out_file, "@Number of selected inputs: %d\n", selected_f);

    if (parm.feature_selection) {
        fprintf(parm.out_file, "Selected inputs (%d/%d): ", selected_f, inputs);
        for (int i = 0; i < inputs; i++)
            if (S[train_instances+i])
                fprintf(parm.out_file, "%d ", i+1);
        fprintf(parm.out_file, "\n");
    }

    reduction = 1.0 - (((scalar) selected_i)/train_instances)*(((scalar) selected_f)/inputs);
    fprintf(parm.out_file, "@Reduction: %lf\n", reduction);
    fprintf(parm.out_file, "@Instance reduction: %lf\n", 1.0 - ((scalar) selected_i)/train_instances);
    fprintf(parm.out_file, "@Feature reduction: %lf\n", 1.0 - ((scalar) selected_f)/inputs);

    time_t evaluation_time = time(NULL);

    // Test results
    TestAllClassifiers((const t_Parameters *) &parm, train_data, train_instances, train_labels, classes, test_data, test_instances, test_labels, inputs, S+train_instances, S, argv[3], minority,
                       parm.evaluate_train_error, NULL, NULL, NULL, train_index, test_index, NULL);

    // Print token
    fprintf(parm.out_file, "Evaluation time: %ld\n", time(NULL) - evaluation_time);
    fprintf(parm.out_file, "@Total time: %ld\n", time(NULL) - init_time);

    // Redundancy measures
    if (!parm.sparse_format && parm.feature_selection) {
        scalar Wr, MI, average;
        RedundancyMeasureOfSeletedFeatures((const t_Parameters *) &parm, test_data, test_instances, test_labels, S+train_instances, classes, inputs, &Wr, &MI, &average);
        fprintf(parm.out_file, "@Wr redundancy: %6.4lf\n", Wr);
        fprintf(parm.out_file, "@MI redundancy: %6.4lf\n", MI);
        fprintf(parm.out_file, "@Average correlation redundancy: %6.4lf\n", average);
    }


    // End output
    SHOWPROGRESS(parm.show_progress > 0, "@Ok%c", '\n');
    fprintf(parm.out_file, "@Ok\n");
    fclose(parm.out_file);

    return 0;
}
