/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"
#include "ensemble.h"

// Y. Yu, W. Pedrycz and D. Miao (2014) "Multi-label classification by exploiting label correlations", Expert Systems with Applications 41:2989–3004.

void TrainMLBasedOnNeighborhoodRoughSets(const t_Parameters *parm, t_MLRS *mlrs, int **labels, int *n_labels, int classes, int instances, int *subset, scalar **data, int **index, int *subspace,
                                         int inputs)
{
    scalar *count;

    AllocateScalarMatrix(&(mlrs->probs), classes, classes);
    MALLOC(count, scalar, classes, "MLBasedOnNeighborhoodRoughSets: count");

    mlrs->k = parm->mlrs_k;
    mlrs->beta = parm->mlrs_beta;

    ZeroScalarMatrix(mlrs->probs, classes, classes);
    ZeroVector(count, NULL, classes);

    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            for (int j = 0; j < n_labels[i]; j++) {
                count[labels[i][j]]++;
                for (int k = 0; k < j; k++)
                    mlrs->probs[labels[i][k]][labels[i][j]]++;
            }

    for (int i = 0; i < classes; i++) {
        mlrs->probs[i][i] = 1.0;
        for (int j = 0; j < i; j++) {
            mlrs->probs[i][j] = mlrs->probs[j][i]/(count[j]+parm->mlrs_s);
            mlrs->probs[j][i] = mlrs->probs[j][i]/(count[i]+parm->mlrs_s);
        }
    }

    // Obtain calibrated threshold
    mlrs->t = 0.5;  // Bogus threshold
    scalar **f;
    int h[classes], n_predicted;
    bool classification, ranks;
    AllocateScalarMatrix(&f, instances, classes);
    for (int i = 0; i < instances; i++)
        MLRSClassify(mlrs, data[i], f[i], classes, index, h, &n_predicted, &classification, &ranks, data, instances, inputs, index ? index[i] : NULL, subspace, subset, labels, n_labels);

    // Check thresholds in (0, 1) with steps of 0.05
    scalar best = 0.0;
    int best_val = 0;
    for (scalar th = 0.05; th < 1.0; th += 0.05) {
        scalar val = 0.0;
        for (int i = 0; i < instances; i++)
            for (int j = 0; j < classes; j++) {
                bool in = ValueInVector(j, labels[i], n_labels[i]);
                if ((in && f[i][j] >= th) || (!in && f[i][j] < th))
                    val++;
            }

        if (val > best_val) {
            best_val = val;
            best = th;
        }
    }

    SAVEPROGRESS(parm->show_progress > 1, "MLRS calibrated threshold: %lf\n", best);
    mlrs->t = best;

    FreeScalarMatrix(&f, instances);
    FREE(count);
}

void MLRSClassify(t_MLRS *mlrs, scalar *instance, scalar *f, int classes, int **index, int *h, int *n_predicted, bool *classification, bool *ranks, scalar **data, int instances, int inputs,
                  int *query_index, int *subspace, int *subset, int **labels, int *n_labels)
{
    scalar prior[classes], sigma[classes];

    // Compute neighborhood of instance
    int neighbors[mlrs->k];
    GetkNeighborsForML(instance, data, instances, index, inputs, query_index, mlrs->k, neighbors, subset, NULL, -1, false, NULL, classes, NULL, -1, subspace);

    // Prior probabilities
    ZeroVector(sigma, NULL, classes);
    for (int i = 0; i < mlrs->k; i++)
        for (int j = 0; j < n_labels[neighbors[i]]; j++)
            sigma[labels[neighbors[i]][j]]++;

    scalar sum = SumVector(sigma, NULL, classes);
    for (int i = 0; i < classes; i++)
        prior[i] = sigma[i]/sum;

    // Probability of each label
    for (int i = 0; i < classes; i++) {
        if (sigma[i] >= mlrs->k * mlrs->beta)
            f[i] = 1.0;
        else if (sigma[i] <= mlrs->k * (1.0-mlrs->beta))
            f[i] = 0.0;
        else {
            f[i] = 0.0;
            for (int j = 0; j < classes; j++)
                f[i] += mlrs->probs[i][j]*prior[j];
        }
    }

    NormalizeVector(f, classes);

    // Relevant labels
    *n_predicted = 0;
    for (int i = 0; i < classes; i++)
        if (f[i] >= mlrs->t)
            h[(*n_predicted)++] = i;

    *classification = *ranks = true;
}

void MLRSFree(t_MLRS *mlrs, int classes)
{
    FreeScalarMatrix(&(mlrs->probs), classes);
}
