/***************************************************************************
 *   Copyright (C) 2022 by Nicolas Garcia-Pedrajas                         *
 *   npedrajas@uco.es                                                      *
 *                                                                         *
 *   This program is FREE software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "classifiers.h"
#include "macros.h"
#include "maths.h"
#include "ml.h"
#include "ensemble.h"

// Y. Yu, W. Pedrycz and D. Miao (2014) "Multi-label classification by exploiting label correlations", Expert Systems with Applications 41:2989–3004.

void TrainMLBasedOnNeighborhoodRoughSetsLocalCorrelation(const t_Parameters *parm, t_MLRSLC *mlrslc, int **labels, int *n_labels, int classes, int instances, int *subset, scalar **data, int **index,
        int *subspace, int inputs)
{
    scalar **count;

    mlrslc->k = parm->mlrs_k;
    mlrslc->beta = parm->mlrs_beta;
    mlrslc->n_clusters = classes;

    // Perform clustering in labels space
    // First get instances in labels space
    scalar **l_data;
    AllocateScalarMatrix(&l_data, instances, classes);
    ZeroScalarMatrix(l_data, instances, classes);
    for (int i = 0; i < instances; i++)
        for (int j = 0; j < n_labels[i]; j++)
            l_data[i][labels[i][j]] = 1.0;

    int ccount[mlrslc->n_clusters];
    MALLOC(mlrslc->partition, int, instances, "TrainMLBasedOnNeighborhoodRoughSetsLocalCorrelation: partition");

    kMeans(mlrslc->partition, ccount, l_data, instances, mlrslc->n_clusters, classes, parm->kmeans_max_iterations, NULL, subset);

    // Obtain probabilities for every cluster
    Allocate3DScalarMatrix(&(mlrslc->probs), mlrslc->n_clusters, classes, classes);
    AllocateScalarMatrix(&count, mlrslc->n_clusters, classes);
    Zero3DScalarMatrix(mlrslc->probs, mlrslc->n_clusters, classes, classes);
    ZeroScalarMatrix(count, mlrslc->n_clusters, classes);
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            for (int j = 0; j < n_labels[i]; j++) {
                count[mlrslc->partition[i]][labels[i][j]]++;
                for (int k = 0; k < j; k++)
                    mlrslc->probs[mlrslc->partition[i]][labels[i][k]][labels[i][j]]++;
            }

    for (int l = 0; l < mlrslc->n_clusters; l++)
        for (int i = 0; i < classes; i++) {
            mlrslc->probs[l][i][i] = 1.0;
            for (int j = 0; j < i; j++) {
                mlrslc->probs[l][i][j] = mlrslc->probs[l][j][i]/(count[l][j]+parm->mlrs_s);
                mlrslc->probs[l][j][i] = mlrslc->probs[l][j][i]/(count[l][i]+parm->mlrs_s);
            }
        }

    // Obtain calibrated threshold
    mlrslc->t = 0.5;    // Bogus threshold
    scalar **f;
    int h[classes], n_predicted;
    bool classification, ranks;
    AllocateScalarMatrix(&f, instances, classes);
    for (int i = 0; i < instances; i++)
        if (!subset || subset[i])
            MLRSLCClassify(mlrslc, data[i], f[i], classes, index, h, &n_predicted, &classification, &ranks, data, instances, inputs, index ? index[i] : NULL, subspace, subset, labels, n_labels);

    // Check thresholds in (0, 1) with steps of 0.05
    scalar best = 0.0;
    int best_val = 0;
    for (scalar th = 0.05; th < 1.0; th += 0.05) {
        scalar val = 0.0;
        for (int i = 0; i < instances; i++)
            if (!subset || subset[i])
                for (int j = 0; j < classes; j++) {
                    bool in = ValueInVector(j, labels[i], n_labels[i]);
                    if ((in && f[i][j] >= th) || (!in && f[i][j] < th))
                        val++;
                }
        if (val > best_val) {
            best_val = val;
            best = th;
        }
    }

    SHOWPROGRESS(parm->show_progress > 1, "MLRS calibrated threshold: %lf\n", best);
    mlrslc->t = best;

    FreeScalarMatrix(&f, instances);
    FreeScalarMatrix(&count, mlrslc->n_clusters);
}

void MLRSLCClassify(t_MLRSLC *mlrslc, scalar *instance, scalar *f, int classes, int **index, int *h, int *n_predicted, bool *classification, bool *ranks, scalar **data, int instances, int inputs,
                    int *query_index, int *subspace, int *subset, int **labels, int *n_labels)
{
    scalar **prior, **sigma, *p_t;

    // Compute neighborhood of instance
    int neighbors[mlrslc->k];
    GetkNeighborsForML(instance, data, instances, index, inputs, query_index, mlrslc->k, neighbors, subset, NULL, -1, false, NULL, classes, NULL, -1, subspace);

    // Prior probabilities
    AllocateScalarMatrix(&prior, mlrslc->n_clusters, classes);
    AllocateScalarMatrix(&sigma, mlrslc->n_clusters, classes);
    MALLOC(p_t, scalar, mlrslc->n_clusters, "MLRSLCClassify: p_t");
    ZeroScalarMatrix(sigma, mlrslc->n_clusters, classes);
    ZeroScalarMatrix(prior, mlrslc->n_clusters, classes);
    ZeroVector(p_t, NULL, mlrslc->n_clusters);

    for (int i = 0; i < mlrslc->k; i++) {
        p_t[mlrslc->partition[neighbors[i]]]++;
        for (int j = 0; j < n_labels[neighbors[i]]; j++)
            sigma[mlrslc->partition[neighbors[i]]][labels[neighbors[i]][j]]++;
    }

    for (int j = 0; j < mlrslc->n_clusters; j++) {
        p_t[j] /= mlrslc->k;
        scalar sum = SumVector(sigma[j], NULL, classes);
        for (int i = 0; i < classes; i++)
            prior[j][i] = sum > ZERO ? sigma[j][i]/sum : 0.0;
    }

    // Probability of each label
    scalar **F;
    AllocateScalarMatrix(&F, mlrslc->n_clusters, classes);
    for (int l = 0; l < mlrslc->n_clusters; l++) {
        for (int i = 0; i < classes; i++) {
            if (sigma[l][i] >= mlrslc->k * mlrslc->beta)
                F[l][i] = 1.0;
            else if (sigma[l][i] <= mlrslc->k *(1.0-mlrslc->beta))
                F[l][i] = 0.0;
            else {
                F[l][i] = 0.0;
                for (int j = 0; j < classes; j++)
                    F[l][i] += mlrslc->probs[l][i][j]*prior[l][j];
            }
        }
    }

    // Global probabilities
    for (int l = 0; l < classes; l++) {
        f[l] = 0.0;
        for (int d = 0; d < mlrslc->n_clusters; d++)
            for (int j = 0; j < classes; j++)
                f[l] += mlrslc->probs[d][l][j]*F[d][j]*p_t[d];
    }

    NormalizeVector(f, classes);

    // Relevant labels
    *n_predicted = 0;
    for (int i = 0; i < classes; i++)
        if (f[i] >= mlrslc->t)
            h[(*n_predicted)++] = i;

    *classification = *ranks = true;

    FreeScalarMatrix(&sigma, mlrslc->n_clusters);
    FreeScalarMatrix(&prior, mlrslc->n_clusters);
    FreeScalarMatrix(&F, mlrslc->n_clusters);
    FREE(p_t);
}


void MLRSLCFree(t_MLRSLC *mlrslc, int classes)
{
    Free3DScalarMatrix(&(mlrslc->probs), mlrslc->n_clusters, classes);
    FREE(mlrslc->partition);
}
